// $Id: 

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _NOISE_MMC_
#define _NOISE_MMC_

#include <time.h>
#include <sys/time.h>
#include <sys/times.h>

#include "ocs.hh"
#include "mmc_ocs.hh"

//! Enumeration type used to determine the particular system
//! to be modeled.

enum nameSystem { TYCO_B2B=1, TYCO=2, MARCO=3, OLEG=4 };

//! NoiseMMC

/*!

We assume a single channel signal (for now)

*/

class NoiseMMC
{

  // ########### Public Methods ###################

  public:

 //! Constructor

 NoiseMMC(string InFileName);

 void ReadInSystemParameters(string InFileName);
 void AllocateSystemMemory(void);
 
 void ReadInMMCParameters(string InFileName);
 void AllocateMMCMemory(void);

 void ReadInRestartParameters(string RestartingDataInFileName);

 ~NoiseMMC(void);

  void RunMMC(void);

  void RunNoiseFree(void);
  void RestartMMC(void);
  void IterateMMC(void);

  double PropagateThroughSystem(double * z_try);
  double PropagateTycoB2B(double * z_try);
  double PropagateTyco(double * z_try);
  double PropagateMarco(double * z_try);
  double PropagateOleg(double * z_try);

 
  double ReceiveSignal(void);

  // ########### Private Methods ###################

  private:

  // ########### Private Data ###################

  private:

  string Job;

  nameSystem NameSystem;

  int NumMaps;
  int TypeStepSizes;
  double RelativeErrorGoal;
  double WalkOffParameter;
  double ConstantStepSize;

  int OutputStepSizesFlag;
  bool TestStepSizesFlag;
  int Beta3ZeroFlag;
  int NonLinearityZeroFlag;
  int TypeAmp;
  int AmpsNoiseOffFlag;
  int DebugLevelAmps;

  int NumAmpsPerMap;

  double ClockRecoveryTime;

  RanNumGen * RNGSignal;
  RanNumGen * RNGAmps;
  RanNumGen * RNGFibers;
  
  OptSignal * Signal; 
  OptSignal * InputSignal; 

  bool LoadSignalFileFlag;
  string SignalFileName;
 
  bool DoStdMCFlag;
  ElecSignalStat * oElecSignalStat;
 
  OptDataWriter* DataWriter;

  OptFiberLocalError * PreDCF;
  OptFiberLocalError * PostDCF;

  //! Tyco Fibers

  OptFiberLocalError * DSF35;
  OptFiberLocalError * DSF45;
  OptFiberLocalError * SMF;

  //! Marco system fiber

  OptFiberLocalError * Fiber;

  //! Oleg System Fibers

  OptFiberLocalError * Dplus;
  OptFiberLocalError * Dminus;


  OptAmplifier * Amp;
  OptAmplifier * PAmp;
  LumpedLoss* Loss;
  LumpedLoss* PLoss;
  
  OptFilter *oOptFilter; 
  Photodetector * oPhotodetector;
  ElecFilter * oElecFilter;

  // MMC parameters

  //! 1 means make independent decisions to accept/reject each component of noise vector
  //! 0 means accept/reject entire vector

  bool IndependentDecisionsFlag;

  //! The number of freqs (per amp) to be included in the noise state space

  int NumRelevantFreqs;

  //! The dimension of the noise state space

  unsigned DimStateSpace; 

  //! The number of bit are calculating pdf for. Between 1 and NumBits

  int ControlBit;
  bool AddBiasedNoiseFlag;

  double stan_dev_real_imag;
  

  //!  The algorithm only stops once  the lowest probability is less than
  //!  MinimumPdfValue. There are also other stopping criteria that must
  //! be met.

  double MinimumPdfValue;

  //! The algorithm only stops once  
  //! max_k(abs((P_k^j-P_k^(j+1))/P_k^(j+1)) < MaxRelError.
  //! (But if MaxRelError = 0.0, then we stop after NumIterations.)
  //! There are also other stopping criteria that must be met.

  double MaxRelError;

  //! Max number of iterations. Only used if MaxRelError = 0.0

  unsigned NumIterations;

  //! Name of output file to write
  //! auxiliary statistics re performance of MMC algorithm


  bool RestartJobFlag;

  RanNumGen * RNG_MMC;

  bool RNGSkipSamplesFlag;
  int NumRNGSamplesToSkip;

  MMC* mmc;

};

#endif // _NOISE_MMC_

