// $Id: StartupTyco.hh,v 1.1 2001/05/05 02:09:38 zweck Exp $ 

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _STARTUP_RON_NOISE_HH_
#define _STARTUP_RON_NOISE_HH_

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsTools.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"

class StartupZweck
{

public:

  StartupZweck(string InFileName);

  string GetJobDir() { return JobDir;};
  string GetJobName() {return JobName;};
  string GetJobPrefix() {return JobPrefix;};
  string GetInputFilesDir() {return InputFilesDir;};
  
  int GetSimType() {return SimType;};
  int GetQtStatSamples() {return qtStatSamples;};
  int GetQtMapSpans() {return qtMapSpans;};
  int GetBeta3ZeroFlag() {return Beta3ZeroFlag;};
  int GetNonLinearityZeroFlag() {return NonLinearityZeroFlag;};
  int GetFiberLossZeroFlag() {return FiberLossZeroFlag;};
  int GetTypeAmp() {return TypeAmp;}; 
  int GetAmpsNoiseOffFlag() {return AmpsNoiseOffFlag;}; 


  
 
  double GetRelativeErrorGoal() {return RelativeErrorGoal;};
  
  double GetConstantStepSize() {return ConstantStepSize;};

  int GetOutputStepSizesFlag() {return OutputStepSizesFlag;};

  int GetDebugLevelAmps() {return DebugLevelAmps;};

  int GetTypeStepSizes() { return TypeStepSizes;};

  double GetWalkOffParameter() { return WalkOffParameter;};
  
  //added by Walter
  double GetBoostLastIterationFactor() {return BoostLastIterationFactor;};
  bool GetWriteToFileVoltagesFlag() {return WriteToFileVoltagesFlag;};
  bool GetStopCriterionOnAllBitsFlag() {return StopCriterionOnAllBitsFlag;};
  bool GetAcceptIfAllBitsInRangeFlag() {return AcceptIfAllBitsInRangeFlag;};
  int GetControlQuantityBit() {return ControlQuantityBit;};
  bool GetSameNoiseSeedsFlag() {return SameNoiseSeedsFlag;};
  int GetNumSamplesMC() {return MC_samples;};
  bool GetDoB2BSimulation() {return DoB2BSimulation;};
  bool GetLoadSignalFileFlag() {return LoadSignalFileFlag;};
  bool GetWriteComplexFieldsFlag() {return WriteComplexFieldsFlag;};
  string GetSignalFileName() {return SignalFileName;};
  double GetMinPdfValue() {return Minimum_pdf_value;};
  int GetPdfOfWhatFlag() {return PdfOfWhat;};
  int GetNumRelevantFrequencies() {return NumRelevantFrequencies;};
  bool GetWriteFileOutputFlag() {return Write_file_output;};
  char GetVerbosityLevel() {return verbose;};
  bool GetDoInterruptedMetropolisFlag() {return DoInterruptedMetropolisFlag;};  
  int GetNumSamplesBetweenTimeShift() {return NumSamplBetwTimeShift;};
  double GetStdNoisePhaseShift() {return StdNoisePhaseShiftInDegrees;};
  double GetWhenToStopIterations() {return WhenToStopIterations;};
  bool GetRestartFlag() {return RestartFlag;};
  unsigned GetNumIterations(int what)
     {
     if (what)
        return NumIterations_1;
     else
        return NumIterations_0;
     };
  double GetVoltageMin(int what)
     {
     if (what)
        return V_min_1;
     else
        return V_min_0;
     }; 
  double GetVoltageMax(int what) 
     {
     if (what)
        return V_max_1;
     else
        return V_max_0;
     };
  unsigned GetNumVoltageBins(int what) 
     {
     if (what)
        return M_1;
     else
        return M_0;
     };
  double GetInitialEpsilon(int what) 
     {
     if (what)
        return Epsilon_1;
     else
        return Epsilon_0;
     };
  double GetTargetAlpha(int what) 
     {
     if (what)
        return Alpha_1;
     else
        return Alpha_0;
     };
  unsigned GetNumProductionSamplesFirstIteration(int what) 
     {
     if (what)
        return N_0_1;
     else
        return N_0_0;
     };
  unsigned GetNumSamplesStage1(int what) 
     {
     if (what)
        return N_stage1_1;
     else
        return N_stage1_0;
     };
  double GetNMultiplicationFactor(int what) 
     {
     if (what)
        return N_factor_1;
     else
        return N_factor_0;
     };
  bool GetFitPWithGaussianFlag(int what) 
     {
     if (what)
        return Fit_P_1;
     else
        return Fit_P_0;
     };
  
    
private:

string JobDir;
string JobName;
string JobPrefix;
string InputFilesDir;
string LogFileName;


int SimType;
int qtStatSamples;
int qtMapSpans;

double RelativeErrorGoal;
double WalkOffParameter;
double ConstantStepSize;

int TypeStepSizes;

  // Flags

int Beta3ZeroFlag;
int NonLinearityZeroFlag;
int FiberLossZeroFlag;	
int TypeAmp;
int AmpsNoiseOffFlag;

int OutputStepSizesFlag;

bool WriteComplexFieldsFlag;

int DebugLevelAmps;

  // Parameters

bool SameNoiseSeedsFlag;

// ######### parameters for MMC added by Walter ##########

// common parameters
double Minimum_pdf_value;
int PdfOfWhat;
int NumRelevantFrequencies;
bool Write_file_output;
char verbose;
bool DoInterruptedMetropolisFlag;
int NumSamplBetwTimeShift;
double StdNoisePhaseShiftInDegrees;
double WhenToStopIterations;
bool RestartFlag;
string SignalFileName;
bool DoB2BSimulation;
bool LoadSignalFileFlag;
int ControlQuantityBit;
bool StopCriterionOnAllBitsFlag;
bool AcceptIfAllBitsInRangeFlag;
bool WriteToFileVoltagesFlag;
double BoostLastIterationFactor;
 
// parameters MMC class for computing pdf of zeros
unsigned NumIterations_0;
double V_min_0;
double V_max_0;
unsigned M_0;
double Epsilon_0;
double Alpha_0;
unsigned N_0_0;
unsigned N_stage1_0;
double N_factor_0;
bool Fit_P_0;


// parameters MMC class for computing pdf of zeros

unsigned NumIterations_1;
double V_min_1;
double V_max_1;
unsigned M_1;
double Epsilon_1;
double Alpha_1;
unsigned N_0_1;
unsigned N_stage1_1;
double N_factor_1;
bool Fit_P_1;

// for MC simulation
int MC_samples;

};

#endif /* _STARTUP_RON_NOISE_HH_ */
