// $Id:
// #####################################################################
//
//    ocsAppA_LimaMMC.cc
//    by: Aurenice O. Lima 2004/2005
//    Simulation Module of the ocs system
//    Optical Communication Systems Simulator
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################



#include <iostream.h>
#include <complex.h>
#include "ocs.hh"

#include "ocsStartupLima.hh"

extern ofstream LogFile;

void GetDGDpdfMMC_Berg(MyStartup *oMyStartup) ;


//#############################################################
//##### Creates a file with the pdf of the DGD
//##    using the Multicanonical Monte Carlo implementation by Berg
//## By: Aurenice Lima in 05/19/04
void GetDGDpdfMMC_Berg(MyStartup *oMyStartup) //#$%#$%
////////void GetDGDpdfMMC_BergMonteCarloWithTransitionProbabilityMatrix051904(MyStartup *oMyStartup) 
{
   Histogram oHistDGD("HistDGD.in");
   Histogram oHistActualDGD("HistActualDGD.in");   
   Histogram oHistSndOrderDGD("HistSndOrdDGD.in");    
   Histogram oHistThirdOrderDGD("HistThirdOrdDGD.in");       

   int qtIterationsMMC = oMyStartup->GetIntParam("qtIterationsMMC");
   cout << "qtIterationsMMC = " << qtIterationsMMC << endl; 

   double RelStepSizeMMC = oMyStartup->GetDoubleParam("RelStepSizeMMC");
   cout << "RelStepSizeMMC = " << RelStepSizeMMC << endl; 

   int qtStatSamples = oMyStartup->GetQtStatSamples();
   cout << "qtStatSamples = " << qtStatSamples << endl;    
 
   int ErrorPropagationFlag = oMyStartup->GetIntParam("ErrorPropagationFlag");
   cout << "ErrorPropagationFlag = " << ErrorPropagationFlag << endl;   

   int NumSamplesForErrorEstim = oMyStartup->GetIntParam("NumSamplesForErrorEstim");  
   if (NumSamplesForErrorEstim == 0) {
      NumSamplesForErrorEstim = 1000;
   }
   cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;

   int ErrorOfErrorFlag = oMyStartup->GetIntParam("ErrorOfErrorFlag");
   cout << "ErrorOfErrorFlag = " << ErrorOfErrorFlag << endl;   

   int NumBootstrapTransMatrices = oMyStartup->GetIntParam("NumBootstrapTransMatrices");  
   if (NumBootstrapTransMatrices == 0) {
      NumBootstrapTransMatrices = 100;
   }
   cout << "NumBootstrapTransMatrices = " << NumBootstrapTransMatrices << endl;

   int DoNotSpoolErrorFlag = oMyStartup->GetIntParam("DoNotSpoolErrorFlag");
   cout << "DoNotSpoolErrorFlag = " << DoNotSpoolErrorFlag << endl;   

   int FiberRotatorModelFlag = oMyStartup->GetIntParam("FiberRotatorModelFlag");
   cout << "FiberRotatorModelFlag = " << FiberRotatorModelFlag << endl;    

   cout << "Creating a file with the pdf of the DGD." << endl;
   cout << "Plot the result with the Gnuplot script: nDGDpdf.gp" << endl;      

   //## This parameter should not be changed if one wants to use the covariance
   int qtStatSamplesStep2 = oMyStartup->GetIntParam("qtStatSamplesStep2");      
   if (qtStatSamplesStep2 == 0) {
      cout << "  *** Forcing a value for qtStatSamplesStep2 = qtStatSamples" <<endl;
      qtStatSamplesStep2 = qtStatSamples;
   }
   cout << "qtStatSamplesStep2 = " << qtStatSamplesStep2 << endl;

   int ComputeAutocorrFlag = oMyStartup->GetIntParam("ComputeAutocorrFlag");
   cout << "ComputeAutocorrFlag = " << ComputeAutocorrFlag << endl;    

   int BinAutocorrelation;   
   if (ComputeAutocorrFlag) {
      BinAutocorrelation = oMyStartup->GetIntParam("BinAutocorrelation");
      cout << "BinAutocorrelation = " << BinAutocorrelation << endl;          
   }
   
   // ## Initialize the random generators
   RanNumGen oRanPMD("RanPMD.in");
   RanNumGen oRanBitString("RanBitString.in");       
   
   // ## Initiating the optical pulses
   OptSignal oOptSignal("OptSignal.in", &oRanBitString);   
   OptFiber oOptFiber("OptFiber.in", &oOptSignal, &oRanPMD);    
                        
   cout << "ExpectedDGD [ps] = " 
          << oOptFiber.GetExpectedDGD()*1e12 << endl;             
   double LikelihoodRatio;       
   double OutPD_Vector[3]; 
   double SndOrderOutPD_Vector[3];
   double NumSndOrderOutPD_Vector[3];
   double ThirdOrderOutPD_Vector[3];
   double NumThirdOrderOutPD_Vector[3];   
   double nDGD, nSndOrderDGD, nThirdOrderDGD;          
   double nDGD_Before = 0;  
   double nSndOrderDGD_Before = 0;
   double nThirdOrderDGD_Before = 0;

   int qtBins = oHistDGD.GetQtBins();
   if (  oHistDGD.GetQtBins() != oHistActualDGD.GetQtBins() 
       || oHistDGD.GetMaxRange() != oHistActualDGD.GetMaxRange()
       || oHistDGD.GetMinRange() != oHistActualDGD.GetMinRange() ) {
      cout << "Error: HistDGD.in and HistActualDGD.in must have " << endl
           << "the same range." << endl
	   << "Aborting now!!!!!" << endl;
      exit(1);	         
   }                 
   if (FiberRotatorModelFlag) {
      //## The random mode coupling will be produced by a single
      //## rotation Rz, which is the simplest one.     
      oOptFiber.GetFiberSampleUnifRz();
   } else {      
      oOptFiber.GetFiberSampleCompleteScatt();
   }    
   int qtStepsFiber = oOptFiber.GetQtStepsFiber();     
   double *MMC_pdf = new double[qtBins];    
   double *MMC_pdfTmp = new double[qtBins];          
   double *GlobalStatSignifMMC = new double[qtBins];      

   double *MMC_pdfStep2 = new double[qtBins]; 
   double *MMC_cdfStep2 = new double[qtBins];

   double *MMC_cdfStep2_BootResamp = new double[qtBins];

   
   double *NumSamplesBinTotal = new double[qtBins];
   
   double *NormHistogram = new double[qtBins];
   
   double *VarianceBinMMC = new double[qtBins];   
   double *VarianceBinThisIterMMC = VarianceBinMMC;   

   double *MeanStdDevBinMMC = new double[qtBins];
   double *VarianceStdDevBinMMC = new double[qtBins];   

   double *SumBinMMC = new double[qtBins];
   double *SumSqBinMMC = new double[qtBins];

   double *SumStdDevBinMMC = new double[qtBins];
   double *SumSqStdDevBinMMC = new double[qtBins];
   
   double *TransProbabilityMatrix = new double [qtBins*qtBins];
   double *NumTransitionsFromBin = new double [qtBins];
   double *CumulativeTransProbabilityMatrix = new double [qtBins*qtBins];   
   double *VarTransProbabilityMatrix = new double [qtBins*qtBins];

   double *TransProbabilityMatrixBootResamp = new double [qtBins*qtBins];
   double *NumTransitionsFromBinBootResamp = new double [qtBins];
   double *SumTransProbabilityMatrixBoot = new double [qtBins*qtBins];   
   double *SumSqTransProbabilityMatrixBoot = new double [qtBins*qtBins];   
   double *VarTransProbabilityMatrixBootEnsemble = new double [qtBins*qtBins];
   double *CumulativeTransProbabilityMatrixBootResamp = new double [qtBins*qtBins];      
        
   double *ThetaZ = new double[qtStepsFiber];    
   double *DeltaThetaZ = new double[qtStepsFiber];     
   double *PhiZ = new double[qtStepsFiber];
   double *DeltaPhiZ = new double[qtStepsFiber];     
   double *PsiZ = new double[qtStepsFiber];
   double *DeltaPsiZ = new double[qtStepsFiber];   
   for (int ii = 0; ii < qtStepsFiber; ii++) {    
      //## Initialize the angles
      double CosThetaZ_Tmp = 2*(oRanPMD.GetRanNum() - 0.5); //## -1 to 1
      ThetaZ[ii] = acos(CosThetaZ_Tmp);
      PhiZ[ii] = pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi      
      PsiZ[ii] = pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi      
   }
   oHistActualDGD.Obtain_pdf(); //## Fill in the vector pdfX   
   oHistDGD.Obtain_pdf();   
   for (int ii = 0; ii < qtBins; ii++) {   
      //## Initialize the iterative pdf 
      MMC_pdf[ii] = 1./qtBins;
      GlobalStatSignifMMC[ii] = 0;              
      NumTransitionsFromBin[ii] = 0;
      NumSamplesBinTotal[ii] = 0;    
   }   

   for (int ii = 0; ii < qtBins*qtBins; ii++) {   
      TransProbabilityMatrix[ii] = 0;
      CumulativeTransProbabilityMatrix[ii] = 0;      
   }

   //## Step I
   //## Computation of the MMC weights
   //## Major loop: MMC iterations
   //int SignThetaFromCos = +1;  
   for (int kk = 0; kk < qtIterationsMMC; kk++) {      
      oHistActualDGD.ClearHistogram();            
      double qtAcceptedSamples = 0;         
      for (int ii = 0; ii < qtStatSamples; ii++) {        
         for (int jj = 0; jj < qtStepsFiber; jj++){ 	 
            if (kk == 0 ) {
               //## Set a random angle for the first iteration        
               DeltaThetaZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi  	 	 
               DeltaPhiZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi  
               DeltaPsiZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi                       
            } else {
               //## Set the ii-th fiber with the perturbation
               DeltaThetaZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //##        
               DeltaPhiZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## 
               DeltaPsiZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //##                       
            }	    	 		    
	    
            if (FiberRotatorModelFlag) {
               oOptFiber.SetScatteringMatrix( jj,   PhiZ[jj] 
	                                    +DeltaPhiZ[jj],
                                           pi/2., -pi/2.);
            } else {            
	       //## Assure the correct statistics in the angle (cosine fo theta is uniform)
               if ( oRanPMD.GetRanNum() >
                    sqrt(1 - sq(cos(ThetaZ[jj] + DeltaThetaZ[jj])))/sqrt(1 - sq(cos(ThetaZ[jj])) ) ) {    
   	         //## Do not accept
   	         DeltaThetaZ[jj] = 0;
	       }  
               oOptFiber.SetScatteringMatrix( jj,  ThetaZ[jj] + DeltaThetaZ[jj], 
	                                    PhiZ[jj] + DeltaPhiZ[jj],
                                            PsiZ[jj] + DeltaPsiZ[jj]);
            }                                                                       
         }                              
         nDGD = oOptFiber.GetDGD(oOptSignal.GetCenterFreq())
                       /oOptFiber.GetExpectedDGD();      
         nSndOrderDGD = oOptFiber.GetSndOrderDGD(oOptSignal.GetCenterFreq())
                      /oOptFiber.GetExpectedSndOrderDGD(); 
         nThirdOrderDGD = oOptFiber.GetThirdOrderDGD(oOptSignal.GetCenterFreq())
                      /oOptFiber.GetExpectedThirdOrderDGD();                                       
         if (kk == 0 && ii == 0)  {
            nDGD_Before = nDGD;     
	    nSndOrderDGD_Before = nSndOrderDGD;
	    nThirdOrderDGD_Before = nThirdOrderDGD;
         }            
	 if (   nDGD > oHistActualDGD.GetMinRange()     
	     && nDGD < oHistActualDGD.GetMaxRange() ) {               
   	    //## I intend to compute the true transition probability
	    //## Therefore, this must come before the decision on the acceptance
            TransProbabilityMatrix[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)
                  +qtBins*oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;	     
            NumTransitionsFromBin[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
            if ( oRanPMD.GetRanNum() <
               MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)] 
              /MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)] ) {      
               //## Accept the transition  
               qtAcceptedSamples++;	
               nDGD_Before = nDGD; 
               for (int jj = 0; jj < qtStepsFiber; jj++){ 	    
		  ThetaZ[jj] += DeltaThetaZ[jj];
                  PhiZ[jj] += DeltaPhiZ[jj];
                  PsiZ[jj] += DeltaPsiZ[jj];                  
               } 
            } 	 
         }	 
         oHistActualDGD.AddImportantSample(nDGD_Before,1);                                    
         NumSamplesBinTotal[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++; 	 	 
         oHistSndOrderDGD.AddImportantSample(nSndOrderDGD_Before,1);   
         oHistThirdOrderDGD.AddImportantSample(nThirdOrderDGD_Before,1);	 
      } //## End of qtSamples
      cout << "Iteration "<< kk << " ->  AcceptanceRate = " 
           << (1.0*qtAcceptedSamples)/qtStatSamples << endl; 
      
      //## Here comes Berg's modification
      
      //## Make a copy of the previous MMC_pdf
      for (int ii = 0; ii < qtBins; ii++) {          
         MMC_pdfTmp[ii] = MMC_pdf[ii];	 
         //## Attention: the normalization mus be made by qtStatSamplesStep2
         //#  That is necessary to have a coherent statistical significance
	 //#  in steps 1 and 2
         NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
      }  
          
      //## Start the computation of MMC_pdf
      for (int ii = 0; ii < qtBins - 1; ii++) {          
         if ( NormHistogram[ii+1] 
	     *NormHistogram[ii] > 0 ) {
            double StatSignifMMC =  NormHistogram[ii+1] 
	                        *NormHistogram[ii]
		              /( NormHistogram[ii+1] 
	                        +NormHistogram[ii] );
            GlobalStatSignifMMC[ii] += StatSignifMMC;
	    double RelStatSignifMMC = StatSignifMMC
                                     /GlobalStatSignifMMC[ii];
	    //RelStatSignifMMC = 1; //#$#$#$#$#$#$#$#   
            MMC_pdf[ii+1] = MMC_pdf[ii] 
	                   *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			   *pow( NormHistogram[ii+1] 
       	                        /NormHistogram[ii],
		                 RelStatSignifMMC);
         } else { //## This iteration had no stat signific in this bin
            MMC_pdf[ii+1] = MMC_pdf[ii] 
	                   *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	 }	    
      }                             
      
      //## Finally, the normalization      
      double NormMMC_pdf = 0;
      for (int ii = 0; ii < qtBins; ii++) {          
         NormMMC_pdf += MMC_pdf[ii];	 
      }                      
      for (int ii = 0; ii < qtBins; ii++) {   
         MMC_pdf[ii]  /= NormMMC_pdf;      	 
      }   
       
      //##
      //## Now, plot the histogram after the kk iteration
      //##      
      oHistDGD.ClearHistogram();
      for (int ii = 0; ii < qtBins; ii++) {   
         oHistDGD.AddImportantSample( oHistActualDGD.Get_pdfX(ii),
               MMC_pdf[ii]*qtBins); //## qtBins is here to workk around the normalization
      }                      
      //## Creates output files with the data
      oHistActualDGD.WriteFile_pdf("nActualDGDpdf.dat");   
      oHistDGD.WriteFile_pdf("nDGDpdfInvWeights.dat");                          
   } //## End of qtIterationsMMC

   //######################
   //## Step II: Computation of the probability using the MMC weights
   //## Now, use the MMC weights computed in Step I to make one single 
   //## simulation with the Metropolis Algorithm using the weights
   //##

   //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
   oHistActualDGD.ClearHistogram();        
   oHistDGD.ClearHistogram();
  
   double qtAcceptedSamples = 0;   
   for (int ii = 0; ii < qtStatSamplesStep2; ii++) {        
      for (int jj = 0; jj < qtStepsFiber; jj++){ 
            DeltaThetaZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi              
            DeltaPhiZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //## 
            DeltaPsiZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //##                    

            if (FiberRotatorModelFlag) {
               oOptFiber.SetScatteringMatrix( jj,   PhiZ[jj] 
	                                    +DeltaPhiZ[jj],
                                           pi/2., -pi/2.);
            } else {            
	       //## Assure the correct statistics in the angle (cosine fo theta is uniform)
               if ( oRanPMD.GetRanNum() >
                    sqrt(1 - sq(cos(ThetaZ[jj] + DeltaThetaZ[jj])))/sqrt(1 - sq(cos(ThetaZ[jj])) ) ) {    
   	         //## Do not accept
   	         DeltaThetaZ[jj] = 0;
	       }              
               oOptFiber.SetScatteringMatrix( jj,  ThetaZ[jj] + DeltaThetaZ[jj], 
	                                    PhiZ[jj] + DeltaPhiZ[jj],
                                            PsiZ[jj] + DeltaPsiZ[jj]);
            }            
      }                              
      nDGD = oOptFiber.GetDGD(oOptSignal.GetCenterFreq())
                    /oOptFiber.GetExpectedDGD();      
      nSndOrderDGD = oOptFiber.GetSndOrderDGD(oOptSignal.GetCenterFreq())
                    /oOptFiber.GetExpectedSndOrderDGD(); 
      nThirdOrderDGD = oOptFiber.GetThirdOrderDGD(oOptSignal.GetCenterFreq())
                   /oOptFiber.GetExpectedThirdOrderDGD();          
      if (   nDGD > oHistActualDGD.GetMinRange()     
          && nDGD < oHistActualDGD.GetMaxRange() ) {          
	 //## I intend to compute the true transition probability
	 //## Therefore, this must come before the decision on the acceptance
         TransProbabilityMatrix[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)
               +qtBins*oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
         NumTransitionsFromBin[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
         if ( oRanPMD.GetRanNum() <
            MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)] 
           /MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)] ) {         
            //## Accept the transition  
            qtAcceptedSamples++;	
            nDGD_Before = nDGD; 
	    nSndOrderDGD_Before = nSndOrderDGD;
	    nThirdOrderDGD_Before = nThirdOrderDGD;	    
            for (int jj = 0; jj < qtStepsFiber; jj++){     
	       ThetaZ[jj] += DeltaThetaZ[jj];               
               PhiZ[jj] += DeltaPhiZ[jj];
               PsiZ[jj] += DeltaPsiZ[jj];   
            } 
         } 	 
      }	 
      oHistActualDGD.AddImportantSample(nDGD_Before,1);                                    
      NumSamplesBinTotal[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;            
      oHistSndOrderDGD.AddImportantSample(nSndOrderDGD_Before,1);   
      oHistThirdOrderDGD.AddImportantSample(nThirdOrderDGD_Before,1);	 
   } //## End of qtStatSamplesStep2   
   cout << "Step II  ->  AcceptanceRate = " 
        << (1.0*qtAcceptedSamples)/qtStatSamplesStep2 << endl; 

   //## Step II-B: Computation of the histogram of the DGD       
   //## First, load the histogram and normalize it 
   //## MMC_pdfTmp[] now contains the weights,
   //## and MMC_pdf contains the last histogram   
        
   //## Make a copy of the previous MMC_pdf
   for (int ii = 0; ii < qtBins; ii++) {          
      MMC_pdfTmp[ii] = MMC_pdf[ii];	 
      NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
   }  
   //## Start the computation of MMC_pdf
   for (int ii = 0; ii < qtBins - 1; ii++) {          
      if ( NormHistogram[ii+1] 
          *NormHistogram[ii] > 0 ) {
         double StatSignifMMC =  NormHistogram[ii+1] 
                                *NormHistogram[ii]
	                  /( NormHistogram[ii+1] 
	                        +NormHistogram[ii] );			
         GlobalStatSignifMMC[ii] += StatSignifMMC;
         double RelStatSignifMMC = StatSignifMMC
                                  /GlobalStatSignifMMC[ii];
         MMC_pdf[ii+1] = MMC_pdf[ii] 
	                *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                *pow( NormHistogram[ii+1] 
       	                     /NormHistogram[ii],
	                     RelStatSignifMMC);
      } else { //## This iteration had no stat signific in this bin
         MMC_pdf[ii+1] = MMC_pdf[ii] 
                 *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
      }	    
   }              
   //## Finally, the normalization      
   double NormMMC_pdfStep2 = 0;   
   for (int ii = 0; ii < qtBins; ii++) {          
      NormMMC_pdfStep2 += MMC_pdf[ii];	    
   }                      
   double Cumulative_pdf_Tmp = 0;
   for (int ii = 0; ii < qtBins; ii++) {   
      MMC_pdf[ii]  /= NormMMC_pdfStep2;   
      MMC_pdfStep2[ii] = MMC_pdf[ii];
      Cumulative_pdf_Tmp += MMC_pdfStep2[ii];
      MMC_cdfStep2[ii] = Cumulative_pdf_Tmp;
      //cout << " *$*$*$*$* " << ii << " -> MMC_pdf = " << MMC_pdf[ii]
      //     << "  MMC_cdf = " << MMC_cdfStep2[ii] << endl; 
   }         

   //##############################################################
   //########### Compute an estimate for the confidence interval         
   if (ErrorPropagationFlag && ! ErrorOfErrorFlag) {
      cout << "Computing the confidence interval" << endl;

//      //## Create a file with the transition matrix
//      ofstream OutFile;
//      string OutFileName = "TransitionMatrix.dat";   
//      //## Plot the pdf of the DGD with its standard deviation
//      OutFile.open(OutFileName.c_str());
//      if (!OutFile) {
//         cout << "Error: " << OutFileName << " could not be created."
//              << endl;
//         exit(1);
//      }
      //## Now, normalize each row of the transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
         if ( !DoNotSpoolErrorFlag) {      
            //cout <<kk<<" -> NumTransitionsFromBin = "<<NumTransitionsFromBin[kk]<<endl;
         }   
	 double CumulativeProb = 0;
         for (int ii = 0; ii < qtBins; ii++) {   
	    if (NumTransitionsFromBin[kk] > 1) {
               VarTransProbabilityMatrix[ii + kk*qtBins] = 
                  ( TransProbabilityMatrix[ii + kk*qtBins] 
                   -sq(TransProbabilityMatrix[ii + kk*qtBins])/NumTransitionsFromBin[kk] )
                  /(NumTransitionsFromBin[kk]*(NumTransitionsFromBin[kk] -1) );
               TransProbabilityMatrix[ii + kk*qtBins] /= NumTransitionsFromBin[kk];	       
	    }   	    
//            OutFile << kk << " " << ii << " " << TransProbabilityMatrix[ii + kk*qtBins] << endl;
	    CumulativeProb += TransProbabilityMatrix[ii + kk*qtBins];
   	    CumulativeTransProbabilityMatrix[ii + kk*qtBins] = CumulativeProb;      
            if ( !DoNotSpoolErrorFlag) {
   	       //cout << kk << " " << ii << " -> " << TransProbabilityMatrix[ii + kk*qtBins]
	       //     << " " << CumulativeProb << " Relative Variation: "
               //     << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins])
               //       /TransProbabilityMatrix[ii + kk*qtBins] << endl;
            }                      
         }	
      }
//      OutFile.close();
      
      //## Now, generate a Monte Carlo ensamble, where the actual transitions
      //## will be done according to the final transition probability
      //## The confidence interval is computed using the Monte Carlo method 

      for (int ii = 0; ii < qtBins; ii++) { 
         SumBinMMC[ii] = 0;
         SumSqBinMMC[ii] = 0;   
      }


      for (int ii = 0; ii < qtBins*qtBins; ii++) {       
         SumTransProbabilityMatrixBoot[ii] = 0;   
         SumSqTransProbabilityMatrixBoot[ii] = 0;   
      }         

      //int NumSamplesForErrorEstim = 1000;  
      //cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;
      
      double *BinProbabilityBootResamp;
      if (ComputeAutocorrFlag) {        
         BinProbabilityBootResamp = new double [qtBins*NumSamplesForErrorEstim];
      }      

      //## Start loop for bootstrap resampling      
      for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {

         for (int ii = 0; ii < qtBins*qtBins; ii++) { 
            TransProbabilityMatrixBootResamp[ii] = 0;	     
         }         
         for (int ii = 0; ii < qtBins; ii++) {       
            NumTransitionsFromBinBootResamp[ii] = 0;  
         }

         for (int ii = 0; ii < qtBins; ii++) {   
            //## Initialize the iterative pdf 
            MMC_pdf[ii] = 1./qtBins;
            GlobalStatSignifMMC[ii] = 0;                    
         }         
         int BinBefore;
         int NewBin;        

         //## Get a randow inicial bin                            
         double RanNumTmp = oRanPMD.GetRanNum();                     
         for (int jj = 0; jj < qtBins; jj++) {                         
            if (   RanNumTmp <= MMC_cdfStep2[jj]) {
               BinBefore = jj;      
   	       break;
   	    }
         }          

         //## Step I
         //## Computation of the MMC weights
         //## Major loop: MMC iterations
         for (int vv = 0; vv < qtIterationsMMC; vv++) {      
            oHistActualDGD.ClearHistogram();
            for (int ii = 0; ii < qtStatSamples; ii++) {  
               //## Get a randow transition. This is the complicated part                             
               double RanNumTmp = oRanPMD.GetRanNum();          
               for (int jj = 0; jj < qtBins; jj++) {                         
	          if (vv == 0) {
                     if (   RanNumTmp <= MMC_cdfStep2[jj]) {
                        NewBin = jj;      
   	                break;
	             }	    
	          } else {
                     if (   RanNumTmp  <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                        NewBin = jj;      
   	                break;
	             }
	          }
               }         
               //## For the Bootstrap estimate of the transition probability matrix               
               TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               NumTransitionsFromBinBootResamp[BinBefore]++;                              
               if ( oRanPMD.GetRanNum() < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
                  //## Accept the transition  
                  BinBefore = NewBin; 
               } 	                                  
               oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
            } //## End of ii (up to qtStatSamples)
   
            //## Here comes Berg's modification      
            //## Make a copy of the previous MMC_pdf
            for (int ii = 0; ii < qtBins; ii++) {          
               MMC_pdfTmp[ii] = MMC_pdf[ii];	 
               //## Attention: the normalization mus be made by qtStatSamplesStep2
               //#  That is because the Step II is the one actually used in the 
               //#  Normalization
               NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
            }  
          
            //## Start the computation of MMC_pdf
            for (int ii = 0; ii < qtBins - 1; ii++) {          
               if ( NormHistogram[ii+1] 
	           *NormHistogram[ii] > 0 ) {
                  double StatSignifMMC =  NormHistogram[ii+1] 
	                              *NormHistogram[ii]
		                    /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );
                  GlobalStatSignifMMC[ii] += StatSignifMMC;
	          double RelStatSignifMMC = StatSignifMMC
                                           /GlobalStatSignifMMC[ii];
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			         *pow( NormHistogram[ii+1] 
       	                              /NormHistogram[ii],
		                       RelStatSignifMMC);
               } else { //## This iteration had no stat signific in this bin
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	       }	    
            }  
                           
            //## Finally, the normalization      
            double NormMMC_pdf = 0;
            for (int ii = 0; ii < qtBins; ii++) {          
               NormMMC_pdf += MMC_pdf[ii];	 
            }                      
            for (int ii = 0; ii < qtBins; ii++) {   
               MMC_pdf[ii]  /= NormMMC_pdf;      	 
            }         
         } //## End of vv (qtIterations)
       
         //#$#$#$#$#$ This is for the computation of the variance
         //######################
         //## Step II: Computation of the probability using the MMC weights
         //## Now, use the MMC weights computed in Step I to make one single 
         //## simulation with the Metropolis Algorithm using the weights
         //##

         //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
         oHistActualDGD.ClearHistogram();             
         for (int ii = 0; ii < qtStatSamplesStep2; ii++) {         
            //## Get a randow transition. This is the complicated part                             
            double RanNumTmp = oRanPMD.GetRanNum();              
            for (int jj = 0; jj < qtBins; jj++) {                         
               if (   RanNumTmp <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                  NewBin = jj;      
	          break;
	       }
            }            
            //## For the Bootstrap estimate of the transition probability matrix      
            TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
            NumTransitionsFromBinBootResamp[BinBefore]++;               
            if ( oRanPMD.GetRanNum()  < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
               //## Accept the transition  
               BinBefore = NewBin; 
            } 	                                  
            oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
         } //## End of qtStatSamplesStep2   

         //## Step II-B: Computation of the histogram of the DGD       
         //## First, load the histogram and normalize it 
         //## MMC_pdfTmp[] now contains the weights,
         //## and MMC_pdf contains the last histogram   
        
         //## Make a copy of the previous MMC_pdf
         for (int ii = 0; ii < qtBins; ii++) {          
            MMC_pdfTmp[ii] = MMC_pdf[ii];	 
            NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
         }  
         //## Start the computation of MMC_pdf
         for (int ii = 0; ii < qtBins - 1; ii++) {          
            if ( NormHistogram[ii+1] 
                *NormHistogram[ii] > 0 ) {
               double StatSignifMMC =  NormHistogram[ii+1] 
                                      *NormHistogram[ii]
	                        /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );			
               GlobalStatSignifMMC[ii] += StatSignifMMC;
               double RelStatSignifMMC = StatSignifMMC
                                        /GlobalStatSignifMMC[ii];
               MMC_pdf[ii+1] = MMC_pdf[ii] 
	                      *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                      *pow( NormHistogram[ii+1] 
       	                           /NormHistogram[ii],
	                           RelStatSignifMMC);
            } else { //## This iteration had no stat signific in this bin
               MMC_pdf[ii+1] = MMC_pdf[ii] 
                       *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
            }	    
         }       
         double NormMMC_pdf = 0;       
         //## Finally, the normalization      
         for (int ii = 0; ii < qtBins; ii++) {          
            NormMMC_pdf += MMC_pdf[ii];	      
         }                      
         for (int ii = 0; ii < qtBins; ii++) {   
            MMC_pdf[ii]  /= NormMMC_pdf;          
            SumBinMMC[ii] += MMC_pdf[ii];
            SumSqBinMMC[ii] += sq(MMC_pdf[ii]);                        
         }   

         if (ComputeAutocorrFlag) {       
            for (int ii = 0; ii < qtBins; ii++) {             
               BinProbabilityBootResamp[ii + kk*qtBins] =  MMC_pdf[ii];
            }   
         }           

         //## Now, normalize each row of the Bootstrap resampled transition probability matrix
         for (int kk = 0; kk < qtBins; kk++) {        
            //cout <<kk<<" -> NumTransitionsFromBinBootResamp = "<<NumTransitionsFromBinBootResamp[kk]<<endl;
	    double CumulativeProb = 0;
            for (int ii = 0; ii < qtBins; ii++) {   
	       if (NumTransitionsFromBinBootResamp[kk] != 0) {
                  TransProbabilityMatrixBootResamp[ii + kk*qtBins] 
                        /= NumTransitionsFromBinBootResamp[kk];	       
	       }   	    
            }	
         }

         for (int ii = 0; ii < qtBins*qtBins; ii++) {        
            SumTransProbabilityMatrixBoot[ii] += TransProbabilityMatrixBootResamp[ii];
            SumSqTransProbabilityMatrixBoot[ii] += sq(TransProbabilityMatrixBootResamp[ii]);       
         }            
         
                             
      } //## End of kk      

      for (int nn = 0; nn < qtBins; nn++) { 
         //MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
         MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
	 VarianceBinThisIterMMC[nn] = NumSamplesForErrorEstim
	       /(NumSamplesForErrorEstim - 1.)
	       *( SumSqBinMMC[nn]/NumSamplesForErrorEstim
	         -sq(SumBinMMC[nn]/NumSamplesForErrorEstim) ); 
         if ( !DoNotSpoolErrorFlag) {		 	    
            cout << nn << " -> MeanBinMMC = " << MMC_pdf[nn] << "  VarianceBinThisIterMMC = " 
                 << VarianceBinThisIterMMC[nn] << endl;	    	    	 
            cout << nn << " ####  -> RelativeVariationMMC = " 
                 <<  sqrt(VarianceBinThisIterMMC[nn])/MMC_pdf[nn] << endl;   
         }       
      } //## End of bin loop: nn	

      {
      //## Create a file with the transition matrix
      ofstream OutFile;
      string OutFileName = "TransitionMatrix.dat";   
      //## Plot the pdf of the DGD with its standard deviation
      OutFile.open(OutFileName.c_str());
      if (!OutFile) {
         cout << "Error: " << OutFileName << " could not be created."
              << endl;
         exit(1);
      }

      //## Compute the variance of the bootstrap resampled transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
         for (int ii = 0; ii < qtBins; ii++) {   
	    if (SumTransProbabilityMatrixBoot[ii + kk*qtBins] != 0) {
               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] =             
                  NumSamplesForErrorEstim/(NumSamplesForErrorEstim - 1.)
	       *( SumSqTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim
	         -sq(SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim) ); 
//               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] = 
//                  ( SumSqTransProbabilityMatrixBoot[ii + kk*qtBins] 
//                   -sq(SumTransProbabilityMatrixBoot[ii + kk*qtBins])/NumSamplesForErrorEstim )
//                  /(NumSamplesForErrorEstim*(NumSamplesForErrorEstim -1) );	       
	    }  else {
               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] = 0;
            }
            if ( !DoNotSpoolErrorFlag) {
	       cout << " ** " << kk << " " << ii << " -> " 
                    << TransProbabilityMatrix[ii + kk*qtBins]                 
	            << " Relative_Variation: " << sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins])
                      /(SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim) << endl;
            }      
            double pdfRelativeStdDevTmp = 99;
            if (SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim != 0) {
               pdfRelativeStdDevTmp = sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins]) 
                                     / (SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim);
            }               
            //## Layout: kk ii Estimated_From_1_MMC MeanBootStrap StdDevBootStrap RelStdDev
            OutFile << kk << " " << ii << " " 
                    << TransProbabilityMatrix[ii + kk*qtBins] << "   "
                    << SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim << "  "
                    << sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins]) << "  "
                    << pdfRelativeStdDevTmp << endl;
         }	
      }
      OutFile.close();
      }
           

      if (ComputeAutocorrFlag) {
         double *AutocorrelationBins = new double[qtBins];
         for (int ii = 0; ii < qtBins; ii++) {             
            AutocorrelationBins[ii] = 0;
         }   
         for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {            
            for (int ii = 0; ii < qtBins; ii++) {             
               AutocorrelationBins[ii] += 
                  (BinProbabilityBootResamp[BinAutocorrelation + kk*qtBins]
                   -SumBinMMC[BinAutocorrelation]/NumSamplesForErrorEstim)
                 *(BinProbabilityBootResamp[ii + kk*qtBins]
                   -SumBinMMC[ii]/NumSamplesForErrorEstim);
            }   
         }            
         //## Normalization of autocorrelation         
         //double NormAutocorrCoeff = AutocorrelationBins[BinAutocorrelation];          
         for (int ii = 0; ii < qtBins; ii++) {           
            AutocorrelationBins[ii] /= (NumSamplesForErrorEstim - 1)
                 *sqrt(VarianceBinThisIterMMC[BinAutocorrelation])
                 *sqrt(VarianceBinThisIterMMC[ii]);
         }                        
         
         string OutFileName = "AutocorrMMC_Prob.dat";
         ofstream OutFile(OutFileName.c_str(),ios::out);
         if (!OutFile) {
            cout << "Error: " << OutFileName << " could not be created."
                 << endl;
            exit(1);
         }
         for (int ii = 0; ii < qtBins; ii++) {   
            OutFile << oHistActualDGD.Get_pdfX(ii) << "  " 
                    << AutocorrelationBins[ii] << endl;
         }         
         OutFile.close();
         
         delete [] BinProbabilityBootResamp;
         delete [] AutocorrelationBins;
      }  
      
      
   } //## End of if ErrorPropagationFlag AND NOT ErrorOfErrorFlag
   


   //##############################################################
   //########### Compute an estimate for the confidence interval  
   //## and for the confidence interval of the error       
   if (ErrorPropagationFlag && ErrorOfErrorFlag) {
      cout << "Computing the confidence interval of probability and error" << endl;   


     //## Now, normalize each row of the transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
//         if ( !DoNotSpoolErrorFlag) {      
//            //cout <<kk<<" -> NumTransitionsFromBin = "<<NumTransitionsFromBin[kk]<<endl;
//         }   
	 double CumulativeProb = 0;
         for (int ii = 0; ii < qtBins; ii++) {    
	    if (NumTransitionsFromBin[kk] > 0) {
  	       if (NumTransitionsFromBin[kk] > 1) {            
                  VarTransProbabilityMatrix[ii + kk*qtBins] = 
                     ( TransProbabilityMatrix[ii + kk*qtBins] 
                      -sq(TransProbabilityMatrix[ii + kk*qtBins])/NumTransitionsFromBin[kk] )
                     /(NumTransitionsFromBin[kk]*(NumTransitionsFromBin[kk] -1) );
               }   
               TransProbabilityMatrix[ii + kk*qtBins] /= NumTransitionsFromBin[kk];	       
	    }   	    
//            OutFile << kk << " " << ii << " " << TransProbabilityMatrix[ii + kk*qtBins] << endl;
	    CumulativeProb += TransProbabilityMatrix[ii + kk*qtBins];
   	    CumulativeTransProbabilityMatrix[ii + kk*qtBins] = CumulativeProb;      
            if ( !DoNotSpoolErrorFlag) {
   	       //cout << kk << " " << ii << " -> " << TransProbabilityMatrix[ii + kk*qtBins]
	       //     << " " << CumulativeProb << " Relative Variation: "
               //     << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins])
               //       /TransProbabilityMatrix[ii + kk*qtBins] << endl;
            }                      
         }	
      }      
      //## Now, the estimated transition matrix and its cdf are available for use
      //## I should not generate one pseudo-transition matrix and compute 
      //## NumSamplesForErrorEstim MMC ensembles for that transition matrix
      //## to obtain an error estimate. I should repeat this process NumBootstrapTransMatrices      
      //## times to obtain a mean and a standard deviation for the error estimate.

      
//      OutFile.close();
      
      //## Now, generate a Monte Carlo ensamble, where the actual transitions
      //## will be done according to the final transition probability
      //## The confidence interval is computed using the Monte Carlo method 

      for (int ii = 0; ii < qtBins; ii++) {          
         SumStdDevBinMMC[ii] = 0;
         SumSqStdDevBinMMC[ii] = 0;         
      }


//      for (int ii = 0; ii < qtBins*qtBins; ii++) {       
//         SumTransProbabilityMatrixBoot[ii] = 0;   
//         SumSqTransProbabilityMatrixBoot[ii] = 0;   
//      }         

      //int NumSamplesForErrorEstim = 1000;  
      //cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;
      
//      double *BinProbabilityBootResamp;
//      if (ComputeAutocorrFlag) {        
//         BinProbabilityBootResamp = new double [qtBins*NumSamplesForErrorEstim];
//      }      

      //## Start loop for bootstrap resampling for each bootstraped transition matrix     
//      for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {
      for (int kk = 0; kk < NumBootstrapTransMatrices; kk++) {
         cout << "Computing matrix # " << kk << " out of " << NumBootstrapTransMatrices <<endl;

         for (int ii = 0; ii < qtBins*qtBins; ii++) { 
            TransProbabilityMatrixBootResamp[ii] = 0;	     
         }         
         for (int ii = 0; ii < qtBins; ii++) {       
            NumTransitionsFromBinBootResamp[ii] = 0;  
         }

         for (int ii = 0; ii < qtBins; ii++) {   
            //## Initialize the iterative pdf 
            MMC_pdf[ii] = 1./qtBins;
            GlobalStatSignifMMC[ii] = 0;                    
         }         
         int BinBefore;
         int NewBin;        

         //## Get a randow inicial bin                            
         double RanNumTmp = oRanPMD.GetRanNum();                     
         for (int jj = 0; jj < qtBins; jj++) {                         
            if (   RanNumTmp <= MMC_cdfStep2[jj]) {
               BinBefore = jj;      
   	       break;
   	    }
         }          

         //## Step I
         //## Computation of the MMC weights
         //## Major loop: MMC iterations
         for (int vv = 0; vv < qtIterationsMMC; vv++) {      
            oHistActualDGD.ClearHistogram();
            for (int ii = 0; ii < qtStatSamples; ii++) {  
               //## Get a randow transition. This is the complicated part                             
               double RanNumTmp = oRanPMD.GetRanNum();          
               for (int jj = 0; jj < qtBins; jj++) {                         
	          if (vv == 0) {
                     if (   RanNumTmp <= MMC_cdfStep2[jj]) {
                        NewBin = jj;      
   	                break;
	             }	    
	          } else {
                     if (   RanNumTmp  <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                        NewBin = jj;      
   	                break;
	             }
	          }
               }         
               //## For the Bootstrap estimate of the transition probability matrix               
               TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               NumTransitionsFromBinBootResamp[BinBefore]++;                              
               if ( oRanPMD.GetRanNum() < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
                  //## Accept the transition  
                  BinBefore = NewBin; 
               } 	                                  
               oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
            } //## End of ii (up to qtStatSamples)
   
            //## Here comes Berg's modification      
            //## Make a copy of the previous MMC_pdf
            for (int ii = 0; ii < qtBins; ii++) {          
               MMC_pdfTmp[ii] = MMC_pdf[ii];	 
               //## Attention: the normalization mus be made by qtStatSamplesStep2
               //#  That is because the Step II is the one actually used in the 
               //#  Normalization
               NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
            }  
          
            //## Start the computation of MMC_pdf
            for (int ii = 0; ii < qtBins - 1; ii++) {          
               if ( NormHistogram[ii+1] 
	           *NormHistogram[ii] > 0 ) {
                  double StatSignifMMC =  NormHistogram[ii+1] 
	                              *NormHistogram[ii]
		                    /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );
                  GlobalStatSignifMMC[ii] += StatSignifMMC;
	          double RelStatSignifMMC = StatSignifMMC
                                           /GlobalStatSignifMMC[ii];
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			         *pow( NormHistogram[ii+1] 
       	                              /NormHistogram[ii],
		                       RelStatSignifMMC);
               } else { //## This iteration had no stat signific in this bin
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	       }	    
            }  
                           
            //## Finally, the normalization      
            double NormMMC_pdf = 0;
            for (int ii = 0; ii < qtBins; ii++) {          
               NormMMC_pdf += MMC_pdf[ii];	 
            }                      
            for (int ii = 0; ii < qtBins; ii++) {   
               MMC_pdf[ii]  /= NormMMC_pdf;      	 
            }         
         } //## End of vv (qtIterations)
       
         //#$#$#$#$#$ This is for the computation of the variance
         //######################
         //## Step II: Computation of the probability using the MMC weights
         //## Now, use the MMC weights computed in Step I to make one single 
         //## simulation with the Metropolis Algorithm using the weights
         //##

         //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
         oHistActualDGD.ClearHistogram();             
         for (int ii = 0; ii < qtStatSamplesStep2; ii++) {         
            //## Get a randow transition. This is the complicated part                             
            double RanNumTmp = oRanPMD.GetRanNum();              
            for (int jj = 0; jj < qtBins; jj++) {                         
               if (   RanNumTmp <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                  NewBin = jj;      
	          break;
	       }
            }            
            //## For the Bootstrap estimate of the transition probability matrix      
            TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
            NumTransitionsFromBinBootResamp[BinBefore]++;               
            if ( oRanPMD.GetRanNum()  < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
               //## Accept the transition  
               BinBefore = NewBin; 
            } 	                                  
            oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
         } //## End of qtStatSamplesStep2   

         //## Step II-B: Computation of the histogram of the DGD       
         //## First, load the histogram and normalize it 
         //## MMC_pdfTmp[] now contains the weights,
         //## and MMC_pdf contains the last histogram   
        
         //## Make a copy of the previous MMC_pdf
         for (int ii = 0; ii < qtBins; ii++) {          
            MMC_pdfTmp[ii] = MMC_pdf[ii];	 
            NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
         }  
         //## Start the computation of MMC_pdf
         for (int ii = 0; ii < qtBins - 1; ii++) {          
            if ( NormHistogram[ii+1] 
                *NormHistogram[ii] > 0 ) {
               double StatSignifMMC =  NormHistogram[ii+1] 
                                      *NormHistogram[ii]
	                        /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );			
               GlobalStatSignifMMC[ii] += StatSignifMMC;
               double RelStatSignifMMC = StatSignifMMC
                                        /GlobalStatSignifMMC[ii];
               MMC_pdf[ii+1] = MMC_pdf[ii] 
	                      *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                      *pow( NormHistogram[ii+1] 
       	                           /NormHistogram[ii],
	                           RelStatSignifMMC);
            } else { //## This iteration had no stat signific in this bin
               MMC_pdf[ii+1] = MMC_pdf[ii] 
                       *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
            }	    
         }       
         double NormMMC_pdf = 0;       
         //## Finally, the normalization      
         for (int ii = 0; ii < qtBins; ii++) {          
            NormMMC_pdf += MMC_pdf[ii];	      
         }                      
         
         double Cumulative_pdf_Tmp = 0;
         for (int ii = 0; ii < qtBins; ii++) {   
            MMC_pdf[ii]  /= NormMMC_pdf;          
            //SumBinMMC[ii] += MMC_pdf[ii];
            //SumSqBinMMC[ii] += sq(MMC_pdf[ii]);                        
            //MMC_cdfStep2_BootResamp[ii] = +MMC_pdf[ii];            
            Cumulative_pdf_Tmp += MMC_pdf[ii];
            MMC_cdfStep2_BootResamp[ii] = Cumulative_pdf_Tmp;            
         }   

         //if (ComputeAutocorrFlag) {       
         //   for (int ii = 0; ii < qtBins; ii++) {             
         //      BinProbabilityBootResamp[ii + kk*qtBins] =  MMC_pdf[ii];
         //   }   
         //}           

         //## Now, normalize each row of the Bootstrap resampled transition probability matrix
         for (int jj = 0; jj < qtBins; jj++) {        
            //cout <<jj<<" -> NumTransitionsFromBinBootResamp = "<<NumTransitionsFromBinBootResamp[jj]<<endl;
	    double CumulativeProb = 0;
            for (int ii = 0; ii < qtBins; ii++) {   
	       if (NumTransitionsFromBinBootResamp[jj] != 0) {
                  TransProbabilityMatrixBootResamp[ii + jj*qtBins] 
                        /= NumTransitionsFromBinBootResamp[jj];	       
	       }   	    
   	       CumulativeProb += TransProbabilityMatrixBootResamp[ii + jj*qtBins];
   	       CumulativeTransProbabilityMatrixBootResamp[ii + jj*qtBins] = CumulativeProb;                   
            }	
         }

         //for (int ii = 0; ii < qtBins*qtBins; ii++) {        
         //   SumTransProbabilityMatrixBoot[ii] += TransProbabilityMatrixBootResamp[ii];
         //   SumSqTransProbabilityMatrixBoot[ii] += sq(TransProbabilityMatrixBootResamp[ii]);       
         //}          

      for (int ii = 0; ii < qtBins; ii++) { 
         SumBinMMC[ii] = 0;
         SumSqBinMMC[ii] = 0;   
      }

      for (int tt = 0; tt < NumSamplesForErrorEstim; tt++) {         
         //## Now, I should compute the bootstrap variance using the kk-th bootstraped 
         //## transition matrix                                                                 
  
         for (int ii = 0; ii < qtBins; ii++) {   
            //## Initialize the iterative pdf 
            MMC_pdf[ii] = 1./qtBins;
            GlobalStatSignifMMC[ii] = 0;                    
         }         
         int BinBefore;
         int NewBin;        

         //## Get a randow inicial bin                            
         double RanNumTmp = oRanPMD.GetRanNum();                     
         for (int jj = 0; jj < qtBins; jj++) {                         
            if (   RanNumTmp <= MMC_cdfStep2_BootResamp[jj]) {
               BinBefore = jj;      
   	       break;
   	    }
         }          

         //## Step I
         //## Computation of the MMC weights
         //## Major loop: MMC iterations
         for (int vv = 0; vv < qtIterationsMMC; vv++) {      
            oHistActualDGD.ClearHistogram();
            for (int ii = 0; ii < qtStatSamples; ii++) {  
               //## Get a randow transition. This is the complicated part                             
               double RanNumTmp = oRanPMD.GetRanNum();          
               for (int jj = 0; jj < qtBins; jj++) {                         
	          if (vv == 0) {
                     if (   RanNumTmp <= MMC_cdfStep2_BootResamp[jj]) {
                        NewBin = jj;      
   	                break;
	             }	    
	          } else {
                     if (   RanNumTmp  <= CumulativeTransProbabilityMatrixBootResamp[jj + BinBefore*qtBins]) {
                        NewBin = jj;      
   	                break;
	             }
	          }
               }         
               //## For the Bootstrap estimate of the transition probability matrix               
               ///TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               ///NumTransitionsFromBinBootResamp[BinBefore]++;                              
               if ( oRanPMD.GetRanNum() < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
                  //## Accept the transition  
                  BinBefore = NewBin; 
               } 	                                  
               oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
            } //## End of ii (up to qtStatSamples)
   
            //## Here comes Berg's modification      
            //## Make a copy of the previous MMC_pdf
            for (int ii = 0; ii < qtBins; ii++) {          
               MMC_pdfTmp[ii] = MMC_pdf[ii];	 
               //## Attention: the normalization mus be made by qtStatSamplesStep2
               //#  That is because the Step II is the one actually used in the 
               //#  Normalization
               NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
            }  
          
            //## Start the computation of MMC_pdf
            for (int ii = 0; ii < qtBins - 1; ii++) {          
               if ( NormHistogram[ii+1] 
	           *NormHistogram[ii] > 0 ) {
                  double StatSignifMMC =  NormHistogram[ii+1] 
	                              *NormHistogram[ii]
		                    /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );
                  GlobalStatSignifMMC[ii] += StatSignifMMC;
	          double RelStatSignifMMC = StatSignifMMC
                                           /GlobalStatSignifMMC[ii];
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			         *pow( NormHistogram[ii+1] 
       	                              /NormHistogram[ii],
		                       RelStatSignifMMC);
               } else { //## This iteration had no stat signific in this bin
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	       }	    
            }  
                           
            //## Finally, the normalization      
            double NormMMC_pdf = 0;
            for (int ii = 0; ii < qtBins; ii++) {          
               NormMMC_pdf += MMC_pdf[ii];	 
            }                      
            for (int ii = 0; ii < qtBins; ii++) {   
               MMC_pdf[ii]  /= NormMMC_pdf;      	 
            }         
         } //## End of vv (qtIterations)
       
         //#$#$#$#$#$ This is for the computation of the variance
         //######################
         //## Step II: Computation of the probability using the MMC weights
         //## Now, use the MMC weights computed in Step I to make one single 
         //## simulation with the Metropolis Algorithm using the weights
         //##

         //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
         oHistActualDGD.ClearHistogram();             
         for (int ii = 0; ii < qtStatSamplesStep2; ii++) {         
            //## Get a randow transition. This is the complicated part                             
            double RanNumTmp = oRanPMD.GetRanNum();              
            for (int jj = 0; jj < qtBins; jj++) {                         
               if (   RanNumTmp <= CumulativeTransProbabilityMatrixBootResamp[jj + BinBefore*qtBins]) {
                  NewBin = jj;      
	          break;
	       }
            }            
            //## For the Bootstrap estimate of the transition probability matrix      
            //TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
            //NumTransitionsFromBinBootResamp[BinBefore]++;               
            if ( oRanPMD.GetRanNum()  < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
               //## Accept the transition  
               BinBefore = NewBin; 
            } 	                                  
            oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
         } //## End of qtStatSamplesStep2   

         //## Step II-B: Computation of the histogram of the DGD       
         //## First, load the histogram and normalize it 
         //## MMC_pdfTmp[] now contains the weights,
         //## and MMC_pdf contains the last histogram   
        
         //## Make a copy of the previous MMC_pdf
         for (int ii = 0; ii < qtBins; ii++) {          
            MMC_pdfTmp[ii] = MMC_pdf[ii];	 
            NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
         }  
         //## Start the computation of MMC_pdf
         for (int ii = 0; ii < qtBins - 1; ii++) {          
            if ( NormHistogram[ii+1] 
                *NormHistogram[ii] > 0 ) {
               double StatSignifMMC =  NormHistogram[ii+1] 
                                      *NormHistogram[ii]
	                        /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );			
               GlobalStatSignifMMC[ii] += StatSignifMMC;
               double RelStatSignifMMC = StatSignifMMC
                                        /GlobalStatSignifMMC[ii];
               MMC_pdf[ii+1] = MMC_pdf[ii] 
	                      *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                      *pow( NormHistogram[ii+1] 
       	                           /NormHistogram[ii],
	                           RelStatSignifMMC);
            } else { //## This iteration had no stat signific in this bin
               MMC_pdf[ii+1] = MMC_pdf[ii] 
                       *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
            }	    
         }       
         double NormMMC_pdf = 0;       
         //## Finally, the normalization      
         for (int ii = 0; ii < qtBins; ii++) {          
            NormMMC_pdf += MMC_pdf[ii];	      
         }                      
         for (int ii = 0; ii < qtBins; ii++) {   
            MMC_pdf[ii]  /= NormMMC_pdf;          
            SumBinMMC[ii] += MMC_pdf[ii];
            SumSqBinMMC[ii] += sq(MMC_pdf[ii]);                        
         }            
         //## End of the computation of a value of MMC_pdf

      } //## End tt  

         for (int nn = 0; nn < qtBins; nn++) { 
            //MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
            //MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
   	    VarianceBinMMC[nn] = NumSamplesForErrorEstim
   	       /(NumSamplesForErrorEstim - 1.)
	       *( SumSqBinMMC[nn]/NumSamplesForErrorEstim
	         -sq(SumBinMMC[nn]/NumSamplesForErrorEstim) ); 
            //cout << kk << " ##### RelVar = " << sqrt(VarianceBinMMC[nn])
            //      /(SumBinMMC[nn]/NumSamplesForErrorEstim) << endl;
            //## Standard deviations     
            SumStdDevBinMMC[nn] += sqrt(VarianceBinMMC[nn]);
            SumSqStdDevBinMMC[nn] += VarianceBinMMC[nn];

//            if ( !DoNotSpoolErrorFlag) {		 	    
//               cout << nn << " -> MeanBinMMC = " << MMC_pdf[nn] << "  VarianceBinThisIterMMC = " 
//                    << VarianceBinThisIterMMC[nn] << endl;	    	    	 
//               cout << nn << " ####  -> RelativeVariationMMC = " 
//                    <<  sqrt(VarianceBinThisIterMMC[nn])/MMC_pdf[nn] << endl;   
//            }       
         } //## End of bin loop: nn	          
      
      
      } //## End of kk   


      for (int nn = 0; nn < qtBins; nn++) { 
            //## Do not use this line ##//  MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
            MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
            MeanStdDevBinMMC[nn] = SumStdDevBinMMC[nn]/NumBootstrapTransMatrices;
            VarianceBinMMC[nn] = sq(MeanStdDevBinMMC[nn]);
   	    VarianceStdDevBinMMC[nn] = NumBootstrapTransMatrices
   	       /(NumBootstrapTransMatrices - 1.)
	       *( SumSqStdDevBinMMC[nn]/NumBootstrapTransMatrices
	         -sq(SumStdDevBinMMC[nn]/NumBootstrapTransMatrices) );                  
 
      } //## End of bin loop: nn	                      
      
   } //## End of if ErrorPropagationFlag AND ErrorOfErrorFlag
   //#####################################################################3

   //##
   //## Now, plot the histogram after the kk iteration
   //##      
   string OutFileName = "nDGDpdf.dat";
   //## Plot the pdf of the DGD with its standard deviation
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }

   oHistDGD.ClearHistogram();      
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfX = oHistDGD.Get_pdfX(ii);
      double NormConstant = (oHistDGD.GetMaxRange() - oHistDGD.GetMinRange() )
	                   /qtBins;
		   
      double pdfYtmp = MMC_pdf[ii]/NormConstant;
      double pdfConfidence = sqrt(VarianceBinMMC[ii])/NormConstant;
      
      double pdfConfidenceOfConfidence = sqrt(VarianceStdDevBinMMC[ii])/NormConstant;
      
      
            
      double pdfRelativeStdDevTmp = 99;
      if (pdfYtmp != 0) {
         pdfRelativeStdDevTmp = pdfConfidence / pdfYtmp;
      }   
      double pdfRelativeStdDevStdDevTmp = 99;
      if (pdfConfidence != 0) {
         pdfRelativeStdDevStdDevTmp = pdfConfidenceOfConfidence / pdfConfidence;
      }   
      if (! ErrorOfErrorFlag) {
         OutFile << pdfX << "   " 
                 << pdfYtmp << "   "  
                 << pdfConfidence << "   "
   	         << pdfRelativeStdDevTmp << endl;   
      } else{
         OutFile << pdfX << "   " 
                 << pdfYtmp << "   "  
                 << pdfConfidence << "   "
   	         << pdfRelativeStdDevTmp << "  "
                 << pdfConfidenceOfConfidence << "   "
   	         << pdfRelativeStdDevStdDevTmp << endl;         
      }                 
   }           
   OutFile.close();     

   //## Creates a file with the Maxwellian pdf with the same mean
   int qtPointsMaxwell = 1000;     
   oOptFiber.WriteFileMaxwell_NormDGD_pdf("Maxwell_pdf.dat",oHistDGD.GetMinRange(),
         oHistDGD.GetMaxRange(),qtPointsMaxwell);
   oOptFiber.WriteFileTrueCSM_NormDGD_pdf("TrueCSM_DGD_pdf.dat",oHistDGD.GetMinRange(),
         oHistDGD.GetMaxRange(),qtPointsMaxwell);  
   oOptFiber.WriteFileTrueCSM_NormDGD_pdf("TrueCSM_DGD_pdf.dat",oHistDGD.GetMinRange(),
         oHistDGD.GetMaxRange(),oHistDGD.GetQtBins());  

   //## Plot the pdf of the histogram with the confidence interval
   OutFileName = "nActualDGDpdf.dat";   
   //## Plot the pdf of the DGD with its standard deviation
   OutFile.open(OutFileName.c_str());
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfX = oHistActualDGD.Get_pdfX(ii);
      double NormConstant = ( oHistActualDGD.GetMaxRange() 
                             -oHistActualDGD.GetMinRange() )
	                    /qtBins;
      double pdfYtmp = NumSamplesBinTotal[ii]
                      /(1.*qtIterationsMMC*qtStatSamples + qtStatSamplesStep2)
                      /NormConstant;    
      OutFile << pdfX << "   " 
              << pdfYtmp << "   "  
              << endl;   
   }           
   OutFile.close();
   
   oHistActualDGD.WriteFile_pdf("nActualDGDpdf2.dat");   
   oHistSndOrderDGD.WriteFile_pdf("nSndOrderDGDpdf.dat");   
   oHistThirdOrderDGD.WriteFile_pdf("nThirdOrderDGDpdf.dat");  
            
   double nDGD_Mean = oHistDGD.GetMean();
   cout << "nActualDGDmean = " << oHistActualDGD.GetMean() << endl;
   cout << "qtStatSamples = " << oHistActualDGD.GetQtSamples() << endl;
   cout << " Percentage inside the range nDGD = "<<oHistActualDGD.GetMinRange()
        << " to " << oHistActualDGD.GetMaxRange() << " = " 
        << oHistActualDGD.GetQtSamplesInRange()/ 
           (double) oHistActualDGD.GetQtSamples()*100. << "%." << endl;
   cout << "nSndOrderDGDmean = " << oHistSndOrderDGD.GetMean() << endl;
   cout << "nThirdOrderDGDmean = " << oHistThirdOrderDGD.GetMean() << endl;
   
   delete [] MMC_pdf;
   delete [] MMC_pdfTmp;
   delete [] GlobalStatSignifMMC;
   delete [] MMC_pdfStep2; 
   delete [] MMC_cdfStep2;

   delete [] MMC_cdfStep2_BootResamp;

   delete [] NormHistogram;   
   delete [] VarianceBinMMC;
   delete [] SumBinMMC;
   delete [] SumSqBinMMC;

   delete [] SumStdDevBinMMC;
   delete [] SumSqStdDevBinMMC;
   
   delete [] MeanStdDevBinMMC;
   delete [] VarianceStdDevBinMMC;   

   delete [] TransProbabilityMatrix;
   delete [] NumTransitionsFromBin;
   delete [] CumulativeTransProbabilityMatrix;
   delete [] VarTransProbabilityMatrix;   

   delete [] TransProbabilityMatrixBootResamp;
   delete [] NumTransitionsFromBinBootResamp;
   delete [] SumTransProbabilityMatrixBoot;   
   delete [] SumSqTransProbabilityMatrixBoot;   
   delete [] VarTransProbabilityMatrixBootEnsemble; 
   delete [] CumulativeTransProbabilityMatrixBootResamp;
   
   delete [] NumSamplesBinTotal;   
   delete [] ThetaZ;   
   delete [] DeltaThetaZ;
   delete [] PhiZ;
   delete [] DeltaPhiZ;   
   delete [] PsiZ;
   delete [] DeltaPsiZ;  

}  //## End of: void GetDGDpdfMMC_Berg(MyStartup *oMyStartup) 

























//#############################################################
//##### Creates a file with the pdf of the DGD
//##    using the Multicanonical Monte Carlo implementation by Berg
//##    I am implementing the two-step procedure that was introduced
//##    by Berg.
//##    I am implementing the enhanced version of the two-step procedure.
//##    THIS IS THE MONTE CARLO ERROR PROPAGATION METHOD
//##    The error of the error is computed using the Jackknife method
//## By: Aurenice Lima in 05/19/04
void GetDGDpdfMMC_Berg_ErrorOfErrorUsingJackknife(MyStartup *oMyStartup) //#$%#$%
////////void GetDGDpdfMMC_BergMonteCarloWithTransitionProbabilityMatrix051904(MyStartup *oMyStartup) 
{
   Histogram oHistDGD("HistDGD.in");
   Histogram oHistActualDGD("HistActualDGD.in");   
   Histogram oHistSndOrderDGD("HistSndOrdDGD.in");    
   Histogram oHistThirdOrderDGD("HistThirdOrdDGD.in");       

   int qtIterationsMMC = oMyStartup->GetIntParam("qtIterationsMMC");
   cout << "qtIterationsMMC = " << qtIterationsMMC << endl; 

   double RelStepSizeMMC = oMyStartup->GetDoubleParam("RelStepSizeMMC");
   cout << "RelStepSizeMMC = " << RelStepSizeMMC << endl; 

   int qtStatSamples = oMyStartup->GetQtStatSamples();
   cout << "qtStatSamples = " << qtStatSamples << endl;    
 
   int ErrorPropagationFlag = oMyStartup->GetIntParam("ErrorPropagationFlag");
   cout << "ErrorPropagationFlag = " << ErrorPropagationFlag << endl;   

   int NumSamplesForErrorEstim = oMyStartup->GetIntParam("NumSamplesForErrorEstim");  
   if (NumSamplesForErrorEstim == 0) {
      NumSamplesForErrorEstim = 1000;
   }
   cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;

   int ErrorOfErrorFlag = oMyStartup->GetIntParam("ErrorOfErrorFlag");
   cout << "ErrorOfErrorFlag = " << ErrorOfErrorFlag << endl;   
   if (! ErrorOfErrorFlag) {
      cout << "You must set ErrorOfErrorFlag = 1 to run this simulation" << endl;
      exit(1);
   }
   
//   int NumBootstrapTransMatrices = oMyStartup->GetIntParam("NumBootstrapTransMatrices");  
//   if (NumBootstrapTransMatrices == 0) {
//      NumBootstrapTransMatrices = 100;
//   }
//   cout << "NumBootstrapTransMatrices = " << NumBootstrapTransMatrices << endl;

   int DoNotSpoolErrorFlag = oMyStartup->GetIntParam("DoNotSpoolErrorFlag");
   cout << "DoNotSpoolErrorFlag = " << DoNotSpoolErrorFlag << endl;   

   int FiberRotatorModelFlag = oMyStartup->GetIntParam("FiberRotatorModelFlag");
   cout << "FiberRotatorModelFlag = " << FiberRotatorModelFlag << endl;    

   cout << "Creating a file with the pdf of the DGD." << endl;
   cout << "Plot the result with the Gnuplot script: nDGDpdf.gp" << endl;      

   //## This parameter should not be changed if one wants to use the covariance
   int qtStatSamplesStep2 = oMyStartup->GetIntParam("qtStatSamplesStep2");      
   if (qtStatSamplesStep2 == 0) {
      cout << "  *** Forcing a value for qtStatSamplesStep2 = qtStatSamples" <<endl;
      qtStatSamplesStep2 = qtStatSamples;
   }
   cout << "qtStatSamplesStep2 = " << qtStatSamplesStep2 << endl;

   int ComputeAutocorrFlag = oMyStartup->GetIntParam("ComputeAutocorrFlag");
   cout << "ComputeAutocorrFlag = " << ComputeAutocorrFlag << endl;    

   int BinAutocorrelation;   
   if (ComputeAutocorrFlag) {
      BinAutocorrelation = oMyStartup->GetIntParam("BinAutocorrelation");
      cout << "BinAutocorrelation = " << BinAutocorrelation << endl;          
   }

   int JackknifeBlockSize = oMyStartup->GetIntParam("JackknifeBlockSize");  
   if (JackknifeBlockSize == 0) {
      JackknifeBlockSize = 100;
   }
   cout << "JackknifeBlockSize = " << JackknifeBlockSize << endl;
   
   //## Computation of the actual Jackknife block size
   int EstimatedBlockSize = JackknifeBlockSize; 
   //cout << "EstimatedBlockSize = " << EstimatedBlockSize << endl;

   int qtTotalStatSamples = qtStatSamples*qtIterationsMMC 
                              +qtStatSamplesStep2;

   //## Computation of the variance using the Jackknife method
   int EstimateNumJackknifeBlocks =  (int) ( qtTotalStatSamples
                                            /EstimatedBlockSize );
   int NumJackknifeBlocks;					 
   for (int ii = 0; ii < 100*EstimateNumJackknifeBlocks; ii++) {
      NumJackknifeBlocks = EstimateNumJackknifeBlocks + ii;
      if ( qtTotalStatSamples/(1.*NumJackknifeBlocks)
          == (int) (qtTotalStatSamples/(1.*NumJackknifeBlocks)) ) {
          break;
      }   	    
   }   
   
   JackknifeBlockSize =  qtTotalStatSamples/NumJackknifeBlocks;
   cout << "   NumJackknifeBlocks = " << NumJackknifeBlocks << endl;
   cout << "   Actual JackknifeBlockSize = " << JackknifeBlockSize << endl;   
   
   // ## Initialize the random generators
   RanNumGen oRanPMD("RanPMD.in");
   RanNumGen oRanBitString("RanBitString.in");       
   
   // ## Initiating the optical pulses
   OptSignal oOptSignal("OptSignal.in", &oRanBitString);   
   OptFiber oOptFiber("OptFiber.in", &oOptSignal, &oRanPMD);    
                        
   cout << "ExpectedDGD [ps] = " 
          << oOptFiber.GetExpectedDGD()*1e12 << endl;             
   double LikelihoodRatio;       
   double OutPD_Vector[3]; 
   double SndOrderOutPD_Vector[3];
   double NumSndOrderOutPD_Vector[3];
   double ThirdOrderOutPD_Vector[3];
   double NumThirdOrderOutPD_Vector[3];   
   double nDGD, nSndOrderDGD, nThirdOrderDGD;          
   double nDGD_Before = 0;  
   double nSndOrderDGD_Before = 0;
   double nThirdOrderDGD_Before = 0;

   int qtBins = oHistDGD.GetQtBins();
   if (  oHistDGD.GetQtBins() != oHistActualDGD.GetQtBins() 
       || oHistDGD.GetMaxRange() != oHistActualDGD.GetMaxRange()
       || oHistDGD.GetMinRange() != oHistActualDGD.GetMinRange() ) {
      cout << "Error: HistDGD.in and HistActualDGD.in must have " << endl
           << "the same range." << endl
	   << "Aborting now!!!!!" << endl;
      exit(1);	         
   }                 
   if (FiberRotatorModelFlag) {
      //## The random mode coupling will be produced by a single
      //## rotation Rz, which is the simplest one.     
      oOptFiber.GetFiberSampleUnifRz();
   } else {      
      oOptFiber.GetFiberSampleCompleteScatt();
   }    
   int qtStepsFiber = oOptFiber.GetQtStepsFiber();     
   double *MMC_pdf = new double[qtBins];    
   double *MMC_pdfTmp = new double[qtBins];          
   double *GlobalStatSignifMMC = new double[qtBins];      

   double *MMC_pdfStep2 = new double[qtBins]; 
   double *MMC_cdfStep2 = new double[qtBins];

   double *MMC_cdfStep2_BootResamp = new double[qtBins];

   
   double *NumSamplesBinTotal = new double[qtBins];
   
   double *NormHistogram = new double[qtBins];
   
   double *VarianceBinMMC = new double[qtBins];   
   double *VarianceBinThisIterMMC = VarianceBinMMC;   
   double *VarianceMeanBinMMC_Jackknife = new double[qtBins];

   double *MeanStdDevBinMMC = new double[qtBins];
   double *VarianceStdDevBinMMC = new double[qtBins];   

   double *SumBinMMC = new double[qtBins];
   double *SumSqBinMMC = new double[qtBins];

   double *SumStdDevBinMMC = new double[qtBins];
   double *SumSqStdDevBinMMC = new double[qtBins];

   double *SumMeanBinMMC = new double[qtBins];
   double *SumSqMeanBinMMC = new double[qtBins];

   double *MeanBinMMC_JackknifeSamples    = new double[NumJackknifeBlocks*qtBins];
   double *StdDevBinMMC_JackknifeSamples = new double[NumJackknifeBlocks*qtBins];

   double *TransProbabilityMatrix = new double [qtBins*qtBins];
   double *NumTransitionsFromBin = new double [qtBins];
   double *CumulativeTransProbabilityMatrix = new double [qtBins*qtBins];   
   double *VarTransProbabilityMatrix = new double [qtBins*qtBins];
   
   int *TransitionsRegistered = new int[2*qtTotalStatSamples]; //## Origin Bin -> Destination Bin
   int counterTransitionsRegistered = 0;

   double *TransProbabilityMatrixBootResamp = new double [qtBins*qtBins];
   double *NumTransitionsFromBinBootResamp = new double [qtBins];
   double *SumTransProbabilityMatrixBoot = new double [qtBins*qtBins];   
   double *SumSqTransProbabilityMatrixBoot = new double [qtBins*qtBins];   
   double *VarTransProbabilityMatrixBootEnsemble = new double [qtBins*qtBins];
   double *CumulativeTransProbabilityMatrixBootResamp = new double [qtBins*qtBins];      
        
   double *ThetaZ = new double[qtStepsFiber];    
   double *DeltaThetaZ = new double[qtStepsFiber];     
   double *PhiZ = new double[qtStepsFiber];
   double *DeltaPhiZ = new double[qtStepsFiber];     
   double *PsiZ = new double[qtStepsFiber];
   double *DeltaPsiZ = new double[qtStepsFiber];   
   for (int ii = 0; ii < qtStepsFiber; ii++) {    
      //## Initialize the angles
      double CosThetaZ_Tmp = 2*(oRanPMD.GetRanNum() - 0.5); //## -1 to 1
      ThetaZ[ii] = acos(CosThetaZ_Tmp);
      PhiZ[ii] = pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi      
      PsiZ[ii] = pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi      
   }
   oHistActualDGD.Obtain_pdf(); //## Fill in the vector pdfX   
   oHistDGD.Obtain_pdf();   
   for (int ii = 0; ii < qtBins; ii++) {   
      //## Initialize the iterative pdf 
      MMC_pdf[ii] = 1./qtBins;
      GlobalStatSignifMMC[ii] = 0;              
      NumTransitionsFromBin[ii] = 0;
      NumSamplesBinTotal[ii] = 0;    
   }   

   for (int ii = 0; ii < qtBins*qtBins; ii++) {   
      TransProbabilityMatrix[ii] = 0;
      CumulativeTransProbabilityMatrix[ii] = 0;      
   }

   //## Step I
   //## Computation of the MMC weights
   //## Major loop: MMC iterations
   //int SignThetaFromCos = +1;  
   for (int kk = 0; kk < qtIterationsMMC; kk++) {      
      oHistActualDGD.ClearHistogram();            
      double qtAcceptedSamples = 0;         
      for (int ii = 0; ii < qtStatSamples; ii++) {        
         for (int jj = 0; jj < qtStepsFiber; jj++){ 	 
            if (kk == 0 ) {
               //## Set a random angle for the first iteration        
               DeltaThetaZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi  	 	 
               DeltaPhiZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi  
               DeltaPsiZ[jj] = 1.0
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi                       
            } else {
               //## Set the ii-th fiber with the perturbation
               DeltaThetaZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //##        
               DeltaPhiZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //## 
               DeltaPsiZ[jj] = RelStepSizeMMC
                     *pi*2*(oRanPMD.GetRanNum() - 0.5); //##                       
            }	    	 		    
	    
            if (FiberRotatorModelFlag) {
               oOptFiber.SetScatteringMatrix( jj,   PhiZ[jj] 
	                                    +DeltaPhiZ[jj],
                                           pi/2., -pi/2.);
            } else {            
	       //## Assure the correct statistics in the angle (cosine fo theta is uniform)
               if ( oRanPMD.GetRanNum() >
                    sqrt(1 - sq(cos(ThetaZ[jj] + DeltaThetaZ[jj])))/sqrt(1 - sq(cos(ThetaZ[jj])) ) ) {    
   	         //## Do not accept
   	         DeltaThetaZ[jj] = 0;
	       }  
               oOptFiber.SetScatteringMatrix( jj,  ThetaZ[jj] + DeltaThetaZ[jj], 
	                                    PhiZ[jj] + DeltaPhiZ[jj],
                                            PsiZ[jj] + DeltaPsiZ[jj]);
            }                                                                       
         }                              
         nDGD = oOptFiber.GetDGD(oOptSignal.GetCenterFreq())
                       /oOptFiber.GetExpectedDGD();      
         nSndOrderDGD = oOptFiber.GetSndOrderDGD(oOptSignal.GetCenterFreq())
                      /oOptFiber.GetExpectedSndOrderDGD(); 
         nThirdOrderDGD = oOptFiber.GetThirdOrderDGD(oOptSignal.GetCenterFreq())
                      /oOptFiber.GetExpectedThirdOrderDGD();                                       
         if (kk == 0 && ii == 0)  {
            nDGD_Before = nDGD;     
	    nSndOrderDGD_Before = nSndOrderDGD;
	    nThirdOrderDGD_Before = nThirdOrderDGD;
         }            
	 if (   nDGD > oHistActualDGD.GetMinRange()     
	     && nDGD < oHistActualDGD.GetMaxRange() ) {               
   	    //## I intend to compute the true transition probability
	    //## Therefore, this must come before the decision on the acceptance
            TransProbabilityMatrix[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)
                  +qtBins*oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;	     
            NumTransitionsFromBin[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
            TransitionsRegistered[counterTransitionsRegistered] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD);
            TransitionsRegistered[counterTransitionsRegistered + qtTotalStatSamples] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before);
            counterTransitionsRegistered++;
            if ( oRanPMD.GetRanNum() <
               MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)] 
              /MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)] ) {      
               //## Accept the transition  
               qtAcceptedSamples++;	
               nDGD_Before = nDGD; 
               for (int jj = 0; jj < qtStepsFiber; jj++){ 	    
		  ThetaZ[jj] += DeltaThetaZ[jj];
                  PhiZ[jj] += DeltaPhiZ[jj];
                  PsiZ[jj] += DeltaPsiZ[jj];                  
               } 
            } 	 
         }	 
         oHistActualDGD.AddImportantSample(nDGD_Before,1);                                    
         NumSamplesBinTotal[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++; 	 	 
         oHistSndOrderDGD.AddImportantSample(nSndOrderDGD_Before,1);   
         oHistThirdOrderDGD.AddImportantSample(nThirdOrderDGD_Before,1);	 
      } //## End of qtSamples
      cout << "Iteration "<< kk << " ->  AcceptanceRate = " 
           << (1.0*qtAcceptedSamples)/qtStatSamples << endl; 
      
      //## Here comes Berg's modification
      
      //## Make a copy of the previous MMC_pdf
      for (int ii = 0; ii < qtBins; ii++) {          
         MMC_pdfTmp[ii] = MMC_pdf[ii];	 
         //## Attention: the normalization mus be made by qtStatSamplesStep2
         //#  That is necessary to have a coherent statistical significance
	 //#  in steps 1 and 2
         NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
      }  
          
      //## Start the computation of MMC_pdf
      for (int ii = 0; ii < qtBins - 1; ii++) {          
         if ( NormHistogram[ii+1] 
	     *NormHistogram[ii] > 0 ) {
            double StatSignifMMC =  NormHistogram[ii+1] 
	                        *NormHistogram[ii]
		              /( NormHistogram[ii+1] 
	                        +NormHistogram[ii] );
            GlobalStatSignifMMC[ii] += StatSignifMMC;
	    double RelStatSignifMMC = StatSignifMMC
                                     /GlobalStatSignifMMC[ii];
	    //RelStatSignifMMC = 1; //#$#$#$#$#$#$#$#   
            MMC_pdf[ii+1] = MMC_pdf[ii] 
	                   *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			   *pow( NormHistogram[ii+1] 
       	                        /NormHistogram[ii],
		                 RelStatSignifMMC);
         } else { //## This iteration had no stat signific in this bin
            MMC_pdf[ii+1] = MMC_pdf[ii] 
	                   *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	 }	    
      }                             
      
      //## Finally, the normalization      
      double NormMMC_pdf = 0;
      for (int ii = 0; ii < qtBins; ii++) {          
         NormMMC_pdf += MMC_pdf[ii];	 
      }                      
      for (int ii = 0; ii < qtBins; ii++) {   
         MMC_pdf[ii]  /= NormMMC_pdf;      	 
      }   
       
      //##
      //## Now, plot the histogram after the kk iteration
      //##      
      oHistDGD.ClearHistogram();
      for (int ii = 0; ii < qtBins; ii++) {   
         oHistDGD.AddImportantSample( oHistActualDGD.Get_pdfX(ii),
               MMC_pdf[ii]*qtBins); //## qtBins is here to workk around the normalization
      }                      
      //## Creates output files with the data
      oHistActualDGD.WriteFile_pdf("nActualDGDpdf.dat");   
      oHistDGD.WriteFile_pdf("nDGDpdfInvWeights.dat");                          
   } //## End of qtIterationsMMC

   //######################
   //## Step II: Computation of the probability using the MMC weights
   //## Now, use the MMC weights computed in Step I to make one single 
   //## simulation with the Metropolis Algorithm using the weights
   //##

   //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
   oHistActualDGD.ClearHistogram();        
   oHistDGD.ClearHistogram();
  
   double qtAcceptedSamples = 0;   
   for (int ii = 0; ii < qtStatSamplesStep2; ii++) {        
      for (int jj = 0; jj < qtStepsFiber; jj++){ 
            DeltaThetaZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //## -pi to pi              
            DeltaPhiZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //## 
            DeltaPsiZ[jj] = RelStepSizeMMC
                  *pi*2*(oRanPMD.GetRanNum() - 0.5); //##                    

            if (FiberRotatorModelFlag) {
               oOptFiber.SetScatteringMatrix( jj,   PhiZ[jj] 
	                                    +DeltaPhiZ[jj],
                                           pi/2., -pi/2.);
            } else {            
	       //## Assure the correct statistics in the angle (cosine fo theta is uniform)
               if ( oRanPMD.GetRanNum() >
                    sqrt(1 - sq(cos(ThetaZ[jj] + DeltaThetaZ[jj])))/sqrt(1 - sq(cos(ThetaZ[jj])) ) ) {    
   	         //## Do not accept
   	         DeltaThetaZ[jj] = 0;
	       }              
               oOptFiber.SetScatteringMatrix( jj,  ThetaZ[jj] + DeltaThetaZ[jj], 
	                                    PhiZ[jj] + DeltaPhiZ[jj],
                                            PsiZ[jj] + DeltaPsiZ[jj]);
            }            
      }                              
      nDGD = oOptFiber.GetDGD(oOptSignal.GetCenterFreq())
                    /oOptFiber.GetExpectedDGD();      
      nSndOrderDGD = oOptFiber.GetSndOrderDGD(oOptSignal.GetCenterFreq())
                    /oOptFiber.GetExpectedSndOrderDGD(); 
      nThirdOrderDGD = oOptFiber.GetThirdOrderDGD(oOptSignal.GetCenterFreq())
                   /oOptFiber.GetExpectedThirdOrderDGD();          
      if (   nDGD > oHistActualDGD.GetMinRange()     
          && nDGD < oHistActualDGD.GetMaxRange() ) {          
	 //## I intend to compute the true transition probability
	 //## Therefore, this must come before the decision on the acceptance
         TransProbabilityMatrix[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)
               +qtBins*oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
         NumTransitionsFromBin[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;
         TransitionsRegistered[counterTransitionsRegistered] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD);
         TransitionsRegistered[counterTransitionsRegistered + qtTotalStatSamples] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before);
         counterTransitionsRegistered++;
         if ( oRanPMD.GetRanNum() <
            MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)] 
           /MMC_pdf[oHistActualDGD.GetAdjustedIndex_pdf(nDGD)] ) {         
            //## Accept the transition  
            qtAcceptedSamples++;	
            nDGD_Before = nDGD; 
	    nSndOrderDGD_Before = nSndOrderDGD;
	    nThirdOrderDGD_Before = nThirdOrderDGD;	    
            for (int jj = 0; jj < qtStepsFiber; jj++){     
	       ThetaZ[jj] += DeltaThetaZ[jj];               
               PhiZ[jj] += DeltaPhiZ[jj];
               PsiZ[jj] += DeltaPsiZ[jj];   
            } 
         } 	 
      }	 
      oHistActualDGD.AddImportantSample(nDGD_Before,1);                                    
      NumSamplesBinTotal[oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before)]++;            
      oHistSndOrderDGD.AddImportantSample(nSndOrderDGD_Before,1);   
      oHistThirdOrderDGD.AddImportantSample(nThirdOrderDGD_Before,1);	 
   } //## End of qtStatSamplesStep2   
   cout << "Step II  ->  AcceptanceRate = " 
        << (1.0*qtAcceptedSamples)/qtStatSamplesStep2 << endl; 

   //## Step II-B: Computation of the histogram of the DGD       
   //## First, load the histogram and normalize it 
   //## MMC_pdfTmp[] now contains the weights,
   //## and MMC_pdf contains the last histogram   
        
   //## Make a copy of the previous MMC_pdf
   for (int ii = 0; ii < qtBins; ii++) {          
      MMC_pdfTmp[ii] = MMC_pdf[ii];	 
      NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
   }  
   //## Start the computation of MMC_pdf
   for (int ii = 0; ii < qtBins - 1; ii++) {          
      if ( NormHistogram[ii+1] 
          *NormHistogram[ii] > 0 ) {
         double StatSignifMMC =  NormHistogram[ii+1] 
                                *NormHistogram[ii]
	                  /( NormHistogram[ii+1] 
	                        +NormHistogram[ii] );			
         GlobalStatSignifMMC[ii] += StatSignifMMC;
         double RelStatSignifMMC = StatSignifMMC
                                  /GlobalStatSignifMMC[ii];
         MMC_pdf[ii+1] = MMC_pdf[ii] 
	                *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                *pow( NormHistogram[ii+1] 
       	                     /NormHistogram[ii],
	                     RelStatSignifMMC);
      } else { //## This iteration had no stat signific in this bin
         MMC_pdf[ii+1] = MMC_pdf[ii] 
                 *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
      }	    
   }              
   //## Finally, the normalization      
   double NormMMC_pdfStep2 = 0;   
   for (int ii = 0; ii < qtBins; ii++) {          
      NormMMC_pdfStep2 += MMC_pdf[ii];	    
   }                      
   double Cumulative_pdf_Tmp = 0;
   for (int ii = 0; ii < qtBins; ii++) {   
      MMC_pdf[ii]  /= NormMMC_pdfStep2;   
      MMC_pdfStep2[ii] = MMC_pdf[ii];
      Cumulative_pdf_Tmp += MMC_pdfStep2[ii];
      MMC_cdfStep2[ii] = Cumulative_pdf_Tmp;
      //cout << " *$*$*$*$* " << ii << " -> MMC_pdf = " << MMC_pdf[ii]
      //     << "  MMC_cdf = " << MMC_cdfStep2[ii] << endl; 
   }         

   //##############################################################
   //########### Compute an estimate for the confidence interval         
   if (ErrorPropagationFlag && ! ErrorOfErrorFlag) {
      cout << "Computing the confidence interval" << endl;

//      //## Create a file with the transition matrix
//      ofstream OutFile;
//      string OutFileName = "TransitionMatrix.dat";   
//      //## Plot the pdf of the DGD with its standard deviation
//      OutFile.open(OutFileName.c_str());
//      if (!OutFile) {
//         cout << "Error: " << OutFileName << " could not be created."
//              << endl;
//         exit(1);
//      }
      //## Now, normalize each row of the transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
         if ( !DoNotSpoolErrorFlag) {      
            //cout <<kk<<" -> NumTransitionsFromBin = "<<NumTransitionsFromBin[kk]<<endl;
         }   
	 double CumulativeProb = 0;
         for (int ii = 0; ii < qtBins; ii++) {   
	    if (NumTransitionsFromBin[kk] > 1) {
               VarTransProbabilityMatrix[ii + kk*qtBins] = 
                  ( TransProbabilityMatrix[ii + kk*qtBins] 
                   -sq(TransProbabilityMatrix[ii + kk*qtBins])/NumTransitionsFromBin[kk] )
                  /(NumTransitionsFromBin[kk]*(NumTransitionsFromBin[kk] -1) );
               TransProbabilityMatrix[ii + kk*qtBins] /= NumTransitionsFromBin[kk];	       
	    }   	    
//            OutFile << kk << " " << ii << " " << TransProbabilityMatrix[ii + kk*qtBins] << endl;
	    CumulativeProb += TransProbabilityMatrix[ii + kk*qtBins];
   	    CumulativeTransProbabilityMatrix[ii + kk*qtBins] = CumulativeProb;      
            if ( !DoNotSpoolErrorFlag) {
   	       //cout << kk << " " << ii << " -> " << TransProbabilityMatrix[ii + kk*qtBins]
	       //     << " " << CumulativeProb << " Relative Variation: "
               //     << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins])
               //       /TransProbabilityMatrix[ii + kk*qtBins] << endl;
            }                      
         }	
      }
//      OutFile.close();
      
      //## Now, generate a Monte Carlo ensamble, where the actual transitions
      //## will be done according to the final transition probability
      //## The confidence interval is computed using the Monte Carlo method 

      for (int ii = 0; ii < qtBins; ii++) { 
         SumBinMMC[ii] = 0;
         SumSqBinMMC[ii] = 0;   
      }


      for (int ii = 0; ii < qtBins*qtBins; ii++) {       
         SumTransProbabilityMatrixBoot[ii] = 0;   
         SumSqTransProbabilityMatrixBoot[ii] = 0;   
      }         

      //int NumSamplesForErrorEstim = 1000;  
      //cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;
      
      double *BinProbabilityBootResamp;
      if (ComputeAutocorrFlag) {        
         BinProbabilityBootResamp = new double [qtBins*NumSamplesForErrorEstim];
      }      

      //## Start loop for bootstrap resampling      
      for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {

         for (int ii = 0; ii < qtBins*qtBins; ii++) { 
            TransProbabilityMatrixBootResamp[ii] = 0;	     
         }         
         for (int ii = 0; ii < qtBins; ii++) {       
            NumTransitionsFromBinBootResamp[ii] = 0;  
         }

         for (int ii = 0; ii < qtBins; ii++) {   
            //## Initialize the iterative pdf 
            MMC_pdf[ii] = 1./qtBins;
            GlobalStatSignifMMC[ii] = 0;                    
         }         
         int BinBefore;
         int NewBin;        

         //## Get a randow inicial bin                            
         double RanNumTmp = oRanPMD.GetRanNum();                     
         for (int jj = 0; jj < qtBins; jj++) {                         
            if (   RanNumTmp <= MMC_cdfStep2[jj]) {
               BinBefore = jj;      
   	       break;
   	    }
         }          

         //## Step I
         //## Computation of the MMC weights
         //## Major loop: MMC iterations
         for (int vv = 0; vv < qtIterationsMMC; vv++) {      
            oHistActualDGD.ClearHistogram();
            for (int ii = 0; ii < qtStatSamples; ii++) {  
               //## Get a randow transition. This is the complicated part                             
               double RanNumTmp = oRanPMD.GetRanNum();          
               for (int jj = 0; jj < qtBins; jj++) {                         
	          if (vv == 0) {
                     if (   RanNumTmp <= MMC_cdfStep2[jj]) {
                        NewBin = jj;      
   	                break;
	             }	    
	          } else {
                     if (   RanNumTmp  <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                        NewBin = jj;      
   	                break;
	             }
	          }
               }         
               //## For the Bootstrap estimate of the transition probability matrix               
               TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               NumTransitionsFromBinBootResamp[BinBefore]++;                              
               if ( oRanPMD.GetRanNum() < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
                  //## Accept the transition  
                  BinBefore = NewBin; 
               } 	                                  
               oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
            } //## End of ii (up to qtStatSamples)
   
            //## Here comes Berg's modification      
            //## Make a copy of the previous MMC_pdf
            for (int ii = 0; ii < qtBins; ii++) {          
               MMC_pdfTmp[ii] = MMC_pdf[ii];	 
               //## Attention: the normalization mus be made by qtStatSamplesStep2
               //#  That is because the Step II is the one actually used in the 
               //#  Normalization
               NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
            }  
          
            //## Start the computation of MMC_pdf
            for (int ii = 0; ii < qtBins - 1; ii++) {          
               if ( NormHistogram[ii+1] 
	           *NormHistogram[ii] > 0 ) {
                  double StatSignifMMC =  NormHistogram[ii+1] 
	                              *NormHistogram[ii]
		                    /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );
                  GlobalStatSignifMMC[ii] += StatSignifMMC;
	          double RelStatSignifMMC = StatSignifMMC
                                           /GlobalStatSignifMMC[ii];
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			         *pow( NormHistogram[ii+1] 
       	                              /NormHistogram[ii],
		                       RelStatSignifMMC);
               } else { //## This iteration had no stat signific in this bin
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	       }	    
            }  
                           
            //## Finally, the normalization      
            double NormMMC_pdf = 0;
            for (int ii = 0; ii < qtBins; ii++) {          
               NormMMC_pdf += MMC_pdf[ii];	 
            }                      
            for (int ii = 0; ii < qtBins; ii++) {   
               MMC_pdf[ii]  /= NormMMC_pdf;      	 
            }         
         } //## End of vv (qtIterations)
       
         //#$#$#$#$#$ This is for the computation of the variance
         //######################
         //## Step II: Computation of the probability using the MMC weights
         //## Now, use the MMC weights computed in Step I to make one single 
         //## simulation with the Metropolis Algorithm using the weights
         //##

         //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
         oHistActualDGD.ClearHistogram();             
         for (int ii = 0; ii < qtStatSamplesStep2; ii++) {         
            //## Get a randow transition. This is the complicated part                             
            double RanNumTmp = oRanPMD.GetRanNum();              
            for (int jj = 0; jj < qtBins; jj++) {                         
               if (   RanNumTmp <= CumulativeTransProbabilityMatrix[jj + BinBefore*qtBins]) {
                  NewBin = jj;      
	          break;
	       }
            }            
            //## For the Bootstrap estimate of the transition probability matrix      
            TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
            NumTransitionsFromBinBootResamp[BinBefore]++;               
            if ( oRanPMD.GetRanNum()  < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
               //## Accept the transition  
               BinBefore = NewBin; 
            } 	                                  
            oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
         } //## End of qtStatSamplesStep2   

         //## Step II-B: Computation of the histogram of the DGD       
         //## First, load the histogram and normalize it 
         //## MMC_pdfTmp[] now contains the weights,
         //## and MMC_pdf contains the last histogram   
        
         //## Make a copy of the previous MMC_pdf
         for (int ii = 0; ii < qtBins; ii++) {          
            MMC_pdfTmp[ii] = MMC_pdf[ii];	 
            NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
         }  
         //## Start the computation of MMC_pdf
         for (int ii = 0; ii < qtBins - 1; ii++) {          
            if ( NormHistogram[ii+1] 
                *NormHistogram[ii] > 0 ) {
               double StatSignifMMC =  NormHistogram[ii+1] 
                                      *NormHistogram[ii]
	                        /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );			
               GlobalStatSignifMMC[ii] += StatSignifMMC;
               double RelStatSignifMMC = StatSignifMMC
                                        /GlobalStatSignifMMC[ii];
               MMC_pdf[ii+1] = MMC_pdf[ii] 
	                      *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                      *pow( NormHistogram[ii+1] 
       	                           /NormHistogram[ii],
	                           RelStatSignifMMC);
            } else { //## This iteration had no stat signific in this bin
               MMC_pdf[ii+1] = MMC_pdf[ii] 
                       *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
            }	    
         }       
         double NormMMC_pdf = 0;       
         //## Finally, the normalization      
         for (int ii = 0; ii < qtBins; ii++) {          
            NormMMC_pdf += MMC_pdf[ii];	      
         }                      
         for (int ii = 0; ii < qtBins; ii++) {   
            MMC_pdf[ii]  /= NormMMC_pdf;          
            SumBinMMC[ii] += MMC_pdf[ii];
            SumSqBinMMC[ii] += sq(MMC_pdf[ii]);                        
         }   

         if (ComputeAutocorrFlag) {       
            for (int ii = 0; ii < qtBins; ii++) {             
               BinProbabilityBootResamp[ii + kk*qtBins] =  MMC_pdf[ii];
            }   
         }           

         //## Now, normalize each row of the Bootstrap resampled transition probability matrix
         for (int kk = 0; kk < qtBins; kk++) {        
            //cout <<kk<<" -> NumTransitionsFromBinBootResamp = "<<NumTransitionsFromBinBootResamp[kk]<<endl;
	    double CumulativeProb = 0;
            for (int ii = 0; ii < qtBins; ii++) {   
	       if (NumTransitionsFromBinBootResamp[kk] != 0) {
                  TransProbabilityMatrixBootResamp[ii + kk*qtBins] 
                        /= NumTransitionsFromBinBootResamp[kk];	       
	       }   	    
            }	
         }

         for (int ii = 0; ii < qtBins*qtBins; ii++) {        
            SumTransProbabilityMatrixBoot[ii] += TransProbabilityMatrixBootResamp[ii];
            SumSqTransProbabilityMatrixBoot[ii] += sq(TransProbabilityMatrixBootResamp[ii]);       
         }            
         
                             
      } //## End of kk      

      for (int nn = 0; nn < qtBins; nn++) { 
         //MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
         MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
	 VarianceBinThisIterMMC[nn] = NumSamplesForErrorEstim
	       /(NumSamplesForErrorEstim - 1.)
	       *( SumSqBinMMC[nn]/NumSamplesForErrorEstim
	         -sq(SumBinMMC[nn]/NumSamplesForErrorEstim) ); 
         if ( !DoNotSpoolErrorFlag) {		 	    
            cout << nn << " -> MeanBinMMC = " << MMC_pdf[nn] << "  VarianceBinThisIterMMC = " 
                 << VarianceBinThisIterMMC[nn] << endl;	    	    	 
            cout << nn << " ####  -> RelativeVariationMMC = " 
                 <<  sqrt(VarianceBinThisIterMMC[nn])/MMC_pdf[nn] << endl;   
         }       
      } //## End of bin loop: nn	

      {
      //## Create a file with the transition matrix
      ofstream OutFile;
      string OutFileName = "TransitionMatrix.dat";   
      //## Plot the pdf of the DGD with its standard deviation
      OutFile.open(OutFileName.c_str());
      if (!OutFile) {
         cout << "Error: " << OutFileName << " could not be created."
              << endl;
         exit(1);
      }

      //## Compute the variance of the bootstrap resampled transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
         for (int ii = 0; ii < qtBins; ii++) {   
	    if (SumTransProbabilityMatrixBoot[ii + kk*qtBins] != 0) {
               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] =             
                  NumSamplesForErrorEstim/(NumSamplesForErrorEstim - 1.)
	       *( SumSqTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim
	         -sq(SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim) ); 
//               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] = 
//                  ( SumSqTransProbabilityMatrixBoot[ii + kk*qtBins] 
//                   -sq(SumTransProbabilityMatrixBoot[ii + kk*qtBins])/NumSamplesForErrorEstim )
//                  /(NumSamplesForErrorEstim*(NumSamplesForErrorEstim -1) );	       
	    }  else {
               VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins] = 0;
            }
            if ( !DoNotSpoolErrorFlag) {
	       cout << " ** " << kk << " " << ii << " -> " 
                    << TransProbabilityMatrix[ii + kk*qtBins]                 
	            << " Relative_Variation: " << sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins])
                      /(SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim) << endl;
            }      
            double pdfRelativeStdDevTmp = 99;
            if (SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim != 0) {
               pdfRelativeStdDevTmp = sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins]) 
                                     / (SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim);
            }               
            //## Layout: kk ii Estimated_From_1_MMC MeanBootStrap StdDevBootStrap RelStdDev
            OutFile << kk << " " << ii << " " 
                    << TransProbabilityMatrix[ii + kk*qtBins] << "   "
                    << SumTransProbabilityMatrixBoot[ii + kk*qtBins]/NumSamplesForErrorEstim << "  "
                    << sqrt(VarTransProbabilityMatrixBootEnsemble[ii + kk*qtBins]) << "  "
                    << pdfRelativeStdDevTmp << endl;
         }	
      }
      OutFile.close();
      }
           

      if (ComputeAutocorrFlag) {
         double *AutocorrelationBins = new double[qtBins];
         for (int ii = 0; ii < qtBins; ii++) {             
            AutocorrelationBins[ii] = 0;
         }   
         for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {            
            for (int ii = 0; ii < qtBins; ii++) {             
               AutocorrelationBins[ii] += 
                  (BinProbabilityBootResamp[BinAutocorrelation + kk*qtBins]
                   -SumBinMMC[BinAutocorrelation]/NumSamplesForErrorEstim)
                 *(BinProbabilityBootResamp[ii + kk*qtBins]
                   -SumBinMMC[ii]/NumSamplesForErrorEstim);
            }   
         }            
         //## Normalization of autocorrelation         
         //double NormAutocorrCoeff = AutocorrelationBins[BinAutocorrelation];          
         for (int ii = 0; ii < qtBins; ii++) {           
            AutocorrelationBins[ii] /= (NumSamplesForErrorEstim - 1)
                 *sqrt(VarianceBinThisIterMMC[BinAutocorrelation])
                 *sqrt(VarianceBinThisIterMMC[ii]);
         }                        
         
         string OutFileName = "AutocorrMMC_Prob.dat";
         ofstream OutFile(OutFileName.c_str(),ios::out);
         if (!OutFile) {
            cout << "Error: " << OutFileName << " could not be created."
                 << endl;
            exit(1);
         }
         for (int ii = 0; ii < qtBins; ii++) {   
            OutFile << oHistActualDGD.Get_pdfX(ii) << "  " 
                    << AutocorrelationBins[ii] << endl;
         }         
         OutFile.close();
         
         delete [] BinProbabilityBootResamp;
         delete [] AutocorrelationBins;
      }  
      
      
   } //## End of if ErrorPropagationFlag AND NOT ErrorOfErrorFlag
   


   //##############################################################
   //########### Compute an estimate for the confidence interval  
   //## and for the confidence interval of the error       
   if (ErrorPropagationFlag && ErrorOfErrorFlag) {
      cout << "Computing the confidence interval of probability and error" << endl;   


     //## Now, normalize each row of the transition probability matrix
      for (int kk = 0; kk < qtBins; kk++) {        
//         if ( !DoNotSpoolErrorFlag) {      
//            //cout <<kk<<" -> NumTransitionsFromBin = "<<NumTransitionsFromBin[kk]<<endl;
//         }   
	 double CumulativeProb = 0;
         for (int ii = 0; ii < qtBins; ii++) {    
	    if (NumTransitionsFromBin[kk] > 0) {
  	       if (NumTransitionsFromBin[kk] > 1) {            
                  VarTransProbabilityMatrix[ii + kk*qtBins] = 
                     ( TransProbabilityMatrix[ii + kk*qtBins] 
                      -sq(TransProbabilityMatrix[ii + kk*qtBins])/NumTransitionsFromBin[kk] )
                     /(NumTransitionsFromBin[kk]*(NumTransitionsFromBin[kk] -1) );
               }   
               TransProbabilityMatrix[ii + kk*qtBins] /= NumTransitionsFromBin[kk];	       
	    }   	    
//            OutFile << kk << " " << ii << " " << TransProbabilityMatrix[ii + kk*qtBins] << endl;
	    CumulativeProb += TransProbabilityMatrix[ii + kk*qtBins];
   	    CumulativeTransProbabilityMatrix[ii + kk*qtBins] = CumulativeProb;      
            if ( !DoNotSpoolErrorFlag) {
   	       //cout << kk << " " << ii << " -> " << TransProbabilityMatrix[ii + kk*qtBins]
	       //     << " " << CumulativeProb << " Relative Variation: "
               //     << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins])
               //       /TransProbabilityMatrix[ii + kk*qtBins] << endl;
            }                      
         }	
      }      
      //## Now, the estimated transition matrix and its cdf are available for use
      //## I should not generate one pseudo-transition matrix and compute 
      //## NumSamplesForErrorEstim MMC ensembles for that transition matrix
      //## to obtain an error estimate. I should repeat this process NumBootstrapTransMatrices      
      //## times to obtain a mean and a standard deviation for the error estimate.

      
//      OutFile.close();
      
      //## Now, generate a Monte Carlo ensamble, where the actual transitions
      //## will be done according to the final transition probability
      //## The confidence interval is computed using the Monte Carlo method 

      for (int ii = 0; ii < qtBins; ii++) {          
         SumStdDevBinMMC[ii] = 0;
         SumSqStdDevBinMMC[ii] = 0;     
             
         SumMeanBinMMC[ii] = 0;
         SumSqMeanBinMMC[ii] = 0;     
      }


//      for (int ii = 0; ii < qtBins*qtBins; ii++) {       
//         SumTransProbabilityMatrixBoot[ii] = 0;   
//         SumSqTransProbabilityMatrixBoot[ii] = 0;   
//      }         

      //int NumSamplesForErrorEstim = 1000;  
      //cout << "NumSamplesForErrorEstim = " << NumSamplesForErrorEstim << endl;
      
//      double *BinProbabilityBootResamp;
//      if (ComputeAutocorrFlag) {        
//         BinProbabilityBootResamp = new double [qtBins*NumSamplesForErrorEstim];
//      }      

      //## Start loop for bootstrap resampling for each bootstraped transition matrix     
//      for (int kk = 0; kk < NumSamplesForErrorEstim; kk++) {
//      for (int kk = 0; kk < NumBootstrapTransMatrices; kk++) {
      for (int kk = 0; kk < NumJackknifeBlocks; kk++) {
         cout << "Computing matrix # " << kk << " out of " << NumJackknifeBlocks <<endl;

         for (int ii = 0; ii < qtBins*qtBins; ii++) { 
            TransProbabilityMatrixBootResamp[ii] = 0;	     
         }         
         for (int ii = 0; ii < qtBins; ii++) {       
            NumTransitionsFromBinBootResamp[ii] = 0;  
         }


         int counter = 0;
         for (int ii = 0; ii < qtTotalStatSamples; ii++) {
            //## Now, exclude non-contiguous samples
//            if ( (ii+kk)/JackknifeBlockSize !=  (ii+kk)/(1.*JackknifeBlockSize) )  {
            if ( (ii+kk)/NumJackknifeBlocks !=  (ii+kk)/(1.*NumJackknifeBlocks) )  {
               int NewBin = TransitionsRegistered[ii];
               int BinBefore = TransitionsRegistered[ii + qtTotalStatSamples];            
               TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               NumTransitionsFromBinBootResamp[BinBefore]++;    
               counter++;
            }   
            //else {
            //   cout << "   ### Sample excluded -> " << ii << endl;
            //}
         }   
         //cout << "#$#$# Number of samples included = " << counter;

//         TransitionsRegistered[counterTransitionsRegistered] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD);
//         TransitionsRegistered[counterTransitionsRegistered + qtTotalStatSamples] = oHistActualDGD.GetAdjustedIndex_pdf(nDGD_Before);



         //## Now, normalize each row of the Bootstrap resampled transition probability matrix
         for (int jj = 0; jj < qtBins; jj++) {        
            //cout <<jj<<" -> NumTransitionsFromBinBootResamp = "<<NumTransitionsFromBinBootResamp[jj]<<endl;
	    double CumulativeProb = 0;
            for (int ii = 0; ii < qtBins; ii++) {   
	       if (NumTransitionsFromBinBootResamp[jj] != 0) {
                  TransProbabilityMatrixBootResamp[ii + jj*qtBins] 
                        /= NumTransitionsFromBinBootResamp[jj];	       
	       }   	    
   	       CumulativeProb += TransProbabilityMatrixBootResamp[ii + jj*qtBins];
   	       CumulativeTransProbabilityMatrixBootResamp[ii + jj*qtBins] = CumulativeProb;                   
            }	
         }

         //for (int ii = 0; ii < qtBins*qtBins; ii++) {        
         //   SumTransProbabilityMatrixBoot[ii] += TransProbabilityMatrixBootResamp[ii];
         //   SumSqTransProbabilityMatrixBoot[ii] += sq(TransProbabilityMatrixBootResamp[ii]);       
         //}          

      for (int ii = 0; ii < qtBins; ii++) { 
         SumBinMMC[ii] = 0;
         SumSqBinMMC[ii] = 0;   
      }

      for (int tt = 0; tt < NumSamplesForErrorEstim; tt++) {         
         //## Now, I should compute the bootstrap variance using the kk-th bootstraped 
         //## transition matrix                                                                 
  
         for (int ii = 0; ii < qtBins; ii++) {   
            //## Initialize the iterative pdf 
            MMC_pdf[ii] = 1./qtBins;
            GlobalStatSignifMMC[ii] = 0;                    
         }         
         int BinBefore;
         int NewBin;        

         //## Get a randow inicial bin                            
         double RanNumTmp = oRanPMD.GetRanNum();                     
         for (int jj = 0; jj < qtBins; jj++) {                         
            if (   RanNumTmp <= MMC_cdfStep2_BootResamp[jj]) {
               BinBefore = jj;      
   	       break;
   	    }
         }          

         //## Step I
         //## Computation of the MMC weights
         //## Major loop: MMC iterations
         for (int vv = 0; vv < qtIterationsMMC; vv++) {      
            oHistActualDGD.ClearHistogram();
            for (int ii = 0; ii < qtStatSamples; ii++) {  
               //## Get a randow transition. This is the complicated part                             
               double RanNumTmp = oRanPMD.GetRanNum();          
               for (int jj = 0; jj < qtBins; jj++) {                         
	          if (vv == 0) {
                     if (   RanNumTmp <= MMC_cdfStep2_BootResamp[jj]) {
                        NewBin = jj;      
   	                break;
	             }	    
	          } else {
                     if (   RanNumTmp  <= CumulativeTransProbabilityMatrixBootResamp[jj + BinBefore*qtBins]) {
                        NewBin = jj;      
   	                break;
	             }
	          }
               }         
               //## For the Bootstrap estimate of the transition probability matrix               
               ///TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
               ///NumTransitionsFromBinBootResamp[BinBefore]++;                              
               if ( oRanPMD.GetRanNum() < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
                  //## Accept the transition  
                  BinBefore = NewBin; 
               } 	                                  
               oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
            } //## End of ii (up to qtStatSamples)
   
            //## Here comes Berg's modification      
            //## Make a copy of the previous MMC_pdf
            for (int ii = 0; ii < qtBins; ii++) {          
               MMC_pdfTmp[ii] = MMC_pdf[ii];	 
               //## Attention: the normalization mus be made by qtStatSamplesStep2
               //#  That is because the Step II is the one actually used in the 
               //#  Normalization
               NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;
            }  
          
            //## Start the computation of MMC_pdf
            for (int ii = 0; ii < qtBins - 1; ii++) {          
               if ( NormHistogram[ii+1] 
	           *NormHistogram[ii] > 0 ) {
                  double StatSignifMMC =  NormHistogram[ii+1] 
	                              *NormHistogram[ii]
		                    /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );
                  GlobalStatSignifMMC[ii] += StatSignifMMC;
	          double RelStatSignifMMC = StatSignifMMC
                                           /GlobalStatSignifMMC[ii];
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
			         *pow( NormHistogram[ii+1] 
       	                              /NormHistogram[ii],
		                       RelStatSignifMMC);
               } else { //## This iteration had no stat signific in this bin
                  MMC_pdf[ii+1] = MMC_pdf[ii] 
	                         *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii];	 
	       }	    
            }  
                           
            //## Finally, the normalization      
            double NormMMC_pdf = 0;
            for (int ii = 0; ii < qtBins; ii++) {          
               NormMMC_pdf += MMC_pdf[ii];	 
            }                      
            for (int ii = 0; ii < qtBins; ii++) {   
               MMC_pdf[ii]  /= NormMMC_pdf;      	 
            }         
         } //## End of vv (qtIterations)
       
         //#$#$#$#$#$ This is for the computation of the variance
         //######################
         //## Step II: Computation of the probability using the MMC weights
         //## Now, use the MMC weights computed in Step I to make one single 
         //## simulation with the Metropolis Algorithm using the weights
         //##

         //### Step II-A: Computation of the histogram with qtStatSamples*qtIterationsmMMC samples
         oHistActualDGD.ClearHistogram();             
         for (int ii = 0; ii < qtStatSamplesStep2; ii++) {         
            //## Get a randow transition. This is the complicated part                             
            double RanNumTmp = oRanPMD.GetRanNum();              
            for (int jj = 0; jj < qtBins; jj++) {                         
               if (   RanNumTmp <= CumulativeTransProbabilityMatrixBootResamp[jj + BinBefore*qtBins]) {
                  NewBin = jj;      
	          break;
	       }
            }            
            //## For the Bootstrap estimate of the transition probability matrix      
            //TransProbabilityMatrixBootResamp[NewBin + qtBins*BinBefore]++;	     
            //NumTransitionsFromBinBootResamp[BinBefore]++;               
            if ( oRanPMD.GetRanNum()  < MMC_pdf[BinBefore]/MMC_pdf[NewBin] ) {         
               //## Accept the transition  
               BinBefore = NewBin; 
            } 	                                  
            oHistActualDGD.SetBin(BinBefore, oHistActualDGD.GetBin(BinBefore) + 1 );      
         } //## End of qtStatSamplesStep2   

         //## Step II-B: Computation of the histogram of the DGD       
         //## First, load the histogram and normalize it 
         //## MMC_pdfTmp[] now contains the weights,
         //## and MMC_pdf contains the last histogram   
        
         //## Make a copy of the previous MMC_pdf
         for (int ii = 0; ii < qtBins; ii++) {          
            MMC_pdfTmp[ii] = MMC_pdf[ii];	 
            NormHistogram[ii] = oHistActualDGD.GetBin(ii)/qtStatSamplesStep2;      
         }  
         //## Start the computation of MMC_pdf
         for (int ii = 0; ii < qtBins - 1; ii++) {          
            if ( NormHistogram[ii+1] 
                *NormHistogram[ii] > 0 ) {
               double StatSignifMMC =  NormHistogram[ii+1] 
                                      *NormHistogram[ii]
	                        /( NormHistogram[ii+1] 
	                              +NormHistogram[ii] );			
               GlobalStatSignifMMC[ii] += StatSignifMMC;
               double RelStatSignifMMC = StatSignifMMC
                                        /GlobalStatSignifMMC[ii];
               MMC_pdf[ii+1] = MMC_pdf[ii] 
	                      *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]
	                      *pow( NormHistogram[ii+1] 
       	                           /NormHistogram[ii],
	                           RelStatSignifMMC);
            } else { //## This iteration had no stat signific in this bin
               MMC_pdf[ii+1] = MMC_pdf[ii] 
                       *MMC_pdfTmp[ii+1]/MMC_pdfTmp[ii]; 	 
            }	    
         }       
         double NormMMC_pdf = 0;       
         //## Finally, the normalization      
         for (int ii = 0; ii < qtBins; ii++) {          
            NormMMC_pdf += MMC_pdf[ii];	      
         }                      
         for (int ii = 0; ii < qtBins; ii++) {   
            MMC_pdf[ii]  /= NormMMC_pdf;          
            SumBinMMC[ii] += MMC_pdf[ii];
            SumSqBinMMC[ii] += sq(MMC_pdf[ii]);                        
         }            
         //## End of the computation of a value of MMC_pdf

      } //## End tt  

         for (int nn = 0; nn < qtBins; nn++) { 
            //MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
            //MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
   	    VarianceBinMMC[nn] = NumSamplesForErrorEstim
   	       /(NumSamplesForErrorEstim - 1.)
	       *( SumSqBinMMC[nn]/NumSamplesForErrorEstim
	         -sq(SumBinMMC[nn]/NumSamplesForErrorEstim) ); 
            //cout << kk << " ##### RelVar = " << sqrt(VarianceBinMMC[nn])
            //      /(SumBinMMC[nn]/NumSamplesForErrorEstim) << endl;
            //## Standard deviations     
            SumStdDevBinMMC[nn] += sqrt(VarianceBinMMC[nn]);
            SumSqStdDevBinMMC[nn] += VarianceBinMMC[nn];

            SumMeanBinMMC[nn] += SumBinMMC[nn]/NumSamplesForErrorEstim;
            SumSqMeanBinMMC[nn] += sq(SumBinMMC[nn]/NumSamplesForErrorEstim);

            MeanBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
            StdDevBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn] = sqrt(VarianceBinMMC[nn]);   


//            if ( !DoNotSpoolErrorFlag) {		 	    
//               cout << nn << " -> MeanBinMMC = " << MMC_pdf[nn] << "  VarianceBinThisIterMMC = " 
//                    << VarianceBinThisIterMMC[nn] << endl;	    	    	 
//               cout << nn << " ####  -> RelativeVariationMMC = " 
//                    <<  sqrt(VarianceBinThisIterMMC[nn])/MMC_pdf[nn] << endl;   
//            }       
         } //## End of bin loop: nn	          
      
      
      } //## End of kk   


      for (int nn = 0; nn < qtBins; nn++) { 
            //## Do not use this line ##//  MMC_pdf[nn] = SumBinMMC[nn]/NumSamplesForErrorEstim;
            MMC_pdf[nn] = MMC_pdfStep2[nn];  //## Restore value from MMC simulation
            MeanStdDevBinMMC[nn] = SumStdDevBinMMC[nn]/NumJackknifeBlocks;
            VarianceBinMMC[nn] = sq(MeanStdDevBinMMC[nn]);            

            double BinomialNormCoefficient = bico(qtTotalStatSamples,JackknifeBlockSize);
            //cout << "BinomialNormCoefficient = " << BinomialNormCoefficient 
            //     << "  <- n = " << qtTotalStatSamples << " k = " << JackknifeBlockSize << endl;
            double Mean_MeanBinMMC_JackknifeSamples = 0;
            double Mean_StdDevBinMMC_JackknifeSamples = 0;
            for (int kk = 0; kk < NumJackknifeBlocks; kk++) {
               Mean_MeanBinMMC_JackknifeSamples += MeanBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn];
               Mean_StdDevBinMMC_JackknifeSamples += StdDevBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn];               
            }            
            //Mean_MeanBinMMC_JackknifeSamples /= BinomialNormCoefficient;
            //Mean_StdDevBinMMC_JackknifeSamples /= BinomialNormCoefficient;             
            Mean_MeanBinMMC_JackknifeSamples /= NumJackknifeBlocks;
            Mean_StdDevBinMMC_JackknifeSamples /= NumJackknifeBlocks;
            VarianceMeanBinMMC_Jackknife[nn] = 0;            
            VarianceStdDevBinMMC[nn] = 0;
            for (int kk = 0; kk < NumJackknifeBlocks; kk++) {
               VarianceMeanBinMMC_Jackknife[nn] += 
                     sq( MeanBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn]
                        -Mean_MeanBinMMC_JackknifeSamples );
               VarianceStdDevBinMMC[nn] += 
                     sq( StdDevBinMMC_JackknifeSamples[kk + NumJackknifeBlocks*nn]
                        -Mean_StdDevBinMMC_JackknifeSamples );                        
            }            
            //VarianceMeanBinMMC_Jackknife[nn] /= (NumJackknifeBlocks/BinomialNormCoefficient);
            //VarianceStdDevBinMMC[nn] /= (NumJackknifeBlocks/BinomialNormCoefficient);            
            VarianceMeanBinMMC_Jackknife[nn] *= (NumJackknifeBlocks - 1.)/NumJackknifeBlocks;
            VarianceStdDevBinMMC[nn] *= (NumJackknifeBlocks - 1.)/NumJackknifeBlocks;
            
      } //## End of bin loop: nn	                      
      
   } //## End of if ErrorPropagationFlag AND ErrorOfErrorFlag
   //#####################################################################3

   //##
   //## Now, plot the histogram after the kk iteration
   //##      
   string OutFileName = "nDGDpdf.dat";
   //## Plot the pdf of the DGD with its standard deviation
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }

   oHistDGD.ClearHistogram();      
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfX = oHistDGD.Get_pdfX(ii);
      double NormConstant = (oHistDGD.GetMaxRange() - oHistDGD.GetMinRange() )
	                   /qtBins;
		   
      double pdfYtmp = MMC_pdf[ii]/NormConstant;
      double pdfConfidence = sqrt(VarianceBinMMC[ii])/NormConstant;
      
      double pdfConfidenceOfConfidence = sqrt(VarianceStdDevBinMMC[ii])/NormConstant;

      double pdfConfidenceJackknife = sqrt(VarianceMeanBinMMC_Jackknife[ii])/NormConstant;     
            
      double pdfRelativeStdDevTmp = 99;
      if (pdfYtmp != 0) {
         pdfRelativeStdDevTmp = pdfConfidence / pdfYtmp;
      }   
      double pdfRelativeStdDevStdDevTmp = 99;
      if (pdfConfidence != 0) {
         pdfRelativeStdDevStdDevTmp = pdfConfidenceOfConfidence / pdfConfidence;
      }   
      double pdfRelativeStdDevJackknifeTmp = 99;
      if (pdfYtmp != 0) {
         pdfRelativeStdDevJackknifeTmp = pdfConfidenceJackknife / pdfYtmp;
      }         
      
      if (! ErrorOfErrorFlag) {
         OutFile << pdfX << "   " 
                 << pdfYtmp << "   "  
                 << pdfConfidence << "   "
   	         << pdfRelativeStdDevTmp << endl;   
      } else{
         OutFile << pdfX << "   " 
                 << pdfYtmp << "   "  
                 << pdfConfidence << "   "
   	         << pdfRelativeStdDevTmp << "  "
                 << pdfConfidenceOfConfidence << "   "
   	         << pdfRelativeStdDevStdDevTmp << "  "
                 << pdfConfidenceJackknife << "   "
   	         << pdfRelativeStdDevJackknifeTmp << endl;         
      }                 
   }           
   OutFile.close();     

   //## Creates a file with the Maxwellian pdf with the same mean
   int qtPointsMaxwell = 1000;     
   oOptFiber.WriteFileMaxwell_NormDGD_pdf("Maxwell_pdf.dat",oHistDGD.GetMinRange(),
         oHistDGD.GetMaxRange(),qtPointsMaxwell);
   oOptFiber.WriteFileTrueCSM_NormDGD_pdf("TrueCSM_DGD_pdf.dat",oHistDGD.GetMinRange(),
         oHistDGD.GetMaxRange(),qtPointsMaxwell);  

   //## Plot the pdf of the histogram with the confidence interval
   OutFileName = "nActualDGDpdf.dat";   
   //## Plot the pdf of the DGD with its standard deviation
   OutFile.open(OutFileName.c_str());
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfX = oHistActualDGD.Get_pdfX(ii);
      double NormConstant = ( oHistActualDGD.GetMaxRange() 
                             -oHistActualDGD.GetMinRange() )
	                    /qtBins;
      double pdfYtmp = NumSamplesBinTotal[ii]
                      /(1.*qtIterationsMMC*qtStatSamples + qtStatSamplesStep2)
                      /NormConstant;    
      OutFile << pdfX << "   " 
              << pdfYtmp << "   "  
              << endl;   
   }           
   OutFile.close();
   
   oHistActualDGD.WriteFile_pdf("nActualDGDpdf2.dat");   
   oHistSndOrderDGD.WriteFile_pdf("nSndOrderDGDpdf.dat");   
   oHistThirdOrderDGD.WriteFile_pdf("nThirdOrderDGDpdf.dat");  
            
   double nDGD_Mean = oHistDGD.GetMean();
   cout << "nActualDGDmean = " << oHistActualDGD.GetMean() << endl;
   cout << "qtStatSamples = " << oHistActualDGD.GetQtSamples() << endl;
   cout << " Percentage inside the range nDGD = "<<oHistActualDGD.GetMinRange()
        << " to " << oHistActualDGD.GetMaxRange() << " = " 
        << oHistActualDGD.GetQtSamplesInRange()/ 
           (double) oHistActualDGD.GetQtSamples()*100. << "%." << endl;
   cout << "nSndOrderDGDmean = " << oHistSndOrderDGD.GetMean() << endl;
   cout << "nThirdOrderDGDmean = " << oHistThirdOrderDGD.GetMean() << endl;
   
   delete [] MMC_pdf;
   delete [] MMC_pdfTmp;
   delete [] GlobalStatSignifMMC;
   delete [] MMC_pdfStep2; 
   delete [] MMC_cdfStep2;

   delete [] MMC_cdfStep2_BootResamp;

   delete [] NormHistogram;   
   delete [] VarianceBinMMC;
   
   delete [] VarianceMeanBinMMC_Jackknife;
   
   delete [] SumBinMMC;
   delete [] SumSqBinMMC;

   delete [] SumStdDevBinMMC;
   delete [] SumSqStdDevBinMMC;

   delete [] SumMeanBinMMC;
   delete [] SumSqMeanBinMMC;
   
   delete [] MeanStdDevBinMMC;
   delete [] VarianceStdDevBinMMC;   
   
   delete [] MeanBinMMC_JackknifeSamples;
   delete [] StdDevBinMMC_JackknifeSamples;   

   delete [] TransProbabilityMatrix;
   delete [] NumTransitionsFromBin;
   delete [] CumulativeTransProbabilityMatrix;
   delete [] VarTransProbabilityMatrix;   

   delete [] TransitionsRegistered;

   delete [] TransProbabilityMatrixBootResamp;
   delete [] NumTransitionsFromBinBootResamp;
   delete [] SumTransProbabilityMatrixBoot;   
   delete [] SumSqTransProbabilityMatrixBoot;   
   delete [] VarTransProbabilityMatrixBootEnsemble; 
   delete [] CumulativeTransProbabilityMatrixBootResamp;
   
   delete [] NumSamplesBinTotal;   
   delete [] ThetaZ;   
   delete [] DeltaThetaZ;
   delete [] PhiZ;
   delete [] DeltaPhiZ;   
   delete [] PsiZ;
   delete [] DeltaPsiZ;  

}  //## End of: void GetDGDpdfMMC_Berg(MyStartup *oMyStartup) 




//#############################################################
//##### Creates a file with the pdf of the DGD with its
//##    respective variance from multiple simulations using
//##    the STANDARD NUMERICAL METHOD. Therefore, 
//##    IT MUST BE USED FOR VALIDATION PURPOSES, ONLY.
//##    Attention, the seed must be 
//##    set to the continued mode.
//##    The pdf in each simulation is computed using 
//##    the Multicanonical Monte Carlo as proposed by Berg.
//## By: Aurenice Lima
void GetStdDevDGDpdfMMC_Berg(MyStartup *oMyStartup) 
{
   int qtSimulations = oMyStartup->GetIntParam("qtSimulations");
   cout << "qtSimulations = " << qtSimulations << endl; 

   int ComputeAutocorrFlag = oMyStartup->GetIntParam("ComputeAutocorrFlag");
   cout << "ComputeAutocorrFlag = " << ComputeAutocorrFlag << endl;    

   Histogram *oHistDGDtmp = new Histogram("HistDGD.in");   
   int qtBins = oHistDGDtmp->GetQtBins();
   delete oHistDGDtmp;

   int BinAutocorrelation;   
   if (ComputeAutocorrFlag) {
      BinAutocorrelation = oMyStartup->GetIntParam("BinAutocorrelation");
      cout << "BinAutocorrelation = " << BinAutocorrelation << endl;          
   }

   double *BinProbability;
   if (ComputeAutocorrFlag) {        
      BinProbability = new double [qtBins*qtSimulations];
   }         
   
   double *pdfX = new double[qtBins];
   double *SumTmp = new double[qtBins]; 
   double *SumSquareTmp = new double[qtBins];     
//   double *SumBootRelVarTmp = new double[qtBins]; 
//   double *SumSquareBootRelVarTmp = new double[qtBins];   
   double *SumBootStdDevTmp = new double[qtBins]; 
   double *SumSquareBootStdDevTmp = new double[qtBins]; 
   double *VarianceBinMMC = new double[qtBins];  

   //double *ActualHistogram_pdf = new double[qtBins];
   double *SumActualHistogramTmp = new double[qtBins];
   double *SumSqActualHistogramTmp = new double[qtBins];

   double *SumTransProbabilityMatrix = new double [qtBins*qtBins];   
   double *SumSqTransProbabilityMatrix = new double [qtBins*qtBins];   
   double *VarTransProbabilityMatrix = new double [qtBins*qtBins];     
//   double *SumTransProbabilityMatrixBootRelVar = new double [qtBins*qtBins];   
//   double *SumSqTransProbabilityMatrixBootRelVar = new double [qtBins*qtBins]; 
   double *SumTransProbabilityMatrixBootStdDev = new double [qtBins*qtBins];   
   double *SumSqTransProbabilityMatrixBootStdDev = new double [qtBins*qtBins]; 


   for (int ii = 0; ii < qtBins; ii++) {
      SumTmp[ii] = 0;
      SumSquareTmp[ii] = 0;
//      SumBootRelVarTmp[ii] = 0;
//      SumSquareBootRelVarTmp[ii] = 0;
      SumBootStdDevTmp[ii] = 0;
      SumSquareBootStdDevTmp[ii] = 0;
      VarianceBinMMC[ii] = 0;
      
      SumActualHistogramTmp[ii] = 0;
      SumSqActualHistogramTmp[ii] = 0;      
   }   

   for (int ii = 0; ii < qtBins*qtBins; ii++) {       
      SumTransProbabilityMatrix[ii] = 0;   
      SumSqTransProbabilityMatrix[ii] = 0;   
//      SumTransProbabilityMatrixBootRelVar[ii];   
//      SumSqTransProbabilityMatrixBootRelVar[ii];   
      SumTransProbabilityMatrixBootStdDev[ii] = 0;   
      SumSqTransProbabilityMatrixBootStdDev[ii] = 0;         
   }  
      

   string InFileName = "nDGDpdf.dat";
   string OutFileName = "nDGDpdfStdDev.dat";
//   string OutFileNameBootRelativeVar = "nDGDpdfBootRelVar.dat";

   string InFileNameActualRatio = "nActualDGDpdf.dat";  
   string OutFileNameActualRatio = "nActualDGDpdfStdDev.dat";      

   string InFileNameTransitionMatrix = "TransitionMatrix.dat";
   string OutFileNameTransitionMatrix = "TransitionMatrixStdDev.dat";
   
   //ifstream InFile(InFileName.c_str(),ios::in);
      
   for (int ii = 0; ii < qtSimulations; ii++) {
      //## Carry out one simulation
      GetDGDpdfMMC_Berg(oMyStartup);
     
      //## Read the pdf data
      {
      ifstream InFile(InFileName.c_str(),ios::in);
      //InFile.open(InFileName.c_str(),ios::in);         
      if (!InFile) {
         cout << "Error: " << InFileName << " could not be open."
              << endl;
         exit(1);
      }      
      double pdfYtmp;
      double varTmp;
      double pdfY_StdDevTmp;
      int counter = 0;
      while ( InFile >> varTmp ) {   
         pdfX[counter] = varTmp;
         InFile >> pdfYtmp;
	 InFile >> pdfY_StdDevTmp;    
	 InFile >> varTmp;
	 SumTmp[counter] += pdfYtmp;
	 SumSquareTmp[counter] += sq(pdfYtmp);	 
         
         SumBootStdDevTmp[counter] += pdfY_StdDevTmp; 
         SumSquareBootStdDevTmp[counter] += sq(pdfY_StdDevTmp); 
         
         if (ComputeAutocorrFlag) {                 
            BinProbability[counter + ii*qtBins] =  pdfYtmp;
         }                  
	 counter++;         
	 //cout << "counter = "<< counter << endl;    
      }      
      InFile.close();
      }
      
      //## Read the H data
      {   
         ifstream InFile(InFileNameActualRatio.c_str(),ios::in);      
         if (!InFile) {
            cout << "Error: " << InFileNameActualRatio << " could not be open."
                 << endl;
            exit(1);
         }         
         double pdfYtmp, varTmp;
         int counter = 0;            	 	 
         while ( InFile >> varTmp ) {   
            //pdfX[counter] = varTmp;
            InFile >> pdfYtmp;  
   	    InFile >> varTmp;
   	    InFile >> varTmp;	 	 
	    SumActualHistogramTmp[counter] += pdfYtmp;
	    SumSqActualHistogramTmp[counter] += sq(pdfYtmp);		    
            counter++;	 
         }
         InFile.close();
      }  
      
      //## Read the Transition Probability Matrix
      {
      ifstream InFile(InFileNameTransitionMatrix.c_str(),ios::in);
      //InFile.open(InFileName.c_str(),ios::in);         
      if (!InFile) {
         cout << "Error: " << InFileNameTransitionMatrix << " could not be open."
              << endl;
         exit(1);
      }      
      double var_kk, var_ii;
      double varTransProbTmp;
      double varTransProbBootStdDevTmp;
      double varTmp;
      //int counter = 0;
      //double test1, test2, test3;
      //InFile >> test1; 
      //InFile >> test2; 
      //InFile >> test3;       
      //cout << test1 << " " << test2 << " " << test3 << endl; 
      while ( InFile >> var_kk ) {   
         InFile >> var_ii;
	 InFile >> varTransProbTmp;
         InFile >> varTmp;
	 InFile >> varTransProbBootStdDevTmp;         
         InFile >> varTmp;
	 SumTransProbabilityMatrix[(int) (var_ii + var_kk*qtBins)] += varTransProbTmp;
	 SumSqTransProbabilityMatrix[(int) (var_ii + var_kk*qtBins)] += sq(varTransProbTmp);	 
	 SumTransProbabilityMatrixBootStdDev[(int) (var_ii + var_kk*qtBins)] += varTransProbBootStdDevTmp;
	 SumSqTransProbabilityMatrixBootStdDev[(int) (var_ii + var_kk*qtBins)] += sq(varTransProbBootStdDevTmp);	 
         
         //cout << " ##### " << var_kk << " " << var_ii << "  -> " << varTmp << endl;
	 //counter++;
	 //cout << "counter = "<< counter << endl;
      }      
      InFile.close();      
      }
      
      cout << "*** End simulation # " << ii << endl;      
   }
   
     
   //## Plot the pdf of the DGD with its standard deviation
   {
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }
   //double *pdfConfidence = new double[qtBins];
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfYtmp = SumTmp[ii]/qtSimulations;
      double pdfConfidence = 0;
      if (SumSquareTmp[ii] > sq(SumTmp[ii])/qtSimulations )  {
         pdfConfidence = sqrt( ( SumSquareTmp[ii] 
                              -qtSimulations*sq(SumTmp[ii]/qtSimulations) )
                               /( qtSimulations-1) 
			       );
////////                               /qtSimulations );   
         VarianceBinMMC[ii] = ( SumSquareTmp[ii] 
                              -qtSimulations*sq(SumTmp[ii]/qtSimulations) )
                               /( qtSimulations-1); 
      } 
      double BootStdDevTmp = SumBootStdDevTmp[ii]/qtSimulations;
      double BootStdDevConfidence = 0;
      if (SumSquareBootStdDevTmp[ii] > sq(SumBootStdDevTmp[ii])/qtSimulations )  {
         BootStdDevConfidence = sqrt( ( SumSquareBootStdDevTmp[ii] 
                              -qtSimulations*sq(SumBootStdDevTmp[ii]/qtSimulations) )
                               /( qtSimulations-1) 
			       );
      }      
         
      double pdfRelativeStdDevTmp = 99;
      if (pdfYtmp != 0) {
         pdfRelativeStdDevTmp = pdfConfidence / pdfYtmp;
      }  
      double RelativeBootStdDevTmp = 99;
      if (BootStdDevTmp != 0) {
         RelativeBootStdDevTmp = BootStdDevConfidence / BootStdDevTmp;
      }  
       
      OutFile << pdfX[ii] << "   " 
              << pdfYtmp << "   "  
              << pdfConfidence << "   "
	      << pdfRelativeStdDevTmp << "  "
              << SumBootStdDevTmp[ii]/qtSimulations << "  "
              << BootStdDevConfidence << "  "
              << RelativeBootStdDevTmp << endl;   
   }           
   OutFile.close();
   }



   //## Plot the H histogram with its standard deviation
   {
   ofstream OutFile(OutFileNameActualRatio.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileNameActualRatio << " could not be created."
           << endl;
      exit(1);
   }
   //double *pdfConfidence = new double[qtBins];
   for (int ii = 0; ii < qtBins; ii++) {
      double pdfYtmp = SumActualHistogramTmp[ii]/qtSimulations;
      double pdfConfidence = 0;
      if (SumSqActualHistogramTmp[ii] > sq(SumActualHistogramTmp[ii])/qtSimulations )  {
         pdfConfidence = sqrt( ( SumSqActualHistogramTmp[ii] 
                              -qtSimulations*sq(SumActualHistogramTmp[ii]/qtSimulations) )
                               /( qtSimulations-1) 
			       );
////////                               /qtSimulations );   
      } 
         
      double pdfRelativeStdDevTmp = 99;
      if (pdfYtmp != 0) {
         pdfRelativeStdDevTmp = pdfConfidence / pdfYtmp;
      }   
      OutFile << pdfX[ii] << "   " 
              << pdfYtmp << "   "  
              << pdfConfidence << "   "
	      << pdfRelativeStdDevTmp << endl;   
   }           
   OutFile.close();
   }


   {
      ofstream OutFile(OutFileNameTransitionMatrix.c_str());
      if (!OutFile) {
         cout << "Error: " << OutFileNameTransitionMatrix << " could not be created."
              << endl;
         exit(1);
      }

      cout << "Transition matrix of ensemble" << endl;
      //## Compute the variance of the transition probability matrix from the ensemble
      for (int kk = 0; kk < qtBins; kk++) {        
         for (int ii = 0; ii < qtBins; ii++) {   
	    if (SumTransProbabilityMatrix[ii + kk*qtBins] != 0) {
               VarTransProbabilityMatrix[ii + kk*qtBins] =             
                 qtSimulations/(qtSimulations - 1.)
	       *( SumSqTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations
	         -sq(SumTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations) );       
	    } else {
               VarTransProbabilityMatrix[ii + kk*qtBins] = 0;
            } 
            double pdfRelativeStdDevTmp = 99;
            if (SumTransProbabilityMatrix[ii + kk*qtBins] != 0) {
               pdfRelativeStdDevTmp = sqrt(VarTransProbabilityMatrix[ii + kk*qtBins]) 
                                     / (SumTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations);
            }  
            
            double VarTransProbabilityMatryxBootStdDev = 0;
	    if (SumTransProbabilityMatrixBootStdDev[ii + kk*qtBins] != 0) {
               VarTransProbabilityMatryxBootStdDev =             
                 qtSimulations/(qtSimulations - 1.)
	       *( SumSqTransProbabilityMatrixBootStdDev[ii + kk*qtBins]/qtSimulations
	         -sq(SumTransProbabilityMatrixBootStdDev[ii + kk*qtBins]/qtSimulations) );       
            } 
            double RelativeBootStdDevTmp = 99;
            if (SumTransProbabilityMatrixBootStdDev[ii + kk*qtBins] != 0) {
               RelativeBootStdDevTmp = sqrt(VarTransProbabilityMatryxBootStdDev) 
                                     / (SumTransProbabilityMatrixBootStdDev[ii + kk*qtBins]/qtSimulations);
            }  

            //## Layout: kk ii MeanMultiple_MMC StdDevMultiple_MMC RelativeVarMultiple_MMC
            //#          MeanMultiple_BootStrapStdDev StdDevMuitipleBootStrapStdDev RelativeVarMultipleBootStrapStdDev
                                     
            OutFile << kk << " " << ii << " " 
                    << SumTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations << "   "
                    << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins]) << "  "                    
                    << pdfRelativeStdDevTmp << "  "                     
                    << SumTransProbabilityMatrixBootStdDev[ii + kk*qtBins]/qtSimulations << "  "
                    << sqrt(VarTransProbabilityMatryxBootStdDev) << "  "
                    << RelativeBootStdDevTmp << endl;
         }	                 
	    //cout << " ***** " << kk << " " << ii << " -> " 
            //     << SumTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations                 
	    //     << " Relative_Variation: " << sqrt(VarTransProbabilityMatrix[ii + kk*qtBins])
            //       /(SumTransProbabilityMatrix[ii + kk*qtBins]/qtSimulations) << endl;
      }    
      OutFile.close();      
   }


   if (ComputeAutocorrFlag) {
      double *AutocorrelationBins = new double[qtBins*qtBins];
      for (int jj = 0; jj < qtBins; jj++) {       
         for (int ii = 0; ii < qtBins; ii++) {                     
            AutocorrelationBins[ii + jj*qtBins] = 0;
         }   
      }   
      for (int kk = 0; kk < qtSimulations; kk++) {            
         for (int jj = 0; jj < qtBins; jj++) {        
            for (int ii = 0; ii < qtBins; ii++) {             
               AutocorrelationBins[ii + jj*qtBins] += 
                  (BinProbability[jj + kk*qtBins]
                   -SumTmp[jj]/qtSimulations)
                 *(BinProbability[ii + kk*qtBins]
                   -SumTmp[ii]/qtSimulations);
            }                   
         }   
      }            
      //## Normalization of autocorrelation         
      //double NormAutocorrCoeff = AutocorrelationBins[BinAutocorrelation];          
      //for (int ii = 0; ii < qtBins; ii++) {           
      //   AutocorrelationBins[ii] /= NormAutocorrCoeff
      //}                        
      for (int jj = 0; jj < qtBins; jj++) {  
         for (int ii = 0; ii < qtBins; ii++) {           
            AutocorrelationBins[ii + jj*qtBins] /= (qtSimulations - 1.)
                 *sqrt(VarianceBinMMC[jj])
                 *sqrt(VarianceBinMMC[ii]);
         }              
      }   
      
      string OutFileName = "AutocorrTrueEnsembleMMC_ProbMatr.dat";
      ofstream OutFile(OutFileName.c_str(),ios::out);
      if (!OutFile) {
         cout << "Error: " << OutFileName << " could not be created."
              << endl;
         exit(1);
      }
      //Histogram oHistDGDtmp("HistDGD.in");
      Histogram *oHistDGDtmp = new Histogram("HistDGD.in"); 
      oHistDGDtmp->Obtain_pdf();         
      for (int jj = 0; jj < qtBins; jj++) {  
         for (int ii = 0; ii < qtBins; ii++) {   
            OutFile << oHistDGDtmp->Get_pdfX(jj) << "  " 
                    << oHistDGDtmp->Get_pdfX(ii) << "  " 
                    << AutocorrelationBins[ii + jj*qtBins] << endl;
         }   
         OutFile << endl; //## Separation of block required by GnuPlot                 
      }         
      OutFile.close();
      delete oHistDGDtmp;  
      
      delete [] BinProbability;
      delete [] AutocorrelationBins;
   }  
   
   delete [] pdfX;
   delete [] SumTmp;
   delete [] SumSquareTmp;   
   delete [] VarianceBinMMC; 
//   delete [] SumBootRelVarTmp;
//   delete [] SumSquareBootRelVarTmp; 
   delete [] SumBootStdDevTmp;
   delete [] SumSquareBootStdDevTmp;
   
   delete [] SumTransProbabilityMatrix;   
   delete [] SumSqTransProbabilityMatrix;      
   delete [] VarTransProbabilityMatrix;  
//   delete [] SumTransProbabilityMatrixBootRelVar;   
//   delete [] SumSqTransProbabilityMatrixBootRelVar;     
   delete [] SumTransProbabilityMatrixBootStdDev;
   delete [] SumSqTransProbabilityMatrixBootStdDev;
   
} 





//#############################################################
//##### Compute the the autocorrelation function and its error
//##    using independent statistical ensembles
//##    Each simulation is computed using 
//##    the Multicanonical Monte Carlo as proposed by Berg.
//## By: Aurenice Lima
void GetAutocorrStdDevDGDpdfMMC_Berg(MyStartup *oMyStartup) 
{
   int qtSubEnsembles = oMyStartup->GetIntParam("qtSubEnsembles");
   cout << "qtSubEnsembles = " << qtSubEnsembles << endl; 

   if (qtSubEnsembles < 1 ) { 
      cout << "You must set qtSubEnsenbles > 0 to obtain the Autocorrelation" << endl;
      exit(1);
   }

   int ComputeAutocorrFlag = oMyStartup->GetIntParam("ComputeAutocorrFlag");
   cout << "ComputeAutocorrFlag = " << ComputeAutocorrFlag << endl;    

   if (!ComputeAutocorrFlag) { 
      cout << "You must set ComputeAutocorrFlag = 1 to obtain the Autocorrelation" << endl;
      exit(1);
   }

   Histogram *oHistDGDtmp = new Histogram("HistDGD.in");   
   int qtBins = oHistDGDtmp->GetQtBins();
   double MinRange = oHistDGDtmp->GetMinRange();
   double MaxRange = oHistDGDtmp->GetMaxRange();
   delete oHistDGDtmp;


   string InFileName = "AutocorrTrueEnsembleMMC_ProbMatr.dat";
   string OutFileName = "AutocorrTrueEnsembleMMC_ProbMatrStdDev.dat";


   double *SumAutocorrelationBins = new double[qtBins*qtBins];
   double *SumSqAutocorrelationBins = new double[qtBins*qtBins];   

   for (int ii = 0; ii < qtBins*qtBins; ii++) {       
      SumAutocorrelationBins[ii] = 0;   
      SumSqAutocorrelationBins[ii] = 0;   
    
   }  

   for (int ii = 0; ii < qtSubEnsembles; ii++) {
      GetStdDevDGDpdfMMC_Berg(oMyStartup);      

      //## Read the Autocorrelation Matrix
      {
      ifstream InFile(InFileName.c_str(),ios::in);
      //InFile.open(InFileName.c_str(),ios::in);         
      if (!InFile) {
         cout << "Error: " << InFileName << " could not be open."
              << endl;
         exit(1);
      }      
      double var_kk, var_ii;
      double varAutocorrelationTmp;
      int varInt_kk, varInt_ii;
      //int counter = 0;
      //double test1, test2, test3;
      //InFile >> test1; 
      //InFile >> test2; 
      //InFile >> test3;       
      //cout << test1 << " " << test2 << " " << test3 << endl; 
      while ( InFile >> var_kk ) {   
         InFile >> var_ii;
	 InFile >> varAutocorrelationTmp;         
         varInt_kk = int ( (var_kk - MinRange)*qtBins/(MaxRange - MinRange) -0.5);
         varInt_ii = int ( (var_ii - MinRange)*qtBins/(MaxRange - MinRange) -0.5);         
         //pdfX[ii] = (ii + 0.5)*(MaxRange - MinRange)/qtBins + MinRange;         
	 SumAutocorrelationBins[varInt_ii + varInt_kk*qtBins] += varAutocorrelationTmp;
	 SumSqAutocorrelationBins[varInt_ii + varInt_kk*qtBins] += sq(varAutocorrelationTmp);	 
        
         //cout << " ##### " << var_kk << " " << var_ii << "  -> " << varTmp << endl;
	 //counter++;
	 //cout << "counter = "<< counter << endl;
      }      
      InFile.close();      
      }      
      cout << "***** End sub-simulation # " << ii << endl;              
   }


   {
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }
   //Histogram oHistDGDtmp("HistDGD.in");
   Histogram *oHistDGDtmp = new Histogram("HistDGD.in"); 
   oHistDGDtmp->Obtain_pdf();     
   double VarAutocorrelationBins;    
   for (int kk = 0; kk < qtBins; kk++) {  
      for (int ii = 0; ii < qtBins; ii++) {         
         if (SumAutocorrelationBins[ii + kk*qtBins] != 0) {
            VarAutocorrelationBins =             
               qtSubEnsembles/(qtSubEnsembles - 1.)
	       *( SumSqAutocorrelationBins[ii + kk*qtBins]/qtSubEnsembles
	         -sq(SumAutocorrelationBins[ii + kk*qtBins]/qtSubEnsembles) )
               /qtSubEnsembles;       
	 } else {
            VarAutocorrelationBins = 0;
         }       
      
         OutFile << oHistDGDtmp->Get_pdfX(kk) << "  " 
                 << oHistDGDtmp->Get_pdfX(ii) << "  " 
                 << SumAutocorrelationBins[ii + kk*qtBins]/qtSubEnsembles << "  "
                 << VarAutocorrelationBins << endl;
      }   
      OutFile << endl; //## Separation of block required by GnuPlot                 
   }         
   OutFile.close();
   delete oHistDGDtmp;  
   }
   
   delete [] SumAutocorrelationBins;   
   delete [] SumSqAutocorrelationBins;      
}





//#############################################################
//##### Creates a file with the Expected value for 
//##    the normalized H. This will give an insight on the 
//##    bias.
//##    THIS METHOD IS OUT OF DATE
//## By: Aurenice Lima
void GetTrueHistogramStep2MMC(MyStartup *oMyStartup) 
{

   int qtIterationsMMC = oMyStartup->GetIntParam("qtIterationsMMC");
   cout << "qtIterationsMMC = " << qtIterationsMMC << endl; 

   int qtStatSamples = oMyStartup->GetQtStatSamples();
   cout << "qtStatSamples = " << qtStatSamples << endl; 

   int qtStatSamplesStep2 = oMyStartup->GetIntParam("qtStatSamplesStep2");      
   if (qtStatSamplesStep2 == 0) {
      qtStatSamplesStep2 = qtStatSamples*qtIterationsMMC;
   }
   cout << "qtStatSamplesStep2 = " << qtStatSamplesStep2 << endl;

//   int qtSimulations = oMyStartup->GetIntParam("qtSimulations");
//   cout << "qtSimulations = " << qtSimulations << endl; 

   // ## Initialize the random generators
   RanNumGen oRanPMD("RanPMD.in");
   RanNumGen oRanBitString("RanBitString.in");       
   
   // ## Initiating the optical pulses
   OptSignal oOptSignal("OptSignal.in", &oRanBitString);   
   OptFiber oOptFiber("OptFiber.in", &oOptSignal, &oRanPMD);   
   
   Histogram *oHistDGD = new Histogram("HistDGD.in");   
   int qtBins = oHistDGD->GetQtBins();
/*
   double *pdfX = new double[qtBins];
   double *SumTmp = new double[qtBins]; 
   double *SumSquareTmp = new double[qtBins];     
   for (int ii = 0; ii < qtBins; ii++) {
      SumTmp[ii] = 0;
      SumSquareTmp[ii] = 0;
   }   
   delete oHistDGDtmp;
*/      

   string InFileName = "nDGDpdfInvWeights.dat";
   string InFileNameActualRatio = "nActualDGDpdf.dat";   
   string OutFileName = "TrueHistogramStep2.dat";
   string OutFileNameRatio = "RatioHistogramStep2.dat";   


   double *pdfX = new double[qtBins];
   double *pdfY = new double[qtBins];
   double *pdfY_True = new double[qtBins];
   double *Histogram_pdf = new double[qtBins];
   double *ActualHistogram_pdf = new double[qtBins];


   {   
      ifstream InFile(InFileNameActualRatio.c_str(),ios::in);      
      if (!InFile) {
         cout << "Error: " << InFileName << " could not be open."
              << endl;
         exit(1);
      }         

      double MeanTmp = 1.; //## The normalized mean DGD of the fiber
      int qtStepsFiber = oOptFiber.GetQtStepsFiber();
      
      double varTmp;
      
      int qtStepsFiberOrig = qtStepsFiber;

      int counter = 0;            
      while ( InFile >> varTmp ) {   
         //pdfX[counter] = varTmp;
         InFile >> ActualHistogram_pdf[counter];  
	 InFile >> varTmp;
	 InFile >> varTmp;	 	 
         counter++;	 
      }
   }   


   {   
      ifstream InFile(InFileName.c_str(),ios::in);      
      if (!InFile) {
         cout << "Error: " << InFileName << " could not be open."
              << endl;
         exit(1);
      }         

      double MeanTmp = 1.; //## The normalized mean DGD of the fiber
      int qtStepsFiber = oOptFiber.GetQtStepsFiber();
      
      double varTmp;
      
      int qtStepsFiberOrig = qtStepsFiber;

      int counter = 0;            
      while ( InFile >> varTmp ) {   
         pdfX[counter] = varTmp;
         InFile >> pdfY[counter];  
	 InFile >> varTmp;
	 InFile >> varTmp;
	 //## Computing true pdf. Code from ocsOptFiber.cc
         double vSample = pdfX[counter];

         if (vSample < 1.5 && qtStepsFiberOrig > 50) {
	    qtStepsFiber = 15;
	 } else {
	    qtStepsFiber = qtStepsFiberOrig;
	 }	 
	 
         double DGDsec = MeanTmp/sqrt((8/(3*pi))*qtStepsFiber);
         double mParam = 0.5*( 1 - vSample/(qtStepsFiber*DGDsec));
         //cout << vSample << " -> " << mParam << endl;
         double True_pdf = 0;      
         if (mParam >= 0) {
            for (int jj = 0; jj <= floor(qtStepsFiber*mParam); jj++) {      
               True_pdf += pow(-1.,jj)*bico(qtStepsFiber,jj)
      	            *pow(qtStepsFiber*mParam - jj,qtStepsFiber-2);
            }
            True_pdf *= vSample/(2*sq(DGDsec)*factrl(qtStepsFiber-2));
         } 
          pdfY_True[counter] = True_pdf;	 
	 
         counter++;
      }
   }   

   for (int ii = 0; ii < qtBins; ii++) {
      Histogram_pdf[ii] = 0;
   }   

   Histogram_pdf[0] = 1;   
   for (int ii = 0; ii < qtBins-1; ii++) {
      if ( pdfY_True[ii] * pdfY[ii] > 0) {
         Histogram_pdf[ii+1] = Histogram_pdf[ii]
               *(pdfY_True[ii+1]/pdfY_True[ii])
	       /(pdfY[ii+1]/pdfY[ii]);
      }       
      else {
         cout << pdfX[ii] << "  " << pdfY[ii] << "  " << pdfY_True[ii] << endl;
         Histogram_pdf[ii+1] = Histogram_pdf[ii];
      }
   }

   double Norm_pdf = 0;
   for (int ii = 0; ii < qtBins; ii++) {
      Norm_pdf += Histogram_pdf[ii];
   }      

   for (int ii = 0; ii < qtBins; ii++) {
      Histogram_pdf[ii] /= Norm_pdf;
   }   

   {
   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileName << " could not be created."
           << endl;
      exit(1);
   }   
   
   for (int ii = 0; ii < qtBins; ii++) {
      OutFile << pdfX[ii] << " " 
              << Histogram_pdf[ii]
	      *qtBins/(oHistDGD->GetMaxRange() - oHistDGD->GetMinRange())
              << endl;
   }

//      OutFile << pdfX[ii] << " " 
//              << pdfY_True[ii]/pdfY[ii]
//	         /(oHistDGD->GetMaxRange() - oHistDGD->GetMinRange())
//             << endl;
   
   OutFile.close();
   }


   {
   ofstream OutFile(OutFileNameRatio.c_str(),ios::out);
   if (!OutFile) {
      cout << "Error: " << OutFileNameRatio << " could not be created."
           << endl;
      exit(1);
   }   
   
   for (int ii = 1; ii < qtBins; ii++) {
      OutFile << pdfX[ii] << "  " 
              << ActualHistogram_pdf[ii]/ActualHistogram_pdf[ii-1] << "  "
              << Histogram_pdf[ii]/Histogram_pdf[ii-1] << "  "
              << endl;
   }

//      OutFile << pdfX[ii] << " " 
//              << pdfY_True[ii]/pdfY[ii]
//	         /(oHistDGD->GetMaxRange() - oHistDGD->GetMinRange())
//             << endl;
   
   OutFile.close();
   }


   delete [] pdfX;
   delete [] pdfY;  
   delete [] pdfY_True;     
   delete [] Histogram_pdf;
} 



