//Id: Raman.cc


//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//##################################################################


// ###############################################
// 
//
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "StartupJH.hh"
#include "ocsOptDataWriter.hh"
#include "MachZehnder.hh"
//#include "ocsRamanAmp.hh"





extern ofstream *LogFile;

void HuWDMBasic(StartupJH * MyStartup);

// ####################################################

// ## Application Wrapper
 
void MyApplication(int argc, char *argv[])  
{

  StartupJH MyStartup("Startup.in");

  switch(MyStartup.GetSimType())
   {
    
     case 1: 
       HuWDMBasic(&MyStartup);
       break;
 

	   
	 default:
      cerr << "MyApplication: Invalid choice of application in SimType" 
			<< MyStartup.GetSimType()
           << endl;
      exit(1);
   } // end switch
} // end MyAppn
// ####################################################


 // oRaman->WriteFileWaveLthPower(FileName);

// ##############################################
// ####################################################


// ####################################################

void HuWDMBasic(StartupJH * MyStartup)
{
  double CLOCK_PER_SEC = 1e+6;
  clock_t Start, Finish;
  double Duration;
  char  C_String[40];
  int QtMapSpans;
  int NumChannels;
  double AveragePower;
  int NumRoundTrip=8;
  double SignalPower;

  
  string FileName,FileName1,FileName2,PowerOutName,PolynomialOutName;
  string SignalAveragePowerOutName,SignalTiltOutName,NoiseAveragePowerOutName, NoiseTiltOutName;
  string OSNR0OutName, OSNR1OutName, OSNR2OutName, OSNR3OutName, OSNR4OutName;
  
  Duration=0;

  LogFileSeparator();

  *LogFile <<  endl << "HuWDMBasic" << endl;
 
  
  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();
  QtMapSpans = MyStartup->GetQtMapSpans();
  RanNumGen * RNG;
  RNG = new RanNumGen(InDir + "RanNumGen.in");



//  RanNumGen * RNG;
//  RNG = new RanNumGen(InDir + "RanNumGen.in");

  cout << "run to here 0" << endl;

  OptSignal * Signal;
  OptSignal * NoiseFreeSignal;
  OptSignal * Noise;
  MachZehnder * SignalMZ;
  MachZehnder * NoiseFreeSignalMZ;
  MachZehnder * NoiseMZ; 

 
  if(MyStartup->GetHybridSignalFlag())
  {    
		cerr << "Error: NoHybridSignal now!!!" << endl 
           << " Abort!!!"<< endl;
    	exit(1);

  }
  else
  {

 	Signal = new OptSignal(InDir + "Signal.in",RNG);
    if(Signal->GetTypePulse() == NRZ)
    {
      SignalMZ = new MachZehnder(InDir + "MZ.in",Signal);
      SignalMZ->MZModulator();
    }
	
	 
	NoiseFreeSignal = new OptSignal(InDir + "Signal.in",RNG);
    if(NoiseFreeSignal->GetTypePulse() == NRZ)
    {
       NoiseFreeSignalMZ = new MachZehnder(InDir + "MZ.in",NoiseFreeSignal);
       NoiseFreeSignalMZ->MZModulator();        
    } 
	 
    Noise = new OptSignal(InDir + "Noise.in",RNG);
    if(Noise->GetTypePulse() == NRZ)
    {
       NoiseMZ = new MachZehnder(InDir + "MZ.in",Noise);
       NoiseMZ->MZModulator();
    }	  
   } 

 	sprintf(C_String,"SignalfoutBeforeEDFA.dat",1);
	FileName = Job + C_String;
  	Signal->WriteFileWavelengthPower(FileName);
	
  sprintf(C_String,"NoiseFreeSignalfoutBeforeEDFA.dat");
  FileName = Job + C_String;
  NoiseFreeSignal->WriteFileWavelengthPower(FileName); 

  sprintf(C_String,"NoisefoutBeforeEDFA.dat");
  FileName = Job + C_String;
  Noise->WriteFileWavelengthPower(FileName); 
  
  
  
    
  OptDataWriter DataWriter(InDir + "DataWriter.in", InDir, Job, Signal);
  DataWriter.WriteInitialData();

  NumChannels = Signal->GetNumChannels();
  int Lp=0;	


int NumFibersPumped = 3;

   
  int NumAmps = QtMapSpans;
  OptAmplifier Ampm0(InDir,InDir + "Amp.in",Signal,RNG);
  Ampm0.ReadInErParameter(InDir +"Er24Linear");
  Ampm0.ReadInGainFlatenFilter(InDir +"EDFALinearGainFlatenFilter.in");
 // Ampm0.ReadInPolynomialFilter(InDir +"EDFAPolynomialFilter.in");
  
	Ampm0.SetDebugLevel(MyStartup->GetDebugLevelAmps());
    Ampm0.SetTypeAmplifier(typeAmplifier(MyStartup->GetTypeAmp()));
    Ampm0.SetTypeAmplifierNoise(typeAmplifierNoise(MyStartup->GetTypeAmplifierNoise()));
 	 

	cout.precision(10);   
	
	Ampm0.SetPumpPower(0.02); // input pump power 20 mW
 
         cout << "Power = " << Linear2dBm(Signal->GetAveragePower())<< endl;
		 
		 


	
	   Ampm0.AmplifyOptSignal();
   	   Ampm0.PropagateSignalandNoise(NoiseFreeSignal,Noise);
		   
//	   DataWriter.WriteEndMap(1,kk,PropagatedLength);

    cout << " After one EDFA power = " << Linear2dBm(Signal->GetAveragePower()) << endl;



	sprintf(C_String,"SignalfoutAfterEDFA.dat",1);
	FileName = Job + C_String;
  	Signal->WriteFileWavelengthPower(FileName);

  sprintf(C_String,"NoiseFreeSignalfoutAfterEDFA.dat");
  FileName = Job + C_String;
  NoiseFreeSignal->WriteFileWavelengthPower(FileName); 

  sprintf(C_String,"NoisefoutAfterEDFA.dat");
  FileName = Job + C_String;
  Noise->WriteFileWavelengthPower(FileName); 


//  DataWriter.WriteEndTransmission(expt_num,kk,PropagatedLength);


  cout << '\a' ;

 
  delete RNG;
 
  cout << "THE END" << endl;
  *LogFile << "THE END" << endl;

 //exit(1);	
} // ## end HuWDMBasic
// ##############################################

