// MyApp.cc

#include "ocs.hh"
#include "StartupHua.hh"

void AnalyticalFormulaUnpolarizedNoise(StartupExample * MyStartup);
void AnalyticalFormulaNoPMD_NoisePartiallyPolarized(
			  StartupExample * MyStartup);

void AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarized(
				    StartupExample * MyStartup);

void AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarizedExperiment(
					     StartupExample * MyStartup);

void MonteCarloUnpolarizedNoise(StartupExample * MyStartup);

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[]) 
{
  

  StartupExample MyStartup("Startup.in",argc,argv);
 
  switch(MyStartup.GetSimType())
   {
     
   case 2:
      AnalyticalFormulaUnpolarizedNoise(&MyStartup);
      break;

   case 3:
      AnalyticalFormulaNoPMD_NoisePartiallyPolarized(&MyStartup);
      break;

   case 4:
      AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarized(&MyStartup);
      break;

   case 5:
      AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarizedExperiment(
							      &MyStartup);
      break;

   case 6: MonteCarloUnpolarizedNoise(&MyStartup);
     break;

    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);

   } // end switch

   cout << "THE END" << endl;


} // ## end MyApplication

// ################################################################

void AnalyticalFormulaUnpolarizedNoise(StartupExample * MyStartup)
{
  // Must set TypeAmplifierNoise = 4 in Amps infiles

 string Job = MyStartup->GetJobPrefix();

 
 RanNumGen *RNGSignal;
 RanNumGen *RNGAmp;

 OptSignal * Signal;

 RNGSignal =  new RanNumGen("./RanNumGenSignal.in");
 RNGAmp = new RanNumGen("./RanNumGenSignal.in");

 Signal = new OptSignal("./Signal.in", RNGSignal);

 LumpedLoss Loss("./Loss.in",Signal);
 OptAmplifier Amp("./Amp.in",Signal,RNGAmp);

 int DemuxChannArrayIndex = 0;
 string InDir = "./";

 int OSAAppendFlag = 0;

  OptSpectrumAnalyzer OSA("OSA.in",Job,OSAAppendFlag,Signal);
 
 GaussianReceiver oGaussianReceiver(Signal,RNGSignal,
				   DemuxChannArrayIndex,InDir,Job);

 Loss.ApplyLumpedLoss();
 Amp.AmplifyOptSignal();

 Signal->WriteFileTimePower(Job + "tInOpt.dat");

 OSA.Update(0);
 OSA.OutputOSA(0);
 OSA.OutputSNR_OSA(0);

 string FileID = "";

 oGaussianReceiver.ReceiveSignal(FileID);

}

// ######################################################

void AnalyticalFormulaNoPMD_NoisePartiallyPolarized(StartupExample * MyStartup)
{

  // Must set TypeAmplifierNoise = 4 in Amps infiles

 string Job = MyStartup->GetJobPrefix();

 string JobDir = Job; // Hua: Is this what you need?? 

 RanNumGen *RNGSignal;
 RanNumGen *RNGAmp;

 OptSignal * Signal;

 RNGSignal =  new RanNumGen("./RanNumGenSignal.in");
 RNGAmp =  new RanNumGen("./RanNumGenAmp.in");

 Signal = new OptSignal("./Signal.in", RNGSignal);

 //OptAmplifier * Amp1;
 //OptAmplifier * Amp2;

 //Amp1 = new OptAmplifier("./PolAmpASE.in",Signal,RNGAmp);
 //Amp2 = new OptAmplifier("./PolAmp.in",Signal,RNGAmp);

 int DemuxChannArrayIndex = 0;
 string InDir = "./";

 GaussianReceiver oGaussianReceiver(Signal,RNGSignal,
				   DemuxChannArrayIndex,InDir,Job);

 Signal->WriteFileTimePower(Job + "tInOpt.dat");

  // Used in conjunction with GaussianReceiver class:
   // We use SignalNoiseBandwidth = 1.0 since
   // TotalNoiseSpectralDensity is noise power per Hertz

 double SignalNoiseBandwidth = 1.0;

 ReduOptSignal ReduUnpolNoise("./Noise.in",SignalNoiseBandwidth);
 ReduOptSignal ReduPolNoise("./Noise.in",SignalNoiseBandwidth);
  
 ReduOptAmplifier ReduUnpolAmpASE("./UnpolAmpASE.in",&ReduUnpolNoise);

 ReduOptAmplifier ReduUnpolAmp("./UnpolAmp.in",&ReduUnpolNoise);
 ReduOptAmplifier ReduPolAmpASE("./PolAmpASE.in",&ReduPolNoise);
 ReduOptAmplifier ReduPolAmp("./PolAmp.in",&ReduPolNoise);

 
 int TypeScattering = 1; //## TypeScattering: 1:Euler rotations
 ReduPolarizTransformer ReduPCase(TypeScattering, &ReduPolNoise);
 
 ReduPolarizer oReduPolarizer(&ReduPolNoise);

 double PolStokes[3];

 // Just for now set:

 PolStokes[0] = 1;
 PolStokes[1] = 2;
 PolStokes[2] = 3;

 // All above here goes before any loop over parameters.

 oReduPolarizer.SetPolarization(PolStokes[1],PolStokes[2],PolStokes[3]); 
 
 ReduUnpolNoise.RegenerateReduOptSignal();
 ReduPolNoise.RegenerateReduOptSignal();
	      	

 ReduUnpolAmpASE.AmplifyOptSignal();
 ReduUnpolAmp.AmplifyOptSignal();
	
 ReduPolAmpASE.AmplifyOptSignal();
 ReduPolAmp.AmplifyOptSignal();
	   
 oReduPolarizer.PolarizeOptSignal();

 ReduPolNoise.AddReducedNoise(&ReduUnpolNoise);

 cout << "APP: TotalNoiseSpectralDensity (reduced Model) [W/Hz] = " <<
   ReduPolNoise.GetTotalNoiseSpectralDensity() << endl;
 

 /*Signal->RegenerateSignalString();
 Amp1->AmplifyOptSignal();
 Amp2->AmplifyOptSignal();
 */

 Signal->AddReducedNoiseToNoiseFreeSignal(&ReduPolNoise);

 string FileID = "";

 oGaussianReceiver.ReceiveSignal(FileID);
}

// ######################################################

void AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarized(
				    StartupExample * MyStartup)
{

  // Must set TypeAmplifierNoise = 4 in Amps infiles


 string Job = MyStartup->GetJobPrefix();

 string JobDir = Job; // Hua: Is this what you need?? 

 RanNumGen *RNGSignal;

 OptSignal * Signal;

 RNGSignal =  new RanNumGen("./RanNumGenSignal.in");

 Signal = new OptSignal("./Signal.in", RNGSignal);

 int DemuxChannArrayIndex = 0;
 string InDir = "./";

 GaussianReceiver oGaussianReceiver(Signal,RNGSignal,
				   DemuxChannArrayIndex,InDir,Job);

 Signal->WriteFileTimePower(Job + "tInOpt.dat");

 // FIBER STUFF

 RanNumGen * RNGFiber;
 RNGFiber =  new RanNumGen("./RanNumGenFiber.in");

 OptFiber *PMF;
 PMF = new OptFiber("./PMF.in", Signal, RNGFiber);

 PMF->GetFiberSampleCompleteScatt();

 oGaussianReceiver.ComputeFirstOrderPMDParameters(PMF);

  // NOISE STUFF

  // Used in conjunction with GaussianReceiver class:
   // We use SignalNoiseBandwidth = 1.0 since
   // TotalNoiseSpectralDensity is noise power per Hertz

 double SignalNoiseBandwidth = 1.0;

 ReduOptSignal ReduUnpolNoise("./Noise.in",SignalNoiseBandwidth);
 ReduOptSignal ReduPolNoise("./Noise.in",SignalNoiseBandwidth);
  
 ReduOptAmplifier ReduUnpolAmpASE("./UnpolAmpASE.in",&ReduUnpolNoise);

 ReduOptAmplifier ReduUnpolAmp("./UnpolAmp.in",&ReduUnpolNoise);
 ReduOptAmplifier ReduPolAmpASE("./PolAmpASE.in",&ReduPolNoise);
 ReduOptAmplifier ReduPolAmp("./PolAmp.in",&ReduPolNoise);

 int TypeScattering = 1; //## TypeScattering: 1:Euler rotations
 ReduPolarizTransformer ReduPCase(TypeScattering, &ReduPolNoise);

 ReduPolarizer oReduPolarizer(&ReduPolNoise);

 double PolStokes[3];

 // Just for now set:

 PolStokes[0] = 1;
 PolStokes[1] = 0;
 PolStokes[2] = 0;

 // All above here goes before any loop over parameters.

  double PropagatedLength = 0;
  PMF->PropagateFiberCoupledNLS(&PropagatedLength);

 oReduPolarizer.SetPolarization(PolStokes[1],PolStokes[2],PolStokes[3]); 

 ReduUnpolNoise.RegenerateReduOptSignal();
 ReduPolNoise.RegenerateReduOptSignal();
	      	
 ReduUnpolAmpASE.AmplifyOptSignal();
 ReduUnpolAmp.AmplifyOptSignal();

 ReduPolAmpASE.AmplifyOptSignal();
 ReduPolAmp.AmplifyOptSignal();
	   
 oReduPolarizer.PolarizeOptSignal();

 ReduPolNoise.AddReducedNoise(&ReduUnpolNoise);

 Signal->AddReducedNoiseToNoiseFreeSignal(&ReduPolNoise);

 string FileID = "";

 oGaussianReceiver.ReceiveSignal(FileID);

}

// ###########################################################

void AnalyticalFormulaFirstOrderPMDAndNoisePartiallyPolarizedExperiment(
				    StartupExample * MyStartup)
{

 string Job = MyStartup->GetJobPrefix();

 string JobDir = Job; // Hua: Is this what you need?? 

 RanNumGen *RNGSignal;

 OptSignal * Signal;

 RNGSignal =  new RanNumGen("./RanNumGenSignal.in");

 Signal = new OptSignal("./Signal.in", RNGSignal);

 int DemuxChannArrayIndex = 0;
 string InDir = "./";

 GaussianReceiver oGaussianReceiver(Signal,RNGSignal,
 				    DemuxChannArrayIndex,InDir,Job);

 double DGD = 30e-12;
 double ClockRecoveryTime = 50e-12;
 double TotalNoiseSpectralDensity = 3e-16;
 double DOP_Noise = 0.5;
 double DOP_Signal = 0.9;
 string OutFileID = "";
 
 double StokesPlusPlus[3];
 double UnitAverageStokesSignal[3];
 double UnitStokesVectorPolarizedPartNoise[3];

 // Unit length

 StokesPlusPlus[0] = 1.0/sqrt(3.0);
 StokesPlusPlus[1] = 1.0/sqrt(3.0);
 StokesPlusPlus[2] = 1.0/sqrt(3.0);

UnitAverageStokesSignal[0] = 0.0;
UnitAverageStokesSignal[1] = 1.0/sqrt(2.0);
UnitAverageStokesSignal[2] = 1.0/sqrt(2.0);

UnitStokesVectorPolarizedPartNoise[0] = 1.0/sqrt(2.0);
UnitStokesVectorPolarizedPartNoise[1] = 1.0/sqrt(2.0);
UnitStokesVectorPolarizedPartNoise[2] = 0.0;

oGaussianReceiver.ReceiveSignalDGD_NoisePartiallyPolarized(
     DGD, 
     ClockRecoveryTime, 
     StokesPlusPlus, 
     UnitAverageStokesSignal,
     DOP_Signal,
     TotalNoiseSpectralDensity,
     DOP_Noise, 
     UnitStokesVectorPolarizedPartNoise,
     OutFileID);

} 

// ###########################################

void MonteCarloUnpolarizedNoise(StartupExample * MyStartup)
{

  // Must set TypeAmplifierNoise = 1 in Amp.in

 string Job = MyStartup->GetJobPrefix();

 int NumMCExpts = MyStartup->GetNumMonteCarloExpts();

 string JobDir = Job; 

 RanNumGen *RNGSignal;

 OptSignal * Signal;

 RNGSignal =  new RanNumGen("./RanNumGenSignal.in");

 Signal = new OptSignal("./Signal.in", RNGSignal);


 LumpedLoss * Loss;

 Loss = new LumpedLoss("./Loss.in",Signal);
 OptAmplifier * Amp;

 Amp = new OptAmplifier("./Amp.in",Signal,RNGSignal);

 OptFilter * oOptFilter;

  oOptFilter = new OptFilter("TunableOptFilter.in", Signal);

 Photodetector * oPhotodetector;

 oPhotodetector = new Photodetector("Photodetector.in",Signal,RNGSignal);

 ElecFilter * oElecFilter;

 oElecFilter = new ElecFilter("ElecFilter.in",oPhotodetector);

 ElecSignalStat * oElecSignalStat;

 oElecSignalStat = new ElecSignalStat("ElecSignalStat.in",Signal, 
				     oPhotodetector);


 int OSAAppendFlag = 0;
 OptSpectrumAnalyzer * OSA;
 OSA = new OptSpectrumAnalyzer("OSA.in", Job, OSAAppendFlag, Signal);
 double PropagatedLength=0.0;

 for(int ii=0;ii<NumMCExpts;ii++)
   {
     Signal->RegenerateSignalString();
     Loss->ApplyLumpedLoss();
     Amp->AmplifyOptSignal();

     OSA->Update(PropagatedLength);

     if(ii==0)
      Signal->WriteFileTimePower(Job+"tOutOpt.dat");

     oOptFilter->FilterOptSignal(); 
     oPhotodetector->DetectOptSignal();  
     oElecFilter->FilterElecSignal();

     if(ii==0)
      oPhotodetector->WriteFileTimePower(Job+"tOutElec.dat");

     oElecSignalStat->UpdatePerformanceStatistics();
    
   }

   OSA->OutputOSA(MyStartup->GetNumMonteCarloExpts());
   OSA->OutputSNR_OSA(MyStartup->GetNumMonteCarloExpts());

   oElecSignalStat->OutputGaussianStatisticsFromMonteCarlo(Job);

}
