// MyApp.cc
#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "ocsOptDataWriter.hh"
#include "ocsOptFiberLocalError.hh"
#include "StartupExampleHuaPMF.hh"
#include "ocsReceiverSubsystem.hh"

/*
Declare all functions in this file 
(except for the function MyApplication) here:
*/

void MyPMF(StartupExample * MyStartup);

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[]) 
{
 
  StartupExample MyStartup("Startup.in",argc,argv);

  /*
  Read O'Reilly pp. 110-115 for info on switch statements. 
  switch must take an integer argument, in this case
  MyStartup.GetSimulationType()
  */

  switch(MyStartup.GetProjIndex())
   {
    case 1:
      MyPMF(&MyStartup);
      break;
	  
    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);
   } // end switch
} // ## end MyApplication


void MyPMF(StartupExample * MyStartup)
{
 // ####### Get Enviroment parameters ###########
 
 string JobDir = MyStartup->GetJobDir();
 cout << "JobDir = " << JobDir << endl;
 string JobPrefix = MyStartup->GetJobPrefix();
 cout << "JobPrefix = " << JobPrefix << endl << endl << endl;
 
 cout << "Start of Pulses through PMF..." << endl << endl;
 LogFileSeparator();
 *LogFile << "Start of Pulses through PMF" << endl;
 
 //## Files to save SOP in time and freq domain ##
 ofstream *FileSopFreq;
 ofstream *FileSopTime;
 ofstream *FileStokes;
 
 string FilenameStokes = JobDir + "SignalTotalSOP.dat"; //## save the average SOP of the signal and the noise to file
 FileStokes = new ofstream(FilenameStokes.c_str(),ios::out);
 
 // ###### Initialize Transmitter ###########
 
 RanNumGen RNG(JobDir+"RanNumGen.in");
 RanNumGen RNGAmps(JobDir + "RanNumGenAmps.in");
 RanNumGen RNGSignal(JobDir + "RanNumGenSig.in");
 
 OptSignal SignalToPMF(JobDir + "Signal.in",&RNG);
 OptSignal SignalAfterPMF(JobDir + "Signal.in",&RNG);
 OptSignal UnpolNoise(JobDir + "Noise.in",&RNG);
 OptSignal PolNoise(JobDir + "Noise.in",&RNG);
 OptSignal SignalToReceiver(JobDir + "Signal.in",&RNG);
 
 // ## Unpolarized noise and amp
 OptAmplifier UnpolAmpASE(JobDir + "UnpolAmpASE.in",&UnpolNoise,&RNGAmps);
 OptAmplifier UnpolAmp(JobDir + "UnpolAmp.in",&UnpolNoise,&RNGAmps);
 
 // ## Polarized noise and amp
 OptAmplifier PolAmpASE(JobDir + "PolAmpASE.in",&PolNoise,&RNGAmps);
 OptAmplifier PolAmp(JobDir + "PolAmp.in",&PolNoise,&RNGAmps);
 Polarizer oPolarizer(&PolNoise);
 
 int TypeScattering = 1; //## TypeScattering: 1:Euler rotations
 PolarizTransformer PCsig(TypeScattering, &SignalAfterPMF);
 PolarizTransformer PCase(TypeScattering, &PolNoise);
 
 SignalAfterPMF.WriteFileTimePower(JobPrefix + "tInOpt.dat");    //*****: Save initial pulse train

 // ### Monitor
 OptDataWriter DataWriter(JobDir + "DataWriter.in",JobDir,JobPrefix,&SignalAfterPMF);
 //OptDataWriter DataWriterReceiver(JobDir + "DataWriter.in",JobDir,JobPrefix,&SignalToReceiver);
 
 DataWriter.WriteInitialData();
 

 // ############# Initialize Fibers ############################

 int NumFibers = 1;
 double DGD_PMF = 0.0;
  
 //OptFiber PMF(JobDir + "PMF.in",&SignalAfterPMF, &RNG);
 
 OptFiber *PMF;
 PMF = new OptFiber("./PMF.in", &SignalAfterPMF, &RNG);


 // ######### Initialize OSAs ############################     

 int OSAAppendFlag = 0;
 OptSpectrumAnalyzer OSAFinal(JobDir + "OSA.in", JobPrefix, OSAAppendFlag, &SignalToReceiver);


// ####### Intialize receivers ###########
 OptFilter oOptFilter(JobDir + "OptFilter.in", &SignalToReceiver);

 Photodetector oPhotodetector(JobDir + "Photodetector.in", &SignalToReceiver,&RNGSignal);

 ElecFilter oElecFilter(JobDir + "ElecFilter.in",&oPhotodetector);

 ElecSignalStat oElecSignalStat(JobDir + "ElecSignalStat.in", &SignalToReceiver, &oPhotodetector);

 Histogram2D RonEyeDiagram(JobDir + "Histogram2D.in",&SignalToReceiver);

 //int DemuxChannArrayIndex = 1;
 //ReceiverSubsystem oReceiverSubsystem(&SignalToReceiver, &RNG,"./","RZ_PMF",DemuxChannArrayIndex,&DataWriterReceiver);
  
  	   
 // ##### Variables to store SOP of the signal
  	   
 double StokesInFreq[4];
 double StokesInTime[4];
 double *SopFreq, *SopTime;
 double DOP_Time, DOP_Freq;
 SopFreq = StokesInFreq;
 SopTime = StokesInTime;
   
 double DeltaFreq = SignalAfterPMF.GetDeltaFreq();
 double DeltaTime = SignalAfterPMF.GetDeltaTime();
 

 //PMF->SetScatteringMatrix(IndexStep,0,0,0); // Set TransM as unit matrix
 
 // int NumLaunchState = 2; // ## Number of launching state to PMF ##


// ######## Read in parameter table from file ########
/*
 string FilenameTable = JobDir + "Table.in";

 if (MyStartup->GetReadParameterTableFlag())
 {
	 int NumParameterGroups =  MyStartup->GetNumParameterGroups();
	 int NumParameters =  MyStartup->GetNumParameters();
 
	 for (int ParameterGroup=0;ParameterGroup<NumParameterGroups;ParameterGroup++)
	 {
	 	cout << ParameterGroup << " ";
 		for (int NumParameter = 0; NumParameter<NumParameters; NumParameter++)
		{	
		cout << MyStartup->ReadTableFromFile(FilenameTable,ParameterGroup,NumParameter)<<" ";
		}
		cout <<endl;
	 }
 }	
 */
 
  // #### File to save performance data ####
   
	  string outFile_performance;
	  outFile_performance = JobDir + "Performance_PulseAfterPMF.dat";
	  ofstream OutFile_performance;
   
	  OutFile_performance.open(outFile_performance.c_str(),ios::out);
	 
	  if (!OutFile_performance) {
	     cerr << "Error: " << outFile_performance << " could not be open."<< endl;
	     exit(1);
	     }      
		 
	  OutFile_performance << "Q factor" << endl;
	    
  	
	
 // ########################################################################
 // ########################### Start of loops #############################
 
 for (int LaunchState = 0; LaunchState <  MyStartup->GetNumSignalLaunchState(); LaunchState++)
 //for (int LaunchState = 19; LaunchState < 21; LaunchState++)
 { 
 	char C_String[4];
	sprintf(C_String,"%i",LaunchState);
	 
	string FilenameSopFreq = JobDir + "SignalSopFreq_SOP" + C_String + ".dat"; //## SOP of signal as a function of freq
	string FilenameSopTime = JobDir + "SignalSopTime_SOP" + C_String + ".dat"; //## SOP of signal as a function of time
   
  
	FileSopFreq = new ofstream(FilenameSopFreq.c_str(),ios::out);
 	FileSopTime = new ofstream(FilenameSopTime.c_str(),ios::out);
 
 	if (!FileSopFreq)
 	{
	 cerr << "Error: FileSopFreq could not be open." << endl << flush;
	 exit(1);
	 }
	 
	cout << "Launching SOP = " << LaunchState << " .............." << endl << endl; 
    //cout << "To scatter fiber .. " << endl << endl;
	
	SignalAfterPMF = SignalToPMF; // ## Initialize launching state to PMF
	
    //PMF->GetFiberSampleCompleteScatt();
	int IndexStep = 0;
	PMF->SetScatteringMatrix(IndexStep,0,0,0); // Set TransM as unit matrix
	
	SignalAfterPMF.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
	DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];
    
	if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		{
		   cout << "Total Noiseless Signal SOP in Time before entering PMF:" << endl;
		   cout << "S0 = " << StokesInTime[0] << endl;
		   cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
		   cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
		   cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
		   cout << " Total Noiseless Signal DOP in Time before entering PMF = " << DOP_Time << endl << endl;
		}	
	
	
	//PCsig.TransformPolarization(pi/4,pi/2,pi/4);
 
	// #### Set signal SOP ####	
	double Stokes[4];
	double SplittingFactorGama = 0.5; // LaunchState*1.0/(MyStartup->GetNumSignalLaunchState()-1.0); //##: should be between 0 and 1.
	cout << "SplittingFactorGama = " << SplittingFactorGama << endl << endl;
	
	double c_plus = sqrt(SplittingFactorGama);
	double c_minus = sqrt(1- SplittingFactorGama);
	double AngleBetweenPspSignalStokes = 2*acos(c_plus);
	cplx Jones[2];

    Stokes[0] = StokesInTime[0];
    //Stokes[1] = cos(pi - AngleBetweenPspSignalStokes)*Stokes[0];
    //Stokes[2] = 0.0*Stokes[0];
    //Stokes[3] = sin(pi - AngleBetweenPspSignalStokes)*Stokes[0];
	
	double AngleSig = pi*0/8.0;
	
	Stokes[1] = cos(pi - AngleBetweenPspSignalStokes);
	Stokes[2] = sin(AngleSig)*sin(pi - AngleBetweenPspSignalStokes);
	Stokes[3] = cos(AngleSig)*sin(pi - AngleBetweenPspSignalStokes);
	
	cout <<"Angle between PSP and Signal SOP = " << AngleBetweenPspSignalStokes << endl << endl;
 	//cout << "SOP of signal = " << Stokes[1] << " " << Stokes[2] << " " << Stokes[3] << endl << endl;
	
	//### Convert Stokes vector to Jones vector) ####

	Stokes2Jones(Stokes, Jones); // ## This function doesn't keep power, so the Jones vector is normalized here
	
//	cout << "Jones Vector = " << Jones[0] << " " << Jones[1] << endl << endl;
	
	// ### Set signal SOP by setting the Jones vector
	SignalAfterPMF.SetInJones(0,0,Jones[0]); // (IndexStep,JonesIndex,JonesValue)
	SignalAfterPMF.SetInJones(0,1,Jones[1]);
	
	SignalAfterPMF.RegenerateSignalString(); // ### Apply the Jones vector
		
	
	if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		{	
		   SignalAfterPMF.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
		   DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];
	
		   cout << "Total Noiseless Signal SOP in Time after setting SOP:" << endl;
		   cout << "S0 = " << StokesInTime[0] << endl;
		   cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
		   cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
		   cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
		   cout << " Total Noiseless Signal DOP in Time after setting SOP = " << DOP_Time << endl << endl;
		}	
	
	// ############# Propagate signal in the fiber ##############################
	
	double PropagatedLength = 0;
    PMF->PropagateFiberCoupledNLS(&PropagatedLength);  

 	 
 	if (LaunchState == 0)
	{
		DGD_PMF = PMF->GetDGD(SignalAfterPMF.GetCenterWavelength());
		cout << "DGD_PMF = " << DGD_PMF << endl;
		
		SignalAfterPMF.WriteFileTimePower(JobPrefix + "tInOptAfterPMF.dat");    //*****: Save noiseless signal pulse train
		
	}
    
	// ############## Calcuate signal SOP within the pulse in time and frequency ##########
	int zz;
    int zz_min = 0;
    int zz_max = SignalAfterPMF.GetQtPoints();
   
    double FreqMin = 0; // FreqMin and FreqMax define the freq window to calculate the SOP and they are real frequencies
	double FreqMax = 0;
    int TimeMin = 0; // TimeMin and TimeMax define the time window and are the index of sampling point [0,qtPoints]
	int TimeMax = 0;
 
    *LogFile << "SOP of the pulse" << endl;
    
		
		FileSopFreq = new ofstream(FilenameSopFreq.c_str(),ios::out);
	 	FileSopTime = new ofstream(FilenameSopTime.c_str(),ios::out);
 
 		if (!FileSopFreq)
	 	{
		   cerr << "Error: FileSopFreq could not be open." << endl << flush;
		   exit(1);
		}
	 
		for (zz =  zz_min; zz < zz_max; zz ++)
	   	{
   
		  FreqMin = (zz - zz_max/2)*DeltaFreq;
		  FreqMax = (zz - zz_max/2 + 1)*DeltaFreq;

		  TimeMin = zz;
		  TimeMax = zz + 1;  
	
		  SignalAfterPMF.GetTotalStokesParametersInBandwidth(SopFreq,FreqMin,FreqMax);
		  SignalAfterPMF.GetTotalStokesParametersInTime(SopTime, TimeMin,TimeMax);
   
		  DOP_Freq = sqrt(sq(StokesInFreq[1]) + sq(StokesInFreq[2]) + sq(StokesInFreq[3]))/StokesInFreq[0];
		  DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];
   
		  // *LogFile << "Stokes parameter of the pulse: " << endl;
		  //*LogFile << StokesInFreq[0] <<" "<< StokesInFreq[1] << " "<< StokesInFreq[2] << " "<< StokesInFreq[3]<< endl;
		  *FileSopFreq << StokesInFreq[0] << " " << StokesInFreq[1] << " " << StokesInFreq[2] << " " << StokesInFreq[3] << " " << DOP_Freq << endl;
		  *FileSopTime << StokesInTime[0] << " " << StokesInTime[1] << " " <<StokesInTime[2]<< " " <<StokesInTime[3]<< " " << DOP_Time << endl;
		 }
   
		  // ###### Calculate the average SOP of the signal in time and freq domain #####
		  SignalAfterPMF.GetTotalStokesParametersInBandwidth(SopFreq,-1*DeltaFreq*SignalAfterPMF.GetQtPoints()/2,DeltaFreq*SignalAfterPMF.GetQtPoints()/2);
		  SignalAfterPMF.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());

		  DOP_Freq = sqrt(sq(StokesInFreq[1]) + sq(StokesInFreq[2]) + sq(StokesInFreq[3]))/StokesInFreq[0];
		  DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];   

	  		
		  *FileStokes << LaunchState << " " << StokesInTime[0] << " " << StokesInTime[1] << " " 
		  		<<StokesInTime[2]<< " " <<StokesInTime[3]<< " " << DOP_Time << " " ;
   
		
		  if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		  {
		   	   cout << "Total Noiseless Signal SOP in Time:" << endl;
		  	   cout << "S0 = " << StokesInTime[0] << endl;
			   cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
			   cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
			   cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
			   cout << " Total Noiseless Signal DOP in Time = " << DOP_Time << endl << endl;
		  }	
	 
	 
	  // #### Check PDVector #####
	  
	  double OutPDVectorPlus[3],OutPDVectorMinus[3];
	  double StokesPlusPlus[4],StokesMinusMinus[4];
	  
	  for (int kk = 0; kk < 3; kk++)
	  {
	  	OutPDVectorPlus[kk] = 0;
	  }
	
	  PMF->GetOutPD_Vector(SignalAfterPMF.GetCenterFreq(),OutPDVectorPlus); 


	  // ######### Calculate S++, S--, S+-, S45 #########	  
	  for (int kk = 0; kk < 3; kk++)
	  {
	  	OutPDVectorMinus[kk] = -1*OutPDVectorPlus[kk];
		StokesPlusPlus[kk+1] = OutPDVectorPlus[kk]/DGD_PMF;
		StokesMinusMinus[kk+1] = OutPDVectorMinus[kk]/DGD_PMF;
	  }
	  
	  StokesPlusPlus[0] = DGD_PMF;
	  StokesMinusMinus[0] = DGD_PMF;
/*	  
	  cout << "PDVectorPlus = " << OutPDVectorPlus[0] << " " <<OutPDVectorPlus[1] << " " << OutPDVectorPlus[2] << endl ;
	  cout << "PDVectorMinus = " << OutPDVectorMinus[0] << " " <<OutPDVectorMinus[1] << " " << OutPDVectorMinus[2] << endl << endl;
	  cout << "S++ = " << StokesPlusPlus[0] << " " <<StokesPlusPlus[1] << " " << StokesPlusPlus[2] << StokesPlusPlus[3] << endl ;
	  cout << "S-- = " << StokesMinusMinus[0] << " " <<StokesMinusMinus[1] << " " << StokesMinusMinus[2] << StokesMinusMinus[3] <<endl << endl;
  */
	  cplx JonesVectorPlus[2], JonesVectorMinus[2];
	  
	  Stokes2Jones(StokesPlusPlus,JonesVectorPlus);
	  Stokes2Jones(StokesMinusMinus,JonesVectorMinus);
	  
//	  cout << "JonesVectorPlus = " << JonesVectorPlus[0] << " " << JonesVectorPlus[1] << endl;
//	  cout << "JonesVectorMinus = " << JonesVectorMinus[0] << " " << JonesVectorMinus[1] << endl << endl;
	  
	  cplx StokesPlusMinus[4];
	  
	  StokesPlusMinus[0] = JonesVectorPlus[0]*JonesVectorMinus[0] + JonesVectorPlus[1]*JonesVectorMinus[1];
	  StokesPlusMinus[1] = JonesVectorPlus[0]*JonesVectorMinus[0] - JonesVectorPlus[1]*JonesVectorMinus[1];
	  StokesPlusMinus[2] = JonesVectorPlus[1]*JonesVectorMinus[0] + JonesVectorPlus[0]*JonesVectorMinus[1];
	  StokesPlusMinus[3] = -1*jc*(JonesVectorPlus[1]*JonesVectorMinus[0] - JonesVectorPlus[0]*JonesVectorMinus[1]);

	  //cout << "S+- = " << StokesPlusMinus[0] << " " << StokesPlusMinus[1] << " " << StokesPlusMinus[2] << " " << StokesPlusMinus[3] << endl;
	 
		  
	  // ## Clear RonED and stats before running a new MC simulation

  	  oElecSignalStat.ClearElecSignalStat();
	  //oElecSignalStat.ClearPatternStats();
	  
      RonEyeDiagram.Clear();
	  
	  
	  // ######## Set Noise power to get certain OSNR #############
	   
	  OSAFinal.Clear();
	  
	  double AmpGain = 15.0; //LaunchState*0.3 + 12.0;
	  cout << " AmpGain for LaunchState " << LaunchState << " = " << AmpGain << endl << endl;
	  
	  UnpolAmp.SetGain(AmpGain);
	  PolAmp.SetGain(AmpGain + 3);
		
	//oPolarizer.SetPolarization(0,0,1); //### Set polarizer direction
	
	// ### Rotate the Noise SOP from S-- to S++ #####
	
	double StokesNoiseTemp[4];
	double AngleSmmSnoise;
	
	// AngleSmmSnoise = LaunchState*1.0/(MyStartup->GetNumSignalLaunchState()-1.0)*2*pi;
		
	// StokesNoiseTemp[0] = StokesInTime[0];
    //StokesNoiseTemp[1] = cos(AngleSmmSnoise);
    //StokesNoiseTemp[2] = -1*sin(AngleSmmSnoise);
    //StokesNoiseTemp[3] = 0;
	
	double StokesSqSum = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]));
	double VectorDirection = -1; //1 is parallel, -1 is anti-podal
	
	StokesNoiseTemp[1] = VectorDirection*StokesInTime[1]/StokesSqSum;
	StokesNoiseTemp[2] = VectorDirection*StokesInTime[2]/StokesSqSum;
	StokesNoiseTemp[3] = VectorDirection*StokesInTime[3]/StokesSqSum;
	
	cout << "Signal SOP = " << StokesInTime[1] << " " << StokesInTime[2] << " " << StokesInTime[3] << endl;
	cout << "Noise SOP = " << StokesNoiseTemp[1] << " " << StokesNoiseTemp[2] << " " << StokesNoiseTemp[3] << endl << endl;
	
	//cout <<"Angle between PSP and Signal SOP = " << AngleBetweenPspSignalStokes << endl << endl;
		
	 oPolarizer.SetPolarization(StokesNoiseTemp[1],StokesNoiseTemp[2],StokesNoiseTemp[3]); //### Set polarizer direction
	
	
	
	  // ################################ Start of MC loop ##############################################
	  for (int expt_num=0;expt_num<MyStartup->GetNumMonteCarloExpts();expt_num++)
      {
  
      	// cout << " Num of MC Expt = " << MyStartup->GetNumMonteCarloExpts() << endl;
	  	 //  cout << " Current MC Expt = " << expt_num << endl;
	  
	   	SignalToReceiver = SignalAfterPMF;
	   	
	   	UnpolNoise.RegenerateSignalString();
	   	PolNoise.RegenerateSignalString();
	   
	   	// ##### Unpolarized signal arm #####
	   	UnpolAmpASE.AmplifyOptSignal();
  	   	UnpolAmp.AmplifyOptSignal();
  	   
	   	UnpolNoise.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
	   	DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];   
	
	   
	   // ##### Polarized signal arm #####
	   PolAmpASE.AmplifyOptSignal();
  	   PolAmp.AmplifyOptSignal();
	   
	   oPolarizer.PolarizeOptSignal();

		if (expt_num == 0)
		{	  
		   PolNoise.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
		   DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];   

//	 	  *FileStokes << StokesInTime[0] << " " << StokesInTime[1] << " " <<StokesInTime[2]<< " " <<StokesInTime[3]<< " " << DOP_Time << " ";
		
		  if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		  {  
		    cout << "Polarized noise SOP in Time:" << endl;
		    cout << "S0 = " << StokesInTime[0] << endl;
		    cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
		    cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
		    cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
		    cout << " Polarized noise DOP in Time = " << DOP_Time << endl << endl;
		  }  
	    }
		
		
	   // ##### Total noise = UnpolNoise + PolNoise #####
	    UnpolNoise.AddFreqDomain(&PolNoise);
	   
	   
		if (expt_num == 0)
		{
		   UnpolNoise.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
		   DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];   
		   
//		   *FileStokes << StokesInTime[0] << " " << StokesInTime[1] << " " <<StokesInTime[2]<< " " <<StokesInTime[3]<< " " << DOP_Time << endl; 
		
 		   if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		   {  
		     cout << "Total noise SOP in Time:" << endl;
		     cout << "S0 = " << StokesInTime[0] << endl;
		     cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
		     cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
		     cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
		     cout << " Total noise DOP in Time = " << DOP_Time << endl << endl;
	       }
	    }
		 
		 
	    SignalToReceiver.AddFreqDomain(&UnpolNoise);
		
	
  
	 // ### Final OSA and SNR before entering receiver ##########
	   if (MyStartup->GetWriteSNRandOSAFlag())
	   {
	 //    cout << "WriteSNRandOSAFlag = " << MyStartup->GetWriteSNRandOSAFlag() << endl;
	 	OSAFinal.Update(PropagatedLength);
	 
	   }

	
    // ######### Enter Reciever Subsystem and output Eye and Q information   #########
	
       oOptFilter.FilterOptSignal();
	 
	   oPhotodetector.DetectOptSignal();  
	   oElecFilter.FilterElecSignal();    

    // oElecSignalStat.GetSingleStringElecSignalStat();
	
	   //oElecSignalStat.AddSingleStringSampleNEW();

		// ## added on Nov. 15, 2005...
		
		int qtPoints = SignalToReceiver.GetQtPoints();
  		int StringLength = SignalToReceiver.GetStringLength();
	    int TargetIndexTime = int(MyStartup->GetClockTime()/SignalToReceiver.GetDeltaTime());

 		for (int jj = 0; jj < StringLength; jj++)
		    oElecSignalStat.BitIntensity[jj] =real(oPhotodetector.sfftPM.tPowerSignal[Periodicity(TargetIndexTime
                                  +qtPoints/StringLength*jj,qtPoints)]);

		
//		int StringLength = SignalToReceiver.GetStringLength();
  		for(int ii = 0;ii < StringLength; ii++)

			oElecSignalStat.UpdatePatternStats(ii,oElecSignalStat.BitIntensity[ii]);
		
		
  	   // Output stats every Startup->GetFreqOutputStats() Monte Carlo expts
/*
 		if((expt_num+1) % MyStartup->GetFreqOutputStats() == 0)
        {
          char  C_String3[40];

          sprintf(C_String3,"PatternStats_expt_num%i_LaunchState%i.dat",expt_num+1,LaunchState);
          string OutFileName = JobDir + C_String3;

          oElecSignalStat.WriteFilePatternStats(OutFileName);

        } // end if

*/ 
	   RonEyeDiagram.UpdateEyeDiagram(&SignalToReceiver,&oPhotodetector);
	
	  //oElecSignalStat.WriteFileEyeDiagram(JobDir + "Eye.dat"); 

	} // ################# End MC simulations ########################
	

  
	cout << endl << "End of MC simulation for input SOP " << LaunchState << ".........." << endl << endl;
	
	char  C_String3[40];

    sprintf(C_String3,"PatternStats_ExptNum%i_LaunchState%i.dat",MyStartup->GetNumMonteCarloExpts(),LaunchState);
    string OutFileName = JobDir + C_String3;

    oElecSignalStat.WriteFilePatternStats(OutFileName);
	oElecSignalStat.WriteFilePatternStats(JobDir + "PatternStatsFinal.dat");


    oPhotodetector.WriteFileTimePower(JobDir + "tReceivedCurrent.dat");   
	int OmitDCFlag = 0;
	oPhotodetector.WriteFileFreqPower(JobDir + "fReceivedCurrent.dat",OmitDCFlag);
	

	OSAFinal.OutputOSA(MyStartup->GetNumMonteCarloExpts(),C_String);
	OSAFinal.OutputSNR_OSA(MyStartup->GetNumMonteCarloExpts(),C_String);
	
	double Q_factor = oElecSignalStat.GetQ_FactorFromIsolatedBits();
/*	cout << "IsolatedBit1_PowerMean = " << oElecSignalStat.GetIsolatedBit1_PowerMean() << endl;
	cout << "IsolatedBit0_PowerMean = " << oElecSignalStat.GetIsolatedBit0_PowerMean() << endl;
	cout << "IsolatedBit1_PowerStdDev = " << oElecSignalStat.GetIsolatedBit1_PowerStdDev()  << endl;
	cout << "IsolatedBit0_PowerStdDev = " << oElecSignalStat.GetIsolatedBit0_PowerStdDev() << endl << endl;
	*/
	
	if (MyStartup->GetDisplayStokesParameterOnScreenFlag() == 1)
		{  
		   UnpolNoise.GetTotalStokesParametersInTime(SopTime,0,SignalAfterPMF.GetQtPoints());
		   DOP_Time = sqrt(sq(StokesInTime[1]) + sq(StokesInTime[2]) + sq(StokesInTime[3]))/StokesInTime[0];
		   
		   cout << "Total noise SOP for LaunchState = " << LaunchState << " in Time:" << endl;
		   cout << "S0 = " << StokesInTime[0] << endl;
		   cout << "S1 = " << StokesInTime[1]/StokesInTime[0] << endl;
		   cout << "S2 = " << StokesInTime[2]/StokesInTime[0] << endl;
		   cout << "S3 = " << StokesInTime[3]/StokesInTime[0] << endl;
		   cout << "Total noise DOP in Time = " << DOP_Time << endl << endl;
	  	}					  
   
	 *FileStokes << " " << StokesNoiseTemp[1] << " " <<StokesNoiseTemp[2]<< " " <<StokesNoiseTemp[3] << " " << endl;
   
   // #### Output performance data to file ####
   
   	  OutFile_performance << AngleSmmSnoise << " " 
	  		<< oElecSignalStat.GetQ_FactorFromIsolatedBits() << " "
	  		<< oElecSignalStat.GetIsolatedBit1_PowerMean() << " "
			<< oElecSignalStat.GetIsolatedBit0_PowerMean() << " "
	  		<< oElecSignalStat.GetIsolatedBit1_PowerStdDev() << " " 
			<< oElecSignalStat.GetIsolatedBit0_PowerStdDev() << endl;
	  
	  cout << "Q factor = " << oElecSignalStat.GetQ_FactorFromIsolatedBits() << endl;
	
	  *LogFile << "THE END of pulse through PMF" << endl;
	  cout << "THE END of pulse though PMF" << endl << endl;

	 	
     char C_String2[80];
	 sprintf(C_String2,"EyeMC_%iLaunchState.dat",LaunchState);
     string FileName = C_String2;
	 RonEyeDiagram.EyeDiagramWriteHistogram(FileName);
	 
 	 
	} //####### End of loop of all Launching states
  
    OutFile_performance.close(); 
	
} //########## End of MyPMF




