// MyApp.cc
#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "ocsOptDataWriter.hh"
#include "ocsOptFiberLocalError.hh"
#include "StartupExample.hh"

/*
Declare all functions in this file 
(except for the function MyApplication) here:
*/

void MyPMF(StartupExample * MyStartup);

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[]) 
{
  /*
  Construct an instance  MyStartup of the class StartupExample
  See MyStartup.hh for the definition of this class 

  Startup.in is the name of the input parameter file for the class
  StartupExample
  */

  StartupExample MyStartup("Startup.in",argc,argv);

  /*
  Read O'Reilly pp. 110-115 for info on switch statements. 
  switch must take an integer argument, in this case
  MyStartup.GetSimulationType()
  */

   cout << " Going to switch ..." << endl;
   
  switch(MyStartup.GetSimType())
   {
    case 1:
      MyPMF(&MyStartup);
      break;
	  
    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);
   } // end switch
} // ## end MyApplication



// ##########################################################################################
// ############# DMS RZ system with PMD - version 2 #########################################
// ############# with John and Brian's recommendations ##################################
void MyPMF(StartupExample * MyStartup)
{
 // ####### Get Enviroment parameters ###########
 
 string JobDir = MyStartup->GetJobDir();
 cout << "JobDir = " << JobDir << endl;
 string JobPrefix = MyStartup->GetJobPrefix();
 cout << "JobPrefix = " << JobPrefix << endl;
 
 cout << "Start of Pulses through PMF..." << endl << endl;
 LogFileSeparator();
 *LogFile << "Start of Pulses through PMF" << endl;
 
 
 // ###### Initialize Transmitter ###########
 
 RanNumGen RNG(JobDir+"RanNumGen.in");
 //OptSignal Signal(JobDir+"Signal.in", &RNG);
 OptSignal SignalPulse(JobDir+"Signal.in", &RNG);
 
 SignalPulse.WriteFileTimePower(JobPrefix + "tInOpt.dat");    //*****: Save initial pulse train

// ### Transmitter Loss ###
// LumpedLoss TransmitterLoss(JobDir + "TransmitterLoss.in",&SignalPulse);
//TransmitterLoss.ApplyLumpedLoss();
 
 // ### Transmitter Amp ###
 //OptAmplifier TransmitterAmp(JobDir + "TransmitterAmp.in",&SignalPulse,&RNG);
 
 //if (MyStartup->GetAmpsNoiseOffFlag())
 //{
 //  TransmitterAmp.SetTypeAmplifierNoise(NOISE_OFF);
 //  cout << "Tansmitter amplifier noise off!" << endl;
 //}
 
// TransmitterAmp.AmplifyOptSignal();
 
 // ### Monitor
 OptDataWriter DataWriter(JobDir + "DataWriter.in",JobDir,JobPrefix,&SignalPulse);
 DataWriter.WriteInitialData();
 
 //cout << " Initial AverageSigmaMarks in time [ps] = " << 1e12*Signal.ComputeAverageSigmaMarks()
 //	<< endl;

 // ############# Initialize Fibers ############################

 int NumFibers = 1;

// OptFiberLocalError PMF(JobDir + "PMF.in", &SignalPulse, &RNG, JobPrefix);

 OptFiber PMF(JobDir + "PMF.in",&SignalPulse, &RNG);
 
// OptFiberLocalError LoopSMF2(JobDir + "LoopSMF2.in", &Signal, &RNG, JobPrefix);

 
 // ############## Initialize EDFAs #############################
 //int NumAmps = 5;
   
 //OptAmplifier LoopEDFA1(JobDir + "LoopEDFA1.in", &Signal, &RNG);
 
//OptAmplifier *LoopAmpPtr[NumAmps];
// LoopAmpPtr[0] = &LoopEDFA1;

 // ## Set parameters of Amplifiers: DebugLevel and Noise
 
 //if (MyStartup->GetDebugLevelAmps())
 //{
 //  for (int i=0;i<NumAmps;i++)
 //     LoopAmpPtr[i]->SetDebugLevel(MyStartup->GetDebugLevelAmps());
 //}
 
// if (MyStartup->GetAmpsNoiseOffFlag())
// {
//   cout << " ## Loop EDFA NOISE OFF! " << endl;   
//   for (int i=0;i<NumAmps;i++)
//      LoopAmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
// }
 
 // ## Set parameters of fibers: Beta3Zero, NonLinearityZero and FiberLossZero
 
 //if (MyStartup->GetBeta3ZeroFlag())
 //{
  // cout << " ## Second order Dispersion Zero! " << endl;
  //cout << " Beta3ZeroFlag = " << MyStartup->GetBeta3ZeroFlag() << endl;
   
  //for (int i=0;i<NumFibers;i++)
  //   FiberPtr->SetSndOrDispFiber(0);
 //}

 //if (MyStartup->GetNonLinearityZeroFlag())
 //{
 // cout << " ## Nonlinearity Zero! " << endl;
 //cout << " Nonlinearity Gamma = " << MyStartup->GetNonLinearityZeroFlag() << endl;
   
 //  for (int i=0;i<NumFibers;i++)
 //    FiberPtr->SetGammaFiber(0);
 //}	   
	   
// if (MyStartup->GetFiberLossZeroFlag())
 //{
//   cout << " ## Fiber Loss Zero! " << endl;
//   cout << " FiberLossZeroFlag = " << MyStartup->GetFiberLossZeroFlag() << endl;
   
//   for (int i=0;i<NumFibers;i++)
//      FiberPtr->SetAttenuation(0);
// }	   

// ## set relative error goal

// cout << endl;
// cout << " Relative Error Goal = " << MyStartup->GetRelativeErrorGoal() <<endl;
// for (int i=0;i<NumFibers;i++)
//      FiberPtr->SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());

	
 // ######### Initialize OSAs ############################     

 int OSAAppendFlag = 0;

 OptSpectrumAnalyzer OSAInitial(JobDir + "OSA.in", JobPrefix, OSAAppendFlag, &SignalPulse);
 OptSpectrumAnalyzer OSAFinal(JobDir + "OSA.in", JobPrefix, OSAAppendFlag, &SignalPulse);

 // # Back to back OSA

 OSAInitial.Update(0);
 OSAInitial.Output(0);


 // ######### Initialize Receiver #######################
//  OptFilter oOptFilter("OptFilter.in",&Signal);
  Photodetector  oPhotodetector("Photodetector.in",&SignalPulse); 
//  ElecFilter oElecFilter("ElecFilter.in",&oPhotodetector);
//  ElecSignalStat oElecSignalStat(&Signal,&oPhotodetector);  

 //ReceiverSubsystem Receiver(&Signal, &RNG, JobDir, JobPrefix, MyStartup.GetDemuxChannelIndex());
 //Receiver.ReceiveSignalBackToBack(Signal);


// ##################################################################
// ########### Propagation in the PMF #######################
 
 for (int expt_num=0;expt_num<MyStartup->GetNumMonteCarloExpts();expt_num++)
 {
  
//  cout << " Num of MC Expts = " << MyStartup->GetNumMonteCarloExpts() << endl;
  double PropagatedLength = 0;
  int ZStepNum = 0;
  
  //SMF.Propagate(&PropagatedLength, &ZStepNumber);
  
  int kk = 0;
  
 // DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);
  
  cout << " NumMapSpans = " << MyStartup->GetQtMapSpans() <<endl;
  
  for (kk = 0; kk < MyStartup->GetQtMapSpans(); kk++)
  {
   // ### Span #1: SMF (3.5km) + DSF(25km) + PC (Loss=1.5dB) + EDFA #1  ###

    DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);
//PMF->PropagateFiberCoupledNLS(&PropagatedLength);  
 	
	//LoopExtraLossSpan1.ApplyLumpedLoss();
 	
	DataWriter.WriteWithinMap(expt_num,kk,PropagatedLength);
	
	//LoopEDFA1.AmplifyOptSignal();
	
	// ### Span #2: DSF(25km) + EDFA #2 ###

    
  }	// ## End for kk - loop


// ########### End of propagation in the PMF ######################
// #################################################################

  
  DataWriter.WriteEndTransmission(expt_num,kk,PropagatedLength);
  
  // ### Final OSA and SNR ##########
  
//  if (MyStartup->GetWriteSNRandOSAFlag())
//  {
//    cout << " WriteSNRandOSAFlag = " << MyStartup->GetWriteSNRandOSAFlag() << endl;
//	OSAFinal.Update(PropagatedLength);
//	OSAFinal.Output(expt_num+1);
//  }
	
  // ###############################################################################
  // ######### Enter Reciever Subsystem and output Eye and Q information   #########

//   oOptFilter.FilterOptSignal();

   RanNumGen RNGSignal(JobDir + "RanNumGenSig.in");

   Photodetector oPhotodetector(JobDir + "Photodetector.in", &SignalPulse,&RNGSignal);
   oPhotodetector.DetectOptSignal();   
//   oElecFilter.FilterElecSignal();
	ElecSignalStat oElecSignalStat(&SignalPulse, &oPhotodetector);
	
   oElecSignalStat.GetSingleStringElecSignalStat();

   oPhotodetector.WriteFileTimePower(JobDir + "tReceivedCurrent.dat");   
   int OmitDCFlag = 0;
   oPhotodetector.WriteFileFreqPower(JobDir + "fReceivedCurrent.dat",OmitDCFlag);
   oElecSignalStat.WriteFileEyeDiagram(JobDir + "Eye.dat"); 


   // #### Ouput performance data to file ####
   
   string outFile_performance;
   outFile_performance = JobDir + "Pulse_PMF.dat";
   ofstream OutFile_performance;
   
   OutFile_performance.open(outFile_performance.c_str(),ios::out);
      if (!OutFile_performance) {
         cerr << "Error: " << outFile_performance << " could not be open."<< endl;
         exit(1);
      }      
   OutFile_performance << "Q factor = " << oElecSignalStat.GetQ_Factor() << endl;
   OutFile_performance << "BER = " << oElecSignalStat.GetBitErrorRate() << endl;
   OutFile_performance.close();   
   
   cout << "Q factor = " << oElecSignalStat.GetQ_Factor() << endl;
   cout << "BER = " << oElecSignalStat.GetBitErrorRate() << endl;

   cout << "Back to Back Eye Opening [mV] = " 
	<< 1e3*oElecSignalStat.GetMinAmplitudeMargin() << endl;

   cout << "OutputPDVector = " << SignalPulse.OutputPDVectorOFLE[0] << endl <<
   SignalPulse.OutputPDVectorOFLE[1] <<endl<< SignalPulse.OutputPDVectorOFLE[2] << endl;
   cout << "DGD = " << sqrt(sq(SignalPulse.OutputPDVectorOFLE[0]) +
   sq(SignalPulse.OutputPDVectorOFLE[1]) + sq(SignalPulse.OutputPDVectorOFLE[2])) << endl;
   
   *LogFile << "THE END" << endl;
   cout << "THE END" << endl;

  } // ##### End of Propagation in PMF ########################3
  
	
 cout << endl << "Propagation in PMF ends normally" << endl;
 
} // #################### END of propagation in PMF#############################




