// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2006):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "NoiseMMC.hh"
#include "ocsOptAmplifier.hh"

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[]) 
{

 time_t StartTime = time(NULL);
 cout << "Start Time: " << ctime(&StartTime) << endl;

 // Call NoiseMMC constructor here

  NoiseMMC * NoiseMMCptr;

  NoiseMMCptr = new NoiseMMC("./Startup.in");

 // Call methods to run NoiseMMC here

  NoiseMMCptr->RunMMC();
  NoiseMMCptr->~NoiseMMC();

 time_t EndTime = time(NULL);
 cout << "End Time: " << ctime(&EndTime) << endl;
 
 double TotalTime = double(EndTime - StartTime);

 struct tms my_tms;

 times(&my_tms);

 double CPUtime = (double(my_tms.tms_utime))/(double(sysconf(_SC_CLK_TCK)));

 cout << "Total Run Time [hrs] = " << TotalTime/3600.0 << endl
      << "Total CPU Time [hrs] = " << CPUtime/3600.0 << endl;
  
  cout << "THE END" << endl;
  *LogFile << "End Time= " << ctime(&EndTime) << endl; 
  *LogFile << "Total Run Time [hrs]= " << TotalTime/3600.0 << endl; 
  *LogFile << "THE END" << endl;


} // end MyApplication

// ###############################################################

NoiseMMC::NoiseMMC(string InFileName)
{

 ReadInSystemParameters(InFileName);
 AllocateSystemMemory();

 ReadInMMCParameters(InFileName);
 AllocateMMCMemory();
 
}
 
// #################################################################

void NoiseMMC::ReadInSystemParameters(string InFileName)
{

 ifstream InFile(InFileName.c_str(),ios::in);

 if (!InFile)
  {
    cerr << "Error: Input file "<< InFileName
         << " could not be opened."<< endl;
    exit(1);
  }

string JobDir = ReadString("JobDir",&InFile);
string JobName =  ReadString("JobName",&InFile);

string LogFileName = JobDir + JobName + ".log";

cout << "Log File = " << LogFileName << endl << endl;

LogFile = new ofstream(LogFileName.c_str(),ios::out);

if (!LogFile)
  {
    cerr << "Error: ocs.log could not be open." << endl << flush;
    exit(1);
  }

Job = JobDir + JobName + "_";

*LogFile << "####################################" << endl;

*LogFile << "NoiseMMC Class" << endl;

*LogFile << InFileName << endl << endl;

*LogFile << "Job  = " << Job << endl;

NameSystem = nameSystem(ReadInt("NameSystem"," ",1,4,LOWER_AND_UPPER,MANDATORY,
				&InFile));

NumMaps = ReadInt("qtMapSpans"," ",0,-1,LOWER_ONLY,MANDATORY,&InFile);

TypeStepSizes = ReadInt("TypeStepSizes"," ",1,4,
                         LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile);

switch(TypeStepSizes)
   {
    case 1: case 2:
     RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,LOWER_ONLY,
				    &InFile);
     break;

    case 3:
     WalkOffParameter = ReadDouble("WalkOffParameter"," ",0,-1,
                                   LOWER_ONLY,MANDATORY,1,&InFile);
     break;

    case 4:
      ConstantStepSize = ReadDouble("ConstantStepSize"," ",0,-1,LOWER_ONLY,
				    &InFile);
      break;
        
    default:
            cerr << "NoiseMMC::TypeStepSize incorrect! Aborting!" << endl;
            exit(-1);
        }      
                                                                   

OutputStepSizesFlag = ReadInt("OutputStepSizes"," ",0,1,
			      LOWER_AND_UPPER,&InFile);

TestStepSizesFlag = bool(ReadInt("TestStepSizesFlag"," ",0,1,
                                  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile));


Beta3ZeroFlag = ReadInt("Beta3ZeroFlag"," ",0,1,LOWER_AND_UPPER,&InFile);
NonLinearityZeroFlag = ReadInt("NonLinearityZeroFlag"," ",0,1,
                                LOWER_AND_UPPER,&InFile);

TypeAmp = ReadInt("TypeAmp"," ",0,5,LOWER_AND_UPPER,&InFile);

AmpsNoiseOffFlag = ReadInt("AmpsNoiseOffFlag"," ",0,1,LOWER_AND_UPPER,&InFile);

DebugLevelAmps = ReadInt("DebugLevelAmps"," ",0,3,LOWER_AND_UPPER,
                          OPTIONAL_NO_WARNING,1,&InFile);

ClockRecoveryTime = ReadDouble("ClockRecoveryTime"," ",0,-1,LOWER_ONLY,
				    MANDATORY,&InFile);
 
LoadSignalFileFlag = ReadInt("LoadSignalFileFlag"," ",0,1,
                                LOWER_AND_UPPER,&InFile);

if(LoadSignalFileFlag)
 SignalFileName = ReadString("SignalFileName",&InFile);


DoStdMCFlag = bool(ReadInt("DoStdMCFlag"," ",0,1,
                            LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile));


 InFile.close();

} //## end ReadInSystemParameters

// #########################################################

void NoiseMMC::AllocateSystemMemory(void)
{

  RNGSignal = new RanNumGen("./RanNumGenSig.in");
  RNGAmps = new RanNumGen("./RanNumGenAmps.in");
  RNGFibers = new RanNumGen("./RanNumGenFibers.in");

  if (LoadSignalFileFlag)
     {
      *LogFile << "Loading Signal from File" << endl;

      Signal = new OptSignal("./Signal.in",RNGSignal,SignalFileName);
     }
  else
    Signal = new OptSignal("./Signal.in",RNGSignal);
  
  // Create a deep copy of the signal
  InputSignal=new OptSignal(*Signal); 

  DataWriter=new OptDataWriter("./DataWriter.in","./",Job,Signal);

  Amp=new OptAmplifier("./Amp.in",Signal,RNGAmps);
  PAmp=new OptAmplifier("./PAmp.in",Signal,RNGAmps);
  
  int NumAmps = 1;
  OptAmplifier * AmpPtr[NumAmps];
  AmpPtr[0] =  Amp;
  AmpPtr[1] =  PAmp;

  for(int ii=0;ii<NumAmps;ii++)
    {
     AmpPtr[ii]->SetDebugLevel(DebugLevelAmps);
     AmpPtr[ii]->SetTypeAmplifier(typeAmplifier(TypeAmp));

     if(AmpsNoiseOffFlag)
       AmpPtr[ii]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   PAmp->SetTypeAmplifierNoise(NOISE_OFF);

  // JZ: Probably best not to have any noise in PAmp. Check vs AWGN case
   // If do put noise in PAmp change  NumRelevantFreqs below
 
   double sigma_ASE_2_per_freq = (Amp->GetSpontaneousEmissionFactor())
               *hPlanck*Signal->GetCenterFreq()*Signal->GetDeltaFreq()
	       *(Amp->GetLinearGainOptAmplif()-1);

  stan_dev_real_imag = sqrt (sigma_ASE_2_per_freq/2.0);

  cout << "stan_dev_real_imag = " << stan_dev_real_imag << endl;

 if(NameSystem == OLEG)
  {
     PreDCF= new OptFiberLocalError("./PreDCF.in",Signal,RNGFibers,Job);
     PostDCF=new  OptFiberLocalError("./PostDCF.in",Signal,RNGFibers,Job);

     Dplus=new OptFiberLocalError("./Dplus.in",Signal,RNGFibers,Job);
     Dminus=new OptFiberLocalError("./Dminus.in",Signal,RNGFibers,Job);

     int NumFibers = 4;

     OptFiberLocalError * FiberPtr[NumFibers];

     FiberPtr[0] = PreDCF;
     FiberPtr[1] = PostDCF;
     FiberPtr[2] = Dplus;
     FiberPtr[3] = Dminus;
   
     if(Beta3ZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

     if(NonLinearityZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

    for(int ii = 0; ii< NumFibers;ii++)
       {
        FiberPtr[ii]->SetOutputStepSizesFlag(OutputStepSizesFlag);
        FiberPtr[ii]->SetTypeStepSizes(TypeStepSizes);

        switch(TypeStepSizes)
          {
           case 1: case 2:

              FiberPtr[ii]->SetRelativeErrorGoal(RelativeErrorGoal);
              break;

           case 3:
             FiberPtr[ii]->SetWalkOffParameter(WalkOffParameter);
             break;

            default:
             cerr << "Wrong choice of TypeStepSizes. It should be 1,2,3."
                  << endl;
             exit(1);
          } // end switch

       } // end for

  } // end NameSystem == OLEG

  if(NameSystem == TYCO)
    {
     PreDCF= new OptFiberLocalError("./PreDCF.in",Signal,RNGFibers,Job);
     PostDCF=new  OptFiberLocalError("./PostDCF.in",Signal,RNGFibers,Job);

     DSF35=new OptFiberLocalError("./DSF35.in",Signal,RNGFibers,Job);
     DSF45=new OptFiberLocalError("./DSF45.in",Signal,RNGFibers,Job);
     SMF=new OptFiberLocalError("./SMF.in",Signal,RNGFibers,Job);

     int NumFibers = 5;
    
     OptFiberLocalError * FiberPtr[NumFibers];

     FiberPtr[0] = PreDCF;
     FiberPtr[1] = PostDCF;
     FiberPtr[2] = DSF45;
     FiberPtr[3] = DSF35;
     FiberPtr[4] = SMF;

     if(Beta3ZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

     if(NonLinearityZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

    for(int ii = 0; ii< NumFibers;ii++)
       {
	FiberPtr[ii]->SetOutputStepSizesFlag(OutputStepSizesFlag);
	FiberPtr[ii]->SetTypeStepSizes(TypeStepSizes);

        switch(TypeStepSizes)
	  {
	   case 1: case 2:

	      FiberPtr[ii]->SetRelativeErrorGoal(RelativeErrorGoal);
              break;

	   case 3:
             FiberPtr[ii]->SetWalkOffParameter(WalkOffParameter);
             break;
	   
	    default:
             cerr << "Wrong choice of TypeStepSizes. It should be 1,2,3."
                  << endl;
             exit(1);
          } // end switch
 
       } // end for
       
    } //end if NameSystem==TYCO

  if(NameSystem == MARCO)
    {
     Fiber= new OptFiberLocalError("./Fiber.in",Signal,RNGFibers,Job);
  
     Fiber->SetGammaFiber(2e-3);
 
     if(Beta3ZeroFlag)
     {
         Fiber->SetSndOrDispFiber(0);
     }

     if(NonLinearityZeroFlag)
       {
         Fiber->SetGammaFiber(0);
       }

	Fiber->SetOutputStepSizesFlag(OutputStepSizesFlag);
	Fiber->SetTypeStepSizes(TypeStepSizes);

        switch(TypeStepSizes)
	  {
	   case 1: case 2:

	      Fiber->SetRelativeErrorGoal(RelativeErrorGoal);
              break;

	   case 3:
             Fiber->SetWalkOffParameter(WalkOffParameter);
             break;
	   
	    default:
             cerr << "Wrong choice of TypeStepSizes. It should be 1,2,3."
                  << endl;
             exit(1);
          } // end switch
       
    Loss = new LumpedLoss(Amp->GetdBGain(),Signal);

    } //end if NameSystem==MARCO


  if(NameSystem==TYCO_B2B)
    {
     Loss = new LumpedLoss(Amp->GetdBGain(),Signal);
     PLoss = new LumpedLoss(PAmp->GetdBGain(),Signal);
    } 

  oOptFilter = new OptFilter("./TunableOptFilter.in",Signal);
  oPhotodetector=new Photodetector("./Photodetector.in",Signal,RNGSignal);
  oElecFilter = new ElecFilter("./ElecFilter.in",oPhotodetector);

  if(DoStdMCFlag)
    oElecSignalStat = new ElecSignalStat("ElecSignalStat.in",Signal,
                                          oPhotodetector);

} //## end AllocateSystemMemory

// #########################################################

void NoiseMMC::ReadInMMCParameters(string InFileName)
{

ifstream InFile(InFileName.c_str(),ios::in);

 if (!InFile)
  {
    cerr << "Error: Input file "<< InFileName
         << " could not be opened."<< endl;
    exit(1);
  }

 LogFileSeparator();
 *LogFile << "NoiseMMC::ReadInMMCParameters" << endl;

 ControlBit = ReadInt("ControlBit"," ",0,Signal->oBitString->GetStringLength(),
		      LOWER_AND_UPPER,MANDATORY,&InFile);

 // Internally we use an integer in range 0 to StringLength-1

 ControlBit -= 1;

 IndependentDecisionsFlag = bool(ReadInt("IndependentDecisionsFlag"," ",0,1,
					  LOWER_AND_UPPER,MANDATORY,&InFile));

 

 RestartJobFlag = bool(ReadInt("RestartJobFlag"," ",0,1,LOWER_AND_UPPER,
                       MANDATORY,0,&InFile));
 
 MinimumPdfValue = ReadDouble("MinimumPdfValue"," ",0,-1,LOWER_ONLY,
			      MANDATORY,&InFile);

 MaxRelError = ReadDouble("MaxRelError"," ",0,-1,LOWER_ONLY,
			   MANDATORY,&InFile);

  int TotalNumAmps;
  int NumAmpsPerMap;

 switch(NameSystem)
   {

   case OLEG:

    NumAmpsPerMap = 1;
    TotalNumAmps = NumMaps; // We Made PAmps noiseless

   case TYCO_B2B: case TYCO:

     NumAmpsPerMap = 4; 
     TotalNumAmps = NumMaps*NumAmpsPerMap; // We Made PAmps noiseless.

     break;

   case MARCO:

     NumAmpsPerMap = 1;
     NumMaps =1;
     TotalNumAmps = NumMaps*NumAmpsPerMap;

     break; 

   default:

     cerr << "NoiseMMC::ReadInParameters switch error. Aborting." << endl;
     exit(1);

   }


 int NumRelevantFreqsInFile = ReadInt("NumRelevantFreqs"," ",0,
				  Signal->GetQtPoints(),LOWER_AND_UPPER,
				  MANDATORY,&InFile);

 NumRelevantFreqs = 0;

 if(NumRelevantFreqsInFile==1)   
    NumRelevantFreqs=Signal->GetQtPoints();
 else
  {   	
      if(NumRelevantFreqsInFile > 1 &&
         NumRelevantFreqsInFile <= Signal->GetQtPoints())

	 NumRelevantFreqs = NumRelevantFreqsInFile;
      
      else
         {
          cerr << "Number of relevant complex frequencies must be within "
	       << "range [1,QtPoints]. \n";
         exit(1);
         }
  }

  if (NumRelevantFreqs%2)
      {
      cerr << "Number of relevant complex frequencies must an even number. \n";
      exit(1);
      }
	 

  //! Factor of two is for real and imag parts of noise
  //! We include pre and post amps here too (see above).

  DimStateSpace = unsigned(2*NumRelevantFreqs*TotalNumAmps);

  NumIterations = unsigned(ReadInt("NumIterations"," ",0,-1,LOWER_ONLY,
				   MANDATORY,&InFile));

  RNGSkipSamplesFlag = bool(ReadInt("RNGSkipSamplesFlag"," ",0,-1,LOWER_ONLY,
                                    OPTIONAL_WARNING,0,&InFile));

  if(RNGSkipSamplesFlag)
   {
     NumRNGSamplesToSkip = ReadInt("NumRNGSamplesToSkip"," ",0,-1,LOWER_ONLY,
                                    MANDATORY,&InFile);

   }

  InFile.close();

  if(DoStdMCFlag)
   {
    IndependentDecisionsFlag = 0;
    MaxRelError = 0.0;
    NumIterations = 1;

   }
   

} // ## end ReadInMMCParameters

// #############################################################

void NoiseMMC::AllocateMMCMemory(void)
{

  RNG_MMC = new RanNumGen("./RanNumGenMMC.in");

  if(RNGSkipSamplesFlag)
    {
     for(int ii = 0; ii < NumRNGSamplesToSkip*(2*DimStateSpace + 1);ii++)
      RNG_MMC->GetRanNum();

     cout << "RNG_MMC: Skipped over " << NumRNGSamplesToSkip << " MMC samples" << endl;
    }

  mmc = new MMC(DimStateSpace,RNG_MMC,Job,"./Startup.in",RestartJobFlag);

  AddBiasedNoiseFlag=1;

} // ## end AllocateMMCMemory

// ###########################################################

NoiseMMC::~NoiseMMC(void)
{

  switch(NameSystem)
   {
    case TYCO_B2B:
      Loss->~LumpedLoss();
      PLoss->~LumpedLoss();
    break;

    case TYCO:
      SMF->~OptFiberLocalError();
      DSF35->~OptFiberLocalError();
      DSF45->~OptFiberLocalError();
      PreDCF->~OptFiberLocalError();
      PostDCF->~OptFiberLocalError();      
    break;

   case MARCO:
      Loss->~LumpedLoss();
      Fiber->~OptFiberLocalError();
      break;

   case OLEG:
      Dplus->~OptFiberLocalError();
      Dminus->~OptFiberLocalError();
      PreDCF->~OptFiberLocalError();
      PostDCF->~OptFiberLocalError();
      break;

    default:
      cerr << "Switch error in NoiseMMC::~NoiseMMC. Abort." <<endl;
      exit(1);

    }

 Amp->~OptAmplifier();
 PAmp->~OptAmplifier();

 oOptFilter->~OptFilter();
 oElecFilter->~ElecFilter();

 oPhotodetector->~Photodetector();

 DataWriter->~OptDataWriter();
 Signal->~OptSignal();
 InputSignal->~OptSignal();

 RNGSignal->~RanNumGen();
 RNGAmps->~RanNumGen();
 RNGFibers->~RanNumGen();

 mmc->~MMC();

 RNG_MMC->~RanNumGen();

} // end Destructor

// #################################################################
   
void NoiseMMC::RunMMC(void)
{

  if(!RestartJobFlag)
    RunNoiseFree();

  IterateMMC();
  
  if(DoStdMCFlag)
   oElecSignalStat->OutputGaussianStatisticsFromMonteCarlo(Job);

}

// #################################################################
     
void NoiseMMC::RunNoiseFree(void)
{

  AddBiasedNoiseFlag=0;
  double V_try = PropagateThroughSystem((double*)mmc->z_try);
  *Signal = *InputSignal;
   
  if(V_try<mmc->V_min || V_try>mmc->V_max)
   {
    cerr << "V_try = " << V_try << endl;
    cerr << "Warning:V_try is not in range [ " << mmc->V_min
	         << " , " << mmc->V_max << " ]! \n";
    cerr << "Aborting" << endl;
    exit(1);
   }

  AddBiasedNoiseFlag=1;

}

// ###############################################################

void NoiseMMC::RestartMMC(void)
{
}

// ###############################################################

void NoiseMMC::IterateMMC(void)
{

  mmc->CurrentMaxRelError = MaxRelError + 1;
 
  // While loop over iterations

  while((MaxRelError==0) ? (mmc->IterationNumber < NumIterations) :
        (mmc->CurrentMaxRelError >  MaxRelError || 
	 mmc->LowestPDFValue > MinimumPdfValue))
    {

      // While loop over metropolis samples

      *LogFile << "Starting Iteration " << mmc->IterationNumber + 1 << endl;

      while(!mmc->AtEndOfStage2)
	{

           if(IndependentDecisionsFlag)
            mmc->GaussianPerturbationIndependentDecisions(stan_dev_real_imag);
           else
            mmc->GaussianPerturbationSingleDecision(stan_dev_real_imag);

           mmc->V_try = PropagateThroughSystem((double*)mmc->z_try);

           /*
           if(TestStepSizesFlag)
            {
              Signal->WriteFileTimePower(Job + "tOutStep.dat");
              Signal->WriteFileFreqPower(Job + "fOutStep.dat");
              exit(1);
            }
           */

	   // Evaluate the sample z_try and V_try
  
           mmc->EvaluateSample();
      	
            // Restoring the original signal without noise

	   *Signal=*InputSignal;

       } // end while loop over samples
	 
  
       // update bias, that is P_k^j

        mmc->Iterate();

    } // end while loop over iterations
     
} // ## end IterateMMC


// ###############################################################

double NoiseMMC::PropagateThroughSystem(double * z_try)
{
 double V_try;

 switch(NameSystem)
  {

   case TYCO_B2B:
    V_try = PropagateTycoB2B(z_try);
    break;

   case TYCO:
    V_try = PropagateTyco(z_try);
    break;

    case MARCO:
    V_try = PropagateMarco(z_try);
    break;

    case OLEG:
    V_try = PropagateOleg(z_try);
    break;

   default:
    cerr << "NoiseMMC::PropagateThroughSystem switch error. Abort." << endl;
    exit(1);
  }
 
return V_try;

} // ## end PropagateThroughSystem

// ################################################################

double NoiseMMC::PropagateTycoB2B(double * z_try)
{

  if(!AddBiasedNoiseFlag)
   DataWriter->WriteInitialData();

  // JZ 16/6/06 PLoss->ApplyLumpedLoss();

  // This amp is noiseless

  // JZ 16/6/06 PAmp->AmplifyOptSignal();

  for(int kk=0;kk<NumMaps;kk++)
   for(int aa=0;aa<NumAmpsPerMap;aa++)
     {
      
      Loss->ApplyLumpedLoss();
 
      Amp->SetBiasedNoiseVector(
	    z_try+(kk*4+aa)*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();

     } // end for loops

  // JZ 16/6/06 PLoss->ApplyLumpedLoss();

  // This amp is noiseless

  // JZ 16/6/06 PAmp->AmplifyOptSignal();
  
  double PropagatedLength = 1.0;

  if(!AddBiasedNoiseFlag) 
    {
      DataWriter->WriteEndTransmission(0,NumMaps,PropagatedLength);

     cout << "Output Av Power [dBm] = "
          << Linear2dBm(Signal->GetAveragePower()) << endl;
    }

  double V_try = ReceiveSignal();

  return V_try;

} // end PropagateTycoB2B

// ################################################################

double NoiseMMC::PropagateTyco(double * z_try)
{
  double PropagatedLength = 0;
  int ZStepNum = 0;

  if(!AddBiasedNoiseFlag)
   DataWriter->WriteInitialData();


  // JZ June 14, 2006. For compatibility with LinearnTyco.cc
  // method TycoNoiseCovariance() used for AWGN and Linearization
  // methods we make pre and post fiber lossless and turn off PAmp.

  PreDCF->SetAttenuationdB(0);
  PreDCF->SetGammaFiber(0);
  PreDCF->Propagate(&PropagatedLength,&ZStepNum);

  // PAmp->AmplifyOptSignal();

  for(int kk=0;kk<NumMaps;kk++)
     {
      
      SMF->Propagate(&PropagatedLength,&ZStepNum);
      DSF35->Propagate(&PropagatedLength,&ZStepNum);

      Amp->SetBiasedNoiseVector(
	    z_try+(kk*4)*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();

      DSF45->Propagate(&PropagatedLength,&ZStepNum);

      Amp->SetBiasedNoiseVector(
	    z_try+(kk*4+1)*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();

      DSF45->Propagate(&PropagatedLength,&ZStepNum);

      Amp->SetBiasedNoiseVector(
	    z_try+(kk*4+2)*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();


      DSF35->Propagate(&PropagatedLength,&ZStepNum);
      SMF->Propagate(&PropagatedLength,&ZStepNum);

      Amp->SetBiasedNoiseVector(
	    z_try+(kk*4+3)*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();

    if (!AddBiasedNoiseFlag)
        DataWriter->WriteEndMap(0,kk,PropagatedLength);
     

     } // end for loops


  PostDCF->SetAttenuationdB(0);
  PostDCF->SetGammaFiber(0);
  PostDCF->Propagate(&PropagatedLength,&ZStepNum);

  //PAmp->AmplifyOptSignal();
  

  if(!AddBiasedNoiseFlag) 
    {
     DataWriter->WriteEndTransmission(0,NumMaps,PropagatedLength);
   
     cout << "Output Av Power [dBm] = "
          << Linear2dBm(Signal->GetAveragePower()) << endl;
    }

  double V_try = ReceiveSignal();

  return V_try;

} // end PropagateTyco

// ################################################################

double NoiseMMC::PropagateOleg(double * z_try)
{
  double PropagatedLength = 0;
  int ZStepNum = 0;

  if(!AddBiasedNoiseFlag)
   DataWriter->WriteInitialData();

  // JZ June 14, 2006. For compatibility with LinearnTyco.cc
  // method TycoNoiseCovariance() used for AWGN and Linearization
  // methods we make pre and post fiber lossless and turn off PAmp.

  PreDCF->SetAttenuationdB(0);
  PreDCF->SetGammaFiber(0);
  PreDCF->Propagate(&PropagatedLength,&ZStepNum);

  for(int kk=0;kk<NumMaps;kk++)
     {

      Dplus->Propagate(&PropagatedLength,&ZStepNum);
      Dminus->Propagate(&PropagatedLength,&ZStepNum);

      Amp->SetBiasedNoiseVector(
            z_try+kk*2*NumRelevantFreqs,
            2*NumRelevantFreqs,AddBiasedNoiseFlag); 

      Amp->AmplifyOptSignal();

      if (!AddBiasedNoiseFlag)
        DataWriter->WriteEndMap(0,kk,PropagatedLength);

     } // end for loops


  PostDCF->SetAttenuationdB(0);
  PostDCF->SetGammaFiber(0);
  PostDCF->Propagate(&PropagatedLength,&ZStepNum);

  if(!AddBiasedNoiseFlag)
    {
     DataWriter->WriteEndTransmission(0,NumMaps,PropagatedLength);
  
     cout << "Output Av Power [dBm] = "
          << Linear2dBm(Signal->GetAveragePower()) << endl;
    }

  double V_try = ReceiveSignal();

  return V_try;

} // end PropagateOleg

// ################################################################

double NoiseMMC::PropagateMarco(double * z_try)
{

  double PropagatedLength = 0;
  int ZStepNum = 0;

  if(!AddBiasedNoiseFlag)
   DataWriter->WriteInitialData();

  Loss->ApplyLumpedLoss();

  Amp->SetBiasedNoiseVector(
	    z_try,2*NumRelevantFreqs,AddBiasedNoiseFlag); 

  Amp->AmplifyOptSignal();

  Fiber->Propagate(&PropagatedLength,&ZStepNum);

  if(!AddBiasedNoiseFlag) 
    {
     DataWriter->WriteEndTransmission(0,NumMaps,PropagatedLength);
   
     cout << "Output Av Power [dBm] = "
          << Linear2dBm(Signal->GetAveragePower()) << endl;
    }

  double V_try = ReceiveSignal();

  return V_try;

} // end PropagateMarco

// ################################################################

double NoiseMMC::ReceiveSignal(void)
{

  oOptFilter->FilterOptSignal();
  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  if(DoStdMCFlag)
    oElecSignalStat->UpdatePerformanceStatistics();
 
  if(!AddBiasedNoiseFlag)
    {       
     oPhotodetector->WriteFileTimePower(Job + "tPowerNoiseFreeElecSignal.dat");
     oPhotodetector->WriteFileFreqPower(Job + "fPowerNoiseFreeElecSignal.dat");
    }

  double TimeForPDF = ClockRecoveryTime + 
                      ControlBit*Signal->GetBitLength();

  while(TimeForPDF < 0)
        TimeForPDF += Signal->GetTimeWindow();

  while(TimeForPDF >= Signal->GetTimeWindow())
        TimeForPDF -= Signal->GetTimeWindow();

  int TimeForPDFIndex =  int(floor(TimeForPDF/Signal->GetDeltaTime()));

  double V_try = (double) real(oPhotodetector->
			       sfftPM.tPowerSignal[TimeForPDFIndex]);

  //cout << "TimeForPDF, V_try = " << TimeForPDF << " " << V_try << endl;

  return V_try;

} // end ReceiveSignal

// ################################################################
