//Id: Noise.cc


//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//##################################################################


// ###############################################
// 
//
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "StartupJH.hh"
#include "ocsOptDataWriter.hh"
#include "MachZehnder.hh"
//#include "ocsRamanAmp.hh"

 



extern ofstream *LogFile;

void RamanBasic(StartupJH * MyStartup);


// ####################################################

// ## Application Wrapper
 
void MyApplication(int argc, char *argv[])  
{
 
  StartupJH MyStartup("Startup.in");

  switch(MyStartup.GetSimType())
   {
    
     case 1: 
	      RamanBasic(&MyStartup);
       break;
 
	   
	 default:
      cerr << "MyApplication: Invalid choice of application in SimType" 
			<< MyStartup.GetSimType()
           << endl;
      exit(1);
   } // end switch
} // end MyAppn
// ####################################################


 // oRaman->WriteFileWaveLthPower(FileName);

void RamanBasic(StartupJH * MyStartup)
{
  double CLOCK_PER_SEC = 1e+6;
  clock_t Start, Finish;
  double Duration;
  char  C_String[40];
  int NumChannels;
  double AveragePower;
  double SignalPower;
	string FileName;
  


  LogFileSeparator();


 
  
  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();
  RanNumGen * RNG;
  RNG = new RanNumGen(InDir + "RanNumGen.in");



//  RanNumGen * RNG;
//  RNG = new RanNumGen(InDir + "RanNumGen.in");


  OptSignal * Signal;
  OptSignal * NoiseFreeSignal;
  OptSignal * Noise;


  MachZehnder * SignalMZ;
  MachZehnder * NoiseFreeSignalMZ;
 


 
 if(MyStartup->GetHybridSignalFlag())
    {    
		cerr << "Error: NoHybridSignal now!!!" << endl 
           << " Abort!!!"<< endl;
    	exit(1);

   }
  else
    {
	 	Signal = new OptSignal(InDir + "Signal.in",RNG);
    if(Signal->GetTypePulse() == NRZ)
    {
      SignalMZ = new MachZehnder(InDir + "MZ.in",Signal);
      SignalMZ->MZModulator();
    }
	
	 
	NoiseFreeSignal = new OptSignal(InDir + "Signal.in",RNG);
    if(NoiseFreeSignal->GetTypePulse() == NRZ)
    {
       NoiseFreeSignalMZ = new MachZehnder(InDir + "MZ.in",NoiseFreeSignal);
       NoiseFreeSignalMZ->MZModulator();        
    } 
	 
    Noise = new OptSignal(InDir + "Noise.in",RNG);
  
  

    } 
  

 	sprintf(C_String,"SignalfoutBeforeRaman.dat",1);
	FileName = Job + C_String;
  	Signal->WriteFileWavelengthPower(FileName);
	
  sprintf(C_String,"NoiseFreeSignalfoutBeforeRaman.dat");
  FileName = Job + C_String;
  NoiseFreeSignal->WriteFileWavelengthPower(FileName); 

  sprintf(C_String,"NoisefoutBeforeRaman.dat");
  FileName = Job + C_String;
  Noise->WriteFileWavelengthPower(FileName); 
  
  


  

//  Signal->WriteFileFreqPower(Job + "FreqOutput.dat", 0); 
//  Signal->WriteFileTimePower(Job + "TimeOutput.dat");


OptFiberLocalError SMF10(InDir + "SMF10.in", Signal, RNG, Job);
OptFiberLocalError DCM(InDir + "DCM.in", Signal, RNG, Job);
OptFiberLocalError SMF25(InDir + "SMF25.in", Signal, RNG, Job);

OptFiberLocalError SMF10NFSignal(InDir + "SMF10.in", NoiseFreeSignal, RNG, Job);
OptFiberLocalError DCMNFSignal(InDir + "DCM.in", NoiseFreeSignal, RNG, Job);
OptFiberLocalError SMF25NFSignal(InDir + "SMF25.in", NoiseFreeSignal, RNG, Job);


int NumFibersPumped = 3;

RamanAmp Raman(Job, InDir + "Raman.in", RNG, Signal, NumFibersPumped);


    Raman.ReadRamanGainLossWavelengthGrid(SMF10.GetFiberLength(),InDir + "SMF28RamanGainLossLinear.in");
    Raman.ReadRamanGainLossWavelengthGrid(DCM.GetFiberLength(),InDir + "DCMRamanGainLossLinear.in"); 
    Raman.ReadRamanGainLossWavelengthGrid(SMF25.GetFiberLength(),InDir + "SMF28RamanGainLossLinear.in"); 
    Raman.AllocateMemory();
	SMF10.SetRamanAmpPointer(&Raman);
    DCM.SetRamanAmpPointer(&Raman);
    SMF25.SetRamanAmpPointer(&Raman);
	
	SMF10NFSignal.SetRamanAmpPointer(&Raman);
    DCMNFSignal.SetRamanAmpPointer(&Raman);
    SMF25NFSignal.SetRamanAmpPointer(&Raman);



//OptFiberLocalError SMF10m0WNon(InDir + "SMF10.in", SignalWNon, RNG, Job);
//OptFiberLocalError DCMm0WNon(InDir + "DCM.in", SignalWNon, RNG, Job);
//OptFiberLocalError SMF25m0WNon(InDir + "SMF25.in", SignalWNon, RNG, Job);


 
// 	SMF10WNon.SetRamanAmpPointer(RamanPtr[ii]);
 //   DCMWNon.SetRamanAmpPointer(RamanPtr[ii]);
 //   SMF25WNon.SetRamanAmpPointer(RamanPtr[ii]);
 



   	 Raman.SetTypeRamanNoise(typeRamanNoise(MyStartup->GetTypeAmplifierNoise()));
	 

   if(MyStartup->GetBeta3ZeroFlag())
     {
         SMF10.SetSndOrDispFiber(0);
	     DCM.SetSndOrDispFiber(0);
	   	 SMF25.SetSndOrDispFiber(0);
         SMF10NFSignal.SetSndOrDispFiber(0);
	     DCMNFSignal.SetSndOrDispFiber(0);
	   	 SMF25NFSignal.SetSndOrDispFiber(0);
    }

	   
	   
    if(MyStartup->GetNonLinearityZeroFlag())
     {

         SMF10.SetGammaFiber(0);
		 DCM.SetGammaFiber(0);
		 SMF25.SetGammaFiber(0);
         SMF10NFSignal.SetGammaFiber(0);
		 DCMNFSignal.SetGammaFiber(0);
		 SMF25NFSignal.SetGammaFiber(0);
     }

    if(MyStartup->GetFiberLossZeroFlag()) {
 		    SMF10.SetAttenuationdB(0);
			DCM.SetAttenuationdB(0);
			SMF25.SetAttenuationdB(0);
		    SMF10NFSignal.SetAttenuationdB(0);
			DCMNFSignal.SetAttenuationdB(0);
			SMF25NFSignal.SetAttenuationdB(0);
		
	    SMF10.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
	  	DCM.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
	  	SMF25.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
	    SMF10NFSignal.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
	  	DCMNFSignal.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
	  	SMF25NFSignal.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());

          SMF10.SetWritePulseWidthFlag(MyStartup->GetWritePulseWidthFlag());
          DCM.SetWritePulseWidthFlag(MyStartup->GetWritePulseWidthFlag());	  
          SMF25.SetWritePulseWidthFlag(MyStartup->GetWritePulseWidthFlag());   
	      
		  SMF10.SetWriteAveragePowerdBmFlag(MyStartup->GetWriteAveragePowerdBmFlag());
	      DCM.SetWriteAveragePowerdBmFlag(MyStartup->GetWriteAveragePowerdBmFlag());
	      SMF25.SetWriteAveragePowerdBmFlag(MyStartup->GetWriteAveragePowerdBmFlag());
          
		  SMF10.SetOutputStepSizesFlag(MyStartup->GetOutputStepSizesFlag());
          DCM.SetOutputStepSizesFlag(MyStartup->GetOutputStepSizesFlag());
          SMF25.SetOutputStepSizesFlag(MyStartup->GetOutputStepSizesFlag());

 
	}


 
	    
	cout.precision(10);  
     double PropagatedLengthAttenuateFiber = 0.0;
	int ZStepNumAttenuateFiber = 0;
  
 	





double RamanStartLength;
double PropagatedLength = 0;
int ZStepNum = 0;
double PropagatedLengthNFSignal = 0;
int ZStepNumNFSignal = 0;
int ZStepNumWNon = 0;




 	    	Raman.SetPumpPower(1);
	
	        cout << "Power = " <<
		
Linear2dBm(Signal->GetAveragePower())<< " dBm" << endl;

	
   
    PropagatedLength = 0;
    ZStepNum = 0; 
    PropagatedLengthNFSignal = 0;
    ZStepNumNFSignal = 0; 	

		RamanStartLength = PropagatedLength;

		Raman.RamanIteration();
        SMF10.PropagateWithRaman(&PropagatedLength,&ZStepNum,RamanStartLength);
        SMF10NFSignal.PropagateWithRaman(&PropagatedLengthNFSignal,&ZStepNumNFSignal,RamanStartLength);

 
        DCM.PropagateWithRaman(&PropagatedLength,&ZStepNum,RamanStartLength);
        DCMNFSignal.PropagateWithRaman(&PropagatedLengthNFSignal,&ZStepNumNFSignal,RamanStartLength);

        SMF25.PropagateWithRaman(&PropagatedLength,&ZStepNum,RamanStartLength);
        SMF25NFSignal.PropagateWithRaman(&PropagatedLengthNFSignal,&ZStepNumNFSignal,RamanStartLength);

	   Raman.AddNoise();
 	   Raman.PropagateNoise(Noise);

		cout << "Power = " <<
		Linear2dBm(Signal->GetAveragePower())<< " dBm" << endl;

//	   DataWriter.WriteEndMap(expt_num,kk,PropagatedLength);
	   
		sprintf(C_String,"SignalfoutAfterRaman.dat",1);
	FileName = Job + C_String;
  	Signal->WriteFileWavelengthPower(FileName);

  sprintf(C_String,"NoiseFreeSignalfoutAfterRaman.dat");
  FileName = Job + C_String;
  NoiseFreeSignal->WriteFileWavelengthPower(FileName); 

  sprintf(C_String,"NoisefoutAfterRaman.dat");
  FileName = Job + C_String;
  Noise->WriteFileWavelengthPower(FileName); 
	
 
  delete RNG;
 
  cout << "THE END" << endl;
  *LogFile << "THE END" << endl;

 //exit(1);	


} // ## end RamanBasic


