// $Id: ReduExampleApplications.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $
// #####################################################################
//
//    ocsAppLima.cc
//
//    Simulation Module of the ocs system
//    Optical Communication Systems Simulator
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################

#include <iostream.h>
#include <complex.h>

#include "ReduMain.hh"
#include "ReduStartup.hh"

// ## Ivan Lima's Applications

void GetReduRelativeQfactor_pdf(MyStartup *oMyStartup);
void GetReduQfactor_pdf(MyStartup *oMyStartup);
void GetReduQfactorWithPeriodicity_pdf(MyStartup *oMyStartup);
void SweepReduPDL(MyStartup *oMyStartup);

void GetReduOutageProb(MyStartup *oMyStartup);
// void GetDGDpdf(MyStartup *oMyStartup);

void TestPDLandPDG(MyStartup *oMyStartup);

//void Merge_pdfs(int Convert2dB_Scale);
void MergeConditional_pdfs(int Convert2dB_Scale);
void MergeWeighting_pdfs(void);
void MergeSelectMaxRelNumHits_pdfs(int Convert2dB_Scale);

void GetSignalFreeNoiseEvolution(MyStartup *oMyStartup); 
void Generate_pdf(MyStartup *oMyStartup);

void GetOutageProbFrom_cdfc(string InFileName_cdfc);

//extern InputData in;  // ## global object
extern ofstream *LogFile;

// #### The optical pulses must be local globals 


//! This function is called from main.cc plays the role of the main 
//! program

/*!

   The function reads in the type of simulation from the file
   StartupLima.in and then calls the appropriate function either
   in ReducedModel.cc or in this file.

   If the program is run from the command line with an argument list
   then the function reads in histogram .dat files whose names
   have been specified in the argument list and  computes the statistics
   of these pdfs, namely min, max, mean and std deviation. 

   The reason you might want to do this is that the function
   void MergeSelectMaxRelNumHits_pdfs(int Convert2dB_Scale)
   in ocsAppLima.cc generates merged pdf's but because of the way
   the merging is done the stats of the pdfs must be computed 
   by running the program again as just described in the previous 
   paragragh. I guess we could streamline all of this....

 */


void MyApplication(int argc, char *argv[])
{       

   if (argc == 2 || argc == 4) 
     {
      //WriteFileHistogram(char *argv[]);
      Histogram oHist_pdf("Hist_pdf.in");
      cout << endl;      
      if (argc == 2) {
         oHist_pdf.ImportData(argv[1]);
         cout << "Histogram of the data from the file "
              << argv[1] << " was generated in Hist_pdf.dat." << endl;         
      }   
      else if (argc == 4) 
       {
         oHist_pdf.ImportData(argv[1], atoi(argv[2]),atoi(argv[3]));
         cout << "Histogram of the column # " << argv[2] <<" out of "
              << argv[3] <<endl
              << "from the file "<< argv[1] 
              << " was generated in Hist_pdf.dat." << endl;         
      }       
 
      cout << "Min  sample   = " << oHist_pdf.GetMinSample() << endl;
      cout << "Max  sample   = " << oHist_pdf.GetMaxSample() << endl;      
      cout << "Mean sample   = " << oHist_pdf.GetMean() << endl;      
      cout << "StdDev sample = " << oHist_pdf.GetStdDev() << endl;      
      oHist_pdf.WriteFile_pdf("Hist_pdf.dat");
      exit(1);
   }    
        
   MyStartup oMyStartup("StartupLima.in", argc, argv);
       
   cout << "Simulation Type  "<< oMyStartup.GetSimType()  
        << endl;
   
   switch(oMyStartup.GetSimType()) 
    {

      //case  2: GetDGDpdf(&oMyStartup); break;         

      case 30: GetReduQfactor_pdf(&oMyStartup); break;  
      case 31: GetReduRelativeQfactor_pdf(&oMyStartup); break;  
      case 32: SweepReduPDL(&oMyStartup); break;        
      case 33: GetReduOutageProb(&oMyStartup); break;
      case 34: GetSignalFreeNoiseEvolution(&oMyStartup); break; 
      case 35: TestPDLandPDG(&oMyStartup); break;      
      case 36: GetReduQfactorWithPeriodicity_pdf(&oMyStartup); break;        

	//  case 70: Merge_pdfs(0); break;
      case 71: MergeConditional_pdfs(0); break;      
      case 72: MergeWeighting_pdfs(); break;
      case 73: MergeSelectMaxRelNumHits_pdfs(0); break;
      case 74: MergeSelectMaxRelNumHits_pdfs(-1); break;
      case 75: MergeConditional_pdfs(-1); break;  
	//      case 76: Merge_pdfs(-1); break;             

      case 111: Generate_pdf(&oMyStartup); break;

      default: cout << endl << "SimType: " << oMyStartup.GetSimType() 
                    << " inexistent." << endl << endl;
   }   
}

// ###################################################################

//### Merge pdfs. It is useful for merging multiple simulations
//## By: Ivan Lima


/***********************

void Merge_pdfs(int Convert2dB_Scale)
{
   Histogram oHist_pdf("Hist_pdf.in");
   oHist_pdf.MultipleImport("Merge_pdfs.in",Convert2dB_Scale);   
   //oHist_pdf.MultipleSelectiveImport("Merge_pdfs.in");
   oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   oHist_pdf.WriteFile_cdf("Merged_cdf.dat");    
   oHist_pdf.WriteFile_cdfc("Merged_cdfc.dat");      


   oHist_pdf.WriteFileQtSamplesBin("MergedQtSamplesBin.dat");  

   oHist_pdf.WriteFileGaussian_pdf("Merged_pdfGaussian.dat",1000); 
   oHist_pdf.WriteFileGaussian_cdf("Merged_cdfGaussian.dat",1000);      
   oHist_pdf.WriteFileGaussian_cdfc("Merged_cdfcGaussian.dat",1000); 

   cout << "qtSamples = " << oHist_pdf.GetQtSamples() << endl;
   cout << "The merged pdf of the files contained in Merge_pdfs.in"
        << endl << "were succesfully created in Merged_pdf.dat." 
        << endl << "The pdf was obtained based upon Hist_pdf.in."
        << endl;
}

*********************/

// ##################################################################

// ### Merge pdfs. It is useful for importance sampling
// #   This function builds a pdf of a random parameter 
// #   selectivelly based on another parameter, which 
// #   was the one biased. It is assumed that there is 
// #   a substantial correlation between the two  
// #   quantities, but there are other parameters 
// #   that determine the dependencie.

// #####################################################################

//! NOT USED I THINK

void MergeConditional_pdfs(int Convert2dB_Scale)
{
   Histogram oHist_pdf("Hist_pdf.in");
   oHist_pdf.ConditionalMultipleSelectiveImport("MergeCond_pdfs.in",
                                                Convert2dB_Scale);
   oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   oHist_pdf.WriteFile_cdfc("Merged_cdfc.dat");      
   oHist_pdf.WriteFileQtSamplesBin("MergedQtSamplesBin.dat");      

   if (Convert2dB_Scale) {
      GetOutageProbFrom_cdfc("Merged_cdfc.dat");
   }   

   cout << "Mean   = " << oHist_pdf.GetMean() << endl;
   cout << "StdDev = " << oHist_pdf.GetStdDev() << endl;
   cout << "Min    = " << oHist_pdf.GetMinSample() << endl;   
   cout << "Max    = " << oHist_pdf.GetMaxSample() << endl;   
   cout << "qtSamples = " << oHist_pdf.GetQtSamples() << endl;
   cout << "SumLikelihoodRatios = " << oHist_pdf.GetSumLikelihoodRatio() << endl; 
   cout << "SumProbabilities = " << oHist_pdf.GetSumProbabilities() << endl;   
   cout << "The merged pdf of the files contained in Merge_pdfs.in"
        << endl << "were succesfully created in Merged_pdf.dat." 
        << endl << "The pdf was obtained based upon Hist_pdf.in."
        << endl;
}

// ###################################################################

//#############################################################
// ### Merge pdfs. It is useful for importance sampling
// #   This function builds a pdf of a random parameter 
// #   selectivelly based on another parameter, which 
// #   was the one biased. It is assumed that there is 
// #   a substantial correlation between the two  
// #   quantities, but there are other parameters 
// #   that determine the dependencie.

//! NOT USED I THINK

void MergeWeighting_pdfs(void)
{
   Histogram oHist_pdf("Hist_pdf.in");
   oHist_pdf.MultipleWeightedImport("Merge_pdfs.in");
   oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   //cout << "Mean   = " << oHist_pdf.GetMean() << endl;
   //cout << "StdDev = " << oHist_pdf.GetStdDev() << endl;
   //cout << "Min    = " << oHist_pdf.GetMinSample() << endl;   
   //cout << "Max    = " << oHist_pdf.GetMaxSample() << endl;   
   //cout << "qtSamples = " << oHist_pdf.GetQtSamples() << endl;
   //cout << "SumLikelihoodRatios = " << oHist_pdf.GetSumLikelihoodRatio() << endl; 
   cout << "SumProbabilities = " << oHist_pdf.GetSumProbabilities() << endl;
   cout << "The merged pdf of the files contained in Merge_pdfs.in"
        << endl << "were succesfully created in Merged_pdf.dat." 
        << endl << "The pdf was obtained based upon Hist_pdf.in."
        << endl;
}




//#############################################################

// ### Merge pdfs. It is useful for importance sampling

//! This method uses Ivan Lima's heuristic for merging biased
//! pdfs obtained using importanc sampling.

/*!

 \param Convert2dB_Scale: Flag that specifies whether or not
            to convert the pdfs to be merged
            from a linear to dB scale.
      
The method does the following.

-# It constructs a Histogram object for the merged pdf by reading
   in parameters from the input parameter file Hist_pdf.in

-# It calls Histogram::MultipleSelectMaxRelNumHitsImport and passes
   it the file name Megre_pdfs.in and the flag Convert2dB_Scale.

-# It outputs data as follows:
 -# Merged_pdf.dat is obtained by calling Histogram::WriteFile_pdf
 -# Merged_cdf.dat is obtained by calling Histogram::WriteFile_cdf
 -# Merged_cdfc.dat is obtained by calling Histogram::WriteFile_cdfc
    if Convert2dB_Scale == 0 and by calling GetOutageProbFrom_cdfc
    if Convert2dB_Scale == 1.


The merged pdf's are written to the files 


\sa Histogram::MultipleSelectMaxRelNumHitsImport

 */

void MergeSelectMaxRelNumHits_pdfs(int Convert2dB_Scale)
{
   Histogram oHist_pdf("Hist_pdf.in");
   oHist_pdf.MultipleSelectMaxRelNumHitsImport("Merge_pdfs.in",
                                                Convert2dB_Scale);

   oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   oHist_pdf.WriteFile_cdf("Merged_cdf.dat");   
   oHist_pdf.WriteFile_cdfc("Merged_cdfc.dat");      

   if (Convert2dB_Scale) 
    {
      GetOutageProbFrom_cdfc("Merged_cdfc.dat");
    }   

   cout << endl << "SumProbabilities = " 
        << oHist_pdf.GetSumProbabilities() << endl;
   cout << "The merged pdf, cdf and the cdfc of the files contained in"
        << "Merge_pdfs.in were succesfully created in Merged_pdf.dat," 
        << endl << "Merged_cdf.dat and Merged_cdfc.dat, respectivelly."
        << endl << "These functions were obtained based upon Hist_pdf.in."
        << endl;
        
} 

// ###########################################################


//## Get the outage probability at specified levels

void GetOutageProbFrom_cdfc(string InFileName_cdfc)
{

   cout << endl << "Obtaining the outage probabilities: "<< endl;
   //string InFileName_cdfc = "Merged_cdfc.dat";
   ifstream In_cdfc(InFileName_cdfc.c_str(),ios::in);
   if ( ! In_cdfc) {
      cerr << "Error: input file " <<InFileName_cdfc 
           << " could not be opened."<< endl;
      exit(1);
   }
   
   int SizeVectors = 1500;
   double In_cdfcArray[SizeVectors][2];

   int counter = 0;   
   char cBuffer[100];
   while (In_cdfc >> cBuffer) { 
      In_cdfcArray[counter][0] = atof(cBuffer);
      In_cdfc >> cBuffer; 
      In_cdfcArray[counter][1] = atof(cBuffer);         
      In_cdfc >> cBuffer; //## Extra field  (confidence) not used (9/24/01)
      counter++;
   } 

   In_cdfc.close();

   double PenaltyLevel[] = {1, 2, 3};

   for ( int ii = 0; ii  < counter - 1; ii++) {
      //cout << In_cdfcArray[ii][0] << "  "
      //     << In_cdfcArray[ii+1][0] << endl;
      for (int jj = 0; jj < 3; jj++) {
         if (  In_cdfcArray[ii][0] < PenaltyLevel[jj]
            && In_cdfcArray[ii+1][0] > PenaltyLevel[jj] ) { //## > 3 db 
            cout << "Outage probability at "
                 << PenaltyLevel[jj] << " dB = " << 
                  (In_cdfcArray[ii+1][1] - In_cdfcArray[ii][1])
                 /(In_cdfcArray[ii+1][0] - In_cdfcArray[ii][0])
                 *( PenaltyLevel[jj] - In_cdfcArray[ii][0]) 
                 + In_cdfcArray[ii][1] << endl;
         }                  
      }
   }

}


/*
//#############################################################
// ### Merge pdfs. It is useful for importance sampling

void ImportGaussian_pdf(void)
{
   Histogram oHist_pdf("Hist_pdf.in");
   //oHist_pdf.MultipleSelectMaxRelNumHitsImport("Merge_pdfs.in",Convert2dB_Scale);
   //oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   //oHist_pdf.WriteFile_cdf("Merged_cdf.dat");   
   //oHist_pdf.WriteFile_cdfc("Merged_cdfc.dat");      
   //cout << "Mean   = " << oHist_pdf.GetMean() << endl;
   //cout << "StdDev = " << oHist_pdf.GetStdDev() << endl;
   //cout << "Min    = " << oHist_pdf.GetMinSample() << endl;   
   //cout << "Max    = " << oHist_pdf.GetMaxSample() << endl;   
   //cout << "qtSamples = " << oHist_pdf.GetQtSamples() << endl;
   //cout << "SumLikelihoodRatios = " << oHist_pdf.GetSumLikelihoodRatio() << endl; 

   string InFileName = "Gaussian_pdf.dat";
   ifstream In_pdf(InFileName.c_str(),ios::in);
   if ( ! In_pdf) {
      cerr << "Error: input file " <<InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }
   
   int SizeVectors = 1500;
   double In_pdfArray[SizeVectors][2];

   int counter = 0;   
   char cBuffer[100];
   while (In_pdf >> cBuffer) { 
      In_pdfArray[counter][0] = atof(cBuffer);
      In_pdf >> cBuffer; 
      In_pdfArray[counter][1] = atof(cBuffer);         
      //In_pdf >> cBuffer; //## Extra field  (confidence) not used (9/24/01)
      
      oHist_pdf.AddImportantSample(-Linear2dB(In_pdfArray[counter][0]),
            In_pdfArray[counter][0]*In_pdfArray[counter][1]*log(10.)/10. );

      //       In_pdfArray[counter][1] );


      
      counter++;
   } 

   In_pdf.close();

   oHist_pdf.WriteFile_pdf("Merged_pdf.dat");   
   oHist_pdf.WriteFile_cdf("Merged_cdf.dat");   
   oHist_pdf.WriteFile_cdfc("Merged_cdfc.dat");  

   cout << endl << "Obtaining the outage probabilities: "<< endl;
   InFileName = "Merged_cdfc.dat";
   ifstream In_cdfc(InFileName.c_str(),ios::in);
   if ( ! In_cdfc) {
      cerr << "Error: input file " <<InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }
   
   //int SizeVectors = 1500;
   //double In_pdfArray[SizeVectors][2];

   counter = 0;   
   //char cBuffer[100];
   while (In_cdfc >> cBuffer) { 
      In_pdfArray[counter][0] = atof(cBuffer);
      In_cdfc >> cBuffer; 
      In_pdfArray[counter][1] = atof(cBuffer);         
      In_cdfc >> cBuffer; //## Extra field  (confidence) not used (9/24/01)
      
      counter++;
   } 

   In_cdfc.close();

   double PenaltyLevel[] = {1, 2, 3};

   for ( int ii = 0; ii  < counter - 1; ii++) {
      //cout << In_pdfArray[ii][0] << "  "
      //     << In_pdfArray[ii+1][0] << endl;
      for (int jj = 0; jj < 3; jj++) {
         if (  In_pdfArray[ii][0] < PenaltyLevel[jj]
            && In_pdfArray[ii+1][0] > PenaltyLevel[jj] ) { //## > 3 db 
            cout << "Outage probability at "
                 << PenaltyLevel[jj] << " dB = " << 
                  (In_pdfArray[ii+1][1] - In_pdfArray[ii][1])
                 /(In_pdfArray[ii+1][0] - In_pdfArray[ii][0])
                 *( PenaltyLevel[jj] - In_pdfArray[ii][0]) 
                 + In_pdfArray[ii][1] << endl;
         }                  
      }
   }


   cout << endl << "SumProbabilities = " 
        << oHist_pdf.GetSumProbabilities() << endl;
   cout << "The merged pdf, cdf and the cdfc of the files contained in"
        << "Merge_pdfs.in were succesfully created in Merged_pdf.dat," 
        << endl << "Merged_cdf.dat and Merted_cdfc.dat, respectivelly."
        << endl << "These functions were obtained based upon Hist_pdf.in."
        << endl;
        
}
*/


//#############################################################
//### Generate the pdf of the desired random variable
//##  to test the random number generator.

void Generate_pdf(MyStartup *oMyStartup)
{
   RanNumGen *oRanNumGen = new RanNumGen("RanNumGen.in");  
   Histogram *oHist_pdf  = new Histogram("Hist_pdf.in");
   
   double LikelihoodRatio;
   
   double ValFunction;
   
   double alpha = oMyStartup->GetDoubleParam(1);
   double beta  = oMyStartup->GetDoubleParam(2);

   int qtStatSamples = oMyStartup->GetQtStatSamples();
   
   for (int ii = 0; ii < qtStatSamples; ii++) {
      ValFunction = oRanNumGen->GetRanCosineBiased(alpha
                                                   ,&LikelihoodRatio);
      oHist_pdf->AddSample(ValFunction);   
      //oHist_pdf->AddImportantSample(ValFunction,LikelihoodRatio);
   }
   oHist_pdf->WriteFile_pdf("Hist_pdf.dat");   
   cout << "Mean   = " << oHist_pdf->GetMean() << endl;
   cout << "StdDev = " << oHist_pdf->GetStdDev() << endl;
   cout << "Min    = " << oHist_pdf->GetMinSample() << endl;   
   cout << "Max    = " << oHist_pdf->GetMaxSample() << endl;   
   cout << "qtSamples = " << oHist_pdf->GetQtSamples() << endl;
   //cout << "SumLikelihoodRatios = " << oHist_pdf.GetSumLikelihoodRatio() << endl; 
   //cout << "SumProbabilities = " << oHist_pdf.GetSumProbabilities() << endl;
   cout << "The pdf was obtained based upon Hist_pdf.in "
        << "and the function coded in." << endl;
        
   delete oRanNumGen, oHist_pdf;        
}

//#############################################################

/***********************

// ### Plot the pdf of desired function
void Generate_pdf(MyStartup *oMyStartup)
{
   RanNumGen *oRanNumGen = new RanNumGen("RanNumGen.in");  
   Histogram *oHist_pdf  = new Histogram("Hist_pdf.in");
   
   
   double ValFunction, Z, W;
   
   double alpha = oMyStartup->GetDoubleParam(1);
   double beta  = oMyStartup->GetDoubleParam(2);

   if ( abs(alpha) < 1e-5 ) {
      //## Prevent numerical problems
      alpha = 1e-5;
   }     
   
   if (beta < -1) {
      beta = -1;
   }
   else if (beta > 1) {
      beta = 1;
   }
   
   int qtStatSamples = oMyStartup->GetQtStatSamples();
   
   for (int ii = 0; ii < qtStatSamples; ii++) {
      Z = oRanNumGen->GetRanNum();      
      W = oRanNumGen->GetRanNum();
      if (W < (1-exp(-alpha*(beta+1)))
             /(2-exp(-alpha*(beta+1))-exp(-alpha*(1-beta)) ) ) {
         ValFunction = beta + 1./alpha * log( exp(-alpha*(beta+1)) 
                       +Z*(1-exp(-alpha*(beta+1)) ) );   //## A
      }                 
      else {                 
         ValFunction = beta - 1./alpha * log( 1 
                       -Z*(1-exp(-alpha*(1-beta)) ) );    //## B                   
      }                 
      oHist_pdf->AddSample(ValFunction);   
   }
   oHist_pdf->WriteFile_pdf("Hist_pdf.dat");   
   cout << "Mean   = " << oHist_pdf->GetMean() << endl;
   cout << "StdDev = " << oHist_pdf->GetStdDev() << endl;
   cout << "Min    = " << oHist_pdf->GetMinSample() << endl;   
   cout << "Max    = " << oHist_pdf->GetMaxSample() << endl;   
   cout << "qtSamples = " << oHist_pdf->GetQtSamples() << endl;
   //cout << "SumLikelihoodRatios = " << oHist_pdf.GetSumLikelihoodRatio() << endl; 
   //cout << "SumProbabilities = " << oHist_pdf.GetSumProbabilities() << endl;
   cout << "The pdf was obtained based upon Hist_pdf.in "
        << "and the function coded in." << endl;
        
   delete oRanNumGen, oHist_pdf;        
}


******************/




