// $Id: 
// #####################################################################
//
//    ocsStartupLima.hh
//
//    Startup the ocs system for the user Ivan Lima
//    Optical Communication Systems Simulator
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################
//  ## Adapted from Zweck's startup

# include "ReduStartup.hh"

extern ofstream *LogFile;

// ## Constructor
// #  By: Ivan T. Lima Jr.


//! This class reads in parameters for your particular application
//! from the file named InFileName.

/*!

This class could be made more user frienly, but I have not changed
it because I want you to be able to test the code by rerunning 
the jobs Ivan already ran when he validated the code.

*/

MyStartup::MyStartup(string InFileName, int argc, char *argv[])
{
   ifstream InFile(InFileName.c_str(),ios::in);       
   if (!InFile) {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
   }

   LogFileName = "ocs.log";
   cout << endl
         << "Log File = " << LogFileName << endl << endl;
   LogFile->open(&LogFileName[0],ios::out);
   if (!LogFile) {
      cerr << "Error: ocs.log could not be open." << endl << flush;
      exit(1);
   }    

   // ##########################
   // ## Start of personal stuff
   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;

   ProcessControlMode = ReadInt("ProcessControlMode"," ",0,-1,LOWER_ONLY,
                                OPTIONAL_NO_WARNING,&InFile);

   string ProcessControlFileName = "ProcessControl/ProcessControl.txt";
   oProcessControl.InitProcessControl(ProcessControlMode
                                      ,ProcessControlFileName
                                      ,argv[0]);      
                                 
   SimType = ReadInt("SimType"," ",0,0,NO_BOUNDS,
                     MANDATORY,&InFile);

   qtStatSamples = ReadInt("qtStatSamples"," ",0,-1,LOWER_ONLY,
                            MANDATORY,&InFile);
                      
   qtMapSpans = ReadInt("qtMapSpans"," ",0,-1,LOWER_ONLY,
                        OPTIONAL_NO_WARNING,&InFile);

    char FieldName[30];                        
   // ## I should create a loop here from 0 to 9 for the Int and the DOuble
   for (int ii = 0; ii < 10; ii++) {
      sprintf(FieldName,"IntParam%d",ii);   
      IntParam[ii] = ReadInt(FieldName," ",0,0,NO_BOUNDS,
                              OPTIONAL_NO_WARNING,&InFile);
   }
   for (int ii = 0; ii < 10; ii++) {   
      sprintf(FieldName,"DoubleParam%d",ii);      
      DoubleParam[ii] = ReadDouble(FieldName," ",0,0,NO_BOUNDS,
                        OPTIONAL_NO_WARNING,&InFile);
   }   
   InFile.close();
}


// #####################################################################
