#include "ocs.hh"
#include "StartupSechLaser.hh"

void SechLaser(StartupSechLaser *MyStartup);

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[])
{

 StartupSechLaser MyStartup("Startup.in");

 SechLaser(&MyStartup);

 cout << "THE END" << endl;


} // ## end MyApplication

// ########################################################

void SechLaser(StartupSechLaser *MyStartup)
{

 string Job = MyStartup->GetJobPrefix();
 
 RanNumGen *RNG;
 OptSignal * Signal;

 RNG =  new RanNumGen("./RanNumGen.in");
 Signal = new OptSignal("./Signal.in", RNG);

 OptDataWriter DataWriter("./DataWriter.in","./",Job,Signal);
 DataWriter.WriteInitialData();

 OptFiberLocalError HausMasterEquationFiber("./HausMasterEquationFiber.in",
                                              Signal,RNG,Job);


 HausMasterEquationFiber.SetNepperAttenuation(
                           MyStartup->GetAttenuationLinear());

 HausMasterEquationFiber.SetFiberAmplifierUnsaturatedNepperGain(
                           MyStartup->GetUnsaturatedLinearGain());

 HausMasterEquationFiber.SetFstOrDispFiber(MyStartup->GetFstOrDisp());
 HausMasterEquationFiber.SetGammaFiber(MyStartup->GetGamma());

 HausMasterEquationFiber.SetSndOrDispFiber(0); 

 cout << "Input Average Power [W] = " << Signal->GetAveragePower() << endl;

 cout << "Input Pulse Energy [J] = " << Signal->GetTotalEnergy() << endl;

 cout << "Input Max Amplitude [W^{1/2}] = "
      << sqrt(Signal->ComputeMaxPower()) << endl;

 cout << "Input Pulse Width (std dev) [s] = "
      << sqrt(Signal->GetSigmaSqOptSignal()) << endl;

 double PropagatedLength = 0.0;
 int ZStepNum = 0;

 HausMasterEquationFiber.Propagate(&PropagatedLength,&ZStepNum);

 cout  << "Output Average Power [W] = "
       << Signal->GetAveragePower() << endl;

 cout  << "Output Pulse Energy [J] = "
       << Signal->GetTotalEnergy() << endl;

 cout << "Output Max Amplitude [W^{1/2}] = "
      << sqrt(Signal->ComputeMaxPower()) << endl;

 cout << "Output Pulse Width (std dev) [s] = "
      << sqrt(Signal->GetSigmaSqOptSignal()) << endl;

 DataWriter.WriteEndTransmission(0,1,PropagatedLength);
 
 Signal->~OptSignal();
 RNG->~RanNumGen();

} // ## end SechLaser

