
#include "ocs.hh"
#include "StartupSimilariton.hh"

void SimilaritonFiberAmplifier(StartupSimilariton *MyStartup);
void TestFiberAmplifierNoise(StartupSimilariton *MyStartup);
void CompensateFiberAmp(StartupSimilariton *MyStartup);
void FermannLaserOct09(StartupSimilariton *MyStartup);
void SimilaritonFiberAmplifierDispMatrix(StartupSimilariton *MyStartup);

extern ofstream *LogFile;

void MyApplication(int argc, char *argv[])
{

 StartupSimilariton MyStartup("Startup.in");

  switch(MyStartup.GetSimType())
   {

    case 1:
      SimilaritonFiberAmplifier(&MyStartup);
      break;

    case 2:
      TestFiberAmplifierNoise(&MyStartup);
      break;

    case 3: 
      CompensateFiberAmp(&MyStartup);
      break;

    case 4:
      FermannLaserOct09(&MyStartup);
      break;

    case 5: SimilaritonFiberAmplifierDispMatrix(&MyStartup);
      break;

    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);

   } // end switch

   cout << "THE END" << endl;


} // ## end MyApplication

// ################################################################

void FermannLaserOct09(StartupSimilariton *MyStartup)
{

 string Job = MyStartup->GetJobPrefix();

 RanNumGen *RNG;
 OptSignal * Signal;

 RNG =  new RanNumGen("./RanNumGen.in");
 Signal = new OptSignal("./Signal.in", RNG);

 int OSAAppendFlag = 0;

 OptSpectrumAnalyzer OSA("OSA.in",Job,OSAAppendFlag,Signal);

 OptDataWriter DataWriter("./DataWriter.in","./",Job,Signal);
 DataWriter.WriteInitialData();

 OptFiberLocalError FiberAmplifier("./FiberAmplifier.in",Signal,RNG,Job);
 OptFiberLocalError DCF("./DCF.in",Signal,RNG,Job);
 SaturableAbsorber  SatAb("./SatAbs.in",Signal);
 LumpedLoss LossSatAbs("./LossSatAbs.in",Signal);

int NumFibers=1;

if(FiberAmplifier.GetFiberLength() > 0)
 {
  NumFibers =
    int(round(MyStartup->GetTotalFiberLength()/
         double(FiberAmplifier.GetFiberLength())));
 }

 int NumLaserRoundTrips = MyStartup->GetNumLaserRoundTrips();

 if(MyStartup->GetTotalFiberLength() == FiberAmplifier.GetFiberLength())
    NumFibers = 1;

 cout << "NumFibers = " << NumFibers << endl;

 cout << "TotalFiberLength (Must agree) = "
      << MyStartup->GetTotalFiberLength() << " "
      << NumFibers*FiberAmplifier.GetFiberLength() << endl;

 cout << "Round Trip Dispersion [fs^2] = "
      << 1e30*(MyStartup->GetTotalFiberLength()*
         2*FiberAmplifier.GetFirstOrderDispersion()
         + DCF.GetFiberLength()*DCF.GetFirstOrderDispersion()) << endl;


 cout << "Input Pulse Energy [pJ] = " << 1e12*Signal->GetTotalEnergy() << endl;

 cout << "Input Max Amplitude [W^{1/2}] = "
      << sqrt(Signal->ComputeMaxPower()) << endl;

 cout << "Input Pulse Width (std dev) [fs] = "
      << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << endl;


 double PropagatedLength = 0.0;
 int ZStepNum = 0;

 string RMSPulseWidthsFileName = Job + "RMSPulseWidths.dat";

 ofstream RMSPulseWidthsFile(RMSPulseWidthsFileName.c_str(),ios::out);

 RMSPulseWidthsFile
 << "% (1) PropagatedLength (2) RMS Pulse Width in Time [fs]"
 <<  " (3) RMS Pulse Width in Freq [Thz]"
 << endl;


 string FWHMFileName = Job + "FWHM.dat";

 ofstream FWHMFile(FWHMFileName.c_str(),ios::out);

 FWHMFile << "% (1) PropagatedLength (2) FWHM of Power in time [fs]"
            << endl;

 string EnergyFileName = Job + "Energy.dat";

 ofstream EnergyFile(EnergyFileName.c_str(),ios::out);

 EnergyFile << "% (1) PropagatedLength (2) Energy [pJ]"
            << endl;


/*
 string LocFreqFileNamePrefix = Job + "LocFreq"; 

 string LocFreqFileName;
*/

for(int MCIndex = 0; MCIndex < MyStartup->GetNumMonteCarloSamples(); MCIndex++)
{
 PropagatedLength = 0.0;
 ZStepNum = 0;

  if(MCIndex == 0)
     {

       RMSPulseWidthsFile << PropagatedLength << "\t"
            << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << "\t"
            << 1e-12*Signal->ComputeRMSSpectralWidth()
            << endl;

      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

       FWHMFile << PropagatedLength << "\t"
            << 1e15*Signal->ComputeFWHMPulsePower()
            << endl;


//std::stringstream stm;
//stm << 0;

 //     LocFreqFileName = LocFreqFileNamePrefix + stm.str();

//      cout << "LocFreqFileName = " << LocFreqFileNamePrefix << endl; 

//      Signal->WriteLocalFrequencyAndChirp(LocFreqFileName);

     }



 for(int rr=1;rr<=NumLaserRoundTrips;rr++)
 {

  for(int ii=0;ii<NumFibers;ii++)
    {

     FiberAmplifier.Propagate(&PropagatedLength,&ZStepNum);

     if(MCIndex == 0)
      {

       RMSPulseWidthsFile << PropagatedLength << "\t"
            << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << "\t"
            << 1e-12*Signal->ComputeRMSSpectralWidth()
            << endl;


      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;
     }
    }

    SatAb.Apply(&PropagatedLength);

     if(MCIndex == 0)
      {

      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

     }

    LossSatAbs.ApplyLumpedLoss();

     if(MCIndex == 0)
      {

       RMSPulseWidthsFile << PropagatedLength << "\t"
            << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << "\t"
            << 1e-12*Signal->ComputeRMSSpectralWidth()
            << endl;


      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;
     }

  for(int ii=0;ii<NumFibers;ii++)
   {

    FiberAmplifier.Propagate(&PropagatedLength,&ZStepNum);

    if(MCIndex == 0)
     {

       RMSPulseWidthsFile << PropagatedLength << "\t"
            << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << "\t"
            << 1e-12*Signal->ComputeRMSSpectralWidth()
            << endl;


      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;
    }

   } // end NumFibers loop

  DataWriter.WriteWithinMap(0,rr,PropagatedLength);

  DCF.Propagate(&PropagatedLength,&ZStepNum);

      if(MCIndex == 0)
       {
         DataWriter.WriteEndMap(0,rr,PropagatedLength);

         RMSPulseWidthsFile << PropagatedLength << "\t"
            << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << "\t"
            << 1e-12*Signal->ComputeRMSSpectralWidth()
            << endl;


         EnergyFile << PropagatedLength << "\t"
                   << 1e12*Signal->GetTotalEnergy() << endl;

         FWHMFile << PropagatedLength << "\t"
                  << 1e15*Signal->ComputeFWHMPulsePower()
                  << endl;

          cout  << "Round trip = " << rr << endl;


         }

 } // end NumLaserRoundTrips loop

 if(MCIndex == 0)
  {
    DataWriter.WriteEndTransmission(0,NumLaserRoundTrips,PropagatedLength);


    RMSPulseWidthsFile.close();
    EnergyFile.close();
    FWHMFile.close();


    if(MyStartup->GetWriteComplexFieldsFlag())
     Signal->WriteFileComplexFieldsTimeAndFreq(Job + "tComplexFields.dat");

  }

 //OSA.Update(PropagatedLength);

 Signal->RegenerateSignalString();


 //if((MCIndex+1)%50 == 0)
 // OSA.Output(MCIndex+1);

  if((MCIndex+1)%10 == 0)
    cout << "NumSamples Completed = " << MCIndex+1 << endl;


} // end for MCIndex


 //OSA.Output(MyStartup->GetNumMonteCarloSamples());

 Signal->~OptSignal();
 RNG->~RanNumGen();

} // ## end FermannLaserOct09


// ############################################################

void SimilaritonFiberAmplifier(StartupSimilariton *MyStartup)
{

 // If NumRoundTripsLaser = 1 then this code models Fermann's fiber amp
 // Otherwise it models a simple laser with fiberamp followed by
 // DCF to compensate for gain and chromatic dispersion in Fiber Amp.

 string Job = MyStartup->GetJobPrefix();
 
 RanNumGen *RNG;
 OptSignal * Signal;

 RNG =  new RanNumGen("./RanNumGen.in");


if(MyStartup->GetReadSignalData())
  Signal = new OptSignal("./Signal.in", RNG,MyStartup->GetSignalFileName());
else
  Signal = new OptSignal("./Signal.in", RNG);

 int OSAAppendFlag = 0;

// OptSpectrumAnalyzer OSA("OSA.in",Job,OSAAppendFlag,Signal);

 OptDataWriter DataWriter("./DataWriter.in","./",Job,Signal);
 DataWriter.WriteInitialData();

 OptFiberLocalError FiberAmplifier("./FiberAmplifier.in",Signal,RNG,Job);
 OptFiberLocalError DCF("./DCF.in",Signal,RNG,Job);
// OptFiberLocalError FSA("./FSA.in",Signal,RNG,Job);
 SaturableAbsorber  SatAb("./SatAbs.in",Signal);

 int NumLaserRoundTrips = MyStartup->GetNumLaserRoundTrips();
 int FreqOutputStats = MyStartup->GetFreqOutputStats();

 if(NumLaserRoundTrips%FreqOutputStats)
  {
   cerr << "FreqOutputStats must be factor of NumLaserRoundTrips" << endl;
   exit(1);
  }

 int NumHists = int(NumLaserRoundTrips/FreqOutputStats);

 cout << "NumHists = " << NumHists << endl;

 Histogram * CentralTimeHist =  new Histogram[NumHists];
 Histogram * CentralFreqHist =  new Histogram[NumHists];
 Histogram * EnergyHist =  new Histogram[NumHists];
 Histogram * PhaseHist =  new Histogram[NumHists];
 Histogram * PulseWidthHist =  new Histogram[NumHists];

 for(int hh=0;hh<NumHists;hh++)
  {
    CentralTimeHist[hh].InitHistogram("./CentralTimeHist.in");
    CentralFreqHist[hh].InitHistogram("./CentralFreqHist.in");
    EnergyHist[hh].InitHistogram("./EnergyHist.in");
    PhaseHist[hh].InitHistogram("./PhaseHist.in");
    PulseWidthHist[hh].InitHistogram("./PulseWidthHist.in");
  }

int NumFibers=1;

if(FiberAmplifier.GetFiberLength() > 0)
 { 
  NumFibers =
    int(round(MyStartup->GetTotalFiberLength()/
         double(FiberAmplifier.GetFiberLength())));
 }


 if(MyStartup->GetTotalFiberLength() == FiberAmplifier.GetFiberLength())
    NumFibers = 1;

 cout << "NumFibers = " << NumFibers << endl;

 cout << "TotalFiberLength (Must agree) = "
      << MyStartup->GetTotalFiberLength() << " "
      << NumFibers*FiberAmplifier.GetFiberLength() << endl;

 cout << "Round Trip Dispersion [fs^2] = " 
      << 1e30*(MyStartup->GetTotalFiberLength()*
         FiberAmplifier.GetFirstOrderDispersion()
         + DCF.GetFiberLength()*DCF.GetFirstOrderDispersion()) << endl;

 LogFileSeparator();

 *LogFile << "FIBER AMP:" << endl;
 *LogFile << "Gamma [(mW)^{-1}] = " << FiberAmplifier.GetGamma() << endl;
 *LogFile << "FstOrDisp [s^2/m] = " 
          << FiberAmplifier.GetFirstOrderDispersion() << endl;
 *LogFile << "SndOrDisp = " 
          << FiberAmplifier.GetSecondOrderDispersion() << endl;

 *LogFile << "Nepper Attenuation = " 
          << FiberAmplifier.GetNepperAttenuationFiber() << endl;
 LogFileSeparator();

 *LogFile << "COMPENSATOR:" << endl;
 *LogFile << "TotalFstOrDisp [ps^2] = "
          << 1e24*DCF.GetFirstOrderDispersion()*DCF.GetFiberLength() << endl;

 
 *LogFile << "Input Pulse Energy [pJ] = "
          << 1e12*Signal->GetTotalEnergy() << endl;

 LogFileSeparator();

 //cout << "Input Average Power [W] = " << Signal->GetAveragePower() << endl;

 cout << "Input Pulse Energy [pJ] = " << 1e12*Signal->GetTotalEnergy() << endl;

 cout << "Input Max Amplitude [W^{1/2}] = "
      << sqrt(Signal->ComputeMaxPower()) << endl;

 //cout << "Input Pulse Width (std dev) [fs] = "
 //     << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << endl;

 double PropagatedLength = 0.0;
 int ZStepNum = 0;

 string MaxAmpFileName = Job + "MaxAmplitude.dat";

 ofstream MaxAmpFile(MaxAmpFileName.c_str(),ios::out);    

 MaxAmpFile << "% (1) PropagatedLength (2) Max Amplitude in time [W^1/2]"
            << endl;


 MaxAmpFile << PropagatedLength << "\t"
            << sqrt(Signal->ComputeMaxPower())
            << endl;


 string FWHMFileName = Job + "FWHM.dat";

 ofstream FWHMFile(FWHMFileName.c_str(),ios::out);

 FWHMFile << "% (1) PropagatedLength (2) FWHM of Power in time [fs]"
            << endl;

 FWHMFile << PropagatedLength << "\t"
            << 1e15*(Signal->ComputeFWHMPulsePower())
            << endl;

int ArgMaxPowerIndex = 0;

double MaxPower=Signal->ComputeMaxPower(ArgMaxPowerIndex);

cout << "ArgMaxPowerIndex = " << ArgMaxPowerIndex << endl;

 string ChirpFileName = Job + "Chirp.dat";

 ofstream ChirpFile(ChirpFileName.c_str(),ios::out);

 ChirpFile << "% (1) PropagatedLength (2) Chirp [THz^2]"
            << endl;

 ChirpFile << PropagatedLength << "\t"
            << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
            << endl;



/*
 string CentralTimeFileName = Job + "CentralTimeSamples.dat";

 ofstream CentralTimeFile(CentralTimeFileName.c_str(),ios::out);

 CentralTimeFile 
   << "%Row i, Col j: Central time [sec] of pulse at end jth RT, MC sample i"  
   << endl; 
*/

 string EnergyFileName = Job + "Energy.dat";

 ofstream EnergyFile(EnergyFileName.c_str(),ios::out);

 EnergyFile << "% (1) PropagatedLength (2) Energy [pJ]" 
            << endl;

string LocFreqFileNamePrefix = Job + "LocFreq";
std::stringstream rtstm;
std::stringstream fnstm;

for(int MCIndex = 0; MCIndex < MyStartup->GetNumMonteCarloSamples(); MCIndex++)
{
 PropagatedLength = 0.0;
 ZStepNum = 0;

/* CentralTimeFile << Signal->ComputeMeanTimePulse(0.0,Signal->GetTimeWindow()) 
		 << "\t";
*/

  if(MCIndex == 0)
     {

      MaxAmpFile << PropagatedLength << "\t"
                 << sqrt(Signal->ComputeMaxPower())
                 << endl;

      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

       FWHMFile << PropagatedLength << "\t"
            << 1e15*Signal->ComputeFWHMPulsePower()
            << endl;

      ChirpFile << PropagatedLength << "\t"
            << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
            << endl;

     }

 
 for(int rr=1;rr<=NumLaserRoundTrips;rr++)
 {

      //cout << "Before Saturabale Absorber" << endl;


    //FSA.Propagate(&PropagatedLength,&ZStepNum);
    SatAb.Apply(&PropagatedLength);

    //cout << "After Saturabale Absorber" << endl;

  for(int ii=0;ii<NumFibers;ii++)
   {

 if(MCIndex == 0)
     {

      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;




    }



    FiberAmplifier.Propagate(&PropagatedLength,&ZStepNum);

     if(MCIndex == 0)
     {
      DataWriter.WriteEndMap(0,(rr-1)*NumFibers + ii,PropagatedLength);

      MaxAmpFile << PropagatedLength << "\t" 
                 << sqrt(Signal->ComputeMaxPower())
                 << endl;

      EnergyFile << PropagatedLength << "\t"
		<< 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;


/*
     rtstm.clear();
     rtstm.str("");
     rtstm << rr;

     fnstm.clear();
     fnstm.str("");
     fnstm << ii;

      Signal->WriteLocalFrequencyAndChirp(
               LocFreqFileNamePrefix + "RT" + rtstm.str() + 
                                       "FA" + fnstm.str() + ".dat");
*/

     }
    } // end NumFibers loop

   if(NumLaserRoundTrips > 1)
    {

      if(MCIndex == 0)
       {

        char C_String[5];
        sprintf(C_String,"%d",rr);

        string Suffix = C_String;


        /*Signal->WriteFileTimePower(Job + "tPreDCFRT" + 
                                 Suffix + ".dat");

        Signal->WriteFileFreqPower(Job + "fPreDCFRT" +
                                 Suffix + ".dat",0);
        */
       }


 for(int ii=0;ii<NumFibers;ii++)
   {

      DCF.Propagate(&PropagatedLength,&ZStepNum);

      if(MCIndex == 0)
       {

        char C_String[5];
        sprintf(C_String,"%d",rr);

        string Suffix = C_String;
/*
        Signal->WriteFileTimePower(Job + "tPostDCFRT" +
                                 Suffix + ".dat");

        Signal->WriteFileFreqPower(Job + "fPostDCFRT" +
                                 Suffix + ".dat",0);
*/
        //DataWriter.WriteEndMap(0,rr*NumFibers,PropagatedLength);

         MaxAmpFile << PropagatedLength << "\t"
                    << sqrt(Signal->ComputeMaxPower())
                    << endl;


         EnergyFile << PropagatedLength << "\t"
                   << 1e12*Signal->GetTotalEnergy() << endl;

         FWHMFile << PropagatedLength << "\t"
                  << 1e15*Signal->ComputeFWHMPulsePower()
                  << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;

//          cout  << "Round trip = " << rr << endl;

         // cout  << "Output Average Power [W] = " 
         //	  << Signal->GetAveragePower() << endl;

         // cout  << "Output Pulse Energy [pJ] = " 
	//	<< 1e12*Signal->GetTotalEnergy() << endl;

	  //cout << "Output Max Amplitude [W^{1/2}] = "
      	  //	<< sqrt(Signal->ComputeMaxPower()) << endl;

          //cout << "Output Pulse Width (std dev) [fs] = "
            //   << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << endl;

/*
     rtstm.clear();
     rtstm.str("");
     rtstm << rr;

     fnstm.clear();
     fnstm.str("");
     fnstm << ii;

      Signal->WriteLocalFrequencyAndChirp(
               LocFreqFileNamePrefix + "RT" + rtstm.str() +
                                       "DCF" + fnstm.str() + ".dat");
*/
         }

      } // end for numfibers

     if(MCIndex == 0)
        DataWriter.WriteEndMap(0,rr*NumFibers,PropagatedLength);


     } // end if NumLaserRoundTrips > 1

/*     CentralTimeFile
         << Signal->ComputeMeanTimePulse(0.0,Signal->GetTimeWindow()) << "\t";
*/
//         cout  << "Round trip = " << rr << endl;

   if(!(rr%FreqOutputStats))
    {
     int hh=int(rr/FreqOutputStats) - 1;

     CentralTimeHist[hh].AddSample(
                  Signal->ComputeMeanTimePulse(0.0,Signal->GetTimeWindow()));
     CentralFreqHist[hh].AddSample(Signal->ComputeMeanFreqSignal());
     EnergyHist[hh].AddSample(Signal->GetTotalEnergy());
     PhaseHist[hh].AddSample(Signal->ComputeMeanPhasePulse());
     PulseWidthHist[hh].AddSample(Signal->ComputeRMSPulseWidth());
    }


 } // end NumLaserRoundTrips loop

 if(MCIndex == 0)
  {
    DataWriter.WriteEndTransmission(0,NumFibers,PropagatedLength);

  //cout << "Output Average Power [W] = " << Signal->GetAveragePower() << endl;

/**
     cout << "Output Pulse Energy [pJ] = " 
	  << 1e12*Signal->GetTotalEnergy() << endl;

    cout << "Output Max Amplitude [W^{1/2}] = " 
         << sqrt(Signal->ComputeMaxPower()) << endl;

**/

    MaxAmpFile.close();
    EnergyFile.close();
    FWHMFile.close();
    ChirpFile.close();

    if(MyStartup->GetWriteComplexFieldsFlag())
     Signal->WriteComplexFieldsX(Job + "tComplexFields.dat",0);

  }

 //OSA.Update(PropagatedLength);

 Signal->RegenerateSignalString();
 
 //CentralTimeFile << endl;

 //if((MCIndex+1)%50 == 0)
  //OSA.Output(MCIndex+1);

 if((MCIndex+1)%10 == 0)
    cout << "NumSamples Completed = " << MCIndex+1 << endl;  


} // end for MCIndex

 string CentralTimeFileName = Job + "CentralTimeStats.dat";

 ofstream CentralTimeFile(CentralTimeFileName.c_str(),ios::out);

 CentralTimeFile 
        << "% (1) RoundTripNum (2) Mean [ps] (3) StdDev [fs] of CentralTimes"
        << endl;

 for(int hh=0;hh<NumHists;hh++)
  {
    CentralTimeFile << (hh+1)*FreqOutputStats  << "\t"
                    << 1e12*CentralTimeHist[hh].GetMean() << "\t"
                    << 1e15*CentralTimeHist[hh].GetStdDev() << endl;
  }

 CentralTimeFile.close();

 string CentralFreqFileName = Job + "CentralFreqStats.dat";

 ofstream CentralFreqFile(CentralFreqFileName.c_str(),ios::out);

 CentralFreqFile
        << "% (1) RoundTripNum (2) Mean [GHz] (3) StdDev [GHz] of CentralFreqs"
        << endl;

 for(int hh=0;hh<NumHists;hh++)
  {
    CentralFreqFile << (hh+1)*FreqOutputStats  << "\t"
                    << 1e-9*CentralFreqHist[hh].GetMean() << "\t"
                    << 1e-9*CentralFreqHist[hh].GetStdDev() << endl;
  }

 CentralFreqFile.close();

 string EnergyStatsFileName = Job + "EnergyStats.dat";

 ofstream EnergyStatsFile(EnergyStatsFileName.c_str(),ios::out);

 EnergyStatsFile
        << "% (1) RoundTripNum (2) Mean [pJ] (3) StdDev [fJ] of Energys"
        << endl;

 for(int hh=0;hh<NumHists;hh++)
  {
    EnergyStatsFile << (hh+1)*FreqOutputStats  << "\t"
                    << 1e12*EnergyHist[hh].GetMean() << "\t"
                    << 1e15*EnergyHist[hh].GetStdDev() << endl;
  }

 EnergyStatsFile.close();

 string PhaseStatsFileName = Job + "PhaseStats.dat";

 ofstream PhaseStatsFile(PhaseStatsFileName.c_str(),ios::out);

 PhaseStatsFile
        << "% (1) RoundTripNum (2) Mean (3) StdDev of Phases"
        << endl;

 for(int hh=0;hh<NumHists;hh++)
  {
    PhaseStatsFile << (hh+1)*FreqOutputStats  << "\t"
                    << PhaseHist[hh].GetMean() << "\t"
                    << PhaseHist[hh].GetStdDev() << endl;
  }

 PhaseStatsFile.close();

 string PWStatsFileName = Job + "PulseWidthStats.dat";

 ofstream PWStatsFile(PWStatsFileName.c_str(),ios::out);

 PWStatsFile
    << "% (1) RoundTripNum (2) Mean [fs] (3) StdDev [fs] of RMS Pulse Widths"
        << endl;

 for(int hh=0;hh<NumHists;hh++)
  { 
    PWStatsFile << (hh+1)*FreqOutputStats  << "\t"
                    << 1e15*PulseWidthHist[hh].GetMean() << "\t"
                    << 1e15*PulseWidthHist[hh].GetStdDev() << endl;
  }

 PWStatsFile.close();

 std::stringstream IndexStm;

 for(int hh=0;hh<NumHists;hh++)
  {

    IndexStm.clear();
    IndexStm.str("");
    IndexStm << (hh+1)*FreqOutputStats;

    cout << "CentralTimePDF_RT" << IndexStm.str() << endl;

    string OutFileName = Job+"CentralTimePDF_RT" + 
                         IndexStm.str() + ".dat";

    CentralTimeHist[hh].WriteFile_pdf(OutFileName);

    cout << "CentralFreqPDF_RT" << IndexStm.str() << endl;


    OutFileName = Job+"CentralFreqPDF_RT" +
                         IndexStm.str() + ".dat";

    CentralFreqHist[hh].WriteFile_pdf(OutFileName);

    cout << "EnergyPDF_RT" << IndexStm.str() << endl;

    OutFileName = Job+"EnergyPDF_RT" +
                         IndexStm.str() + ".dat";

    EnergyHist[hh].WriteFile_pdf(OutFileName);

    cout << "PhasePDF_RT" << IndexStm.str() << endl;


    OutFileName = Job+"PhasePDF_RT" +
                         IndexStm.str() + ".dat";

    PhaseHist[hh].WriteFile_pdf(OutFileName);

    cout << "PulseWidthPDF_RT" << IndexStm.str() << endl;

    OutFileName = Job+"PulseWidthPDF_RT" +
                         IndexStm.str() + ".dat";

    PulseWidthHist[hh].WriteFile_pdf(OutFileName);

  }


 MaxAmpFile.close();
 EnergyFile.close();
 
// OSA.Output(MyStartup->GetNumMonteCarloSamples());

 delete [] CentralTimeHist;

 Signal->~OptSignal();
 RNG->~RanNumGen();

} // ## end SimilaritonFiberAmplifier

// ###################################################################

void SimilaritonFiberAmplifierDispMatrix(StartupSimilariton *MyStartup)
{

 string JobSeries = MyStartup->GetJobPrefix();

 RanNumGen *RNG;
 OptSignal * Signal;

 RNG =  new RanNumGen("./RanNumGen.in");

 Signal = new OptSignal("./Signal.in", RNG);

 int OSAAppendFlag = 0;

 OptSpectrumAnalyzer OSA("OSA.in",JobSeries,OSAAppendFlag,Signal);

 OptFiberLocalError FiberAmplifier("./FiberAmplifier.in",Signal,RNG,JobSeries);
 OptFiberLocalError DCF("./DCF.in",Signal,RNG,JobSeries);
 SaturableAbsorber  SatAb("./SatAbs.in",Signal);


 int NumLaserRoundTrips = MyStartup->GetNumLaserRoundTrips();

 double DispSwingMin = MyStartup->GetDispSwingMin();
 double DispSwingMax = MyStartup->GetDispSwingMax();
 int DispSwingN = MyStartup->GetDispSwingN();

 double DispSwingDelta = (DispSwingMax - DispSwingMin)/(DispSwingN - 1);

 double DispAvMin = MyStartup->GetDispAvMin();
 double DispAvMax = MyStartup->GetDispAvMax();
 int DispAvN = MyStartup->GetDispAvN();

 double DispAvDelta = (DispAvMax - DispAvMin)/(DispAvN - 1);

 string DispMatrixFileName = JobSeries + "DispMatrix.dat";

 ofstream DispMatrixFile(DispMatrixFileName.c_str(),ios::out);

 DispMatrixFile << "%% DispAvMin,Max,N; DispSwingMin,Max,N;  "
            << endl;

 DispMatrixFile <<  DispAvMin << "\t"
                <<  DispAvMax << "\t"
                <<  DispAvN 
                << endl;

 DispMatrixFile <<  DispSwingMin << "\t"
                <<  DispSwingMax << "\t"
                <<  DispSwingN 
                << endl;

 DispMatrixFile.close();


 std::stringstream DispSwingStm;
 std::stringstream DispAvStm;

string FWHMFileName;
ofstream FWHMFile;
string ChirpFileName;
ofstream ChirpFile;
string EnergyFileName;
ofstream EnergyFile;
double PropagatedLength;
int ZStepNum;


 for(int DispAvInd = 0;DispAvInd < DispAvN;DispAvInd++)
  for(int DispSwingInd = 0;DispSwingInd < DispSwingN;DispSwingInd++)
   {

     Signal->RegenerateSignalString();

     PropagatedLength = 0.0;
     ZStepNum = 0;

     double DispSwing = DispSwingMin + DispSwingInd*DispSwingDelta;
     double DispAv = DispAvMin + DispAvInd*DispAvDelta;

     double DispFA  = 0.5*(DispAv + DispSwing)/
                                      FiberAmplifier.GetFiberLength();

     double DispDCF = 0.5*(DispAv - DispSwing)/DCF.GetFiberLength();

     FiberAmplifier.SetFstOrDispFiber(DispFA);

     DCF.SetFstOrDispFiber(DispDCF);

     cout << "Av,Swing,FA,DCF = " << DispAv << " " << DispSwing << " "
          << DispFA << " " << DispDCF << endl;

     DispSwingStm.clear();
     DispSwingStm.str("");
     DispSwingStm << DispSwingInd;

     DispAvStm.clear();
     DispAvStm.str("");
     DispAvStm << DispAvInd;

     string Job = JobSeries + "Av" + DispAvStm.str() + 
                              "Sw" + DispSwingStm.str() + "_";

     cout << Job << endl;

 FWHMFileName = Job + "FWHM.dat";

 FWHMFile.open(FWHMFileName.c_str(),ios::out);

 FWHMFile << "% (1) PropagatedLength (2) FWHM of Power in time [fs]"
            << endl;

 FWHMFile << PropagatedLength << "\t"
            << 1e15*(Signal->ComputeFWHMPulsePower())
            << endl;


 int ArgMaxPowerIndex = 0;

 double MaxPower=Signal->ComputeMaxPower(ArgMaxPowerIndex);

 ChirpFileName = Job + "Chirp.dat";

 ChirpFile.open(ChirpFileName.c_str(),ios::out);

 ChirpFile << "% (1) PropagatedLength (2) Chirp [THz^2]"
            << endl;

 ChirpFile << PropagatedLength << "\t"
            << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
            << endl;

 EnergyFileName = Job + "Energy.dat";

 EnergyFile.open(EnergyFileName.c_str(),ios::out);

 EnergyFile << "% (1) PropagatedLength (2) Energy [pJ]"
            << endl;
    
 int WriteRT = int(floor(0.8*NumLaserRoundTrips));
 

 for(int rr=1;rr<=NumLaserRoundTrips;rr++)
 {

    SatAb.Apply(&PropagatedLength);


      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;


    FiberAmplifier.Propagate(&PropagatedLength,&ZStepNum);


      EnergyFile << PropagatedLength << "\t"
                << 1e12*Signal->GetTotalEnergy() << endl;

      FWHMFile << PropagatedLength << "\t"
               << 1e15*Signal->ComputeFWHMPulsePower()
               << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;


      DCF.Propagate(&PropagatedLength,&ZStepNum);


         EnergyFile << PropagatedLength << "\t"
                   << 1e12*Signal->GetTotalEnergy() << endl;

         FWHMFile << PropagatedLength << "\t"
                  << 1e15*Signal->ComputeFWHMPulsePower()
                  << endl;

      ChirpFile << PropagatedLength << "\t"
                << 1e-24*(Signal->GetChirp(ArgMaxPowerIndex))
                << endl;

 if(rr==WriteRT)
    Signal->WriteComplexFieldsX(Job + "tComplexFieldsB4End.dat",1);

 } // end NumLaserRoundTrips loop

 Signal->WriteComplexFieldsX(Job + "tComplexFieldsEnd.dat",1);

   EnergyFile.close();
   FWHMFile.close();
   ChirpFile.close();


   } // end double DispMatrix for loop

 Signal->~OptSignal();
 RNG->~RanNumGen();

} // end SimilaritonFiberAmplifierDispMatrix


// #####################################################################

void TestFiberAmplifierNoise(StartupSimilariton *MyStartup)
{

 string Job = MyStartup->GetJobPrefix();

 RanNumGen *RNG;
 OptSignal * Signal;

 RNG =  new RanNumGen("./RanNumGen.in");
 Signal = new OptSignal("./Signal.in", RNG);

 int OSAAppendFlag = 0;

 OptSpectrumAnalyzer OSA("OSA.in",Job,OSAAppendFlag,Signal);

 OptDataWriter DataWriter("./DataWriter.in","./",Job,Signal);
 DataWriter.WriteInitialData();

 OptAmplifier Amp("./Amp.in",Signal,RNG);
 OptFiberLocalError FiberAmplifier("./FiberAmplifier.in",Signal,RNG,Job);

 int NumFibers =
    int(round(MyStartup->GetTotalFiberLength()/
         double(FiberAmplifier.GetFiberLength())));

 if(MyStartup->GetTotalFiberLength() == FiberAmplifier.GetFiberLength())
    NumFibers = 1;

 cout << "NumFibers = " << NumFibers << endl;

 cout << "TotalFiberLength (Must agree) = "
      << MyStartup->GetTotalFiberLength() << " "
      << NumFibers*FiberAmplifier.GetFiberLength() << endl;


//FiberAmplifier.SetGammaFiber(1e-8);
FiberAmplifier.SetGammaFiber(0);

cout << "Input power = " << Signal->GetAveragePower() << endl;

 double PropagatedLength = 0.0;
 int ZStepNum = 0;

for(int MCIndex = 0; MCIndex < MyStartup->GetNumMonteCarloSamples(); MCIndex++)
{
 PropagatedLength = 0.0;
 ZStepNum = 0;

 Amp.AmplifyOptSignal();

 for(int ii=0;ii<NumFibers;ii++)
  FiberAmplifier.Propagate(&PropagatedLength,&ZStepNum);

 if(MCIndex == 0)
  {
    DataWriter.WriteEndTransmission(0,0,PropagatedLength);
    cout << "Output power = " << Signal->GetAveragePower() << endl;
  }

 OSA.Update(PropagatedLength);

 Signal->RegenerateSignalString();

 if((MCIndex+1)%100 == 0)
  OSA.Output(MCIndex+1);

  if((MCIndex+1)%20 == 0)
    cout << "NumSamples Completed = " << MCIndex+1 << endl;

} // end for MCIndex

 OSA.Output(MyStartup->GetNumMonteCarloSamples());

 Signal->~OptSignal();
 RNG->~RanNumGen();

}

// ###########################################################

void CompensateFiberAmp(StartupSimilariton *MyStartup)
{

 string Job = MyStartup->GetJobPrefix();

 RanNumGen *RNG;
 OptSignal * Signal;
 
 RNG =  new RanNumGen("./RanNumGen.in");
 Signal = new OptSignal("./Signal.in", RNG,MyStartup->GetSignalFileName());
 
 OptDataWriter DataWriter("./DataWriter.in","./",Job,Signal);
 DataWriter.WriteInitialData();

 OptFiberLocalError DCF("./DCF.in",Signal,RNG,Job);

 OptFilter oOptFilter("./OptFilter.in",Signal);

LogFileSeparator();
 *LogFile << "TotalFstOrDisp [ps^2] = "
          << 1e24*DCF.GetFirstOrderDispersion()*DCF.GetFiberLength() << endl;
 *LogFile << "Input Pulse Energy [pJ] = "
          << 1e12*Signal->GetTotalEnergy() << endl;
 LogFileSeparator();

 cout << "TotalFstOrDisp [ps^2] = "
      << 1e24*DCF.GetFirstOrderDispersion()*DCF.GetFiberLength() << endl;

 cout << "Input Pulse Energy [pJ] = " << 1e12*Signal->GetTotalEnergy() << endl;

 cout << "Input Max Amplitude [W^{1/2}] = " 
      << sqrt(Signal->ComputeMaxPower()) << endl;

 cout << "Input Pulse Width (std dev) [fs] = " 
      << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << endl;

 cout << "Input FWHM [fs] = " 
      << 1e15*Signal->ComputeFWHMPulsePower() << endl;

 double PropagatedLength = 0.0;
 int ZStepNum = 0;

 DCF.Propagate(&PropagatedLength,&ZStepNum);

 oOptFilter.FilterOptSignal();

 DataWriter.WriteEndTransmission(0,1 ,PropagatedLength);

 cout << "Output Pulse Energy [pJ] = " << 1e12*Signal->GetTotalEnergy() << endl;

 cout << "Output Max Amplitude [W^{1/2}] = "
      << sqrt(Signal->ComputeMaxPower()) << endl;

 cout << "Output Pulse Width (std dev) [fs] = "
      << 1e15*sqrt(Signal->GetSigmaSqOptSignal()) << endl;

 cout << "Output FWHM [fs] = " 
      << 1e15*Signal->ComputeFWHMPulsePower() << endl;

 Signal->~OptSignal();
 RNG->~RanNumGen();

} // ## end CompensateFiberAmp
