
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "StartupSechLaser.hh"

extern ofstream *LogFile;

StartupSechLaser::StartupSechLaser(string InFileName)
{

ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
 {
    cerr << "Error: Input file "<< InFileName 
         << " could not be opened."<< endl;
    exit(1);
 }

JobDir = ReadString("JobDir",&InFile);
JobName =  ReadString("JobName",&InFile);

LogFileName = JobDir + JobName + ".log";

 cout << endl
      << "Log File = " << LogFileName << endl << endl;

LogFile = new ofstream(LogFileName.c_str(),ios::out);
   
if (!LogFile) 
  {
    cerr << "Error: ocs.log could not be open." << endl << flush;
    exit(1);
  }    

JobPrefix = JobDir + JobName + "_";



LogFileSeparator();
*LogFile << InFileName << endl << endl;

*LogFile << "Job Prefix = " << JobPrefix << endl;

FstOrDisp = ReadDouble("FstOrDisp","s^2",1,-1,NO_BOUNDS,
			     OPTIONAL_WARNING,0,&InFile);

Gamma = ReadDouble("Gamma","(Wm)^(-1)",1,-1,NO_BOUNDS,
                             OPTIONAL_WARNING,0,&InFile);


AttenuationLinear =  ReadDouble("AttenuationLinear","m^(-1)",1,-1,NO_BOUNDS,
                             OPTIONAL_WARNING,0,&InFile);


UnsaturatedLinearGain = ReadDouble("UnsaturatedLinearGain","m^(-1)",
                             1,-1,NO_BOUNDS,
                             OPTIONAL_WARNING,0,&InFile);

InFile.close();


} // end constructor

// #####################################################################







