
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "StartupSimilariton.hh"

extern ofstream *LogFile;

StartupSimilariton::StartupSimilariton(string InFileName)
{

ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
 {
    cerr << "Error: Input file "<< InFileName 
         << " could not be opened."<< endl;
    exit(1);
 }

JobDir = ReadString("JobDir",&InFile);
JobName =  ReadString("JobName",&InFile);

LogFileName = JobDir + JobName + ".log";

 cout << endl
      << "Log File = " << LogFileName << endl << endl;

LogFile = new ofstream(LogFileName.c_str(),ios::out);
   
if (!LogFile) 
  {
    cerr << "Error: ocs.log could not be open." << endl << flush;
    exit(1);
  }    

JobPrefix = JobDir + JobName + "_";



LogFileSeparator();
*LogFile << InFileName << endl << endl;

*LogFile << "Job Prefix = " << JobPrefix << endl;

SimType = ReadInt("SimType"," ",0,0,NO_BOUNDS,&InFile);
TotalFiberLength = ReadDouble("TotalFiberLength","m",0,-1,LOWER_ONLY,&InFile);
NumLaserRoundTrips = ReadInt("NumLaserRoundTrips","",1,-1,
                                 LOWER_ONLY,OPTIONAL_WARNING,1,&InFile);

FreqOutputStats = ReadInt("FreqOutputStats","",1,-1,
                                 LOWER_ONLY,OPTIONAL_WARNING,1,&InFile);

NumMonteCarloSamples = ReadInt("NumMonteCarloSamples"," ",1,-1,
                                LOWER_ONLY,OPTIONAL_WARNING,1,&InFile);


WriteComplexFieldsFlag = bool(ReadInt("WriteComplexFieldsFlag"," ",0,-1,
                                LOWER_ONLY,OPTIONAL_WARNING,0,&InFile));

ReadSignalData = bool(ReadInt("ReadSignalData"," ",0,-1,
                                LOWER_ONLY,OPTIONAL_WARNING,0,&InFile));

SignalFileName = ReadString("SignalFileName",OPTIONAL_WARNING," ",&InFile);

SignalFileName = JobDir + SignalFileName;

cout << "SignalFileName = " << SignalFileName << endl;


DispSwingMin = ReadDouble("DispSwingMin","s^2",0,-1,LOWER_ONLY,
                           OPTIONAL_NO_WARNING,&InFile);

DispSwingMax = ReadDouble("DispSwingMax","s^2",DispSwingMin,-1,LOWER_ONLY,
                           OPTIONAL_NO_WARNING,&InFile);


DispSwingN = ReadInt("DispSwingN"," ",2,-1,
                      LOWER_ONLY,OPTIONAL_NO_WARNING,&InFile);

DispAvMin = ReadDouble("DispAvMin","s^2",0,-1,LOWER_ONLY,
                           OPTIONAL_NO_WARNING,&InFile);

DispAvMax = ReadDouble("DispAvMax","s^2",DispAvMin,-1,LOWER_ONLY,
                           OPTIONAL_NO_WARNING,&InFile);


DispAvN = ReadInt("DispAvN"," ",2,-1,
                      LOWER_ONLY,OPTIONAL_NO_WARNING,&InFile);


InFile.close();


} // end constructor

// #####################################################################







