// $Id: StartupTyco.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "StartupTyco.hh"

extern ofstream *LogFile;

StartupZweck::StartupZweck(string InFileName)
{


ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
 {
    cerr << "Error: Input file "<< InFileName 
         << " could not be opened."<< endl;
    exit(1);
 }

JobDir = ReadString("JobDir",&InFile);
JobName =  ReadString("JobName",&InFile);

LogFileName = JobDir + JobName + ".log";

 cout << endl
      << "Log File = " << LogFileName << endl << endl;

LogFile = new ofstream(LogFileName.c_str(),ios::out);
   
if (!LogFile) 
  {
    cerr << "Error: ocs.log could not be open." << endl << flush;
    exit(1);
  }    

JobPrefix = JobDir + JobName + "_";



LogFileSeparator();
*LogFile << InFileName << endl << endl;

*LogFile << "Job Prefix = " << JobPrefix << endl;
InputFilesDir = ReadString("InputFilesDir", &InFile);
SimType = ReadInt("SimType"," ",0,0,NO_BOUNDS,&InFile);
qtStatSamples = ReadInt("qtStatSamples"," ",0,-1,LOWER_ONLY,&InFile);
qtMapSpans = ReadInt("qtMapSpans"," ",0,-1,LOWER_ONLY,&InFile);


WriteAveragePowerdBmFlag = ReadInt("WriteAveragePowerdBm"," ",0,1,
                                 LOWER_AND_UPPER,&InFile);



WriteSNRandOSAFlag = ReadInt("WriteSNRandOSAFlag"," ",0,1,
                              LOWER_AND_UPPER,&InFile);


DebugLevelAmps = ReadInt("DebugLevelAmps"," ",0,3,LOWER_AND_UPPER,&InFile);



FreqOutputSNR = ReadInt("FreqOutputSNR"," ",-1,-2,
                              LOWER_ONLY,&InFile);

FreqOutputReceiverStats =  ReadInt("FreqOutputReceiverStats"," ",-1,-2,
                              LOWER_ONLY,&InFile);


RandomizeBitStringsForMC = ReadInt("RandomizeBitStringsForMC"," ",0,1,
                                    LOWER_AND_UPPER,&InFile);

RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,LOWER_ONLY,
                                &InFile);

OutputStepSizesFlag = ReadInt("OutputStepSizes"," ",0,1,
                               LOWER_AND_UPPER,&InFile);


Beta3ZeroFlag = ReadInt("Beta3ZeroFlag"," ",0,1,LOWER_AND_UPPER,&InFile);
NonLinearityZeroFlag = ReadInt("NonLinearityZeroFlag"," ",0,1,
                                LOWER_AND_UPPER,&InFile);
FiberLossZeroFlag = ReadInt("FiberLossZeroFlag"," ",0,1,
                             LOWER_AND_UPPER,&InFile);
TypeAmp = ReadInt("TypeAmp"," ",0,5,LOWER_AND_UPPER,&InFile);
AmpsNoiseOffFlag = ReadInt("AmpsNoiseOffFlag"," ",0,1,LOWER_AND_UPPER,&InFile);

TypeAmplifierNoise = ReadInt("TypeAmplfierNoise"," ",0,1,LOWER_AND_UPPER,&InFile);


NumExptsForIvansED =  ReadInt("NumExptsForIvansED"," ",0,-1,
                               LOWER_ONLY,&InFile);


DemuxChannelIndex =  -1 +ReadInt("DemuxChannelIndex"," ",1,-1,
                               LOWER_ONLY,MANDATORY,&InFile);

HybridSignalFlag =  ReadInt("HybridSignalFlag"," ",0,1,
                               LOWER_AND_UPPER,MANDATORY,0,&InFile);


int UseMosixFlag = 0;

UseMosixFlag = ReadInt("UseMosixFlag"," ",0,1,LOWER_AND_UPPER,
                        OPTIONAL_WARNING,&InFile);

if(UseMosixFlag)
  {
   lock_or_unlock_mosix(MOSIX_UNLOCK);
   cpujob_mosix();
   slow_mosix();
  }

InFile.close();


} // end constructor

// #####################################################################







