// $Id: StartupZweck.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


# include "StartupZweck.hh"

extern ofstream *LogFile;

StartupZweck::StartupZweck(string InFileName)
{


ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
 {
    cerr << "Error: Input file "<< InFileName 
         << " could not be opened."<< endl;
    exit(1);
 }

JobDir = ReadString("JobDir",&InFile);
JobName =  ReadString("JobName",&InFile);

LogFileName = JobDir + JobName + ".log";

 cout << endl
      << "Log File = " << LogFileName << endl << endl;

LogFile = new ofstream(LogFileName.c_str(),ios::out);
   
if (!LogFile) 
  {
    cerr << "Error: ocs.log could not be open." << endl << flush;
    exit(1);
  }    

JobPrefix = JobDir + JobName + "_";



LogFileSeparator();
*LogFile << InFileName << endl << endl;

*LogFile << "Job Prefix = " << JobPrefix << endl;
InputFilesDir = ReadString("InputFilesDir", &InFile);
SimType = ReadInt("SimType"," ",0,0,NO_BOUNDS,&InFile);
qtStatSamples = ReadInt("qtStatSamples"," ",0,-1,LOWER_ONLY,&InFile);
qtMapSpans = ReadInt("qtMapSpans"," ",0,-1,LOWER_ONLY,&InFile);
WriteDTPWithinMap = ReadInt("WriteDTPWithinMap"," ",0,1,
                             LOWER_AND_UPPER,&InFile);

WriteDTP = ReadInt("WriteDTP"," ",0,1,
                             LOWER_AND_UPPER,&InFile);


WritePulseWidthFlag = ReadInt("WritePulseWidth"," ",0,1,
                             LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

WriteAveragePowerdBmFlag = ReadInt("WriteAveragePowerdBm"," ",0,1,
                             LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);


OutputStepSizesFlag = ReadInt("OutputStepSizes"," ",0,1,
                             LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

WiggleFlag = ReadInt("WiggleFlag"," ",0,1,LOWER_AND_UPPER,
                      OPTIONAL_WARNING,&InFile);

WigglePeriod = ReadDouble("WigglePeriod"," ",0,0,NO_BOUNDS,
                           OPTIONAL_WARNING,&InFile);

WiggleShift = ReadDouble("WiggleShift"," ",0,WigglePeriod,NO_BOUNDS,
                          OPTIONAL_WARNING,&InFile);


KeepDCFlag = ReadInt("KeepDCFlag"," ",0,1,LOWER_AND_UPPER,OPTIONAL_WARNING,
                      &InFile);

SndOrDispZeroFlag = ReadInt("SndOrDispZeroFlag"," ",0,1,
                                LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

GhostPulseBit = ReadInt("GhostPulseBit"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

ii_z = ReadInt("ii_z"," ",0,-1,
                             NO_BOUNDS,OPTIONAL_WARNING,&InFile);

RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
                                LOWER_ONLY,OPTIONAL_WARNING,&InFile);

FiberLossZeroFlag = ReadInt("FiberLossZeroFlag"," ",0,1,
                             LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);


DispMin = ReadDouble("DispMin"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

DispDelta = ReadDouble("DispDelta"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

NumDispPts = ReadInt("NumDispPts"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

LeftTime = ReadDouble("LeftTime"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

RightTime = ReadDouble("RightTime"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

LeftTimeZero = ReadDouble("LeftTimeZero"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

RightTimeZero = ReadDouble("RightTimeZero"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

LeftTimeOne = ReadDouble("LeftTimeOne"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);

RightTimeOne = ReadDouble("RightTimeOne"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);


FreqOutputStats = ReadInt("FreqOutputStats"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);
FreqOutputMC = ReadInt("FreqOutputMC"," ",0,-1,
                             LOWER_ONLY,OPTIONAL_WARNING,&InFile);


NumOnes = ReadInt("NumOnes"," ",3,6,
                   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);


int UseMosixFlag = 0;

UseMosixFlag = ReadInt("UseMosixFlag"," ",0,1,LOWER_AND_UPPER,
                        OPTIONAL_WARNING,&InFile);

if(UseMosixFlag)
  {
   lock_or_unlock_mosix(MOSIX_UNLOCK);
   cpujob_mosix();
   slow_mosix();
  }

InFile.close();





InFile.close();


} // end constructor

// #####################################################################




