// $Id: ocsAppZweck.cc,v 1.5 2000/10/20 18:40:21 zweck
// ###############################################
// 
// The applications in this file were written by John Zweck
// and are not to be modified by anyone else for now.
// 
// ###############################################

#include <iostream.h>
#include <complex.h>

#include "ocs.hh"
#include "MachZehnder.hh"
#include "StartupCiena.hh"
#include "PulseEvolutionAnalyzer.hh"

// ## John Zweck's Applications

void TestHistogram2D(StartupZweck * MyStartup); 
void CienaLoopWDM(StartupZweck * MyStartup); 
void TestMZ(StartupZweck * MyStartup); 
void TestSatAmp(StartupZweck * MyStartup); 
void Ciena450(StartupZweck * MyStartup); 

void TestDispersion(StartupZweck * MyStartup); 
void TestAverageChirp(StartupZweck * MyStartup);
void TestFixedOutPowerAmp(StartupZweck * MyStartup);
void TestRegenerateSignal(StartupZweck * MyStartup); 
void TestGainProfile(StartupZweck * MyStartup); 


extern ofstream *LogFile;

// #### The optical pulses must be local globals 



// #######################################################

// ## Application Wrapper

void MyApplication(int argc, char *argv[])  
{

 StartupZweck MyStartup("Startup.in");


  switch(MyStartup.GetSimType())
   {
    
   case 1: TestFixedOutPowerAmp(&MyStartup);
           break;

   case 2:
      TestHistogram2D(&MyStartup);
      break;

   case 3:
      TestDispersion(&MyStartup);
      break;


   case 4:
      CienaLoopWDM(&MyStartup);
      break;

   case 5:
      TestMZ(&MyStartup);
      break;

   case 6:
      TestSatAmp(&MyStartup);
      break;

   case 7:
      Ciena450(&MyStartup);
      break;


   case 8:
      TestAverageChirp(&MyStartup);
      break;

    case 9:
      TestRegenerateSignal(&MyStartup);
      break;

   case 10:
      TestGainProfile(&MyStartup);
      break;


    default:
      cerr << "Invalid choice of application in SimType" << endl;
      exit(1);
          
   } // end switch

} // end ZweckAppn

// ########################################################

void TestHistogram2D(StartupZweck * MyStartup)
{

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  *LogFile << "TestHistogram" << endl;

  Histogram2D oHistogram2D(InDir + "Histogram2D.in",0,2.2,0,3.01);

  oHistogram2D.AddSample(0.1,1.1);
  oHistogram2D.AddSample(1.1,2.1);
  oHistogram2D.AddSample(1.1,0.1);
  oHistogram2D.AddSample(1.5,2.1);
  oHistogram2D.AddSample(4.0,2.1);
  oHistogram2D.AddSample(5.0,6.1);
  oHistogram2D.AddSample(1.0,7.1);

  oHistogram2D.WriteHistogram(Job + "Hist.dat",Job + "HistParameters.dat");

} // ## end TestHistogram 


// #########################################################

void CienaLoopWDM(StartupZweck * MyStartup)
{


  // #################################################################
  // ########### CienaLoopWDM CienaLoopWDM CienaLoopWDM ##############
  // ###############################################################

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  *LogFile << "CienaLoopWDM" << endl;

  // ################## Initialize signal #################


  RanNumGen RNGSig(InDir + "RanNumGenSignal.in");
  RanNumGen RNGAmp(InDir + "RanNumGenAmps.in");
  RanNumGen RNGFiber(InDir + "RanNumGenFiber.in");

  OptSignal WDMSignal(InDir + "WDMSignal.in",&RNGSig);

 
  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,&WDMSignal);

  OptAmplifier TransmitterAmp1(InDir +"TransmitterAmp1.in",&WDMSignal,&RNGAmp);
  OptAmplifier TransmitterAmp2(InDir +"TransmitterAmp2.in",&WDMSignal,&RNGAmp);

  if(MyStartup->GetAmpsNoiseOffFlag())
     {
       TransmitterAmp1.SetTypeAmplifierNoise(NOISE_OFF);
       TransmitterAmp2.SetTypeAmplifierNoise(NOISE_OFF);
     }


  TransmitterAmp1.AmplifyOptSignal();
  TransmitterAmp2.AmplifyOptSignal();

  LumpedLoss TransmitterLoss(InDir + "TransmitterLoss.in",&WDMSignal);

  TransmitterLoss.ApplyLumpedLoss();

  LumpedLoss LumpedLossCoupler(InDir + "LumpedLossCoupler.in",&WDMSignal);

  if(!MyStartup->GetFiberLossZeroFlag())
     LumpedLossCoupler.ApplyLumpedLoss();

  DataWriter.WriteInitialData();

   // ########## Initialize OSA ###########################


    int OSAAppendFlag = 0;

    OptSpectrumAnalyzer OSA0(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
  
    OptSpectrumAnalyzer OSA1(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);

    if(MyStartup->GetOutputAllOSAs())
       OSAAppendFlag = 1;

    OptSpectrumAnalyzer OSA3(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA5(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA7(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA9(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA11(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
  OptSpectrumAnalyzer OSA13(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
  OptSpectrumAnalyzer OSA15(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
  OptSpectrumAnalyzer OSA17(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSAFinal(InDir +"OSA.in",Job,OSAAppendFlag,&WDMSignal);

   

    // ##### Back to Back OSA ##########################

    // Let's always output Back to Back OSA to get initial SNR correct!
   
   OSA0.Update(0);
   OSA0.Output(0);

   //cout << "THE END ZZZ" << endl;
   //exit(1);

// ############ Initialize Receiver susbsystems  #####
 

 int NumReceivers = MyStartup->GetNumChannsDemux();

  ReceiverSubsystem * Receiver = new ReceiverSubsystem[NumReceivers];


  for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {

 
       Receiver[dd].ReceiverSubsystemInit(&WDMSignal,
                  &RNGAmp,InDir,Job, 
                  MyStartup->GetDemuxChannsArray(dd),
                  &DataWriter);

      Receiver[dd].ReceiveSignalBackToBack(&WDMSignal);
    }

  // ################## Initialize Fibers ################

   OptFiberLocalError SMF50A(InDir + "SMF50A.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError SMFs3A(InDir + "SMFs3A.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError SMFs4A(InDir + "SMFs4A.in",&WDMSignal,&RNGFiber,Job);

   OptFiberLocalError SMF750mA(InDir + "SMF0.75A.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError SMF1500mA(InDir + "SMF1.5A.in",&WDMSignal,&RNGFiber,Job);

   
   OptFiberLocalError SMF2200mA(InDir + "SMF2.2A.in",&WDMSignal,&RNGFiber,Job);

   OptFiberLocalError SMF75B(InDir + "SMF75B.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError SMF75C(InDir + "SMF75C.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError SMF75D(InDir + "SMF75D.in",&WDMSignal,&RNGFiber,Job);


   OptFiberLocalError SMF50D(InDir + "SMF50D.in",&WDMSignal,&RNGFiber,Job);

   OptFiberLocalError DCF10A(InDir + "DCF10A.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF20A(InDir + "DCF20A.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF40A(InDir + "DCF40A.in",&WDMSignal,&RNGFiber,Job);

   OptFiberLocalError DCF10B(InDir + "DCF10B.in",&WDMSignal,&RNGFiber,Job); 
                 
   OptFiberLocalError DCF20B(InDir + "DCF20B.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF40B(InDir + "DCF40B.in",&WDMSignal,&RNGFiber,Job);
                 

   OptFiberLocalError DCF10C(InDir + "DCF10C.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF20C(InDir + "DCF20C.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF40C(InDir + "DCF40C.in",&WDMSignal,&RNGFiber,Job);
                  
   OptFiberLocalError DCF20D(InDir + "DCF20D.in",&WDMSignal,&RNGFiber,Job);
   OptFiberLocalError DCF80D(InDir + "DCF80D.in",&WDMSignal,&RNGFiber,Job);

   // Extra fiber to get total dispersion in loop correct
   // in the pos/neg/zero disp cases  

   OptFiberLocalError DCFPosCaseB(InDir + "DCFPosCaseB.in",&WDMSignal,
                                  &RNGFiber,Job);

   OptFiberLocalError DCFNegCaseB(InDir + "DCFNegCaseB.in",&WDMSignal,
                                  &RNGFiber,Job);


   OptFiberLocalError DCFZeroCaseB(InDir + "DCFZeroCaseB.in",&WDMSignal,
                                  &RNGFiber,Job);



   OptFiberLocalError PreDCF(InDir + "PreDCF.in",&WDMSignal,
                                &RNGFiber,Job);

   OptFiberLocalError PostDCF(InDir + "PostDCF.in",&WDMSignal,
                                &RNGFiber,Job);

   // ################ Set Array of Pointers to Fibers ###########

   const int NumFibers = 26;

   OptFiberLocalError * FiberPtr[NumFibers];

   FiberPtr[0] = &SMF50A; 
   FiberPtr[1] = &SMF75B;
   FiberPtr[2] = &SMF75C;
   FiberPtr[3] = &SMF75D;

   FiberPtr[4] = &DCF10A;
   FiberPtr[5] = &DCF20A;
   FiberPtr[6] = &DCF40A;
   FiberPtr[7] = &DCF10B;
   FiberPtr[8] = &DCF20B;
   FiberPtr[9] = &DCF40B;
   FiberPtr[10] = &DCF10C;
   FiberPtr[11] = &DCF20C;
   FiberPtr[12] = &DCF40C;
   FiberPtr[13] = &DCF20D;
   FiberPtr[14] = &DCF80D;

   FiberPtr[15] = &SMFs3A;
   FiberPtr[16] = &SMFs4A;
   FiberPtr[17] = &SMF750mA;
   FiberPtr[18] = &SMF1500mA;

   FiberPtr[19] = &SMF2200mA;
   FiberPtr[20] = &SMF50D;
   FiberPtr[21] = &DCFPosCaseB;
   FiberPtr[22] = &DCFNegCaseB;
   FiberPtr[23] = &DCFZeroCaseB;

   FiberPtr[24] = &PreDCF;
   FiberPtr[25] = &PostDCF;

   // ############ Check for Special Fiber Cases ####################

   if(MyStartup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(MyStartup->GetGammaZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(MyStartup->GetFiberLossZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetAttenuation(0);
     }


   if(!MyStartup->GetRelativeErrorGoal())
     {
       //cout << "Do nothing" << endl;
     }
   else
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
     }

   
     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetOutputStepSizesFlag(
                            MyStartup->GetOutputStepSizesFlag());

    
  

     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetWriteAveragePowerdBmFlag(
                         MyStartup->GetWriteAveragePowerFlag());

     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetWritePulseWidthFlag(
                         DataWriter.GetWritePulseWidthInFiberFlag());


   // ################ Initialize Optical Amplifiers ############

   
   OptAmplifier AmpA1(InDir + "AmpA1.in",&WDMSignal,&RNGAmp);
   OptAmplifier AmpA2(InDir + "AmpA2.in",&WDMSignal,&RNGAmp);

   OptAmplifier AmpB1(InDir + "AmpB1.in",&WDMSignal,&RNGAmp);
   OptAmplifier AmpB2(InDir + "AmpB2.in",&WDMSignal,&RNGAmp);
                  
   OptAmplifier AmpC1(InDir + "AmpC1.in",&WDMSignal,&RNGAmp);
   OptAmplifier AmpC2(InDir + "AmpC2.in",&WDMSignal,&RNGAmp);
                
   OptAmplifier AmpD1(InDir + "AmpD1.in",&WDMSignal,&RNGAmp);
   OptAmplifier AmpD2(InDir + "AmpD2.in",&WDMSignal,&RNGAmp);

   OptAmplifier PreDCFAmp(InDir + "PreDCFAmp.in",&WDMSignal,&RNGAmp);
   OptAmplifier PostDCFAmp(InDir + "PostDCFAmp.in",&WDMSignal,&RNGAmp);
         

   // #################### Set array of pointers to Amps ########

   int NumAmps = 10;

   OptAmplifier * AmpPtr[NumAmps];

   AmpPtr[0] = &AmpA1;
   AmpPtr[1] = &AmpA2;
   AmpPtr[2] = &AmpB1;
   AmpPtr[3] = &AmpB2;
   AmpPtr[4] = &AmpC1;
   AmpPtr[5] = &AmpC2;
   AmpPtr[6] = &AmpD1;
   AmpPtr[7] = &AmpD2;
   AmpPtr[8] = &PreDCFAmp;
   AmpPtr[9] = &PostDCFAmp;

  if(MyStartup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(MyStartup->GetDebugLevelAmps());
     }


   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(MyStartup->GetTypeAmp()));
     


   if(MyStartup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

   // ###################### Construct LumpedLoss objects ##########

     LumpedLoss LumpedLossA(InDir + "LumpedLossA.in",&WDMSignal);
     LumpedLoss LumpedLossB(InDir + "LumpedLossB.in",&WDMSignal);
     LumpedLoss LumpedLossC(InDir + "LumpedLossC.in",&WDMSignal);
     LumpedLoss LumpedLossD(InDir + "LumpedLossD.in",&WDMSignal);
     LumpedLoss PreDCFLoss(InDir + "PreDCFLoss.in",&WDMSignal);
     LumpedLoss PostDCFLoss(InDir + "PostDCFLoss.in",&WDMSignal);





 // ################ Propagate through loop ############

  LogFileSeparator();
  *LogFile << "Propagating through loop" << endl;

  double PropagatedLength; 
  int ZStepNum; 

  int DispMap = MyStartup->GetDispMap();
  // DispMap = -1 Neg disp map
  // DispMap = 0 Zero disp map
  // DispMap = +1 Pos disp map
  
  for(int expt_index = 0; expt_index < MyStartup->GetQtStatSamples();
      expt_index++)
   {

     int kk;

     //if(expt_index % 10 == 0)
     cout << "Commensing experiment " << expt_index + 1 << endl << flush;

     if(expt_index > 0)
       {
        for(int i=0;i<NumFibers;i++)
	  FiberPtr[i]->SetOutputStepSizesFlag(0);
                            
       }
   
    PropagatedLength = 0.0;
    ZStepNum = 0;


    if(MyStartup->GetDoPreComp())
      {
	PreDCFLoss.ApplyLumpedLoss();
	DataWriter.WriteEndMap(expt_index,kk,PropagatedLength);
       PreDCF.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteEndMap(expt_index,kk ,PropagatedLength);
       PreDCFAmp.AmplifyOptSignal();
       DataWriter.WriteEndMap(expt_index,kk,PropagatedLength);
      }

    for(kk = 0; kk < MyStartup->GetQtMapSpans(); kk++) 
      {

       switch(DispMap)
	 {
           case 0:
            SMF2200mA.Propagate(&PropagatedLength,&ZStepNum); 
            break;

	   case -1:
            SMFs3A.Propagate(&PropagatedLength,&ZStepNum);
            SMFs4A.Propagate(&PropagatedLength,&ZStepNum);
            SMF750mA.Propagate(&PropagatedLength,&ZStepNum);
            break;

	   case 1:
            SMFs3A.Propagate(&PropagatedLength,&ZStepNum);
            SMFs4A.Propagate(&PropagatedLength,&ZStepNum);
            SMF1500mA.Propagate(&PropagatedLength,&ZStepNum);
            break;

	 default:
	   cerr << "CienaLoopWDM: DispMap error. ABORTING!!" << endl;
           exit(1);
	 }


       SMF50A.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
       AmpA1.AmplifyOptSignal();
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       DCF20A.Propagate(&PropagatedLength,&ZStepNum);
       DCF40A.Propagate(&PropagatedLength,&ZStepNum);
       
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
       AmpA2.AmplifyOptSignal();
       
       if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossA.ApplyLumpedLoss();
      
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       SMF75B.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);     
       AmpB1.AmplifyOptSignal();
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
   
      switch(DispMap)
	 {
           case 0:
            DCF10B.Propagate(&PropagatedLength,&ZStepNum);
            DCFZeroCaseB.Propagate(&PropagatedLength,&ZStepNum);
            break;

	   case -1:
            DCF20D.Propagate(&PropagatedLength,&ZStepNum);
            DCFNegCaseB.Propagate(&PropagatedLength,&ZStepNum);
            break;

	   case 1:
             DCF10B.Propagate(&PropagatedLength,&ZStepNum);
             DCFPosCaseB.Propagate(&PropagatedLength,&ZStepNum);
	     break;

	 default:
	   cerr << "CienaLoopWDM: DispMap error. ABORTING!!" << endl;
           exit(1);
	 }

       
       DCF20B.Propagate(&PropagatedLength,&ZStepNum);
       DCF40B.Propagate(&PropagatedLength,&ZStepNum);

       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       AmpB2.AmplifyOptSignal();

       if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossB.ApplyLumpedLoss();

      DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       SMF75C.Propagate(&PropagatedLength,&ZStepNum);

       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
       AmpC1.AmplifyOptSignal();
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
      
       DCF80D.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
  
       AmpC2.AmplifyOptSignal();
      
       if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossC.ApplyLumpedLoss();

       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       SMF50D.Propagate(&PropagatedLength,&ZStepNum);
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

     
       AmpD1.AmplifyOptSignal();
     
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);

       DCF20C.Propagate(&PropagatedLength,&ZStepNum);
       DCF40C.Propagate(&PropagatedLength,&ZStepNum);
    
       DataWriter.WriteWithinMap(expt_index,kk,PropagatedLength);
       AmpD2.AmplifyOptSignal();
       
       if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossD.ApplyLumpedLoss();

       if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossCoupler.ApplyLumpedLoss();

       DataWriter.WriteEndMap(expt_index,kk,PropagatedLength);


    if(MyStartup->GetOutputAllOSAs() && MyStartup->GetWriteSNRandOSAFlag())
      {
        switch(kk+1)
	  {
	   case 1:
             OSA1.Update(PropagatedLength);
             break;
	   case 3:
             OSA3.Update(PropagatedLength);
             break;
	   case 5:
             OSA5.Update(PropagatedLength);
             break;
	   case 7:
             OSA7.Update(PropagatedLength);
             break; 
           case 9:
             OSA9.Update(PropagatedLength);
             break;
           case 11:
             OSA11.Update(PropagatedLength);
             break; 
	   case 13:
             OSA13.Update(PropagatedLength);
             break; 
           case 15:
             OSA15.Update(PropagatedLength);
             break;
           case 17:
             OSA17.Update(PropagatedLength);
             break; 

	  }

      } // end if




   if(expt_index == 0 && kk % 3 == 0)
    cout << "Completed " << kk << " periods of dispersion map " << endl;

     } // ############ end for kk-loop #################


    // $$$$$$$$$$$$$$$$$$$$ end for kk-loop $$$$$$$$$$$$$$


    if(MyStartup->GetDoPostComp())
      {
	PostDCFLoss.ApplyLumpedLoss();
	DataWriter.WriteEndMap(expt_index,kk,PropagatedLength);
        PostDCF.Propagate(&PropagatedLength,&ZStepNum);
        DataWriter.WriteEndMap(expt_index,kk,PropagatedLength);
        PostDCFAmp.AmplifyOptSignal();
      }


    DataWriter.WriteEndTransmission(expt_index,kk,PropagatedLength);

    int UpdateIvansEDFlag = 0;
    int OutputPD_DataFlag = 0;

    if(expt_index == 0)
      OutputPD_DataFlag = 1;

    if(expt_index <= MyStartup->GetNumExptsForIvansED())
      UpdateIvansEDFlag = 1;

    for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
       {
         Receiver[dd].ReceiveSignal(
                 &WDMSignal,UpdateIvansEDFlag,OutputPD_DataFlag,
                 expt_index+1,MyStartup->GetQtMapSpans(),PropagatedLength);
       }


    if(MyStartup->GetWriteSNRandOSAFlag())
      {
           OSAFinal.Update(PropagatedLength);  
      } // end if


  
    if(expt_index == 0)
      {
       cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
       
      }

 if(MyStartup->GetWriteSNRandOSAFlag()  
    &&(expt_index == 0 || expt_index % MyStartup->GetFreqOutputSNR() == 0) )
    {

     if(MyStartup->GetOutputAllOSAs())
      {
         OSA1.Output(expt_index+1);
	 OSA3.Output(expt_index+1);
	 OSA5.Output(expt_index+1);
         OSA7.Output(expt_index+1);
	 OSA9.Output(expt_index+1);
         OSA11.Output(expt_index+1);
         OSA13.Output(expt_index+1);
	 OSA15.Output(expt_index+1);
         OSA17.Output(expt_index+1);
      } // end if

     OSAFinal.Output(expt_index+1);
    }

// ################## PostProcess Signal #################

if(expt_index == MyStartup->GetQtStatSamples() - 1 ||
   expt_index % MyStartup->GetFreqOutputReceiverStats() == 0)
{

 for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      int StatsAppendFlag = 0;

	{
          Receiver[dd].PostProcessReceiverStats(
                                      expt_index+1,MyStartup->GetQtMapSpans(),
                                     StatsAppendFlag);  
        }
    }

} // end if #######

    // Regenerate a (random) initial signal for next expt.
    

    for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->ResetVolatileVariables();

    if(MyStartup->GetRandomizeBitStringsForMC())
        WDMSignal.GetNewRandomSignalString();

    WDMSignal.RegenerateSignalString();

    TransmitterAmp1.AmplifyOptSignal();
    TransmitterAmp2.AmplifyOptSignal();
    TransmitterLoss.ApplyLumpedLoss();

    if(!MyStartup->GetFiberLossZeroFlag())
          LumpedLossCoupler.ApplyLumpedLoss();



     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetWriteAveragePowerdBmFlag(0);
    

   } // ########## end for expt_index loop ###########


 LogFileSeparator();

 *LogFile << "THE END!!" << endl;
 cout <<   "THE END!!" << endl;

} // ## end CienaLoopWDM

// #########################################################

void TestMZ(StartupZweck * MyStartup)
{

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  *LogFile << "TestMZ" << endl;


  // ################## Initialize signal #################


  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal Signal(InDir + "WDMSignal.in",&RNG);

  MachZehnder oMachZehnder(InDir + "MZ.in",&Signal);

  oMachZehnder.MZModulator();
  
  Signal.WriteFileTimePower(Job + "tMZ.dat");
  Signal.WriteFileFreqPower(Job + "fMZ.dat",1);


  /*
  Photodetector oPhoto("Photodetector.in",&Signal);
 
  oPhoto.DetectOptSignal();

  ElecFilter oElec3("ElecFilter.in",&oPhoto);

  oElec3.ComputeMatchAndFilter();

  oPhoto.WriteFileTimePower(Job+"tFiltMZ.dat");
  oPhoto.WriteFileFreqPower(Job+"fFiltMZ.dat",1);

  oElec3.WriteTimePowerFilter(Job+"tRecFilter.dat");
  oElec3.WriteFreqPowerFilter(Job+"fRecFilter.dat");
  */



  /*

  OptSignal CompSignal(InDir + "WDMSignal.in",&RNG);

  double LeftTime = Signal.GetTimeWindow()/2.0 - 2*Signal.GetBitLength();
  double RightTime = Signal.GetTimeWindow()/2.0 + Signal.GetBitLength();

  cout << "LeftTime = " << LeftTime*1e12 << endl;
  cout << "RightTime = " << RightTime*1e12 << endl;

 cout << "MZTotalEnergyRange = " 
       << Signal.GetEnergyInTimeRange(LeftTime,RightTime)  << endl;
 cout << "MZTotalEnergy = " 
       << Signal.GetTotalEnergy()  << endl;




  double NormalizedPower =  CompSignal.GetOptPowerChann()*
                       Signal.GetEnergyInTimeRange(LeftTime,RightTime)/
                       CompSignal.GetEnergyInTimeRange(LeftTime,RightTime);


  CompSignal.SetOptPowerChannel(NormalizedPower);

  CompSignal.RegenerateSignalString();


  cout << "CompSignalTotalEnergyRange = " 
       << CompSignal.GetEnergyInTimeRange(LeftTime,RightTime) << endl;
 cout << "MZTotalEnergy = " 
       << CompSignal.GetTotalEnergy()  << endl;

  */

 /*
  CompSignal.WriteFileTimePower(Job + "tComp.dat");
  CompSignal.WriteFileFreqPower(Job + "fComp.dat",1);
 */

  /*
OptFiberLocalError SMF90(InDir + "SMF90.in",&Signal,&RNG,Job);
OptFiberLocalError CompSMF90(InDir + "SMF90.in",&CompSignal,&RNG,Job);


 double PropagatedLength = 0;
 int NumZSteps = 0;

 SMF90.Propagate(&PropagatedLength,&NumZSteps);
 CompSMF90.Propagate(&PropagatedLength,&NumZSteps);

  CompSignal.WriteFileTimePower(Job + "tCompEnd.dat");
  Signal.WriteFileTimePower(Job + "tEnd.dat");
  */

  //oMachZehnder.GeneratePerfectRectBinaryElecData(0);
  oMachZehnder.GenerateSmoothedRectBinaryElecDataSuperGauss(0);

   
  oMachZehnder.WriteFileTimePower(Job + "tBD.dat");
  oMachZehnder.WriteFileFreqPower(Job + "fBD.dat");
   

  oMachZehnder.FilterBinaryElecData();

  
  oMachZehnder.WriteFileTimePower(Job + "tFBD.dat");
  oMachZehnder.WriteFileFreqPower(Job + "fFBD.dat");
  

  /*

 ReceiverSubsystem * Receiver = new 
                      ReceiverSubsystem[MyStartup->GetNumChannsDemux()];


  for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {

     Receiver[dd].ReceiverSubsystemInit(&Signal,&RNG,InDir,Job, 
                  MyStartup->GetDemuxChannsArray(dd));

    }

  //  Receiver[0].oElecFilter->WriteFreqPowerFilter(Job + "F.dat");

 for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      Receiver[dd].ReceiveSignalBackToBack(&Signal);
    }

  */
 
 
// This piece of code outputs the Filter profiles

 //Photodetector oPhoto("Photodetector.in",&Signal);

 /*

 ElecFilter oElec("TransElecFilter.in",&oPhoto);

 
 oElec.WriteTimePowerFilter(Job+"tTransFilter.dat");
 oElec.WriteFreqPowerFilter(Job+"fTransFilter.dat");
 
 

 ElecFilter oElec2("ElecFilter.in",&oPhoto);


 oElec2.WriteTimePowerFilter(Job+"tRecFilter.dat");
 oElec2.WriteFreqPowerFilter(Job+"fRecFilter.dat");
 */



} // ## end TestMZ

// #############################################################

void TestSatAmp(StartupZweck * MyStartup)
{

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  *LogFile << "TestSatAmp" << endl;


  // ################## Initialize signal #################


  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal Signal(InDir + "SignalSC.in",&RNG);

  OptAmplifier SatAmp(InDir + "SatAmp.in",&Signal,&RNG);

  SatAmp.WriteSteadyStateGainVersusPower(Job + "GainVPower.dat",
         5*SatAmp.GetSaturatingPowerLinear(),30000);


} //## end TestSatAmp

// ###########################################################

void Ciena450(StartupZweck * MyStartup)
{

  // #################################################################
  // ########### Ciena450 Ciena450 Ciena450 ##########################
  // #################################################################

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  LogFileSeparator();
  *LogFile << "Ciena450" << endl;


  // ################## Initialize signal #################


  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal WDMSignal(InDir + "WDMSignal.in",&RNG);

  OptDataWriter DataWriter(InDir + "DataWriter.in",InDir,Job,&WDMSignal);

  MachZehnder MZ(InDir + "MZ.in",&WDMSignal);

  MZ.MZModulator();


  // To get initial OSNR correct we fudge things by preamplifying
  // to get the correct SNR and then applying some lumped loss to get
  // max signal power correct

  OptAmplifier TransmitterAmp(InDir + "TransmitterAmp.in",&WDMSignal,&RNG);
  

  if(MyStartup->GetAmpsNoiseOffFlag())
     {
       TransmitterAmp.SetTypeAmplifierNoise(NOISE_OFF);
     }


  TransmitterAmp.AmplifyOptSignal();

  LumpedLoss TransmitterLoss(InDir + "TransmitterLoss.in",&WDMSignal);
  TransmitterLoss.ApplyLumpedLoss();

  //WDMSignal.WriteFileTimePower(Job + "tInOpt.dat");
  //WDMSignal.WriteFileFreqPower(Job + "fInOpt.dat",0);

  
  // ################# Initialize Fibers ####################

 OptFiberLocalError NZDSF90(InDir + "NZDSF90.in",&WDMSignal,&RNG,Job);
 OptFiberLocalError DCFA(InDir + "DCFA.in",&WDMSignal,&RNG,Job);
 OptFiberLocalError DCFB(InDir + "DCFB.in",&WDMSignal,&RNG,Job);
 OptFiberLocalError DCFC(InDir + "DCFC.in",&WDMSignal,&RNG,Job);
 OptFiberLocalError DCFD(InDir + "DCFD.in",&WDMSignal,&RNG,Job);
 OptFiberLocalError DCFE(InDir + "DCFE.in",&WDMSignal,&RNG,Job);
 

 int NumFibers = 6;

 OptFiberLocalError * FiberPtr[NumFibers];

   FiberPtr[0] = &NZDSF90; 
   FiberPtr[1] = &DCFA;
   FiberPtr[2] = &DCFB;
   FiberPtr[3] = &DCFC;
   FiberPtr[4] = &DCFD;
   FiberPtr[5] = &DCFE;


   // ############ Check for Special Fiber Cases ####################

   if(MyStartup->GetBeta3ZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetSndOrDispFiber(0);
     }

   if(MyStartup->GetGammaZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(MyStartup->GetFiberLossZeroFlag())
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetAttenuation(0);
     }


   if(!MyStartup->GetRelativeErrorGoal())
     {
       //cout << "Do nothing" << endl;
     }
   else
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
     }

      for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetOutputStepSizesFlag(
                            MyStartup->GetOutputStepSizesFlag());

     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetWriteAveragePowerdBmFlag(
                         MyStartup->GetWriteAveragePowerFlag());


  // ################ Initialize Optical Amplifiers ############

   
   OptAmplifier AmpA1(InDir + "AmpA1.in",&WDMSignal,&RNG);
   OptAmplifier AmpA2(InDir + "AmpA2.in",&WDMSignal,&RNG);

   OptAmplifier AmpB1(InDir + "AmpB1.in",&WDMSignal,&RNG);
   OptAmplifier AmpB2(InDir + "AmpB2.in",&WDMSignal,&RNG);
                  
   OptAmplifier AmpC1(InDir + "AmpC1.in",&WDMSignal,&RNG);
   OptAmplifier AmpC2(InDir + "AmpC2.in",&WDMSignal,&RNG);
                
   OptAmplifier AmpD1(InDir + "AmpD1.in",&WDMSignal,&RNG);
   OptAmplifier AmpD2(InDir + "AmpD2.in",&WDMSignal,&RNG);

   OptAmplifier AmpE1(InDir + "AmpE1.in",&WDMSignal,&RNG);
   OptAmplifier AmpE2(InDir + "AmpE2.in",&WDMSignal,&RNG);


   // #################### Set array of pointers to Amps ########

   OptAmplifier * AmpPtr[10];

   int NumAmps = 10;


   AmpPtr[0] = &AmpA1;
   AmpPtr[1] = &AmpA2;
   AmpPtr[2] = &AmpB1;
   AmpPtr[3] = &AmpB2;
   AmpPtr[4] = &AmpC1;
   AmpPtr[5] = &AmpC2;
   AmpPtr[6] = &AmpD1;
   AmpPtr[7] = &AmpD2;
   AmpPtr[8] = &AmpE1;
   AmpPtr[9] = &AmpE2;


  if(MyStartup->GetDebugLevelAmps())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetDebugLevel(MyStartup->GetDebugLevelAmps());
     }


   
  for(int i=0;i<NumAmps;i++)
    AmpPtr[i]->SetTypeAmplifier(typeAmplifier(MyStartup->GetTypeAmp()));
     


   if(MyStartup->GetAmpsNoiseOffFlag())
     {
       for(int i=0;i<NumAmps;i++)
         AmpPtr[i]->SetTypeAmplifierNoise(NOISE_OFF);
     }

 
   // ############### Set up for DTP output ###############

    // ## Write DTP-data at distance = 0.0, creating file (appendFlag = 0)
    // ## DTP = Distance-Time-Power

    int TimeIndexIncrement = 1;

    *LogFile << "TimeIndexIncrement for writing out data  = " 
            << TimeIndexIncrement << endl;
    
    int NumSlicesZ;

    if(MyStartup->GetWriteDTPFlag())
      {
       WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat", 
                                       TimeIndexIncrement,0.0,0);
    
       NumSlicesZ = 1; 
      }
    

  // ########## Initialize OSA ###########################


    int OSAAppendFlag = 0;

    OptSpectrumAnalyzer OSA0(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA_A(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);

    OSAAppendFlag = 1;

    OptSpectrumAnalyzer OSA_B(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA_C(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA_D(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
    OptSpectrumAnalyzer OSA_E(InDir + "OSA.in",Job,OSAAppendFlag, &WDMSignal);
   
    OSAAppendFlag = 0;

    //OptSpectrumAnalyzer OSA_1st(InDir + "OSA.in",Job,OSAAppendFlag, 
    //                          &WDMSignal);

   
    // ##### Back to Back OSA ##########################

    // Let's always output Back 2 back OSA to make sure initial SNR
    // is OK.
 if(MyStartup->GetWriteSNRandOSAFlag())
   {
     OSA0.Update(0);
     OSA0.Output(0);
   }

    // exit(1); // ZZZZZZZ


  // ############ Initialize Receiver susbsystems  #####
 
   ReceiverSubsystem * Receiver = new 
                      ReceiverSubsystem[MyStartup->GetNumChannsDemux()];


  for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {

     Receiver[dd].ReceiverSubsystemInit(&WDMSignal,&RNG,InDir,Job, 
                  MyStartup->GetDemuxChannsArray(dd),&DataWriter);

    }


  // ################ Initialize Back to Back Pulses ##############

  
 for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      Receiver[dd].ReceiveSignalBackToBack(&WDMSignal);
    }

 //exit(1); // ZZZZZZZZZZ

 // Don't need these lines as Receiver doesn't change WDMSignal.

 //MZ.MZModulator();
 //TransmitterAmp.AmplifyOptSignal();
 //TransmitterLoss.ApplyLumpedLoss();

 // ########

 if(MyStartup->GetWriteAveragePowerFlag())
     WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",0.0,1,0);

 //exit(1); // ZZZZZZZZZZZ

   // ################ Propagate through system  ############

  WDMSignal.SetWriteAccumulatedFirstOrderDispersionFlag(
			MyStartup->GetWriteAccDispFlag());

  WDMSignal.OpenAccumulatedFirstOrderDispersionFile(Job + "AccDisp.dat");

  int WriteAveragePulseWidthFlag = MyStartup->GetWriteAveragePulseWidthFlag();

  WDMSignal.OpenAveragePulseWidthFile(Job + "PulseWidth.dat",
                                      WriteAveragePulseWidthFlag);


  LogFileSeparator();
  *LogFile << "Propagating through system" << endl;

  double PropagatedLength; 
  int ZStepNum; 

   
   // ######## Initialize PowerMeter ############

   OptSpectrumAnalyzer PowerMeter(InDir + "PowerMeter.in",Job,0, &WDMSignal);

   PowerMeter.RunPowerMeter(0.0);


  for(int expt_index = 0; expt_index < MyStartup->GetQtStatSamples();
      expt_index++)
   {

     cout << "Commensing experiment " << expt_index + 1 << endl << flush;

     if(expt_index > 0)
       {
        for(int i=0;i<NumFibers;i++)
	  FiberPtr[i]->SetOutputStepSizesFlag(0);
                            
       }

    PropagatedLength = 0.0;
    ZStepNum = 0;

 
       if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	 {
          WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
                 	      		      PropagatedLength,1,0);
	 }

       WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);


    NZDSF90.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

     if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }

     if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


     //OSA_1st.Update(PropagatedLength);

     WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



    AmpA1.AmplifyOptSignal();

 if(expt_index == 0)
    PowerMeter.RunPowerMeter(PropagatedLength);

     if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


    DCFA.Propagate(&PropagatedLength,&ZStepNum);
 
  if(expt_index == 0)
    PowerMeter.RunPowerMeter(PropagatedLength);

     if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }


     if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


     WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);


    
      AmpA2.AmplifyOptSignal();

  if(expt_index == 0)
    PowerMeter.RunPowerMeter(PropagatedLength);

     if(MyStartup->GetWriteSNRandOSAFlag())
      OSA_A.Update(PropagatedLength);


      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }



    if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }


    NZDSF90.Propagate(&PropagatedLength,&ZStepNum);
 
    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

    WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


     if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }

    AmpB1.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);


      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


    DCFB.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);


    WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


    AmpB2.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);


  if(MyStartup->GetWriteSNRandOSAFlag())
    OSA_B.Update(PropagatedLength);


    if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    NZDSF90.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    AmpC1.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


      WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



    DCFC.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    AmpC2.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

   if(MyStartup->GetWriteSNRandOSAFlag())
    OSA_C.Update(PropagatedLength);



   WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);


      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }


    NZDSF90.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


      WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);




    AmpD1.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    DCFD.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


      WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



    AmpD2.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

  if(MyStartup->GetWriteSNRandOSAFlag())
    OSA_D.Update(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }

    NZDSF90.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


      WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



    AmpE1.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

    DCFE.Propagate(&PropagatedLength,&ZStepNum);

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }

      WDMSignal.WriteAccumulatedFirstOrderDispersion(PropagatedLength);

       WDMSignal.WriteAveragePulseWidth(PropagatedLength,
                                         WriteAveragePulseWidthFlag);



    AmpE2.AmplifyOptSignal();

    if(expt_index == 0)
     PowerMeter.RunPowerMeter(PropagatedLength);

 if(MyStartup->GetWriteSNRandOSAFlag())
    OSA_E.Update(PropagatedLength);

      if(expt_index == 0 && MyStartup->GetWriteAveragePowerFlag())
	{
	  WDMSignal.WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength,1,0);
	 }


    if(expt_index == 0 &&  MyStartup->GetWriteDTPFlag())
	 {
           WDMSignal.WriteDistTimePowerPhase(Job + "DTP.dat",
                                             TimeIndexIncrement,
                                             PropagatedLength,1);
           NumSlicesZ++; 

	 }

    // ############### End of Propagation #################


     if(expt_index == 0)
      {
       cout <<"Total PropagatedLength [km] = "<<PropagatedLength*1e-3<< endl;
       cout <<"Number of Slices in Z = " << NumSlicesZ << endl;
      }

     int UpdateIvansEDFlag = 0;
     int OutputPD_DataFlag = 0;

     if(expt_index == 0)
        OutputPD_DataFlag = 1;

     if(expt_index <= MyStartup->GetNumExptsForIvansED())
         UpdateIvansEDFlag = 1;

     for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
    {
      Receiver[dd].ReceiveSignal(&WDMSignal,UpdateIvansEDFlag,
                                  OutputPD_DataFlag,0,0,PropagatedLength);
    }


    if(expt_index == 0)
      {
	//WDMSignal.WriteFileTimePower(Job + "tOutOpt.dat");
	//WDMSignal.WriteFileFreqPowerdBm(Job + "fOutOpt.dat");

       if(MyStartup->GetWriteComplexFieldsFlag())
         WDMSignal.WriteComplexFields(Job + "tComplexFields.dat",0);

      }

 if(MyStartup->GetWriteSNRandOSAFlag()  
    &&(expt_index == 0 || expt_index % MyStartup->GetFreqOutputSNR() == 0) )
    {
      //OSA_1st.Output(-1);
     OSA_A.Output(expt_index+1);
     OSA_B.Output(expt_index+1);
     OSA_C.Output(expt_index+1);
     OSA_D.Output(expt_index+1);
     OSA_E.Output(expt_index+1);
    }

   // ################## PostProcess Signal #################

   if(expt_index % MyStartup->GetFreqOutputReceiverStats() == 0)
    {

       for(int dd = 0; dd < MyStartup->GetNumChannsDemux(); dd++)
         {
          int StatsAppendFlag = 0;

          Receiver[dd].PostProcessReceiverStats(
                                     expt_index+1,0,StatsAppendFlag);
          StatsAppendFlag = 1;   
        }
    } // ## end if

   
   // ######### Prepare for next expt ##############

    for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->ResetVolatileVariables();

    if(MyStartup->GetRandomizeBitStringsForMC())
        WDMSignal.GetNewRandomSignalString();


    MZ.MZModulator();
    TransmitterAmp.AmplifyOptSignal();
    TransmitterLoss.ApplyLumpedLoss();

     for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetWriteAveragePowerdBmFlag(0);
    

      WDMSignal.CloseAccumulatedFirstOrderDispersionFile();
      WDMSignal.ResetAccumulatedFirstOrderDispersionToZero();
      WDMSignal.SetWriteAccumulatedFirstOrderDispersionFlag(0);

      WDMSignal.CloseAveragePulseWidthFile(
                                      WriteAveragePulseWidthFlag);

      
      WriteAveragePulseWidthFlag = 0; // Only write out for 1st MC expt
      

   } // end for expt_index loop


 LogFileSeparator();

 *LogFile << "THE END!!" << endl;
 cout <<   "THE END!!" << endl;

} // ## end Ciena450

// #######################################################

void TestDispersion(StartupZweck * MyStartup)
{


  LogFileSeparator();
  *LogFile << "TestDispersion" << endl;

  // ################## Initialize signal #################

  
  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();


  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal WDMSignal(InDir + "WDMSignal.in",&RNG);

  WDMSignal.WriteFileTimePower(Job + "tIn.dat");
  WDMSignal.WriteFileFreqPowerdBm(Job + "fIn.dat");

  //OptFiberLocalError SMF50A(InDir + "SMF50A.in",&WDMSignal,&RNG,Job);
  OptFiber SMF50A(InDir + "SMF50A.in",&WDMSignal,&RNG);



   if(MyStartup->GetBeta3ZeroFlag())
     {
        SMF50A.SetSndOrDispFiber(0);
     }

   if(MyStartup->GetGammaZeroFlag())
     {
       
         SMF50A.SetGammaFiber(0);
     }

   if(MyStartup->GetFiberLossZeroFlag())
     {
         SMF50A.SetAttenuation(0);
     }

   /*
    if(!MyStartup->GetRelativeErrorGoal())
     {
       cout << "Do nothing" << endl;
     }
   else
     {
       
         SMF50A.SetRelativeErrorGoal(MyStartup->GetRelativeErrorGoal());
     }

    
     
   SMF50A.SetOutputStepSizesFlag(
                            MyStartup->GetOutputStepSizesFlag());

   SMF50A.SetWriteAveragePowerdBmFlag(
                         MyStartup->GetWriteAveragePowerFlag());

   */

       double PropagatedLength = 0.0;
       int ZStepNum = 0;

       //SMF50A.Propagate(&PropagatedLength,&ZStepNum);

       SMF50A.PropagateFiberScalarNLS(&PropagatedLength);

       WDMSignal.WriteFileTimePower(Job + "tOut.dat");
       WDMSignal.WriteFileFreqPowerdBm(Job + "fOut.dat");


       cout << "ZStepNum = " << ZStepNum << " Length = " 
            << PropagatedLength << endl;

} // ## end TestDispersion

// #####################################################

void TestAverageChirp(StartupZweck * MyStartup)
{

  LogFileSeparator();
  *LogFile << "TestAverageChirp" << endl;

  // ################## Initialize signal #################

  
  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();


  RanNumGen RNG(InDir + "RanNumGenSignal.in");

  OptSignal Signal(InDir + "SCSignal.in",&RNG);

  /*
  Signal.WriteFileTimePower(Job + "tIn.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fIn.dat");

  double Chirp = Signal.ComputeAverageChirpOfMarks(Signal.GetBitLength()/2.0);

  cout << "Chirp = " << Chirp << endl;
  */

  OptFilter oOptFilter(InDir+"OptFilter.in", &Signal);

  oOptFilter.WriteFreqPowerGaussianFilter(Job+"OptFilter.dat");

  cout << "THE END!!" << endl;  //ZZZ

}// ## end TestAverageChirp

// ################################################

void TestFixedOutPowerAmp(StartupZweck * MyStartup)
{

  LogFileSeparator();
  *LogFile << "TestFixedOutPowerAmp" << endl;

  // ################## Initialize signal #################

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal Signal(InDir + "WDMSignal.in",&RNG);

  Signal.WriteFileTimePower(Job + "tIn.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fIn.dat");

  OptAmplifier Amp(InDir + "Amp.in",&Signal,&RNG);

  Amp.AmplifyOptSignal();

  Signal.WriteFileTimePower(Job + "tOut.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fOut.dat");



} // ## end TestFixedOutPowerAmp

// ###########################################################
 
void TestRegenerateSignal(StartupZweck * MyStartup) 
{

  LogFileSeparator();
  *LogFile << "TestRegenerateSignal" << endl;

  // ################## Initialize signal #################

  
  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();


  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal Signal(InDir + "SCSignal.in",&RNG);



  //Signal.TestGetRandomJones(Job+"Stokes.dat",20);

  //exit(1);

  Signal.RegenerateSignalString();

  Signal.WriteFileTimePower(Job + "tIn.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fIndBm.dat");
  Signal.WriteFileFreqPower(Job + "fInLin.dat",1);


  cout << "Average Pulse Width [ps] = " 
       << 1e12*Signal.ComputeAverageSigmaMarks() << endl;


  Signal.WriteComplexFields(Job + "CF.dat",1);

  OptSignal SignalNew(InDir + "SCSignal.in",&RNG);

  for(int ii=0;ii<SignalNew.GetQtPoints();ii++)
    {
      SignalNew.sfftM.tOptSignalX[ii] = 0;
      SignalNew.sfftM.tOptSignalY[ii] = 0;
      SignalNew.sfftM.fOptSignalX[ii] = 0;
      SignalNew.sfftM.fOptSignalY[ii] = 0;

    }

  SignalNew.ReadComplexFields(Job + "CF.dat");

  SignalNew.WriteFileTimePower(Job + "tInNew.dat");
  SignalNew.WriteFileFreqPower(Job + "fInLinNew.dat",0);

OptSignal SignalDiff(InDir + "SCSignal.in",&RNG);

 SignalDiff = SignalNew -Signal;

 SignalDiff.WriteFileTimePower(Job + "tInDiff.dat");
  SignalDiff.WriteFileFreqPower(Job + "fInLinDiff.dat",0);




} // ## end TestRegenerateSignal

// ###################################################




void TestGainProfile(StartupZweck * MyStartup)
{

  LogFileSeparator();
  *LogFile << "TestGainProfile" << endl;

  // ################## Initialize signal #################

  string InDir = MyStartup->GetInputFilesDir();
  string Job = MyStartup->GetJobPrefix();

  RanNumGen RNG(InDir + "RanNumGen.in");

  OptSignal Signal(InDir + "WDMSignal.in",&RNG);
  MachZehnder MZ(InDir + "MZ.in",&Signal);

  MZ.MZModulator();


  // To get initial OSNR correct we fudge things by preamplifying
  // to get the correct SNR and then applying some lumped loss to get
  // max signal power correct

  OptAmplifier TransmitterAmp(InDir + "TransmitterAmp.in",&Signal,&RNG);
  

  if(MyStartup->GetAmpsNoiseOffFlag())
     {
       TransmitterAmp.SetTypeAmplifierNoise(NOISE_OFF);
     }


  TransmitterAmp.AmplifyOptSignal();

  LumpedLoss TransmitterLoss(InDir + "TransmitterLoss.in",&Signal);
  TransmitterLoss.ApplyLumpedLoss();

  Signal.WriteFileTimePower(Job + "tIn.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fIn.dat");

  OptAmplifier Amp(InDir + "AmpGainProfile.in",&Signal,&RNG);

  Amp.AmplifyOptSignal();

  Signal.WriteFileTimePower(Job + "tOut.dat");
  Signal.WriteFileFreqPowerdBm(Job + "fOut.dat");

  Amp.WriteGainProfile(Job + "GainProfile.dat");


} // ## end TestGainProfile

// ###########################################
