#ifndef __INLINES_H__
#define __INLINES_H__


#include "ron_base.hh"
#include <math.h>
//#include "tools.hh"


//#include "/home/home1/walter/ocs_cvs/optics/C++/ocs/include/ocs.hh"


// CAUTION: TO use good math inlines, install
// the additional mathinlines package, DO NOT USE
// see /usr/include/bits/mathinlines.h  !!



// cyclic left shift by 'shift' of array in[] of length 'len'.
template<class T>
void lshift(T *in, int len, int shift)
{
  if(shift < 0)
    shift += (2 - shift/len)*len;
  shift = shift % len;
  T *temp = new T[shift];
  memcpy(temp, in, sizeof(T)*shift);
  memmove(in, in+shift, sizeof(T)*(len-shift));
  memcpy(in+len-shift, temp, sizeof(T)*shift);
  delete [] temp;
}


template<class T>
void lshift(T *in, int len, double shift)
{
  // take care of integral part first.
  if(fabs(floor(shift)) >= 1.0)
    lshift(in, len, int(floor(shift)));
    
  // fractional part left?
  const double shift_l = shift - floor(shift);
  if(shift_l > EPS) // shift_l is always > 0
  {
    for(int i=0; i < len-1; i++)
      in[i] = in[i]*(1.0-shift_l) + in[i+1]*shift_l;
    in[len-1] = in[len-1]*(1.0-shift_l) + in[0]*shift_l;
  }
}





//============================================================================

// computes exp(i * x), where x is real
// these routines are ok for values in thew range [1e-302, 1e18]
inline cplx exp_i_real(double __x)
{
#if !defined __NO_MATH_INLINES && defined __GNUC__
  register long double __cosr;
  register long double __sinr;
  __asm __volatile__
    ("fsincos\n\t"
     : "=t" (__cosr), "=u" (__sinr) : "0" (__x));
// alternatively:
//  double sin_x, cos_x;
//  __sincos(x, &sin_x, &cos_x);
  return cplx(double(__cosr), double(__sinr));
#else
  return cplx(cos(x), sin(x));
#endif
}

inline void exp_i_real2(double __x, cplx *out)
{
#if !defined __NO_MATH_INLINES && defined __GNUC__
  register long double __cosr;
  register long double __sinr;
  __asm __volatile__
    ("fsincos\n\t"
     : "=t" (__cosr), "=u" (__sinr) : "0" (__x));
//  double sin_x, cos_x;
//  __sincos(x, &sin_x, &cos_x);
  *out = cplx(double(__cosr), double(__sinr));
#else
  *out = cplx(cos(x), sin(x));
#endif
}


//============================================================================

// computes exp(x), where x is complex
inline cplx exp_complex(cplx x)
{
  return exp(real(x)) * exp_i_real(imag(x));
}

//============================================================================

// the normal distribution
// p(x) = 1/(sqrt(2pi)*sigma)  exp(-x^2/(2*sigma^2))
// Normalization: int_{-\infty}^{\infty} p(x) dx = 1
#define INV_SQRT_2PI	0.39894228040143
inline double normal_std(double x, double x0, double sigma)
{
  return exp(-0.5*sqr((x-x0)/sigma)) * INV_SQRT_2PI / sigma;
}
// var: variance, var == sqr(sigma)
inline double normal_var(double x, double x0, double var_l)
{
  return exp(-0.5*sqr(x-x0)/var_l) * INV_SQRT_2PI / sqrt(var_l);
}


//============================================================================

/*
inline bool simeq(double a, double b)
{
  return (fabs(a - b) <= EPS*fabs(b));
}
*/

//============================================================================
// 'factor' is power ratio

inline double dB_to_factor(double G_dB)
{
  return pow(10.0, G_dB/10.0);
}

//============================================================================
// 'factor' is power ratio

inline double factor_to_dB(double factor)
{
  return 10.0 * log10(factor);
}

//--------------------------------------------------------------------------------------


// The following four routines linearly map [a1, a2] --> [b1, b2].
// If exactly one of the intervals is dicrete (type int), mapping is done
// so that each 'slot' (each of the numbers a1, a1+1, a1+2..., a2, or b1...b2)
// corresponds to equal subintervals in the continuous domain
// For the double to int case, this means
//   [a1, a1+da) --> b1, [a1+da, a1+2da) --> b1+1, ..., [a2-da, a2) --> b2,
//   with   da = diff_a / num_b
//
// CAUTION: NO BOUNDS CHECKING IS DONE !!

//--------------------------------------------------------------------------------------
// double to double, [a1, a2] --> [b1, b1+diff_b]
// diff_a = a2-a1 (can be negative)
// diff_b = b2-b1 (can be negative)
inline double lin_map(double x, double a1, double diff_a, double b1, double diff_b)
{
  return (x-a1)/diff_a * diff_b + b1;
}

//--------------------------------------------------------------------------------------
// double to int, [a1, a2) --> [b1, b1+num_b-1]
// CAUTION: a2 is mapped to b1+num_b, x = a2 is illegal value!!
// diff_a = a2-a1 (can be negative)
// num_b = b2-b1+1 (can be negative)
inline int lin_map(double x, double a1, double diff_a, int b1, int num_b)
{
  return  int(floor(  (x-a1)/diff_a * num_b + b1 ));
}

//--------------------------------------------------------------------------------------
// int to double, [a1, a2] --> [b1, b1+diff_b]
inline double lin_map(int x, int a1, int num_a, double b1, double diff_b)
{
  return double(x-a1)/double(num_a) * diff_b + b1;
}

//--------------------------------------------------------------------------------------
// int to int, [a1, a2] --> [b1, b1+num_b-1]
inline double lin_map(int x, int a1, int num_a, int b1, int num_b)
{
  return  ((x-a1)*num_b)/num_a + b1; // all in integer arithmetics; saves floor()
}

//--------------------------------------------------------------------------------------
// lin_interpolate interpolates between two array entries using triangular interpolation
// the array is of type double
// CAUTION: NO BOUNDS CHECKING IS DONE !!  x = a2 is illegal value!!
//
// double to int, [a1, a1+diff_a) --> [b1, b1+num_b-1]  (CAUTION: a2 is mapped to b1+num_b !!)
// diff_a = a2-a1 (can be negative)
// num_b = b2-b1+1 (can be negative)

inline double lin_interpolate(double x, double a1, double diff_a,
 int b1, int num_b, double *data)
{
  // 'exact' index
  double x2 = lin_map(x, a1, diff_a, double(b1), double(num_b));
  double x_floor = floor(x2);
  double *p = &(data[int(x_floor)]);

  // triangular interpolation
  // f(x) \approx f(a) + (f(b)-f(a))/(b-a) * (x-a),  here: b-a = 1
  return  (*p) + ((*(p+1))-(*p)) * (x2-x_floor);
}
// complex array
inline cplx lin_interpolate(double x, double a1, double diff_a, int b1, int num_b, cplx *data)
{
  // 'exact' index
  double x2 = lin_map(x, a1, diff_a, double(b1), double(num_b));
  double x_floor = floor(x2);
  cplx *p = &(data[int(x_floor)]);

  // triangular interpolation
  // f(x) \approx f(a) + (f(b)-f(a))/(b-a) * (x-a),  here: b-a = 1
  return  (*p) + ((*(p+1))-(*p)) * (x2-x_floor);
}



//------------------------------------------------------------------------------
// CAUTION!! FWHM in norm(gaussian()), not in output of gaussian() !!!
// freq is an offset frequency
// FWHM_MULTIPLIER = 0.5 / sqrt(log(2))
#define FWHM_MULTIPLIER   0.600561204393225
inline cplx gaussian(double t, double t0, double FWHM_norm, double freq, double phase)
{
  // sigma = t_FWHM / 2*sqrt(-2*ln(1/2)) == t_FWHM / 2*sqrt(2*ln(2)), use ln(2) == M_LN2
  // we want FWHM in POWER, not field!!
  double  sigma = FWHM_norm * FWHM_MULTIPLIER;
  double  t_l = t-t0;

  return  exp(cplx(-0.5*sqr(t_l / sigma), 2.0*M_PI*freq*t + phase));
}


//------------------------------------------------------------------------------

inline double soliton(double t, double t0, double eta)
{
  return  eta / cosh(eta*(t-t0));
}


//------------------------------------------------------------------------------
//
// Butterworth window function: 1 / ( 1 + |2(x-x0)/FWHM|^k ),
// where `k' is the order
inline double butterworth(double t, double t0, double FWHM_norm, double order)
{
  return  1.0 / ( 1.0+pow(fabs(2.0*(t-t0)/FWHM_norm),order) );
}


//------------------------------------------------------------------------------
// CRZ (chirped raised cosine) pulse centered at t0 and of duration
//  T_bit = 2pi/omega, with chirp coefficient `chirp' and offset frequency
//  `freq'. Max amplitude is 1.0.
//
//  t:  time argument
//  t0: central time
//  omega: pulse frequency parameter (determines pulse width, usually 1/T_bit)
//  chirp: you guessed it...
//  freq: carrier frequency
//
inline cplx crz(double t, double t0, double omega, double chirp, double freq)
{
  double  phase = omega*(t-t0);
  if(fabs(phase) > M_PI)
    return 0.0; // only consider center peak
  else
  {
    // OLD CODE, revised 27 Aug 01, R.H., according to 
    // Tian-Shiang Yang, W.L. Kath: Optimal prechirping for dispersion-m...
    // OFC ??, ThQ4-1, p. 249
    //
    //double  cos_l = cos(phase);
    //return  sqrt(0.5*(1.0+cos_l))
    //  * exp_i_real(M_PI*(chirp*cos_l + 2.0*freq*t)); // Caution: factor pi before ()
    
    // CAUTION: WE USE OPPOSITE SIGN IN chirp parameter from the above reference!
    //  (chirp := -m, where 'm' is the chirp parameter in the paper)
    return  sqrt(0.5*(1.0+cos(M_PI*sin(0.5*phase))))
      * exp_i_real(M_PI*(chirp*cos(phase) + 2.0*freq*t)); // Caution: factor pi before ()
    
  }
}





#endif // __INLINES_H__
/**/

