// $Id: ron_base.hh,v 1.1 2007/10/24 14:39:21 zweck Exp $

#ifndef __RON_BASE_H__
#define __RON_BASE_H__

#include <iostream.h>
#include <stdlib.h>
#include <errno.h>
#include <complex.h>
#include <math.h>
#include <values.h>
#include <limits.h>
#include <malloc.h>
#include <new>
//#include "vector.h"

#ifdef __GNUC__
#  ifndef __i686__
#    define __i686__
#  endif
typedef complex<double> cplx;
typedef complex<long double> lcplx;
// dat type long double gives problems with egcs 1.1.2
#  define LONG_DOUBLE double
#else
typedef complex  cplx;
#  define LONG_DOUBLE double
#endif


#include <bits/nan.h>
/*
  #ifndef NAN
  #ifdef NaN
  // for Linux 6.0
  #define NAN NaN
  #else
  #include <nan.h>
  cplx  NAN_dummy; // for SGI
  #define NAN NaN(NAN_dummy)
  #endif
  #endif
*/


#ifndef EPS
#  define EPS 2.2e-16
#endif


/*
// pointer type to a function which yields a uniformly
// distributed random number in the interval [0,1)
typedef double (*uniform_rand_function) (void);


// use this code to get time info based on cycles:
// cycle_type cycles1 = cycle_get_time();
// // ...compute something...
// f();
// cycle_type cycles2 = cycle_get_time();
// fill in the right value here from /proc/cpuinfo
// double tt = cycle_time_to_sec(cycles2-cycles1, 398000000L);
// cout << "Function f() took " << tt << " seconds to complete" << endl;

// real time clock access, gives number of clock ticks
typedef unsigned long long cycle_type;


static __inline__ cycle_type read_tsc()
{
     cycle_type ret;

     __asm__ __volatile__("rdtsc": "=A" (ret)); 
     // no input, nothing else clobbered 
     return ret;
}


#define cycle_get_time()  read_tsc()
#define cycle_time_diff(t1,t2) ((t1) - (t2))
// 'cycles_per_sec' is e.g. 398000000L for a 398 MHz Pentium II,
// look up the value in /proc/cpuinfo
#define cycle_time_to_sec(t,cycles_per_sec) (((double)(t)) / cycles_per_sec)
//#define FFTW_TIME_MIN (1.0e-4)	// for Pentium TSC register 
*/


/*
#ifdef GLOBAL_NEW_MEMORY
extern volatile unsigned long GLOBAL_NEW_MEMORY_VAL;
#endif //GLOBAL_NEW_MEMORY

// overloaded "new" and "new[]" operator
// purpose: allocate memory aligned to 16-byte boundaries
inline void* operator new[](size_t t);
inline void* operator new[](size_t t)
{
  void *p = memalign(16, t);
  if(!p)
    throw bad_alloc();
  if(((unsigned int)(p))%16)
  {
    cerr << "\n\noperator new[]: bad pointer alignment (ron_base.hh)\n\n";
    exit(-1);
  }
#ifdef GLOBAL_NEW_MEMORY
  GLOBAL_NEW_MEMORY_VAL += (unsigned long)(t);
#endif //GLOBAL_NEW_MEMORY
#ifdef DEBUG_NEW
  cout << "\nAllocating " << t << " bytes! (new[])" << endl;
#ifdef GLOBAL_NEW_MEMORY
  cout << "Total allocated memory: " << GLOBAL_NEW_MEMORY_VAL << " bytes" << endl;
#endif //GLOBAL_NEW_MEMORY
#endif // DEBUG_NEW
  return(p);
}
*/

/*
inline void* operator new(size_t t);
inline void* operator new(size_t t)
{
  void *p = memalign(16, t);
  if(!p)
    throw bad_alloc();
  if(((unsigned int)(p))%16)
  {
    cerr << "\n\noperator new: bad pointer alignment (ron_base.hh)\n\n";
    exit(-1);
  }
#ifdef GLOBAL_NEW_MEMORY
  GLOBAL_NEW_MEMORY_VAL += (unsigned long)(t);
#endif //GLOBAL_NEW_MEMORY
#ifdef DEBUG_NEW
  cout << "\nAllocating " << t << " bytes! (new)" << endl;
#ifdef GLOBAL_NEW_MEMORY
  cout << "Total allocated memory: " << GLOBAL_NEW_MEMORY_VAL << " bytes" << endl;
#endif //GLOBAL_NEW_MEMORY
#endif // DEBUG_NEW
  return(p);
}
*/

/*
inline void operator delete[](void *p);
inline void operator delete[](void *p)
{
#ifdef DEBUG_NEW
  cout << "Freeing memory (delete[])" << endl;
#endif // DEBUG_NEW
  free(p);
}
*/

/*
inline void operator delete(void *p);
inline void operator delete(void *p)
{
#ifdef DEBUG_NEW
  cout << "Freeing memory (delete)" << endl;
#endif // DEBUG_NEW
  free(p);
}
*/




template<class T> inline void init_zero(T* in, size_t t);
template<class T> inline void init_zero(T* in, size_t t)
{
  register T* T_end = in+t;
  for(; in<T_end; in++)
    *in = T(0);
}

template<class T> inline void init_to_value(T* in, size_t t, T value);
template<class T> inline void init_to_value(T* in, size_t t, T value)
{
  register T* T_end = in+t;
  for(; in<T_end; in++)
    *in = value;
}

bool simeq(double a, double bb);

template<class T> inline T sqr(T);
template<class T> inline T sqr(T in)
{
  return in*in;
}

/*
template<class T> inline T cube(T);
template<class T> inline T cube(T in)
{
  return in*in*in;
}
*/
/*
template <class T> inline T max(T,T);
template <class T> inline T max(T a, T bb)
{
  return  ((a >= bb)? a:bb);
}
inline cplx max(cplx &a, cplx &bb);
inline cplx max(cplx &a, cplx &bb)
{
  return (norm(a) >= norm(bb))? a:bb;
}
*/

template <class T> inline T min(T,T);
template <class T> inline T min(T a, T bb)
{
  return  ((a < bb)? a:bb);
}
inline cplx min(cplx &a, cplx &bb);
inline cplx min(cplx &a, cplx &bb)
{
  return (norm(a) < norm(bb))? a:bb;
}



template<class T>
  inline int inrange (const T x, const T a, const T bb)
  {
    return (a < bb) ? (x>=a && x<=bb) : (x>=bb && x<=a);
  }

template<class T>
  inline int inrange_bb (const T x, const T a, const T bb)
  {
    return (a < bb) ? (x>a && x<bb) : (x>bb && x<a);
  }

template<class T>
  inline void torange (T& x, const T a, const T bb)
  {
    if (a<bb)
    {
      if   (x<a) x = a;
      else if (x>bb) x = bb;
      return;
    }

    if (a>bb)
    {
      if   (x<bb) x = bb;
      else if (x>a) x = a;
      return;
    }
  }

template<class T>
  inline void swap (T& x, T& y)
  {
    T tmp = x; x = y; y = tmp;
  }

template<class T>
  inline T sum (T *in, int N, int skip = 1)
  {
    T val = T(0.0);
    T *p = in;
    T *const p_end = in+N;
    for(; p < p_end; p+=skip)
      val += *p;

    return  val;
  }

// returns the location of the maximum element in an array
template<class T>
  inline T findmax (T *in, int N, int *loc)
  {
    T *p = in;
    *loc = 0;
    T this_max = *p;
    T *const p_end = in+N;
    for(p++; p < p_end; p++)
      if(*p > this_max)
      {
	this_max = *p;
	*loc = int(p - in);
      }
    return  this_max;
  }

// returns the location of the maximum *magnitude* element in an array
template<class T>
  inline T findmax_abs (T *in, int N, int *loc)
  {
    T *p = in;
    *loc = 0;
    T this_max_abs = abs(*p);
    T *const p_end = in+N;
    for(p++; p < p_end; p++)
    {
      register const T p_abs = abs(*p);
      if(p_abs > this_max_abs)
      {
	this_max_abs = p_abs;
	*loc = int(p - in);
      }
    }
    return  this_max_abs;
  }



template<class T>
  inline T findmin (T *in, int N, int *loc)
  {
    T *p = in;
    *loc = 0;
    T this_min = *p;
    T *const p_end = in+N;
    for(p++; p < p_end; p++)
      if(*p < this_min)
      {
	this_min = *p;
	*loc = int(p - in);
      }

    return  this_min;
  }

#define BUS_SIZE 8
template<class T> inline T*  ron_alloc (int size, T **p_orig);
template<class T>
  inline T*  ron_alloc (int len, T **p_orig)
  {
    if(sizeof(T) < BUS_SIZE)
      return *p_orig = new T[len];
    *p_orig = new T[len+1];
    cout << "p_orig: " << int(*p_orig) << ", p: "
      << ((int(*p_orig) % (BUS_SIZE))) << endl;
    return *p_orig;// + (*p_orig % bus_size);
  }




#endif  /*  __RON_BASE_H__ */
/**/

