// $Id: MachZehnder.hh,v 1.2 2008/08/08 02:19:51 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _MACH_ZEHNDER_
#define _MACH_ZEHNDER_

#include "ocsOptSignal.hh"
#include "ocsElecFilter.hh"


enum typeMachZehnder { BINARY_DATA_PERFECT_RECT = 1,
                       BINARY_DATA_ARCTANH_SMOOTHING = 2,
		       BINARY_DATA_SUPERGAUSS_SMOOTHING = 3,
		       SINUSOIDAL_SIGNAL = 4 };


class MachZehnder
{

public:

MachZehnder(string InFileName, OptSignal * oOptSignal2);

void GeneratePerfectRectBinaryElecData(int ChannelArrayIndex);
void GenerateSmoothedRectBinaryElecDataArctanh(int ChannelArrayIndex);
void GenerateSmoothedRectBinaryElecDataSuperGauss(int ChannelArrayIndex);
void GenerateSinusoidalElecSignal(int ChannelArrayIndex);

double SuperGauss(double Time);

void FilterBinaryElecData(void);

void MZModulator(void);
void PhaseModulator(void);

double EAM34PulseShape(double TimeRelPulseCenter);
double EAMPoly(double time);


void WriteFileTimePower(string outFile);
void WriteFileFreqPower(string outFile);

  typeMachZehnder GetTypeMachZehnder(void) { return TypeMachZehnder; };


private:

typeMachZehnder TypeMachZehnder;

BitString * oBitString;
OptSignal * oOptSignal;
ElecFilter * oElecFilter;

cfftw * fft;
fftPowerMatrices sfftPM;
fftMatrices sfftM; 

typeSimulation TypeSimulation;

int NumChannels;
double CenterFreq;

double BitLength; // The time length in seconds of a bit
double BitRateChann;

int StringLength;
double OptPowerChann;

int qtPoints;
double TimeWindow,DeltaTime;
double FreqWindow,DeltaFreq;  
      
double V_pi, V_pi1, V_pi2;
double Phi_10,Phi_20;
double DeltaPhaseModulator;
double OptModulationDepth;
double SlopeFactor;
double ShiftAmt;
bool ChirpFreeFlag;

int SuperGaussExponent;

int RandomPolarizationFlag;

  // Parameters for UMBC_RZ

bool DoEOMFlag;
bool EAMWithHump;

};

#endif /* _MACH_ZEHNDER_ */




