// $Id: 

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _POST_DISP_COMP_SUBSYSTEM_HH_
#define _POST_DISP_COMP_SUBSYSTEM_HH_


#include "ocsOptSignal.hh"
#include "ocsOptAmplifier.hh"
#include "ocsOptFilter.hh"
#include "ocsLumpedLoss.hh"
#include "ocsOptFiberLocalError.hh"
#include "ocsRanNumGen.hh"
#include "ocsOptSpectrumAnalyzer.hh"
#include "ocsOptDataWriter.hh"
#include "NoiseCovariance.hh"

class PostDispCompSubSystem
{

  // ########### Methods ###################

public:

//! Constructor for Ciena system PostDisperionComepsnation, April 2003

   PostDispCompSubSystem(void);

   void InitApril2003(string InFileName,OptSignal * Signal2,RanNumGen * RNG2,
                 OptSpectrumAnalyzer * PowerMeter2,string Job2);

   void InitApril2003(string InFileName,OptSignal * Signal2,
                      NoiseCovariance * NoiseCov,RanNumGen * RNG2,
                      OptSpectrumAnalyzer * PowerMeter2,string Job2);

  PostDispCompSubSystem(string InFileName,OptSignal * SCSignal2,
                        RanNumGen * RNG2,
                        string InDir2,string Job2,
                        int DemuxChannArrayIndex2,OptDataWriter * DataWriter2);

  PostDispCompSubSystem(string InFileName,OptSignal * SCSignal2,
                        RanNumGen * RNG2,
                        string InDir2,string Job2,
                        int DemuxChannArrayIndex2,OptDataWriter * DataWriter2,
			OptSpectrumAnalyzer * PowerMeter2);


  void SetRecMapNum(int Value) {RecMapNum = Value;};

  void ApplyPostDispComp(void);
  void ApplyPostDispCompNoiseCov(void);

private:

  void ApplyPostDispCompUMBCCienaLoopApr2001(void);

  void AllocateMemoryUMBCCienaLoopJuly2001(void);
  void SetAmountPostDispCompJuly2001(void);
  void ApplyPostDispCompUMBCCienaLoopJuly2001(void);
  void ApplyPostDispCompUMBCCienaLoopJuly2001NoiseCov(void);


  void AdditionalDemuxFiltering(void);

  // ########## Data #####################

private:

  string InDir;
  string Job;
  RanNumGen * RNG;

  int PostDispCompMethod;
  int AmountPostDispComp;  // In ps/nm; an int so can use a switch
  int DemuxChannArrayIndex;
  double FreqFWHM_OptFilter;
  int ApplyDemuxFilterFlag;

  OptSignal * SCSignal;
  OptDataWriter * DataWriter;

  // Variables for the July method

  int TypePostAmps;
  int TypePostAmpsNoise;
  int GammaZeroFlag;
  int FiberLossZeroFlag;
  double RelativeErrorGoal;
  int PreAmt;
  int RecMapNum;
  int DispType;
 

  double PostExtraLength;
  double PostExtraLength7,PostExtraLength10,PostExtraLength13;
  double PostExtraLength16,PostExtraLength19;

  NoiseCovariance * NoiseCov; 

  OptSpectrumAnalyzer * PowerMeter;
  OptAmplifier * PostAmpDCF;

  OptAmplifier * PostAmpSMF;
  OptAmplifier * PostAmpFinal;


  OptAmplifier * AmpPtr[3];

  OptFiberLocalError * PostLTS1;
  OptFiberLocalError * PostLTS2;
  OptFiberLocalError * PostLTS3;
  OptFiberLocalError * PostLTS4;
  OptFiberLocalError * PostLTS5;

  OptFiberLocalError * PostLycom1;
  OptFiberLocalError * PostLycom2;
  OptFiberLocalError * PostLycom3;
  OptFiberLocalError * PostLycom4;
  OptFiberLocalError * PostLycom5;

  OptFiberLocalError * PostSMF5807;
  OptFiberLocalError * PostSMF5809;
  OptFiberLocalError * PostSMF5810;

  OptFiberLocalError * PostDCF4_20;
  OptFiberLocalError * PostDCF3_10;
  OptFiberLocalError * PostSMF3_25;
  OptFiberLocalError * PostSMF186;
  OptFiberLocalError * PostSMFs3;
  OptFiberLocalError * PostSMFs4;
  OptFiberLocalError * PostSMF1500;

  OptFiberLocalError * PostExtraSMF;
  OptFiberLocalError * PostExtraDCF;

  int NumFibers;

  OptFiberLocalError * FiberPtr[22];

};

#endif // _POST_DISP_COMP_SUBSYSTEM_HH_











