// $Id: PulseEvolutionAnalyzer.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _PULSE_EVOLUTION_ANALYZER_HH_
#define _PULSE_EVOLUTION_ANALYZER_HH_

#include "ocsBitString.hh"
#include "ocsOptSignal.hh"
#include "ocsOptFiberLocalError.hh"
#include "ocsHistogram.hh"

class PulseEvolutionAnalyzer
{

public:

  PulseEvolutionAnalyzer(OptSignal * PropagatingSignal2, 
                         string InDir, string Job2,
                         bool ComputeTimeFreqShifts2,
                         bool ComputeTimingJitter2,
                         bool ComputeMeanZerosStdDevOnes2,
			 int ComputeEnergyInBitSlot2);

  void SetBitString(BitString * oBitString2);

  void ApplyPulseAnalyzer(double PropagatedLength2);

  void LinearlyCompressPulses(void);

  void IsolatePulse(OptSignal * FullSignal, OptSignal * SinglePulse, 
		    int BitSlotSinglePulse);

  void SeparatePulses(void);

  void ComputeTimeShift(void);
  void ComputeFreqShift(void);
  void ComputeMaxPower(void);
  void WriteTimeFreqShifts(string OutFileName);

void WriteMeanZerosStdDevOnes(string OutFileName);
void ComputePulseEnergyStats(void);
  
  void WriteEnergyInBitSlot(string OutFileName);

bool GetComputeMeanZerosStdDevOnesFlag(void)
  {return ComputeMeanZerosStdDevOnes;};

 
  // ### Private Data ###

private:

OptSignal * PropagatingSignal;
OptSignal * CompressedSignal; // Doubles as LeadingPulse
OptSignal * TrailingPulse;

Histogram * HistogramZeros;
Histogram * HistogramOnes;
BitString * oBitString;

OptFiberLocalError * LinearFiber;

string Job;

double PropagatedLength;

  bool ComputeTimingJitter;
  bool ComputeTimeFreqShifts;
  bool ComputeMeanZerosStdDevOnes;
 
  int ComputeEnergyInBitSlot;

int LeadingPulseBitSlot,TrailingPulseBitSlot;

double LeadPulseTimeShift,TrailPulseTimeShift;
double LeadPulseFreqShift,TrailPulseFreqShift;

double LeadPulseMaxPower,TrailPulseMaxPower;

double LeftTimeZero,LeftTimeOne,RightTimeZero,RightTimeOne;

  double MeanZeros,MeanOnes,StdDevOnes;

}; // end class PulseEvolutionAnalyzer

#endif // _PULSE_EVOLUTION_ANALYZER_HH_


