// $Id: StatisticsOfArray.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


// Created by John Zweck. Sept 25th, 2000.

// This class extracts simple statistics from a double array.
// It's role is different from the Histogram class since in that class
// each time you want to add a sample you make a separate function call.
// In the StatisticsOfArrray class we are computing the statistics 
// of a given array.

// The statistics are computed when the constructor is called.
// They may be accessed directly as the min,max,mean,std_dev and 
// variance are public variables (for simplicity and spped of coding).


#ifndef _STATISTICS_OF_ARRAY_HH_
#define _STATISTICS_OF_ARRAY_HH_

class StatisticsOfArray
{

  // ### Public Methods ###

 public:

  StatisticsOfArray(int arrayLengthIn,  double *arrayIn);
  ~StatisticsOfArray();

  double GetMinimum() {return minimum;};
  double GetMaximum() {return maximum;};
  double GetMean() {return mean;};
  double GetStdDev() {return standardDeviation;};
  double GetVariance() {return variance;};
  double GetMedian() {return median;};

  // ### Private Data ###

 private:

  int arrayLength;

  double * array;
  double * sortedArray;

  double minimum,maximum;
  double mean,standardDeviation,variance;
  double median;


 // ### Private Methods ###

private:

  double Minimum(void);
  double Maximum(void);
  double Mean(void);
  double MeanSquares(void);
  double Variance(void);
  double Median(double * SortedArray, int Length);
 
  void SelectionSort(double * Array, double * SortedArray, int Length);
  int MaximumIndex(double * Array, int Length);

}; // ## end declaration of class  StatisticsOfArray


#endif /* _STATISTICS_OF_ARRAY_HH_ */
