// $Id: ocsElecEqualizer.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################
#ifndef _OCS_ELEC_EQUALIZER_HH_
#define _OCS_ELEC_EQUALIZER_HH_

#include "ocsPhotodetector.hh"
#include "ocsElecSignalStat.hh"

// ##### Class ElecEqualizer
class ElecEqualizer
{
   public:
      ElecEqualizer(string InFileName, 
                    ElecSignalStat *oElecSignalStat2 );
      ElecEqualizer(string InFileName, 
                    ElecSignalStat *oElecSignalStat2,
                    ElecSignalStat *oElecSignalStatImag2);
      void InitElecEqualizer(string InFileName, 
                             ElecSignalStat *oElecSignalStat2);
      ~ElecEqualizer();
      void AllocateMemory(void);
      void ResetEqualizer(void);
      void EqualizeBitIntensityWiener(void); 
      void EqualizeBitIntensityWienerReal(void); 
      void EqualizeBitIntensityWienerCplx(void); 

      void EqualizeBitIntensityLMS( int StringLength 
                                   ,double *DistortedBitIntensity
                                   ,double *EqualizedBitIntensity 
                                   ,double *Back2BackBitIntensity
                                   ,double StepSize); 
      void EqualizeBitIntensityWiener( int StringLength 
                                   ,double *DistortedBitIntensity
                                   ,double *EqualizedBitIntensity); 
      void EqualizeBitIntensityWiener( int StringLength 
                                   ,double *DistortedBitIntensity
                                   ,double *EqualizedBitIntensity 
                                   ,double *Back2BackBitIntensity); 
      double GetEffectiveAmpMargin(ElecSignalStat *oElecSignalStat2);
      void WriteFileEyeDiagram(string outFile);
      double GetTapWeight(int IndexTap);

      ElecSignalStat *oElecSignalStat;
      ElecSignalStat *oElecSignalStatIm;
   private:
      int qtInputTaps;
      int EqualizationTap;
      int FeedBackCorrectedBits;
      int UseFinalWeights;
      int NormalizeWeights;
      int SymmetricEqualization;
      int IsComplexEqualization;
      int UseExactDesired;
 
      int StringLength;
      double MeanBitIntensity;
      double MeanBitIntensityIm;
      int    *CorrelationVectorQtAverage;
 
      //## Pure real equalizer
      double *TapWeight;
      double **CorrelationMatrix;
      double *CorrelationVector;
      double *CrossCorrelationVector;
      double **TapWeightTmp;

      //## Complex equalizer
      cplx *TapWeightCplx;
      cplx **CorrelationMatrixCplx;
      cplx *CorrelationVectorCplx;
      cplx *CrossCorrelationVectorCplx;
      cplx **TapWeightTmpCplx;
};
#endif /* _OCS_ELEC_EQUALIZER_HH_ */
