// $Id: ocsElecFilter.hh,v 1.2 2005/07/06 16:27:46 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_ELEC_FILTER_HH_
#define  _OCS_ELEC_FILTER_HH_

#include "ocsOptSignal.hh"
#include "ocsPhotodetector.hh"


//! The type of Electric filter
/*! An enumeration type
    -# NO_ELEC_FILTER = 0
    -# GAUSSIAN = 1, Must specify width.
    -# BESSEL = 2,  Must specify order and width.
    -# MEASURED = 3, Must specify file name of measured data.
    -# INTEGRATE_AND_DUMP = 4 
*/

enum typeElecFilter { NO_ELEC_FILTER = 0, 
                      GAUSSIAN = 1, 
		      BESSEL = 2,
		      MEASURED = 3,
                      INTEGRATE_AND_DUMP = 4,
		      ANRITSU = 5,
                      GAUSSIAN_AND_ANRITSU = 6}; 

// ##### Input parameter to initialize Electrical Filters
struct ElecFilterParam
{
   int TypeElecFilter;
   int OrderElecFilter;
   double FreqFWHM_ElecFilter;
   double CenterFreqElecFilter;
};

// ##### Class Electrical filter

class ElecFilter
{
   public:
      ElecFilter();
      void InitElecFilter(cfftw *fft2
                      ,fftPowerMatrices sfftPM2
                      , ElecFilterParam ElecFilter);

      ElecFilter(string InFileName, Photodetector *oPhotodetector2);

      ElecFilter(ElecFilter *oElecFilter2, Photodetector *oPhotodetector2);

  //! Use this constructor if you have a measured electric filter and 
  //! many receivers and want to save memory.

      ElecFilter(string InFileName, Photodetector *oPhotodetector2,
		 cplx * MeasuredFilterFreq2,bool ReadMeasuredFilterFlag);


      ElecFilter(string InFileName, cfftw *fft2,fftPowerMatrices sfftPM2);
  
      void FilterElecSignal(void);
 
      void SetPhotodetector(Photodetector *oPhotodetector2);

	 // add by Jonathan Hu
	  string GetAnritsuFilterFileName(void)
            {return AnritsuFilterFileName;};
			
      typeElecFilter GetTypeElecFilter(void)
            {return TypeElecFilter;};

      int GetOrderElecFilter(void)
            {return OrderElecFilter;};

      double GetCenterFreqElecFilter(void)
            {return CenterFreqElecFilter;};	    	    	    

      double GetFreqFWHM_ElecFilter(void)
            {return FreqFWHM_ElecFilter;};
      cplx GetMeasuredFilterFreq(int ii) 
            {return MeasuredFilterFreq[ii];};
      void SetFreqFWHM_ElecFilter(double FreqFWHM_ElecFilter2);
      
      void Set_f_3dB_ElecFilter(double f_3dB_ElecFilter2);
      
      double Get_f_3dB_ElecFilter(void)
            {return FreqFWHM_ElecFilter/2.;};

      void WriteFreqPowerGaussianFilter(string OutFileName);
      
      //writes to file the impulse response of the filter
      //that is the output signal when the input signal is an impulse
      //at the center of the time window
      void WriteTimePowerFilter(string OutFileName);
      // writes to file the frequency response of the filter
      void WriteFreqPowerFilter(string OutFileName);

      double ComputeMatchAndFilter(void);

      
     void GetClockRecoveryTime(BitString *oBitString); // By Wei Wang
     void WriteFreqCplxFilter(string OutFileName);

     double GetCurrentEquivalentSpectralWidth(void);
     void WriteFilterImpulseResponse(string FileNameImpulseResponseTime,
                                     string FileNameImpulseResponseFreq);

  int GetDebugLevel(void) {return DebugLevel;};

  // ######## PUBLIC DATA ##################

  //! Array of length StringLength to store the Average Current in each bit.
  //! Only used with the INTEGRATE_AND_DUMP filter

   double * AverageCurrentInBit;

  // ########## PRIVATE METHODS ##################
   
   private:

     void ElectricalFilterB5(cplx *my_power_time,cfftw *fft, 
                             double omega_0, double fwhm_norm);
     void ElectricalFilterB4(cplx *my_power_time,cfftw *fft, 
                             double omega_0, double fwhm_norm);
     void ElectricalFilterGaussian(cplx *my_power_time, 
				   cfftw *fft, double omega_0, 
                                   double fwhm_norm);

     //! Added by Anshul June'03.
     
     void IntegrateAndDumpFilter(cplx *my_power_time,cfftw *fft);

  //! This method filters the electrical signal using a measured
  //! electrical filter
  /*! The filter shape is read in from the file 
      string MeasuredElecFilterFileName using the method 
      ReadMeasuredFilterData. \n
      The filter is applied in the freq domain after which we
      convert to the time domain,take the real part and update
      in frequency just to be safe.
  */

     void ElectricalFilterMeasured(void);

  //! Reads in the measured data of an electrical filter 
 
     void ReadMeasuredFilterData(void);

     void ReadMeasuredFilterData(cplx *MeasuredFilterFreq2);

   // the following function is writen by Jonathan Hu
	 double GetPhase(double Frequency);
	 void ReadElecPhase(string InFileName);
	 void ReadElecPhasePaola(string InFileName);
	
	 double GetElecGain(double Frequency);
	 void ReadElecGain(string InFileName);

	 double * xElecGain;
	 double * yElecGain;
	 double * y2ElecGain;
	 double ElecGainStartFreq;
	 double ElecGainStepFreq;
	 int NumElecGainWidth;

	 string AnritsuFilterFileName;

	 double * xElecPhase;
	 double * yElecPhase;
	 double * y2ElecPhase;
	double ElecPhaseStartFreq;
	double ElecPhaseStepFreq;
	int NumElecPhaseWidth;
	int RunAnritsuFilter;
	void AnritsuFilter(cplx *my_power_time, cfftw *fft, double omega_0, 
                double fwhm_norm);
	void AnritsuFilterPaola(cplx *my_power_time, cfftw *fft, 
                double omega_0, 
                double fwhm_norm);
	double GainPeak;
 	double FreqGain3dB;
	
	double splint( double *xa,  double *ya,  double *y2a, int n, double x);
	void spline(double *x,  double *y,  int n, double yp1, double ypn,  double *y2);
	double GaussianSmooth(double *y, int Order, int N );

	// up to here (JH)
	

 //! An integrate and dump electric filter as used in Marcuse paper

  /*!
     This filter replaces the signal by a piecewise constant function
     whose values are constant in the bit slots. The k-th bit slot
     is assumed to go from [k*BitWindow, (k+1)*BitWindow]. So we need 
     the channel we are filtering to have its carrier frequency equal 
     to the center frequency of the discretized signal.  
     In a given bit slot the constant value is
     equal to the average of the  current within that bit slot, which is
     the time integral of the photo-current over the bit slot divided
     by the length of the bit window. (Here we follow Marcuse JLT Vol 9 No 4
     1991, equation (19). 

     This method also computes the array AverageCurrentinBit of length 
     StringLength which stores the constant average currents

     We need qtPoints = integer * StringLength

     We assume that the width of the recangular optical filter is equal to 
     the FreqWindow of the signal. This is OK as long as there are no
     non-linear effects and as long as there is only one channel
     and the optical filter is not too narrow. You could always write
     your own recangular optical filter in the class OptFilter if these
     assumptions are not satisfied in your application.  
   */

     void IntegrateAndDumpFilter_OLD(void);
 
     void AllocateMemoryForAverageCurrentinBit(void);

     double GetAverageCurrentInTimeRange(double LeftTime, 
                                         double RightTime);
//	 int GetBitWeightCenter(int index);


     cplx Bessel5(double omega, double omega_0, double FWHM);
     cplx Bessel4(double omega, double omega_0, double FWHM);
     double Gaussian(double omega, double omega_0, double FWHM);

    //! Added by Anshul

    cplx IntegrateAndDump(double freq); 

  //! Converts the ElecFilter's sfftPM data from time to frequency.

     void Time2Freq(void);
 
  //! Converts the ElecFilter's sfftPM data from frequency to time.

    void Freq2Time(void);

   // ########## PRIVATE DATA ##################
  
      fftPowerMatrices sfftPM;
      cfftw *fft;   
      int ObjectInitialized;                  
      int qtPoints;

      double DeltaFreq, DeltaTime,FreqWindow, TimeWindow, BitLength; 
      double ClockRecoveryTimeForIntegrateAndDump;
      int StringLength;


  //! The type of Electric filter. 
  /*! Possible values of this enumeration type are:
    -# NO_ELEC_FILTER = 0.
    -# GAUSSIAN = 1. Must specify width.
    -# BESSEL = 2  Must specify order and width.
    -# MEASURED = 3 Must specify file name of measured data.

*/
                 
      typeElecFilter TypeElecFilter;
      int OrderElecFilter;
      double FreqFWHM_ElecFilter; 

      // FreqFWHM_ElecFilter is the FWHM of the power of the filter
      // in the frequency domain. (JZ)

      double CenterFreqElecFilter;  

  //! Name of the file containing the data for the measured filter 
  /*! \sa TypeElecFilter */

      string MeasuredElecFilterFileName;

  //! Complex array to store frequency data for the Measured Filter.

      cplx * MeasuredFilterFreq;
      Photodetector *oPhotodetector;

      int DebugLevel;

   
};

#endif /*  _OCS_ELEC_FILTER_HH_ */







