// $Id: ocsHistogram.hh,v 1.2 2005/07/06 16:27:58 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################


#ifndef _OCS_HISTOGRAM_HH_
#define  _OCS_HISTOGRAM_HH_


#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsReducedModelTools.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"


// ##### Class Histogram

//! The Histogram class is used to compute statistics

/*! The Histogram class is used to compute statistics and 
 probability density functions. It can be used for 
 standard Monte Carlo simulations or Monte Carlo simulations
 with importance sampling
*/

class Histogram
{
   public:

  //! Vacuous constructor that can be used when constructing an array
  //! of Histogram objects

      Histogram() {};

  //! Histogram constructor

      Histogram(int qtBins2,double MinRange2, double MaxRange2);

 //! Histogram constructor: Reads parameters from a file

      Histogram(string InFileName);

 //! Like Histogram::Histogram(string InFileName), but can
 //! be used to initialize an element of an array of Histograms

      void InitHistogram(string InFileName);

 //! Histogram destructor
  /*!
     \sa ReleaseMemory
   */
      ~Histogram();

  //! Called by destructor ~Histogram()
      void ReleaseMemory(void);


  //! Reinitializes histogram. 
      void ClearHistogram(void);


  //! Adds a sample to the histogram. 
  /*! Used for standard Monte Carlo simulation.
      Calls AddImportantSample with parameter LikelihoodRatio = 1.
   \param vSample: the value of the sample to be added (double)
   \sa AddImportantSample
  */

      void AddSample(double vSample);

  //! Adds a sample to the histogram. Used in importance sampling.

      void AddImportantSample(double vSample, double LikelihoodRatio);

  //! Imports data from a file. Used when merge biased pdfs in 
  //! importance sampling.

      void ImportData(string InFileName, int Column,int qtColumns);

  //! Imports data from a file. Used when merge biased pdfs in 
  //! importance sampling.

      void ImportData(string InFileName);


      void SetFileNameWithSamples(string FileNameWithSamples2);


  //! Converts a histogram into a probability density function (pdf)

      void Obtain_pdf(void); // the pdf will be pdfY:  pdfX x pdfY



      //void InitBalanceHeuristicDataStorage(double BiasingParam2,
      //                                     int qtRamdomSourcesPerSample2); 

     //! The method of choice for merging in multiple importance sampling

      void MultipleHeuristicBalanceImport(string InFileName, 
                                          int Convert2dB_Scale);

/*! 
     Computes the probability density function 
     f(X=x| z_min < Z < z_max), 
     Samples of the random variables X,Z have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

      void MergeMultipleConditionalHeuristicBalance(string InFileName, 
					  double z_min, double z_max,
                                          int Convert2dB_Scale);


/*! 
     Computes the probability density function 
     f(X=x| z_min < Z < z_max, w_min < W < w_max), 
     Samples of the random variables X,Z,W have been obtained from an 
     importance sampling simulation, and that the names of the files containing
     the samples are stored in the file InFileName. 
   */

      void MergeMultipleConditionalHeuristicBalance(string InFileName, 
					  double z_min, double z_max,
					  double w_min, double w_max,
                                          int Convert2dB_Scale);


      //void MultipleBalanceHeuristicImport(string InFileName); //## Best


  //! NOT USED AT MOMENT

      void MultipleWeightedImport(string InFileName);

  //! Merges biased pdfs created using importance sampling into a 
  //! single pdf. This is Ivan Lima's Maximum Number of Hits Heuristic.

  /*! 

      The biased pdf's are assumed to have been previously 
      generated by calls to methods like
      Histogram::WriteFile_pdf(string OutFileName).
      

     \param InFileName: Name of file that contains a list of the files 
            containing the biased pdf's. All biased pdf's are assumed to
            have the same discretized domain.

     \param Convert2dB_Scale: Flag that specifies whether or not
            to convert the pdfs to be merged
            from a linear to dB scale.

      The method does the following:

      -# Reads in the file names of the pdf's from InFileName.
         You can have up to 100 files.
      -# For each input pdf construct a Histogram object and 
         add (Sample, LikelihoodRatio) pairs to the histogram
         by reading in data from the file
      -# Call Histogram::Obtain_pdf for each input pdf.

      Next for each bin in the domain of the histograms:

      -# For each of the input biased pdf's compute the number of samples
         in the bin relative to the total number of samples in the
         input biased pdf. Call this number the Relative Number of Hits
         in the given bin for the given biased pdf.
      -# For this histogram, (the one that is calling this method)
         set the value of the bin in question to be the maximum of the
         Relative Number of Hits, where we maximize over all the biased pdfs.

      Finally, we normalize the histogram by forcing the number of samples
      to be 1. This will give us the correct statistics WHY???.

      Note that because of the way the histogram has been constructed
      we can't use the methods GetMean(), GetMax(), GetQtSamples(), etc.

   */

      void MultipleSelectMaxRelNumHitsImport(string InFileName,
                                             int Convert2dB_Scale);

  //! This method appears to be outdated code and is not used by the 
  //! reduced model

      void MultipleSelectiveImport(string InFileName);

  //! NOT USED AT MOMENT

      void ConditionalMultipleSelectiveImport(string InFileName,
                                              int Convert2dB_Scale);
  //! NOT USED AT MOMENT

      void MultipleImport(string InFileName, int Convert2dB_Scale);

  //! By JZ Aug 2004

  /*! Imports the data from files created by the histogram class
    and merge multiple sets of files.
    This method is useful to merge all the samples of independent 
    unbiased simulations, i.e., we assume LikelihoodRatio=1.0

    By: Ivan Lima (12/5/01). Adapted by JZ Aug 2004
  */

      void MultipleUnbiasedImport(string InFileName);


  //! Writes the probability density function (pdf) computed by the
  //! histogram object to a file

      void WriteFile_pdf(string OutFileName);

  //! Writes the cumulative density function (cdf) computed by the
  //! histogram object to a file

      void WriteFile_cdf(string OutFileName);

  //! Writes the complement of the cumulative density function (cdf) 
  //! computed by the
  //! histogram object to a file

      void WriteFile_cdfc(string OutFileName);

  //! Write the number of hits per bin to a file

      void WriteFileQtSamplesBin(string OutFileName);

  //! Write the values in the bins to a file

      void WriteFileBinValues(string OutFileName);


  //! Old code. Do not use!

      void WriteFilePDF(string OutFileName);

  //! Computes a  Maxwellian pdf with same mean as Histogram's pdf 
  //! and writes it to a file

      void WriteFileMaxwell_pdf(string OutFileName, int qtPoints);    
 
  //! Not used by Reduced Model!

     void WriteFileSndOrderPDV_pdf(string OutFileName, int qtPoints);
  
  //! Computes a  Gaussian pdf with same mean and standard devaition
  //! as Histogram's pdf 
  //! and writes it to a file

      void WriteFileGaussian_pdf(string OutFileName, int qtPoints);      
      void WriteFileGaussian_cdf(string OutFileName, int qtPoints);
      void WriteFileGaussian_cdfc(string OutFileName, int qtPoints);

      int GetQtBins(void) {return qtBins;};
      double GetMinRange(void) {return MinRange;};
      double GetMaxRange(void) {return MaxRange;};
      long GetQtSamples(void) {return qtSamples;};     
      long GetQtSamplesInRange(void) {return qtSamplesInRange;};        

      void SetBin(int NumBin, double ValueBin) 
            { Bin[NumBin] = ValueBin;}; //## Be careful with this!!!
      double GetBin(int NumBin) {return Bin[NumBin];};
      double Get_cdfBin(int NumBin) {return cdfBin[NumBin];};
      double Get_cdfcBin(int NumBin) {return cdfcBin[NumBin];};            
      int GetIndex_pdf(double vSample);      
      long GetQtSamplesBin(int NumBin) {return qtSamplesBin[NumBin];};
      double GetSumLikelihoodRatio(void) {return SumLikelihoodRatio;};
      double GetMean(void) {return MeanTmp
                            /SumLikelihoodRatio;};
      double GetMeanSquare(void) {return MeanSquareTmp
                                  /SumLikelihoodRatio;};

  //! Get the standard deviation where normalize by qtSamples - 1

      double GetStdDev(void);      
      double GetMinSample() {return MinSample;};
      double GetMaxSample() {return MaxSample;};

  //! Get the sum of all the probabilities. Use for debugging

      double GetSumProbabilities(void); 

      double Get_pdfX(int ii) {return pdfX[ii];};
      double Get_pdfY(int ii) {return pdfY[ii];};
      double Get_pdfConfidence(int ii) {return pdfConfidence[ii];};
      double Get_cdfConfidence(int ii) {return cdfConfidence[ii];};      
      double Get_cdfcConfidence(int ii) {return cdfcConfidence[ii];};       
   private:   

  //! Private method called by Histogram constructors
  /*! Reads in parameters and calls method ClearHistogram  
      \sa qtBins MinRange MaxRange
          ClearHistogram    
          Histogram(int qtBins2 ,double MinRange2,double MaxRange2)
          Histogram(string InFileName)
  */

      void InitHistogram(int qtBins2,double MinRange2, double MaxRange2);



      int ObjectInitialized;

      int Index_pdf;

      //! Flag that controls whether (sample,likelihood ratio)-pairs are
      //! written to the output file named string FileNameWithSamples.
      //! Used in importance sampling
      //! \sa Histogram(string InFileName)
      //! \sa FileNameWithSamples AddImportantSample

      int MakeFileWithSamples;

      
      //! Name of the output file to which (sample,likelihood ratio)-pairs
      //! are written. 
      //! Used in importance sampling.
      //! \sa MakeFileWithSamples AddImportantSample

      string FileNameWithSamples;

  //! Name of the input parameter file for Histogram class
  /*! \sa Histogram(string InFileName)
   */

      string HistInFileName;

  //! The number of bins in the histogram
      int qtBins;

      int qtBinsLast;

  //! The minimum value in the domain of the histogram (double)

      double MinRange;

 //! The maximum value in the domain of the histogram (double)

      double MaxRange;
  
  //! Double array of length qtBins used to store the binned samples
  //! of the histogram, weighted by their likelihood.
  /*! 
      Each bin of the histogram has an associated indicator function.
      The indicator function I_i of the i-th bin is a binary function
      on samples. It takes the value 1 if the sample is placed
      in the i-th bin and the value 0 otherwise. In other words I_i is
      a binary random variable. The mean of I_i gives the probability 
      of being in the bin. The standard deviation of I_i is a measure
      of how close the mean of I_i is to the true value of the pdf.
      The confidence interval of the i-th bin in a histogram is 
      [mean(I_i) - std_dev(I_i), mean(I_i) + std_dev(I_i)]. 
      There is a large chance that the true probability of being in
      the i-th bin lies in the confidence interval. So the smaller
      std_dev(I_i)/mean(I_i) the more accurately the histogram estimates
      the true pdf. 

      In standard Monte Carlo the probability of being in [a,b] is the
      sum over all samples of the indicator function of [a,b] evaluated
      at the sample, divided by the number of samples.

      In importance sampling the probability of being in [a,b] is the
      sum over all samples of the product of the 
      indicator function of [a,b] evaluated
      at the sample and the likelihood that the sample
      would actually occur, divided by the number of samples.
   */
      double *Bin;


  //! Double array of length qtBins used to store the squares of the
  //! binned samples of the histogram, weighted by their likelihood.
  /*! Used to compute confidence intervals. 
      \sa Bin
  */

      double *BinSquare;

  //! Integer array containing the number of samples that have been 
  //! placed in a bin.
  /*! For standard Monte Carlo qtSamplesBin[i] = Bin[i], but this
      is not true for importance sampling, since the binned samples
      are weighted by their Likelihood ratio.
  */

   double *cdfBin;
   
   double *cdfcBin;  //## For the computation of the outage probability
   
   double *cdfBinSquare;
   
   double *cdfcBinSquare;

      long *qtSamplesBin;

  //! The total number of samples used to construct the histogram

      long qtSamples;

  //! The total number of samples that fall in the domain of the histogram

      long qtSamplesInRange; 

  //! The sum of all likelihood ratios. Used in importance sampling
      double SumLikelihoodRatio;  

  //! Used only in standard Monte Carlo. Is the same as number of samples.
  //! Not used in code.
      double SumUnitaryLikelihoodRatio;

      

  //! Double array of X-values of pdf. Computed by Obtain_pdf()
      double *pdfX;

  //! Double array of Y-values of pdf. Computed by Obtain_pdf()
  //! \sa Bin
       double  *pdfY;
       
  //! Double array giving size of confidence interval. Computed by Obtain_pdf()
  /*! The confidence intervals are not computed in all cases.
      \sa Bin
  */
       double  *pdfConfidence;             
       int PreviouslyComputed_pdfConfidenceFlag;
       
       double *cdfConfidence;
       
       double *cdfcConfidence;

       double MeanTmp, MeanSquareTmp;

  //! The minimum value of all samples. May lie outside domain of histogram!
      double MinSample;
  //! The maximum value of all samples. May lie outside domain of histogram!
      double  MaxSample;   
      ofstream *OutFileWithSamples;
      double *ExpPolynomialCoef;

  //! DebugLevel = 0 gives no output to LogFile

      int DebugLevel;
      
       /*
      //## Extra variables for the Heuristic Balance
      double BiasingParam;
      int qtRandomSourcesPerSample; 
      int BalanceHeuristicInitialized;
      */
};      


#endif /*  _OCS_HISTOGRAM_HH_ */







