// $Id: ocsLumpedLoss.hh,v 1.2 2006/05/11 23:44:58 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_LUMPED_LOSS_HH_
#define _OCS_LUMPED_LOSS_HH_


#include "ocsOptSignal.hh"
#include "ocsReduLumpedLoss.hh"

// Written by John Zweck Jan 5, 2001.

// Sometimes the only thing you want to model about an optical device
// is its loss. 
// This class is used to model the loss of such optical devices 
 
// It could be easily extended to model the loss of electrical devices
// by adding another constructor which is passed a Photodetector
// rather than an OptSignal.



class LumpedLoss
{
   public:

   LumpedLoss(double LossdB2, OptSignal *oOptSignal2);
   LumpedLoss(string InFileName, OptSignal *oOptSignal2);
   
   void InitLumpedLoss(double LossdB2, OptSignal *oOptSignal2);   

   void SwitchOptSignalData(OptSignal *oOptSignal2);
   void ApplyLumpedLoss();
   void SetLossdB(double LossdB2);
   double GetLossdB(void) {return LossdB;};   
   double GetLossLinear(void) {return LossLinear;};

   private:
   
   typeSimulation TypeSimulation;
   double LossdB;
   double LossLinear;
   double LossLinearSqrt;
   OptSignal * oOptSignal;
   fftMatrices sfftM;
   ReduLumpedLoss * oReduLumpedLoss;
};



#endif /* _OCS_LUMPED_LOSS_HH_ */
