// $Id: ocsOptDataWriter.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_OPT_DATA_WRITER_HH_
#define _OCS_OPT_DATA_WRITER_HH_

#include "ocsOptSignal.hh"
#include "ocsOptSpectrumAnalyzer.hh"
#include "PulseEvolutionAnalyzer.hh"

// ##### Class OptDataWriter ############

/* This class was written on May 2nd by John Zweck to control the output
   of optical data. The flags and methods declared below should give you
   a good idea of the different sorts of optical data which are handled
   by this class. Note that OSA output and SNR data are not handled by this
   class. Neither is any electrically related output, which is already
   nicely handled by the ReceiverSubsystem class. (JZ)
   The methods assume that the first Monte Carlo experiment is numbered 0.
*/


class OptDataWriter
{

  // ####### Public Methods ########

   public:

  OptDataWriter(string InFileName,string InDir2,
                string Job2, OptSignal * Signal2);

  void WriteInitialData(void);
  void WriteWithinMap(int expt_num, int map_number, double PropagatedLength);
  void WriteEndMap(int expt_num, int map_number, double PropagatedLength);
  void WriteEndTransmission(int expt_num, int map_number, 
                            double PropagatedLength);
  void TurnOffWriteFlags(int expt_num);    

  int GetWritePulseWidthInFiberFlag() {return WritePulseWidthInFiberFlag;};
  int GetOutputStepSizesFlag() {return OutputStepSizesFlag;};

  bool GetComputeMeanZerosStdDevOnes() {return ComputeMeanZerosStdDevOnes;};
  int GetComputeEnergyInBitSlot() {return ComputeEnergyInBitSlot;};

  //#### Public Data ########


  PulseEvolutionAnalyzer * PEA;

  // ########## Private Data ##########

private:

OptSignal * Signal;
string Job;

OptSpectrumAnalyzer * PowerMeter;


bool WriteComplexFieldsFlag;
bool WritePowerChannelFlag;
int  ChannelArrayIndexForPower;
bool WritePowerFlag; 

bool WriteDistanceTimePowerFlag;
bool WriteDistanceTimePowerWithinMapFlag;
bool WriteDistanceTimePowerPhaseFlag;
bool WriteDistanceTimePowerPhaseWithinMapFlag;
 
bool WritePowerMapFlag;
bool WriteAccDispFlag;

bool WritePulseWidthFlag;
bool WriteAverageChirpFlag;

bool WriteLocalFreqFlag;

bool WriteInitialOptDataFlag;
bool WriteFinalOptDataFlag;  
       // To write optical data at end of first Monte Carlo experiment

int NumSlicesZ; // Required for DistanceTime and DistanceTimePhase output
int TimeIndexIncrement;

bool OmitDCFlag; // For writing out frequency data

  bool ComputeTimeFreqShifts;
  bool ComputeTimingJitter;
  bool ComputeMeanZerosStdDevOnes;
  int ComputeEnergyInBitSlot;

int OutputStepSizesFlag;
int  WritePulseWidthInFiberFlag;

bool WriteDOPFlag;
double RelMinFreqDOP,RelMaxFreqDOP;


};

#endif /* _OCS_OPT_DATA_WRITER_HH_ */









