// $Id: ocsOptFilter.hh,v 1.2 2005/07/06 16:28:18 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_OPT_FILTER_HH_
#define _OCS_OPT_FILTER_HH_

#include "ocsOptSignal.hh"


enum typeOptFilter { NOFILTER = 0, SUPERGAUSS = 1, 
                     SANTEC = 2, BOTH_SANTEC = 3 ,
                     BANDPASS = 4,FABRY_PEROT = 5, 
		     CIENA_DEMUX = 6 };


// ##### Input parameter to initialize Optical Filters
struct OptFilterParam
{
   typeOptFilter TypeOptFilter;
   int OrderOptFilter;
   double FreqFWHM_OptFilter;
   double CenterFreqOptFilter;
};

// ##### Class OptFilter
class OptFilter
{
   public:

    //#### Methods by Ivan Lima:

      OptFilter();
      void InitOptFilter(cfftw *fft2, fftMatrices sfftM2
                          ,OptFilterParam soOptFilter);

      //######################################
      //## Copy constructor 
      OptFilter(OptFilter *oOrigOptFilter, OptSignal *oOptSignal2);
      
      void FilterOptSignal(); 


      void FilterOptSignal_OLD();
      void SwitchOptSignalData(OptSignal *oOptSignal2);

     typeOptFilter GetTypeOptFilter(void)
           {return TypeOptFilter;};
     int GetOrderOptFilter(void)
           {return OrderOptFilter;};
     double GetRelativeCenterFreqOptFilter()
           {return RelativeCenterFreqOptFilter;};
     double GetFreqFWHM_OptFilter(void)
           { return FreqFWHM_OptFilter;};
     void SetFreqFWHM_OptFilter(double FreqFWHM_OptFilter2);	      

     //###### Methods by John Zweck:

     // Use this constructor:

     OptFilter(string InFileName, OptSignal *oOptSignal2);

     // Two Constructors for tunable optical filter designed to be used in 
     // conjunction with OptDemuxer (Added by JZ):

     OptFilter(string InFileName,typeSimulation TypeSimulation2); 

  
     OptFilter(int TypeOptFilter2,int OrderOptFilter2,
               double FreqFWHM_OptFilter2,typeSimulation TypeSimulation2);
  
 
     double GetPowerEquivalentSpectralWidth(void);
     void WriteFilterImpulseResponse(string FileNameImpulseResponseTime,
                                     string FileNameImpulseResponseFreq);
	       
     void SetRelativeCenterFreq(double RelativeCenterFreqOptFilter2); 
                                                         
 
     void SetOpticalData(cfftw * fft2, fftMatrices sfftM2); 

     void GaussianTunableOpticalFilter(cplx * TimeSignal, cplx * FreqSignal);

  /*! By Oleg Sinkin. The frequency response function of this filter
      is H(f) = exp( -1/2*((f-fc)/sigma)^(2n)), where fc is the
      center frequency, sigma is a width parameter related to FWHM
      by sigma = 0.5*FWHM/((ln(2))^(1/(2n))) and n is the order of the super
      Gaussian function. The choice n=1 corresponds to a standard Gaussian
      
      To use, call the constructor     
      OptFilter::OptFilter(string InFileName,typeSimulation TypeSimulation2)
      and set OptFilter::TypeOptFilter = 1 OptFilter::OrderOptFilter = n
  */
 
     void SuperGaussianTunableOpticalFilter(cplx * TimeSignal, 
					    cplx * FreqSignal);


    /*! Added by Anshul June'03
     The Bandwidth of the Bandpass filter is assumed to be equal to its
     FWHM. 
    */

     void OpticalBandpassFilter(cplx * TimeSignal, cplx * FreqSignal);
	  
     /*! Added by Anshul Apr 04.
	   
		3 mirror fabry perot filter.
		Reference : PTL vol 8 no 9 Sep 96. Freire, Carvalho 
		and da Silva 
	   
	*/  

     void FabryPerotFilter(cplx * TimeSignal, cplx * FreqSignal);

     void SantecTunableOpticalFilter(cplx * TimeSignal, cplx * FreqSignal);
   
     //! Added by Paola Griggio for Ciena experiment

     void DemuxLoopTunableOpticalFilter(cplx * TimeSignal, cplx * FreqSignal);

     void TunableFilterOptSignal();

     void WriteFreqPowerGaussianFilter(string OutFileName);

     //! By Oleg Sinkin. Outputs the frequency response |H(f)|^2 of
     //! a super Gaussian optical filter

     void WriteFreqPowerSuperGaussianFilter(string OutFileName);
     
     //! by Anshul. writes |H(f)|^2 for a bandpass filter.
     void WriteFreqPowerBandpassFilter(string OutFileName);

     //! by Anshul. writes |H(f)|^2 for a FabryPerot filter.
     void WriteFreqPowerFabryPerotFilter(string OutFileName);

     void WriteFreqPowerSantecFilter(string OutFileName);
     double SantecPoly(double x);

     //! Added by Paola Griggio for Ciena experiment      

     void WriteFreqPowerDemuxLoopFilter(string OutFileName);

     //! Added by Paola Griggio for Ciena experiment 

     double DemuxLoopPoly(double x);
     
     bool GetDoAdditionalFiltering(void) {return DoAdditionalFilteringFlag;};


   private:
      OptSignal *oOptSignal;
      cfftw *fft;
      fftMatrices sfftM;
      int ObjectInitialized;                          
      typeOptFilter TypeOptFilter;
      int OrderOptFilter;
      typeSimulation TypeSimulation;
      double FreqFWHM_OptFilter;
      double RelativeCenterFreqOptFilter;

     // For Ciena work:   
    bool DoAdditionalFilteringFlag; 
   
  int DebugLevel;

};

#endif /* _OCS_OPT_FILTER_HH_ */
