// $Id: ocsPhotodetector.hh,v 1.2 2008/08/08 02:22:28 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_PHOTODETECTOR_HH_
#define _OCS_PHOTODETECTOR_HH_

#include "ocsOptSignal.hh"

// ##### This structure simplifies the access to the field matrices

#ifndef _PHOSSIL_FFT_POWER_MATRICES_
#define _PHOSSIL_FFT_POWER_MATRICES_

struct fftPowerMatrices 
{
      cplx* tPowerSignal;    
      cplx* fPowerSignal;
};

#endif /* _PHOSSIL_FFT_POWER_MATRICES_ */

// ##### Class Photodetector
class Photodetector
{
   public:
      Photodetector();
      Photodetector(string InFileName,OptSignal *oOptSignal2);
      Photodetector(string InFileName,OptSignal *oOptSignal2,RanNumGen *RNG2);

     // Use this constructor when you have multiple receivers and you
     // want to save memory:

      Photodetector(string InFileName,OptSignal *oOptSignal2,RanNumGen *RNG2,
                    fftPowerMatrices & sfftPM2,fftPowerMatrices & WorksfftPM2);

      Photodetector(Photodetector & OrigPhotodetector);
	                                           // Copy constructor

      Photodetector & operator=(Photodetector & RHS_Photodetector);
      
      void SetOptSignal(OptSignal *oOptSignal2);

      ~Photodetector();
      void ReleaseMemory();
      void InitPhotodetector();
      void DetectOptSignal();

      void AddShotNoise(void);
      void AddShotNoiseOLD(void);

      double GetRFPower(double Freq);
      double GetRFPhase(double Freq);
      double GetRFPowerFromFreqIndex(int FreqIndex);
      double GetRFPhaseFromFreqIndex(int FreqIndex);

      void ComputeWorkDifferentiatedThenSquaredElecSignal(void);
      void ComputeWorkContEXORElecSignal(void);
      void ComputeWorkBinaryEXORElecSignal(void);
      void ComputeWorkSumDelayedHalfBitPeriodThenSquare(void);

      //double GetAverageCurrent(void);
      double GetAverageElecPower(void);
      double GetAverageElecCurrent(void);

      double GetAverageAbsElecPower(void);

      double GetMaxCurrentIsolatedBit1(void);
      double GetSignalBandwidth(void);

      void InitPhotodetector(cfftw *fft2
                           ,fftMatrices sfftM2
                           ,double ConvFactOptElecPhotod2);      

  //! Used in ChiSquareReceiver class to construct the
  //! transfer function of the electric filter. JZ May 2004

      void GenerateImpulseCurrent(void);
  
      void Time2Freq(void);
      void Freq2Time(void);
      void WorkTime2Freq(void);
      void fftshift(void);
      void ifftshift(void);
      void TimeShiftSignal(double dTime);

      double GetConvFactOptElecPhotod()
            {return ConvFactOptElecPhotod;};
      int GetQtPoints(void) {return qtPoints;};
      double GetDeltaTime(void) {return DeltaTime;};
      double GetDeltaFreq(void) {return DeltaFreq;};
      double GetTimeWindow(void) {return TimeWindow;};
      double GetFreqWindow(void) {return FreqWindow;};
      int GetStringLength(void) {return (int) (TimeWindow/BitPeriod );};
      
  int GetDebugLevel(void) {return DebugLevel;};

      void WriteFileTimePower(string OutFile);
      void WriteFileFreqPower(string OutFile,int OmitDCFlag);
      void WriteFileFreqPower(string OutFile);
      
      // // ## Methods suitable for electrical equalization
      //void ClearElecSignal(void);
      //void AddElecSignal(Photodetector *oPhotodetectorOrig);
      //void AddElecSignal(Photodetector *oPhotodetectorOrig,
      //                            double MultiplFactor, 
      //                            double AmpShiftBeforeMultipl,
      //                            double TimeShift);
      //void AddConstantElecSignal(double ConstantIntensity);
                              
      fftPowerMatrices sfftPM; 
      fftPowerMatrices WorksfftPM; 
           //  Required in ComputeWork... methods of Photodetector that are
           // called by double ElecSignalStat::ComputePhaseTargetFreq(void)

      cfftw *fft;
      OptSignal *oOptSignal;

   private:      
      
      
      fftMatrices sfftM;  
       
      int ObjectInitialized;             
      int qtPoints,qtPointsLast;
      double ConvFactOptElecPhotod;
      double DeltaFreq, DeltaTime,FreqWindow,TimeWindow;
      double BitPeriod;

     // Parameters related to shot noise

      int AddShotNoiseFlag;
      double ShotNoiseCoeff;

      double ShotNoiseVarianceCoeff; // Out of date

      RanNumGen * RNG;

  int DebugLevel;

     };

#endif /* _OCS_PHOTODETECTOR_HH_ */








