// $Id: ocsRamanAmp.hh,v 1.3 2008/08/08 02:23:11 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_RAMAN_AMP_HH_
#define _OCS_RAMAN_AMP_HH_

//#include "ocsGlobalStructures.hh"
#include "ocsOptSignal.hh"

enum typeRamanNoise { RAMAN_NOISE_OFF = 0, 
                          RAMAN_NOISE_ON_RANDOM_PHASE = 1,
			  			  RAMAN_NOISE_ON_GAUSSIAN_WHITE = 2,
			  			  RAMAN_NOISE_ON_NO_RNG = 3 }; 
// ##### Class RamanAmp

class RamanAmp
{
   public:

    //#### Methods by Jonathan Hu:
    // Please see algorithm in: Bumki Min, PTL vol.12 No. 11
	// P.1486-1488 Nov. 2000 also see Howard Kidorf, PTL vol 11. No. 5. pp. 530-532
	// May 1999
	//
	
      RamanAmp();
      ~RamanAmp();
      void ReleaseMemory();
	  RamanAmp(string Job, string InFileName, RanNumGen *RNG2, int TNumFiber);
	  
	  RamanAmp(string Job, string InFileName, RanNumGen *RNG2, OptSignal *oOptSignal1,
	  		 int TNumFiber);
	  
	  void AllocateMemory();
	  
	  void PrepareSignal();
	  void PrepareSignalCW();
	  void PrepareSignalNonFourier();
	  void SetGainProfile(double FiberLength, string InFileName);
	  void ReadPlainFileGainLoss(double FiberLength,string InFileName);
	  void ReadRamanGainLossWavelengthGrid(double FiberLength,string InFileName);
	  void PrepareNextStep();
		//	asdg
//      void InitRamanIteration();
	  void RamanIteration();
	  void AddNoise();
	  void PropagateSignalandNoise(OptSignal *NFSignal, OptSignal *Noise);
 	  void PropagateNoise(OptSignal *Noise);
	  
	  void ShootingGeneticAlgorithmsVictor();
	  void GeneticAlgorithmsVictor(double * Vector);
	
	
	
	  void PrepareSignalStart();
	  void PrepareSignalLastRound();
	  void CallLastRound();
	  void ForwardLastRound(int StepI);
	  double OptimizeForwardIntegrationLastRound(int IntStepI, double * PreSig);
	  void SingalRayleighBScattteringBackward(int StepI);
	  void DoubleRayleighBScattteringForward(int StepI);
	  void SetPumpPower(double Value);
	  
	  void ShootingGAVictor();
	  void ShootingGA();
	  void ShootingGAVaryCW();
	  void ShootingGeneticAlgorithms(double VectorEqSpace);
	  void ShootingGeneticAlgorithmsVaryCW(double VectorEqSpace); 
	  void ShootingIntegration(int NExtItr);
	  void ShootingIntegrationExample(int NExtItr);
	  void GetPumpInitGuessFromIntg(double * TargetAlpha);
  	  void UpdateIndividualWavelength(double VectorSpaceUpdate);
	  void UpdateIndividualWavelength(int Num, double * VectorUpdate, 
	  			double Space, double CW);
	  void GeneticAlgorithmsVaryCW(double * Vector, double PumpSpace, double * CW);
	  void GeneticAlgorithms(double * Vector);
	  void CalculateFrel();

	  void InitRamanIteration();
	  void RelaxRamanIteration();	  
      void ForwardStep(int StepI);
      void BackwardStep(int StepI);
	  double OptimizeForwardIntegration(int IntStepI, double * PreSig);
	  double OptimizeBackwardIntegration(int IntStepI, double * PrePum);

	  void FirstGuessForwardShooting(int IntStepI, double * PreSig, double * PrePum);
	  void ShootingRamanIteration();
      void PrePumpShooting(int PreShootingStepI);
  //    void PrePumpShootingL2R(int PreShootingStepI);
	  void SignalOnlyPropagation(int PreShootingStepI);
      void PreSignalShooting(int PreShootingStepI);
 	  void PreSignalShootingWOA(int PreShootingStepI);
	//  void PrePumpShootingLastOne(int StepI);
	  void ForwardShooting(int StepI);
//	  void RKShooting(int IntStepI);
//	  void RKShootingSimple(int IntStepI);
//	  void RKShootingtext(int IntStepI);
	  double OptimizeForwardShootingIntegration(int IntStepI, double * PreSig, double * PrePum);
	  double OptimizeForwardPumpShootingIntegration(int IntStepI, double * PreSig, double * PrePum);
  	  double OptimizeForwardSignalShootingIntegration(int IntStepI, double * PreSig, double * PrePum);
	  void InternalRelaxationInForwardShooting(int IntStepI, double * PreSig, double * PrePum);
	  
	  void ParallelCorrection(double * ShootingValue, double * Target);	  
	  void ParallelCorrectionNewton(double * ShootingValue, double * Target);
	  void OneByOneCorrection(double * ShootingValue, double * Target);
	  void ParallelIntgCorrection(double * ShootingValue, double * Target);	
	  void ParallelIntgCorrectionNewton(double *ShootingValue, double * Target);
	  void UEvolution(int StepI, double * MatrixU);
	  
	  void GetMatrixW(int WStepI, double * MatrixW);
      void GaussianElimination(double * MatrixU, double * CorrectP, double * EpxirP);
	  void MatrixCopy(double * M1, double * M2, int M, int N);
	  void MatrixMultiply(double * M1, double * M2, int N);
	  void MatrixPlus(double * M1, double * M2, int N);
	  void MatrixPlus(double * M1, double * M2, int M, int N);
	  void DisplayM(double * M1 , int M , int N);
	  void MatrixDivid(double * M1, double M, int N);
	  void MatrixDivid(double * M1, double Num, int M, int N);
	  void MatrixMultiply(double * M1, double M2, int N);
	  void MatrixSubtract(double * M1, double * M2, int M, int N);
	  void MatrixReducedMultiply(double * M1, double * M2, int N , int P);
	  //double F2W(double Frequency) {  return( 3e+8 / Frequency);};
	  //double W2F(double WaveLength) { return( 3e+8 / WaveLength); };

      
	  double GetTempDepFactor(double Wavelength1,double Wavelength2);
      double GetRamanGainFreqGrid(double Wavelength1, double Wavelength2, int NumSteps);
      double GetRamanGainFreqGrid(double Wavelength1, double Wavelength2);
      double GetRamanGain(double Wavelength1, double Wavelength2, int NumSteps);
	  double GetRamanGain(double Wavelength1, double Wavelength2);
	  double GetRamanGainClose(double Wave1, double Wave2, int Index);
	  double GetAttenuationFiber(double Wavelength);
      double GetAttenuationFiber(double Wavelength, int NumSteps );
 //  double GetRamanPowerFreq(double Z, double Freq);
	  double GetRamanPowerFreqIndex(double Z, int SignalFreqIndex);
	  void SetAmpLength(double Value);
	  void SetTypeRamanNoise(typeRamanNoise TypeRamanNoise2);
	  void DoubleRamanPumpPower();
	
	  string WriteTypeRamanNoise(void);
	  void WriteSignalWaveLthPower(string OutFileName, int Num);
      void WriteSignalWaveLthPowerWith$(string OutFileName, int Num);
      void WritePumpWaveLthPower(string OutFileName, int Num);
      void WritePumpPower(string OutFileName);
	  void WritePumpPowerdB(string OutFileName);
	  void WriteSignalOutputPower(string OutFileName);
  	  void WriteSignalOutputPowerdB(string OutFileName);
      void WriteSignalConvergence(string OutFileName);
      void WriteSignalTrack(string OutFileName, int Num);
	  void WriteSignalTrack(string OutFileName);
 	  void WriteAveragePowerdBm(int loop);
	  void WriteEqIntgData(string InFileName);	  
	  void WriteGainShapeData(string InFileName,string OutFileName1,string OutFileName2);  
	  void WriteCentralWavelength(string InFileName,string OutFileName1);
      void WriteSignalNoise(OptSignal *Signal1, OptSignal *NFSignal, OptSignal *N, string
	  					OutSignalNoiseFileName);
          

	//  void GetAccuratePump();
	
	
	  void RamanTransient();

   private:
   	  void Calculatedsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt);
	  void Mac24BF(double * oSignalTemp,double * oPumpTemp,double * dPsdt,double * dPpdt);
	  void Mac24FB(double * oSignalTemp,double * oPumpTemp,double * dPsdt,double * dPpdt);
	  void CalculateForwarddsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt);
	  void CalculateBackwarddsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt);



// just for debug

 	  void ReadPlainFileGain(double FiberLength,string InFileName);
 	  void ReadPlainFileLoss(double FiberLength, string InFileName);
      double  WaveLengthStart;
      double WavelengthStep;



   private:
      typeRamanNoise TypeRamanNoise;
	
	  double DeltaZ(int NumSteps);
      
   
      double * oSignalWaveLth;  
      double * oSignalPower;
      double * oSignalPowerPre;
	  double * oSignalSample;
	  double * oSignalTrack;
	  double * oIntgSigPower;
	  double * SigFEnergy;
	  
	  double * oASEWaveLth;
	  double * oSRBWaveLth;
	  double * oDRBWaveLth;
	  double * oASEPower;
	  double * oSRBPower;
	  double * oDRBPower;
	  
      double * oPumpWaveLth;
      double * oPumpPower;
	  double * oPumpTarget;
	  double * oPumpTrack;
	  double * oPumpTargetBackup;
      double * oPumpPowerPre;
	  double * oPumpPower45Converted;
	  double * oPumpWaveLthInput;
	  double * oPumpPowerInput;
	  
	  double * oPumpIntgTarget;
//	  double * IntgPumPower;
      
      double * oRamanGain;
	  double * oLoss;
      double PumpAttenuation;
	  
      double * DeltaPsignal;
      double DeltaPsignalsign;
      double * DeltaPpump;
      double DeltaPpumpsign;
	  
	  double * PreCorrectP;

      double CompCoeff;
      double * DeltaPpumpTotal;
      double * DeltaPsignalTotal;
      
      int TypeIntergration;
	  //int TypePerturbation;
	  int TypeAlgorithm;
	  int TypeCorrect;
	  double WeightCorrect;
	  double IterationAccurateTh;
	  double ShootingAccurateTh;

	  //double PerturbationRatio;
	  
	  double RamanStartLength;
      double AmpLength;
      double DeltaZConst;
	  double * DeltaZArray;
      int NumSigCh;
      int NumPumCh;
	  int NumASECh;
	  int NumSRBCh;
	  int NumDRBCh;
	  
	  int NumFibersPumped;
	  int NumFiberIndex;
	  int * FiberLocation;
	  
	  
	  int PumpWavelengthForGain;
	  double  WaveLengthStartGain;
      double WavelengthStepGain;
      int  NumGainWidth;
	  double  WaveLengthStartLoss;
      double WavelengthStepLoss;
      int  NumLossWidth;
	  double FreqStartGain;
  	  double FreqStepGain;

	  
	  double GlobleRipple;

	  int NumPumChTotal;
      int NumCh;
      int NumIteration;
	  int NumIntItr;
	  
	  int TypeGeneticAlgorithms;
	  int Population;
	  int SelectionFixedNum;
	  double CrossRate;
	  double MutationRate;
	  int StartNumSpace;
	  int EndNumSpace;
	  
//	  int MultiFiberFlag;
	  int AccurateSign;
//	  double AccurateTh;
	  int BlowupSign;
	  int FixSignalShootingSign;
	  int FixPumpShootingSign;
//      double AttenuationFiber;
      double RayleighBackscatterCoeff;
      int NumZSteps;
      
      string JobRaman;

   	  int NumTSteps ; 
 	  double DeltaT ;  

   
      OptSignal *oOptSignal;
      RanNumGen *RNG; 

      cfftw *fft;
      fftMatrices sfftM;
	  int qtPoints;
 	  double DeltaFreq, FreqWindow, DeltaTime, TimeWindow;
      int ObjectInitialized;                          
      int TypeOptFilter;
      int OrderOptFilter;
      double FreqFWHM_OptFilter;
      double RelativeCenterFreqOptFilter;
//         double SpliterLossFactor;
	  double PumpPowerFactor;
	string InFileNameSave;
		int NoFourierSignal;
		int  ReadPlainFileGainLossFlag;
 		int ReadRamanGainLossWavelengthGridFlag;

};

#endif /* _OCS_RAMAN_AMP_HH_ */





