// $Id: ocsReadWriteTools.hh,v 1.2 2005/07/06 16:45:54 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_READ_WRITE_TOOLS_HH_
#define _OCS_READ_WRITE_TOOLS_HH_

#ifndef _OCS_READ_WRITE_ENUMERATORS_
#define _OCS_READ_WRITE_ENUMERATORS_

enum BoundsFlagEnumerator { LOWER_ONLY = 1, UPPER_ONLY = 2, 
                            LOWER_AND_UPPER = 3, NO_BOUNDS = 4};


enum InputParameterOptionsEnumerator { MANDATORY = 1, OPTIONAL_WARNING = 2,
                                       OPTIONAL_NO_WARNING = 3};


#endif /* _OCS_READ_WRITE_ENUMERATORS_ */

#ifndef _OCS_TYPE_SIMULATION_
#define _OCS_TYPE_SIMULATION_

enum typeSimulation { SCALAR = 1, VECTOR = 2 };

#endif /*_OCS_TYPE_SIMULATION_ */

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;



#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsReducedModelTools.hh"




// The first two functions added by JZ May 4th 2001.
// They improve on the functions discussed below in that 
// they allow a default value to be specified.


// ## The next five functions were added on March 1st, 2001 by JZ.
// ## They allow the user to specify whether the parameter being read
// ## is mandatory or optional.

// ## The reason for doing this is to allow new parameters to be added
// ## to the class and only add them to the input files if the user
// ## will be using these parameters in his/her application.
 
// ## If InputParameterOption ==  MANDATORY
// ##  and the parameter is not
// ## present in the .in file then the code will abort.

// ## If InputParameterOption == OPTIONAL_WARNING then the if the 
// ## parameter is not present the code will not abort but will
// ## output a warning message. 

// ## If InputParameterOption == OPTIONAL_NO_WARNING then 
// ## the code will not abort but will not
// ## output a warning message. It is recommended that this option
// ## be only  used when reading in an array of parameters.

// ## Over time the other versions of these functions (see below) 
// ## will be phased out.


int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            int DefaultValue,
            ifstream * InFile, bool DebugLevel); 

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    int DefaultValue,
            ifstream * InFile, bool DebugLevel); 

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,
            ifstream * InFile, bool DebugLevel); 

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  string DefaultValue,
                  ifstream * InFile, bool DebugLevel);


//#######

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            int DefaultValue,
            ifstream * InFile); 

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    int DefaultValue,
            ifstream * InFile); 

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,
            ifstream * InFile); 

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  string DefaultValue,
                  ifstream * InFile);


//#######

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel); 

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel); 

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel); 

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  ifstream * InFile, bool DebugLevel);


void MissingParameter(string ParameterName,
                      InputParameterOptionsEnumerator InputParameterOption, 
		      bool DebugLevel);

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile); 

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile); 

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile); 

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  ifstream * InFile);


void MissingParameter(string ParameterName,
                      InputParameterOptionsEnumerator InputParameterOption);

// ########## Older versions of Read parameter functions, still very
// ########## much in use in the code 

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile, 
	    bool DebugLevel); 

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
		 BoundsFlagEnumerator BoundsFlag, ifstream * InFile,
		 bool DebugLevel); 



double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile,
		  bool DebugLevel); 

string ReadString(string ParameterName, ifstream * InFile, 
		  bool DebugLevel);

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile); 

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
		 BoundsFlagEnumerator BoundsFlag, ifstream * InFile); 



double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile); 

string ReadString(string ParameterName, ifstream * InFile);

void cerror(char * msg);
void LogFileSeparator();
void LogFileStars();
void LogFileAbortWithErrorMsg(string ErrorMsg);

string WriteTypeSimulation(typeSimulation TypeSimulation);

// added by walter
double* ReadVectorOfDoubles(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,unsigned NumbersOfElements,
            ifstream * InFile);

double* ReadVectorOfDoubles(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue, unsigned NumberOfElements,
            ifstream * InFile, bool DebugLevel);
void WrongTypeParameter(string ParameterName);
void WrongTypeParameter(string ParameterName, bool DebugLevel);

unsigned ReadUnsignedInt(string ParameterName, string UnitsString,
            unsigned LowerBound,unsigned UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            unsigned DefaultValue,
            ifstream * InFile, bool DebugLevel); 

unsigned ReadUnsignedInt(string ParameterName, string UnitsString,
            unsigned LowerBound,unsigned UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            unsigned DefaultValue,
            ifstream * InFile); 


//end added by walter
	    
	    

#endif /* OCS_READ_WRITE_TOOLS_HH_ */
