// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2003):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_RECEIVER_SEMI_ANALYTICAL_HH_
#define _OCS_RECEIVER_SEMI_ANALYTICAL_HH_

#include "ocsOptSignal.hh"
#include "ocsOptFilter.hh"
#include "ocsOptDemuxer.hh"
#include "ocsPhotodetector.hh"
#include "ocsElecFilter.hh"
#include "ocsHistogram2D.hh"
#include "ocsElecSignalStat.hh"
#include "ocsOptDataWriter.hh"

enum typeReceiver { MONTE_CARLO = 1, SEMI_ANALYTICAL = 2 };

class ReceiverSemiAnalytical
{

  // ########### Public Methods ###################

public:

  // Written by John Zweck, April 2003

 //! Use this constructor followed by calls to ReceiverSubsystemInit
 //! if you are defining an array of ReceiverSubsystems

  ReceiverSemiAnalytical();  
  
  ReceiverSemiAnalytical(typeReceiver TypeReceiver2,
			 OptSignal * WDMSignal,
			 RanNumGen *RNG2,
			 int DemuxChannArrayIndex2,
			 string InDir, 
			 string Job2);
                              
  //! Use this Init method if you have many receivers and want to save memory.

  void ReceiverSemiAnalyticalInit(OptSignal * WDMSignal,
			     RanNumGen *RNG2,
                             string InDir, 
			     string Job2, 
                             int DemuxChannArrayIndex2,
                             fftPowerMatrices & sfftPM2,
                             fftPowerMatrices & WorksfftPM2,
			     OptSignal * GlobalSCSignal,  
                             OptDemuxer *Demuxer2);

  void SetReceiverMapNumber(int Value) { RecMapNum = Value;};
  void SetJobName(string Name) { Job = Name;};

  void SetDemuxChannArrayIndex(int Value) { DemuxChannArrayIndex = Value;};

  string WriteTypeReceiver(void);

  ~ReceiverSemiAnalytical();

  //! Calls either ReceiveSignalBackToBackMonteCarlo
  //! or ReceiveSignalSemiAnalytical

  void ReceiveSignalBackToBack(void);

  void ReceiveSignalBackToBackMonteCarlo(void);
  
  //! Calls ReceiveSignal(MapNum,0) 

  void ReceiveSignal(int MapNum);

  //! Calls either ReceiveSignalMonteCarlo or ReceiveSignalSemiAnalytical
  void ReceiveSignal(int MapNum,int MonteCarloSystemRealizationNum);
  void ReceiveSignalMonteCarlo(int MapNum,int MonteCarloSystemRealizationNum);
  void ReceiveSignalSemiAnalytical(string FileID);

  void PostProcessReceiverStats(int MapNum,int MonteCarloSystemRealizationNum,
				double PropagatedLength,int AppendFlag);

  void WriteStats(int MapNum,int MonteCarloSystemRealizationNum,
		   double PropagatedLength,int AppendFlag);

  void WriteNoiseFreeOpticalSignalFreq(string OutFileName);

  void GetLimaParameters(bool AppendOutFileFlag,
                         double FreqFWHM_OptFilter, 
			 double FreqFWHM_ElecFilter,double OSA_Bandwidth);


  // ########## Data ###########

public:

  OptSignal * WDMSignal;
  OptSignal * SingleChannelSignal;

  OptFilter * TunableOptFilter;
  OptDemuxer *oOptDemuxer;

  Photodetector * oPhotodetector;
  ElecFilter * oElecFilter;

  ElecSignalStat * oElecSignalStat;
  Histogram2D *EyeDiagramMonteCarlo;

private:

  typeReceiver TypeReceiver;

  int DemuxChannArrayIndex;
  string Job;
  int RecMapNum;

  int ObjectInitialized;

  double * StdDev_Signal_ASE;

};

#endif  




