// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#ifndef _OCS_REDU_OPT_SIGNAL_HH_
#define _OCS_REDU_OPT_SIGNAL_HH_

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
using namespace std;


#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsReducedModelTools.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"
#include "ModulationFormats.hh"

#ifndef _OVERLOADED_OPERATORS_
#define _OVERLOADED_OPERATORS_
//## Added by Ivan Lima to make the source compatible with g++
//## Most of the new code consists of overloading operators

inline double log (int a);
inline double log(int a)
{
   return log( (double) a );
}

inline double sqrt(int a);
inline double sqrt(int a)
{
   return sqrt( (double) a );
}

// End of extra 
#endif   // _OVERLOADED_OPERATORS_


//! The reduced optical signal class represents the reduced optical signal.

/*! The reduced optical signal data consists of the power and three Stokes
    parameters of the signal and noise within each channel of a WDM system
    and possibly additional unpolarized noise that is  not contained 
    within any of the channels. This class also contains methods to 
    compute the $\Delta Q$-factor of a given channel due to the polarization
    effects.
*/

// ##### Class ReduOptSignal
class ReduOptSignal
{
   public:

  //! Constructor
      ReduOptSignal(string InFileName, 
                    double SignalNoiseBandwidth2);


  //! JZAug02

      ReduOptSignal(string InFileName2,string JobName);


  //! Destructor
      ~ReduOptSignal();

  //! Copy Constructor

      ReduOptSignal(ReduOptSignal & OrigReduOptSignal);

   //! Assignment operator

   ReduOptSignal & operator=(ReduOptSignal & RHS_ReduOptSignal);


  //! The noise in ReduOptSignal1 is added incoherently to the
  //! current ReduOptSignal. For each channel, the Stokes vectors of the noise 
  //! in ReduOptSignal1 are added to those of the current ReduOptSignal

  void AddReducedNoise(ReduOptSignal * ReduOptSignal1);

  //! (Re)-initializes the Stokes 4-vector for the signal and noise
  //! in each channel

      void RegenerateReduOptSignal(void);

  //! JZAug02
  //! Reinitializes the Stokes 4-vector for the signal and noise
  //! in each channel. This special purpose method allows
  //! for different optical power levels in the different channels

      void RegenerateReduOptSignal(double * OptPowerChannArray2);

  //! JZAug02
  //! Reinitializes the Stokes 4-vector for the signal and noise
  //! in each channel. This special purpose method allows
  //! for different optical power levels in the different channels
  //! \sa OptPowerChannArray

      void RegenerateReduOptSignalVaryingPower(void);



  //! Depolarizes the noise in each channel

      void DepolarizeReduNoise(void);

  //! Writes the power and Stokes parameters of the signal and noise
  //! in each channel to a file.

      void WriteFileOptSignal(char *outFile);

  //! JZAug02

      void WriteMatlabFileOptSignal(string OutFileName);


  //! Set the optical and electrical filter bandwidths
 
 
     void SetOptElecFilterBandwidth
                   ( double OptFilterBandwidthTmp
                    ,double ElecFilterBandwidthTmp);

  //! Set the optical and electrical filter bandwidths
 
     void SetOptElecFilterBandwidth
                   ( string FileNameOptFilter
                    ,string FileNameElecFilter);

      void SetEnhancementFactor(double EnhancementFactorTmp)
           {EnhancementFactor = EnhancementFactorTmp;};    

      double GetChannSpacing(void) {return ChannSpacing;}; 

      double GetSignalNoiseBandwidth(void) 
             {return SignalNoiseBandwidth;};


  //! Returns the degree of polarization of the total Stokes parameters.
  /*! \sa     GetTotalStokes */

      double GetDegreeOfPolarization(void);

  //! Returns the total power 

      double GetTotalPower(void);

 //! Returns the total signal power. 

      double GetSignalPower(void);

  //! Returns the pulse type. 

      const char * WriteTypePulse(void);



  //! Returns the Q factor of the channel with index ChannelArrayIndex 

      double GetQ_Factor(int ChannelArrayIndex);

      double GetBitErrorRate(double Q_Factor) {return 
                exp(-sq(Q_Factor)/2)/(Q_Factor*sqrt(2*pi));};

  //! Returns the optical SNR of the channel with index ChannelArrayIndex

      double GetOptSNR(int ChannelArrayIndex);

      double GetSignal(int ChannelArrayIndex) 
         {return StokesSignal[0][ChannelArrayIndex];};

  //! Computes the Stokes parameters of the total signal and noise

      void GetTotalStokes( double *StokesTotal);

  //! Returns a normalized Stokes parameter of the Signal in a given channel.
  /*! Returns the normalized Stokes parameter with index StokesIndex 
  (in range 0,1,2,3) for the channel with index ChannelArrayIndex
  The Stokes parameter is normalized using the power of the signal
      in the given channel.

*/

      double GetStokesSignal(int StokesIndex,
                             int ChannelArrayIndex)
         {return StokesSignal[StokesIndex][ChannelArrayIndex]
                /StokesSignal[0][ChannelArrayIndex];
         };
       
  //! Returns the noise in the channel with index ChannelArrayIndex

      double GetNoiseInsideChannel(int ChannelArrayIndex);
      double GetScalarProdSignalNoise(int ChannelArrayIndex);      
      double GetCoPolarizedNoiseInsideChannel(int ChannelArrayIndex);
      double GetOrthogPolarizedNoiseInsideChannel(int ChannelArrayIndex);            
 
//! Returns a normalized Stokes parameter of the noise in a given channel.
  /*! Returns the normalized Stokes parameter with index StokesIndex 
  (in range 0,1,2,3) for the channel with index ChannelArrayIndex
  The Stokes parameter is normalized using the power of the polarized noise
      in the given channel.

*/

      double GetStokesNoiseInsideChannel(int StokesIndex,
                             int ChannelArrayIndex)
         {
	 //return StokesNoise[StokesIndex][ChannelArrayIndex]
         //       /StokesNoise[0][ChannelArrayIndex];
	    return GetStokesNoise(StokesIndex,ChannelArrayIndex);
         };

      double GetStokesNoise(int StokesIndex,
                             int ChannelArrayIndex)
         {return StokesNoise[StokesIndex][ChannelArrayIndex]
                /( StokesNoise[0][ChannelArrayIndex] );
            //      +StokesExtraUnpolChannNoise/NumChannels );
         };

  //! Returns the noise in the bandwidth BandwidthNoise about the
  //! channel with index ChannelArrayIndex
  

      double GetNoiseInsideBandwidth(int ChannelArrayIndex,
                                     double BandwidthNoise);

  //! To be safe, use only for single-channel simulations.
  //! Note: StokesNoise[0][0] contains total noise in frequency
  //! window given by ChannSpacing centered on channel zero.

      double GetTotalNoiseSpectralDensity(void)
        { return GetNoiseInsideBandwidth(0,1.0);};

      double GetCoPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise );
      double GetOrthogPolarizedNoiseInsideBandwidth(int ChannelArrayIndex,
                                              double BandwidthNoise );
					      
      double GetNormSignalNoiseBeatingCopolarized(int ChannelArrayIndex);
					      
      double GetNormSignalNoiseBeatingOrthogPolarized(int ChannelArrayIndex);


  //! Returns the electrical signal-to-noise ratio of a channel 

      double GetElecTimeDomainSNR(int ChannelArrayIndex);

  //! Computes the probability that a penalty is larger than a
  //! maximum allowed penalty assuming that the penalty is 
  //! Gaussian distributed

      double GetOutageProbGaussian(double MeanPenalty, 
                   double StdDevPenalty, double MaxAcceptablePenalty);

//! This method enables the user to change the number of channels on the fly

      void SetNumChannels(int NumChannels2);
      void SetOptPowerChannel(double OptPowerChann2) 
            {OptPowerChann = OptPowerChann2;};      

      int GetNumChannels(void) {return NumChannels;};
      double GetOptPowerChann(void) {return OptPowerChann;};
      double GetBitRateChann(void)  {return BitRateChann;};

  //! Converts index of channel from ChannelPhysicalIndex to ChannelArrayIndex

      int GetChannelArrayIndex(int ChannelPhysicalIndex);

 //! Converts index of channel from ChannelArrayIndex to ChannelPhysicalIndex

      int GetChannelPhysicalIndex(int ChannelArrayIndex);

  //! Returns the central frequency of the reduced signal in Hz

      double GetCenterFreq(void) {return CenterFreq;};

  //! Returns the frequency of the channel with index ChannelArrayIndex.
  /*! 
      ChannelArrayIndex in range 0,1,2,...,NumChannels-1.
   */
      double GetFrequency(int ChannelArrayIndex) 
             {return Frequency[ChannelArrayIndex];};

  
      double GetEnhancementFactor(void)
             {return EnhancementFactor;};    


 //! A helper function for InitializeInStokes. 

      void SetInStokes( const int ChannelArrayIndex
                       ,double * StokesVector);

 //! A helper function for InitializeInStokes. 

      double GetInStokes( const int ChannelArrayIndex
                         ,const int StokesIndex);



      // ## These vectors contain the actual data

  //! Array used to store the Stokes parameters of the signal in each channel.
  /*! StokesSignal[ii][jj] is the ii-th Stokes parameter of signal in the
      the jj-th channel.
  */
      double *StokesSignal[4];

 //! Array used to store the Stokes parameters of the noise in each channel.
  /*! StokesSignal[ii][jj] is the ii-th Stokes parameter of noise in the
      the jj-th channel.
  */
      double *StokesNoise[4];

   private:

      // ## Private Methods

  //! Allocates memory for the Stokes parameters of the signal and noise
  //! in each channel

      void AllocateMemory(void);

  //! Releases allocated memory

      void ReleaseMemory(void);

  //! Initializes the frequencies and input Stokes parameters of the reduced
  //! optical signal. Called by the class constructor.

      void ProcessReduOptSignalParameters(ifstream *InFile);

  //! Initializes the input Stokes parameters of the reduced
  //! optical signal. Called by ProcessReduOptSignalParameters.

      void InitializeInStokes(ifstream * InFile);


  //! A helper function for InitializeInStokes. 
  /*! In the parameter file the input stokes parameters of, for example,
      channel 1 are specified using the syntax

      $InStokesChann1 S_0 S_1 S_2 S_3

      where S_0 S_1 S_2 S_3 are 4 doubles.
  */

      int ReadInStokes( string InStokesName
                       ,int ch
                       ,ifstream * InFile);


  //! A helper function for InitializeInStokes. 

      void SetInStokes( const int ChannelArrayIndex 
                       ,const int StokesIndex 
                       ,double StokesValue);  

  //! A helper function for InitializeInStokes. 

      double *GetInStokes(const int ChannelArrayIndex);


  //! JZAug02

  void ReadFileFreqPower(string FrequencyDataInFileName);


      // ## Private data

      //! Enumeration variable which gives pulse format.
      /*! Used to compute efficiency factor of receiver.
          \sa WriteTypePulse
      */

      typePulse TypePulse;

      //! Central Frequency of the simulation.
      /*! Either CenterFreq  or CenterWavelength must be zero.
          Whichever of these parameters is not zero is used to set
          the center of the simulation window. If the flag
          EvenlySpacedFreqsFlag in the input parameter file is 1 then
          the channel frequencies are set using the  center frequency
          and the channel spacing.
          As I understand it code is only guarenteed to work
          when EvenlySpacedFreqsFlag=1.
	  \sa CenterWavelength ChannSpacing
      */

      double CenterFreq;

      //! Central wavelength of the simulation.
      /*! Either CenterFreq  or CenterWavelength must be zero.
          See CenterFreq for more info
          \sa CenterFreq ChannSpacing
      */

      double CenterWavelength;

      //! The number of WDM channels
      int NumChannels;

      //! The channel spacing. 
      /*! Only meaningful when EvenlySpacedFreqsFlag=1 in parameter input
          file. As I understand it the REduced Model 
          code is only guarenteed to work
          when EvenlySpacedFreqsFlag=1. (??)
          \sa CenterFreq CenterWavelength
      */

      double ChannSpacing;

      //! The bandwidth of the signal and noise in each channel
      /*! Must be less than ChannSpacing.
          
          This parameter is the bandwidth of the noise (and of the signal)
          that is used in the computation of the 
          signal-to-noise ratio of a channel using the function GetOptSNR. 

          The larger SignalNoiseBandwidth the more the noise that is
          included in a channel, the lower the optical SNR.


      */

      double SignalNoiseBandwidth;


      //! The bit rate of a channel.
      /*! e.g. 10 Gb/s. Not used in reduced model ??
       */


      double BitRateChann;

      //! The total optical power in each channel

      double OptPowerChann; 

      //! Array of optical powers for each channel/frequency
      //! JZAug02

      double * OptPowerChannArray; 


      //! Double array of initial Stokes parameters for each channel.

      double *InStokes;

      //! Bandwidth of the optical filter in the receiver.
     /*! Used in calculation of Q-factor from ESNR. 
         \sa ElecFilterBandwidth EnhancementFactor
     */


      double OptFilterBandwidth;

     //! Bandwidth of the electrical filter in the receiver.
     /*! Used in calculation of Q-factor from ESNR. 
         \sa OptFilterBandwidth EnhancementFactor
     */

      double ElecFilterBandwidth;

      //! The enhancement factor of the receiver. 
      /*! Depends on the pulse format (TypePulse) and the 
          bandwidths of the optical and electrical filters.
          Is an input parameter to the reduced model.

          See the article "Polarization Effects in Long-Haul Undersea Systems"
          by Curtis Menyuk et al. to appear as a chapter in a book in 2002
          for a discussion of the enhancement factor. In that paper
          it may be called the efficiency factor. 
      */

      double EnhancementFactor;

  //! Name of the input parameter file for the Reduced Optical Signal
  /*! This parameter is probably only used by the constuctor and 
      so could be changed to be local to the constructor 
  */

      string InFileName;

  //! Double array of size NumChannels used to store 
  //! frequencies of channels relative to the CenterFreq.

      double *Frequency;

};

#endif // _OCS_REDU_OPT_SIGNAL_HH_












