// $Id: ocsReducedModelTools.hh,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

// #####################################################################
//
//    ocsTools.hh
//
//    Header file for ocsTools.cc.
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//    The copyright applies to functions written by the OFCL members
// #####################################################################


#ifndef _OCS_TOOLS_HH_
#define _OCS_TOOLS_HH_

#include <unistd.h>

#ifndef _OCS_SYSTEM_HDRS_
#define _OCS_SYSTEM_HDRS_

#include <iostream.h>
#include <string.h>
#include <math.h> 
#include <fstream.h>
#include <complex.h>
#include <stdlib.h>
#include <stdio.h>
#include <strstream>

#include <string>
  using namespace std;

#endif /* _OCS_SYSTEM_HDRS_ */

#include "ocsConst.hh"
#include "ocsReadWriteTools.hh"
#include "ocsRanNumGen.hh"

#define NEAR_ZERO 2.2e-13

typedef complex<double> cplx; 

// #### square
template<class T> inline T sq(T);
template<class T> inline T sq(T in)
{
  return (in*in);
}

// ##### cube
template<class T> inline T cube(T);
template<class T> inline T cube(T in)
{
  return (in*in*in);
}

// ##### Sign function
inline int sgn(double x); 
inline int sgn(double x) 
{
   if (x >= 0)
      return(1);
   else   
      return(-1);
}

inline double min(double a, double b);
inline double min(double a, double b)
{
  if(a<b)
    return a;
  else
    return b;
}


// ##############################################################

// Functions added by Ivan


void TransRzRyV3to_Xhat( double *Vector, double *thetaZthetaY);
//                        ,double Rotation[][3]);
void TransRyV3to_YZ( double *Vector, double *thetaY);
void TransRzRxV3to_Xhat( double *Vector, double *thetaZthetaX);
//                    ,double Rotation[][3]);
void RotatesAboutZ(double angle, double *Vector); //, double Rz[][3]);
void RotatesAboutX(double angle, double *Vector); //, double Rx[][3]);
void RotatesAboutY(double angle, double *Vector); //, double Rx[][3]);
void Multiply( double Left[][3], double Right[][3]);  
void Multiply(double Left[][3],double Right[3]);
double ScalarProduct(double *VectorA, double *VectorB);
void CrossProduct(double Left[4], double Right[4], int FirstCoord);
double GetVectorLength(double *Vector, int Dimension);

double dB2Linear(double dBvalue);
double dBm2Linear(double dBmPower);
double Linear2dB(double LinearValue);
double Linear2dBm(double LinearPower);

double GetDegreeOfPolarization(double *StokesVector);
double GetPhase(const cplx Phasor);
void Stokes2Jones(const double *Stokes, cplx *Jones);
void Jones2Stokes(const cplx *Jones, double *Stokes);
void Stokes2JonesKeepPower(const double *Stokes, cplx *Jones);
void Jones2StokesKeepPower(const cplx *Jones, double *Stokes);

// ##########################################################

// Functions added by John

double Frequency2Wavelength(double Frequency); // Freq in Hz
double Wavelength2Frequency(double Wavelength); // Freq in Hz


// ##############################################################

// Functions added by Ron

enum{MOSIX_UNLOCK, MOSIX_LOCK};
int lock_or_unlock_mosix(int what);
int cpujob_mosix(void);
int slow_mosix(void);
inline double dB_to_factor(double G_dB);
inline double factor_to_dB(double factor);

// ##############################################################

// ## Numerical Recipes Functions

void nrerror(char error_text[]);
float erffc(float x);
float gammp(float a, float x);
void gcf(float *gammcf, float a, float x, float *gln);
void gser(float *gamser, float a, float x, float *gln);
float gammln(float xx);
float gammq(float a, float x);

// ## End of Numerical Recipes Functions

#endif /* OCS_TOOLS_HH_ */
