// $Id: 

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "PostDispComp.hh"

extern ofstream *LogFile;

// ################################################################

//! Use this constructor followed by calls to InitApril2003()
//! if you are defining an array of PostDispCompSubSystems

PostDispCompSubSystem::PostDispCompSubSystem(void)
{
  
}

void PostDispCompSubSystem::InitApril2003(string InFileName,
                                     OptSignal * Signal2,
                                     NoiseCovariance * NoiseCov2,
                                     RanNumGen * RNG2,
                                     OptSpectrumAnalyzer * PowerMeter2,
                                     string Job2)
{

NoiseCov = NoiseCov2;
InitApril2003(InFileName,Signal2,RNG2,PowerMeter2,Job2);

}

// ##########################################################

//! Initialization method
//!  for Ciena system PostDisperionCompensation, April 2003

void PostDispCompSubSystem::InitApril2003(string InFileName,
				     OptSignal * Signal2,
                                     RanNumGen * RNG2,
				     OptSpectrumAnalyzer * PowerMeter2,
                                     string Job2)
{


  // THERE ARE TWO CONSTRUCTORS

 ifstream InFile(InFileName.c_str(),ios::in);    
   
 /*  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    }
 */

 LogFileSeparator();
 *LogFile << InFileName << endl;


 PostDispCompMethod = 0; // Default value is don't post comp

 if(InFile)
   {

    PostDispCompMethod = ReadInt("PostDispCompMethod"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);

    ApplyDemuxFilterFlag = 0;

    TypePostAmps  = ReadInt("TypePostAmps"," ",0,7,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,5,&InFile);

    TypePostAmpsNoise = ReadInt("TypePostAmpsNoise"," ",0,4,
                               LOWER_AND_UPPER,MANDATORY,0,&InFile);

    GammaZeroFlag = ReadInt("GammaZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    FiberLossZeroFlag = ReadInt("FiberLossZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);

    //OutputStepSizesFlag = ReadInt("OutputStepSizesFlag"," ",0,1,
    //                        LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

    PreAmt = ReadInt("PreAmt"," ",1,-1,
                                   NO_BOUNDS,OPTIONAL_WARNING,&InFile);
    DispType = ReadInt("DispType"," ",-1,1,
                                   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

    PostExtraLength7 = ReadDouble("PostExtraLength7","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength10 = ReadDouble("PostExtraLength10","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength13 = ReadDouble("PostExtraLength13","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength16 = ReadDouble("PostExtraLength16","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength19 = ReadDouble("PostExtraLength19","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

    InFile.close();

   } // ## end if(InFile)

 SCSignal = Signal2;
 RNG = RNG2;
 InDir = "./";
 Job = Job2 + "Post"; // Do this so that files not written over

 if(PostDispCompMethod==2)
   {
     PowerMeter =   PowerMeter2;
     AllocateMemoryUMBCCienaLoopJuly2001();
   }
}// ## end constructor

// #################################################################

PostDispCompSubSystem::PostDispCompSubSystem(string InFileName,
					     OptSignal * SCSignal2,
                                             RanNumGen * RNG2,
                                       string InDir2,string Job2,
                                       int DemuxChannArrayIndex2,
                                       OptDataWriter * DataWriter2)
{


  // THERE ARE TWO CONSTRUCTORS

 ifstream InFile(InFileName.c_str(),ios::in);    
   
 /*  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    }
 */

 LogFileSeparator();
 *LogFile << InFileName << endl;


 PostDispCompMethod = 0; // Default value is don't post comp

 if(InFile)
   {

    PostDispCompMethod = ReadInt("PostDispCompMethod"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);


    AmountPostDispComp  = ReadInt("AmountPostDispComp"," ",1,-1,
                                   NO_BOUNDS,OPTIONAL_WARNING,&InFile);

    FreqFWHM_OptFilter  = ReadDouble("FreqFWHM_OptFilter"," ",0,-1,
                              LOWER_ONLY,OPTIONAL_WARNING,&InFile);
   
    ApplyDemuxFilterFlag = ReadInt("ApplyDemuxFilterFlag"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);


    TypePostAmps  = ReadInt("TypePostAmps"," ",0,7,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,5,&InFile);

    TypePostAmpsNoise = ReadInt("PostAmpsNoiseOffFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    GammaZeroFlag = ReadInt("GammaZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    FiberLossZeroFlag = ReadInt("FiberLossZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);

    //OutputStepSizesFlag = ReadInt("OutputStepSizesFlag"," ",0,1,
    //                        LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

    PreAmt = ReadInt("PreAmt"," ",1,-1,
                                   NO_BOUNDS,OPTIONAL_WARNING,&InFile);
    DispType = ReadInt("DispType"," ",-1,1,
                                   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);


    PostExtraLength7 = ReadDouble("PostExtraLength7","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength10 = ReadDouble("PostExtraLength10","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength13 = ReadDouble("PostExtraLength13","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength16 = ReadDouble("PostExtraLength16","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength19 = ReadDouble("PostExtraLength19","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

    InFile.close();

    

   } // ## end if(InFile)

 SCSignal = SCSignal2;
 RNG = RNG2;
 InDir = InDir2;
 Job = Job2 + "Post"; // Do this so that files not written over

 DemuxChannArrayIndex = DemuxChannArrayIndex2;
 DataWriter = DataWriter2;

 if(PostDispCompMethod==2)
   {  

      AllocateMemoryUMBCCienaLoopJuly2001();
   }
}// ## end constructor

// ##############################################################

PostDispCompSubSystem::PostDispCompSubSystem(string InFileName,
					     OptSignal * SCSignal2,
                                             RanNumGen * RNG2,
                                       string InDir2,string Job2,
                                       int DemuxChannArrayIndex2,
                                       OptDataWriter * DataWriter2,
				       OptSpectrumAnalyzer * PowerMeter2)
{

 ifstream InFile(InFileName.c_str(),ios::in);    
   
 /*  if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    }
 */

 LogFileSeparator();
 *LogFile << InFileName << endl;


 PostDispCompMethod = 0; // Default value is don't post comp

 if(InFile)
   {

    PostDispCompMethod = ReadInt("PostDispCompMethod"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);


    AmountPostDispComp  = ReadInt("AmountPostDispComp"," ",1,-1,
                                   NO_BOUNDS,OPTIONAL_WARNING,&InFile);

    FreqFWHM_OptFilter  = ReadDouble("FreqFWHM_OptFilter"," ",0,-1,
                              LOWER_ONLY,OPTIONAL_WARNING,&InFile);
   
    ApplyDemuxFilterFlag = ReadInt("ApplyDemuxFilterFlag"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);


    TypePostAmps  = ReadInt("TypePostAmps"," ",0,7,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,5,&InFile);

    TypePostAmpsNoise = ReadInt("PostAmpsNoiseOffFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    GammaZeroFlag = ReadInt("GammaZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    FiberLossZeroFlag = ReadInt("FiberLossZeroFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

    RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,0,&InFile);

    //OutputStepSizesFlag = ReadInt("OutputStepSizesFlag"," ",0,1,
    //                        LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

    PreAmt = ReadInt("PreAmt"," ",1,-1,
                                   NO_BOUNDS,OPTIONAL_WARNING,&InFile);
    DispType = ReadInt("DispType"," ",-1,1,
                                   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);


    PostExtraLength7 = ReadDouble("PostExtraLength7","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength10 = ReadDouble("PostExtraLength10","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength13 = ReadDouble("PostExtraLength13","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength16 = ReadDouble("PostExtraLength16","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);
    PostExtraLength19 = ReadDouble("PostExtraLength19","m",1,-1,
                            NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

    InFile.close();

    

   } // ## end if(InFile)

 SCSignal = SCSignal2;
 RNG = RNG2;
 InDir = InDir2;
 Job = Job2 + "Post"; // Do this so that files not written over

 DemuxChannArrayIndex = DemuxChannArrayIndex2;
 DataWriter = DataWriter2;

 if(PostDispCompMethod==2)
   {  // The next line is only difference from other constructor:
      PowerMeter = PowerMeter2; 
      AllocateMemoryUMBCCienaLoopJuly2001();
   }
}// ## end constructor

// ##############################################################

void PostDispCompSubSystem::ApplyPostDispComp(void)
{

  switch(PostDispCompMethod)
   {
    case 0:
      // Do Nothing
      cout << "NO Post Compensation performed by PostDispComp class" << endl;
      break;

    case 1:
      ApplyPostDispCompUMBCCienaLoopApr2001();
      cout << "Post Compensation IS performed by PostDispComp class" << endl;
      break;

    case 2:
      ApplyPostDispCompUMBCCienaLoopJuly2001();
      //cout << "Post Compensation IS performed by PostDispComp class" << endl;
      break;

   default:
     cerr << "PostDispCompSubSystem::ApplyPostDispComp:" << endl
	  << "switch(PostDispCompMethod) error. ABORTING NOW!!" << endl;
     exit(1);

   }

}// ## end ApplyPostDispComp

// ###########################################################


void PostDispCompSubSystem::ApplyPostDispCompNoiseCov(void)
{

  switch(PostDispCompMethod)
   {
    case 0:
      // Do Nothing
      cout << "NO Post Compensation performed by PostDispComp class" << endl;
      break;

    case 1:
      //ApplyPostDispCompUMBCCienaLoopApr2001();
      cout << "NO Post Compensation performed by PostDispComp class" << endl;
      break;

    case 2:
      ApplyPostDispCompUMBCCienaLoopJuly2001NoiseCov();
      //cout << "Post Compensation IS performed by PostDispComp class" << endl;
      break;

   default:
     cerr << "PostDispCompSubSystem::ApplyPostDispComp:" << endl
          << "switch(PostDispCompMethod) error. ABORTING NOW!!" << endl;
     exit(1);

   }

}// ## end ApplyPostDispComp


// #############################################################

void PostDispCompSubSystem::AdditionalDemuxFiltering(void)
{

  // Adapted from OptDemuxer::ExtractChannel

  
  OptFilter TunableFilter(1,1,FreqFWHM_OptFilter,
			  SCSignal->GetTypeSimulation());

  double RelativeCenterFreq =  SCSignal->GetFrequency(0)
                             - SCSignal->GetCenterFreq();

  TunableFilter.SetRelativeCenterFreq(RelativeCenterFreq);
  
  TunableFilter.SetOpticalData(SCSignal->fft,
                                SCSignal->sfftM);

  TunableFilter.TunableFilterOptSignal();


} // ## end AdditionalDemuxFiltering

// #############################################################

void PostDispCompSubSystem::ApplyPostDispCompUMBCCienaLoopApr2001(void)
{

  //  cout << "DOING POST DISP COMP from PostDispComp.in" << endl;

  LumpedLoss PreLoss(InDir + "PostDispPreLoss.in",SCSignal);
  LumpedLoss PostLoss(InDir + "PostDispPostLoss.in",SCSignal);

  
  OptSpectrumAnalyzer PowerMeter(InDir +"PostPowerMeter.in",Job,0,SCSignal);

  PowerMeter.RunPowerMeter(0.0);

  PreLoss.ApplyLumpedLoss();

  PowerMeter.RunPowerMeter(0.0);

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  switch(AmountPostDispComp)
   {
    case 0:
      {  
       OptAmplifier PostDispAmp5_86(InDir + "PostDispAmp5_8.6.in",
                                    SCSignal,RNG);



       PostDispAmp5_86.AmplifyOptSignal();
      }
      break;

case 1638:
      {  
       

        OptFiberLocalError PostSMF1638(InDir + "PostSMF1638.in",
                                    SCSignal,RNG,Job);

        PostSMF1638.Propagate(&PropagatedLength,&ZStepNum);

      }
      break;

case 819:
      {  
       

        OptFiberLocalError PostSMF819(InDir + "PostSMF819.in",
                                    SCSignal,RNG,Job);

        PostSMF819.Propagate(&PropagatedLength,&ZStepNum);

      }
      break;

case 983:
      {  
       

        OptFiberLocalError PostSMF983(InDir + "PostSMF983.in",
                                    SCSignal,RNG,Job);

        PostSMF983.Propagate(&PropagatedLength,&ZStepNum);

      }
      break;


    case -172:

      {
        OptAmplifier PostDispAmp4_05(InDir + "PostDispAmp4_0.5.in",
                                      SCSignal,RNG);

        OptAmplifier PostDispAmp5_86(InDir + "PostDispAmp5_8.6.in",
                                      SCSignal,RNG);


       PostDispAmp4_05.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

       OptFiberLocalError PostDCF172(InDir + "PostDCF172.in",
                                    SCSignal,RNG,Job);


       PostDCF172.SetWritePulseWidthFlag(
                         DataWriter->GetWritePulseWidthInFiberFlag());
       PostDCF172.SetOutputStepSizesFlag(
                         DataWriter->GetOutputStepSizesFlag());



       PostDCF172.Propagate(&PropagatedLength,&ZStepNum);

       PowerMeter.RunPowerMeter(PropagatedLength);

       PostDispAmp5_86.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

      }
      break;

    case -526:

      {
	OptAmplifier PostDispAmp4_10(InDir + "PostDispAmp4_1.0.in",
                                      SCSignal,RNG);

        OptAmplifier PostDispAmp5_86(InDir + "PostDispAmp5_8.6.in",
                                      SCSignal,RNG);

       PostDispAmp4_10.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

       OptFiberLocalError PostDCF526(InDir + "PostDCF526.in",
                                    SCSignal,RNG,Job);

       PostDCF526.SetWritePulseWidthFlag(
                         DataWriter->GetWritePulseWidthInFiberFlag());
       PostDCF526.SetOutputStepSizesFlag(
                         DataWriter->GetOutputStepSizesFlag());


       PostDCF526.Propagate(&PropagatedLength,&ZStepNum);

       PowerMeter.RunPowerMeter(PropagatedLength);

       PostDispAmp5_86.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

      }
      break;

    case -848:

      {
	
        OptAmplifier PostDispAmp4_15(InDir + "PostDispAmp4_1.5.in",
                                      SCSignal,RNG);

        OptAmplifier PostDispAmp5_76(InDir + "PostDispAmp5_7.6.in",
                                      SCSignal,RNG);


       PostDispAmp4_15.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

       cout << "Hi from DCF848" << endl;

       OptFiberLocalError PostDCF848(InDir + "PostDCF848.in",
                                    SCSignal,RNG,Job);

       PostDCF848.SetWritePulseWidthFlag(
                         DataWriter->GetWritePulseWidthInFiberFlag());
       PostDCF848.SetOutputStepSizesFlag(
                         DataWriter->GetOutputStepSizesFlag());



       PostDCF848.Propagate(&PropagatedLength,&ZStepNum);

       PowerMeter.RunPowerMeter(PropagatedLength);

       PostDispAmp5_76.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

      }
      break;

     case 865:

      {

        OptAmplifier PostDispAmp4_10(InDir + "PostDispAmp4_1.0.in",
                                      SCSignal,RNG);

        OptAmplifier PostDispAmp5_76(InDir + "PostDispAmp5_7.6.in",
                                      SCSignal,RNG);

	
       PostDispAmp4_10.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

       OptFiberLocalError PostSMF445_865A(InDir + "PostSMF445_865A.in",
                                    SCSignal,RNG,Job);

       PostSMF445_865A.SetWritePulseWidthFlag(
                         DataWriter->GetWritePulseWidthInFiberFlag());
       PostSMF445_865A.SetOutputStepSizesFlag(
                         DataWriter->GetOutputStepSizesFlag());



       PostSMF445_865A.Propagate(&PropagatedLength,&ZStepNum);

       OptFiberLocalError PostSMF865B(InDir + "PostSMF865B.in",
                                    SCSignal,RNG,Job);

       PostSMF865B.SetWritePulseWidthFlag(
                         DataWriter->GetWritePulseWidthInFiberFlag());
       PostSMF865B.SetOutputStepSizesFlag(
                         DataWriter->GetOutputStepSizesFlag());


       PostSMF865B.Propagate(&PropagatedLength,&ZStepNum);

       PowerMeter.RunPowerMeter(PropagatedLength);

       PostDispAmp5_76.AmplifyOptSignal();

       PowerMeter.RunPowerMeter(PropagatedLength);

      }
      break;

    default:
      cerr << "PostDispCompSubSystem::ApplyPostDispComp:" << endl
	   << "switch(AmountPostDispComp) error. ABORTING" << endl;
      exit(1);

    }


  PowerMeter.RunPowerMeter(PropagatedLength);

  if(ApplyDemuxFilterFlag)
   AdditionalDemuxFiltering();

  PowerMeter.RunPowerMeter(PropagatedLength);
  PostLoss.ApplyLumpedLoss();

  PowerMeter.RunPowerMeter(PropagatedLength);
 
} // ## end ApplyPostDispCompUMBCCienaLoopApr2001

// ###################################################

void PostDispCompSubSystem::SetAmountPostDispCompJuly2001(void)
{

  //cout << "DispType = " << DispType << endl;
  //cout << "PreAmt = " << PreAmt << endl;
  //cout << "RecMapNum = " << RecMapNum << endl;

  if(DispType == -1 && PreAmt == 0) // D- case, No pre
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 631; break;        
	 case 10: AmountPostDispComp = 1070; break;                          
	 case 13: AmountPostDispComp = 1325; break;  
	 case 16: AmountPostDispComp = 1511; break;  
	 case 19: AmountPostDispComp = 1760; break;  
	}

  if(DispType == -1 && PreAmt == 441) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 445; break;        
	 case 10: AmountPostDispComp = 631; break;                          
	 case 13: AmountPostDispComp = 1070; break;  
	 case 16: AmountPostDispComp = 1325; break;  
	 case 19: AmountPostDispComp = 1505; break;  
	}


  if(DispType == -1 && PreAmt == 875) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = -170; break;        
	 case 10: AmountPostDispComp = 186; break;                          
	 case 13: AmountPostDispComp = 626; break;  
	 case 16: AmountPostDispComp = 885; break;  
	 case 19: AmountPostDispComp = 1070; break;  
	}

  if(DispType == 0 && PreAmt == 0) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 186; break;        
	 case 10: AmountPostDispComp = 322; break;                          
	 case 13: AmountPostDispComp = 322; break;  
	 case 16: AmountPostDispComp = 527; break;  
	 case 19: AmountPostDispComp = 527; break;  
	}

  if(DispType == 0 && PreAmt == -341) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 322; break;        
	 case 10: AmountPostDispComp = 322; break;                          
	 case 13: AmountPostDispComp = 322; break;  
	 case 16: AmountPostDispComp = 322; break;  
	 case 19: AmountPostDispComp = 322; break;  
	}

  if(DispType == 0 && PreAmt == -506) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 527; break;        
	 case 10: AmountPostDispComp = 527; break;                          
	 case 13: AmountPostDispComp = 527; break;  
	 case 16: AmountPostDispComp = 527; break;  
	 case 19: AmountPostDispComp = 527; break;  
	}


  if(DispType == 1 && PreAmt == 0) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = -754; break;        
	 case 10: AmountPostDispComp = -1184; break;                          
	 case 13: AmountPostDispComp = -1354; break;  
	 case 16: AmountPostDispComp = -1605; break;  
	 case 19: AmountPostDispComp = -1945; break;  
	}

  if(DispType == 1 && PreAmt == -851) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 186; break;        
	 case 10: AmountPostDispComp = -170; break;                          
	 case 13: AmountPostDispComp = -510; break;  
	 case 16: AmountPostDispComp = -850; break;  
	 case 19: AmountPostDispComp = -924; break;  
	}

 if(DispType == 1 && PreAmt == -1260) 
      switch(RecMapNum)
	{
	 case 7: AmountPostDispComp = 445; break;        
	 case 10: AmountPostDispComp = 186; break;                          
	 case 13: AmountPostDispComp = -170; break;  
	 case 16: AmountPostDispComp = -340; break;  
	 case 19: AmountPostDispComp = -680; break;  
	}


 
 //cout << "AmountPostDispComp = " << AmountPostDispComp << endl;

   switch(RecMapNum)
     {

         case 0:
           PostExtraLength = 0.0;
           break;

       case 7:
         PostExtraLength = PostExtraLength7;
         break;

       case 10:
	 PostExtraLength = PostExtraLength10;
         break;

       case 13:
         PostExtraLength = PostExtraLength13;
         break;

       case 16:
	 PostExtraLength = PostExtraLength16;
         break;

       case 19:
         PostExtraLength = PostExtraLength19;
         break;

       default:
         cerr << "RecMapNum switch error in " << endl
	      << "PostDispCompSubSystem::"
              << "ApplyPostDispCompUMBCCienaLoopJuly2001" << endl;
         exit(1);
     }

 
   if(PostExtraLength >=0)
     {
       PostExtraSMF->SetFiberLength(PostExtraLength);
       PostExtraDCF->SetFiberLength(0);  
     }
   else
     {
       PostExtraSMF->SetFiberLength(0);
       PostExtraDCF->SetFiberLength(abs(PostExtraLength));  
     }


} // ## end SetAmountPostDispCompJuly2001

// ########################################################

void PostDispCompSubSystem::AllocateMemoryUMBCCienaLoopJuly2001(void)
{
  PostAmpDCF = new OptAmplifier(InDir + "PostAmpDCF.in",SCSignal,RNG,
				PowerMeter);
  PostAmpSMF = new OptAmplifier(InDir + "PostAmpSMF.in",SCSignal,RNG,
                                PowerMeter);
  PostAmpFinal = new OptAmplifier(InDir + "PostAmpFinal.in",SCSignal,RNG,
                                  PowerMeter);
  int NumAmps = 3; // If change this must change .hh file

   AmpPtr[0] = PostAmpDCF;
   AmpPtr[1] = PostAmpSMF;
   AmpPtr[2] = PostAmpFinal;
       
  for(int i=0;i<NumAmps;i++)
    {
     AmpPtr[i]->SetTypeAmplifier(typeAmplifier(TypePostAmps));
     AmpPtr[i]->SetTypeAmplifierNoise(typeAmplifierNoise(TypePostAmpsNoise));
    }

    PostLTS1 = new OptFiberLocalError(InDir + "PostLTS1.in",SCSignal,RNG,Job);
    PostLTS2 = new OptFiberLocalError(InDir + "PostLTS2.in",SCSignal,RNG,Job);
    PostLTS3 = new OptFiberLocalError(InDir + "PostLTS3.in",SCSignal,RNG,Job);
    PostLTS4 = new OptFiberLocalError(InDir + "PostLTS4.in",SCSignal,RNG,Job);
    PostLTS5 = new OptFiberLocalError(InDir + "PostLTS5.in",SCSignal,RNG,Job);

PostLycom1 = new OptFiberLocalError(InDir + "PostLycom1.in",SCSignal,RNG,Job);
PostLycom2 = new OptFiberLocalError(InDir + "PostLycom2.in",SCSignal,RNG,Job);
PostLycom3 = new OptFiberLocalError(InDir + "PostLycom3.in",SCSignal,RNG,Job);
PostLycom4 = new OptFiberLocalError(InDir + "PostLycom4.in",SCSignal,RNG,Job);
PostLycom5 = new OptFiberLocalError(InDir + "PostLycom5.in",SCSignal,RNG,Job);

PostSMF5807 = new OptFiberLocalError(InDir + "PostSMF5807.in",SCSignal,RNG,
                                     Job);
PostSMF5809 = new OptFiberLocalError(InDir + "PostSMF5809.in",SCSignal,RNG,
                                      Job);
PostSMF5810 = new OptFiberLocalError(InDir + "PostSMF5810.in",SCSignal,RNG,
                                     Job);
PostDCF4_20 = new OptFiberLocalError(InDir + "PostDCF4_20.in",SCSignal,RNG,
                                     Job);
PostDCF3_10 = new OptFiberLocalError(InDir + "PostDCF3_10.in",SCSignal,RNG,
                                     Job);

PostSMF3_25 = new OptFiberLocalError(InDir + "PostSMF3_25.in",SCSignal,RNG,
                                     Job);
PostSMF186 = new OptFiberLocalError(InDir + "PostSMF186.in",SCSignal,RNG,Job);
PostSMFs3 = new OptFiberLocalError(InDir + "SMFs3A.in",SCSignal,RNG,Job);
PostSMFs4 = new OptFiberLocalError(InDir + "SMFs4A.in",SCSignal,RNG,Job);
PostSMF1500 = new OptFiberLocalError(InDir + "SMF1.5A.in",SCSignal,RNG,Job);

PostExtraSMF = new OptFiberLocalError(InDir + "PostExtraSMF.in",
                                      SCSignal,RNG,Job);

PostExtraDCF = new OptFiberLocalError(InDir + "PostExtraDCF.in",
                                      SCSignal,RNG,Job);

NumFibers = 22;  // WARNING: If you change this change .hh file too:
                 // OptFiberLocalError * FiberPtr[22];

FiberPtr[0] = PostLTS1;
FiberPtr[1] = PostLTS2;
FiberPtr[2] = PostLTS3;
FiberPtr[3] = PostLTS4;
FiberPtr[4] = PostLTS5;
FiberPtr[5] = PostLycom1;
FiberPtr[6] =  PostLycom2;
FiberPtr[7] = PostLycom3;
FiberPtr[8] = PostLycom4;
FiberPtr[9] =  PostLycom5;
FiberPtr[10] = PostSMF5807;
FiberPtr[11] = PostSMF5809;
FiberPtr[12] = PostSMF5810;
FiberPtr[13] = PostDCF4_20;
FiberPtr[14] = PostDCF3_10;
FiberPtr[15] = PostSMF3_25;
FiberPtr[16] = PostSMF186;
FiberPtr[17] = PostSMFs3;
FiberPtr[18] = PostSMFs4;
FiberPtr[19] = PostSMF1500;
FiberPtr[20] = PostExtraSMF;
FiberPtr[21] = PostExtraDCF;

   // ############ Check for Special Fiber Cases ####################

   

   if(GammaZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetGammaFiber(0);
     }

   if(FiberLossZeroFlag)
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetAttenuationdB(0);
     }


   if(!RelativeErrorGoal)
     {
       //cout << "Do nothing" << endl;
     }
   else
     {
       for(int i=0;i<NumFibers;i++)
         FiberPtr[i]->SetRelativeErrorGoal(RelativeErrorGoal);
     }


   for(int i=0;i<NumFibers;i++)
       FiberPtr[i]->SetOutputStepSizesFlag(0);

} // ## end AllocateMemoryUMBCCienaLoopJuly2001

// ################################################################

void PostDispCompSubSystem::ApplyPostDispCompUMBCCienaLoopJuly2001(void)
{

   SetAmountPostDispCompJuly2001();

  double PropagatedLength = 0.0;
  int ZStepNum = 0;

  //PowerMeter->RunPowerMeter(0.0);

  /* The following are correct:
  cout << "PostAmps OSA NumChanns = " << 
       PostAmpDCF->PowerMeter->GetNumChanns() 
                                 << endl;
  cout << "PostSig FreqChann = " << SCSignal->GetFrequency(0) << endl;
  cout << "PostSig CenterFreq = " << SCSignal->GetCenterFreq() << endl;
  */

  switch(AmountPostDispComp)
   {
    case -754:
        
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);

      break;

    case -1184:
        
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom2->Propagate(&PropagatedLength,&ZStepNum);
      break;

    case -1354:

        PostAmpDCF->AmplifyOptSignal();
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLycom2->Propagate(&PropagatedLength,&ZStepNum);
       PostLTS1->Propagate(&PropagatedLength,&ZStepNum);

      break;

    case -1605:
        
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLycom2->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom4->Propagate(&PropagatedLength,&ZStepNum);
      break;

    case -1945:
       PostDCF4_20->Propagate(&PropagatedLength,&ZStepNum);
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLycom2->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom4->Propagate(&PropagatedLength,&ZStepNum);
      break;

   case 186:
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -170:
       PostLTS1->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -510:
        PostAmpDCF->AmplifyOptSignal();
       PostLTS1->Propagate(&PropagatedLength,&ZStepNum);
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -850:
       PostLTS1->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostDCF4_20->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -924:
       PostLTS1->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpDCF->AmplifyOptSignal();
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostLycom3->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case 445:
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -340:
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case -680:
        PostAmpDCF->AmplifyOptSignal();
       PostLTS3->Propagate(&PropagatedLength,&ZStepNum);
       PostDCF4_20->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case 631:
       
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
       break;

  case 1070:
       
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);

       break;

  case 1325:
       
       PostSMF5807->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);

       break;


  case 1505:
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5809->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);

       break;


  case 1511:
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5807->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);

       break;

  case 1760:
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5809->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5807->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);

       break;

 case 626:
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);
       break;

 case 885:
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
        PostAmpSMF->AmplifyOptSignal();
       PostSMF5810->Propagate(&PropagatedLength,&ZStepNum);
       break;

 case 322:
        PostAmpSMF->AmplifyOptSignal();
       PostSMF186->Propagate(&PropagatedLength,&ZStepNum);
       PostSMFs3->Propagate(&PropagatedLength,&ZStepNum);
       PostSMFs4->Propagate(&PropagatedLength,&ZStepNum);
       PostSMF1500->Propagate(&PropagatedLength,&ZStepNum);

       break;

 case 527:
        PostAmpSMF->AmplifyOptSignal();
       PostSMF3_25->Propagate(&PropagatedLength,&ZStepNum);
       PostSMFs4->Propagate(&PropagatedLength,&ZStepNum);
       PostSMF1500->Propagate(&PropagatedLength,&ZStepNum);

       break;


    default:
      cerr << "PostDispCompSubSystem::ApplyPostDispComp:" << endl
           << "AmountPostDispComp = " << AmountPostDispComp << endl
	   << "switch(AmountPostDispComp) error. ABORTING" << endl;
      exit(1);

    }


  PostExtraSMF->Propagate(&PropagatedLength,&ZStepNum);
  PostExtraDCF->Propagate(&PropagatedLength,&ZStepNum);

  //PowerMeter->RunPowerMeter(PropagatedLength);
  PostAmpFinal->AmplifyOptSignal();
  //PowerMeter->RunPowerMeter(PropagatedLength);


  /*
  cout << "Total Accd Disp in PostDispComp [ps/nm] =  " <<
           -1e3*2*pi*LightSpeed*
           SCSignal->GetAccumulatedFirstOrderDispersion()/
           pow(SCSignal->GetCenterWavelength(),2) << endl;
  */

} // ## end ApplyPostDispCompUMBCCienaLoopJuly2001

// ###################################################

void PostDispCompSubSystem::ApplyPostDispCompUMBCCienaLoopJuly2001NoiseCov(
                                                                      void)
{

  SetAmountPostDispCompJuly2001();

  
  //PowerMeter->RunPowerMeter(0.0);

  /* The following are correct:
  cout << "PostAmps OSA NumChanns = " << 
       PostAmpDCF->PowerMeter->GetNumChanns() 
                                 << endl;
  cout << "PostSig FreqChann = " << SCSignal->GetFrequency(0) << endl;
  cout << "PostSig CenterFreq = " << SCSignal->GetCenterFreq() << endl;
  */

  switch(AmountPostDispComp)
   {
    case -754:
        
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLycom3);

      break;

    case -1184:
        
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLycom3);
       NoiseCov->Propagate(PostLycom2);
      break;

    case -1354:

        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostLycom3);
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLycom2);
       NoiseCov->Propagate(PostLTS1);

      break;

    case -1605:
        
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostLycom3);
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLycom2);
       NoiseCov->Propagate(PostLycom4);
      break;

    case -1945:
       NoiseCov->Propagate(PostDCF4_20);
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostLycom3);
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLycom2);
       NoiseCov->Propagate(PostLycom4);
      break;

   case 186:
       NoiseCov->Propagate(PostSMF186);
       break;

  case -170:
       NoiseCov->Propagate(PostLTS1);
       break;

  case -510:
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLTS1);
       NoiseCov->Propagate(PostLTS3);
       break;

  case -850:
       NoiseCov->Propagate(PostLTS1);
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostDCF4_20);
       break;

  case -924:
       NoiseCov->Propagate(PostLTS1);
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostLycom3);
       break;

  case 445:
       NoiseCov->Propagate(PostSMF3_25);
       break;

  case -340:
       NoiseCov->Propagate(PostLTS3);
       break;

  case -680:
        NoiseCov->Amplify(PostAmpDCF);
       NoiseCov->Propagate(PostLTS3);
       NoiseCov->Propagate(PostDCF4_20);
       break;

  case 631:
       
       NoiseCov->Propagate(PostSMF186);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
       break;

  case 1070:
       
       NoiseCov->Propagate(PostSMF186);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);

       break;

  case 1325:
       
       NoiseCov->Propagate(PostSMF5807);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);

       break;


  case 1505:
       NoiseCov->Propagate(PostSMF186);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5809);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);

       break;


  case 1511:
       NoiseCov->Propagate(PostSMF186);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5807);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);

       break;

  case 1760:
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5809);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5807);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);

       break;

 case 626:
       NoiseCov->Propagate(PostSMF186);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);
       break;

 case 885:
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF5810);
       break;

 case 322:
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF186);
       NoiseCov->Propagate(PostSMFs3);
       NoiseCov->Propagate(PostSMFs4);
       NoiseCov->Propagate(PostSMF1500);

       break;

 case 527:
        NoiseCov->Amplify(PostAmpSMF);
       NoiseCov->Propagate(PostSMF3_25);
       NoiseCov->Propagate(PostSMFs4);
       NoiseCov->Propagate(PostSMF1500);

       break;


    default:
      cerr << "PostDispCompSubSystem::ApplyPostDispComp:" << endl
           << "AmountPostDispComp = " << AmountPostDispComp << endl
	   << "switch(AmountPostDispComp) error. ABORTING" << endl;
      exit(1);

    }


  NoiseCov->Propagate(PostExtraSMF);
  NoiseCov->Propagate(PostExtraDCF);

  //PowerMeter->RunPowerMeter(PropagatedLength);
  NoiseCov->Amplify(PostAmpFinal);
  //PowerMeter->RunPowerMeter(PropagatedLength);


  /*
  cout << "Total Accd Disp in PostDispComp [ps/nm] =  " <<
           -1e3*2*pi*LightSpeed*
           SCSignal->GetAccumulatedFirstOrderDispersion()/
           pow(SCSignal->GetCenterWavelength(),2) << endl;
  */

} // ## end ApplyPostDispCompUMBCCienaLoopJuly2001NoiseCov

// ###################################################













