// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2009):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "SaturableAbsorber.hh"

extern ofstream *LogFile;

SaturableAbsorber::SaturableAbsorber(string InFileName, OptSignal *oOptSignal2)
{

oOptSignal = oOptSignal2;
TypeSimulation = oOptSignal->GetTypeSimulation();

ifstream InFile(InFileName.c_str(),ios::in);

if (!InFile)
    {
      cerr << "Error: Input file "<< InFileName
           << " could not be opened."<< endl;
      exit(1);
    }


LogFileSeparator();
*LogFile << InFileName << endl << endl;

DebugFlag = ReadInt("DebugFlag"," ",0,1,LOWER_AND_UPPER,&InFile);

TypeSaturableAbsorption = typeSaturableAbsorption(
                          ReadInt("TypeSaturableAbsorption"," ",
                                   0,6,LOWER_AND_UPPER,&InFile));

*LogFile << WriteTypeSaturableAbsorption();

SaturableAbsorberLength = ReadDouble("SaturableAbsorberLength","m",0,-1,
                                    LOWER_ONLY,&InFile);

if(TypeSaturableAbsorption == QUADRATIC_NO_LOSS_ANALYTIC || QUADRATIC_NO_LOSS)
{

FastSaturableAbsorptionNonlinearCoefficient =
  ReadDouble("FastSaturableAbsorptionNonlinearCoefficient","(Wm)^{-1}",0,-1,
                                    LOWER_ONLY,&InFile);

}


if(TypeSaturableAbsorption == QUADRATIC_WITH_LOSS ||
   TypeSaturableAbsorption == QUADRATIC_WITH_LOSS_ANALYTIC ||
   TypeSaturableAbsorption == RATIONAL_QUADRATIC)
{

FastSaturableAbsorptionNonlinearCoefficient = 
  ReadDouble("FastSaturableAbsorptionNonlinearCoefficient","(Wm)^{-1}",0,-1,
                                    LOWER_ONLY,&InFile);

FastSaturableAbsorptionAttenuationCoefficient =
  ReadDouble("FastSaturableAbsorptionAttenuationCoefficient","W",0,-1,
                                    LOWER_ONLY,&InFile);

if(FastSaturableAbsorptionAttenuationCoefficient <= 0)
 {
  cout << "FastSaturableAbsorptionAttenuationCoefficient must be positive."
       << "ABORTING" << endl;
  exit(1);

 }

}

if(TypeSaturableAbsorption == COUPLED_ODE)
{

CoupledODEUnsaturatedNepperLoss =
  ReadDouble("CoupledODEUnsaturatedNepperLoss","(m)^{-1}",0,-1,
              LOWER_ONLY,&InFile);

CoupledODERelaxationTime =
  ReadDouble("CoupledODERelaxationTime","sec",0,-1,LOWER_ONLY,&InFile);

CoupledODESaturationPower =
  ReadDouble("CoupledODESaturationPower","W",0,-1,LOWER_ONLY,&InFile);

}



if(TypeSaturableAbsorption != QUADRATIC_WITH_LOSS_ANALYTIC &&
   TypeSaturableAbsorption != QUADRATIC_NO_LOSS_ANALYTIC)
{

   TypeStepSizes = typeStepSizes(ReadInt("TypeStepSizes"," ",1,4,
                                 LOWER_AND_UPPER,&InFile));

   *LogFile << "TypeStepSizes = " << WriteTypeStepSizes() << endl;


   if(TypeStepSizes == 1 || TypeStepSizes == 3)
    {
     cout << "TypeStepSizes = " << WriteTypeStepSizes() << endl;
     cout << "SaturableAbsorber TypeStepSizes must be 2 or 4" << endl
          << "ABORTING!" << endl;
     exit(1);

    }

   if(TypeStepSizes==CONSTANT)
    {
     NumSteps = ReadInt("NumSteps"," ",0,-1,LOWER_ONLY,&InFile);
    }
   
   if(TypeStepSizes==LOCAL_ERROR_2ND)
    {
     DeltaZMaxInfile = ReadDouble("DeltaZMax","m",0,-1,
                                   LOWER_ONLY,&InFile);

     RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
                                     LOWER_ONLY,&InFile);
    } 


} // end TypeSaturableAbsorption != QUADRATIC_WITH_LOSS_ANALYTIC


InFile.close();
  
fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
qtPoints = fft->give_N();

InitializedDeltaZInitialFlag = false;

ComputeDeltaZMax();

// ### Initialize memory #####

 if(!oOptSignal->GetUsingOptFiberLocalError())
   {
     cerr << "OptFiberLocalError constructor: " << endl
          << "Must set UsingOptFiberLocalError = 1 in OptSignal.in" << endl
          << "ABORTING NOW!!" << endl;
     exit(1);
   }


switch(TypeSimulation)
 {
  case SCALAR:
    LaserWorkX = oOptSignal->OFLEWork.LaserWorkX;
    SAF1 = new double[qtPoints];
    SAF2 = new double[qtPoints];

    // We use f data on RHS as that is already malloced in OptSignal
    // But we interpret it as t data

    sfftM_Start.tOptSignalX = oOptSignal->OFLEWork.sfftM_Start.fOptSignalX;
    sfftM_Coarse.tOptSignalX = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalX;

    if(!LaserWorkX || !sfftM_Start.tOptSignalX 
       || !sfftM_Coarse.tOptSignalX || !SAF1 || !SAF2)
       {
        cerr << "Malloc error in SaturableAbsorber constructor"
             << "Aborting now!!" << endl;

        exit(1);
       }

    break;

  case VECTOR:

    SAF1 = new double[qtPoints];
    SAF2 = new double[qtPoints];

    LaserWorkX = oOptSignal->OFLEWork.LaserWorkX;
    LaserWorkY = oOptSignal->OFLEWork.LaserWorkY;
    sfftM_Start.tOptSignalX = oOptSignal->OFLEWork.sfftM_Start.fOptSignalX;
    sfftM_Coarse.tOptSignalX = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalX;
    sfftM_Start.tOptSignalY = oOptSignal->OFLEWork.sfftM_Start.fOptSignalY;
    sfftM_Coarse.tOptSignalY = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalY;

    if(!LaserWorkX || !LaserWorkY || 
       !sfftM_Start.tOptSignalX || !sfftM_Start.tOptSignalY ||
       !sfftM_Coarse.tOptSignalX || !sfftM_Coarse.tOptSignalY ||
       !SAF1 || !SAF2)
       {
        cerr << "Malloc error in SaturableAbsorber constructor"
             << "Aborting now!!" << endl;

        exit(1);
       }

    break;

 } // end switch 

} // end SaturableAbsorber::SaturableAbsorber

// ##############################################################

void SaturableAbsorber::ComputeDeltaZMax(void)
{
 DeltaZMax = DeltaZMaxInfile;
 InitializedDeltaZInitialFlag = false;
}

// ##############################################################

void SaturableAbsorber::Apply(double  * PropagatedLength)
{

if(TypeSaturableAbsorption != NO_SAT_ABS)
 *PropagatedLength += SaturableAbsorberLength;

if(DebugFlag)
  cout << "Total Energy Before SaturableAbsorber [pJ] = " 
       << 1e12*oOptSignal->GetTotalEnergy() << endl;

UpdateTime();

switch(TypeSaturableAbsorption)
{
  case NO_SAT_ABS:
       break;

  case QUADRATIC_NO_LOSS_ANALYTIC:
       ApplyQuadraticNoLossAnalytic();
       break;

  case QUADRATIC_WITH_LOSS_ANALYTIC:
       ApplyQuadraticWithLossAnalytic();
       break;

  case QUADRATIC_WITH_LOSS: case COUPLED_ODE: 
  case RATIONAL_QUADRATIC:  case QUADRATIC_NO_LOSS:
       ApplyHeunsMethod();
       break;

  default:
       cout << "Illegal TypeSaturableAbsorption ABORTING!" << endl;
       exit(1);

}  

UpdateFreq();

if(DebugFlag)
  cout << "Total Energy After SaturableAbsorber [pJ] = "
       << 1e12*oOptSignal->GetTotalEnergy() << endl;



} // end :Apply

// ##############################################################

string SaturableAbsorber::WriteTypeSaturableAbsorption(void)
{
 string Type;

 switch(TypeSaturableAbsorption)
 {
  case QUADRATIC_WITH_LOSS_ANALYTIC: 
       Type = "QUADRATIC_WITH_LOSS_ANALYTIC";
       break;

  case QUADRATIC_WITH_LOSS:
       Type = "QUADRATIC_WITH_LOSS";
       break;

  case COUPLED_ODE:
       Type = "COUPLED_ODE";
       break;

  case RATIONAL_QUADRATIC:
       Type = "RATIONAL_QUADRATIC";
       break;

  case QUADRATIC_NO_LOSS:
       Type = "QUADRATIC_NO_LOSS";
       break;

  case QUADRATIC_NO_LOSS_ANALYTIC:
       Type = "QUADRATIC_NO_LOSS_ANALYTIC";
       break;


  case NO_SAT_ABS:
       Type = "NO_SAT_ABS";
       break;



  default:
       Type = "Illegal TypeSaturableAbsorption ABORTING!";
       cout << Type << endl;
       exit(1); 

 } // end switch

 return Type;
}

// ##############################################################

string SaturableAbsorber::WriteTypeStepSizes(void)
{
 string Type;

 switch(TypeStepSizes)
 {
  case LOCAL_ERROR_2ND:
       Type = "LOCAL_ERROR_2ND";
       break;

  case CONSTANT:
       Type = "CONSTANT";
       break;

  default:
       Type = "SaturableAbsorber::WriteTypeStepSizes: Illegal TypeStepSizes ABORTING!";
       cout << Type << endl;
       exit(1);

 } // end switch

 return Type;
}


// ##############################################################

void SaturableAbsorber::ApplyQuadraticNoLossAnalytic(void)
{


 switch(TypeSimulation)
 {
  case SCALAR:
  {
  register cplx *t_X_ptr = sfftM.tOptSignalX;

  cplx * t_X_end_ptr = t_X_ptr + qtPoints;

  double Factor = 2*SaturableAbsorberLength*
                  FastSaturableAbsorptionNonlinearCoefficient;

  // Ignoring 8/9-ths factor here for now at least as would have to
  // think about theory and probbaly not relevant over short lengths
  // typically have in lasers??

  for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++)
   {

    register double Base = 1-Factor*sq(abs(*t_X_ptr));

  if(Base <= 0)
  {
   cout << "SaturableAbsorber::ApplyQuadraticNoLossAnalytic"
        << endl
        << "Singularity in analytic ODE solver. ABORT!" << endl;
   exit(1);
  }

    *t_X_ptr *= pow(Base, -0.5);
   }

  } // end SCALAR block
  break;


  case VECTOR:
  {

   register cplx *t_X_ptr = sfftM.tOptSignalX;
   register cplx *t_Y_ptr = sfftM.tOptSignalY;

   cplx * t_X_end_ptr = t_X_ptr + qtPoints;

   double Factor = 2*SaturableAbsorberLength*
                    FastSaturableAbsorptionNonlinearCoefficient;

   for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++, t_Y_ptr++)
    {
     register double Base = 1-Factor*(sq(abs(*t_X_ptr)) + sq(abs(*t_Y_ptr)));

     if(Base <= 0)
      {
        cout << "SaturableAbsorber::"
             << "ApplyQuadraticNoLossAnalytic"
             << endl
              << "Singularity in analytic ODE solver. ABORT!" << endl;
        exit(1);
       }

     *t_X_ptr *= pow(Base, -0.5);
     *t_Y_ptr *= pow(Base, -0.5);

    }

  } // end VECTOR block
  break;

  default:

   cerr << "SaturableAbsorber::ApplyQuadraticAnalyticNoLoss"
        << endl
        << "Switch error. ABORT!" << endl;
   exit(1);
 } // end switch

} // end ApplyQuadraticAnalyticNoLoss




// #############################################################

void SaturableAbsorber::ApplyQuadraticWithLossAnalytic(void)
{

 switch(TypeSimulation)
 {
  case SCALAR:
  {
  register cplx *t_X_ptr = sfftM.tOptSignalX;
  
  cplx * t_X_end_ptr = t_X_ptr + qtPoints;
  
  double Exponential = exp(2*FastSaturableAbsorptionAttenuationCoefficient*
                             FastSaturableAbsorptionNonlinearCoefficient*
                             SaturableAbsorberLength);

 double Factor = (1-Exponential)/FastSaturableAbsorptionAttenuationCoefficient;
  
  for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++)
   {
   
    register double Base = Factor*sq(abs(*t_X_ptr)) + Exponential;

    if(Base <= 0)
     {
      cout << "SaturableAbsorber::ApplyQuadraticAnalytic"
           << endl
           << "Singularity in analytic ODE solver. ABORT!" << endl;
      exit(1);
     }
  
    *t_X_ptr *= pow(Base, -0.5);
   }
   
  } // end SCALAR block
  break;

  case VECTOR:
  {

   register cplx *t_X_ptr = sfftM.tOptSignalX;
   register cplx *t_Y_ptr = sfftM.tOptSignalY;

   cplx * t_X_end_ptr = t_X_ptr + qtPoints;

  double Exponential = exp(2*FastSaturableAbsorptionAttenuationCoefficient*
                             FastSaturableAbsorptionNonlinearCoefficient*
                             SaturableAbsorberLength);

 double Factor = (1-Exponential)/FastSaturableAbsorptionAttenuationCoefficient;

   for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++, t_Y_ptr++)
    {

     register double Base = Factor*(sq(abs(*t_X_ptr)) + sq(abs(*t_Y_ptr)))
                            + Exponential;

     *t_X_ptr *= pow(Base, -0.5);
     *t_Y_ptr *= pow(Base, -0.5);

    }

  } // end VECTOR block
  break;

  default:

   cerr << "SaturableAbsorber:: ApplyQuadraticAnalytic"
        << endl
        << "Switch error. ABORT!" << endl;

   exit(1);

 } // end switch


} // end ApplyQuadraticAnalytic

// ##############################################################

void SaturableAbsorber::ApplyHeunsMethod(void)
{

 switch(TypeStepSizes)
  {

   case LOCAL_ERROR_2ND: 
     ApplyHeunsMethodLocalError();
     break;

   case CONSTANT:
     ApplyHeunsMethodConstantSteps();
     break;

   default:
     cerr << "ApplyHeunsMethod switch error. ABORT." << endl;
     exit(1);

  }

}

// ##############################################################

void SaturableAbsorber::ApplyHeunsMethodConstantSteps(void)
{
 double StepSizeZ = SaturableAbsorberLength/NumSteps;

 for(int k=0;k<NumSteps;k++)
  ApplyHeunsSingleStep(StepSizeZ);

}

// ##############################################################

void SaturableAbsorber::ApplyHeunsMethodLocalError(void)
{

double DeltaZ;
int  ZStepNumThisFiber = 0;
double RelativeError = 0.0;

  // #### Determine the Initial Step Size in Z ####
  
  // We use the initial step size as computed from the last time we propagated
  // through this particular fiber. If this is the first time
  // through the fiber then DeltaZInitial = 0 will hold, having been
  // set in the constructor, and we choose the initial step so that
  // DeltaZInitial = 0.5*DeltaZMax.



if(!InitializedDeltaZInitialFlag)
 {

  DeltaZInitial = 0.5*DeltaZMax;

  DeltaZ = DeltaZInitial;

  InitializedDeltaZInitialFlag = true;
 }
else
 {
  DeltaZ = DeltaZInitial;
 }

 double CurrentZValueWithinSatAbs = 0;

 // DeltaZ is size of one fine step.
 // 2*DeltaZ is size of one coarse step.

 while(CurrentZValueWithinSatAbs < SaturableAbsorberLength)
  {
    if(DeltaZ > DeltaZMax)
     {
      DeltaZ = DeltaZMax;
     }

   // Work out if next coarse step would exceed the SaturableAbsorber length

   if(2*DeltaZ + CurrentZValueWithinSatAbs > SaturableAbsorberLength)
    {
      DeltaZ =  (SaturableAbsorberLength - CurrentZValueWithinSatAbs)/2.0;
    }
   else
     {
       // Trick to avoid having the very last step being very small

       if(4*DeltaZ + CurrentZValueWithinSatAbs > SaturableAbsorberLength)
         {
          // ## Make the next to last step slightly larger than half of the 
          // ## remainder

          double MaxNextToLastCoarseStepSize =
                  0.55*(SaturableAbsorberLength - CurrentZValueWithinSatAbs);

          if(2*DeltaZ > MaxNextToLastCoarseStepSize)
            {
              DeltaZ =  MaxNextToLastCoarseStepSize/2.0;
            }
         } //## end if

     } // ## end else

  // ######## Propagation starts here ########   

  CopyFftMatricesTimeData(&sfftM,&sfftM_Start);

  // ##### A Coarse step (step size 2*DeltaZ) ### 

  ApplyHeunsSingleStep(2*DeltaZ);

  // ###### Data Management ######

  CopyFftMatricesTimeData(&sfftM,&sfftM_Coarse);

  // Reinstate starting Freq Field

  CopyFftMatricesTimeData(&sfftM_Start,&sfftM);

  // ####  A Fine step ( 2 x step size DeltaZ) 

  ApplyHeunsSingleStep(DeltaZ);
  ApplyHeunsSingleStep(DeltaZ);


  RelativeError = ComputeRelativeError();

  if(RelativeError > 2.0*RelativeErrorGoal) // Failure 
    {
      DeltaZ *= 0.5; // Halve the step size

      CopyFftMatricesTimeData(&sfftM_Start,&sfftM); // Throw away solution

    } // ## end if
  else // Success
    {

      CurrentZValueWithinSatAbs += 2*DeltaZ; // We went two fine steps

      double StepSizeMultiplier = sqrt(2.);

      if(RelativeError > RelativeErrorGoal) // Reduce step size
         {
          DeltaZ /= StepSizeMultiplier;
         }
      else if(RelativeError < 0.5*RelativeErrorGoal) // Increase step size
           {
             DeltaZ *= StepSizeMultiplier;
           }

      // After taking the first step in a fiber we set 
      // DeltaZInitial to be the new DeltaZValue
      // This value will be used next time we go through this fiber

      if(ZStepNumThisFiber==0)
        {
          DeltaZInitial = DeltaZ;
        }

      ZStepNumThisFiber += 2;

    } // ## end else

  } // ## end while(CurrentZValueWithinSatAbs < SatAbsLength)

} // end ApplyHeunsLocalError

// ##############################################################

/*! If S_c and S_f denote the coarse and fine solutions then 
The relative error is defined to be
RelativeError = | S_c - S_f | / | S_f |
*/

double SaturableAbsorber::ComputeRelativeError(void)
{

 double NumSum = 0.0;
 double DenSum = 0.0;
 double RelativeError = 0;

 switch(TypeSimulation)
   {
     case SCALAR:
     {
       cplx * t_X_Coarse_ptr = sfftM_Coarse.tOptSignalX;
       cplx * t_X_Fine_ptr = sfftM.tOptSignalX;

       for(int jj = 0; jj < qtPoints; jj++, t_X_Coarse_ptr++, t_X_Fine_ptr++)
        {
          // sfftM.tOptSignalX contains the fine step result

          NumSum += sq(abs( (*t_X_Coarse_ptr) - (*t_X_Fine_ptr) ));
          DenSum += sq(abs(*t_X_Fine_ptr));

         } // end for jj

       RelativeError = sqrt(NumSum/DenSum);
    }
       break;

    case VECTOR:
     {
       cplx * t_X_Coarse_ptr = sfftM_Coarse.tOptSignalX;
       cplx * t_X_Fine_ptr = sfftM.tOptSignalX;
       cplx * t_Y_Coarse_ptr = sfftM_Coarse.tOptSignalY;
       cplx * t_Y_Fine_ptr = sfftM.tOptSignalY;

       for(int jj = 0; jj < qtPoints;
         jj++,t_X_Coarse_ptr++,t_X_Fine_ptr++,t_Y_Coarse_ptr++,t_Y_Fine_ptr++)
        {
          NumSum += sq(abs( (*t_X_Coarse_ptr) - (*t_X_Fine_ptr) ))
                   +sq(abs( (*t_Y_Coarse_ptr) - (*t_Y_Fine_ptr) ));

          DenSum += sq(abs(*t_X_Fine_ptr)) + sq(abs(*t_Y_Fine_ptr));
 
        } // end for jj

       RelativeError = sqrt(NumSum/DenSum);
    }
       break;

    default:
      cerr << "TypeSolver switch error in"
           << "SaturableAbsorber::ComputeRelativeError"
           << endl;

      exit(1);

     } // ## end switch


   return RelativeError;

} // end ComputeRelativeError

// ############################################################

void SaturableAbsorber::CopyFftMatricesTimeData(fftMatrices * OrigData,
                                                 fftMatrices * CopyOfData)
{
  // void * memcpy(void * dest, const void * src, size_t NumBytes);

 switch(TypeSimulation)
   {

    case SCALAR:

     memcpy(CopyOfData->tOptSignalX,OrigData->tOptSignalX,
            qtPoints*sizeof(cplx));

     break;

   case VECTOR:

     memcpy(CopyOfData->tOptSignalX,OrigData->tOptSignalX,
            qtPoints*sizeof(cplx));

     memcpy(CopyOfData->tOptSignalY,OrigData->tOptSignalY,
            qtPoints*sizeof(cplx));

     break;

   default:
     cerr << "TypeSolver switch error in "
          << "SaturableAbsorber::::CopyFftMatricesTimeData" << endl;

     exit(1);

   } // ## end switch
}

// ##############################################################

void SaturableAbsorber::ApplyHeunsSingleStep(double StepSizeZ)
{

switch(TypeSimulation)
{

 case SCALAR:
  {
   cplx * UInX_ptr = sfftM.tOptSignalX;
   cplx * UInY_ptr = NULL;
   cplx * UOutX_ptr = LaserWorkX;
   double * SAF1_ptr = SAF1; 

   cplx * UInX_end_ptr = UInX_ptr + qtPoints;

   ComputeSAFunction(UInX_ptr,UInY_ptr,SAF1_ptr);

   UInX_ptr = sfftM.tOptSignalX;
   SAF1_ptr = SAF1;

   // First stage of Heun's method.
   // For du/dz = F(u)*u
   // v_{n+1} = u_n*(1+h*F(u_n)) 

   for( ; UInX_ptr < UInX_end_ptr; UInX_ptr++,UOutX_ptr++,SAF1_ptr++ )
       *UOutX_ptr = (*UInX_ptr)*(1 + StepSizeZ*(*SAF1_ptr));

   UInX_ptr = LaserWorkX;
   UInX_end_ptr = UInX_ptr + qtPoints;

   UOutX_ptr = sfftM.tOptSignalX;

   double * SAF2_ptr = SAF2;

   ComputeSAFunction(UInX_ptr,UInY_ptr,SAF2_ptr);

   UInX_ptr = LaserWorkX;
   SAF1_ptr = SAF1;
   SAF2_ptr = SAF2;

  // Second stage of Heun's method.
   // For du/dz = F(u)*u
   // u_{n+1} = u_n*(1+0.5*h*F(u_n)) +0.5*h*F(v_{n+1})v_{n+1} 

   for( ; UInX_ptr < UInX_end_ptr; 
          UInX_ptr++,UOutX_ptr++,SAF1_ptr++,SAF2_ptr++ )
       *UOutX_ptr = *UOutX_ptr + 0.5*StepSizeZ*(
                    (*SAF1_ptr)*(*UOutX_ptr) + (*SAF2_ptr)*(*UInX_ptr));

  } //end case SCALAR

 break;

 case VECTOR:
  {
   cplx * UInX_ptr = sfftM.tOptSignalX;
   cplx * UInY_ptr = sfftM.tOptSignalY;

   cplx * UInX_end_ptr = UInX_ptr + qtPoints;

   cplx * UOutX_ptr = LaserWorkX;
   cplx * UOutY_ptr = LaserWorkY;

   double * SAF1_ptr = SAF1;            
   double * SAF2_ptr = SAF2;    

   ComputeSAFunction(UInX_ptr,UInY_ptr,SAF1_ptr);

   UInX_ptr=sfftM.tOptSignalX;
   UInY_ptr=sfftM.tOptSignalY;
   SAF1_ptr = SAF1;

   for( ; UInX_ptr < UInX_end_ptr;
          UInX_ptr++,UInY_ptr++,UOutX_ptr++,UOutY_ptr++,SAF1_ptr++ )
    {

     *UOutX_ptr = (*UInX_ptr)*(1+ StepSizeZ*(*SAF1_ptr));
     *UOutY_ptr = (*UInY_ptr)*(1+ StepSizeZ*(*SAF1_ptr));

    }

   UInX_ptr = LaserWorkX;
   UInY_ptr = LaserWorkY;

   UInX_end_ptr = UInX_ptr + qtPoints;

   UOutX_ptr = sfftM.tOptSignalX;
   UOutY_ptr = sfftM.tOptSignalY;

   ComputeSAFunction(UInX_ptr,UInY_ptr,SAF2_ptr);

   UInX_ptr = LaserWorkX;
   UInY_ptr = LaserWorkY;
   SAF1_ptr = SAF1;
   SAF2_ptr = SAF2;

   for( ; UInX_ptr < UInX_end_ptr;
          UInX_ptr++,UInY_ptr++,UOutX_ptr++,UOutY_ptr++,SAF1_ptr++,SAF2_ptr++ )
    {

      *UOutX_ptr = *UOutX_ptr + 0.5*StepSizeZ*(
                    (*SAF1_ptr)*(*UOutX_ptr) + (*SAF2_ptr)*(*UInX_ptr));

      *UOutY_ptr = *UOutY_ptr + 0.5*StepSizeZ*(
                    (*SAF1_ptr)*(*UOutY_ptr) + (*SAF2_ptr)*(*UInY_ptr));

    }

  } //end case VECTOR

} // end switch

} // end ApplyHeunsSingleStep

// ############################################################

void SaturableAbsorber::ComputeSAFunction(cplx * U_XIn, cplx * U_YIn, 
                                          double * SAFOut)
{

switch(TypeSaturableAbsorption)
{

case QUADRATIC_WITH_LOSS: case RATIONAL_QUADRATIC:
case QUADRATIC_NO_LOSS:

if(TypeSimulation == SCALAR)
 {
  for(int kk=0;kk<qtPoints;kk++)
   SAFOut[kk] = FastSaturableAbsorptionFunction(U_XIn[kk]);
 }
else
 {
  for(int kk=0;kk<qtPoints;kk++)
   SAFOut[kk] = FastSaturableAbsorptionFunction(U_XIn[kk]) + 
                FastSaturableAbsorptionFunction(U_YIn[kk]);
 }

break;

case COUPLED_ODE:

  // Need Power close to 0 at index 0.
  SAFOut[0] = CoupledODEUnsaturatedNepperLoss;

  for(int kk=0;kk<qtPoints;kk++)
   {
     SAFOut[kk+1] = RungaKutta4Step(U_XIn,U_YIn,SAFOut[kk],kk); 
   }
 
break;

default:
 cerr << "SaturableAbsorber::ComputeSAFunction: Switch error. ABORTING" 
      << endl;

 exit(1);

}

}

// ###########################################################

double SaturableAbsorber::RungaKutta4Step(cplx* U_XIn, cplx * U_YIn,
                          double loss_in, int time_index_in)
{

double loss_out;

double h = oOptSignal->GetDeltaTime();


double PowerStart, PowerMid, PowerEnd;

int t_index[6];

for(int jj=0;jj<6;jj++)
{
  t_index[jj] = oOptSignal->
                WrapPeriodically<int>(time_index_in-2+jj,0,qtPoints);
}

// Use quintic interpolation of Power for midpoint to ensure local
// order of accuracy is DeltaTime^6 for interpolation which is
// no worse than DeltaTime^5 for RK4. Formula derived in JZ note book
// and check in matlab for quintics. 

switch(TypeSimulation)
 { 
  case SCALAR:
   PowerStart = sq(abs(U_XIn[t_index[2]]));
   PowerEnd =   sq(abs(U_XIn[t_index[3]]));
   PowerMid = 3*(sq(abs(U_XIn[t_index[0]])) + sq(abs(U_XIn[t_index[5]])))
             -25*(sq(abs(U_XIn[t_index[1]])) + sq(abs(U_XIn[t_index[4]])))
             +150*(sq(abs(U_XIn[t_index[2]])) + sq(abs(U_XIn[t_index[3]])));
  break;

  case VECTOR:

  PowerStart = sq(abs(U_XIn[2]))+sq(abs(U_YIn[2]));

  PowerEnd =   sq(abs(U_XIn[t_index[3]]))+sq(abs(U_YIn[t_index[3]]));

  PowerMid = 3*(   sq(abs(U_XIn[t_index[0]])) + sq(abs(U_XIn[t_index[5]])))
             -25*( sq(abs(U_XIn[t_index[1]])) + sq(abs(U_XIn[t_index[4]])))
             +150*(sq(abs(U_XIn[t_index[2]])) + sq(abs(U_XIn[t_index[3]])))
             +3*(  sq(abs(U_YIn[t_index[0]])) + sq(abs(U_YIn[t_index[5]])))
             -25*( sq(abs(U_YIn[t_index[1]])) + sq(abs(U_YIn[t_index[4]])))
             +150*(sq(abs(U_YIn[t_index[2]])) + sq(abs(U_YIn[t_index[3]])));

  break;
 }

double F1 = h*CoupledODESatAbsFunction(PowerStart,loss_in);
double F2 = h*CoupledODESatAbsFunction(PowerMid,loss_in+0.5*F1);
double F3 = h*CoupledODESatAbsFunction(PowerMid,loss_in+0.5*F2);
double F4 = h*CoupledODESatAbsFunction(PowerEnd,loss_in+F3);

loss_out = loss_in + (F1+2*F2+2*F3+F4)/6.0;

return loss_out;

}

// ###########################################################

double SaturableAbsorber::CoupledODESatAbsFunction(double Power,double loss)
{

double Value;

Value = -(loss-CoupledODEUnsaturatedNepperLoss)/CoupledODERelaxationTime 
        -(loss*Power)/(CoupledODERelaxationTime*CoupledODESaturationPower);

return Value;

}

// ###########################################################

//! Fast Saturable Absortion is du/dz = F(u)u where
//! F=FastSaturableAbsorptionFunction

double SaturableAbsorber::FastSaturableAbsorptionFunction(cplx u)
{

double OutValue;

switch(TypeSaturableAbsorption)
{

case QUADRATIC_NO_LOSS:

     OutValue = FastSaturableAbsorptionNonlinearCoefficient*sq(abs(u));

     break;


case QUADRATIC_WITH_LOSS:

     OutValue = FastSaturableAbsorptionNonlinearCoefficient*
                (sq(abs(u)) - FastSaturableAbsorptionAttenuationCoefficient);

     break;

case RATIONAL_QUADRATIC:

     OutValue = -FastSaturableAbsorptionNonlinearCoefficient*
                 FastSaturableAbsorptionAttenuationCoefficient/
                (1+ sq(abs(u))/ FastSaturableAbsorptionAttenuationCoefficient);

     break;

default:

cerr << "SaturableAbsorber::FastSaturableAbsorptionFunction" << endl
     << "Switch error. ABORT." << endl;

exit(1);

} // end switch

return OutValue;


}

// ##############################################################

void inline  SaturableAbsorber::UpdateTime(void)
{

  fft->IFFT(sfftM.fOptSignalX, sfftM.tOptSignalX);

  if(TypeSimulation==VECTOR)
    fft->IFFT(sfftM.fOptSignalY, sfftM.tOptSignalY);

}// ## end UpdateTime

// ############################################################

void inline  SaturableAbsorber::UpdateFreq(void)
{

  fft->FFT(sfftM.tOptSignalX, sfftM.fOptSignalX);

  if(TypeSimulation==VECTOR)
    fft->FFT(sfftM.tOptSignalY, sfftM.fOptSignalY);

}// ## end UpdateFreq

// ##############################################################



