// $Id: StatisticsOfArray.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include<stdlib.h>
#include<math.h>

#include "StatisticsOfArray.hh"


StatisticsOfArray::StatisticsOfArray(int arrayLengthIn, double *arrayIn)
  {
    arrayLength = arrayLengthIn;
    array = arrayIn;
    minimum = Minimum();
    maximum = Maximum();
    mean = Mean();
    variance = Variance();
    standardDeviation = sqrt(variance);

    sortedArray = new double[arrayLength];

    SelectionSort(array,sortedArray,arrayLength);

    median = Median(sortedArray,arrayLength);

  }// ## end StatisticsOfArray constructor


// ########################################################

StatisticsOfArray::~StatisticsOfArray()
  {
    //delete [] sortedArray;
    //delete array;
  }
  
// ########################################################

double StatisticsOfArray::Minimum(void)
{

  int i;
  double min;

  min = array[0];

  for(i=0;i<arrayLength;i++)
    {
      if(array[i] < min)
         min = array[i];
    }

  return min;

} // ## end Minimum

// ########################################################

double StatisticsOfArray::Maximum(void)
{

  int i;
  double max;

  max = array[0];

  for(i=0;i<arrayLength;i++)
    {
      if(array[i] > max)
         max = array[i];
    }

  return max;

} // ## end Maximum

// ########################################################

double StatisticsOfArray::Mean(void)
{

  int i;
  double temp_mean = 0.0;

  for(i=0;i<arrayLength;i++)
     temp_mean += array[i];

  return temp_mean/ (double(arrayLength));

} // ## end Mean

// #####################################################

double StatisticsOfArray::MeanSquares(void)
{

  int i;
  double temp_mean = 0.0;

  for(i=0;i<arrayLength;i++)
     temp_mean += array[i]*array[i];

  return temp_mean/ (double(arrayLength));

} // ## end MeanSquares


// ########################################################

double StatisticsOfArray::Variance(void)
{
  int i;
  double temp_var; 

  temp_var=0;

  /*
   for(i=0;i<arrayLength;i++)
    temp_var += (array[i]-mean)*(array[i]-mean);

  return temp_var/ (double(arrayLength) - 1.0);
  */

  return double(arrayLength)*(MeanSquares()-mean*mean)/ 
         double(arrayLength-1.0);

  // This gives an "unbiased" computation of variance

} // ## end StandardDeviation

// ########################################################

double StatisticsOfArray::Median(double * SortedArray, int Length)
{
  
  double Median;

    if(Length%2) // Length odd
      Median = SortedArray[(Length+1)/2 - 1];
    else // Length even
      Median = (SortedArray[Length/2 - 1] + SortedArray[Length/2])/2.0;

  return Median;

} // ## end Median

// #############################################

void StatisticsOfArray::SelectionSort(double * Array, double * SortedArray, 
                                      int Length)
{

  for(int ii=0;ii<Length;ii++)
    SortedArray[ii] = Array[ii];

  for(int SubLength = Length; SubLength >=1; SubLength --)
    {
      int Index = MaximumIndex(SortedArray,SubLength);

      if(Index < SubLength - 1) // If Index == SubLength do nothing
	{
          double Temp = SortedArray[SubLength-1];
          SortedArray[SubLength-1] = SortedArray[Index];
          SortedArray[Index] = Temp;
	} // ## end if

    } // ## end for

} // ## end SelectionSort

// ##############################################

int StatisticsOfArray::MaximumIndex(double * Array, int Length)
{

  double Max = Array[0];
  int MaxIndex = 0;

  for(int ii=0;ii<Length;ii++)
    {
      if(Array[ii] > Max)
	{
          Max = Array[ii];
          MaxIndex = ii;
	}
    }

  return MaxIndex;

} // ## end MaximumIndex

// ##########################################







