// $Id: ocsElecFilter.cc,v 1.4 2008/08/08 02:26:07 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsElecFilter.hh"

extern ofstream *LogFile;

// ################### Methods for the class ElecFilter

// #### Constructor of the class ElecFilter
// ## By: Ivan Lima
ElecFilter::ElecFilter()
{
   ObjectInitialized = 0;
}

// #### Initializes the object ElecFilter
// ## By: Ivan Lima
void ElecFilter::InitElecFilter(cfftw *fft2
                         ,fftPowerMatrices sfftPM2
                         , ElecFilterParam soElecFilter)
{

  // THIS METHOD IS OUT OF DATE (JZ 01/29/01)


   fft = fft2;
   sfftPM = sfftPM2;
   //sfftM = sfftM2;
   //FWHM_ElecBessFilter = FWHM_ElecBessFilter2;
   //OrderElecBessFilter = OrderElecBessFilter2;
   //TypeElecFilter = soElecFilter.TypeElecFilter;
   OrderElecFilter = soElecFilter.OrderElecFilter;
   FreqFWHM_ElecFilter = soElecFilter.FreqFWHM_ElecFilter;
   CenterFreqElecFilter = soElecFilter.CenterFreqElecFilter;     
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;
   ObjectInitialized = 1;      
}

// #########################################################

ElecFilter::ElecFilter(string InFileName, Photodetector *oPhotodetector2)
{
  // NOTE (JZ): If you change this constructor you also need to change
  // the following constructor 
  // ElecFilter(string InFileName,cfftw *fft2,fftPowerMatrices sfftPM2)

  ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
   DebugLevel = oPhotodetector2->GetDebugLevel();

   if(DebugLevel)
     {
     LogFileSeparator();
     *LogFile << InFileName << endl << endl;
     }

   TypeElecFilter = typeElecFilter(ReadInt("TypeElecFilter"," ",0,4,
                       LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel));

   OrderElecFilter = 0;
   FreqFWHM_ElecFilter = 0.0;
   CenterFreqElecFilter = ReadDouble("CenterFreqElecFilter","Hz",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile,
				     DebugLevel);

   SetPhotodetector(oPhotodetector2);

   ObjectInitialized = 1;      

   switch(TypeElecFilter)
     {

      case GAUSSIAN: case BESSEL:

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);

        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

	if(DebugLevel)
	  {
           *LogFile << "This electrical filter is a low-pass filter." << endl
	   << "The input value, f_3dB [GHz] = " 
	   << 0.5*FreqFWHM_ElecFilter*1e-9 << "," << endl
	   << "is the HALF width at half maximum of the filter."
	   << endl
           << "This means that the amplitude of the " << endl
           << "filter transfer function has height 1/sqrt(2) at f_3dB."
	     << endl << endl;
	  }

        break;

       case MEASURED:

	 MeasuredElecFilterFileName = ReadString("MeasuredElecFilterFileName",
						 MANDATORY,"",&InFile,
						 DebugLevel);
        
         ReadMeasuredFilterData();
       
         break;
 
 

       case  INTEGRATE_AND_DUMP:
         //AllocateMemoryForAverageCurrentinBit();
	 *LogFile << " Electrical Filter is Integrate and Dump Filter" << endl;
         break;
 
       case ANRITSU:   // add by Jonathan Hu       

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);
        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);
	 		
        AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
				MANDATORY,"",&InFile,DebugLevel);

	ReadElecGain(AnritsuFilterFileName);
	ReadElecPhase(AnritsuFilterFileName);
	RunAnritsuFilter = 1;
            
         break;

        case GAUSSIAN_AND_ANRITSU:  // add by Paola and John     

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);
        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);
	 		
        AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
				MANDATORY,"",&InFile,DebugLevel);

	ReadElecGain(AnritsuFilterFileName);
	ReadElecPhasePaola(AnritsuFilterFileName);
	RunAnritsuFilter = 1;
 
        break;

     }// ## end switch

   InFile.close();


} // ## end ElecFilter constructor


//#########################################################
//## By: Ivan Lima (5/3/02) 

ElecFilter::ElecFilter(ElecFilter *oElecFilter2, 
		       Photodetector *oPhotodetector2)
{

   //## You need to update the other constructors if you change this

     DebugLevel = oElecFilter2->GetDebugLevel();

   TypeElecFilter  = oElecFilter2->GetTypeElecFilter();
   OrderElecFilter = oElecFilter2->GetOrderElecFilter();
   FreqFWHM_ElecFilter = oElecFilter2->GetFreqFWHM_ElecFilter();
   CenterFreqElecFilter = oElecFilter2->GetCenterFreqElecFilter();   

   SetPhotodetector(oPhotodetector2);

   ObjectInitialized = 1;      

   if ( TypeElecFilter == MEASURED ) {
     MeasuredFilterFreq = NewCplx(qtPoints);
      for(int ii= 0; ii < qtPoints; ii++)  {
         MeasuredFilterFreq[ii] = oElecFilter2->GetMeasuredFilterFreq(ii);
      }    
   }
   
   if( TypeElecFilter == ANRITSU)
      {  // add by Jonathan Hu

        AnritsuFilterFileName = oElecFilter2->GetAnritsuFilterFileName();

        ReadElecGain(AnritsuFilterFileName);
        ReadElecPhase(AnritsuFilterFileName);
       
     }

   if( TypeElecFilter == GAUSSIAN_AND_ANRITSU)
      {  // add by Paola and Zweck

        AnritsuFilterFileName = oElecFilter2->GetAnritsuFilterFileName();

        ReadElecGain(AnritsuFilterFileName);
        ReadElecPhasePaola(AnritsuFilterFileName);
       
     }


} // ## end ElecFilter copy constructor


// #######################################################

ElecFilter::ElecFilter(string InFileName, Photodetector *oPhotodetector2,
		       cplx * MeasuredFilterFreq2, bool ReadMeasuredFilterFlag)
{

  // NOTE (JZ): If you change this constructor you also need to change
  // the following constructor 
  // ElecFilter(string InFileName,cfftw *fft2,fftPowerMatrices sfftPM2)

  ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

  DebugLevel = oPhotodetector2->GetDebugLevel();

   if(DebugLevel)
     {
     LogFileSeparator();
     *LogFile << InFileName << endl << endl;
     }
    
   TypeElecFilter = typeElecFilter(ReadInt("TypeElecFilter"," ",0,4,
                       LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel));

   OrderElecFilter = 0;
   FreqFWHM_ElecFilter = 0.0;
   CenterFreqElecFilter = ReadDouble("CenterFreqElecFilter","Hz",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile,
				     DebugLevel);

   SetPhotodetector(oPhotodetector2);

   ObjectInitialized = 1;      


   switch(TypeElecFilter)
     {

      case GAUSSIAN: case BESSEL:

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);

        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

	if(DebugLevel)
	  {
        *LogFile << "This electrical filter is a low-pass filter." << endl
	   << "The input value, f_3dB [GHz] = " 
	   << 0.5*FreqFWHM_ElecFilter*1e-9 << "," << endl
	   << "is the HALF width at half maximum of the filter."
	   << endl
           << "This means that the amplitude of the " << endl
           << "filter transfer function has height 1/sqrt(2) at f_3dB."
	     << endl << endl;
	  }

        break;

       case MEASURED:

	 MeasuredElecFilterFileName = ReadString("MeasuredElecFilterFileName",
						 MANDATORY,"",&InFile,
						 DebugLevel);


	 // If the pointer MeasuredFilterFreq2 is null then read in
	 // the data from the file named MeasuredElecFilterFileName
	 // Otherwise we assume that the data has already been read in

         if(ReadMeasuredFilterFlag)
	   {
             ReadMeasuredFilterData(MeasuredFilterFreq2);
	   }
         else
	     MeasuredFilterFreq = MeasuredFilterFreq2;
         
         break;

     case INTEGRATE_AND_DUMP:
       AllocateMemoryForAverageCurrentinBit();
       break;

 	   
       case ANRITSU:   // add by Jonathan Hu

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);

        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

	AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
						 MANDATORY,"",&InFile);

	ReadElecGain(AnritsuFilterFileName);
	ReadElecPhase(AnritsuFilterFileName);
       
         break;
		 
        case GAUSSIAN_AND_ANRITSU:  // add by Paola and Zweck

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);

        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

	AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
						 MANDATORY,"",&InFile);

	ReadElecGain(AnritsuFilterFileName);
	ReadElecPhasePaola(AnritsuFilterFileName);
       
         break;

     }// ## end switch

   InFile.close();

 } // ## end ElecFilter constructor


//#######################################################
//## By: Ivan Lima (5/3/02)

void ElecFilter::SetPhotodetector(Photodetector *oPhotodetector2)
{
   oPhotodetector = oPhotodetector2;
   fft = oPhotodetector->fft;
   sfftPM = oPhotodetector->sfftPM;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;

   TimeWindow = oPhotodetector->oOptSignal->GetTimeWindow();
   StringLength = oPhotodetector->oOptSignal->GetStringLength();
   BitLength = oPhotodetector->oOptSignal->GetBitLength();

   //ClockRecoveryTimeForIntegrateAndDump = BitLength/2.0;
   ClockRecoveryTimeForIntegrateAndDump = BitLength; // Anshul
}


// #######################################################

ElecFilter::ElecFilter(string InFileName,cfftw *fft2,fftPowerMatrices sfftPM2)
{

  ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


   DebugLevel = ReadInt("DebugLevel"," ",0,4,
			LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile,0);
   
     if(DebugLevel)
       {
        LogFileSeparator();
        *LogFile << InFileName << endl << endl;
        *LogFile << "DebugLevel = " << DebugLevel << endl;
       }

   TypeElecFilter = typeElecFilter(ReadInt("TypeElecFilter"," ",0,4,
                       LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel));

   OrderElecFilter = 0;
   FreqFWHM_ElecFilter = 0.0;
   CenterFreqElecFilter = ReadDouble("CenterFreqElecFilter","Hz",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile,
				     DebugLevel );



   fft = fft2;
   sfftPM = sfftPM2;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;

   ObjectInitialized = 1;      

   switch(TypeElecFilter)
     {
      case GAUSSIAN: case BESSEL:

        OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);

        FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

	if(DebugLevel)
	  {
        *LogFile << "This electrical filter is a low-pass filter." << endl
	   << "The input value, f_3dB [GHz] = " 
	   << 0.5*FreqFWHM_ElecFilter*1e-9 << "," << endl
	   << "is the HALF width at half maximum of the filter."
	   << endl
           << "This means that the amplitude of the " << endl
           << "filter transfer function has height 1/sqrt(2) at f_3dB."
	     << endl << endl;

	  }

        break;

       case MEASURED:

	 MeasuredElecFilterFileName = ReadString("MeasuredElecFilterFileName",
						 MANDATORY,"",&InFile);

         ReadMeasuredFilterData();

         break;

     case INTEGRATE_AND_DUMP:

       cerr << "ElecFilter::ElecFilter, Don't use this constructor for"
            << " INTEGRATE_AND_DUMP. Aborting Now!" << endl;

       exit(1);

       break;
	  
	  
	 case ANRITSU:  // add by Jonathan Hu       

       OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);
       FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

       AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
						 MANDATORY,"",&InFile);

       ReadElecGain(AnritsuFilterFileName);
       ReadElecPhase(AnritsuFilterFileName);
       
         break;

       case GAUSSIAN_AND_ANRITSU: // add by Paola and Zweck       

       OrderElecFilter = ReadInt("OrderElecFilter"," ",4,5,
                              LOWER_AND_UPPER,MANDATORY,&InFile,DebugLevel);
       FreqFWHM_ElecFilter = 2.0*ReadDouble("f_3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);

       AnritsuFilterFileName = ReadString("AnritsuFilterFileName",
						 MANDATORY,"",&InFile);

       ReadElecGain(AnritsuFilterFileName);
       ReadElecPhasePaola(AnritsuFilterFileName);
       
         break;

     }// ## end switch

   InFile.close();

  

} // ## end ElecFilter constructor

//#######################################################
//## By: Ivan Lima (4/27/02)

void ElecFilter::SetFreqFWHM_ElecFilter(double FreqFWHM_ElecFilter2)
{
  FreqFWHM_ElecFilter = FreqFWHM_ElecFilter2;

   if(DebugLevel)
     {
   LogFileSeparator();

   *LogFile << "WARNING: FreqFWHM_ElecFilter reset to " 
           <<  FreqFWHM_ElecFilter << endl << endl;

     }
}


//#######################################################
//## By: Ivan Lima (5/10/02)
void ElecFilter::Set_f_3dB_ElecFilter(double f_3dB_ElecFilter2)
{
   FreqFWHM_ElecFilter = 2*f_3dB_ElecFilter2;

   if(DebugLevel)
     {
   LogFileSeparator();

   *LogFile << "WARNING: f_3dB_ElecFilter reset to " 
           << f_3dB_ElecFilter2 <<endl
	   << "Thus, FreqFWHM_ElecFilter = "  
           <<  FreqFWHM_ElecFilter << endl << endl;
     }

}


// #########################################################

void ElecFilter::FilterElecSignal(void)
{

// #### Filters the electrical signal
// ## By: Ivan Lima 
// ## Adapted by JZ, Nov 29, 2000

  // FreqFWHM_ElecFilter is the FWHM of the power of the filter
  // in the frequency domain. (JZ)

  if(TypeElecFilter != 0)
    {

   if (! ObjectInitialized) {
      cerr << "Error: Object of class ElecFilter not initialized " 
           << endl;
      exit(1);
   }   

  switch(TypeElecFilter)
    {
      case GAUSSIAN:

        ElectricalFilterGaussian(sfftPM.tPowerSignal, fft, 
                                 CenterFreqElecFilter, FreqFWHM_ElecFilter);
   
        break;

      case BESSEL:

	switch(OrderElecFilter)
	  {

	  case 4:
            ElectricalFilterB4(sfftPM.tPowerSignal, fft, 
                               CenterFreqElecFilter,FreqFWHM_ElecFilter);
            break;
          case 5:
	    ElectricalFilterB5(sfftPM.tPowerSignal, fft, 
                               CenterFreqElecFilter,FreqFWHM_ElecFilter);
            break;

          default:
           cerr << "Error: Bessel ElecFilter of OrderElecFilter " 
		<< OrderElecFilter 
                << " not implemented."      << endl;
           exit(1); 

	  } // ## end switch(OrderElecFilter)

	 break;


      case MEASURED:
          
	ElectricalFilterMeasured();
        break;


    case INTEGRATE_AND_DUMP:

        //IntegrateAndDumpFilter_OLD();
        IntegrateAndDumpFilter(sfftPM.tPowerSignal, fft); // added by Anshul
        break;

    case ANRITSU:

	AnritsuFilter(sfftPM.tPowerSignal, fft,CenterFreqElecFilter,
		      FreqFWHM_ElecFilter);

	break;


    case GAUSSIAN_AND_ANRITSU:

      // add by Paola 13/05: to emulate the filter at the photodetector
	      	        
      ElectricalFilterGaussian(sfftPM.tPowerSignal, fft, 
	  		      CenterFreqElecFilter,FreqFWHM_ElecFilter);
			
      AnritsuFilterPaola(sfftPM.tPowerSignal, fft, 
                   CenterFreqElecFilter,FreqFWHM_ElecFilter);
		
      break;
		
      default:
 
        cerr << "Error: ElecFilter of TypeElecFilter " << TypeElecFilter
             << " not implemented."      << endl;
        exit(1); 
  
    } // ## end TypeElecFilter switch                   

  fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);   // Time2Freq
                
  } // end if Type != 0
  
} // ## end FilterElecSignal

// ###########################################################


void ElecFilter::ElectricalFilterB5(cplx *my_power_time, 
                                    cfftw *fft, double omega_0, 
                                    double fwhm_norm)
{
  
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();
  double phase0125 = arg(Bessel5(fwhm_norm/8.0, omega_0, fwhm_norm)); 
                     // go almost 0.5*fwhm
  for(int i=0; i < N; i++, power_f_p++)
  {
    // delta_f is the same for the splitstep data vector (length N_SS)
    // and for downsampled values b[] since delta_f = 1/T_total
    // CAUTION: delta_f is independent of step number, delta_t not!!
    double freq = fft->freq(i);
    double phase = -freq/(fwhm_norm/8.0) * phase0125; // minus to compensate
    *power_f_p *= cplx(cos(phase), sin(phase))
                  * Bessel5(freq, omega_0, fwhm_norm);

  }
  
  // transform back to time domain
  fft->IFFT(0, my_power_time);

  // Take real part of time signal (i.e. set imag part to be 0)
  // This ensures that the electrical signal is real in time
  // From a theoretical point of view since the Bessel5 filter 
  // is Hermitian: H(omega) = conj(H(-omega)) it follows that
  // if an input signal is real in time (Hermitian in freq)
  // then the output signal will also be Hermitian in freq
  // and hence should be real in time. However the IFFT will
  // introduce a small numerical error, so we correct for that
  // since we're trying to keep all time domain elec signals purely real

    // Uncommented following line to correct amplitude of filtered signal
    // when Gaussian filter not centered at 0. SEM 5/10/02

   if (omega_0 == 0.0) 
    { 
      for(int i=0; i < N; i++)
         my_power_time[i] = real(my_power_time[i]);
    }        
  
  
//    cout << "\n  Values at cutoff (should be 1/sqrt(2)): " << 
//    abs(Bessel5(bessel_FWHM/2, 0, bessel_FWHM)) << "\t" <<
//    abs(Bessel5(-bessel_FWHM/2, 0, bessel_FWHM)) << endl;

} //## end  ElectricalFilterB5()

// ##############################################################

void ElecFilter::ElectricalFilterB4(cplx *my_power_time, 
                                    cfftw *fft, double omega_0, 
                                    double fwhm_norm)
{
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();

  // Written by Ron.
  // Adapted from Bessel5 by John Zweck, Nov 29, 2000

  // The Bessel Filter is not centered. Centered filters make
  // it easier analyse pulse positions 
  // To overcome this problem Ron has shifted the center
  // of the peak of the filter in time by multiplying
  // by exp(i*phase) where phase = phase0125 is the phase
  // of the Bessel filter at a freq equal to 1/8th of its 
  // FWHM. This has the effect of making the magnitude of 
  // the Bessel filter at +- FWHM/2 to be 1/sqrt(2).
  // Because the Bessel filter is asymmetric in time
  // the peak is still a little to the left of the 
  // origin in the time domain. (JZ)

  double phase0125 = arg(Bessel4(fwhm_norm/8.0, omega_0, fwhm_norm)); 
           // go almost 0.5*fwhm

  for(int i=0; i < N; i++, power_f_p++)
  {
    // delta_f is the same for the splitstep data vector (length N_SS)
    // and for downsampled values b[] since delta_f = 1/T_total
    // CAUTION: delta_f is independent of step number, delta_t not!!

    double freq = fft->freq(i);
    double phase = -freq/(fwhm_norm/8.0) * phase0125; // minus to compensate
    *power_f_p *= cplx(cos(phase), sin(phase))*
                Bessel4(freq, omega_0, fwhm_norm);

     }

  // transform back to time domain

  fft->IFFT(0, my_power_time);
  
  // Take real part of time signal (i.e. set imag part to be 0)
  // This ensures that the electrical signal is real in time
  // From a theoretical point of view since the Bessel5 filter 
  // is Hermitian: H(omega) = conj(H(-omega)) it follows that
  // if an input signal is real in time (Hermitian in freq)
  // then the output signal will also be Hermitian in freq
  // and hence should be real in time. However the IFFT will
  // introduce a small numerical error, so we correct for that
  // since we're trying to keep all time domain elec signals purely real

   // Uncommented following line to correct amplitude of filtered signal
   // when Gaussian filter not centered at 0. SEM 5/10/02

   if (omega_0 == 0.0) 
   { 
      for(int i=0; i < N; i++)
         my_power_time[i] = real(my_power_time[i]);
    }

  //  cout << "\n  Values at cutoff (should be 1/sqrt(2)): " << 
  // abs(Bessel4(fwhm_norm/2, 0, fwhm_norm)) << "\t" <<
  //  abs(Bessel4(-fwhm_norm/2, 0, fwhm_norm)) << endl;

} //## end  ElectricalFilterB4()

//###############################################

//by Jonathan Hu
void ElecFilter::AnritsuFilter(cplx *my_power_time, 
                                    cfftw *fft, double omega_0, 
                                    double fwhm_norm)
{
 
/* if(RunAnritsuFilter == 1){

 		LogFile << " data TOutPut here " << RunAnritsuFilter << endl;
   for(int ii=0; ii<fft->give_N();ii++){
		LogFile << ii * fft->delta_t << "  " <<  real(my_power_time[ii]) << endl;
	} 
  		LogFile <<endl <<  " data FOutPut here" << endl;

//   for(int ii=0; ii<fft->give_N();ii++){
//   		if( ii < N/2)
//			LogFile << fft->freq(ii) << "  " <<  abs(power_f_p[ii]) << endl;
//		else
//			LogFile << fft->freq(ii) + -2 *  fft->freq(N/2) << "  " <<  abs(power_f_p[ii]) << endl;
//	}
}
*/  


   
   
   
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  
  const int N = fft->give_N();
  double freq;
  	double G ;	  

/*  for(int i=0; i < N; i++, power_f_p++)
  {
  // Please make sure it's in the right convention 
  //
  //
  //

 		freq = fft->freq(i);			
		if(i==0){
				G = (3*GetElecGain(abs(freq))+GetElecGain(abs(fft->freq(i+1))))/4;
			}else{if(i==N){
				G = (GetElecGain(abs(fft->freq(i-1)))+3*GetElecGain(abs(freq)))/4;
				}else{
					G = (GetElecGain(abs(fft->freq(i-1)))+2*GetElecGain(abs(freq))
					+GetElecGain(abs(fft->freq(i+1))))/4;
					}
				}
				G = GetElecGain(abs(freq));
       *LogFile << freq << "	" << 10*log10(G) 
	   			<< "   " <<GetPhase(abs(freq)) <<  endl;


	
   }
   
   exit(1);
*/
   for(int i=0; i < N; i++, power_f_p++)
  {
  // Please make sure it's in the right convention 

  	
 		freq = fft->freq(i);
			G = GetElecGain(abs(freq));
			
//           *power_f_p *= Gaussian(freq, omega_0, fwhm_norm);
//		   cout <<  Gaussian(freq, omega_0, fwhm_norm) << "  " ;
	   	   if(freq >= 0){
		    	*power_f_p *= G *exp(-1.0 * jc * GetPhase(abs(freq)));
//				LogFile << freq << "  " << GetElecGain(abs(freq)) << "   " << GetPhase(abs(freq)) << endl;
			}else {
				*power_f_p *= G *exp(jc * GetPhase(abs(freq)));
//				LogFile << freq << "  " << GetElecGain(abs(freq)) << "   " << -1.0* GetPhase(abs(freq)) << endl;
			}			

	
   } 
/*  for(int i=0; i < N; i++, power_f_p++)
  {
  // Please make sure it's in the right convention 

  	
 		freq = fft->freq(i);
//           *power_f_p *= Gaussian(freq, omega_0, fwhm_norm);
//		   cout <<  Gaussian(freq, omega_0, fwhm_norm) << "  " ;
	   	   if(freq >= 0){
		    	*power_f_p *= GetElecGain(abs(freq))*exp(-1.0 * jc * GetPhase(abs(freq)));
//				LogFile << freq << "  " << GetElecGain(abs(freq)) << "   " << GetPhase(abs(freq)) << endl;
			}else {
				*power_f_p *= GetElecGain(abs(freq)) *exp(jc * GetPhase(abs(freq)));
//				LogFile << freq << "  " << GetElecGain(abs(freq)) << "   " << -1.0* GetPhase(abs(freq)) << endl;
			}			


	
   }
*/
  fft->IFFT(0, my_power_time);
  
  // Take real part of time signal (i.e. set imag part to be 0)
  // This ensures that the electrical signal is real in time
  // From a theoretical point of view since the Bessel5 filter 
  // is Hermitian: H(omega) = conj(H(-omega)) it follows that
  // if an input signal is real in time (Hermitian in freq)
  // then the output signal will also be Hermitian in freq
  // and hence should be real in time. However the IFFT will
  // introduce a small numerical error, so we correct for that
  // since we're trying to keep all time domain elec signals purely real

   // Uncommented following line to correct amplitude of filtered signal
   // when Gaussian filter not centered at 0. SEM 5/10/02
//cout << " run to here 8 " << endl;

   if (omega_0 == 0.0) 
   { 
      for(int i=0; i < N; i++) {
         my_power_time[i] = -1. * real(my_power_time[i]);
  		 }
   }

} //## end AnritsuFilter()

// #########################################################

void ElecFilter::AnritsuFilterPaola(cplx *my_power_time, 
                                    cfftw *fft, double omega_0, 
                                    double fwhm_norm)
{

  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  
  const int N = fft->give_N();
  double freq;

  

  for(int i=0; i < N; i++, power_f_p++)
  {
  // Please make sure it's in the right convention 
  //
  // by Paola Griggio
  // I added -pi to the phase to not invert the 
  // current of the signal in the time domain 
  // and I added 2*pi*50e-12*abs(freq) to set delay introduced by the filter 0
	
     freq = fft->freq(i);

     // Added by JZ: Normalization by GetElecGain(0) to make filter have no overall gain/loss 

     if(freq >= 0)
      {
         *power_f_p *= GetElecGain(abs(freq))*exp(-1.0 * jc * 
                       (-pi+GetPhase(abs(freq))-(2*pi*50e-12)*abs(freq)))/GetElecGain(0);
  
        // *LogFile << freq << "  " <<-pi+GetPhase(abs(freq))-(50e-12)*
	 //abs(freq)<<"   " << GetPhase(abs(freq)) << endl;
			
      }
     else 
      {
         *power_f_p *= GetElecGain(abs(freq)) *exp(jc * 
                        (-pi+GetPhase(abs(freq))-(2*pi*50e-12)*abs(freq)))/GetElecGain(0);

         //LogFile << freq << "  " << GetElecGain(abs(freq)) << "   " 
	 //<< -1.0* GetPhase(abs(freq)) << endl;
      }			
   }

  fft->IFFT(0, my_power_time);
  
  
   if (omega_0 == 0.0) 
   { 
      for(int i=0; i < N; i++) {
         my_power_time[i] = real(my_power_time[i]);
  		 }
   }

} //## end AnritsuFilterPaola()

// #########################################################

void ElecFilter::ElectricalFilterGaussian(cplx *my_power_time, 
                         cfftw *fft, double omega_0, double fwhm_norm)
{
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();

  // Written by Ron.
  // Adapted from Bessel5 by John Zweck, Nov 29, 2000

  for(int i=0; i < N; i++, power_f_p++)
  {
    // delta_f is the same for the splitstep data vector (length N_SS)
    // and for downsampled values b[] since delta_f = 1/T_total
    // CAUTION: delta_f is independent of step number, delta_t not!!

    double freq = fft->freq(i);
    
    *power_f_p *= Gaussian(freq, omega_0, fwhm_norm);
 }

  // transform back to time domain

  fft->IFFT(0, my_power_time);
  
  // Take real part of time signal (i.e. set imag part to be 0)
  // This ensures that the electrical signal is real in time
  // From a theoretical point of view since the Bessel5 filter 
  // is Hermitian: H(omega) = conj(H(-omega)) it follows that
  // if an input signal is real in time (Hermitian in freq)
  // then the output signal will also be Hermitian in freq
  // and hence should be real in time. However the IFFT will
  // introduce a small numerical error, so we correct for that
  // since we're trying to keep all time domain elec signals purely real

   // Uncommented following line to correct amplitude of filtered signal
   // when Gaussian filter not centered at 0. SEM 5/10/02

    if (omega_0 == 0.0) 
     { 
      for(int i=0; i < N; i++)
         my_power_time[i] = real(my_power_time[i]);
     }     

      /*
  cout << "\n  Values at cutoff (should be 1/sqrt(2)): " << 
  abs(Gaussian(fwhm_norm/2, omega_0, fwhm_norm)) << "\t" <<
  abs(Gaussian(-fwhm_norm/2, omega_0, fwhm_norm)) << endl;
      */

} //## end  ElectricalFilterGaussian()

// ##########################################################

// 5th order normalized Bessel function, see bessel5.m

/*

JZFeb03 

For Physics convention we need to change sign of const cplx s.
This makes sense as Bessel5 is adapted from Matlab code which
uses the Math convention. 

With this change the filter is causal. 

*/

cplx ElecFilter::Bessel5(double omega, double omega_0, double FWHM)
{
  // 5th order
  const double mult = 4.8548214044214655;

  const cplx s = cplx( 0.0, -(omega-omega_0)/FWHM * mult );

  // 5th order normalized polynomial
  return 945.0  /  (945.0 + s * (945.0 + s * (420.0+s * (105.0+s * (15.0+s)  ))));
}

// ############################################################

/*

JZFeb03 

For Physics convention we need to change sign of const cplx s.

With this change the filter is causal. 

*/

//! \todo Check FWHM for elec filters: Should it be for |H| or |H|^2??

cplx ElecFilter::Bessel4(double omega, double omega_0, double FWHM)
{

  // Adapted by JZ from Ron's bessel4.m 

// bessel4() Implements a complex Bessel filter function of degree 4.
// Author: Ronald Holzloehner,  19 May 99
//
// Usage:  B4(omega, omega_0, FWHM)
// Inputs: omega:   (angular) frequency
//         omega_0: center frequency
//         FWHM:    FWHM bandwidth (half max. of |H(s)|^2)
//
// The 4th Bessel polynomial is
// B_4 = s^4 + 10 s^3 + 45 s^2 + 105 s + 105
//   with  s = 2.113917675*2*i*(omega-omega_0)/FWHM
//
// Bandwidth definiton:
//
//  |H(s)|^2 = 1/2  at s = 2.113917675
//
//  follow-up from Ruppert: cut-off freq omega_c is defined as 
//  "-3 dB-point" in |H(omega)|, that is with  dB = 20 log_10(R)
//  (factor 20 due to voltage, not power here), 
//  R = |H(omega_c)|/|H(omega_0)|
//  ==> gives R = 10^(-0.15) = sqrt(2)   (-3 dB should be -3.0103 dB)
//

  const double mult = 4.2278353500419810;

  // The negative sign is required for Physics convention

  const cplx s = cplx( 0.0, -(omega-omega_0)/FWHM * mult );

  return 105.0 /  (105.0 + s*(105.0 + s*(45.0 + s*(10.0+s))));

}

// ############################################################

double ElecFilter::Gaussian(double omega, double omega_0, double FWHM)
{

  // Written by JZ from Ron's optical_filter in ocsTools.cc

// Usage:  Gaussian(omega, omega_0, FWHM)
// Inputs: omega:   (angular) frequency
//         omega_0: center frequency
//         FWHM:    FWHM bandwidth (half max. of |H(s)|^2)
// CAUTION: gives FWHM in intensity, not field !! (checked 21 Jan 00)


  double sigma = FWHM / 2.35482 * M_SQRT2;

  return exp(-0.5*sq((omega-omega_0)/sigma));

}

// ##########################################################

cplx ElecFilter::IntegrateAndDump(double freq)
{

  // Written by Anshul . June '03 
  // H(w) = (exp(iwT) - 1)/iwT.
  
  //BitLength = oPhotodetector->oOptsignal->GetBitLength(); not reqd
  // since SetPhotodetector already does this.
  
  double omega = 2*pi*freq;
   
  if ( omega == 0.0)
      return 1.0; 
  else{
      cplx x = cplx(0.0,omega*BitLength);
      return ((exp(x)-1.0)/x);
  }

}

// ##########################################################

void ElecFilter::WriteFreqPowerGaussianFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }

  
   double freq;

 for(int ii= 0; ii < qtPoints; ii++)
 {

     freq = fft->freq(ii);
   
     OutFile << freq << " " <<
           sq(abs( Gaussian(freq,0.0,FreqFWHM_ElecFilter)))
				 << endl;
  }

   OutFile.close();

} // ## end WriteFreqPowerGaussianFilter

 // #########################################################

void ElecFilter::WriteTimePowerFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }


   for(int ii=0;ii<qtPoints;ii++)
     sfftPM.tPowerSignal[ii] = 0.0;

  
   sfftPM.tPowerSignal[qtPoints/2] = 1.0;

 

    fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);

   FilterElecSignal();
  

OutFile <<  "%(1) Time (2) Abs value of Elec Current (3) Real part Elec Current"
	<< endl;

 for(int ii= 0; ii < qtPoints; ii++)
 {

     
   
     OutFile << ii*DeltaTime << " " 
             << abs(sfftPM.tPowerSignal[ii]) << " "
             << real(sfftPM.tPowerSignal[ii]) << endl;

          
  }

   OutFile.close();

} // ## end WriteTimePowerFilter

 // #########################################################

void ElecFilter::WriteFreqPowerFilter(string OutFileName)
{

  // Written by John Zweck, Feb 14th 2001

 ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }

  
   double freq;

 fft->fftshift(MeasuredFilterFreq);

 for(int ii= 0; ii < qtPoints; ii++)
 {


     freq = ii*DeltaFreq - FreqWindow/2.0;
   
    
     switch(TypeElecFilter)
       {

         case GAUSSIAN:

         OutFile << freq  << " " <<
           sq(abs( Gaussian(freq,0.0,FreqFWHM_ElecFilter)))
				 << endl;
         break;


       case BESSEL:

         if(OrderElecFilter == 4)
             OutFile << freq << " " <<
                sq(abs( Bessel4(freq,0.0,FreqFWHM_ElecFilter)))
				 << endl;

         if(OrderElecFilter == 5)
             OutFile << freq << " " <<
                sq(abs( Bessel5(freq,0.0,FreqFWHM_ElecFilter)))
				 << endl;

         break;

       case MEASURED:

	 OutFile << freq << " " << sq(abs(MeasuredFilterFreq[ii]));

         break;

       case INTEGRATE_AND_DUMP: // Addded by Anshul

          OutFile << freq << '\t' << sq(abs(IntegrateAndDump(freq)))
                 << endl;  	 
      	 break; 

       } // end switch

 } // end for

 fft->ifftshift(MeasuredFilterFreq);

   OutFile.close();

} // ## end WriteFreqPowerFilter

 // #########################################################

double ElecFilter::ComputeMatchAndFilter(void)
{

  // Written by John Zweck, Feb 15th 2001.

  // This method can be used to see how well a filter is
  // matched to a signal. 

  // The method is still very rough and should be used with great
  // caution. Improvements are welcome.

  // The idea is as follows.

  // Suppose the received signal is r = s + n where n is the noise
  // which we assume to be white.

  // The filter response is y = y_s + y_n where y_s = Conv(s,h)
  // where h is the filter function.

  // We define the output SNR to be the ratio of the total energy of y_s
  // divided by the expected value of the total energy of the noise.

  // (This may not be a good definition???)

  // Using Cauchy Schwarz inequality you can show that SNR <= Energy of s.

  // So Match is defined to be sqrt(SNR/Energy(s)) <=1.

  // Match = sqrt(Energy(Conv(h,s))/(Energy(h).Energy(s)))

  // All energies are computed in the freq domain.

  double SignalEnergy = 0.0;

  for(int ii=0;ii<qtPoints;ii++)
    {
      SignalEnergy += sq(abs(sfftPM.fPowerSignal[ii]));
    }


  FilterElecSignal();

   double FilteredSignalEnergy = 0.0;

  for(int ii=0;ii<qtPoints;ii++)
    {
      FilteredSignalEnergy += sq(abs(sfftPM.fPowerSignal[ii]));
    }


  // Compute energy in filter

 double FilterEnergy = 0.0;

 

 for(int ii= 0; ii < qtPoints; ii++)
 {
     double freq = fft->freq(ii);
   
     switch(TypeElecFilter)
       {

         case 1:

         FilterEnergy +=
           sq(abs( Gaussian(freq,0.0,FreqFWHM_ElecFilter)));
			
         break;


       case 2:

         if(OrderElecFilter == 4)
          FilterEnergy +=
	   sq(abs( Bessel4(freq,0.0,FreqFWHM_ElecFilter)));
				

         if(OrderElecFilter == 5)
           FilterEnergy +=
	   sq(abs( Bessel5(freq,0.0,FreqFWHM_ElecFilter)));
	

         break;
      
       case 4:

         FilterEnergy += sq(abs( IntegrateAndDump(freq)));

         break;
       } // end switch

 } // end for


 cout << "SignalEnergy = " << SignalEnergy << endl;
 cout << "FilterEnergy = " << FilterEnergy << endl;
 cout << "FilteredSignalEnergy = " << FilteredSignalEnergy << endl;

 double Match =  sqrt(FilteredSignalEnergy/(SignalEnergy*FilterEnergy));

 cout << "Match = " << Match << endl;

 return Match;


} // ## end ComputeMatchingCoefficientAndFilter

// ################################################

 /*! Called by the constructor

     The input data is assumed to consist of three columns: \n
     (1) Freq relative to central freq with same discretization as 
         in our code.\n
     (2) Real part of Measured Filter at freq given by col (1)\n
     (3) Imag part of Measured Filter at freq given by col (1)\n

      \sa string MeasuredElecFilterFileName,
          WriteFreqCplxFilter(string OutFileName),
          ElecFilter(string InFileName, Photodetector *oPhotodetector),
          ElecFilter(string InFileName, Photodetector *oPhotodetector,
                     cplx * MeasuredFilterFreq),
          ElecFilter(string InFileName, cfftw *fft2,fftPowerMatrices sfftPM2)
  */

void ElecFilter::ReadMeasuredFilterData(void)
{
  // Written by John Zweck, Mar 13th 2002

  cout << "In ElecFilter::ReadMeasuredFilterData" << endl;

  MeasuredFilterFreq = NewCplx(qtPoints);

  ifstream InFile(MeasuredElecFilterFileName.c_str(),ios::in);

   if (!InFile) 
     {
      cerr << "Error: " << MeasuredElecFilterFileName
	   << "could not be opened."<<endl;
      exit(1);
     }

 double CodeFreq, InFileFreq;
 double RealPart,ImagPart;

  for(int ii= 0; ii < qtPoints; ii++)
  {

   CodeFreq = ii*DeltaFreq - FreqWindow/2.0; // Hz
       
   InFile >> InFileFreq >>  RealPart >> ImagPart;
                
   if( abs(InFileFreq-CodeFreq) > DeltaFreq/2.0)
     {
       cerr << "InFileFreq = " << InFileFreq << " "
	    << "CodeFreq = " << CodeFreq << endl;

       exit(1);
     }

   MeasuredFilterFreq[ii] = RealPart + jc*ImagPart;

  } // end for

 
 fft->ifftshift(MeasuredFilterFreq);

 InFile.close();

 //WriteFreqCplxFilter("MeasuredFilterCheck.dat");

} // ## end ReadMeasuredFilterData

// ##########################################################

void ElecFilter::ReadMeasuredFilterData(cplx * MeasuredFilterFreq2)
{

  ReadMeasuredFilterData();

  for(int ii= 0; ii < qtPoints; ii++)
    MeasuredFilterFreq2[ii]  = MeasuredFilterFreq[ii];
}

// #####################################################

void ElecFilter::WriteFreqCplxFilter(string OutFileName)
{

  // Written by John Zweck, Mar 13th 2002

 ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }

  
   OutFile << "% (1) Freq rel central frequency of simulation window [GHz] " 
	   << endl
	   << "% (2) Real part of filter   (3) Imag part of filter  " << endl;


 if(TypeElecFilter == MEASURED)
  fft->fftshift(MeasuredFilterFreq);

 double freq;

 for(int ii= 0; ii < qtPoints; ii++)
 {

     freq = ii*DeltaFreq - FreqWindow/2.0;
       
     switch(TypeElecFilter)
       {

         case GAUSSIAN:

         OutFile << 1e-9*freq  << " " 
                 << Gaussian(freq,0.0,FreqFWHM_ElecFilter) << " "
                 << 0 << endl;

         break;


       case BESSEL:

         if(OrderElecFilter == 4)
             OutFile << freq << " " 
                     << real(Bessel4(freq,0.0,FreqFWHM_ElecFilter)) << " "
                     << imag(Bessel4(freq,0.0,FreqFWHM_ElecFilter)) << endl;

         if(OrderElecFilter == 5)
             OutFile << freq << " " 
                     << real(Bessel5(freq,0.0,FreqFWHM_ElecFilter)) << " "
                     << imag(Bessel5(freq,0.0,FreqFWHM_ElecFilter)) << endl;

         break;

       case MEASURED:
	 OutFile << freq << " " 
                 << real(MeasuredFilterFreq[ii]) << " "
                 << imag(MeasuredFilterFreq[ii]) << endl;


         break;


       case INTEGRATE_AND_DUMP: // Added by Anshul
          OutFile << freq << '\t' << real(IntegrateAndDump(freq)) << '\t'
	                          << imag(IntegrateAndDump(freq)) << endl;
                                    

         break;


       } // end switch

 } // end for

 if(TypeElecFilter == MEASURED)
  fft->ifftshift(MeasuredFilterFreq);

 

   OutFile.close();

} // ## end WriteFreqCplxFilter

// #####################################################

void ElecFilter::ElectricalFilterMeasured(void)
{

  Time2Freq();
  
  for(int ii=0;ii<qtPoints;ii++)
    {
      sfftPM.fPowerSignal[ii] *= MeasuredFilterFreq[ii];
    }

  Freq2Time();
  
  // Take real part of time signal (i.e. set imag part to be 0)
  // This ensures that the electrical signal is real in time
 
  for(int ii=0; ii < qtPoints; ii++)
    sfftPM.tPowerSignal[ii] = real(sfftPM.tPowerSignal[ii]);

  //Time2Freq();

} // ## end ElectricalFilterMeasured

// ##################################################

void ElecFilter::GetClockRecoveryTime(BitString *oBitString)
{
	int ch=0; //channel number, always equal to 0 in this case

	int b[3];
	
	int i;
	int pat[3];
	pat[0]=0;
	pat[1]=1;
	pat[2]=0;

	int PatternLength=3;

	int pos=0,kk=0,CurBit;
	while(pos<3)
	{
		if(kk==StringLength)kk=0;
		
		CurBit=oBitString->GetBit(ch,kk);
		if(CurBit==pat[pos])	pos++;
		else 
		{
			if(CurBit==pat[0])	pos=1;
			else pos=0;
		}
		kk++;
	}

	int CentralBitPos=kk-2;
	if(CentralBitPos<0)CentralBitPos+=8;

	double sum1=0;
	double sum2=0;

	int left;//BUGGY!!!! =CentralBitPos*BitLength-BitLength/2;
	int right;// =(CentralBitPos+1)*BitLength+BitLength/2;

	for(int i=0;i<2*BitLength;i++)
	{
		sum1+=(i+1)*real(sfftPM.tPowerSignal[left+i]);
		sum2+=real(sfftPM.tPowerSignal[left+i]);
	}

	ClockRecoveryTimeForIntegrateAndDump=abs(sum1)/abs(sum2);

} // ## end GetClockRecoveryTime

// ###########################################################

void ElecFilter::IntegrateAndDumpFilter_OLD(void)
{

  if(!AverageCurrentInBit)
    {
      cerr << "ElecFilter::IntegrateAndDumpFilter, Malloc error. ABORTING"
	   << endl;
      exit(1);
    }


  int left,right;
  for(int kk=0;kk<StringLength;kk++)
  {
   // BUGGY left=kk*BitLength+ClockRecoveryTimeForIntegrateAndDump-BitLength/2;
    // BUGGY  right=left+BitLength;

         //AverageCurrentInBit[kk] = 
	 //GetAverageCurrentInTimeRange(kk*BitLength,(kk+1)*BitLength);

	 AverageCurrentInBit[kk] = 
		 GetAverageCurrentInTimeRange(left,right);
  }

  int NumPtsPerBit = qtPoints/StringLength;

  for(int ii=0;ii<qtPoints;ii++)
      sfftPM.tPowerSignal[ii] = AverageCurrentInBit[ii/NumPtsPerBit];

  // Finally take FFT !!

  Time2Freq();

} // ## end IntegrateAndDumpFilter_OLD

// ################################################

// Added by Anshul .. June 2003. 
// adapted from ElectricalFilterGaussian.

void ElecFilter::IntegrateAndDumpFilter(cplx *my_power_time,cfftw *fft)
{
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();


  for(int i=0; i < N; i++, power_f_p++)
  {
    double freq = fft->freq(i);
    
    *power_f_p *= IntegrateAndDump(freq);
  }

  // transform back to time domain

  fft->IFFT(0, my_power_time);
 
  for(int i=0; i < N;i++){   
      my_power_time[i] = real(my_power_time[i]);//..not sure Anshul
  }
  
} //## end  IntegrateAndDumpFilter


// ###############################################

void ElecFilter::AllocateMemoryForAverageCurrentinBit(void)
{
  AverageCurrentInBit = new double[StringLength];

} // ## end AllocateMemoryForAverageCurrentinBit

// ##################################################

//! Returns the integral of the current in a time interval divided
//! by the width of the time interval

/*! We assume LeftTime >= 0 and RightTime >= LeftTime.
    LeftTime >= TimeWindow is OK though.
    If RightTime >= TimeWindow we wrap.
    We also check that the width of the time window is less than the
    TimeWindow for the signal.
*/


double ElecFilter::GetAverageCurrentInTimeRange(double LeftTime, 
                                                double RightTime)
{

  while(LeftTime >= TimeWindow)
    {
      LeftTime -= TimeWindow;
      RightTime -= TimeWindow;
    }

  double LengthTimeRange = RightTime - LeftTime;

  if(LengthTimeRange <= 0 || LengthTimeRange > TimeWindow || LeftTime < 0)
    {
      cerr << "ElecFilter::GetAverageCurrentInTimeRange, error. Aborting Now"
	   << endl;
    }


  int NumPtsInRange = int(floor(LengthTimeRange/DeltaTime));

  int left_ii = int(ceil(LeftTime/DeltaTime));
  int right_ii = left_ii + NumPtsInRange;

  //cout << "left_ii = " << left_ii << " right_ii = " << right_ii << endl;

  int ii;

  double Current = 0.0;

  if(right_ii < qtPoints)
    {
      for (ii = left_ii; ii < right_ii; ii++) 
        Current += real(sfftPM.tPowerSignal[ii]);
    }
  else    //## right_ii > qtPoints
    {
      for (ii = left_ii; ii < qtPoints; ii++) 
        Current += real(sfftPM.tPowerSignal[ii]);

      for (ii = 0; ii < right_ii - qtPoints; ii++) 
        Current += real(sfftPM.tPowerSignal[ii]);

    }

 return Current*DeltaTime/LengthTimeRange;


} // ## end GetAverageCurrentInTimeRange

// ##################################################

void ElecFilter::Time2Freq(void)
{
   fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);    
}

// ###################################################

void ElecFilter::Freq2Time(void)
{
   fft->IFFT(sfftPM.fPowerSignal,sfftPM.tPowerSignal);    
}

// ###################################################

//## By: Ivan Lima (8/20/02)
void ElecFilter::WriteFilterImpulseResponse(
                 string FileNameImpulseResponseTime,
                 string FileNameImpulseResponseFreq)
{
   Photodetector *oPhotodetector2 = new Photodetector(*oPhotodetector);
   //*oPhotodetector2 = *oPhotodetector;  

   oPhotodetector->GenerateImpulseCurrent(); 
   
   FilterElecSignal();
   
   /*
   //cplx ElecFilterDC_Component = oPhotodetector2->sfftPM.fPowerSignal[0];
  
   //##  Get the effective electrical filter bandwidth
   double EffectiveElecFilterBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      EffectiveElecFilterBandwidth += 
         real(oPhotodetector->sfftPM.fPowerSignal[ii]);
   }
   EffectiveElecFilterBandwidth *= oPhotodetector->GetDeltaFreq();
   cout << "Effective ElecFilter bandwidth (GHz) = " 
        << EffectiveElecFilterBandwidth*1e-9 << endl;   
   */
   oPhotodetector->WriteFileTimePower(FileNameImpulseResponseTime);
   oPhotodetector->WriteFileFreqPower(FileNameImpulseResponseFreq);  
   	
   *oPhotodetector = *oPhotodetector2;
   delete oPhotodetector2;
}

//####################################################################3

//## The spectral width includes the negative frequencies
//## By: Ivan Lima (8/20/02)
double ElecFilter::GetCurrentEquivalentSpectralWidth(void)
{
   Photodetector *oPhotodetector2 = new Photodetector(*oPhotodetector);
   //*oPhotodetector2 = *oPhotodetector;  

   oPhotodetector->GenerateImpulseCurrent(); 
   
   FilterElecSignal();
   
   cplx ElecFilterDC_Component = oPhotodetector->sfftPM.fPowerSignal[0];
  
   int qtPoints = oPhotodetector->GetQtPoints();
   //##  Get the effective electrical filter bandwidth
   double CurrentEquivalentSpectralWidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      CurrentEquivalentSpectralWidth += 
         real(oPhotodetector->sfftPM.fPowerSignal[ii]);
   }
   CurrentEquivalentSpectralWidth *= oPhotodetector->GetDeltaFreq();

   *oPhotodetector = *oPhotodetector2;
   delete oPhotodetector2;
   return CurrentEquivalentSpectralWidth
          /abs(ElecFilterDC_Component);
}


// ###########################################################

// add phase information for Elec. Amp.
// by: Jonathan Hu


// ############################################################


void ElecFilter::ReadElecPhase(string InFileName)
{

 char C_String[40];
 sprintf(C_String,"Phase.in");
 InFileName = InFileName + C_String;
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
    

 InFile >>  ElecPhaseStartFreq;
 InFile >>  ElecPhaseStepFreq;
 InFile >>  NumElecPhaseWidth;
  
 *LogFile << "ElecPhaseStartFreq = " << ElecPhaseStartFreq << " Hz" << endl; 
 *LogFile << "ElecPhaseStepFreq = " << ElecPhaseStepFreq << " Hz" << endl; 
 *LogFile << "NumElecPhaseWidth = " << NumElecPhaseWidth  << endl; 

 yElecPhase = new double [NumElecPhaseWidth+1];

 
 for(int ii=1;ii<=NumElecPhaseWidth;ii++)
    {  
		InFile >> temp;
	   if(temp < ElecPhaseStartFreq || temp > ElecPhaseStartFreq+ ElecPhaseStepFreq*  NumElecPhaseWidth){
	   		cerr << "Error: in  ElecGain Freq input."<< endl;
    		exit(1);
    	}
       InFile >> yElecPhase[ii];
//	   LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
 InFile >> temp;
 *LogFile << "File end. " << temp << endl;
 if(temp >0 ){
	   		cerr << "Error: in  ElecPhase Freq input."<< endl;
	   		cerr << " temp = "<< temp <<  endl;
    		exit(1);
    	}

 InFile.close(); 
 xElecPhase = new double [NumElecPhaseWidth+1];
 for(int ii=1;ii<=NumElecPhaseWidth;ii++)
    {  
		xElecPhase[ii] = ElecPhaseStartFreq + (ii-1) * ElecPhaseStepFreq;
//	   LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
 y2ElecPhase = new double [NumElecPhaseWidth+1];
 GaussianSmooth(yElecPhase, 3, NumElecPhaseWidth);
 spline(xElecPhase,  yElecPhase, NumElecPhaseWidth ,  0,  0,  y2ElecPhase);




} // end  ReadElecPhase

// ######################################################################


void ElecFilter::ReadElecPhasePaola(string InFileName)
{

 char C_String[40];
 sprintf(C_String,"Phase.in");
 InFileName = InFileName + C_String;
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
    

 InFile >>  ElecPhaseStartFreq;
 InFile >>  ElecPhaseStepFreq;
 InFile >>  NumElecPhaseWidth;
  
 *LogFile << "ElecPhaseStartFreq = " << ElecPhaseStartFreq << " Hz" << endl; 
 *LogFile << "ElecPhaseStepFreq = " << ElecPhaseStepFreq << " Hz" << endl; 
 *LogFile << "NumElecPhaseWidth = " << NumElecPhaseWidth  << endl; 

 yElecPhase = new double [NumElecPhaseWidth+1];

 
 for(int ii=1;ii<=NumElecPhaseWidth;ii++)
    {  
		InFile >> temp;
	   if(temp < 0 || temp > 2.1e10){
	   		cerr << "Error: in  ElecGain Freq input."<< endl;
    		exit(1);
    	}
       InFile >> yElecPhase[ii];
//	   LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
 InFile >> temp;
 *LogFile << "File end. " << temp << endl;
 if(temp >0 ){
	   		cerr << "Error: in  ElecPhase Freq input."<< endl;
	   		cerr << " temp = "<< temp <<  endl;
    		exit(1);
    	}

 InFile.close(); 
 xElecPhase = new double [NumElecPhaseWidth+1];
 for(int ii=1;ii<=NumElecPhaseWidth;ii++)
    {  
		xElecPhase[ii] = ElecPhaseStartFreq + (ii-1) * ElecPhaseStepFreq;
//	   LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
 y2ElecPhase = new double [NumElecPhaseWidth+1];
 GaussianSmooth(yElecPhase, 3, NumElecPhaseWidth);
 spline(xElecPhase,  yElecPhase, NumElecPhaseWidth ,  0,  0,  y2ElecPhase);


} // end  ReadElecPhase

// ########################################################################

double ElecFilter::GetPhase(double Frequency)
{ 
    if(Frequency<0){
		cerr << "Error: negative Frequency."<< endl;
    	exit(1);
	}
	if(Frequency<0||Frequency>ElecPhaseStartFreq+ElecPhaseStepFreq*NumElecPhaseWidth){
		return( (Frequency-ElecPhaseStartFreq)/ElecPhaseStepFreq/NumElecPhaseWidth*
			(GetPhase(ElecPhaseStartFreq+ElecPhaseStepFreq*NumElecPhaseWidth)
			-GetPhase(ElecPhaseStartFreq)) + GetPhase(ElecPhaseStartFreq));
		
//		return( ( Frequency - 1e9 ) * ( -92.332 + 1.511 ) / 19e9 -1.511 ) ; // for Anritsu
 	}

	return(	splint( xElecPhase, yElecPhase ,  y2ElecPhase, NumElecPhaseWidth, Frequency));

 
// this is for linear interpolation
/*
 double Interpolate;
 int InterpolateIndex;
 ElecPhaseStartFreq = 0;
 ElecPhaseStepFreq = 50000000;
 NumElecPhaseWidth = 402;
 InterpolateIndex = int((Frequency - ElecPhaseStartFreq)/ ElecPhaseStepFreq)+1;
 
 if(InterpolateIndex < -1e-9 ){
      cout << "Error: Frequency is not in the right range"<<endl;
	  cout << " Frequency= " << Frequency << " ElecGainStepFreq = " << ElecGainStepFreq 
	  		<< " ElecGainStartFreq =" << ElecGainStartFreq << endl;
	  cout << " InterpolateIndex = " << InterpolateIndex << endl;
      exit(1);	
 }     
// cout << " run to here GetPhase" << endl;
 if(Frequency>ElecPhaseStartFreq+ElecPhaseStepFreq*NumElecPhaseWidth)
 	return( ( Frequency - 1e9 ) * ( -92.332 + 1.511 ) / 19e9 -1.511 ) ;
 
 Interpolate = yElecPhase[InterpolateIndex] + 
 	(yElecPhase[InterpolateIndex + 1] - yElecPhase[InterpolateIndex])
	* ( ( Frequency - ElecPhaseStartFreq  -  ElecPhaseStepFreq * ( InterpolateIndex-1) ) 
	/ ElecPhaseStepFreq  ) ;
// cout << " run to here end GetPhase" << endl;
 
 return( Interpolate ) ;

// return( ( Frequency - 1e9 ) * ( -92.332 + 1.511 ) / 19e9 -1.511 ) ;
// return( ( Frequency - 1e9 ) * ( -92.332 + 1.511 ) / 19e9 -1.511 ) ;
*/
}  // Get Phase

// ############################################################

void ElecFilter::ReadElecGain(string InFileName)
{

 char C_String[40];
 sprintf(C_String,"Gain.in");
 InFileName = InFileName + C_String;
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
    



 InFile >>  ElecGainStartFreq;
 InFile >>  ElecGainStepFreq;
 InFile >>  NumElecGainWidth;
  
 *LogFile << "ElecGainStartFreq = " << ElecGainStartFreq << " Hz" << endl; 
 *LogFile << "ElecGainStepFreq = " << ElecGainStepFreq << " Hz" << endl; 
 *LogFile << "NumElecGainWidth = " << NumElecGainWidth  << endl; 

 yElecGain = new double [NumElecGainWidth+1];

 
 for(int ii=1;ii<NumElecGainWidth+1;ii++)
    {  
		InFile >> temp;
	   if(temp < ElecGainStartFreq || temp > ElecGainStartFreq + ElecGainStepFreq * NumElecGainWidth){
	   		cerr << "Error: in  ElecGain Freq input."<< endl;
    		exit(1);
    	}
       InFile >> yElecGain[ii];
//	   *LogFile << temp << "  " << ElecGain[ii] << endl;
    }

 InFile >> temp;
 if(temp >0 ){
	   		cerr << "Error: in  ElecGain Freq input."<< endl;
	   		cerr << " temp = "<< temp <<  endl;
    		exit(1);
    	}
  GainPeak= ReadDouble("GainPeak","",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);
  FreqGain3dB = ReadDouble("FreqGain3dB","Hz",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);


 *LogFile << "File end. " << temp << endl;

 InFile.close();

 xElecGain = new double [NumElecGainWidth+1];
 for(int ii=1;ii<=NumElecGainWidth;ii++)
    {  
		xElecGain[ii] = ElecGainStartFreq+ (ii-1) * ElecGainStepFreq;
//	   *LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
 y2ElecGain = new double [NumElecGainWidth+1];
 GaussianSmooth(yElecGain, 3, NumElecGainWidth);

 spline(xElecGain,  yElecGain, NumElecGainWidth ,  0,  0,  y2ElecGain);

} // end  ReadElecGain

// #########################################################

double ElecFilter::GetElecGain(double Frequency)
{   



/*double j;
for(double i = 0;i<=50e9;i+=1e8) {
	j = i;
	*LogFile << i << "  " << splint( xElecGain, yElecGain ,  y2ElecGain, NumElecGainWidth, j) << "  " << j << endl;
	}
exit(1);
*/	
	if(Frequency<0){
		cerr << "Frequency calcuation wrong "
           << " aborting."<< endl;
      	exit(1);
	}

//	return(exp((-3./sq(50e9)*sq(Frequency)+31)/10*log(10)));
//		return(exp((-3./sq(FreqGain3dB)*sq(Frequency)+GainPeak)/10*log(10)));
		
//		return(exp((-3./sq(25e9)*sq(Frequency)+9.5)/10*log(10)));
		
	if(Frequency>ElecGainStartFreq+ElecGainStepFreq*NumElecGainWidth){
/*		if(AnritsuFilterFileName[0]=='S'){
			return(exp((-3./sq(50e9)*sq(Frequency)+31)/10*log(10)));
		}else{
			cerr << " not SHF Amp aborting" << endl;
			exit(1);
		}
*/		
		return(exp((-3./sq(FreqGain3dB)*sq(Frequency)+GainPeak)/10*log(10)));


		return(GetElecGain(ElecGainStartFreq+ElecGainStepFreq*NumElecGainWidth));
	}
	double splt;
	splt = splint( xElecGain, yElecGain ,  y2ElecGain, NumElecGainWidth, Frequency);
	return(exp(splt/10*log(10)) ) ;
// This is for linear interpolation
/* double Interpolate;
 int InterpolateIndex;
 ElecGainStartFreq = 0;
 ElecGainStepFreq = 50000000;
 NumElecGainWidth = 402;
 InterpolateIndex = int((Frequency - ElecGainStartFreq)/ ElecGainStepFreq) + 1;
// cout << "run to here 10" << endl;
 if(InterpolateIndex < -1e-9 ){
      cout << "Error: Frequency is not in the right range"<<endl;
	  cout << " Frequency= " << Frequency << " ElecGainStepFreq = " << ElecGainStepFreq 
	  		<< " ElecGainStartFreq =" << ElecGainStartFreq << endl;
	  cout << " InterpolateIndex = " << InterpolateIndex << endl;
      exit(1);	
	  
	  
 }
 
 if(Frequency>ElecGainStartFreq+ElecGainStepFreq*NumElecGainWidth)
 	return(0);
 
 Interpolate = yElecGain[InterpolateIndex] + 
 	(yElecGain[InterpolateIndex + 1] - yElecGain[InterpolateIndex])
	* ( ( Frequency - ElecGainStartFreq  -  ElecGainStepFreq * (InterpolateIndex -1)) 
	/ ElecGainStepFreq  ) ;
//cout <<  yElecGain[InterpolateIndex] << " + " 
//	<<   (yElecGain[InterpolateIndex + 1] - yElecGain[InterpolateIndex]) << " * "
//	<<  "( " << ( Frequency - ElecGainStartFreq  -  ElecGainStepFreq * (InterpolateIndex-1) ) 
//	<<	"	/  " << ElecGainStepFreq  << ")" << endl ;
//cout << " Interpolate = " << Interpolate << endl;
// return (Interpolate);
 return( exp(Interpolate/10*log(10)) ) ;
*/
} // endl 

// ######################################3
void ElecFilter::spline(double *x,  double *y,  int n, double yp1, double ypn,  double *y2)
{
	int i,k;
	double p,qn,sig,un;

	double * u ;
	u = new double [n];

//	u=vector(1,n-1);
	if (yp1 > 0.99e30)
		y2[1]=u[1]=0.0;
	else {
		y2[1] = -0.5;
		u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}
	for (i=2;i<=n-1;i++) {
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2[i-1]+2.0;
		y2[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}
	if (ypn > 0.99e30)
		qn=un=0.0;
	else {
		qn=0.5;
		un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
	}
	y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
	for (k=n-1;k>=1;k--)
		y2[k]=y2[k]*y2[k+1]+u[k];
//	free_vector(u,1,n-1);
	delete [] u;
}
// ################################################################

double ElecFilter::splint( double *xa,  double *ya,  double *y2a, int n, double x)
{
	void nrerror(char error_text[]);
	int klo,khi,k;
	double h,b,a;

	klo=1;
	khi=n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
	}
	h=xa[khi]-xa[klo];
	if (h == 0.0) nrerror("Bad xa input to routine splint");
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	return(a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0);
}

// #########################################################

double ElecFilter::GaussianSmooth(double *y, int Order, int N )
{	double sigma = 2;
	double Normalize=0;
	double * result = new double [N+1];
	for (int ii=-1*Order;ii<= Order; ii ++){
		Normalize += exp(-1*sq(ii)/2/sq(sigma));
	}
//	cout << "Normalize = " << Normalize << endl;
	for (int ii=Order+1;ii<= N-Order-1; ii ++){
		result[ii] = 0;
		for (int jj=-1*Order;jj<= Order; jj ++){
			result[ii] += y[ii+jj] * exp(-1*sq(jj)/2/sq(sigma));
		}		
	}	
	for (int ii=Order+1;ii<= N-Order-1; ii ++){
//		LogFile << ii << "   " << y[ii] << "   " << result[ii]/Normalize << endl;
		y[ii] = result[ii]/Normalize;
	}	
	delete [] result;

}
// #########################################################

















