// $Id: ocsLumpedLoss.cc,v 1.2 2006/05/11 22:30:33 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsLumpedLoss.hh"

extern ofstream *LogFile;

//###############################################################
//## By: Ivan Lima (08/08/02)

LumpedLoss::LumpedLoss(double LossdB2, OptSignal *oOptSignal2)
{
   InitLumpedLoss(LossdB2, oOptSignal2);
}


//###############################################################
//## By: Ivan Lima (08/08/02)

LumpedLoss::LumpedLoss(string InFileName, OptSignal *oOptSignal2)
{
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;
   ifstream InFile(InFileName.c_str(),ios::in);    
   if (!InFile) {
      cerr << "Error: Input file "<< InFileName 
           << "could not be opened."<< endl;
      exit(1);
   }
   double LossdB2 = ReadDouble("LossdB"," ",1,-1,NO_BOUNDS,&InFile);
   InFile.close();
   InitLumpedLoss(LossdB2, oOptSignal2);
} // ## end constructor

//###############################################################
//## By: Ivan Lima (08/08/02)

void LumpedLoss::InitLumpedLoss(double LossdB2, OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   sfftM = oOptSignal->sfftM;
   TypeSimulation = oOptSignal->GetTypeSimulation();
   *LogFile << "TypeSimulation = " << WriteTypeSimulation(TypeSimulation) 
	   << endl;
   LossdB = LossdB2;
   LossLinear = dB2Linear(LossdB);
   LossLinearSqrt = sqrt(LossLinear);
   *LogFile << "LossLinear = " << LossLinear << endl;

   oReduLumpedLoss = new ReduLumpedLoss(LossdB,oOptSignal->oReduOptSignal);

}

// #####################################################

void LumpedLoss::ApplyLumpedLoss(void)
{

double InPower,OutPower;

oOptSignal->GainOfAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
            1.0/LossLinear);

 oReduLumpedLoss->ApplyLumpedLoss();

switch(TypeSimulation)
  {

    case SCALAR:

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]/LossLinearSqrt;

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.fOptSignalX[jj] = sfftM.fOptSignalX[jj]/LossLinearSqrt;

   
    break;

     case VECTOR:

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.tOptSignalX[jj] = sfftM.tOptSignalX[jj]/LossLinearSqrt;

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.tOptSignalY[jj] = sfftM.tOptSignalY[jj]/LossLinearSqrt;

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.fOptSignalX[jj] = sfftM.fOptSignalX[jj]/LossLinearSqrt;

     for(int jj=0;jj<oOptSignal->GetQtPoints();jj++)
      sfftM.fOptSignalY[jj] = sfftM.fOptSignalY[jj]/LossLinearSqrt;

    break;

  default: 

    cerr << "TypeSimulation switch error in LumpedLoss::ApplyLumpedLoss" 
         << endl;
     exit(1);         


  } // ## end switch

} // ## end ApplyLumpedLoss

//###########################################################
//## Set a new value for the loss in dB
//## By: Ivan Lima (7/13/02)
void LumpedLoss::SetLossdB(double LossdB2)
{
   LossdB = LossdB2;
   LossLinear = dB2Linear(LossdB);
   LossLinearSqrt = sqrt(LossLinear);
}


//##########################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (8/08/02)

// Not set up for Reduced stuff

void LumpedLoss::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   sfftM = oOptSignal2->sfftM;   
}






