// $Id:
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "ocsOptDataWriter.hh"

extern ofstream *LogFile;

// #####################################################

OptDataWriter::OptDataWriter(string InFileName, string InDir2,string Job2, 
                             OptSignal * Signal2)
{

  Job = Job2;
  Signal = Signal2;

ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
 {
    cerr << "Error: Input file "<< InFileName 
         << " could not be opened."<< endl;
    exit(1);
 }

 LogFileSeparator();

 *LogFile << "OptDataWriter Constructor" << endl << endl;


WriteComplexFieldsFlag = ReadInt("WriteComplexFields"," ",0,1,
                              LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

WritePowerChannelFlag = ReadInt("WritePowerChannel"," ",0,1,
                              LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


ChannelArrayIndexForPower = ReadInt("ChannelArrayIndexForPower",
                          " ",1,-1, LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);


WriteDistanceTimePowerFlag =  ReadInt("WriteDistanceTimePower"," ",0,1,
                                LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


WriteDistanceTimePowerWithinMapFlag = 
                                ReadInt("WriteDistanceTimePowerWithinMap",
                                               " ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);



WriteDistanceTimePowerPhaseFlag =  ReadInt("WriteDistanceTimePowerPhase",
                                            " ",0,1,
                                LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


WriteDistanceTimePowerPhaseWithinMapFlag = 
                     ReadInt("WriteDistanceTimePowerPhaseWithinMap"," ",0,1,
                             LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


 if(WriteDistanceTimePowerPhaseWithinMapFlag)
   {
     WriteDistanceTimePowerWithinMapFlag = 0;
   }


 if(WriteDistanceTimePowerPhaseFlag)
   WriteDistanceTimePowerFlag = 0;




TimeIndexIncrement =  ReadInt("TimeIndexIncrementForDistanceTime"," ",1,-1,
                             LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);

 *LogFile << "TimeIndexIncrement for DTP = " << TimeIndexIncrement << endl;


WriteInitialOptDataFlag = ReadInt("WriteInitialOptData"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


WriteFinalOptDataFlag = ReadInt("WriteFinalOptData"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

OmitDCFlag = !(ReadInt("KeepDCFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,1,&InFile));


WritePulseWidthFlag = ReadInt("WritePulseWidth"," ",0,1,
                              LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


WriteAverageChirpFlag = ReadInt("WriteAverageChirp"," ",0,1,
                              LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

WriteLocalFreqFlag = ReadInt("WriteLocalFreq"," ",0,1,
                              LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);


WriteAccDispFlag = ReadInt("WriteAccDisp"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);
 
OutputStepSizesFlag = ReadInt("OutputStepSizesFlag"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

WritePulseWidthInFiberFlag =  ReadInt("WritePulseWidthInFiberFlag"," ",0,1,
                                       LOWER_AND_UPPER,OPTIONAL_NO_WARNING,
                                       0,&InFile);

WriteDOPFlag =  bool(ReadInt("WriteDOPFlag"," ",0,1,
                                       LOWER_AND_UPPER,OPTIONAL_NO_WARNING,
                                       0,&InFile));

 
Signal->OpenDegreeOfPolarizationFile(Job + "DOP.dat",WriteDOPFlag);

 if(WriteDOPFlag)
   {

    double AbsMinFreqDOP = ReadDouble("AbsMinFreqDOP","Hz",0,-1,
                               LOWER_ONLY,OPTIONAL_NO_WARNING,&InFile);

    double AbsMaxFreqDOP = ReadDouble("AbsMaxFreqDOP","Hz",0,-1,
                               LOWER_ONLY,OPTIONAL_NO_WARNING,&InFile);

    RelMinFreqDOP =  AbsMinFreqDOP - Signal->GetCenterFreq();
    RelMaxFreqDOP =  AbsMaxFreqDOP - Signal->GetCenterFreq();

    *LogFile << "[RelMinFreqDOP,RelMaxFreqDOP] [GHz] = [" 
	    << 1e-9*RelMinFreqDOP << "," << 1e-9*RelMaxFreqDOP << "]" << endl;

   } // ## end if(WriteDOPFlag)

WritePowerMapFlag = ReadInt("WritePowerMap"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

 if(WritePowerMapFlag)
   {
     //24.828e9 = 0.2 nm

     double PowerMeterResolutionBandWidthHz = 
                 ReadDouble("ResolutionBandWidthHz"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,24.828e9,&InFile);

     double SuperGaussExponent  = ReadDouble("SuperGaussExponent"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);

     int  TypeAmplifierNoise =  ReadInt("TypeAmplifierNoise"," ",0,4,
                               LOWER_ONLY,OPTIONAL_NO_WARNING,2,&InFile);   

     PowerMeter = new OptSpectrumAnalyzer(Signal,
                      PowerMeterResolutionBandWidthHz,1,SuperGaussExponent,
					  TypeAmplifierNoise);
     
     PowerMeter->SetJob(Job);

   }

  Signal->SetWriteAccumulatedFirstOrderDispersionFlag(WriteAccDispFlag);

  Signal->OpenAccumulatedFirstOrderDispersionFile(Job + "AccDisp.dat");
  Signal->OpenAveragePulseWidthFile(Job +"PulseWidth.dat",WritePulseWidthFlag);

ComputeTimeFreqShifts = ReadInt("ComputeTimeFreqShifts"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

ComputeTimingJitter = ReadInt("ComputeTimingJitter"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

ComputeMeanZerosStdDevOnes = ReadInt("ComputeMeanZerosStdDevOnes"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile);

ComputeEnergyInBitSlot = ReadInt("ComputeEnergyInBitSlot"," ",-1,-2,
                               LOWER_ONLY,OPTIONAL_NO_WARNING,-1,&InFile);

// ComputeEnergyInBitSlot < 0 means don't do it. Otherwise it gives 
// Index of Bit Slot to compute energy in


  if(ComputeTimeFreqShifts || ComputeTimingJitter || 
     ComputeMeanZerosStdDevOnes|| ComputeEnergyInBitSlot >=0)
    {

      PEA = new PulseEvolutionAnalyzer(Signal,InDir2,Job,
                                       ComputeTimeFreqShifts,
                                       ComputeTimingJitter,
                                       ComputeMeanZerosStdDevOnes,
                                       ComputeEnergyInBitSlot);
    } // end if


  InFile.close();

} // end OptDataWriter

// ####################################################

void OptDataWriter::WriteInitialData(void)
{

  if(WriteInitialOptDataFlag)
    {
       Signal->WriteFileTimePower(Job + "tInOpt.dat");
       //Signal->WriteFileFreqPowerdBm(Job + "fInOpt.dat");
       Signal->WriteFileFreqPower(Job + "fInOpt.dat",OmitDCFlag);
       //Either use WriteFileFreqPower for linear scale output or 
       //WriteFileFreqPowerdBm for log scale output
    }

  if(WritePowerMapFlag)
     PowerMeter->RunPowerMeter(0.0);

  if(WriteDistanceTimePowerPhaseFlag)
      {
       Signal->WriteDistTimePowerPhase(Job + "DTPP.dat", 
                                       TimeIndexIncrement,0.0,0);
    
       NumSlicesZ = 1; 
      }


  if(WriteDistanceTimePowerFlag)
      {
       Signal->WriteDistTimePower(Job + "DTP.dat", 
                                  TimeIndexIncrement,0.0,0);
    
       NumSlicesZ = 1; 
      }


 if(WritePowerChannelFlag)
   Signal->WritePowerdBmChannel(Job + "PowerInChannel.dat",
				      0.0,0,ChannelArrayIndexForPower-1);


 if(WriteDOPFlag)
   Signal->WriteDegreeOfPolarization(0.0,RelMinFreqDOP,RelMaxFreqDOP);

if(WriteAccDispFlag)
    Signal->WriteAccumulatedFirstOrderDispersion(0.0);

if(ComputeTimeFreqShifts || ComputeTimingJitter || 
   ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0)
   PEA->ApplyPulseAnalyzer(0.0);

 if(WriteLocalFreqFlag)
   Signal->WriteLocalFrequencyAndChirp(Job + "LocalFreqChirpIn.dat");


} // ## end WriteInitialData

// ####################################################

void OptDataWriter::WriteWithinMap(int expt_num, int map_number, 
                              double PropagatedLength)
{

  // The first expt_num must be 0 and first map_number must be 0

  if(expt_num == 0)
    {
      if(WriteAccDispFlag)
       Signal->WriteAccumulatedFirstOrderDispersion(PropagatedLength);

      if(WritePulseWidthFlag)
       Signal->WriteAveragePulseWidth(PropagatedLength,
                                      int(WritePulseWidthFlag));


      if(WritePowerMapFlag)
       PowerMeter->RunPowerMeter(PropagatedLength);

     if(map_number==0 && WritePowerFlag)
	 {
	   // Signal->WriteAveragePowerdBm(Job + "Power.dat",
	   //                        PropagatedLength,0);
	  // The last argument in this function call is not used any more

	 }

     if(map_number==0 && WritePowerChannelFlag)
	 {
          Signal->WritePowerdBmChannel(Job + "PowerInChannel.dat",
				       PropagatedLength,1,
                                       ChannelArrayIndexForPower-1);

       // The 2nd to last argument in this function call is not used any more
	 }

     if(WriteDistanceTimePowerWithinMapFlag && WriteDistanceTimePowerFlag)
	 {
           Signal->WriteDistTimePower(Job + "DTP.dat",
                                        TimeIndexIncrement,
                                        PropagatedLength,1);
           NumSlicesZ++; 

	 }

     if(WriteDistanceTimePowerPhaseWithinMapFlag && 
        WriteDistanceTimePowerPhaseFlag)
	 {
           Signal->WriteDistTimePowerPhase(Job + "DTPP.dat",
                                        TimeIndexIncrement,
                                        PropagatedLength,1);
           NumSlicesZ++; 

	 }


 if(WriteDOPFlag)
   Signal->WriteDegreeOfPolarization(PropagatedLength,
				     RelMinFreqDOP,RelMaxFreqDOP);

if(ComputeTimeFreqShifts || ComputeTimingJitter||
   ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0)
      PEA->ApplyPulseAnalyzer(PropagatedLength);

    } // end if expt_index

} // ## end OptDataWriter::WriteWithinMap

// ########################################################

void OptDataWriter::WriteEndMap(int expt_num, int map_number, 
                           double PropagatedLength)
{
  // The first expt_num must be 0 and first map_number must be 0

  if(expt_num == 0)
    {
      if(WriteAccDispFlag)
       Signal->WriteAccumulatedFirstOrderDispersion(PropagatedLength);

      if(WritePulseWidthFlag)
       Signal->WriteAveragePulseWidth(PropagatedLength,
                                      int(WritePulseWidthFlag));


      if(WritePowerMapFlag)
       PowerMeter->RunPowerMeter(PropagatedLength);

     if(map_number==0 && WritePowerFlag)
	 {
	   //Signal->WriteAveragePowerdBm(Job + "Power.dat",
	   //                        PropagatedLength,0);
	  // The last argument in this function call is not used any more

	 }

     if(map_number==0 && WritePowerChannelFlag)
	 {
          Signal->WritePowerdBmChannel(Job + "PowerInChannel.dat",
				       PropagatedLength,1,
                                       ChannelArrayIndexForPower-1);

       // The 2nd to last argument in this function call is not used any more
	 }

     if(WriteDistanceTimePowerFlag)
	 {
           Signal->WriteDistTimePower(Job + "DTP.dat",
                                        TimeIndexIncrement,
                                        PropagatedLength,1);
           NumSlicesZ++; 

	 }

     if(WriteDistanceTimePowerPhaseFlag)
	 {
           Signal->WriteDistTimePowerPhase(Job + "DTPP.dat",
                                        TimeIndexIncrement,
                                        PropagatedLength,1);
           NumSlicesZ++; 

	 }

 if(WriteDOPFlag)
   Signal->WriteDegreeOfPolarization(PropagatedLength,
				     RelMinFreqDOP,RelMaxFreqDOP);


if(ComputeTimeFreqShifts || ComputeTimingJitter || 
   ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0)
    PEA->ApplyPulseAnalyzer(PropagatedLength);

    } // end if expt_index


} // ## end WriteEndMap

//###############################################

void OptDataWriter::WriteEndTransmission(int expt_num,int map_num,
                                    double PropagatedLength)
{


  WriteEndMap(expt_num,map_num,PropagatedLength);
  TurnOffWriteFlags(expt_num);

  if(expt_num == 0)
    {

     if(WriteFinalOptDataFlag)
       {
        Signal->WriteFileTimePower(Job + "tOutOpt.dat");
        //Signal->WriteFileFreqPowerdBm(Job + "fOutOpt.dat");
	Signal->WriteFileFreqPower(Job + "fOutOpt.dat",OmitDCFlag);
       //Either use WriteFileFreqPower for linear scale output or 
       //WriteFileFreqPowerdBm for log scale output
       
    if(WriteDOPFlag)
       Signal->WriteDegreeOfPolarization(PropagatedLength,
                                         RelMinFreqDOP,RelMaxFreqDOP);

     if(WriteComplexFieldsFlag)
         Signal->WriteComplexFields(Job + "tComplexFields.dat",0);

       }
   
     /*
if(ComputeTimeFreqShifts || ComputeTimingJitter || 
   ComputeMeanZerosStdDevOnes || ComputeEnergyInBitSlot>=0
)
     PEA->ApplyPulseAnalyzer(PropagatedLength);
     */

    } // end if 


 if(WriteLocalFreqFlag)
   Signal->WriteLocalFrequencyAndChirp(Job + "LocalFreqChirpOut.dat");

} // ## end OptDataWriter::WriteEndTransmission

// ############################################################

void OptDataWriter::TurnOffWriteFlags(int expt_num)
{

  if(expt_num == 0)
    {
     Signal->CloseAccumulatedFirstOrderDispersionFile();
     Signal->SetWriteAccumulatedFirstOrderDispersionFlag(0);
     Signal->CloseAveragePulseWidthFile(WritePulseWidthFlag);
     Signal->CloseAveragePulseWidthFile(WriteDOPFlag);

     WritePulseWidthFlag = 0; // Only write out for 1st MC expt
     WriteDOPFlag = 0;

    } 

   Signal->ResetAccumulatedFirstOrderDispersionToZero();

} // ## end OptDataWriter::TurnOffWriteFlags

// ############################################################









