// $Id: ocsOptDemuxer.cc,v 1.2 2006/05/11 22:31:30 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptDemuxer.hh"

extern ofstream *LogFile;


// ######################################

OptDemuxer::OptDemuxer(OptSignal* WDM_Signal2, OptFilter* TunableFilter2)
{

  WDM_Signal = WDM_Signal2;
  TunableFilter = TunableFilter2;
  //  SingleChannelSignal = NULL;
  SingleChannelSignal = new OptSignal(*WDM_Signal);

  // For Ciena work we need:
  DoAdditionalFilteringFlag = TunableFilter->GetDoAdditionalFiltering();

} // end constructor

// ######################################
//## By: Ivan Lima, adapting from John Zweck's constructor
OptDemuxer::OptDemuxer(OptSignal* WDM_Signal2, OptFilter* TunableFilter2,
                       OptSignal* SingleChannelSignal2)
{
   WDM_Signal = WDM_Signal2;
   TunableFilter = TunableFilter2;

/*
   if(SingleChannelSignal2 == NULL) {
      SingleChannelSignal = new OptSignal(*WDM_Signal);
   }
   else {
      SingleChannelSignal = SingleChannelSignal2;          
   }  
*/

SingleChannelSignal = SingleChannelSignal2;

  // For Ciena work we need (JZ):
  DoAdditionalFilteringFlag = TunableFilter->GetDoAdditionalFiltering();


} // end constructor


// #########################################

OptDemuxer::~OptDemuxer()
{
}// end destructor

// #########################################

OptSignal& OptDemuxer::ExtractChannel(int ChannelArrayIndex)
{

  // Note (JZ Dec 4th, 2000) ExtractChannel method altered to take
  // a channel array index as input!!!
 
  /*  
  if(SingleChannelSignal != NULL)
    {
      SingleChannelSignal->~OptSignal();
      SingleChannelSignal = NULL;
    }
 

  SingleChannelSignal = new OptSignal(*WDM_Signal);
  */


  
  SingleChannelSignal->SetNumChannels(WDM_Signal->GetNumChannels());
  *SingleChannelSignal = *WDM_Signal;

  // FIX:  Later I should change the array Frequency to be of length one
  // and extract the right Frequency. Same for InJones and InStokes.
  // However for now we don't need these parameters so leave as for
  // multichannel signal


  // These two commented out lines are incorrect as we are not
  // changing the center freq of the slowly-varying envelope apprxn.
  // JZ April 6th 2001.
  //  SingleChannelSignal->SetCenterFreq(
  //				 WDM_Signal->GetFrequency(ChannelArrayIndex));
   
  SingleChannelSignal->SetNumberChannelsForDemuxer(1);
   

  SingleChannelSignal->SetFrequency(0,
                       WDM_Signal->GetFrequency(ChannelArrayIndex) );
   


  SingleChannelSignal->oBitString->ReduceToSingleChannelBitString(
                                   ChannelArrayIndex); 

  SingleChannelSignal->SetAccumulatedFirstOrderDispersion(
       WDM_Signal->GetAccumulatedFirstOrderDispersion());

  SingleChannelSignal->SetAccumulatedSecondOrderDispersion(
       WDM_Signal->GetAccumulatedSecondOrderDispersion());


  // cout << "ExtractChannel" << endl;

  // SingleChannelSignal->oBitString->WriteBitString();
  

  // See April 6th comment above by JZ for why these two lines
  // also commented out and replaced below.
  //double RelativeCenterFreq = SingleChannelSignal->GetCenterFreq()
  //                            - WDM_Signal->GetCenterFreq();

   double RelativeCenterFreq = WDM_Signal->GetFrequency(ChannelArrayIndex)
                             - WDM_Signal->GetCenterFreq();

  TunableFilter->SetRelativeCenterFreq(RelativeCenterFreq);
  
  TunableFilter->SetOpticalData(SingleChannelSignal->fft,
                                SingleChannelSignal->sfftM);

  TunableFilter->TunableFilterOptSignal();

  //TunableFilter->WriteFreqPowerGaussianFilter("OptFilter.dat");
  //exit(1);

  //TunableFilter->WriteFreqPowerSantecFilter("OptFilterSantec.dat");
  //exit(1);


  return *SingleChannelSignal;

} // end ExtractChannel

// #########################################

/*! 

This method calls OptDemuxer::PrepareToDemuxChannel 
and then optically filters the signal. 

To use the method make the following calls:

SingleChannelSignal = new OptSignal(*WDMSignal);

TunableOptFilter = new OptFilter(InDir + "TunableOptFilter.in",
			         SingleChannelSignal);
  
oOptDemuxer = new OptDemuxer(WDMSignal,TunableOptFilter,SingleChannelSignal);

oOptDemuxer->DemuxChannel(ChannelArrayIndex);

The last call can be made multiple times throughout the code to demux
the current WDMSignal, and can have different ChannelArrayIndexes
for the different calls.

The method is more robust than calling 
OptSignal& OptDemuxer::ExtractChannel(int ChannelArrayIndex)

It is very similar to 

void OptDemuxer::ExtractSingleChannel(int ChannelArrayIndex)

The main difference is that in the semianalytical receiver
you want to call 

void OptDemuxer::PrepareToDemuxChannel(ChannelArrayIndex)

separately from calling void OptFilter::FilterOptSignal(void)

JZ April 2003.

*/


void OptDemuxer::DemuxChannel(int ChannelArrayIndex)
{
  PrepareToDemuxChannel(ChannelArrayIndex);
  TunableFilter->FilterOptSignal();
}

// #####################################

  /*! This method sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver
  */

void OptDemuxer::PrepareToDemuxChannel(int ChannelArrayIndex)
{
  
  SingleChannelSignal->SetNumChannels(WDM_Signal->GetNumChannels());
  *SingleChannelSignal = *WDM_Signal;

  // FIX:  Later I should change the array Frequency to be of length one
  // and extract the right Frequency. Same for InJones and InStokes.
  // However for now we don't need these parameters so leave as for
  // multichannel signal


  SingleChannelSignal->SetNumberChannelsForDemuxer(1);
   
  SingleChannelSignal->SetFrequency(0,
                       WDM_Signal->GetFrequency(ChannelArrayIndex) );
   
  SingleChannelSignal->oBitString->ReduceToSingleChannelBitString(
                                   ChannelArrayIndex); 

   double RelativeCenterFreq = WDM_Signal->GetFrequency(ChannelArrayIndex)
                             - WDM_Signal->GetCenterFreq();

  TunableFilter->SetRelativeCenterFreq(RelativeCenterFreq);
  

} // end PrepareToDemuxChannel


// #########################################
//## There is a difference in which the single 
//#  channel is being extracted. The main 
//#  difference is that there is no dynamical
//#  realocation of the SingleChannelOptSignal. (IL)
//## By: John Zweck and Ivan Lima  
void OptDemuxer::ExtractSingleChannel(int ChannelArrayIndex)
{

  // Note (JZ Dec 4th, 2000) ExtractChannel method altered to take
  // a channel array index as input!!!
 

  SingleChannelSignal->SetNumChannels(WDM_Signal->GetNumChannels() );

  *SingleChannelSignal = *WDM_Signal;
  
  SingleChannelSignal->ReduceToSingleChannel(ChannelArrayIndex);  
 
  double RelativeCenterFreq = WDM_Signal->GetFrequency(ChannelArrayIndex)
                             - WDM_Signal->GetCenterFreq();

  TunableFilter->SetRelativeCenterFreq(RelativeCenterFreq);
  
  TunableFilter->SetOpticalData(SingleChannelSignal->fft,
                                SingleChannelSignal->sfftM);

  TunableFilter->TunableFilterOptSignal();

} // end ExtractChannel

// #########################################

OptSignal& OptDemuxer::ExtractChannelCenterFreq(double CenterFreqOptFilter,
                                                BitString * SCBitString,
						bool DoFilterFlag)
{

  // Written by JZ July 20th 2001
  // Used when want to extract channel using it's center freq 
  // rather than its channel index  
  
  /*
  if(SingleChannelSignal != NULL)
    {
      SingleChannelSignal->~OptSignal();
      SingleChannelSignal = NULL;
    }
 

  SingleChannelSignal = new OptSignal(*WDM_Signal);
  */


  SingleChannelSignal->SetNumChannels(WDM_Signal->GetNumChannels());
  *SingleChannelSignal = *WDM_Signal;   
  SingleChannelSignal->SetNumberChannelsForDemuxer(1);

  SingleChannelSignal->SetFrequency(0,CenterFreqOptFilter);


  SingleChannelSignal->oBitString->ReduceToSingleChannelBitString(
                                   SCBitString); 

 
 
  if(DoFilterFlag)
    {
 

  // See April 6th comment above by JZ for why thesde two lines
  // also commented out and replaced below.
  //double RelativeCenterFreq = SingleChannelSignal->GetCenterFreq()
  //                            - WDM_Signal->GetCenterFreq();

   double RelativeCenterFreq = CenterFreqOptFilter
                             - WDM_Signal->GetCenterFreq();

  TunableFilter->SetRelativeCenterFreq(RelativeCenterFreq);
  
  //cout << "RelativeCenterFreq = " << RelativeCenterFreq << endl;

  TunableFilter->SetOpticalData(SingleChannelSignal->fft,
                                SingleChannelSignal->sfftM);

  TunableFilter->TunableFilterOptSignal();

  //TunableFilter->WriteFreqPowerGaussianFilter("OptFilter.dat");
  //exit(1);


    }// end if DoFilterFlag

  return *SingleChannelSignal;

} // end ExtractChannelCenterFreq

// #########################################





