// $Id: ocsOptFiberLocalError.cc,v 1.7 2008/08/08 02:26:53 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptFiberLocalError.hh"

extern ofstream *LogFile;

OptFiberLocalError::OptFiberLocalError(string InFileName, 
                                       OptSignal *oOptSignal2, RanNumGen *RNG2,
                                       string Job2) 
{

  // Written by John Zweck Dec 19, 2000

oOptSignal = oOptSignal2;
OutputPDVector = oOptSignal->OutputPDVectorOFLE;
TypeSimulation = oOptSignal->GetTypeSimulation();
Job = Job2;
RNG = RNG2;

// added by Hai 7-1-03

JonesMatrix=oOptSignal->JonesMatrixChannel;
NumberOfChannels=oOptSignal->GetNumChannels();
FreqChannels = new double[NumberOfChannels];

for (int ii=0;ii<NumberOfChannels;ii++)
{
        FreqChannels[ii]=oOptSignal->GetFrequency(ii); // real freq.
}


ifstream InFile(InFileName.c_str(),ios::in);    

if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
LogFileSeparator();
*LogFile << InFileName << endl << endl;


TypeSolver = typeSolver(ReadInt("TypeSolver"," ",1,3,LOWER_AND_UPPER,&InFile));

*LogFile << "Solver Type = " << WriteTypeSolver() << endl;

 switch(TypeSolver)
   {
    case SCALAR_NLS:

      if(TypeSimulation==VECTOR)
	{
          cerr << "OptFiberLocalError constructor: "
	       << endl
	       << "TypeSimulation and TypeSolver in conflict"
	       << endl;
          exit(1);

	}

      break;

    case VECTOR_MANAKOV_PMD: case VECTOR_MANAKOV_NO_PMD:

       if(TypeSimulation==SCALAR)
	{
          cerr << "OptFiberLocalError constructor: "
	       << endl
	       << "TypeSimulation and TypeSolver in conflict"
	       << endl;
          exit(1);

	}

      break;

   } // end switch


TypeStepSizes = typeStepSizes(
		ReadInt("TypeAlgorithm"," ",1,4,LOWER_AND_UPPER,
			MANDATORY,1,&InFile));

*LogFile << "TypeStepSize = " << WriteTypeStepSizes() << endl;

DoTimeOperatorExactlyFlag = ReadInt("DoTimeOperatorExactlyFlag"," ",0,1,
				    LOWER_AND_UPPER,
				    OPTIONAL_NO_WARNING,0,&InFile);

TypeRamanAmplification = typeRamanAmplification(
                          ReadInt("TypeRamanAmplification"," ",0,2,
                                  LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,
								  &InFile));

ScalarSolverEightNinthsFlag = ReadInt("ScalarSolverEightNinthsFlag"," ",
				      0,1,LOWER_AND_UPPER,&InFile);

LengthFiber = ReadDouble("LengthFiber","m",0,-1,
                                    LOWER_ONLY,&InFile);
ReferenceFreq = 0;
ReferenceWavelength = 0;

ReferenceFreq = ReadDouble("ReferenceFreq","Hz",0,-1,
			LOWER_ONLY,&InFile);

ReferenceWavelength = ReadDouble("ReferenceWavelength","m",0,-1,
			LOWER_ONLY,&InFile);


if(ReferenceFreq == 0)
   {
     ReferenceFreq = Wavelength2Frequency(ReferenceWavelength);
  
     *LogFile << endl 
             << "Since ReferenceFreq is zero we use ReferenceWavelength:"
             << endl
             << endl
             << "ReferenceWavelength [nm] = " << ReferenceWavelength*1e9 
             << endl
             << "ReferenceFreq [GHz] = " << ReferenceFreq*1e-9 << endl << endl;
   }
  else
   if(ReferenceWavelength == 0)
    {
      ReferenceWavelength = Frequency2Wavelength(ReferenceFreq);
  
      *LogFile << endl 
              << "Since ReferenceWavelength is zero we use ReferenceFreq:" 
              << endl << endl 
              << "ReferenceWavelength [nm] = " << ReferenceWavelength*1e9 
              << endl
              << "ReferenceFreq [GHz] = " << ReferenceFreq*1e-9 << endl 
              << endl;
    }
   else
     {
       *LogFile << endl 
	       << "ERROR: "<< endl
               << "Make one of ReferenceFreq and ReferenceWavelength zero!" 
	       << endl << "The code will then use the other value." 
	       << endl <<  endl << "ABORTING NOW" << endl;

       exit(1);

     }


double DispersionFiberRefWavelength = 
   1e-12/(1e-9*1e3)*ReadDouble("DispersionFiber","ps/(nm-km)",0,0,
		                NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile); 

// DispersionFiberRefWavelength now in SI units s/m^2

double DispSlopeFiberRefWavelength =  
   1e-12/(1e-18*1e3)*ReadDouble("DispSlopeFiber","ps/(nm^2-km)",0,0,
		                 NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile); 

// DispSlopeFiberRefWavelength now in SI units s/m^3

SetFstSndOrDispDirectlyFlag = false;

SetFstSndOrDispDirectlyFlag = 
   bool(ReadInt("SetFstSndOrDispDirectlyFlag"," ",0,1,LOWER_AND_UPPER,
            OPTIONAL_NO_WARNING,0,&InFile));

if(SetFstSndOrDispDirectlyFlag)
  {
   FstOrDispFiber = ReadDouble("FstOrDisp","s^2/m",0,0,
                                NO_BOUNDS,MANDATORY,&InFile);
   SndOrDispFiber = ReadDouble("SndOrDisp","s^3/m",0,0,
                                NO_BOUNDS,MANDATORY,&InFile);
 }

NonLinIndexFiber = 0;
double NonlinearCoefficientGamma = 0;

NonLinIndexFiber  =  ReadDouble("NonLinIndexFiber","m^2/W",0,-1,
			LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile); 

EffectAreaFiber  =  ReadDouble("EffectAreaFiber","m^2",0,-1,
			LOWER_ONLY,&InFile); 

NonlinearCoefficientGamma = ReadDouble("NonlinearCoefficientGamma",
                                       "(Wm)^(-1)",0,-1,LOWER_ONLY,
                                       OPTIONAL_NO_WARNING,0,&InFile);

if(NonLinIndexFiber != 0 && NonlinearCoefficientGamma != 0)
   {
     cerr << "Only one of NonLinIndexFiber and NonlinearCoefficientGamma"
          << endl
          << "can be zero. ABORTING!" << endl;
     exit(1);
   }


TypeFastSaturableAbsorption = typeFastSaturableAbsorption( 
    ReadInt("TypeFastSaturableAbsorption"," ",0,-1,
                        LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile));

FastSaturableAbsorptionCoefficient = 
    ReadDouble("FastSaturableAbsorptionCoefficient","(Wm)^(-1)",0,-1,
                        LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);


DoFastSaturableAbsorptionFlag = 0;

if(FastSaturableAbsorptionCoefficient != 0)
   DoFastSaturableAbsorptionFlag = 1;

if(DoFastSaturableAbsorptionFlag && TypeStepSizes == LOCAL_ERROR_3RD)
      {
        cerr << "OptFiberLocalError class:"
             << endl
             << "FastSaturableAbsorptionCoefficient != 0"
             << endl
             << "and TypeStepSizes = LOCAL_ERROR_3RD"
             << endl
             << "not implemented. ABORTING NOW!"
             << endl;
        exit(1);
      }
    


// We allow AttenuationFiber to be negative to model gain in
// a fiber amplifier

double InAttenuationFiber = 0;
double InNepperAttenuation = 0;
double InTotalAttenuation = 1;

InAttenuationFiber = ReadDouble("AttenuationFiber","dB/m",1,-1,
			NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

InNepperAttenuation = ReadDouble("NepperAttenuationFiber","m^(-1)",1,-1,
                        NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

InTotalAttenuation = ReadDouble("TotalLinearAttenuationFiber"," ",0,-1,
                        LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);

if((InAttenuationFiber != 0 && InNepperAttenuation != 0) ||
   (InAttenuationFiber != 0 && InTotalAttenuation != 1) ||
   (InNepperAttenuation != 0 && InTotalAttenuation != 1) )
 {
  cerr << "Only one of AttenuationFiber, NepperAttenuationFiber"
       << "and TotalAttenuationFiber can be non-trivial" << endl
       << "ABORTING" << endl;
  exit(1);
 }

if(InAttenuationFiber != 0)
   SetAttenuationdB(InAttenuationFiber);
else
 if(InNepperAttenuation !=0)
   SetNepperAttenuation(InNepperAttenuation);
 else
  if(InTotalAttenuation != 1)
     SetNepperAttenuation(log(InTotalAttenuation)/LengthFiber);
  else
   SetNepperAttenuation(0);

DeltaZMaxInfile = ReadDouble("DeltaZMax","m",0,-1,
			LOWER_ONLY,&InFile);

RelativeErrorGoal = ReadDouble("RelativeErrorGoal"," ",0,-1,
			LOWER_ONLY,&InFile);

// JZFeb03: This won't work well when have one channel which
// is not at center frequency!! 

LocalErrorFreqLimit = oOptSignal->GetRelevantNumModesOFLE()/2;


OutputStepSizesFlag = ReadInt("OutputStepSizesFlag"," ",0,1,
			LOWER_AND_UPPER,&InFile);

StepSizesFileMode = ReadInt("StepSizesFileMode"," ",1,2,
			LOWER_AND_UPPER,&InFile);

// StepSizesFileMode = 1 if opening for output, = 2 if appending output

 if(TypeStepSizes == CONSTANT) // Const Step size second order
  {
    NumStepsBetweenScatterings = ReadInt("NumStepsBetweenScatterings"," ",0,-1,
			LOWER_ONLY,MANDATORY,&InFile);
  }

 // Put this line in in case use SetTypeStepSizes later

    NumStepsBetweenScatterings = ReadInt("NumStepsBetweenScatterings"," ",0,-1,
					 LOWER_ONLY,OPTIONAL_WARNING,&InFile);

    DeltaZ = LengthFiber/float(NumStepsBetweenScatterings);
    *LogFile << "DeltaZ [m] = " << DeltaZ << endl;


 if(TypeStepSizes == WALK_OFF)
   {
     WalkOffParameter = ReadDouble("WalkOffParameter","1/ps",0,-1,
                                    LOWER_ONLY,MANDATORY,1,&InFile);
   }

 // Put this line in in case use SetTypeStepSizes later

     WalkOffParameter = ReadDouble("WalkOffParameter","1/ps",0,-1,
                                    LOWER_ONLY,OPTIONAL_WARNING,1,&InFile);


if(TypeSolver == VECTOR_MANAKOV_PMD) // ## Parameters related to PMD 
{

  ScatteringStepSize = ReadDouble("ScatteringStepSize","m",0,-1,
                       LOWER_ONLY,MANDATORY,1e3,&InFile);

  // Make sure pointers to Euler angle arrays are null

  Theta = 0;
  Phi = 0;
  Psi = 0;

  SetFiberLengthsForPMD();
 
  PMD =  ReadDouble("PMD","ps/sqrt(km)",0,-1,
			LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile); 

  MeanDGDPerSqrtLength =  ReadDouble("MeanDGDPerSqrtLength","ps/sqrt(km)",0,-1,
			LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);  
                       
 if (PMD != 0 && MeanDGDPerSqrtLength != 0) 
  {
   cout << "OptFiberLocalError constructor:" << endl
        << "At least one of PMD_Fiber and MeanDGD_SqrtLength must be zero"
	<< endl;
   exit(1);
  }

  if(MeanDGDPerSqrtLength != 0)
    SetMeanDGDPerSqrtLength(MeanDGDPerSqrtLength);
  else
    SetPMD(PMD);

} // end if(TypeSolver == VECTOR_MANAKOV_PMD)

TypeFiberLossGain = typeFiberLossGain(ReadInt("TypeFiberLossGain"," ",0,1,
                                      LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,
                                      &InFile));

*LogFile << "TypeFiberLossGain = " << WriteTypeFiberLossGain() << endl;

if(TypeSolver != SCALAR_NLS  && TypeFiberLossGain == FIBER_AMPLIFIER)
{
 
 cerr << "WARNING: In OptFiberLocalError object" << endl;
 cerr << "TypeSolver != SCALAR_NLS  && TypeFiberLossGain == FIBER_AMPLIFIER" << endl;
 cerr << "The FIBER_AMPLIFIER has so far only been VERIFIED for TypeSolver == SCALAR_NLS" << endl;


}

if( (TypeStepSizes == WALK_OFF )  && 
     TypeFiberLossGain == FIBER_AMPLIFIER)
{

 cerr << "In OptFiberLocalError object" << endl;
 cerr << "The FIBER_AMPLIFIER is not implemented for" << endl
      << "TypeStepSizes == WALK_OFF. Just NT check DeltaZMax stuff" <<endl;
 cerr << "ABORTING NOW!" << endl;
 exit(1);

 // The only issue is to make sure "DeltaZMax" is set correctly.

}


TypeFiberAmplifierNoise = FIBER_AMP_NOISE_OFF;
TypeFiberAmplifierGain = FIBER_AMP_UNSATURATED_FLAT_GAIN;

FiberAmplifierSpontaneousEmissionFactor = 0.0;
FiberAmplifierNoiseMaxStepSizeFactor = 0.0;

if(TypeFiberLossGain == FIBER_AMPLIFIER)
{
  TypeFiberAmplifierGain = typeFiberAmplifierGain(
      ReadInt("TypeFiberAmplifierGain"," ",0,1,
               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile));

  if(TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
    {

     if(TypeStepSizes == LOCAL_ERROR_3RD)
      {
        cerr << "OptFiberLocalError class:" 
	     << endl
             << "TypeFiberAmplifierGain = FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN"
             << endl
             << "and TypeStepSizes = LOCAL_ERROR_3RD"
             << endl
             << "not implemented. ABORTING NOW!" 
             << endl;
        exit(1);
      }


     FiberAmplifierSuperGaussFiltGainCoefficient = 
       ReadDouble("FiberAmplifierSuperGaussFiltGainCoefficient","Hz=s^{-1}",
                 0.0,-1.0,LOWER_ONLY,MANDATORY,&InFile);

      FiberAmplifierSuperGaussFiltOrder =
       ReadInt("FiberAmplifierSuperGaussFiltOrder"," ",
                 0,-1,LOWER_ONLY,MANDATORY,1,&InFile);



     double InFiberAmplifierUnsaturatedGaindB = 
              ReadDouble("FiberAmplifierUnsaturatedGaindBPerLength","dB/m",
                 0.0,-1.0,LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

     double InFiberAmplifierUnsaturatedTotalLinearGain =
              ReadDouble("FiberAmplifierUnsaturatedTotalLinearGain"," ",
                 0.0,-1.0,LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile);

     double InFiberAmplifierUnsaturatedNepperGain =
              ReadDouble("FiberAmplifierUnsaturatedNepperGainPerLength",
                 "m^(-1)",0.0,-1.0,LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

     if((InFiberAmplifierUnsaturatedGaindB !=0 &&
         InFiberAmplifierUnsaturatedNepperGain != 0) ||
         (InFiberAmplifierUnsaturatedGaindB !=0 &&
          InFiberAmplifierUnsaturatedTotalLinearGain != 1) ||
         (InFiberAmplifierUnsaturatedNepperGain != 0 && 
          InFiberAmplifierUnsaturatedTotalLinearGain != 1))
        {
         cerr << "Only one of FiberAmplifierUnsaturatedGaindBPerLength,"
              << endl
              << "FiberAmplifierUnsaturatedNepperGainPerLength and" 
              << "FiberAmplifierUnsaturatedTotalLinearGain can be non-trivial."
              << endl << "ABORTING!"
              << endl;
         exit(1);
        }
     
     {
     if(InFiberAmplifierUnsaturatedGaindB != 0)
      FiberAmplifierUnsaturatedNepperGain = 
        0.1*log(10)*InFiberAmplifierUnsaturatedGaindB;
     else
      {
      if(InFiberAmplifierUnsaturatedNepperGain != 0)
         FiberAmplifierUnsaturatedNepperGain = 
          InFiberAmplifierUnsaturatedNepperGain;
      else
       {
        if(InFiberAmplifierUnsaturatedTotalLinearGain != 1)
         FiberAmplifierUnsaturatedNepperGain = 
           log(InFiberAmplifierUnsaturatedTotalLinearGain)/LengthFiber;  
        else
         FiberAmplifierUnsaturatedNepperGain = 0;
       }
       }
       }

     *LogFile << "FiberAmplifierUnsaturatedNepperGain = "
              <<  FiberAmplifierUnsaturatedNepperGain << endl;

     FiberAmplifierSaturationEnergy = 
               ReadDouble("FiberAmplifierSaturationEnergy","J",
                 0.0,-1.0,LOWER_ONLY,MANDATORY,&InFile);

      FiberAmplifierGainSaturationMaxGainPerStep =
        ReadDouble("FiberAmplifierGainSaturationMaxGainPerStep"," ",
                 1.0,-1.0,LOWER_ONLY,OPTIONAL_WARNING,1.01,&InFile);



    }

  TypeFiberAmplifierNoise = typeFiberAmplifierNoise(
                             ReadInt("TypeFiberAmplifierNoise"," ",0,2,
                                      LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,
                                      &InFile));

  if(TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF)
   {
     FiberAmplifierSpontaneousEmissionFactor = 
      ReadDouble("FiberAmplifierSpontaneousEmissionFactor"," ",
                 1.0,-1.0,LOWER_ONLY,MANDATORY,&InFile);

     FiberAmplifierNoiseMaxStepSizeFactor =
          ReadDouble("FiberAmplifierNoiseMaxStepSizeFactor"," ",
            0.0,0.2,LOWER_AND_UPPER,OPTIONAL_WARNING,0.1,&InFile);
   }

}// end if(TypeFiberLossGain == FIBER_AMPLIFIER)


*LogFile << "TypeFiberAmplifierGain = " 
	 << WriteTypeFiberAmplifierGain() << endl;

*LogFile << "TypeFiberAmplifierNoise = " 
         << WriteTypeFiberAmplifierNoise() << endl;

InFile.close();

fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
CenterFreq = oOptSignal->GetCenterFreq();
FWHM_Pulse = oOptSignal->GetFWHM_Pulse();
      
WriteAveragePowerdBmFlag = 0;
WritePulseWidthFlag = 0;

// Set derived parameters

double CenterWavelength = Frequency2Wavelength(CenterFreq);

SetDispersion(DispersionFiberRefWavelength,DispSlopeFiberRefWavelength);
  
qtPoints         = fft->give_N();        
DeltaFreq        = fft->delta_f;         

WithNonLin       = (NonLinIndexFiber !=0 || NonlinearCoefficientGamma !=0 ||
                    FastSaturableAbsorptionCoefficient != 0);   
         
ComputeDeltaZMax();

gamma = 0;     

cout << WriteTypeFiberLossGain() << " " << WriteTypeFiberAmplifierNoise() << endl;

 
if (WithNonLin || 
   (TypeFiberLossGain == FIBER_AMPLIFIER &&
    TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF) ||
   (TypeFiberLossGain == FIBER_AMPLIFIER &&
    TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN))
 {
   gamma = 0;

  if(NonLinIndexFiber != 0 && NonlinearCoefficientGamma != 0)
   {
     cerr << "Only one of NonLinIndexFiber and NonlinearCoefficientGamma"
          << endl
          << "can be zero. ABORTING!" << endl;
     exit(1);  
   }

  if(NonLinIndexFiber != 0)
   gamma = NonLinIndexFiber*2*pi*CenterFreq/(EffectAreaFiber*LightSpeed);  
  if(NonlinearCoefficientGamma != 0)
   SetGammaFiber(NonlinearCoefficientGamma);
   
  InitializedDeltaZInitialFlag = false;
  DeltaZInitial = 0.0; 
 }      
  
// ### Initialize memory #####

 if(!oOptSignal->GetUsingOptFiberLocalError())
   {
     cerr << "OptFiberLocalError constructor: " << endl
	  << "Must set UsingOptFiberLocalError = 1 in OptSignal.in" << endl
	  << "ABORTING NOW!!" << endl;
     exit(1);
   }



switch(TypeSolver)
  {

    // Note: We only need the freq data not the time data 
    // and so we do not malloc the time fields

    case SCALAR_NLS:

      sfftM_Start.fOptSignalX = oOptSignal->OFLEWork.sfftM_Start.fOptSignalX;
      sfftM_Coarse.fOptSignalX = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalX;
      LaserWorkX = oOptSignal->OFLEWork.LaserWorkX;     
 
      if(!sfftM_Start.fOptSignalX || !sfftM_Coarse.fOptSignalX ||
         !LaserWorkX)
       {
        cerr << "Malloc error in ocsOptFiberLocalError constructor" 
             << "Aborting now!!" << endl;
    
        exit(1);
       }

    FreqDomainOperatorX = oOptSignal->OFLEWork.FreqDomainOperatorX;
    SqFreqDomainOperatorX = oOptSignal->OFLEWork.SqFreqDomainOperatorX;

    if(!FreqDomainOperatorX || !SqFreqDomainOperatorX)
      {
        cerr << "Malloc error in ocsOptFiberLocalError constructor" 
             << "Aborting now!!" << endl;
    
        exit(1);
      }

     break;

  case VECTOR_MANAKOV_PMD: case VECTOR_MANAKOV_NO_PMD:

    sfftM_Start.fOptSignalX = oOptSignal->OFLEWork.sfftM_Start.fOptSignalX;
    sfftM_Start.fOptSignalY = oOptSignal->OFLEWork.sfftM_Start.fOptSignalY;
    sfftM_Coarse.fOptSignalX = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalX;
    sfftM_Coarse.fOptSignalY = oOptSignal->OFLEWork.sfftM_Coarse.fOptSignalY;

    LaserWorkX = oOptSignal->OFLEWork.LaserWorkX;
    LaserWorkY = oOptSignal->OFLEWork.LaserWorkY;


      if(!sfftM_Start.fOptSignalX || !sfftM_Start.fOptSignalY ||
         !sfftM_Coarse.fOptSignalX || !sfftM_Coarse.fOptSignalY||
         !LaserWorkX || !LaserWorkY)
       {
        cerr << "Malloc error in ocsOptFiberLocalError constructor" 
             << "Aborting now!!" << endl;
    
        exit(1);
       }


    FreqDomainOperatorX = oOptSignal->OFLEWork.FreqDomainOperatorX;
    FreqDomainOperatorY = oOptSignal->OFLEWork.FreqDomainOperatorY;
    SqFreqDomainOperatorX = oOptSignal->OFLEWork.SqFreqDomainOperatorX;
    SqFreqDomainOperatorY = oOptSignal->OFLEWork.SqFreqDomainOperatorY;

    if(!FreqDomainOperatorX ||!FreqDomainOperatorY ||
       !SqFreqDomainOperatorX ||!SqFreqDomainOperatorY)
      {
        cerr << "Malloc error in ocsOptFiberLocalError constructor" 
             << "Aborting now!!" << endl;
    
        exit(1);
      }

     break;

  default:

    cerr << "Error in TypeSolver switch in ocsOptFiberLocalError constructor"
         << endl;

    exit(1);


  } // ## end switch

switch(TypeRamanAmplification)
{

   case NO_RAMAN:
        break;

   case FULL_RAMAN:

        NepperAttenuationFiber = 0;
	   
        break;


  case SIMPLE_RAMAN:


    // Parameters for simple Raman amplifier with no pump depletion
   // Only suitable for a single pump and a single signal wavelength
  // This is for Backward Pumping only right now.
  // The theory in the forward pumping case is explained in
  // Agrawal p 319. To do the backward case change z to -z
   // in the pump equation, and solve the BVP.
	// If you want to add noise use the equivalent noise figure
	// of Hansen et al PTL Vol 10 Num 1 1998. Put this in by 
	// adding lumped loss and an EDFA after the Raman amplified 
	// fiber. The lumped loss has loss equal to the on-off gain of 
	// the RA and the EDFA has a gain equal to the on-off gain of the RA
	// and a noise figure equal to the equivalent noise figure as 
	// in Hansen. (They give measured values you could use).

	RamanOnOffGaindB =  ReadDouble("RamanOnOffGaindB"," ",0,-1,
			LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

	// The default value RamanOnOffGaindB = 0 doesn't do any Raman amplification.

	NepperRamanPumpLoss = (log(10.)/10)*
                        ReadDouble("RamanPumpLoss","dB/m",0,-1,
			LOWER_ONLY,OPTIONAL_NO_WARNING,0.2e-3,&InFile);

	NepperRamanGain = RamanOnOffGaindB*log(10.)/10;  
	

	//cout << "NepperRamanGain = " << NepperRamanGain << endl;
	//cout << "NepperRamanPumpLoss = " << NepperRamanPumpLoss << endl;

     break;
	 
	 default:
	   cerr << "Switch error in OptFiberLocalError constructor" << endl;
	   exit(1);
	 
}


LogFileSeparator();

*LogFile << "Processing Fiber Parameters" << endl << endl;

*LogFile <<  "LengthFiber [km] = " << LengthFiber*1e-3 << endl;
*LogFile <<  "NonLinIndexFiber [m^2/W] = " << NonLinIndexFiber << endl;
*LogFile <<  "EffectAreaFiber [m^2] = " << EffectAreaFiber << endl;
*LogFile <<  "AttenuationFiber [dB/km] = " <<  AttenuationFiber*1e3 << endl;
*LogFile <<  "Center Freq of Signal [GHz] = " << CenterFreq*1e-9 << endl;
*LogFile <<  "Center Wavelength of Signal [nm] = " << CenterWavelength*1e9 
          << endl;
  
*LogFile <<  "qtPoints = " << qtPoints << endl; 
*LogFile <<  "DeltaFreq [GHz] = " <<   DeltaFreq*1e-9 << endl; 
*LogFile << endl;
*LogFile <<  "DispersionFiber at Reference Wavelength [ps/(nm-km)] = " 
        << DispersionFiberRefWavelength*1e12/(1e9*1e-3)  << endl; 
*LogFile <<  "DispSlopeFiber at Reference Wavelength [ps/(nm^2-km)] = " 
          << DispSlopeFiberRefWavelength*1e12/(1e18*1e-3) << endl; 

*LogFile <<  "FstOrDispFiber at Center Wavelength [ps^2/km] = " 
          << FstOrDispFiber*1e27 << endl; 
*LogFile <<  "SndOrDispFiber  at Center Wavelength [ps^3/km] = " 
          << SndOrDispFiber*1e39 << endl; 
               
*LogFile << "Gamma [m^(-1)W^(-1)] = " << gamma << endl;

*LogFile << "FWHM_Pulse = " << FWHM_Pulse << endl;

if(TypeSolver == VECTOR_MANAKOV_PMD)
{
  *LogFile << "PMD [ps/sqrt(km)] = " << 1e12*sqrt(1e3)*PMD << endl;
  *LogFile << "MeanDGDPerSqrtLength [ps/sqrt(km)] = " 
           << 1e12*sqrt(1e3)*MeanDGDPerSqrtLength<< endl;

   *LogFile << "MeanDGD [ps] = " << 
                1e12*MeanDGDPerSqrtLength*sqrt(LengthFiber) << endl;
}   

ResetVolatileVariables();

PDVectorHai = new double[3];                //added by Hai 8/14/03

}// ## end OptFiberLocalError constructor

// ###############################################################

void OptFiberLocalError::SetRamanAmpPointer(RamanAmp * oRaman2)
{
  oRaman = oRaman2;
}

// ##################################################################

void OptFiberLocalError::SetTypeStepSizes(int TypeStepSizes2)
{

  TypeStepSizes = typeStepSizes(TypeStepSizes2);
  *LogFile << "SetTypeStepSizes = " << WriteTypeStepSizes() << endl;
  SetFiberLengthsForPMD();
}

// ##################################################################

void OptFiberLocalError::SetWalkOffParameter(double WalkOffParameter2)
{ 
  WalkOffParameter = WalkOffParameter2;
}

// ###################################################################

//! Sets the lengths of each section in the coarse-step method
//! for propagation with PMD.

/*!


  We divide the fiber into lengths of size ScatteringStepSize.
  The method  

  void OptFiberLocalError::PropagateWithoutPMD(double * PropagatedLength,
					   int * ZStepNum)

  will be applied to each section of length 
  OptFiberLocalError::ScatteringStepSize. 
  For each ScatteringStep before calling PropagateWithoutPMD the 
  Signal will undergo a random rotation on the Poincare sphere. 

  If TotalLengthFiber < ScatteringStepSize
      -# ScatteringStepSize = TotalLengthFiber
      -# NumberScatterings = 1.

  If TotalLengthFiber >= ScatteringStepSize
       -# If TotalLengthFiber = NumberScatterings*ScatteringStepSize
             -# LengthFiberExceptLast = LengthFiberLast =  ScatteringStepSize
       -# Else 
             -# NumberScatterings = floor(TotalLengthFiber/ScatteringStepSize)
             -# LengthFiberExceptLast = ScatteringStepSize;
             -# LengthFiberLast = TotalLengthFiber - 
                                  NumberScatterings*ScatteringStepSize
             -# NumberScatterings += 1
             -# But if LengthFiberLast < 0.5*ScatteringStepSize we 
                make one less step and make the last step longer.

*/

void OptFiberLocalError::SetFiberLengthsForPMD(void)
{


// Must have LengthFiber and ScatteringStepSize correct when enter
// this method.

// The method computes NumberScatterings, LengthFiberExceptLast,
// and LengthFiberLast which are required for 
// TypeSolver ==VECTOR_MANAKOV_PMD

  // Also allocates memory for Theta,Phi,Psi arrays

  // It works for both the Constant step size and variable step size methods.

 if(TypeSolver == VECTOR_MANAKOV_PMD)
   {
     
     TotalLengthFiber = LengthFiber;

     if(TotalLengthFiber <= ScatteringStepSize)
       {
         NumberScatterings = 1;
         LengthFiberLast = TotalLengthFiber;
         LengthFiberExceptLast = 0.0;
       }
     else
       {
         NumberScatterings = int(floor(TotalLengthFiber/ScatteringStepSize));
         LengthFiberLast = TotalLengthFiber - 
	                   NumberScatterings*ScatteringStepSize;

         if(LengthFiberLast < 1e-10) // LengthFiberLast == 0 
           {
	     LengthFiberExceptLast = ScatteringStepSize;
	     LengthFiberLast =  ScatteringStepSize;
	   }
         else
	   {
             
             LengthFiberExceptLast = ScatteringStepSize;
             LengthFiberLast = TotalLengthFiber - 
	                       NumberScatterings*ScatteringStepSize;
	     NumberScatterings += 1;

	       if(LengthFiberLast < 0.5*ScatteringStepSize)
		 {
		   NumberScatterings -= 1;
                   LengthFiberLast += ScatteringStepSize;
		 }

	   } // end else

       } // end else

   
     *LogFile << "TotalLengthFiber [km] = " << 1e3*TotalLengthFiber << endl;
     *LogFile << "ScatteringStepSize [km] = " << 1e3*ScatteringStepSize << endl;
     *LogFile << "NumberScatterings = " << NumberScatterings << endl;
     *LogFile << "LengthFiberExceptLast [km] = " 
             << 1e3*LengthFiberExceptLast << endl;
     *LogFile << "LengthFiberLast [km] = " << 1e3*LengthFiberLast << endl;
     

 if(TypeStepSizes == CONSTANT) // Const Step size second order
   {

     DeltaZ = LengthFiberExceptLast/(double(NumStepsBetweenScatterings));
     
   } // ## end if(TypeStepSizes == CONSTANT)

 AllocateMemoryForEulerAngles();
 GetNewFiberRealization();

} // end if

} // ## end SetFiberLengthsForPMD(void)

// ######################################################

void OptFiberLocalError::AllocateMemoryForEulerAngles(void)
{

 // Allocate memory for Euler angle arrays

  
 if(Theta)
   {
     delete [] Theta;
     Theta = 0;
   }

 if(Phi)
   {
     delete [] Phi;
     Phi = 0;
   }

 if(Psi)
   {
     delete [] Psi;
     Psi = 0;
   }

  

 Theta = new double[NumberScatterings];
 Phi = new double[NumberScatterings];
 Psi = new double[NumberScatterings];

 if(!Theta || !Phi || !Psi)
   {
     cerror("OptFiberLocalError::AllocateMemoryForEulerAngles Memory allocation error");
     
     exit(1);
   }


} // ## end Allocate....

// ###############################################################

void OptFiberLocalError::SetFiberLength(double Value) 
{
  LengthFiber=Value;
  SetFiberLengthsForPMD();
} 

// ###############################################################

void OptFiberLocalError::SetScatteringStepSize(double Value)
{
  ScatteringStepSize = Value;
  SetFiberLengthsForPMD();
}

// ################################################################

void OptFiberLocalError::PropagateWithRaman(double * PropagatedLength,
					    int * ZStepNum,
					    double  RamanStartingLength2)
{
  RamanStartingLength = RamanStartingLength2;
  PropagatedLengthStartFiber = *PropagatedLength;

  Propagate(PropagatedLength,ZStepNum);

}

// ####################################################################

void OptFiberLocalError::Propagate(double * PropagatedLength,int * ZStepNum)
{
  if(TypeRamanAmplification==SIMPLE_RAMAN)
    {
      RamanStartingLength = *PropagatedLength;
    }

  if(TypeRamanAmplification==FULL_RAMAN)
  {
    /****
      char  C_String[40];
      sprintf(C_String,"SignalTrack.dat");
      string FileName = Job + C_String;
      oRaman->WriteSignalTrack(FileName,qtPoints/2.0);
    ********/
  }


  // We lump the entire fiber loss for NoiseSpectralDensity here

  // Note: This won't work for Raman or for when 
  // TypeFiberLossGain == FIBER_AMPLIFIER !!

  oOptSignal->GainOfAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
			       exp(-LengthFiber*NepperAttenuationFiber));

 
switch(TypeSolver)
  {

    case SCALAR_NLS: case VECTOR_MANAKOV_NO_PMD:

      PropagateWithoutPMD(PropagatedLength,ZStepNum);
      break;
  
    case VECTOR_MANAKOV_PMD:

      PropagateWithPMD(PropagatedLength,ZStepNum);
      break;
 
    default:
     cerr << "Switch error in  OptFiberLocalError::Propagate" << endl;
     exit(1);

  }

} // ## end Propagate

// #################################################

void OptFiberLocalError::PropagateWithoutPMD(double * PropagatedLength,
                                             int * ZStepNum)
{

  // Must be called from Propagate() !!

  TotalLengthFiber = LengthFiber; // Used in Simple and Full Raman

switch(TypeStepSizes)
  {

     case LOCAL_ERROR_3RD: 
     case LOCAL_ERROR_2ND:

      //LogFileSeparator();
      //*LogFile << "PropagateLocalError" << endl;

     PropagateLocalError(PropagatedLength,ZStepNum);

     break;

    case WALK_OFF:
    case CONSTANT:

      //LogFileSeparator();
      //*LogFile << "PropagateConstStepSizeOrWalkOff" << endl;

     PropagateConstStepSizeOrWalkOff(PropagatedLength,ZStepNum);

     break;

   default:
     cerr << "Switch error in  OptFiberLocalError::Propagate" << endl;
     exit(1);

  }

} // ## end PropagateWithoutPMD

// #################################################

void OptFiberLocalError::PropagateWithPMD(double * PropagatedLength,
                                             int * ZStepNum)
{

  // Must be called from Propagate() !!
  
  double ScatteringStepSizeTmp =   ScatteringStepSize;
  double DeltaZInitialWithPMD;
  TotalLengthFiber = LengthFiber;

  LengthFiber = LengthFiberExceptLast;
  ScatteringStepSize = LengthFiberExceptLast;

  //cout << "NumberScatterings = " << NumberScatterings << endl;

 if(TypeStepSizes == CONSTANT) // Const Step size second order
   {

     DeltaZ = LengthFiber/(double(NumStepsBetweenScatterings));
     
   } // ## end if(TypeAlgorithm == 2)

  for(int ii=0;ii<NumberScatterings;ii++)
    {
      if(ii==NumberScatterings-1)
        {
           LengthFiber = LengthFiberLast;
           ScatteringStepSize = LengthFiberLast;

           if(TypeStepSizes == CONSTANT) // Const Step size second order
              DeltaZ = LengthFiber/(double(NumStepsBetweenScatterings));

	}

      //cout << "LengthFiber = " << LengthFiber << endl;

      ApplyPMDCompleteScatteringFreqDomainOperator(ii); // Scatter

      UpdateJonesMatrixAfterBiregringence(LengthFiber);

      PropagateWithoutPMD(PropagatedLength,ZStepNum);

      // For the next time we call this method we want to have 
      // DeltaZInitial equal to the value we had in the first step
      // of the first section. We save this value in DeltaZInitialWithPMD

      if(ii==0)
        DeltaZInitialWithPMD = DeltaZInitial;

      // For the next section of fiber we want the first DeltaZ to be
      // the same as the final DeltaZ value in the previous section.
      // So we set:

      DeltaZInitial = DeltaZ;
              
    }

  ScatteringStepSize = ScatteringStepSizeTmp;
  LengthFiber = TotalLengthFiber;
  DeltaZInitial = DeltaZInitialWithPMD;

}

// ###################################################

void OptFiberLocalError::PropagateLocalError(double * PropagatedLength, 
                                             int * ZStepNum)
{

int ZStepNumThisFiber = 0; 

ResetVolatileVariables();

static int NotFirstTimeCalled = 0;

if(NotFirstTimeCalled)
  OpenStepSizesFile(Job + "StepSizes.dat",2); //append mode
else 
 OpenStepSizesFile(Job + "StepSizes.dat",StepSizesFileMode);

NotFirstTimeCalled = 1;

// Uncomment if want data (eg power value) at start of map.
// The problem is that then the step size is wrong (it's 0)
//WriteStepSizes(*PropagatedLength);

WriteAveragePowerdBm(*PropagatedLength);


if(!WithNonLin && 
   TypeFiberAmplifierNoise == FIBER_AMP_NOISE_OFF &&
   TypeFiberAmplifierGain == FIBER_AMP_UNSATURATED_FLAT_GAIN)
{

  // Since there is no nonlinearity and dispersion is being 
  // solved exactly we can do propagation in a single step.

  ComputeDispersionOperator(LengthFiber);
  ApplyFreqDomainOperator(APPLY_ONCE,LengthFiber);
  
  CurrentZValueWithinFiber += LengthFiber;
  *PropagatedLength += LengthFiber;

  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber,LengthFiber);

//  ApplyRamanOperator(0,*PropagatedLength);


  *ZStepNum += 1;
  ZStepNumThisFiber += 1;

  UpdateAccumulatedFirstOrderDispersion(LengthFiber);
  UpdateAccumulatedSecondOrderDispersion(LengthFiber);

} // ## end if(!WithNonLin)

else // WithNonLin
{

  // #### Determine the Initial Step Size in Z ####
 
  // We use the initial step size as computed from the last time we propagated
  // through this particular fiber. If this is the first time
  // through the fiber then DeltaZInitial = 0 will hold, having been
  // set in the constructor, and we choose the initial step so that
  // DeltaZInitial = 0.5*DeltaZMax.

 
  if(!InitializedDeltaZInitialFlag)
    {
      //cout << "DeltaZInitial = " << DeltaZInitial << endl;

      // This choice is very ad-hoc:

      DeltaZInitial = 0.5*DeltaZMax;

      DeltaZ = DeltaZInitial;

      InitializedDeltaZInitialFlag = true;
      DispersionOperatorValid = false;
    }
  else 
    {
     DeltaZ = DeltaZInitial;
     DispersionOperatorValid = false;
    }

 while(CurrentZValueWithinFiber < LengthFiber)
  {

   if(2*DeltaZ > DeltaZMax) 
     // As our coarse step is of size 2*DeltaZ. This will work well
     // for the FIBER_AMPLIFIER case
     {
      DeltaZ = DeltaZMax/2.0;
      DispersionOperatorValid = false;
     }

   // Work out if next coarse step would exceed the fiber length

   if(2*DeltaZ + CurrentZValueWithinFiber > LengthFiber)
    {
      DeltaZ =  (LengthFiber - CurrentZValueWithinFiber)/2.0;
      DispersionOperatorValid = false;
    }
   else
     {
       // Trick to avoid having the very last step being very small

       if(4*DeltaZ + CurrentZValueWithinFiber > LengthFiber)
	 {
          // ## Make the next to last step slightly larger than half of the 
	  // ## remainder
       
          double MaxNextToLastCoarseStepSize = 
                  0.55*(LengthFiber - CurrentZValueWithinFiber);

          if(2*DeltaZ > MaxNextToLastCoarseStepSize)
	    {
              DeltaZ =  MaxNextToLastCoarseStepSize/2.0;
	      DispersionOperatorValid = false;
	    }
	 } //## end if

     } // ## end else

  
  // ######## Propagation starts here ########   

  // Calculate saturable gain per unit length once per step
  // for speed and better local error method error calculation

  FiberAmplifierComputeSaturableGain();

  FiberAmplifierAddMonteCarloNoise(2*DeltaZ);

  // Compute the linear operator for a half-size fine step only, thus
  // saving time later on

  ComputeDispersionOperator(DeltaZ/2.0);
  UpdateFreq(); // Make sure freq field is valid

  CopyFftMatricesFreqData(&sfftM,&sfftM_Start);

  // ##### A Coarse step (step size 2*DeltaZ) ### 

  ApplyFreqDomainOperator(APPLY_TWICE,DeltaZ); // Advance to z + DeltaZ
                                          
  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber + 
		       CurrentZValueWithinFiber,DeltaZ);


  ApplyTimeDomainOperator(2*DeltaZ);

  ApplyFreqDomainOperator(APPLY_TWICE,DeltaZ); // Advance to z + 2*DeltaZ
  
  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber + 
		       CurrentZValueWithinFiber+DeltaZ,DeltaZ);

  // ###### Data Management ######

  CopyFftMatricesFreqData(&sfftM,&sfftM_Coarse);

  // Reinstate starting Freq Field
  
  CopyFftMatricesFreqData(&sfftM_Start,&sfftM);

  FreqDataValid = true;
  TimeDataValid = false;

  // ####  A Fine step ( 2 x step size DeltaZ) 

  ApplyFreqDomainOperator(APPLY_ONCE,0.5*DeltaZ); // Advance to z + DeltaZ/2
  
  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber + 
		       CurrentZValueWithinFiber,DeltaZ/2.0);

  ApplyTimeDomainOperator(DeltaZ);

  ApplyFreqDomainOperator(APPLY_ONCE,0.5*DeltaZ); // Advance to z + DeltaZ
  
  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber + 
		       CurrentZValueWithinFiber + DeltaZ/2.0,DeltaZ/2.0);

  // Now we've gone one fine step of length DeltaZ

  ApplyFreqDomainOperator(APPLY_ONCE,0.5*DeltaZ); // Advance to z + 3*DeltaZ/2
  
  if(TypeRamanAmplification != NO_RAMAN)
   ApplyRamanOperator(PropagatedLengthStartFiber + 
		      CurrentZValueWithinFiber + DeltaZ,DeltaZ/2.0);

  ApplyTimeDomainOperator(DeltaZ);

  ApplyFreqDomainOperator(APPLY_ONCE,0.5*DeltaZ); // Advance to z + 2*DeltaZ
  
  if(TypeRamanAmplification != NO_RAMAN)
   ApplyRamanOperator(PropagatedLengthStartFiber + 
		      CurrentZValueWithinFiber + 1.5*DeltaZ,DeltaZ/2.0);

  // Now we've gone a second  fine step of length DeltaZ

  // ##### Compute relative error and third order solution ####

  // The third order solution is the one which is stored in sfftM

  ComputeThirdOrderSolutionAndRelativeError();
 
  if(RelativeError > 2.0*RelativeErrorGoal) // Failure 
    {
      DeltaZ *= 0.5; // Halve the step size
 
      DispersionOperatorValid = false;

      CopyFftMatricesFreqData(&sfftM_Start,&sfftM); // Throw away solution

      FreqDataValid = true;
      TimeDataValid = false;

    } // ## end if

  else // Success
    {

      CurrentZValueWithinFiber += 2*DeltaZ; // We went two fine steps
      *PropagatedLength += 2*DeltaZ;

      UpdateAccumulatedFirstOrderDispersion(2*DeltaZ);
      UpdateAccumulatedSecondOrderDispersion(2*DeltaZ);

      double StepSizeMultiplier = sqrt(2.);

      if(RelativeError > RelativeErrorGoal) // Reduce step size
         {
          DeltaZ /= StepSizeMultiplier;
	  DispersionOperatorValid = false;
	 }
      else if(RelativeError < 0.5*RelativeErrorGoal) // Increase step size
           {
	     DeltaZ *= StepSizeMultiplier;
	     DispersionOperatorValid = false;
	   }

      // After taking the first step in a fiber we set 
      // DeltaZInitial to be the new DeltaZValue
      // This value will be used next time we go through this fiber
      
      if(ZStepNumThisFiber==0)
	{
          DeltaZInitial = DeltaZ;
	}

      *ZStepNum += 2;
      ZStepNumThisFiber += 2;

      WriteStepSizes(*PropagatedLength);
      WriteAveragePowerdBm(*PropagatedLength);

    } // ## end else

  } // ## end while(CurrentZValueWithinFiber < LengthFiber)
  

} // ## end else WithNonLin


UpdateTime();
UpdateFreq();

if(OutputStepSizesFlag)
   StepSizesFile.close();

} // ## end  PropagateLocalError

// #############################################################

void OptFiberLocalError::PropagateConstStepSizeOrWalkOff
                         (double * PropagatedLength,int * ZStepNum)
{

ResetVolatileVariables();

int ZStepNumThisFiber = 0;

static int NotFirstTimeCalled = 0;

if(NotFirstTimeCalled)
 OpenStepSizesFile(Job + "StepSizes.dat",2); //append mode
else 
 OpenStepSizesFile(Job + "StepSizes.dat",StepSizesFileMode);

NotFirstTimeCalled = 1;

// Uncomment the next two lines if you want max power at start of map
// Problem is that step size is wrong (0)

//WriteStepSizes(*PropagatedLength);
WriteAveragePowerdBm(*PropagatedLength);

if(!WithNonLin && 
   TypeFiberAmplifierNoise == FIBER_AMP_NOISE_OFF &&
   TypeFiberAmplifierGain == FIBER_AMP_UNSATURATED_FLAT_GAIN)
{
  // Since there is no nonlinearity and dispersion is being 
  // solved exactly we can do propagation in a single step.

//  FiberAmplifierAddMonteCarloNoise(LengthFiber);

  ComputeDispersionOperator(LengthFiber);
  ApplyFreqDomainOperator(APPLY_ONCE,LengthFiber);
  
  if(TypeRamanAmplification != NO_RAMAN)
    ApplyRamanOperator(PropagatedLengthStartFiber + 
		       CurrentZValueWithinFiber,LengthFiber);

  CurrentZValueWithinFiber += LengthFiber;
  *PropagatedLength += LengthFiber;
  *ZStepNum += 1;
  ZStepNumThisFiber += 1;

   UpdateAccumulatedFirstOrderDispersion(LengthFiber);
   UpdateAccumulatedSecondOrderDispersion(LengthFiber);

   WriteStepSizes(*PropagatedLength);
   WriteAveragePowerdBm(*PropagatedLength);

} // ## end if(!WithNonLin)

else // WithNonLin
{

  int NumSteps;

  switch(TypeStepSizes)
    {
     case CONSTANT: 
       NumSteps = NumStepsBetweenScatterings;
       break;

     case WALK_OFF:
       ComputeWalkOffStepSize();
       NumSteps = WalkOffNumSteps;
       break;

    default:
      cerr << "OptFiberLocalError::PropagateConstStepSizeSecondOrder error"
	   << endl
	   << "ABORTING" << endl;

      exit(1);
    }
     

  for(int ll = 0; ll < NumSteps; ll++)
    {

      if(TypeStepSizes == WALK_OFF)
	{
          if(ll < NumSteps - 1)
	    DeltaZ = WalkOffStepSize;  // DispersionOperatorValid = true;
          else
	    {     
             DeltaZ = WalkOffLastStepSize;
	     DispersionOperatorValid = false;
	    }

	}

      FiberAmplifierComputeSaturableGain();

      FiberAmplifierAddMonteCarloNoise(DeltaZ);

      ComputeDispersionOperator(DeltaZ/2.0);

      ApplyFreqDomainOperator(APPLY_ONCE,DeltaZ/2.0); // Advance to z + DeltaZ/2
	  
	  if(TypeRamanAmplification != NO_RAMAN)
       ApplyRamanOperator(PropagatedLengthStartFiber + 
		          CurrentZValueWithinFiber,DeltaZ/2.0); 

      ApplyTimeDomainOperator(DeltaZ);

      ApplyFreqDomainOperator(APPLY_ONCE,DeltaZ/2.0); // Advance to z + DeltaZ
	  
	  if(TypeRamanAmplification != NO_RAMAN)
        ApplyRamanOperator(PropagatedLengthStartFiber + 
		           CurrentZValueWithinFiber + DeltaZ/2.0,DeltaZ/2.0); 
                                 // Advance to z + DeltaZ/

      CurrentZValueWithinFiber += DeltaZ;
      *PropagatedLength += DeltaZ;
      *ZStepNum += 1;
      ZStepNumThisFiber += 1;

       UpdateAccumulatedFirstOrderDispersion(DeltaZ);
       UpdateAccumulatedSecondOrderDispersion(DeltaZ);

      WriteStepSizes(*PropagatedLength);
      WriteAveragePowerdBm(*PropagatedLength);

    } // end for ll loop


} // ## end else WithNonLin

UpdateTime();
UpdateFreq();

if(OutputStepSizesFlag)
   StepSizesFile.close();


} // ## end PropagateConstStepSizeOrWalkOff

// #########################################################

void OptFiberLocalError::OpenStepSizesFile(string StepSizesFileName, int Mode)
{

  if(OutputStepSizesFlag)
    {

      switch(Mode)
	{

	  case 1:

           StepSizesFile.open(StepSizesFileName.c_str(),ios::out);    

           if (!StepSizesFile) 
            {
             cerr << "Error: StepSizesFile  "<< StepSizesFileName 
                   << " could not be opened."<< endl;
             exit(1);
            }

           StepSizesFile << "% (1) PropagatedLength [km], "
			 << "(2) DeltaZ [m] (Fine Step Size for LOCAL_ERROR)," 
			 << " (3) MaxPower [mW], "
			 << " (4) RelativeError "
			 << " (5) DispersionFiberCenterWavelength [ps/(nm-km)]"
			 << " (6) Power in Central Frequency [W]"
			 << endl;
			 
			
           if(WritePulseWidthFlag)
	     StepSizesFile << endl 
               << "% Also output in last column Average Pulse Width [ps]" 
		<< " and Average power [mW] and AccdDisp [ps^2]" 
                << endl << endl;

           if(TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
            StepSizesFile << "%(7) SaturablePowerGainPerLength [1/m]" << endl;

           break;

	  case 2: 

           StepSizesFile.open(StepSizesFileName.c_str(),ios::app);    

           if (!StepSizesFile) 
            {
             cerr << "Error: StepSizesFile  "<< StepSizesFileName 
                   << " could not be opened."<< endl;
             exit(1);
            }

           break;

	default: 
	  cerr << "Switch error in  OptFiberLocalError::OpenStepSizesFile" 
               << endl;
          exit(1);


	} // ## end switch

    } // ## end if

} // ## end OpenStepSizesFile

// ################################################

void OptFiberLocalError::ResetVolatileVariables(void)
{

  //*LogFile << "Resetting Volatile Variables" << endl;

  TimeDataValid = true;  
  FreqDataValid = true;

  DispersionOperatorValid = false;
  CurrentZValueWithinFiber = 0.0; // Correct for PMD?
  RelativeError = 0.0;

} // ## end ResetVolatileVariables

// ####################################################

OptFiberLocalError::~OptFiberLocalError(void)
{
 
} // ## end Destructor

// #####################################################

string OptFiberLocalError::WriteTypeSolver(void)
{

string name;

switch(TypeSolver)
  {
  case SCALAR_NLS: name = "SCALAR_NLS"; break;
  case VECTOR_MANAKOV_PMD: name = "VECTOR_MANAKOV_PMD"; break;
  case VECTOR_MANAKOV_NO_PMD: name = "VECTOR_MANAKOV_NO_PMD"; break;
  default: name = "Invalid Solver Type";
  }

return name;

} // ## end WriteTypeSolver

// ################################################

string OptFiberLocalError::WriteTypeStepSizes(void)
{

string name;

switch(TypeStepSizes)
  {
   case LOCAL_ERROR_3RD: name = "LOCAL_ERROR_3RD"; break;
   case LOCAL_ERROR_2ND: name = "LOCAL_ERROR_2ND"; break;
   case WALK_OFF: name = "WALK_OFF"; break;
   case CONSTANT: name = "CONSTANT"; break;
   default: name = "Invalid StepSize Type";
  }

return name;

} // ## end WriteTypeStepSizes

// ########################################################

string OptFiberLocalError::WriteTypeFiberLossGain(void)
{

string name;

switch(TypeFiberLossGain)
 {
   case LOSSY_FIBER: name = "LOSSY_FIBER"; break;
   case FIBER_AMPLIFIER: name = "FIBER_AMPLIFIER"; break;
   default: name = "Invalid FiberLossGain Type";
 }

return name;

}

// #######################################################

string OptFiberLocalError::WriteTypeFiberAmplifierNoise(void)
{

string name;

switch(TypeFiberAmplifierNoise)
 {
   case FIBER_AMP_NOISE_OFF: name = "FIBER_AMP_NOISE_OFF"; break;
   case FIBER_AMP_NOISE_GAUSSIAN_WHITE: 
         name = "FIBER_AMP_NOISE_GAUSSIAN_WHITE"; break;
   default: name = "Invalid FiberAmplifierNoise Type";
 }

return name;

}

// ######################################################

string OptFiberLocalError::WriteTypeFiberAmplifierGain(void)
{

string name;

switch(TypeFiberAmplifierGain)
 {
   case FIBER_AMP_UNSATURATED_FLAT_GAIN: 
        name = "FIBER_AMP_UNSATURATED_FLAT_GAIN"; break;
   case FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN:
         name = "FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN"; break;
   default: name = "Invalid FiberAmplifierGain Type";
 }

return name;

}

// ######################################################

void inline  OptFiberLocalError::UpdateTime(void)
{

 if(!TimeDataValid)
   {
     if(FreqDataValid)
       {
         fft->IFFT(sfftM.fOptSignalX, sfftM.tOptSignalX); 
         
         if(TypeSolver==VECTOR_MANAKOV_PMD ||TypeSolver==VECTOR_MANAKOV_NO_PMD)
          fft->IFFT(sfftM.fOptSignalY, sfftM.tOptSignalY);     
	         
         TimeDataValid = true;

       }
     else
       {
       cerr << "OptFiberLocalError::UpdateTime(): Both Time and Freq Invalid!" 
            << "Aborting Now!!" << endl;

       exit(1);
       }

   } // end if


}// ## end UpdateTime

// ########################################################

void inline  OptFiberLocalError::UpdateFreq(void)
{

 if(!FreqDataValid)
   {
     if(TimeDataValid)
       {
         fft->FFT(sfftM.tOptSignalX, sfftM.fOptSignalX); 
         
         if(TypeSolver==VECTOR_MANAKOV_PMD ||TypeSolver==VECTOR_MANAKOV_NO_PMD)
          fft->FFT(sfftM.tOptSignalY, sfftM.fOptSignalY);     
         
         FreqDataValid = true;

       }
     else
       {
       cerr << "OptFiberLocalError::UpdateFreq(): Both Time and Freq Invalid!" 
            << "Aborting Now!!" << endl;

       exit(1);
       }

   } // end if


}// ## end UpdateFreq

// ########################################################

/*!

    For the WALK_OFF method we choose the step size to be constant
    within each fiber as follows.

    OptFiberLocalError::DeltaZ = 
    OptFiberLocalError::WalkOffParameter / abs( D_1\lambda_1 - D_2 \lambda_2 )

    where D_1 and D_2 are the dispersions in ps/nm-km at wavelengths
    \lambda_1 and \lambda_2.

    We choose \lambda_1,2 as follows:

    WDM Case: \lambda_1,2 are the central wavelengths of the edge channels

    Single Channel Case: \lambda_1,2 correspond to frequencies

                          \freq_1,2 =  CenterFreq +- 
                                       RootMeanSquare spectral width of signal.

    Of course for the last step we may need to choose a smaller step size
    so that the overall propagation distance is correct.

    Note: For simulations with PMD, we recompute WalkOff step size for each
    section. For single channel simulations since 
    RootMeanSquare spectral width of signal may change slightly it's possible
    that the different sections of the same fiber have different step sizes. 

 */

void OptFiberLocalError::ComputeWalkOffStepSize(void)
{

  double MinFreq,MaxFreq;

  if(oOptSignal->GetNumChannels() == 1)
    {

      double RMSWidth = oOptSignal->ComputeRMSSpectralWidth();

      //cout << "RMSWidth = " << RMSWidth << endl;

      MinFreq = oOptSignal->GetCenterFreq() - 2*RMSWidth;
      MaxFreq = oOptSignal->GetCenterFreq() + 2*RMSWidth;
    }
  else
    {
      for(int ch=0;ch<oOptSignal->GetNumChannels();ch++)
	{

         if(ch==0)
           {
             MinFreq = oOptSignal->GetFrequency(ch);
             MaxFreq = oOptSignal->GetFrequency(ch);
	   }
         else
	    {
              if(MinFreq > oOptSignal->GetFrequency(ch))
                 MinFreq = oOptSignal->GetFrequency(ch);

              if(MaxFreq < oOptSignal->GetFrequency(ch))
                 MaxFreq = oOptSignal->GetFrequency(ch);
	    }
	} // end for
    } // end else

  //cout << "MinFreq = " << MinFreq << " MaxFreq = " << MaxFreq << endl;

 double MinWavelength = Frequency2Wavelength(MaxFreq);
 double MaxWavelength = Frequency2Wavelength(MinFreq);

 /*
 cout << "MinWavelength = " << MinWavelength 
      << " MaxWavelength = " << MaxWavelength << endl;
 */

 double DispersionMinWavelength = GetDispersion(MinWavelength);
 double DispersionMaxWavelength = GetDispersion(MaxWavelength);

 // With this normalization denominator is in units of ps/km
 // So WalkOffParameter is in units of 1000/ps

 WalkOffStepSize = 1e3*1e-12*WalkOffParameter/
                   abs(DispersionMinWavelength*MinWavelength -
		   DispersionMaxWavelength*MaxWavelength);


 // Want the following for TypeFiberLossGain == FIBER_AMPLIFIER

 if(TypeFiberLossGain == FIBER_AMPLIFIER && 
    TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF &&
    WalkOffStepSize > DeltaZMax)
     WalkOffStepSize = DeltaZMax;
                   
 //cout << "WalkOffStepSize = " << WalkOffStepSize << endl;

 if(WalkOffStepSize >= LengthFiber)
   {
     WalkOffNumSteps = 1;
     WalkOffStepSize = 0;
     WalkOffLastStepSize = LengthFiber;
   }
 else
   {
     if(WalkOffStepSize >= LengthFiber/2.0)
       {
        WalkOffNumSteps = 2;
        WalkOffStepSize = LengthFiber/2.0;
        WalkOffLastStepSize = LengthFiber/2.0;
       }
      else
	{
          WalkOffNumSteps = int(floor( LengthFiber / WalkOffStepSize ));
 
          WalkOffLastStepSize = LengthFiber - WalkOffNumSteps*WalkOffStepSize;

          if(WalkOffLastStepSize > 0)
             WalkOffNumSteps += 1;
          else
	    { 

             if(WalkOffLastStepSize < 0)
              {
                cerr << "OptFiberLocalError::ComputeWalkOffStepSize error"
	               << endl
	               << "ABORTING" << endl;

                exit(1);
              }
             else // WalkOffLastStepSize = 0
	       {
		 WalkOffLastStepSize = WalkOffStepSize;
	       }

	    } // end else
	} // end else
   } // end else

DispersionOperatorValid = false;

 /****
 cout << "WalkOffNumSteps = " << WalkOffNumSteps << endl;
 cout << "LengthFiber = " << LengthFiber << endl;
 cout << "WalkOffStepSize = " << WalkOffStepSize << endl;
 cout << "WalkOffLastStepSize = " << WalkOffLastStepSize << endl;
 ****/

} // ## end void OptFiberLocalError::ComputeWalkOffStepSize(void)

// ########################################################

void OptFiberLocalError::ComputeDispersionOperator(double InputStepSizeZ)
{

  /* This method is now correct for the Physics convention.
     See FFTandNLSConventions.pdf and the tests therein.

     The fact that we have the combination
     FstOrDispDivTwo + omega*SndOrDispDivSix
     comes from the fact that this is the start of the Taylor's series
     for beta(CenterFreq + freq) - beta(CenterFreq).
     See Curtis Menyuk's Multiple Length Scales paper in J. Engr Math
     equation (25). 
  */

if(!DispersionOperatorValid)
 {

  double StepSizeZ = InputStepSizeZ;

  if(TypeFiberLossGain == FIBER_AMPLIFIER &&
     TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
   {
     StepSizeZ /= 2.0; 
     // We do half step of dispersion operator before and after
     // parabolic saturable gain operator
   }


  switch(TypeSolver)
    {

    case SCALAR_NLS: case VECTOR_MANAKOV_NO_PMD:
      {
	// FAST METHOD: (Debugged JZ Oct 20 2002)

      cplx * FreqOpX_ptr = FreqDomainOperatorX;
      cplx * SqFreqOpX_ptr = SqFreqDomainOperatorX;

      register const double FstOrDispDivTwo = StepSizeZ*FstOrDispFiber/2.0;
      register const double SndOrDispDivSix = StepSizeZ*SndOrDispFiber/6.0;
 
      register const double Attenuation = 
	                         exp(-StepSizeZ*NepperAttenuationFiber/2.0);

      if(TypeStepSizes == LOCAL_ERROR_3RD || TypeStepSizes == LOCAL_ERROR_2ND)
	{
           for(int jj=0;jj<qtPoints;jj++, FreqOpX_ptr++, SqFreqOpX_ptr++)
            {
             double omega = 2.0*pi*fft->freq(jj); // omega is freq in radians
	
	     *FreqOpX_ptr = Attenuation*
	       exp_jc_real(sq(omega)*(FstOrDispDivTwo + omega*SndOrDispDivSix));

             *SqFreqOpX_ptr = (*FreqOpX_ptr)*(*FreqOpX_ptr);
      
            } // end for-jj loop
    
	} // end  if(TypeStepSizes == LOCAL_ERROR_3RD...)
      else
	{
            for(int jj=0;jj<qtPoints;jj++, FreqOpX_ptr++)
             {
              double omega = 2.0*pi*fft->freq(jj); // omega is freq in radians
	
	      *FreqOpX_ptr = Attenuation*
	      exp_jc_real(sq(omega)*(FstOrDispDivTwo + omega*SndOrDispDivSix));
      
             } // end for-jj loop
          
	} // end else
       
      }

     break;
 
      case VECTOR_MANAKOV_PMD:
	{

	  // STILL TO BE DEBUGGED FINALLY

      register const double BirefringenceTerm  =  
	                       StepSizeZ*GetFreqDerivativeOfBirefringence();

      register const double FstOrDispDivTwo = StepSizeZ*FstOrDispFiber/2.0;
      register const double SndOrDispDivSix = StepSizeZ*SndOrDispFiber/6.0; 
      register const double Attenuation = 
	                         exp(-StepSizeZ*NepperAttenuationFiber/2.0);

      cplx * FreqOpX_ptr = FreqDomainOperatorX;
      cplx * FreqOpY_ptr = FreqDomainOperatorY;
      cplx * SqFreqOpX_ptr = SqFreqDomainOperatorX;
      cplx * SqFreqOpY_ptr = SqFreqDomainOperatorY;

      if(TypeStepSizes == LOCAL_ERROR_3RD || TypeStepSizes == LOCAL_ERROR_2ND)
	{
         for(int jj=0;jj<qtPoints; 
	  jj++, FreqOpX_ptr++, FreqOpY_ptr++, SqFreqOpX_ptr++, SqFreqOpY_ptr++)
          {
            register double omega = 2.0*pi*fft->freq(jj); 
	                                 // omega is freq in radians

            register double CDandLossTerm = sq(omega)*
	                      (FstOrDispDivTwo + omega*SndOrDispDivSix);

	 // We include the CenterFreq here since if we ever use WavePlates
	 // we'll need to use absolute frequency since in that case
	 // we'd just be doing a rotation about the z-axis and if we used 
	 // relative freq then at rel freq = 0 you'd never get off the
	 // equator. However for coarse step which is what we're doing here
	 // it doesn't matter statistically. Of course for a given fiber
	 // realization the answers will be different when you use
	 // rel and absolute freqs.

	 /************

         // SLOW METHOD:
	    
            // JZFeb03: Changed signs of terms involving Birefringence
	    // as required for change from Math to Physics convention

            // That way fast axis is the same one as it was with Math 
            // convention

         FreqDomainOperatorX[jj] = Attenuation*exp_jc_real(
			 -(omega+2.0*pi*CenterFreq)*BirefringenceTerm+
	                 CDandLossTerm);

         FreqDomainOperatorY[jj] = Attenuation*exp_jc_real(
			 (omega+2.0*pi*CenterFreq)*BirefringenceTerm+
	                 CDandLossTerm);

            
	 **************/

	 // FAST METHOD:

	    // JZFeb03: Changed signs of terms involving Birefringence
	    // as required for change from Math to Physics convention

         *FreqOpX_ptr = Attenuation*exp_jc_real(
			 -(omega+2.0*pi*CenterFreq)*BirefringenceTerm+
	                 CDandLossTerm);

         *FreqOpY_ptr = Attenuation*exp_jc_real(
			 (omega+2.0*pi*CenterFreq)*BirefringenceTerm+
	                 CDandLossTerm);
            
         *SqFreqOpX_ptr = (*FreqOpX_ptr)*(*FreqOpX_ptr);
         *SqFreqOpY_ptr = (*FreqOpY_ptr)*(*FreqOpY_ptr);

        } // end for-jj loop

      } // end if
    else
      {
       for(int jj=0;jj<qtPoints; jj++, FreqOpX_ptr++, FreqOpY_ptr++)
          {
            register double omega = 2.0*pi*fft->freq(jj); 
	                                 // omega is freq in radians

            register double CDandLossTerm = sq(omega)*
	                      (FstOrDispDivTwo + omega*SndOrDispDivSix);
	
	 // FAST METHOD:

             // JZFeb03: Changed signs of terms involving Birefringence
	    // as required for change from Math to Physics convention


         *FreqOpX_ptr = Attenuation*exp_jc_real(
			 -(omega+2.0*pi*CenterFreq)*BirefringenceTerm+
			 CDandLossTerm);

         *FreqOpY_ptr = Attenuation*exp_jc_real(
			 (omega+2.0*pi*CenterFreq)*BirefringenceTerm+
                         CDandLossTerm);
            
         } // end for-jj loop

      } // end else	

	}	
      break;

    default:

    cerr << "TypeSolver switch error in "
         << "OptFiberLocalError:ComputeDispersionOperator " << endl;

    exit(1);

  } // ## end switch

  DispersionOperatorValid = true;

 } // end if(!DispersionOperatorValid)

}// ## end ComputeDispersionOperator

// #####################################################

void OptFiberLocalError::ApplyDispersionOperator(bool ApplyTwiceFlag)
{

 switch(TypeSolver)
  {

   case SCALAR_NLS:
     {

     /*********

     // SLOW (OLD) METHOD:

    for(int jj=0;jj<qtPoints;jj++)
     {
       sfftM.fOptSignalX[jj] *= FreqDomainOperatorX[jj];
     }

     ***********/

     // FAST METHOD: (Debugged JZ Oct 20 2002)

     register cplx * f_X_ptr = sfftM.fOptSignalX;
     cplx * f_X_end_ptr = f_X_ptr + qtPoints;

     register cplx * FreqOp_X_ptr;

     if(ApplyTwiceFlag)
        FreqOp_X_ptr = SqFreqDomainOperatorX;
     else
        FreqOp_X_ptr = FreqDomainOperatorX;

     for( ; f_X_ptr < f_X_end_ptr; f_X_ptr++, FreqOp_X_ptr++)
       *f_X_ptr *= (*FreqOp_X_ptr);

     }

    break;

    case VECTOR_MANAKOV_NO_PMD:   //(Debugged JZ Oct 20 2002)
     {

     /************* 

     SLOW (OLD) METHOD:

    for(int jj=0;jj<qtPoints;jj++)
     {
       sfftM.fOptSignalX[jj] *= FreqDomainOperatorX[jj];
       sfftM.fOptSignalY[jj] *= FreqDomainOperatorX[jj];
     }

     **************/

     register cplx * f_X_ptr = sfftM.fOptSignalX;
     register cplx * f_Y_ptr = sfftM.fOptSignalY;
     cplx * f_X_end_ptr = f_X_ptr + qtPoints;

     register cplx * FreqOp_X_ptr;
     
     if(ApplyTwiceFlag)
       {
        FreqOp_X_ptr = SqFreqDomainOperatorX;
       }
     else
       {
        FreqOp_X_ptr = FreqDomainOperatorX;
       }

     for( ; f_X_ptr < f_X_end_ptr; f_X_ptr++, f_Y_ptr++, FreqOp_X_ptr++)
       {
         *f_X_ptr *= (*FreqOp_X_ptr);
         *f_Y_ptr *= (*FreqOp_X_ptr);
       }

     }

    break;

    case VECTOR_MANAKOV_PMD:
      {
      // NEED TO DEBUG FAST AND SLOW VERSIONS OF THIS CASE

      /************

      SLOW (OLD) METHOD: 

      for(int jj=0;jj<qtPoints;jj++)
      {
        sfftM.fOptSignalX[jj] *= FreqDomainOperatorX[jj];
        sfftM.fOptSignalY[jj] *= FreqDomainOperatorY[jj];
      }

      ***************/

     register cplx * f_X_ptr = sfftM.fOptSignalX;
     register cplx * f_Y_ptr = sfftM.fOptSignalY;

     cplx * f_X_end_ptr = f_X_ptr + qtPoints;

     register cplx * FreqOp_X_ptr;
     register cplx * FreqOp_Y_ptr;

     if(ApplyTwiceFlag)
       {
        FreqOp_X_ptr = SqFreqDomainOperatorX;
        FreqOp_Y_ptr = SqFreqDomainOperatorY;
       }
     else
       {
        FreqOp_X_ptr = FreqDomainOperatorX;
        FreqOp_Y_ptr = FreqDomainOperatorY;
       }

     for( ; f_X_ptr < f_X_end_ptr; 
	  f_X_ptr++, f_Y_ptr++, FreqOp_X_ptr++, FreqOp_Y_ptr++)
       {
         *f_X_ptr *= (*FreqOp_X_ptr);
         *f_Y_ptr *= (*FreqOp_Y_ptr);

       }

      }

      break;

   default:

    cerr << "TypeSolver switch error in "
         << "OptFiberLocalError::ApplyDispersionOperator " << endl;

    exit(1);

  } // ## end switch
   
 //TimeDataValid = false;

} // ## end ApplyDispersionOperator

// #######################################################

void OptFiberLocalError::ApplySaturableGainOperator(double StepSizeZ)
{

switch(TypeSolver)
{

 case SCALAR_NLS:
  {
   bool FirstStepFlag = true;
   double Gain1 = 
           FiberAmplifierComputeSaturableGainForHeunsMethod(FirstStepFlag);

   register cplx * InX_ptr = sfftM.fOptSignalX;
   cplx * InX_end_ptr = InX_ptr + qtPoints;

   register cplx * OutX_ptr = LaserWorkX;

   for(int jj=0 ; InX_ptr < InX_end_ptr; jj++,InX_ptr++,OutX_ptr++ )
    {
       double omega = 2.0*pi*fft->freq(jj); // in radians

       *OutX_ptr = (1 + 0.5*Gain1*StepSizeZ*
                   (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                          2*FiberAmplifierSuperGaussFiltOrder)))
                   *(*InX_ptr);
    }

   FirstStepFlag = false;
   double Gain2 =
           FiberAmplifierComputeSaturableGainForHeunsMethod(FirstStepFlag);

   InX_ptr = LaserWorkX;
   InX_end_ptr = InX_ptr + qtPoints;

   OutX_ptr = sfftM.fOptSignalX;

   for(int jj=0 ; InX_ptr < InX_end_ptr; jj++,InX_ptr++,OutX_ptr++ )
    {
       double omega = 2.0*pi*fft->freq(jj); // in radians

       *OutX_ptr = (1 + 0.25*Gain1*StepSizeZ*
                   (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                          2*FiberAmplifierSuperGaussFiltOrder)))
                    *(*OutX_ptr) + 
                   0.25*Gain2*StepSizeZ*
                   (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                          2*FiberAmplifierSuperGaussFiltOrder))
                    *(*InX_ptr);
    }


  } //end case SCALAR_NLS
  
 break;

 case VECTOR_MANAKOV_NO_PMD:  case VECTOR_MANAKOV_PMD:
  {
   bool FirstStepFlag = true;
   double Gain1 =
           FiberAmplifierComputeSaturableGainForHeunsMethod(FirstStepFlag);

   register cplx * InX_ptr = sfftM.fOptSignalX;
   register cplx * InY_ptr = sfftM.fOptSignalY;

   cplx * InX_end_ptr = InX_ptr + qtPoints;

   register cplx * OutX_ptr = LaserWorkX;
   register cplx * OutY_ptr = LaserWorkY;

   for(int jj=0 ; InX_ptr < InX_end_ptr; 
       jj++,InX_ptr++,InY_ptr++,OutX_ptr++,OutY_ptr++ )
    {
       double omega = 2.0*pi*fft->freq(jj); // in radians

       double Factor = 1 + 0.5*Gain1*StepSizeZ*
                       (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                              2*FiberAmplifierSuperGaussFiltOrder));

       *OutX_ptr = Factor*(*InX_ptr);
       *OutY_ptr = Factor*(*InY_ptr);
    }

   FirstStepFlag = false;
   double Gain2 =
           FiberAmplifierComputeSaturableGainForHeunsMethod(FirstStepFlag);

   InX_ptr = LaserWorkX;
   InY_ptr = LaserWorkY;

   InX_end_ptr = InX_ptr + qtPoints;

   OutX_ptr = sfftM.fOptSignalX;
   OutY_ptr = sfftM.fOptSignalY;

   for(int jj=0 ; InX_ptr < InX_end_ptr; 
       jj++,InX_ptr++,InY_ptr++,OutX_ptr++,OutY_ptr++ )
    {
       double omega = 2.0*pi*fft->freq(jj); // in radians

       double Factor1 = 1 + 0.25*Gain1*StepSizeZ*
               (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                             2*FiberAmplifierSuperGaussFiltOrder));

       double Factor2 = 0.25*Gain2*StepSizeZ*
               (1-pow(omega/FiberAmplifierSuperGaussFiltGainCoefficient,
                      2*FiberAmplifierSuperGaussFiltOrder));

       *OutX_ptr = Factor1*(*OutX_ptr) + Factor2*(*InX_ptr);
       *OutY_ptr = Factor1*(*OutY_ptr) + Factor2*(*InY_ptr);

    }

  } //end cases VECTOR_MANAKOV_NO_PMD, VECTOR_MANAKOV_PMD
 
 break;

 default:

 cerr << "TypeSolver switch error in "
      << "OptFiberLocalError::ApplySaturableGainOperator " << endl;

 exit(1);

} // end switch

} // end OptFiberLocalError::ApplySaturableGainOperator

// ######################################################

void OptFiberLocalError::ApplyFreqDomainOperator(bool ApplyDispersionTwiceFlag,
                                                 double GainSatStepSizeZ)
{

UpdateFreq();

if(TypeFiberLossGain == LOSSY_FIBER ||
   (TypeFiberLossGain == FIBER_AMPLIFIER &&
    TypeFiberAmplifierGain == FIBER_AMP_UNSATURATED_FLAT_GAIN))
 {
  ApplyDispersionOperator(ApplyDispersionTwiceFlag);
 }
else
{
 if(TypeFiberLossGain == FIBER_AMPLIFIER && 
    TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
  {
    ApplyDispersionOperator(ApplyDispersionTwiceFlag);
    ApplySaturableGainOperator(GainSatStepSizeZ);
    ApplyDispersionOperator(ApplyDispersionTwiceFlag);
  }
} // end else

TimeDataValid = false;

}

// ######################################################

void OptFiberLocalError::ApplyRamanOperator(double TotalPropagatedLength,
                                            double ZStep2)
{
	
	if(TypeSolver == VECTOR_MANAKOV_PMD)
	  {
	    cerr << "PMD and Full Raman not implemented yet. ABORTING" 
		     << endl;
			 
	    exit(1);		 
	  
	  }
	

   switch(TypeRamanAmplification)
   {
   
     case NO_RAMAN:
	       break;
		     
     case SIMPLE_RAMAN:
           ApplySimpleRamanOperator(TotalPropagatedLength,ZStep2);
		   break;
		   
     case FULL_RAMAN:
	       ApplyFullRamanOperator(TotalPropagatedLength,ZStep2);
		   break;
   
      default:
	     cerr << "Switch error in OptFiberLocalError::ApplyRamanOperator. ABORTING"
		      << endl;
	      exit(1);
   } // end switch
   
} // end ApplyRamanOperator

// #############################################################################
   
void OptFiberLocalError::ApplySimpleRamanOperator(double TotalPropagatedLength,
                                            double ZStep2)
{  

  // Added simple Raman amplifier on May 21, 2002 (JZ).

// Only suitable for a single pump and a single signal wavelength
// This is for Backward Pumping only right now.
// The theory in the forward pumping case is explained in
// Agrawal p 319. To do the backward case change z to -z
// in the pump equation, and solve the BVP.
// If you want to add noise use the equivalent noise figure
// of Hansen et al PTL Vol 10 Num 1 1998. Put this in by 
// adding lumped loss and an EDFA after the Raman amplified 
// fiber. The lumped loss has loss equal to the on-off gain of 
// the RA and the EDFA has a gain equal to the on-off gain of the RA
// and a noise figure equal to the equivalent noise figure as 
// in Hansen. (They give measured values you could use).

  double CurrentZValueWithinFiber2 = TotalPropagatedLength - 
                                     RamanStartingLength;

    UpdateFreq();
    double RamanFactor =  exp(0.5*NepperRamanGain*
                          exp(NepperRamanPumpLoss*CurrentZValueWithinFiber2)*
			  (exp(NepperRamanPumpLoss*ZStep2)-1)/
			  (exp(NepperRamanPumpLoss*TotalLengthFiber)-1));

    /***********
    cout << "CurrentZValueWithinFiber2 = " 
         << CurrentZValueWithinFiber2 << endl;
    cout << "ZStep2 = " << ZStep2 << endl;
    cout << "TotalLengthFiber = " << TotalLengthFiber << endl;
    cout << "RamanFactor = "  << RamanFactor << endl;
    *****************/

    switch(TypeSolver)
     {

       case SCALAR_NLS:

        for(int jj=0;jj<qtPoints;jj++)
         {
           sfftM.fOptSignalX[jj] *= RamanFactor;
         }

       break;

       case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:

        for(int jj=0;jj<qtPoints;jj++)
         {
          sfftM.fOptSignalX[jj] *= RamanFactor;
          sfftM.fOptSignalY[jj] *= RamanFactor;
         }


       break;

       default:

        cerr << "TypeSolver switch error in "
             << "OptFiberLocalError::ApplyRamanOperator " << endl;

        exit(1);

      } // ## end switch
   

    TimeDataValid = false;

} // ## end ApplySimpleRamanOperator

// #######################################################################

void OptFiberLocalError::ApplyFullRamanOperator(double TotalPropagatedLength ,
                                                double ZStep2)
{
    int jjshift;
	
    UpdateFreq();

    double LengthWithinRamanAmp = TotalPropagatedLength - RamanStartingLength;

    /*
    cout << LengthWithinRamanAmp + ZStep2 << " " 
	 << LengthWithinRamanAmp <<  " " << RamanStartingLength << " "
	 << TotalPropagatedLength << endl;
    */
     
    switch(TypeSolver)
     {

       case SCALAR_NLS:

        for(int jj=0;jj<qtPoints;jj++)
         {	
		  // in ocsRamanAmp class I (JH) use the FFTshift before 
		 //transfer the signal to the Raman class.
		 
		if(jj>=qtPoints/2)
		  {
		   jjshift = jj - qtPoints/2;
		  }
		else
		  {
		   jjshift = jj + qtPoints/2;
		  }
//				 	cout << TotalPropagatedLength << "  " << RamanStartingLength << "  " ;
					
//					cout << LengthWithinRamanAmp << "  " << ZStep2 << " jj="  << jj <<endl;
				//	exit(1);
					
		sfftM.fOptSignalX[jj] *= sqrt(oRaman->GetRamanPowerFreqIndex(
                                  LengthWithinRamanAmp + ZStep2,jjshift)
      / oRaman->GetRamanPowerFreqIndex(LengthWithinRamanAmp, jjshift));
				
         }

       break;

       case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:

        for(int jj=0;jj<qtPoints;jj++)
         {
		 // use Full Raman Model
		// in ocsRamanAmp class I (JH) use the FFTshift before 
		//transfer the signal to the Raman class.
		
				 	if(jj>=qtPoints/2){
						jjshift = jj-qtPoints/2;
					}else{
						jjshift = jj+qtPoints/2;
					}
	      sfftM.fOptSignalX[jj] *= sqrt(oRaman->GetRamanPowerFreqIndex(LengthWithinRamanAmp + ZStep2,
		                                                         jjshift)
						/ oRaman->GetRamanPowerFreqIndex(LengthWithinRamanAmp,jjshift));
						
	      sfftM.fOptSignalY[jj] *= sqrt(oRaman->GetRamanPowerFreqIndex(LengthWithinRamanAmp + ZStep2,
		                                                         jjshift)
						/ oRaman->GetRamanPowerFreqIndex(LengthWithinRamanAmp, jjshift));
					
		
         }


       break;

       default:

        cerr << "TypeSolver switch error in "
             << "OptFiberLocalError::ApplyRamanOperator " << endl;

        exit(1);

      } // ## end switch
   

    TimeDataValid = false; 

} // ## end ApplyFullRamanOperator

// ####################################################

void OptFiberLocalError::ApplyTimeDomainOperator(double StepSizeZ)
{

 UpdateTime(); 

 if(!DoFastSaturableAbsorptionFlag)
  {
   ApplyKerrNonlinearityOperator(StepSizeZ);
  }
 else
  {
   ApplyKerrNonlinearityOperator(StepSizeZ/2.0);
   ApplyFastSaturableAbsorptionOperator(StepSizeZ);
   ApplyKerrNonlinearityOperator(StepSizeZ/2.0);
  }

 FreqDataValid = false;

}

// ##################################################

void OptFiberLocalError::ApplyKerrNonlinearityOperator(double StepSizeZ)
{

  /* 
     This method is now correct for the Physics convention.
     See FFTandNLSConventions.pdf and the tests therein.
  */


 switch(TypeSolver)
  {

   case SCALAR_NLS:
   {

     /******

    //SLOW METHOD: 

    cplx Factor = jc*gamma*StepSizeZ;

    for(int jj=0;jj<qtPoints;jj++)
     {
      sfftM.tOptSignalX[jj] *= exp(Factor*sq(abs(sfftM.tOptSignalX[jj])));
     }
    *********/

     // FAST METHOD: (Debugged JZ Oct 20 2002)

    register cplx *t_X_ptr = sfftM.tOptSignalX;

    cplx * t_X_end_ptr = t_X_ptr + qtPoints;

    double Factor = gamma*StepSizeZ;

    if(ScalarSolverEightNinthsFlag)
       Factor *= 8.0/9.0;

    if(DoTimeOperatorExactlyFlag)
      {
          for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++)
           {
           *t_X_ptr *= exp_jc_real(Factor*sq(abs(*t_X_ptr)));
           }

      } // end if
    else
      {
	// RON's method:
        // Use a series expansion of sine and cosine to compute exp(x).
        // The point is that the nonlinear phase rotation for x < 3e-4
        // is so small that an expansion up to 4th order is more than 
        // enough and always yields relative errors smaller than roundoff.
        // Otherwise we use the exact formula.

        for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++)
         {

           register const double x = Factor*sq(abs(*t_X_ptr));

	   if(x > 3e-4)
	     {
              *t_X_ptr *= exp_jc_real(x);
	      //cout << x << endl;
	     }
           else
	   {
             //cout << x << endl;
             /***** 
		4th-order Taylor:
                exp(i*x) = cos(x) +i*sin(x),
                cos(x) = 1 - x^2/2 + x^4/24 + O(x^6)
                sin(x) = x - x^3/6 + O(x^5)
	     *****/

	     register const double xx = x*x;
 
             *t_X_ptr *= cplx( 1.0+(0.0416666666666667*xx-0.5)*xx,
                               (1.0-0.1666666666666667*xx)*x );

	   } // end else
	   
	 } // end for

      } // end else DoTimeOperatorExactlyFlag

    }        
    break;

  case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:
    {

    /****************

    // SLOW METHOD

    cplx Factor = jc*(8.0/9.0)*gamma*StepSizeZ;

    for(int jj=0;jj<qtPoints;jj++)
     {

      PowerTemp = sq(abs(sfftM.tOptSignalX[jj])) + 
                  sq(abs(sfftM.tOptSignalY[jj])); 

      sfftM.tOptSignalX[jj] *= exp(Factor*PowerTemp);
      sfftM.tOptSignalY[jj] *= exp(Factor*PowerTemp);
     }

    ***************/

    // FAST METHOD (Debugged JZ Oct 20 2002,NO_PMD at least)

    register cplx *t_X_ptr = sfftM.tOptSignalX;
    register cplx *t_Y_ptr = sfftM.tOptSignalY;

    register const double Factor = (8.0/9.0)*gamma*StepSizeZ;

    cplx * t_X_end_ptr = t_X_ptr +qtPoints;

    if(DoTimeOperatorExactlyFlag)
      {
        for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++, t_Y_ptr++)
         {

          register double PowerTemp = sq(abs(*t_X_ptr)) + sq(abs(*t_Y_ptr)); 

          *t_X_ptr *= exp_jc_real(Factor*PowerTemp);
          *t_Y_ptr *= exp_jc_real(Factor*PowerTemp);
	 }

      } // end if
    else
      {
	// RON's method:
        // Use a series expansion of sine and cosine to compute exp(x).
        // The point is that the nonlinear phase rotation for x < 3e-4
        // is so small that an expansion up to 4th order is more than 
        // enough and always yields relative errors smaller than roundoff.
        // Otherwise we use the exact formula.
	
        for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++, t_Y_ptr++)
         {

           register const double x = Factor*
	                             (sq(abs(*t_X_ptr)) + sq(abs(*t_Y_ptr))); 

	   if(x > 3e-4)
	     {
              *t_X_ptr *= exp_jc_real(x);
              *t_Y_ptr *= exp_jc_real(x);
	     }
           else
	   {

             /***** 
		4th-order Taylor:
                exp(i*x) = cos(x) +i*sin(x),
                cos(x) = 1 - x^2/2 + x^4/24 + O(x^6)
                sin(x) = x - x^3/6 + O(x^5)
	     *****/

	     register const double xx = x*x;
 
             *t_X_ptr *= cplx( 1.0+(0.0416666666666667*xx-0.5)*xx,
                               (1.0-0.1666666666666667*xx)*x );

             *t_Y_ptr *= cplx( 1.0+(0.0416666666666667*xx-0.5)*xx,
                               (1.0-0.1666666666666667*xx)*x );

	   } // end else
	   
	 } // end for

      } // end else DoTimeOperatorExactlyFlag

    }
    break;


  default:

    cerr << "TypeSolver switch error in "
         << "OptFiberLocalError::ApplyTimeDomainOperator " << endl;

    exit(1);

  } //## end switch


 FreqDataValid = false;

} // ## end ApplyKerrNonlinearityOperator

// ########################################################

void OptFiberLocalError::ApplyFastSaturableAbsorptionOperator(double StepSizeZ)
{

if(TypeFastSaturableAbsorption == QUADRATIC_ANALYTIC)
   ApplyFastSaturableAbsorptionQuadraticAnalytic(StepSizeZ);
else
   ApplyFastSaturableAbsorptionHeunsMethod(StepSizeZ);

} // end ApplyFastSaturableAbsorptionOperator

// ######################################################

void OptFiberLocalError::ApplyFastSaturableAbsorptionQuadraticAnalytic(
                         double StepSizeZ)
{


 switch(TypeSolver)
 {
  case SCALAR_NLS:
  {
  register cplx *t_X_ptr = sfftM.tOptSignalX;

  cplx * t_X_end_ptr = t_X_ptr + qtPoints;

  double Factor = 2*StepSizeZ*FastSaturableAbsorptionCoefficient;

  // Ignoring 8/9-ths factor here for now at least as would have to
  // think about theory and probbaly not relevant over short lengths
  // typically have in lasers??

  for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++)
   {

    register double Base = 1-Factor*sq(abs(*t_X_ptr));

  if(Base <= 0)
  {
   cout << "OptFiberLocalError::ApplyFastSaturableAbsorptionQuadraticAnalytic"
        << endl
        << "Singularity in analytic ODE solver. ABORT!" << endl;
   exit(1);
  }

    *t_X_ptr *= pow(Base, -0.5);
   }
  
  } // end SCALAR_NLS block
  break;

  case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:
  {

   register cplx *t_X_ptr = sfftM.tOptSignalX;
   register cplx *t_Y_ptr = sfftM.tOptSignalY;

   cplx * t_X_end_ptr = t_X_ptr + qtPoints;

   double Factor = 2*StepSizeZ*FastSaturableAbsorptionCoefficient;

   for( ; t_X_ptr < t_X_end_ptr; t_X_ptr++, t_Y_ptr++)
    {
     register double Base = 1-Factor*(sq(abs(*t_X_ptr)) + sq(abs(*t_Y_ptr)));

     if(Base <= 0)
      {
        cout << "OptFiberLocalError::"
             << "ApplyFastSaturableAbsorptionQuadraticAnalytic"
             << endl
              << "Singularity in analytic ODE solver. ABORT!" << endl;
        exit(1);
       }

     *t_X_ptr *= pow(Base, -0.5);
     *t_Y_ptr *= pow(Base, -0.5);

    }

  } // end VECTOR block
  break;

  default: 

   cerr << "OptFiberLocalError:: ApplyFastSaturableAbsorptionQuadraticAnalytic"
        << endl
        << "Switch error. ABORT!" << endl;
   exit(1);
 } // end swicth

} // end ApplyFastSaturableAbsorptionQuadraticAnalytic

// #######################################################

/*!
Scalar Case:

du/dz = F(u)*u

where in QUADRATIC case F(x)= sq(abs(x))

with Heun's Method (aka Modified Euler, RK2) is

v_{n+1} = u_n + h*F(u_n)*u_n
u_{n+1}= u_n + 0.5*h*(F(u_n)+F(v_{n+1}).

Vector Case:

U=(u_X,u_Y)^T

dU/dz = F(U)*U

where in QUADRATIC case F(U) = F(u_X) + F(u_Y)

*/

void OptFiberLocalError::ApplyFastSaturableAbsorptionHeunsMethod(
                          double StepSizeZ)
{

switch(TypeSolver)
{

 case SCALAR_NLS:
  {
   register cplx * InX_ptr = sfftM.tOptSignalX;
   cplx * InX_end_ptr = InX_ptr + qtPoints;

   register cplx * OutX_ptr = LaserWorkX;

   for( ; InX_ptr < InX_end_ptr; InX_ptr++,OutX_ptr++ )
       *OutX_ptr = (*InX_ptr)*(1+ 
                   + StepSizeZ*FastSaturableAbsorptionFunction(*InX_ptr));

   InX_ptr = LaserWorkX;
   InX_end_ptr = InX_ptr + qtPoints;

   OutX_ptr = sfftM.tOptSignalX;

   for( ; InX_ptr < InX_end_ptr; InX_ptr++,OutX_ptr++ )
      *OutX_ptr = *OutX_ptr + 
                  0.5*StepSizeZ*(
                    FastSaturableAbsorptionFunction(*OutX_ptr)*(*OutX_ptr) +
                    FastSaturableAbsorptionFunction(*InX_ptr)*(*InX_ptr));

  } //end case SCALAR_NLS

 break;

 case VECTOR_MANAKOV_NO_PMD:  case VECTOR_MANAKOV_PMD:
  {
   register cplx * InX_ptr = sfftM.tOptSignalX;
   register cplx * InY_ptr = sfftM.tOptSignalY;

   cplx * InX_end_ptr = InX_ptr + qtPoints;

   register cplx * OutX_ptr = LaserWorkX;
   register cplx * OutY_ptr = LaserWorkY;

   for( ; InX_ptr < InX_end_ptr;
          InX_ptr++,InY_ptr++,OutX_ptr++,OutY_ptr++ )
    {
     cplx InFSA = FastSaturableAbsorptionFunction(*InX_ptr) +
                  FastSaturableAbsorptionFunction(*InY_ptr);

     *OutX_ptr = (*InX_ptr)*(1+ StepSizeZ*InFSA);
     *OutY_ptr = (*InY_ptr)*(1+ StepSizeZ*InFSA);

    }

   InX_ptr = LaserWorkX;
   InY_ptr = LaserWorkY;
 
   InX_end_ptr = InX_ptr + qtPoints;

   OutX_ptr = sfftM.tOptSignalX;
   OutY_ptr = sfftM.tOptSignalY;

   for( ; InX_ptr < InX_end_ptr; 
          InX_ptr++,InY_ptr++,OutX_ptr++,OutY_ptr++ )
    {
      cplx OutFSA = FastSaturableAbsorptionFunction(*OutX_ptr) +
                    FastSaturableAbsorptionFunction(*OutY_ptr);

      cplx InFSA = FastSaturableAbsorptionFunction(*InX_ptr) +
                   FastSaturableAbsorptionFunction(*InY_ptr);

      *OutX_ptr = *OutX_ptr +
                  0.5*StepSizeZ*(OutFSA*(*OutX_ptr) + InFSA*(*InX_ptr));

      *OutY_ptr = *OutY_ptr +
                  0.5*StepSizeZ*(OutFSA*(*OutY_ptr) + InFSA*(*InY_ptr));

    }

  } //end case VECTOR

} // end switch

} // end ApplyFastSaturableAbsorptionHeunsMethod

// ######################################################

//! Fast Saturable Absortion is du/dz = F(u)u where
//! F=FastSaturableAbsorptionFunction

cplx OptFiberLocalError::FastSaturableAbsorptionFunction(cplx u)
{

cplx OutValue;

switch(TypeFastSaturableAbsorption)
{

case QUADRATIC:
     
     OutValue = FastSaturableAbsorptionCoefficient*sq(abs(u));

     break;

default:

cerr << "OptFiberLocalError::FastSaturableAbsorptionFunction" << endl
     << "Switch error. ABORT." << endl;
exit(1);

} // end switch

return OutValue;

}

// #######################################################

double OptFiberLocalError::ComputeMaxPower(void)
{

UpdateTime();

MaxPower = 0.0;

double Power;

switch(TypeSolver)
      {

      case SCALAR_NLS:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj]));

           if(Power > MaxPower)  
              MaxPower = Power;
              
	 } // end for

        break;

      case VECTOR_MANAKOV_PMD: case VECTOR_MANAKOV_NO_PMD:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj])) + 
                   sq(abs(sfftM.tOptSignalY[jj]));

           if(Power > MaxPower)
              MaxPower = Power;
	 }

        break;

      default:

	cerr << "TypeSolver error in OptFiberLocalError::ComputeMaxPower"
             << endl;
        exit(1);

      } // ## end switch


return MaxPower;

}// ## end ComputeMaxPower

// #####################################################

void OptFiberLocalError::CopyFftMatricesFreqData(fftMatrices * OrigData, 
                                                 fftMatrices * CopyOfData)
{

  // void * memcpy(void * dest, const void * src, size_t NumBytes);

 switch(TypeSolver)
   {

    case SCALAR_NLS:
      
     memcpy(CopyOfData->fOptSignalX,OrigData->fOptSignalX,
	    qtPoints*sizeof(cplx));

     break;

   case VECTOR_MANAKOV_PMD: case VECTOR_MANAKOV_NO_PMD:

     memcpy(CopyOfData->fOptSignalX,OrigData->fOptSignalX,
	    qtPoints*sizeof(cplx));

     memcpy(CopyOfData->fOptSignalY,OrigData->fOptSignalY,
	    qtPoints*sizeof(cplx));

     break;

   default:
     cerr << "TypeSolver switch error in "
          << "OptFiberLocalError::CopyFftMatricesFreqData" << endl;

     exit(1);

   } // ## end switch

} // ## end CopyFftMatricesFreqData

// #######################################################

double OptFiberLocalError::ComputeThirdOrderSolutionAndRelativeError(void)
{

  // This method computes the solution up to 3rd order 
  // and the relative error.

  // If S_c and S_f denote the coarse and fine solutions then 
  // up to third order the solution S_third is 

  // S_third = 4/3 S_f - 1/3 S_c.

  // The relative error is defined to be
  
  // RelativeError = | S_c - S_f | / | S_f |

 double NumSum = 0.0;
 double DenSum = 0.0;

 // JZFeb03: LocalErrorFreqLimit will not work well when have one channel
 // and it's not at center frequency!! 

 int LocalErrorFreqLimit = oOptSignal->GetRelevantNumModesOFLE()/2;

 switch(TypeSolver)
   {
     case SCALAR_NLS:
     {
       cplx * f_X_Coarse_ptr = sfftM_Coarse.fOptSignalX;
       cplx * f_X_Fine_ptr = sfftM.fOptSignalX;

      for(int jj = 0; jj < qtPoints; jj++, f_X_Coarse_ptr++, f_X_Fine_ptr++)
	{
          // sfftM.fOptSignalX contains the fine step result
	  // in the frequency domain.

	  // ASE NOISE FIX OF RON

	  if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	    {
               NumSum += 
               sq(abs( (*f_X_Coarse_ptr) - (*f_X_Fine_ptr) ));
	    }

          if(jj <  LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	    {
              DenSum +=  sq(abs(*f_X_Fine_ptr));
	    }

	  if(TypeStepSizes == LOCAL_ERROR_3RD)
	    {
              *f_X_Fine_ptr = (4.0*(*f_X_Fine_ptr) - (*f_X_Coarse_ptr))/3.0;
	    }

       // if TypeStepSizes == LOCAL_ERROR_2ND we keep the fine step solution
       // which is already in sfftM.fOptSignal. So do nothing here

	} // ## end for jj loop

       RelativeError = sqrt(NumSum/DenSum);
       TimeDataValid = false;
     }
       break;

    case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:
      {

       cplx * f_X_Coarse_ptr = sfftM_Coarse.fOptSignalX;
       cplx * f_X_Fine_ptr = sfftM.fOptSignalX;
       cplx * f_Y_Coarse_ptr = sfftM_Coarse.fOptSignalY;
       cplx * f_Y_Fine_ptr = sfftM.fOptSignalY;

     for(int jj = 0; jj < qtPoints; 
	 jj++,f_X_Coarse_ptr++,f_X_Fine_ptr++,f_Y_Coarse_ptr++,f_Y_Fine_ptr++)
      {
          // sfftM.fOptSignal[X,Y] contains the fine step result
	  // in the frequency domain.
	 
       if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	 {
           NumSum += 
	             sq(abs((*f_X_Coarse_ptr) - (*f_X_Fine_ptr)))
	            +sq(abs((*f_Y_Coarse_ptr) - (*f_Y_Fine_ptr)));

	 }

       if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	 {
          DenSum +=  sq(abs(*f_X_Fine_ptr)) + 
                     sq(abs(*f_Y_Fine_ptr));
	 }


       if(TypeStepSizes == LOCAL_ERROR_3RD)
	 { 
          *f_X_Fine_ptr = (4.0*(*f_X_Fine_ptr) -(*f_X_Coarse_ptr))/3.0;
	  *f_Y_Fine_ptr = (4.0*(*f_Y_Fine_ptr) -(*f_Y_Coarse_ptr))/3.0;
     	 }

       // if TypeStepSizes == LOCAL_ERROR_2ND we keep the fine step solution
       // which is already in sfftM.fOptSignal. So do nothing here.


	} // ## end for jj loop

       RelativeError = sqrt(NumSum/DenSum);
       TimeDataValid = false;
      } 
       break;

    default:
      cerr << "TypeSolver switch error in" 
           << "OptFiberLocalError::ComputeThirdOrderSolutionAndRelativeError"
	   << endl;

      exit(1);

     } // ## end switch


   return RelativeError;

} // ## end ComputeThirdOrderSolutionAndRelativeError

// ######################################################

double OptFiberLocalError::ComputeThirdOrderSolutionAndRelativeErrorOLD(void)
{

  // This method computes the solution up to 3rd order 
  // and the relative error.

  // If S_c and S_f denote the coarse and fine solutions then 
  // up to third order the solution S_third is 

  // S_third = 4/3 S_f - 1/3 S_c.

  // The relative error is defined to be
  
  // RelativeError = | S_c - S_f | / ( 3 S_third )

 

 double NumSum = 0.0;
 double DenSum = 0.0;

 int LocalErrorFreqLimit = oOptSignal->GetRelevantNumModesOFLE()/2;

 switch(TypeSolver)
   {
     case SCALAR_NLS:
      
      for(int jj = 0; jj < qtPoints; jj++)
	{
          // sfftM.fOptSignalX contains the fine step result
	  // in the frequency domain.

	  // FIX: ASE NOISE FIX OF RON

	  if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	    {
               NumSum += 
               sq(abs(sfftM_Coarse.fOptSignalX[jj] - sfftM.fOptSignalX[jj]));
	    }

          sfftM.fOptSignalX[jj] = (4.0*sfftM.fOptSignalX[jj] - 
                                sfftM_Coarse.fOptSignalX[jj])/3.0;

          if(jj <  LocalErrorFreqLimit||jj >= qtPoints-LocalErrorFreqLimit)
	    {
              DenSum +=  sq(abs(sfftM.fOptSignalX[jj]));
	    }

	} // ## end for jj loop

       RelativeError = sqrt(NumSum/DenSum) / 3.0;
       TimeDataValid = false;

       break;

    case VECTOR_MANAKOV_NO_PMD: case VECTOR_MANAKOV_PMD:

        for(int jj = 0; jj < qtPoints; jj++)
	{
          // sfftM.fOptSignal[X,Y] contains the fine step result
	  // in the frequency domain.

	 
       if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	 {
           NumSum += 
                 sq(abs(sfftM_Coarse.fOptSignalX[jj] - sfftM.fOptSignalX[jj]))
	        +sq(abs(sfftM_Coarse.fOptSignalY[jj] - sfftM.fOptSignalY[jj]));

	 }

       sfftM.fOptSignalX[jj] = (4.0*sfftM.fOptSignalX[jj] - 
                                sfftM_Coarse.fOptSignalX[jj])/3.0;


       sfftM.fOptSignalY[jj] = (4.0*sfftM.fOptSignalY[jj] - 
                                sfftM_Coarse.fOptSignalY[jj])/3.0;

	 
       if(jj < LocalErrorFreqLimit || jj >= qtPoints-LocalErrorFreqLimit)
	 {
          DenSum +=  sq(abs(sfftM.fOptSignalX[jj])) + 
                     sq(abs(sfftM.fOptSignalY[jj]));
	 }

	} // ## end for jj loop

       RelativeError = sqrt(NumSum/DenSum) / 3.0;
       TimeDataValid = false;

       break;

    default:
      cerr << "TypeSolver switch error in" 
           << "OptFiberLocalError::ComputeThirdOrderSolutionAndRelativeError"
	   << endl;

      exit(1);

     } // ## end switch


   return RelativeError;

} // ## end ComputeThirdOrderSolutionAndRelativeErrorOLD

// ######################################################

void OptFiberLocalError::UpdateAccumulatedFirstOrderDispersion(double StepSize)
{

 oOptSignal->SetAccumulatedFirstOrderDispersion(
 oOptSignal->GetAccumulatedFirstOrderDispersion() + StepSize*FstOrDispFiber);

}
// ######################################################################

void OptFiberLocalError::UpdateAccumulatedSecondOrderDispersion
                                                       (double StepSize)
{

 oOptSignal->SetAccumulatedSecondOrderDispersion(
 oOptSignal->GetAccumulatedSecondOrderDispersion() + StepSize*SndOrDispFiber);

}

// ################################################################

void OptFiberLocalError::FiberAmplifierComputeSaturableGain(void)
{

  FiberAmplifierSaturableGainPerLength =
    FiberAmplifierUnsaturatedNepperGain/
    (1+oOptSignal->GetTotalEnergy()/FiberAmplifierSaturationEnergy);

 //cout << "FiberAmplifierSaturableGainPerLength = "
 //     << FiberAmplifierSaturableGainPerLength << endl;

}

// ###############################################################

double OptFiberLocalError::FiberAmplifierComputeSaturableGainForHeunsMethod(
     bool FirstStep)
{

  double GainPerLength =
    FiberAmplifierUnsaturatedNepperGain/
    (1+FiberAmplifierGetEnergyForHeunsMethod(FirstStep)/
       FiberAmplifierSaturationEnergy);

  return GainPerLength;

}


// #################################################################

double OptFiberLocalError::FiberAmplifierGetEnergyForHeunsMethod(bool FirstStep)
{

  double Energy = 0.0;

 if(FirstStep)
  {

  if(TypeSimulation == VECTOR)
    {

     for (int ii = 0; ii < qtPoints; ii++)
     {
         Energy += sq(abs(sfftM.fOptSignalX[ii]))
                  +sq(abs(sfftM.fOptSignalY[ii]));

     } // ##end for ii loop
    }
  else
    {
     for (int ii = 0; ii < qtPoints; ii++)
       {
         Energy += sq(abs(sfftM.fOptSignalX[ii]));

       } // ##end for ii loop

    }
  } // end if(FirstStep)
 else
  {

  if(TypeSimulation == VECTOR)
    {

     for (int ii = 0; ii < qtPoints; ii++)
     {
         Energy += sq(abs(LaserWorkX[ii]))
                  +sq(abs(LaserWorkY[ii]));

     } // ##end for ii loop
    }
  else
    {
     for (int ii = 0; ii < qtPoints; ii++)
       {
         Energy += sq(abs(LaserWorkX[ii]));

       } // ##end for ii loop

    }
  } // end else(FirstStep)

 return Energy*oOptSignal->GetTimeWindow();

}

// ##################################################################

void OptFiberLocalError::FiberAmplifierComputeNoiseAmplitudeFactor(
                                                   double StepSize)
{

  if(TypeSolver == SCALAR_NLS && 
     TypeFiberLossGain == FIBER_AMPLIFIER &&
     TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF)
    {


      double LinearPowerGain;

      switch(TypeFiberAmplifierGain)
      {
       case FIBER_AMP_UNSATURATED_FLAT_GAIN:
        LinearPowerGain = exp(-StepSize*NepperAttenuationFiber);
        break;

       case FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN:
        LinearPowerGain = exp(FiberAmplifierSaturableGainPerLength*StepSize);
        break;

       default:
        cerr << "Switch error in method" 
             << "OptFiberLocalError::FiberAmplifierComputeNoiseAmplitudeFactor"
             << endl;
        exit(1);

      }

      FiberAmplifierNoiseAmplitudeFactor = 0.0;

     // Use SpontaneousEmissionFactor - See Desurvire p. 77 (2.33)
     // Also see Curtis Short Course Slide No. 86. 
     // For us a node point is 1 complex mode.

     // We use the sqrt because we add average energy of n_sph\nu(G-1)
     // to each complex mode

     // We only have noise if there is gain (rather than loss)

      if (LinearPowerGain >= 1) 
      { 
        FiberAmplifierNoiseAmplitudeFactor = 
                      sqrt( hPlanck*FiberAmplifierSpontaneousEmissionFactor
                            *DeltaFreq*(LinearPowerGain-1) );
      }
    }

}

// #################################################################

void OptFiberLocalError::FiberAmplifierAddMonteCarloNoise(double StepSize)
{
   if(TypeSolver == SCALAR_NLS &&
      TypeFiberLossGain == FIBER_AMPLIFIER &&
      TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF)
    {

      UpdateFreq();
  
      FiberAmplifierComputeNoiseAmplitudeFactor(StepSize);
       
      register cplx * f_X_ptr = sfftM.fOptSignalX;
      cplx * f_X_end_ptr = f_X_ptr + qtPoints;
 
      switch(TypeFiberAmplifierNoise)
         {
           case FIBER_AMP_NOISE_GAUSSIAN_WHITE:
      
            for (int jj=0 ; f_X_ptr < f_X_end_ptr; f_X_ptr++, jj++)
             {

               *f_X_ptr += FiberAmplifierNoiseAmplitudeFactor*
                           sqrt(CenterFreq + fft->freq(jj))*
                           ( RNG->GetGaussianDeviate()
                         +jc*RNG->GetGaussianDeviate() )/sqrt(2.);

             }

           break;

           default: 

            cerr << "OptFiberLocalError::FiberAmplifierAddMonteCarloNoise"
                 << endl
                 << "Invalid TypeFiberAmplifierNoise. Aborting." 
                 << endl;
            exit(1);

         }

     TimeDataValid = false;
      
     }  // end if

}

// ##############################################################

void OptFiberLocalError::WriteStepSizes(double PropagatedLength)
{

  // Note that we are outputting this data every second fine step
  // so that the sum of the DeltaZ's will only be half of the 
  //  PropagatedLength at the end of the run.

  double StepSize = DeltaZ;

  if(TypeStepSizes == LOCAL_ERROR_2ND || TypeStepSizes == LOCAL_ERROR_3RD)
    StepSize /= 2;

if(OutputStepSizesFlag)
  {

    ComputeMaxPower();

    StepSizesFile << PropagatedLength*1e-3 << "\t" 
		  << StepSize << "\t" 
                  << MaxPower*1e3 << "\t" << RelativeError  
                  << "\t" 
                  << DispersionFiberCenterWavelength*1e12/(1e9*1e-3) 
                  << "\t"
                  << sq(abs(sfftM.fOptSignalX[0])) +   
                     sq(abs(sfftM.fOptSignalY[0]));
					 
					 
    if(WritePulseWidthFlag)
      {      
        StepSizesFile << "\t" << 1e12*sqrt(oOptSignal->GetSigmaSqOptSignal());
        StepSizesFile << "\t" << 1e3*oOptSignal->GetTotalEnergy();
        StepSizesFile << "\t" 
                      << 1e24*oOptSignal->GetAccumulatedFirstOrderDispersion();
        StepSizesFile << "\t" 
                   << 1e36*oOptSignal->GetAccumulatedSecondOrderDispersion();

      }


    if(TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
      {
        StepSizesFile << "\t" << FiberAmplifierSaturableGainPerLength;
      }

    StepSizesFile << endl;
    
   } // ## end if

} // end ## WriteStepSizes

// ##############################################################

void OptFiberLocalError::SetOutputStepSizesFlag(int OutputStepSizesFlag2)
{

  OutputStepSizesFlag = OutputStepSizesFlag2;

  /*********
  LogFileSeparator();

  *LogFile << "WARNING: OutputStepSizesFlag reset to " 
         <<  OutputStepSizesFlag << endl << endl;
  ***********/
}

// ##############################################################

void OptFiberLocalError::SetRelativeErrorGoal(double RelativeErrorGoal2)
{

 RelativeErrorGoal = RelativeErrorGoal2;

 /************
 LogFileSeparator();

 *LogFile << "WARNING: RelativeErrorGoal reset to " 
         <<  RelativeErrorGoal << endl << endl;
 ************/

} // ## end SetRelativeErrorGoal

// #########################################################

void OptFiberLocalError::SetFstOrDispFiber(double FstOrDispFiber2)
{

 FstOrDispFiber = FstOrDispFiber2;

 /************
 LogFileSeparator();

 *LogFile << "WARNING: First order dispersion reset to " 
         <<  FstOrDispFiber << endl << endl;

 ***********/
}

// ###################################

void OptFiberLocalError::SetSndOrDispFiber(double SndOrDispFiber2)
{
 SndOrDispFiber_Save = SndOrDispFiber; // for recover
 SndOrDispFiber = SndOrDispFiber2;

 /******************
 LogFileSeparator();

 *LogFile << "WARNING: Second order dispersion reset to " 
         <<  SndOrDispFiber << endl << endl;
 ***********/
}

// ###################################

void OptFiberLocalError::RecoverSndOrDispFiber()
{
 SndOrDispFiber = SndOrDispFiber_Save; // for recover

 /******************
 LogFileSeparator();

 *LogFile << "WARNING: Second order dispersion reset to " 
         <<  SndOrDispFiber << endl << endl;
 ***********/
}

// ###########################################

void OptFiberLocalError::SetDispersion(double DispersionFiberRefWavelength,
				       double DispSlopeFiberRefWavelength)
{

  // Value must be in  SI units s/m^2

 // ######### Compute FstOrDispFiber and SndOrDispFiber #######

 // The  FstOrDispFiber and SndOrDispFiber are to be computed
 // at the CenterWavelength of the optical signal which is passing through
 // the fiber. 
 // For this we need to convert the values of DisperionFiber and 
 // DisperionSlopeFiber given in the input file at the ReferenceWavelength
 // to values at the CenterWavelength. 
 // This is done using a linear approximation.

if(!SetFstSndOrDispDirectlyFlag)
{
 if (ReferenceFreq == 0)
   {
     *LogFile << "ERROR: ReferenceFreq = 0" << endl
              << "ABORTING NOW"
              << endl; 
      exit(1);   
   }   

double CenterWavelength = Frequency2Wavelength(CenterFreq);

DispersionFiberCenterWavelength = DispersionFiberRefWavelength + 
         DispSlopeFiberRefWavelength*(CenterWavelength - ReferenceWavelength);

DispSlopeFiberCenterWavelength = DispSlopeFiberRefWavelength;


// See Agrawal's blue book for the following formulae

FstOrDispFiber = -DispersionFiberCenterWavelength 
                  *LightSpeed/(2*pi*sq(CenterFreq) );

SndOrDispFiber = LightSpeed/(sq(2*pi)*cube(CenterFreq) )*
                (LightSpeed/CenterFreq*DispSlopeFiberCenterWavelength
                +  2*DispersionFiberCenterWavelength);

}

} // ## end SetDispersion

// ########################################################

//! Returns value of Dispersion in s/m^2 at Wavelength in m

double OptFiberLocalError::GetDispersion(double Wavelength)
{
  return DispersionFiberCenterWavelength + 
         DispSlopeFiberCenterWavelength*
          (Wavelength - Frequency2Wavelength(CenterFreq));

}

// #######################################

void OptFiberLocalError::SetGammaFiber(double gamma2)
{
 gamma_Save =  gamma; // add by Jonathan for recover
  // Written by John Zweck, Nov 2nd, 2000

 gamma = gamma2;
 WithNonLin  = (gamma != 0); 

 /*
 LogFileSeparator();

 *LogFile << "WARNING: Nonlinearity coefficient  reset to Gamma = " 
         <<  gamma << endl << endl;
      

*/

} // end SetGammaFiber

// #######################################

void OptFiberLocalError::RecoverGammaFiber()
{

 gamma = gamma_Save;
 WithNonLin  = (gamma != 0); 

 /*
 LogFileSeparator();

 *LogFile << "WARNING: Nonlinearity coefficient  reset to Gamma = " 
         <<  gamma << endl << endl;
      

*/

} // end SetGammaFiber
// ######################################################

void OptFiberLocalError::SetAttenuationdB(double AttenuationFiber2)
{

  // Written by John Zweck, Nov 2nd, 2000

AttenuationFiber = AttenuationFiber2;
NepperAttenuationFiber = AttenuationFiber*log(10.0)/10.0; 

ComputeDeltaZMax();

}

//#################################################################

void OptFiberLocalError::SetNepperAttenuation(double NepperAttenuationFiber2)
{

NepperAttenuationFiber = NepperAttenuationFiber2;
AttenuationFiber = NepperAttenuationFiber*10.0/log(10.0);

ComputeDeltaZMax();

}

// #########################################################

void OptFiberLocalError::SetFiberAmplifierUnsaturatedNepperGain(
           double FiberAmplifierUnsaturatedNepperGain2)
{

  FiberAmplifierUnsaturatedNepperGain = 
   FiberAmplifierUnsaturatedNepperGain2;

  ComputeDeltaZMax();
}

// #########################################################

void OptFiberLocalError::ComputeDeltaZMax(void)
{

// See docs on enum typeFiberAmplifierNoise in .hh file

DeltaZMax = DeltaZMaxInfile;

if(TypeFiberLossGain == FIBER_AMPLIFIER &&
   TypeFiberAmplifierNoise != FIBER_AMP_NOISE_OFF)
    {


      switch(TypeFiberAmplifierGain)
      {
       case FIBER_AMP_UNSATURATED_FLAT_GAIN:
         DeltaZMax = min(DeltaZMax,
          FiberAmplifierNoiseMaxStepSizeFactor/(abs(NepperAttenuationFiber)));

        break;

       case FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN:

         // If FiberAmplifierUnsaturatedNepperGain is a lot different
         // from actual gain in a given step this may give step sizes
         // that are much smaller than necessary.

         DeltaZMax = min(DeltaZMax,
                     FiberAmplifierNoiseMaxStepSizeFactor/
                      (abs(FiberAmplifierUnsaturatedNepperGain)));

        break;

       default:
        cerr << "Switch error in method"
             << "OptFiberLocalError::ComputeDeltaZMax"
             << endl;
        exit(1);

      }


     *LogFile << "DeltaZMax [reset for fiber amp noise] = "
              << DeltaZMax << endl;
     cout << "DeltaZMax [reset for fiber amp noise] = "
              << DeltaZMax << endl;
    }


if(TypeFiberLossGain == FIBER_AMPLIFIER &&
   TypeFiberAmplifierGain == FIBER_AMP_SATURABLE_SUPERGAUSSFILT_GAIN)
    {
     DeltaZMax = min(DeltaZMax,
                log(FiberAmplifierGainSaturationMaxGainPerStep)*2.0/
                 FiberAmplifierUnsaturatedNepperGain);
     *LogFile << "DeltaZMax [reset for fiber amp saturable gain] = "
              << DeltaZMax << endl;
     cout << "DeltaZMax [reset for fiber amp saturable gain] = "
              << DeltaZMax << endl;
    }

}

// ##################################################

void OptFiberLocalError::SetWriteAveragePowerdBmFlag(int Value)
{

  // Written by John Zweck, Nov 2nd, 2000

 WriteAveragePowerdBmFlag = Value;

 /**********
 *LogFile << "WARNING: Reset WriteAveragePowerdBmFlag to " 
         << WriteAveragePowerdBmFlag  << endl;
 *********/
}

// ##################################################

void OptFiberLocalError::WriteAveragePowerdBm(double PropagatedLength2)
{

  // This method is a bit ad-hoc 

if(WriteAveragePowerdBmFlag)
  {
    //oOptSignal->WriteAveragePowerdBm(Job + "AveragePower.dat",
    //                                  PropagatedLength2,1);

   oOptSignal->WritePowerdBmChannel(Job + "AveragePowerCh1.dat",
				      PropagatedLength2,1,0);
	 
  }

}

// ####################################################
// 
// The following functions are adapted by JZ from those written by
// Ivan Lima in ocsOptFiber.cc, ocsOptFiberPMDTools.cc
//
// ###########################################################

void OptFiberLocalError::ApplyPMDCompleteScatteringFreqDomainOperator
                         (int ScatteringSectionIndex)
{
   cplx TmpX;
   cplx TmpY;

   cplx TmpJM[2][2];

  switch(TypeSolver)
    {

      case SCALAR_NLS: case VECTOR_MANAKOV_NO_PMD:

        // DO NOTHING!!

       break;
 
      case VECTOR_MANAKOV_PMD:

	 UpdateFreq();

	 ComputeEulerMatrixAndUpdateOutputPDVector(ScatteringSectionIndex);

	 //cout << EulerMatrix[0][0] << " " << EulerMatrix[0][1] << " " 
	 //     << EulerMatrix[1][0] << " " << EulerMatrix[1][1] << endl;


         for(int jj=0;jj<qtPoints;jj++)
	   {

             TmpX = sfftM.fOptSignalX[jj];
             TmpY = sfftM.fOptSignalY[jj];

             sfftM.fOptSignalX[jj] = EulerMatrix[0][0]*TmpX + 
                                     EulerMatrix[0][1]*TmpY;

             sfftM.fOptSignalY[jj] = EulerMatrix[1][0]*TmpX + 
                                     EulerMatrix[1][1]*TmpY;
	   }


          for (int ii=0;ii<NumberOfChannels;ii++)
           {
             TmpJM[0][0]=JonesMatrix[ii*4+0];
             TmpJM[0][1]=JonesMatrix[ii*4+1];
             TmpJM[1][0]=JonesMatrix[ii*4+2];
             TmpJM[1][1]=JonesMatrix[ii*4+3];

             // [00, 01]    [0, 1]
             // [10, 11]    [2, 3]

             JonesMatrix[ii*4+0]=  EulerMatrix[0][0]*TmpJM[0][0]+
                                   EulerMatrix[0][1]*TmpJM[1][0];

             JonesMatrix[ii*4+1]=  EulerMatrix[0][0]*TmpJM[0][1]+
                                   EulerMatrix[0][1]*TmpJM[1][1];

             JonesMatrix[ii*4+2]=  EulerMatrix[1][0]*TmpJM[0][0]+
                                   EulerMatrix[1][1]*TmpJM[1][0];

             JonesMatrix[ii*4+3]=  EulerMatrix[1][0]*TmpJM[0][1]+
                                   EulerMatrix[1][1]*TmpJM[1][1];
            }

         TimeDataValid = false;

      break;

    default:

    cerr << "TypeSolver switch error in "
         << "OptFiberLocalError::ApplyPMDFreqDomainOperator " << endl;

    exit(1);

  } // ## end switch


} // ## end ApplyPMDFreqDomainOperator

// ########################################################

void OptFiberLocalError::GetNewFiberRealization(void)
{

 
  for(int ii=0;ii<NumberScatterings;ii++)
    {
 
      
      Theta[ii] = acos(2*(RNG->GetRanNum() - 0.5));                     
      Phi[ii]   = 2*pi*RNG->GetRanNum();
      Psi[ii]   = 2*pi*RNG->GetRanNum();

      //cout << Theta[ii] << " " << Phi[ii] << " " << Psi[ii] << endl;

    }

} // ### end GetNewFiberRealization

// ########################################################

void OptFiberLocalError::SetFiberRealization(int SectionNumber, 
			 double Theta2, double Phi2, double Psi2)
{

  if(SectionNumber < 0 || SectionNumber >= NumberScatterings)
    {
      cerr << "OptFiberLocalError::SetFiberRealization: Invalid SectionNumber"
	   << endl
	   << "ABORTING" << endl;
      exit(1);
    }

  Theta[SectionNumber] = Theta2;
  Phi[SectionNumber]   = Phi2;
  Psi[SectionNumber]   = Psi2;


} // ### end SetFiberRealization

// ########################################################

void OptFiberLocalError::ComputeEulerMatrixAndUpdateOutputPDVector(
			 int ScatteringSectionIndex)
{
 
  GetEulerMatrix(Theta[ScatteringSectionIndex],Phi[ScatteringSectionIndex],
		 Psi[ScatteringSectionIndex]);

  // Updating Polarization dispersion vector.
  // At this stage we have this set up for one frequency.
  // We could make multiple calls to do it for several freqs.

  double Freq = oOptSignal->GetFreqForPDVectorOFLE(); 
  // Freq = RelFreq + CenterFreq

  PropagateOutputPDVector(Freq,Theta[ScatteringSectionIndex],
		          Phi[ScatteringSectionIndex],
		          Psi[ScatteringSectionIndex]);
}

// ########################################################

void OptFiberLocalError::GetEulerMatrix(double theta,double phi,double psi)  
{ 

   theta=-theta;    // added by Hai
   phi=-phi;        // added by Hai
   psi=-psi;       // added by Hai



   EulerMatrix[0][0] = exp( jc*(psi + phi)/2.0)*cos(theta/2.0);

   EulerMatrix[0][1] = jc*exp( jc*(psi - phi)/2.0)*sin(theta/2.0);

   EulerMatrix[1][0] = jc*exp(-jc*(psi - phi)/2.0)*sin(theta/2.0);

   EulerMatrix[1][1] = exp(-jc*(psi + phi)/2.0)*cos(theta/2.0); 
}

// ######################################################

void OptFiberLocalError::SetMeanDGDPerSqrtLength(double MeanDGDPerSqrtLength2)
{
   // ## Input MeanDGDPerSqrtLength2 must be in [ps/sqrt(km)]

  MeanDGDPerSqrtLength = 1e-12/sqrt(1e3)*MeanDGDPerSqrtLength2; //[s/sqrt(km)]
  PMD = sqrt(3.0*pi/8.0)*MeanDGDPerSqrtLength;  //[s/sqrt(km)]
}

// ########################################################

void OptFiberLocalError::SetPMD(double PMD2)
{
   // ## Input PMD2 must be in [ps/sqrt(km)]

   PMD = 1e-12/sqrt(1e3)*PMD2;  //[s/sqrt(km)]
   MeanDGDPerSqrtLength = sqrt(8./(3.*pi))*PMD;  //[s/sqrt(km)]

}

// ########################################################

double OptFiberLocalError::GetFreqDerivativeOfBirefringence(void)
{
  /*

   Can derive this from JLT 1997, p. 1774
 
   If distance between random rotations is ScatteringStepSize
   and we do one birefringence step in that distance then
   we should return b' = PMD/sqrt(4*ScatteringStepSize).
   If instead we take several small split-steps between random rotations
   and we want to put birefringence in each, then we still return 
   b' = PMD/sqrt(4*ScatteringStepSize). The reason is that for two steps
   of size Delta1 and Delta2 so that ScatteringStepSize = Delta1 + Delta2
   the method ComputeDispersionOperator multiplies by exp( \pm i Delta b')
   so that if you do two steps you just add the Deltas. Of course this argument
   only works between random rotations!! (See handwritten notes by JZ for more
   details.) 25/1/02

  */


  return PMD/sqrt(4*ScatteringStepSize); 
}

// ########################################################

//! Returns the Differential Group Delay.

/*! Computes the DGD at the absolute frequency given in 
 oOptSignal::FreqForPDVectorOFLE for propagation over
 all fibers from where oOptSignal::ResetPDVectorOFLE()
 was called to the current position. 

 The method simply returns the length of the vector given by
 OptFiberLocalError::OutputPDVector. 

 See also OptFiberLocalError::PropagateOutputPDVector

*/

double OptFiberLocalError::GetDGD()
{

   PDVectorHai[0]=OutputPDVector[0];       // added by Hai 8/14/03
   PDVectorHai[1]=OutputPDVector[1];       // added by Hai 8/14/03
   PDVectorHai[2]=OutputPDVector[2];       // added by Hai 8/14/03

   return(sqrt(sq(OutputPDVector[0])+sq(OutputPDVector[1])
                                  +sq(OutputPDVector[2])) ); 
}

// ########################################################

//! Propagates output polarization dispersion vector through a single
//! step of the coarse-step method for PMD. 

/*!

Input parameters are

    -# Freq: The absolute frequency at which 
       OptFiberLocalError::OutputPDVector is to be computed
    -# theta, phi, psi: The Euler angles of the current step

Algorithm:

Let PDV_{n} denote the polarization dispersion vector after the n-th
section in the coarse step method. From Gordon and Kogelnick's review
paper on PMD in Proc. Nat. Acad. Sci., we have that

PDV_{n} = DeltaPDV + T_{n} PDV_{n-1}

where DeltaPDV is the PDV due to the n-th section and T_{n} is the 
matrix which transforms the Stokes vector at the input to the n-th section
to the Stokes vector at the output of the n-th section.

For the case we have of a random rotation given by an Euler matrix followed
by a birefringence rotation in Stokes space we have

DeltaPDV = -2*b'*2*pi*Freq*ScatteringStepSize, where 
b' = OptFiberLocalError::GetFreqDerivativeOfBirefringence()

and 

T_{n} = R_x(2b'ScatteringStepSize*Freq) R_x(Psi)R_y(-Theta)R_x(Phi)

where the first factor in the product is the birefringence rotation
and the other three factors are the rotation by the Euler angles 
in Stokes space.

At the start of the system (z=0) we have OutputPDVector = (0,0,0)
since there DGD = length(OutputPDVector) = 0. 

OptFiberLocalError::PropagateOutputPDVector is called by 
OptFiberLocalError::ComputeEulerMatrixAndUpdateOutputPDVector(int 
ScatteringSectionIndex) which is called  by 
OptFiberLocalError::PropagateWithPMD().

The method OptSignal::ResetPDVectorOFLE() resets the PDVectors to (0,0,0).
This method is called by OptSignal::RegenerateSignalString();

 */

void OptFiberLocalError::PropagateOutputPDVector(double Freq,
                         double theta, double phi, double psi)
{
  double DeltaAngle = 2*pi*Freq*
                      2*GetFreqDerivativeOfBirefringence()*ScatteringStepSize;
 
  RotatesAboutX( phi,  OutputPDVector); 
  RotatesAboutY(-theta,OutputPDVector); 
  RotatesAboutX( psi,  OutputPDVector); 
             
  RotatesAboutX(DeltaAngle,OutputPDVector);   
  OutputPDVector[0] += -2*GetFreqDerivativeOfBirefringence()*
                        ScatteringStepSize;         
}

// ############################################################

/********

void OptFiberLocalError::PropagateStokesTransferMatrix(double Freq,
                         double theta, double phi, double psi, 
			 double *TransferMatrix)
{

  


} // ## end PropagateStokesTransferMatrix

      *******/

// ##########################################################

//! 

/********

double OptFiberLocalError::GetStokesTransferMatrix(int row, int col)
{

  if(row < 1 || row > 3 || col < 1 || col > 3)
    {
      cerr << "OptFiberLocalError::GetStokesTransferMatrix indexing error"
	   << "(row, col) = " << row << " " << col << endl
           << "ABORTING" << endl;

      exit(1);
    }

 return StokesTransferMatrixTranspose[col-1][row-1]
   
} // ## end GetTransferMatrix

// ############################################################

	 **********/


 // /***** 

// Commented out by JZ July 2005

// below is added by Hai at 7-1-03
void OptFiberLocalError::UpdateJonesMatrixAfterBiregringence(double StepSizeZ)
{	
	double BirefringenceAmount;
	double BirefringencePhase;
	cplx BirefringenceFactorX;
	cplx BirefringenceFactorY;
	cplx TmpJM[2][2];
	
	BirefringenceAmount=StepSizeZ*GetFreqDerivativeOfBirefringence();
	for (int ii=0;ii<NumberOfChannels;ii++)
	{

	     BirefringencePhase=FreqChannels[ii]*2.0*pi*BirefringenceAmount;
	     BirefringenceFactorX=exp_jc_real(-BirefringencePhase);
	     BirefringenceFactorY=exp_jc_real(BirefringencePhase);	     

// notice here assume no chromatic dispersion and loss !!!
	     
    
	     JonesMatrix[ii*4+0]*=BirefringenceFactorX;
					   
	     JonesMatrix[ii*4+1]*=BirefringenceFactorX;
				   		
	     JonesMatrix[ii*4+2]*=BirefringenceFactorY;
				   
	     JonesMatrix[ii*4+3]*=BirefringenceFactorY;
	 }
}

// ****/

//####################################################

// added by Walter
void OptFiberLocalError::SetConstantDeltaZ(double DeltaZValue)
{
// set DeltaZ in meters
if (DeltaZValue<=0 || DeltaZValue>LengthFiber)
   {
   cerr << "Error using OptFiberLocalError::SetConstantDeltaZ " << endl
   << "DeltaZValue<=0 || DeltaZValue>FiberLength " << endl
   << "ABORTING!!! " <<endl;
   exit(-1);
   }
else
   DeltaZ=DeltaZValue;
}

/*
void OptFiberLocalError::SetNumStepsBetweenScatterings(int Value)
{
if (Value>0)
 {
   NumStepsBetweenScatterings=Value;
}
*/














