// $Id: ocsOptFilter.cc,v 1.3 2008/08/08 02:26:53 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptFilter.hh"

extern ofstream *LogFile;


// ################### Methods for the class OptFilter

// #### Constructor of the class OptFilter
// ## By: Ivan Lima
OptFilter::OptFilter()
{
   ObjectInitialized = 0;
}

// ##########################################################

// #### Initializes the object OptGaussFilter
// ## By: Ivan Lima
void OptFilter::InitOptFilter(cfftw *fft2, fftMatrices sfftM2
                                      ,OptFilterParam soOptFilter)
{
   fft = fft2;
   sfftM = sfftM2; 
   TypeOptFilter = soOptFilter.TypeOptFilter;
   OrderOptFilter = soOptFilter.OrderOptFilter;
   FreqFWHM_OptFilter = soOptFilter.FreqFWHM_OptFilter;
   RelativeCenterFreqOptFilter = soOptFilter.CenterFreqOptFilter;

   // FIX!! The RelativeCenterFreqOptFilter should be set to be 
   // RelativeCenterFreqOptFilter = soOptFilter.CenterFreqOptFilter - 
   //                         oOptSignal.CenterFreq
   // This is only an issue if are using the method      


   ObjectInitialized = 1;
}

// ########################################################

//! Filters optical signal

/*! This method was changed by JZ, April 2003.
    It now calls TunableFilterOptSignal()
    The effect is same as the previous version 
    which is now in FilterOptSignal_OLD()
    except that if a SUPERGAUSS filter is centered at
    RelativeCenterFreqOptFilter rather than at 0.
*/

void OptFilter::FilterOptSignal(void)
{
  TunableFilterOptSignal();
}

// ####################################################

void OptFilter::FilterOptSignal_OLD()
{
  // Created by John Zweck Aug 8th 2001
  // by adapting FilterOptSignal() method written by Ivan Lima


   if (! ObjectInitialized) {
      cerr << "Error: Object of class OptGaussFilter not initialized" 
           << endl;
      exit(1);
   }   


   switch(TypeOptFilter)
     {

     case NOFILTER:
       break;

     case SUPERGAUSS:

       if (OrderOptFilter == 1 )
	 {
          // ## Gaussian filter
      
          // ## The input is in freq, but the time is updated by optical_filter
          // ## It is not using the order and the central frequency yet.

        optical_filter(sfftM.tOptSignalX, sfftM.fOptSignalX, fft
                     ,FreqFWHM_OptFilter);


        if(TypeSimulation == VECTOR)
           optical_filter(sfftM.tOptSignalY, sfftM.fOptSignalY, fft
                         ,FreqFWHM_OptFilter);
            
	 }
        else
	  {
            cerr << "OptFilter::FilterOptSignal" << endl 
	        << "Error: Gaussian OptFilter of OrderOptFilter " 
                << OrderOptFilter 
                << " not implemented."      << endl;
            exit(1);  
	  }

        break;

     
    case CIENA_DEMUX:
       // ## Filter used in the WDM loop experiment

       DemuxLoopTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);


       if(TypeSimulation == VECTOR)
        DemuxLoopTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);
      
       break;


     case SANTEC:
       // ## Santec filter from Gary's Lab

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);


       if(TypeSimulation == VECTOR)
        SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       break;

     case BOTH_SANTEC:

       OrderOptFilter = 4;

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);


       if(TypeSimulation == VECTOR)
        SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       OrderOptFilter = 12;

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

       if(TypeSimulation == VECTOR)
        SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

     default:                
       cerr << "OptFilter::FilterOptSignal" << endl 
	    << "Error: OptFilter of TypeOptFilter " << TypeOptFilter
           << " Or OrderOptFilter " << OrderOptFilter 
           << " not implemented."      << endl;
      exit(1);   
   }    
              
}// end FilterOptSignal

// ######################################################

OptFilter::OptFilter(string InFileName, OptSignal *oOptSignal2)
{

  // Created by John Zweck 24 Oct, 2000
  //## You should update the copy constructor if you change this (IL)

  ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
 
   oOptSignal = oOptSignal2;
 
   DebugLevel = oOptSignal->GetDebugLevel();

   if(DebugLevel)
     {
      LogFileSeparator();
      *LogFile << InFileName << endl << endl;
     }

   TypeOptFilter = typeOptFilter(
                   ReadInt("TypeOptFilter"," ",0,6,LOWER_AND_UPPER,&InFile,
			   DebugLevel));

   OrderOptFilter = ReadInt("OrderOptFilter"," ",1,12,
                              LOWER_AND_UPPER,&InFile,DebugLevel);


 if(DebugLevel)
  {
   switch(TypeOptFilter)
     {

       case NOFILTER: 
            *LogFile << "Not using an optical filter" << endl;
            break;

       case SUPERGAUSS: 
           *LogFile << "Using Gaussian Optical Filter of order " 
		   << OrderOptFilter << endl;
         
        break;

       case CIENA_DEMUX: 
           *LogFile << "Using Demux filter for WDM loop experiment"<< endl;
         
       break;

       case SANTEC:
	 
         switch(OrderOptFilter)
	   {
	   case 4:
             *LogFile << "Using Santec04 Filter in Gary's Lab" << endl;
             break;
           case 12:
             *LogFile << "Using Santec12 Filter in Gary's Lab" << endl;
             break;
   
	   default:
             cerr << "With Santec Filter OrderOptFilter must be 4 or 12" 
                  << endl
                  << "Aborting now!" << endl;
	     exit(1);
	   } //end switch Order

      case BOTH_SANTEC:

      *LogFile << "Using both Santec Optical Filters" << endl;

      break;

      case BANDPASS: 
	 
	   *LogFile << "Using Optical Bandpass Filter" << endl;
      break;

      case FABRY_PEROT: 
	 
	   *LogFile << "Using Fabry-Perot Filter" << endl;
      break;

      default:
         cerr << " Erroneous choice of optical filter" << endl;


     }// end switch Type
  }

   FreqFWHM_OptFilter = ReadDouble("FreqFWHM_OptFilter","Hz",0,-1,
                                    LOWER_ONLY,&InFile,DebugLevel);

   RelativeCenterFreqOptFilter = ReadDouble("RelativeCenterFreq",
                                      "Hz",0,-1,NO_BOUNDS,&InFile,DebugLevel);

   InFile.close();

   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   TypeSimulation = oOptSignal->GetTypeSimulation();

   ObjectInitialized = 1;      

} // ## end OptFilter constructor


//######################################################
//##  Copy constructor for the optical filter
//## By: Ivan Lima (5/3/02)   (adapted from the counstructor by John Zweck)

OptFilter::OptFilter(OptFilter *oOrigOptFilter, OptSignal *oOptSignal2)
{
  //## You should update the default constructor if you change this (IL)
   
   LogFileSeparator();
   *LogFile << "Initializing OptFilter object from copy constructor" 
           << endl << endl;
 
   oOptSignal = oOptSignal2;
 
   TypeOptFilter =  oOrigOptFilter->GetTypeOptFilter();
   OrderOptFilter = oOrigOptFilter->GetOrderOptFilter();

   if(DebugLevel)
     {
   switch(TypeOptFilter)
     {
       case NOFILTER:
        *LogFile << "Not using an optical filter" << endl;
        break;

       case SUPERGAUSS: 
        *LogFile << "Using Gaussian Optical Filter of order " 
		<< OrderOptFilter << endl;
        break;

      case CIENA_DEMUX:

      *LogFile << "Using demux filter for the WDM loop experiment" << endl;

      break;

       case SANTEC:
	 
         switch(OrderOptFilter)
	   {
	   case 4:
             *LogFile << "Using Santec04 Filter in Gary's Lab" << endl;
             break;
           case 12:
             *LogFile << "Using Santec12 Filter in Gary's Lab" << endl;
             break;
   
	   default:
             cerr << "With Santec Filter OrderOptFilter must be 4 or 12" 
                  << endl
                  << "Aborting now!" << endl;
	     exit(1);
	   } //end switch Order
          
      case BOTH_SANTEC:

      *LogFile << "Using both Santec Optical Filters" << endl;

      break;


      case BANDPASS: 

	   *LogFile << "Using Optical Bandpass Filter" << endl;
       break;

     case FABRY_PEROT: 

	   *LogFile << "Using Fabry Perot Filter" << endl;
       break;
	  
      default:
         cerr << " Erroneous choice of optical filter" << endl;
     }// end switch Type
     }

   FreqFWHM_OptFilter = oOrigOptFilter->GetFreqFWHM_OptFilter();
   RelativeCenterFreqOptFilter =   
         oOrigOptFilter->GetRelativeCenterFreqOptFilter();


   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   TypeSimulation = oOptSignal->GetTypeSimulation();

   ObjectInitialized = 1;      

} // ## end OptFilter copy constructor


// #########################################################

OptFilter::OptFilter(string InFileName,typeSimulation TypeSimulation2)
{

  TypeSimulation = TypeSimulation2;

  // Created by John Zweck 25 Oct, 2000
  // Tunable Optical Filter constructor

  ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;
 

   TypeOptFilter = typeOptFilter(
                   ReadInt("TypeOptFilter"," ",0,5,LOWER_AND_UPPER,&InFile));

   OrderOptFilter = ReadInt("OrderOptFilter"," ",1,-1,
                              LOWER_ONLY,&InFile);


   if(DebugLevel)
     {
   switch(TypeOptFilter)
     {

       case NOFILTER: 
            *LogFile << "Not using an optical filter" << endl;
             break;
   
       case SUPERGAUSS: 
	 
            *LogFile << "Using SuperGaussian Optical Filter of order" 
		    << OrderOptFilter << endl;
	   
        break;
       case SANTEC:
	 
         switch(OrderOptFilter)
	   {
	   case 4:
             *LogFile << "Using Santec04 Filter in Gary's Lab" << endl;
             break;
           case 12:
             *LogFile << "Using Santec12 Filter in Gary's Lab" << endl;
             break;
   
	   default:
             cerr << "With Santec Filter OrderOptFilter must be 4 or 12" 
                  << endl
                  << "Aborting now!" << endl;
	     exit(1);
	   } //end switch Order
          
     case BOTH_SANTEC:

      *LogFile << "Using both Santec Optical Filters" << endl;

      break;

     case BANDPASS: 

	   *LogFile << "Using Optical Bandpass Filter" << endl;
       break;
     
	  case FABRY_PEROT: 

	   *LogFile << "Using Fabry Perot Filter" << endl;
       break;
	  
      default:
       cerr << " Erroneous choice of optical filter" << endl;




     }// end switch Type
     }

   FreqFWHM_OptFilter = ReadDouble("FreqFWHM_OptFilter","Hz",0,-1,
                                    LOWER_ONLY,&InFile);

      
  DoAdditionalFilteringFlag = ReadInt("DoAdditionalFiltering"," ",0,1,
                                    LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,
                                    &InFile);

   InFile.close();

   ObjectInitialized = 0;      


} // ## end OptFilter constructor


// ######################################################


OptFilter::OptFilter(int TypeOptFilter2,int OrderOptFilter2,
                      double FreqFWHM_OptFilter2, 
		     typeSimulation TypeSimulation2)
{
  // Constructor for tunable optical filter designed to be used in 
  // conjunction with OptDemuxer (Added by JZ 16th Oct 2000)

  TypeOptFilter =  typeOptFilter(TypeOptFilter2);
  OrderOptFilter = OrderOptFilter2;
  FreqFWHM_OptFilter = FreqFWHM_OptFilter2;
  ObjectInitialized = 0;
  TypeSimulation = TypeSimulation2;

}


// #####################################################

void OptFilter::SetRelativeCenterFreq(double RelativeCenterFreqOptFilter2) 
  { 

    // Added by JZ 16th Oct 2000
    // To be used in conjunction with with OptDemuxer

   RelativeCenterFreqOptFilter =  RelativeCenterFreqOptFilter2;
   

   //*LogFile << "RelativeCenterFreqOptFilter = " << 
   //            RelativeCenterFreqOptFilter << endl;

  } 


// ####################################################

void OptFilter::SetOpticalData(cfftw * fft2, fftMatrices sfftM2)
   {

    // Added by JZ 16th Oct 2000
    // To be used in conjunction with with OptDemuxer

     fft = fft2;
     sfftM = sfftM2;
     ObjectInitialized = 1;
     
   }


// ################################################

void OptFilter::GaussianTunableOpticalFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{

  // Created by John Zweck Oct 19th 2000
  // Adapted from Ron's optical_filter in ocsTools
  // The new feature is to incorporate the central frequency
 
  double sigma =  FreqFWHM_OptFilter / 2.35482 * M_SQRT2;

  // Ron says this choice of sigma is used because fwhm_norm is for 
  // intensity of gaussian filter not for the gaussian filter itself 
  // I believe him.

  const int N = fft->give_N();

  // Note: fft->freq(i) is frequency relative to the central frequency 
  // of the signal used to originally define the fft object measured in Hertz

  // FIX: If we end up using Optical Filters with RelativeCenterFreqOptFilter
  // close to the end of the FrequencyWindow we will need to make sure that
  // fft->freq(i) - RelativeCenterFreqOptFilter is "wrapped" (JZ)

  for(int i=0; i < N; i++)
    {
      FreqSignal[i] *= exp(-0.5*sq(
                          (fft->freq(i) - RelativeCenterFreqOptFilter)
                                     /sigma));

      // Both these frequencies are relative to the Central Frequency of
      // the signal.

    }

  fft->IFFT(FreqSignal,TimeSignal);

} // end GaussianTunableOpticalFilter

// ###############################################

void OptFilter::SuperGaussianTunableOpticalFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{
	// Created by Oleg Sinkin Feb.25 2003
	// Adapted from John's GaussianTunableOpticalFilter
	// H(f)=exp(-1/2*((f-fc)/sigma)^2n) -- frequency response function
	// where fc is the center frequency, 
        // sigma is a width parameter related to
	//	FWHM by sigma=0.5*FWHM/( (ln(2))^(1/2n) )
	// and n is the order of super Gaussian function, n=1 corresponds to
	// standard Gaussian

	int n =OrderOptFilter;
	double sigma =  0.5*FreqFWHM_OptFilter /pow(0.693147181,0.5/n);

	// fwhm_norm is for intensity of the filter |H(f)|^2
	// not for the gaussian filter itself 

	const int N = fft->give_N();

	// Note: fft->freq(i) is frequency relative to the central frequency 
	// of the signal used to originally define the fft object 
	//measured in Hertz

	// FIX: If we end up using Optical Filters with 
	//RelativeCenterFreqOptFilter
	// close to the end of the FrequencyWindow we will 
	// need to make sure that
	// fft->freq(i) - RelativeCenterFreqOptFilter is "wrapped" (JZ)

	for(int i=0; i < N; i++)
	{
		FreqSignal[i] *= exp(-0.5*
		pow( (fft->freq(i) - RelativeCenterFreqOptFilter)/sigma, 2*n));
		
           // Both these frequencies are relative to the Central Frequency of
	   // the signal.
	}

	fft->IFFT(FreqSignal,TimeSignal);

} // end SuperGaussianTunableOpticalFilter

// ################################################

void OptFilter::OpticalBandpassFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{
// Added by Anshul June '03.
// assuming FreqFWHM_OptFilter to be Bandwidth of the bandpass filter.
  
  const int N = fft->give_N();

// For some unknown reason ,this doesn't work for the 
//case when relativecentrefreq = 0
// e.g if  Bw = 60e9 , one would expect H(f) = 1.0 for -30e9 
// but this code results in H(f) = 0.0 for -30e9. 


  for(int i=0; i < N; i++)
    {
      if( (( fft->freq(i) - RelativeCenterFreqOptFilter) >= 
	   -FreqFWHM_OptFilter/2)&&
          (( fft->freq(i) - RelativeCenterFreqOptFilter) <  
	   FreqFWHM_OptFilter/2)  )
      
         FreqSignal[i] *= 1.0;
      else
      {
         if (RelativeCenterFreqOptFilter == 0.0 ){
   
             if ( ((fft->freq(i) - 1) < -FreqFWHM_OptFilter/2)&&
                  ((fft->freq(i) + 1) > -FreqFWHM_OptFilter/2)  )
         
                  FreqSignal[i] *= 1.0;
             else
                  FreqSignal[i] *= 0.0;			    
	 }
         else 
            FreqSignal[i] *= 0.0;			    
      }
    }

 
  fft->IFFT(FreqSignal,TimeSignal);

} // end OpticalBandpassFilter

// ################################################
// ################################################

void OptFilter::FabryPerotFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{
// Added by Anshul  Apr '04.
//	3 mirror fabry perot filter.
//	Reference : PTL vol 8 no 9 Sep 96. Freire, Carvalho and da Silva 

//R0 and R1 are the reflectances of the center and the outer mirrors
//T0 and T1 are the corresponding transmittances.
// by definiton R0 + T0 = 1 , R1 + T1 = 1.

/*IMPORTANT :Choosing  R1 = 0.8 results in the impulse response 
 of the filter to be a second order butterworth type*/

  const double R1 =  0.8; 
  double R0 = (4*R1)/(sq(1 + R1));   // eqn 3 in paper
  double T0 = 1 - R0;
  double T1 = 1 - R1;

// This is required since in the paper all equations are in terms of omega
// where omega = 2*pi*freq; so FWHM changes by a factor of 2*pi.
  FreqFWHM_OptFilter = FreqFWHM_OptFilter/(2*pi); 

  double beta = 2*T0*(sq(T1)) + R1*(2 - R1- (4*R0)) - 1; //eqn 5
  double FSR = (FreqFWHM_OptFilter*pi)/acos(1 - sqrt(1 + beta/(4*R1))); //eqn 4

  double numerator = sqrt(T0)*T1;

  const int N = fft->give_N();
  for(int i=0; i < N; i++)
	{
		cplx denominator = 1.0-2*sqrt(R0*R1)*
		  exp((-jc*(fft->freq(i)))/FSR)
		                 +R1*exp((-2*jc*(fft->freq(i)))/FSR);
		
                    
		FreqSignal[i] *= numerator/denominator;   // eqn 2 
		
	}

	fft->IFFT(FreqSignal,TimeSignal);

} // end FabryPerotFilter

// ##############################################

void OptFilter::TunableFilterOptSignal()
{

  // Created by John Zweck 20 Oct 2000
  // by adapting FilterOptSignal()method written by Ivan Lima


   if (! ObjectInitialized) {
      cerr << "Error: Object of class OptGaussFilter not initialized" 
           << endl;
      exit(1);
   }   

   switch(TypeOptFilter)
     {

     case NOFILTER:

       break;

     case SUPERGAUSS:
        
       if(OrderOptFilter==1)// ## Gaussian filter
	 {

          GaussianTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

          if(TypeSimulation == VECTOR)
           GaussianTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);
	 }
        else // ## SuperGaussian filter
	  {
           SuperGaussianTunableOpticalFilter(sfftM.tOptSignalX, 
					     sfftM.fOptSignalX);

           if(TypeSimulation == VECTOR)
            SuperGaussianTunableOpticalFilter(sfftM.tOptSignalY, 
					      sfftM.fOptSignalY);


	  }

        break;


     case CIENA_DEMUX:
       // ## Demux filter used for Jiping's experiment

       DemuxLoopTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

       if(TypeSimulation == VECTOR)
          DemuxLoopTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       break;
     
     case SANTEC:
       // ## Santec filter from Gary's Lab

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

       if(TypeSimulation == VECTOR)
          SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       break;

     case BOTH_SANTEC:
       OrderOptFilter = 4;

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

       if(TypeSimulation == VECTOR)
          SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       OrderOptFilter = 12;

       SantecTunableOpticalFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

       if(TypeSimulation == VECTOR)
          SantecTunableOpticalFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);

       break;


    
	  case BANDPASS: // Added by Anshul June'03.
     
       OpticalBandpassFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

          if(TypeSimulation == VECTOR)
             OpticalBandpassFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);
    
       break;

     case FABRY_PEROT: // Added by Anshul Apr'04.
     
       FabryPerotFilter(sfftM.tOptSignalX, sfftM.fOptSignalX);

          if(TypeSimulation == VECTOR)
            FabryPerotFilter(sfftM.tOptSignalY, sfftM.fOptSignalY);
       break;



     default:                
       cerr << "OptFilter::TunableFilterOptSignal" << endl 
	    << "Error: OptFilter of TypeOptFilter " << TypeOptFilter
           << " Or OrderOptFilter " << OrderOptFilter 
           << " not implemented."      << endl;
      exit(1);   
   }    
              
}// end TunableFilterOptSignal

// ###################################################

void OptFilter::WriteFreqPowerGaussianFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }


   double sigma =  FreqFWHM_OptFilter / 2.35482 * M_SQRT2;

  // Ron says this choice of sigma is used because fwhm_norm is for 
  // intensity of gaussian filter not for the gaussian filter itself 
  // I believe him.

  const int N = fft->give_N();

   double freq;

 for(int ii= 0; ii < N; ii++)
 {

     freq = fft->freq(ii);
   
     OutFile << freq << " " <<
           sq(abs(exp(-0.5*sq((freq - RelativeCenterFreqOptFilter)
			      /sigma))))
                                 << endl;
  }

   OutFile.close();

} // ## end WriteFreqPowerGaussianFilter

// ##########################################################

void OptFilter::WriteFreqPowerSuperGaussianFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }
	int n =OrderOptFilter;

	double sigma =  0.5*FreqFWHM_OptFilter /pow(0.693147181,0.5/n);
	cout<<"sigma SG= "<<sigma<<endl;
	const int N = fft->give_N();

   double freq;

 for(int ii= 0; ii < N; ii++)
   {

     freq = fft->freq(ii);
     // next line: factor of 0.5 is replaced by 1.0 due to squaring of H(f)
     OutFile << freq << "\t" 
             << sq(abs  (exp(-0.5*
	  pow( (fft->freq(ii) - RelativeCenterFreqOptFilter)/sigma, 2*n)) ))
             << endl;
  }

   OutFile.close();

} // ## end WriteFreqPowerSuperGaussianFilter

// ##########################################################

void OptFilter::WriteFreqPowerBandpassFilter(string OutFileName)
{
    // Added by Anshul June'03.
    
    ofstream OutFile(OutFileName.c_str(),ios::out);
    if (!OutFile) 
      {
         cerr << "Error: " << OutFileName << "could not be created."<<endl;
         exit(1);
      }

    const int N = fft->give_N();

    double freq;

    OutFile << "% frequency   |H(f)|^2" << endl; 
      

// For some unknown reason ,this doesnt work 
//for the case when relativecentrefreq = 0
// e.g if  Bw = 60e9 , one would expect H(f) = 1.0 for -30e9 
// but this code results in H(f) = 0.0 for -30e9. 
 
  for(int i=0; i < N; i++)
    {
      freq = fft->freq(i);

      if( (( freq - RelativeCenterFreqOptFilter) >= -FreqFWHM_OptFilter/2)&&
          (( freq - RelativeCenterFreqOptFilter) <   FreqFWHM_OptFilter/2)  )
               
	    OutFile << freq << '\t'  << 1.0 << endl;
      else
      {
         if (RelativeCenterFreqOptFilter == 0.0 ){
   
             if ( ((freq - 1) < -FreqFWHM_OptFilter/2)&&
                  ((freq + 1) > -FreqFWHM_OptFilter/2)  )
                     
	          OutFile << freq << '\t'  << 1.0 << endl;
             else
                  OutFile << freq << '\t'  << 0.0 << endl;
	 }
         else 
             OutFile << freq << '\t'  << 0.0 << endl;
      }
    }


    OutFile.close();

} // ## end WriteFreqPowerBandpassFilter

// #########################################################

void OptFilter::WriteFreqPowerFabryPerotFilter(string OutFileName)
{
// Added by Anshul Apr'04.
//	3 mirror fabry perot filter.
//	Reference : PTL vol 8 no 9 Sep 96. Freire, Carvalho and da Silva 

  ofstream OutFile(OutFileName.c_str(),ios::out);
    if (!OutFile) 
      {
         cerr << "Error: " << OutFileName << "could not be created."<<endl;
         exit(1);
      }
  OutFile << "% frequency   |H(f)|^2" << endl; 

//R0 and R1 are the reflectances of the center and the outer mirrors
//T0 and T1 are the corresponding transmittances.
// by definiton R0 + T0 = 1 , R1 + T1 = 1.

/*IMPORTANT :Choosing  R1 = 0.8 results in the impulse response 
 of the filter to be a second order butterworth type*/

  const double R1 =  0.8; 
  double R0 = (4*R1)/(sq(1 + R1));   // eqn 3 in paper
  double T0 = 1 - R0;
  double T1 = 1 - R1;

// This is required since in the paper all equations are in terms of omega
// where omega = 2*pi*freq; so FWHM changes by a factor of 2*pi.
  FreqFWHM_OptFilter = FreqFWHM_OptFilter/(2*pi); 

  double beta = 2*T0*(sq(T1)) + R1*(2 - R1- (4*R0)) - 1; //eqn 5
  double FSR = (FreqFWHM_OptFilter*pi)/acos(1 - sqrt(1 + beta/(4*R1))); //eqn 4

  double numerator = sqrt(T0)*T1;

  const int N = fft->give_N();
  double freq;
  
  for(int i=0; i < N; i++)
	{
	   freq = fft->freq(i);
	    
	   cplx denominator = 1.0-2*sqrt(R0*R1)*exp((-jc*freq)/FSR) 
		                  + R1*exp((-2*jc*freq)/FSR);
		
      OutFile << freq << '\t' << sq(abs(numerator/denominator)) << endl;
		
	}

   OutFile.close();

} // ## end WriteFreqPowerFabryPerotFilter


// #########################################################

void OptFilter::WriteFreqPowerSantecFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }


  // Ron says this choice of sigma is used because fwhm_norm is for 
  // intensity of gaussian filter not for the gaussian filter itself 
  // I believe him.

   const int N = fft->give_N();

   double freq;

   double SantecFilterValue;

 for(int ii= 0; ii < N; ii++)
 {

   freq = 1e-12*fft->freq(ii); // Put into THz

     if(ii==0) // freq=0
       SantecFilterValue = 1;
     else
       SantecFilterValue = dB2Linear(-exp(SantecPoly(log(abs(freq))))); 
   
     OutFile << freq << " " << SantecFilterValue << endl;
  }

   OutFile.close();

} // ## end WriteFreqPowerSantecFilter

 // #########################################################

double OptFilter::SantecPoly(double x)
{

  if(OrderOptFilter==4)
    return 0.0633*pow(x,3) + 0.1566*pow(x,2) + 0.4707*x+3.5069;
  if(OrderOptFilter==12)
    return 0.0455*pow(x,3) - 0.0596*pow(x,2) + 0.3934*x+3.063;

} // ## end SantecPoly

// ############################################################

void OptFilter::SantecTunableOpticalFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{

  // Created by John Zweck Aug 1st 2001

  // The Santec filters are optical WDM demuxing filters 
  // in Gary's Lab. I fitted this filter using a Transfer function
  // provided to me by Hai Xu. There are two filters, whose orders are
  // 4 and 12. The Santec4 filter has FWHM = 0.545 nm = 67.656 GHz
  // but you don't need to put the FWHM in the input file.
  // The Santec12 filter is somewhat wider.
  // The fit is very good. I can show you the fitted data if you like. (JZ)

  const int N = fft->give_N();

  // Note: fft->freq(i) is frequency relative to the central frequency 
  // of the signal used to originally define the fft object measured in Hertz

  // FIX: If we end up using Optical Filters with RelativeCenterFreqOptFilter
  // close to the end of the FrequencyWindow we will need to make sure that
  // fft->freq(i) - RelativeCenterFreqOptFilter is "wrapped" (JZ)

  double RelativeFreq;
  double SantecFilterValue;

  for(int ii=0; ii < N; ii++)
    {
      // For Santec filter need Freq in THz
      RelativeFreq = 1e-12*(fft->freq(ii) - RelativeCenterFreqOptFilter);

     if(RelativeFreq==0) 
       SantecFilterValue = 1;
     else
       SantecFilterValue = dB2Linear(-exp(SantecPoly(log(
                                      abs(RelativeFreq))))); 
      
     FreqSignal[ii] *= SantecFilterValue;

      // Both these frequencies are relative to the Central Frequency of
      // the signal.

    }

  fft->IFFT(FreqSignal,TimeSignal);

} // end SantecTunableOpticalFilter

// ###################################################################
//##
//# added by Paola Griggio 03/17/04 to emulate the Demux filter in the
//# WDM loop experiment.

void OptFilter::WriteFreqPowerDemuxLoopFilter(string OutFileName)
{

 ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
     }

   const int N = fft->give_N();

   double freq;

   double DemuxLoopFilterValue;

 for(int ii= 0; ii < N; ii++)
 {

   freq = 1e-12*fft->freq(ii); // Put into THz

     if(ii==0) // freq=0
       DemuxLoopFilterValue = 1;
     else
       DemuxLoopFilterValue = dB2Linear(-exp(DemuxLoopPoly(log(abs(freq))))); 
   
     OutFile << freq << " " << DemuxLoopFilterValue << endl;
  }

   OutFile.close();

} // ## end WriteFreqPowerDemuxLoopFilter


//########################################################
//# filter to demux the channels in the WDM loop experiment
//# by Paola Griggio 03/17/04. See Jiping's data

double OptFilter::DemuxLoopPoly(double x)
{
 return 0.0011*pow(x,6)+ 0.0347*pow(x,5)+0.3987*pow(x,4)+2.1246*pow(x,3)+5.0023*pow(x,2)+5.0339*x+5.1905;
}

// ############################################################

void OptFilter::DemuxLoopTunableOpticalFilter
                (cplx * TimeSignal, cplx * FreqSignal)
{

  //  by Paola Griggio 03/17/04
  // The DemuxLoop filter is used in the WDM loop at 10 Gbit/s, 
  // to select the channels
  // I fitted this filter using a Transfer function
  // provided to me by Jiping Wen. 
  // To verify the fit, refer to matlab program : "Demux_Filter.m"

  const int N = fft->give_N();

  // Note: fft->freq(i) is frequency relative to the central frequency 
  // of the signal used to originally define the fft object measured in Hertz

  double RelativeFreq;
  double DemuxLoopFilterValue;

  for(int ii=0; ii < N; ii++)
    {
      // Freq in THz
      RelativeFreq = 1e-12*(fft->freq(ii) - RelativeCenterFreqOptFilter);

     if(RelativeFreq==0) 
       DemuxLoopFilterValue = 1;
     else
       DemuxLoopFilterValue = dB2Linear(-exp(DemuxLoopPoly(log(
                                      abs(RelativeFreq))))); 
      
     FreqSignal[ii] *= DemuxLoopFilterValue;

      // Both these frequencies are relative to the Central Frequency of
      // the signal.

    }

  fft->IFFT(FreqSignal,TimeSignal);

} // end DemuxLoopTunableOpticalFilter

//#######################################################
//## By: Ivan Lima (4/27/02)
void OptFilter::SetFreqFWHM_OptFilter(double FreqFWHM_OptFilter2)
{
   FreqFWHM_OptFilter = FreqFWHM_OptFilter2;

   LogFileSeparator();

   *LogFile << "WARNING: FreqFWHM_OptFilter reset to " 
           <<  FreqFWHM_OptFilter << endl << endl;
}



//####################################################################3
//## By: Ivan Lima (8/20/02)
void OptFilter::WriteFilterImpulseResponse(
                 string FileNameImpulseResponseTime,
                 string FileNameImpulseResponseFreq)
{
   OptSignal *oOptSignal2 = new OptSignal(*oOptSignal);
   //*oOptSignal2 = *oOptSignal;  

   oOptSignal->GenerateImpulseSignal(); 
   
   FilterOptSignal();
   

   oOptSignal->WriteFileTimePower(FileNameImpulseResponseTime);
   oOptSignal->WriteFileFreqPower(FileNameImpulseResponseFreq);  
   	
   *oOptSignal = *oOptSignal2;
   delete oOptSignal2;
}




//####################################################################3
//## The spectral width includes the negative frequencies
//## By: Ivan Lima (8/20/02)
double OptFilter::GetPowerEquivalentSpectralWidth(void)
{
   OptSignal *oOptSignal2 = new OptSignal(*oOptSignal);
   //*oOptSignal2 = *oOptSignal;  

   oOptSignal->GenerateImpulseSignal(); 
   
   FilterOptSignal();
   
   
   cplx OptFilterDC_Component = oOptSignal->sfftM.fOptSignalX[0];   
     
   int qtPoints = oOptSignal->GetQtPoints();  
   //##  Get the effective electrical filter bandwidth
   double PowerEquivalentSpectralWidth = 0;
   for (int ii = 0; ii < qtPoints; ii++) {
      PowerEquivalentSpectralWidth += 
            sq(abs(oOptSignal->sfftM.fOptSignalX[ii]));
   }
   PowerEquivalentSpectralWidth *= oOptSignal->GetDeltaFreq();

   //cout << "OptFilterDC_Component = " << OptFilterDC_Component << endl;
   //cout << "#$#$#$ Equivalent Spectral Width Optical Filter (GHz) = " 
   //     << PowerEquivalentSpectralWidth/abs(OptFilterDC_Component)*1e-9 << endl;        

   *oOptSignal = *oOptSignal2;
   delete oOptSignal2;
   return PowerEquivalentSpectralWidth
          /abs(OptFilterDC_Component);
}



//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/13/02)
void OptFilter::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   //oOptSignal = oOptSignal2;
   fft = oOptSignal2->fft;
   sfftM = oOptSignal2->sfftM;   
}

//####################################















