// $Id: ocsOptSpectrumAnalyzer.cc,v 1.2 2006/05/11 22:33:00 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "ocsOptSpectrumAnalyzer.hh"

extern ofstream *LogFile;

// ######################################################

OptSpectrumAnalyzer::OptSpectrumAnalyzer(string InFileName,string Job2,
                                         int AppendFlag2,
                                         OptSignal *oOptSignal2)
{

  // Written by John Zweck, Jan 15th 2001
 // IF YOU CHANGE THIS CONSTRUCTOR CHANGE THE OTHER ONE TOO 

oOptSignal = oOptSignal2;
Job = Job2;
AppendFlag = AppendFlag2;


LogFileSeparator();

*LogFile << "Job name for OptSpectrumAnalyzer = " << Job << endl;

ifstream InFile(InFileName.c_str(),ios::in);    
   
if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
*LogFile << InFileName << endl << endl;


ResolutionBandwidthHz = ReadDouble("ResolutionBandwidthHz","Hz",0,-1,
                                    LOWER_ONLY,&InFile);

ModeOSA = ReadInt("ModeOSA"," ",1,2,LOWER_AND_UPPER,&InFile);

*LogFile << "ResolutionBandwidthHz [GHz] = " << ResolutionBandwidthHz*1e-9
	 << endl;


SuperGaussExponent = ReadDouble("SuperGaussExponent"," ",0,-1,
                                    LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);


PowerMeterFlag = ReadInt("PowerMeterFlag"," ",0,1,
                           LOWER_AND_UPPER,OPTIONAL_WARNING,0,&InFile);

TypeAmplifierNoise = ReadInt("TypeAmplifierNoise"," ",0,4,
                           LOWER_AND_UPPER,OPTIONAL_NO_WARNING,2,&InFile);

 NumChannels = oOptSignal->GetNumChannels();
 qtPoints = oOptSignal->GetQtPoints();
 fft = oOptSignal->fft;
 DeltaFreq = oOptSignal->GetDeltaFreq();

 AllocateMemory(); //## Ivan Lima (01/06/20)

 InFile.close();

} // ## end constructor 

// ######################################################

OptSpectrumAnalyzer::OptSpectrumAnalyzer(OptSignal *oOptSignal2,
                                         double ResolutionBandwidthHz2,
                                         int ModeOSA2,
                                         double SuperGaussExponent2)
{

  // Written by John Zweck, April 10th 2001
  // For use in conjunction with FIXED_OUTPUT_POWER amplification
  // method of OptAmplifier

  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE THE OTHER ONE TOO 

 oOptSignal = oOptSignal2;

 ResolutionBandwidthHz = ResolutionBandwidthHz2;
                              
 ModeOSA = ModeOSA2;
 TypeAmplifierNoise = 2; // Default value
 SuperGaussExponent = SuperGaussExponent2;
 PowerMeterFlag = 1;

 NumChannels = oOptSignal->GetNumChannels();
 qtPoints = oOptSignal->GetQtPoints();
 fft = oOptSignal->fft;
 DeltaFreq = oOptSignal->GetDeltaFreq();

 AllocateMemory();

} // ## end 2nd constructor

// ##################################################
 
OptSpectrumAnalyzer::OptSpectrumAnalyzer(OptSignal *oOptSignal2,
                                         double ResolutionBandwidthHz2,
                                         int ModeOSA2,
                                         double SuperGaussExponent2,
					 int TypeAmplifierNoise2)
{

  // Written by John Zweck, April 10th 2001
  // For use in conjunction with FIXED_OUTPUT_POWER amplification
  // method of OptAmplifier

  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE THE OTHER ONE TOO 

 TypeAmplifierNoise = TypeAmplifierNoise2;

 oOptSignal = oOptSignal2;

 ResolutionBandwidthHz = ResolutionBandwidthHz2;
                              
 ModeOSA = ModeOSA2;
 SuperGaussExponent = SuperGaussExponent2;
 PowerMeterFlag = 1;

 NumChannels = oOptSignal->GetNumChannels();
 qtPoints = oOptSignal->GetQtPoints();
 fft = oOptSignal->fft;
 DeltaFreq = oOptSignal->GetDeltaFreq();

 AllocateMemory();
   
} // ## end 3rd constructor 

// ######################################################

OptSpectrumAnalyzer::~OptSpectrumAnalyzer()
{
   delete [] fPowerSignal;
   delete [] fOSA;
   delete [] tPowerSignal;
   delete [] tOSA;
   delete [] SignalPeakLin;
   delete [] NoiseAverageLin;
   
} // ## end ~OptSpectrumAnalyzer()

// #######################################################

//## By: Ivan Lima
void OptSpectrumAnalyzer::AllocateMemory(void)
{

   fPowerSignal = new cplx[qtPoints];
   fOSA = new cplx[qtPoints];
   tPowerSignal = new cplx[qtPoints];
   tOSA = new cplx[qtPoints];
 
   SignalPeakLin = new double[NumChannels];
   NoiseAverageLin = new double[NumChannels];

   for(int jj=0;jj<qtPoints;jj++) 
    {
      fPowerSignal[jj] = 0.0;
      fOSA[jj] = 0.0;
      tPowerSignal[jj] = 0.0;
      tOSA[jj] = 0.0;
   }

   for(int ch=0;ch<NumChannels;ch++) 
   {
     SignalPeakLin[ch] = 0.0;
     NoiseAverageLin[ch] = 0.0;
   }   

}

// ############################################################

void OptSpectrumAnalyzer::Update(double PropagatedLength2)
{

  switch(ModeOSA)
    {

     case 1:
      if(SuperGaussExponent > 2 || SuperGaussExponent < 2)
       {
	 //cout << "Using UpdateOSA_OLD as SuperGaussExponent != 2" << endl;
        UpdateOSA_OLD(PropagatedLength2);
       }
      else
       {
	 //cout << "Using UpdateOSA" << endl;
        UpdateOSA(PropagatedLength2);
       }
      break;

     case 2:
      UpdateOSA(PropagatedLength2);
      break;
    }

   UpdateSNR_OSA(PropagatedLength2);

} // ## end Update

// ################################################

void OptSpectrumAnalyzer::Output(int NumExpts)
{

   OutputOSA(NumExpts);
   OutputSNR_OSA(NumExpts);

} // ## end Output

//#################################################

double  OptSpectrumAnalyzer::ComputeAveragePowerFreqWindow(double LeftFreq, 
                                                           double RightFreq)
{

  int LeftIndex,RightIndex;

  
  int qtPoints = oOptSignal->GetQtPoints();

  if(RightFreq < LeftFreq)
    {
      cout << "Error in ComputeAveragePowerFreqWindow" << endl;          
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;         
      exit(1);
    }

  if(LeftFreq >= 0)
    {
      LeftIndex = int(LeftFreq/DeltaFreq);
    }     
  else
    {
      LeftIndex = int(LeftFreq/DeltaFreq) +  qtPoints;
    }

  // *LogFile << "LeftFreq [GHz] = " << LeftFreq*1e-9 << endl;
  //*LogFile << "LeftIndex = " << LeftIndex << endl;
  //*LogFile << "Freq(LeftIndex) [GHz] = " << oOptSignal->fft->freq(LeftIndex) 
  //  << endl;


  if(LeftIndex < 0 || LeftIndex > qtPoints)
    {
      cout << "OptSpectrumAnalyzer::ComputeAveragePowerFreqWindow:" 
	   << "LeftIndex error (1) !" << endl;
      cout << "LeftIndex = " << LeftIndex << endl;  
      cout << "RightIndex = " << RightIndex << endl;           
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;        
      exit(1);
    }

  if(RightFreq >= 0)
    {
      RightIndex = int(RightFreq/DeltaFreq);
    }     
  else
    {
      RightIndex = int(RightFreq/DeltaFreq) +  qtPoints;
    }

  //*LogFile << "RightFreq [GHz] = " << RightFreq*1e-9 << endl;
  //*LogFile << "RightIndex = " << RightIndex << endl << flush;
  //*LogFile << "Freq(RightIndex) [GHz] = "
  // << oOptSignal->fft->freq(RightIndex) 
  //  << endl;



  if(RightIndex < 0 || RightIndex > qtPoints)
    {
      cout <<  "OptSpectrumAnalyzer::ComputeAveragePowerFreqWindow:"
	   << "RightIndex error (1) !" << endl;
      cout << "LeftIndex = " << LeftIndex << endl;  
      cout << "RightIndex = " << RightIndex << endl;           
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;               
      exit(1);
    }


  double Average = 0.0;

  if(LeftIndex <= RightIndex)
    {

      if(oOptSignal->GetTypeSimulation() == VECTOR)
      {
       for(int ii=LeftIndex; ii < RightIndex; ii++)
	Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii])) 
	         + sq(abs(oOptSignal->sfftM.fOptSignalY[ii])); 
      }
     else
      {
       for(int ii=LeftIndex; ii < RightIndex; ii++)
        Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii]));
      }
  
 
    }
  else
    {
      int ii;

      if(oOptSignal->GetTypeSimulation() == VECTOR)
      {
       for(ii=LeftIndex;ii<qtPoints;ii++)
          Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii])) 
	         + sq(abs(oOptSignal->sfftM.fOptSignalY[ii])); 

       for(ii=0;ii<RightIndex;ii++)
          Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii])) 
	         + sq(abs(oOptSignal->sfftM.fOptSignalY[ii]));   
      }
     else
      {
       for(ii=LeftIndex;ii<qtPoints;ii++)
          Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii]));

       for(ii=0;ii<RightIndex;ii++)
          Average += sq(abs(oOptSignal->sfftM.fOptSignalX[ii]));
      }
     
    }


  //*LogFile << "Power sum [W] = " << Average << endl;

  //Average = Average*DeltaFreq/(RightFreq-LeftFreq);

  return Average;


} // ## end ComputeAveragePowerFreqWindow

// ##############################################
// ## By: John Zweck and Ivan Lima (ModeOSA == 2)

void OptSpectrumAnalyzer::UpdateOSA(double PropagatedLength2)
{

  if(TypeAmplifierNoise == 4)
    Clear();

  PropagatedLength = PropagatedLength2;

   for(int jj=0;jj<qtPoints;jj++)
    {
      if(oOptSignal->GetTypeSimulation() == VECTOR)
      {
       fPowerSignal[jj] = sq(abs(oOptSignal->sfftM.fOptSignalX[jj])) 
	               + sq(abs(oOptSignal->sfftM.fOptSignalY[jj]));
      }
      else
      {
       fPowerSignal[jj] = sq(abs(oOptSignal->sfftM.fOptSignalX[jj]));
      }


      if(TypeAmplifierNoise == 4)
         fPowerSignal[jj] += 
	  oOptSignal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()
	  *DeltaFreq;
  

    }

   /****************

   //## Uncomment these lines to see what the "transfer function" is
   for(int jj=0;jj<qtPoints;jj++)
    {
      fPowerSignal[jj] = 0.0; 
    }


    for(int jj=0;jj<1;jj++)
      fPowerSignal[jj] = 1.0; 
    ***************/

  fft->IFFT(fPowerSignal,tPowerSignal);  // Freq2Time
  fft->IFFT(fOSA,tOSA);  // Freq2Time
 
    double Gaussian_sigma_freq = 2*pi*ResolutionBandwidthHz/
                          (2*pow(2*log(2),1.0/(SuperGaussExponent)));
 
  // The Gaussian filters a power signal directly
  // so we want exp(- sq(FWHM/2) / 2 sq(sigma) ) = 0.5
  // which when solved for sigma gives the above equation.
  // I checked this by looking at output when fPowerSignal = Dirac.
   // You need to 2*pi factor to convert from Hz to radians.   

   // For speed we implement the convolution with a Gaussian using
   // the convolution theorem.

   double FilterEnvelopTime;
   double time;
   
   //cout << "ModeOSA = " << ModeOSA << endl;   
   
   for(int jj=0;jj<qtPoints;jj++) {

       time = jj*oOptSignal->GetDeltaTime();
       if(time >= 0.5*oOptSignal->GetTimeWindow()) {
	 time -= oOptSignal->GetTimeWindow();
       }  

       switch(ModeOSA) {
          case 1:
             FilterEnvelopTime = Gaussian_sigma_freq*
                                    oOptSignal->GetTimeWindow()/
                                    (sqrt(2*pi))*
                            exp(-pow(time*Gaussian_sigma_freq,2)/2);
             break;               
          case 2:
             if (time != 0) {
                FilterEnvelopTime = oOptSignal->GetTimeWindow()
                    *ResolutionBandwidthHz
                    *sin(time*pi*ResolutionBandwidthHz)
                    /(time*pi*ResolutionBandwidthHz);
             }
             else {
                FilterEnvelopTime = oOptSignal->GetTimeWindow()
                    *ResolutionBandwidthHz; 
             }   
             break;
       }
       
       tOSA[jj] += FilterEnvelopTime*tPowerSignal[jj];

   }
   

/*******  
   for(int jj=0;jj<qtPoints;jj++)
     {

       double time = jj*oOptSignal->GetDeltaTime();

       if(time >= 0.5*oOptSignal->GetTimeWindow())
	 time -= oOptSignal->GetTimeWindow();

       double GaussianFactorTime = Gaussian_sigma_freq*
                                    oOptSignal->GetTimeWindow()/
                                    (sqrt(2*pi))*
                            exp(-pow(time*Gaussian_sigma_freq,2)/2);

       tOSA[jj] += GaussianFactorTime*tPowerSignal[jj];

     }
*******/

    fft->FFT(tOSA,fOSA);  // Time2Freq

} // end UpdateOSA

// #####################################################

void OptSpectrumAnalyzer::UpdateOSA_OLD(double PropagatedLength2)
{

  if(TypeAmplifierNoise == 4)
    Clear();

  PropagatedLength = PropagatedLength2;

  for(int jj=0;jj<qtPoints;jj++)
   {
      if(oOptSignal->GetTypeSimulation() == VECTOR)
      {
      fPowerSignal[jj] = sq(abs(oOptSignal->sfftM.fOptSignalX[jj])) 
	               + sq(abs(oOptSignal->sfftM.fOptSignalY[jj])); 
      }
      else
      {
      fPowerSignal[jj] = sq(abs(oOptSignal->sfftM.fOptSignalX[jj]));
      }



      if(TypeAmplifierNoise == 4)
        fPowerSignal[jj] +=  
	  oOptSignal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()
	  *DeltaFreq;
 

    }

   /*****************
   for(int jj=0;jj<qtPoints;jj++)
    {
      fPowerSignal[jj] = 0.0; 
    }
   

    for(int jj=0;jj<1;jj++)
      fPowerSignal[jj] = 1.0; 
  
     ****************/
 
  fft->fftshift(fPowerSignal);
  fft->fftshift(fOSA);

  double Gaussian_sigma = ResolutionBandwidthHz/
                          (2*pow(2*log(2),1.0/(SuperGaussExponent)));
 
  // The Gaussian filters a power signal directly
  // so we want exp(- sq(FWHM/2) / 2 sq(sigma) ) = 0.5
  // which when solved for sigma gives the above equation.
  // I checked this by looking at output when fPowerSignal = Dirac.


  double omega,eta;
  double eta_min,eta_max;

  double LeftRelFreq = -0.5*oOptSignal->GetFreqWindow();

  double RightRelFreq = 0.5*oOptSignal->GetFreqWindow();

  int LeftIndex,RightIndex;
  double GaussianFactor;
  double FreqDiff;

  for(int jj=0;jj<qtPoints;jj++)
    {

      // Compute omega(jj)

      omega = jj*DeltaFreq - oOptSignal->GetFreqWindow()/2.0;

      // Compute range of eta within 4*Gaussian_sigma of omega
      // Changed to 2 sigma May 31 2001 by JZ


      eta_min = omega - 4*Gaussian_sigma;
      eta_max = omega + 4*Gaussian_sigma;

  
      if(eta_min < LeftRelFreq)
	eta_min = LeftRelFreq;

      if(eta_max > RightRelFreq)
	eta_max = RightRelFreq;

       
       LeftIndex = int(eta_min/DeltaFreq + 0.5*qtPoints);

       if(LeftIndex < 0)
	 LeftIndex = 0;

       if(LeftIndex >= qtPoints)
	 LeftIndex = qtPoints-1;

       RightIndex = int(eta_max/DeltaFreq + 0.5*qtPoints);
      
       // cout << Right Index = " << RightIndex << endl;

        if(RightIndex < 0)
	 RightIndex = 0;

       if(RightIndex >= qtPoints)
	 RightIndex = qtPoints-1;

       // Now Left and Right Indices are in range.

       if(LeftIndex > RightIndex)
	 {
           
	   cerr << "OptSpectrumAnalyzer::UpdateOSA:" << endl
                << "Index Error A " 
                << LeftIndex << "\t" << RightIndex 
                << "\t" << jj << endl;
          exit(1);
	 }

   
        for(int kk=LeftIndex;kk<=RightIndex;kk++)
	  {
	    // Compute eta(kk)

            eta = kk*DeltaFreq - oOptSignal->GetFreqWindow()/2.0;

            FreqDiff = abs(omega - eta);

            /* if( FreqDiff > 4*Gaussian_sigma )
	      FreqDiff -= oOptSignal->GetFreqWindow();
	    */

            GaussianFactor = exp(- 0.5*pow(FreqDiff/Gaussian_sigma,
                                           SuperGaussExponent));

            fOSA[jj] += GaussianFactor*fPowerSignal[kk];


	  } // end for kk loop
    
    } // ## end for jj loop


  fft->ifftshift(fOSA);

} // ## end  UpdateOSA_OLD

// #######################################################

double OptSpectrumAnalyzer::GetNoiseEquivalentBandwidth(void)
{

double B_OSA = 0;

if(ModeOSA==1)
{

double Gaussian_sigma = ResolutionBandwidthHz/
                          (2*pow(2*log(2),1.0/(SuperGaussExponent)));

int MaxIndex = int(ceil(4*Gaussian_sigma/DeltaFreq));

// Contribution from peak of supergaussian:

B_OSA = 1.0;

// Contributions from positive and negative rel freqs 
// are the same. Hence we compute one side and multiply by 2

for(int ii=1;ii<=MaxIndex;ii++)
{
 
double Freq = ii*DeltaFreq;

B_OSA += 2*exp(- 0.5*pow(Freq/Gaussian_sigma,SuperGaussExponent));

}

B_OSA *= DeltaFreq;

}

if(ModeOSA==2)
 B_OSA = ResolutionBandwidthHz;

return B_OSA;

}

// ######################################################

void OptSpectrumAnalyzer::OutputOSA(int ExptNum)
{

  int Distance = int(PropagatedLength/1000);

  string OutFileName;

  char C_String[20];

  sprintf(C_String,"OSA_E%d_%d.dat",ExptNum,Distance);

  string Suffix = C_String;

  OutFileName = Job + Suffix;

  ofstream OutFile;

  OutFile.open(OutFileName.c_str(),ios::out);

  fft->fftshift(fOSA);

 if(ExptNum == 0)
    ExptNum++;


 streamsize temp = OutFile.precision(8);

  OutFile << "%(1) Wavelength [nm], (2) RelFreq [GHz], (3) OSA power [dBm]"
          << endl; 

  for (int ii = 0; ii < qtPoints; ii++)
    { 
     OutFile <<  1e9*Frequency2Wavelength(
                 oOptSignal->GetCenterFreq() + 
                 ii*DeltaFreq - oOptSignal->GetFreqWindow()/2.0)  
             << "\t"
             << 1e-9*(ii*DeltaFreq - oOptSignal->GetFreqWindow()/2.0)
             << "\t"  
             << Linear2dBm(real(fOSA[ii])/ExptNum)
             << " " 
             << real(fOSA[ii])/ExptNum
             << endl;   
     }           
  
  OutFile.precision(temp);

  fft->ifftshift(fOSA); 
 
  OutFile.close();

} // ## end  OutputOSA

// #########################################################

double OptSpectrumAnalyzer::GetOSAOutputLinear(int ExptNum, double RelFreq)
{

  fft->fftshift(fOSA);

  int Index = int(
              floor((RelFreq + oOptSignal->GetFreqWindow()/2.0)/DeltaFreq));

  double Value = real(fOSA[Index])/ExptNum;

  fft->ifftshift(fOSA);

  return Value;

} // ## end  OutputOSA


// ############################################################
// Written by Jonathan Hu Feb 25
// #######################################################

void OptSpectrumAnalyzer::OutputOSA(int ExptNum, string Str)
{ // writen by Jonathan Hu

  int Distance = int(PropagatedLength/1000);

  string OutFileName;

  char C_String[20];

  sprintf(C_String,"OSA_E%d_%d",ExptNum,Distance);

  string Suffix = C_String;

  OutFileName = Job + Suffix + Str + ".dat";

  ofstream OutFile;

  OutFile.open(OutFileName.c_str(),ios::out);

  fft->fftshift(fOSA);

 if(ExptNum == 0)
    ExptNum++;


 streamsize temp = OutFile.precision(8);

  for (int ii = 0; ii < qtPoints; ii++)
    { 
     OutFile <<  1e9*Frequency2Wavelength(
                 oOptSignal->GetCenterFreq() + 
                 ii*DeltaFreq - oOptSignal->GetFreqWindow()/2.0)  
             << "\t" 
             << Linear2dBm(real(fOSA[ii])/ExptNum)
       //       << real(fOSA[ii])/ExptNum
             << endl;   
     }           
   cout << "ExptNum=" << ExptNum << endl;
  OutFile.precision(temp);

  fft->ifftshift(fOSA); 
 
  OutFile.close();

} // ## end  OutputOSA

// #######################################################3

double OptSpectrumAnalyzer::GetPeakPowerdBm(int ChannelArrayIndex, int ExptNum)
{   // writen by Jonathan Hu
	fft->fftshift(fOSA);
	int kk = qtPoints/2 + (int) ((oOptSignal->GetFrequency(ChannelArrayIndex) 
				- oOptSignal->GetCenterFreq())/DeltaFreq);
//	cout << "kk = " << kk << "  " << " osa=" << Linear2dBm(real(fOSA[kk]));
// 	for(int i=0;i<qtPoints;i++)
//		{
//			*LogFile << i << "  " <<  Linear2dBm(real(fOSA[i])) << endl;
//		}
 	double value;
	value = Linear2dBm(real(fOSA[kk])/ExptNum);
	fft->ifftshift(fOSA);
	return(value);


}  // ## endl of GetPeakPower

// ######################################################

double OptSpectrumAnalyzer::GetValleyPowerdBm(int ChannelArrayIndex, int ExptNum)
{   // writen by Jonathan Hu
	fft->fftshift(fOSA);
	int kk = qtPoints/2 + (int) (((oOptSignal->GetFrequency(ChannelArrayIndex)
				+oOptSignal->GetFrequency(ChannelArrayIndex+1))/2
				- oOptSignal->GetCenterFreq())/DeltaFreq);
//	cout << "kk = " << kk << "  " << " osa=" << Linear2dBm(real(fOSA[kk]));
// 	for(int i=0;i<qtPoints;i++)
//		{
//			*LogFile << i << "  " <<  Linear2dBm(real(fOSA[i])) << endl;
//		}
 	double value;
	value = Linear2dBm(real(fOSA[kk])/ExptNum);
	fft->ifftshift(fOSA);
	return(value);

}  // ## endl of GetValleyPower



// ###########################################################
// Written by Jonathan Hu Feb 25
void OptSpectrumAnalyzer::Clear()
{
  for (int ii = 0; ii < qtPoints; ii++)
    { 
		fOSA[ii] = 0;
		tOSA[ii] = 0;
    } 

}

// #######################################################
// ## By: Ivan Lima (06/21/01)
void OptSpectrumAnalyzer::WriteFileOutputOSA( string OutFileName, int ExptNum)
{ 
  //## ExtNum should be used to normalize the intensity by the number
  //#  of times the fOSA was accumulated. If there is only one, set 
  //#  it to 1 (0 also works).

  //int Distance = int(PropagatedLength/1000);
  //string OutFileName;
  //char C_String[20];
  //sprintf(C_String,"OSA_E%d_%d.dat",ExptNum,Distance);
  //string Suffix = C_String;
  //OutFileName = Job + Suffix;
  //ofstream OutFile;
  ofstream OutFile(OutFileName.c_str(),ios::out);

  fft->fftshift(fOSA);

  if(ExptNum == 0) {
     ExptNum++;
  }   

  for (int ii = 0; ii < qtPoints; ii++) { 
     OutFile <<  ii*DeltaFreq - oOptSignal->GetFreqWindow()/2.0  
             << "  " 
             << real(fOSA[ii])/ExptNum
             << endl;   
  }           
   
  fft->ifftshift(fOSA); 
 
  OutFile.close();

} // ## end  OutputOSA

// #####################################################


void OptSpectrumAnalyzer::ComputeAveragePowerFreqWindowOSA
          (double LeftFreq,double RightFreq,
           double *Average_Watts)
{

  int LeftIndex,RightIndex;

  
  int qtPoints = oOptSignal->GetQtPoints();

  int NumPointsWindow = int(abs(RightFreq-LeftFreq)/DeltaFreq);

  //cout << "NumPointsWindow = " << NumPointsWindow << endl;

  if(RightFreq < LeftFreq)
    {
      cerr << "Error in ComputeAveragePowerFreqWindow" << endl;
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;        
      exit(1);
    }

  if(LeftFreq >= 0)
    {
      LeftIndex = int(LeftFreq/DeltaFreq);
    }     
  else
    {
      LeftIndex = int(LeftFreq/DeltaFreq) +  qtPoints;
    }

  // *LogFile << "LeftFreq [GHz] = " << LeftFreq*1e-9 << endl;
  //*LogFile << "LeftIndex = " << LeftIndex << endl;
  //*LogFile << "Freq(LeftIndex) [GHz] = " << oOptSignal->fft->freq(LeftIndex) 
  //  << endl;

  double CenterFreq = oOptSignal->GetCenterFreq();

  // cout << "LeftWave [nm] = " << Frequency2Wavelength(
  //                            CenterFreq +RightFreq)*1e9 << endl;
  
  

  if(LeftIndex < 0 || LeftIndex > qtPoints)
    {

      cerr <<  "OptSpectrumAnalyzer::ComputeAveragePowerFreqWindow:"      
	   << "LeftIndex error (2) !" << endl;
      cout << "LeftIndex = " << LeftIndex << endl;        
      cout << "RightIndex = " << RightIndex << endl;             
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;  

      exit(1);
    }

  if(RightFreq >= 0)
    {
      RightIndex = int(RightFreq/DeltaFreq);
    }     
  else
    {
      RightIndex = int(RightFreq/DeltaFreq) +  qtPoints;
    }

  //*LogFile << "RightFreq [GHz] = " << RightFreq*1e-9 << endl;
  //*LogFile << "RightIndex = " << RightIndex << endl << flush;
  //*LogFile << "Freq(RightIndex) [GHz] = "
  // << oOptSignal->fft->freq(RightIndex) 
  //  << endl;

  //cout << "RightWave [nm] = " << Frequency2Wavelength(
  //                           CenterFreq + LeftFreq)*1e9 << endl;
  

  if(RightIndex < 0 || RightIndex > qtPoints)
    {
      cerr << "OptSpectrumAnalyzer::ComputeAveragePowerFreqWindow:"
           << "RightIndex error (2) !" << endl;
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;             
      cout << "LeftIndex = " << LeftIndex << endl;        
      cout << "RightIndex = " << RightIndex << endl;             
      exit(1);
    }

  *Average_Watts = 0.0;

  if(LeftIndex <= RightIndex)
    {
      for(int ii=LeftIndex; ii < RightIndex; ii++)
	{
           *Average_Watts += real(fOSA[ii]);
	}


    }
  else
    {
      int ii;

      for(ii=LeftIndex;ii<qtPoints;ii++)
	{
         *Average_Watts += real(fOSA[ii]);
	}


      for(ii=0;ii<RightIndex;ii++)
         {
         *Average_Watts += real(fOSA[ii]);
       	} 
    }


*Average_Watts = *Average_Watts/NumPointsWindow;
 
} // ## end ComputeAveragePowerFreqWindowOSA


// #####################################################

void  OptSpectrumAnalyzer::OutputSNR_OSA(int NumExpts)
{

  string OutFileName;

  char C_String[20];

  sprintf(C_String,"SNR_OSA_%d.dat",NumExpts);

  string Suffix = C_String;

  OutFileName = Job + Suffix;

  ofstream OutFile;

  if(AppendFlag)
    OutFile.open(OutFileName.c_str(),ios::app);
  else
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "% PropagatedLength [km] and for each channel:" << endl
	      << "% SNR [dB], Peak Signal Power [dBm]  Noise Power [mW]"
	      << endl;
    }


  if(NumExpts == 0)
     NumExpts++;

  OutFile << PropagatedLength/1e3 << "\t";


  // Outputs SNR [dB], Peak Signal Power [mW]  Noise Power [mW]    


   for(int ch=0;ch<NumChannels;ch++)
    {
      OutFile << Linear2dB(SignalPeakLin[ch]/NoiseAverageLin[ch]) 
              << "\t"
	      << Linear2dBm(SignalPeakLin[ch]/NumExpts) << "\t" 
	      << 1e3*(NoiseAverageLin[ch]/NumExpts) << "\t"; 


    } // end for ch loop

   OutFile << endl;

   OutFile.close();

} // ## end OutputSNR_OSA

// ##################################################

void  OptSpectrumAnalyzer::OutputSNR_OSA(int NumExpts, string Str)
{

  string OutFileName;

  char C_String[20];

  sprintf(C_String,"SNR_OSA_%d",NumExpts);

  string Suffix = C_String;

  OutFileName = Job + Suffix + Str + ".dat";

  ofstream OutFile;

  if(AppendFlag)
    OutFile.open(OutFileName.c_str(),ios::app);
  else
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "% PropagatedLength [km] and for each channel:" << endl

              << "% SNR [dB], Peak Signal Power [dBm]  Noise Power [mW]"
              << endl;
    }


  if(NumExpts == 0)
     NumExpts++;

  OutFile << PropagatedLength/1e3 << "\t";

  // Outputs SNR [dB], Peak Signal Power [mW]  Noise Power [mW]    

   for(int ch=0;ch<NumChannels;ch++)
    {
      OutFile << Linear2dB(SignalPeakLin[ch]/NoiseAverageLin[ch])
              << "\t"
              << Linear2dBm(SignalPeakLin[ch]/NumExpts) << "\t"
              << 1e3*(NoiseAverageLin[ch]/NumExpts) << "\t";

       cout << "OSNR = " 
	    << Linear2dB(SignalPeakLin[ch]/NoiseAverageLin[ch]) << endl;
        cout << "Psig = " << Linear2dBm(SignalPeakLin[ch]/NumExpts) << endl;
        cout << "Pase = " << Linear2dBm(NoiseAverageLin[ch]/NumExpts) << endl;

    } // end for ch loop

   OutFile << endl;

   OutFile.close();

} // ## end OutputSNR_OSA

//######################################################


void  OptSpectrumAnalyzer::UpdateSNR_OSA(double PropagatedLength2)
{

  PropagatedLength = PropagatedLength2;

  for(int ch=0;ch<NumChannels;ch++)
    {
   

      // We are not really computing averages but rather total
      // power within a given interval of frequencies.
      // The fact that by summing the freq data over an interval
      // gives you the total power in the interval comes from
      // Parseval's relation. See comment in OptSignal class
      // where there is code to test this. (Search under Parseval).


      //LogFileSeparator();

      //*LogFile << "Updating OSNR in channel " << ch+1 << endl;


      double RelFreqChannel = oOptSignal->GetFrequency(ch) - 
	                      oOptSignal->GetCenterFreq();

      //  cout << "RelFreqChannel [GHz] = " << RelFreqChannel*1e-9 << endl;


      //cout << " ResolutionBandwidthHz = " <<  ResolutionBandwidthHz << endl;

      double SignalWindowLeft = RelFreqChannel - ResolutionBandwidthHz/2.0;
      double SignalWindowRight = RelFreqChannel + ResolutionBandwidthHz/2.0;


      //cout << "SignalWindowLeft [GHz] = " << SignalWindowLeft*1e-9 << endl;
      //cout << "SignalWindowRight [GHz] = " << SignalWindowRight*1e-9 
      //                      << endl;


      //double Average_Watts;
      
      
      SignalPeakLin[ch] = ComputeMaxPowerLinFreqWindowOSA(
                                      SignalWindowLeft,SignalWindowRight);
	 
      double NoiseWindow1Center = RelFreqChannel - 
                                  0.5*oOptSignal->GetChannSpacing();

      double NoiseWindow1Left = NoiseWindow1Center - ResolutionBandwidthHz/2.0;
      double NoiseWindow1Right = NoiseWindow1Center+ResolutionBandwidthHz/2.0;


      //cout << "NoiseWindow1Left [GHz] = " << NoiseWindow1Left*1e-9 
      //   << endl;

      //cout << "NoiseWindow1Right [GHz] = " << NoiseWindow1Right*1e-9 
      //   << endl;



      // ComputeAveragePowerFreqWindowOSA(NoiseWindow1Left,NoiseWindow1Right,
      //                                 &Average_Watts);

      
      //NoiseAverageLin[ch] += 0.5*Average_Watts;

       NoiseAverageLin[ch] = ComputeMaxPowerLinFreqWindowOSA(
                                      NoiseWindow1Left,NoiseWindow1Right);



      double NoiseWindow2Center = RelFreqChannel + 
                                  0.5*oOptSignal->GetChannSpacing();

      double NoiseWindow2Left = NoiseWindow2Center - ResolutionBandwidthHz/2.0;
      double NoiseWindow2Right = NoiseWindow2Center+ResolutionBandwidthHz/2.0;


      //cout << "NoiseWindow2Left [GHz] = " << NoiseWindow2Left*1e-9 
      //           << endl;

      //cout << "NoiseWindow2Right [GHz] = " << NoiseWindow2Right*1e-9 
      //     << endl;


      //ComputeAveragePowerFreqWindowOSA(NoiseWindow2Left,NoiseWindow2Right,
      //                               &Average_Watts);

      
      //NoiseAverageLin[ch] += 0.5*Average_Watts;

      double temp = ComputeMaxPowerLinFreqWindowOSA(
                                      NoiseWindow2Left,NoiseWindow2Right);

      if(temp > NoiseAverageLin[ch])
         NoiseAverageLin[ch] = temp;

 

      //cout << NoiseAverageLin[ch] << "\t" << SignalPeakLin[ch] << endl;


    } // ## end ch loop

} // ## end UpdateSNR_OSA

//#################################################

double OptSpectrumAnalyzer::ComputeMaxPowerLinFreqWindowOSA(
                            double LeftFreq,double RightFreq)
{


  int LeftIndex,RightIndex;

  int qtPoints = oOptSignal->GetQtPoints();

 
  if(RightFreq < LeftFreq)
    {
      cout << "Error in ComputeMaxPowerdBmFreqWindow" << endl;          
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;         
      exit(1);
    }

  if(LeftFreq >= 0)
    {
      LeftIndex = int(floor(LeftFreq/DeltaFreq));
    }     
  else
    {
      LeftIndex = int(floor(LeftFreq/DeltaFreq)) +  qtPoints;
    }
 
 /*
  cout << "LeftFreq [GHz] = " << LeftFreq*1e-9 << endl;
  cout << "LeftIndex = " << LeftIndex << endl;
  cout << "Freq(LeftIndex) [GHz] = " << oOptSignal->fft->freq(LeftIndex)*1e-9 
    << endl;
 */

  double CenterFreq = oOptSignal->GetCenterFreq();

 
  //cout << "RightWave [nm] = " << Frequency2Wavelength(
  //        CenterFreq + oOptSignal->fft->freq(LeftIndex))*1e9 << endl;
  
  
  if(LeftIndex < 0 || LeftIndex > qtPoints)
    {
      cout << "OptSpectrumAnalyzer::ComputeMaxPowerLinFreqWindowOSA:"
           << "LeftIndex error (3) !" << endl;
      cout << "LeftIndex = " << LeftIndex << endl;  
      cout << "RightIndex = " << RightIndex << endl;           
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;              
      exit(1);
    }

  if(RightFreq >= 0)
    {
      RightIndex = int(RightFreq/DeltaFreq);
    }     
  else
    {
      RightIndex = int(RightFreq/DeltaFreq) +  qtPoints;
    }

 

  /*
  cout << "RightFreq [GHz] = " << RightFreq*1e-9 << endl;
  cout << "RightIndex = " << RightIndex << endl << flush;
  cout << "Freq(RightIndex) [GHz] = "
   << oOptSignal->fft->freq(RightIndex)*1e-9
    << endl;
  */
  

  //cout << "LeftWave [nm] = " << Frequency2Wavelength(
  //              CenterFreq +oOptSignal->fft->freq(RightIndex))*1e9 << endl;
  

  if(RightIndex < 0 || RightIndex > qtPoints)
    {
      cout << "OptSpectrumAnalyzer::ComputeMaxPowerLinFreqWindowOSA:"
                     << "RightIndex error (3) !" << endl;
      cout << "LeftIndex = " << LeftIndex << endl;  
      cout << "RightIndex = " << RightIndex << endl;           
      cout << "LeftFreq = " << LeftFreq << endl;        
      cout << "RightFreq = " << RightFreq << endl;              
      exit(1);
    }



  double MaxPowerLin = 0.0;
  double Candidate;

  if(LeftIndex <= RightIndex)
    {
      for(int ii=LeftIndex; ii < RightIndex; ii++)
	{
                     
         Candidate = real(fOSA[ii]);

         if(Candidate > MaxPowerLin)
	   MaxPowerLin = Candidate;


	}

    }
  else
    {
      int ii;

      for(ii=LeftIndex;ii<qtPoints;ii++)
	{
         Candidate = real(fOSA[ii]);

         if(Candidate > MaxPowerLin)
	   MaxPowerLin = Candidate;
	}


      for(ii=0;ii<RightIndex;ii++)
         {
          Candidate = real(fOSA[ii]);

         if(Candidate > MaxPowerLin)
	   MaxPowerLin = Candidate;
       	} 
    } // end else


  //cout << "MaxPowerLin = " << MaxPowerLin << endl;

  return MaxPowerLin;

} // ## ComputeMaxPowerLinFreqWindowOSA

// ############################################

void  OptSpectrumAnalyzer::RunPowerMeter(double PropagatedLength2)
{

  // Written by John Zweck Apr 9th 2001
  // The OSA can also be used as a simple power meter
  // Experimentalists often measure the power map using
  // the peak signal power per channel in an OSA.
  // The power meter is designed to do just this.


  if(PowerMeterFlag)
    {
     ClearPowerMeter();

     switch(ModeOSA)
      {

        case 1:
         if(SuperGaussExponent > 2 || SuperGaussExponent < 2)
          {
	 //cout << "PM:Using UpdateOSA_OLD as SuperGaussExponent != 2" << endl;
           UpdateOSA_OLD(PropagatedLength2);
          }
         else
          {
	    //cout << "PM:Using UpdateOSA" << endl;
           UpdateOSA(PropagatedLength2);
          }

        break;

        case 2:
         UpdateOSA(PropagatedLength2);
         break;
      }

     UpdateSNR_OSA(PropagatedLength2);
     OutputPowerMeter();
     //OutputOSA(1); // for debugging
    }

} //## end RunPowerMeter

// ############################################

double  OptSpectrumAnalyzer::AmpPowerMeterInputLinear(void)
{

double AverageSignalPowerLin;
double PropagatedLength2 = 0.0; // Irrelevant for Amp

if(PowerMeterFlag)
    {
     ClearPowerMeter();

     switch(ModeOSA)
      {

        case 1:
         if(SuperGaussExponent > 2 || SuperGaussExponent < 2)
          {
	 //cout << "PM:Using UpdateOSA_OLD as SuperGaussExponent != 2" << endl;
           UpdateOSA_OLD(PropagatedLength2);
          }
         else
          {
	    //cout << "PM:Using UpdateOSA" << endl;
           UpdateOSA(PropagatedLength2);
          }

        break;

        case 2:
         UpdateOSA(PropagatedLength2);
         break;
      }

     UpdateSNR_OSA(PropagatedLength2);
     AverageSignalPowerLin = PowerMeterComputeAverageSignalPowerLin();
    }

return AverageSignalPowerLin;

} // ## end AmpPowerMeterInputLinear

// #################################################

void  OptSpectrumAnalyzer::ClearPowerMeter(void)
{

  for(int ii=0;ii<qtPoints;ii++)
    fOSA[ii] = 0.0;

  for(int ii=0;ii<qtPoints;ii++)
    tOSA[ii] = 0.0;
}

// #####################################################

void  OptSpectrumAnalyzer::OutputPowerMeter(void)
{

  string OutFileName;

  char C_String[20];

  sprintf(C_String,"PowerMap.dat");

  string Suffix = C_String;

  OutFileName = Job + Suffix;

  ofstream OutFile;

  static int FirstTimeCalled = 1;


  if(FirstTimeCalled)
    {
      OutFile.open(OutFileName.c_str(),ios::out);
      OutFile << "%(1) PropagatedLength [km] (2) Average power per channel, "
	      << "% averaged over all channels using OSA, then output in [dBm]"
	       << endl
	      << "% (3) Power in Channel 1 [dBm] " 
	      << endl;
    }
   else
    OutFile.open(OutFileName.c_str(),ios::app);
  
  FirstTimeCalled = 0;

  OutFile << PropagatedLength/1e3 << "\t";

  // Outputs Peak Signal Power [dBm]

  /*
   for(int ch=0;ch<NumChannels;ch++)
    {
      OutFile << Linear2dBm(SignalPeakLin[ch]) << "\t"; 
	  
    } // end for ch loop

 OutFile << endl;

  */


  double AverageSignalPeakPower = 0;

 for(int ch=0;ch<NumChannels;ch++)
   AverageSignalPeakPower += SignalPeakLin[ch];

 AverageSignalPeakPower = AverageSignalPeakPower/NumChannels;

 OutFile << Linear2dBm(AverageSignalPeakPower) << "\t";
   
 for(int ch=0;ch<NumChannels;ch++)
   OutFile << Linear2dBm(SignalPeakLin[ch]) << "\t";
 OutFile << endl;
 
   OutFile.close();

} // ## end OutputPowerMeter

//#####################################################

double OptSpectrumAnalyzer::PowerMeterComputeAverageSignalPowerLin(void)
{

  double AverageSignalPower = 0.0;

  for(int ch=0;ch<NumChannels;ch++)
    {
      AverageSignalPower += SignalPeakLin[ch]; 
    } // end for ch loop

 return AverageSignalPower/double(NumChannels);

} // ## end PowerMeterComputeAverageSignalPowerLin

// #######################################################



//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/24/01)
void OptSpectrumAnalyzer::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   fft = oOptSignal->fft;
   qtPoints = oOptSignal->GetQtPoints();
}








