// $Id: ocsPhotodetector.cc,v 1.3 2007/10/24 14:27:11 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsPhotodetector.hh"

extern ofstream *LogFile;

// ################### Methods for the class Photodetector
// # This class contains the fft structure, sfftM, that carries out
// # the optical signal in the electrical domain throughout the 
// # simulation.

// #### Constructor of the class Photodetector
// ## By: Ivan Lima
Photodetector::Photodetector()
{
   ObjectInitialized = 0;
}

// #### Destructor of the class Photodetector
// ## By: Ivan Lima
Photodetector::~Photodetector()
{
   if (ObjectInitialized)
      ReleaseMemory();   
}

// #### Releases the memory (internal method)
// ## By: Ivan Lima
void Photodetector::ReleaseMemory()
{
   FreeCplx(sfftPM.tPowerSignal);  
   FreeCplx(sfftPM.fPowerSignal); 
  
   if(oOptSignal->GetTypePulse() == NRZ)
     {
      FreeCplx(WorksfftPM.tPowerSignal);  
      FreeCplx(WorksfftPM.fPowerSignal); 
     }

}

// #### Initializes the photodetector
// ## By: Ivan Lima

void Photodetector::InitPhotodetector(cfftw *fft2
                              ,fftMatrices sfftM2
                              ,double ConvFactOptElecPhotod2)
{

   fft = fft2;
   sfftM = sfftM2;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;      
   ConvFactOptElecPhotod = ConvFactOptElecPhotod2;

   AddShotNoiseFlag = 0;
   ShotNoiseVarianceCoeff = 0.0; // Out of date
   ShotNoiseCoeff = 0.0;

   if (ObjectInitialized && qtPoints != qtPointsLast) {
      ReleaseMemory();
      ObjectInitialized = 0;
   }   
   if (! ObjectInitialized ) {
      ObjectInitialized = 1;   
      sfftPM.tPowerSignal = NewCplx(qtPoints);
      sfftPM.fPowerSignal = NewCplx(qtPoints);
      qtPointsLast = qtPoints;          

      if(oOptSignal->GetTypePulse() == NRZ)
	{
         WorksfftPM.tPowerSignal = NewCplx(qtPoints);
         WorksfftPM.fPowerSignal = NewCplx(qtPoints);
	}
             
   } 
}

// ######################################################

Photodetector::Photodetector(string InFileName,OptSignal *oOptSignal2,
                             RanNumGen *RNG2)
{
  // IF YOU CHANGE THIS METHOD CHANGE THE NEXT TWO CONSTRCUTORS AND
  // COPY CONSTRUCTOR  AND ASSIGNMENT OPERATOR TOO

  // Use this constructor if you might add shot noise

  // Written by John Zweck 25th Oct, 2000, modified Jan 18th 2002 (RNG2)
  // Adapted from an out of date method by Ivan Lima

 
 ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
   {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }

   
 
   DebugLevel = oOptSignal2->GetDebugLevel();

   if(DebugLevel)
     {
      LogFileSeparator();
      *LogFile << InFileName << endl << endl;
     }

   //oOptSignal = oOptSignal2; 

   ConvFactOptElecPhotod = ReadDouble("ConvFactOptElecPhotod"," ",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile,
				      DebugLevel);

   AddShotNoiseFlag = 
           ReadInt("AddShotNoiseFlag"," ",0,1,LOWER_AND_UPPER,
                    OPTIONAL_NO_WARNING,
                    0,&InFile,DebugLevel);


   //ShotNoiseVarianceCoeff is out of date, 10 Jan 2002.
   // Use ShotNoiseCoeff instead

   ShotNoiseVarianceCoeff = ReadDouble("ShotNoiseVarianceCoeff"," ",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

   ShotNoiseCoeff = ReadDouble("ShotNoiseCoeff","sqrt(Hz)",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

  
   InFile.close();

   SetOptSignal(oOptSignal2);
/*
   oOptSignal = oOptSignal2; 
   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;      
   TimeWindow = oOptSignal->GetTimeWindow();
   BitPeriod = TimeWindow/double(oOptSignal->GetStringLength());
*/


   RNG = RNG2; // JZ Jan 18th 2002
   
   sfftPM.tPowerSignal = NewCplx(qtPoints);
   sfftPM.fPowerSignal = NewCplx(qtPoints);

   if(oOptSignal->GetTypePulse() == NRZ)
    {
      WorksfftPM.tPowerSignal = NewCplx(qtPoints);
      WorksfftPM.fPowerSignal = NewCplx(qtPoints);
    }

   ObjectInitialized = 1;

} // end constructor

// #################################################

Photodetector::Photodetector(string InFileName,OptSignal *oOptSignal2,
                             RanNumGen *RNG2,
                             fftPowerMatrices & sfftPM2,
                             fftPowerMatrices & WorksfftPM2)
{
  // IF YOU CHANGE THIS METHOD CHANGE THE OTHER TWO CONSTRCUTORS AND
  // COPY CONSTRUCTOR TOO

  // Use this constructor if you might add shot noise

  // Written by John Zweck 25th Oct, 2000, modified Jan 18th 2002 (RNG2)
  // Adapted from an out of date method by Ivan Lima

 
 ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
   {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


   DebugLevel = oOptSignal2->GetDebugLevel();

   if(DebugLevel)
     {
      LogFileSeparator();
      *LogFile << InFileName << endl << endl;
     }

   ConvFactOptElecPhotod = ReadDouble("ConvFactOptElecPhotod"," ",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,1,&InFile,
				      DebugLevel);

   AddShotNoiseFlag = 
           ReadInt("AddShotNoiseFlag"," ",0,1,LOWER_AND_UPPER,
                    OPTIONAL_NO_WARNING,
                    0,&InFile,DebugLevel);


   //ShotNoiseVarianceCoeff is out of date, 10 Jan 2002.
   // Use ShotNoiseCoeff instead

   ShotNoiseVarianceCoeff = ReadDouble("ShotNoiseVarianceCoeff"," ",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

   ShotNoiseCoeff = ReadDouble("ShotNoiseCoeff","sqrt(Hz)",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

  
   InFile.close();
   
   SetOptSignal(oOptSignal2);

/*
   oOptSignal = oOptSignal2; 
   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;      
   TimeWindow = oOptSignal->GetTimeWindow();
   BitPeriod = TimeWindow/double(oOptSignal->GetStringLength());
*/

   RNG = RNG2; // JZ Jan 18th 2002

   sfftPM = sfftPM2;
   WorksfftPM = WorksfftPM2;

   ObjectInitialized = 1;

} // end constructor

// #################################################

Photodetector::Photodetector(string InFileName,OptSignal *oOptSignal2)
{
  // IF YOU CHANGE THIS METHOD CHANGE THE PREVIOUS CONSTRCUTOR AND
  // COPY CONSTRUCTOR TOO

  // Written by John Zweck 25th Oct, 2000
  // Adapted from an out of date method by Ivan Lima

 
 ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
   {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


   DebugLevel = oOptSignal2->GetDebugLevel();

   if(DebugLevel)
     {
      LogFileSeparator();
      *LogFile << InFileName << endl << endl;
     }
   
 
   ConvFactOptElecPhotod = ReadDouble("ConvFactOptElecPhotod"," ",0,-1,
                                    LOWER_ONLY,OPTIONAL_WARNING,1,&InFile,
				      DebugLevel);

   AddShotNoiseFlag = 
           ReadInt("AddShotNoiseFlag"," ",0,1,LOWER_AND_UPPER,
		   OPTIONAL_NO_WARNING,
                    0,&InFile,DebugLevel);


   //ShotNoiseVarianceCoeff is out of date, 10 Jan 2002.
   // Use ShotNoiseCoeff instead

   ShotNoiseVarianceCoeff = ReadDouble("ShotNoiseVarianceCoeff"," ",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

   ShotNoiseCoeff = ReadDouble("ShotNoiseCoeff","sqrt(Hz)",0,-1,
				       LOWER_ONLY,OPTIONAL_NO_WARNING,0.0,
                                     &InFile,DebugLevel);

  
   InFile.close();

   SetOptSignal(oOptSignal2);
/*
   oOptSignal = oOptSignal2; 
   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;      
   TimeWindow = oOptSignal->GetTimeWindow();
   BitPeriod = TimeWindow/double(oOptSignal->GetStringLength());
*/

   RNG = oOptSignal->RNG; 
   
   sfftPM.tPowerSignal = NewCplx(qtPoints);
   sfftPM.fPowerSignal = NewCplx(qtPoints);

   if(oOptSignal->GetTypePulse() == NRZ)
     {
      WorksfftPM.tPowerSignal = NewCplx(qtPoints);
      WorksfftPM.fPowerSignal = NewCplx(qtPoints);
     }

   ObjectInitialized = 1;

} // end constructor


// #################################################
//### Set the parameters for the optical signal
//# By: Ivan Lima

void Photodetector::SetOptSignal(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2; 
   fft = oOptSignal->fft;
   sfftM = oOptSignal->sfftM;
   qtPoints = fft->give_N();
   DeltaFreq = fft->delta_f;
   FreqWindow = DeltaFreq*qtPoints;
   DeltaTime = 1/FreqWindow;      
   TimeWindow = oOptSignal->GetTimeWindow();
   BitPeriod = TimeWindow/double(oOptSignal->GetStringLength());
}



// #################################################

Photodetector::Photodetector(Photodetector & OrigPhotodetector)
{

  // Copy Constructor
  // Written by John Zweck, April 5th 2001.

  oOptSignal = OrigPhotodetector.oOptSignal; 
  qtPoints = OrigPhotodetector.qtPoints;
  qtPointsLast  = OrigPhotodetector.qtPointsLast;
  ConvFactOptElecPhotod = OrigPhotodetector.ConvFactOptElecPhotod;
  DeltaFreq = OrigPhotodetector.DeltaFreq;
  DeltaTime = OrigPhotodetector.DeltaTime;
  FreqWindow = OrigPhotodetector.FreqWindow;
  TimeWindow = OrigPhotodetector.TimeWindow;
  BitPeriod = OrigPhotodetector.BitPeriod;
  ObjectInitialized = OrigPhotodetector.ObjectInitialized;
  AddShotNoiseFlag = OrigPhotodetector.AddShotNoiseFlag;
  ShotNoiseVarianceCoeff = OrigPhotodetector.ShotNoiseVarianceCoeff;
  ShotNoiseCoeff = OrigPhotodetector.ShotNoiseCoeff;

  RNG = OrigPhotodetector.RNG; // Keep same RNG
  fft = OrigPhotodetector.fft; //Keep same fft

     
  sfftPM.tPowerSignal = NewCplx(qtPoints);
  sfftPM.fPowerSignal = NewCplx(qtPoints);

  for(int ii=0;ii<qtPoints;ii++)
    sfftPM.tPowerSignal[ii] = OrigPhotodetector.sfftPM.tPowerSignal[ii];

  for(int ii=0;ii<qtPoints;ii++)
    sfftPM.fPowerSignal[ii] = OrigPhotodetector.sfftPM.fPowerSignal[ii];


  if(oOptSignal->GetTypePulse() == NRZ)
    {
     WorksfftPM.tPowerSignal = NewCplx(qtPoints);
     WorksfftPM.fPowerSignal = NewCplx(qtPoints);

     for(int ii=0;ii<qtPoints;ii++)
      WorksfftPM.tPowerSignal[ii] = 
       OrigPhotodetector.WorksfftPM.tPowerSignal[ii];

     for(int ii=0;ii<qtPoints;ii++)
      WorksfftPM.fPowerSignal[ii] = 
       OrigPhotodetector.WorksfftPM.fPowerSignal[ii];
    }

}// ## end Copy Constructor

// #####################################################

Photodetector & Photodetector::operator=(Photodetector & RHS_Photodetector)
{
 // Copy Constructor
  // Written by John Zweck, Feb 20th 2002.

  // We should really insist that RHS_Photodetector.oOptSignal 
  // and oOptSignal are really the same
  // as in the assignment operator
  //  OptSignal & OptSignal::operator=(OptSignal & RHS_Signal)

  if(qtPoints != RHS_Photodetector.qtPoints)
     {
       *LogFile << "Invalid use of Photodetector::operator=" << endl
                << "Aborting Now!!" << endl << flush;
       exit(1);

     }

  oOptSignal = RHS_Photodetector.oOptSignal; 
  
  qtPointsLast  = RHS_Photodetector.qtPointsLast;
  ConvFactOptElecPhotod = RHS_Photodetector.ConvFactOptElecPhotod;
  DeltaFreq = RHS_Photodetector.DeltaFreq;
  DeltaTime = RHS_Photodetector.DeltaTime;
  FreqWindow = RHS_Photodetector.FreqWindow;
  TimeWindow = RHS_Photodetector.TimeWindow;
  BitPeriod = RHS_Photodetector.BitPeriod;
  ObjectInitialized = RHS_Photodetector.ObjectInitialized;
  AddShotNoiseFlag = RHS_Photodetector.AddShotNoiseFlag;
  ShotNoiseVarianceCoeff = RHS_Photodetector.ShotNoiseVarianceCoeff;
  ShotNoiseCoeff = RHS_Photodetector.ShotNoiseCoeff;

  RNG = RHS_Photodetector.RNG; // Keep same RNG
  fft = RHS_Photodetector.fft; //Keep same fft

  for(int ii=0;ii<qtPoints;ii++)
    sfftPM.tPowerSignal[ii] = RHS_Photodetector.sfftPM.tPowerSignal[ii];

  for(int ii=0;ii<qtPoints;ii++)
    sfftPM.fPowerSignal[ii] = RHS_Photodetector.sfftPM.fPowerSignal[ii];

  if(oOptSignal->GetTypePulse() == NRZ)
    {
      for(int ii=0;ii<qtPoints;ii++)
       WorksfftPM.tPowerSignal[ii] = 
        RHS_Photodetector.WorksfftPM.tPowerSignal[ii];

      for(int ii=0;ii<qtPoints;ii++)
       WorksfftPM.fPowerSignal[ii] = 
        RHS_Photodetector.WorksfftPM.fPowerSignal[ii];
    }

}// ## end Photodetector::operator=

// #######################################################

// ##### Convert from the Optical to the Electrical domain
// ## By: Ivan Lima

void Photodetector::DetectOptSignal()
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class Photodetector not initialized" 
           << endl;
      exit(1);
   } 
   for ( int ii = 0; ii < qtPoints; ii++) {
      sfftPM.tPowerSignal[ii] = ConvFactOptElecPhotod
                    *( sq(abs(sfftM.tOptSignalX[ii])) 
                      +sq(abs(sfftM.tOptSignalY[ii])) );
   }
  
   //fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);         
   
   Time2Freq();

   AddShotNoise();

}   

// ############################################################

// ##### Compute the RF power at a given frequency
// ## By: Brian Marks

double Photodetector::GetRFPower(double Freq)
{
  int FreqIndex = (int)(abs(Freq)*TimeWindow + 0.1);

  return sq(abs(sfftPM.fPowerSignal[FreqIndex]));

} // ## End GetRFPower

// #######################################################

// ##### Compute the RF phase at a given frequency
// ## By: Brian Marks

double Photodetector::GetRFPhase(double Freq)
{
  int FreqIndex = (int)(abs(Freq)*TimeWindow + 0.1);

  return atan2(imag(sfftPM.fPowerSignal[FreqIndex]),
	       real(sfftPM.fPowerSignal[FreqIndex]));

} // ## End GetRFPhase

// #######################################################

// ##### Compute the RF power at a given frequency
// ## By: Brian Marks

double Photodetector::GetRFPowerFromFreqIndex(int FreqIndex)
{

  return sq(abs(sfftPM.fPowerSignal[FreqIndex]));

} // ## End GetRFPowerFromFreqIndex

// #######################################################

// ##### Compute the RF phase at a given frequency
// ## By: Brian Marks

double Photodetector::GetRFPhaseFromFreqIndex(int FreqIndex)
{

  return atan2(imag(sfftPM.fPowerSignal[FreqIndex]),
	       real(sfftPM.fPowerSignal[FreqIndex]));

} // ## End GetRFPhaseFromFreqIndex

// ############################################################

void Photodetector::AddShotNoise(void)
{

  // Written by John Zweck, April 3rd 2001, modified Jan 10th 2002
  // By default AddShotNoiseFlag == 0

  // For the theory of shot noise in a photodetector see 
  // Agrawal "Fiber-optic communication systems", p 163 ff,
  // VPI's manual and tek.com and agilent.com 
  // under oscilloscopes and agilent.com search for 83440C 83440D
  //for data sheets.
 
  // Algorithm:
  // We input the shot noise coefficient, called q by Agrawal, in sqrt(Hz).
  // The two-sided spectral noise density is then
  // N_shot = q Average Current in A/sqrt(Hz).
  // We add noise in the frequency interval [f,f+DeltaFreq] of 
  // n_shot([f,f+DeltaFreq]) = N_shot sqrt(DeltaFreq) exp(2 pi j phi_random)
  // for all positive frequencies.
  // For the negative frequencies we us 
  // n_shot(-f) = conj(n_shot(f)) 
  // to ensure Hermitian symmetry of n_shot(f) and hence that
  // the shot noise current in time, n_shot(t), is real. 

  //WriteFileTimePower("InitialTime.dat");
  //WriteFileFreqPower("InitialFreq.dat",0);

  if(AddShotNoiseFlag)
    {
      double NoiseAmplitude = ShotNoiseCoeff*GetAverageElecCurrent()*
                              sqrt(DeltaFreq);

      //cout << "Average Current = " << GetAverageElecCurrent() << endl;
      //cout << "NoiseAmplitude = " << NoiseAmplitude << endl;

      // Frequencies go like  0 1 2 3 -4 -3 -2 -1  (N=8)
      // Treat DC term (0) and -N/2 term (eg -4) as special
      // cases. The other terms are related by index i and N-i
      // and need to have the Hermitian symmetry.
  
      for(int ii=0;ii<qtPoints/2;ii++)
	{
          cplx Noise = NoiseAmplitude*exp(2*pi*jc*RNG->GetRanNum());

           sfftPM.fPowerSignal[ii] = sfftPM.fPowerSignal[ii] + Noise;
          
	   // For hermitian symmetry need DC term to be real

          if(ii==0)
	    sfftPM.fPowerSignal[0] = real(sfftPM.fPowerSignal[0]);

          if(ii!=0)
           sfftPM.fPowerSignal[qtPoints-ii] = 
           sfftPM.fPowerSignal[qtPoints-ii] + conj(Noise);
 
	}

      //sfftPM.fPowerSignal[qtPoints/2]  =   sfftPM.fPowerSignal[qtPoints/2] +
      //	 NoiseAmplitude*exp(2*pi*jc*RNG->GetRanNum());

      sfftPM.fPowerSignal[qtPoints/2]  = sfftPM.fPowerSignal[qtPoints/2];

      Freq2Time();

      //cout << "Post Shot Noise Average Current = " 
      //     << GetAverageElecCurrent() << endl;


    }// ## end if(AddShotNoiseFlag)

  //WriteFileTimePower("FinalTime.dat");
  //WriteFileFreqPower("FinalFreq.dat",0);


} // #### end Photodetector::AddShotNoise(void)

// ###################################################################

void Photodetector::AddShotNoiseOLD(void)
{

  // Written by John Zweck, April 3rd 2001
  // By default AddShotNoiseFlag == 0

  // For the theory of shot noise in a photodetector see 
  // Agrawal "Fiber-optic communication systems", p 163 ff.
  // The parameter ShotNoiseVarianceCoeff := q times ReceiverBandwidth.
  // For our purposes I set ReceiverBandwidth = FreqWindow.
  // So ShotNoiseVariance = sq(sigma_s) = ShotNoiseVarianceCoeff times I_p
  // where I_p = Electric current in photodetector.


  if(AddShotNoiseFlag)
    {

       double q = ShotNoiseVarianceCoeff/FreqWindow;

      double NoiseAmplitude = sqrt(q*DeltaFreq*GetAverageElecPower());

      //cout << "Average Power = " << GetAverageElecPower() << endl;
      //cout << "NoiseAmplitude = " << NoiseAmplitude << endl;

      for(int ii=0;ii<qtPoints;ii++)
       sfftPM.fPowerSignal[ii] = sfftPM.fPowerSignal[ii] + 
	                        NoiseAmplitude*exp(2*pi*jc*RNG->GetRanNum());

      Freq2Time();

      // We want an output signal which is real in time.
      // The simplest (though not necessarily best) way to do this
      // is to take the real part of the time signal and then use Time2Freq.
      // Note that Noise is real in time iff Noise is hermitian in freq
      // It may be better (to ensure no statistical correlations??) to either 
      // (1) Add shot noise directly in time domain, which you can do or
      // (2) When add noise in freq domain make sure that added noise
      // is hermitian: N(omega) = conj(N(-omega)).

      for(int ii=0;ii<qtPoints;ii++)
	sfftPM.tPowerSignal[ii] = real(sfftPM.tPowerSignal[ii]);

      Time2Freq();


      //cout << "Post Shot Noise Average Power = " 
      //     << GetAverageElecPower() << endl;


    }// ## end if(AddShotNoiseFlag)

} // #### end Photodetector::AddShotNoiseOLD(void)

// ####################################################################

void Photodetector::ComputeWorkDifferentiatedThenSquaredElecSignal(void)
{

  WorksfftPM.tPowerSignal[0] = 
      (sfftPM.tPowerSignal[1] - sfftPM.tPowerSignal[qtPoints-1])/(2*DeltaTime);


  for(int ii=1;ii<qtPoints-1;ii++)
    WorksfftPM.tPowerSignal[ii] = 
      (sfftPM.tPowerSignal[ii+1] - sfftPM.tPowerSignal[ii-1])/(2*DeltaTime);

  WorksfftPM.tPowerSignal[qtPoints-1] = 
      (sfftPM.tPowerSignal[0] - sfftPM.tPowerSignal[qtPoints-2])/(2*DeltaTime);


  // We want to square the real part and set the imag part to 0:

  for(int ii=0;ii<qtPoints;ii++)
    WorksfftPM.tPowerSignal[ii] = 
      real(WorksfftPM.tPowerSignal[ii])*
      real(WorksfftPM.tPowerSignal[ii]);

  WorkTime2Freq();

} // ## end ComputeWorkDifferentiatedThenSquaredElecSignal(void)

// ###########################################################

void Photodetector::ComputeWorkContEXORElecSignal(void)
{

  for(int ii=0;ii<qtPoints;ii++)
    {
      int Shifted_ii = int(ii + 0.5*BitPeriod/DeltaTime + 0.5);

      while(Shifted_ii >= qtPoints)
	Shifted_ii -= qtPoints;


      //cout << ii << "\t" << Shifted_ii << endl;

      WorksfftPM.tPowerSignal[ii] = 
	  sfftPM.tPowerSignal[ii] - sfftPM.tPowerSignal[Shifted_ii];

    }

  // We want to square the real part and set the imag part to 0:

  for(int ii=0;ii<qtPoints;ii++)
    WorksfftPM.tPowerSignal[ii] = 
      real(WorksfftPM.tPowerSignal[ii])*
      real(WorksfftPM.tPowerSignal[ii]);

  WorkTime2Freq();

} // ## end ComputeWorkContEXORElecSignal

// ################################################################

void Photodetector::ComputeWorkBinaryEXORElecSignal(void)
{

  double AveragePower = GetAverageElecPower();

  //cout << "AveragePower = " <<  AveragePower << endl;

  for(int ii=0;ii<qtPoints;ii++)
    {

      double Term1, Term2;

      if(sq(abs(sfftPM.tPowerSignal[ii])) > AveragePower)
	Term1 = 1.0;
      else
        Term1 = 0.0;

      int Shifted_ii = int(ii + 0.5*BitPeriod/DeltaTime + 0.5);

      while(Shifted_ii >= qtPoints)
	Shifted_ii -= qtPoints;


      if(sq(abs(sfftPM.tPowerSignal[Shifted_ii])) > AveragePower)
	Term2 = 1.0;
      else
        Term2 = 0.0;

     
      WorksfftPM.tPowerSignal[ii] = Term1 - Term2;

    }

  // We want to square the real part and set the imag part to 0:

  for(int ii=0;ii<qtPoints;ii++)
    WorksfftPM.tPowerSignal[ii] = 
      real(WorksfftPM.tPowerSignal[ii])*
      real(WorksfftPM.tPowerSignal[ii]);

  WorkTime2Freq();

} // ## end BinaryEXORElecSignal

// ##############################################################

void Photodetector::ComputeWorkSumDelayedHalfBitPeriodThenSquare(void)
{

//## Make a copy of the ElecSignal, differentiate it and
//#  then square it. The effect of this is to double 
//#  the frequency of the signal to get a strong 
//#  component to recover the clock on the frequency
//#  of the signal.
//## By: Ivan Lima, based on another function by John Zweck
   
   int StringLength = oOptSignal->GetStringLength();

   for (int ii = 0; ii < qtPoints; ii++) {
      WorksfftPM.tPowerSignal[ii] = 
          sfftPM.tPowerSignal[ii] 
          -sfftPM.tPowerSignal[Periodicity(
               ii+qtPoints/(2*StringLength),qtPoints)];
   }      

  // We want to square the real part and set the imag part to 0:

  for(int ii=0;ii<qtPoints;ii++)
     WorksfftPM.tPowerSignal[ii] = 
      real(WorksfftPM.tPowerSignal[ii])*
      real(WorksfftPM.tPowerSignal[ii]);

  WorkTime2Freq();

} // ## end DifferentiateElecSignal

// ###################################################################

double Photodetector::GetAverageElecCurrent(void)
{

 double AvCurrent = 0;
 
 for (int ii = 0; ii < qtPoints; ii++) 
  {   
    AvCurrent +=  real(sfftPM.tPowerSignal[ii]);            
  }
 return(AvCurrent*DeltaTime/TimeWindow);

} // ## end GetAverageElecCurrent

// ###################################################################

double Photodetector::GetAverageElecPower(void)
{
   double MassEnergy = 0;
 
   for (int ii = 0; ii < qtPoints; ii++) {   
      MassEnergy +=  sq(real(sfftPM.tPowerSignal[ii]));            
   }

   return(MassEnergy*DeltaTime/TimeWindow);

} // ## end GetAverageElecPower

//#######################################################
//## Get the signal bandwidth
//## By: Ivan Lima (5/3/02)

double Photodetector::GetSignalBandwidth(void)
{
   //double FilterTransfDC = sq(abs(sfftPM.fPowerSignal[0]));
   //cout << "%$%$%$%$ = " << sq(abs(sfftPM.fPowerSignal[0])) << endl;
   double EffectiveSignalBandwidth = 0;
   for (int ii = 0; ii < qtPoints; ii++)  {   
      EffectiveSignalBandwidth +=  sq(abs(sfftPM.fPowerSignal[ii]));
      //sq(abs(sfftPM.fPowerSignal[ii]));
//                                   /FilterTransfDC;            
   }
   return EffectiveSignalBandwidth*DeltaFreq;

}  //## End GetSignalBandwidth

// ################################################################

double Photodetector::GetAverageAbsElecPower(void)
{
   double MassEnergy = 0;
 
   for (int ii = 0; ii < qtPoints; ii++) {   
      MassEnergy +=  sq(abs(sfftPM.tPowerSignal[ii]));            
   }

   return(MassEnergy*DeltaTime/TimeWindow);

} //## end GetAverageAbsElecPower

//###################################################################
//##### Converts from time to frequency domain
//## By: Ivan Lima
void Photodetector::Time2Freq(void)
{
   fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);    
}


// ##############################################################

// ##### Converts from frequency to time domain
// ## By: Ivan Lima
void Photodetector::Freq2Time(void)
{
   fft->IFFT(sfftPM.fPowerSignal,sfftPM.tPowerSignal);    
}


//################################################
//##### Run the FFT shift 
//##    It is convenient to visualize the signal
//#     in the spectrum domain.
//#     This is one interface method.
//## By: Ivan Lima
void Photodetector::fftshift(void)
{
   fft->fftshift(sfftPM.fPowerSignal); 
}


//################################################
//##### Run the IFFT shift 
//##    It undo the shift produced by FFT_shift
//#     This is one interface method.
void Photodetector::ifftshift(void)
{
   fft->ifftshift(sfftPM.fPowerSignal); 
}


//################################################

//! This method shifts the signal (JZFeb03: to right) 
//! by the amount dTime using the fft
//! By: Brian Marks

void Photodetector::TimeShiftSignal(double dTime)
{
   fftshift();
      
   for (int ii=0; ii<qtPoints; ii++) {
     sfftPM.fPowerSignal[ii] *= 
        exp(jc*2.*pi*(ii-qtPoints/2)*dTime/TimeWindow);

       // JZFeb03: Previous line changed since changed to Physics convention 
       //exp(-jc*2.*pi*(ii-qtPoints/2)*dTime/TimeWindow);
   }
   
   ifftshift();
   
   Freq2Time();
   
}


//##########################################################

void Photodetector::WorkTime2Freq(void)
{
   fft->FFT(WorksfftPM.tPowerSignal,WorksfftPM.fPowerSignal);    
}

// #############################################################

// ##### Write electrical signal current in time domain to a file
// ## By: Ivan Lima
void Photodetector::WriteFileTimePower(string outFile)
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class Photodetector not initialized" 
           << endl;
      exit(1);
   }   
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

OutFile << "%(1) Time (2) Abs value of Elec Current (3) Real part Elec Current"
	<< endl;

   for (int ii = 0; ii < qtPoints; ii++) 
      OutFile <<  ii*DeltaTime  << "   " 
              << abs(sfftPM.tPowerSignal[ii]) <<  " " 
	      << real(sfftPM.tPowerSignal[ii]) 
              << endl; 
                 
   OutFile.close();
}


// ###############################################################

/*!

Outputs power spectral density to a file

 */

// ##### Write electrical signal current in frequency domain to a file
// ## By: Ivan Lima
void Photodetector::WriteFileFreqPower(string outFile,int OmitDCFlag)
{

   if (! ObjectInitialized) {
      cerr << "Error: Object of class Photodetector not initialized " 
           << endl;
      exit(1);
   }   
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

   OutFile << "%(1) Freq (2) Abs value of Elec Current "
	   << "(3) Real part Elec Current (2&3 are power spectral densities)"
	<< endl;


   //////fft->FFT(sfftPM.tPowerSignal,sfftPM.fPowerSignal);
   fft->fftshift(sfftPM.fPowerSignal);
   for (int ii = 0; ii < qtPoints; ii++) 
     {

       if(OmitDCFlag)
	 {
           // The DC frequency is dropped out
           if (ii != qtPoints/2) 
            {  
              OutFile <<  ii*DeltaFreq - FreqWindow/2  << "   " 
                 << abs(sfftPM.fPowerSignal[ii])/DeltaFreq <<  " " 
   		 << real(sfftPM.fPowerSignal[ii])/DeltaFreq 
                 << endl;
	

             } 
	 }
       else
	 {
            OutFile <<  ii*DeltaFreq - FreqWindow/2  << "   " 
                    << abs(sfftPM.fPowerSignal[ii])/DeltaFreq <<  " "
	            << real(sfftPM.fPowerSignal[ii])/DeltaFreq
		    << endl; 

	 }

          
   }   
   fft->ifftshift(sfftPM.fPowerSignal);                 
   OutFile.close();
}


//##############################################################
//##### Write electrical signal current in frequency domain to a file
//##     It is suitable to visualize the data in log scale.
//## By: Ivan Lima
void Photodetector::WriteFileFreqPower(string outFile)
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class Photodetector not initialized " 
           << endl;
      exit(1);
   }   
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

   //## You should use the fields (1*) and (3*)
   OutFile << "%(1*) Time (2) Abs value of Elec Current "
	   << " (3*) Real part Elec Current"
	   << endl;

   fft->fftshift(sfftPM.fPowerSignal);
   for (int ii = 0; ii < qtPoints; ii++) {
      OutFile <<  ii*DeltaFreq - FreqWindow/2  << "   " 
              << abs(sfftPM.fPowerSignal[ii])/DeltaFreq <<  " "
	      << real(sfftPM.fPowerSignal[ii])/DeltaFreq 
	      << endl; 
   }   
   fft->ifftshift(sfftPM.fPowerSignal);                 
   OutFile.close();
}



//#########################################################
//## Generate impulse response that is useful to study filters
//## By: Ivan Lima (8/20/02)


  //! Used in ChiSquareReceiver class to construct the
  //! transfer function of the electric filter. 

void Photodetector::GenerateImpulseCurrent(void)   
{
   OptSignal *oOptSignal2 = new OptSignal(*oOptSignal);
   oOptSignal2->GenerateImpulseSignal();

   for (int ii = 0; ii < qtPoints; ii++) {
      sfftPM.tPowerSignal[ii] = 
            oOptSignal2->sfftM.tOptSignalX[ii];
   }   
   Time2Freq();   
   delete oOptSignal2;
}   










