// $Id: 
//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsPolDepAttenuator.hh" // ## In order to access the friend methods

extern ofstream *LogFile;

// ################### Methods for the class PolDepAttenuator

// #### Constructor of PolDepAttenuator
// ## By: Ivan Lima
PolDepAttenuator::PolDepAttenuator(OptSignal *oOptSignalTmp)
{
   InitPolDepAttenuator( oOptSignalTmp);
}


// #### Constructor of PolDepAttenuator from an input file
// ## By: Ivan Lima
PolDepAttenuator::PolDepAttenuator(string InFileName, 
                                   OptSignal *oOptSignalTmp)  
{
   InitPolDepAttenuator( oOptSignalTmp);
   
   ifstream InFile(InFileName.c_str(),ios::in);    
   if (!InFile) {
      cerr << "Error: Input file "<< InFileName 
           << "could not be opened."<< endl;
      exit(1);
   }   
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;
   
   double PolDepAttenuationX_Tmp = ReadDouble("PolDepAttenuationX",
                                            " ",1,-1,NO_BOUNDS,&InFile);
   double PolDepAttenuationY_Tmp = ReadDouble("PolDepAttenuationY",
                                            " ",1,-1,NO_BOUNDS,&InFile);
      
   SetPolDepAttenuation( PolDepAttenuationX_Tmp
                        ,PolDepAttenuationY_Tmp);

   InFile.close();

   // added by Hai 7-1-03

   JonesMatrix=oOptSignal->JonesMatrixChannel; 
   NumberOfChannels=oOptSignal->GetNumChannels();

			
}


// #### Constructor of PolDepAttenuator w/ initialization
// ## By: Ivan Lima
PolDepAttenuator::PolDepAttenuator(OptSignal *oOptSignalTmp
                          ,double PolDepAttenuationX_Tmp
                          ,double PolDepAttenuationY_Tmp)  
{
   InitPolDepAttenuator( oOptSignalTmp);
   SetPolDepAttenuation( PolDepAttenuationX_Tmp
                        ,PolDepAttenuationY_Tmp);
}


// ## Initializes the parameters
void PolDepAttenuator::InitPolDepAttenuator
                      (OptSignal *oOptSignalTmp)
{
   oOptSignal = oOptSignalTmp;
   //NumChannels = oOptSignal->GetNumChannels();
   qtPoints = oOptSignal->GetQtPoints();
   oReduPolDepAttenuator = new ReduPolDepAttenuator
                               (oOptSignal->oReduOptSignal);
   ObjectInitialized = 0;
}               


// ## Allows for a change in the loss at both polarizations
void PolDepAttenuator::SetPolDepAttenuation(
                             double PolDepAttenuationX_Tmp
                            ,double PolDepAttenuationY_Tmp)
{
   SetPolDepAttenuationX(PolDepAttenuationX_Tmp);
   SetPolDepAttenuationY(PolDepAttenuationY_Tmp);

   oReduPolDepAttenuator->SetPolDepAttenuation(PolDepAttenuationX_Tmp,
					       PolDepAttenuationY_Tmp);
              
   ObjectInitialized = 1;   
}                            
     
           
// ## Allows for a change in the loss at the X polarization
void PolDepAttenuator::SetPolDepAttenuationX(
                              double PolDepAttenuationX_Tmp)
{
   // ## The factor /2. comes from the fact that 
   // #  the LinarPolDepAttenuationX is actually
   // #  from the field, not from the power.
   // #  But the attenuation is done correctly in
   // #  the method PolDepAttenuator.
   LinearPolDepAttenuationX 
           = dB2Linear(-PolDepAttenuationX_Tmp/2.);
   //GetTransmMatrixCoefficients();
}


// ## Allows for a change in the loss at the Y polarization
void PolDepAttenuator::SetPolDepAttenuationY(
                              double PolDepAttenuationY_Tmp)
{
   // ## The factor /2. comes from the fact that 
   // #  the LinarPolDepAttenuationX is actually
   // #  from the field, not from the power.
   // #  But the attenuation is done correctly in
   // #  the method PolDepAttenuator.
   LinearPolDepAttenuationY 
           = dB2Linear(-PolDepAttenuationY_Tmp/2.);
   //GetTransmMatrixCoefficients();   
}


// #### Propagates through the PolDepAttenuator
// ## By: Ivan Lima
void PolDepAttenuator::PolDepAttenuate(void)
{
   if (! ObjectInitialized) {
      cerr << "Error: Object of class PolDepAttenuator"
           << " not initialized" 
           << endl;
      exit(1);
   }                      
   //cout << "LinearPolDepAttenuationX = " << LinearPolDepAttenuationX << endl;
   //cout << "LinearPolDepAttenuationY = " << LinearPolDepAttenuationY << endl;
   
   for (int ii = 0; ii < qtPoints; ii++) {
      oOptSignal->sfftM.fOptSignalX[ii] *= LinearPolDepAttenuationX;
      oOptSignal->sfftM.fOptSignalY[ii] *= LinearPolDepAttenuationY;      
   } 
   oOptSignal->Freq2Time();  

   oReduPolDepAttenuator->PolDepAttenuate();
}


//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/24/01)

// Not set up for oReduPolDepAttenuator

void PolDepAttenuator::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   qtPoints = oOptSignal->GetQtPoints();
   //CenterFreq = oOptSignal->GetCenterFreq();
}

//####################################################

//! By Hai

void PolDepAttenuator::UpdateJonesMatrixPDL(void)
{
	for (int ii=0;ii<NumberOfChannels;ii++)
	{

	     JonesMatrix[ii*4+0]*=LinearPolDepAttenuationX;
					   
	     JonesMatrix[ii*4+1]*=LinearPolDepAttenuationX;
				   		
	     JonesMatrix[ii*4+2]*=LinearPolDepAttenuationY;
				   
	     JonesMatrix[ii*4+3]*=LinearPolDepAttenuationY;
	    }
}
	    
//####################################################
