// $Id: ocsPolarizTransformer.cc,v 1.2 2006/05/11 22:34:30 zweck Exp $
//###################################################################
//
//       Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsPolarizTransformer.hh"

extern ofstream *LogFile;

// ########### Methods for the class PolarizTransformer         

//################################################################
// #### Initializes the object PolarizTransformer
// ## By: Ivan Lima
PolarizTransformer::PolarizTransformer( int TypeScattering2,  
                                        OptSignal *oOptSignalTmp)
{
   TypeScattering = typeScattering(TypeScattering2);
   oOptSignal = oOptSignalTmp;
   qtPoints = oOptSignal->GetQtPoints();

   LogFileSeparator();   
   *LogFile << "Initializing PolarizTransformer" << endl;
   *LogFile << "Type Scattering = " << WriteTypeScattering() << endl;   

   //## Default values -> Do nothing
   psiLast = thetaLast = phiLast = 0;

   
   oReduPolarizTransformer = new ReduPolarizTransformer(
				 TypeScattering,
				 oOptSignal->oReduOptSignal);
}



//################################################################
// #### Returns the TypeScattering
// ## By: Ivan Lima
string PolarizTransformer::WriteTypeScattering(void)
{
   string sTypeScattering;

   switch(TypeScattering) {
      case RX_RY_RX:
         sTypeScattering = "RX_RY_RX == 1: The Euler Matrix";
         break;
      case RY_RZ_RY: 
         sTypeScattering = "RY_RZ_RY == 2";
         break;
      case  RX_RZ_RY:
         sTypeScattering = "RX_RZ_RY == 3";  
         break;
      default:
         cout << "From: ReduPolarizTransformer"<< endl;
         cout << "ERROR: Invalid choice of TypeScattering" << endl;
         cout << "TypeScattering = " << TypeScattering << endl;
         cout << "ABORTING NOW!!" << endl;
         exit(1);
         break;	 
   }
   return sTypeScattering;
}


//################################################################
//#### Polarization Transformation
//## It causes the following rotations based on the 
//#  selected sequence of rotation
//## By: Ivan Lima
void PolarizTransformer::TransformPolarization(
                 double psi
                ,double theta
                ,double phi )
{
   psiLast   = psi;
   thetaLast = theta;
   phiLast   = phi;

   cplx JonesVector[2];
   
   for (int ii = 0; ii < qtPoints; ii++) {   
      JonesVector[0] = oOptSignal->sfftM.tOptSignalX[ii];
      JonesVector[1] = oOptSignal->sfftM.tOptSignalY[ii]; 
      switch (TypeScattering) {
         case RX_RY_RX: // ## : Euler Matrix
            //TransformPolarizationRxRyRx();  
            RotatesAboutX( psiLast,  JonesVector);
            RotatesAboutY( -thetaLast, JonesVector);          	     
            RotatesAboutX( phiLast,   JonesVector);   	    
            break;
         case RY_RZ_RY: // ## RyRzRy      
            RotatesAboutY( -psiLast,  JonesVector);
            RotatesAboutZ( thetaLast, JonesVector);          	     
            RotatesAboutY( -phiLast,   JonesVector);  
            break;
         case RX_RZ_RY: // ## RxRzRy      
            //TransformPolarizationRxRzRy();
            RotatesAboutY( -psiLast,  JonesVector);
            RotatesAboutZ( thetaLast, JonesVector);          	     
            RotatesAboutX( phiLast,   JonesVector);	    
            break;
         default:
            cout << "From: ReduPolarizTransformer"<< endl;
            cout << "ERROR: Invalid choice of TypeScattering" << endl;
            cout << "TypeScattering = " << TypeScattering << endl;
            cout << "ABORTING NOW!!" << endl;
            exit(1);	       
   	    break;
      }                                              
      oOptSignal->sfftM.tOptSignalX[ii] = JonesVector[0];
      oOptSignal->sfftM.tOptSignalY[ii] = JonesVector[1];    
   } //## end for
   oOptSignal->Time2Freq();


 oReduPolarizTransformer->TransformPolarization(psi,theta,phi);
} 





//################################################################
//#### Undo the last Polarization Transformation
//## By: Ivan Lima
void PolarizTransformer::UndoLastPolarizTransformation(void)
{
   cplx JonesVector[2];
   
   for (int ii = 0; ii < qtPoints; ii++) {   
      JonesVector[0] = oOptSignal->sfftM.tOptSignalX[ii];
      JonesVector[1] = oOptSignal->sfftM.tOptSignalY[ii]; 
      switch (TypeScattering) {
         case RX_RY_RX: // ## : Euler Matrix
            RotatesAboutX( -phiLast,   JonesVector); 	    	   
            RotatesAboutY( thetaLast, JonesVector);          	      	    
            RotatesAboutX( -psiLast,  JonesVector);	    
            break;
         case RY_RZ_RY: // ## RyRzRy      
            RotatesAboutY( phiLast,   JonesVector); 
            RotatesAboutZ( -thetaLast, JonesVector);          	     
            RotatesAboutY( psiLast,  JonesVector); 
            break;
         case RX_RZ_RY: // ## RxRzRy      
            RotatesAboutX( -phiLast,   JonesVector);	
            RotatesAboutZ( -thetaLast, JonesVector);          	     
            RotatesAboutY( psiLast,  JonesVector);  
            break;
         default:
            cout << "From: ReduPolarizTransformer"<< endl;
            cout << "ERROR: Invalid choice of TypeScattering" << endl;
            cout << "TypeScattering = " << TypeScattering << endl;
            cout << "ABORTING NOW!!" << endl;
            exit(1);	       
   	    break;
      }                                              
      oOptSignal->sfftM.tOptSignalX[ii] = JonesVector[0];
      oOptSignal->sfftM.tOptSignalY[ii] = JonesVector[1];    
   } //## end for
   oOptSignal->Time2Freq();


   oReduPolarizTransformer->UndoLastPolarizTransformation();
} 

  
//#### Polarization transformation
//##   It produces the last rotation
//## By: Ivan Lima
void PolarizTransformer::TransformPolarization(void)
{ 
   TransformPolarization(psiLast,thetaLast,phiLast);
}  
  
 
 
//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/24/01)

// Not set up for oReduPolarizTransformer

void PolarizTransformer::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   qtPoints = oOptSignal->GetQtPoints();
}

 
 //################################################################
//#### Polarization Transformation
//## It sets the Jones vector of the signal
//## By: Hua Jiao on Nov. 28, 2005

void PolarizTransformer::TransformPolarization(cplx JonesVector0,
					       cplx JonesVector1)
{
   for (int ii = 0; ii < qtPoints; ii++) 
     {   
       oOptSignal->sfftM.tOptSignalX[ii] = 
	 JonesVector0/sqrt(oOptSignal->GetQtPoints());
      oOptSignal->sfftM.tOptSignalY[ii] = 
	JonesVector1/sqrt(oOptSignal->GetQtPoints());    
      	  
	   } //## end for
  
   oOptSignal->Time2Freq();
} 
               
