// $Id: 
// #####################################################################
//
//    ocsProcessControl.cc
//
//    Optical Communication Systems Simulator
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
// #####################################################################

# include "ocsProcessControl.hh"

extern ofstream *LogFile;

// ## Constructor
// #  By: Ivan Lima
void ProcessControl::InitProcessControl( int ProcessControlModeTmp 
                                   ,string ProcessControlFileName
                                   ,char *ProcessName)
{

   // ## Description of the modes:
   // #     ProcessControlMode == 0
   // #        -> No control at all;
   // #     ProcessControlMode == 1   
   // #        -> List processes in $HOME/ProcessControl.txt;
   // #     ProcessControlMode >= 2   
   // #        -> List processes in $HOME/ProcessControl.txt;   
   // #           Send an email to $(USER)@(HOSTNAME)
   // #           after the process is finished.   

   ProcessControlMode = ProcessControlModeTmp;
 
   PathListProc =  getenv("HOME");
   PathListProc += "/";
   PathListProc += ProcessControlFileName; //ProcessControl.txt";
   ListProcMessage = getenv("HOSTNAME"); 
   ListProcMessage +=  "\t ";  
   char cBuffer[30]; // ## It should be enough     
   sprintf(cBuffer,"%d",getpid());    
   ListProcMessage += cBuffer;   
   ListProcMessage +=  "\t ";                
   ListProcMessage += ProcessName;
   ListProcMessage +=  "\t ";           
   ListProcMessage += getenv("PWD");      
   InitTime = time(NULL);
   char *cTime =   ctime(&InitTime);
   cout << ListProcMessage << "\t" << cTime;
   //cout << ListProcMessage << "\t" << getpid() << endl;      
   if (ProcessControlMode > 0)   {
      // #### Creates a list of running processes
      OpenListProc( PathListProc);
      ListProc << ListProcMessage << "\t" << cTime;      
      //ListProc << ListProcMessage << endl;
      //ListProc << ListProcMessage << "\t" << getpid() << endl;   
      //ListProc.flush();
      ListProc.close();
   }
}

// ## Open ListProc file
void ProcessControl::OpenListProc(string ListProcFileName)
{
   ListProc.open(ListProcFileName.c_str(),ios::app);
   if (! ListProc) {
      cerr << "Error: "<< ListProcFileName
           <<" could not be open." 
      << endl << flush;
      exit(1);
   }    
}

// ## Destructor
// #  By: Ivan Lima
ProcessControl::~ProcessControl()
{
   FinalTime = time(NULL);
   char *cTime = ctime(&FinalTime);
   cout     << "END\t"
            << ListProcMessage  
            << "\t" << "Duration: " << FinalTime-InitTime << " sec"
            <<"\t" << cTime;      
   if (    ProcessControlMode == 1 
        || ProcessControlMode == 2 ) {
      OpenListProc(PathListProc);   
      ListProc << "END\t"
               << ListProcMessage
               << "\t" << "Duration: " << FinalTime-InitTime << " sec"
               <<"\t" << cTime;  
      //ListProc << ListProcMessage<< "\t" << getpid() 
      //         << "\t END" << endl;
      ListProc.close();
   }
   if (ProcessControlMode == 2) {
      // ## Now sends an e-mail, if required
      // ## ListProcTmp contains the name of the file
      // ## that will be e-mailed to the user
      
      string ListProcTmp = "ProcessControlTmp.txt";
      string DeleteTempFile = "rm ";
      DeleteTempFile += ListProcTmp;      
      system(DeleteTempFile.c_str());
      
      OpenListProc(ListProcTmp);
      //ListProc.open(ListProcTmp.c_str(),ios::out);
      //if (! ListProc) {
      //   cerr << "Error: "<< ListProcTmp
      //        <<" could not be open." 
      //   << endl << flush;
      //   exit(1);
      //}    
      //ListProc << ListProcMessage<< "\t" << getpid() 
      //         << "\t END" << endl;
      ListProc << "END\t"
               << ListProcMessage
               << "\t" << "Duration: " << FinalTime-InitTime << " sec"
               <<"\t" << cTime;        
      ListProc.close();   
      
      char cBuffer[20]; // ## It should be enough
      sprintf(cBuffer,"%d", getpid());    
      ListProcMessage  = "mail -s Process_Finished:_"; 
      ListProcMessage += cBuffer;
      ListProcMessage += " ";
      ListProcMessage += getenv("USER");
      ListProcMessage += " < "+ListProcTmp;
      system(ListProcMessage.c_str());                  
      //system("$OCS/mail -s ocsProcFinish lima < ocsListProcTmp.txt");   
   }      
}

// #####################################################################
