// $Id: ocsReadWriteTools.cc,v 1.2 2005/07/06 16:50:16 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReadWriteTools.hh"


extern ofstream *LogFile;

// ##################################################################

// ## Functions added by John Zweck

// ##############################################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            int DefaultValue, 
            ifstream * InFile)
{

int Value = ReadInt(ParameterName,UnitsString,LowerBound,UpperBound,
		    BoundsFlag,InputParameterOption,DefaultValue, InFile, 1);

 return Value;
}

// ###############################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            int DefaultValue, 
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  int ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      ParameterValue = DefaultValue;
    }

  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


} // ## end ReadInt

// ##########################################################

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound, int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
	    int DefaultValue,ifstream * InFile)
{
long Value = ReadLongInt(ParameterName,UnitsString,LowerBound,
			 UpperBound,BoundsFlag, 
			 InputParameterOption,DefaultValue, InFile, 1);

 return Value;
}

// #############################################################

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound, int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
	    int DefaultValue,
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  long ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      ParameterValue = DefaultValue;
    }

  if(DebugLevel)
   *LogFile << ParameterName << " = " << ParameterValue << " " 
           << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// ##############################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,
            ifstream * InFile)
{

double Value = 
  ReadDouble(ParameterName,UnitsString,LowerBound,UpperBound,BoundsFlag, 
           InputParameterOption,DefaultValue,InFile,1);

 return Value;

}

// ##########################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      ParameterValue =  DefaultValue;
    }

  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


}  // ## end ReadDouble

// ##############################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  string DefaultValue,
                  ifstream * InFile)
{

string Value = ReadString(ParameterName,InputParameterOption,DefaultValue,
	  InFile,1);

 return Value;


}

// ##############################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  string DefaultValue,
                  ifstream * InFile, bool DebugLevel)
{
  int ParameterRead = 0;
  string ParameterValue ="";
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      ParameterValue =  DefaultValue;
    }
  
    

  if(DebugLevel)
    {
      if(LogFile)
        *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
      else
        cout << endl 
             << ParameterName << " = " << ParameterValue << " " << endl 
             << endl;
    }

  return ParameterValue;

} // ## end ReadString

// ##############################################################


// ### Now follow version of ReadInt and ReadDouble without the 
// ### defult value parameter

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile)
{

int Value = 
ReadInt(ParameterName,UnitsString, LowerBound,UpperBound,
        BoundsFlag,InputParameterOption,InFile,1);

 return Value;

}


// ######################################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  int ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      return 0;
    }

  if(DebugLevel)
   *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


} // ## end ReadInt



// #############################################################

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound, int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile)
{

long Value = 
ReadLongInt(ParameterName,UnitsString,LowerBound,UpperBound,BoundsFlag, 
	    InputParameterOption,InFile,1);

 return Value;
}

// ##################################################

long ReadLongInt(string ParameterName, string UnitsString,
            int LowerBound, int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  long ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      return 0;
    }

  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// ##############################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
		  ifstream * InFile)
{

double Value = 
ReadDouble(ParameterName,UnitsString,LowerBound,UpperBound,
                  BoundsFlag,InputParameterOption,InFile,1);

 return Value;

}

// ##################################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      return 0;
    }

 if(DebugLevel)
   *LogFile << ParameterName << " = " << ParameterValue << " " 
           << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


}  // ## end ReadDouble

// ##############################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  ifstream * InFile)
{

string Value =
  ReadString(ParameterName,InputParameterOption,InFile,1);
 return Value;

}

// #######################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  ifstream * InFile, bool DebugLevel)
{
  int ParameterRead = 0;
  string ParameterValue ="";
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      return " ";
    }
  else
    {

      if(DebugLevel)
	{
         if(LogFile)
           *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
         else
          cout << endl 
               << ParameterName << " = " << ParameterValue << " " << endl 
               << endl;
	}
    }

  return ParameterValue;

} // ## end ReadString

// ##############################################################

void MissingParameter(string ParameterName,
                      InputParameterOptionsEnumerator InputParameterOption)
{

MissingParameter(ParameterName,InputParameterOption,1);

}

// ###############################################################

void MissingParameter(string ParameterName,
                      InputParameterOptionsEnumerator InputParameterOption,
		      bool DebugLevel)
{


  switch(InputParameterOption)
    {

      case OPTIONAL_NO_WARNING:
          break;

      case OPTIONAL_WARNING:
	if(LogFile)
	  {
          LogFileStars();
          *LogFile << "           WARNING" << endl 
		  << ParameterName << " not found!!" << endl;
          LogFileStars();
	  }
        else
	  {
           cout    << "           WARNING" << endl 
		   << ParameterName << " not found!!" << endl;
	  }

          break;

      case MANDATORY:

         if(LogFile)
	  {
          LogFileStars();
          *LogFile << "             WARNING" << endl 
		  << ParameterName << " not found!!" << endl
                  << "             ABORTING NOW" << endl;
          LogFileStars();
	  }

          cerr << "             WARNING" << endl 
		  << ParameterName << " not found!!" << endl
                  << "             ABORTING NOW" << endl;

          exit(1);
          break;


    default:
        cerr << "Switch option error in MissingParameter call for"
	   <<  ParameterName << endl;
        exit(1);

    } // ## end switch
         
}// ## end MissingParameter

// ###############################################################
// ###############################################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{

int Value = 
ReadInt(ParameterName,UnitsString, LowerBound,UpperBound,
        BoundsFlag,InFile,1);

 return Value;
}

// ################################################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile, 
	    bool DebugLevel)
{

  int ParameterRead = 0;
  int ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl
               << "             ABORTING NOW" << endl;
       LogFileStars();
       exit(1);
    }
  
  if(DebugLevel)
   *LogFile << ParameterName << " = " << ParameterValue << " " 
           << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadInt

// #########################################################

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{
long Value = 
ReadLongInt(ParameterName,UnitsString, LowerBound,UpperBound,
        BoundsFlag,InFile,1);

 return Value;
}

// ############################################################

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile, 
		 bool DebugLevel)
{

  int ParameterRead = 0;
  long ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl
               << "             ABORTING NOW" << endl;
       LogFileStars();
       exit(1);
    }
  
  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// #########################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{

double value = 
ReadDouble( ParameterName,UnitsString,
            LowerBound, UpperBound,
            BoundsFlag,InFile,1);
 return value;
}

// ##############################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile, 
		  bool DebugLevel)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl;
       LogFileStars();
      
    }
  
  if(DebugLevel)
     *LogFile << ParameterName << " = " << ParameterValue << " " 
             << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadDouble

// #######################################################

string ReadString(string ParameterName, ifstream * InFile)
{
string value = 
  ReadString(ParameterName,InFile,1);
 return value;
}

// ########################################################

string ReadString(string ParameterName, ifstream * InFile, 
		  bool DebugLevel) 
{

  int ParameterRead = 0;
  string ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      cerr << "ERROR in ReadString." << endl 
           << "Parameter " << ParameterName    
           << " not found." << endl << "ABORTING NOW" << endl;

      exit(1);

    }
  else
    {
      if(DebugLevel)
	{
         if(LogFile)
           *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
         else
          cout << endl 
               << ParameterName << " = " << ParameterValue << " " << endl 
               << endl;
	}
    }

  return ParameterValue;

} // end ReadString();

// #########################################################


void cerror(char * msg) {cout << msg << endl << flush;}

// #########################################################

void LogFileSeparator()
{

  *LogFile << endl << endl<<
          "#########################################" << endl << endl;

}

// #####################################################

void LogFileStars()
{
  *LogFile << endl << "**************************************"
          <<endl << endl;
}

// #########################################################
// ## Default error handler function
// ## By: Ivan Lima
void LogFileAbortWithErrorMsg(string ErrorMsg)
{ 
   LogFileSeparator();
   *LogFile << "Error:" << endl << ErrorMsg << endl;
   //## I should include the name of the log file here, if I can
   cout << "Error: See *.log for details on the error." << endl;
   exit(1);
}

// #########################################################

string WriteTypeSimulation(typeSimulation TypeSimulation)
{

 switch(TypeSimulation)
   {

    case SCALAR:

           return "SCALAR";
           break;

    case  VECTOR:
            return "VECTOR";
            break;


   default:
     cerr << "ERROR: Invalid choice of TypeSimulation" << endl;
     cerr << "ABORTING NOW!!" << endl;

     exit(1);


    }

} // ## end WriteTypeSimulation

// ######################################




// added by Walter to read vectors of doubles from an input file
// As input parameter one needs to pass the NumberOfElements of the
// vector who wants to load from file to memory
// The output parameter is the pointer to the first element of the 
// vector loaded in memory
// if the parameter is missing and it is not mandatory a vector
// with element values = DefaultValue will be created
// if set on, the boundary check is done for each of the elemnts of the vector

// ##############################################################

double* ReadVectorOfDoubles(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,unsigned NumbersOfElements,
            ifstream * InFile)
{

double* Pointer = 
  ReadVectorOfDoubles(ParameterName,UnitsString,LowerBound,UpperBound,BoundsFlag, 
           InputParameterOption,DefaultValue,NumbersOfElements,InFile,1);

 return Pointer;

}

// ##########################################################

double* ReadVectorOfDoubles(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue, unsigned NumberOfElements,
            ifstream * InFile, bool DebugLevel)
{
  double* Start_pointer= new double[NumberOfElements];
  double* Running_pointer=Start_pointer-1;
  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;
   
  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
      for (unsigned kk=0;kk<NumberOfElements;kk++)
         {
	 Running_pointer+=1;
         try
	    {
	    *InFile >> ParameterValue;
	    }
	 // this catches all the exceptions even the one that
	 // occurs when the ParameterValue that has been already read is not
	 // a double
	 catch ( ... ) 
	    {
	    cout << "passo per catch\n";           
	    ParameterRead=-1;
	    break;
	    }
	 *Running_pointer=ParameterValue; 
	  ParameterRead = 1;
         }  // end for
     }  // end if   
      
    } // end while


  if(ParameterRead==0 || ParameterRead==-1)
    {
      if (ParameterRead==0)
         MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      else
         WrongTypeParameter(ParameterName,DebugLevel);
     	 
      for (unsigned kk=0;kk<NumberOfElements;kk++)
         {
	 Running_pointer+=1;
         *Running_pointer=DefaultValue; 
         }
    }
    

  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;
  
  Running_pointer=Start_pointer-1;
  for (unsigned kk=0;kk<NumberOfElements;kk++)
     {
     Running_pointer+=1;    
     ParameterValue=*Running_pointer;
     switch(BoundsFlag)
        {
        case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The double parameter " 
                          << ParameterName << ", element " << kk << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << ", element " << kk << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << ", element " << kk << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch
 }   
  
 return Start_pointer;


}  // ## end ReadVectorOfDoubles

// ##############################################################

// ##############################################################

void WrongTypeParameter(string ParameterName)
{

WrongTypeParameter(ParameterName,1);

}

// ###############################################################

void WrongTypeParameter(string ParameterName, bool DebugLevel)
{

	if(LogFile)
	  {
          LogFileStars();
          *LogFile << "           WARNING" << endl 
		  << ParameterName << " is not of the defined type" << endl;
	  *LogFile << ParameterName << " will be set to DefaulValue " << endl;
		  
          LogFileStars();
	  }
        else
	  {
           cout    << "           WARNING" << endl 
		   << ParameterName << " is not ot the defined type" << endl;
	  cout << ParameterName << " will be set to DefaulValue " << endl;
	  }

         
}// ## end WrongTypeParameter


// ##############################################################

unsigned ReadUnsignedInt(string ParameterName, string UnitsString,
            unsigned LowerBound,unsigned UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            unsigned DefaultValue, 
            ifstream * InFile)
{

unsigned Value = ReadUnsignedInt(ParameterName,UnitsString,LowerBound,UpperBound,
		    BoundsFlag,InputParameterOption,DefaultValue, InFile, 1);

 return Value;
}

// ###############################################

unsigned ReadUnsignedInt(string ParameterName, string UnitsString,
            unsigned LowerBound,unsigned UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            unsigned DefaultValue, 
            ifstream * InFile, bool DebugLevel)
{

  int ParameterRead = 0;
  unsigned ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption,DebugLevel);
      ParameterValue = DefaultValue;
    }

  if(DebugLevel)
    *LogFile << ParameterName << " = " << ParameterValue << " " 
            << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


} // ## end ReadUnsignedInt






// end part added by Walter





