// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2003):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

// Written by John Zweck, April 2003

#include "ocsReceiverSemiAnalytical.hh"

extern ofstream *LogFile;

// #######################################################

ReceiverSemiAnalytical::ReceiverSemiAnalytical(void)
{
  ObjectInitialized = 0;
}

// #########################################################

ReceiverSemiAnalytical::ReceiverSemiAnalytical(typeReceiver TypeReceiver2,
					       OptSignal * WDMSignal2,
					       RanNumGen *RNG2,
					       int DemuxChannArrayIndex2,
					       string InDir, 
					       string Job2)
{

LogFileSeparator();

*LogFile << "ReceiverSemiAnalytical Constructor" << endl;

ObjectInitialized = 1;
 
TypeReceiver = TypeReceiver2;
WDMSignal = WDMSignal2;
DemuxChannArrayIndex = DemuxChannArrayIndex2;
Job = Job2;

SingleChannelSignal = new OptSignal(*WDMSignal);

TunableOptFilter = new OptFilter(InDir + "TunableOptFilter.in",
			         SingleChannelSignal);
  
oOptDemuxer = new OptDemuxer(WDMSignal,TunableOptFilter,SingleChannelSignal);

// Make this call so that in constructor the call to 
// oElecSignalStat->SetOptElecFiltersForGaussianStatistics(
//		                   TunableOptFilter,oElecFilter);
// is correct.


oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

oPhotodetector = new Photodetector (InDir + "Photodetector.in",
				    SingleChannelSignal,RNG2);

oElecFilter = new ElecFilter(InDir + "ElecFilter.in",oPhotodetector);

oElecSignalStat = new  ElecSignalStat(SingleChannelSignal,oPhotodetector);
	

// Later I'll want to save memory by rewriting following call

if(TypeReceiver == SEMI_ANALYTICAL)
 {
   oElecSignalStat->SetOptElecFiltersForGaussianStatistics(
   	                   TunableOptFilter,oElecFilter);

   StdDev_Signal_ASE = NewDouble(SingleChannelSignal->GetQtPoints());
 }

if(TypeReceiver == MONTE_CARLO)
  EyeDiagramMonteCarlo = new  Histogram2D(InDir + "Histogram2D.in", 
				                 SingleChannelSignal);

LogFileSeparator();
*LogFile << "ReceiverSemiAnalytical Constructor" << endl;

 *LogFile << "TypeReceiver = " << WriteTypeReceiver() << endl;


} // ## end ReceiverSemiAnalytical

// ########################################################## 

string ReceiverSemiAnalytical::WriteTypeReceiver(void)
{

  string TypeReceiverString;

  switch(TypeReceiver)
    {
     case MONTE_CARLO: 
      TypeReceiverString = "MONTE_CARLO";
      break;

     case SEMI_ANALYTICAL: 
      TypeReceiverString = "SEMI_ANALYTICAL";
      break;

     default:
       cerr << "Switch error in ReceiverSemiAnalytical::WriteTypeReceiver"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);

    }

  return TypeReceiverString;

} // ## end WriteTypeReceiver

// ############################################################

ReceiverSemiAnalytical::~ReceiverSemiAnalytical()
{
  
  if(ObjectInitialized == 1)
    {
     oOptDemuxer->~OptDemuxer();
     TunableOptFilter->~OptFilter();    
     oElecFilter->~ElecFilter();
     oElecSignalStat->~ElecSignalStat();
     //EyeDiagramMonteCarlo->~Histogram2D();
     oPhotodetector->~Photodetector();
     SingleChannelSignal->~OptSignal();
    }
  
} // ## end ~ReceiverSemiAnalytical()

// ###########################################################

void ReceiverSemiAnalytical::ReceiveSignalBackToBack(void)
{

  switch(TypeReceiver)
    {
     case MONTE_CARLO:
      ReceiveSignalBackToBackMonteCarlo();
      break;

     case SEMI_ANALYTICAL:
      ReceiveSignalSemiAnalytical("B2B_");
      break;

     default:
       cerr << "Switch error in "
	    << "ReceiverSemiAnalytical::ReceiveSignalBackToBack"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);
     
    }

}

// ########################################################

void ReceiverSemiAnalytical::ReceiveSignal(int MapNum)
{
 ReceiveSignal(MapNum,0);
}

// #########################################################

void ReceiverSemiAnalytical::ReceiveSignal(int MapNum,
				      	   int MonteCarloSystemRealizationNum)
{
  char C_String[10];
  string FileID;

  switch(TypeReceiver)
    {
     case MONTE_CARLO:
      ReceiveSignalMonteCarlo(MapNum,MonteCarloSystemRealizationNum);
      break;

     case SEMI_ANALYTICAL:

      sprintf(C_String,"RT_%i_Ch_%i_",MapNum,DemuxChannArrayIndex+1); 
      FileID = C_String;

      ReceiveSignalSemiAnalytical(FileID);
      break;

     default:
       cerr << "Switch error in "
	    << "ReceiverSemiAnalytical::ReceiveSignal"
	    << endl
	    << "ABORTING NOW!!" << endl;
       exit(1);
     
    }

}

// ########################################################

void ReceiverSemiAnalytical::ReceiveSignalBackToBackMonteCarlo(void)
{

  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex); 
  oPhotodetector->DetectOptSignal();  
  oElecFilter->FilterElecSignal();
  oElecSignalStat->ComputePerformanceStatistics();

  char  C_String[40];
  string FileName;

  sprintf(C_String,"InEye_%i.dat",DemuxChannArrayIndex+1);  
  FileName = Job + C_String;

  oElecSignalStat->WriteFileEyeDiagram(FileName); 

  EyeDiagramMonteCarlo->UpdateEyeDiagram(SingleChannelSignal,oPhotodetector);

  sprintf(C_String,"InEyeHistogram_%i.dat",DemuxChannArrayIndex+1); 
  FileName = Job + C_String;

  EyeDiagramMonteCarlo->WriteEyeDiagramHistogram(FileName);
  EyeDiagramMonteCarlo->Clear();

  LogFileSeparator();

  double AmplitudeMarginBtoB = oElecSignalStat->GetAmplitudeMargin();

   cout << "In Array Channel " << DemuxChannArrayIndex +1 << endl;
	
   cout 
	<< "Margin = " << oElecSignalStat->GetAmplitudeMargin() << endl  
	<< "Q factor = " << oElecSignalStat->GetQ_Factor() << endl
	<< "Space mean = " << oElecSignalStat->GetBit0_PowerMean() << endl
	<< "Mark mean = " << oElecSignalStat->GetBit1_PowerMean() << endl
        << "Space StdDev = " << oElecSignalStat->GetBit0_PowerStdDev() << endl
        << "Mark StdDev = " << oElecSignalStat->GetBit1_PowerStdDev() << endl
        << "Normalized Space StdDev = " 
        << oElecSignalStat->GetBit0_PowerStdDev()/
           (oElecSignalStat->GetBit1_PowerMean()- 
            oElecSignalStat->GetBit0_PowerMean()) << endl
        << "Normalized Mark StdDev = " 
        << oElecSignalStat->GetBit1_PowerStdDev()/
           (oElecSignalStat->GetBit1_PowerMean()- 
            oElecSignalStat->GetBit0_PowerMean()) << endl
	<< "Power mean = " << oElecSignalStat->GetMeanPower() <<endl
        << "Num Zeros = " <<  oElecSignalStat->GetQtBit0() << endl
        << "Num Ones = " <<  oElecSignalStat->GetQtBit1() << endl
        << "SNR_Time[dB] = " << Linear2dB(oElecSignalStat->GetBit1_PowerMean()/
                            oElecSignalStat->GetBit0_PowerMean())  << endl;

   oElecSignalStat->ClearElecSignalStat();

} // ## end ReceiveSignalBackToBackMonteCarlo

// #########################################################

void ReceiverSemiAnalytical::ReceiveSignalMonteCarlo(int MapNum,
			      int MonteCarloSystemRealizationNum)
{
  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex); 
  oPhotodetector->DetectOptSignal();  
  oElecFilter->FilterElecSignal();
  oElecSignalStat->UpdatePerformanceStatistics();

  EyeDiagramMonteCarlo->UpdateEyeDiagram(SingleChannelSignal,oPhotodetector);

} // ## end ReceiveSignalMonteCarlo

// #########################################################

void ReceiverSemiAnalytical::ReceiveSignalSemiAnalytical(string FileID)
{

  /* The following call to PrepareToDemuxChannel
     sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver
  */

  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

  //WriteNoiseFreeOpticalSignalFreq(Job+FileID);

  // The following call does not allocate any new memory
  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();

  double TotalNoiseSpectralDensity = 
  SingleChannelSignal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics();

  cout << "TotalNoiseSpectralDensity [W/Hz] = " 
       << TotalNoiseSpectralDensity << endl;

 string OutFileName = Job + FileID + "TotalNoiseSpectralDensity.dat";

 ofstream OutFileN_ASE(OutFileName.c_str(),ios::out);

  if (!OutFileN_ASE)
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

 OutFileN_ASE << "%TotalNoiseSpectralDensity [W/Hz] " << endl
              << TotalNoiseSpectralDensity << endl;

 OutFileN_ASE.close();

 
  // We are assuming that noise is unpolarized here.

  double NoiseSpectralDensityX, NoiseSpectralDensityY;

  if(SingleChannelSignal->GetTypeSimulation()==SCALAR)
    {
      NoiseSpectralDensityX = TotalNoiseSpectralDensity;
      NoiseSpectralDensityY = 0.0;
    }
  else
    {
      NoiseSpectralDensityX = TotalNoiseSpectralDensity/2.0;
      NoiseSpectralDensityY = TotalNoiseSpectralDensity/2.0;
 
    }

 
 double MeanNoise = oElecSignalStat->GetMeanASE_ASE(NoiseSpectralDensityX,
						   NoiseSpectralDensityY);

 double StdDev_ASE_ASE = 
   oElecSignalStat->GetStdDevASE_ASE(NoiseSpectralDensityX,
				    NoiseSpectralDensityY);
 
 for(int jj=0;jj < SingleChannelSignal->GetQtPoints(); jj++)
   {
     
    StdDev_Signal_ASE[jj] = oElecSignalStat->GetStdDevSignal_ASE_TimeIndex(jj,
	   NoiseSpectralDensityX,NoiseSpectralDensityY);
     
   }


  
  OutFileName = Job + FileID + "StdDevCurrent.dat";

  ofstream OutFile(OutFileName.c_str(),ios::out);
  
  if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int jj=0;jj< SingleChannelSignal->GetQtPoints()  ;jj++)
    {
      OutFile <<  jj*SingleChannelSignal->GetDeltaTime() << "\t" 
	      << sqrt( sq(StdDev_ASE_ASE) + sq(StdDev_Signal_ASE[jj])) << endl;
    }

  OutFile.close();

  // In case we have changed the data in SingleChannelSignal in the 
  //   previous function calls we re-initialize the SingleChannelSignal
  //   from the WDM signal, and we then apply the demuxing filter to it.
  

  oOptDemuxer->DemuxChannel(DemuxChannArrayIndex);

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();

  OutFileName = Job + FileID + "MeanCurrent.dat";

  ofstream OutFile2(OutFileName.c_str(),ios::out);

  if (!OutFile2) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


  for(int jj=0;jj<SingleChannelSignal->GetQtPoints();jj++)
    {
      OutFile2 <<  jj*SingleChannelSignal->GetDeltaTime() << "\t" 
	      <<  MeanNoise + real(oPhotodetector->sfftPM.tPowerSignal[jj])
	      <<  endl;
    }

  OutFile2.close();

  OutFileName = Job + FileID + "Clock.dat";

  // Need the following lines to get clock time and 
  // then to undo Brian's time shift so that MeanCurrent and StdDevCurrent
  // are lined up in time

  oElecSignalStat->ComputePerformanceStatistics();
 //oPhotodetector->TimeShiftSignal(oElecSignalStat->GetClockRecoveryTimeNEW());

  ofstream OutFile3(OutFileName.c_str(),ios::out);

  if (!OutFile3) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

  OutFile3 << "% (1) ClockTime [ps] (2) ClockCurrent [mV]" << endl
	   <<  1e12*oElecSignalStat->GetClockRecoveryTime() << "\t" 
	   <<  oElecSignalStat->GetClockCurrent() << endl;
 
  OutFile3.close();


} // ## end ReceiveSignalSemiAnalytical

// #########################################################

void ReceiverSemiAnalytical::PostProcessReceiverStats(
                              int MapNum,
			      int MonteCarloSystemRealizationNum,
			      double PropagatedLength,
			      int AppendFlag)
{

   char C_String[80];

   sprintf(C_String,"EyeMC_%iRT_%iCh_%i.dat",
         MonteCarloSystemRealizationNum, MapNum,DemuxChannArrayIndex+1);
   
   string FileName = Job + C_String;

   EyeDiagramMonteCarlo->WriteEyeDiagramHistogram(FileName); 
 
   WriteStats(MapNum,MonteCarloSystemRealizationNum,PropagatedLength,
	      AppendFlag);

} // ## end PostProcessReceiverStats

// #############################################################

void ReceiverSemiAnalytical::WriteStats(int MapNum,
					int MonteCarloSystemRealizationNum,
					double PropagatedLength,
					int AppendFlag)
{


 char C_String[40];

 sprintf(C_String,"StatsMC_%iCh_%i.dat",
	 MonteCarloSystemRealizationNum,DemuxChannArrayIndex+1);

 string FileName = Job + C_String;

 ofstream OutFile;

 if(AppendFlag)
  OutFile.open(FileName.c_str(),ios::app);
 else
   OutFile.open(FileName.c_str(),ios::out);
 
  if (!OutFile) 
   {
      cerr << "Error: " << FileName << "could not be created."<<endl;
      exit(1);
   }

if(!AppendFlag)
  {
    OutFile << "%(1) PropagatedLength [km]  (2) AmplitudeMargin (3) Q_Factor"
	    << endl 
	    <<"%(4) BitErrorRate (5) Bit0_PowerMean (6) Bit1_PowerMean" 
	    << endl 
	    << "%(7) SpaceStdDev (8) MarkStdDev (9) NormalizedSpaceStdDev"
	    << endl 
            << "%(10) NormalizedMarkStdDev (11) MeanPower "
	    << endl 
	    << "%(12) QtBit0 (13) QtBit1 (14) SNR_Time[dB]" << endl
	    << "%(15) ClockTime [ps] (16) ClockCurrent [mV]" << endl;


  }
 
  OutFile << int(PropagatedLength/1000.0) << "\t" 
          <<oElecSignalStat->GetAmplitudeMargin() << "\t"
          << oElecSignalStat->GetQ_Factor() << "\t"
          << oElecSignalStat->GetBitErrorRate() << "\t"
          << oElecSignalStat->GetBit0_PowerMean() << "\t"
	  << oElecSignalStat->GetBit1_PowerMean() << "\t"
          << oElecSignalStat->GetBit0_PowerStdDev() <<  "\t"
          << oElecSignalStat->GetBit1_PowerStdDev() <<  "\t"
          << oElecSignalStat->GetBit0_PowerStdDev()/
             (oElecSignalStat->GetBit1_PowerMean()- 
              oElecSignalStat->GetBit0_PowerMean()) << "\t"
          << oElecSignalStat->GetBit1_PowerStdDev()/
             (oElecSignalStat->GetBit1_PowerMean()- 
              oElecSignalStat->GetBit0_PowerMean()) << "\t"
          << oElecSignalStat->GetMeanPower() << "\t" 
	  << oElecSignalStat->GetQtBit0() << "\t"
	  << oElecSignalStat->GetQtBit1() << "\t"
          << Linear2dB(oElecSignalStat->GetBit1_PowerMean()/
                            oElecSignalStat->GetBit0_PowerMean())  << "\t"
	  << oElecSignalStat->GetClockTimeForMonteCarloEye() << "\t"
	  << oElecSignalStat->GetClockCurrent() << endl;

  
  OutFile.close();

} //## end WriteStats

// ############################################################

//! Writes the noise free optical signal in frequency before optical filtering
//! to a file. Useful as input to a SCALAR Chi-square receiver model.
//! Not implemented for VECTOR simulations yet. JZ 12/11/03 

void ReceiverSemiAnalytical::WriteNoiseFreeOpticalSignalFreq(string OutFileName)
{

 // Not implemented for VECTOR simulations yet!

 if(WDMSignal->GetTypeSimulation() == SCALAR)
  {  

   OutFileName += "fNoiseFreeSignal.dat";

   ofstream File(OutFileName.c_str(),ios::out);

   if (!File)
    {
      cerr << "Error: " << OutFileName
           << "could not be created. ABORTING"<<endl;
      exit(1);
    }

   File
     << "% (1) Freq [GHz] (2) Power [W] (3) Real Part (4) Imag Part"
     << endl;

  for(int jj=0;jj<WDMSignal->GetQtPoints();jj++)
     {

       File
         << 1e-9*jj*WDMSignal->GetDeltaFreq() << "\t"
         << sq(abs(WDMSignal->sfftM.fOptSignalX[jj])) << "\t"
         << real(WDMSignal->sfftM.fOptSignalX[jj]) << "\t"
         << imag(WDMSignal->sfftM.fOptSignalX[jj]) << endl;
     }

   File.close();

  } // end if

} // ## end WriteNoiseFreeOpticalSignal


// #########################################################

void ReceiverSemiAnalytical::GetLimaParameters(bool AppendOutFileFlag,
                                               double FreqFWHM_OptFilter, 
					       double FreqFWHM_ElecFilter,
					       double OSA_Bandwidth)
{

  /* This call sets the number of channels of the SingleChannelSignal
     to 1, reduces the WDM BitString to a SingleChannel BitString,
     sets the carrier frequency of the only channel in the SingleChannel 
     signal and sets the center frequency of the optical filter. It does
     not do the actual filtering since we need to do that in the 
     semianalytical receiver
  */


  oOptDemuxer->PrepareToDemuxChannel(DemuxChannArrayIndex);

  //WriteNoiseFreeOpticalSignalFreq(Job+FileID);

  // The following call does not allocate any new memory
  
  oElecSignalStat->ResetOptElecFiltersForGaussianStatistics();

  oElecSignalStat->ResetOptElecFilterBandwidthsForGaussianStatistics(
                                  FreqFWHM_OptFilter,FreqFWHM_ElecFilter); 
  
  // We need the signal to be noiseless here. It can either be 
  // a SCALAR or VECTOR signal.

  // The coefficients are computed using the worst mark and worst space

 

  double K1 = oElecSignalStat->GetNormKcoeff1_GaussianStatistics(); 
  double K0 = oElecSignalStat->GetNormKcoeff0_GaussianStatistics();
  double M = oElecSignalStat->GetNormNumberNoiseModes_GaussianStatistics(); 
  double alpha_e = oElecSignalStat->
    GetCurrentExtinctionRatio_GaussianStatistics();
 
  double EnhancementFactor = oElecSignalStat->GetEnhancementFactor(
                                                 OSA_Bandwidth); 
  double NormalizedEnhancementFactor = 
          oElecSignalStat->GetNormalizedEnhancementFactor();

  string OutFileName = Job + "LimaParameters.dat";
  
  ofstream OutFile;

  if(!AppendOutFileFlag)
    {
     
     OutFile.open(OutFileName.c_str(),ios::out);

     if (!OutFile) 
      {
       cerr << "Error: " << OutFileName << "could not be created."<<endl;
       exit(1);
      }

     OutFile << "% (1) Optical FWHM [GHz] (2) Electrical FWHM [GHz] " << endl
	     << "%(3) M (4) K0 (5) K1 (6) NormalizedEnhancementFactor" << endl
	     << "%(7) EnhancementFactor (8) alpha_e [dB] " << endl;
    }
  else
    {
     OutFile.open(OutFileName.c_str(),ios::app);

     if (!OutFile) 
      {
       cerr << "Error: " << OutFileName << "could not be created."<<endl;
       exit(1);
      }

    } // end else

  OutFile << 1e-9*FreqFWHM_OptFilter << "\t" 
	  << 1e-9*FreqFWHM_ElecFilter << "\t"
          << M << "\t" << K0 << "\t" << K1 << "\t" 
	  << NormalizedEnhancementFactor << "\t"
	  << EnhancementFactor << "\t" << Linear2dB(alpha_e) << endl;

  OutFile.close();

} // ## end GetLimaParameters

// #########################################################







