// $Id: ocsReceiverSubsystem.cc,v 1.3 2008/08/08 02:28:04 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReceiverSubsystem.hh"

extern ofstream *LogFile;

// Written by John Zweck, Nov 21, 2000
// This class is designed to make it easy to
// receive signals and statistically analyse them in a WDM system

// For each channel which you want to demux and collect data
// on, you need a separate ReceiverSubsystem object.

ReceiverSubsystem::ReceiverSubsystem()
{
  // Use this constructor followed by calls to ReceiverSubsystemInit
  // if you are defining an array of ReceiverSubsystems

  ObjectInitialized = 0;
}

// ###############################################

ReceiverSubsystem::ReceiverSubsystem(OptSignal * WDMSignal,RanNumGen *RNG2,
                                     string InDir, string Job2,
                                     int DemuxChannArrayIndex2,
                                     OptDataWriter * DataWriter2)
{

ReceiverSubsystemInit(WDMSignal,RNG2,InDir,Job2,DemuxChannArrayIndex2,
                      DataWriter2);

}



//###########################################################
//## Simplified receiver
//## By: Ivan Lima
ReceiverSubsystem::ReceiverSubsystem(OptSignal *WDMSignal, 
                    int DemuxChannArrayIndex2,
                    string OptFilterFileName,
                    string ElecFilterFileName,
                    string PhotodetectorFileName, 
                    string ElecSignalStatFileName)
{ 
   InitSimplifiedReceiver(WDMSignal, 
                              DemuxChannArrayIndex2,
                              OptFilterFileName,
                              ElecFilterFileName,
                              PhotodetectorFileName,
                              ElecSignalStatFileName);
}



// ########################################

void ReceiverSubsystem::ReceiverSubsystemInit(OptSignal * WDMSignal,
					      RanNumGen *RNG2,
                                     string InDir, string Job2,
                                     int DemuxChannArrayIndex2,
                                     OptDataWriter * DataWriter2)
{
  ObjectInitialized = 1;
  FlagSimplifiedReceiver = 0;
  LogFileSeparator();

  *LogFile << "ReceiverSubsystem Constructor" << endl;

  DataWriter = DataWriter2;
  Job = Job2;
  PostDispCompFileNameSuffix = "0";

  DemuxChannArrayIndex =  DemuxChannArrayIndex2;

  *LogFile << "DemuxChannArrayIndex = " << DemuxChannArrayIndex << endl;

  TunableFilter = new OptFilter(InDir + "TunableOptFilter.in",
				WDMSignal->GetTypeSimulation());

  oOptDemuxer = new OptDemuxer(WDMSignal, TunableFilter);

  SCSignal = new OptSignal(oOptDemuxer->ExtractChannel(DemuxChannArrayIndex));

    // Here we are using the OptSignal copy constructor
    // The data in SCSignal need not be correct at this stage.
    // We just need to allocate fixed memory for it, so
    // that the components initialized below know where to look for the 
    // SCSignal data when their methods are called

  oOptFilter = new OptFilter(InDir + "OptFilter.in",SCSignal);
  //  oOptFilterPD = new OptFilter(InDir + "OptFilterPD.in",SCSignal);

  // Note that oOptFilter may not be necessary since we are already using
  // a TunableFilter.

  /******
  oPostDispComp = new PostDispCompSubSystem(InDir + "PostDispComp.in",
                                         SCSignal,RNG2,InDir,Job,
                                         DemuxChannArrayIndex,DataWriter);
  *******/         

  oPhotodetector = new Photodetector (InDir + "Photodetector.in",SCSignal,
                                      RNG2);
  oElecFilter = new ElecFilter(InDir + "ElecFilter.in",oPhotodetector);

  //oElecFilter->WriteFreqCplxFilter(Job+"CplxElecFilter.dat");

  oElecSignalStat = new  ElecSignalStat(SCSignal,oPhotodetector);
			   
  RonEyeDiagram = new  Histogram2D(InDir + "Histogram2D.in", SCSignal);

  LogFileSeparator();
  *LogFile << "ReceiverSubsystem Constructor" << endl;

}

// ## end ReceiverSubsystemInit

// ######################################################

void ReceiverSubsystem::ReceiverSubsystemInit(OptSignal * WDMSignal,
					      RanNumGen *RNG2,
                                     string InDir, string Job2, 
                                     int DemuxChannArrayIndex2,
                                     OptDataWriter * DataWriter2,
                                     string PostDispCompFileNameSuffix2,
                                     fftPowerMatrices & sfftPM2,
                                     fftPowerMatrices & WorksfftPM2,
				     OptSignal * GlobalSCSignal,
                                     OptSpectrumAnalyzer * PowerMeter2,
                                     OptDemuxer *Demuxer2,
				     cplx * MeasuredFilterFreq)
{

  ObjectInitialized = 1;
  FlagSimplifiedReceiver = 0;
  LogFileSeparator();

  *LogFile << "ReceiverSubsystem Constructor" << endl;

  DataWriter = DataWriter2;
  Job = Job2;
  DemuxChannArrayIndex =  DemuxChannArrayIndex2;

  *LogFile << "DemuxChannArrayIndex = " << DemuxChannArrayIndex << endl;

  TunableFilter = new OptFilter(InDir + "TunableOptFilter.in",
				GlobalSCSignal->GetTypeSimulation());
  oOptDemuxer = Demuxer2;
  
  SCSignal = GlobalSCSignal;

    // Here we are using the OptSignal copy constructor
    // The data in SCSignal need not be correct at this stage.
    // We just need to allocate fixed memory for it, so
    // that the components initialized below know where to look for the 
    // SCSignal data when their methods are called

  oOptFilter = new OptFilter(InDir + "OptFilter.in",SCSignal);
  //  oOptFilterPD = new OptFilter(InDir + "OptFilterPD.in",SCSignal);

  // Note that oOptFilter may not be necessary since we are already using
  // a TunableFilter.

  PostDispCompFileNameSuffix = PostDispCompFileNameSuffix2;
  /******
  oPostDispComp = new PostDispCompSubSystem(
             InDir + "PostDispComp" + PostDispCompFileNameSuffix + ".in",
                                         SCSignal,RNG2,InDir,Job,
	                                 DemuxChannArrayIndex,DataWriter,
					 PowerMeter2);
  ******/
                

  oPhotodetector = new Photodetector (InDir + "Photodetector.in",SCSignal,
                                      RNG2,sfftPM2,WorksfftPM2);

  bool ReadMeasuredFilterFlag = 0;

  oElecFilter = new ElecFilter(InDir + "ElecFilter.in",oPhotodetector,
                               MeasuredFilterFreq,ReadMeasuredFilterFlag);

  oElecSignalStat = new  ElecSignalStat(SCSignal,oPhotodetector);
  RonEyeDiagram = new  Histogram2D(InDir + "Histogram2D.in", SCSignal);

  LogFileSeparator();
  *LogFile << "ReceiverSubsystem Constructor" << endl;

}



// ## end ReceiverSubsystemInit

// ######################################################

ReceiverSubsystem::~ReceiverSubsystem()
{
  if (ObjectInitialized) {
     if (! FlagSimplifiedReceiver) {
        TunableFilter->~OptFilter();
        //oPostDispComp->~PostDispCompSubSystem();     
     }
     oOptDemuxer->~OptDemuxer();
     SCSignal->~OptSignal();
     oOptFilter->~OptFilter();
     oPhotodetector->~Photodetector();
     oElecFilter->~ElecFilter();
     oElecSignalStat->~ElecSignalStat();
     RonEyeDiagram->~Histogram2D();
     // oOptFilterPD->~OptFilter();
  }   

} // ## end ~ReceiverSubsystem()

// ########################################

void ReceiverSubsystem::ReceiveSignalBackToBack(OptSignal * WDMSignal) 
{

   LogFileSeparator();
   *LogFile << "ReceiveSignalBackToBack Channel " << DemuxChannArrayIndex + 1
           << endl;

   *SCSignal = oOptDemuxer->ExtractChannel(DemuxChannArrayIndex);

    
   /**************
     if(WDMSignal->GetNumChannels() > 1)
     {
      *SCSignal = oOptDemuxer->ExtractChannel(emuxChannArrayIndex);
     }
   else // NumChannels == 1, so no need to demux. 
        // NB Later may want to OptFilter for ASE!!!
     {
      *SCSignal = *WDMSignal;
     }

     cout << "ReceiverSubsystem::ReceiveSignalBackToBack TEMP DEMUX CHANGE" 
       << endl;
   **********/
   

  char  C_String2[40];

  sprintf(C_String2,"tInOD_%i.dat",DemuxChannArrayIndex+1);
  string FileName2 = Job + C_String2;

  //SCSignal->WriteFileTimePower(FileName2);

  sprintf(C_String2,"fInOD_%i.dat",DemuxChannArrayIndex+1);
  FileName2 = Job + C_String2;

  //SCSignal->WriteFileFreqPower(FileName2,0);

  if(oOptDemuxer->GetDoAdditionalFiltering())
    {
      cout << "DOING ADD FILTER ABORT" << endl;
      exit(1);

     double RelativeCenterFreq = WDMSignal->GetFrequency(DemuxChannArrayIndex)
                             - WDMSignal->GetCenterFreq();

     oOptFilter->SetRelativeCenterFreq(RelativeCenterFreq);
     oOptFilter->FilterOptSignal(); // Additional Demux Filtering

     //oOptFilterPD->SetRelativeCenterFreq(RelativeCenterFreq);
     //oOptFilterPD->TunableFilterOptSignal(); // Additional Demux Filtering

    }

  
  oPhotodetector->DetectOptSignal();   
  
   char  C_String[40];
   sprintf(C_String,"fInBF_%i.dat",DemuxChannArrayIndex+1);
   string FileName = Job + C_String;
   //oPhotodetector->WriteFileFreqPower(FileName,0);
   
   sprintf(C_String,"tInBF_%i.dat",DemuxChannArrayIndex+1);
   FileName = Job + C_String;
   //oPhotodetector->WriteFileTimePower(FileName);


  oElecFilter->FilterElecSignal();     

//  oElecSignalStat->GetSingleStringElecSignalStatConvolutingNEW();  
//  oPhotodetector->TimeShiftSignal(oElecSignalStat->GetClockRecoveryTimeNEW());

  oElecSignalStat->ComputePerformanceStatistics();
  oPhotodetector->TimeShiftSignal(oElecSignalStat->GetClockRecoveryTime());


  PropagatedLength = 0.0;

  //char  C_String[40];

  sprintf(C_String,"InED_%i.dat",DemuxChannArrayIndex+1);
  //string FileName = Job + C_String;
  FileName = Job + C_String;
  oElecSignalStat->WriteFileEyeDiagram(FileName); 

  RonEyeDiagram->UpdateEyeDiagram(SCSignal,oPhotodetector);

  sprintf(C_String,"EyeMC_%iRT_%iCh_%i.dat",
                    0,0,DemuxChannArrayIndex+1);
   
   FileName = Job + C_String;

   RonEyeDiagram->WriteEyeDiagramHistogram(FileName);
   RonEyeDiagram->Clear();

   /***
   sprintf(C_String,"tReceivedCurrent_%i.dat",DemuxChannArrayIndex+1);
   FileName = Job + C_String;
   oPhotodetector->WriteFileTimePower(FileName);
   ***/

  // ## Get Back to Back Signal data

   LogFileSeparator();

   

   double AmplitudeMarginBtoB = oElecSignalStat->GetAmplitudeMargin();

   cout << "In Array Channel " << DemuxChannArrayIndex +1 << endl;
	
   cout << "Distance = " << int(PropagatedLength/1000.0) << endl 
	<< "Margin = " << oElecSignalStat->GetAmplitudeMargin() << endl  
	<< "Q factor = " << oElecSignalStat->GetQ_Factor() << endl
	<< "Space mean = " << oElecSignalStat->GetBit0_PowerMean() << endl
	<< "Mark mean = " << oElecSignalStat->GetBit1_PowerMean() << endl
        << "Space StdDev = " << oElecSignalStat->GetBit0_PowerStdDev() << endl
        << "Mark StdDev = " << oElecSignalStat->GetBit1_PowerStdDev() << endl
        << "Normalized Space StdDev = " 
        << oElecSignalStat->GetBit0_PowerStdDev()/
           (oElecSignalStat->GetBit1_PowerMean()- 
            oElecSignalStat->GetBit0_PowerMean()) << endl
        << "Normalized Mark StdDev = " 
        << oElecSignalStat->GetBit1_PowerStdDev()/
           (oElecSignalStat->GetBit1_PowerMean()- 
            oElecSignalStat->GetBit0_PowerMean()) << endl
	<< "Power mean = " << oElecSignalStat->GetMeanPower() <<endl
        << "Num Zeros = " <<  oElecSignalStat->GetQtBit0() << endl
        << "Num Ones = " <<  oElecSignalStat->GetQtBit1() << endl
        << "SNR_Time[dB] = " << Linear2dB(oElecSignalStat->GetBit1_PowerMean()/
                            oElecSignalStat->GetBit0_PowerMean())  << endl;

   oElecSignalStat->ClearElecSignalStat();

   sprintf(C_String,"tInPD_%i.dat",DemuxChannArrayIndex+1);
   FileName = Job + C_String;
   //oPhotodetector->WriteFileTimePower(FileName); 

   sprintf(C_String,"fInPD_%i.dat",DemuxChannArrayIndex+1);
   FileName = Job + C_String;
   //oPhotodetector->WriteFileFreqPower(FileName,0);
  
} // ## end ReceiveSignalBack2Back

// ########################################

void ReceiverSubsystem::ReceiveSignal(OptSignal * WDMSignal, 
                                      int UpdateIvansEDFlag,
                                      int OutputPD_DataFlag, 
                                      int ExptNum, int MapNum,
                                      double PropagatedLength2)
{

  /**********
   LogFileSeparator();
  
   *LogFile << "ReceiveSignal Channel " << DemuxChannArrayIndex + 1
           << endl;
  *********/

   PropagatedLength = PropagatedLength2;

   // WDMSignal->oBitString->WriteBitString();

   /*
   if(WDMSignal->GetNumChannels() > 1)
     {
      *SCSignal = oOptDemuxer->ExtractChannel(DemuxChannArrayIndex);
     }
   else // NumChannels == 1, so no need to demux. 
        // NB Later may want to OptFilter for ASE!!!
     {
      *SCSignal = *WDMSignal;
     }

   */

*SCSignal = oOptDemuxer->ExtractChannel(DemuxChannArrayIndex);

  //cout << "Channel " <<  DemuxChannArrayIndex + 1 << endl;
  //SCSignal->oBitString->WriteBitString();

  
     
 if(oOptDemuxer->GetDoAdditionalFiltering())
    {

      cout << "oOptDemuxer->GetDoAdditionalFiltering() ABORT" << endl;
      exit(1);
     double RelativeCenterFreq = WDMSignal->GetFrequency(DemuxChannArrayIndex)
                             - WDMSignal->GetCenterFreq();

     oOptFilter->SetRelativeCenterFreq(RelativeCenterFreq);
     oOptFilter->FilterOptSignal(); // Additional Demux Filtering

     //oOptFilterPD->SetRelativeCenterFreq(RelativeCenterFreq);
     //oOptFilterPD->TunableFilterOptSignal(); // Additional Demux Filtering

    }

 /******
      oPostDispComp->SetRecMapNum(RecMapNum);
      oPostDispComp->ApplyPostDispComp();
 ******/

     // cout << "PostDispComp in Receiversubsystem!!" << endl;

  if(OutputPD_DataFlag)  
      {

	// Note that the way I have this coded if you output this data
	// for multiple expts then the files is written over.

         char C_String[40];

         sprintf(C_String,"tOutPDC_Map_%iCh_%i.dat",MapNum,
                 DemuxChannArrayIndex+1);

         string FileName = Job + C_String;
	 //SCSignal->WriteFileTimePower(FileName); 
      }

  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();     
  oElecSignalStat->UpdatePerformanceStatistics();
  oPhotodetector->TimeShiftSignal(oElecSignalStat->GetClockRecoveryTime());
// Update Ivan's Eyediagram

  if(UpdateIvansEDFlag) 
    {   
      char C_String[40];

      sprintf(C_String,"ED_Map_%iCh_%i.dat",MapNum,DemuxChannArrayIndex+1);
      string FileName = Job + C_String;
      oElecSignalStat->WriteFileEyeDiagram(FileName); 
 
    }// end if
            
      
    // Update Ron's Eyediagram
  // Uncomment this line if you want Ron's ED
  RonEyeDiagram->UpdateEyeDiagram(SCSignal,oPhotodetector);

  /******
   char C_String[40];

   sprintf(C_String,"tReceivedCurrent_%iCh_%i.dat",MapNum,
                 DemuxChannArrayIndex+1);

   string FileName = Job + C_String;
   oPhotodetector->WriteFileTimePower(FileName);
  ******/

    if(OutputPD_DataFlag)  
      {

	// Note that the way I have this coded if you output this data
	// for multiple expts then the files is written over.

         char C_String[40];

         sprintf(C_String,"tOutPD_Map_%iCh_%i.dat",MapNum,
                 DemuxChannArrayIndex+1);

         string FileName = Job + C_String;
	 //     oPhotodetector->WriteFileTimePower(FileName); 

         sprintf(C_String,"fOutPD_Map_%iCh_%i.dat",MapNum,
                           DemuxChannArrayIndex+1);
         FileName = Job + C_String;

	 // DONT   oPhotodetector->WriteFileFreqPower(FileName,0);

      } // end if 


} // ## end ReceiveSignal

// ##########################################

void ReceiverSubsystem::PostProcessReceiverStats(int ExptNum,int MapNum,
                                                 int AppendFlag)
{
  /************
   LogFileSeparator();
   *LogFile << "PostProcessReceiverStats" << endl;
  *********/

   char C_String[80];

   sprintf(C_String,"EyeMC_%iRT_%iCh_%i.dat",
                     ExptNum, MapNum,DemuxChannArrayIndex+1);
   
   string FileName = Job  + PostDispCompFileNameSuffix + "_"+ C_String;

   RonEyeDiagram->WriteEyeDiagramHistogram(FileName);
 
    // ## Creating the file with the eye diagram
           

   WriteStats(ExptNum,MapNum,AppendFlag);

} // ## end  PostProcessReceiverStats

// ########################################


void ReceiverSubsystem::WriteStats(int ExptNum, int MapNum, int AppendFlag)
{

 char C_String[40];

 sprintf(C_String,"StatsMC_%iCh_%i.dat",ExptNum,DemuxChannArrayIndex+1);
 string FileName = Job +  PostDispCompFileNameSuffix + "_" + C_String;

 ofstream OutFile;

 if(AppendFlag)
  OutFile.open(FileName.c_str(),ios::app);
 else
   OutFile.open(FileName.c_str(),ios::out);
 
  if (!OutFile) 
   {
      cerr << "Error: " << FileName << "could not be created."<<endl;
      exit(1);
   }

if(!AppendFlag)
  {
    OutFile << "%(1) PropagatedLength [km]  (2) AmplitudeMargin (3) Q_Factor"
	    << endl 
	    <<"%(4) BitErrorRate (5) Bit0_PowerMean (6) Bit1_PowerMean" 
	    << endl 
	    << "%(7) SpaceStdDev (8) MarkStdDev (9) NormalizedSpaceStdDev"
	    << endl 
            << "%(10) NormalizedMarkStdDev (11) MeanPower "
	    << endl 
	    << "%(12) QtBit0 (13) QtBit1 (14) SNR_Time[dB]" << endl
	    << "%(15) ClockTime [ps] (16) ClockCurrent [mV]" << endl;


  }
 
  OutFile << int(PropagatedLength/1000.0) << "\t" 
          <<oElecSignalStat->GetAmplitudeMargin() << "\t"
          << oElecSignalStat->GetQ_Factor() << "\t"
          << oElecSignalStat->GetBitErrorRate() << "\t"
          << oElecSignalStat->GetBit0_PowerMean() << "\t"
	  << oElecSignalStat->GetBit1_PowerMean() << "\t"
          << oElecSignalStat->GetBit0_PowerStdDev() <<  "\t"
          << oElecSignalStat->GetBit1_PowerStdDev() <<  "\t"
          << oElecSignalStat->GetBit0_PowerStdDev()/
             (oElecSignalStat->GetBit1_PowerMean()- 
              oElecSignalStat->GetBit0_PowerMean()) << "\t"
          << oElecSignalStat->GetBit1_PowerStdDev()/
             (oElecSignalStat->GetBit1_PowerMean()- 
              oElecSignalStat->GetBit0_PowerMean()) << "\t"
          << oElecSignalStat->GetMeanPower() << "\t" 
	  << oElecSignalStat->GetQtBit0() << "\t"
	  << oElecSignalStat->GetQtBit1() << "\t"
          << Linear2dB(oElecSignalStat->GetBit1_PowerMean()/
                            oElecSignalStat->GetBit0_PowerMean())  << "\t"
	  << 1e12*oElecSignalStat->GetClockRecoveryTime() << "\t"
	  << oElecSignalStat->GetClockCurrent() << endl;

  
  OutFile.close();

}



//############################################################
//## Methods for the simplified version of the receiver
//#  These methods do no produce pos-compensation and do 
//#  not spool files. Thus, it is up to the user to produce them,
//#  whenever they need it. These modules were written
//#  by: Ivan Lima, based on modules written by John Zweck, above.
  
//############################################################
//## Initialization module of the simplified receiver.
//#  The constructor is defined in the header file. 
//## By: Ivan Lima 
void ReceiverSubsystem::InitSimplifiedReceiver(OptSignal  *WDMSignal, 
                                         int DemuxChannArrayIndex2,
                                         string OptFilterFileName,
                                         string ElecFilterFileName,                                         
                                         string PhotodetectorFileName, 
                                         string ElecSignalStatFileName)
{                                         
   FlagSimplifiedReceiver = 1;
   LogFileSeparator();
   *LogFile << "ReceiverSubsystem Constructor" << endl;
   DemuxChannArrayIndex =  DemuxChannArrayIndex2;

   *LogFile << "DemuxChannArrayIndex = " << DemuxChannArrayIndex << endl;
   WDM_OrigSignal = WDMSignal; 
   SCSignal = new OptSignal(*WDM_OrigSignal);  
   oOptFilter = new OptFilter(OptFilterFileName, SCSignal);
   oOptDemuxer = new OptDemuxer(WDM_OrigSignal,oOptFilter, SCSignal );   
   oPhotodetector = new Photodetector (PhotodetectorFileName, SCSignal);
   oElecFilter = new ElecFilter(ElecFilterFileName, oPhotodetector);
   oElecSignalStat = new  ElecSignalStat(ElecSignalStatFileName, SCSignal, 
                                         oPhotodetector);                                  

  //## Allow the use of the semi-analytical Statistics
  //## By: Ivan Lima (6/8/02)
  oElecSignalStat->SetOptElecFiltersForGaussianStatistics(
                           oOptFilter, oElecFilter);					 
					 
}                                               


//################################################################
//## Extract the channel defined in the initialization
//## By: Ivan Lima
void ReceiverSubsystem::ReceiveSingleChannel(void)
{
   ReceiveSingleChannel(DemuxChannArrayIndex);
}


//################################################################                                                
//##  Extract the specified channel
//## By: Ivan Lima 
void ReceiverSubsystem::ReceiveSingleChannel(int DemuxChannArrayIndex2)
{
   DemultiplexSingleChannnel(DemuxChannArrayIndex);
   ReceivePreviouslyDemultiplexedSingleChannnel();
//  oOptDemuxer->ExtractSingleChannel(DemuxChannArrayIndex2);  
//  oPhotodetector->DetectOptSignal();
//  oElecFilter->FilterElecSignal();     
//  //oElecSignalStat->AddSignalStringSampleConvoluting(); 
//  oElecSignalStat->GetSingleStringElecSignalStatConvoluting();
}


//################################################################                                                
//##  Demultiplex the specified channel
//##  It allows, for instance, post dispersion compensation.
//## By: Ivan Lima 
void ReceiverSubsystem::DemultiplexSingleChannnel(int DemuxChannArrayIndex2)
{
  oOptDemuxer->ExtractSingleChannel(DemuxChannArrayIndex2);  
}


//################################################################                                                
//##  Extract the electrical domain signal from a 
//##  previously demultiplexed optical signal.
//## By: Ivan Lima 
void ReceiverSubsystem::ReceivePreviouslyDemultiplexedSingleChannnel(void)
{
  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();     
  oElecSignalStat->ComputePerformanceStatistics();
}

// ############################################################

void ReceiverSubsystem::ReceiveSignalBackToBackCenterFreq(
                                                   OptSignal * WDMSignal,
                                                   double CenterFreqOptFilter,
                                                   BitString * SCBitString,
						   bool OutputEye,
                                                   bool DoDemuxFilter) 
{

  // Written by J Zweck for a specific application.
  // Specifies the channel to be demuxed not by its channel number
  // but by a center freq. 



                           // We still need to use demuxer method 
                         //so as not to loose
                         // optical data which may still propagate.
                         // Also need to set SCBitString correctly.  
  // The  DoDemuxFilter just says whether we optical filter when receive.

 

  //WDMSignal->WriteFileFreqPower(Job + "fPreDemux.dat",0);
  //WDMSignal->WriteFileTimePower(Job + "tPreDemux.dat");

  *SCSignal = oOptDemuxer->ExtractChannelCenterFreq(CenterFreqOptFilter,
						    SCBitString,DoDemuxFilter);

  //SCSignal->WriteFileFreqPower(Job + "fPostDemux.dat",0);
  //SCSignal->WriteFileTimePower(Job + "tPostDemux.dat");

  oPhotodetector->DetectOptSignal();   
  
  oElecFilter->FilterElecSignal();     
  oElecSignalStat->ComputePerformanceStatistics();
 
  PropagatedLength = 0.0;

  if(OutputEye)
    {
     char  C_String[40];

     sprintf(C_String,"InED.dat");
     string FileName = Job + C_String;
     oElecSignalStat->WriteFileEyeDiagram(FileName); 
    }

  bool StatsAppendFlag=0;
  WriteStatsCenterFreq(0,StatsAppendFlag);

  oElecSignalStat->ClearElecSignalStat();

} // ## end ReceiveSignalBack2BackCenterFreq

// ########################################

void ReceiverSubsystem::ReceiveSignalCenterFreq(OptSignal * WDMSignal, 
						double PropagatedLength2,
                                                int MapNum,
                                      double CenterFreqOptFilter,
				      BitString *SCBitString, bool OutputEye,
                                      bool DoDemuxFilter) 
{

  // Written by J Zweck for a VERY specific application.
  // We assume that there is only a single MC expt, but we may want to
  // compute stats for different propagation distances.
  // Specifies the channel to be demuxed not by its channel number
  // but by a center freq. 

  
  *SCSignal = oOptDemuxer->ExtractChannelCenterFreq(CenterFreqOptFilter,
						    SCBitString,DoDemuxFilter);
 


  PropagatedLength = PropagatedLength2;
 
  oPhotodetector->DetectOptSignal();
  oElecFilter->FilterElecSignal();     
  oElecSignalStat->ComputePerformanceStatistics();


  if(OutputEye)
    {

      char C_String[40];

      sprintf(C_String,"OutED_%d.dat",MapNum);
      string FileName = Job + C_String;
      oElecSignalStat->WriteFileEyeDiagram(FileName); 
    }
        
  bool AppendFlag = 1;

  WriteStatsCenterFreq(MapNum,AppendFlag);

  oElecSignalStat->ClearElecSignalStat();

} // ## end ReceiveSignalCenterFreq

// ##########################################

void ReceiverSubsystem::WriteStatsCenterFreq(int MapNum, bool AppendFlag)
{


 char C_String[40];

 sprintf(C_String,"MinAmpMargin.dat");
 string FileName = Job + C_String;

 ofstream OutFile;

 if(AppendFlag)
  OutFile.open(FileName.c_str(),ios::app);
 else
   OutFile.open(FileName.c_str(),ios::out);
 
  if (!OutFile) 
   {
      cerr << "Error: " << FileName << "could not be created."<<endl;
      exit(1);
   }

if(!AppendFlag)
  {
    OutFile << "% (1) MapNum (2) MinAmplitudeMargin (3) MaxZero (4) MinOne" 
            << endl
	    << "% (5) MeanZero (6) MeanOne (7) StdDevZero (8) StdDevOne (9) Q"
	    << endl;
  }


  OutFile << MapNum << "\t" 
          << oElecSignalStat->GetMinAmplitudeMargin() << "\t"
          << oElecSignalStat->GetBit0_PowerMax()   << "\t"
	  << oElecSignalStat->GetBit1_PowerMin()   << "\t"
	  << oElecSignalStat->GetBit0_PowerMean()  << "\t"
	  << oElecSignalStat->GetBit1_PowerMean()  << "\t"
	  << oElecSignalStat->GetBit0_PowerStdDev() << "\t"
	  << oElecSignalStat->GetBit1_PowerStdDev() << "\t"
	  << oElecSignalStat->GetQ_Factor() << endl;

  OutFile.close();

}

// ###################################################













