// $Id: ocsReduLumpedLoss.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReduLumpedLoss.hh"

extern ofstream *LogFile;

//###############################################################
//## By: Ivan Lima (08/17/02)
ReduLumpedLoss::ReduLumpedLoss(double LossdB2, ReduOptSignal *oReduOptSignal2)
{
   InitReduLumpedLoss(LossdB2, oReduOptSignal2);
}


//###############################################################
//## By: Ivan Lima (08/17/02)
ReduLumpedLoss::ReduLumpedLoss(string InFileName, 
                               ReduOptSignal *oReduOptSignal2)
{
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;
   ifstream InFile(InFileName.c_str(),ios::in);    
   if (!InFile) {
      cerr << "Error: Input file "<< InFileName 
           << "could not be opened."<< endl;
      exit(1);
   }
   double LossdB2 = ReadDouble("LossdB"," ",1,-1,NO_BOUNDS,&InFile);
   InFile.close();
   InitReduLumpedLoss(LossdB2, oReduOptSignal2);
} // ## end constructor



//###############################################################
//## By: Ivan Lima (08/17/02)
void ReduLumpedLoss::InitReduLumpedLoss(double LossdB2, 
                                        ReduOptSignal *oReduOptSignal2)
{
   oReduOptSignal = oReduOptSignal2;
   LossdB = LossdB2;
   LossLinear = dB2Linear(LossdB);
   *LogFile << "LossLinear = " << LossLinear << endl;
}



//###########################################################
//## By: Ivan Lima (8/17/02)
void ReduLumpedLoss::ApplyLumpedLoss(void)
{                  
   int NumChannels = oReduOptSignal->GetNumChannels();
   for (int jj = 0; jj < NumChannels; jj++)  {                     
      // ## Actual propagation through the LumpedLoss
      for (int ii = 0; ii < 4; ii++)  {      
         oReduOptSignal->StokesSignal[ii][jj] /= LossLinear; 
         oReduOptSignal->StokesNoise[ii][jj]  /= LossLinear;
      }                            
   } 
} //## end ApplyLumpedLoss

//############################################################


//###########################################################
//## Set a new value for the loss in dB
//## By: Ivan Lima (8/17/02)
void ReduLumpedLoss::SetLossdB(double LossdB2)
{
   LossdB = LossdB2;
   LossLinear = dB2Linear(LossdB);
}
