// $Id: ocsReduReadWriteTools.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsReadWriteTools.hh"


extern ofstream *LogFile;

// ##################################################################

// ## Functions added by John Zweck

// ##############################################################





int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            int DefaultValue,
            ifstream * InFile)
{

  int ParameterRead = 0;
  int ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      ParameterValue = DefaultValue;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


} // ## end ReadInt

// ##########################################################

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
	    long DefaultValue,
            ifstream * InFile)
{

  int ParameterRead = 0;
  long ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      ParameterValue = DefaultValue;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// ##############################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
	    double DefaultValue,
            ifstream * InFile)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      ParameterValue =  DefaultValue;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


}  // ## end ReadDouble

// ##############################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  string DefaultValue,
                  ifstream * InFile)
{
  int ParameterRead = 0;
  string ParameterValue ="";
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      ParameterValue =  DefaultValue;
    }
  
    
      if(*LogFile)
        *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
      else
        cout << endl 
             << ParameterName << " = " << ParameterValue << " " << endl 
             << endl;
    

  return ParameterValue;

} // ## end ReadString

// ##############################################################


// ### Now follow version of ReadInt and ReadDouble without the 
// ### defult value parameter


int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile)
{

  int ParameterRead = 0;
  int ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      return 0;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


} // ## end ReadInt



// #############################################################

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
	    InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile)
{

  int ParameterRead = 0;
  long ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      return 0;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// ##############################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, 
            InputParameterOptionsEnumerator InputParameterOption,
            ifstream * InFile)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      return 0;
    }

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;


}  // ## end ReadDouble

// ##############################################################

string ReadString(string ParameterName, 
                  InputParameterOptionsEnumerator InputParameterOption,
                  ifstream * InFile)
{
  int ParameterRead = 0;
  string ParameterValue ="";
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      MissingParameter(ParameterName,InputParameterOption);
      return " ";
    }
  else
    {
      if(LogFile)
        *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
      else
        cout << endl 
             << ParameterName << " = " << ParameterValue << " " << endl 
             << endl;
    }

  return ParameterValue;

} // ## end ReadString

// ##############################################################

void MissingParameter(string ParameterName,
                      InputParameterOptionsEnumerator InputParameterOption)
{


  switch(InputParameterOption)
    {

      case OPTIONAL_NO_WARNING:
          break;

      case OPTIONAL_WARNING:
	if(LogFile)
	  {
          LogFileStars();
          *LogFile << "           WARNING" << endl 
		  << ParameterName << " not found!!" << endl;
          LogFileStars();
	  }
        else
	  {
           cout    << "           WARNING" << endl 
		   << ParameterName << " not found!!" << endl;
	  }

          break;

      case MANDATORY:

         if(LogFile)
	  {
          LogFileStars();
          *LogFile << "             WARNING" << endl 
		  << ParameterName << " not found!!" << endl
                  << "             ABORTING NOW" << endl;
          LogFileStars();
	  }

          cerr << "             WARNING" << endl 
		  << ParameterName << " not found!!" << endl
                  << "             ABORTING NOW" << endl;

          exit(1);
          break;


    default:
        cerr << "Switch option error in MissingParameter call for"
	   <<  ParameterName << endl;
        exit(1);

    } // ## end switch
         
}// ## end MissingParameter

// ###############################################################
// ###############################################################

int ReadInt(string ParameterName, string UnitsString,
            int LowerBound,int UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{

  int ParameterRead = 0;
  int ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl
               << "             ABORTING NOW" << endl;
       LogFileStars();
       exit(1);
    }
  

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadInt

// #########################################################

long ReadLongInt(string ParameterName, string UnitsString,
            long LowerBound, long UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{

  int ParameterRead = 0;
  long ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

         // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl
               << "             ABORTING NOW" << endl;
       LogFileStars();
       exit(1);
    }
  

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:
           if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;


    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadLongInt

// #########################################################

double ReadDouble(string ParameterName, string UnitsString,
            double LowerBound, double UpperBound,
            BoundsFlagEnumerator BoundsFlag, ifstream * InFile)
{

  int ParameterRead = 0;
  double ParameterValue = 0;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
       LogFileStars();
       *LogFile << "             WARNING" << endl 
               << ParameterName << " not found!!" << endl;
       LogFileStars();
      
    }
  

  *LogFile << ParameterName << " = " << ParameterValue << " " 
          << UnitsString << endl;

  switch(BoundsFlag)
    {
    case LOWER_ONLY:
	      if(ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , inf)" 
                          << endl;
                  LogFileStars();
		} // end if
           break;

    case UPPER_ONLY:

             if(ParameterValue > UpperBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "(-inf , " << UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if
           
	     break;

    case LOWER_AND_UPPER:
             if(ParameterValue > UpperBound || ParameterValue < LowerBound)
		{
                  LogFileStars();
	          *LogFile << "             WARNING" << endl 
                          << "The integer parameter " 
                          << ParameterName << endl 
                          << "is not in the required range " 
                          << "[" << LowerBound << " , " <<UpperBound << "]" 
                          << endl;
                  LogFileStars();
		} // end if

	    break;

    case NO_BOUNDS:
            break;

    default: 
             cerror("Invalid BoundsFlag Enumerator Choice");
	     exit(1);


    } // end switch


  return ParameterValue;

} // ### end ReadDouble

// #######################################################

string ReadString(string ParameterName, ifstream * InFile) 
{

  int ParameterRead = 0;
  string ParameterValue;
  string NameBuffer;

  string ParameterNameDollar = "$"+ParameterName;

  InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if a Read function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
    if (ParameterNameDollar==NameBuffer) 
     {
         *InFile >> ParameterValue; 

	 // FIX: Check that have really read value in correctly!

         ParameterRead = 1;
     }  // end if   
      
    } // end while


  if(!ParameterRead)
    {
      cerr << "ERROR in ReadString." << endl 
           << "Parameter " << ParameterName    
           << " not found." << endl << "ABORTING NOW" << endl;

      exit(1);

    }
  else
    {
      if(LogFile)
        *LogFile << ParameterName << " = " << ParameterValue << " " << endl;
      else
        cout << endl 
             << ParameterName << " = " << ParameterValue << " " << endl 
             << endl;
    }

  return ParameterValue;

} // end ReadString();

// #########################################################


void cerror(char * msg) {cout << msg << endl << flush;}

// #########################################################

void LogFileSeparator()
{

  *LogFile << endl << endl<<
          "#########################################" << endl << endl;

}

// #####################################################

void LogFileStars()
{
  *LogFile << endl << "**************************************"
          <<endl << endl;
}

// #########################################################
// ## Default error handler function
// ## By: Ivan Lima
void LogFileAbortWithErrorMsg(string ErrorMsg)
{ 
   LogFileSeparator();
   *LogFile << "Error:" << endl << ErrorMsg << endl;
   //## I should include the name of the log file here, if I can
   cout << "Error: See *.log for details on the error." << endl;
   exit(1);
}

