// $Id: ocsReducedModelTools.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $
// #####################################################################
//
//    ocsTools.cc
//
//    Functions from several authors and external libraries. 
//
//    Copyright:
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//    The copyright applies to functions written in the OFCL members
// #####################################################################

#include "ocsReducedModelTools.hh"
#include "ocsConst.hh"

extern ofstream *LogFile;


// #################################################################
// #################################################################

// Functions written by Ivan Lima Jr.




// ## It produces the angles and the matrices that rotates the 
// #  vector to (-1,0,0)
void TransRzRyV3to_Xhat( double *Vector, double *thetaZthetaY)
//                        ,double Rotation[][3])
{
   double delta,gamma;                              
   //double Ry[3][3], Rz[3][3];

   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;      
   
   if (Vector[0] != 0)  
      delta = atan(Vector[2]/Vector[0]); 
   else 
      delta = pi/2.; 
   RotatesAboutY(delta,Vector); //,Ry); //## 01/11/27
   if (Vector[0] < 0)  
      gamma = 1/2.*atan(Vector[1]/Vector[0]);  
   else if (Vector[0] > 0)   {      
      if (atan(Vector[1]/Vector[0]) < 0) 
         gamma = 1/2.*(atan(Vector[1]/Vector[0]) + pi );
      else
         gamma = 1/2.*(atan(Vector[1]/Vector[0]) - pi );               
   }      
   else {
      if (Vector[1] > 0) 
         gamma = -pi/4;
      else
         gamma =  pi/4;       
   }      
   RotatesAboutZ(2*gamma,Vector); //,Rz); 

   thetaZthetaY[0] = 2*gamma; // ## Rz       
   thetaZthetaY[1] = delta;   // ## Ry   
        
   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;       
   //Multiply(Rotation,Rz,Ry);     
}


// ## It produces the angles and the matrices that rotates the 
// #  vector to (0, y, z > 0)
void TransRyV3to_YZ( double *Vector, double *thetaY)
{
   double delta;                              

   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;      
   if (Vector[2] != 0)  {
      delta = -atan(Vector[0]/Vector[2]); 
   }   
   else {
      delta = -sgn(Vector[0])*pi/2.; 
   }   
   if (Vector[2] < 0 ) {
      delta += -pi;
    }  
      
   RotatesAboutY(delta,Vector); //,Ry);

   *thetaY = delta;   // ## Ry   
   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;       
   //Multiply(Rotation,Rz,Ry);     
}


// ## Transforms from (vx,vy,vz) to (-1,0,0) 
// ## Returning the angles and the Trans. Matrix RzRx
void TransRzRxV3to_Xhat( double *Vector, double *thetaZthetaX)
//                    ,double Rotation[][3])
{
   double delta,gamma;                              
   //double Rx[3][3], Rz[3][3];

   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;      
   if (Vector[1] != 0)  
      delta = atan(Vector[2]/Vector[1]); 
   else 
      delta = pi/2.; 
   RotatesAboutX(delta,Vector); //,Rx);
   if (Vector[0] < 0)  
      gamma = 1/2.*atan(Vector[1]/Vector[0]);  
   else if (Vector[0] > 0)   {      
      if (atan(Vector[1]/Vector[0]) < 0) 
         gamma = 1/2.*(atan(Vector[1]/Vector[0]) + pi );
      else
         gamma = 1/2.*(atan(Vector[1]/Vector[0]) - pi );               
   }      
   else {
      if (Vector[1] > 0) 
         gamma = -pi/4;
      else
         gamma =  pi/4;       
   }      
   RotatesAboutZ(2*gamma,Vector); //,Rz); 

   thetaZthetaX[0] = 2*gamma; // ## Rz angle       
   thetaZthetaX[1] = delta;   // ## Rx angle    
   //cout<<"Trans "<<Vector[0]<<" "<<Vector[1]<<" "<<Vector[2]<<endl;       
   //Multiply(Rotation,Rz,Rx);     
}   


// ## Solve the 3x3 multiplication: Right3x3 = Left3x3*Right3x3
void Multiply( double Left[][3], double Right[][3])
{
   double RightTmp[3][3];
   for (int ii = 0; ii < 3; ii++) {
      for (int jj = 0; jj < 3; jj++) {
         RightTmp[ii][jj] = Right[ii][jj];             
      }   
   }   
   for (int ii = 0; ii < 3; ii++) {
      for (int jj = 0; jj < 3; jj++) {
         for (int kk = 0; kk < 3; kk++) {
            Right[ii][jj] += Left[ii][kk]*RightTmp[kk][jj];             
         }   
         //cout << "  "<<  ii << " " << jj << "  " << Left[ii][jj] << endl;          
         //cout << "  "<<  ii << " " << jj << "  " << Right[ii][jj] << endl;          
         //cout << ii << " " << jj << "  " << Result[ii][jj] << endl;          
      }   
   }   
}


// ## Solve the matrix3x3 vector multipl.: Right3 = Left3x3*Right3
void Multiply(double Left[][3],double Right[3])
{
   double RightTmp[3];
   for (int ii = 0; ii < 3; ii++)
      RightTmp[ii] = Right[ii];
   for (int ii = 0; ii < 3; ii++) {   
      Right[ii] = 0;
      for (int jj = 0; jj < 3; jj++) {
         Right[ii] += Left[ii][jj]*RightTmp[jj];             
      }   
  }       
}

//## Make the CrossProduct

//## The result will be returned in Right
void CrossProduct(double *Left, double *Right, int FirstCoord)
{
   //## FirstCoord determines where the vector starts
   //#  0 -> 0,1,2
   //#  1 -> 1,2,3
   double RightTmp[4];
   for (int ii = 0; ii < 3; ii++) {
      RightTmp[ii+FirstCoord] = Right[ii+FirstCoord];
   }
   Right[0+FirstCoord] = Left[1+FirstCoord]*RightTmp[2+FirstCoord] 
                      -Left[2+FirstCoord]*RightTmp[1+FirstCoord];
   Right[1+FirstCoord] = Left[2+FirstCoord]*RightTmp[0+FirstCoord] 
                      -Left[0+FirstCoord]*RightTmp[2+FirstCoord];
   Right[2+FirstCoord] = Left[0+FirstCoord]*RightTmp[1+FirstCoord] 
                      -Left[1+FirstCoord]*RightTmp[0+FirstCoord];   
}


// ## Rotates about Z
void RotatesAboutZ(double angle, double *Vector) //, double Rz[][3])
{
   double Rz[3][3];
   //for (int ii=0; ii < 3; ii++)
   //   for (int jj = 0; jj < 3; jj++)
   //      Rz[ii][jj] = 0;    
   Rz[0][0] =  cos(angle);   Rz[0][1] =  sin(angle);
   Rz[1][0] = -sin(angle);   Rz[1][1] =  cos(angle);
   //Rz[2][2] = 1.;
   
   double tmpVector[3];
   tmpVector[0] = Vector[0];
   tmpVector[1] = Vector[1];    
   Vector[0] = Rz[0][0]*tmpVector[0] + Rz[0][1]*tmpVector[1];
   Vector[1] = Rz[1][0]*tmpVector[0] + Rz[1][1]*tmpVector[1];      
}


// ## Rotates about X
void RotatesAboutX(double angle, double *Vector) //, double Rx[][3])
{
   double Rx[3][3];
   //for (int ii = 0; ii < 3; ii++)
   //   for (int jj=0; jj < 3; jj++)
   //      Rx[ii][jj] = 0;    
   //Rx[0][0] = 1.;         
   Rx[1][1] =  cos(angle);    Rx[1][2] =  sin(angle);
   Rx[2][1] = -sin(angle);    Rx[2][2] =  cos(angle);
   
   double tmpVector[3];
   tmpVector[1] = Vector[1];   
   tmpVector[2] = Vector[2];   
   Vector[1] = Rx[1][1]*tmpVector[1] + Rx[1][2]*tmpVector[2];
   Vector[2] = Rx[2][1]*tmpVector[1] + Rx[2][2]*tmpVector[2];   
}


// ## Rotates about Y
void RotatesAboutY(double angle, double *Vector) //, double Ry[][3])
{
   double Ry[3][3];
   //for (int ii = 0; ii < 3; ii++)
   //   for (int jj=0; jj < 3; jj++)
   //      Ry[ii][jj] = 0;  
   Ry[0][0] =  cos(angle);              Ry[0][2] =  sin(angle);           
   //                        Ry[1][1] = 1.;         
   Ry[2][0] = -sin(angle);               Ry[2][2] =  cos(angle);
   
   double tmpVector[3];
   tmpVector[0] = Vector[0];   
   tmpVector[2] = Vector[2];   
   Vector[0] = Ry[0][0]*tmpVector[0] + Ry[0][2]*tmpVector[2];
   Vector[2] = Ry[2][0]*tmpVector[0] + Ry[2][2]*tmpVector[2];   
}


// ## Scalar product
//## The coordinates are from 0 to 2
double ScalarProduct(double *VectorA, double *VectorB) // from 0 to 2
{
   return( VectorA[0]*VectorB[0]
          +VectorA[1]*VectorB[1]
          +VectorA[2]*VectorB[2]);
} 

// ## Gets the length of a vector
double GetVectorLength(double *Vector, int Dimension) // from 0 to dimension-1
{
   double VectorLength = 0;
   for (int ii = 0; ii < Dimension; ii++) {
      VectorLength += sq(Vector[ii]);
   }   
   return sqrt(VectorLength);   
} 


// ############################################
 
// ## Converts from dB to the linear scale
double dB2Linear(double dBvalue)
{
   return(pow(10.,dBvalue/10.) ); 
}


// ########################################

double dBm2Linear(double dBmPower)
{

  // Written by John Zweck, Dec 15th 2000

  // Returns power is in Watts, W

  return(pow(10.0,dBmPower/10.0 - 3.0) );


}

// ####################################

double Linear2dB(double LinearValue)
{
 // Written by John Zweck, Dec 15th 2000

  return 10.0*log10(LinearValue);  


}

// #####################################

double Linear2dBm(double LinearPower)
{
 // Written by John Zweck, Dec 15th 2000

  // The input power LinearValue must be power in Watts, W.

   return 10.0*log10(LinearPower*1000.0);  

}

// #####################################
double GetDegreeOfPolarization(double *StokesVector)
{
   if ( StokesVector[0]*(1 + 2*NEAR_ZERO) < 
         sqrt(sq(StokesVector[1]) + sq(StokesVector[2])
        +sq(StokesVector[3])) ) {
      cout << "From: DegreeOfPolarization()" << endl 
           << "Error: The  Stokes  Vector has DOP > 1." 
           << endl
           << "Please, check simulation algorithm" 
           << endl;
      for (int ii = 0; ii < 4; ii++)
         cout << "StokesVector[" << ii << "] = " 
              << StokesVector[ii] << endl;     
      exit(1);
   }   
   return(sqrt( sq(StokesVector[1]) + sq(StokesVector[2])
               +sq(StokesVector[3]) )/StokesVector[0] );
}      
     
//######################################################     
double GetPhase(const cplx Phasor) 
//## Garanteed from -pi to pi
{
   if (abs(Phasor) == 0) 
      return(0.);
   else   // phase = sgn( Aim ) * acos( Are/Ao)
      return( sgn(real(-jc*Phasor))*acos(real(Phasor)/abs(Phasor)) );      
}


//##################################################
void Stokes2Jones(const double *Stokes, cplx *Jones)
//## The Stokes vector should have dimension 4: from 1 to 3   
//## The Jones  vector should have dimension 2: from 0 to 1  
//## The Jones  vector is returned normalized
{   
   //## Ths step provides robustness agains lack of normalization
   //#  Obviously, it assumes that the Stokes parameters
   //#  are totaly polarized.
   double Stokes0 = sqrt(sq(Stokes[1])+sq(Stokes[2])+sq(Stokes[3]) );
   //if (Stokes0 < NEAR_ZERO) {
   //   cout << "Error: the Stokes parameters cannot be zero." << endl;
   //   cout << "Exiting from Stokes2Jones()" << endl;
   //   exit(1);
   //}             
   if (Stokes0 == 0) {
      Jones[0] = Jones[1] = 0;
   }
   else {
      double StokesNorm[4];   
      StokesNorm[1] = Stokes[1]/Stokes0;
      StokesNorm[2] = Stokes[2]/Stokes0;
      StokesNorm[3] = Stokes[3]/Stokes0;   
      double PhaseXY; 
      PhaseXY = atan2( -StokesNorm[3], StokesNorm[2] );
      Jones[0] = sqrt((1. + StokesNorm[1])/2.);
      Jones[1] = sqrt((1. - StokesNorm[1])/2.) * exp(jc*PhaseXY);

      // 6 Sept. 2002 - Bug fix (BM)
      //if (sq(StokesNorm[1]) >= 1 - NEAR_ZERO) {
      //if (sq(StokesNorm[1]) == 1) {      
      //      PhaseXY = 0;
      //}      
      //else {
      //   PhaseXY = acos( StokesNorm[2]/sqrt(1-sq(StokesNorm[1]) ) );
      //}   
      //Jones[0] = sqrt((1. + StokesNorm[1])/2.);
      //Jones[1] = sqrt((1. - StokesNorm[1])/2.)
      //          *exp(-jc*sgn(StokesNorm[3])*PhaseXY);         
   }          
}

//#####################################################
void Stokes2JonesKeepPower(const double *Stokes, cplx *Jones)
{
   Stokes2Jones(Stokes,Jones);
   double TotPower = sqrt(sqrt( sq(Stokes[1])
                          +sq(Stokes[2])+sq(Stokes[3])) );
   Jones[0] *= TotPower; 
   Jones[1] *= TotPower;                      
}


//##################################################
void Jones2Stokes(const cplx *Jones, double *Stokes)
//## The Jones  vector should have dimension 2: from 0 to 1  
//## The Stokes vector should have dimension 4: from 0 to 3 
//## The Stokes vector is returned normalized
{   
   Stokes[0]   = sq(abs(Jones[0])) + sq(abs(Jones[1]));
   if (Stokes[0] != 0) {
      Stokes[1] = (sq(abs(Jones[0]))-sq(abs(Jones[1])) )
                  /Stokes[0];                           //##  |uc|^2-|vc|^2
      Stokes[2] = 2*real(Jones[0]*conj(Jones[1]))
                  /Stokes[0];                           //##  2 Re{uc . vc*}
      Stokes[3] = 2*real(-jc*Jones[0]*conj(Jones[1]))  
                  /Stokes[0];                           //## 2 Im{uc . vc*}         
      Stokes[0] = 1;             
   }
   else  {
      Stokes[1] = Stokes[2] = Stokes[3] = 0;
   }                     
}


//##################################################
void Jones2StokesKeepPower(const cplx *Jones, double *Stokes)
{
   Jones2Stokes(Jones,Stokes);
   double TotPower = sq(abs(Jones[0])) + sq(abs(Jones[1]));
   for (int ii = 0; ii < 4; ii++) {
      Stokes[ii] *= TotPower;
   }
}


// ##### End of functions written by Ivan T. Lima Jr.


// #################################################################
// #################################################################

// Functions written by John Zweck

double Frequency2Wavelength(double Frequency)
{
  // Frequency is in Hz

  return LightSpeed/Frequency;  
}

// ############################################################

double Wavelength2Frequency(double Wavelength)
{
  return LightSpeed/Wavelength; // returns Frequency  in Hz
}


// ##############################################################
// ##############################################################

// ##### The functions below were written by Ronald Holzloehner



//-----------------------------------------------------------------------
// enable program to be migrated by MOSIX
// Options: MOSIX_UNLOCK or MOSIX_LOCK
int lock_or_unlock_mosix(int what)
{
  ofstream *lock_file;

  char lock_name[50];
  sprintf(lock_name, "/proc/%d/lock", getpid());
  lock_file = new ofstream( lock_name, ios::out );
  if(lock_file->bad())
    return 1;
  *lock_file << ((what==MOSIX_LOCK)?"1":"0") << endl;
  lock_file->close();
  return 0;
}


//-----------------------------------------------------------------------
// tell MOSIX this is a pure CPU job (as opposed to one with lots of I/O)

int cpujob_mosix(void)
{
  ofstream *cpu_file;

  cpu_file = new ofstream( "/proc/mosix/decay/cpujob", ios::out );
  if(cpu_file->bad())
    return 1;
  *cpu_file << 1 << endl;
  cpu_file->close();
  return 0;
}


//-----------------------------------------------------------------------
// tell MOSIX that slow statistical decay applies to this program
int slow_mosix(void)
{
  ofstream *slow_file;

  slow_file = new ofstream( "/proc/mosix/decay/slow", ios::out );
  if(slow_file->bad())
    return 1;
  *slow_file << 1 << endl;
  slow_file->close();
  return 0;
}

// ##### End functions written by Ronald Holzloehner


// ##### Functions from Numerical Recipes

// ## Functions from nrutil.c

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>

#define NR_END 1
#define FREE_ARG char*

void nrerror(char error_text[])
/* Numerical Recipes standard error handler */
{
	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}


float erffc(float x)
{
	float gammp(float a, float x);
	float gammq(float a, float x);

	return x < 0.0 ? 1.0+gammp(0.5,x*x) : gammq(0.5,x*x);
}
/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */

float gammp(float a, float x)
{
	void gcf(float *gammcf, float a, float x, float *gln);
	void gser(float *gamser, float a, float x, float *gln);
	void nrerror(char error_text[]);
	float gamser,gammcf,gln;

	if (x < 0.0 || a <= 0.0) nrerror("Invalid arguments in routine gammp");
	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return 1.0-gammcf;
	}
}
/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */

#include <math.h>
#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void gcf(float *gammcf, float a, float x, float *gln)
{
	float gammln(float xx);
	void nrerror(char error_text[]);
	int i;
	float an,b,c,d,del,h;

	*gln=gammln(a);
	b=x+1.0-a;
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) {
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) < EPS) break;
	}
	if (i > ITMAX) nrerror("a too large, ITMAX too small in gcf");
	*gammcf=exp(-x+a*log(x)-(*gln))*h;
}
#undef ITMAX
#undef EPS
#undef FPMIN
/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */

#include <math.h>
#define ITMAX 100
#define EPS 3.0e-7

void gser(float *gamser, float a, float x, float *gln)
{
	float gammln(float xx);
	void nrerror(char error_text[]);
	int n;
	float sum,del,ap;

	*gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) nrerror("x less than 0 in routine gser");
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		nrerror("a too large, ITMAX too small in routine gser");
		return;
	}
}
#undef ITMAX
#undef EPS
/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */


#include <math.h>

float gammln(float xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}
/* (C) Copr. 1986-92 Numeric*/


float gammq(float a, float x)
{
	void gcf(float *gammcf, float a, float x, float *gln);
	void gser(float *gamser, float a, float x, float *gln);
	void nrerror(char error_text[]);
	float gamser,gammcf,gln;

	if (x < 0.0 || a <= 0.0) nrerror("Invalid arguments in routine gammq");
	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return 1.0-gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return gammcf;
	}
}
/* (C) Copr. 1986-92 Numerical Recipes Software *1%:?!$:V3+). */


// ##### End of functions obtained from Numerical Recipes



