


//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

# include "ocsOptAmplifier.hh"
# include <values.h>

extern ofstream *LogFile;

// ################### Methods for the class OptAmplifier

// ####################################################

// #### Constructor of the class
// ## By: Ivan Lima

/*

OptAmplifier::OptAmplifier()
{
   ObjectInitialized = 0;
}

*/

// #####################################################             
 
OptAmplifier::OptAmplifier(string InDir2, string InFileName, 
                           OptSignal *oOptSignal2, double * SignalIn, 
						   RanNumGen *RNG2, double F, double TotalTime2,
						   int DataPointInCye2)
{

  // Written by Jonathan Hu, May 21th, 2003 based on John's constructor
// this is used for transient
// I use time average model to solve transient effect
// please refer to paper 
// Y. Sun et al., J. Sel. Top. Quant. Electron. 3, 991-1007 (1997)
  // USE THIS CONSTRUCTOR OR NEXT ONE
  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE NEXT ONE TOO (JZ)
  
  
InDir = InDir2;
oOptSignal = oOptSignal2;
CenterFreq = oOptSignal->GetCenterFreq();
fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
TypeSimulation = oOptSignal->GetTypeSimulation();

RNG = RNG2;

qtPoints  = fft->give_N();
DeltaFreq = fft->delta_f; 


ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


LogFileSeparator();
*LogFile << InFileName << endl;

DebugLevel  = ReadInt("DebugLevel"," ",1,3,LOWER_AND_UPPER,MANDATORY,&InFile);


TypeAmplifier  = typeAmplifier(ReadInt("TypeAmplifier"," ",0,6,
                                        LOWER_AND_UPPER,MANDATORY,&InFile));
 

 switch(TypeAmplifier)
   {
    case SCALAR_SIMPLE: case SCALAR_SATURATED:
      if(TypeSimulation == VECTOR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_SIMPLE: case VECTOR_SATURATED: case VECTOR_FIXED_OUTPUT_POWER:
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:

    if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;
   
   } // ## end switch

*LogFile << "TypeAmplifier = " << WriteTypeAmplifier() << endl;

TypeAmplifierNoise  = typeAmplifierNoise(ReadInt("TypeAmplifierNoise"," ",0,6,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

*LogFile << "TypeAmplifierNoise = " << WriteTypeAmplifierNoise() << endl; 




TypeGainProfile = FLAT_GAIN; // Default value

int temp = 0;

temp = ReadInt("TypeGainProfile"," ",1,6,
		   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

if(temp)
  TypeGainProfile = typeGainProfile(temp);


*LogFile << "TypeGainProfile = " << WriteTypeGainProfile() << endl; 

*LogFile << " here" ;



if (TypeGainProfile == FLAT_GAIN_INSIDE_BANDWIDTH) {
   GainBandwidthHz = ReadDouble("GainBandwidthHz","Hz",0,0,
                               LOWER_ONLY,MANDATORY,&InFile);

   if (GainBandwidthHz == 0 ) {
      //## Enough to fit all the channels
      //## If you want the whole bandwidth, use the FLAT_GAIN option
      //GainBandwidthHz = 1/fft->delta_t;        
      GainMinRelFreqHz = oOptSignal->GetFrequency(0)
                        -oOptSignal->GetCenterFreq()     
                        -oOptSignal->GetChannSpacing()/2.;

      GainMaxRelFreqHz = 
         oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1)
                          -oOptSignal->GetCenterFreq()       
                          +oOptSignal->GetChannSpacing()/2.;

      //cout << "ChannSpacing " << oOptSignal->GetChannSpacing() << endl;                                           
      //cout << "GainMinRelFreqHz " << GainMinRelFreqHz << endl;
      //cout << "GainMaxRelFreqHz " << GainMaxRelFreqHz << endl;
      //cout <<  oOptSignal->GetFrequency(0) -oOptSignal->GetCenterFreq() << endl;                        
      //cout <<  oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1) -oOptSignal->GetCenterFreq()  << endl;                  
   }
   else {
      GainMinRelFreqHz = -GainBandwidthHz/2.;
      GainMaxRelFreqHz =  GainBandwidthHz/2.;  
   }
   *LogFile << "GainBandwidthHz = " << GainBandwidthHz << endl;   
   *LogFile << "   -> GainMinRelFreqHz = " << GainMinRelFreqHz << endl;   
   *LogFile << "   -> GainMaxRelFreqHz = " << GainMaxRelFreqHz << endl;   
}   

LinearGainProfile = 0;
AllocateGainProfileMemory(); 

dBGain = ReadDouble("GainOptAmplifier","dB",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);

LinearGainOptAmplif = dB2Linear(dBGain); 

LinearGainFixedTotalOutputPower = LinearGainOptAmplif;

 if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER 
    || TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER  )
   {

     double OutputPowerPerChanneldBm = ReadDouble("OutputPowerPerChanneldBm",
                                              "dBm",0,0,NO_BOUNDS,
                                               OPTIONAL_WARNING,&InFile);

    //OutputPowerLinear = oOptSignal->GetNumChannels()*
    //                     dBm2Linear(OutputPowerPerChanneldBm);


   OutputPowerLinear = dBm2Linear(OutputPowerPerChanneldBm);

   double ResolutionBandwidthHz =  ReadDouble("ResolutionBandwidthHz","Hz",
                                        0,-1,
                                       LOWER_ONLY, OPTIONAL_WARNING,&InFile);

   double  SuperGaussExponent  = ReadDouble("SuperGaussExponent"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);

       
    PowerMeter = new OptSpectrumAnalyzer(oOptSignal,ResolutionBandwidthHz,1,
                                 SuperGaussExponent,int(TypeAmplifierNoise));

   } // ## end if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER) 

/// We either input NoiseFigOptAmplifier (in dB) or SpontaneousEmissionFactor
/// SpontaneousEmissionFactor = n_{sp} is in linear units (not dB)
/// See Desurvire p. 77 for definition.
/// The choice is made by setting the other value to 0.0.
/// The NoiseAmplitudeFactor is computed from either NoiseFigOptAmplifier  
/// or SpontaneousEmissionFactor

NoiseFigOptAmplifier = 0;
SpontaneousEmissionFactor = 0;
 

NoiseFigOptAmplifier = ReadDouble("NoiseFigOptAmplifier","dB",0,0,
				  NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SpontaneousEmissionFactor = ReadDouble("SpontEmissionFactor"," ",
                                        1.0,-1.0,
				        LOWER_ONLY,OPTIONAL_WARNING,&InFile);

ComputeNoiseAmplitudeFactor();

PolDepGainOptAmplifier = ReadDouble(
                               "PolDepGainOptAmplifier","dB"
                               ,0,-1,LOWER_ONLY,OPTIONAL_WARNING,0.0,&InFile);
   
                               
//PolDepGainOrigPhase = new double[qtPoints];                               

//AmpLength = ReadDouble("AmpLength","m",0,-1,LOWER_ONLY,&InFile);

AmpLength = 1.0;

 if(DebugLevel >=3)
   {

    *LogFile << "AmpLength = 1.0." 
	 << endl 
	 << "This value doesn't have any physical meaning at moment."
	 << endl
	 << "The Saturated Amplifier gives same output regardless of"
	 << "value of AmpLength" << endl;
   }

NumZSteps = ReadInt("NumZSteps"," ",1,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);

SaturatingPowerdBm = ReadDouble("SaturatingPowerdBm","dBm",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SaturatingPowerLinear = dBm2Linear(SaturatingPowerdBm);

UnsaturatedGaindB = ReadDouble("UnsaturatedGaindB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

UnsaturatedGainLinear = dB2Linear(UnsaturatedGaindB);

UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;

// Input parameters relating to rippled gain


RippleAmplitudeLinear = dB2Linear(ReadDouble("RippleAmplitudedB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile));



RipplePeriod = ReadDouble("RipplePeriod","Hz",
                     0,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);


  // WARNING: If FreqWindow is not a multiple of RipplePeriod 
  // then will have some sort of discontinuity in frequency
  // which will be transformed into a lack of smoothness in time.
  // This probably isn't a problem if noise is already being added
  // to the system. However it's important to be aware of the issue.

 if(DebugLevel >= 2 && TypeGainProfile == RIPPLED_GAIN)
   {
     if(floor(oOptSignal->GetFreqWindow()/RipplePeriod) !=
        oOptSignal->GetFreqWindow()/RipplePeriod)


     *LogFile << "BEWARE OptAmplifier::OptAmplifier " << endl
	     << "FreqWindow is not a multiple of RipplePeriod" << endl;
	  
   }



RelativeRippleShift = ReadDouble("RelativeRippleShift","Hz",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);


ObjectInitialized = 1;

 LogFileSeparator();

 *LogFile << "Processing Opt Amplifier Input Parameters" << endl << endl;

 *LogFile << "LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
 *LogFile << "OutputPowerLinear [mW] = " << 1e3*OutputPowerLinear << endl;
 *LogFile << "LinearNoiseFigOptAmplif = " << LinearNoiseFigOptAmplif << endl;
 *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;
 *LogFile << "CenterFreq = " << CenterFreq << endl;
 *LogFile << "qtPoints = " << qtPoints << endl;
 *LogFile << "DeltaFreq = " << DeltaFreq << endl;


 *LogFile << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
 *LogFile << "UnsaturatedGainLinear = " << UnsaturatedGainLinear << endl;
 *LogFile << "UnsaturatedDifferentialGainLinear = " 
	 << UnsaturatedDifferentialGainLinear << endl;

 *LogFile << "RippleAmplitudeLinear = " << RippleAmplitudeLinear << endl;

 oPolDepAttenuator.InitPolDepAttenuator(oOptSignal);
 oPolDepAttenuator.SetPolDepAttenuation(0.,0.);

   
 TotPumpPowerIndex = ReadInt("TotPumpPowerIndex","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 TotSignalPowerIndex = ReadInt("TotSignalPowerIndex","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 NumGainWidth = ReadInt("NumGainWidth","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 NumGainWidthDep = ReadInt("NumGainWidthDep","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 AmplifierPumpPowerC = ReadDouble("AmplifierPumpPowerC","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

 oEDFAGain = new double[TotSignalPowerIndex * TotPumpPowerIndex * NumGainWidth];
 oEDFAGainPowerDependent = new double[NumGainWidthDep];

 ObjectInitializedJH = 1;

 AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 PumpPowerFactor = ReadDouble("PumpPowerFactor"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 AmpLength =  ReadDouble("AmpLength"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 DeltaZ = ReadDouble("DeltaZ", " ", 0,-1,
 									LOWER_ONLY,&InFile);
 TypeIntergration = ReadInt("TypeIntergration", " ", 0,-1,
 									LOWER_ONLY,&InFile);
 NumIntItr = ReadInt("NumIntItr", " ", 0,-1,
 									LOWER_ONLY,&InFile); 
 NumZSteps = int (AmpLength/DeltaZ + 0.00001 ) + 1; 
 
// signal build up
  char str[40];
// NumSigCh =  ReadInt("NumSigCh"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
 NumSigCh =  oOptSignal->GetQtPoints();
 NumNoiCh =  NumSigCh;
 NumPumCh =  ReadInt("NumPumCh"," ",0,-1,
                                    LOWER_ONLY,&InFile);
// NumNoiCh =  ReadInt("NumNoiCh"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
									
 oSignalWaveLth = new double [NumSigCh];
 oSignalPower = new double [NumSigCh*(NumZSteps+1)];
 oPumpWaveLth = new double [NumPumCh];
 oPumpPower = new double [NumPumCh*(NumZSteps+1)];
 
 oNoiseWaveLth = new double [NumNoiCh];									
 oNoiseBPower = new double [NumNoiCh*(NumZSteps+1)];
 oNoiseFPower = new double [NumNoiCh*(NumZSteps+1)];


 UpperNBackward = new double [NumZSteps];
 
 ConfinementFactor = ReadDouble("ConfinementFactor"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 DopantConcentration = ReadDouble("DopantConcentration"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 EffectArea = ReadDouble("EffectArea"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 LifeTime = ReadDouble("LifeTime"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 SaturationFactorAdjuctment = ReadDouble("SaturationFactorAdjuctment"," ",0,-1,
		    						 LOWER_ONLY,&InFile); 
 PumpABS = ReadDouble("PumpABS"," ",0,-1,
                                     LOWER_ONLY,&InFile);

// AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
//                                    LOWER_ONLY,&InFile);

	





   /*
   for(int ii=0;ii<NumSigCh;ii++)  {
   
      sprintf(str,"SignalWaveLth%d",ii);
      oSignalWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] << endl;
      sprintf(str,"SignalPower%d",ii);
      oSignalPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
   }
   */
   
/* fft->fftshift(sfftM.fOptSignalX);
 for(int ii=0;ii<NumSigCh;ii++)  {
   	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() );
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] ;
      oSignalPower[ii*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[ii*NumZSteps] < 1e-100)
			oSignalPower[ii*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
	  *LogFile << "    oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
 }
 fft->ifftshift(sfftM.fOptSignalX);
 */
 for(int ii=0;ii<NumPumCh;ii++)
    {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpCurrent%d",ii);				    
      oPumpPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
//	  cout << " OPumpCurrent = " << oPumpPower[ii*NumZSteps]  << "A ->" ;
	  oPumpPower[ii*NumZSteps] = PumpCurrent2Power(  oPumpPower[ii*NumZSteps] );
	  
	  sprintf(str,"PumpCurrent%dSystem0dB",ii);
	  PumpPowerSystem0dB = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile); 
	  PumpPowerSystem0dB = PumpCurrent2Power(PumpPowerSystem0dB);
	  
	  
	  sprintf(str,"PumpCurrent%dTransparency",ii);
      PumpPowerTransparency = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
	  PumpPowerTransparency = PumpCurrent2Power(PumpPowerTransparency);
	  
//	  cout << " OPumpPower = " << oPumpPower[ii*NumZSteps] << endl;			
      *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLth[ii] << endl;   
      *LogFile << "    oPumpPower" << ii << " = " << 
                          oPumpPower[ii*NumZSteps] << endl;  
						  
   	}
 	DeltaWave = ReadDouble("DeltaWave"," ",0,-1,
                                    LOWER_ONLY,&InFile);
    InFile.close();

//	cout << " contruction middle1" << endl;
// the Following part is for Channel Drop Transient
 	IntrinsicSaturationPowerPump = new double [1]; 
	IntrinsicSaturationPowerSignal = new double [8];
	*LogFile << "run to here l3 " << endl; 
	GainSig = new double [8*NumTSteps];
	GainPum = new double [1*NumTSteps];
	AbsorptionLossSignal = new double [8];
	AbsorptionLossPump = new double [1]; 
	*LogFile << "run to here l4 " << endl; 
// the following part is for Amplitidue Fl
    TotalTime = TotalTime2;
	DataPointInCye = DataPointInCye2;
	DeltaT = 1 / F /DataPointInCye;
//	cout << " contruction middle2" << endl;
	NumTSteps = (int)(TotalTime / DeltaT);
	*LogFile << "NumTSteps = " << NumTSteps << endl; 
  //  cout << " NumTSteps = " << NumTSteps << endl;
 	AverageN2T = new double[NumTSteps];
// 	SignalInput = new double[NumTSteps];
	PumpInput  = new double[NumTSteps];

	SignalInput = SignalIn;	  
	AverageN2T[0]=0.607239;
/*		for(int ii=0;ii < NumTSteps;ii++){
		cout << ii << "   " << SignalInput[ii] << endl ;
//		cout << " run to here1" << endl;	
 	}

*/	
//	cout << " contruction middle3" << endl;
//	cout << "run to constructor" << endl;
	SaturationFactor = DopantConcentration * EffectArea / LifeTime;
	SaturationFactor *= SaturationFactorAdjuctment;
	for(int StepT =0;StepT < NumTSteps;StepT ++){
		PumpInput[0 * NumTSteps + StepT] = dBm2Linear(19);
 // 22 is the value I used for AmpFlu
 	}
//	cout << PumpPowerSystem0dB << endl;
	NumSigCh = 1;
	NumPumCh = 1;

	oSignalWaveLth[0] = 1550e-9;	
	oPumpWaveLth[0] = 980e-9;
	ABSAdjustGA=1;
	EMSAdjustGA=1;
//	cout << " contruction end" << endl;
	PolynomialFilterAppliedFlag = 0;

UserHasSpecifiedNoiseAmplitudeFactorFlag = false;

} // ## end OptAmplifier constructor used by Jonathan Hu

// #####################################################

OptAmplifier::OptAmplifier(string InDir2, string InFileName, 
                           OptSignal *oOptSignal2, RanNumGen *RNG2)
{

  // Written by Jonathan Hu, Oct.26th, 2002 based on John's constructor
  // this is for Giles' model
  // please refer to following material
 //  "Erbium-Doped Fiber Amplifiers Fundamentals and Technology" by Becker. mainly in Chaper 6
 //  "Modeling Erbium-Dopted Fiber Amplifiers" by Giles and Desurvire, JLT p271-283 1991
 //  "Characterization of Erbium-Dopted Fibers and Application to Modeling 980-nm and 1480-nm Pumped
 // 		Amplifiers"	Giles et al., PTL p.363-365 1991
 //  cutback method(used for measure parameters) can be found in Desurvire's book " 
  // Erbium-Dopted Fiber Amplifiers Principles and
 //			Applications"
 
  // USE THIS CONSTRUCTOR OR NEXT ONE
  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE NEXT ONE TOO (JZ)
  
  // Currently, the model only work for scalar case.
InDir = InDir2;
oOptSignal = oOptSignal2;
CenterFreq = oOptSignal->GetCenterFreq();
fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
TypeSimulation = oOptSignal->GetTypeSimulation();

RNG = RNG2;

qtPoints  = fft->give_N();
DeltaFreq = fft->delta_f; 


ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


LogFileSeparator();
*LogFile << InFileName << endl;

DebugLevel  = ReadInt("DebugLevel"," ",1,3,LOWER_AND_UPPER,MANDATORY,&InFile);


TypeAmplifier  = typeAmplifier(ReadInt("TypeAmplifier"," ",0,6,
                                        LOWER_AND_UPPER,MANDATORY,&InFile));
 

 switch(TypeAmplifier)
   {
    case SCALAR_SIMPLE: case SCALAR_SATURATED:
      if(TypeSimulation == VECTOR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_SIMPLE: case VECTOR_SATURATED: case VECTOR_FIXED_OUTPUT_POWER:
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:

    if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;
   
   } // ## end switch

*LogFile << "TypeAmplifier = " << WriteTypeAmplifier() << endl;

TypeAmplifierNoise  = typeAmplifierNoise(ReadInt("TypeAmplifierNoise"," ",0,6,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

*LogFile << "TypeAmplifierNoise = " << WriteTypeAmplifierNoise() << endl; 


TypeGainProfile = FLAT_GAIN; // Default value

int temp = 0;

temp = ReadInt("TypeGainProfile"," ",1,6,
		   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

if(temp)
  TypeGainProfile = typeGainProfile(temp);


*LogFile << "TypeGainProfile = " << WriteTypeGainProfile() << endl; 

*LogFile << " here" ;



if (TypeGainProfile == FLAT_GAIN_INSIDE_BANDWIDTH) {
   GainBandwidthHz = ReadDouble("GainBandwidthHz","Hz",0,0,
                               LOWER_ONLY,MANDATORY,&InFile);

   if (GainBandwidthHz == 0 ) {
      //## Enough to fit all the channels
      //## If you want the whole bandwidth, use the FLAT_GAIN option
      //GainBandwidthHz = 1/fft->delta_t;        
      GainMinRelFreqHz = oOptSignal->GetFrequency(0)
                        -oOptSignal->GetCenterFreq()     
                        -oOptSignal->GetChannSpacing()/2.;

      GainMaxRelFreqHz = 
         oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1)
                          -oOptSignal->GetCenterFreq()       
                          +oOptSignal->GetChannSpacing()/2.;

      //cout << "ChannSpacing " << oOptSignal->GetChannSpacing() << endl;                                           
      //cout << "GainMinRelFreqHz " << GainMinRelFreqHz << endl;
      //cout << "GainMaxRelFreqHz " << GainMaxRelFreqHz << endl;
      //cout <<  oOptSignal->GetFrequency(0) -oOptSignal->GetCenterFreq() << endl;                        
      //cout <<  oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1) -oOptSignal->GetCenterFreq()  << endl;                  
   }
   else {
      GainMinRelFreqHz = -GainBandwidthHz/2.;
      GainMaxRelFreqHz =  GainBandwidthHz/2.;  
   }
   *LogFile << "GainBandwidthHz = " << GainBandwidthHz << endl;   
   *LogFile << "   -> GainMinRelFreqHz = " << GainMinRelFreqHz << endl;   
   *LogFile << "   -> GainMaxRelFreqHz = " << GainMaxRelFreqHz << endl;   
}   

LinearGainProfile = 0;
AllocateGainProfileMemory(); 

dBGain = ReadDouble("GainOptAmplifier","dB",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);

LinearGainOptAmplif = dB2Linear(dBGain); 

LinearGainFixedTotalOutputPower = LinearGainOptAmplif;

 if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER 
    || TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER  )
   {

     double OutputPowerPerChanneldBm = ReadDouble("OutputPowerPerChanneldBm",
                                              "dBm",0,0,NO_BOUNDS,
                                               OPTIONAL_WARNING,&InFile);

    //OutputPowerLinear = oOptSignal->GetNumChannels()*
    //                     dBm2Linear(OutputPowerPerChanneldBm);


   OutputPowerLinear = dBm2Linear(OutputPowerPerChanneldBm);

   double ResolutionBandwidthHz =  ReadDouble("ResolutionBandwidthHz","Hz",
                                        0,-1,
                                       LOWER_ONLY, OPTIONAL_WARNING,&InFile);

   double  SuperGaussExponent  = ReadDouble("SuperGaussExponent"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);

       
    PowerMeter = new OptSpectrumAnalyzer(oOptSignal,ResolutionBandwidthHz,1,
                                 SuperGaussExponent,int(TypeAmplifierNoise));

   } // ## end if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER) 

/// We either input NoiseFigOptAmplifier (in dB) or SpontaneousEmissionFactor
/// SpontaneousEmissionFactor = n_{sp} is in linear units (not dB)
/// See Desurvire p. 77 for definition.
/// The choice is made by setting the other value to 0.0.
/// The NoiseAmplitudeFactor is computed from either NoiseFigOptAmplifier  
/// or SpontaneousEmissionFactor

NoiseFigOptAmplifier = 0;
SpontaneousEmissionFactor = 0;
 

NoiseFigOptAmplifier = ReadDouble("NoiseFigOptAmplifier","dB",0,0,
				  NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SpontaneousEmissionFactor = ReadDouble("SpontEmissionFactor"," ",
                                        1.0,-1.0,
				        LOWER_ONLY,OPTIONAL_WARNING,&InFile);

ComputeNoiseAmplitudeFactor();

PolDepGainOptAmplifier = ReadDouble(
                               "PolDepGainOptAmplifier","dB"
                               ,0,-1,LOWER_ONLY,OPTIONAL_WARNING,0.0,&InFile);
   
                               
//PolDepGainOrigPhase = new double[qtPoints];                               

//AmpLength = ReadDouble("AmpLength","m",0,-1,LOWER_ONLY,&InFile);

AmpLength = 1.0;

 if(DebugLevel >=3)
   {

    *LogFile << "AmpLength = 1.0." 
	 << endl 
	 << "This value doesn't have any physical meaning at moment."
	 << endl
	 << "The Saturated Amplifier gives same output regardless of"
	 << "value of AmpLength" << endl;
   }

NumZSteps = ReadInt("NumZSteps"," ",1,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);

SaturatingPowerdBm = ReadDouble("SaturatingPowerdBm","dBm",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SaturatingPowerLinear = dBm2Linear(SaturatingPowerdBm);

UnsaturatedGaindB = ReadDouble("UnsaturatedGaindB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

UnsaturatedGainLinear = dB2Linear(UnsaturatedGaindB);

UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;

// Input parameters relating to rippled gain


RippleAmplitudeLinear = dB2Linear(ReadDouble("RippleAmplitudedB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile));



RipplePeriod = ReadDouble("RipplePeriod","Hz",
                     0,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);


  // WARNING: If FreqWindow is not a multiple of RipplePeriod 
  // then will have some sort of discontinuity in frequency
  // which will be transformed into a lack of smoothness in time.
  // This probably isn't a problem if noise is already being added
  // to the system. However it's important to be aware of the issue.

 if(DebugLevel >= 2 && TypeGainProfile == RIPPLED_GAIN)
   {
     if(floor(oOptSignal->GetFreqWindow()/RipplePeriod) !=
        oOptSignal->GetFreqWindow()/RipplePeriod)


     *LogFile << "BEWARE OptAmplifier::OptAmplifier " << endl
	     << "FreqWindow is not a multiple of RipplePeriod" << endl;
	  
   }



RelativeRippleShift = ReadDouble("RelativeRippleShift","Hz",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);


ObjectInitialized = 1;

 LogFileSeparator();

 *LogFile << "Processing Opt Amplifier Input Parameters" << endl << endl;

 *LogFile << "LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
 *LogFile << "OutputPowerLinear [mW] = " << 1e3*OutputPowerLinear << endl;
 *LogFile << "LinearNoiseFigOptAmplif = " << LinearNoiseFigOptAmplif << endl;
 *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;
 *LogFile << "CenterFreq = " << CenterFreq << endl;
 *LogFile << "qtPoints = " << qtPoints << endl;
 *LogFile << "DeltaFreq = " << DeltaFreq << endl;


 *LogFile << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
 *LogFile << "UnsaturatedGainLinear = " << UnsaturatedGainLinear << endl;
 *LogFile << "UnsaturatedDifferentialGainLinear = " 
	 << UnsaturatedDifferentialGainLinear << endl;

 *LogFile << "RippleAmplitudeLinear = " << RippleAmplitudeLinear << endl;

 oPolDepAttenuator.InitPolDepAttenuator(oOptSignal);
 oPolDepAttenuator.SetPolDepAttenuation(0.,0.);

   
 TotPumpPowerIndex = ReadInt("TotPumpPowerIndex","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 TotSignalPowerIndex = ReadInt("TotSignalPowerIndex","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 NumGainWidth = ReadInt("NumGainWidth","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 NumGainWidthDep = ReadInt("NumGainWidthDep","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);
 AmplifierPumpPowerC = ReadDouble("AmplifierPumpPowerC","",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

 oEDFAGain = new double[TotSignalPowerIndex * TotPumpPowerIndex * NumGainWidth];
 oEDFAGainPowerDependent = new double[NumGainWidthDep];

 ObjectInitializedJH = 1;

 AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 PumpPowerFactor = ReadDouble("PumpPowerFactor"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 AmpLength =  ReadDouble("AmpLength"," ",0,-1,
                                 LOWER_ONLY,&InFile);
 DeltaZ = ReadDouble("DeltaZ", " ", 0,-1,
 									LOWER_ONLY,&InFile);
 TypeIntergration = ReadInt("TypeIntergration", " ", 0,-1,
 									LOWER_ONLY,&InFile);
 NumIntItr = ReadInt("NumIntItr", " ", 0,-1,
 									LOWER_ONLY,&InFile); 
 NumZSteps = int (AmpLength/DeltaZ + 0.00001 ) + 1; 
 
// signal build up
  char str[40];
// NumSigCh =  ReadInt("NumSigCh"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
 NumSigCh =  oOptSignal->GetQtPoints();
 NumNoiCh =  NumSigCh;
 NumPumCh =  ReadInt("NumPumCh"," ",0,-1,
                                    LOWER_ONLY,&InFile);
// NumNoiCh =  ReadInt("NumNoiCh"," ",0,-1,
//                                    LOWER_ONLY,&InFile);									
 oSignalWaveLth = new double [NumSigCh];
 oSignalPower = new double [NumSigCh*(NumZSteps+1)];
 oPumpWaveLth = new double [NumPumCh];
 oPumpPower = new double [NumPumCh*(NumZSteps+1)];
 
 oNoiseWaveLth = new double [NumNoiCh];									
 oNoiseBPower = new double [NumNoiCh*(NumZSteps+1)];
 oNoiseFPower = new double [NumNoiCh*(NumZSteps+1)];


 UpperNBackward = new double [NumZSteps];
 
 ConfinementFactor = ReadDouble("ConfinementFactor"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 DopantConcentration = ReadDouble("DopantConcentration"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 EffectArea = ReadDouble("EffectArea"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 LifeTime = ReadDouble("LifeTime"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 SaturationFactorAdjuctment = ReadDouble("SaturationFactorAdjuctment"," ",0,-1,
		    						 LOWER_ONLY,&InFile); 
 PumpABS = ReadDouble("PumpABS"," ",0,-1,
                                     LOWER_ONLY,&InFile);

// AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
//                                    LOWER_ONLY,&InFile);

	





   /*
   for(int ii=0;ii<NumSigCh;ii++)  {
   
      sprintf(str,"SignalWaveLth%d",ii);
      oSignalWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] << endl;
      sprintf(str,"SignalPower%d",ii);
      oSignalPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
   }
   */
   
/* fft->fftshift(sfftM.fOptSignalX);
 for(int ii=0;ii<NumSigCh;ii++)  {
   	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() );
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] ;
      oSignalPower[ii*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[ii*NumZSteps] < 1e-100)
			oSignalPower[ii*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
	  *LogFile << "    oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
 }
 fft->ifftshift(sfftM.fOptSignalX);
 */
 for(int ii=0;ii<NumPumCh;ii++)
   {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpCurrent%d",ii);				    
      oPumpPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
//	  cout << " OPumpCurrent = " << oPumpPower[ii*NumZSteps]  << "A ->" ;
	  oPumpPower[ii*NumZSteps] = PumpCurrent2Power(  oPumpPower[ii*NumZSteps] );
	  
	  sprintf(str,"PumpCurrent%dSystem0dB",ii);
	  PumpPowerSystem0dB = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile); 
	  PumpPowerSystem0dB = PumpCurrent2Power(PumpPowerSystem0dB);
	  
	  
	  sprintf(str,"PumpCurrent%dTransparency",ii);
      PumpPowerTransparency = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
	  PumpPowerTransparency = PumpCurrent2Power(PumpPowerTransparency);
	  
//	  cout << " OPumpPower = " << oPumpPower[ii*NumZSteps] << endl;			
      *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLth[ii] << endl;   
      *LogFile << "    oPumpPower" << ii << " = " << 
                          oPumpPower[ii*NumZSteps] << endl;  
						  
   }

	

 DeltaWave = ReadDouble("DeltaWave"," ",0,-1,
                                    LOWER_ONLY,&InFile);


    InFile.close();

// the Following part is for Channel Drop Transient
// I use other construction for transient effect, 
	NumTSteps = 200;
 	IntrinsicSaturationPowerPump = new double [1]; 
	IntrinsicSaturationPowerSignal = new double [8];
	GainSig = new double [8*NumTSteps];
	GainPum = new double [1*NumTSteps];
	SignalInput = new double [8*NumTSteps];
	PumpInput  = new double [1*NumTSteps];
	AbsorptionLossSignal = new double [8];
	AbsorptionLossPump = new double [1]; 


	PolynomialFilterAppliedFlag = 0;

UserHasSpecifiedNoiseAmplitudeFactorFlag = false;

} // ## end OptAmplifier constructor used by Jonathan Hu

// #####################################################             

OptAmplifier::OptAmplifier(string InFileName, 
                           OptSignal *oOptSignal2, RanNumGen *RNG2)
{

  // Written by John Zweck, Nov 7th 2000
  // USE THIS CONSTRUCTOR OR NEXT ONE
  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE NEXT ONE TOO (JZ)

oOptSignal = oOptSignal2;
oReduOptAmplifier = new ReduOptAmplifier(InFileName,oOptSignal->oReduOptSignal);

CenterFreq = oOptSignal->GetCenterFreq();
fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
TypeSimulation = oOptSignal->GetTypeSimulation();
RNG = RNG2;

qtPoints  = fft->give_N();
DeltaFreq = fft->delta_f; 


ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


LogFileSeparator();
*LogFile << InFileName << endl;

DebugLevel  = ReadInt("DebugLevel"," ",1,3,LOWER_AND_UPPER,MANDATORY,&InFile);


TypeAmplifier  = typeAmplifier(ReadInt("TypeAmplifier"," ",0,6,
                                        LOWER_AND_UPPER,MANDATORY,&InFile));


 switch(TypeAmplifier)
   {
    case SCALAR_SIMPLE: case SCALAR_SATURATED:
      if(TypeSimulation == VECTOR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_SIMPLE: case VECTOR_SATURATED: case VECTOR_FIXED_OUTPUT_POWER:
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:

    if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;
   
   } // ## end switch

*LogFile << "TypeAmplifier = " << WriteTypeAmplifier() << endl;

TypeAmplifierNoise  = typeAmplifierNoise(ReadInt("TypeAmplifierNoise"," ",0,6,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

*LogFile << "TypeAmplifierNoise = " << WriteTypeAmplifierNoise() << endl; 


// added by Walter for MMC and Importance Sampling simulation


if (TypeAmplifierNoise == NOISE_ON_BIASED )
   {
   BiasedNoise= NewCplx(qtPoints);
   }
// end stuff added by Walter

TypeGainProfile = FLAT_GAIN; // Default value

int temp = 0;

temp = ReadInt("TypeGainProfile"," ",1,4,
		   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

if(temp)
  TypeGainProfile = typeGainProfile(temp);


*LogFile << "TypeGainProfile = " << WriteTypeGainProfile() << endl; 

if (TypeGainProfile == FLAT_GAIN_INSIDE_BANDWIDTH) {
   GainBandwidthHz = ReadDouble("GainBandwidthHz","Hz",0,0,
                               LOWER_ONLY,MANDATORY,&InFile);

   if (GainBandwidthHz == 0 ) { 
      //## I think I added this piece of code (Ivan Lima)
      //## Enough to fit all the channels
      //## If you want the whole bandwidth, use the FLAT_GAIN option
      //GainBandwidthHz = 1/fft->delta_t;        
      GainMinRelFreqHz = oOptSignal->GetFrequency(0)
                        -oOptSignal->GetCenterFreq()     
                        -oOptSignal->GetChannSpacing()/2.;

      GainMaxRelFreqHz = 
         oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1)
                          -oOptSignal->GetCenterFreq()       
                          +oOptSignal->GetChannSpacing()/2.;

      //cout << "ChannSpacing " << oOptSignal->GetChannSpacing() << endl;                                           
      //cout << "GainMinRelFreqHz " << GainMinRelFreqHz << endl;
      //cout << "GainMaxRelFreqHz " << GainMaxRelFreqHz << endl;
      //cout <<  oOptSignal->GetFrequency(0) -oOptSignal->GetCenterFreq() << endl;                        
      //cout <<  oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1) -oOptSignal->GetCenterFreq()  << endl;                  
   }
   else {
      GainMinRelFreqHz = -GainBandwidthHz/2.;
      GainMaxRelFreqHz =  GainBandwidthHz/2.;  
   }
   *LogFile << "GainBandwidthHz = " << GainBandwidthHz << endl;   
   *LogFile << "   -> GainMinRelFreqHz = " << GainMinRelFreqHz << endl;   
   *LogFile << "   -> GainMaxRelFreqHz = " << GainMaxRelFreqHz << endl;   
}   

LinearGainProfile = 0;
AllocateGainProfileMemory(); 

dBGain = ReadDouble("GainOptAmplifier","dB",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);

LinearGainOptAmplif = dB2Linear(dBGain); 

LinearGainFixedTotalOutputPower = LinearGainOptAmplif;

 if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER 
    || TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER  )
   {

     double OutputPowerPerChanneldBm = ReadDouble("OutputPowerPerChanneldBm",
                                              "dBm",0,0,NO_BOUNDS,
                                               OPTIONAL_WARNING,&InFile);

    //OutputPowerLinear = oOptSignal->GetNumChannels()*
    //                     dBm2Linear(OutputPowerPerChanneldBm);


   OutputPowerLinear = dBm2Linear(OutputPowerPerChanneldBm);

   double ResolutionBandwidthHz =  ReadDouble("ResolutionBandwidthHz","Hz",
                                        0,-1,
                                       LOWER_ONLY, OPTIONAL_WARNING,&InFile);

   double  SuperGaussExponent  = ReadDouble("SuperGaussExponent"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);
  
    PowerMeter = new OptSpectrumAnalyzer(oOptSignal,ResolutionBandwidthHz,1,
                                  SuperGaussExponent,int(TypeAmplifierNoise));
  
   } // ## end if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER) 

/// We either input NoiseFigOptAmplifier (in dB) or SpontaneousEmissionFactor
/// SpontaneousEmissionFactor = n_{sp} is in linear units (not dB)
/// See Desurvire p. 77 for definition.
/// The choice is made by setting the other value to 0.0.
/// The NoiseAmplitudeFactor is computed from either NoiseFigOptAmplifier  
/// or SpontaneousEmissionFactor

NoiseFigOptAmplifier = 0;
SpontaneousEmissionFactor = 0;
 

NoiseFigOptAmplifier = ReadDouble("NoiseFigOptAmplifier","dB",0,0,
				  NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SpontaneousEmissionFactor = ReadDouble("SpontEmissionFactor"," ",
                                        1.0,-1.0,
				        LOWER_ONLY,OPTIONAL_WARNING,&InFile);

ComputeNoiseAmplitudeFactor();

PolDepGainOptAmplifier = ReadDouble(
                               "PolDepGainOptAmplifier","dB"
                               ,0,-1,LOWER_ONLY,OPTIONAL_WARNING,0.0,&InFile);
   
                               
//PolDepGainOrigPhase = new double[qtPoints];                               

//AmpLength = ReadDouble("AmpLength","m",0,-1,LOWER_ONLY,&InFile);

AmpLength = 1.0;

 if(DebugLevel >=3)
   {

    *LogFile << "AmpLength = 1.0." 
	 << endl 
	 << "This value doesn't have any physical meaning at moment."
	 << endl
	 << "The Saturated Amplifier gives same output regardless of"
	 << "value of AmpLength" << endl;
   }

NumZSteps = ReadInt("NumZSteps"," ",1,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);

SaturatingPowerdBm = ReadDouble("SaturatingPowerdBm","dBm",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SaturatingPowerLinear = dBm2Linear(SaturatingPowerdBm);

UnsaturatedGaindB = ReadDouble("UnsaturatedGaindB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

UnsaturatedGainLinear = dB2Linear(UnsaturatedGaindB);

UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;

// Input parameters relating to rippled gain



RippleAmplitudeLinear = dB2Linear(ReadDouble("RippleAmplitudedB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile));



RipplePeriod = ReadDouble("RipplePeriod","Hz",
                     0,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);


  // WARNING: If FreqWindow is not a multiple of RipplePeriod 
  // then will have some sort of discontinuity in frequency
  // which will be transformed into a lack of smoothness in time.
  // This probably isn't a problem if noise is already being added
  // to the system. However it's important to be aware of the issue.

 if(DebugLevel >= 2 && TypeGainProfile == RIPPLED_GAIN)
   {
     if(floor(oOptSignal->GetFreqWindow()/RipplePeriod) !=
        oOptSignal->GetFreqWindow()/RipplePeriod)


     *LogFile << "BEWARE OptAmplifier::OptAmplifier " << endl
	     << "FreqWindow is not a multiple of RipplePeriod" << endl;
	  
   }



RelativeRippleShift = ReadDouble("RelativeRippleShift","Hz",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);


ObjectInitialized = 1;

 LogFileSeparator();

 *LogFile << "Processing Opt Amplifier Input Parameters" << endl << endl;

 *LogFile << "LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
 *LogFile << "OutputPowerLinear [mW] = " << 1e3*OutputPowerLinear << endl;
 *LogFile << "LinearNoiseFigOptAmplif = " << LinearNoiseFigOptAmplif << endl;
 *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;
 *LogFile << "CenterFreq = " << CenterFreq << endl;
 *LogFile << "qtPoints = " << qtPoints << endl;
 *LogFile << "DeltaFreq = " << DeltaFreq << endl;


 *LogFile << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
 *LogFile << "UnsaturatedGainLinear = " << UnsaturatedGainLinear << endl;

 *LogFile << "UnsaturatedDifferentialGainLinear = " 
	 << UnsaturatedDifferentialGainLinear << endl;

 *LogFile << "RippleAmplitudeLinear = " << RippleAmplitudeLinear << endl;

 oPolDepAttenuator.InitPolDepAttenuator(oOptSignal);
 oPolDepAttenuator.SetPolDepAttenuation(0.,0.);
   
InFile.close();

UserHasSpecifiedNoiseAmplitudeFactorFlag =  false;

} // ## end OptAmplifier constructor

// ########################################################

OptAmplifier::OptAmplifier(string InFileName, 
                           OptSignal *oOptSignal2, RanNumGen *RNG2,
                           OptSpectrumAnalyzer * PowerMeter2)
{

  // Written by John Zweck, Nov 7th 2000 ;
  // USE THIS CONSTRUCTOR OR PREVIOUS ONE
  // IF YOU CHANGE THIS CONSTRUCTOR CHANGE PREVIOUS ONE TOO (JZ)

oOptSignal = oOptSignal2;
oReduOptAmplifier = new ReduOptAmplifier(InFileName,oOptSignal->oReduOptSignal);

CenterFreq = oOptSignal->GetCenterFreq();
fft = oOptSignal->fft;
sfftM = oOptSignal->sfftM;
TypeSimulation = oOptSignal->GetTypeSimulation();
RNG = RNG2;

qtPoints  = fft->give_N();
DeltaFreq = fft->delta_f; 


ifstream InFile(InFileName.c_str(),ios::in);    
   
   if (!InFile) 
    {
      cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
      exit(1);
    }


LogFileSeparator();
*LogFile << InFileName << endl;

DebugLevel  = ReadInt("DebugLevel"," ",1,3,LOWER_AND_UPPER,MANDATORY,&InFile);


TypeAmplifier  = typeAmplifier(ReadInt("TypeAmplifier"," ",0,6,
                                        LOWER_AND_UPPER,MANDATORY,&InFile));

 switch(TypeAmplifier)
   {
    case SCALAR_SIMPLE: case SCALAR_SATURATED:
      if(TypeSimulation == VECTOR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_SIMPLE: case VECTOR_SATURATED: case VECTOR_FIXED_OUTPUT_POWER:
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:

    if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;
   
   } // ## end switch

*LogFile << "TypeAmplifier = " << WriteTypeAmplifier() << endl;

TypeAmplifierNoise  = typeAmplifierNoise(ReadInt("TypeAmplifierNoise"," ",0,6,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

*LogFile << "TypeAmplifierNoise = " << WriteTypeAmplifierNoise() << endl; 

// added by Walter for MMC and Importance Sampling simulation

if (TypeAmplifierNoise == NOISE_ON_BIASED )
   {
   BiasedNoise= NewCplx(qtPoints);
   }
// end stuff added by Walter


TypeGainProfile = FLAT_GAIN; // Default value

int temp = 0;

temp = ReadInt("TypeGainProfile"," ",1,4,
		   LOWER_AND_UPPER,OPTIONAL_WARNING,&InFile);

if(temp)
  TypeGainProfile = typeGainProfile(temp);


*LogFile << "TypeGainProfile = " << WriteTypeGainProfile() << endl; 

if (TypeGainProfile == FLAT_GAIN_INSIDE_BANDWIDTH) {
   GainBandwidthHz = ReadDouble("GainBandwidthHz","Hz",0,0,
                               LOWER_ONLY,MANDATORY,&InFile);
   if (GainBandwidthHz == 0 ) { //## I think I added this piece of code (Ivan Lima)
      //## Enough to fit all the channels
      //## If you want the whole bandwidth, use the FLAT_GAIN option
      //GainBandwidthHz = 1/fft->delta_t;        
      GainMinRelFreqHz = oOptSignal->GetFrequency(0)
                        -oOptSignal->GetCenterFreq()     
                        -oOptSignal->GetChannSpacing()/2.;

      GainMaxRelFreqHz = 
         oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1)
                          -oOptSignal->GetCenterFreq()       
                          +oOptSignal->GetChannSpacing()/2.;

      //cout << "ChannSpacing " << oOptSignal->GetChannSpacing() << endl;                                           
      //cout << "GainMinRelFreqHz " << GainMinRelFreqHz << endl;
      //cout << "GainMaxRelFreqHz " << GainMaxRelFreqHz << endl;
      //cout <<  oOptSignal->GetFrequency(0) -oOptSignal->GetCenterFreq() << endl;                        
      //cout <<  oOptSignal->GetFrequency(oOptSignal->GetNumChannels()-1) -oOptSignal->GetCenterFreq()  << endl;                  
   }
   else {
      GainMinRelFreqHz = -GainBandwidthHz/2.;
      GainMaxRelFreqHz =  GainBandwidthHz/2.;  
   }
   *LogFile << "GainBandwidthHz = " << GainBandwidthHz << endl;   
   *LogFile << "   -> GainMinRelFreqHz = " << GainMinRelFreqHz << endl;   
   *LogFile << "   -> GainMaxRelFreqHz = " << GainMaxRelFreqHz << endl;   
}   

LinearGainProfile = 0;
AllocateGainProfileMemory(); 

dBGain = ReadDouble("GainOptAmplifier","dB",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);

LinearGainOptAmplif = dB2Linear(dBGain); 
LinearGainFixedTotalOutputPower = LinearGainOptAmplif;

 if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER 
    || TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER  )
   {

     double OutputPowerPerChanneldBm = ReadDouble("OutputPowerPerChanneldBm",
                                              "dBm",0,0,NO_BOUNDS,
                                               OPTIONAL_WARNING,&InFile);

    //OutputPowerLinear = oOptSignal->GetNumChannels()*
    //                     dBm2Linear(OutputPowerPerChanneldBm);


   OutputPowerLinear = dBm2Linear(OutputPowerPerChanneldBm);

   double ResolutionBandwidthHz =  ReadDouble("ResolutionBandwidthHz","Hz",
                                        0,-1,
                                       LOWER_ONLY, OPTIONAL_WARNING,&InFile);

   double  SuperGaussExponent  = ReadDouble("SuperGaussExponent"," ",0,-1,
                               LOWER_ONLY,OPTIONAL_WARNING,2,&InFile);

       
   PowerMeter = PowerMeter2;

   PowerMeter->SetResolutionBandwidthHz(ResolutionBandwidthHz);
   PowerMeter->SetSuperGaussExponent(SuperGaussExponent);
   PowerMeter->SetPowerMeterFlag(1);
   PowerMeter->SetModeOSA(1);
   PowerMeter->SetTypeAmplifierNoise(TypeAmplifierNoise);

   } // ## end if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER) 

/// We either input NoiseFigOptAmplifier (in dB) or SpontaneousEmissionFactor
/// SpontaneousEmissionFactor = n_{sp} is in linear units (not dB)
/// See Desurvire p. 77 for definition.
/// The choice is made by setting the other value to 0.0.
/// The NoiseAmplitudeFactor is computed from either NoiseFigOptAmplifier  
/// or SpontaneousEmissionFactor

NoiseFigOptAmplifier = 0;
SpontaneousEmissionFactor = 0;
 

NoiseFigOptAmplifier = ReadDouble("NoiseFigOptAmplifier","dB",0,0,
				  NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SpontaneousEmissionFactor = ReadDouble("SpontEmissionFactor"," ",
                                        1.0,-1.0,
				        LOWER_ONLY,OPTIONAL_WARNING,&InFile);

ComputeNoiseAmplitudeFactor();

PolDepGainOptAmplifier = ReadDouble(
                               "PolDepGainOptAmplifier","dB"
                               ,0,-1,LOWER_ONLY,OPTIONAL_WARNING,0.0,&InFile);
   
                               
//PolDepGainOrigPhase = new double[qtPoints];                               

//AmpLength = ReadDouble("AmpLength","m",0,-1,LOWER_ONLY,&InFile);

AmpLength = 1.0;

 if(DebugLevel >=3)
   {

    *LogFile << "AmpLength = 1.0." 
	 << endl 
	 << "This value doesn't have any physical meaning at moment."
	 << endl
	 << "The Saturated Amplifier gives same output regardless of"
	 << "value of AmpLength" << endl;
   }

NumZSteps = ReadInt("NumZSteps"," ",1,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);

SaturatingPowerdBm = ReadDouble("SaturatingPowerdBm","dBm",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

SaturatingPowerLinear = dBm2Linear(SaturatingPowerdBm);

UnsaturatedGaindB = ReadDouble("UnsaturatedGaindB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);

UnsaturatedGainLinear = dB2Linear(UnsaturatedGaindB);

UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;

// Input parameters relating to rippled gain



RippleAmplitudeLinear = dB2Linear(ReadDouble("RippleAmplitudedB","dB",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile));



RipplePeriod = ReadDouble("RipplePeriod","Hz",
                     0,-1,LOWER_ONLY,OPTIONAL_WARNING,&InFile);


  // WARNING: If FreqWindow is not a multiple of RipplePeriod 
  // then will have some sort of discontinuity in frequency
  // which will be transformed into a lack of smoothness in time.
  // This probably isn't a problem if noise is already being added
  // to the system. However it's important to be aware of the issue.

 if(DebugLevel >= 2 && TypeGainProfile == RIPPLED_GAIN)
   {
     if(floor(oOptSignal->GetFreqWindow()/RipplePeriod) !=
        oOptSignal->GetFreqWindow()/RipplePeriod)


     *LogFile << "BEWARE OptAmplifier::OptAmplifier " << endl
	     << "FreqWindow is not a multiple of RipplePeriod" << endl;
	  
   }



RelativeRippleShift = ReadDouble("RelativeRippleShift","Hz",
                     1,-1,NO_BOUNDS,OPTIONAL_WARNING,&InFile);


ObjectInitialized = 1;

 LogFileSeparator();

 *LogFile << "Processing Opt Amplifier Input Parameters" << endl << endl;

 *LogFile << "LinearGainOptAmplif = " << LinearGainOptAmplif << endl;
 *LogFile << "OutputPowerLinear [mW] = " << 1e3*OutputPowerLinear << endl;
 *LogFile << "LinearNoiseFigOptAmplif = " << LinearNoiseFigOptAmplif << endl;
 *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;
 *LogFile << "CenterFreq = " << CenterFreq << endl;
 *LogFile << "qtPoints = " << qtPoints << endl;
 *LogFile << "DeltaFreq = " << DeltaFreq << endl;


 *LogFile << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
 *LogFile << "UnsaturatedGainLinear = " << UnsaturatedGainLinear << endl;

 *LogFile << "UnsaturatedDifferentialGainLinear = " 
	 << UnsaturatedDifferentialGainLinear << endl;

 *LogFile << "RippleAmplitudeLinear = " << RippleAmplitudeLinear << endl;

 oPolDepAttenuator.InitPolDepAttenuator(oOptSignal);
 oPolDepAttenuator.SetPolDepAttenuation(0.,0.);
   
InFile.close();

UserHasSpecifiedNoiseAmplitudeFactorFlag = false;

}


//#################################
//## By: Ivan Lima (7/12/01)
OptAmplifier::~OptAmplifier()
{
   //delete [] PolDepGainOrigPhase;
}


// #################################################

void OptAmplifier::AmplifyOptSignal(void)
{

  // This method was rewritten on Feb 9th 2001 by John Zweck.
  // Users are strongly encouraged to use this method for 
  // all calls to amplifiers of all types. If you add another 
  // amplifier model this method should be updated. 

  // The enum types typeAmplifier and typeAmplifierNoise are 
  // defined in ocsGlobalStructures.hh

  double InPower;

  if(DebugLevel >=3)
    {
      LogFileSeparator();

      InPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                 oOptSignal->GetTimeWindow())
              + pulse_energy(sfftM.tOptSignalY,qtPoints,
                             oOptSignal->GetTimeWindow());

     InPower = Linear2dBm(InPower/oOptSignal->GetTimeWindow());
    }


  if(TypeGainProfile == USER_DEFINED_GAIN_PROFILE)
    {
      cerr << "OptAmplifier::AmplifyOptSignal" << endl
	   << "WARNING: USER_DEFINED_GAIN_PROFILE not yet implmented"
	   << endl << "ABORTING NOW!" 
	   << endl;
   
      exit(1);
       
     }

   //## Add the PDG effect, if it exist. (IL)

   if (  TypeAmplifier != NO_AMPLIFICATION
       && PolDepGainOptAmplifier != 0 ) 
    { 
      AddPolDepGainEffect();
    }   
 
switch(TypeGainProfile)
{

 case FLAT_GAIN:
 case FLAT_GAIN_INSIDE_BANDWIDTH:
      AmplifyWithFlatGain();
      break;
 case RIPPLED_GAIN:
      AmplifyWithRippledGain();
      break;
 case USER_DEFINED_GAIN_PROFILE:
      //AmplifyWithUserDefinedGain();
      break;
 case HOMOGENEOUS_MODEL:
      cout << "here" << endl;
	  AmplifyHomogeneous();
	  
      break;
 default:

   *LogFile << " WARNING: OptAmplifier::AmplifyOptSignal"
	   << endl << "Error in TypeGainProfile" << endl;

} // ## end switch(TypeGainProfile)
  if(DebugLevel >=3)
    {
       double OutPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                  oOptSignal->GetTimeWindow())
               + pulse_energy(sfftM.tOptSignalY,qtPoints,
                              oOptSignal->GetTimeWindow());

        OutPower = Linear2dBm(OutPower/oOptSignal->GetTimeWindow());

        *LogFile << "InPower [dBm] = " << InPower << endl;
        *LogFile << "OutPower [dBm] = " << OutPower << endl;
        *LogFile << "Gain [dB] = " << OutPower - InPower << endl;
    }


oReduOptAmplifier->AmplifyOptSignal();

} // ## end AmplifyOptSignal


//########################################################
//#### Add the PDG effect
//## By: Ivan Lima
void OptAmplifier::AddPolDepGainEffect(void)
{
   SetPolDepGainPolarization();
   PolDepAmplification();
}

//########################################################
//#### Set the angle of the PDL element
//## By: Ivan Lima
void OptAmplifier::SetPolDepGainPolarization(void)
{  
   double StokesTotal[4];
   double StokesTotalShifted[3]; 
   //double StokesTmp[4];
   //double StokesShifted[3];    
   //double OrigPhase;
   cplx JonesTmp[2];
   //double thetaZthetaY[2];   
   

   oOptSignal->GetTotalStokesParameters(StokesTotal);           
   
   //## PolDepAttenuatioY was set to 0 => Linear == 1

   //## It will produce a differential gain (that's -loss)                              
   oPolDepAttenuator.SetPolDepAttenuationX(
                  -PolDepGainOptAmplifier
                 *GetDegreeOfPolarization(StokesTotal) );  

   //cout << "PolDepGainOptAmplifier = "<< PolDepGainOptAmplifier << endl;
   //cout << "DOP = " << GetDegreeOfPolarization(StokesTotal) << endl;
   //cout << "PDL = " << -PolDepGainOptAmplifier
   //              *GetDegreeOfPolarization(StokesTotal) <<endl;

   for (int ii = 1; ii < 4; ii++) {
      StokesTotalShifted[ii-1] = StokesTotal[ii]; 
   }   
   
   //## Get the direction of the Polarization Dependent Gain. 
   //## The total polarized power defines the direction.
   //cout << "StokesTotalShifted -> " << StokesTotalShifted[0] << "  "
   //     << StokesTotalShifted[1] << "  " << StokesTotalShifted[2] << endl;
        
   TransRzRyV3to_Xhat( StokesTotalShifted, PolDepGain_thetaZthetaY);
   
}


//########################################################
//#### Set the angle of the PDL element
//## By: Ivan Lima
void OptAmplifier::PolDepAmplification(void)
{  
   //double StokesTotal[4];
   //double StokesTotalShifted[3]; 
   double StokesTmp[4];
   double StokesShifted[3];    
   //double OrigPhase;
   cplx JonesTmp[2];
   //double thetaZthetaY[2];   
   
   //for (int ii = 1; ii < 4; ii++) { 
   //   cout  << ii << "+ StokesTotalShifted ->" << StokesTotalShifted[ii-1] << endl;
   //} 
   //cout << "PolDepGain_thetaZthetaY -> " << PolDepGain_thetaZthetaY[0]
   //     << "  " << PolDepGain_thetaZthetaY[1] << endl;

   //oOptSignal->GetTotalStokesParameters(StokesTotal); 
                        
   //## Rotate all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             

      //if (jj < 2) {
      //   cout << jj << " A-> " << JonesTmp[0] << "  " 
      //        << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      //}	  

      RotatesAboutY(PolDepGain_thetaZthetaY[1], JonesTmp);
      RotatesAboutZ(PolDepGain_thetaZthetaY[0], JonesTmp);                           

      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];
      
      //if (jj < 2) {
      //   cout << jj << " B-> " << JonesTmp[0] << "  " 
      //        << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      //}	        
   }
   oOptSignal->Freq2Time();   
   
   //oOptSignal->GetTotalStokesParameters(StokesTotal); 
   //cout << "Ending *******" << endl;
   //exit(1);   

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   // ## The polarization of the signal is now (S0,-1,0,0)
   // ## Here comes the PDG acting as a PDAttenuator
   oPolDepAttenuator.PolDepAttenuate();

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   //## Rotate back all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             

      //if (jj < 2) {
      //   cout << jj << " C-> " << JonesTmp[0] << "  " 
      //        << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      //}	  


      RotatesAboutZ(-PolDepGain_thetaZthetaY[0], JonesTmp);
      RotatesAboutY(-PolDepGain_thetaZthetaY[1], JonesTmp);      
                    
      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];

      //if (jj < 2) {
      //   cout << jj << " D-> " << JonesTmp[0] << "  " 
      //        << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      //}	        
   }
   oOptSignal->Freq2Time();     

   //oOptSignal->GetTotalStokesParameters(StokesTotal);
   //cout << "End Cycle ***" << endl;
   
} //## End PolDepAmplification()


/*
//########################################################
//#### Set the angle of the PDL element
//## By: Ivan Lima
void OptAmplifier::PolDepAmplification(void)
{  
   //double StokesTotal[4];
   //double StokesTotalShifted[3]; 
   double StokesTmp[4];
   double StokesShifted[3];    
   //double OrigPhase;
   cplx JonesTmp[2];
   //double thetaZthetaY[2];   
   
   //for (int ii = 1; ii < 4; ii++) { 
   //   cout  << ii << "+ StokesTotalShifted ->" << StokesTotalShifted[ii-1] << endl;
   //} 
   cout << "PolDepGain_thetaZthetaY -> " << PolDepGain_thetaZthetaY[0]
        << "  " << PolDepGain_thetaZthetaY[1] << endl;

                        
   //## Rotate all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             
   
      if (abs(JonesTmp[0]) != 0) {
         PolDepGainOrigPhase[jj] = GetPhase(JonesTmp[0]);
      }
      else {
         PolDepGainOrigPhase[jj] = GetPhase(JonesTmp[1]);      
      }   

      if (jj < 2) {
         cout << jj << " A-> " << JonesTmp[0] << "  " 
              << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      }	   
      
      
      //JonesOrigTmp[0] = JonesTmp[0];
      //JonesOrigTmp[1] = JonesTmp[1];      
      
      Jones2StokesKeepPower(JonesTmp,StokesTmp);
      
      for (int ii = 1; ii < 4; ii++) {
         StokesShifted[ii-1] = StokesTmp[ii];    
         //if (jj > qtPoints*0.99)         
         //   cout  << jj <<"  "<< ii << "* Stokes ->" << StokesTmp[ii] << endl;
      }               

      RotatesAboutY( PolDepGain_thetaZthetaY[1],StokesShifted);
      RotatesAboutZ( PolDepGain_thetaZthetaY[0],StokesShifted);     
      
      //## Put together the whole rotated Jones parameters
      for (int ii = 1; ii < 4; ii++) {
         StokesTmp[ii] = StokesShifted[ii-1]; 
      }   
      Stokes2JonesKeepPower(StokesTmp,JonesTmp);
      
      //## I need to restore the phase because it is 
      //#  lost in the Stokes space.
      //#  It actually does not make a lot of difference
      //#  here, since the PolDepAttenuator does 
      //#  its effect in the frequency domain.
      //#  After the polarization in the Jones 
      //#  space is restored, the correct phases 
      //#  will be reconstituted.
      JonesTmp[0] *= exp(jc*PolDepGainOrigPhase[jj]);
      JonesTmp[1] *= exp(jc*PolDepGainOrigPhase[jj]);
      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];


      if (jj < 2) {
         cout << jj << " B-> " << JonesTmp[0] << "  " 
              << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      }	  
      
      //if (   JonesTmp[0] != JonesOrigTmp[0]
      //    || JonesTmp[1] != JonesOrigTmp[1]) {
      //   cout << "Problem: " << jj<<endl
      //        << JonesTmp[0] <<" "<<JonesOrigTmp[0] << endl
      //        << JonesTmp[1] <<" "<<JonesOrigTmp[1] << endl;
      //}    
      
   }
   oOptSignal->Freq2Time();   

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   // ## The polarization of the signal is now (S0,-1,0,0)
   // ## Here comes the PDG acting as a PDAttenuator
   oPolDepAttenuator.PolDepAttenuate();

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   //## Rotate back all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             

      if (jj < 2) {
         cout << jj << " C-> " << JonesTmp[0] << "  " 
              << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      }	  
      
      //OrigPhase = GetPhase(JonesTmp[0]);
      
      Jones2StokesKeepPower(JonesTmp,StokesTmp);
      for (int ii = 1; ii < 4; ii++) {
         StokesShifted[ii-1] = StokesTmp[ii];    
      }               

      RotatesAboutZ(-PolDepGain_thetaZthetaY[0],StokesShifted);
      RotatesAboutY(-PolDepGain_thetaZthetaY[1],StokesShifted);   
      
      //## Put together the whole rotated Jones parameters

      for (int ii = 1; ii < 4; ii++) {
         StokesTmp[ii] = StokesShifted[ii-1]; 
         //if (jj > qtPoints*0.99)
         //   cout << jj <<"  "<< ii << "+ Stokes ->" << StokesTmp[ii] << endl;         
      }   

      Stokes2JonesKeepPower(StokesTmp,JonesTmp);
      
      //## I need to restore the phase because it is 
      //#  lost in the Stokes space
      JonesTmp[0] *= exp(jc*PolDepGainOrigPhase[jj]);
      JonesTmp[1] *= exp(jc*PolDepGainOrigPhase[jj]);      
      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];                  
      
      if (jj < 2) {
         cout << jj << " D-> " << JonesTmp[0] << "  " 
              << JonesTmp[1] << "  " << PolDepGainOrigPhase[jj] << endl;
      }	    
   }

   oOptSignal->Freq2Time();      
   
} //## End of AddPolDepGainEffect()
*/



/*
//########################################################
//#### Add the PDG effect
//## By: Ivan Lima
void OptAmplifier::AddPolDepGainEffect(void)
{  
   double StokesTotal[4];
   double StokesTotalShifted[3]; 
   double StokesTmp[4];
   double StokesShifted[3];    
   //double OrigPhase;
   cplx JonesTmp[2];
   double thetaZthetaY[2];
   
   //double AverageSpectralDensity 
   //      = oOptSignal->GetTotalEnergy()/qtPoints;   
   
   //cplx JonesOrigTmp[2];

   oOptSignal->GetTotalStokesParameters(StokesTotal);
   
   //for (int ii = 1; ii < 4; ii++) { 
   //   cout  << ii << "* StokesTotal ->" << StokesTotal[ii] << endl;
   //}          
   
   //## PolDepAttenuatioY was set to 0 => Linear == 1

   //## It will produce a differential gain (that's -loss)                              
   oPolDepAttenuator.SetPolDepAttenuationX(
                  -PolDepGainOptAmplifier
                 *GetDegreeOfPolarization(StokesTotal) );  

   //cout << "PolDepGainOptAmplifier = "<< PolDepGainOptAmplifier << endl;
   //cout << "DOP = " << GetDegreeOfPolarization(StokesTotal) << endl;
   //cout << "PDL = " << -PolDepGainOptAmplifier
   //              *GetDegreeOfPolarization(StokesTotal) <<endl;

   for (int ii = 1; ii < 4; ii++) {
      StokesTotalShifted[ii-1] = StokesTotal[ii]; 
   }   
   
   //## Get the direction of the Polarization Dependent Gain. 
   //## The total polarized power defines the direction.
   TransRzRyV3to_Xhat( StokesTotalShifted, thetaZthetaY);

   //for (int ii = 1; ii < 4; ii++) { 
   //   cout  << ii << "+ StokesTotalShifted ->" << StokesTotalShifted[ii-1] << endl;
   //} 
                        
   //## Rotate all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             
   
      if (abs(JonesTmp[0]) != 0) {
         PolDepGainOrigPhase[jj] = GetPhase(JonesTmp[0]);
      }
      else {
         PolDepGainOrigPhase[jj] = GetPhase(JonesTmp[1]);      
      }   
      
      //JonesOrigTmp[0] = JonesTmp[0];
      //JonesOrigTmp[1] = JonesTmp[1];      
      
      Jones2StokesKeepPower(JonesTmp,StokesTmp);
      
      for (int ii = 1; ii < 4; ii++) {
         StokesShifted[ii-1] = StokesTmp[ii];    
         //if (jj > qtPoints*0.99)         
         //   cout  << jj <<"  "<< ii << "* Stokes ->" << StokesTmp[ii] << endl;
      }               

      RotatesAboutY( thetaZthetaY[1],StokesShifted);
      RotatesAboutZ( thetaZthetaY[0],StokesShifted);     
      
      //## Put together the whole rotated Jones parameters
      for (int ii = 1; ii < 4; ii++) {
         StokesTmp[ii] = StokesShifted[ii-1]; 
      }   
      Stokes2JonesKeepPower(StokesTmp,JonesTmp);
      
      //## I need to restore the phase because it is 
      //#  lost in the Stokes space.
      //#  It actually does not make a lot of difference
      //#  here, since the PolDepAttenuator does 
      //#  its effect in the frequency domain.
      //#  After the polarization in the Jones 
      //#  space is restored, the correct phases 
      //#  will be reconstituted.
      JonesTmp[0] *= exp(jc*PolDepGainOrigPhase[jj]);
      JonesTmp[1] *= exp(jc*PolDepGainOrigPhase[jj]);
      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];
      
      //if (   JonesTmp[0] != JonesOrigTmp[0]
      //    || JonesTmp[1] != JonesOrigTmp[1]) {
      //   cout << "Problem: " << jj<<endl
      //        << JonesTmp[0] <<" "<<JonesOrigTmp[0] << endl
      //        << JonesTmp[1] <<" "<<JonesOrigTmp[1] << endl;
      //}    
      
   }
   oOptSignal->Freq2Time();   

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   // ## The polarization of the signal is now (S0,-1,0,0)
   // ## Here comes the PDG acting as a PDAttenuator
   oPolDepAttenuator.PolDepAttenuate();

   //cout << "GetTotalEnergy() = " << oOptSignal->GetTotalEnergy() << endl;   

   //## Rotate back all the frequency domain components     
   for (int jj = 0; jj < qtPoints; jj++) {   
      JonesTmp[0] = sfftM.fOptSignalX[jj];
      JonesTmp[1] = sfftM.fOptSignalY[jj];             
      
      //OrigPhase = GetPhase(JonesTmp[0]);
      
      Jones2StokesKeepPower(JonesTmp,StokesTmp);
      for (int ii = 1; ii < 4; ii++) {
         StokesShifted[ii-1] = StokesTmp[ii];    
      }               

      RotatesAboutZ(-thetaZthetaY[0],StokesShifted);
      RotatesAboutY(-thetaZthetaY[1],StokesShifted);   
      
      //## Put together the whole rotated Jones parameters

      for (int ii = 1; ii < 4; ii++) {
         StokesTmp[ii] = StokesShifted[ii-1]; 
         //if (jj > qtPoints*0.99)
         //   cout << jj <<"  "<< ii << "+ Stokes ->" << StokesTmp[ii] << endl;         
      }   

      Stokes2JonesKeepPower(StokesTmp,JonesTmp);
      
      //## I need to restore the phase because it is 
      //#  lost in the Stokes space
      JonesTmp[0] *= exp(jc*PolDepGainOrigPhase[jj]);
      JonesTmp[1] *= exp(jc*PolDepGainOrigPhase[jj]);      
      sfftM.fOptSignalX[jj] = JonesTmp[0];
      sfftM.fOptSignalY[jj] = JonesTmp[1];                  
   }

   oOptSignal->Freq2Time();      
   
} //## End of AddPolDepGainEffect()
*/



// ##########################################################

void OptAmplifier::AmplifyWithFlatGain(void)
{

  switch(TypeAmplifier)
   {

   case NO_AMPLIFICATION:

           break;

    case SCALAR_SIMPLE: 

            ComputeNoiseAmplitudeFactor();
            AmplifyGaussianStatisticsNoise();
            AmplifyOptSignalScalar();
            break;

    case  VECTOR_SIMPLE:
            ComputeNoiseAmplitudeFactor();
            AmplifyGaussianStatisticsNoise();
            AmplifyOptSignalVector();
            break;

    case SCALAR_SATURATED: 
           AmplifyOptSignalSteadyStateSaturatedGain(1); 
           // 1 is the Scalar option
	   // ComputeNoiseAmplitudeFactor called from within 
	   // AmplifyOptSignalSteadyStateSaturatedGain
           break;


   case VECTOR_SATURATED:
           AmplifyOptSignalSteadyStateSaturatedGain(0); 
           // 0 is the Vector option
           // ComputeNoiseAmplitudeFactor called from within 
	   // AmplifyOptSignalSteadyStateSaturatedGain
           break;

 
   case VECTOR_FIXED_OUTPUT_POWER:
            //## The output signal power is fixed at the output
            ComputeGainForFixedOutputPower();
            ComputeNoiseAmplitudeFactor();
	    AmplifyGaussianStatisticsNoise();
            AmplifyOptSignalVector();
            break;
            
   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:
            //## The total power is fixed at the output (Ivan Lima 6/13/01)
            //#  If using FLAT_GAIN_INSIDE_BANDWIDTH make sure 
            //#  that there is no signal comming out from 
            //#  the bandwidth without gain, because this 
            //#  algorithm is based on the power in the whole bandwidth.
            ComputeGainForFixedTotalOutputPower();
	    LinearGainFixedTotalOutputPower = LinearGainOptAmplif;
            ComputeNoiseAmplitudeFactor();
            AmplifyOptSignalVector();    
            //## Limits the total output power
	    //## Is not adjusting for the power from the added noise
            //ComputeGainForFixedTotalOutputPower();                    
   	    //LinearGainFixedTotalOutputPower *= LinearGainOptAmplif;
            //AdjustGainFixedTotalOutputPowerVector();            
            break;

   default: 
           *LogFile << " WARNING: OptAmplifier::AmplifyOptSignal"
                   << endl <<"Error in TypeAmplifier" << endl;

   }//## end TypeAmplifierswitch


} // ## end  AmplifyWithFlatGain

// ###################################################

void OptAmplifier::AmplifyWithRippledGain(void)
{

  if(TypeAmplifierNoise == NOISE_ON_SEMI_ANALYTICAL)
    {
      cerr << "RIPPLED_GAIN with NOISE_ON_SEMI_ANALYTICAL" << endl
	   << "not implemented. ABORTING!!" << endl;
      exit(1);

    }


  AllocateGainProfileMemory();


  switch(TypeAmplifier)
   {

   case NO_AMPLIFICATION:

           break;

    case SCALAR_SIMPLE: 
          cerr << "OptAmplifier::AmplifyWithRippledGain" << endl
	       << "SCALAR_SIMPLE Amplifier with RIPPLED_GAIN " 
	       << " not yet implemented" << endl
	       << "ABORTING NOW!!" << endl;
          exit(1);

	  // To implement just write AmplifyOptSignalScalarWithGainProfile
	  // and uncomment these lines:

      //ComputeNoiseAmplitudeFactorGainProfile();
      //AddRippleToGain();
      //AmplifyOptSignalScalarWithGainProfile();

            break;

    case  VECTOR_SIMPLE:
            ComputeNoiseAmplitudeFactorGainProfile();
            AddRippleToGain();
            AmplifyOptSignalVectorWithGainProfile();
            break;

    case SCALAR_SATURATED: 
          cerr << "OptAmplifier::AmplifyWithRippledGain" << endl
	       << "SCALAR_SATURATED Amplifier with RIPPLED_GAIN " 
	       << " not yet implemented" << endl
	       << "ABORTING NOW!!" << endl;

          exit(1);
          break;


   case VECTOR_SATURATED:
           cerr << "OptAmplifier::AmplifyWithRippledGain" << endl
	       << "VECTOR_SATURATED Amplifier with RIPPLED_GAIN " 
	       << " not yet implemented" << endl
	       << "ABORTING NOW!!" << endl;

           exit(1);
           break;

 
   case VECTOR_FIXED_OUTPUT_POWER:
            ComputeGainForFixedOutputPower();
            ComputeNoiseAmplitudeFactorGainProfile();
            AddRippleToGain();
            AmplifyOptSignalVectorWithGainProfile();
            break;


   default: 
           *LogFile << " WARNING: OptAmplifier::AmplifyOptSignal"
                   << endl <<"Error in TypeAmplifier" << endl;

   }//## end TypeAmplifierswitch


} // ## end  AmplifyWithRippledGain

// ################################################

void  OptAmplifier::ComputeNoiseAmplitudeFactor(void)
{

  // Written by John Zweck, January 2001

  if(TypeAmplifierNoise == NOISE_OFF)
    NoiseAmplitudeFactor = 0.0;
  else  //if(TypeAmplifierNoise == NOISE_ON)
   {

     if(NoiseFigOptAmplifier == 0)
      {
     // Use SpontaneousEmissionFactor - See Desurvire p. 77 (2.33)
     // Also see Curtis Short Course Slide No. 86. 
     // For us a node point is 1 complex mode.

     // We use the sqrt cause we add average energy of n_sph\nu(G-1)
     // to each complex mode

     if (LinearGainOptAmplif >= 1) {  //## Avoid indeterminations (IL)
        NoiseAmplitudeFactor = sqrt( hPlanck*SpontaneousEmissionFactor
                                     *DeltaFreq*(LinearGainOptAmplif-1) );
      }				     
      else {
         NoiseAmplitudeFactor = 0; //## No gain (attenuation) -> No noise
      }
       
      }
    else
    if(SpontaneousEmissionFactor == 0)
     {

      // Use NoiseFigOptAmplifier - See Desurvire p. 99-100 (2.119)
      // Here we are using the large gain approximation 
      // LinearGainOptAmplif >> 1

      LinearNoiseFigOptAmplif = dB2Linear(NoiseFigOptAmplifier);

      if (LinearGainOptAmplif >= 1) {  //## Avoid indeterminations (IL)
         NoiseAmplitudeFactor = sqrt( (hPlanck/2.0)*LinearNoiseFigOptAmplif
                                      *DeltaFreq*(LinearGainOptAmplif-1) );
      }				
      else {
         NoiseAmplitudeFactor = 0;  //## No gain (attenuation) -> No noise
      }      
    }
   else
     {
       cerr << "One of NoiseFigOptAmplifier and SpontaneousEmissionFactor"
	    << endl
	    << "must be zero so we know to use the other one!!";
	 exit(1);
     }
   } // ## end if NOISE_ON


 if(UserHasSpecifiedNoiseAmplitudeFactorFlag)
    NoiseAmplitudeFactor = UserSpecifiedNoiseAmplitudeFactor;

 if(DebugLevel >=2)
       *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;


}// ## end ComputeNoiseAmplitiudeFactor

// ####################################################

void OptAmplifier::SetNoiseAmplitudeFactor(double Value)
{

UserHasSpecifiedNoiseAmplitudeFactorFlag = true;
UserSpecifiedNoiseAmplitudeFactor = Value;

}

// ##################################################

void  OptAmplifier::ComputeNoiseAmplitudeFactorGainProfile(void)
{

  // Written by John Zweck, March 1st 2001
  // NOTE: In the case of a non-flat gain profile
  // we use a different definition of NoiseAmplitudeFactor
  // than is given in ComputeNoiseAmplitudeFactor.
  // We should probably change definition of NoiseAmplitudeFactor
  // in the flat gain case to match the one given here.


  if(TypeAmplifierNoise == NOISE_OFF)
    NoiseAmplitudeFactor = 0.0;
  else //if(TypeAmplifierNoise == NOISE_ON)
   {

     if(NoiseFigOptAmplifier == 0)
      {
     // Use SpontaneousEmissionFactor - See Desurvire p. 77 (2.33)
     // Also see Curtis Short Course Slide No. 86. 
     // For us a node point is 1 complex mode.

     // We use the sqrt cause we add average energy of n_sph\nu(G-1)
     // to each complex mode


     NoiseAmplitudeFactor = sqrt( hPlanck*SpontaneousEmissionFactor
                                  *DeltaFreq);
       
      }
    else
    if(SpontaneousEmissionFactor == 0)
     {

      // Use NoiseFigOptAmplifier - See Desurvire p. 99-100 (2.119)
      // Here we are using the large gain approximation 
      // LinearGainOptAmplif >> 1

      LinearNoiseFigOptAmplif = dB2Linear(NoiseFigOptAmplifier);

      NoiseAmplitudeFactor = sqrt( (hPlanck/2.0)*LinearNoiseFigOptAmplif
                                   *DeltaFreq);
    }
   else
     {
       cerr << "One of NoiseFigOptAmplifier and SpontaneousEmissionFactor"
	    << endl
	    << "must be zero so we know to use the other one!!";
	 exit(1);
     }
   } // ## end if NOISE_ON

 if(DebugLevel >=2)
       *LogFile << "NoiseAmplitudeFactor = " << NoiseAmplitudeFactor << endl;

}// ## end ComputeNoiseAmplitiudeFactorGainProfile

//#####################################################

void OptAmplifier::AmplifyOptSignalVector()
{

   //#### Amplify the signal in the optical frequency domain
   //## By: Ivan Lima, modified and renamed by John Zweck, Feb 9th 2001
   //## Gaussian white noise generator by Ivan Lima (29/5/02)

   // This method used to be called AmplifyOptSignal()

   if (! ObjectInitialized) {
      cerr << "Error: Object of class OptAmplifier not initialized" 
           << endl;
      exit(1);
   }   
   //cout << sq(NoiseAmplitudeFactor)*CenterFreq*2 << endl;

   cplx NormalizedNoiseModeX = 0;
   cplx NormalizedNoiseModeY = 0;
 
   double Freq;
   //## If FLAT_GAIN_INSIDE_BANDWIDTH, it does just that (Ivan Lima 6/13/01)
   for ( int jj = 0; jj < qtPoints; jj++) {
      Freq = CenterFreq + fft->freq(jj);       

      //DeltaFreq*FreqFactorFFT(jj,qtPoints);        
      if ( TypeGainProfile != FLAT_GAIN_INSIDE_BANDWIDTH 
           || (    fft->freq(jj) > GainMinRelFreqHz  
                && fft->freq(jj) < GainMaxRelFreqHz ) ) {  
		
         if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
            NormalizedNoiseModeX = exp(jc*2*pi*RNG->GetRanNum());
            NormalizedNoiseModeY = exp(jc*2*pi*RNG->GetRanNum());	 
         }
         else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
            NormalizedNoiseModeX = ( RNG->GetGaussianDeviate()
   		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);
            NormalizedNoiseModeY = ( RNG->GetGaussianDeviate()
   		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.); 
         }		
	 
         sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    *NormalizedNoiseModeX;

         sfftM.fOptSignalY[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalY[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    *NormalizedNoiseModeY;		
/*
         sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    //*RNG->GetGaussianDeviate()
	             *exp(jc*2*pi*RNG->GetRanNum());

         sfftM.fOptSignalY[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalY[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    //*RNG->GetGaussianDeviate()
                      *exp(jc*2*pi*RNG->GetRanNum() );   
*/                                                    
      }                 
                             
   }   
   //     cout << "NoiseDensity = "<<  2*sq(NoiseAmplitudeFactor*sqrt(CenterFreq)
   //              /sqrt(DeltaFreq)) << endl;     

   fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);


} // ## end AmplifyOptSignalVector


//#########################################################
//## By: Ivan Lima (6/2/2001)
void OptAmplifier::AdjustGainFixedTotalOutputPowerVector(void)
{ 
   //## The adjusted gain has to be in LinearGainOptAmplif
   double Freq;
   for ( int jj = 0; jj < qtPoints; jj++) {
      Freq = CenterFreq + fft->freq(jj); 
      //DeltaFreq*FreqFactorFFT(jj,qtPoints);        

      sfftM.fOptSignalX[jj] 
	= sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj];

      sfftM.fOptSignalY[jj] 
	= sqrt(LinearGainOptAmplif)*sfftM.fOptSignalY[jj];                             
   }   
   oOptSignal->Freq2Time();
   //fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   //fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);   
} //## end AdjustGainFixedTotalOutputPower()




// ############################################################

void OptAmplifier::AmplifyOptSignalVectorWithGainProfile()
{

//  Amplifys the signal in the optical frequency domain
//  with a previously defined gain profile

// By John zweck, March 1st 2001.
//## Gaussian white noise by Ivan Lima


cplx NormalizedNoiseModeX = 0;
cplx NormalizedNoiseModeY = 0;  
double Freq;

for ( int jj = 0; jj < qtPoints; jj++) 
 {
   Freq = CenterFreq + fft->freq(jj); 

      if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
         NormalizedNoiseModeX = exp(jc*2*pi*RNG->GetRanNum());
         NormalizedNoiseModeY = exp(jc*2*pi*RNG->GetRanNum());	 
      }
      else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
         NormalizedNoiseModeX = ( RNG->GetGaussianDeviate()
		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);
         NormalizedNoiseModeY = ( RNG->GetGaussianDeviate()
		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.); 
      }


      sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    *NormalizedNoiseModeX;

      sfftM.fOptSignalY[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalY[jj]
         + NoiseAmplitudeFactor*sqrt(Freq) 
                    *NormalizedNoiseModeY;			    

   /*
   sfftM.fOptSignalX[jj] 
	= sqrt(LinearGainProfile[jj])*sfftM.fOptSignalX[jj]
	+ NoiseAmplitudeFactor*sqrt(Freq*(LinearGainProfile[jj]-1)) 
                    //*RNG->GetGaussianDeviate()
	            *exp(jc*2*pi*RNG->GetRanNum());

       sfftM.fOptSignalY[jj] 
	= sqrt(LinearGainProfile[jj])*sfftM.fOptSignalY[jj]
	+ NoiseAmplitudeFactor*sqrt(Freq*(LinearGainProfile[jj]-1)) 
                    //*RNG->GetGaussianDeviate()
                      *exp(jc*2*pi*RNG->GetRanNum() );   
   */		      
                             
 } //## end for jj loop   

fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
fft->IFFT(sfftM.fOptSignalY,sfftM.tOptSignalY);

} // ## end AmplifyOptSignalVectorWithGainProfile


//#############################################################

void OptAmplifier::AmplifyOptSignalScalar()
{

// #### Amplify the signal in the optical frequency domain
// ## By: Ivan Lima
// Modified by John Zweck
//## Gaussian white noise by Ivan Lima


   if (! ObjectInitialized) 
   {
      cerr << "Error: Object of class OptAmplifier not initialized" 
           << endl;
      exit(1);
   }   
 
 
   cplx NormalizedNoiseModeX = 0; 
   double Freq;

   for ( int jj = 0; jj < qtPoints; jj++) 
    {
     Freq = CenterFreq + fft->freq(jj); 

      if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
         NormalizedNoiseModeX = exp(jc*2*pi*RNG->GetRanNum()); 
      }
      else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
         NormalizedNoiseModeX = ( RNG->GetGaussianDeviate()
		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);
      }

      else if(TypeAmplifierNoise == NOISE_ON_NO_RNG ) 
      {	  // add by Jonathon
         NormalizedNoiseModeX =  sfftM.fOptSignalX[jj] / 
	                          abs(sfftM.fOptSignalX[jj]);
      }

      if (TypeAmplifierNoise == NOISE_ON_BIASED )
        {
	// added by Walter
	sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + BiasedNoise[jj];

	}
      else
        {	 
        sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + NoiseAmplitudeFactor*sqrt(Freq)*NormalizedNoiseModeX;

	}			    
   }	


   fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
  
   
   for ( int jj = 0; jj < qtPoints; jj++) 
    {
    
     sfftM.fOptSignalY[jj] = 0.0;
     sfftM.tOptSignalY[jj] = 0.0;
    }


} // ## end AmplifyOptSignalScalar()

// ####################################

// ## The following methods added by John Zweck

// ###########################################

void  OptAmplifier::SetGain(double dBGain2)
{

 dBGain = dBGain2;
 LinearGainOptAmplif = dB2Linear(dBGain); 
 //## The following line is required for the noiseless simulation
 LinearGainFixedTotalOutputPower = LinearGainOptAmplif; 
 if(DebugLevel >=3)
   {

     *LogFile << "WARNING: Gain reset to dBGain [dB] = "
         << dBGain << endl;
   }

}

// ###############################################



void OptAmplifier::SetTypeAmplifier(typeAmplifier TypeAmplifier2)
{

  TypeAmplifier = TypeAmplifier2;

  if(DebugLevel >=2)
    {
     *LogFile << "WARNING: Setting type of amplifier to " 
	  << WriteTypeAmplifier()
	  << endl; 
    }

 switch(TypeAmplifier)
   {
    case SCALAR_SIMPLE: case SCALAR_SATURATED:
      if(TypeSimulation == VECTOR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_SIMPLE: case VECTOR_SATURATED: 

    if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
      break;

   case VECTOR_FIXED_OUTPUT_POWER:   case VECTOR_FIXED_TOTAL_OUTPUT_POWER: 

       if(TypeSimulation == SCALAR)
	{
	 cerr << "OptAmplifier constructor: TypeAmplifier and TypeSimulation"
	     << endl
	     << "in conflict" << endl;
	 exit(1);
	}
     //  cerr << "SetTypeAmplifier = VECTOR_FIXED_OUTPUT_POWER is ILLEGAL. "
     //  << "Do it in constructor. ABORTING" << endl;
     //exit(1);

     break;

   } // ## end switch

} // ## end SetTypeAmplifier

// #############################################

void OptAmplifier::SetTypeAmplifierNoise(
                                typeAmplifierNoise TypeAmplifierNoise2)
{

 TypeAmplifierNoise = TypeAmplifierNoise2;

 if(TypeAmplifier == VECTOR_FIXED_OUTPUT_POWER || 
    TypeAmplifier == VECTOR_FIXED_TOTAL_OUTPUT_POWER)
      PowerMeter->SetTypeAmplifierNoise(int(TypeAmplifierNoise));

 ComputeNoiseAmplitudeFactor();

  if(DebugLevel >=2)
    {
      *LogFile << "WARNING: Setting type of amplifier noise to " 
	  << WriteTypeAmplifierNoise()
	  << endl;
    } 
}

// ###########################################

void OptAmplifier::SetTypeGainProfile(typeGainProfile TypeGainProfile2)
{

  TypeGainProfile = TypeGainProfile2;
  
  if(DebugLevel >=2)
    {
      *LogFile << "WARNING: Setting type of amplifier gain profile to " 
	  << WriteTypeGainProfile()
	  << endl;
    }

  AllocateGainProfileMemory();

}

void OptAmplifier::SetAmpLength(double AmpLength2)
{
// cout << " SetAmpLength =" << AmpLength2 << endl;
  AmpLength = AmpLength2;

}

// ############################################

void OptAmplifier::SetPumpPowerTime(double PumpPowerT)
{

	for(int StepT =0;StepT < NumTSteps;StepT ++){
		PumpInput[0 * NumTSteps + StepT] = PumpPowerT ;
 	} 
//	cout << "SetPumpPowerTime  =" << PumpPowerT  << endl;

}

// ############################################

void OptAmplifier::SetSaturationFactorAdjuctment(double SA)
{	
	SaturationFactorAdjuctment = SA;
	SaturationFactor = DopantConcentration * EffectArea / LifeTime;
	SaturationFactor *= SaturationFactorAdjuctment;
//	cout << "SetSaturationFactorAdjuctment  ="
//		 << SaturationFactorAdjuctment  << endl;

}


// ##############################################

string OptAmplifier::WriteTypeAmplifier(void)
{

 switch(TypeAmplifier)
   {

    case NO_AMPLIFICATION:

           return "NO_AMPLIFICATION";
           break;

    case SCALAR_SIMPLE: 

            return "SCALAR_SIMPLE";
            break;
	    
    case  VECTOR_SIMPLE:
            return "VECTOR_SIMPLE";
            break;

    case SCALAR_SATURATED: 
           return "SCALAR_SATURATED";
           break;


   case VECTOR_SATURATED:
           return "VECTOR_SATURATED";
           break;

   case VECTOR_FIXED_OUTPUT_POWER:
          return "VECTOR_FIXED_OUTPUT_POWER";
          break;

   case VECTOR_FIXED_TOTAL_OUTPUT_POWER:
          return "VECTOR_FIXED_TOTAL_OUTPUT_POWER";
          break;          

   default:
     cerr << "ERROR: Invalid choice of TypeAmplifier" << endl;
     cerr << "ABORTING NOW!!" << endl;

     exit(1);


    }

}

// ######################################

string OptAmplifier::WriteTypeAmplifierNoise(void)
{

 switch(TypeAmplifierNoise)
   {

    case NOISE_OFF:

           return "NOISE_OFF";
           break;

    case NOISE_ON_CONST_POWER_RANDOM_PHASE:

           return "NOISE_ON_CONST_POWER_RANDOM_PHASE";
           break;

    case NOISE_ON_GAUSSIAN_WHITE:

           return "NOISE_ON_GAUSSIAN_WHITE";
           break;

    case NOISE_ON_NO_RNG:

	   return "NOISE_ON_NO_RNG";
	   break;
		     
    case NOISE_ON_SEMI_ANALYTICAL:

           return "NOISE_ON_SEMI_ANALYTICAL";
           break;

    case NOISE_ON_BIASED:

           return "NOISE_ON_BIASED";
           break;



 	default:
		cerr << "With TypeAmplifierNoise must between 0 and 5" << endl 
		<< "Aborting now! " << endl;
		exit(1); 

    }

}

// ##################################

string OptAmplifier::WriteTypeGainProfile(void)
{

 switch(TypeGainProfile)
   {

    case FLAT_GAIN:    
           return "FLAT_GAIN";
           break;

    case FLAT_GAIN_INSIDE_BANDWIDTH:
           return "FLAT_GAIN_INSIDE_BANDWIDTH";
           break;                     

    case RIPPLED_GAIN:

           return "RIPPLED_GAIN";
           break;


    case USER_DEFINED_GAIN_PROFILE:

           return "USER_DEFINED_GAIN_PROFILE";
           break;
	
    case HOMOGENEOUS_MODEL:
			
          return "HOMOGENEOUS_MODEL";
	  break;
	
	default:
		cerr << "Wrong TypeGainProfile " << endl 
		<< "Aborting now! " << endl;
		exit(1); 

    }

}

// #########################################


double OptAmplifier::AmplifyOptSignalSteadyStateSaturatedGain
                     (int ScalarNotVectorFlag)
{

  //LogFileSeparator();

  //*LogFile << "AmplifyOptSignalSteadyStateSaturatedGain" << endl;

  double InputPower;

  if(ScalarNotVectorFlag) // Using a scalar NLS solver
    {
      InputPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                oOptSignal->GetTimeWindow())/
                                              oOptSignal->GetTimeWindow();
    }
  else // Using a vector-NLS or Manakov Solver
    {
      InputPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                oOptSignal->GetTimeWindow())
                 + pulse_energy(sfftM.tOptSignalY,qtPoints,
                                oOptSignal->GetTimeWindow());

      InputPower = InputPower/oOptSignal->GetTimeWindow();

    }

   
  if(TypeAmplifierNoise == NOISE_ON_SEMI_ANALYTICAL)
    {
     InputPower += 
          oOptSignal->GetAccumulatedTotalNoiseSpectralDensityGaussianStatistics()*
          oOptSignal->GetNoiseBandwidthGaussianStatistics();
    }

  //*LogFile << "InputPower [dBm] = " << Linear2dBm(InputPower) << endl;

  LinearGainOptAmplif = SteadyStateSaturatedGain(InputPower);

  ComputeNoiseAmplitudeFactor();

  AmplifyGaussianStatisticsNoise();

  //*LogFile << "Gain [dB] = " << Linear2dB(LinearGainOptAmplif) << endl;

   if(ScalarNotVectorFlag) // Using a scalar NLS solver
    {
      AmplifyOptSignalScalar();
    }
  else // Using a vector-NLS or Manakov Solver
    {
      AmplifyOptSignalVector();
    }

} // ## end AmplifyOptSignalSteadyStateSaturatedGain

// ###############################################

double OptAmplifier::SteadyStateSaturatedGain(double InputPower)
{

  // Dec 14th, 2000: This method still needs more testing

  // Written by John Zweck, Nov 30th, 2000

  // For theory see Agrawal [NLO,p476ff] and Ruo Mei's 
  // "Comparison of theory and experiment" paper.

  // This routine returns the linear saturated gain G for which
  // P_out = G*P_in, where P_in is the input power to
  // the amplifier and P_out is the output power.
  // The power is computed by taking the average power
  // of the pulse train. 

  // G = LinearGainOptAmplif

  // The ODE which describes the evolution of the power in the 
  // amplifier is dP/dz = g(z,P(z))*P
  // where z is distance along the amplifier.

  // Ignoring frequency dependence of the gain we choose 
  //
  //  g(z,P) = g_0/( 1+ tau*P(z)/P_sat )

  // We solve the ODE using Euler's method.

  // Variables set in the constructor are:

  // AmpLength
  // NumZSteps
  // SaturatingPower (P_sat)
  

  // LinearUnsaturatedDifferentialGain , g_0 = g(z,0)
  // Note that g_0 = (1+tau)*g(z,P_sat) also
  // This could be a way to compute tau I guess.....

  // CHECK: denominator in Eulers method eqn below is correct,
  // and work out how to get SaturatingPower.

  double DeltaZ =  AmpLength/float(NumZSteps);
  
  double OutputPower = InputPower;

  for(int ii=0;ii<NumZSteps;ii++)
    {
      OutputPower += UnsaturatedDifferentialGainLinear*DeltaZ*OutputPower/
                     (1+ (OutputPower/SaturatingPowerLinear));
    }
  
  return OutputPower/InputPower; 

}// ## end SteadyStateSaturatedGain

// ####################################################

void OptAmplifier::WriteSteadyStateGainVersusPower(string OutFileName,
                   double InputPowerMaxLinear, int NumPowerSteps)
{

  // Dec 14th, 2000: This method still needs more testing

  // Written by John Zweck, Nov 30th, 2000

  // This method writes out (10*log10(G) , 10*log10(P_in/1mW)) pairs
  // in units (dB, dBm)
  // where G_0 = exp(g_0 AmpLength) is the UnsaturatedGain
  // and G is the SteadyStateSaturatedGain

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile) 
   {
      *LogFile << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }



 double DeltaPower = InputPowerMaxLinear/NumPowerSteps;
 
 LogFileSeparator();

 *LogFile << "WriteSteadyStateGainVersusPower" << endl << endl;


 *LogFile << "InputPowerMaxLinear [W] = " << InputPowerMaxLinear << endl;
 *LogFile << "NumPowerSteps = " << NumPowerSteps << endl;
 *LogFile << "DeltaPower = " <<  DeltaPower << endl;


 double SaturatedGain;
 double InputPower;

 for(int ii = 1; ii <= NumPowerSteps; ii++) 
     // Start at ii==1 as need InputPower > 0
   {
     InputPower = ii*DeltaPower;
     SaturatedGain =  SteadyStateSaturatedGain(InputPower);

     OutFile << 10*log10(InputPower/1e-3)  << " " 
	     << 10*log10(SaturatedGain) << endl;

   } // end for loop

 OutFile.close();


}// ## end WriteSteadyStateGainVersusPower

// #########################################################

void OptAmplifier::AmplifyGaussianStatisticsNoise(void)
{

  if(TypeAmplifierNoise == NOISE_ON_SEMI_ANALYTICAL)
    {

      // First increment the AccumulatedTotalNoiseSpectralDensity
      // Then set the NoiseAmplificationFactor to 0 so noise-free signal
      // doesn't have noise added to it.

      if(TypeGainProfile == FLAT_GAIN_INSIDE_BANDWIDTH)
	{
          cerr << "FLAT_GAIN_INSIDE_BANDWIDTH with NOISE_ON_SEMIANALYTICAL"
	       << "is not debugged. ABORTING!" << endl;
	  exit(1);

	}

      switch(TypeAmplifier)
	{

	 case NO_AMPLIFICATION:
	  break;

         case SCALAR_SIMPLE: case SCALAR_SATURATED:

          oOptSignal->GainOfAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                      LinearGainOptAmplif);

          oOptSignal->AddToAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                      GetSpectralNoisePowerDensityAddedPerPolarization());


          break;

         case VECTOR_SIMPLE: case VECTOR_SATURATED: 
         case VECTOR_FIXED_OUTPUT_POWER:
              
          oOptSignal->GainOfAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                      LinearGainOptAmplif);

          oOptSignal->AddToAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
                         2*GetSpectralNoisePowerDensityAddedPerPolarization());


          
          break;

	 case VECTOR_FIXED_TOTAL_OUTPUT_POWER:

	  // Not implemented yet
          break;

	default:
          cerr << "Switch error in OptAmplifier::AmplifyGaussianStatisticsNoise"
	       << endl
	       << "ABORTING NOW!!" << endl;
          exit(1);


	} // end switch

      NoiseAmplitudeFactor = 0;
    
    } // end if

} // ## end OptAmplifier::AmplifyGaussianStatisticsNoise

//#########################################################

//## This method should be executed when the optical 
//#  signal object contains the intensity that one wants 
//#  to have after the amplification.
//## This method should be executed only once at 
//#  the beggining
//## By: Ivan Lima, 5/14/01
void OptAmplifier::SetFixedOutputPowerFromCurrentOptSignal(void)
{

   OutputPowerLinear = PowerMeter->AmpPowerMeterInputLinear();
	
}

//#########################################################
//## This method should be executed when the optical 
//#  signal object contains the intensity that one wants 
//#  to have after the amplification.
//## This method should be executed only once at 
//#  the beggining.
//#  Different from SetFixedOutputPowerFrom..., this method gets
//#  the total power (/NumChannels) from the frequency domain using a flat 
//#  frequency dependence transmission.
//## By: Ivan Lima, 5/14/01
void OptAmplifier::SetFixedTotalOutputPowerFromCurrentOptSignal(void)
{
   OutputPowerLinear = oOptSignal->GetAveragePowerFreqDomain()
                      /oOptSignal->GetNumChannels();
   //cout << "Freq domain = "<<  oOptSignal->GetAveragePowerFreqDomain()<< endl;
   //cout << "Time domain = "<<  oOptSignal->GetAveragePower()<< endl;
   //exit(1);
}

// ###########################################################

void OptAmplifier::ComputeGainForFixedOutputPower(void)
{

  // Written by John Zweck, Apr 10th 2001.
  // This method computes the Gain required to achieve a fixed
  // output power level. 
 
  // The output power level is required to be the peak power
  // of a channel as measured by an OSA (PowerMeter) of the
  // specified input ResolutionBandwidth

  // This method computes the input power level to the amplifier
  // using a PowerMeter of the same ResolutionBandwidth and 
  // then computes the gain from the computed input and required
  // output powers

   double InputPowerLinear = PowerMeter->AmpPowerMeterInputLinear();

   if(InputPowerLinear == 0)
     {
      cerr << "Divide by zero error in " << endl
  	   << "OptAmplifier::ComputeGainForFixedOutputPower" << endl
	   << "ABORTING NOW!!" << endl;

      exit(1);
     }

   if(InputPowerLinear > OutputPowerLinear)
     {
       cerr << "OptAmplifier::ComputeGainForFixedOutputPower" << endl
	    << "InputPower > OutputPower" << endl
	    << "InputPowerdBm = " << Linear2dBm(InputPowerLinear)
	    << "OutputPowerdBm = " << Linear2dBm(OutputPowerLinear)
	    << "SETTING LinearGainOptAmplif = 1 SO YOU CAN FIX PARAMETERS"
	    << endl;

       LinearGainOptAmplif = 1;

     }

   LinearGainOptAmplif = OutputPowerLinear/InputPowerLinear;

   if(DebugLevel >= 3)
     *LogFile << "ComputeGainForFixedOutputPower: LinearGainOptAmplif = " 
	     << LinearGainOptAmplif << endl;

} // ## end ComputeGainForFixedOutputPower

//##########################################################
//## This method is Based on the method 
//#  ComputeGainForFixedOutPower(), but the former is not 
//#  frequency selective (favoring the signal) as the latter.
//## by: Ivan Lima
void OptAmplifier::ComputeGainForFixedTotalOutputPower(void)
{

  //## This method computes the Gain required to achieve a fixed
  //#  TOTAL output power level. 
  //#  The total power determines the saturation level.

  if(TypeAmplifierNoise==NOISE_ON_SEMI_ANALYTICAL)
    {
      cerr << "VECTOR_FIXED_TOTAL_OUTPUT_POWER with NOISE_ON_SEMI_ANALYTICAL"
	   << "not implemented. ABORTING!" << endl;
      exit(1);
    }

   double InputPowerLinear = oOptSignal->GetAveragePowerFreqDomain()
                            /oOptSignal->GetNumChannels();
   //cout <<"InputPowerLinear = " <<  oOptSignal->GetAveragePowerFreqDomain() << endl;

   if(InputPowerLinear == 0)
     {
      cerr << "Divide by zero error in " << endl
  	   << "OptAmplifier::ComputeGainForFixedTotalOutputPower" << endl
	   << "ABORTING NOW!!" << endl;

      exit(1);
     }

//   if(InputPowerLinear > OutputPowerLinear)
//     {
//       cerr << "OptAmplifier::ComputeGainForFixedTotalOutputPower" << endl
//   	    << "InputPower > OutputPower" << endl
//   	    << "InputPowerdBm = " << Linear2dBm(InputPowerLinear)
//	    << "OutputPowerdBm = " << Linear2dBm(OutputPowerLinear)
//	    << "SETTING LinearGainOptAmplif = 1 SO YOU CAN FIX PARAMETERS"
//	    << endl;
//       LinearGainOptAmplif = 1;
//     }

   LinearGainOptAmplif = OutputPowerLinear/InputPowerLinear;

   if(DebugLevel >= 3)
     *LogFile << "ComputeGainForFixedOutputPower: LinearGainOptAmplif = " 
	     << LinearGainOptAmplif << endl;


} // ## end ComputeGainForFixedOutputPower


// ###########################################################

void OptAmplifier::ComputeGainForFixedOutputPower_OLD(void)
{

  // Written by John Zweck, Feb 19th 2001.
  // This method computes the Gain required to achieve a fixed
  // output power level. 
  // At the moment we compute the total input power.

  // Because the noise builds up over distance to specify
  // a fixed output power level per channel we should really
  // compute the total power in each channel (over a specified
  // frequency window) and then average overth channels to 
  // get an estimate of input signal (as oppsoed to noise) power
  // from which to compute the gain. Let's add this feature
  // when we find we need to do so!!

  // The performance of the methods needs to be evaluated 
  // more.

   double InputPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                oOptSignal->GetTimeWindow())
                     + pulse_energy(sfftM.tOptSignalY,qtPoints,
                                oOptSignal->GetTimeWindow());

   InputPower = InputPower/oOptSignal->GetTimeWindow();

   if(InputPower == 0)
     {
      cerr << "Divide by zero error in " << endl
  	   << "OptAmplifier::ComputeGainForFixedOutputPower" << endl
	   << "ABORTING NOW!!" << endl;

      exit(1);
     }

   if(InputPower > OutputPowerLinear*oOptSignal->GetNumChannels())
     {
       cerr << "OptAmplifier::ComputeGainForFixedOutputPower" << endl
	    << "InputPower > OutputPower" << endl
	    << "ABORTING NOW!!" << endl;

       exit(1);

     }

   LinearGainOptAmplif = OutputPowerLinear*oOptSignal->GetNumChannels()/
                         InputPower;

   if(DebugLevel >= 3)
     *LogFile << "ComputeGainForFixedOutputPower: LinearGainOptAmplif = " 
	     << LinearGainOptAmplif << endl;

} // ## end ComputeGainForFixedOutputPower_OLD

// ###########################################################

void OptAmplifier::AddRippleToGain(void)
{


  // This method assumes that LinearGainOptAmplif has been 
  // computed.
  // The method changes the gain from being flat to having a 
  // gain profile which is given by a sine function in frequency
  // with amplitude RippleAmplitude, period RipplePeriod (input in Hz)
  // and shift RelativeRippleShift.

  // Since we are aiming for a given  RippleAmplitude in dB let's
  // add the ripple in dB rather than linearly.

  // RelativeRippleShift shifts sine wave to right.

  // WARNING: If FreqWindow is not a multiple of RipplePeriod 
  // then will have some sort of discontinuity in frequency
  // which will be transformed into a lack of smoothness in time.
  // This probably isn't a problem if noise is already being added
  // to the system. However it's important to be aware of the issue.

  for(int jj=0;jj<qtPoints;jj++) 
    {
      double Freq = fft->freq(jj);  // Freq is in Hz

      LinearGainProfile[jj] = dB2Linear( Linear2dB(LinearGainOptAmplif) + 
	        Linear2dB(RippleAmplitudeLinear)*
	        sin(2*pi*(Freq - RelativeRippleShift)/RipplePeriod) );
      
    }  // ## end for jj loop

} // ## end AddRippleToGain

// ###########################################################

void OptAmplifier::AllocateGainProfileMemory()
{

  if(    TypeGainProfile != FLAT_GAIN 
      && TypeGainProfile != FLAT_GAIN_INSIDE_BANDWIDTH)
    {
      if(LinearGainProfile)
	delete [] LinearGainProfile;

          LinearGainProfile = new double[qtPoints];

          if(!LinearGainProfile)
	    {
              cerr << "OptAmplifier::AllocateGainProfileMemory" << endl
		   << "Malloc error.ABORTING NOW!!"
		   << endl;
              exit(1);
	    }

      

    } // ## end if


} // ## end AllocateGainProfileMemory

// ##############################################

void OptAmplifier::WriteGainProfile(string OutFileName)
{

  // Written by John Zweck, March 2nd, 2001


if(LinearGainProfile) // To ensure memory is allocated!
{

 ofstream OutFile(OutFileName.c_str(),ios::out);

 if (!OutFile) 
   {
     *LogFile << "OptAmplifier::WriteGainProfile Error: " << endl
              << OutFileName << "could not be created."<<endl;
      exit(1);
   }

OutFile << "% Freq [GHz] , Gain [dB] " << endl;

for(int jj = qtPoints/2; jj < qtPoints; jj++) 
  {
    OutFile <<  1e-9*fft->freq(jj)
            << "\t" 
            <<  Linear2dB(LinearGainProfile[jj])
            << endl; 
  }

for(int jj = 0; jj < qtPoints/2; jj++) 
  {
    OutFile <<  1e-9*fft->freq(jj)
            << "\t" 
            <<  Linear2dB(LinearGainProfile[jj])
            << endl; 
  }


 OutFile.close();

} // ## end if(LinearGainProfile)

}// ## end WriteGainProfile

// #########################################################

//####################################################################
//## This method is useful for computation of the Q-factor using
//#  the semi-analytical algorithm
//## By: Ivan Lima
 void OptAmplifier::AmplifyAlternateOptSignalWithoutNoiseWithPrevioslyComputedGain(
            OptSignal *oOptSignal2)
{
   if (TypeAmplifier != NO_AMPLIFICATION) {
   if (PolDepGainOptAmplifier != 0 )  {
      OptSignal *oOptSignalTmp = oOptSignal;
      SwitchOptSignalData(oOptSignal2);
      PolDepAmplification();
      SwitchOptSignalData(oOptSignalTmp);
   }      
   double Freq;
   for ( int jj = 0; jj < qtPoints; jj++) {
      Freq = CenterFreq + fft->freq(jj);       
      //DeltaFreq*FreqFactorFFT(jj,qtPoints);        
      if ( TypeGainProfile != FLAT_GAIN_INSIDE_BANDWIDTH 
           || (    fft->freq(jj) > GainMinRelFreqHz  
                && fft->freq(jj) < GainMaxRelFreqHz ) ) {  

         oOptSignal2->sfftM.fOptSignalX[jj] 
            = sqrt(LinearGainOptAmplif)
	      *oOptSignal2->sfftM.fOptSignalX[jj];
         oOptSignal2->sfftM.fOptSignalY[jj] 
            = sqrt(LinearGainOptAmplif)
	      *oOptSignal2->sfftM.fOptSignalY[jj];  
		
//         oOptSignal2->sfftM.fOptSignalX[jj] 
//            = sqrt(LinearGainFixedTotalOutputPower)
//	      *oOptSignal2->sfftM.fOptSignalX[jj];
//         oOptSignal2->sfftM.fOptSignalY[jj] 
//            = sqrt(LinearGainFixedTotalOutputPower)
//	      *oOptSignal2->sfftM.fOptSignalY[jj];                                                 
      }                                              
   }      

   oOptSignal2->Freq2Time();
   }   
}	    

//####################################################
//## Set the parameters for a saturated amplifier operating
//## in gain compression
//## By: Ivan Lima (9/3/2002)
void OptAmplifier::SetGainCompressionSaturatedAmplifier(
                   double GainTarget_dB, double GainCompression_dB, 
		   double OutputPowerTarget)
{
   double GainTargetLinear = dB2Linear(GainTarget_dB);
   double GainCompressionLinear = dB2Linear(GainCompression_dB);

   //## If the amplifier was not saturated, set it now
   SetTypeAmplifier(VECTOR_SATURATED);
   UnsaturatedGainLinear =  GainTargetLinear * GainCompressionLinear;  
   UnsaturatedDifferentialGainLinear = log(UnsaturatedGainLinear)/AmpLength;   
   //SaturatingPowerLinear = OutputPowerTarget/GainTargetLinear
   //                       *(UnsaturatedGainLinear - 2)
   //			  /(2*log(2));   
   SaturatingPowerLinear = (GainTargetLinear - 1.)/GainTargetLinear
                           *OutputPowerTarget
			   /log(UnsaturatedGainLinear/GainTargetLinear);
   cout << "UnsaturatedDifferentialGainLinear = " 
        << UnsaturatedDifferentialGainLinear << endl;  
   cout << "UnsaturatedGainLinear = " << UnsaturatedGainLinear << endl;  
   cout << "SaturatingPowerLinear = " << SaturatingPowerLinear << endl;
}		   

//####################################################
//## This function is used to change the optical signal
//#  on the fly.
//#  Check the constructor before using it, since 
//#  it might be out of date.
//##  It only cares about the data, while leaving the
//#   other parameters of the original OptSignal.
//##   Thus, the OptSignal has to have the same number of 
//#   FFT points and everything else.
//## By: Ivan Lima (7/24/01)
void OptAmplifier::SwitchOptSignalData(OptSignal *oOptSignal2)
{
   oOptSignal = oOptSignal2;
   fft = oOptSignal2->fft;
   sfftM = oOptSignal2->sfftM;   
   qtPoints = oOptSignal2->GetQtPoints();
   oPolDepAttenuator.SwitchOptSignalData(oOptSignal2);
   //CenterFreq = oOptSignal->GetCenterFreq();
}


//############################################################
//############################################################
//############################################################
//############################################################
//############################################################
//############################################################
//############################################################
//############################################################
// Following function is written by Jonathan Hu


void OptAmplifier::AmplifyWithUserDefinedGain()
{
// this function is written by Jonathan Hu
// out of date
  double InputSignalPower;

  InputSignalPower = pulse_energy(sfftM.tOptSignalX,qtPoints,
                                oOptSignal->GetTimeWindow())/
                                              oOptSignal->GetTimeWindow();
  InputSignalPower = Linear2dBm(InputSignalPower);
  cout << "Wavelength=" << oOptSignal->GetCenterWavelength()<<  "  Input=" << InputSignalPower
  		<< "  AmplifierPumpPowerC=" << AmplifierPumpPowerC << endl; 
  
  LinearGainOptAmplif = GetEDFAGain(oOptSignal->GetCenterWavelength(),InputSignalPower,AmplifierPumpPowerC);
  
  cout << " Gain=" << LinearGainOptAmplif << endl;
  //  ComputeNoiseAmplitudeFactor();

  //*LogFile << "Gain [dB] = " << Linear2dB(LinearGainOptAmplif) << endl;
  
  AmplifyOptSignalScalar();

}  // end of AmplifyWithUserDefinedGain

// ############################################################
// this function is written by Jonathan Hu
// read in the gain coeff data
// talk to me to get a sample
void OptAmplifier::ReadGainProfile(string InFileName, int SignalPowerInputIndex)
{
 char str[40];
 double Wavelength;
 

 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }
 ifstream InFile2(InFileName.c_str(),ios::in);   
 if (!InFile2) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }
 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   

 StartWavelength = ReadDouble("StartWavelength","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 DeltaWavelength = ReadDouble("DeltaWavelength","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 StartPumpPower = ReadDouble("StartPumpPower","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 DeltaPumpPower = ReadDouble("DeltaPumpPower","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 MidPower = ReadDouble("MidPower","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 NumGainWidth = ReadInt("NumGainWidth","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 
 for(int ii=0;ii<NumGainWidth;ii++){   
	for (int jj=0;jj<TotPumpPowerIndex;jj++){	
       InFile2  >> Wavelength;		
	   Wavelength *= 1e-9;	    
	   if(Wavelength>1900e-9) {
		    cerr << "Error: wavelength "
        		   << " read in Wavelength error"<< endl;
  			cerr << "in -" << SignalPowerInputIndex << "dBm file" ;
			exit(1);
       }
	   InFile2 >> oEDFAGain[SignalPowerInputIndex* TotPumpPowerIndex * NumGainWidth 
	   						+ ii * TotPumpPowerIndex + jj];
		cout << "  " << 	oEDFAGain[SignalPowerInputIndex* TotPumpPowerIndex * NumGainWidth 
			   						+ ii * TotPumpPowerIndex + jj];
	}	
 }
 InFile.close();
 InFile2.close(); 
   
} // ## end ReadGainProfile

//############################################################
// this function is written by Jonathan Hu
// out of date
void OptAmplifier::ReadGainPowerDependent(string InFileName)
{
 char str[40];
 double Wavelength,data, Power;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   

/* StartPumpPowerC = ReadDouble("StartPumpPowerC","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 DeltaPumpPowerC = ReadDouble("DeltaPumpPowerC","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);
 */
 NumGainWidthDep = ReadInt("NumGainWidthDep","",0,0,NO_BOUNDS,
                     OPTIONAL_WARNING,&InFile);

 for(int ii=0;ii<NumGainWidthDep;ii++){ 
       InFile >> data;
	   InFile >> data; 
	   InFile >> data;
	   InFile >> Power;
	   Power *= 1e-3; // chance to A units

	   if(Power>1) {
		    cerr << "Error: power "
        		   << " read in file" << InFileName << " power error"<< endl;
  			exit(1);
       }

	   InFile >> data;
	   InFile >> oEDFAGainPowerDependent[ii];
	   InFile >> data;
	   InFile >> data; 
	   InFile >> data;
 }	
 
 InFile.close();
   
} // ## end ReadGainPowerDependent

//############################################################
// this function is written by Jonathan Hu
// out of date
void OptAmplifier::ReadInEDFAGain(string InDir3)
{  
 char str[40];
// int * SignalPowerdBm;
// SignalPowerdBm = new int[TotSignalPowerIndex];
int SignalPowerdBm[6] = {-4, -7, -10, -13, -15, -17};

 	for(int jj=0;jj<TotSignalPowerIndex;jj++){
		sprintf(str,"n%ddbm.in",-1*SignalPowerdBm[jj]);
		ReadGainProfile(InDir3 + str, jj);
		sprintf(str,"PowerDepandend.in");
		ReadGainPowerDependent(InDir3 + str); 

	}
	
// delete [] SignalPowerdBm;

} // ## end ReadInEDFAGain





//############################################################
// out of date
double OptAmplifier::GetEDFAGain(double Wavel, double SignalPower, double PumpPowerC)
// this function is written by Jonathan Hu
// SignalPower in units of dBm, PumpPowerC in units of A
{
//	double StartWavelength = 1539.5e-9;
//	double DeltaWavelength = 0.5e-9;
//	double StartPumpPower = 17e-3;
//	double DeltaPumpPower = 2e-3;
//	double MidPower = 100e-3;

	int index50 = int((50e-3 - StartPumpPower) /DeltaPumpPower);
	int index100 = int((100e-3 - StartPumpPower) /DeltaPumpPower);
	int index150 = int((150e-3 - StartPumpPower) /DeltaPumpPower);
	int index;
	double ratio;
	double GainBeforePumpPowerAdj[2], GainBeforeSignalPowerAdj[2];
	int NumPumpPowerIndex;
	int SignalPowerdBm[6] = {-4,-7,-10,-13,-15,-17};	
	int NumSignalPowerIndex=-1;
	int offsetIndex;

	do{
		NumSignalPowerIndex++;
	}while(SignalPower<SignalPowerdBm[NumSignalPowerIndex]);
	NumSignalPowerIndex--;
	cout << " NSignalPowIndex="	<< NumSignalPowerIndex;
	offsetIndex = NumSignalPowerIndex * NumGainWidth * TotPumpPowerIndex;
	index = int((Wavel - StartWavelength) / DeltaWavelength);
	ratio = (Wavel - (index * DeltaWavelength + StartWavelength) ) / DeltaWavelength;
	index = offsetIndex + index * TotPumpPowerIndex ; 
	if(PumpPowerC > MidPower)
		{ index++; cout << " large";}
	cout << " index=" << index << " ratio=" << ratio;	
	GainBeforePumpPowerAdj[0] = oEDFAGain[index] + ( oEDFAGain[index+TotPumpPowerIndex] - oEDFAGain[index] ) * ratio;
	index++;
	GainBeforePumpPowerAdj[1] = oEDFAGain[index] + ( oEDFAGain[index+TotPumpPowerIndex] - oEDFAGain[index] ) * ratio;
	cout << " index=" << index << "  " << TotPumpPowerIndex << " ratio=" << ratio;
	cout << " GainBeforeP=" << GainBeforePumpPowerAdj[0] << " GainBeforeP=" << GainBeforePumpPowerAdj[1] ;
	// for pump power adjustment
	NumPumpPowerIndex = int((PumpPowerC - StartPumpPower) / DeltaPumpPower);
	if(PumpPowerC>MidPower){
		ratio = (oEDFAGainPowerDependent[NumPumpPowerIndex] - oEDFAGainPowerDependent[index100])
					/ (oEDFAGainPowerDependent[index150] - oEDFAGainPowerDependent[index100]);
	}else{
		ratio = (oEDFAGainPowerDependent[NumPumpPowerIndex] - oEDFAGainPowerDependent[index50])
					/ (oEDFAGainPowerDependent[index100] - oEDFAGainPowerDependent[index50]);
	}
	GainBeforeSignalPowerAdj[0] = GainBeforePumpPowerAdj[0] 
				+ ratio * (GainBeforePumpPowerAdj[1] -GainBeforePumpPowerAdj[0]);
	
	cout << " GainbeofreSignalP=" << GainBeforeSignalPowerAdj[0] << endl;

	// do it again when NumSignalPowerIndex increase by one.
	NumSignalPowerIndex++;			
	cout << " NSignalPowIndex="	<< NumSignalPowerIndex;
	offsetIndex = NumSignalPowerIndex * NumGainWidth * TotPumpPowerIndex;
	index = int((Wavel - StartWavelength) / DeltaWavelength);
	ratio = (Wavel - (index * DeltaWavelength + StartWavelength) ) / DeltaWavelength;
	index = offsetIndex + index * TotPumpPowerIndex ; 
	if(PumpPowerC > MidPower)
		{ index++; cout << " large";}
	cout << " index=" << index << " ratio=" << ratio;	
	GainBeforePumpPowerAdj[0] = oEDFAGain[index] + ( oEDFAGain[index+TotPumpPowerIndex] - oEDFAGain[index] ) * ratio;
	index++;
	GainBeforePumpPowerAdj[1] = oEDFAGain[index] + ( oEDFAGain[index+TotPumpPowerIndex] - oEDFAGain[index] ) * ratio;
	cout << " index=" << index << "  " << TotPumpPowerIndex << " ratio=" << ratio;
	cout << " GainBeforeP=" << GainBeforePumpPowerAdj[0] << " GainBeforeP=" << GainBeforePumpPowerAdj[1] ;
	// for pump power adjustment
	NumPumpPowerIndex = int((PumpPowerC - StartPumpPower) / DeltaPumpPower);
	if(PumpPowerC>MidPower){
		ratio = (oEDFAGainPowerDependent[NumPumpPowerIndex] - oEDFAGainPowerDependent[index100])
					/ (oEDFAGainPowerDependent[index150] - oEDFAGainPowerDependent[index100]);
	}else{
		ratio = (oEDFAGainPowerDependent[NumPumpPowerIndex] - oEDFAGainPowerDependent[index50])
					/ (oEDFAGainPowerDependent[index100] - oEDFAGainPowerDependent[index50]);
	}
	GainBeforeSignalPowerAdj[1] = GainBeforePumpPowerAdj[0] 
				+ ratio * (GainBeforePumpPowerAdj[1] -GainBeforePumpPowerAdj[0]);
	
	cout << " GainbeofreSignalP=" << GainBeforeSignalPowerAdj[1] << endl;
			
	NumSignalPowerIndex--;			
	return(	GainBeforeSignalPowerAdj[0] + (GainBeforeSignalPowerAdj[1]	- GainBeforeSignalPowerAdj[0])
				* (SignalPower - SignalPowerdBm[NumSignalPowerIndex] ) 
				/ (SignalPowerdBm[NumSignalPowerIndex+1] - SignalPowerdBm[NumSignalPowerIndex]));
	
} // end GetEDFAGain



//############################################################

// this function is written by Jonathan Hu
// read in ReadInErParameter

void OptAmplifier::ReadInErParameter(string InDir3)
{  
 
 char str[40];
 sprintf(str,"SignalCrossSection.in");
 ReadSignalCrossSection(InDir3 + str);
// sprintf(str,"GainFlatenFilter.in");
// ReadGainFlatenFilter(InDir3 + str); 
// sprintf(str,"PumpCrossSection.in");
// ReadPumpCrossSection(InDir3 + str); 
} // ## end ReadInErParameter

// ############################################################
// read in the measured data for Gain flaten filter
// by Jonathan
void OptAmplifier::ReadInGainFlatenFilter(string InFileName)
{
 char str[40];
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 


 InFile >>  WavelengthStartFilter;
 InFile >>  WavelengthStepFilter;
 InFile >>  NumFilterWidth;
 
 oGainFlatenFilter = new double [NumFilterWidth];


 for(int ii=0;ii<NumFilterWidth;ii++)
    {  
		InFile >> temp;
	   if(temp < 1400 || temp > 1700){
	   		cerr << "Error: in wavelength input."<< endl;
    		exit(1);
    	}
       InFile >> oGainFlatenFilter[ii];
	  
    }


 InFile.close();

} // end ReadInGainFlatenFilter 


// #########################################################
// get Gain Flaten filter using linear interpolation
double OptAmplifier::GetGainFlatenFilter(double Wavelength)
{
 double Interpolate;
 int InterpolateIndex;

 InterpolateIndex = int((Wavelength - WavelengthStartFilter)/WavelengthStepFilter);
 
 if(InterpolateIndex < 0 || InterpolateIndex > NumFilterWidth ){
      *LogFile << "Error: Signal Wavelength is not in the right range"<<endl;
      exit(1);	
	  
	  return (0);
 
 
 }
 
 Interpolate = oGainFlatenFilter[InterpolateIndex] + 
 	(oGainFlatenFilter[InterpolateIndex + 1] - oGainFlatenFilter[InterpolateIndex])
	* ( ( Wavelength - WavelengthStartFilter - WavelengthStepFilter * InterpolateIndex ) / WavelengthStepFilter ) ;

 return( exp(Interpolate/10*log(10)) ) ;

} // endl GetSignalAbs

// ############################################################
// Inorder to get gain flat for 28 round trip, I use a Polynomial function to correct the gain profile.
// Attention: I only correct by 0.6% which is reasonable in measured data
void OptAmplifier::ReadInPolynomialFilter(string InFileName)
{
 char str[40];
 double temp;
 PolynomialFilterAppliedFlag = 1;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 


 InFile >>  WavelengthStartPolynomialFilter;
 InFile >>  WavelengthStepPolynomialFilter;
 InFile >>  NumPolynomialFilterWidth;
 
 oPolynomialFilter = new double [NumPolynomialFilterWidth];


 for(int ii=0;ii<NumPolynomialFilterWidth;ii++)
    {  
		InFile >> temp;
	   if(temp < 1400 || temp > 1700){
	   		cerr << "Error: in wavelength input."<< endl;
    		exit(1);
    	}
       InFile >> oPolynomialFilter[ii];
	  
    }


 InFile.close();

} // end ReadInPolynomialFilter 


// #########################################################

double OptAmplifier::GetPolynomialFilter(double Wavelength)
{
 double Interpolate;
 int InterpolateIndex;

 InterpolateIndex = int((Wavelength - WavelengthStartPolynomialFilter)/WavelengthStepPolynomialFilter);
 
 if(InterpolateIndex < 0 || InterpolateIndex > NumPolynomialFilterWidth ){
      *LogFile << "Error: Signal Wavelength is not in the right range"<<endl;
	  *LogFile << "Wavelength = " << Wavelength 
	  		<< "  WavelengthStartPolynomialFilter=" << WavelengthStartPolynomialFilter << endl;
      exit(1);	
	  
	  return (0);
 
 
 }
 
 Interpolate = oPolynomialFilter[InterpolateIndex] + 
 	(oPolynomialFilter[InterpolateIndex + 1] - oPolynomialFilter[InterpolateIndex])
	* ( ( Wavelength - WavelengthStartPolynomialFilter - WavelengthStepPolynomialFilter * InterpolateIndex ) 
	/ WavelengthStepPolynomialFilter ) ;

 return( exp(Interpolate/10*log(10)) ) ;

} // endl 

// ############################################################
// read in Signal cross section
// please take a look at sample file
void OptAmplifier::ReadSignalCrossSection(string InFileName)
{
 char str[40];
 double temp, CorrectABSFactor, CorrectEMSFactor;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
// NumCrossSectionWidth = 1005;
 InFile >> PumpWavelengthForGain;
 InFile >>  WavelengthStartCrossSection;
 InFile >>  WavelengthStepCrossSection;
 InFile >>  NumCrossSectionWidth;
 InFile >>  CorrectABSFactor;
 InFile >>  CorrectEMSFactor; 
 oSignalAbs = new double [NumCrossSectionWidth];
 oSignalEms = new double [NumCrossSectionWidth];

// this is the read in file for OASIX or HuWDM
 for(int ii=0;ii<NumCrossSectionWidth;ii++)
    {   
       InFile >> temp;
	   if(temp < 1400 || temp > 1700){
	   		cerr << "Error: in wavelength input."<< endl;
    		exit(1);
    	}
       InFile >> oSignalAbs[ii]; 
	   oSignalAbs[ii] *= CorrectABSFactor;
//	   	oSignalAbs[ii] *= 3.0/2.31;
	   InFile >> oSignalEms[ii]; 
	   oSignalEms[ii] *= CorrectEMSFactor;   
//	   	oSignalEms[ii] *= 3.0/2.31;
	  
    }



 InFile >> temp;
 if(temp>0){
 	cerr << "SignalCrossSection Input Error"<< endl
		<< "Aborting now!!!";
	exit(1);
 }
	


 InFile.close();

} // end ReadSignalCrossSection

// ############################################################

/*void OptAmplifier::ReadPumpCrossSection(string InFileName)
{
 char str[40];
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 if (!InFile) 
  	{
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
 int Width = 1005;
 oPumpAbs = new double [Width]; 
 for(int ii=0;ii<Width;ii++)
    {   
		InFile >> temp;
 	   if(temp < 800 || temp > 1100){
	   		cerr << "Error: in wavelength input."<< endl;
    		exit(1);
    	}      	
		InFile >> oPumpAbs[ii]; 
	   

    }


 InFile.close();

} // end ReadPumpCrossSection
*/
// #######################################################
// get signal absorption coeff in particular wavelength
// by Jonathan Hu
double OptAmplifier::GetSignalAbs(double Wavelength)
{
 double Interpolate;
 
 
 
 int InterpolateIndex;

 if( Wavelength > 1700e-9 || Wavelength < 1400e-9){
  	  *LogFile << "Wavelength = " << Wavelength << endl;
      *LogFile << "Error: 11Signal Wavelength is not in the right range"<<endl;
      exit(1);
   }

 InterpolateIndex = int((Wavelength - WavelengthStartCrossSection)/WavelengthStepCrossSection);
 
 if(InterpolateIndex < 0 || InterpolateIndex > NumCrossSectionWidth ){
      *LogFile << "Error: Signal Wavelength is not in the right range"<<endl;
      exit(1);
 }
 
 Interpolate = oSignalAbs[InterpolateIndex] + 
 	(oSignalAbs[InterpolateIndex + 1] - oSignalAbs[InterpolateIndex])
	* ( ( Wavelength - WavelengthStartCrossSection - WavelengthStepCrossSection * InterpolateIndex ) / WavelengthStepCrossSection ) ;

 return( Interpolate/4.343 );

} // endl GetSignalAbs

// #######################################################
// get signal emission coeff in particular wavelength
// by Jonathan Hu
double OptAmplifier::GetSignalEms(double Wavelength)
{
 double Interpolate;

 int InterpolateIndex;

 if( Wavelength > 1700e-9 || Wavelength < 1400e-9){
      *LogFile << "Error: Signal Wavelength is not in the right range"<<endl;
      exit(1);
   }

 InterpolateIndex = int((Wavelength - WavelengthStartCrossSection)/WavelengthStepCrossSection);
 
 if(InterpolateIndex < 0 || InterpolateIndex > NumCrossSectionWidth ){
      *LogFile << "Error: Signal Wavelength is not in the right range"<<endl;
      exit(1);
 }
  
 Interpolate = oSignalEms[InterpolateIndex] + 
 	(oSignalEms[InterpolateIndex + 1] - oSignalEms[InterpolateIndex])
	* ( ( Wavelength - WavelengthStartCrossSection - WavelengthStepCrossSection * InterpolateIndex ) / WavelengthStepCrossSection ) ;

 return( Interpolate/4.343 );

} // endl GetSignalAbs

// #######################################################
// get pump absorption
// by Jonathan Hu
double OptAmplifier::GetPumpAbs(double Wavelength)
{
 double Interpolate;

	if(Wavelength==980e-9){
		return(PumpABS/4.343);
	}else{
		cerr << "Error: Pump Wavelength is not in the right range" << endl;
        exit(1);
	}
	
/* int InterpolateIndex;

 if( Wavelength < 875e-9 || Wavelength > 1075e-9){
   
      cerr << "Error: Pump Wavelength is not in the right range"<<endl;
      exit(1);
   }

 InterpolateIndex = int((Wavelength - WavelengthStart)/WavelengthStep);
 
 Interpolate = oPumpAbs[InterpolateIndex] + 
 	(oPumpAbs[InterpolateIndex + 1] - oPumpAbs[InterpolateIndex])
	* ( ( Wavelength - WavelengthStart - WavelengthStep * InterpolateIndex ) / WavelengthStep ) ;

 return( Interpolate/4.343 );
*/

} // endl GetPumpAbs

// ########################################################
// calcuate the gain for Signal (6.6) in Becker's book
// calculate the power for noise (6.7) in Becker's book
// by Jonathan Hu 
void OptAmplifier::PropagateSignalandNoise(OptSignal *NoiseFreeSignal,OptSignal * Noise)
{   

double Freq;
if(TypeAmplifierNoise != NOISE_OFF)
{
 cplx NormalizedNoiseMode; 
 cplx NormalizedNoiseModeX;
 fft->fftshift(NoiseFreeSignal->sfftM.fOptSignalX);
 fft->fftshift(Noise->sfftM.fOptSignalX);
 
 *LogFile << " run to here 1" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }

 switch(TypeGainProfile)
 {

 case FLAT_GAIN:
 	for ( int jj = 0; jj < qtPoints; jj++) {
         NoiseFreeSignal->sfftM.fOptSignalX[jj] *= sqrt(LinearGainOptAmplif); 
         Noise->sfftM.fOptSignalX[jj] *= sqrt(LinearGainOptAmplif);  
  	}
		 *LogFile << " run to here FG2" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }
	
    for ( int jj = 0; jj < qtPoints; jj++) 
    {
      Freq = CenterFreq + fft->freq(jj); 

      if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
         NormalizedNoiseModeX = exp(jc*2*pi*RNG->GetRanNum()); 
      }
      else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
         NormalizedNoiseModeX = ( RNG->GetGaussianDeviate()
		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);
      }
      else if(TypeAmplifierNoise == NOISE_ON_NO_RNG ) 
      {	  // add by Jonathon
         NormalizedNoiseModeX =  1;
      }
      Noise->sfftM.fOptSignalX[jj] 
         = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj])) + sq(NoiseAmplitudeFactor) * Freq * NormalizedNoiseModeX);			    
    }  		 \
	*LogFile << " run to here FG3" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }
	break;


 case HOMOGENEOUS_MODEL:
 	for ( int jj = 0; jj < qtPoints; jj++) {
         NoiseFreeSignal->sfftM.fOptSignalX[jj] *= sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps]);  
         Noise->sfftM.fOptSignalX[jj] = abs(Noise->sfftM.fOptSignalX[jj] * sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1]
		 								/ oSignalPower[jj*NumZSteps]));  
  	}
 *LogFile << " run to here 2" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }

   		for ( int jj = 0; jj < qtPoints; jj++) {

   		NormalizedNoiseMode = 0;
   

         if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
   	  		LinearGainOptAmplif = oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps];	     
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
 			Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
           	
         }else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
	        NoiseAmplitudeFactor = sqrt(oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
   	  		LinearGainOptAmplif = oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps];
            NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
   		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);
 			Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
			

         }else if(	TypeAmplifierNoise == NOISE_ON_NO_RNG ) {	
	        NoiseAmplitudeFactor = 0;
   	  		LinearGainOptAmplif = (oSignalPower[jj*NumZSteps + NumZSteps - 1] + oNoiseFPower[jj*NumZSteps + NumZSteps - 1])
							 / oSignalPower[jj*NumZSteps];
          	//Noise->sfftM.fOptSignalX[jj] *= sqrt((sq(abs(Noise->sfftM.fOptSignalX[jj])) 
					//+ oNoiseFPower[jj*NumZSteps + NumZSteps - 1]) / sq(abs(Noise->sfftM.fOptSignalX[jj])));
   	 		Noise->sfftM.fOptSignalX[jj] = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj])) 
				+ oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
			}

 		}
	 *LogFile << " run to here 3" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }
	 for(int ii=0;ii<NumSigCh;ii++)  {
	 NoiseFreeSignal->sfftM.fOptSignalX[ii]  *= sqrt(GetGainFlatenFilter(oSignalWaveLth[ii]));
	 Noise->sfftM.fOptSignalX[ii]  *= sqrt(GetGainFlatenFilter(oSignalWaveLth[ii]));
 	 if(PolynomialFilterAppliedFlag==1){
 	 	NoiseFreeSignal->sfftM.fOptSignalX[ii]  *= sqrt(GetPolynomialFilter(oSignalWaveLth[ii]));
 	 	Noise->sfftM.fOptSignalX[ii]  *= sqrt(GetPolynomialFilter(oSignalWaveLth[ii]));
	 }
 } 
 break;
 default:

   cout << " WARNING: OptAmplifier::AmplifyOptSignal"
	   << endl << "Error in TypeGainProfile" << endl;

 } // ## end switch(TypeGainProfile) 
 *LogFile << " run to here 4" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }
 fft->ifftshift(NoiseFreeSignal->sfftM.fOptSignalX);
 fft->IFFT(NoiseFreeSignal->sfftM.fOptSignalX,NoiseFreeSignal->sfftM.tOptSignalX);
 fft->ifftshift(Noise->sfftM.fOptSignalX);
// fft->IFFT(Noise->sfftM.fOptSignalX,Noise->sfftM.tOptSignalX);
 } 
 *LogFile << " run to here 5" << endl;
 for ( int jj = 0; jj < qtPoints; jj++) {
     *LogFile <<    jj << "  " << Noise->sfftM.fOptSignalX[jj] << endl;  
 }
} // endl of pr
// ########################################################
// Here, Homogeneous model is the same as Giles' model
// the whole process is solving rate equation and propagate the signal in Er doped fiber
// by Jonathan Hu
void OptAmplifier::AmplifyHomogeneous()
{

 fft->fftshift(sfftM.fOptSignalX);

 for(int ii=0;ii<NumSigCh;ii++)  {
   	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() );
//      *LogFile << "oSignalWaveLth" << ii << " = " << 
//                          oSignalWaveLth[ii] ;
      oSignalPower[ii*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[ii*NumZSteps] < 1e-100)
			oSignalPower[ii*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
//	  *LogFile << "    oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps] << endl;
 }



 if(TypeAmplifierNoise != NOISE_OFF)
 {

  for(int ii=0;ii<NumNoiCh;ii++)
    {
      oNoiseWaveLth[ii] = oSignalWaveLth[ii];  
//	  cout << "oNoiseWaveLth[ii]=" << oNoiseWaveLth[ii]  << endl;
	}
  for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
 	for(int ii=1;ii<NumZSteps;ii++){
   		oNoiseBPower[NNoi * NumZSteps + ii]=0;
		oNoiseFPower[NNoi * NumZSteps + ii]=0;
	}  
  }
} 


 PropagatSignal();

/*
NumSigCh = 1;
oSignalWaveLth[0] = 1550e-9;
//  oSignalWaveLth[0] = LightSpeed / 193.413e12;
//   oSignalWaveLth[0] = LightSpeed / 305.913e12;
 oSignalPower[0] = 0.5e-3; 
  oPumpPower[0] = 150e-3; 
// cout << "Power in = " << oOptSignal->GetAveragePowerFreqDomainInsideChannels() << endl;
cout << "NumSigCh = " << NumSigCh << "  NumNoiCh = " << NumNoiCh << endl;
cout << "oSignalWaveLth = " << oSignalWaveLth[0]  << endl;
 PropagatSignal();

//exit(1);	
 cout 	<< AmpLength << "  " << oNoiseFPower[qtPoints/2*NumZSteps + NumZSteps - 1]
									/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[qtPoints/2])/LightSpeed)) 	
			<< "  " << oNoiseBPower[qtPoints/2*NumZSteps + NumZSteps  - 1 ] 
						/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[qtPoints/2])/LightSpeed)) << endl;

//for(int ii = 0;ii < qtPoints; ii++){ 
//	*LogFile <<  oNoiseWaveLth[ii]  << "   " << oNoiseFPower[ii*NumZSteps + NumZSteps - 1]
//				/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[ii])/LightSpeed)) << endl; 
//}

*LogFile << " NoiseWaveLth = " << oNoiseWaveLth[qtPoints/2] << endl ;

//for(int ii = 1;ii < 9; ii++){ 
//	*LogFile <<  ii  << "   " << oNoiseFPower[NumZSteps*qtPoints/2 + ii * 100]
//				/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[qtPoints/2])/LightSpeed)) << endl;
//
//} 

cout << "speedlight = " << LightSpeed << endl;
cout << "DeltaFreq = " << DeltaFreq << endl;



for (int ii = 1; ii<9 ; ii ++)
{ cout 	<< ii << "  "
		<< "  " << 10 * log10(oSignalPower[ii * 100 ] / oSignalPower[0]) <<	endl;

}
for (int ii = 0; ii<NumZSteps ; ii ++)
{ 
	cout 	<< ii*DeltaZ << "  " << oSignalPower[ii ] 	
			<< "  " << oPumpPower[ ii  ]  << endl;
			
 }
for (int ii = 0; ii<NumZSteps ; ii ++)
{ 
	cout 	<< ii*DeltaZ << "  " << oNoiseFPower[qtPoints/2*NumZSteps + ii ]
									/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[qtPoints/2])/LightSpeed)) 	
			<< "  " << oNoiseBPower[qtPoints/2*NumZSteps + NumZSteps - ii - 1 ] 
						/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[qtPoints/2])/LightSpeed)) << endl;
			
 }
  cout 	<< "SignalWave=" << oSignalWaveLth[0] 
      	<< "   SignalPower start= " << oSignalPower[0] 
		<< "   SignalPower end= " << oSignalPower[NumZSteps-1] 
		<< " Gain(dB)=" << 10 * log10(oSignalPower[NumZSteps-1 ] / oSignalPower[0]) <<	endl;

// for(int ii=0;ii<NumNoiCh;ii++)  {
//   	  cout << "NoiWave=" << oNoiseWaveLth[ii]  
//      		<< "   NoiFPower=" << oNoiseFPower[ii*NumZSteps + NumZSteps - 1 ] << "   mw/nm=" 
//			<< 1.0e3*oNoiseFPower[ii*NumZSteps + NumZSteps - 1 ]/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[ii])/LightSpeed)) 	
//			<< "   NoiBPower=" << oNoiseBPower[ii*NumZSteps + NumZSteps - 1 ] << "   mw/nm=" 
//			<< 1.0e3*oNoiseBPower[ii*NumZSteps + NumZSteps - 1 ]/(1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[ii])/LightSpeed)) << endl;
// }
// exit(1); 
 
// cout << "Power out = " << oOptSignal->GetAveragePowerFreqDomainInsideChannels() << endl;
*/
/*
  NumSigCh = 1; 
  oSignalWaveLth[0] = 1550.12e-9 ;
  
  oSignalPower[0] =  -9.95;
  oSignalPower[0] = exp(oSignalPower[0] / 10 * log(10)) * 1e-3;
  
  SaturationFactorAdjuctment =1.0;
  
  PropagatSignal();

  cout  <<  " Signal gain = "  << 10 * log(oSignalPower[NumZSteps-1]*1e3)/ log(10) +9.95  << "  " << endl;
 *LogFile  <<  " Signal gain = "  << 10 * log(oSignalPower[NumZSteps-1]*1e3)/ log(10) +9.95  << "  " << endl;

 cout << endl;
*/


/*
for(int Ntest=170;Ntest<1800;Ntest+=80)
{
  *LogFile << (double)Ntest/10 << "  " ;
//  if(Ntest == 410 ||Ntest ==1770)
	cout  <<  (double)Ntest/10 << "  " ;
// for(SaturationFactorAdjuctment=0.7;SaturationFactorAdjuctment<1.4;SaturationFactorAdjuctment +=0.1)
// SaturationFactorAdjuctment = 0.7;

for(int ii=3;ii<=9;ii += 3){ 
//  Ntest = 500;
  oPumpPower[0] = PumpCurrent2Power((double)Ntest /10 * 1e-3);
//  cout << oPumpPower[0] ;
//  exit(1);
   
  NumSigCh = 1;
  oSignalWaveLth[0] = 1549.12e-9 ;
  
//  oSignalPower[0] =  -9.95;
  oSignalPower[0] = -1.0 * (double)ii;
  oSignalPower[0] = exp(oSignalPower[0] / 10 * log(10)) * 1e-3;
  
  PropagatSignal();

//  cout << "Input = " << oSignalPower[0] << "=dbm" << 10 * log(oSignalPower[0]*1000)/ log(10) << endl
//   		<< "Output = " << oSignalPower[NumZSteps-1] << "=dBm" << 10 * log(oSignalPower[NumZSteps-1]*1000)/ log(10) << endl
//  		<< " Amplength=" << AmpLength  << endl
//		<< " Gain or Loss =" << 10 * log(oSignalPower[NumZSteps-1] /  oSignalPower[0]   )/ log(10) << endl;

  *LogFile  <<  10 * log(oSignalPower[NumZSteps-1]*1000)/ log(10) << "   ";
// if(Ntest == 410 ||Ntest ==1770)
	  cout  <<    10 * log(oSignalPower[NumZSteps-1]*1e3)/ log(10) 
	  				- 10 * log( oSignalPower[0]*1e3 )/ log(10)  << "  " ;

 
	}  
	
  *LogFile << endl;
  cout << endl;
}

cout << endl; 

*/

/*
// SaturationFactorAdjuctment = 1.05;
for(int Ntest=0;Ntest<41;Ntest++)
{
  *LogFile << (double)Ntest/10 << "  " ;
//  if(Ntest == 410 ||Ntest ==1770)
//  		cout  <<  endl << (double)Ntest/10 << "  " ;
//  for(SaturationFactorAdjuctment=1.4;SaturationFactorAdjuctment<1.6;SaturationFactorAdjuctment +=0.01)
// { 
  
//  oPumpPower[0] =( 15.00001 * 1e-3 - 0.015 ) * 0.08257 / ( 0.2 - 0.015 ) ;
  oPumpPower[0] = PumpCurrent2Power(15.00001 * 1e-3);
  NumSigCh = 1;
  oSignalWaveLth[0] = 1530e-9 + Ntest * 1e-9;
  
  oSignalPower[0] =  -40.75;
  oSignalPower[0] = exp(oSignalPower[0] / 10 * log(10)) * 1e-3;
   
  PropagatSignal();

//  cout << "Input = " << oSignalPower[0] << "=dbm" << 10 * log(oSignalPower[0]*1000)/ log(10) << endl
//   		<< "Output = " << oSignalPower[NumZSteps-1] << "=dBm" << 10 * log(oSignalPower[NumZSteps-1]*1000)/ log(10) << endl
//  		<< " Amplength=" << AmpLength  << endl
//		<< " Gain or Loss =" << 10 * log(oSignalPower[NumZSteps-1] /  oSignalPower[0]   )/ log(10) << endl;

//  *LogFile  <<  10 * log(oSignalPower[NumZSteps-1]*1000)/ log(10) << "   ";
//	if(Ntest == 410 ||Ntest ==1770)
	  cout  <<    10 * log(oSignalPower[NumZSteps-1]*1e3)/ log(10) +40.75 << endl;
  
//  *LogFile << endl;
 // cout << endl;
  }
*/



   cplx NormalizedNoiseMode = 0;
   

 
   //## If FLAT_GAIN_INSIDE_BANDWIDTH, it does just that (Ivan Lima 6/13/01)
   for ( int jj = 0; jj < qtPoints; jj++) {

      //DeltaFreq*FreqFactorFFT(jj,qtPoints);        
      if ( TypeAmplifierNoise != NOISE_OFF ) {  
		
         if (TypeAmplifierNoise == NOISE_ON_CONST_POWER_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
   	  		LinearGainOptAmplif = oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps];	     
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
	
         }
         else if (TypeAmplifierNoise == NOISE_ON_GAUSSIAN_WHITE ) {
	        NoiseAmplitudeFactor = sqrt(oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
   	  		LinearGainOptAmplif = oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps];
            NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
   		      +jc*RNG->GetGaussianDeviate() )/sqrt(2.);

         }else if(	TypeAmplifierNoise == NOISE_ON_NO_RNG ) {	

	        NoiseAmplitudeFactor = 0;
   	  		LinearGainOptAmplif = (oSignalPower[jj*NumZSteps + NumZSteps - 1] + oNoiseFPower[jj*NumZSteps + NumZSteps - 1])
							 / oSignalPower[jj*NumZSteps];
//			NormalizedNoiseMode =  sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
   
//	        NoiseAmplitudeFactor = sqrt(oNoiseFPower[jj*NumZSteps + NumZSteps - 1]);
//   	  		LinearGainOptAmplif = oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps];
//			NormalizedNoiseMode =  sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
   	 	}
		 
	 
         sfftM.fOptSignalX[jj] 
         = sqrt(LinearGainOptAmplif)*sfftM.fOptSignalX[jj]
         + NoiseAmplitudeFactor * NormalizedNoiseMode;

		
                                                   
      }else{                 
		sfftM.fOptSignalX[jj] *= sqrt(oSignalPower[jj*NumZSteps + NumZSteps - 1] 
  							/ oSignalPower[jj*NumZSteps]);
   	  }
 }
// cout << 10*log10(sq(abs(sfftM.fOptSignalX[NumSigCh/2]))*1000) << endl;

 for(int ii=0;ii<NumSigCh;ii++)  {
	 sfftM.fOptSignalX[ii]  *= sqrt(GetGainFlatenFilter(oSignalWaveLth[ii]));
 	 if(PolynomialFilterAppliedFlag==1){
	  	 sfftM.fOptSignalX[ii]  *= sqrt(GetPolynomialFilter(oSignalWaveLth[ii]));
 	}
 }
 

//  cout << "  G" << GetGainFlatenFilter(oSignalWaveLth[NumSigCh/2]) ;
//  cout << "  sqrt" << sqrt(GetGainFlatenFilter(oSignalWaveLth[NumSigCh/2]))<< "  " ;
//  cout << "  dB" << 10*log10(abs(GetGainFlatenFilter(oSignalWaveLth[NumSigCh/2]))) << "  " ;

// cout << 10*log10(sq(abs(sfftM.fOptSignalX[NumSigCh/2]))*1000) << endl;
// exit(1);

 fft->ifftshift(sfftM.fOptSignalX);

 fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
 

} // endl of AmplifyHomogeneous

// ########################################################
// out of date
void OptAmplifier::ApplyPolynomial()
{ // be careful if you want to use this

 fft->fftshift(sfftM.fOptSignalX);

 for(int ii=0;ii<NumSigCh;ii++)  {
 	 sfftM.fOptSignalX[ii]  *= sqrt(GetPolynomialFilter(Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() )        ));
 }

 fft->ifftshift(sfftM.fOptSignalX);

 fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);



}

// #############################################################
// here, I propagate the signal or signal and noise.
// first solve the rate equation according to (6.3) and than propagate signal according to (6.5) (6.6)
// noise part is solved using (6.7)
// by Jonathan Hu
void OptAmplifier::PropagatSignal()
{



  if(TypeAmplifierNoise == NOISE_OFF)
  {      
  	for(int i=1;i<NumZSteps;i++){
 		ForwardStep(i);
	/*	if(i==5*int(i/5)){
			cout <<   i*DeltaZ << "  signalgain " 
				<< UpperN << "  " << endl ;
			for(int S=0;S<NumSigCh;S++){
				cout << oSignalPower[S * NumZSteps + i ] << "   " 
					 << 10 * log10(oSignalPower[S * NumZSteps + i ] / oSignalPower[S * NumZSteps])
						<< endl;
			}			
		}
	*/
	}
//	cout << "noise OFF" << endl;
	
//  	for(int S=0;S<NumSigCh;S++){
	//			cout << 10 * log10(oSignalPower[S * NumZSteps + NumZSteps - 1 ] / oSignalPower[S * NumZSteps])
	//				<< "     " << oSignalPower[S * NumZSteps + NumZSteps - 1 ]  
	//				<< "     " << oSignalPower[S * NumZSteps]  << endl;
//	}
  }else{
    for(int j=0;j<3;j++){
      for(int i=1;i<NumZSteps;i++){
 		ForwardStep(i);
/*		if(i==5*int(i/5)){
			cout <<   i*DeltaZ << "  signalgain " 
				<< UpperN << "  " ; 
			for(int S=0;S<NumSigCh;S++){
				cout << oSignalPower[S * NumZSteps + i ] << "   " 
					 << 10 * log10(oSignalPower[S * NumZSteps + i ] / oSignalPower[S * NumZSteps])
						<< endl;
			}			
		}
*/   	  }
	  for(int i=1;i<NumZSteps;i++){
		BackwardStep(i);
			
		
   	  }

	   
    }	
	
		
	for(int N=0;N<NumNoiCh;N++){
//	  cout << "NoiseWaveLth=" << oNoiseWaveLth[N];
/*	  for(int i=0;i<NumZSteps;i++){
		if(i==5*int(i/5)){
			cout << endl <<  i*DeltaZ << "  " 
			    << oNoiseFPower[N * NumZSteps + i ] / (1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[N])/LightSpeed)) << "  " 
				<< oNoiseBPower[N * NumZSteps + i ] / (1.0e9*abs(-DeltaFreq*sq(oNoiseWaveLth[N])/LightSpeed)) ;
		}
	  }
*/	//  cout << endl;
	}
  } // end of 	if(TypeAmplifierNoise == NOISE_OFF)
  
  // set the Signal data
// ApplyGainFlatenFilter();




} // end of PropagatSignal

// ########################################################
// out of date
void OptAmplifier::ApplyGainFlatenFilter()  
{

}


// #########################################################
// propagate the signal forward
// by Jonathan Hu

void OptAmplifier::ForwardStep(int StepI)
{
 double g;
 double *PreSigPower = new double[NumSigCh];
 double *PrePumPower = new double[NumPumCh];
 double *PreNoiPower = new double[NumNoiCh];

 double *TempSigPower = new double[NumSigCh];
 double *TempPumPower = new double[NumPumCh];
 double *TempNoiPower = new double[NumNoiCh];
 
 	//double PreSigPower[NumSigCh], PrePumPower[NumPumCh];
//  double SignalStepGain, PumpStepGain;

  double monitor;	
  	for(int S=0;S<NumSigCh;S++){
		TempSigPower[S] = oSignalPower[S * NumZSteps + StepI-1];
	}
	for(int P=0;P<NumPumCh;P++){
		TempPumPower[P] = oPumpPower[P * NumZSteps + StepI-1];
	}
    if(!(TypeAmplifierNoise == NOISE_OFF)){
		for(int N=0;N<NumNoiCh;N++){
			TempNoiPower[N] = oNoiseFPower[N * NumZSteps + StepI - 1] + 
					oNoiseBPower[N * NumZSteps + NumZSteps - StepI - 1 ];
		}
   	}
	
	// calculate N2 using (6.3) in Becker's book
	CalculateUpperN(StepI,TempSigPower,TempPumPower,TempNoiPower);
  	// Guess part in improved Eular's method
	FirstGuessForward(StepI,PreSigPower,PrePumPower,PreNoiPower);
	
	for(int i=0;i<NumIntItr;i++){

 		for(int NSig=0;NSig<NumSigCh;NSig++){ 

      	switch(TypeIntergration)
     	{
       		case 1:   // Left Boundary
         		TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+StepI-1]; 
			//	cout << 
         	break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+StepI] + oSignalPower[NSig*NumZSteps+StepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+StepI] / oSignalPower[NSig*NumZSteps+StepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+StepI-1] *( g - 1 ) /  log(g) ;
         break;
		 


       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }

  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+StepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+StepI] + oPumpPower[NPum*NumZSteps+StepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+StepI] / oPumpPower[NPum*NumZSteps+StepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+StepI-1] *( g - 1 ) /   log(g) ;
         break;


	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
  }

  if(!(TypeAmplifierNoise == NOISE_OFF)){
	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempNoiPower[NNoi]=  oNoiseFPower[NNoi * NumZSteps + StepI-1] + oNoiseBPower[NNoi * NumZSteps + NumZSteps -
			StepI - 1 ]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempNoiPower[NNoi]=  (oNoiseFPower[NNoi*NumZSteps+StepI] + oNoiseFPower[NNoi*NumZSteps+StepI-1]
		 		+ oNoiseBPower[NNoi * NumZSteps + NumZSteps - StepI - 1 ] +oNoiseBPower[NNoi * NumZSteps + NumZSteps - StepI ] ) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         cerr << "With type Intergration must not be 3 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
		// need to be recoded
       //  g = oNoisePower[NNoi*NumZSteps+StepI] / oNoisePower[NNoi*NumZSteps+StepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
       //  TempNoiPower[NNoi] = oNoisePower[NNoi*NumZSteps+StepI-1] *( g - 1 ) /   log(g) ;
         break;


	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
    }
  }
  		CalculateUpperN(StepI,TempSigPower,TempPumPower,TempNoiPower);
//	    cout << StepI * DeltaZ << "  " << UpperN << endl; 	
	// correction part in improved Eular's part
		monitor = OptimizeForward(StepI, PreSigPower, PrePumPower,PreNoiPower);

		if(monitor>0.2){
//			BlowupSign=1;
			cout << "error" ;
			exit(1);
		}
	}
	
 delete [] PreSigPower;
 delete [] PrePumPower;
 delete [] PreNoiPower;

 delete [] TempSigPower;
 delete [] TempPumPower;
 delete [] TempNoiPower;
 


} // end of ForwardStep


// ######################################################
// since our experimental setup only has forward pump, so backward iteration is not needed to solve signal gain
// but it have to be added if one wants to calculate the noise.
// this process is according to (6.7) in becker's book
// by Jonathan Hu 
void OptAmplifier::BackwardStep(int StepI)
{
  double C1, C2;
  if(!(TypeAmplifierNoise == NOISE_OFF)){
	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
		C1 = UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi]) 
				- (1 - UpperNBackward[StepI]) * GetSignalAbs(oNoiseWaveLth[NNoi]) 
				-  GetAttenuationFiber(oNoiseWaveLth[NNoi]);
	    C2 = 2 * UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi]) 
				* hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) * DeltaFreq *
				GetTempDepFactor(oNoiseWaveLth[NNoi]);
		oNoiseBPower[NNoi*NumZSteps+StepI] = oNoiseBPower[NNoi*NumZSteps+StepI - 1] * exp(C1 * DeltaZ) 
				+ exp(C1 * DeltaZ) * (  exp(-1 * C1 * DeltaZ) * C2 * DeltaZ);
//		oNoiseBPower[NNoi*NumZSteps+StepI] = oNoiseBPower[NNoi*NumZSteps+StepI - 1] * exp(C1 * DeltaZ) 
//				+ exp( C1 * DeltaZ) * (  C2 * DeltaZ);
//		oNoiseBPower[NNoi*NumZSteps+StepI] = oNoiseBPower[NNoi*NumZSteps+StepI - 1]
//				+ ( C1 * oNoiseBPower[NNoi*NumZSteps+StepI - 1] + C2)* DeltaZ;




  	}  
  }else{      
     cerr << "BackwardStep is only valid in the NOISE_ON case"<< endl 
             << "Aborting now! " << endl;
         exit(1);
  } 

	


/*  if(!(TypeAmplifierNoise == NOISE_OFF)){

	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
		dpdz =  (UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi]) - (1-UpperNBackward[StepI]) * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* oNoiseBPower[NNoi*NumZSteps+StepI - 1]
				+ UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaWave * sq(Wavelength2Frequency(oNoiseWaveLth[NNoi])) / LightSpeed    // this is deltaFreq
				;
		oNoiseBPower[NNoi*NumZSteps+StepI] = oNoiseBPower[NNoi*NumZSteps+StepI - 1] + dpdz * DeltaZ;
  
 		dpdz =  (UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi]) - (1-UpperNBackward[StepI]) * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* (oNoiseBPower[NNoi*NumZSteps+StepI - 1] + oNoiseBPower[NNoi*NumZSteps+StepI ]) / 2
				+ UpperNBackward[StepI] * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaWave * sq(Wavelength2Frequency(oNoiseWaveLth[NNoi])) / LightSpeed    // this is deltaFreq
				;
		oNoiseBPower[NNoi*NumZSteps+StepI] = oNoiseBPower[NNoi*NumZSteps+StepI - 1] + dpdz * DeltaZ; 
  	}  
  }
*/
} // end of BackwardStep

// #####################################################
// Calculate N2 use exactly the same equation(6.3) in Becker's book 
void OptAmplifier::CalculateUpperN(int IntStepI, double * SigPower, double * PumPower, double * NoiPower)
{
  double Numerator=0,Denominator=0;
// calculate Numerater
  for(int NSig=0;NSig<NumSigCh;NSig++){
	Numerator +=  GetSignalAbs(oSignalWaveLth[NSig]) * SigPower[NSig] 
			/ Wavelength2Frequency(oSignalWaveLth[NSig]);
    
//    cout <<  GetSignalAbs(oSignalWaveLth[NSig]) / DopantConcentration << " * " << SigPower[NSig] <<
//			" / " <<  Wavelength2Frequency(oSignalWaveLth[NSig]);
  }
//  cout << "  Numerator=" << Numerator;
  for(int NPum=0;NPum<NumPumCh;NPum++){
	Numerator += GetPumpAbs(oPumpWaveLth[NPum]) * PumPower[NPum] 
			/ Wavelength2Frequency(oPumpWaveLth[NPum]);
//    cout << "  Numerator=" << Numerator;
  }
   if(!(TypeAmplifierNoise == NOISE_OFF)){
  	  	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
			Numerator += GetSignalAbs(oNoiseWaveLth[NNoi]) * NoiPower[NNoi] 
					/ Wavelength2Frequency(oNoiseWaveLth[NNoi]);
//    cout << "  Numerator=" << Numerator;
  		}
	}
//  cout << LifeTime << " * " <<  ConfinementFactor << " / " <<  EffectArea << " / " << hPlanck;
//  cout << "  Numerator=" << Numerator;
// calculate Denominator
  for(int NSig=0;NSig<NumSigCh;NSig++){
	Denominator +=  (GetSignalAbs(oSignalWaveLth[NSig]) + GetSignalEms(oSignalWaveLth[NSig]) ) * SigPower[NSig]
			 / Wavelength2Frequency(oSignalWaveLth[NSig]);
  }
  for(int NPum=0;NPum<NumPumCh;NPum++){
	Denominator += (GetPumpAbs(oPumpWaveLth[NPum]) + GetPumpEms(oPumpWaveLth[NPum]) ) * PumPower[NPum]
			 / Wavelength2Frequency(oPumpWaveLth[NPum]);
  }
  
  if(!(TypeAmplifierNoise == NOISE_OFF)){
	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
			Denominator += (GetSignalAbs(oNoiseWaveLth[NNoi]) + GetSignalEms(oNoiseWaveLth[NNoi]) ) * NoiPower[NNoi]
				 / Wavelength2Frequency(oNoiseWaveLth[NNoi]);
  	}
  }
  Denominator += SaturationFactorAdjuctment * DopantConcentration * EffectArea * hPlanck / LifeTime;
//	cout << "SaturationFactor = " << SaturationFactorAdjuctment * DopantConcentration * EffectArea / LifeTime << endl;
//  cout << "  Denominator=" << Denominator;

// cout << "Numerator / Denominator " << Numerator / Denominator << endl;
// exit(1);
  UpperN = Numerator / Denominator;
  LowerN = 1 - UpperN;
  UpperNBackward[NumZSteps-IntStepI] = UpperN;
//  cout << " UpperN=" << UpperN  << endl; //<< endl;


//  exit(1);
} // end of CalculateUpperN


// ######################################################
// as I used to solve Raman Amp. I use Improved Eular method to solve numerical integration
void OptAmplifier::FirstGuessForward(int IntStepI, double * PreSig, double * PrePum, double * PreNoi)
{
  double SignalStepGain, PumpStepGain;
  int test;
  for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
     
	SignalStepGain = UpperN * GetSignalEms(oSignalWaveLth[NSig]) - LowerN * GetSignalAbs(oSignalWaveLth[NSig]);
    oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
        exp((SignalStepGain - GetAttenuationFiber(oSignalWaveLth[NSig]))*DeltaZ);
	
	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    PumpStepGain = UpperN * GetPumpEms(oPumpWaveLth[NPum]) - LowerN * GetPumpAbs(oPumpWaveLth[NPum]);
       
    oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI-1] * 
			exp( (PumpStepGain - GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZ);
  	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI];
    if (oPumpPower[NPum*NumZSteps+IntStepI] < 0){
	    cout << " StepI = "  << IntStepI << ", NPum = " << NPum << endl;
		cout << oPumpPower[NPum*NumZSteps+IntStepI] << "=" << oPumpPower[NPum*NumZSteps+IntStepI-1]
			 << "*" << exp( (-1) * (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZ) << endl;  	
	    cerr << "PumpPower can't be negative" << endl 
             << "Aborting now! " << endl;
        exit(1);
		}
  }
  // here I only use trap rule
  
  if(!(TypeAmplifierNoise == NOISE_OFF)){
  	double dpdz, C1, C2;
	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
		C1 = UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi]) 
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]);
	    C2 = 2 * UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) * hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
					* DeltaFreq * GetTempDepFactor(oNoiseWaveLth[NNoi]);
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] * exp(C1 * DeltaZ) 
				+ exp(C1 * DeltaZ) * (  exp(-1 * C1 * DeltaZ) * C2 * DeltaZ);
// cout << "direct = " << oNoiseFPower[NNoi*NumZSteps+IntStepI] << endl;
//		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] * exp(C1 * DeltaZ) 
//				+ exp(C1 * DeltaZ) * ( C2 * DeltaZ );
//		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1]
//				+ ( C1 * oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + C2)* DeltaZ;
/*cout << "trap = " << oNoiseFPower[NNoi*NumZSteps+IntStepI] << endl;
cout << "C1 = " << C1 << "  C2 = " << "  DeltaZ = " << DeltaZ << endl; 

cout << "dir = " << oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] * exp(C1 * DeltaZ) << endl;

cout << "trap = " << oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + C1 * oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] * DeltaZ;
cin >> test;
if (test==0)
	exit(1);
*/
/*cout <<  " Brian=" <<	oNoiseFPower[NNoi*NumZSteps+IntStepI] ;
		dpdz =  (UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* oNoiseFPower[NNoi*NumZSteps+IntStepI - 1]
				+ UpperN * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaFreq   // this is deltaFreq
				;
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + dpdz * DeltaZ;
  
 		dpdz =  (UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* (oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + oNoiseFPower[NNoi*NumZSteps+IntStepI ]) / 2
				+ UpperN * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaFreq   // this is deltaFreq
				;
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + dpdz * DeltaZ; 
cout << "  easy=" <<	oNoiseFPower[NNoi*NumZSteps+IntStepI]  << endl;
*/
  	}  
  }
  
}// end of FirstGuessForward

// ##############################################
// guess-correct process for improved Eular method
// by Jonathan Hu

double OptAmplifier::OptimizeForward(int IntStepI, double * PreSig, double * PrePum, double * PreNoi)
{
  double g;
  double SignalStepGain,PumpStepGain;   
  double RMS;

  RMS = 0;

  for(int NSig=0;NSig<NumSigCh;NSig++){
  
	SignalStepGain = UpperN * GetSignalEms(oSignalWaveLth[NSig]) - LowerN * GetSignalAbs(oSignalWaveLth[NSig]);
    oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
        exp((SignalStepGain  - GetAttenuationFiber(oSignalWaveLth[NSig]))*DeltaZ);
//	cout << endl << "Factor=" << SignalStepGain * ConfinementFactor - GetAttenuationFiber(oSignalWaveLth[NSig]);
//	cout << endl << "Factor=" << SignalStepGain << " * " <<  ConfinementFactor << " - " 
//				<< GetAttenuationFiber(oSignalWaveLth[NSig]);
	RMS +=  abs(oSignalPower[NSig*NumZSteps+IntStepI] - PreSig[NSig]) / oSignalPower[NSig*NumZSteps+IntStepI]; 

	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
//	cout << " RMS=" << " " << RMS ;  
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){
    PumpStepGain = UpperN * GetPumpEms(oPumpWaveLth[NPum]) - LowerN * GetPumpAbs(oPumpWaveLth[NPum]);
       
    oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI-1] * 
			exp( (PumpStepGain  - GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZ);
    //cout << endl << " Factor=" << PumpStepGain  - GetAttenuationFiber(oPumpWaveLth[NPum]);
    if (oPumpPower[NPum*NumZSteps+IntStepI] < 0){
	    cout << IntStepI ;
		cout << oPumpPower[NPum*NumZSteps+IntStepI] << "=" << oPumpPower[NPum*NumZSteps+IntStepI-1]
			 << "*" << exp( (-1) * (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum])) * DeltaZ);  	
	    cerr << "PumpPower can't be negative" << endl 
             << " Aborting now! " << endl;
        exit(1);
		}

	RMS += abs(oPumpPower[NPum*NumZSteps+IntStepI] - PrePum[NPum]) / oPumpPower[NPum*NumZSteps+IntStepI];
	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI];

  }



  			
  RMS = (RMS/(NumSigCh + NumPumCh));

  double C1, C2;
  if(!(TypeAmplifierNoise == NOISE_OFF)){
	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
		C1 = UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi])  
					- GetAttenuationFiber(oNoiseWaveLth[NNoi]);
	    C2 = 2 * UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) * hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
					* DeltaFreq * GetTempDepFactor(oNoiseWaveLth[NNoi]);
//					cout << " NoiseFactor=" <<  GetTempDepFactor(oNoiseWaveLth[NNoi]) << endl;
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] * exp(C1 * DeltaZ) 
				+ exp(C1 * DeltaZ) * (  exp(-1 * C1 * DeltaZ) * C2 * DeltaZ);


//		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1]
//				+ ( C1 * oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + C2)* DeltaZ;


}
	

	/*	for(int NNoi=0;NNoi<NumNoiCh;NNoi++){
		dpdz =  (UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* oNoiseFPower[NNoi*NumZSteps+IntStepI - 1]
				+ UpperN * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaWave * sq(Wavelength2Frequency(oNoiseWaveLth[NNoi])) / LightSpeed    // this is deltaFreq
				;
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + dpdz * DeltaZ;
  
 		dpdz =  (UpperN * GetSignalEms(oNoiseWaveLth[NNoi]) - LowerN * GetSignalAbs(oNoiseWaveLth[NNoi])
				- GetAttenuationFiber(oNoiseWaveLth[NNoi]) ) 
				* (oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + oNoiseFPower[NNoi*NumZSteps+IntStepI ]) / 2
				+ UpperN * GetSignalEms(oNoiseWaveLth[NNoi])*hPlanck * Wavelength2Frequency(oNoiseWaveLth[NNoi]) 
				* DeltaWave * sq(Wavelength2Frequency(oNoiseWaveLth[NNoi])) / LightSpeed    // this is deltaFreq
				;
		oNoiseFPower[NNoi*NumZSteps+IntStepI] = oNoiseFPower[NNoi*NumZSteps+IntStepI - 1] + dpdz * DeltaZ; 
  	}
	*/
  }
  return(RMS);

} // end OptimizeForwardShootingIntegration

void OptAmplifier::Loss(double Loss)
{ // please see 

	for(int ii=0;ii < NumTSteps;ii++){
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			SignalInput[NSig*NumTSteps + ii] /= dB2Linear(Loss);
		}
	}
//	cout << "Loss = " <<  Loss << endl;
} /// lossLoss

// this is for transient model, for a sin modulated signal in time domain,
// what's the modulation after the EDFA?
void OptAmplifier::AmplitudeFluctuation()
{ // please see 


//	    *LogFile << "oSignalWaveLth=" <<  oSignalWaveLth[0] << endl;

/*	for(int StepT =0;StepT < NumTSteps;StepT ++){
		cout << PumpInput[0 * NumTSteps + StepT] << "  ";
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			cout << SignalInput[NSig*NumTSteps + StepT]<< "  ";
		}
	}		
*/
	// setup the droping signal
// 	cout << " run to here0" << endl;

	for(int ii=1;ii < NumTSteps;ii++){
//		*LogFile <<  ii ;
		CalculateAverageN2T(ii);
//		*LogFile <<  "1";
	}
//	cout << " run to here1" << endl;

	CalculateGain();

	AverageN2T[0] = AverageN2T[NumTSteps-1] ; 
//	cout << " run to here3" << endl;
/*	for(int ii=0;ii < NumTSteps;ii+=1){
		*LogFile << ii << "  " <<  10 * log10(exp(GainPum[0 *NumTSteps + ii]))	;		
//						- 10 * log10(exp(GainPum[0 *NumTSteps + 0]));
		for (int NSig = 0;NSig < NumSigCh; NSig ++){ 
			*LogFile << "  " <<  10 * log10(exp(GainSig[NSig *NumTSteps + ii])) ;
//						- 10 * log10(exp(GainSig[NSig *NumTSteps + 0])) ;
		}
		*LogFile << endl;
	}
*/   

}  // end of AmplitudeFluctuation

// ######################################

void OptAmplifier::ResetTransient(double F, double TT)
{	
	TotalTime = TT;
	DeltaT = 1 / F /DataPointInCye;
	NumTSteps = (int)(TotalTime / DeltaT);
	AverageN2T[0]=0.60239;
//	    *LogFile << "oSignalWaveLth=" <<  oSignalWaveLth[0] << endl;



} // ResetTransient

// ########################################

void OptAmplifier::SetEMSAdjustGA(double EMSAdjustGA2)
{
	EMSAdjustGA2 = EMSAdjustGA;
}

// ########################################

void OptAmplifier::SetABSAdjustGA(double ABSAdjustGA2)
{
	ABSAdjustGA2 = ABSAdjustGA;
}



// ######################################
// Transient model
// I use Yan Sun's paper JSTQE 1997 p.991
// the equation I used is (16)
// Jonathan Hu

void OptAmplifier::CalculateAverageN2T(int IntStepT)
{	
	double SumEq16;
	double dNdt;
	double * TempSigIn = new double[NumSigCh];
	double * TempPumIn = new double[NumPumCh];
	double TempAverageN2T;
//	double * AverageN2T = new double[NumTSteps];
	for(int NSig=0;NSig<NumSigCh;NSig++){
		TempSigIn[NSig] = (SignalInput[NSig*NumTSteps+IntStepT-1]
				+SignalInput[NSig*NumTSteps+IntStepT])/2;

	}
//	cout << " run to here2.11" << endl;	

/*	for(int ii=0;ii < NumTSteps;ii++){
		cout << ii << "   " << SignalInput[ii] << "            " ;
//		cout << " run to here1" << endl;	
 	}
	exit(1);
*/	for(int NPum=0;NPum<NumPumCh;NPum++){
		TempPumIn[NPum] = (PumpInput[NPum*NumTSteps+IntStepT-1]
				+PumpInput[NPum*NumTSteps+IntStepT])/2;
	}
// 	cout << " run to here2  " << IntStepT  << "  " << NumTSteps<< endl;	
	
//	TempAverageN2T = AverageN2T[0];
 //   *LogFile << "oSignalWaveLth=" <<  oSignalWaveLth[0] << endl;
// 	cout << " run to here2.3" << endl;	
	SumEq16 = 0;
//	cout << oSignalWaveLth[0] << endl;
// 	cout << " run to here2.5" << endl;	
//	cout << 0 << "  " <<GetSignalAbs(oSignalWaveLth[0])<< endl;
//	cout << GetSignalEms(oSignalWaveLth[0]) << "  " << GetSignalAbs(oSignalWaveLth[0]);
	for(int NSig=0;NSig<NumSigCh;NSig++){
			SumEq16 += TempSigIn[NSig]/1.6e-19*(exp(((ABSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig]) 
				+ EMSAdjustGA*GetSignalEms(oSignalWaveLth[NSig]))*TempAverageN2T-
				ABSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig]))*AmpLength)-1);
//		cout << SumEq16 << "   ";
	}
// 	cout << " run to here2.6" << endl;	
	for(int NPum=0;NPum<NumPumCh;NPum++){
			SumEq16 += TempPumIn[NPum]/1.6e-19*(exp(((GetPumpAbs(oPumpWaveLth[NPum]) 
				+ GetPumpEms(oPumpWaveLth[NPum]))*TempAverageN2T-
				GetPumpAbs(oPumpWaveLth[NPum]))*AmpLength)-1);
//		cout << SumEq16 << "   ";
	}
//	cout << TempAverageN2T << "   " << SumEq16/SaturationFactor/AmpLength << endl;
// 	cout << " run to here3" << endl;	
	dNdt = (TempAverageN2T +SumEq16/SaturationFactor/AmpLength)/LifeTime;
//	cout << "LifeTime =" << LifeTime << endl;
//	cout << "dNdt=" << dNdt << endl;
	AverageN2T[IntStepT] = AverageN2T[IntStepT-1] - dNdt * DeltaT;

	TempAverageN2T = (AverageN2T[IntStepT]+ AverageN2T[IntStepT-1])/2 ;
	SumEq16 = 0;
	for(int NSig=0;NSig<NumSigCh;NSig++){
			SumEq16 += TempSigIn[NSig]/1.6e-19*(exp(((ABSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig]) 
				+ EMSAdjustGA*GetSignalEms(oSignalWaveLth[NSig]))*TempAverageN2T-
				EMSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig]))*AmpLength)-1);
	}
	for(int NPum=0;NPum<NumPumCh;NPum++){
			SumEq16 += TempPumIn[NPum]/1.6e-19*(exp(((GetPumpAbs(oPumpWaveLth[NPum]) 
				+ GetPumpEms(oPumpWaveLth[NPum]))*TempAverageN2T-
				GetPumpAbs(oPumpWaveLth[NPum]))*AmpLength)-1);

	}

//  cout << " run to here4" << endl;	
	dNdt = (TempAverageN2T +SumEq16/SaturationFactor/AmpLength)/LifeTime;
//	cout << "LifeTime = " << LifeTime << endl;
	AverageN2T[IntStepT] = AverageN2T[IntStepT-1] - dNdt * DeltaT;
//	*LogFile <<IntStepT*DeltaT << "  " <<  AverageN2T[IntStepT] 
//				<< "  " << SignalInput[IntStepT]<< "   " << dNdt << "  " << DeltaT << endl;
		 
	delete [] TempSigIn ;
	delete [] TempPumIn ;
}  // end of CalculateAverageN2T

// #########################################
// standard equation for gain
// Jonathan Hu
void OptAmplifier::CalculateGain()
{

	for(int ii=0;ii < NumTSteps;ii++){
		for(int NSig=0;NSig<NumSigCh;NSig++){
			SignalInput[ii] *=	exp(((ABSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig])+ABSAdjustGA*GetSignalEms(oSignalWaveLth[NSig]))
					* AverageN2T[ii]-ABSAdjustGA*GetSignalAbs(oSignalWaveLth[NSig]))*AmpLength);
		}
	}


}  // end of CalculateGain

// ###########################################
// this is to reproduce Yan Sun's  another paper EL 1996 p. 1490
void OptAmplifier::ChannelAddDrop()
{ // please see Electronics Letters P.1490 1996 transient

	DeltaT = 1e-6;
	AmpLength = 23;
	NumSigCh = 8;
	NumPumCh = 1;
	double Freq = 193.5e12 ;
	for(int ii = 0; ii < NumPumCh; ii++){
		IntrinsicSaturationPowerPump[ii] = 0.44e-3/(hPlanck*306.12e12)*1.6e-19; 
	}

	IntrinsicSaturationPowerSignal[0] = 0.197e-3/(hPlanck*Freq)*1.6e-19; 
	for(int NSig = 1; NSig < NumSigCh; NSig++){
		IntrinsicSaturationPowerSignal[NSig] = 0.214e-3/(hPlanck*Freq)*1.6e-19; 
	}
//	cout << " hPlanck * Freq = " << hPlanck * Freq << endl;

	GainPum[0*NumTSteps + 0] = log(dB2Linear(-9.04908 )); //* 2.8418;
		GainSig[0*NumTSteps + 0] = log(dB2Linear(8.98346));
	for(int NSig = 1; NSig < NumSigCh; NSig++){
		GainSig[NSig*NumTSteps + 0] = log(dB2Linear(9.117));
	}	

	for(int StepT =0;StepT < NumTSteps;StepT ++){
		PumpInput[0 * NumTSteps + StepT] = dBm2Linear(16.3);
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			SignalInput[NSig*NumTSteps + StepT] = dBm2Linear(-2);
		}
	}	
	
/*	for(int StepT =0;StepT < NumTSteps;StepT ++){
		cout << PumpInput[0 * NumTSteps + StepT] << "  ";
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			cout << SignalInput[NSig*NumTSteps + StepT]<< "  ";
		}
	}		
*/
	// setup the droping signal
  
	int NumDropSignal = 1;
	for(int NSig = NumSigCh - 1 ; NSig > NumSigCh-1-NumDropSignal; NSig--){
		for(int StepT =20;StepT < NumTSteps;StepT ++){
			SignalInput[NSig*NumTSteps + StepT] = 0;
		}
	}	
	for(int StepT =25;StepT < NumTSteps;StepT ++){
		PumpInput[0*NumTSteps + StepT] *= 0.88;
	}
	
    
	AbsorptionLossPump[0*NumTSteps + 0] = 0.257 * AmpLength;
	AbsorptionLossSignal[0] = 0.145 * AmpLength;
	for(int ii = 1; ii < NumSigCh; ii++){
		AbsorptionLossSignal[ii] = 0.125 * AmpLength;
	}
//	cout << " run to here1" << endl;

	for(int ii = 0; ii < NumSigCh; ii++){
    cout << GainPum[0*NumTSteps +0] * IntrinsicSaturationPowerPump[0] 
   		- GainSig[ii*NumTSteps +0] * IntrinsicSaturationPowerSignal[ii] 
		+ (IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
		- IntrinsicSaturationPowerSignal[ii] * AbsorptionLossSignal[ii]) << endl;
	}  
 
 
 
 
 
 /*	cout << GainPum[0*NumTSteps +0] * IntrinsicSaturationPowerPump[0] << "  "
	   		<< - GainSig[2*NumTSteps +0]  * IntrinsicSaturationPowerSignal[2] << "  "
			<< + IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] << "  "
			<< - IntrinsicSaturationPowerSignal[2] * AbsorptionLossSignal[2] << endl;
	*/


/*	cout << GainPum[0*NumTSteps + 0] << "   " <<    AbsorptionLossPump[0*NumTSteps + 0] ;
	AbsorptionLossSignal[0] = 0.145 * AmpLength;
	for(int ii = 1; ii < NumSigCh; ii++){
		AbsorptionLossSignal[ii] = 0.125 * AmpLength;
	}
*/

//NumSigCh = 1;
	for(int ii=1;ii < NumTSteps;ii++){
//		ForwardTPumpSignalStep(ii);
	}

	for(int ii=1;ii < NumTSteps;ii++){
		ForwardTPumpStep(ii);
	}
	*LogFile << " Start GainPum " << endl;  
		
	for(int ii=0;ii < NumTSteps;ii+=NumTSteps/100){
//		*LogFile << ii << "  " << GainPum[0 *NumTSteps + ii]  
//				<< "  " <<  10 * log10(exp(GainPum[0 *NumTSteps + ii]))<< endl;
	}
//	cout << " run to here2" << endl;
	for(int ii=1;ii < NumTSteps;ii++){
		ForwardTSignalStep(ii);
	}
//	cout << " run to here3" << endl;
	for(int ii=0;ii < NumTSteps;ii+=1){
		*LogFile << ii << "  " <<  10 * log10(exp(GainPum[0 *NumTSteps + ii]))	;		
//						- 10 * log10(exp(GainPum[0 *NumTSteps + 0]));
		for (int NSig = 0;NSig < NumSigCh; NSig ++){ 
			*LogFile << "  " <<  10 * log10(exp(GainSig[NSig *NumTSteps + ii])) ;
//						- 10 * log10(exp(GainSig[NSig *NumTSteps + 0])) ;
		}
		*LogFile << endl;
	}
    


}  // end of ChannelAddDrop

// ##############################
// solve the time domain transient effect, refer Sun's JSTQE paper
// by Jonathan Hu
void OptAmplifier::ForwardTSignalStep(int StepT)
{ //  cout << " NumSigCh = " << NumSigCh ;
	for(int NSig = 0; NSig < NumSigCh; NSig++){
		GainSig[NSig *NumTSteps + StepT] = (IntrinsicSaturationPowerPump[0]*GainPum[0*NumTSteps + StepT]
				+ (IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
					- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))
				/IntrinsicSaturationPowerSignal[NSig];
	//	cout << GainSig[NSig *NumTSteps + StepT] << "  " ;
	  
	}
/*	
	int NSig = 0;
	GainSig[NSig *NumTSteps + StepT] = (IntrinsicSaturationPowerPump[0]*GainPum[0*NumTSteps + StepT]
				+ (IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
					- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))
				/IntrinsicSaturationPowerSignal[NSig];
	if(PumpInput = ){
		GainSig[NSig *NumTSteps + StepT] - 
	
	}
	
*/	
//	*LogFile << "  Gain = " << GainSig[0*NumTSteps+StepT] << endl;	

	double	GainSigEffective;
	GainSigEffective = GainSig[0*NumTSteps+StepT-1];
	double	SumEq8;
	double  dGdt;
	
	SumEq8 = PumpInput[0*NumTSteps+StepT-1]*(exp((GainSigEffective*IntrinsicSaturationPowerSignal[0]+
					(IntrinsicSaturationPowerSignal[0] * AbsorptionLossSignal[0]
						- IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] ))/
					IntrinsicSaturationPowerPump[0])-1);

// cout << PumpInput[0*NumTSteps+StepT-1] << "  " << (exp(GainPumEffective)-1) << endl;
//	*LogFile << StepT << "   " << SumEq8 ;
	SumEq8 += SignalInput[0*NumTSteps+StepT-1]*(exp(GainSigEffective)-1);	
	for(int NSig=1;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainSigEffective*IntrinsicSaturationPowerSignal[0]+
					(IntrinsicSaturationPowerSignal[0] * AbsorptionLossSignal[0]
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig] ))/
					IntrinsicSaturationPowerSignal[NSig])-1);
//	*LogFile << "   " << SumEq8 ;
	}
//	*LogFile << endl;
//	cout << SumEq8 << endl;
	
	dGdt = -1.0 *  (GainSigEffective + AbsorptionLossSignal[0]) - SumEq8 / IntrinsicSaturationPowerSignal[0];
	dGdt /= LifeTime;
//	*LogFile  << "   Gain=" <<  GainSigEffective << " Abs=" <<  AbsorptionLossSignal[0]
//			<< "    2nd term = "  << - SumEq8 / IntrinsicSaturationPowerSignal[0] ;
//	cout << "dGdt = " << dGdt ;

//	exit(1);
	GainSig[0*NumTSteps+StepT] = GainSig[0*NumTSteps+StepT - 1] + dGdt * DeltaT;
//	*LogFile << "  dGdt = " << dGdt << " DeltaT=" << DeltaT << "  Gain = " << GainSig[0*NumTSteps+StepT] << endl;	
	GainSigEffective = (GainSig[0*NumTSteps+StepT] + GainSig[0*NumTSteps+StepT -1]) / 2 ;

	SumEq8 = PumpInput[0*NumTSteps+StepT-1]*(exp((GainSigEffective*IntrinsicSaturationPowerSignal[0]+
					(IntrinsicSaturationPowerSignal[0] * AbsorptionLossSignal[0]
						- IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] ))/
					IntrinsicSaturationPowerPump[0])-1);
	SumEq8 += SignalInput[0*NumTSteps+StepT-1]*(exp(GainSigEffective)-1);

// cout << PumpInput[0*NumTSteps+StepT-1] << "  " << (exp(GainPumEffective)-1) << endl;
//	*LogFile << StepT << SumEq8 ;
	
	for(int NSig=1;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainSigEffective*IntrinsicSaturationPowerSignal[0]+
					(IntrinsicSaturationPowerSignal[0] * AbsorptionLossSignal[0]
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig] ))/
					IntrinsicSaturationPowerSignal[NSig])-1);
//	*LogFile << "   " << SumEq8 ;
	}
//	*LogFile << endl;
//	cout << SumEq8 << endl;
	
	dGdt = -1.0 *  (GainSigEffective + AbsorptionLossSignal[0]) - SumEq8 / IntrinsicSaturationPowerSignal[0];
	dGdt /= LifeTime;
//	*LogFile  << "   Gain=" <<  GainPumEffective << " Abs=" <<  AbsorptionLossPump[0]
//			<< "    2nd term = "  << - SumEq8 / IntrinsicSaturationPowerPump[0] ;
//	cout << "dGdt = " << dGdt ;

//	exit(1);
	GainSig[0*NumTSteps+StepT] = GainSig[0*NumTSteps+StepT - 1] + dGdt * DeltaT;
	
//	exit(1);

}	// end of ForwardTSignalStep

// for pump step
void OptAmplifier::ForwardTPumpStep(int StepT)
{
	double	GainPumEffective;
	GainPumEffective = GainPum[0*NumTSteps+StepT-1];
	double	SumEq8;
	double  dGdt;
	SumEq8 = PumpInput[0*NumTSteps+StepT-1]*(exp(GainPumEffective)-1);
// cout << PumpInput[0*NumTSteps+StepT-1] << "  " << (exp(GainPumEffective)-1) << endl;
	
//	*LogFile << StepT << SumEq8 ;
	for(int NSig=0;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainPumEffective*IntrinsicSaturationPowerPump[0]+
					(IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))/
					IntrinsicSaturationPowerSignal[NSig])-1);
//	*LogFile << "   " << SumEq8 ;
	}
//	*LogFile << endl;
//	cout << SumEq8 << endl;
	
	dGdt = -1.0 *  (GainPumEffective + AbsorptionLossPump[0]) - SumEq8 / IntrinsicSaturationPowerPump[0];
	dGdt /= LifeTime;
//	*LogFile  << "   Gain=" <<  GainPumEffective << " Abs=" <<  AbsorptionLossPump[0]
//			<< "    2nd term = "  << - SumEq8 / IntrinsicSaturationPowerPump[0] ;
//	cout << "dGdt = " << dGdt ;

//	exit(1);
	GainPum[0*NumTSteps+StepT] = GainPum[0*NumTSteps+StepT - 1] + dGdt * DeltaT;
//	*LogFile << "  dGdt = " << dGdt << " DeltaT=" << DeltaT << "  Gain = " << GainPum[0*NumTSteps+StepT] << endl;	
	GainPumEffective = (GainPum[0*NumTSteps+StepT] + GainPum[0*NumTSteps+StepT -1]) / 2 ;

	SumEq8 = PumpInput[0*NumTSteps+StepT-1]*(exp(GainPumEffective)-1);
	for(int NSig=0;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainPumEffective*IntrinsicSaturationPowerPump[0]+
					(IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))/
					IntrinsicSaturationPowerSignal[NSig])-1);
	}
	
	dGdt = -1.0 * IntrinsicSaturationPowerPump[0] * (GainPumEffective + AbsorptionLossPump[0]) - SumEq8;
	dGdt /= (LifeTime*IntrinsicSaturationPowerPump[0]);
	
	GainPum[0*NumTSteps+StepT] = GainPum[0*NumTSteps+StepT - 1] + dGdt * DeltaT;

		
} // end of ForwardTStep

// for Pump and signal propagate togethor
void OptAmplifier::ForwardTPumpSignalStep(int StepT)
{
	double	GainPumEffective;
	GainPumEffective = GainPum[0*NumTSteps+StepT-1];
	double	SumEq8;
	double  dGdt;
	SumEq8 = PumpInput[0*NumTSteps+StepT-1]*(exp(GainPumEffective)-1);
//	cout << PumpInput[0*NumTSteps+StepT-1] << "  " << (exp(GainPumEffective)-1) << endl;
	
//	cout << SumEq8 << endl;
	for(int NSig=0;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainPumEffective*IntrinsicSaturationPowerPump[0]+
					(IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))/
					IntrinsicSaturationPowerSignal[NSig])-1);
	}
//	cout << SumEq8 << endl;
	
	dGdt = -1.0 *  (GainPumEffective + AbsorptionLossPump[0]) - SumEq8 / IntrinsicSaturationPowerPump[0];
	dGdt /= LifeTime;
	*LogFile  << StepT << "  Gain+Ab = " << -1.0 *  (GainPumEffective + AbsorptionLossPump[0]) 
			<< "    2nd term = "  << - SumEq8 / IntrinsicSaturationPowerPump[0] ;
	*LogFile << "   dGdt = " << dGdt ;

//	exit(1);
	GainPum[0*NumTSteps+StepT] = GainPum[0*NumTSteps+StepT - 1] + dGdt * DeltaT;
	*LogFile << "  Gain = " << GainPum[0*NumTSteps+StepT] << endl;	
/*	GainPumEffective = (GainPum[0*NumTSteps+StepT] + GainPum[0*NumTSteps+StepT -1]) / 2 ;
	
	for(int NSig=0;NSig<NumSigCh;NSig++){
		SumEq8 += SignalInput[NSig*NumTSteps+StepT-1]*(exp((GainPumEffective*IntrinsicSaturationPowerPump[0]+
					(IntrinsicSaturationPowerPump[0] * AbsorptionLossPump[0] 
						- IntrinsicSaturationPowerSignal[NSig] * AbsorptionLossSignal[NSig]))/
					IntrinsicSaturationPowerSignal[NSig])-1);
	}
	
	dGdt = -1.0 * IntrinsicSaturationPowerPump[0] * (GainPumEffective + AbsorptionLossPump[0]) - SumEq8;
	dGdt /= (LifeTime*IntrinsicSaturationPowerPump[0]);
	
	GainPum[0*NumTSteps+StepT] = GainPum[0*NumTSteps+StepT - 1] + dGdt * DeltaT;
*/



}  // end of ForwardTPumpSignalStep

double OptAmplifier::GetAttenuationFiber(double Wavelength)
{
 //  if (Wavelength < 1000e-9){
  // 		return(0.005);
//	}else{
//		return(0.0022);
//	}
	return(AttenuationFiber/4.343);
} // end GetAttenuationFiber

// ###############################################

double OptAmplifier::PumpCurrent2Power(double Current)
{    
      if(Current < 14e-3){
         cerr << "Current is too small, = " << Current << ", Pump power much large than 0"<< endl 
             << "Aborting now! " << endl;
         exit(1);
      } 	
	 return( ( Current-0.014 ) * 0.083 / ( 0.2 - 0.014 ) 
	  					* exp( ( -1.0 ) / 10.0 * log(10)));
}

// #################################################
 
double OptAmplifier::GetPumpPowerSystem0dB()
{
	return(PumpPowerSystem0dB);
}

// #################################################
 
double OptAmplifier::GetPumpPowerTransparency()
{
	return(PumpPowerTransparency);
} 


// ####################################################
// used to calculate noise
// by Jonathan Hu
double OptAmplifier::GetTempDepFactor(double Wavelength)
{
 return(1);

 double KBoltzmann = 8.617e-5 * 1.6e-19;   // eV/K
 double Temperature = 273+23;
 double Frequency;
 double Nphon;

 Frequency = Wavelength2Frequency(Wavelength);
 cout << hPlanck*Frequency << endl;
 cout << KBoltzmann*Temperature << endl;
 cout << exp(hPlanck*Frequency/(KBoltzmann*Temperature));
 exit(1);
 
 
 return( 1/(exp(hPlanck*Frequency/(KBoltzmann*Temperature))-1));
 

} // end of GetTempDepFactor

// ############################################
void OptAmplifier::SetPumpPower(double Power)
{
      if(NumPumCh != 1){
         cerr << "NumPumCh = " << NumPumCh << endl;		 
		 cerr << "Currently, this function works only for in the case of NumPumCh is equal to One"<< endl 
             << "Aborting now! " << endl;
         exit(1);
      } 
	  if(Power < 0){
         cerr << "Pump power much large than 0"<< endl 
             << "Aborting now! " << endl;
         exit(1);
      } 
	  oPumpPower[0*NumZSteps] = Power * PumpPowerFactor;
	  *LogFile << "SetPumpPower=" << oPumpPower[0*NumZSteps] << endl;
}




//#############################################################

void OptAmplifier::AmplifyOptSignalScalarSinTim()
{


   if (! ObjectInitialized) 
   {
      cerr << "Error: Object of class OptAmplifier not initialized" 
           << endl;
      exit(1);
   }   
   double omega = 2*pi*10e9; // define omega
   for ( int jj = 0; jj < qtPoints; jj++) 
   {
     sfftM.tOptSignalX[jj] *= 1 + 0.1 * sin(omega/oOptSignal->GetFreqWindow()*jj);			    
   	 // at any time component, multiply a small sin modulated fraction.
   }   

   fft->FFT(sfftM.tOptSignalX,sfftM.fOptSignalX);
	// fft to update freq component   

} // ## end



void OptAmplifier::AmplifyOptSignalScalarSinFreq()
// this is an example if you want to double Freq compnent 
{


   if (! ObjectInitialized) 
   {
      cerr << "Error: Object of class OptAmplifier not initialized" 
           << endl;
      exit(1);
   }   
   
   double Freq;

   for ( int jj = 0; jj < qtPoints; jj++) 
    {
     Freq = CenterFreq + fft->freq(jj); 
     sfftM.fOptSignalX[jj] *= 2;			    
   	 // get double at any freq component,
   }   

   fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
   // inverse FFT to update time component

} // ## end


// ########################################################


void OptAmplifier::SetBiasedNoiseVector(double* zd_try, int dimension,
                                        int AddBiasedNoiseFlag)
{

// added by Walter and modified by John Zweck

// This method copies into the private array of cplx called BiasedNoise
// the noise produced by MMC class and that has to be added to the signal
// amplified by this amplifier 

// Setting AddBiasedNoiseFlag=1 enables this function to work
// normally, while AddBiasedNoiseFlag=0 should be used for a 
// noise-free simulation 

if(TypeAmplifierNoise != NOISE_ON_BIASED)
{
 cerr << "Must have TypeAmplifierNoise == BIASED_NOISE" << endl;
 exit(1);
}

cplx* running_pointer=BiasedNoise;
register const double *const zd_try_end = zd_try + dimension;


if (AddBiasedNoiseFlag)
   {
   if (dimension%2)
      {
      cerr << "OptAmplifier::SetBiasedNoiseVector \n" 
      << "Size of the input vector z_try is not even. \n";
      exit(1);
      }
   int dimension_complex_vector=dimension/2;   
   if (qtPoints%2)
      {
      cerr << "OptAmplifier::SetBiasedNoiseVector \n" 
      << "Size of the signal vector is not even. \n";
      exit(1);
      }
   
   if (qtPoints<dimension_complex_vector)
      {
      cerr << "OptAmplifier::SetBiasedNoiseVector \n" 
      << "Size of the noise input vector bigger than the size of signal vector.\n";
      exit(1);
      }

   //need to consider that dimension/2 could be smaller than qtPoints
   //in this case I set the higher frequency components of the noise to zero
   //BiasedNoise has the same structure of sfftM.fOptSignalX which means
   //that the Fourier Transform is unshifted
   //that is that possible zero frequency componets should be placed in the 
   //center of the BaisedNoise vector
 
   int num_compl_freq_to_be_set_to_zero= (qtPoints-dimension_complex_vector);

   register const double *const zd_try_middle =zd_try +dimension/2;

   for(; zd_try < zd_try_middle; zd_try=zd_try+2, running_pointer++)
      *running_pointer= cplx((*zd_try),(*(zd_try+1)));
      
   for (int i=1;i<=num_compl_freq_to_be_set_to_zero;i++,running_pointer++)
      *running_pointer=cplx(0.0,0.0);
         
   for(; zd_try < zd_try_end; zd_try=zd_try+2, running_pointer++)
      *running_pointer= cplx((*zd_try),(*(zd_try+1)));

   //debugging
   if ((running_pointer)!=(BiasedNoise+qtPoints))
      {
      cerr << "BiasedNoise vector has incorrect size. \n";
      exit(1);
      } 
   }
 else  
   {
   for(; zd_try < zd_try_end; zd_try=zd_try+2, running_pointer++)
      *running_pointer=cplx(0.0,0.0);
   
   }   //end if (!BiasedNoiseOff)  
/*
  //testing 
  for (int i=0;i<qtPoints;i++)
     cout << imag(BiasedNoise[i]) << "\t";
  cout<< endl;   
*/ 
// testing
/*
running_pointer=BiasedNoise;
for (int kk=0; kk<qtPoints;kk++,running_pointer=running_pointer+1)
   {
   cout << real(*running_pointer) << "\t" << imag(*running_pointer) << "\n";
   }
*/   
}  //end SetBiasedNoiseVector


// ################################################################

void OptAmplifier::WriteFileBiasedNoiseTime(string PathAndFileName)
{
ofstream NoiseOutput(PathAndFileName.c_str(),ios::out);
if (!NoiseOutput)
   {
	cerr << "File " << PathAndFileName << " could not be created! "
	     << "Aborting!" << endl;
	exit(-1);
	}
		   
NoiseOutput << "%(1) time (s), (2) real part, (3) imag part" << endl;

double TimeWindow = oOptSignal->GetTimeWindow();
double DeltaT= TimeWindow/qtPoints;

cplx* BiasedNoiseInTime= NewCplx(qtPoints);
fft->IFFT(BiasedNoise,BiasedNoiseInTime);

for (int jj=0;jj<qtPoints;jj++)
   {
   NoiseOutput << jj*DeltaT << "\t" 
	            << real(BiasedNoiseInTime[jj]) << "\t"
	            << imag(BiasedNoiseInTime[jj])	<< endl;
	}
FreeCplx(BiasedNoiseInTime);
NoiseOutput.close();
}

// ################################################################

void OptAmplifier::WriteFileBiasedNoiseFreq(string PathAndFileName)
{
ofstream NoiseOutput(PathAndFileName.c_str(),ios::out);
if (!NoiseOutput)
   {
	cerr << "File " << PathAndFileName << " could not be created! "
	     << "Aborting!" << endl;
	exit(-1);
	}
		   
NoiseOutput << "%(1) freq (Hz), (2) real part, (3) imag part" << endl;

double FreqWindow=qtPoints*DeltaFreq;
fft->fftshift(BiasedNoise);
      
for(int jj=0;jj<qtPoints;jj++)
  {
  NoiseOutput << jj*DeltaFreq - FreqWindow/2.0 << "\t"
              << real(BiasedNoise[jj]) << "\t"
	           << imag(BiasedNoise[jj]) << endl;
  } // end for loop

                 
fft->ifftshift(BiasedNoise);
  
NoiseOutput.close();

}

// #########################################################

//! shifts the BiasedNoise in time by NumOfBitSlots 

void OptAmplifier::TimeShiftBiasedNoise(int NumOfBitSlots)
{
// the BiasedNoise vector is unshifted that means that  frequency 0 is in the
// first position of vector,in the middle we have (FreqWindow/2-deltaF) and
// (-FreqWindow/2) and in the last position of the vector there is frequency
// -deltaF
double TimeWindow = oOptSignal->GetTimeWindow();
double DeltaT= TimeWindow/qtPoints;
double DeltaF=1/TimeWindow;
double FreqWindow=qtPoints*DeltaF;
int NumOfBits= oOptSignal->GetStringLength();
double freq=0;
cplx* running_pointer=BiasedNoise;

// prepare for checking
cplx* BiasedNoiseInTime= NewCplx(qtPoints);

//checking
fft->IFFT(BiasedNoise,BiasedNoiseInTime);
cplx* running_for_time=BiasedNoiseInTime;

ofstream* NoiseOutput =new ofstream("./out/BiasedNoiseInTime.dat",ios::out);
*NoiseOutput << "(1) time, (2) real part, (3) imag part" << endl;
*NoiseOutput << "% Noise before shift of t= " 
             << NumOfBitSlots*TimeWindow/NumOfBits 
				 << " (" << NumOfBitSlots << ") bit slots" << endl;

for (int i=1;i<=qtPoints;i++,running_for_time++)
   {
   *NoiseOutput << (i-1)*DeltaT << "\t" << real(*running_for_time) << "\t"
	             << imag(*running_for_time)	<< endl;
	}
// end checking

for (int i=1;i<=qtPoints;i++,running_pointer++)
   {
   (i<=floor(double(qtPoints/2.0)))
     ?(freq=(i-1)*DeltaF):(freq=-FreqWindow+(i-1)*DeltaF);
   *running_pointer*=(cplx) exp(jc*2.0*pi*freq*(
	      NumOfBitSlots*TimeWindow/NumOfBits)); 
   }
	
// checking again
fft->IFFT(BiasedNoise,BiasedNoiseInTime);
running_for_time=BiasedNoiseInTime;

*NoiseOutput << endl;
*NoiseOutput << "% Noise after shift of t= " 
             << NumOfBitSlots*TimeWindow/NumOfBits 
				 << " (" << NumOfBitSlots << ") bit slots" << endl;

for (int i=1;i<=qtPoints;i++,running_for_time++)
   {
   *NoiseOutput << (i-1)*DeltaT << "\t" << real(*running_for_time) << "\t"
	             << imag(*running_for_time)	<< endl;
	}
*NoiseOutput << endl;
FreeCplx(BiasedNoiseInTime);
NoiseOutput->close();
//end checcking	
	
} //end TimeShiftBiasedNoise











