// $Id:

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2000):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

#include "ocsOptSignal.hh"

extern ofstream *LogFile;


// ################### Methods for the class OptSignal
// # This class contains the fft structure, sfftM, that carries out
// # the optical signal in the optical domain throughout the
// # Simulations.
// ####################################################################

OptSignal::OptSignal(string InFileName, RanNumGen *RNG2)
{
  ConstructOptSignal(InFileName,RNG2);

  RegenerateSignalString();
   
    if(DebugLevel)
         {
     
          //## Ivan Lima (4/28/02)
          //#  I should check this information for a WDM system
         *LogFile << endl << "Mean Power of the string per channel [W] = "
                       << GetAveragePower()/NumChannels
                       << endl
                       << "Mean Power [dBm] = "
                       << Linear2dBm(GetAveragePower()/NumChannels )
                       << endl
                       << "In order to have Mean Power [W] = "
                      << OptPowerChann << endl
                       << "which is Mean Power [dBm] = "
                       << Linear2dBm(OptPowerChann)
                       << endl
                       << "set OptPowerChann [W] = "
                       <<  sq(OptPowerChann)/(GetAveragePower()/NumChannels)
                       << endl << endl;
     
          } // end if DebugLevel


  
} // ## end OptSignal::OptSignal

// ###########################################################################

/*! Use this constructor if you want to quickly initialize signal in time and
   frequency using data in file named DataFileName
*/

OptSignal::OptSignal(string InFileName, RanNumGen *RNG2, string DataFileName)
{
  ConstructOptSignal(InFileName,RNG2);

  ReadFileComplexFieldsTimeAndFreq(DataFileName);

  if(DebugLevel)
         {

          //## Ivan Lima (4/28/02)
          //#  I should check this information for a WDM system
         *LogFile << endl << "Mean Power of the string per channel [W] = "
                       << GetAveragePower()/NumChannels
                       << endl
                       << "Mean Power [dBm] = "
                       << Linear2dBm(GetAveragePower()/NumChannels )
                       << endl
                       << "In order to have Mean Power [W] = "
                       << OptPowerChann << endl
                       << "which is Mean Power [dBm] = "
                       << Linear2dBm(OptPowerChann)
                       << endl
                       << "set OptPowerChann [W] = "
                       <<  sq(OptPowerChann)/(GetAveragePower()/NumChannels)
                       << endl << endl;

          } // end if DebugLevel

} // ## end OptSignal::OptSignal

// ###############################################################

void OptSignal::ConstructOptSignal(string InFileName, RanNumGen *RNG2)
{

  // Created by John Zweck 24th Oct, 2000
  // The basic constructor for the class

  // IF YOU CHANGE THIS CONSTRUCTOR ALSO CHANGE 
  //OptSignal(OptSignal & OrigOptSignal); // Copy Constructor (JZ)
  //     OptSignal & operator=(OptSignal & RHS_Signal);
                  // assignment operator (JZ)
  //   OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
  //                   RanNumGen *RNG3);
  //             // Constructor for sum of signals

 ifstream InFile(InFileName.c_str(),ios::in);    
   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }
 

 RNG = RNG2;


 DebugLevel = ReadInt("DebugLevel"," ",0,4,LOWER_AND_UPPER,
                       OPTIONAL_NO_WARNING,1,&InFile,0);

 if(DebugLevel)
  {
   LogFileSeparator();
   *LogFile << InFileName << endl << endl;
   *LogFile << "DebugLevel = " << DebugLevel << endl;
  }
 
 TypeSimulation = typeSimulation(ReadInt("TypeSimulation"," ",1,2,
		       LOWER_AND_UPPER,MANDATORY,2,&InFile,DebugLevel));

 if(DebugLevel)
    *LogFile << WriteTypeSimulation(TypeSimulation) << " Simulation" << endl;

 
 TypePulse =  typePulse(ReadInt("TypePulse"," ",1,17,LOWER_AND_UPPER,&InFile,
				DebugLevel));

 if(TypePulse == INPUT_SEQUENCE){	 
	InputSequenceFileName = ReadString("InputSequenceFileName",
						 MANDATORY,"",&InFile,
						 DebugLevel);
 }
 

 if(TypePulse == INPUT_SIGNAL_FROM_FILE)
    {
        InputTimeDataFileName = ReadString("InputTimeDataFileName",
                                                 MANDATORY,"",&InFile,
                                                 DebugLevel);
    }


 qtPoints =   ReadInt("qtPoints"," ",0,-1,LOWER_ONLY,&InFile,DebugLevel);
 FWHM_Pulse = ReadDouble("FWHM_Pulse","s",0,-1,LOWER_ONLY,&InFile,DebugLevel);
 ChirpParameter = ReadDouble("ChirpParameter"," ",0,0,NO_BOUNDS,&InFile,
			     DebugLevel);
 double ExtinctionRatio = ReadDouble("ExtinctionRatio"," ",1,-1,
                                      UPPER_ONLY,&InFile,DebugLevel);

  if (ExtinctionRatio >= -1) 
    {
      LogFileStars();
      *LogFile << "            ERROR" << endl 
           << "Error: Extinction Ratio cannot be > then -1 dB." 
           << endl;
      LogFileStars();
      exit(1);
    }   

 RiseTime =  ReadDouble("RiseTime","s",0,-1,LOWER_ONLY,&InFile,DebugLevel);

 CenterFreq = 0;
 CenterWavelength = 0;

 CenterFreq =  ReadDouble("CenterFreq","Hz",0,-1,LOWER_ONLY,&InFile,
			  DebugLevel);
 CenterWavelength = ReadDouble("CenterWavelength","m",0,-1,LOWER_ONLY,
                                &InFile,DebugLevel);

 if(CenterFreq == 0)
   {
     CenterFreq = Wavelength2Frequency(CenterWavelength);
  
     if(DebugLevel)
      *LogFile << endl 
              << "Since CenterFreq is zero we use CenterWavelength:" << endl
              << endl
              << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
              << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
   }
  else
   if(CenterWavelength == 0)
    {
      CenterWavelength = Frequency2Wavelength(CenterFreq);
  
      if(DebugLevel)
       *LogFile << endl << "Since CenterWavelength is zero we use CenterFreq:" 
               << endl << endl 
               << "CenterWavelength [nm] = " << CenterWavelength*1e9 << endl
               << "CenterFreq [GHz] = " << CenterFreq*1e-9 << endl << endl;
    }
   else
     {
       *LogFile << endl 
	       << "ERROR: Make one of CenterFreq and CenterWavelength zero!" 
	       << endl << "The code will then use the other value." 
	       << endl <<  endl << "ABORTING NOW" << endl;

       exit(1);

     }


 ChannSpacing =  ReadDouble("ChannSpacing","Hz",0,-1,LOWER_ONLY,&InFile,
			    DebugLevel);
 BitRateChann =  ReadDouble("BitRateChann","bit/s",0,-1,LOWER_ONLY,&InFile,
			    DebugLevel);
 OptPowerChann = ReadDouble("OptPowerChann","W",0,-1,LOWER_ONLY,&InFile,
			    DebugLevel);

 RandomPolarizationsFlag = ReadInt("RandomPolarizationsFlag","",0,1,
                           LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile,
				   DebugLevel);

 UsingOptFiberLocalError =  ReadInt("UsingOptFiberLocalError"," ",0,1,
                               LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile,
				    DebugLevel);

 oBitString = new BitString(&InFile,RNG);

 NumChannels = oBitString->GetNumChannels();
 StringLength = oBitString->GetStringLength();


 if(DebugLevel)
   {
    LogFileSeparator();
    *LogFile << "Processing Input Signal Parameters and Allocating Memory"
            << endl; 
   }

 BitLength =  1.0/BitRateChann;
 TimeWindow =  BitLength*StringLength;   
 DeltaTime = TimeWindow/qtPoints;

 // DeltaFreq and FreqWindow are in Hz

 DeltaFreq =   1./TimeWindow; 
 FreqWindow = DeltaFreq*qtPoints;   

 LinearExtinctionRatio = dB2Linear(ExtinctionRatio);

 if(UsingOptFiberLocalError)
   {
    ComputeRelevantNumModesOFLEFlag = 
       ReadInt("ComputeRelevantNumModesOFLEFlag"," ",0,1,
                LOWER_AND_UPPER,OPTIONAL_NO_WARNING,0,&InFile,DebugLevel);

    ComputeRelevantNumModesOFLE();

    }

 

 ResetAccumulatedFirstOrderDispersionToZero();
 ResetAccumulatedSecondOrderDispersionToZero();

 NoiseBandwidthGaussianStatistics = 
                           ReadDouble("NoiseBandwidthGaussianStatistics","Hz",
					   0,-1,LOWER_ONLY,OPTIONAL_NO_WARNING,
					   FreqWindow,&InFile,DebugLevel);

 AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 0.0;

 if(DebugLevel)
  {
   *LogFile << endl;
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;      
   *LogFile << "qtPoints = " << qtPoints << endl;
   *LogFile << "StringLength = " << StringLength << endl;
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
   *LogFile << "BitLength [ps] = " << BitLength*1e12 << endl;
   *LogFile << "TimeWindow [ps] = " << TimeWindow*1e12 << endl;
   *LogFile << "FreqWindow [GHz]  = " << FreqWindow*1e-9 << endl;
   *LogFile << "DeltaTime [ps] = " << DeltaTime*1e12 << endl;
   *LogFile << "DeltaFreq [GHz] = " << DeltaFreq*1e-9 << endl;
 
   *LogFile << "FWHM_Pulse Power [ps] = " << FWHM_Pulse*1e12 << endl;
   *LogFile << "FWHM_Pulse Complex Wave [ps] = sqrt(2)*FWHM_Pulse Power = " 
           << sqrt(2)*FWHM_Pulse*1e12 << endl;


   *LogFile << "OptPowerChann [mW] = " << OptPowerChann*1e3 << endl;
   *LogFile << "ChirpParameter = " << ChirpParameter << endl;
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Center Wavelength [nm] = " << CenterWavelength*1e9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;
   *LogFile << "Linear Extinction Ratio = " << LinearExtinctionRatio << endl;
   *LogFile << "Rise Time [ps] = " << RiseTime *1e12 << endl;
   *LogFile << "AccumulatedFirstOrderDispersion [ps^2] = " 
	   << AccumulatedFirstOrderDispersion*1e24 << endl;
   *LogFile << "AccumulatedSecondOrderDispersion [ps^3] = " 
	   << AccumulatedSecondOrderDispersion*1e36 << endl;

   *LogFile << "RandomPolarizationsFlag = " << RandomPolarizationsFlag << endl;
   *LogFile << "UsingOptFiberLocalError = " << UsingOptFiberLocalError << endl;

   *LogFile << "NoiseBandwidthGaussianStatistics [GHz] = " 
	   << 1e-9*NoiseBandwidthGaussianStatistics << endl;
 
  } // ## end if

   //  JZFeb03: Instead of requiring qtPoints to be a power of 2
   // we just insist that it is even. OptSignal::WriteFileFreqPower
   // requires this I think.


   if(qtPoints%2)
     {
       cerr << "ERROR: qtPoints = " << qtPoints 
	    << " should be even. ABORTING NOW."
            << endl;   
       exit(1);
     }

   /********************
  // ## Check that qtPoints is a power of 2
   if( abs(log(qtPoints)/log(2) - (int) (log(qtPoints)/log(2) +0.1) )
        > 1e-8 )  
      {
        *LogFile << log(qtPoints)/log(2) << "   " 
                << (int) (log(qtPoints)/log(2) +0.1) << endl;
        *LogFile << "ERROR: qtPoints should be a power of 2." 
                << "ABORTING NOW"
                << endl;   
       exit(1);
     }              
   **********************/


   MemAllocated = 0;

   if ( ! MemAllocated) {
      MemAllocated = 1;

      fft = new cfftw(qtPoints, TimeWindow);

      /******
      sfftM.tOptSignalX = new cplx[qtPoints];   
      sfftM.fOptSignalX = new cplx[qtPoints];  
      sfftM.tOptSignalY = new cplx[qtPoints];   
      sfftM.fOptSignalY = new cplx[qtPoints];
      ***********/

      sfftM.tOptSignalX = NewCplx(qtPoints);
      sfftM.fOptSignalX = NewCplx(qtPoints);
      sfftM.tOptSignalY = NewCplx(qtPoints);
      sfftM.fOptSignalY = NewCplx(qtPoints);

      AllocateOptFiberLocalErrorWorkSpace();

      Frequency = new double[NumChannels];
      InStokes = new double[NumChannels*4];
      InJones = new cplx[NumChannels*2];
  
      JonesMatrixChannel = new cplx[NumChannels*4];  //added by Hai 7-1-03

      QuadratureWorkSpace = new double[qtPoints];

     }   

   // Initialize the array which contains the Frequency of each channel


 int EvenlySpacedFreqsFlag =  ReadInt("EvenlySpacedFreqsFlag","",0,2,
                                   LOWER_AND_UPPER,OPTIONAL_WARNING,1,&InFile,
				      DebugLevel);

 
// At some point we should change the name of the parameter read in from
// Signal.in to be ChannelAllocationMethod

// can not make it work. Walter Output signal of a propagation is all NaN
//ChannelAllocationMethod = EvenlySpacedFreqsFlag;
int ChannelAllocationMethod = EvenlySpacedFreqsFlag;

// ChannelAllocationMethod = 1: EvenlySpacedFreqs (Default method)
// ChannelAllocationMethod = 0: ReadInWavelengths
// ChannelAllocationMethod = 2: ReadInFrequencies


switch(ChannelAllocationMethod)
{

 case 1:
   {

     if(DebugLevel)
       {
        LogFileSeparator();

        *LogFile << "Evenly spaced channels" << endl << endl;

        *LogFile << "WARNING: Exact channel frequency values may have changed"
                << endl 
	        << "from input values to ensure they are exactly"
	        << endl << "on the frequency grid" << endl << endl;
       }

   // The middle frequency of each channel  is TempFreq. This frequency
   // may not lie exactly on the grid. For some choices of BitString
   // (especially a string consisting of all 1's) the optical spectrum
   // can be extremely narrow. Therefore, if DeltaFreq is not small enough the 
   // optical spectrum may not be accurately discretized. This can mean that
   // the maximum spectral power of the channels can vary by a large amount
   // when it is discretized. Rather making DeltaFreq inordinately small
   // a better fix is to adjust the middle frequencies of the channels
   // a little bit so that they lie exactly on the frequency grid.


   for(int ch = 0;ch<NumChannels;ch++)
     {

       double TempFreq = CenterFreq + GetChannelPhysicalIndex(ch)*ChannSpacing;

       if(TempFreq > CenterFreq)
	 {
           int mm = int(floor( (TempFreq - CenterFreq)/DeltaFreq ) );
           Frequency[ch] = CenterFreq + mm*DeltaFreq;
	 }
      else
	{
	 if(TempFreq < CenterFreq)
	   {
            int mm = int(floor( (CenterFreq - TempFreq)/DeltaFreq ) );
            Frequency[ch] = CenterFreq - mm*DeltaFreq;

	   }
         else
           Frequency[ch] = CenterFreq;
	}


       if(DebugLevel >=3)
	 {
          LogFile->precision(10);

          *LogFile << "Frequency in channel "
               << GetChannelPhysicalIndex(ch) << " = "
               <<  Frequency[ch]*1e-9 << " GHz" << endl;

          *LogFile << "Prior to placing frequency on grid its value was "
               << TempFreq*1e-9 << " GHz" << endl;
   
          LogFile->precision(6);
	 }

     } // ## end for ch loop

   if(DebugLevel)
     {
       *LogFile << endl << endl;

       *LogFile << "Left frequency in window [GHz] = "
               <<(CenterFreq - 0.5*FreqWindow)*1e-9 
               << endl;

       *LogFile << "Right frequency in window [GHz] = "
               <<(CenterFreq + 0.5*FreqWindow)*1e-9 
              << endl;

       *LogFile << "Min frequency of channels [GHz] = " 
               << Frequency[0]*1e-9 << endl;
       *LogFile << "Max frequency of channels [GHz] = " 
               << Frequency[NumChannels-1]*1e-9 << endl;

     }
   }

   break;

   case 0:

     {

       if(DebugLevel)
	 {
          *LogFile << "WARNING: Exact channel frequency values may have changed"
                  << endl 
	          << "from input values to ensure they are exactly"
	          << endl << "on the frequency grid" << endl << endl;
	 }

   // The middle frequency of each channel  is TempFreq. This frequency
   // may not lie exactly on the grid. For some choices of BitString
   // (especially a string consisting of all 1's) the optical spectrum
   // can be extremely narrow. Therefore, if DeltaFreq is not small enough the 
   // optical spectrum may not be accurately discretized. This can mean that
   // the maximum spectral power of the channels can vary by a large amount
   // when it is discretized. Rather making DeltaFreq inordinately small
   // a better fix is to adjust the middle frequencies of the channels
   // a little bit so that they lie exactly on the frequency grid.

      char C_String[40];
      string WavelengthChannName;

      double minFreq;
      double maxFreq;      

      for(int ch=0;ch<NumChannels;ch++)
	{
          sprintf(C_String,"WavelengthChann%d",ch+1);
          WavelengthChannName = C_String;

          double TempFreq =  Wavelength2Frequency(
                                   ReadDouble(WavelengthChannName,"m",0,-1,
                                   LOWER_ONLY,&InFile));


          if(TempFreq > CenterFreq)
	    {
              int mm = int(floor( (TempFreq - CenterFreq)/DeltaFreq ) );
              Frequency[ch] = CenterFreq + mm*DeltaFreq;
	    }
          else
	    {
	     if(TempFreq < CenterFreq)
	      {
               int mm = int(floor( (CenterFreq - TempFreq)/DeltaFreq ) );
               Frequency[ch] = CenterFreq - mm*DeltaFreq;
	      }
             else
              Frequency[ch] = CenterFreq;
	    }

	  if(DebugLevel >=3)
	    {
              LogFile->precision(10);

              *LogFile << "Frequency in channel "
               << GetChannelPhysicalIndex(ch) << " = "
               <<  Frequency[ch]*1e-9 << " GHz = " 
	       <<  Frequency2Wavelength(Frequency[ch])*1e9 << " nm" << endl; 

              *LogFile << "Prior to placing frequency on grid its value was "
               << TempFreq*1e-9 << " GHz" << endl << endl;


              LogFile->precision(6);
	    }

	  // Compute the min and max frequencies

          if(ch==0)
           {
             minFreq = Frequency[ch];
             maxFreq = Frequency[ch];
	   }
          else
	    {
              if(minFreq > Frequency[ch])
                 minFreq = Frequency[ch];

              if(maxFreq < Frequency[ch])
                 maxFreq = Frequency[ch];
	    }
         

	} // end for ch loop


      if(DebugLevel)
	{
            LogFileSeparator();
      
            if(DebugLevel >=2)
	      {
               for(int ch=0;ch<NumChannels;ch++)
                *LogFile << "Frequency[" << ch << "] = " 
                        <<  Frequency[ch] << endl;
	      }

            *LogFile << endl << endl;

	    *LogFile << "Frequency Window [GHz] = ["
		    << (CenterFreq - 0.5*FreqWindow)*1e-9 
		    << " , "
		    << (CenterFreq + 0.5*FreqWindow)*1e-9 
		    << "]" << endl;

            *LogFile << "Min frequency of channels [GHz] = " 
                    << minFreq*1e-9 << endl;
            *LogFile << "Max frequency of channels [GHz] = " 
                    << maxFreq*1e-9 << endl;

            cout << "WARNING: Set channel spacing to be min freq separation"
		 << endl << "between channels and restart job" << endl
		 << "This is vital if writing out power in a channel!" 
		 << endl;

	    // FIX: Automate the above sometime soon........
	    // Easiest way is to insist that channels are numbered
	    // monotonically in .in file
	} // end if DebugLevel
     } 
 break;
 
 case 2:

   {

     if(DebugLevel)
       *LogFile << "WARNING: Exact channel frequency values may have changed"
              << endl 
	      << "from input values to ensure they are exactly"
	      << endl << "on the frequency grid" << endl << endl;

   // The middle frequency of each channel  is TempFreq. This frequency
   // may not lie exactly on the grid. For some choices of BitString
   // (especially a string consisting of all 1's) the optical spectrum
   // can be extremely narrow. Therefore, if DeltaFreq is not small enough the 
   // optical spectrum may not be accurately discretized. This can mean that
   // the maximum spectral power of the channels can vary by a large amount
   // when it is discretized. Rather making DeltaFreq inordinately small
   // a better fix is to adjust the middle frequencies of the channels
   // a little bit so that they lie exactly on the frequency grid.

      char C_String[40];
      string FrequencyChannName;

      double minFreq;
      double maxFreq;      

      for(int ch=0;ch<NumChannels;ch++)
	{
          sprintf(C_String,"FrequencyChann%d",ch+1);
          FrequencyChannName = C_String;

          double TempFreq = ReadDouble(FrequencyChannName,"m",0,-1,
                                       LOWER_ONLY,&InFile);

          if(TempFreq > CenterFreq)
	    {
              int mm = int(floor( (TempFreq - CenterFreq)/DeltaFreq ) );
              Frequency[ch] = CenterFreq + mm*DeltaFreq;
	    }
          else
	    {
	     if(TempFreq < CenterFreq)
	      {
               int mm = int(floor( (CenterFreq - TempFreq)/DeltaFreq ) );
               Frequency[ch] = CenterFreq - mm*DeltaFreq;
	      }
             else
              Frequency[ch] = CenterFreq;
	    }

	  if(DebugLevel >=3)
	    {
              LogFile->precision(10);

              *LogFile << "Frequency in channel "
               << GetChannelPhysicalIndex(ch) << " = "
               <<  Frequency[ch]*1e-9 << " GHz = " 
	       <<  Frequency2Wavelength(Frequency[ch])*1e9 << " nm" << endl; 

              *LogFile << "Prior to placing frequency on grid its value was "
               << TempFreq*1e-9 << " GHz" << endl << endl;


              LogFile->precision(6);
	    }

	  // Compute the min and max frequencies

          if(ch==0)
           {
             minFreq = Frequency[ch];
             maxFreq = Frequency[ch];
	   }
          else
	    {
              if(minFreq > Frequency[ch])
                 minFreq = Frequency[ch];

              if(maxFreq < Frequency[ch])
                 maxFreq = Frequency[ch];
	    }
         

	} // end for ch loop

      if(DebugLevel)
	{
            LogFileSeparator();
      
            if(DebugLevel >=2)
	      {
               for(int ch=0;ch<NumChannels;ch++)
                *LogFile << "Frequency[" << ch << "] = " 
                        <<  Frequency[ch] << endl;
	      }

            *LogFile << endl << endl;

	    *LogFile << "Frequency Window [GHz] = ["
		    << (CenterFreq - 0.5*FreqWindow)*1e-9 
		    << " , "
		    << (CenterFreq + 0.5*FreqWindow)*1e-9 
		    << "]" << endl;

            *LogFile << "Min frequency of channels [GHz] = " 
                    << minFreq*1e-9 << endl;
            *LogFile << "Max frequency of channels [GHz] = " 
                    << maxFreq*1e-9 << endl;

            cout << "WARNING: Set channel spacing to be min freq separation"
		 << endl << "between channels and restart job" << endl
		 << "This is vital if writing out power in a channel!" 
		 << endl;

	    // FIX: Automate the above sometime soon........
	    // Easiest way is to insist that channels are numbered
	    // monotonically in .in file

	} // end if DebugLevel
   }
 
 break;

 default:
   cerr << "Invalid choice of ChannelAllocationMethod. ABORT!" << endl;
   exit(1);

} // ## end switch



   // ## Initialize the polarization

   // ## The Stokes and Jones initialization needs to be FIXed
   // ## when we add new constructor

   // ## First set the Stokes Vector for each Channel

   InitializeInStokes(&InFile);

     // Set the Jones Vector for each Channel

   cplx TempJones[2];

   for(int ch=0;ch<NumChannels;ch++)
     {
       Stokes2Jones(GetInStokes(ch),TempJones); 
       SetInJones(ch,TempJones);   
     }

   if(UsingOptFiberLocalError && TypeSimulation == VECTOR)
     {

      double RelFreqForPDVector = ReadDouble("RelFreqForPDVector","Hz",1,-1,
			             NO_BOUNDS,OPTIONAL_NO_WARNING,0,&InFile);

      FreqForPDVectorOFLE = RelFreqForPDVector + CenterFreq;


      if(DebugLevel)
      *LogFile << "FreqForPDVectorOFLE [THz] = " 
              << 1e-12*FreqForPDVectorOFLE << endl;
      
      OutputPDVectorOFLE = new double[3];


      ComputeInputPDVectorFlagOFLE = ReadInt("ComputeInputPDVectorFlagOFLE",
					     " ",0,1,LOWER_AND_UPPER,
					     OPTIONAL_NO_WARNING,0,&InFile);
  
      
      //ComputeInputPDVectorFlagOFLE = 0; // JZ FIX THIS 4/11/03

      if(ComputeInputPDVectorFlagOFLE)
	{

          InputPDVectorOFLE = new double[3];

          StokesTransferMatrixTransposeOFLE = new double* [3];

          for(int ii=0;ii<3;ii++)
	    StokesTransferMatrixTransposeOFLE[ii] = new double[3];

        }

     } 


 if(TypePulse == QPSK_CW)
  {
     TypeEOMElecFilter = typeEOMElecFilter(ReadInt("TypeEOMElecFilter"," ",
                                           0,2,LOWER_AND_UPPER,&InFile));

     *LogFile << "TypeEOMElecFilter = " << EOM_GetTypeElecFilter() << endl;

     if(TypeEOMElecFilter != EOM_NO_ELEC_FILTER)
        EOM_ElecFilter_FreqFWHM = ReadDouble("EOM_ElecFilter_FreqFWHM","Hz",
                                              0,-1,LOWER_ONLY,&InFile);

     EOM_sfftPM.tPowerSignal = NewCplx(qtPoints);
     EOM_sfftPM.fPowerSignal = NewCplx(qtPoints);

    if(!EOM_sfftPM.tPowerSignal || !EOM_sfftPM.fPowerSignal)
     {
      *LogFile << "Malloc error in OptSignal::EOM_Initialize" << endl
               << "Aborting Now" << endl << flush;

       exit(1);

     }
  } // end TypePulse == QPSK_CW


 if(TypePulse == BPSK_PHASE_MOD || TypePulse == BPSK_MACH_ZEHNDER ||
    TypePulse == CSRZ_MACH_ZEHNDER || TypePulse == RZ33_MACH_ZEHNDER ||
    TypePulse == RZ50_MACH_ZEHNDER || TypePulse == QPSK_SEQ_PM_NRZ ||
    TypePulse == QPSK_DPMZ_PM_NRZ || TypePulse == QPSK_SEQ_PM_AM_RZ33 ||
    TypePulse == QPSK_SEQ_PM_AM_RZ50 || TypePulse == QPSK_DPMZ_PM_AM_RZ33 ||
    TypePulse == QPSK_DPMZ_PM_AM_RZ50 || TypePulse == NRZ_RCOS_MACH_ZEHNDER)
   {

     if(TypePulse == NRZ_RCOS_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
        TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER)
        // These types use optical NRZ data of RCOS type
      {
       TypeEOMElecData = EOM_BINARY_DATA_RCOS;
       TypeEOMElecFilter = EOM_NO_ELEC_FILTER;
      }
     else // All other types use electrical NRZ data
      {       
       TypeEOMElecData = typeEOMElecData(ReadInt("TypeEOMElecData"," ",1,4,
                                       LOWER_AND_UPPER,&InFile));

       TypeEOMElecFilter = typeEOMElecFilter(ReadInt("TypeEOMElecFilter"," ",
                                           0,2,LOWER_AND_UPPER,&InFile));
      }


     *LogFile << "TypeEOMElecData = " << EOM_GetTypeElecData() << endl;
     *LogFile << "TypeEOMElecFilter = " << EOM_GetTypeElecFilter() << endl;

     if(TypeEOMElecFilter != EOM_NO_ELEC_FILTER)
        EOM_ElecFilter_FreqFWHM = ReadDouble("EOM_ElecFilter_FreqFWHM","Hz",
                                              0,-1,LOWER_ONLY,&InFile);

     if(TypeEOMElecData==EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING)
      {
          EOM_SuperGaussExponent =
           ReadInt("EOM_SuperGaussExponent"," ",0,-1,LOWER_ONLY,&InFile);
 
          if(EOM_SuperGaussExponent%2) // SuperGaussExponent is odd
           {
           cerr << "OptSignal::EOM_Initialize: SuperGaussExponent must be even"
            << endl
            << "ABORTING NOW!!"
            << endl;

            exit(1);

           }
       }


     if(TypePulse != BPSK_PHASE_MOD)
      {
       EOM_V_pi_RelError = 0.0;
       EOM_Phi_10_RelError = 0.0;
       EOM_Phi_20_RelError = 0.0;
       EOM_SinusoidalVertShift_RelError = 0.0;

/*       EOM_V_pi_RelError =  ReadDouble("EOM_V_pi_RelError"," ",0,-1,
                                    LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

       EOM_Phi_10_RelError =  ReadDouble("EOM_Phi_0_RelError"," ",0,-1,
                                     LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

       EOM_Phi_20_RelError = EOM_Phi_10_RelError; 

       EOM_SinusoidalVertShift_RelError =  ReadDouble(
                               "EOM_SinusoidalVertShift_RelError"," ",0,-1,
                                LOWER_ONLY,OPTIONAL_NO_WARNING,0,&InFile);

*/

      }

     if(TypeEOMElecData==EOM_BINARY_DATA_ARCTANH_SMOOTHING)
        {
          /* EOM_SlopeFactor =
           ReadDouble("EOM_SlopeFactor"," ",0,-1,LOWER_ONLY,&InFile);

          EOM_SlopeFactor *=  BitRateChann;
          */

          // 10%-90% rise time on electrical data

          EOM_RiseTime = ReadDouble("EOM_RiseTime"," ",0,-1,LOWER_ONLY,&InFile);

          EOM_SlopeFactor = (2.0/EOM_RiseTime)*atanh(0.8);

          EOM_ShiftAmt = 0.0;

        }

     if(TypeEOMElecData==EOM_BINARY_DATA_RCOS)
      {
       // EOM_SlopeFactor = beta*BitLength = beta*T in Ho and Kahn

       *LogFile << "TypeEOMElecData==EOM_BINARY_DATA_RCOS." << endl
                << "Specify either (optical power) RiseTime or" << endl
                << "(electrical) EOM_RiseTime by setting other one to -1"
                 << endl;

      double TmpRiseTime = ReadDouble("RiseTime"," ",-1,-2,LOWER_ONLY,&InFile);       double TmpEOM_RiseTime = ReadDouble("EOM_RiseTime"," ",-1,-2,
                                           LOWER_ONLY,&InFile);

       if((TmpRiseTime < 0 && TmpEOM_RiseTime < 0) || 
          (TmpRiseTime >= 0 && TmpEOM_RiseTime >= 0))
        {
         cerr << "OptSignal, TypeEOMElecData==EOM_BINARY_DATA_RCOS" << endl
              << "Exactly one of RiseTime and EOM_RiseTime must be >= 0"
              << endl;
         exit(1);

        }

       if(TmpRiseTime < 0)
        {
          EOM_RiseTime = TmpEOM_RiseTime;
          EOM_SlopeFactor = EOM_RiseTime/((2/pi)*asin(0.8));
        }

       if(TmpEOM_RiseTime < 0)
        {

         RiseTime = TmpRiseTime;
         EOM_SlopeFactor = RiseTime/((2/pi)*asin(-1+(4/pi)*asin(sqrt(0.9))));
         EOM_RiseTime = (2/pi)*asin(0.8)*EOM_SlopeFactor;

        }
      
       } // end if(TypeEOMElecData==EOM_BINARY_DATA_RCOS)

     EOM_sfftPM.tPowerSignal = NewCplx(qtPoints);
     EOM_sfftPM.fPowerSignal = NewCplx(qtPoints);

    if(!EOM_sfftPM.tPowerSignal || !EOM_sfftPM.fPowerSignal)
     {
      *LogFile << "Malloc error in OptSignal::EOM_Initialize" << endl
               << "Aborting Now" << endl << flush;

       exit(1);

     }

    if(TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ ||
       TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse == QPSK_SEQ_PM_AM_RZ50 ||
       TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50)
    {
     EOM_sfftPM1.tPowerSignal = NewCplx(qtPoints);
     EOM_sfftPM1.fPowerSignal = NewCplx(qtPoints);

    if(!EOM_sfftPM1.tPowerSignal || !EOM_sfftPM1.fPowerSignal)
     {
      *LogFile << "Malloc error in OptSignal::EOM_Initialize" << endl
               << "Aborting Now" << endl << flush;

       exit(1);

     }

     EOM_sfftPM2.tPowerSignal = NewCplx(qtPoints);
     EOM_sfftPM2.fPowerSignal = NewCplx(qtPoints);

    if(!EOM_sfftPM2.tPowerSignal || !EOM_sfftPM2.fPowerSignal)
     {
      *LogFile << "Malloc error in OptSignal::EOM_Initialize" << endl
               << "Aborting Now" << endl << flush;

       exit(1);

     }

    }  // end if


  } // end TypePulse == BPSK_PHASE_MOD || TypePulse == BPSK_MACH_ZEHNDER


   InFile.close();

   // Used in conjunction with GaussianReceiver class:
   // We use SignalNoiseBandwidth = 1.0 since
   // TotalNoiseSpectralDensity is noise power per Hertz
   // Tested and seen to be correct. JZ Jan 17th 2006

 oReduOptSignal = new ReduOptSignal(InFileName,1.0);

} // ## end ConstructOptSignal() 


// ##################################################################

void OptSignal::AllocateOptFiberLocalErrorWorkSpace(void)
{

  if(UsingOptFiberLocalError)
    {
      
      switch(TypeSimulation)
       {

          // Note: We only need the freq data not the time data 
          // and so we do not malloc the time fields

       case SCALAR:

	  OFLEWork.FreqDomainOperatorX = NewCplx(qtPoints);
	  OFLEWork.SqFreqDomainOperatorX = NewCplx(qtPoints);

          OFLEWork.sfftM_Start.fOptSignalX = NewCplx(qtPoints);
          OFLEWork.sfftM_Coarse.fOptSignalX = NewCplx(qtPoints);

          OFLEWork.LaserWorkX = NewCplx(qtPoints);     

          if(!OFLEWork.FreqDomainOperatorX || 
	     !OFLEWork.SqFreqDomainOperatorX || 
	     !OFLEWork.sfftM_Start.fOptSignalX || 
             !OFLEWork.sfftM_Coarse.fOptSignalX || 
             !OFLEWork.LaserWorkX)
            {
              cerr << "Malloc error in " << endl
		   << "OptSignal::AllocateOptFiberLocalErrorWorkSpace" 
                   << "Aborting now!!" << endl;
    
              exit(1);
            }

	  break;

       case VECTOR:

          OFLEWork.FreqDomainOperatorX = NewCplx(qtPoints);
          OFLEWork.FreqDomainOperatorY = NewCplx(qtPoints);
          OFLEWork.SqFreqDomainOperatorX = NewCplx(qtPoints);
          OFLEWork.SqFreqDomainOperatorY = NewCplx(qtPoints);

          OFLEWork.sfftM_Start.fOptSignalX = NewCplx(qtPoints);
          OFLEWork.sfftM_Start.fOptSignalY = NewCplx(qtPoints);
          OFLEWork.sfftM_Coarse.fOptSignalX = NewCplx(qtPoints);
          OFLEWork.sfftM_Coarse.fOptSignalY = NewCplx(qtPoints);

          OFLEWork.sfftM_Coarse.fOptSignalX = NewCplx(qtPoints);
          OFLEWork.sfftM_Coarse.fOptSignalY = NewCplx(qtPoints);

          OFLEWork.LaserWorkX = NewCplx(qtPoints);                 
          OFLEWork.LaserWorkY = NewCplx(qtPoints);

          if(!OFLEWork.FreqDomainOperatorX ||
             !OFLEWork.FreqDomainOperatorY ||
	     !OFLEWork.SqFreqDomainOperatorX ||
             !OFLEWork.SqFreqDomainOperatorY ||
             !OFLEWork.sfftM_Start.fOptSignalX || 
             !OFLEWork.sfftM_Start.fOptSignalY ||
             !OFLEWork.sfftM_Coarse.fOptSignalX || 
             !OFLEWork.sfftM_Coarse.fOptSignalY||
             !OFLEWork.LaserWorkX ||
             !OFLEWork.LaserWorkY)
            {
	      cerr << "Malloc error in " << endl
		   << "OptSignal::AllocateOptFiberLocalErrorWorkSpace" 
                   << "Aborting now!!" << endl;
    
               exit(1);
             }

       break;

       default:

	    cerr << "Error in TypeSimulation switch in  " << endl
		 << "OptSignal::AllocateOptFiberLocalErrorWorkSpace" << endl;

            exit(1);

       } // end switch

    } // end if

} // ## end OptSignal::AllocateOptFiberLocalErrorWorkSpace

// ###############################################################

void OptSignal::DeallocateOptFiberLocalErrorWorkSpace(void)
{

  if(UsingOptFiberLocalError)
    {

     FreeCplx(OFLEWork.FreqDomainOperatorX);
     FreeCplx(OFLEWork.SqFreqDomainOperatorX);
   

     switch(TypeSimulation)
      {
       case SCALAR:
	
          FreeCplx(OFLEWork.sfftM_Start.fOptSignalX);
          FreeCplx(OFLEWork.sfftM_Coarse.fOptSignalX);
          FreeCplx(OFLEWork.LaserWorkX);
	  break;

       case VECTOR:
          FreeCplx(OFLEWork.FreqDomainOperatorY);
          FreeCplx(OFLEWork.SqFreqDomainOperatorY);
          FreeCplx(OFLEWork.sfftM_Start.fOptSignalX);
          FreeCplx(OFLEWork.sfftM_Start.fOptSignalY);
          FreeCplx(OFLEWork.sfftM_Coarse.fOptSignalX);
          FreeCplx(OFLEWork.sfftM_Coarse.fOptSignalY);
          FreeCplx(OFLEWork.LaserWorkX);
          FreeCplx(OFLEWork.LaserWorkY);
          break;
      }// end switch
    }// end if

} // ## end DeallocateOptFiberLocalErrorWorkSpace

// ################################################################

// ##### Destructor: releases the allocated memory
// ## By: Ivan Lima
OptSignal::~OptSignal()
{
   if (MemAllocated) {

     /**********
      delete [] sfftM.tOptSignalX;
      delete [] sfftM.fOptSignalX;
      delete [] sfftM.tOptSignalY;
      delete [] sfftM.fOptSignalY;   
     ***************/

     FreeCplx(sfftM.tOptSignalX);
     FreeCplx(sfftM.fOptSignalX);
     FreeCplx(sfftM.tOptSignalY);
     FreeCplx(sfftM.fOptSignalY);

      delete [] Frequency;
      delete [] InStokes;
      delete [] InJones;

     if(TypePulse == QPSK_CW || TypePulse == BPSK_PHASE_MOD ||
        TypePulse == BPSK_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
        TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER ||
        TypePulse == NRZ_RCOS_MACH_ZEHNDER)
        {
         FreeCplx(EOM_sfftPM.tPowerSignal);
         FreeCplx(EOM_sfftPM.fPowerSignal);
        }

     if(TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ || 
        TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse == QPSK_SEQ_PM_AM_RZ50 ||
        TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50)
       {
         FreeCplx(EOM_sfftPM.tPowerSignal);
         FreeCplx(EOM_sfftPM.fPowerSignal);
         FreeCplx(EOM_sfftPM1.tPowerSignal);
         FreeCplx(EOM_sfftPM1.fPowerSignal);
         FreeCplx(EOM_sfftPM2.tPowerSignal);
         FreeCplx(EOM_sfftPM2.fPowerSignal);
       }

      oBitString->~BitString();
      DeallocateOptFiberLocalErrorWorkSpace();

      delete [] QuadratureWorkSpace;

      delete fft;

      oReduOptSignal->~ReduOptSignal();
   }    
}

// ###############################################################

OptSignal::OptSignal(OptSignal & OrigOptSignal)
{

  // Written by John Zweck 16th Oct 2000
  // Copy Constructor
  // Checked: BitStrings and ReduOptSignals are deep copies.

   RNG = OrigOptSignal.RNG;  // We want to keep the same RNG
   TypeSimulation = OrigOptSignal.TypeSimulation;
   TypePulse       = OrigOptSignal.TypePulse;
   qtPoints        = OrigOptSignal.qtPoints; 
   FWHM_Pulse      = OrigOptSignal.FWHM_Pulse; 
   ChirpParameter  = OrigOptSignal.ChirpParameter;
   CenterFreq      = OrigOptSignal.CenterFreq;
   CenterWavelength =  OrigOptSignal.CenterWavelength;
   BitRateChann    = OrigOptSignal.BitRateChann;
   NumChannels = OrigOptSignal.NumChannels;
   
   ChannSpacing = OrigOptSignal.ChannSpacing;
   LinearExtinctionRatio = OrigOptSignal.LinearExtinctionRatio;
   RiseTime = OrigOptSignal.RiseTime;
   OptPowerChann = OrigOptSignal.OptPowerChann;
   RandomPolarizationsFlag  = OrigOptSignal.RandomPolarizationsFlag;
   UsingOptFiberLocalError = OrigOptSignal.UsingOptFiberLocalError;

   oBitString = new  BitString(*OrigOptSignal.oBitString);
   oReduOptSignal = new ReduOptSignal(*OrigOptSignal.oReduOptSignal);

   StringLength = oBitString->GetStringLength();

   BitLength = OrigOptSignal.BitLength;
   TimeWindow = OrigOptSignal.TimeWindow;
   DeltaTime = OrigOptSignal. DeltaTime;
   DeltaFreq = OrigOptSignal.DeltaFreq;
   FreqWindow = OrigOptSignal.FreqWindow;
   DebugLevel = OrigOptSignal.DebugLevel;
   AccumulatedFirstOrderDispersion = 
           OrigOptSignal.AccumulatedFirstOrderDispersion;
   AccumulatedSecondOrderDispersion = 
           OrigOptSignal.AccumulatedSecondOrderDispersion;


   AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 
       OrigOptSignal.AccumulatedTotalNoiseSpectralDensityGaussianStatistics;

   NoiseBandwidthGaussianStatistics = 
     OrigOptSignal.NoiseBandwidthGaussianStatistics;

   FreqForPDVectorOFLE = OrigOptSignal.FreqForPDVectorOFLE;
   ComputeInputPDVectorFlagOFLE = OrigOptSignal.ComputeInputPDVectorFlagOFLE;

   ComputeRelevantNumModesOFLEFlag = 
                             OrigOptSignal.ComputeRelevantNumModesOFLEFlag;
       

   RelevantNumModesOFLE = OrigOptSignal.RelevantNumModesOFLE;

   TypeEOMElecData = OrigOptSignal.TypeEOMElecData;
   TypeEOMElecFilter = OrigOptSignal.TypeEOMElecFilter;
   EOM_SlopeFactor = OrigOptSignal.EOM_SlopeFactor;
   EOM_ShiftAmt = OrigOptSignal.EOM_ShiftAmt;
   EOM_SuperGaussExponent = OrigOptSignal.EOM_SuperGaussExponent;
   EOM_ElecFilter_FreqFWHM = OrigOptSignal.EOM_ElecFilter_FreqFWHM;

EOM_V_pi_RelError = OrigOptSignal.EOM_V_pi_RelError;
EOM_Phi_10_RelError = OrigOptSignal.EOM_Phi_10_RelError;
EOM_Phi_20_RelError = OrigOptSignal.EOM_Phi_20_RelError;
EOM_SinusoidalVertShift_RelError = 
  OrigOptSignal.EOM_SinusoidalVertShift_RelError;

   /*
   *LogFileSeparator();
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;      
   *LogFile << "qtPoints = " << qtPoints << endl;
   *LogFile << "StringLength = " << StringLength << endl;
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
   *LogFile << "BitLength [ps] = " << BitLength*1e12 << endl;
   *LogFile << "TimeWindow [ps] = " << TimeWindow*1e12 << endl;
   *LogFile << "FreqWindow [GHz]  = " << FreqWindow*1e-9 << endl;
   *LogFile << "DeltaTime [ps] = " << DeltaTime*1e12 << endl;
   *LogFile << "DeltaFreq [GHz] = " << DeltaFreq*1e-9 << endl;
 
   *LogFile << "FWHM_Pulse [ps] = " << FWHM_Pulse*1e12 << endl;
   *LogFile << "OptPowerChann [mW] = " << OptPowerChann*1e3 << endl;
   *LogFile << "ChirpParameter = " << ChirpParameter << endl;
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;
   *LogFile << "Linear Extinction Ratio = " << LinearExtinctionRatio << endl;
   *LogFile << "RandomPolarizationsFlag = " << RandomPolarizationsFlag << endl;
   *LogFile << "UsingOptFiberLocalError = " << UsingOptFiberLocalError << endl;

   */

   MemAllocated = 0;

   if ( ! MemAllocated) {
      MemAllocated = 1;
 
      fft = new cfftw(qtPoints, TimeWindow);

      sfftM.tOptSignalX = NewCplx(qtPoints);
      sfftM.fOptSignalX = NewCplx(qtPoints);
      sfftM.tOptSignalY = NewCplx(qtPoints);
      sfftM.fOptSignalY = NewCplx(qtPoints);
            
      AllocateOptFiberLocalErrorWorkSpace();
 
      Frequency = new double[NumChannels];
      InStokes = new double[NumChannels*4];
      InJones = new cplx[NumChannels*2];

      QuadratureWorkSpace = new double[qtPoints];

      if(TypePulse == QPSK_CW || TypePulse == BPSK_PHASE_MOD ||
         TypePulse == BPSK_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
         TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER ||
         TypePulse == NRZ_RCOS_MACH_ZEHNDER)
      {
      EOM_sfftPM.tPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM.fPowerSignal = NewCplx(qtPoints);
      }

     if(TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ || 
        TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse == QPSK_SEQ_PM_AM_RZ50 ||         TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50)
      {
      EOM_sfftPM.tPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM.fPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM1.tPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM1.fPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM2.tPowerSignal = NewCplx(qtPoints);
      EOM_sfftPM2.fPowerSignal = NewCplx(qtPoints);
      }



      if(UsingOptFiberLocalError && TypeSimulation == VECTOR)
	{
          OutputPDVectorOFLE = new double[3];
         
          for(int ii=0;ii<3;ii++)
	    OutputPDVectorOFLE[ii] = OrigOptSignal.OutputPDVectorOFLE[ii];

          if(ComputeInputPDVectorFlagOFLE)
	   {

            InputPDVectorOFLE = new double[3];

            StokesTransferMatrixTransposeOFLE = new double* [3];

            for(int ii=0;ii<3;ii++)
	      StokesTransferMatrixTransposeOFLE[ii] = new double[3];



            for(int ii=0;ii<3;ii++)
	     InputPDVectorOFLE[ii] = OrigOptSignal.InputPDVectorOFLE[ii];

            for(int ii=0;ii<3;ii++)
             for(int jj=0;jj<3;jj++)
	      StokesTransferMatrixTransposeOFLE[ii][jj] = 
		OrigOptSignal.StokesTransferMatrixTransposeOFLE[ii][jj];

	   } // end if ComputeInputPDVectorFlagOFLE
	} // end if UsingOptFiberLocalError

   } // ## end if !MemAlloc  
   
   for(int ii = 0; ii < qtPoints; ii++)
     sfftM.tOptSignalX[ii] = OrigOptSignal.sfftM.tOptSignalX[ii];

   for(int ii = 0; ii < qtPoints; ii++)
     sfftM.tOptSignalY[ii] = OrigOptSignal.sfftM.tOptSignalY[ii];

   Time2Freq();

   if(TypePulse == QPSK_CW || TypePulse == BPSK_PHASE_MOD ||
      TypePulse == BPSK_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
      TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER ||
      TypePulse == NRZ_RCOS_MACH_ZEHNDER)
    {
     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.tPowerSignal[ii] = 
      OrigOptSignal.EOM_sfftPM.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.fPowerSignal[ii] = 
      OrigOptSignal.EOM_sfftPM.fPowerSignal[ii];

    }


   if(TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ ||
      TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse == QPSK_SEQ_PM_AM_RZ50 ||         TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50)
     {
     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.tPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.fPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM.fPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM1.tPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM1.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM1.fPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM1.fPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM2.tPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM2.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM2.fPowerSignal[ii] =
      OrigOptSignal.EOM_sfftPM2.fPowerSignal[ii];

     }


   for(int ch=0;ch<NumChannels;ch++)
     Frequency[ch] = OrigOptSignal.Frequency[ch];

   // ## Initialize the polarization

   // ## First set the Stokes Vector for each Channel

   for(int ch=0;ch<NumChannels;ch++)
   for(int ss = 0; ss < 4; ss++)
     SetInStokes(ch,ss,OrigOptSignal.GetInStokes(ch,ss));
     
     // ## Second set the Jones Vector for each Channel

   for(int ch=0;ch<NumChannels;ch++)
   for(int  ss = 0; ss < 2; ss++) 
     {
       SetInJones(ch,ss,OrigOptSignal.GetInJones(ch,ss));
     }


}// end copy constructor

// #################################################################

OptSignal & OptSignal::operator=(OptSignal & RHS_Signal)
{

  // Written by John Zweck 21tst Nov 2000
  // Assignment Operator

  // This assignment operator assumes that the Number of Channels, 
  // qtPoints, TimeWindow, StringLength are unchanged so that none
  // of the dynamically allocated data arrays/classes changes size.
  // It was designed to be used in conjunction with the ReceiverSubsystem
  // class.

   if(qtPoints != RHS_Signal.qtPoints || 
      NumChannels != RHS_Signal.NumChannels ||
      StringLength != RHS_Signal.StringLength ||
      TimeWindow != RHS_Signal.TimeWindow)
     {

       cout << qtPoints << " " << RHS_Signal.qtPoints << endl;
       cout <<  NumChannels << " " << RHS_Signal.NumChannels << endl;
       cout <<  StringLength << " " << RHS_Signal.StringLength << endl;
       cout <<  TimeWindow << " " << RHS_Signal.TimeWindow << endl;

       *LogFile << "Invalid use of OptSignal::operator=" << endl
                << "Aborting Now!! G'day" << endl << flush;

       cout << "Invalid use of OptSignal::operator=" << endl
                << "Aborting Now!! G'day" << endl;

       exit(1);

     }


   RNG = RHS_Signal.RNG; // We want to keep the same RNG
   TypeSimulation  = RHS_Signal.TypeSimulation;
   TypePulse       = RHS_Signal.TypePulse;
   qtPoints        = RHS_Signal.qtPoints; 
   FWHM_Pulse      = RHS_Signal.FWHM_Pulse; 
   ChirpParameter  = RHS_Signal.ChirpParameter;
   CenterFreq      = RHS_Signal.CenterFreq;
   CenterWavelength = RHS_Signal.CenterWavelength;
   BitRateChann    = RHS_Signal.BitRateChann;
   NumChannels = RHS_Signal.NumChannels;
   ChannSpacing = RHS_Signal.ChannSpacing;
   
   
   LinearExtinctionRatio = RHS_Signal.LinearExtinctionRatio;
   RiseTime = RHS_Signal.RiseTime;
   OptPowerChann = RHS_Signal.OptPowerChann;
   RandomPolarizationsFlag = RHS_Signal.RandomPolarizationsFlag;
   UsingOptFiberLocalError = RHS_Signal.UsingOptFiberLocalError;
    ComputeRelevantNumModesOFLEFlag = 
                             RHS_Signal.ComputeRelevantNumModesOFLEFlag;
       

   RelevantNumModesOFLE = RHS_Signal.RelevantNumModesOFLE;

   TypeEOMElecData = RHS_Signal.TypeEOMElecData;
   TypeEOMElecFilter = RHS_Signal.TypeEOMElecFilter;
   EOM_SlopeFactor = RHS_Signal.EOM_SlopeFactor;
   EOM_ShiftAmt = RHS_Signal.EOM_ShiftAmt;
   EOM_SuperGaussExponent = RHS_Signal.EOM_SuperGaussExponent;
   EOM_ElecFilter_FreqFWHM = RHS_Signal.EOM_ElecFilter_FreqFWHM;

EOM_V_pi_RelError = RHS_Signal.EOM_V_pi_RelError;
EOM_Phi_10_RelError = RHS_Signal.EOM_Phi_10_RelError;
EOM_Phi_20_RelError = RHS_Signal.EOM_Phi_20_RelError;
EOM_SinusoidalVertShift_RelError = RHS_Signal.EOM_SinusoidalVertShift_RelError;

   FreqForPDVectorOFLE = RHS_Signal.FreqForPDVectorOFLE;
   ComputeInputPDVectorFlagOFLE = RHS_Signal.ComputeInputPDVectorFlagOFLE;


   *oBitString = *RHS_Signal.oBitString;
   *oReduOptSignal = *RHS_Signal.oReduOptSignal;

   StringLength = RHS_Signal.StringLength;

   BitLength = RHS_Signal.BitLength;
   TimeWindow = RHS_Signal.TimeWindow;
   DeltaTime = RHS_Signal. DeltaTime;
   DeltaFreq = RHS_Signal.DeltaFreq;
   FreqWindow = RHS_Signal.FreqWindow;
   DebugLevel = RHS_Signal.DebugLevel;
   AccumulatedFirstOrderDispersion = 
            RHS_Signal.AccumulatedFirstOrderDispersion;
   AccumulatedSecondOrderDispersion = 
            RHS_Signal.AccumulatedSecondOrderDispersion;

   AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 
     RHS_Signal.AccumulatedTotalNoiseSpectralDensityGaussianStatistics;

   NoiseBandwidthGaussianStatistics = 
     RHS_Signal.NoiseBandwidthGaussianStatistics;
  
   /*
   LogFileSeparator();
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;      
   *LogFile << "qtPoints = " << qtPoints << endl;
   *LogFile << "StringLength = " << StringLength << endl;
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
   *LogFile << "BitLength [ps] = " << BitLength*1e12 << endl;
   *LogFile << "TimeWindow [ps] = " << TimeWindow*1e12 << endl;
   *LogFile << "FreqWindow [GHz]  = " << FreqWindow*1e-9 << endl;
   *LogFile << "DeltaTime [ps] = " << DeltaTime*1e12 << endl;
   *LogFile << "DeltaFreq [GHz] = " << DeltaFreq*1e-9 << endl;
 
   *LogFile << "FWHM_Pulse [ps] = " << FWHM_Pulse*1e12 << endl;
   *LogFile << "OptPowerChann [mW] = " << OptPowerChann*1e3 << endl;
   *LogFile << "ChirpParameter = " << ChirpParameter << endl;
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;
   *LogFile << "Linear Extinction Ratio = " << LinearExtinctionRatio << endl;
   *LogFile << "RandomPolarizationsFlag = " << RandomPolarizationsFlag << endl;
   *LogFile << "UsingOptFiberLocalError = " << UsingOptFiberLocalError << endl;

   */

   MemAllocated = 1;
    
   //## Speed up the data transfer (Ivan Lima, 7/24/01)   
   for(int ii = 0; ii < qtPoints; ii++) {
      sfftM.tOptSignalX[ii] = RHS_Signal.sfftM.tOptSignalX[ii];
      sfftM.tOptSignalY[ii] = RHS_Signal.sfftM.tOptSignalY[ii];     
      sfftM.fOptSignalX[ii] = RHS_Signal.sfftM.fOptSignalX[ii];
      sfftM.fOptSignalY[ii] = RHS_Signal.sfftM.fOptSignalY[ii];     
   }

   if(TypePulse == QPSK_CW || TypePulse == BPSK_PHASE_MOD ||
      TypePulse == BPSK_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
      TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER ||
      TypePulse == NRZ_RCOS_MACH_ZEHNDER)
    {
     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.tPowerSignal[ii] = RHS_Signal.EOM_sfftPM.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.fPowerSignal[ii] = RHS_Signal.EOM_sfftPM.fPowerSignal[ii];

    }

   if(TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ ||
      TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse == QPSK_SEQ_PM_AM_RZ50 ||         TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50)
     {
     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.tPowerSignal[ii] = RHS_Signal.EOM_sfftPM.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM.fPowerSignal[ii] = RHS_Signal.EOM_sfftPM.fPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM1.tPowerSignal[ii] = RHS_Signal.EOM_sfftPM1.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM1.fPowerSignal[ii] = RHS_Signal.EOM_sfftPM1.fPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM2.tPowerSignal[ii] = RHS_Signal.EOM_sfftPM2.tPowerSignal[ii];

     for(int ii = 0; ii < qtPoints; ii++)
      EOM_sfftPM2.fPowerSignal[ii] = RHS_Signal.EOM_sfftPM2.fPowerSignal[ii];

     }


   for(int ch=0;ch<NumChannels;ch++)
     Frequency[ch] = RHS_Signal.Frequency[ch];

   // ## Initialize the polarization

   // ## First set the Stokes Vector for each Channel

   for(int ch=0;ch<NumChannels;ch++)
   for(int ss = 0; ss < 4; ss++)
     SetInStokes(ch,ss,RHS_Signal.GetInStokes(ch,ss));
     
     // ## Second set the Jones Vector for each Channel

   for(int ch=0;ch<NumChannels;ch++)
   for(int  ss = 0; ss < 2; ss++) 
     {
       SetInJones(ch,ss,RHS_Signal.GetInJones(ch,ss));
     }

      if(UsingOptFiberLocalError && TypeSimulation == VECTOR)
	{
          for(int ii = 0; ii < 3; ii++)
           OutputPDVectorOFLE[ii] = RHS_Signal.OutputPDVectorOFLE[ii];

          if(ComputeInputPDVectorFlagOFLE)
	   {

            for(int ii=0;ii<3;ii++)
	     InputPDVectorOFLE[ii] = RHS_Signal.InputPDVectorOFLE[ii];

            for(int ii=0;ii<3;ii++)
             for(int jj=0;jj<3;jj++)
	      StokesTransferMatrixTransposeOFLE[ii][jj] = 
		RHS_Signal.StokesTransferMatrixTransposeOFLE[ii][jj];

	   } // end if ComputeInputPDVectorFlagOFLE
	} // end if UsingOptFiberLocalError

   return *this;

}// end assignment operator (=)

// #############################################################
//## The subtraction is done only in frequency domain
//## to remove the noise added to a noiseless signal.
//## By: Ivan Lima (6/9/02)

void OptSignal::SubtractFreqDomain(OptSignal *oOptSignal2)
{
   if (oOptSignal2->GetQtPoints() != qtPoints) {
      cout << "Error from OptSignal::SubtractFrequencyDomain()" << endl
           << "The two OptSignal objects have different FFT points" 
	   << endl << "Aborting now." << endl;
      exit(1);	     
   }
   for(int jj = 0; jj < qtPoints; jj++)  {
      sfftM.fOptSignalX[jj] -= oOptSignal2->sfftM.fOptSignalX[jj];
      sfftM.fOptSignalY[jj] -= oOptSignal2->sfftM.fOptSignalY[jj];
   }
   Freq2Time();
} // ## end




// #############################################################
//## The sum is done only in frequency domain
//## By: Ivan Lima (6/17/02)

void OptSignal::AddFreqDomain(OptSignal *oOptSignal2)
{
   if (oOptSignal2->GetQtPoints() != qtPoints) {
      cout << "Error from OptSignal::AddFrequencyDomain()" << endl
           << "The two OptSignal objects have different FFT points" 
	   << endl << "Aborting now." << endl;
     exit(1);	     
   }
   for(int jj = 0; jj < qtPoints; jj++)  {
      sfftM.fOptSignalX[jj] += oOptSignal2->sfftM.fOptSignalX[jj];
      sfftM.fOptSignalY[jj] += oOptSignal2->sfftM.fOptSignalY[jj];
   }
   Freq2Time();
} // ## end

// ###########################################################

OptSignal & OptSignal::operator-(OptSignal & Signal2)
{

OptSignal * DifferenceSignal;

DifferenceSignal = new OptSignal(*this);

int jj;

for(jj=0;jj<qtPoints;jj++)
  {
    DifferenceSignal->sfftM.tOptSignalX[jj] -= Signal2.sfftM.tOptSignalX[jj];
  }

for(jj=0;jj<qtPoints;jj++)
  {
    DifferenceSignal->sfftM.tOptSignalY[jj] -= Signal2.sfftM.tOptSignalY[jj];
  }

for(jj=0;jj<qtPoints;jj++)
  {
    DifferenceSignal->sfftM.fOptSignalX[jj] -= Signal2.sfftM.fOptSignalX[jj];
  }

for(jj=0;jj<qtPoints;jj++)
  {
    DifferenceSignal->sfftM.fOptSignalY[jj] -= Signal2.sfftM.fOptSignalY[jj];
  }

return *DifferenceSignal;

} // ## end operator-


// ################################################################

OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
                     RanNumGen *RNG3)
{

  // This constructor is used to take the sum of two signals
  // For most parameters we use those of Signal1.

  // Does not work when oReducedOptSignal is important to track!
  // Does not work for EOM data

  RNG = RNG3;

  if(Signal1.GetQtPoints() != Signal2.GetQtPoints() ||
     Signal1.GetStringLength() != Signal2.GetStringLength() ||
     Signal1.GetBitRateChann() != Signal2.GetBitRateChann() ||
     Signal1.GetTypeSimulation() != Signal2.GetTypeSimulation() ||
     Signal1.GetCenterFreq() != Signal2.GetCenterFreq() )
    {
      cerr << "OptSignal::OptSignal(Signal1,Signal2,RNG2)" << endl
	      << "Input Signals incompatable" << endl
	      << "Aborting now!!" << endl;
      exit(1);
    }

  DebugLevel = Signal1.DebugLevel;
  qtPoints = Signal1.qtPoints;
  TypeSimulation = Signal1.GetTypeSimulation();
  TypePulse = Signal1.GetTypePulse();
  FWHM_Pulse = Signal1.GetFWHM_Pulse();
  ChirpParameter = Signal1.ChirpParameter;
  CenterFreq = Signal1.GetCenterFreq();
  CenterWavelength = Signal1.CenterWavelength;
  BitRateChann = Signal1.BitRateChann;

  NumChannels = Signal1.NumChannels + Signal2.NumChannels;  // NB!!
  ChannSpacing =  min(Signal1.ChannSpacing,Signal2.ChannSpacing); // NB!!

  //ChannelAllocationMethod=Signal1.ChannelAllocationMethod; //carefull! 
 
  LinearExtinctionRatio = Signal1.LinearExtinctionRatio;
  RiseTime  = Signal1.RiseTime;
  OptPowerChann = Signal1.OptPowerChann;
  RandomPolarizationsFlag = Signal1.RandomPolarizationsFlag;   
  UsingOptFiberLocalError = Signal1.UsingOptFiberLocalError;
  ComputeRelevantNumModesOFLEFlag = 
                             Signal1.ComputeRelevantNumModesOFLEFlag;
       

   RelevantNumModesOFLE = Signal1.RelevantNumModesOFLE;

  oBitString = new  BitString(*Signal1.oBitString,*Signal2.oBitString,RNG);

  oReduOptSignal = new ReduOptSignal(*Signal1.oReduOptSignal); // WRONG!

  if(DebugLevel >=3)
    {
     Signal1.oBitString->WriteBitString();
     Signal2.oBitString->WriteBitString();
     oBitString->WriteBitString();
    }


  StringLength = oBitString->GetStringLength();

   BitLength = Signal1.BitLength;
   TimeWindow = Signal1.TimeWindow;
   DeltaTime = Signal1. DeltaTime;
   DeltaFreq = Signal1.DeltaFreq;
   FreqWindow = Signal1.FreqWindow;

   AccumulatedFirstOrderDispersion = 
           Signal1.AccumulatedFirstOrderDispersion;
   AccumulatedSecondOrderDispersion = 
           Signal1.AccumulatedSecondOrderDispersion;

   AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 
     Signal1.AccumulatedTotalNoiseSpectralDensityGaussianStatistics;

   NoiseBandwidthGaussianStatistics = Signal1.NoiseBandwidthGaussianStatistics;

   ResetAccumulatedFirstOrderDispersionToZero();
   ResetAccumulatedSecondOrderDispersionToZero();

 if(DebugLevel >= 3)
  {
   LogFileSeparator();
   *LogFile << "SumSignals Constructor" << endl;
   *LogFile << "NumChannels = " << NumChannels << endl;
   *LogFile << "SIGNAL PARAMETERS" << endl << endl;
   *LogFile << "Pulse Type = " << WriteTypePulse() << endl << endl;      
   *LogFile << "qtPoints = " << qtPoints << endl;
   *LogFile << "StringLength = " << StringLength << endl;
   *LogFile << "BitRateChann [Gb/s] = " << BitRateChann*1e-9 << endl;
   *LogFile << "BitLength [ps] = " << BitLength*1e12 << endl;
   *LogFile << "TimeWindow [ps] = " << TimeWindow*1e12 << endl;
   *LogFile << "FreqWindow [GHz]  = " << FreqWindow*1e-9 << endl;
   *LogFile << "DeltaTime [ps] = " << DeltaTime*1e12 << endl;
   *LogFile << "DeltaFreq [GHz] = " << DeltaFreq*1e-9 << endl;
 
   *LogFile << "FWHM_Pulse [ps] = " << FWHM_Pulse*1e12 << endl;
   *LogFile << "OptPowerChann [mW] = " << OptPowerChann*1e3 << endl;
   *LogFile << "ChirpParameter = " << ChirpParameter << endl;
   
   *LogFile << "Center Frequency [GHz] = " << CenterFreq*1e-9  << endl;
   *LogFile << "Channel Spacing [GHz] = " << ChannSpacing*1e-9  << endl;
   *LogFile << "Linear Extinction Ratio = " << LinearExtinctionRatio << endl;
   *LogFile << "RandomPolarizationsFlag  = " <<RandomPolarizationsFlag << endl;
   *LogFile << "UsingOptFiberLocalError = " << UsingOptFiberLocalError << endl;

   } // end if


  MemAllocated = 0;

   if ( ! MemAllocated) {
      MemAllocated = 1;
           
      
      Frequency = new double[NumChannels];
      InStokes = new double[NumChannels*4];
      InJones = new cplx[NumChannels*2];

      fft = new cfftw(qtPoints, TimeWindow);

      sfftM.tOptSignalX = NewCplx(qtPoints);
      sfftM.fOptSignalX = NewCplx(qtPoints);
      sfftM.tOptSignalY = NewCplx(qtPoints);
      sfftM.fOptSignalY = NewCplx(qtPoints);
 
  
      AllocateOptFiberLocalErrorWorkSpace();
   }   

 MemAllocated = 1;


      if(UsingOptFiberLocalError && TypeSimulation == VECTOR)
	{
          FreqForPDVectorOFLE = Signal1.FreqForPDVectorOFLE;
          ComputeInputPDVectorFlagOFLE = Signal1.ComputeInputPDVectorFlagOFLE;

          for(int ii=0;ii<3;ii++)
	    OutputPDVectorOFLE[ii] = Signal1.OutputPDVectorOFLE[ii];

          if(ComputeInputPDVectorFlagOFLE)
	   {
            for(int ii=0;ii<3;ii++)
	     InputPDVectorOFLE[ii] = Signal1.InputPDVectorOFLE[ii];

            for(int ii=0;ii<3;ii++)
             for(int jj=0;jj<3;jj++)
	      StokesTransferMatrixTransposeOFLE[ii][jj] = 
		Signal1.StokesTransferMatrixTransposeOFLE[ii][jj];

	   } // end if ComputeInputPDVectorFlagOFLE
	} // end if UsingOptFiberLocalError

 int ch,mm;


 for(ch=0;ch<Signal1.NumChannels;ch++)
   {
     Frequency[ch] = Signal1.Frequency[ch];
   }

 for(ch=Signal1.NumChannels;ch<NumChannels;ch++)
  {
   Frequency[ch] = Signal2.Frequency[ch-Signal1.NumChannels];
  }


 for(ch=0;ch<Signal1.NumChannels;ch++)
   {
     for(mm=0;mm<4;mm++)
      SetInStokes(ch,mm,Signal1.GetInStokes(ch,mm));

     for(mm=0;mm<2;mm++)
       SetInJones(ch,mm,Signal1.GetInJones(ch,mm));

   }


 for(ch=Signal1.NumChannels;ch<NumChannels;ch++)
  {
     for(mm=0;mm<4;mm++)
      SetInStokes(ch,mm,Signal2.GetInStokes(ch-Signal1.NumChannels,mm));

     for(mm=0;mm<2;mm++)
       SetInJones(ch,mm,Signal2.GetInJones(ch-Signal1.NumChannels,mm));

  }


 if(DebugLevel >= 3)
   {
     for(ch=0;ch<NumChannels;ch++)
      *LogFile << "SumSignal (Channel,Freq) = " << ch << " " 
          << Frequency[ch] << endl;

     for(ch=0;ch<NumChannels;ch++)
      for(mm=0;mm<4;mm++)
       *LogFile << "SumSignal (Channel,mm,Stokes) = " << ch << " "
               << mm << " " 
               << GetInStokes(ch,mm) << endl;

     for(ch=0;ch<NumChannels;ch++)
      for(mm=0;mm<2;mm++)
       *LogFile << "SumSignal (Channel,mm,Jones) = " << ch << " " 
               << mm << " "
               << GetInJones(ch,mm) << endl;

   } // end if Debug



//## Speed up the data transfer (Ivan Lima, 7/24/01)
for(int jj=0;jj<qtPoints;jj++) {
   sfftM.tOptSignalX[jj] = Signal1.sfftM.tOptSignalX[jj] +
                           Signal2.sfftM.tOptSignalX[jj];
   sfftM.tOptSignalY[jj] = Signal1.sfftM.tOptSignalY[jj] +
                           Signal2.sfftM.tOptSignalY[jj];
   sfftM.fOptSignalX[jj] = Signal1.sfftM.fOptSignalX[jj] +
                           Signal2.sfftM.fOptSignalX[jj];
   sfftM.fOptSignalY[jj] = Signal1.sfftM.fOptSignalY[jj] +
                           Signal2.sfftM.fOptSignalY[jj];
}

//for(jj=0;jj<qtPoints;jj++)
//  {
//    sfftM.tOptSignalY[jj] = Signal1.sfftM.tOptSignalY[jj] +
//                            Signal2.sfftM.tOptSignalY[jj];
//  }
//
//for(jj=0;jj<qtPoints;jj++)
//  {
//    sfftM.fOptSignalX[jj] = Signal1.sfftM.fOptSignalX[jj] +
//                            Signal2.sfftM.fOptSignalX[jj];
//  }
//
//for(jj=0;jj<qtPoints;jj++)
//  {
//    sfftM.fOptSignalY[jj] = Signal1.sfftM.fOptSignalY[jj] +
//                            Signal2.sfftM.fOptSignalY[jj];
//  }


} // ## end SumSignal Constructor

// #######################################################

void  AddSignals(OptSignal * Signal1,OptSignal * Signal2,
                 OptSignal * Signal3)
{

  // Written by JZ May 4th 2001

  // This method assumes that 
  //OptSignal::OptSignal(OptSignal & Signal1,OptSignal & Signal2,
  //                   RanNumGen *RNG3)
  // has been previously called

  // I think this method is really a friend of the OptSignal class ?

  if(Signal3->NumChannels != Signal1->NumChannels + Signal2->NumChannels)
    {
      cerr << "OptSignal::AddSignals: Incompatable parameters. ABORT!"
	   << endl;
      exit(1);
    }


  AddBitStrings(Signal1->oBitString,Signal2->oBitString,Signal3->oBitString);

  if(Signal3->DebugLevel >=3)
    {
     Signal1->oBitString->WriteBitString();
     Signal2->oBitString->WriteBitString();
     Signal3->oBitString->WriteBitString();
    }

int jj;

for(jj=0;jj<Signal3->qtPoints;jj++)
  {
    Signal3->sfftM.tOptSignalX[jj] = Signal1->sfftM.tOptSignalX[jj] +
                            Signal2->sfftM.tOptSignalX[jj];
  }

for(jj=0;jj<Signal3->qtPoints;jj++)
  {
    Signal3->sfftM.tOptSignalY[jj] = Signal1->sfftM.tOptSignalY[jj] +
                            Signal2->sfftM.tOptSignalY[jj];
  }

for(jj=0;jj<Signal3->qtPoints;jj++)
  {
    Signal3->sfftM.fOptSignalX[jj] = Signal1->sfftM.fOptSignalX[jj] +
                            Signal2->sfftM.fOptSignalX[jj];
  }

for(jj=0;jj<Signal3->qtPoints;jj++)
  {
    Signal3->sfftM.fOptSignalY[jj] = Signal1->sfftM.fOptSignalY[jj] +
                            Signal2->sfftM.fOptSignalY[jj];
  }


} // ## end AddSignals

// ###############################################################

//! Adds a ReduOptSignal containing noise and no signal
//! to the current OptSignal which is assumed to be noise-free.
//! By John Zweck, Jan 2006

void OptSignal::AddReducedNoiseToNoiseFreeSignal(ReduOptSignal *ReduOptSignal1)
{

 if( NumChannels != ReduOptSignal1->GetNumChannels()) 
   {
       cerr << "Invalid use of OptSignal::AddReducedNoiseToNoiseFreeSignal" 
	    << endl
            << "Aborting." << endl;
       exit(1);

   }

  for(int ch=0;ch<NumChannels;ch++)
   for(int ss = 0; ss < 4; ss++)
    oReduOptSignal->StokesNoise[ss][ch] = ReduOptSignal1->StokesNoise[ss][ch]; 
  
  SetAccumulatedTotalNoiseSpectralDensityGaussianStatistics(
             oReduOptSignal->GetTotalNoiseSpectralDensity());
 
}

// ###############################################################

void OptSignal::ComputeRelevantNumModesOFLE(void)
{
  // By JZ, May 8, 2002, Based on RH code

  // Note: We could improve this method for NumChannels > 1!!

  // If have one channel which is not at center frequency
  // this may not be a good choice!! 

  RelevantNumModesOFLE = qtPoints;

  if(ComputeRelevantNumModesOFLEFlag)
    {
      if(NumChannels == 1)
	{
	  double SignalBandwidthNorm = 0.6*BitRateChann; 
           // Ron claims
	  // that this is good for many pulse formats

          RelevantNumModesOFLE =
              int(min(int(ceil(4.0*SignalBandwidthNorm/DeltaFreq)), qtPoints));

          // Make sure the mode number is even
          if(2*(RelevantNumModesOFLE/2) != RelevantNumModesOFLE)
            RelevantNumModesOFLE++;

	  // For ODE and MC Ron adds 6 for safety.

	}
      
    }

  if(DebugLevel)
  *LogFile << "RelevantNumModesOFLE = " << RelevantNumModesOFLE << endl;

} // ## end ComputeRelevantNumModesOFLE

// ###########################################################

void OptSignal::SetBandwidthForRelevantNumModesOFLE(double BandwidthGHz)
{
   
  RelevantNumModesOFLE = int(min(int(ceil(BandwidthGHz/DeltaFreq)),qtPoints));


 // Make sure the mode number is even

 if(2*(RelevantNumModesOFLE/2) != RelevantNumModesOFLE)
    RelevantNumModesOFLE++;

 
 if(DebugLevel)
  *LogFile << "RelevantNumModesOFLE = " << RelevantNumModesOFLE << endl;

}

// ##############################################################

// ##### Gets a new signal string
// ## By: Ivan Lima

void OptSignal::GetNewRandomSignalString()
{
   oBitString->GetRandomString();
   RegenerateSignalString();
}

// ###########################################################

void OptSignal::ResetJonesMatrix()
{	
	for (int ii=0;ii<NumChannels;ii++)
	{
	 JonesMatrixChannel[ii*4+0]=1.0;
	 JonesMatrixChannel[ii*4+1]=0.0;
	 JonesMatrixChannel[ii*4+2]=0.0;
	 JonesMatrixChannel[ii*4+3]=1.0;
	}
}	// added by Hai at 7-1-03

// ###########################################################

void OptSignal::RegenerateSignalString()
{

  AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 0.0;
  oReduOptSignal->RegenerateReduOptSignal();

  // Tested TypeSimulation changes made in all methods called here
  // Oct 18 2002, JZ

  
   SetRandomJonesVectors();
   // JZ FIX THIS 4/11/03 
   // ResetPDVectorOFLE();

   ResetPDVectorOFLE();
   ResetJonesMatrix();              //by Hai  7-1-03
  
   switch(TypePulse)
    {
       case NRZ: case RZ_GAUSS: case RZ_SECH: case CRZ_GAUSS:
	GenerateSignalStringGaussianOrSech();
       break;

       case CRZ_RCOS: case CRZ_SIN_MZ: case CW:
        GenerateSignalStringRaisedCosinesWithPhaseShift(0);
       break;

       case RZ_GAUSS_PATTERN:
        GenerateSignalStringRZGaussianPattern();
       break;
	  
       case INPUT_SEQUENCE:
	GenerateSignalStringUsingInputSequence();
       break;	  
	  
       case UMBC_RZ:
	// For now must call code from Mach Zehnder class 
	// when write your appln.
        break;

       case QPSK_CW:  
        GenerateSignalStringQPSK_CW();
       break;

       case QPSK_RZ_GAUSS:
        GenerateSignalStringQPSK_Gaussian();
       break;

       case BPSK_MACH_ZEHNDER:
         GenerateSignalStringBPSKMachZehnderModulator();
       break;

       case BPSK_PHASE_MOD:
         GenerateSignalStringBPSKPhaseModulator();
       break;

       case CSRZ_MACH_ZEHNDER:
	 GenerateSignalStringCSRZMachZehnderModulator();
       break;

       case RZ33_MACH_ZEHNDER:
         GenerateSignalStringRZ33MachZehnderModulator();
       break;

       case RZ50_MACH_ZEHNDER:
         GenerateSignalStringRZ50MachZehnderModulator();
       break;

       case INPUT_SIGNAL_FROM_FILE:
         InputTimeDataFromFile();
       break;

       case QPSK_SEQ_PM_NRZ:
         GenerateSignalStringQPSK_SEQ_PM_NRZ();
       break;

       case QPSK_DPMZ_PM_NRZ:
         GenerateSignalStringQPSK_DPMZ_PM_NRZ();
       break;

       case QPSK_SEQ_PM_AM_RZ33:
         GenerateSignalStringQPSK_SEQ_PM_AM_RZ33();
       break;

       case QPSK_SEQ_PM_AM_RZ50:
         GenerateSignalStringQPSK_SEQ_PM_AM_RZ50();
       break;

       case QPSK_DPMZ_PM_AM_RZ33:
         GenerateSignalStringQPSK_DPMZ_PM_AM_RZ33();
       break;

       case QPSK_DPMZ_PM_AM_RZ50:
         GenerateSignalStringQPSK_DPMZ_PM_AM_RZ50();
       break;

       case NRZ_RCOS_MACH_ZEHNDER:
         GenerateSignalStringNRZ_RCOS_MachZehnderModulator();
       break;

     default:
      cerr << "OptSignal::RegenerateSignalString:"
		     << endl
		     << "Invalid TypePulse "
		     << "ABORTING NOW!!"
		     << endl;

        exit(1);


    } // ## end switch(TypePulse)

} // ## end RegenerateSignalString

// ###################################################

void OptSignal::ConstantPhaseShift(double PhaseShiftAmount)
{

  // This phase shift must be frequency and time independent!

  int jj;

for(jj=0;jj<qtPoints;jj++)
  {
    sfftM.tOptSignalX[jj] = exp(2*pi*jc*PhaseShiftAmount/360)*
                            sfftM.tOptSignalX[jj];
                            
  }


if(TypeSimulation == VECTOR)
{
 for(jj=0;jj<qtPoints;jj++)
  {
    sfftM.tOptSignalY[jj] = exp(2*pi*jc*PhaseShiftAmount/360)*
                            sfftM.tOptSignalY[jj];
                            
  }
}

for(jj=0;jj<qtPoints;jj++)
  {
    sfftM.fOptSignalX[jj] = exp(2*pi*jc*PhaseShiftAmount/360)*
                            sfftM.fOptSignalX[jj];
                            
  }

if(TypeSimulation == VECTOR)
{
for(jj=0;jj<qtPoints;jj++)
  {
    sfftM.fOptSignalY[jj] = exp(2*pi*jc*PhaseShiftAmount/360)*
                            sfftM.fOptSignalY[jj];
                            
  }
}

} // end ConstantPhaseShift

// #####################################################

void OptSignal::GenerateSignalStringGaussianOrSech(void)
{

  // Adapted for Physics Convention JZFeb03

  // Written by John Zweck on Feb 19th 2001,
  // by significantly adapting code written by Ivan Lima.

  // Fixed for NRZ with nonzero raise time by Ivan Lima (7/11/02)

  // Fixed for finite extinction ratio by Ivan Lima (9/29/02) 

  // This method should be used to generate pulses which are defined by
  // an analytic formula which decays significantly outside the 
  // bit window eg,  Gaussian, sech, NRZ pulses.

  // However it will not work for Raised cosine type pulses for which
  // a different method is required. 

  // With this way of defining the pulse train it will be as smooth
  // (infinitely differentiable) in the time domain as one could 
  // reasonably hope for, and 
  // consequently have the correct asymptotic decay at large
  // frequencies in the frequency domain.
  // To see how well this works look at a log scale frequency
  // domain plot.

  // FIX: The (Gaussian filtered) NRZ pulse train is going to be 
  // nearly as smooth as we would like. This needs further work.

  // FIX: Also the way the Extinction Ratio is done reduces the 
  // smoothness of the pulse somewhat it seems from looking at 
  // output data. However we might be able to live with this.

  // WARNING: The Extinction Ratio should be used cautiously since it may not
  // give quite the output you expect. This is particularly the case
  // for pulses with a fairly large FWHM relative to the BitLength.

  // The pulses are centered at (kk+0.5)*BitLength for kk = 0 to StringLength

  // The algorithm goes as follows:

  // (1) For each channel we construct a cplx valued SingleChannelWave
  //     which is a smooth function of time. To form the X and Y 
  //     polarizations of the WDM signal we take  linear combinations
  //     of the SingleChannelWaves, one for each channel.
  //     The coefficients are given by the product of 
  //     exp( - jc*ChannelPhaseOffset)
  //     -- which shifts the  central frequency to that of the channel --
  //     and the X or Y  
  //     component of the Jones vector of the channel.

  //  (2) For each channel the SingleChannelWave is computed as 
  //      follows. For each mark ("1") we compute a function 
  //      Envelope(t) exp(i Phase(t) which is defined for all
  //      times t in the time window. Because this function may not
  //      be periodic (and hence infinitely differentiable)
  //     on [0,TimeWindow] we insist that the single point 
  //      where it is not smooth is located exactly half of the 
  //     TimeWindow away from the pulse's center where the amplitude
  //     has effectively gone to zero. 
  //     The Envelope(t) exp(i Phase(t) functions, one for each mark,
  //      are summed together to form the a "proto-SingleChannelWave"
  //    function, which can also be expressed in  the form 
  //     Envelope(t) exp(i Phase(t).
  //    To incorporate the Extinction Ratio we replace this last
  //    Envelope(t) function by sqrt( sq(Envelope(t))*(1-LXR) + LXR )
  //    which will be smooth as LXR > 0 and 1-LXR > 0, which implies
  //    that sq(Envelope(t))*(1-LXR) + LXR > 0 so that it's sqrt is smooth,
  //    or rather as smooth as "proto-SingleChannelWave" is!
  //    The resulting function is SingleChannelWave.



if (! MemAllocated) 
  {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
  }   

    
 double Envelope;
 double Phase;   // For Chirped pulses
 double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system

 cplx SingleChannelWave;

 fftMatrices sfftM_Tmp; 

  if (RiseTime != 0 && TypePulse == NRZ) 
    { 
      sfftM_Tmp.tOptSignalX = NewCplx(qtPoints);
      sfftM_Tmp.fOptSignalX = NewCplx(qtPoints);

      if(TypeSimulation == VECTOR)
	{
         sfftM_Tmp.tOptSignalY = NewCplx(qtPoints);
         sfftM_Tmp.fOptSignalY = NewCplx(qtPoints);
	}        
    }


 for (int ii = 0; ii < qtPoints; ii++) 
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;
    
     // Make Y poln zero even for SCALAR simulations

     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;
 
   } // ## end for ii-loop

 for(int ch=0;ch<NumChannels;ch++)
    {

      for (int ii = 0; ii < qtPoints; ii++)
	{ 
         SingleChannelWave = 0.0;

         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          //## Modified by Ivan Lima (9/4/02) to account for the novel
	  //## extinction ratio implementation
          //#$#$if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")	  
          // {    
	  
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

	 switch(TypePulse)
	   {

	     case NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	       if(TimeRelPulseCenter < BitLength/2.0 && 
                  TimeRelPulseCenter >= -BitLength/2.0)
                        Envelope = 1.0;
                     else
                        Envelope = 0.0;

                     Phase = 0.0;

                     break;

             case RZ_GAUSS:
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

	            Envelope = exp(-sq( TimeRelPulseCenter 
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    Phase = 0.0;
  
                    break;
        
	     case RZ_SECH:  
                     // ## FWHM = 2*log(1+sqrt(2)) * sigma_power  
 
                     Envelope = 1.0/cosh( TimeRelPulseCenter
                              /(FWHM_Pulse/(2*log(1+sqrt(2)) )) );

                     // Phase is for sech pulse solution of Haus's master
                     // equation. See Haus et al., JOSAB, Vol 8, (10), p2068,
                     // 1991.

                     Phase = ChirpParameter*log(Envelope);

                     break;

	   case CRZ_GAUSS:
		     Envelope   = exp( -2*log(2)*
			              sq( TimeRelPulseCenter/FWHM_Pulse ));
                   
		     // The ChirpParameter here is precisely the chirp
		     // parameter, C, in [NLO,p67].

                     Phase = -ChirpParameter*2*log(2)*
			      sq( TimeRelPulseCenter/FWHM_Pulse );
  

	             break;

	
         
              default:
		cerr << "TypePulse Error in " << endl
		     << "OptSignal::GenerateSignalStringGaussianOrSech."
		     << endl;
			 
                    exit(1);

	   } // ## end switch(TypePulse)

         //## Extinction ratio fixed by Ivan Lima (8/29/02)
         if (oBitString->GetBit(ch,kk)) {    
   	    SingleChannelWave += sqrt(OptPowerChann)
	                        *Envelope*exp(jc*Phase);
	 } else { //## A space
   	    SingleChannelWave += sqrt(OptPowerChann*LinearExtinctionRatio)
	                        *Envelope*exp(jc*Phase);
	 }    
   
   } // ## end for kk-loop


      // At this point SingleChannelWave includes the contributions
      // from all the pulses at ii*DeltaTime

      // #### Next add the SingleChannelWave to the WDM Signal:

      // ## Get Jones vector for current channel

      // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

   //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*(ii)*DeltaTime;

    // JZFeb03: Changed and checked for Physics Convention: 

    ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

     //## Fixing for NRZ with nonzero raise time (Ivan Lima)

      if (RiseTime != 0 && TypePulse == NRZ) 
        { 
       
         sfftM_Tmp.tOptSignalX[ii] = GetInJones(ch,0)*SingleChannelWave;

         if(TypeSimulation == VECTOR)                                      
            sfftM_Tmp.tOptSignalY[ii] = GetInJones(ch,1)*SingleChannelWave;
         }	
      else 
        { //## All the other formats (or NRZ without rise time)

	  // JZFeb03: Corect for Physics Convention 

         sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
   	                        exp(- jc*ChannelPhaseOffset);
      
         if(TypeSimulation == VECTOR)                         
            sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	                             exp(- jc*ChannelPhaseOffset);
      }
      
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

      //## Fixing for NRZ with nonzero raise time (Ivan Lima)
     // ## Constraint of the rise time  for NRZ

     if (RiseTime != 0 && TypePulse == NRZ) 
       { 
        fft->FFT(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX); 

        if(TypeSimulation == VECTOR)    
           fft->FFT(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY);  

        optical_filter(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

        if(TypeSimulation == VECTOR)
           optical_filter(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

        for (int ii = 0; ii < qtPoints; ii++) 
	  {

              //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*
	      //                           (ii)*DeltaTime;

             // JZFeb03: Changed and checked for Physics Convention: 

           ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*
	                         (ii)*DeltaTime;


           sfftM.tOptSignalX[ii] += sfftM_Tmp.tOptSignalX[ii]*
   	                        exp(- jc*ChannelPhaseOffset);   

           if(TypeSimulation == VECTOR)
              sfftM.tOptSignalY[ii] += sfftM_Tmp.tOptSignalY[ii]*
   	                               exp(- jc*ChannelPhaseOffset);	
	}
     } // end if NRZ


   } // ## end for ch-loop


   Time2Freq();
 
     if (RiseTime != 0 && TypePulse == NRZ) { 
        FreeCplx(sfftM_Tmp.tOptSignalX);
        FreeCplx(sfftM_Tmp.fOptSignalX);

        if(TypeSimulation == VECTOR)
         {
          FreeCplx(sfftM_Tmp.tOptSignalY);
          FreeCplx(sfftM_Tmp.fOptSignalY);
	 }        

     } // end if RiseTime != 0 && TypePulse == NRZ
  

} // ## end GenerateSignalStringGaussianOrSech()

// #####################################################

void OptSignal::GenerateSignalStringQPSK_Gaussian(void)
{

 double Envelope;
 double Phase;   // For Chirped pulses
 double ChannelPhaseOffset;

 cplx SingleChannelWave;
  
 for (int ii = 0; ii < qtPoints; ii++)
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;

     // Make Y poln zero even for SCALAR simulations

     sfftM.tOptSignalY[ii] = 0.0; 
     sfftM.fOptSignalY[ii] = 0.0;

   } // ## end for ii-loop

 // Assume StringLength is even

for(int ch=0;ch<NumChannels;ch++)
 {
  for(int ii = 0; ii < qtPoints; ii++)
    {
     SingleChannelWave = 0.0;

     for(int kk=0;kk<StringLength/2;kk++)
      {
        double PulseCenter = (2*kk+1)*BitLength;

        double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

        while(TimeRelPulseCenter < -TimeWindow/2)
              TimeRelPulseCenter += TimeWindow;
        while(TimeRelPulseCenter >= TimeWindow/2)
              TimeRelPulseCenter -= TimeWindow;

        Envelope = exp(-sq( TimeRelPulseCenter
                            *sqrt(2*log(2))/FWHM_Pulse ));

        if(oBitString->GetBit(ch,2*kk)==0 && oBitString->GetBit(ch,2*kk+1)==0)
           Phase = 0;
        if(oBitString->GetBit(ch,2*kk)==0 && oBitString->GetBit(ch,2*kk+1)==1)
           Phase=-pi/2.0;
        if(oBitString->GetBit(ch,2*kk)==1 && oBitString->GetBit(ch,2*kk+1)==0)
           Phase = pi/2.0;
        if(oBitString->GetBit(ch,2*kk)==1 && oBitString->GetBit(ch,2*kk+1)==1)
           Phase=pi;

        SingleChannelWave += sqrt(OptPowerChann)*Envelope*exp(jc*Phase);

      } // ## end for kk-loop

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
                           exp(- jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
                                     exp(- jc*ChannelPhaseOffset);

  } // end for ii loop  

} // ## end for ch-loop

Time2Freq();

} // ## end GenerateSignalStringQPSK_Gaussian

// ###############################################################

void OptSignal::GenerateSignalStringQPSK_CW(void)
{

 cplx * Phase;

 Phase = EOM_sfftPM.tPowerSignal;

 double ChannelPhaseOffset;

 cplx SingleChannelWave;

 for (int ii = 0; ii < qtPoints; ii++)
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;

     // Make Y poln zero even for SCALAR simulations

     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;

   } // ## end for ii-loop
 
 // Assume StringLength is even

for(int ch=0;ch<NumChannels;ch++)
{

  // First define piecewise constant Phase function for the channel

 for(int ii = 0; ii < qtPoints; ii++)
  {

      int jj = (int)  (-1 + qtPoints/StringLength
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );

      int kk = (int) (jj/(qtPoints/StringLength));

      if(kk >= StringLength)
         kk -= StringLength;

      // kk is the index of the bit, in range [0,StringLength)
      // We assume StringLength is even

      int FirstIndex, SecondIndex;

      if(kk%2) // kk is odd
       {
        FirstIndex = kk-1;    // FirstIndex is in [0,StringLength-2]
        SecondIndex = kk;     // SecondIndex is in [1,StringLength-1]
       }
      else  // kk is even
       {
        FirstIndex = kk;    // FirstIndex is in [0,StringLength-2]
        SecondIndex = kk+1; // SecondIndex is in [1,StringLength)
       }

      // FirstIndex is always even and SecondIndex is odd

        if(oBitString->GetBit(ch,FirstIndex)==0 && 
           oBitString->GetBit(ch,SecondIndex)==0)
            Phase[ii] = 0;
        if(oBitString->GetBit(ch,FirstIndex)==0 && 
           oBitString->GetBit(ch,SecondIndex)==1)
            Phase[ii]=-pi/2.0;
        if(oBitString->GetBit(ch,FirstIndex)==1 && 
           oBitString->GetBit(ch,SecondIndex)==0)
            Phase[ii] = pi/2.0;
        if(oBitString->GetBit(ch,FirstIndex)==1 && 
           oBitString->GetBit(ch,SecondIndex)==1)
            Phase[ii]=pi;

  } // end piecewise constant Phase function for loop

  // Second filter the Phase function for that channel
   
   EOM_FilterElecData();

  // Third construct the optical signal in the channel assuming 
  // a constant power function.

  for(int ii = 0; ii < qtPoints; ii++)
    {
      SingleChannelWave = sqrt(OptPowerChann)*exp(jc*Phase[ii]);

      ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

      sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
                               exp(- jc*ChannelPhaseOffset);

      if(TypeSimulation == VECTOR)
        sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
                                 exp(- jc*ChannelPhaseOffset);

    } // end for ii loop  

} // ## end for ch-loop

Time2Freq();

} // ## end GenerateSignalStringQPSK_CW

// ######################################################

void OptSignal::GenerateSignalStringNRZ_RCOS_MachZehnderModulator(void)
{

for (int ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop


for(int ch=0;ch<NumChannels;ch++)
 {
   bool ZeroOneNotMinusPlusFlag=0;
   EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);

   EOM_FilterElecData(); // Maybe commented out in conmstructor


  for (int ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  double ChannelPhaseOffset = 
   2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  double Phi_1 = 0.25*pi*(1+ChirpParameter)*real(EOM_sfftPM.tPowerSignal[ii]);

  double Phi_2 = 0.25*pi*(1-ChirpParameter)*real(EOM_sfftPM.tPowerSignal[ii]);

  cplx Envelope = 0.5*(exp(jc*Phi_1) + jc*exp(-jc*Phi_2));


  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop


Time2Freq();

} // ## end GenerateSignalStringNRZ_RCOS_MachZehnderModulator


// #######################################################

void OptSignal::GenerateSignalStringBPSKPhaseModulator(void)
{
 
  EOM_PhaseModulator();

} // ## end GenerateSignalStringBPSK

// ############################################################

void OptSignal::GenerateSignalStringBPSKMachZehnderModulator(void)
{

  double V_pi = -pi;
  double Phi = pi;
  double SinusoidalRate = 0.0;
  double TimeShift = 0.0;
  double VertShift = 0;

  EOM_MachZehnderModulatorElecNRZ(V_pi,V_pi,Phi,Phi,SinusoidalRate,
			   TimeShift,VertShift);

} // ## end GenerateSignalStringBPSK

// ##########################################################

void OptSignal::GenerateSignalStringCSRZMachZehnderModulator(void)
{

GenerateSignalStringNRZ_RCOS_MachZehnderModulator();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.5*pi*sin(pi*ii*DeltaTime*BitRateChann));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ##########################################################

void OptSignal::GenerateSignalStringRZ33MachZehnderModulator(void)
{

GenerateSignalStringNRZ_RCOS_MachZehnderModulator();
 
for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.5*pi*(1+cos(pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ##########################################################

void OptSignal::GenerateSignalStringRZ50MachZehnderModulator(void)
{

GenerateSignalStringNRZ_RCOS_MachZehnderModulator();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.25*pi*(1-cos(2*pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ###########################################################

void OptSignal::GenerateSignalStringCSRZMachZehnderModulatorOLD(void)
{

  double V_pi = pi/2.0;
  double Phi_10 = 0;
  double Phi_20 = pi;
  double SinusoidalRate = 0.5*BitRateChann; 
  double TimeShift = 0.0;
  double VertShift = 0;

  EOM_MachZehnderModulatorElecNRZ(V_pi,V_pi,Phi_10,Phi_20,SinusoidalRate,
                           TimeShift,VertShift);

  for(int ii=0;ii<qtPoints;ii++)
   {
    sfftM.tOptSignalX[ii] *= -jc;
    sfftM.tOptSignalY[ii] *= -jc;
    sfftM.fOptSignalX[ii] *= -jc;
    sfftM.fOptSignalY[ii] *= -jc;  
   }

}
// ##########################################################

void OptSignal::GenerateSignalStringRZ33MachZehnderModulatorOLD(void)
{

  double V_pi = pi/2.0;
  double Phi_10 = 0;
  double Phi_20 = pi;
  double SinusoidalRate = 0.5*BitRateChann;
  double TimeShift = 0.5*pi;
  double VertShift = 1;

  EOM_MachZehnderModulatorElecNRZ(V_pi,V_pi,Phi_10,Phi_20,SinusoidalRate,
                           TimeShift,VertShift);

  for(int ii=0;ii<qtPoints;ii++)
   {
    sfftM.tOptSignalX[ii] *= -jc;
    sfftM.tOptSignalY[ii] *= -jc;
    sfftM.fOptSignalX[ii] *= -jc;
    sfftM.fOptSignalY[ii] *= -jc;
   }


}

// ##########################################################

void OptSignal::GenerateSignalStringRZ50MachZehnderModulatorOLD(void)
{

  double V_pi = pi/4.0;
  double Phi_10 = 0;
  double Phi_20 = pi;
  double SinusoidalRate = BitRateChann;
  double SinusoidalTimeShift = pi/2.0;
  double SinusoidalVertShift = 1.0;

  EOM_MachZehnderModulatorElecNRZ(V_pi,V_pi,Phi_10,Phi_20,
			   SinusoidalRate,SinusoidalTimeShift, 
			   SinusoidalVertShift);

  for(int ii=0;ii<qtPoints;ii++)
   {
    sfftM.tOptSignalX[ii] *= -jc;
    sfftM.tOptSignalY[ii] *= -jc;
    sfftM.fOptSignalX[ii] *= -jc;
    sfftM.fOptSignalY[ii] *= -jc;
   }

}

// ############################################################

void OptSignal::EOM_WriteFileElecData(string OutFileName)
{

if(TypePulse == BPSK_MACH_ZEHNDER || TypePulse == CSRZ_MACH_ZEHNDER ||
   TypePulse == RZ33_MACH_ZEHNDER || TypePulse == RZ50_MACH_ZEHNDER ||
   TypePulse == BPSK_PHASE_MOD ||
   TypePulse == QPSK_SEQ_PM_NRZ || TypePulse == QPSK_DPMZ_PM_NRZ ||
   TypePulse == QPSK_SEQ_PM_AM_RZ33 || TypePulse ==  QPSK_SEQ_PM_AM_RZ50 ||
   TypePulse == QPSK_DPMZ_PM_AM_RZ33 || TypePulse == QPSK_DPMZ_PM_AM_RZ50 ||
   TypePulse == NRZ_RCOS_MACH_ZEHNDER)
{

ofstream OutFile(OutFileName.c_str(),ios::out);

switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(0);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(0);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(0);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag = 1;
       EOM_GenerateRaisedCosineBinaryElecData(0,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

for(int ii=0;ii<qtPoints;ii++)
 OutFile << ii*DeltaTime << "\t" << real(EOM_sfftPM.tPowerSignal[ii]) << endl;

OutFile.close();

}

}

// ##########################################################

void OptSignal::SetEOM_V_pi_RelError(double EOM_V_pi_RelError2)
       {EOM_V_pi_RelError = EOM_V_pi_RelError2;}

void OptSignal::SetEOM_Phi_10_RelError(double EOM_Phi_10_RelError2)
       {EOM_Phi_10_RelError = EOM_Phi_10_RelError2;}

void OptSignal::SetEOM_Phi_20_RelError(double EOM_Phi_20_RelError2)
       {EOM_Phi_20_RelError = EOM_Phi_20_RelError2;}

void OptSignal::SetEOM_SinusoidalVertShift_RelError(
        double EOM_SinusoidalVertShift_RelError2)
     {EOM_SinusoidalVertShift_RelError = EOM_SinusoidalVertShift_RelError2;}


// ###########################################################

void OptSignal::EOM_MachZehnderModulatorElecNRZ(double V_pi1, double V_pi2,
                                         double Phi_10, double Phi_20,
					 double SinusoidalRate,
					 double SinusoidalTimeShift,
					 double SinusoidalVertShift)
{

V_pi1 = V_pi1*(1 + EOM_V_pi_RelError);
V_pi2 = V_pi2*(1 + EOM_V_pi_RelError);
Phi_10 = Phi_10*(1 + EOM_Phi_10_RelError);
Phi_20 = Phi_20*(1 + EOM_Phi_20_RelError);
SinusoidalVertShift = SinusoidalVertShift*
                       (1 + EOM_SinusoidalVertShift_RelError);

int ii;

double ChannelPhaseOffset;
double Phi_1,Phi_2;
cplx Envelope;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop


for(int ch=0;ch<NumChannels;ch++)
 {
   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

   if(SinusoidalRate > 0)
    EOM_SinusoidallyModulateElecData(SinusoidalRate,SinusoidalTimeShift,
                                     SinusoidalVertShift);

  for (ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  Phi_1 = V_pi1*real(EOM_sfftPM.tPowerSignal[ii]) + Phi_10;

  Phi_2 = - V_pi2*real(EOM_sfftPM.tPowerSignal[ii]) - Phi_20;

  Envelope = 0.5*(exp(jc*Phi_1) + exp(jc*Phi_2));


  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop


Time2Freq();

} // ## end EOM_MachZehnderModulatorElecNRZ 

// ##############################################################

void OptSignal::EOM_MachZehnderModulatorOptData(double V_pi1, double V_pi2,
                                         double Phi_10, double Phi_20,
                                         double SinusoidalRate,
                                         double SinusoidalThetaShift)
{

for (int ii = 0; ii < qtPoints; ii++)
{

  double Phi_1 = 
          V_pi1*sin(2*pi*SinusoidalRate*(ii)*DeltaTime -SinusoidalThetaShift)
          + Phi_10;

  double Phi_2 = 
          V_pi2*sin(2*pi*SinusoidalRate*(ii)*DeltaTime -SinusoidalThetaShift)
          + Phi_20;

  cplx Envelope = 0.5*(exp(jc*Phi_1) + exp(-jc*Phi_2));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

} // ## end EOM_MachZehnderModulatorOptNRZ

// ###############################################################

void OptSignal::EOM_MachZehnderModulatorOptNRZOLD(double V_pi1, double V_pi2,
                                         double Phi_10, double Phi_20,
                                         double SinusoidalRate,
                                         double SinusoidalThetaShift)
{

int ii;

double ChannelPhaseOffset;
double Phi_1,Phi_2;
cplx Envelope;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop


for(int ch=0;ch<NumChannels;ch++)
 {
   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

  for (ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  Phi_1 = V_pi1*sin(2*pi*SinusoidalRate*(ii)*DeltaTime -SinusoidalThetaShift) 
          + Phi_10;

  Phi_2 = V_pi2*sin(2*pi*SinusoidalRate*(ii)*DeltaTime -SinusoidalThetaShift)
          + Phi_20;

  Envelope = 0.5*(exp(jc*Phi_1) + exp(-jc*Phi_2));


  sfftM.tOptSignalX[ii] += real(EOM_sfftPM.tPowerSignal[ii])*
                           GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += real(EOM_sfftPM.tPowerSignal[ii])*
                              GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop


Time2Freq();

} // ## end EOM_MachZehnderModulatorOptNRZOLD

// #############################################################


void OptSignal::EOM_SinusoidallyModulateElecData(double SinusoidalRate,
						 double SinusoidalTimeShift,
                                                 double SinusoidalVertShift)
{

  for (int ii = 0; ii < qtPoints; ii++)
   EOM_sfftPM.tPowerSignal[ii] *= 
    SinusoidalVertShift + 
    sin(2*pi*SinusoidalRate*ii*DeltaTime - SinusoidalTimeShift);

 fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);

}

// ##################################################################
    
void OptSignal::EOM_PhaseModulator(void)
{

int ii;

double ChannelPhaseOffset;
double Phi;
cplx Envelope;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop

for(int ch=0;ch<NumChannels;ch++)
 {

   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

  for (ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  // Gives a 0-bit having phase of pi and 1-bit phase of 0

  Phi = pi*(1-real(EOM_sfftPM.tPowerSignal[ii]));

  Envelope = exp(jc*Phi);

  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop

Time2Freq();

} // ## end EOM_PhaseModulator

// ######################################################

void OptSignal::GenerateSignalStringQPSK_SEQ_PM_NRZ(void)
{

int ii;

double ChannelPhaseOffset;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop

for(int ch=0;ch<NumChannels;ch++)
 {

   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,1);
       break;

     } // end switch

   EOM_FilterElecData();

   for(ii=0; ii < qtPoints; ii++)
    EOM_sfftPM1.tPowerSignal[ii] = EOM_sfftPM.tPowerSignal[ii]; 

   oBitString->GetRandomString(); // silly but needed for WDM 

   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

   for(ii=0; ii < qtPoints; ii++)
    EOM_sfftPM2.tPowerSignal[ii] = EOM_sfftPM.tPowerSignal[ii];

  for (ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  // Gives a 0-bit having phase of pi and 1-bit phase of 0

  double Phi1 = pi*(1-real(EOM_sfftPM1.tPowerSignal[ii]));

  double Phi2 = 0.5*pi*(1-real(EOM_sfftPM2.tPowerSignal[ii]));

  cplx Envelope = exp(jc*(Phi1+Phi2));

  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop

Time2Freq();

} // ## end GenerateSignalQPSK_SEQ_PM_NRZ

// #############################################################

void OptSignal::GenerateSignalStringQPSK_DPMZ_PM_NRZ(void)
{

int ii;

double ChannelPhaseOffset;

for (ii = 0; ii < qtPoints; ii++)
 {
    sfftM.tOptSignalX[ii] = 0.0;
    sfftM.tOptSignalY[ii] = 0.0;
    sfftM.fOptSignalX[ii] = 0.0;
    sfftM.fOptSignalY[ii] = 0.0;

 } // ## end for ii-loop

for(int ch=0;ch<NumChannels;ch++)
 {

   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
     } // end switch

   EOM_FilterElecData();

   for(ii=0; ii < qtPoints; ii++)
    EOM_sfftPM1.tPowerSignal[ii] = EOM_sfftPM.tPowerSignal[ii];

   oBitString->GetRandomString(); // silly but needed for WDM 

   switch (TypeEOMElecData)
     {
     case EOM_BINARY_DATA_PERFECT_RECT:
       EOM_GeneratePerfectRectBinaryElecData(ch);
       break;

     case EOM_BINARY_DATA_ARCTANH_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataArctanh(ch);
       break;

     case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(ch);
       break;

     case EOM_BINARY_DATA_RCOS:
       bool ZeroOneNotMinusPlusFlag=1;
       EOM_GenerateRaisedCosineBinaryElecData(ch,ZeroOneNotMinusPlusFlag);
       break;

     } // end switch

   EOM_FilterElecData();

   for(ii=0; ii < qtPoints; ii++)
    EOM_sfftPM2.tPowerSignal[ii] = EOM_sfftPM.tPowerSignal[ii];

  for (ii = 0; ii < qtPoints; ii++)
   {

  // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

  ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

  double Envelope1 = cos(pi*(1-real(EOM_sfftPM1.tPowerSignal[ii])));

  double Envelope2 = cos(pi*(1-real(EOM_sfftPM2.tPowerSignal[ii])));

  cplx Envelope = (Envelope1 + jc*Envelope2)/sqrt(2);

  sfftM.tOptSignalX[ii] += GetInJones(ch,0)*sqrt(OptPowerChann)*
                           Envelope*exp(-jc*ChannelPhaseOffset);

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] += GetInJones(ch,1)*sqrt(OptPowerChann)*
                              Envelope*exp(-jc*ChannelPhaseOffset);

   } // ## end for ii-loop
   } // ## end for ch-loop

Time2Freq();

} // ## end GenerateSignalQPSK_DPMZ_PM_NRZ

// ######################################################################

void OptSignal::GenerateSignalStringQPSK_SEQ_PM_AM_RZ33(void)
{

GenerateSignalStringQPSK_SEQ_PM_NRZ();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.5*pi*(1+cos(pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ####################################

void OptSignal::GenerateSignalStringQPSK_SEQ_PM_AM_RZ50(void)
{

GenerateSignalStringQPSK_SEQ_PM_NRZ();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.25*pi*(1-cos(2*pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ########################################################

void OptSignal::GenerateSignalStringQPSK_DPMZ_PM_AM_RZ33(void)
{

GenerateSignalStringQPSK_DPMZ_PM_NRZ();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.5*pi*(1+cos(pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ####################################

void OptSignal::GenerateSignalStringQPSK_DPMZ_PM_AM_RZ50(void)
{

GenerateSignalStringQPSK_DPMZ_PM_NRZ();

for (int ii = 0; ii < qtPoints; ii++)
{

  double Envelope = sin(0.25*pi*(1-cos(2*pi*ii*DeltaTime*BitRateChann)));

  sfftM.tOptSignalX[ii] *= Envelope;

  if(TypeSimulation == VECTOR)
     sfftM.tOptSignalY[ii] *= Envelope;

} // ## end for ii-loop

Time2Freq();

}

// ########################################################

string OptSignal::EOM_GetTypeElecData(void)
{

string Type;

switch(TypeEOMElecData)
{

 case EOM_BINARY_DATA_PERFECT_RECT: 
       Type = "EOM_BINARY_DATA_PERFECT_RECT";
       break;
 case EOM_BINARY_DATA_ARCTANH_SMOOTHING: 
       Type = "EOM_BINARY_DATA_ARCTANH_SMOOTHING";
       break;
 case EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING:
       Type = "EOM_BINARY_DATA_SUPERGAUSS_SMOOTHING";
       break;
     case EOM_BINARY_DATA_RCOS:
       Type = "EOM_BINARY_DATA_RCOS";
       break;
 default:
       Type = "Invalid TypeEOMElecData";
       cerr << "OptSignal::EOM_GetTypeElecData error. Aborting." << endl;
       exit(1);

}

return Type;

}

// #########################################################

string OptSignal::EOM_GetTypeElecFilter(void)
{

string Type;

switch(TypeEOMElecFilter)
{

 case EOM_NO_ELEC_FILTER:
       Type = "EOM_NO_ELEC_FILTER";
       break;
 case EOM_GAUSSIAN:
       Type = "EOM_GAUSSIAN";
       break;
 case EOM_BESSEL:
       Type = "EOM_BESSEL";
       break;
 default:
       Type = "Invalid TypeEOMElecFilter";
       cerr << "OptSignal::EOM_GetTypeElecFilter error. Aborting." << endl;
       exit(1);

}

return Type;

}

// ########################################################

void OptSignal::EOM_GenerateRaisedCosineBinaryElecData(
      int ChannelArrayIndex, bool ZeroOneNotMinusPlusDataFlag)
{
   
   int SymbolZero, SymbolOne;

   if(ZeroOneNotMinusPlusDataFlag)
    {
     SymbolZero = 0;
     SymbolOne  = 1;
    } 
   else
    {
     SymbolZero = -1;
     SymbolOne =  1;
    }

   for (int ii = 0; ii < qtPoints; ii++) 
      {
        EOM_sfftPM.tPowerSignal[ii] = 0.0;
        EOM_sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop

   for (int ii = 0; ii < qtPoints; ii++) 
   {

      double Time = ii*DeltaTime;

      int kk_min = int(ceil(-1.5 + Time/BitLength));

      int kk_max = kk_min+1;

      double PulseValue_kk_min = 
             RaisedCosineForNRZ_RCOS_MZ(ii*DeltaTime - (kk_min+0.5)*BitLength);

      double PulseValue_kk_max = 
             RaisedCosineForNRZ_RCOS_MZ(ii*DeltaTime - (kk_max+0.5)*BitLength);

      if(kk_min < 0)
        kk_min += StringLength;

      if(kk_max >= StringLength)
        kk_max -= StringLength;

      int BitValue_kk_min, BitValue_kk_max;

      if(oBitString->GetBit(ChannelArrayIndex,kk_min))
         BitValue_kk_min = SymbolOne;
      else
         BitValue_kk_min = SymbolZero;

      if(oBitString->GetBit(ChannelArrayIndex,kk_max))
         BitValue_kk_max = SymbolOne;
      else
         BitValue_kk_max = SymbolZero;

     EOM_sfftPM.tPowerSignal[ii] = 
      BitValue_kk_min*PulseValue_kk_min + BitValue_kk_max*PulseValue_kk_max;

   } // ## end for ii-loop

 fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);

}

// ####################################################

double OptSignal::RaisedCosineForNRZ_RCOS_MZ(double Time)
{

double Beta = EOM_SlopeFactor/BitLength;

double Value;

if(Beta > 0)
{
 if(abs(Time) <= (1-Beta)*0.5*BitLength)
  Value = 1;
 else
  if(abs(Time) >= (1+Beta)*0.5*BitLength)
   Value=0;
  else
   Value = 0.5*(1-sin(pi*(abs(Time) - 0.5*BitLength)/EOM_SlopeFactor));
}
else
{

if(Time >= -0.5*BitLength && Time < 0.5*BitLength)
 Value=1;
else
 Value=0;

}

return Value;

}

// ###########################################################

void OptSignal::EOM_GeneratePerfectRectBinaryElecData(int ChannelArrayIndex)
{

   int ii,jj,kk;
   double Value;

    for (ii = 0; ii < qtPoints; ii++)
      {
        EOM_sfftPM.tPowerSignal[ii] = 0.0;
        EOM_sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop

   for (ii = 0; ii < qtPoints; ii++)
   {

     // Comments by JZ (I think this is correct):
     // (1) Given an index ii we want to know which 
     // bit slot kk the index ii is in, and also
     // what the index jj of the center of this kk-th bit slot is.
     // The reason jj is the index of the center of the bit slot
     // is that 
     // jj = (int) (-1 + (qtPoints/StringLength)*(SomeInteger + 0.5) 
     //
     // and qtPoints/StringLength = Number of points in a bit slot
     // (I'm assuming this ratio is an integer for simplicity).


     // (2) The kk-th bit slot goes from 
     //  ii = kk*M to kk*M + M-1
     // where M =  qtPoints/StringLength is as above.


      jj = (int)  (-1 + qtPoints/StringLength
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );
      kk = (int) (jj/(qtPoints/StringLength));

      // ## It can happen that kk == StringLength in which case 
      // ## it should be set to kk = 0 to be in range for the
      // ## if (oBitString.GetBit(ch,kk)) test (JZ)

      if(kk >= StringLength)
        kk -= StringLength;

      if(oBitString->GetBit(ChannelArrayIndex,kk))
         Value = 1.0;
      else
         Value = 0.0;

     EOM_sfftPM.tPowerSignal[ii] = Value;

   } // ## end for ii-loop

 fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);

} // ## end EOM_GeneratePerfectRectBinaryElecData

// ############################################################

void OptSignal::EOM_GenerateSmoothedRectBinaryElecDataArctanh(
                int ChannelArrayIndex)
{

   int ii,jj,kk;
   int kkm1,kkp1;
   double Value;

    for (ii = 0; ii < qtPoints; ii++)
      {
        EOM_sfftPM.tPowerSignal[ii] = 0.0;
        EOM_sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop


   for (ii = 0; ii < qtPoints; ii++)
   {

     // Comments by JZ (I think this is correct):
     // (1) Given an index ii we want to know which 
     // bit slot kk the index ii is in, and also
     // what the index jj of the center of this kk-th bit slot is.
     // The reason jj is the index of the center of the bit slot
     // is that 
     // jj = (int) (-1 + (qtPoints/StringLength)*(SomeInteger + 0.5) 
     //
     // and qtPoints/StringLength = Number of points in a bit slot
     // (I'm assuming this ratio is an integer for simplicity).


     // (2) The kk-th bit slot goes from 
     //  ii = kk*M to kk*M + M-1
     // where M =  qtPoints/StringLength is as above.


      jj = (int)  (-1 + qtPoints/StringLength
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );
      kk = (int) (jj/(qtPoints/StringLength));

      // ## It can happen that kk == StringLength in which case 
      // ## it should be set to kk = 0 to be in range for the
      // ## if (oBitString.GetBit(ch,kk)) test (JZ)

      if(kk >= StringLength)
        kk -= StringLength;


      if(ii <= jj)
        {
          if(kk==0)
            kkm1 = StringLength-1;
          else
            kkm1 = kk-1;

          if(oBitString->GetBit(ChannelArrayIndex,kk) &&
             oBitString->GetBit(ChannelArrayIndex,kkm1)) // 1,1
               Value = 1.0;
          else
            {
             if(!oBitString->GetBit(ChannelArrayIndex,kk)&&
                !oBitString->GetBit(ChannelArrayIndex,kkm1)) // 0,0
                 Value = 0.0;
             else
             {
               if(oBitString->GetBit(ChannelArrayIndex,kk)&&
                  !oBitString->GetBit(ChannelArrayIndex,kkm1)) // 0,1
                {
                  Value =
                       0.5*
                (1+tanh(EOM_SlopeFactor*
                        (ii*DeltaTime - (kk+EOM_ShiftAmt)*BitLength)));

                } // # end 1,0 if 
               else // 1,0
                {
                  Value = 1 -
                  0.5*(1+ tanh(EOM_SlopeFactor*
                           (ii*DeltaTime - (kk-EOM_ShiftAmt)*BitLength)));


                }
             } // end else
            }//end else

        }// end if ii<=jj

        if(ii > jj)
         {
          if(kk==StringLength-1)
            kkp1 = 0;
          else
            kkp1 = kk+1;

          if(oBitString->GetBit(ChannelArrayIndex,kk) &&
             oBitString->GetBit(ChannelArrayIndex,kkp1)) // 1,1
               Value = 1.0;
          else
            {
             if(!oBitString->GetBit(ChannelArrayIndex,kk)&&
                !oBitString->GetBit(ChannelArrayIndex,kkp1)) // 0,0
                 Value = 0.0;
             else
             {
               if(oBitString->GetBit(ChannelArrayIndex,kk)&&
                  !oBitString->GetBit(ChannelArrayIndex,kkp1)) // 1,0
                {

                 Value =
                          1-0.5*(1+tanh(EOM_SlopeFactor*
                                (ii*DeltaTime - (kk+1-EOM_ShiftAmt)*BitLength)));

                } // # end 1,0 if 
               else // 0,1
                {
                 Value =
                        0.5*(1+tanh(EOM_SlopeFactor*
                             (ii*DeltaTime - (kk+1+EOM_ShiftAmt)*BitLength)));
                }
             } // end else
            }//end else

        }// end if ii>jj   


      EOM_sfftPM.tPowerSignal[ii] = Value;

   } // ## end for ii-loop

fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);

} // ## end EOM_GenerateSmoothedRectBinaryElecDataArctanh

// #############################################################

void OptSignal::EOM_GenerateSmoothedRectBinaryElecDataSuperGauss(
                int ChannelArrayIndex)
{

 for(int ii = 0; ii < qtPoints; ii++)
      {
        EOM_sfftPM.tPowerSignal[ii] = 0.0;
        EOM_sfftPM.fPowerSignal[ii] = 0.0;

      } // ## end for ii-loop

 for(int ii = 0; ii < qtPoints; ii++)
  {

    double Value;

    double Time = ii*DeltaTime;
    int BitSlot = int(floor(Time/BitLength));

    // Claim: 0 <= BitSlot < StringLength

    int LeftBitSlot = BitSlot - 1;
    int RightBitSlot = BitSlot + 1;

    if(LeftBitSlot < 0)
      LeftBitSlot += StringLength;

    if(RightBitSlot >= StringLength)
      RightBitSlot -= StringLength;

    // Work out whether Time is in left or right half of the bit slot.

    if(Time - BitSlot*BitLength < 0.5*BitLength) // ii is in left half
      {
        int BitValue = oBitString->GetBit(ChannelArrayIndex,BitSlot);
        int LeftBitValue = oBitString->GetBit(ChannelArrayIndex,LeftBitSlot);

        if(LeftBitValue == 1 && BitValue == 1)
          Value = 1;
        if(LeftBitValue == 0 && BitValue == 0)
          Value = 0;
        if(LeftBitValue == 1 && BitValue == 0)
          Value = EOM_SuperGauss(Time - (BitSlot-0.5)*BitLength);
        if(LeftBitValue == 0 && BitValue == 1)
          Value = EOM_SuperGauss(Time - (BitSlot+0.5)*BitLength);

      } // ## end ii is in left half
    else // ii is in right half
      {
        int BitValue = oBitString->GetBit(ChannelArrayIndex,BitSlot);
        int RightBitValue = oBitString->GetBit(ChannelArrayIndex,RightBitSlot);

        if(BitValue == 1 && RightBitValue == 1)
          Value = 1;
        if(BitValue == 0 && RightBitValue == 0)
          Value = 0;
        if(BitValue == 1 && RightBitValue == 0)
          Value = EOM_SuperGauss(Time - (BitSlot+0.5)*BitLength);
        if(BitValue == 0 && RightBitValue == 1)
          Value = EOM_SuperGauss(Time - (BitSlot+1.5)*BitLength);

      }// ## end ii is in right half

     EOM_sfftPM.tPowerSignal[ii] = Value;

  } // ## end for ii-loop

fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);

} // ## end EOM_GenerateSmoothedRectBinaryElecDataSuperGauss

// #############################################################

double OptSignal::EOM_SuperGauss(double Time)
{

  while(Time < -0.5*TimeWindow)
    Time += TimeWindow;
  while(Time >= 0.5*TimeWindow)
    Time -= TimeWindow;

  // -0.5*TimeWindow <= Time < 0.5*TimeWindow

  return pow(2,-pow(Time/(0.5*BitLength), EOM_SuperGaussExponent));

}

// ##########################################################

void OptSignal::EOM_FilterElecData(void)
{
  switch(TypeEOMElecFilter)
    {
      case EOM_NO_ELEC_FILTER:

           break;

      case EOM_GAUSSIAN:

        EOM_ElectricalFilterGaussian(EOM_sfftPM.tPowerSignal, 
                                     EOM_ElecFilter_FreqFWHM);

        fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);   

        break;

      case EOM_BESSEL:

           EOM_ElectricalFilterBessel4(EOM_sfftPM.tPowerSignal,
			               EOM_ElecFilter_FreqFWHM);

           fft->FFT(EOM_sfftPM.tPowerSignal,EOM_sfftPM.fPowerSignal);   

           break;

    } 

}

// ##############################################################

void OptSignal::EOM_ElectricalFilterBessel4(cplx *my_power_time,
                                            double fwhm_norm)
{

  double omega_0=0.0;
  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();

  double phase0125 = arg(EOM_Bessel4(fwhm_norm/8.0, fwhm_norm));

  for(int i=0; i < N; i++, power_f_p++)
  {
    
    double freq = fft->freq(i);
    double phase = -freq/(fwhm_norm/8.0) * phase0125; // minus to compensate
    *power_f_p *= cplx(cos(phase), sin(phase))*
                EOM_Bessel4(freq, fwhm_norm);

     }

  fft->IFFT(0, my_power_time);
   
  if (omega_0 == 0.0)
   {
      for(int i=0; i < N; i++)
         my_power_time[i] = real(my_power_time[i]);
    }

} //## end EOM_ElectricalFilterBessel4

// ################################################################

cplx OptSignal::EOM_Bessel4(double omega,double FWHM)
{
  double omega_0=0.0;

  const double mult = 4.2278353500419810;

  // The negative sign is required for Physics convention

  const cplx s = cplx( 0.0, -(omega-omega_0)/FWHM * mult );

  return 105.0 /  (105.0 + s*(105.0 + s*(45.0 + s*(10.0+s))));
}

// ################################################################

void OptSignal::EOM_ElectricalFilterGaussian(cplx *my_power_time,
					     double fwhm_norm)
{
  double omega_0=0.0;

  fft->FFT(my_power_time, 0);
  cplx *power_f_p = fft->work;
  const int N = fft->give_N();

  for(int i=0; i < N; i++, power_f_p++)
  {

    double freq = fft->freq(i);

    *power_f_p *= EOM_Gaussian(freq,fwhm_norm);
 }


  fft->IFFT(0, my_power_time);

    if (omega_0 == 0.0)
     {
      for(int i=0; i < N; i++)
         my_power_time[i] = real(my_power_time[i]);
     }

}

// #########################################################

double OptSignal::EOM_Gaussian(double omega,double FWHM)
{

// CAUTION: gives FWHM in intensity, not field !! (checked 21 Jan 00)

  double sigma = FWHM / 2.35482 * M_SQRT2;

  return exp(-0.5*sq((omega)/sigma));

}

// ############################################################

void OptSignal::GenerateSignalStringUsingInputSequence(void)
{

   // Written by Jonathan on April 1st. 2004,
  // by adapting code written by Jonh Zweck & Ivan Lima.

  // Fixed for NRZ with nonzero raise time by Ivan Lima (7/11/02)

  // Fixed for finite extinction ratio by Ivan Lima (9/29/02) 

 if (! MemAllocated) 
  {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
  }   
 
// for Sequence input
 double MaxX_Seq; // used for normalize    
 int  SequenceNumIndex;  // 4096
 int  SequenceBitLength;	// 
 int ii_Shift, Num_Shift; // 
 double Zerolevel_Shift;

 fftMatrices sfftM_Seq, sfftM_Seq_Tmp; 

 sfftM_Seq_Tmp.tOptSignalX = NewCplx(qtPoints);
 sfftM_Seq.tOptSignalX = NewCplx(qtPoints);

if (NumChannels !=1){
  cerr << " Input Sequence only used for single channel" << endl;
  	exit(1);
	}
	
 char C_String[40];
// sprintf(C_String,"0101_TimeSequence.in");
// sprintf(C_String,"PRBS_TimeSequence.in");
// sprintf(C_String,"128TimeSequence_0101_0405.in");
// sprintf(C_String,"128TimeSequence_PRBS_0405.in");
 double temp;
 ifstream InFile(InputSequenceFileName.c_str(),ios::in);   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InputSequenceFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InputSequenceFileName << endl << endl;   

 InFile >>  SequenceNumIndex;  // 4096
 InFile >>  SequenceBitLength;	// 128

 if (SequenceNumIndex != qtPoints || SequenceBitLength != StringLength){
 	cerr << "Sequence Input must agree with signal input format" << endl;
	exit(1);
 }
 
 *LogFile << "SequenceNumIndex = " << SequenceNumIndex << endl; 
 *LogFile << "SequenceBitLength = " << SequenceBitLength <<  endl; 
 
 double *ySequence, *xSequence, *y2Sequence;

 ySequence = new double [SequenceNumIndex];
 xSequence = new double [SequenceNumIndex];
 y2Sequence = new double [SequenceNumIndex];
 
 for(int ii=0;ii<SequenceNumIndex;ii++)
    {  
		InFile >> temp;
       InFile >> ySequence[ii];

    }

 InFile >> temp;
 if(temp >0 ){
	   		cerr << "Error: in  ElecGain Freq input."<< endl;
	   		cerr << " temp = "<< temp <<  endl;
    		exit(1);
    	}

 Num_Shift = ReadInt("Num_Shift","",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);
 Zerolevel_Shift = ReadInt("Zerolevel_Shift","",0,-1,
                             LOWER_ONLY,MANDATORY,&InFile,DebugLevel);
 for(int ii=0;ii<SequenceNumIndex;ii++)
    {  

	   ySequence[ii] += Zerolevel_Shift;
	   if(ySequence[ii]<0){
	   		ySequence[ii] = 0;
		}else{
		    ySequence[ii] = sqrt(ySequence[ii]);
		}
//	   *LogFile << temp << "  " << ElecGain[ii] << endl;
    }

 *LogFile << "File end. " << temp << endl;

 InFile.close();

 for(int ii=0;ii<SequenceNumIndex;ii++)
    {  
		xSequence[ii] = ii * 3.125e-12;
//	   *LogFile << temp << "  " << ElecPhase[ii] << endl;
    }
// GaussianSmooth(yElecGain, 3, NumElecGainWidth);

 spline(xSequence,  ySequence, SequenceNumIndex ,  0.0,  0.0,  y2Sequence);

// caculate envelop
      for (int ii = 0; ii < qtPoints-32; ii++)
	{ 

		sfftM_Seq_Tmp.tOptSignalX[ii] = 
			splint( xSequence,  ySequence,  y2Sequence, SequenceNumIndex, ii * 3.125e-12/3984*4001);
 
     } 
      for (int ii = qtPoints-32; ii < qtPoints; ii++)
	{ 

		if (InputSequenceFileName == "PRBS_TimeSequence.in" || InputSequenceFileName == "128TimeSequence_PRBS_0405.in"					 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N1.in" 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N2.in" 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N3.in" 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N4.in" 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N5.in" 
					 || InputSequenceFileName == "128TimeSequence_PRBS_N6.in" 
					){
			sfftM_Seq_Tmp.tOptSignalX[ii] = 
				sfftM_Seq_Tmp.tOptSignalX[ii-32] ;
		} else{
			if(InputSequenceFileName == "0101_TimeSequence.in" || InputSequenceFileName == "128TimeSequence_0101_0405.in"
					 || InputSequenceFileName == "128TimeSequence_0101_N1.in" 
					 || InputSequenceFileName == "128TimeSequence_0101_N2.in" 
					 || InputSequenceFileName == "128TimeSequence_0101_N3.in" 
					 || InputSequenceFileName == "128TimeSequence_0101_N4.in" 
					 || InputSequenceFileName == "128TimeSequence_0101_N5.in" 
					 || InputSequenceFileName == "128TimeSequence_0101_N6.in" 
					 ){
				sfftM_Seq_Tmp.tOptSignalX[ii] = 
					sfftM_Seq_Tmp.tOptSignalX[ii-32*2] ;
			}else{
				cerr << " file name incorrect " << endl << " aborting " << endl;
				exit(1);
			}
        } 
     } 
	 
	 // end for ii loop to incorporate Extinction Ratio and update tOptSignal
		MaxX_Seq = 0;
        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

          if(abs(sfftM_Seq_Tmp.tOptSignalX[ii]) > MaxX_Seq)
		  		MaxX_Seq = abs(sfftM_Seq_Tmp.tOptSignalX[ii]) ;
 
  	  	}
        for (int ii = 0; ii < qtPoints; ii++) 
	  	{
		  	sfftM_Seq_Tmp.tOptSignalX[ii] /= MaxX_Seq ;
   	  	}		
// shift
		for (int ii = 0; ii < qtPoints; ii++) 
	  	{	
			if( ii + Num_Shift < qtPoints && ii + Num_Shift >= 0 ) 
				ii_Shift = ii + Num_Shift;
			else if (ii + Num_Shift < 0)
					ii_Shift = ii + Num_Shift + qtPoints;
				else 	
					ii_Shift = ii + Num_Shift - qtPoints;
			sfftM_Seq.tOptSignalX[ii] = sfftM_Seq_Tmp.tOptSignalX[ii_Shift];
   	  	}	

// for simulation data

 double MaxX,MaxY; // used for normalize    
 double Envelope,Envelope_Tmp;
 double Phase;   // For Chirped pulses
 double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system
 cplx SingleChannelWave, SingleChannelWave_Tmp;

 fftMatrices sfftM_Tmp; 


      sfftM_Tmp.tOptSignalX = NewCplx(qtPoints);
      sfftM_Tmp.fOptSignalX = NewCplx(qtPoints);

      if(TypeSimulation == VECTOR)
	{
         sfftM_Tmp.tOptSignalY = NewCplx(qtPoints);
         sfftM_Tmp.fOptSignalY = NewCplx(qtPoints);
	}        



 for (int ii = 0; ii < qtPoints; ii++) 
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;
    
     // Make Y poln zero even for SCALAR simulations

     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;
 
   } // ## end for ii-loop

 for(int ch=0;ch<NumChannels;ch++)
    {

// caculate envelop
      for (int ii = 0; ii < qtPoints; ii++)
	{ 
		 SingleChannelWave_Tmp = 0.0;
		 
         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          //## Modified by Ivan Lima (9/4/02) to account for the novel
	  //## extinction ratio implementation
          //#$#$if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")	  
          // {    
	  
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

//		NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	       if(TimeRelPulseCenter < BitLength/2.0 && 
                  TimeRelPulseCenter >= -BitLength/2.0)
                        Envelope_Tmp = 1.0;
                     else
                        Envelope_Tmp = 0.0;

                     Phase = 0.0;
 	// RZ
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

  

         //## Extinction ratio fixed by Ivan Lima (8/29/02)
        if (oBitString->GetBit(ch,kk)) {    
   	    SingleChannelWave_Tmp += Envelope_Tmp;
		
	 } else { //## A space
  	    SingleChannelWave_Tmp += sqrt(LinearExtinctionRatio)*Envelope_Tmp;
	 }    
   
   } // ## end for kk-loop




      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

     //## Fixing for NRZ with nonzero raise time (Ivan Lima)


       
         sfftM_Tmp.tOptSignalX[ii] = GetInJones(ch,0)*SingleChannelWave_Tmp;

         if(TypeSimulation == VECTOR)                                      
            sfftM_Tmp.tOptSignalY[ii] = GetInJones(ch,1)*SingleChannelWave_Tmp;
	 //## All the other formats (or NRZ with rise time)

	  // JZFeb03: Corect for Physics Convention 
 
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

       fft->FFT(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX); 

        if(TypeSimulation == VECTOR)    
           fft->FFT(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY);  

        optical_filter(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

        if(TypeSimulation == VECTOR)
           optical_filter(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

/*        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

          if(abs(sfftM_Tmp.tOptSignalX[ii]) > MaxX)
		  		MaxX = abs(sfftM_Tmp.tOptSignalX[ii]) ;
 
          if(TypeSimulation == VECTOR)
          		if(abs(sfftM_Tmp.tOptSignalY[ii]) > MaxY)
		  			MaxY = abs(sfftM_Tmp.tOptSignalY[ii]) ;
              	
   	  	}
		
*/

// calculate RZ pulses
      for (int ii = 0; ii < qtPoints; ii++)
	{ 
         SingleChannelWave = 0.0;
		 
         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          //## Modified by Ivan Lima (9/4/02) to account for the novel
	  //## extinction ratio implementation
          //#$#$if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")	  
          // {    
	  
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

//		NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	// RZ
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

	            Envelope = exp(-sq( TimeRelPulseCenter 
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    Phase = 0.0;
  

         //## Extinction ratio fixed by Ivan Lima (8/29/02)
   	    SingleChannelWave += sqrt(OptPowerChann)
	                        *Envelope*exp(jc*Phase);
   
   } // ## end for kk-loop



    ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

     //## Fixing for NRZ with nonzero raise time (Ivan Lima)


       
	 //## All the other formats (or NRZ without rise time)

	  // JZFeb03: Corect for Physics Convention 

         sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
   	                        exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalX[ii];
      
         if(TypeSimulation == VECTOR)                         
            sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	                             exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalX[ii];

/*          sfftM.tOptSignalX[ii] = sqrt(sfftM_Tmp.tOptSignalX[ii]);
      
         if(TypeSimulation == VECTOR)                         
            sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	                             exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalX[ii];
*/ 
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

      //## Fixing for NRZ with nonzero raise time (Ivan Lima)
     // ## Constraint of the rise time  for NRZ


   } // ## end for ch-loop



		MaxX = 0;
        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

          if(abs(sfftM.tOptSignalX[ii]) > MaxX)
		  		MaxX = abs(sfftM.tOptSignalX[ii]) ;
 
  	  	}


		string OutFileName;
 		sprintf(C_String,"CompareSeq.dat");
 		OutFileName = C_String; 
 		ofstream OutFile(OutFileName.c_str(),ios::out);
 		if (!OutFile){
   			cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   			exit(1);
 		}
        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

            OutFile << ii << "  " << abs(sfftM_Seq.tOptSignalX[ii]) 
		  		<< "  " << abs(sfftM.tOptSignalX[ii]) / MaxX << endl;
 
   	  	}	 
		OutFile.close(); 
		
	
        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

			sfftM.tOptSignalX[ii] = sfftM.tOptSignalX[ii] / abs(sfftM.tOptSignalX[ii])
							* sfftM_Seq.tOptSignalX[ii] * MaxX;
 
  	  	}		

 
   Time2Freq();
 
		FreeCplx(sfftM_Seq.tOptSignalX);
        FreeCplx(sfftM_Seq_Tmp.tOptSignalX);

        FreeCplx(sfftM_Tmp.tOptSignalX);
        FreeCplx(sfftM_Tmp.fOptSignalX);

        if(TypeSimulation == VECTOR)
         {
          FreeCplx(sfftM_Tmp.tOptSignalY);
          FreeCplx(sfftM_Tmp.fOptSignalY);
	 	} 
		
		
	       

 delete [] ySequence ;
 delete [] xSequence ;
 delete [] y2Sequence ;



} // ## end GenerateSignalStringUsingInputSequence()


// #####################################################

void OptSignal::GenerateSignalStringRZGaussianPattern(void)
{
  // adapted for RZ Gaussian Paulse with Pattern depandence Jonathan Feb 04
   
  // please see GenerateSignalStringGaussianOrSech function for reference
  // It works fine with one channel, please check the code again 
  // if you want to use WDM. 



if (! MemAllocated) 
  {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
  }   
 
 double MaxX,MaxY; // used for normalize    
 double Envelope,Envelope_Tmp;
 double Phase;   // For Chirped pulses
 double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system
 cplx SingleChannelWave, SingleChannelWave_Tmp;

 fftMatrices sfftM_Tmp; 

 
      sfftM_Tmp.tOptSignalX = NewCplx(qtPoints);
      sfftM_Tmp.fOptSignalX = NewCplx(qtPoints);

      if(TypeSimulation == VECTOR)
	{
         sfftM_Tmp.tOptSignalY = NewCplx(qtPoints);
         sfftM_Tmp.fOptSignalY = NewCplx(qtPoints);
	}        



 for (int ii = 0; ii < qtPoints; ii++) 
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;
    
     // Make Y poln zero even for SCALAR simulations

     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;
 
   } // ## end for ii-loop

 for(int ch=0;ch<NumChannels;ch++)
    {

// caculate envelop
      for (int ii = 0; ii < qtPoints; ii++)
	{ 
		 SingleChannelWave_Tmp = 0.0;
		 
         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          //## Modified by Ivan Lima (9/4/02) to account for the novel
	  //## extinction ratio implementation
          //#$#$if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")	  
          // {    
	  
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

//		NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	       if(TimeRelPulseCenter < BitLength/2.0 && 
                  TimeRelPulseCenter >= -BitLength/2.0)
                        Envelope_Tmp = 1.0;
                     else
                        Envelope_Tmp = 0.0;

                     Phase = 0.0;
 	// RZ
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

  

         //## Extinction ratio fixed by Ivan Lima (8/29/02)
        if (oBitString->GetBit(ch,kk)) {    
   	    SingleChannelWave_Tmp += Envelope_Tmp;
		
	 } else { //## A space
  	    SingleChannelWave_Tmp += sqrt(LinearExtinctionRatio)*Envelope_Tmp;
	 }    
   
   } // ## end for kk-loop




      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

     //## Fixing for NRZ with nonzero raise time (Ivan Lima)


       
         sfftM_Tmp.tOptSignalX[ii] = GetInJones(ch,0)*SingleChannelWave_Tmp;

         if(TypeSimulation == VECTOR)                                      
            sfftM_Tmp.tOptSignalY[ii] = GetInJones(ch,1)*SingleChannelWave_Tmp;
	 //## All the other formats (or NRZ without rise time)

	  // JZFeb03: Corect for Physics Convention 
 
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

       fft->FFT(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX); 

        if(TypeSimulation == VECTOR)    
           fft->FFT(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY);  

        optical_filter(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

        if(TypeSimulation == VECTOR)
           optical_filter(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

/*        for (int ii = 0; ii < qtPoints; ii++) 
	  	{

          if(abs(sfftM_Tmp.tOptSignalX[ii]) > MaxX)
		  		MaxX = abs(sfftM_Tmp.tOptSignalX[ii]) ;
 
          if(TypeSimulation == VECTOR)
          		if(abs(sfftM_Tmp.tOptSignalY[ii]) > MaxY)
		  			MaxY = abs(sfftM_Tmp.tOptSignalY[ii]) ;
              	
   	  	}
		
*/

// calculate RZ pulses
      for (int ii = 0; ii < qtPoints; ii++)
	{ 
         SingleChannelWave = 0.0;
		 
         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          //## Modified by Ivan Lima (9/4/02) to account for the novel
	  //## extinction ratio implementation
          //#$#$if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")	  
          // {    
	  
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

//		NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	// RZ
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

	            Envelope = exp(-sq( TimeRelPulseCenter 
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    Phase = 0.0;
  

         //## Extinction ratio fixed by Ivan Lima (8/29/02)
   	    SingleChannelWave += sqrt(OptPowerChann)
	                        *Envelope*exp(jc*Phase);
   
   } // ## end for kk-loop



    ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;

      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

     //## Fixing for NRZ with nonzero raise time (Ivan Lima)


       
	 //## All the other formats (or NRZ without rise time)

	  // JZFeb03: Correct for Physics Convention 

         sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
   	       exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalX[ii];
      
         if(TypeSimulation == VECTOR)                         
            sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	             exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalY[ii];

/*          sfftM.tOptSignalX[ii] = sqrt(sfftM_Tmp.tOptSignalX[ii]);
      
         if(TypeSimulation == VECTOR)                         
            sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	          exp(- jc*ChannelPhaseOffset)*sfftM_Tmp.tOptSignalY[ii];
*/ 
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

      //## Fixing for NRZ with nonzero raise time (Ivan Lima)
     // ## Constraint of the rise time  for NRZ

		
		



   } // ## end for ch-loop


   Time2Freq();
 

        FreeCplx(sfftM_Tmp.tOptSignalX);
        FreeCplx(sfftM_Tmp.fOptSignalX);

        if(TypeSimulation == VECTOR)
         {
          FreeCplx(sfftM_Tmp.tOptSignalY);
          FreeCplx(sfftM_Tmp.fOptSignalY);
	 }        

  

} // ## end GenerateSignalStringRZGaussianPattern()

// ##########################################################


void OptSignal::GenerateSignalStringRaisedCosines(void)
{

  // JZFeb03: Updated for physics convention

  // Written by John Zweck, Feb 20th 2001, by significantly adapting 
  // code written by Ivan Lima.

  // WARNING: THIS METHOD NEEDS TO BE REWORKED SUBSTANTIALLY
  // TO PRODUCE SMOOTH PULSES. IT IS A STOP-GAP MEASURE FOR NOW

   if (! MemAllocated) 
    {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
    }   

   int ii,jj,kk; 
   
   double Envelope;
   double Phase;   // For Chirped pulses
   double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system


    for (ii = 0; ii < qtPoints; ii++) 
      {

	// Even for SCALAR simns make Y poln zero

        sfftM.tOptSignalX[ii] = 0.0;
        sfftM.tOptSignalY[ii] = 0.0;
        sfftM.fOptSignalX[ii] = 0.0;
        sfftM.fOptSignalY[ii] = 0.0;

      } // ## end for ii-loop

   for(int ch=0;ch<NumChannels;ch++)
   {

   for (ii = 0; ii < qtPoints; ii++) 
   {

     // Comments by JZ (I think this is correct):
     // (1) Given an index ii we want to know which 
     // bit slot kk the index ii is in, and also
     // what the index jj of the center of this kk-th bit slot is.
     // The reason jj is the index of the center of the bit slot
     // is that 
     // jj = (int) (-1 + (qtPoints/StringLength)*(SomeInteger + 0.5) 
     //
     // and qtPoints/StringLength = Number of points in a bit slot
     // (I'm assuming this ratio is an integer for simplicity).


     // (2) The kk-th bit slot goes from 
     //  ii = kk*M to kk*M + M-1
     // where M =  qtPoints/StringLength is as above.


      jj = (int)  (-1 + qtPoints/StringLength 
                 *(floor( (1.*ii)/(qtPoints/StringLength)) + 0.5) );
      kk = (int) (jj/(qtPoints/StringLength)); 

      // ## It can happen that kk == StringLength in which case 
      // ## it should be set to kk = 0 to be in range for the
      // ## if (oBitString.GetBit(ch,kk)) test (JZ)

      if(kk >= StringLength)
	kk -= StringLength;
     
      if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")
       {    
         // ## Attention: the Width FWHM is of the Power.
         // ## The terms below are of the field (not of the Power).
         // ## FWHM does not apply to NRZ, but the RiseTime
         
         switch(TypePulse)
	   {
	     
	      case CRZ_RCOS: //## CRZ Raised Cosine (Ruo-Mei Mu)
 
		// ## FIX: Should we take sqrt of envelope - YES!!
		// ## FIX: Add filtering for RiseTime
                // ## NOTE: Width currently determined by BitRateChann
		// ## rather than by FWHM_PULSE !!

	   	      Envelope = sqrt(0.5*(1.0+cos(2*pi*BitRateChann*
                                              (ii-jj)*DeltaTime)));
		                

                      Phase = ChirpParameter*pi*
		  	      cos(2*pi*BitRateChann*(ii-jj)*DeltaTime);

	              break;

                case CRZ_SIN_MZ: //## Sinusoidally driven Mach-Zehnder
		  //  RZ pulse format
		  // (Kath, Yang and Evangelides)
		  // "Optimal Prechirping for DM transmission of RZ pulses"
		  // A conference paper probably OFC or ECOC, ThQ4-1 / 249
		  // This pulse models the sort of RZ pulse you would get
		  // by using a sinusoidal voltage to modulate a 
		  // Mach-Zehnder interferometer.
		

		      Envelope = sqrt(0.5*(1.0+cos(pi*
                                   sin(pi*(ii-jj)*DeltaTime/BitLength))));
		                

                      Phase = ChirpParameter*pi*
		  	      cos(2*pi*(ii-jj)*DeltaTime/BitLength);

	              break;

         
              default:
		cerr << "OptSignal::GenerateSignalStringRaisedCosines:"
		     << endl
		     << "Invalid TypePulse"
		     << "ABORTING NOW!!"
		     << endl;

                    exit(1);

	   } // ## end switch(TypePulse)


	 // ## Incorporation of power carried by the "0"-bits using the 
	 // ## ExtinctionRatio

	 // ## See [FOCS, p. 176] for theory.

	 // ## The following equation changes the envelope for a "1" (mark)
         // ## so that
	 //
         // ## (1) It has the background power level of a "0" (space):
         // ##     soOptSignal.OptPowerChann*LinearExtinctionRatio, and
         // ## (2) The pulse shape for a "1" is added to this background
	 // ##     power level in such as way that the peak power is still
         // ##     soOptSignal.OptPowerChann.
	 // 
	 // ##  Note that if you want to eliminate the background level
	 // ##  you should choose LinearExtinctionRatio very small and 
	 // ## positive, i.e ExtinctionRatio in dB very large and negative
	 // ## ExtinctionRatio = -1000 dB works fine

         Envelope = sqrt ( sq(Envelope) * OptPowerChann
                   *(1. - LinearExtinctionRatio)
                   +OptPowerChann*LinearExtinctionRatio );     
                
      }                     
      else 
       {       
         Envelope = sqrt(OptPowerChann*LinearExtinctionRatio); 
         
         Phase = 0.0;

       }   

      // ## Get Jones vector for current channel

      // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

   // ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*(ii)*DeltaTime;

    // JZFeb03: Changed and checked for Physics Convention: 

    ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;


   // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.


      if(TypePulse == CW)
	{
          Envelope = 1.0; 
	  Phase = 0.0;
	}	


      sfftM.tOptSignalX[ii] += GetInJones(ch,0)*Envelope*
                               exp(jc*(Phase - ChannelPhaseOffset)); 

      if(TypeSimulation == VECTOR)
         sfftM.tOptSignalY[ii] += GetInJones(ch,1)*Envelope*
                                  exp(jc*(Phase - ChannelPhaseOffset)); 

   } // ## end for ii-loop
   } // ## end for ch-loop


   Time2Freq();

} // ## end GenerateSignalStringRaisedCosines


// ######################################################


//#################################################
//## This method is adequate to obtain the impulse
//#  response of optical filters and other components
//#  of the system. The impulse is generated only at
//#  the CenterFreq (for now).
//## Remember that this signal is periodic. Thus,  
//#  the number of points shoul be as large as
//#  possible so that
//#  this function will be closer to a real impulse.
//#  Note that the FFT definition in cfftw class
//#  imply the the impulse has to be equal to 
//#  qtPoints at the point where the impulse is.
//##  The impulse is generated in both polarizations.
//## By: Ivan Lima

void OptSignal::GenerateImpulseSignal(void)
{
   for (int ii = 0; ii < qtPoints; ii++) {
      sfftM.tOptSignalX[ii] = 0;                                      
      sfftM.tOptSignalY[ii] = 0;     
   }     
 
   sfftM.tOptSignalX[0] = qtPoints;     

   if(TypeSimulation == VECTOR)
      sfftM.tOptSignalY[0] = qtPoints;

   Time2Freq();  
   SetFrequency(0,CenterFreq );     
}

//########################################################
//##  Change the average power of the current signal 
//##  to a given value  of the average power. 
//##  Note that it is the TOTAL optical power of all the channels
//## By: Ivan Lima  (5/20/02)

// Not setup for oReduOptSignal stuff

void OptSignal::ChangeAveragePowerChannCurrentSignal(
      double AveragePowerChann2)
{
   double AveragePowerChannCurrent = GetAveragePower()/GetNumChannels();
   for (int ii = 0; ii < qtPoints; ii++) 
    {
      sfftM.tOptSignalX[ii] *= sqrt( AveragePowerChann2
                                    /AveragePowerChannCurrent);

      if(TypeSimulation == VECTOR)
         sfftM.tOptSignalY[ii] *= sqrt( AveragePowerChann2
                                    /AveragePowerChannCurrent);
    }
     
   Time2Freq();

}

//########################################################
//##  Regenerated the signal string with a given value
//##  of the average power. 
//##  Note that it is the TOTAL optical power of all the channels
//## By: Ivan Lima  (5/20/02)

// Not setup for oReduOptSignal stuff

void OptSignal::RegenerateSignalWithGivenAveragePowerChann(
      double AveragePowerChann2)
{
   RegenerateSignalString();
   ChangeAveragePowerChannCurrentSignal(AveragePowerChann2);
}


//########################################################
//## Change the center frequency absolutely. That means
//## that the spectral components of optical signal
//## will ACTUALLY be shifted.  Thus, the frequencies 
//## of the channels remain the same.
//## Note that this is quite different from SetCenterFreq().
//## By: Ivan Lima  (6/9/02)

void OptSignal::ChangeAbsoluteCenterFreq(double CenterFreq2)
{   


  // JZFeb03: Works correctly with WDM and Physics convention

   cplx *fOptSignalX_Tmp = NewCplx(qtPoints);
   cplx *fOptSignalY_Tmp = NewCplx(qtPoints);

   int FreqIndexShift = (int) ( (CenterFreq2 - CenterFreq)/DeltaFreq);   
   //cout << "#$# FreqIndexShift = " << FreqIndexShift << endl; 
   fftshift();   
   for (int ii = 0; ii < qtPoints; ii++) {
      fOptSignalX_Tmp[ii] = sfftM.fOptSignalX[ii];
      fOptSignalY_Tmp[ii] = sfftM.fOptSignalY[ii];      
   }      
   for (int ii = 0; ii < qtPoints; ii++) {
      sfftM.fOptSignalX[ii] = fOptSignalX_Tmp[Periodicity(ii+FreqIndexShift,
                                                     qtPoints)];     
      sfftM.fOptSignalY[ii] = fOptSignalY_Tmp[Periodicity(ii+FreqIndexShift,
                                                     qtPoints)];  
   }  
   ifftshift();
   Freq2Time();
   FreeCplx(fOptSignalX_Tmp);
   FreeCplx(fOptSignalY_Tmp);
   
   //for (int ii = 0; ii < NumChannels; ii++) {
   //   SetFrequency(ii,GetFrequency(ii) + CenterFreq2 - CenterFreq );      
   //}
   //## Adjust center frequency
   CenterFreq = CenterFreq2;   
}  
  
//#################################################
//## Make the realocation of the BitStringArray, 
//#  if necessary. In that case, obviously, the data
//#  will be lost. Thus, it should only be used when
//#  you want to import data from another OptSignal
//#  right after this.
//#  This method was developed for one particular
//#  application, so that it will probably not be
//#  maintained.
//#  If you want to use a single channel you should
//#  use ReduceToSingleChannel()
//## By: Ivan Lima
void OptSignal::SetNumChannels(const int NumChannels2)
{
   if (NumChannels != NumChannels2) {
      NumChannels = NumChannels2;
      oBitString->ReleaseBitStringArray();
      oBitString->AllocateBitStringArray(NumChannels);
      oReduOptSignal->SetNumChannels(NumChannels);
   }
   else
      NumChannels = NumChannels2;
}



// #######################################################

//! Not set up for ReduOptSignal with more than one channel

void OptSignal::SetNumberChannelsForDemuxer(int NumChannels2)
{
  // Written by JZ July 29th 2001

  NumChannels = NumChannels2;
  oBitString->SetNumChannels(NumChannels);
  //oReduOptSignal->SetNumChannels(NumChannels);
}

// #########################################################
//## Provides a complete reduction to single channel
//#  from a WDM system
//## By: Ivan Lima (7/9/01)

// Not set up for oReduOptSignal

void OptSignal::ReduceToSingleChannel(int ChannelArrayIndex)
{
   NumChannels = 1;
   //cout << "Reducing2SingleChann " << ChannelArrayIndex 
   //     << "  Freq -> " << GetFrequency(ChannelArrayIndex) << endl;
   SetFrequency(0,GetFrequency(ChannelArrayIndex) );   
   oBitString->ReduceToSingleChannelBitString(ChannelArrayIndex);
}

// ###############################################################

void OptSignal::OpenAccumulatedFirstOrderDispersionFile(string FileName)
{
  
  if(WriteAccumulatedFirstOrderDispersionFlag)
    {

    AccumulatedFirstOrderDispersionFile.open(FileName.c_str(),
                                             ios::out);    

    if (! AccumulatedFirstOrderDispersionFile) 
     {
        cerr << "Error: AccumulatedFirstOrderDispersionFile   "
             << FileName
             << " could not be opened."<< endl;
        exit(1);
     }

 
  AccumulatedFirstOrderDispersionFile << "% PropagatedLength [km] "
   << " AccumulatedFirstOrderDispersion at central frequency [ps/nm] "
   << " Accumulated GVD [ps^2]"
   << endl;

    }

} // ## end OpenAccumulatedFirstOrderDispersionFile

// ###########################################################
 
void OptSignal::WriteAccumulatedFirstOrderDispersion(double PropagatedLength)
{

  // Note: What I'm calling FirstOrderDispersion = Beta'' is really the
  // GVD parameter (see Agrawal) and is measured in ps^2/km.
  // The FirstOrderDispersion is measured in ps/(nm-km).

  // Lets output them both!

  if(WriteAccumulatedFirstOrderDispersionFlag)
    {

     AccumulatedFirstOrderDispersionFile << 1e-3*PropagatedLength
				      << "\t" 
    <<                  -1e3*2*pi*LightSpeed*AccumulatedFirstOrderDispersion/
                        (CenterWavelength*CenterWavelength)
    << "\t"
    << 1e24*AccumulatedFirstOrderDispersion
				      << endl;
    }

} // ## end WriteAccumulatedFirstOrderDispersionFile

// ##########################################################

void OptSignal::CloseAccumulatedFirstOrderDispersionFile(void)
{
  if(WriteAccumulatedFirstOrderDispersionFlag)
     AccumulatedFirstOrderDispersionFile.close();

} // ## end CloseAccumulatedFirstOrderDispersionFile

// ############################################################

void OptSignal::OpenDegreeOfPolarizationFile(string FileName,bool WriteDOPFlag)
{
  
 
  if(WriteDOPFlag)
    {
     DegreeOfPolarizationFile.open(FileName.c_str(),ios::out);    

    if (! DegreeOfPolarizationFile) 
     {
        cerr << "Error: DegreeOfPolarizationFile   "
             << FileName
             << " could not be opened."<< endl;
        exit(1);
     }

 
  DegreeOfPolarizationFile << "% PropagatedLength [km] "
   << " DOP "
   << endl;

    } // ## end if WriteDOPFlag

} // ## end OpenDegreeOfPolarizationFile

// ###########################################################
 
void OptSignal::WriteDegreeOfPolarization(double PropagatedLength,
                                          double RelMinFreq, double RelMaxFreq)
{

   DegreeOfPolarizationFile << 1e-3*PropagatedLength
			    << "\t" 
			    << GetDegreeOfPolarization(RelMinFreq,RelMaxFreq)
			    << endl;

} // ## end WriteDegreeOfPolarizationFile

// ##########################################################

void OptSignal::CloseDegreeOfPolarizationFile(bool WriteDOPFlag)
{
  if(WriteDOPFlag)
     DegreeOfPolarizationFile.close();

} // ## end CloseDegreeOfPolarizationFile


//##########################################################
//## Eliminate the X component (horizontal) of the polarization
//## This method is used in the semi-analytical method to compute
//#  the Q-factor
//## By: Ivan Lima
void OptSignal::EliminateX_Polarization(void)
{
   for (int ii = 0; ii < qtPoints; ii++) {   
      sfftM.tOptSignalX[ii] = 0; 
      sfftM.fOptSignalX[ii] = 0;       
   }    
}

//##########################################################
//## Eliminate the Y component (horizontal) of the polarization
//## This method is used in the semi-analytical method to compute
//#  the Q-factor
//## By: Ivan Lima
void OptSignal::EliminateY_Polarization(void)
{
   for (int ii = 0; ii < qtPoints; ii++) {   
      sfftM.tOptSignalY[ii] = 0; 
      sfftM.fOptSignalY[ii] = 0;       
   }    
}

// ######################################################

void OptSignal::ScaleSignal(double ScaleFactor)
{

for(int ii = 0; ii < qtPoints; ii++) 
 sfftM.tOptSignalX[ii] *= ScaleFactor;
      
for(int ii = 0; ii < qtPoints; ii++)  
 sfftM.tOptSignalY[ii] *= ScaleFactor;

for(int ii = 0; ii < qtPoints; ii++)  
 sfftM.fOptSignalX[ii] *= ScaleFactor;
      
for(int ii = 0; ii < qtPoints; ii++)
 sfftM.fOptSignalY[ii] *= ScaleFactor;

}
// #################################################

void OptSignal::SwitchXandYPolarizations(void)
{

  for (int ii = 0; ii < qtPoints; ii++) 
    {   
      cplx temp = sfftM.tOptSignalX[ii];
      sfftM.tOptSignalX[ii] = sfftM.tOptSignalY[ii];  
      sfftM.tOptSignalY[ii] = temp;
   }    

  for (int ii = 0; ii < qtPoints; ii++) 
    {   
      cplx temp = sfftM.fOptSignalX[ii];
      sfftM.fOptSignalX[ii] = sfftM.fOptSignalY[ii];  
      sfftM.fOptSignalY[ii] = temp;
   }    

}

//##########################################################
//## Get the average power from the time domain
//## By: Ivan Lima
//####################

/**********

OLD CODE. NEW VERSION USING SIMPSON'S RULE is below 

double OptSignal::GetAveragePower(void)
{
  // Checked for Physics convention

   double MassEnergy = 0;

   // TypeSimulation change tested Oct 18 2002 by JZ

if(TypeSimulation == VECTOR) 
 {
  for (int ii = 0; ii < qtPoints; ii++) 
    {   
     MassEnergy +=  sq(abs(sfftM.tOptSignalX[ii])) 
                    +sq(abs(sfftM.tOptSignalY[ii]));
    }    
 }
else
  {
   for (int ii = 0; ii < qtPoints; ii++) 
    {   
      MassEnergy +=  sq(abs(sfftM.tOptSignalX[ii])); 
    } 
  }

   return(MassEnergy*DeltaTime/TimeWindow);

}// end GetAveragePower()
**********/

// ##############################################

double OptSignal::ComputeMaxPower(void)
{

  // TypeSimulation change tested Oct 18 2002 by JZ


double MaxPower = 0.0;

double Power;

switch(TypeSimulation)
      {

      case SCALAR:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj]));

           if(Power > MaxPower)  
              MaxPower = Power;
              
	 } // end for

        break;

      case VECTOR:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj])) + 
                   sq(abs(sfftM.tOptSignalY[jj]));

           if(Power > MaxPower)
              MaxPower = Power;
	 }

        break;

      default:

	cerr << "TypeSimulation error in OptSignal::ComputeMaxPower"
             << endl;
        exit(1);

      } // ## end switch


return MaxPower;

}// ## end ComputeMaxPower

// #########################################################

double OptSignal::ComputeMaxPower(int & ArgMaxIndex)
{

  // TypeSimulation change tested Oct 18 2002 by JZ


double MaxPower = 0.0;

ArgMaxIndex=0;

double Power;

switch(TypeSimulation)
      {

      case SCALAR:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj]));

           if(Power > MaxPower)
             {
              MaxPower = Power;
              ArgMaxIndex = jj;
             }

         } // end for

        break;

      case VECTOR:

        for(int jj=0;jj<qtPoints;jj++)
         {

           Power = sq(abs(sfftM.tOptSignalX[jj])) +
                   sq(abs(sfftM.tOptSignalY[jj]));

           if(Power > MaxPower)
              {
               MaxPower = Power;
               ArgMaxIndex = jj;
              }

         }

        break;
      default:

        cerr << "TypeSimulation error in OptSignal::ComputeMaxPower"
             << endl;
        exit(1);

      } // ## end switch


return MaxPower;

}// ## end ComputeMaxPower

//##########################################################
//## This function gets the total power from the frequency 
//## domain.
//## By: Ivan Lima (07/31/01)
double OptSignal::GetAveragePowerFreqDomain(void)
{

  // TypeSimulation change tested Oct 18 2002 by JZ
      
  // JZFeb03: Tested against ptSignal::GetAveragePower()

   double MassEnergy = 0;

 if(TypeSimulation==VECTOR)
  {
   for (int ii = 0; ii < qtPoints; ii++) 
    {   
      MassEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                    +sq(abs(sfftM.fOptSignalY[ii]));
    }
  }
 else
  {
   for (int ii = 0; ii < qtPoints; ii++) 
    {   
      MassEnergy +=  sq(abs(sfftM.fOptSignalX[ii])); 
    }

  }

 return(MassEnergy);

}// end GetAveragePowerFreqDomain()



//##########################################################
//## This function gets the power inside the bandwidth 
//#  of the channels only. Thus, the total spectrum considered
//#  will be equal to the NumChannels * ChannelSpacing
//#  By: Ivan Lima (05/26/01)
double OptSignal::GetAveragePowerFreqDomainInsideChannels(void)
{

 // TypeSimulation change tested Oct 18 2002 by JZ
 
 double MassEnergy = 0;

 if(TypeSimulation == VECTOR)
   {

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);    
   int kk = qtPoints/2 + (int) ((Frequency[0] - ChannSpacing 
                                 - CenterFreq)/DeltaFreq);
   int jj = qtPoints/2 + (int) ((Frequency[NumChannels-1] + ChannSpacing 
                    - CenterFreq)/DeltaFreq );   

   //   cout << "kk = "<< kk << endl;
   //   cout << "jj = "<< jj << endl;                       
   //int kk = (int) ( (FreqWindow/2. - BandWidth/2.)/DeltaFreq );
   //int jj = (int) ( (FreqWindow/2. + BandWidth/2.)/DeltaFreq );        
  
   for (int ii = kk; ii < jj; ii++) {   
      //for (int ii = 0; ii < qtPoints; ii++) {   
      MassEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
   }
   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY); 
   //cout << "MassEnergy = " << MassEnergy << endl;
   }
 else
   {
    fft->fftshift(sfftM.fOptSignalX);
    
   int kk = qtPoints/2 + (int) ((Frequency[0] - ChannSpacing 
                                 - CenterFreq)/DeltaFreq);
   int jj = qtPoints/2 + (int) ((Frequency[NumChannels-1] + ChannSpacing 
                    - CenterFreq)/DeltaFreq );   

   //   cout << "kk = "<< kk << endl;
   //   cout << "jj = "<< jj << endl;                       
   //int kk = (int) ( (FreqWindow/2. - BandWidth/2.)/DeltaFreq );
   //int jj = (int) ( (FreqWindow/2. + BandWidth/2.)/DeltaFreq );        
  
   for (int ii = kk; ii < jj; ii++) {   
      //for (int ii = 0; ii < qtPoints; ii++) {   
     MassEnergy +=  sq(abs(sfftM.fOptSignalX[ii])); 
    
   }
   fft->ifftshift(sfftM.fOptSignalX);
   
   //cout << "MassEnergy = " << MassEnergy << endl;

   }

   return(MassEnergy);

}// end GetAveragePowerFreqDomainInsideChannels()



//##########################################################
//## This function gets only the power inside the bandwidth 
//#  of the given channel. Thus, the total spectrum considered
//#  will be equal to the central frequency of the channel +- ChannelSpacing
//#  By: Ivan Lima (2/5/03)
double OptSignal::GetAveragePowerFreqDomainInsideChannel(int ChannelArrayIndex)
{
   // FIX SCALAR simulations case!!
   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);    
   //int kk = qtPoints/2 + (int) ((Frequency[0] - ChannSpacing 
   //                              - CenterFreq)/DeltaFreq);
   //int jj = qtPoints/2 + (int) ((Frequency[NumChannels-1] + ChannSpacing 
   //                 - CenterFreq)/DeltaFreq );   
   int kk = qtPoints/2 + (int) ((Frequency[ChannelArrayIndex] - ChannSpacing/2. 
                                 - CenterFreq)/DeltaFreq);
   int jj = qtPoints/2 + (int) ((Frequency[ChannelArrayIndex] + ChannSpacing/2. 
                    - CenterFreq)/DeltaFreq );  
//   cout << "kk = "<< kk << endl;
//   cout << "jj = "<< jj << endl;                       
   //int kk = (int) ( (FreqWindow/2. - BandWidth/2.)/DeltaFreq );
   //int jj = (int) ( (FreqWindow/2. + BandWidth/2.)/DeltaFreq );        
   double MassEnergy = 0;
   for (int ii = kk; ii < jj; ii++) {   
      //for (int ii = 0; ii < qtPoints; ii++) {   
      MassEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
   }
   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY); 
   //cout << "MassEnergy = " << MassEnergy << endl;
   return(MassEnergy);

}// end GetAveragePowerFreqDomainInsideChannels()

// ######################################################

//! Returns integral of conj(tOptSignalX(t))*tOptSignalX(t+TimeOffset)dt

//! By John Zweck. Not the best way to code it up, but should do for now

cplx OptSignal::GetAutocorrelationFunctionOptSignalX(double TimeOffset)
{

  cplx AC = 0.0;
  int shift;

  for(int ii=0;ii<qtPoints;ii++)
    {
      if(TimeOffset != 0)
	{
          shift = (int)((ii*DeltaTime +  TimeOffset)/DeltaTime);

          while(shift < 0)
	    shift += qtPoints;
          while(shift >= qtPoints)
	    shift -= qtPoints;
	}
      else
	shift = ii;

      AC += conj(sfftM.tOptSignalX[ii])*sfftM.tOptSignalX[shift]*DeltaTime;
    }

  return AC/TimeWindow;

}

//##############################################
//##### Get the degree of polarization of the signal
//## By: Ivan Lima
double OptSignal::GetDegreeOfPolarization(void)
{
   double tmpStokes[] = {0,0,0,0};
   for (int ii = 0; ii < qtPoints; ii++) {
      tmpStokes[0] += sq(abs(sfftM.tOptSignalX[ii]))
                     +sq(abs(sfftM.tOptSignalY[ii]));
      tmpStokes[1] += sq(abs(sfftM.tOptSignalX[ii]))
                     -sq(abs(sfftM.tOptSignalY[ii]));                      
      tmpStokes[2] += 2*real( sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
      tmpStokes[3] += 2*real(-jc*sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) ); 
                                               
   }
   return( sqrt(sq(tmpStokes[1])+sq(tmpStokes[2])
               +sq(tmpStokes[3]))/tmpStokes[0]);   

   //return( sqrt(sq(tmpStokes[1]/2)+sq(tmpStokes[2])
   //            +sq(tmpStokes[3]))/(tmpStokes[0]/2));                 
}     

// #####################################################

//##### Get the degree of polarization of the signal
//## By: John Zweck after code by Ivan Lima Sept 7th 2002

//! Computes the degree of polarization and total Stokes parameters using 
//! the frequency domain data only. JZ Sep 7th 2002.

/*! You can either caluclate the degree of polarization and the
    total Stokes parameters in
    the frequency domain or the time domain. The answers
    are the same. Depending on your particular application it might
    be quicker to compute it in the time domain. JZ Sep 7th 2002.
*/


double OptSignal::GetDegreeOfPolarizationAndTotalStokesFreqDomain
                  (double * Stokes)
{
   for(int kk = 0; kk < 4; kk++) 
    {
      Stokes[kk] = 0;
    }   

   for (int ii = 0; ii < qtPoints; ii++) 
    {
      Stokes[0] += sq(abs(sfftM.fOptSignalX[ii]))
                     +sq(abs(sfftM.fOptSignalY[ii]));
      Stokes[1] += sq(abs(sfftM.fOptSignalX[ii]))
                     -sq(abs(sfftM.fOptSignalY[ii]));                      
      Stokes[2] += 2*real( sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );
      Stokes[3] += 2*real(-jc*sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) ); 
                                               
   }

   Stokes[1] /= Stokes[0];
   Stokes[2] /= Stokes[0];
   Stokes[3] /= Stokes[0];
   Stokes[0] = 1;
 
   return sqrt( sq(Stokes[1])+sq(Stokes[2])+sq(Stokes[3]) );

}     


// #####################################################

//##### Get the total Stokes parameters of the signal
//## It is normalized
//## By: Ivan Lima
void OptSignal::GetTotalStokesParameters(double *Stokes)
{
   //## *Stokes2 has to have allocated memory from 0 to 3
   //## First, *Stokes2 is cleaned up   
   for (int ii = 0; ii < 4; ii++) {
      Stokes[ii] = 0;
   }   
   for (int ii = 0; ii < qtPoints; ii++) {
      Stokes[0] += sq(abs(sfftM.tOptSignalX[ii]))
                     +sq(abs(sfftM.tOptSignalY[ii]));
      Stokes[1] += sq(abs(sfftM.tOptSignalX[ii]))
                     -sq(abs(sfftM.tOptSignalY[ii]));                      
      Stokes[2] += 2*real( sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
      Stokes[3] += 2*real(-jc*sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );          
                                      
   }
   //cout << "StokesTotal -> " << Stokes[0] << "  " << Stokes[1] <<"  "
   //     << Stokes[2] << "  " << Stokes[3] << endl;
   
   Stokes[1] /= Stokes[0];
   Stokes[2] /= Stokes[0];
   Stokes[3] /= Stokes[0];
   Stokes[0] = 1;
 
}     

// ############################################

void OptSignal::GetTotalStokesParametersInBandwidth(
				 double *Stokes,
				 double RelMinFreq, double RelMaxFreq)
{

  // Adapted by JZ from code of Ivan Lima. Aug 19th 2002.
   
   for (int ii = 0; ii < 4; ii++) 
    {
      Stokes[ii] = 0;
    }
 
   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);   
 
   int MinFreqIndex = (int) ( (RelMinFreq + FreqWindow/2.0)/DeltaFreq );
   int MaxFreqIndex = (int) ( (RelMaxFreq + FreqWindow/2.0)/DeltaFreq );

   cout << "MinFreqIndex = " << MinFreqIndex << endl;
   cout << "MaxFreqIndex = " << MaxFreqIndex << endl;


   if (MinFreqIndex < 0 || MaxFreqIndex > qtPoints) 
    {
      cerr << "Error: Too wide a frequency interval = " 
           << 1e9*RelMinFreq << " " << 1e9*RelMinFreq << endl
           << "In GetTotalStokesParametersInBandwidth(RelMinFreq,RelMaxFreq)"
           << endl;
      exit(1);   
   }

   
   for (int ii = MinFreqIndex; ii < MaxFreqIndex; ii++) 
    {
      
      Stokes[0] += sq(abs(sfftM.fOptSignalX[ii]))
                     +sq(abs(sfftM.fOptSignalY[ii]));
      Stokes[1] += sq(abs(sfftM.fOptSignalX[ii]))
                     -sq(abs(sfftM.fOptSignalY[ii]));                      
      Stokes[2] += 2*real( sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );
      Stokes[3] += 2*real(-jc*sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );
    }

   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY); 
   
   //cout << "StokesTotalbandwidth -> " 
   //     << Stokes[0] << "  " << Stokes[1] <<"  "
   //     << Stokes[2] << "  " << Stokes[3] << endl;
   
   Stokes[1] /= Stokes[0];
   Stokes[2] /= Stokes[0];
   Stokes[3] /= Stokes[0];
   Stokes[0] = 1;

} // ## end GetTotalStokesParameterInBandwidth

// ##############################################################
// ##### Get the Stokes parameters in a time window
// ## by Hua Jiao, adapted from code of John. Nov 2nd 2005
 
void OptSignal::GetTotalStokesParametersInTime(
				 double *Stokes,
				 int RelMinTimeIndex, int RelMaxTimeIndex)
{
   
   for (int ii = 0; ii < 4; ii++) 
    {
      Stokes[ii] = 0;
    }
  
   int MinTimeIndex = RelMinTimeIndex;
   int MaxTimeIndex = RelMaxTimeIndex;

   if (MinTimeIndex < 0 || MaxTimeIndex > qtPoints) 
    {
      cerr << "Error: Too wide a time interval = " 
           << 1e-12*RelMinTimeIndex << " " << 1e-12*RelMaxTimeIndex << endl
           << "In GetTotalStokesParametersInTime(RelMinTime,RelMaxTime)"
           << endl;
      exit(1);   
   }

   
   for (int ii = MinTimeIndex; ii < MaxTimeIndex; ii++) 
    {
      
      Stokes[0] += sq(abs(sfftM.tOptSignalX[ii]))
                     +sq(abs(sfftM.tOptSignalY[ii]));
      Stokes[1] += sq(abs(sfftM.tOptSignalX[ii]))
                     -sq(abs(sfftM.tOptSignalY[ii]));                      
      Stokes[2] += 2*real( sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
      Stokes[3] += 2*real(-jc*sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
    }

} // ## end GetTotalStokesParameterInTime

// ##############################################################

// ##### Get the degree of polarization of the signal in a bandwidth
// ## By: Ivan Lima (I think)

double OptSignal::GetDegreeOfPolarization(double BandWidth)
{
   double tmpStokes[] = {0,0,0,0};
   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);    
   int kk = (int) ( (FreqWindow/2. - BandWidth/2.)/DeltaFreq );
   int jj = (int) ( (FreqWindow/2. + BandWidth/2.)/DeltaFreq );
   if (kk < 0 || jj > qtPoints) {
      cerr << "Error: bandwidth too wide. : " << BandWidth  << endl
           << "From GetDegreeOfPolarization(BandWidth)"
           << endl;
      exit(1);   
   }
   // ## Notice that: SignalFreq = ii*DeltaFreq - FreqWindow/2.;   
   for (int ii = kk; ii < jj; ii++) 
    {
      tmpStokes[0] += sq(abs(sfftM.fOptSignalX[ii]))
                     +sq(abs(sfftM.fOptSignalY[ii]));
      tmpStokes[1] += sq(abs(sfftM.fOptSignalX[ii]))
                     -sq(abs(sfftM.fOptSignalY[ii]));                      
      tmpStokes[2] += 2*real( sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );
      tmpStokes[3] += 2*real(-jc*sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) ); 
   }

   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY);    

 
    return( sqrt(sq(tmpStokes[1])+sq(tmpStokes[2])
               +sq(tmpStokes[3]))/tmpStokes[0]); 
 
}      

// #########################################################
// ##### Get the degree of polarization of the signal in a frequency interval
// ## By: Ivan Lima, adapted by JZ 15/4/2002

double OptSignal::GetDegreeOfPolarization(double RelMinFreq, double RelMaxFreq)
{
   double tmpStokes[] = {0,0,0,0};

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);   
 

   int MinFreqIndex = (int) ( (RelMinFreq + FreqWindow/2.0)/DeltaFreq );
   int MaxFreqIndex = (int) ( (RelMaxFreq + FreqWindow/2.0)/DeltaFreq );

   //cout << "MinFreqIndex = " << MinFreqIndex << endl;
   //cout << "MaxFreqIndex = " << MaxFreqIndex << endl;


   if (MinFreqIndex < 0 || MaxFreqIndex > qtPoints) 
    {
      cerr << "Error: Too wide a frequency interval = " 
           << 1e9*RelMinFreq << " " << 1e9*RelMinFreq << endl
           << "In GetDegreeOfPolarization(RelMinFreq,RelMaxFreq)"
           << endl;
      exit(1);   
   }

   
   for (int ii = MinFreqIndex; ii < MaxFreqIndex; ii++) 
    {
      /******************
       IVAN: This seems to be a bug. You want to integrate
       over frequency not time. JZ

      tmpStokes[0] += sq(abs(sfftM.tOptSignalX[ii]))
                     +sq(abs(sfftM.tOptSignalY[ii]));
      tmpStokes[1] += sq(abs(sfftM.tOptSignalX[ii]))
                     -sq(abs(sfftM.tOptSignalY[ii]));                      
      tmpStokes[2] += 2*real( sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
      tmpStokes[3] += 2*real(-jc*sfftM.tOptSignalX[ii]
                           *conj(sfftM.tOptSignalY[ii]) );
      **********/

      tmpStokes[0] += sq(abs(sfftM.fOptSignalX[ii]))
                     +sq(abs(sfftM.fOptSignalY[ii]));
      tmpStokes[1] += sq(abs(sfftM.fOptSignalX[ii]))
                     -sq(abs(sfftM.fOptSignalY[ii]));                      
      tmpStokes[2] += 2*real( sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );
      tmpStokes[3] += 2*real(-jc*sfftM.fOptSignalX[ii]
                           *conj(sfftM.fOptSignalY[ii]) );


 
    }

   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY); 
   
   return( sqrt(sq(tmpStokes[1])+sq(tmpStokes[2])
               +sq(tmpStokes[3]))/tmpStokes[0]); 
  
}      

// #########################################################

// ##### Get the square of the pulse width of a single pulse

//! Our of date. Use ComputeRMSPulseWidth() instead

double OptSignal::GetSigmaSqOptSignal(void)
{
 
 // TypeSimulation change tested Oct 18 2002 by JZ
  
   // ## This method requires that there would be only one bit
   // ## But it can be improved to provide the average SigmaSq
   // ## The ExtinctionRatio should be = 1000
//   if (StringLength != 1) {
//      cerr << "From GetSignalWidth(): StringLength = " 
//           << StringLength << " ." << endl 
//           << "Warning: This method requires StringLengrh==1." 
//           << endl; 
//   }
      
   double Power   = 0;
   double tPower  = 0; 
   double t2Power = 0;
   double vTmp;

   if(TypeSimulation == VECTOR)
     {

      for (int ii = 0; ii < qtPoints; ii++) 
       {
        vTmp     = sq(abs(sfftM.tOptSignalX[ii]))
   	          +sq(abs(sfftM.tOptSignalY[ii]));
        Power   += vTmp;
        tPower  += vTmp*(ii*DeltaTime);           
        t2Power += vTmp*sq(ii*DeltaTime);               
       }
     }
   else
     {
      for (int ii = 0; ii < qtPoints; ii++) 
       {
	vTmp     = sq(abs(sfftM.tOptSignalX[ii]));
        Power   += vTmp;
        tPower  += vTmp*(ii*DeltaTime);           
        t2Power += vTmp*sq(ii*DeltaTime);               
       }
     }

   Power   *= DeltaTime; // ## Add the common term in the integrals
   tPower  *= DeltaTime; // #  for completeness. Though it cancells out.
   t2Power *= DeltaTime;
   
   return(t2Power/Power - sq(tPower/Power));
}

// #########################################################

double OptSignal::ComputeRMSPulseWidth(void)
{

double MeanSqTimePulse;

switch(TypeSimulation)
{
  case SCALAR:

   for(int ii=0;ii<qtPoints;ii++)
    QuadratureWorkSpace[ii] = sq(ii*DeltaTime)*sq(abs(sfftM.tOptSignalX[ii]));

   MeanSqTimePulse = Quadrature(QuadratureWorkSpace,qtPoints,0,TimeWindow);
  
  break;

  case VECTOR:

   for(int ii=0;ii<qtPoints;ii++)
    QuadratureWorkSpace[ii] = sq(ii*DeltaTime)*(
                              sq(abs(sfftM.tOptSignalX[ii])) + 
                              sq(abs(sfftM.tOptSignalY[ii])));

   MeanSqTimePulse = Quadrature(QuadratureWorkSpace,qtPoints,0,TimeWindow);

  break;

}// end switch

return sqrt(MeanSqTimePulse/GetTotalEnergy() - sq(ComputeMeanTimePulse()));

} // end ComputeRMSPulseWidth

// #################################################

double OptSignal::ComputeFWHMPulsePower(void)
{

double HalfMaxPower = 0.5*ComputeMaxPower();

double PreviousPower = 0;

double RightTime,LeftTime;

for(int jj=0;jj<qtPoints;jj++)
{
  double Power = sq(abs(sfftM.tOptSignalX[jj]))
                +sq(abs(sfftM.tOptSignalY[jj]));

 if(PreviousPower <= HalfMaxPower && Power >= HalfMaxPower)
    LeftTime = jj*DeltaTime;

 if(PreviousPower >= HalfMaxPower && Power <= HalfMaxPower)
    RightTime = jj*DeltaTime;

 PreviousPower = Power;

}

double FWHM=RightTime-LeftTime;

//cout << "LeftTime, RightTime = " << LeftTime << " " << RightTime << endl;

return FWHM;

}

// #########################################################

// ##### Damps the signal. It is not optimized, but we are on the rush
// ## This is an implementation of absorbers in the boundaries
// ## By: Ivan Lima
void OptSignal::DampBounds(double DampTime)
{
   double DampFactor;
   for (int ii = 0; ii < qtPoints; ii++) {
      if (ii*DeltaTime <= 5.*DampTime ) {      
         DampFactor = 1.                               //- exp(-sq(ii*DeltaTime)/(2.*sq(DampTime)));
                     -1/cosh( ii*DeltaTime
                     /(DampTime/(2*log(1+sqrt(2)) ) ) );             
         sfftM.tOptSignalX[ii] *= DampFactor;                                   
         sfftM.tOptSignalY[ii] *= DampFactor;            
      }    
      if ( (qtPoints - ii)*DeltaTime <= 5.*DampTime) {
         DampFactor = 1.                             //- exp(-sq( (qtPoints - ii)*DeltaTime)/(2.*sq(DampTime)));   
                     -1/cosh( (qtPoints - ii)*DeltaTime
                     /(DampTime/(2*log(1+sqrt(2)) ) ) );             
         sfftM.tOptSignalX[ii] *= DampFactor;                                   
         sfftM.tOptSignalY[ii] *= DampFactor;      
      }        
   } 
   Time2Freq();  
}

// ###############################################################

// ##### Get the Fast Fourier Transform
// ## By: Ivan Lima
void OptSignal::Time2Freq()
{


 // TypeSimulation change tested Oct 18 2002 by JZ

  // JZFeb03: Did not need to be changed for Physics Convention
 
   fft->FFT(sfftM.tOptSignalX, sfftM.fOptSignalX); 

   if(TypeSimulation == VECTOR)
      fft->FFT(sfftM.tOptSignalY, sfftM.fOptSignalY);     
}

// ################################################################

// ##### Get the Fast Fourier Transform of the X polarization
// #     It is suitable for the scalar model
// ## By: Ivan Lima

void OptSignal::Time2FreqScalar()
{
  if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::Time2FreqScalar TypeSimulation error"
	   << endl
	   << "ABORTING!!"
	   << endl;
      exit(1);

    }

   fft->FFT(sfftM.tOptSignalX, sfftM.fOptSignalX);    
}

// #########################################################

// ##### Get the Inverse Fast Fourier Transform
// ## By: Ivan Lima
void OptSignal::Freq2Time()
{

 // TypeSimulation change tested Oct 18 2002 by JZ
 
   fft->IFFT(sfftM.fOptSignalX, sfftM.tOptSignalX); 

   if(TypeSimulation == VECTOR)
      fft->IFFT(sfftM.fOptSignalY, sfftM.tOptSignalY);     
}


//################################################
//##### Run the FFT shift 
//##    It is convenient to visualize the signal
//#     in the spectrum domain.
//#     This is one interface method.
//## By: Ivan Lima
void OptSignal::fftshift(void)
{

 // TypeSimulation change tested Oct 18 2002 by JZ
 
   fft->fftshift(sfftM.fOptSignalX);

   if(TypeSimulation == VECTOR)
      fft->fftshift(sfftM.fOptSignalY);    
}


//################################################
//##### Run the IFFT shift 
//##    It undoes the shift produced by FFT_shift
//#     This is one interface method.
//## By: Ivan Lima
void OptSignal::ifftshift(void)
{
 // TypeSimulation change tested Oct 18 2002 by JZ
 
   fft->ifftshift(sfftM.fOptSignalX);

   if(TypeSimulation == VECTOR)
      fft->ifftshift(sfftM.fOptSignalY);    
}

// ##############################################

// ##### Get the Inverse Fast Fourier Transform for the X polarization
// #     It is suitable for the scalar model
// ## By: Ivan Lima
void OptSignal::Freq2TimeScalar()
{
   if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::Time2FreqScalar TypeSimulation error"
	   << endl
	   << "ABORTING!!"
	   << endl;
      exit(1);

    }

   fft->IFFT(sfftM.fOptSignalX, sfftM.tOptSignalX); 
}

// ##################################################

// ##### Write pulse in time domain to a file
// ## By: Ivan Lima

void OptSignal::WriteFileTimePower(string outFile)
{

  // FIX FOR SCALAR CASE

 // TypeSimulation change tested Oct 18 2002 by JZ
 
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) 
    {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

   int ii;


   if(TypeSimulation == VECTOR)
     {
      for (ii = 0; ii < qtPoints; ii++) 
          OutFile << ii*DeltaTime << "   " 
                  <<  sq(abs(sfftM.tOptSignalX[ii]))
	             +sq(abs(sfftM.tOptSignalY[ii]))  <<endl;   
     }
    else
     {
      for (ii = 0; ii < qtPoints; ii++) 
          OutFile << ii*DeltaTime << "   " 
                  <<  sq(abs(sfftM.tOptSignalX[ii])) <<endl;   
     }

   OutFile.close();
}



//##################################################

//##### Write the power spectrum density of the signal
//##    in wavelength to a file.
//##    I recommend plotting that file in the log scale
//## By: Ivan Lima in 6/26/2003

void OptSignal::WriteFileWavelengthPower(string outFile)
{ 
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

 if(TypeSimulation == VECTOR)
  {

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);
      
   for (int ii = 0; ii < qtPoints-1; ii++)  
    { 
      double Wavelength = Frequency2Wavelength(ii*DeltaFreq - FreqWindow/2
                                               +CenterFreq);

     //cout << Wavelength << "  -> freq = " << ii*DeltaFreq - FreqWindow/2
     //                                          +CenterFreq << endl;					       
					       
      double DeltaLambda = Wavelength 
            - Frequency2Wavelength( (ii+1)*DeltaFreq - FreqWindow/2
                                   +CenterFreq);   				   
      OutFile <<  Wavelength << "   " 
              <<( sq(abs(sfftM.fOptSignalX[ii]))
                 +sq(abs(sfftM.fOptSignalY[ii])) 
                    )/DeltaLambda  
               << endl;   
    }

   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY);    
  }
 else
   {
    fft->fftshift(sfftM.fOptSignalX);
       
    for (int ii = 0; ii < qtPoints-1; ii++)  
     {    
      double Wavelength = Frequency2Wavelength(ii*DeltaFreq - FreqWindow/2
                                               +CenterFreq);
      double DeltaLambda = Wavelength 
            - Frequency2Wavelength( (ii+1)*DeltaFreq - FreqWindow/2
                                   +CenterFreq);   
      OutFile <<  Wavelength << "   " 
              <<( sq(abs(sfftM.fOptSignalX[ii])) 
                 )/DeltaLambda  
               << endl;  
    }

   fft->ifftshift(sfftM.fOptSignalX);
   
   }

   OutFile.close();   
}




//##################################################

//##### Write the power spectrum density of the signal
//##    to a file.
//##    I recommend plotting that file in the log scale
//## By: Ivan Lima

void OptSignal::WriteFileFreqPower(string outFile)
{

  // JZFeb03: Checked for Physics convention.
  // DeltaFreq is in Hz as required. 
  // See document FFTandNLSConventions.pdf

 // TypeSimulation change tested Oct 18 2002 by JZ
 
   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
   }

 if(TypeSimulation == VECTOR)
  {

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);
      
   for (int ii = 0; ii < qtPoints; ii++)  
    {    
      OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
              <<( sq(abs(sfftM.fOptSignalX[ii]))
                 +sq(abs(sfftM.fOptSignalY[ii])) 
                    )/DeltaFreq  
               << endl;   
    }

   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY);    
  }
 else
   {
    fft->fftshift(sfftM.fOptSignalX);
       
    for (int ii = 0; ii < qtPoints; ii++)  
     {    
      OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
              <<( sq(abs(sfftM.fOptSignalX[ii]))
                        )/DeltaFreq  
               << endl;   
    }

   fft->ifftshift(sfftM.fOptSignalX);
   
   }

   OutFile.close();   
}



//##################################################
//##### Write pulse in frequency domain to a file
//##    I recommend using the equivalent method
//##    without the OmitDCFlag.
//## By: Ivan Lima

//! Writes frequency data to a file according to the format specified by
//! TypeOutputFlag. 

/*! \param TypeOutputFlag:
           - 0: Power [W]. Keep central frequency\n
	   - 1: Power Spectral Density [W/Hz]. Omit central frequency.
	        Useful only for single channel case where carrier is
                central frequency\n
	   - 2: Power Spectral Density [W/Hz]. Keep central frequency.
                In this case when you keep StringLength same but change
                qtPoints the DC and side-tones have same height and the
                background power that comes from the shape of a single
                pulse also has the same basic shape, though it looks
                noisy and this "noise" looks different because of 
                different patterning in the bit strings.
                If you double StringLength then you double the height
                of the DC and side tones. \n
	   - 3: Option 2 with 12 digits of accuracy

This method is backward compatible with 
void OptSignal::WriteFileFreqPower(string outFile,int OmitDCFlag)
which just did cases 0 and 1 for TypeOutputFlag. \n

\sa void  OptSignal::WriteAveragePowerdBm(string,double,int)
    which has code you can uncomment for a check of Parseval's relation
    and to see that we have the units correct here.\n

JZAug02

*/

void OptSignal::WriteFileFreqPower(string outFile,int TypeOutputFlag)
{

 // TypeSimulation change tested Oct 18 2002 by JZ

  // JZFeb03: Checked for Physics convention 
   
   ofstream OutFile(outFile.c_str(),ios::out);

   if (!OutFile) 
    {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
     }

   fft->fftshift(sfftM.fOptSignalX);
  
   if(TypeSimulation == VECTOR)
      fft->fftshift(sfftM.fOptSignalY);      

   switch(TypeOutputFlag)
     {
       case 0:

         OutFile << "%(1) Rel Freq [Hz] (2) Power [W] Keep DC" 
		 << endl;
         break;

       case 1: 

         OutFile << "%(1) Rel Freq [Hz] (2) Power Spectral Density [W/Hz] "
		 << endl
		 << "%Omit Central Frequency"
		 << endl;
         break;

       case 2:

         OutFile << "%(1) Rel Freq [Hz] (2) Power Spectral Density [W/Hz] "
		 << endl
		 << "%Keep Central Frequency"
		 << endl;
         break;

       case 3:
         break;

       default:

	    cerr << "Switch error in OptSignal::WriteFileFreqPower" << endl
		 << "ABORTING NOW!!" << endl;

            exit(1);


     }

   for (int ii = 0; ii < qtPoints; ii++) 
    {

     switch(TypeOutputFlag)
	{

	  case 0: // Keep DC term and don't divide by DeltaFreq

	    if(TypeSimulation == VECTOR)
	      {

                OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                        << sq(abs(sfftM.fOptSignalX[ii]))
                          +sq(abs(sfftM.fOptSignalY[ii]))  
                        << endl; 
	      }
             else
	       {
                OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                        << sq(abs(sfftM.fOptSignalX[ii]))
                        << endl; 
	       }

	  break;

	  case 1:

          if (ii != qtPoints/2) 
           {   
	     if(TypeSimulation == VECTOR)  
	       {    
              OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                      <<( sq(abs(sfftM.fOptSignalX[ii]))
                         +sq(abs(sfftM.fOptSignalY[ii])) 
                         )/DeltaFreq  
                      << endl;  
	       }
             else
	       {
                OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                        <<( sq(abs(sfftM.fOptSignalX[ii]))
                          )/DeltaFreq  
                    << endl;  
	       }
 
	   }

          break;

	  case 2: // Keep DC term and divide by DeltaFreq

	    if(TypeSimulation == VECTOR)
	      {
               OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                       << (sq(abs(sfftM.fOptSignalX[ii]))
		       +sq(abs(sfftM.fOptSignalY[ii])))/DeltaFreq
                       << endl; 
	      }
            else
	      {
                OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                       << (sq(abs(sfftM.fOptSignalX[ii])))/DeltaFreq
                       << endl; 
	      }

          break;


	  case 3:

	    OutFile.precision(12);
   
            
            if(TypeSimulation == VECTOR)
	      {
               OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                       <<( sq(abs(sfftM.fOptSignalX[ii]))
                          +sq(abs(sfftM.fOptSignalY[ii])) 
                       )/DeltaFreq  
                       << endl;   
	      }
            else
	      {
               OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
                       <<( sq(abs(sfftM.fOptSignalX[ii]))
			 )/DeltaFreq  
                       << endl;   

	      }

          break;


	  default:

	    cerr << "Switch error in OptSignal::WriteFileFreqPower" << endl
		 << "ABORTING NOW!!" << endl;

            exit(1);

	} // end switch

    } // end for ii

   fft->ifftshift(sfftM.fOptSignalX);

   if(TypeSimulation == VECTOR)
      fft->ifftshift(sfftM.fOptSignalY);
    
   OutFile.close();   

} // ## end WriteFileFreqPower

// ####################################################

// ## Extra methods written by John Zweck

// ######################################################

/*!

Outputs power spectral density [W/Hz] in dBm to a file
See FFTandNLSConventions.pdf for definition
 */ 

void OptSignal::WriteFileFreqPowerdBm(string OutFileName)
{


 // TypeSimulation change tested Oct 18 2002 by JZ
  // JZFeb03: Fixed a normalization error. Now it really gives PSD

  // Written by JZ Jan 10, 2001

   ofstream OutFile(OutFileName.c_str(),ios::out);
   if (!OutFile) 
    {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
    }

 if(TypeSimulation == VECTOR)
  {

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);
      
   for (int ii = 0; ii < qtPoints; ii++) {
     { 
         OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
              << Linear2dBm(
                            (sq(abs(sfftM.fOptSignalX[ii]))
			     +sq(abs(sfftM.fOptSignalY[ii]))) /DeltaFreq )
              << endl;   
      }
   }  
                 
   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY);

  }
 else
  {
   fft->fftshift(sfftM.fOptSignalX);
         
   for (int ii = 0; ii < qtPoints; ii++) {
     { 
         OutFile <<  ii*DeltaFreq - FreqWindow/2 << "   " 
		 << Linear2dBm(sq(abs(sfftM.fOptSignalX[ii])) / DeltaFreq )
		 << endl;   
      }
   }  
                 
   fft->ifftshift(sfftM.fOptSignalX);
   
   }
  
   OutFile.close();   
}

// ####################################################

void OptSignal::WriteComplexFields(string OutFileName, int ForRestartFlag)
{

  // This method writes out X and Y polarizations of the 
  // complex field in a form which can be read by matlab.
  // It is useful for computing the difference
  // between the output ComplexFields of two runs
  // and for restarting a job.

ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

   if(ForRestartFlag)
      OutFile.precision(12);
   else
     {
       OutFile 
       << "%% (1) Time (sec) (2) real X, (3) imag X, (4) real Y, (5) imag Y"
       << endl;
     }

for(int jj=0;jj<qtPoints;jj++)
  {
    if(!ForRestartFlag)
      OutFile << jj*DeltaTime << "\t";


      OutFile << real(sfftM.tOptSignalX[jj]) << "\t"
	      << imag(sfftM.tOptSignalX[jj]) << "\t" 
              << real(sfftM.tOptSignalY[jj]) << "\t"
	      << imag(sfftM.tOptSignalY[jj]) << "\t" 	      
	      << endl;

  } // end for loop


OutFile.close();

} // ## end WriteComplexFields

// ##########################################################

void OptSignal::WriteComplexFieldsX(string OutFileName, int ForRestartFlag)
{

  // This method writes out X  polarization of the 
  // complex field in a form which can be read by matlab.
  // It is useful for computing the difference
  // between the output ComplexFields of two runs
  // and for restarting a job.

ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile)
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

   if(ForRestartFlag)
      OutFile.precision(15);
   else
     {
       OutFile
       << "%% (1) time (sec) (2) real X, (3) imag X"
       << endl;
     }

for(int jj=0;jj<qtPoints;jj++)
  {
    if(!ForRestartFlag)
      OutFile << jj*DeltaTime << "\t";


      OutFile << real(sfftM.tOptSignalX[jj]) << "\t"
              << imag(sfftM.tOptSignalX[jj]) << "\t"
              << endl;

  } // end for loop


OutFile.close();

} // ## end WriteComplexFieldsX


// #############################################################

//! Writes all complex fields to a file for quick restart

/*! Writes 8 columns of data to file with name OutFileName
    Real and Imaginery parts of X and Y polarizations for
    time and frequency data.
    Use with OptSignal::ReadFileComplexFieldsTimeAndFreq(string InFileName)
*/

void OptSignal::WriteFileComplexFieldsTimeAndFreq(string OutFileName)
{

ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


   OutFile.precision(15);


  for(int jj=0;jj<qtPoints;jj++)
    {
     OutFile << real(sfftM.tOptSignalX[jj]) << "\t"
	     << imag(sfftM.tOptSignalX[jj]) << "\t" 
             << real(sfftM.tOptSignalY[jj]) << "\t"
	     << imag(sfftM.tOptSignalY[jj]) << "\t" 	
             << real(sfftM.fOptSignalX[jj]) << "\t"
	     << imag(sfftM.fOptSignalX[jj]) << "\t" 
             << real(sfftM.fOptSignalY[jj]) << "\t"
	     << imag(sfftM.fOptSignalY[jj]) << "\t" 	  
	     << endl;

  } // end for loop

OutFile.close();

} // ## end WriteFileComplexFieldsTimeAndFreq

// #######################################################

//! Reads all complex fields to a file for quick restart

/*! Reads 8 columns of data from file with name InFileName:
    Real and Imaginery parts of X and Y polarizations for
    time and frequency data.
    Use with OptSignal::WriteFileComplexFieldsTimeAndFreq(string InFileName)
*/

void OptSignal::ReadFileComplexFieldsTimeAndFreq(string InFileName)
{

ifstream InFile(InFileName.c_str(),ios::in);

   if (!InFile) 
   {
      cerr << "Error: " << InFileName << "not found."<<endl;
      exit(1);
   }


   double tRealX,tImagX,tRealY,tImagY;
   double fRealX,fImagX,fRealY,fImagY;


for(int jj=0;jj<qtPoints;jj++)
  {

    InFile >> tRealX
	   >> tImagX
	   >> tRealY
	   >> tImagY
           >> fRealX
	   >> fImagX
	   >> fRealY
	   >> fImagY;

    sfftM.tOptSignalX[jj] = tRealX + jc*tImagX;
    sfftM.tOptSignalY[jj] = tRealY + jc*tImagY;

    sfftM.fOptSignalX[jj] = fRealX + jc*fImagX;
    sfftM.fOptSignalY[jj] = fRealY + jc*fImagY;

  } // end for loop


InFile.close();

} // ## end ReadFileComplexFieldsTimeAndFreq

// ##############################################

void OptSignal::WriteComplexFieldsFreq(string OutFileName)
{

  // This method writes out X and Y polarizations of the 
  // complex field in a form which can be read by matlab.

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile) 
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }

   
       OutFile 
       << "%% (1) Freq (Hz) (2) real X, (3) imag X, (4) real Y, (5) imag Y"
       << endl;
   

   fft->fftshift(sfftM.fOptSignalX);
   fft->fftshift(sfftM.fOptSignalY);
      
  for(int jj=0;jj<qtPoints;jj++)
    {
 
      OutFile << jj*DeltaFreq - FreqWindow/2.0 << "\t"
              << real(sfftM.fOptSignalX[jj]) << "\t"
	      << imag(sfftM.fOptSignalX[jj]) << "\t" 
              << real(sfftM.fOptSignalY[jj]) << "\t"
	      << imag(sfftM.fOptSignalY[jj]) << "\t" 
	      << endl;

     } // end for loop

                 
   fft->ifftshift(sfftM.fOptSignalX);
   fft->ifftshift(sfftM.fOptSignalY);
  
   OutFile.close();

} // ## end WriteComplexFieldsFreq

// ############################################

void OptSignal::WriteFreqPhaseX(string OutFileName)
{

   ofstream OutFile(OutFileName.c_str(),ios::out);

   if (!OutFile)
   {
      cerr << "Error: " << OutFileName << "could not be created."<<endl;
      exit(1);
   }


       OutFile
       << "%% (1) Freq (Hz) (2) Phase X"
       << endl;


   fft->fftshift(sfftM.fOptSignalX);

  for(int jj=0;jj<qtPoints;jj++)
    {

      OutFile << jj*DeltaFreq - FreqWindow/2.0 << "\t"
              << GetPhase(sfftM.fOptSignalX[jj]) << endl;

     } // end for loop



   fft->ifftshift(sfftM.fOptSignalX);

   OutFile.close();

} // ## end WriteFreqPhaseX

// #############################################

void OptSignal::InputTimeDataFromFile(void)
{

if(TypeSimulation == SCALAR)
   ReadComplexFieldsX(InputTimeDataFileName);
else
   ReadComplexFields(InputTimeDataFileName);

}

// ##############################################

void OptSignal::ReadComplexFields(string InFileName)
{

  // This method reads in the  X and Y polarizations of the 
  // complex field.
  // It is useful  for restarting a job.

ifstream InFile(InFileName.c_str(),ios::in);

   if (!InFile) 
   {      
      cerr << "Error: " << InFileName << " not found."<<endl;
      exit(1);
   }


   double realX,imagX,realY,imagY;


for(int jj=0;jj<qtPoints;jj++)
  {

    InFile >> realX
	   >> imagX
	   >> realY
	   >> imagY;
	   
    sfftM.tOptSignalX[jj] = realX + jc*imagX;
    sfftM.tOptSignalY[jj] = realY + jc*imagY;
   
  } // end for loop


InFile.close();

 Time2Freq();

} // ## end ReadComplexFields

// ################################################

void OptSignal::ReadComplexFieldsX(string InFileName)
{

  // This method reads in the  X polarization of the 
  // complex field.
  // It is useful  for restarting a job.

ifstream InFile(InFileName.c_str(),ios::in);

   if (!InFile)
   {
      cerr << "Error: " << InFileName << " not found."<<endl;
      exit(1);
   }


   double realX,imagX;


for(int jj=0;jj<qtPoints;jj++)
  {

    InFile >> realX
           >> imagX;

    sfftM.tOptSignalX[jj] = realX + jc*imagX;
    sfftM.tOptSignalY[jj] = 0.0;

  } // end for loop


InFile.close();

 Time2Freq();

} // ## end ReadComplexFieldsX

// ###############################################

void  OptSignal::WriteAveragePowerdBm
(string OutFileName,double distance,int appendFlag)
{

 // TypeSimulation change tested Oct 18 2002 by JZ
  // JZFeb03: Checked for Phyics convention 

  // ## The first time the method is called we should create
  // ## the file (i.e. write over old version of file). 
  // ## Thereafter we append data to the file

  ofstream OutFile;

  static int FirstTimeCalled = 1;

  if(FirstTimeCalled) // Create file
    {
     OutFile.open(OutFileName.c_str(),ios::out);
     OutFile << "%Distance [km], Power [dBm]"
	     << endl;
    }
  else          // Append file
     OutFile.open(OutFileName.c_str(),ios::app);

  FirstTimeCalled = 0;

   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be opened."<<endl;
      exit(1);
     }

   
   double Energy = 0;

   if(TypeSimulation == VECTOR)
     {

      for (int ii = 0; ii < qtPoints; ii++) 
       {   
        Energy +=  sq(abs(sfftM.tOptSignalX[ii])) 
                  +sq(abs(sfftM.tOptSignalY[ii]));
       }
     }
    else
      {
       for (int ii = 0; ii < qtPoints; ii++) 
        {   
	  Energy +=  sq(abs(sfftM.tOptSignalX[ii]));
        }

      }

   //OutFile.precision(8);

   OutFile << distance << "\t"
           << Linear2dBm(Energy*DeltaTime/TimeWindow)
           << endl; 

   OutFile.close();


   // ## To check Parseval's relation you can uncomment the remaining
   // ## lines:

   
   /**
  double FreqEnergy = 0;

  if(TypeSimulation == VECTOR)
  {

    for (int ii = 0; ii < qtPoints; ii++) 
    {   
      FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
    }
  }
  else
  {
   for (int ii = 0; ii < qtPoints; ii++) 
    {   
      FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii]));
    }
  }


   cout << "Time domain Average Power [W]: " << Energy*DeltaTime/TimeWindow 
	<< " Freq domain Average Power [W]: " << FreqEnergy << endl << endl;
   **/
   

} // ## end WriteAveragePowerdBm

// ##########################################

void  OptSignal::WritePowerdBmChannel
(string OutFileName,double distance,int appendFlag,int ChannelArrayIndex)
{

  double RelFreqChannel = Frequency[ChannelArrayIndex] - CenterFreq;

  double LeftFreq = RelFreqChannel - ChannSpacing/2.0;

  double RightFreq = RelFreqChannel + ChannSpacing/2.0;

//  WritePowerdBmFreqWindow(OutFileName,distance,appendFlag,
//                          LeftFreq,RightFreq);
//     

}

// ##########################################

void  OptSignal::WritePowerdBmFreqWindow
(string OutFileName,double distance,int appendFlag,
 double LeftFreq,double RightFreq)
{

  // JZFeb03: Checked for Phyics convention. Outputs power in W

  cerror("This routine may contain bugs. ABORTING");
  exit(1);
  
  // ## BUG??? This routine or the way it's been tested 
  // may be buggy: if you want to output power over multiple
  // spans of a dispersion map it'll only keep the most recent map(JZ)

  // ## NOTE: Only for VECTOR data at the moment

  // ## The first time the method is called we should create
  // ## the file (i.e. write over old version of file). 
  // ## Thereafter we append data to the file

  ofstream OutFile;


  static int FirstTimeCalled = 1;

  if(FirstTimeCalled) // Create file
    {
     OutFile.open(OutFileName.c_str(),ios::out);

     OutFile << "%Distance [km], Power in Channel [mW], Power in Channel [dBm]"
	     << endl;
    }

  else          // Append file
     OutFile.open(OutFileName.c_str(),ios::app);
    
  FirstTimeCalled = 0;


   if (!OutFile) 
     {
      cerr << "Error: " << OutFileName << "could not be opened."<<endl;
      exit(1);
     }


  int LeftIndex,RightIndex;

 
  if(RightFreq < LeftFreq)
    {
      cerr << "Error in WritePowerdBmFreqWindow" << endl;
      exit(1);
    }

  if(LeftFreq >= 0)
    {
      LeftIndex = int(LeftFreq/DeltaFreq);
    }     
  else
    {
      LeftIndex = int(LeftFreq/DeltaFreq) +  qtPoints;
    }


  
  if(LeftIndex < 0 || LeftIndex > qtPoints)
    {
      cerr << "OptSignal::WritePowerdBmFreqWindow: LeftIndex error!" << endl;
      exit(1);
    }

  if(RightFreq >= 0)
    {
      RightIndex = int(RightFreq/DeltaFreq);
    }     
  else
    {
      RightIndex = int(RightFreq/DeltaFreq) +  qtPoints;
    }


  if(RightIndex < 0 || RightIndex > qtPoints)
    {
      cerr << "RightIndex error!" << endl;
      exit(1);
    }


  double FreqEnergy = 0;


 if(LeftIndex <= RightIndex)
    {
      for(int ii=LeftIndex; ii < RightIndex; ii++)
	{
          FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
	}


    }
  else
    {
      int ii;

      for(ii=LeftIndex;ii<qtPoints;ii++)
	{
          FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
	}


      for(ii=0;ii<RightIndex;ii++)
         {
           FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii])) 
                     +sq(abs(sfftM.fOptSignalY[ii]));
       	} 
    } // end else

 
 // Outputs Distance [km], Power in Channel [mW], Power in Channel [dBm]

    OutFile.precision(8);
    OutFile << 1e-3*distance << "\t"
	    << 1e3*FreqEnergy << "\t"
	    << Linear2dBm(FreqEnergy)
           << endl; 
// OutFile << distance << "\t"
//	    << sq(abs(sfftM.fOptSignalX[480])) << "   " 
//		<< sq(abs(sfftM.fOptSignalX[512])) << "   " 
//		<< sq(abs(sfftM.fOptSignalX[570])) << endl; 
     OutFile.close();

} // ## end WriteAveragePowerdBmFreqWindow

// #########################################

template<class T>  T OptSignal::WrapPeriodically(T periodicVariable, 
                                     T leftEndPoint, T rightEndPoint) 
{

// Returns  the result of wrapping a periodic variable, periodicVariable,  
// so that it is in the range [leftEndPoint,rightEndPoint)

  T period = rightEndPoint - leftEndPoint;
  
  while(periodicVariable >= rightEndPoint)
    periodicVariable -= period;

  while(periodicVariable < leftEndPoint)
    periodicVariable += period;

  return periodicVariable;

}// end WrapPeriodically

// #######################################################

// ## Write amplitude and phase of X-polarized complex wave
// ## in time domain to a file

// ## By: John Zweck

void OptSignal::WriteFileTimeAmpPhase(string outFile)
{

  if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::WriteFileTimeAmpPhase not implemented for VECTOR"
	   << endl
	   << "ABORTING"
	   << endl;
      exit(1);

    }

   ofstream OutFile(outFile.c_str(),ios::out);
   if (!OutFile) 
     {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
     }

   int ii;

   for (ii = 0; ii < qtPoints; ii++) 
     {
      OutFile << ii*DeltaTime << "   " 
              << abs(sfftM.tOptSignalX[ii]) << "  "
              << GetPhase(sfftM.tOptSignalX[ii]) 
	      <<endl; 
     }  


   OutFile.close();

} // ## end WriteFileTimeAmpPhase

// ################################################################

double OptSignal::GetLocalFrequency(int timeIndex)
{
  if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::GetLocalFrequency not implemented for VECTOR"
	   << endl
	   << "ABORTING"
	   << endl;
      exit(1);

    }

  // JZFeb03: The definitions below are correct for the physics convention
  // since IFFT is u(t) = sum \tilde u(\omega) exp(-i\omega t)
  // If Phi(t) = - Omega_channel t + O(t^2) then 
  // LocalFrequency(t) = - dPhi/dt = Omega_channel + O(t) as required.

  // However I think we still have a bug somehow???????? Numbers seem
  // wrong.


  // Local Frequency of a complex wave E(t) = A(t) exp(i Phi(t)) 
  // is defined to be (see for example [NLO,p91])
  //
  //                  LocalFrequency(t) = - dPhi/dt 

  // The local frequency at t = timeIndex*DeltaTime is computed from the phase
  // by  using a central difference to approximate the derivative
  // (for second order accuracy). 

  int rightIndex = WrapPeriodically(timeIndex + 1, 0, qtPoints);
  int leftIndex = WrapPeriodically(timeIndex - 1, 0, qtPoints);

  // The function GetPhase returns a value in the range [-pi,pi)

  // If the phase at rightIndex is pi - epsilon and that at leftIndex 
  // is -pi + epsilon, we would want phase difference to be 
  // 2*epsilon rather than  2*pi + 2*epsilon.
  // Consequently we wrap the phase difference to be in the range [-pi,pi)

  // Local Frequency is in Hz = Cycles per second
  
  return -(WrapPeriodically(GetPhase(sfftM.tOptSignalX[rightIndex])-
                            GetPhase(sfftM.tOptSignalX[leftIndex]),
			    -pi,pi))/(4*pi*DeltaTime);
  

  /***************
  // RON'S METHOD:
  int rightIndex = WrapPeriodically(timeIndex + 3, 0, qtPoints);
  int leftIndex = WrapPeriodically(timeIndex, 0, qtPoints);

 return -(WrapPeriodically(GetPhase(sfftM.tOptSignalX[rightIndex])-
                          GetPhase(sfftM.tOptSignalX[leftIndex]),
			   -pi,pi))/(2*pi*3*DeltaTime);
  ************/


} // ## end GetLocalFrequency

// ############################################################

double OptSignal::GetChirp(int timeIndex)
{


  // JZFeb03: Could have a bug??

  if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::GetChirp not implemented for VECTOR"
	   << endl
	   << "ABORTING"
	   << endl;
      exit(1);

    }


  // This method assumes the signal is in the X polarization only!!!

  // The Chirp of a complex wave E(t) = A(t) exp(i Phi(t)) 
  // is defined to be 
  //
  //                  Chirp(t) =  -d^2Phi/dt^2 

  // Chirp is in Hz^2 !!
  // LocalFrequency is in Hz 
  // Usually we are interested in computing the Chirp at time 
  // t_0 = central time of a given pulse.
  // The negative sign is included so that the sign of the chirp
  // agrees with the sign of the chirp parameter C in [NLO,p67]

  // The pulse can then be approximated by using a Taylor's series for Phi(t)
  // about t_0

  // E(t) ~ A(t) exp( i [Phi(t_0) - 2*pi*LocalFrequency(t_0)*(t-t_0) - 
  //                     0.5*4*pi^2*Chirp(t_0)*(t-t_0)^2] )  

  // The Chirp at t = timeIndex*DeltaTime is computed from the phase
  // by  using a central difference to approximate the 2nd derivative
  // (for second order accuracy). 


  int rightIndex = WrapPeriodically<int>(timeIndex + 1, 0, qtPoints);
  int leftIndex = WrapPeriodically<int>(timeIndex - 1, 0, qtPoints);

  // Since the right and left phase differences are computed from the 
  // periodic phase Phi(t) they should be in the range [-pi,pi)
  // (See explanation in the method GetLocalFrequency)

  double rightDiff = WrapPeriodically<double>(
                         GetPhase(sfftM.tOptSignalX[rightIndex])-
		         GetPhase(sfftM.tOptSignalX[timeIndex]),-pi,pi);
             
  double leftDiff = WrapPeriodically<double>(
                         GetPhase(sfftM.tOptSignalX[timeIndex])-
		         GetPhase(sfftM.tOptSignalX[leftIndex]),-pi,pi);


  return -(rightDiff - leftDiff)/(sq(2*pi*DeltaTime));

} // ## end GetChirp



// ############################################################

double OptSignal::ComputeAverageChirpOfMarks(double TimeWithinBitSlot)
{
      // This method computes the average chirp of the marks (1's)
      // at a given time within the bit slot. It only works for a single
      // channel signal, whose index is ch=0. (JZ)

  // This method assumes the signal is in the X polarization only!!!

 
 double ChirpSum = 0.0;
 int NumMarks = 0;

 for(int kk=0;kk<StringLength;kk++)
   {     
      if (oBitString->GetBit(0,kk))
	{
          NumMarks++;

          double Time = kk*BitLength + TimeWithinBitSlot;

          int TimeIndex = int(floor(Time/DeltaTime));

          if(TimeIndex < 0)
             TimeIndex = 0;

          if(TimeIndex >= qtPoints)
             TimeIndex = qtPoints-1;

          ChirpSum += GetChirp(TimeIndex);

	} // end if

   } // ## end for kk-loop

 return ChirpSum/double(NumMarks);

} // ## end ComputeAverageChirpOfMarks

// #########################################################

void OptSignal::WriteLocalFrequencyAndChirp(string outFile)
{

  // ONLY FOR SCALAR CODE!!

  if(TypeSimulation == SCALAR)
    {

      ofstream OutFile(outFile.c_str(),ios::out);

      if (!OutFile) 
       {
        cerr << "Error: " << outFile << "could not be created."<<endl;
        exit(1);
       }

      OutFile << "%% (1) Time (2) LocalFreq in Hz (3) Chirp in Hz^2" << endl;

     double PowerThreshold = 1e-3*ComputeMaxPower();

     int ii;

     for (ii = 0; ii < qtPoints; ii++) 
      {

        if(sq(abs(sfftM.tOptSignalX[ii])) > PowerThreshold)
          OutFile << ii*DeltaTime 
                  << "\t" << GetLocalFrequency(ii) 
                  << "\t" << GetChirp(ii)
                  << endl; 
        else          
          OutFile << ii*DeltaTime << "\t" << "nan" << "\t" << "nan" <<endl;

      }  

      OutFile.close();

      } // end if
   else
     {
      cerr << "No OptSignal::WriteLocalFrequencyAndChirp for TypeSimulation == VECTOR" 
	   << endl;
     }

} // ## end WriteLocalFrequency

// #######################################################

void  OptSignal::WriteChirp(string outFile)
{

 ofstream OutFile(outFile.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << outFile << "could not be created."<<endl;
      exit(1);
     }

   int ii;

   for (ii = 0; ii < qtPoints; ii++) 
     {
      OutFile << ii*DeltaTime << "   " 
              << GetChirp(ii)
	      <<endl; 
     }  


   OutFile.close();

} // ## end WriteChirp

// #####################################################

void  OptSignal::WriteDistTimePowerPhase(string outFile, 
					 int timeIndexIncrement, 
                                         double distance,
                                         int appendFlag)
{

  // ## The first time the method is called we should create
  // ## the file (i.e. write over old version of file). 
  // ## Thereafter we append data to the file

  ofstream OutFile;

  if(appendFlag) // Append file
     OutFile.open(outFile.c_str(),ios::app);
  else          // Create file
     OutFile.open(outFile.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << outFile << "could not be opened."<<endl;
      exit(1);
     }


   int jj;


 if(TypeSimulation == VECTOR)
  {

   for (int ii = 0; ii < int(qtPoints/timeIndexIncrement); ii++) 
     {
       jj = ii*timeIndexIncrement;

       OutFile << distance << "  "
               << jj*DeltaTime << "   " 
               << sq(abs(sfftM.tOptSignalX[jj]))
        	 +sq(abs(sfftM.tOptSignalY[jj])) 
	       << "  "
	       << GetPhase(sfftM.tOptSignalX[jj])
	       << GetPhase(sfftM.tOptSignalY[jj])
	       <<endl; 
     }  
  }
 else
   {
    for (int ii = 0; ii < int(qtPoints/timeIndexIncrement); ii++) 
     {
       jj = ii*timeIndexIncrement;

       OutFile << distance << "  "
               << jj*DeltaTime << "   " 
               << sq(abs(sfftM.tOptSignalX[jj]))
               << "  "
	       << GetPhase(sfftM.tOptSignalX[jj])
	       <<endl; 
     }  
   }

   OutFile.close();

} // ## end WriteDistTimePowerPhase

// ############################################

void  OptSignal::WriteDistTimePower(string outFile, 
					 int timeIndexIncrement, 
                                         double distance,
                                         int appendFlag)
{

  // ## The first time the method is called we should create
  // ## the file (i.e. write over old version of file). 
  // ## Thereafter we append data to the file


  ofstream OutFile;

  if(appendFlag) // Append file
     OutFile.open(outFile.c_str(),ios::app);
  else          // Create file
     OutFile.open(outFile.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << outFile << "could not be opened."<<endl;
      exit(1);
     }


   int jj;

 if(TypeSimulation == VECTOR)
  {

   for (int ii = 0; ii < int(qtPoints/timeIndexIncrement); ii++) 
     {
       jj = ii*timeIndexIncrement;

       OutFile << distance << "\t"
               << jj*DeltaTime << "\t" 
               << sq(abs(sfftM.tOptSignalX[jj]))
        	 +sq(abs(sfftM.tOptSignalY[jj])) 
	       <<endl; 
     }  
  }
 else
   {
    for (int ii = 0; ii < int(qtPoints/timeIndexIncrement); ii++) 
     {
       jj = ii*timeIndexIncrement;

       OutFile << distance << "\t"
               << jj*DeltaTime << "\t" 
               << sq(abs(sfftM.tOptSignalX[jj]))
	       <<endl; 
     }  

   }

   OutFile.close();

} // ## end WriteDistTimePower

// ############################################

void  OptSignal::WriteDistTimeCplxWave(string outFile, double distance, 
                                    int appendFlag)
{


  if(TypeSimulation == VECTOR)
    {
      cerr << "OptSignal::WriteDistTimeCplxWave not implemented for VECTOR"
	   << endl
	   << "ABORTING"
	   << endl;
      exit(1);

    }
  // ## LATER ADD CAPABILITY TO ONLY WRITE OUT EVERY mth POINT IN TIME


  // ## The first time the method is called we should create
  // ## the file (i.e. write over old version of file). 
  // ## Thereafter we append data to the file

  ofstream OutFile;

  if(appendFlag) // Append file
     OutFile.open(outFile.c_str(),ios::app);
  else          // Create file
     OutFile.open(outFile.c_str(),ios::out);

   if (!OutFile) 
     {
      cerr << "Error: " << outFile << "could not be opened."<<endl;
      exit(1);
     }


   for (int ii = 0; ii < qtPoints; ii++) 
     {
       OutFile << distance << "  "
               << ii*DeltaTime << "   " 
               << real(sfftM.tOptSignalX[ii]) << "  "
	       << real(jc*sfftM.tOptSignalX[ii])
	       << endl; 
     }  


   OutFile.close();

} // ## end WriteDistTimeCplxWave

// ############################################

int OptSignal::GetChannelArrayIndex(int ChannelPhysicalIndex)
{

  //## Written by John Zweck Oct 11, 2000
  //## 
  //## ChannelPhysicalIndex runs 
  //## -N/2,..... N/2 - 1 for N = NumChannels even and
  //## - (N-1)/2, ... (N-1)/2 for N odd
  //##
  //## ChannelArrayIndex runs 0,1,...,N-1

  // FIX Add bounds checker!!

  oBitString->GetChannelArrayIndex(ChannelPhysicalIndex);

} // ## end GetChannelArrayIndex

// ##############################################

int OptSignal::GetChannelPhysicalIndex(int ChannelArrayIndex)
{

  //## Written by John Zweck Oct 11, 2000
  //## 
  //## ChannelPhysicalIndex runs 
  //## -N/2,..... N/2 - 1 for N = NumChannels even and
  //## - (N-1)/2, ... (N-1)/2 for N odd
  //##
  //## ChannelArrayIndex runs 0,1,...,N-1

  // FIX Add bounds checker!

  oBitString->GetChannelPhysicalIndex(ChannelArrayIndex);

}

// ################################################

double OptSignal::GetFrequency(int ChannelArrayIndex)
{

//## Written by John Zweck Oct 11, 2000

  return Frequency[ChannelArrayIndex];
}



//################################################
//## By: Ivan Lima (6/9/02)
double OptSignal::GetRelativeCenterFrequency(int ChannelArrayIndex)
{
  return Frequency[ChannelArrayIndex] - CenterFreq;
}
     

     
// ##################################################

double OptSignal::GetWavelength(int ChannelArrayIndex)
{

//## Written by John Zweck Oct 11, 2000

  return Frequency2Wavelength(Frequency[ChannelArrayIndex]);

}

// ######################################################

void OptSignal::SetFrequency(int ChannelArrayIndex, double Freq)
{
  //## Written by John Zweck Nov 14, 2000

  Frequency[ChannelArrayIndex] = Freq;

  if(DebugLevel >=2)
    {
     LogFile->precision(10);

     *LogFile << "Frequency in array indexed channel " 
          << ChannelArrayIndex << " = " 
	  << Freq*1e-9 << " GHz" << endl; 

     LogFile->precision(6);
    }

} // ## end SetFrequency

// ###################################################

void OptSignal::SetInStokes(const int ChannelArrayIndex, 
                            const int StokesIndex, 
                            double StokesValue)
{
  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

  InStokes[StokesIndex + ChannelArrayIndex*4]  = StokesValue;

}

// #######################################################

double OptSignal::GetInStokes(const int ChannelArrayIndex,
                              const int StokesIndex)
{

  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

  return InStokes[StokesIndex + ChannelArrayIndex*4];

}
// #######################################################

void OptSignal::SetInJones(const int ChannelArrayIndex, const int JonesIndex, 
                           cplx JonesValue)
{
  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

  InJones[JonesIndex + ChannelArrayIndex*2]  = JonesValue;

}

// #######################################################

cplx OptSignal::GetInJones(const int ChannelArrayIndex, const int JonesIndex)
{

  //## Written by John Zweck Oct 11, 2000

  // ## FIX: Add bounds checking!!

  return InJones[JonesIndex + ChannelArrayIndex*2];

}
// #######################################################

void OptSignal::SetInStokes(const int ChannelArrayIndex,double * StokesVector)
{

  for(int StokesIndex = 0; StokesIndex < 4; StokesIndex ++)
    InStokes[StokesIndex + ChannelArrayIndex*4] = StokesVector[StokesIndex];

}

// #####################################################

double * OptSignal::GetInStokes(const int ChannelArrayIndex)
{
  return &InStokes[ChannelArrayIndex*4];
}

// ######################################################

cplx * OptSignal::GetInJones(const int ChannelArrayIndex)
{
  return &InJones[ChannelArrayIndex*2];
}

// #####################################################

void OptSignal::SetInJones(const int ChannelArrayIndex, cplx * JonesVector)
{

  for(int JonesIndex = 0; JonesIndex < 2; JonesIndex ++)
    InJones[JonesIndex + ChannelArrayIndex*2] = JonesVector[JonesIndex];

}

// #####################################################

void OptSignal::SetRandomJonesVectors(void)
{
  cplx tempJones[2];

  if(RandomPolarizationsFlag && TypeSimulation == VECTOR)
    {

     for(int ch=0;ch<NumChannels;ch++)
      {
        GetRandomJones(tempJones);
        SetInJones(ch,tempJones);
      } // end for

    } // end if

} // ## end SetRandomJonesVectors

// #######################################################

void  OptSignal::GetRandomJones(cplx * Jones)
{
         
  // This method produces a random Jones vector so that
  // the corresponding Stokes vector is uniformily
  // distributed on the Poincare sphere.
  // The method is adapted from OptFiber::GetFiberSampleCompleteScatt()
  // and OptFiber::GetEulerMatrix.

  // The idea is that to get a random Jones vector we let an
  // Euler matrix act on the Jones vector (1,0). If the angles
  // theta,phi,psi which define the Euler matrix are chosen
  // at random as described below then we get a Jones vector
  // as required.


  // cos(theta) is uniformily distributed in (-1,1)
  // phi and psi are uniformily distributed in (0,2*pi)

  // Jones[0] = M_00 and Jones[1] = M_10 (or M_01 ?) where M is the 
  // Euler matrix.

  // Visually the output looks correct using the 
  // next method TestGetRandomJones
           
  double theta = acos(2*(RNG->GetRanNum() - 0.5));                     
  double phi   = 2*pi*RNG->GetRanNum();
  double psi   = 2*pi*RNG->GetRanNum();

  Jones[0] = exp( jc*(psi+phi)/2.0 )*cos(theta/2.0);
  Jones[1] = jc*exp(-jc*(psi-phi)/2.0)*sin(theta/2.0);
  
} // ## end GetRandomJones

// #####################################################

void OptSignal::TestGetRandomJones(string OutFileName, int NumberPoints)
{

ofstream OutFile(OutFileName.c_str(),ios::out);

if (!OutFile) 
  {
   cerr << "Error: " << OutFileName << "could not be created."<<endl;
   exit(1);
  }

 cplx Jones[2];
 double Stokes[4];

 for(int ii=0;ii<NumberPoints;ii++)
   {

     GetRandomJones(Jones);
    
     Jones2Stokes(Jones,Stokes);

     OutFile << Stokes[1] << " "
             << Stokes[2] << " " 
	     << Stokes[3] << endl;

   }

OutFile.close();

} //## end TestGetRandomJones

// ###################################################

/*!
    Resets OutputPDVectorOFLE and InputPDVectorOFLE to (0,0,0)
    and StokesTransferMatrixTransposeOFLE to the 3x3 identity matrix.
 */


void OptSignal::ResetPDVectorOFLE(void)
{

 // cerror("OptSignal::ResetPDVectorOFLE not implemented yet.");


  if(UsingOptFiberLocalError && TypeSimulation == VECTOR &&
     ComputeInputPDVectorFlagOFLE)
    {
      for(int ii=0;ii<3;ii++)
       OutputPDVectorOFLE[ii] = 0.0;

      for(int ii=0;ii<3;ii++)
       InputPDVectorOFLE[ii] = 0.0;

    }

/*

  if(UsingOptFiberLocalError && TypeSimulation == VECTOR &&
     ComputeInputPDVectorFlagOFLE)
    {
      for(int ii=0;ii<3;ii++)
       InputPDVectorOFLE[ii] = 0.0;

      for(int ii=0;ii<3;ii++)
       for(int jj=0;jj<3;jj++)
         StokesTransferMatrixTransposeOFLE[ii][jj] = 0.0;

      for(int ii=0;ii<3;ii++)
       StokesTransferMatrixTransposeOFLE[ii][ii] = 1.0;

    }

*/
}
// ####################################################

const char * OptSignal::WriteTypePulse(void)
{

  char * name;

  switch(TypePulse)
    {
      case NRZ: name = "NRZ"; break;
      case RZ_GAUSS: name = "RZ Gaussian"; break;
      case RZ_SECH: name = "RZ Sech"; break;
      case CRZ_GAUSS: name = "CRZ Gaussian"; break;
      case CRZ_RCOS: name = "CRZ Raised Cosine"; break;
      case CRZ_SIN_MZ: name = "Sinusoidally driven Mach Zehnder RZ"; break;
      case UMBC_RZ: name = "UMBC RZ"; break;
      case CW: name = "CW"; break;
      case RZ_GAUSS_PATTERN: name = "RZ_GAUSS_PATTERN"; break;
      case QPSK_CW: name = "QPSK_CW"; break;
      case QPSK_RZ_GAUSS: name = "QPSK_RZ_GAUSS"; break;
      case BPSK_PHASE_MOD: name = "BPSK_PHASE_MOD"; break;
      case BPSK_MACH_ZEHNDER: name = "BPSK_MACH_ZEHNDER"; break;
      case CSRZ_MACH_ZEHNDER: name = "CSRZ_MACH_ZEHNDER"; break;
      case RZ33_MACH_ZEHNDER: name = "RZ33_MACH_ZEHNDER"; break;
      case RZ50_MACH_ZEHNDER: name = "RZ50_MACH_ZEHNDER"; break;
      case INPUT_SIGNAL_FROM_FILE: name = "INPUT_SIGNAL_FROM_FILE"; break;
      case QPSK_SEQ_PM_NRZ: name = "QPSK_SEQ_PM_NRZ"; break;
      case QPSK_DPMZ_PM_NRZ: name = "QPSK_DPMZ_PM_NRZ"; break;
      case QPSK_SEQ_PM_AM_RZ33: name = "QPSK_SEQ_PM_AM_RZ33"; break;
      case QPSK_SEQ_PM_AM_RZ50: name = "QPSK_SEQ_PM_AM_RZ50"; break;
      case QPSK_DPMZ_PM_AM_RZ33: name = "QPSK_DPMZ_PM_AM_RZ33"; break;
      case QPSK_DPMZ_PM_AM_RZ50: name = "QPSK_DPMZ_PM_AM_RZ50"; break;
      case NRZ_RCOS_MACH_ZEHNDER: name = "NRZ_RCOS_MACH_ZEHNDER"; break;

      default : name = "Invalid Pulse Type";
    }

  return name;  

}

// #######################################################

void OptSignal::InitializeInStokes(ifstream * InFile)
{

char C_String[20]; 
string InStokesName;
int ParameterReadFlag;


if(TypeSimulation == VECTOR)
 {
  for(int ch = 0; ch < NumChannels;ch++)
    {

     // To create a string object from a string and an int
     // it seems simplest to use the sprintf function for
     // C Strings and then initialize the string using the
     // C String.

     sprintf(C_String,"InStokesChann%d",ch+1);
     InStokesName = C_String;

     ParameterReadFlag = ReadInStokes(InStokesName,ch,InFile);


     // If the InStokes parameters for channels 1,2 are not present
     // we abort.
     // If the InStokes parameters for a channel > 2 are not present
     // we use channel 1 parameters for odd channels and channel 2 
     // parameters for even channels

     if(ch < 2 && !ParameterReadFlag)
       {
         *LogFile << "ERROR initializing Optical Signal." << endl
		 << "InStokesChann1 and InStokesChann2 must be specified" 
                 << endl
		 << "ABORTING NOW"
		 << endl;

         exit(1);

       }

     if(ch>=2 && !ParameterReadFlag)
       {
	 SetInStokes(ch,GetInStokes(ch%2));

         if(DebugLevel >=3)
	   {

         *LogFile << endl << "WARNING: "
		 << endl
                 << "Since InStokesChann" << ch+1 << " was not specified"
	         << endl
	         << "set InStokesChann" << ch+1 << " = " << "InStokesChann" 
                 << (ch)%2 +1
                 << endl;
	   }

       }	
     
     
    } // end for ch loop

 } // end if VECTOR
else
  {

    double Stokes[4];

    Stokes[0] = 1.0;
    Stokes[1] = 1.0;
    Stokes[2] = 0.0;
    Stokes[3] = 0.0;

   for(int ch = 0; ch < NumChannels;ch++)
     SetInStokes(ch,&Stokes[0]);

  }


} // ## end InitializeInStokes

// ######################################################

int OptSignal::ReadInStokes(string InStokesName, int ch, ifstream * InFile)
{

  int ParameterRead = 0;

  string NameBuffer;
  string ParameterNameDollar = "$"+InStokesName;

  double Stokes[4];

   InFile->clear(); // Resets ifstream state to good
  
  // In order for an input operation to actually do something its
  // state must be "good". When the end of file is reached the 
  // state is set to "eof" which is not good! (This will have happened
  // if the ReadInt function has been previously called.)
  // The clear() function resets the state to "good"


  InFile->seekg(0); // Resets the file pointer to the start of the file.


  while (*InFile >> NameBuffer)
    {
      if (ParameterNameDollar==NameBuffer) 
          {

            *InFile >> Stokes[0] >> Stokes[1] >> Stokes[2] >> Stokes[3];

	    SetInStokes(ch,&Stokes[0]);

	    ParameterRead = 1;

     }  // end if   
      
    } // end while


  return ParameterRead;

}

// ######################################################

int OptSignal::GetFloorIndex(double Time)
{
 
  double WrappedTime = WrapPeriodically<double>(Time,0.0,TimeWindow);

  // We have WrappedTime in interval [0,TimeWindow)

  int FloorIndex = int(floor(WrappedTime/DeltaTime));

  // FloorIndex in interval [0,qtPoints) as Floor(n) = n for integer n

  return FloorIndex;
}

// ######################################################

int OptSignal::GetCeilIndex(double Time)
{

  double WrappedTime = WrapPeriodically<double>(Time,0.0,TimeWindow);

  // We have WrappedTime in interval [0,TimeWindow)

  int CeilIndex = int(ceil(WrappedTime/DeltaTime));

  // CeilIndex in interval [0,qtPoints] as Ceil(n) = n for integer n

  if(CeilIndex == qtPoints)
     CeilIndex=0;

  return CeilIndex;

}

// #########################################################

double OptSignal::GetRealX(double Time)
{
  int FloorIndex = GetFloorIndex(Time);
  int CeilIndex = GetCeilIndex(Time);

  if(FloorIndex==CeilIndex)
     return real(sfftM.tOptSignalX[FloorIndex]);
  else
   {

   double WrapTime = WrapPeriodically<double>(Time,0.0,TimeWindow);

   return (WrapTime/DeltaTime - FloorIndex)*real(sfftM.tOptSignalX[CeilIndex]) 
     + (1+FloorIndex - WrapTime/DeltaTime)*real(sfftM.tOptSignalX[FloorIndex]);
   }
   
}

// #####################################################

double OptSignal::GetImagX(double Time)
{
  int FloorIndex = GetFloorIndex(Time);
  int CeilIndex = GetCeilIndex(Time);

  if(FloorIndex==CeilIndex)
     return imag(sfftM.tOptSignalX[FloorIndex]);
  else
   { 

    double WrapTime = WrapPeriodically<double>(Time,0.0,TimeWindow);

   return (WrapTime/DeltaTime - FloorIndex)*imag(sfftM.tOptSignalX[CeilIndex])
     + (1+FloorIndex - WrapTime/DeltaTime)*imag(sfftM.tOptSignalX[FloorIndex]);

   }

}
// #########################################################

double OptSignal::GetRealY(double Time)
{
  int FloorIndex = GetFloorIndex(Time);
  int CeilIndex = GetCeilIndex(Time);

  if(FloorIndex==CeilIndex)
     return real(sfftM.tOptSignalY[FloorIndex]);
  else
   {

    double WrapTime = WrapPeriodically<double>(Time,0.0,TimeWindow);

    return (WrapTime/DeltaTime - FloorIndex)*real(sfftM.tOptSignalY[CeilIndex])
      + (1+FloorIndex - WrapTime/DeltaTime)*real(sfftM.tOptSignalY[FloorIndex]);
    }

}

// #####################################################

double OptSignal::GetImagY(double Time)
{
  int FloorIndex = GetFloorIndex(Time);
  int CeilIndex = GetCeilIndex(Time);

  if(FloorIndex==CeilIndex)
     return imag(sfftM.tOptSignalY[FloorIndex]);
  else
   {

   double WrapTime = WrapPeriodically<double>(Time,0.0,TimeWindow);


   return (WrapTime/DeltaTime - FloorIndex)*imag(sfftM.tOptSignalY[CeilIndex])
     + (1+FloorIndex - WrapTime/DeltaTime)*imag(sfftM.tOptSignalY[FloorIndex]);
   }
}

// #####################################################

double OptSignal::GetEnergyInBitSlot(int BitSlotNumber)
{

  // BitSlotNumber must be in range [0,StringLength)

  if(BitSlotNumber<0 || BitSlotNumber>=StringLength)
    {
      cerr << "Error in OptSignal::GetEnergyInBitSlot: " << endl
	   << "BitSlotNumber<0 || BitSlotNumber>=StringLength" << endl;
      exit(1);
    }

  double LeftTime = BitSlotNumber*DeltaTime;
  double RightTime =  (BitSlotNumber+1)*DeltaTime;

return  GetEnergyInTimeRange(LeftTime,RightTime);

} // ## end GetEnergyInBitSlot

// #################################################

/*********

OLD CODE. NEW VERSION BELOW USES SIMPSON'S RULE

double OptSignal::GetTotalEnergy(void)
{

  // JZFeb03: Checked

 // TypeSimulation change tested Oct 18 2002 by JZ
 
  // Created by John Zweck, Dec 9, 2000
  // This method computes the total energy of the pulses 
  
  // We should do this using Simpson's Rule instead too .......

  int ii;

  double Energy = 0.0;

  if(TypeSimulation == VECTOR)
    {

     for (ii = 0; ii < qtPoints; ii++) 
     {
         Energy += sq(abs(sfftM.tOptSignalX[ii]))
                  +sq(abs(sfftM.tOptSignalY[ii]));

     } // ##end for ii loop
    }
  else
    {
     for (ii = 0; ii < qtPoints; ii++) 
       {
         Energy += sq(abs(sfftM.tOptSignalX[ii]));
                  
       } // ##end for ii loop

    }


 return Energy*DeltaTime;

} // # end GetTotalEnergy

**********/

// #################################################

/*********

OLD CODE. NEW VERSION BELOW USES SIMPSON'S RULE

double OptSignal::GetEnergyInTimeRange(double LeftTime, double RightTime)
{
  // JZFeb03: Checked

 // TypeSimulation change tested Oct 18 2002 by JZ
 
  int left_ii, right_ii;

  left_ii = int(ceil(LeftTime/DeltaTime));
  right_ii = int(floor(RightTime/DeltaTime));

  //cout << "left_ii = " << left_ii << " right_ii = " << right_ii << endl;


  if(left_ii < 0) 
    left_ii=0;
  if(right_ii >= qtPoints)
    right_ii = qtPoints-1;

  int ii;

  double Energy = 0.0;

  if(TypeSimulation == VECTOR)
    {

     for (ii = left_ii; ii <= right_ii; ii++) 
      {
         Energy += sq(abs(sfftM.tOptSignalX[ii]))
                  +sq(abs(sfftM.tOptSignalY[ii]));

      } // ##end for ii loop
    }
   else
     {
      for (ii = left_ii; ii <= right_ii; ii++) 
       {
         Energy += sq(abs(sfftM.tOptSignalX[ii]));
       
       } // ##end for ii loop

     }

 return Energy*DeltaTime;

} // ## end GetEnergyInTimeRange

****/

// ####################################

double OptSignal::GetAveragePower(void)
{

double AvPow = GetTotalEnergy()/TimeWindow;
return AvPow;

}

// ##################################

double OptSignal::GetTotalEnergy(void)
{

double Energy = GetEnergyInTimeRange(0,TimeWindow);

return Energy;
}

// #########################################

double OptSignal::GetEnergyInTimeRange(double LeftTime,double RightTime)
{
  int left_ii = int(ceil(LeftTime/DeltaTime));
  int right_ii = int(floor(RightTime/DeltaTime));

  if(left_ii < 0)
    left_ii=0;
  if(right_ii >= qtPoints)
    right_ii = qtPoints-1;

  int NumPts = right_ii - left_ii + 1;

  double Energy;

  if(TypeSimulation == VECTOR)
    {

     for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = sq(abs(sfftM.tOptSignalX[ii+left_ii]));

     Energy = Quadrature(QuadratureWorkSpace,NumPts,LeftTime,RightTime);

     for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = sq(abs(sfftM.tOptSignalY[ii+left_ii]));

     Energy += Quadrature(QuadratureWorkSpace,NumPts,LeftTime,RightTime);

    }
   else
     {
      for (int ii = 0; ii < NumPts; ii++)
         QuadratureWorkSpace[ii] = sq(abs(sfftM.tOptSignalX[ii+left_ii]));

      Energy = Quadrature(QuadratureWorkSpace,NumPts,LeftTime,RightTime);

     }

 return Energy;

} // ## end GetEnergyInTimeRange

// #####################################

/*********************

Currently  not used in code. 
Not checked for Physics convention

double OptSignal::ComputePulseCenter(int ChannelArrayIndex, int BitIndex)
{

  // The ComputePulseCenter method returns the time in seconds of
  // the center of a pulse using the linear dispersive equation
  // approximation and assuming a Gaussian pulse.
  // (The LDE is special case of NLS in which nonlinearity and 
  // higher order dispersion are set to zero.) 

  // The method assumes that the AccumulatedFirstOrderDispersion
  // has been updated (see comments below)  and that the
  // RegenerateSignalString method has been used to initialize the pulse
  // train (JZ, Feb 8th, 2001).


  // At z = 0, the pulse with index BitIndex is centered at 
  // T = (BitIndex + 0.5)*BitLength

  // Assuming a Gaussian pulse centered at  T and having 
  // relative channel frequency Omega in Hz, i.e.

  // At z = 0, pulse is exp( - (t - T)^2 / (2  tau_0^2)) exp(2*pi i Omega t)
  // and that this pulse evolves according to the linear dispersive
  // equation the pulse center at z > 0 is 
  // T - 2*pi*Omega*AccumulatedFirstOrderDispersion. 



  double PulseCenter =  (BitIndex + 0.5)*BitLength - 
                        2*pi*(CenterFreq - GetFrequency(ChannelArrayIndex))*
                        AccumulatedFirstOrderDispersion;

 

   while(PulseCenter < 0)
         PulseCenter += TimeWindow;
   while(PulseCenter >= TimeWindow)
         PulseCenter -= TimeWindow;

   // Now we have : 0 <= PulseCenter < TimeWindow

   return PulseCenter;

} // ## end  ComputePulseCenter

*********************/

// ########################################

double OptSignal::ComputeMeanTimePulse(void)
{
return ComputeMeanTimePulse(0,TimeWindow);
}

// #########################################

double OptSignal::ComputeMeanTimePulse(double StartTime, double EndTime)
{

  // Written by John Zweck, Jan 2010

  // We assume that 0 <= StartTime < EndTime <= TimeWindow

  if(StartTime > EndTime || StartTime < 0 || EndTime > TimeWindow)
    {
      cerr << "Error in OptSignal::ComputeSigmaSqTimeWindow" << endl
           << "StartTime > EndTime"
           << endl;

      exit(1);

    }

   int Start_ii = int(floor(StartTime/DeltaTime));

   if(Start_ii < 0)
      Start_ii = 0;
   if(Start_ii >= qtPoints)
      Start_ii = qtPoints-1;

   int End_ii = int(ceil(EndTime/DeltaTime));

   if(End_ii < 0)
      End_ii = 0;
   if(End_ii >= qtPoints)
      End_ii = qtPoints-1;


   int NumPts = End_ii - Start_ii + 1;

   double TimeMoment;

   if(DebugLevel > 4) 
    cout << "Start_ii = " << Start_ii << "End_ii = " << End_ii << endl;

   if(TypeSimulation == VECTOR)
    {
      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = (ii+Start_ii)*DeltaTime*
                                  sq(abs(sfftM.tOptSignalX[ii+Start_ii]));

      TimeMoment = Quadrature(QuadratureWorkSpace,NumPts,StartTime,EndTime);

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = (ii+Start_ii)*DeltaTime*
                                  sq(abs(sfftM.tOptSignalY[ii+Start_ii]));

      TimeMoment += Quadrature(QuadratureWorkSpace,NumPts,StartTime,EndTime);

    }
   else
    {

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = (ii+Start_ii)*DeltaTime*
                                  sq(abs(sfftM.tOptSignalX[ii+Start_ii]));

      TimeMoment = Quadrature(QuadratureWorkSpace,NumPts,StartTime,EndTime);

    }//end else

   return TimeMoment/GetTotalEnergy();

} // end ComputeMeanTimePulse

// ###########################################


/***

OLD CODE. 

double OptSignal::ComputeMeanTimePulse(double StartTime, double EndTime)
{

  // Written by John Zweck, Feb 19th 2001

  // We assume that 0 <= StartTime < EndTime <= TimeWindow

  if(StartTime > EndTime || StartTime < 0 || EndTime > TimeWindow)
    {
      cerr << "Error in OptSignal::ComputeSigmaSqTimeWindow" << endl
	   << "StartTime > EndTime"
	   << endl;

      exit(1);

    }

   double Power   = 0;
   double tPower  = 0; 
   double Tmp;

   int Start_ii = int(floor(StartTime/DeltaTime));

   if(Start_ii < 0)
      Start_ii = 0;
   if(Start_ii >= qtPoints)
      Start_ii = qtPoints-1;

   int End_ii = int(ceil(EndTime/DeltaTime));

   if(End_ii < 0)
      End_ii = 0;
   if(End_ii >= qtPoints)
      End_ii = qtPoints-1;

   if(DebugLevel > 4)
   cout << "Start_ii = " << Start_ii << "End_ii = " << End_ii << endl;

if(TypeSimulation == VECTOR)
 {

   for (int ii = Start_ii; ii <= End_ii; ii++) 
    {
       Tmp     = sq(abs(sfftM.tOptSignalX[ii]))
	        +sq(abs(sfftM.tOptSignalY[ii]));
       Power   += Tmp;
       tPower  += Tmp*(ii*DeltaTime);           
    }
 
 }
else
  {
   for (int ii = Start_ii; ii <= End_ii; ii++) 
    {
      Tmp     = sq(abs(sfftM.tOptSignalX[ii]));
      Power   += Tmp;
      tPower  += Tmp*(ii*DeltaTime);           
    }

  }
   
   return tPower/Power;

} // ## end ComputeMeanTimePulse

****/

 // #####################################

//! Needs to be updated. \sa OptSignal::ComputeRMSPulseWidth()

double OptSignal::ComputeSigmaSqTimeWindow(double StartTime, double EndTime)
{
 // TypeSimulation change tested Oct 18 2002 by JZ
 
  // Written by John Zweck, Feb 19th 2001

  // We assume that 0 <= StartTime < EndTime <= TimeWindow

  if(StartTime > EndTime)
    {
      cerr << "Error in OptSignal::ComputeSigmaSqTimeWindow" << endl
	   << "StartTime > EndTime"
	   << endl;

      exit(1);

    }

   double Power   = 0;
   double tPower  = 0; 
   double t2Power = 0;
   double Tmp;

   int Start_ii = int(floor(StartTime/DeltaTime));

   if(Start_ii < 0)
      Start_ii = 0;
   if(Start_ii >= qtPoints)
      Start_ii = qtPoints-1;

   int End_ii = int(ceil(EndTime/DeltaTime));

   if(End_ii < 0)
      End_ii = 0;
   if(End_ii >= qtPoints)
      End_ii = qtPoints-1;

   if(DebugLevel > 4)
   cout << "Start_ii = " << Start_ii << "End_ii = " << End_ii << endl;

   if(TypeSimulation == VECTOR)
     {
      for (int ii = Start_ii; ii <= End_ii; ii++) 
      {
       Tmp     = sq(abs(sfftM.tOptSignalX[ii]))
	        +sq(abs(sfftM.tOptSignalY[ii]));
       Power   += Tmp;
       tPower  += Tmp*(ii*DeltaTime);           
       t2Power += Tmp*sq(ii*DeltaTime);               
      }
     }
    else
      {
       for (int ii = Start_ii; ii <= End_ii; ii++) 
        {
	  Tmp     = sq(abs(sfftM.tOptSignalX[ii]));
	 Power   += Tmp;
         tPower  += Tmp*(ii*DeltaTime);           
         t2Power += Tmp*sq(ii*DeltaTime);               
        }
      }

   return(t2Power/Power - sq(tPower/Power));

} // ## end ComputeSigmaSqTimeWindow

// ###########################################################

double OptSignal::ComputeAverageSigmaMarks(void)
{

  // Written by John Zweck, Feb 19th 2001

  // This method asssumes that signal is single channel!!

  // If you want to use it in MC simulations where the BitString
  // is changed you'll need to modify it to return the NumMarks
  // too.

  // The average sigma returned is in seconds

  double Sum = 0.0;
  int NumMarks = 0;

  for(int kk=0; kk < StringLength; kk++)
    {
      if(oBitString->GetBit(0,kk))
	{
         NumMarks++;

         double StartTime = kk*BitLength;
         double EndTime = (kk+1)*BitLength;

         Sum += sqrt(ComputeSigmaSqTimeWindow(StartTime,EndTime));
	} // ## end if

    }// ## end for kk loop

  return Sum/double(NumMarks);

    }// ## end ComputeAverageSigmaMarks

// ##############################################################


/****

OLD VERSION. See below for new one

double OptSignal::ComputeMeanFreqSignal(void)
{
  // TypeSimulation change tested Oct 18 2002 by JZ

  // Written by John Zweck, May 20th 2001


  // Returns  Mean Freq of signal in Hz.

   double Power   = 0;
   double OmegaTimesPower  = 0;
   double Tmp;

 if(TypeSimulation == VECTOR)
  {
   for (int ii = 0; ii < qtPoints; ii++)
    {
       Tmp     = sq(abs(sfftM.fOptSignalX[ii]))
                +sq(abs(sfftM.fOptSignalY[ii]));

       Power   += Tmp;

       double Freq = fft->freq(ii);

       OmegaTimesPower  += Tmp*Freq;

    }
  }
 else
   {
    for (int ii = 0; ii < qtPoints; ii++)
     {
       Tmp     = sq(abs(sfftM.fOptSignalX[ii]));

       Power   += Tmp;

       double Freq = fft->freq(ii);

       OmegaTimesPower  += Tmp*Freq;

    }

   }

   return OmegaTimesPower/Power;

} // ## end ComputeMeanFreqSignal

*****/

// ############################################################

double OptSignal::ComputeMeanFreqSignal(void)
{
 
  // Written and Debugged and Tested by John Zweck, Jan 2010
  // I shifted a signal in freq and checked that calculated 
  // central freq of that shifted signal correctly.

  // Returns  Mean Freq of signal in Hz.

  int NumPts = qtPoints;

  double FreqMoment, Energy;

   if(TypeSimulation == VECTOR)
    {
      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = fft->freq(ii)*(
                                  sq(abs(sfftM.fOptSignalX[ii])) + 
                                  sq(abs(sfftM.fOptSignalY[ii])));

      FreqMoment = Quadrature(QuadratureWorkSpace,NumPts,0,FreqWindow);

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = sq(abs(sfftM.fOptSignalX[ii])) +
                                 sq(abs(sfftM.fOptSignalY[ii]));
    
      Energy = Quadrature(QuadratureWorkSpace,NumPts,0,FreqWindow);

    }
   else
    {
      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = fft->freq(ii)*(
                                  sq(abs(sfftM.fOptSignalX[ii])));

      FreqMoment = Quadrature(QuadratureWorkSpace,NumPts,0,FreqWindow);

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] =  sq(abs(sfftM.fOptSignalX[ii]));

      Energy = Quadrature(QuadratureWorkSpace,NumPts,0,FreqWindow);

    }//end else

   // I checked that actual Energy is really Energy*sq(TimeWindow)
  
   return FreqMoment/Energy;

} // ## end ComputeMeanFreqSignal

// #########################################################

double OptSignal::ComputeRMSSpectralWidth(void)
{
  // TypeSimulation change tested Oct 18 2002 by JZ
 
  // Written by John Zweck, Feb 19th 2001
  // Adapted and renamed by JZ on May 17th 2001

  // Returns Root Mean Square Bandwidth of signal in Hz.

   double Power   = 0;
   double OmegaTimesPower  = 0; 
   double SqOmegaTimesPower = 0;
   double Tmp;

 if(TypeSimulation == VECTOR)
  {

   for (int ii = 0; ii < qtPoints; ii++) 
    {


       Tmp     = sq(abs(sfftM.fOptSignalX[ii]))
	        +sq(abs(sfftM.fOptSignalY[ii]));

       Power   += Tmp;

       double Freq = fft->freq(ii);

       OmegaTimesPower  += Tmp*Freq;           
       SqOmegaTimesPower += Tmp*sq(Freq);               
    }
  }
else
  {
  for (int ii = 0; ii < qtPoints; ii++) 
    {


      Tmp     = sq(abs(sfftM.fOptSignalX[ii]));
	
       Power   += Tmp;

       double Freq = fft->freq(ii);

       OmegaTimesPower  += Tmp*Freq;           
       SqOmegaTimesPower += Tmp*sq(Freq);               
    }
  }
 
   return sqrt(SqOmegaTimesPower/Power - sq(OmegaTimesPower/Power));

} // ## end ComputeRMSSpectralWidth

// #########################################################

double OptSignal::ComputeMeanPhasePulse(void)
{
return ComputeMeanPhasePulse(0,TimeWindow);
}

//##########################################################

//! See McKinstrie and Xie, JSTQE Vol 8, no. 3, p616, 2002

double OptSignal::ComputeMeanPhasePulse(double StartTime, double EndTime)
{

  if(StartTime > EndTime || StartTime < 0 || EndTime > TimeWindow)
    {
      cerr << "Error in OptSignal::ComputeSigmaSqTimeWindow" << endl
           << "StartTime > EndTime"
           << endl;

      exit(1);

    }

   int Start_ii = int(floor(StartTime/DeltaTime));

   if(Start_ii < 0)
      Start_ii = 0;
   if(Start_ii >= qtPoints)
      Start_ii = qtPoints-1;

   int End_ii = int(ceil(EndTime/DeltaTime));

   if(End_ii < 0)
      End_ii = 0;
   if(End_ii >= qtPoints)
      End_ii = qtPoints-1;


   int NumPts = End_ii - Start_ii + 1;

   double RealMoment, ImagMoment;

   switch(TypeSimulation) 
    {
     case SCALAR:

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = real(sfftM.tOptSignalX[ii+Start_ii])*
                                  sq(abs(sfftM.tOptSignalX[ii+Start_ii]));

      RealMoment = Quadrature(QuadratureWorkSpace,NumPts,StartTime,EndTime);

      for(int ii = 0; ii < NumPts; ii++)
       QuadratureWorkSpace[ii] = imag(sfftM.tOptSignalX[ii+Start_ii])*
                                  sq(abs(sfftM.tOptSignalX[ii+Start_ii]));

      ImagMoment = Quadrature(QuadratureWorkSpace,NumPts,StartTime,EndTime);

     break;

     case VECTOR:

     RealMoment = 1;
     ImagMoment = 1;

     cerr << "OptSignal::ComputeMeanPhaseSignal: Not implemented for "
          << endl
          << "VECTOR signals. Aborting." << endl;
     exit(1);

     break;
    }


   return atan2(ImagMoment,RealMoment);

} // end ComputeMeanPhaseSignal

// ###############################################################

void OptSignal::OpenAveragePulseWidthFile(string FileName,int WriteFlag)
{

 // Written by John Zweck Feb 20th 2001
  
  if(WriteFlag)
    {

    AveragePulseWidthFile.open(FileName.c_str(),ios::out);    

    if (! AveragePulseWidthFile) 
     {
        cerr << "Error: OptSignal::OpenAveragePulseWidthFile   "
             << FileName
             << " could not be opened."<< endl;
        exit(1);
     }

 
  AveragePulseWidthFile << "% PropagatedLength [km] "
			<< "  PulseWidth [ps] "
			<< endl;

  
 
  cout << "WARNING: For PulseWidth computation must have "
       << endl 
       << "extinction ratio = -1000 and a single channel at center frequency!"
       << endl;

 cout << "WARNING: FOR NOW: Must have a single pulse not a pulse train. "
      << "BUT you can fix this in  OptSignal::WriteAveragePulseWidth"
      << endl;


    } // end if WriteFlag


} // ## end OpenAveragePulseWidthFile

// ###########################################################
 
void OptSignal::WriteAveragePulseWidth(double PropagatedLength,
                                       int WriteFlag)
{
 // Written by John Zweck Feb 20th 2001

  if(WriteFlag)
    {

     AveragePulseWidthFile << 1e-3*PropagatedLength
			   << "\t" 
       // FIX ME!! << 1e12*ComputeAverageSigmaMarks()
			   << 1e12*sqrt(GetSigmaSqOptSignal())
			   << endl;
    }

} // ## end WriteAveragePulseWidth

// ##########################################################

void OptSignal::CloseAveragePulseWidthFile(int WriteFlag)
{

  // Written by John Zweck Feb 20th 2001

  if(WriteFlag)
     AveragePulseWidthFile.close();

} // ## end CloseAveragePulseWidthFile

// ##########################################################


// ## Import only the data in time and frequency domains
// ## By: Ivan Lima
void OptSignal::ImportOptSignal(OptSignal *oOptSignalOrig)
{
   if ( qtPoints != oOptSignalOrig->GetQtPoints()
       || StringLength != oOptSignalOrig->GetStringLength()
       || NumChannels != oOptSignalOrig->GetNumChannels() ) {
      *LogFile << "From OptSignal::ImportOptSignal(.)" 
              << "ERROR: either qtPoints, StringLength or NumChannels" << endl
	      << "Are mismatching." << endl
	      << "ABORTING NOW"<<endl;
      cout << "Program aborted. See ocs.log for details" << endl;	            	      
      exit(1);      
   }
   //delete oBitString;
   //oBitString = new BitString(*oOptSignalOrig->oBitString);
   for (int jj = 0; jj < NumChannels; jj++) {
      for (int ii = 0; ii < StringLength; ii++) {
         oBitString->SetBit(jj,ii,oOptSignalOrig->oBitString->GetBit(jj,ii) );
      }
   }   
   for (int ii = 0; ii < qtPoints; ii++) {  
      sfftM.tOptSignalX[ii] = oOptSignalOrig->sfftM.tOptSignalX[ii];
      sfftM.tOptSignalY[ii] = oOptSignalOrig->sfftM.tOptSignalY[ii];
      sfftM.fOptSignalX[ii] = oOptSignalOrig->sfftM.fOptSignalX[ii];
      sfftM.fOptSignalY[ii] = oOptSignalOrig->sfftM.fOptSignalY[ii];
   }   
}

// ##########################################################


void OptSignal::GenerateSignalStringGaussianOrSechWithPhaseShift(
                double PhaseShiftAmount)
{

  // JZFeb03: Updated for Physics convention

  // This method is for a particular application.
  // For two channels one channel has an overall phase shift wrt the other

  // Written by John Zweck on Feb 19th 2001, and Oct 10th 2001
  // by significantly adapting code written by Ivan Lima.

  // This method should be used to generate pulses which are defined by
  // an analytic formula which decays significantly outside the 
  // bit window eg,  Gaussian, sech, NRZ pulses.

  // However it will not work for Raised cosine type pulses for which
  // a different method is required. 

  // With this way of defining the pulse train it will be as smooth
  // (infinitely differentiable) in the time domain as one could 
  // reasonably hope for, and 
  // consequently have the correct asymptotic decay at large
  // frequencies in the frequency domain.
  // To see how well this works look at a log scale frequency
  // domain plot.

  // FIX: The (Gaussian filtered) NRZ pulse train is going to be 
  // nearly as smooth as we would like. This needs further work.

  // FIX: Also the way the Extinction Ratio is done reduces the 
  // smoothness of the pulse somewhat it seems from looking at 
  // output data. However we might be able to live with this.

  // WARNING: The Extinction Ratio should be used cautiously since it may not
  // give quite the output you expect. This is particularly the case
  // for pulses with a fairly large FWHM relative to the BitLength.

  // The pulses are centered at (kk+0.5)*BitLength for kk = 0 to StringLength

  // The algorithm goes as follows:

  // (1) For each channel we construct a cplx valued SingleChannelWave
  //     which is a smooth function of time. To form the X and Y 
  //     polarizations of the WDM signal we take  linear combinations
  //     of the SingleChannelWaves, one for each channel.
  //     The coefficients are given by the product of 
  //     exp( - jc*ChannelPhaseOffset)
  //     -- which shifts the  central frequency to that of the channel --
  //     and the X or Y  
  //     component of the Jones vector of the channel.

  //  (2) For each channel the SingleChannelWave is computed as 
  //      follows. For each mark ("1") we compute a function 
  //      Envelope(t) exp(i Phase(t) which is defined for all
  //      times t in the time window. Because this function may not
  //      be periodic (and hence infinitely differentiable)
  //     on [0,TimeWindow] we insist that the single point 
  //      where it is not smooth is located exactly half of the 
  //     TimeWindow away from the pulse's center where the amplitude
  //     has effectively gone to zero. 
  //     The Envelope(t) exp(i Phase(t) functions, one for each mark,
  //      are summed together to form the a "proto-SingleChannelWave"
  //    function, which can also be expressed in  the form 
  //     Envelope(t) exp(i Phase(t).
  //    To incorporate the Extinction Ratio we replace this last
  //    Envelope(t) function by sqrt( sq(Envelope(t))*(1-LXR) + LXR )
  //    which will be smooth as LXR > 0 and 1-LXR > 0, which implies
  //    that sq(Envelope(t))*(1-LXR) + LXR > 0 so that it's sqrt is smooth,
  //    or rather as smooth as "proto-SingleChannelWave" is!
  //    The resulting function is SingleChannelWave.



if (! MemAllocated) 
  {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
  }   

    
 double Envelope;
 double Phase;   // For Chirped pulses
 double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system

 cplx SingleChannelWave;

  fftMatrices sfftM_Tmp; 
  if (RiseTime != 0 && TypePulse == NRZ) { 
      sfftM_Tmp.tOptSignalX = NewCplx(qtPoints);
      sfftM_Tmp.fOptSignalX = NewCplx(qtPoints);

      if(TypeSimulation == VECTOR)
	{  
         sfftM_Tmp.tOptSignalY = NewCplx(qtPoints);
         sfftM_Tmp.fOptSignalY = NewCplx(qtPoints);
	}        
  }


 for (int ii = 0; ii < qtPoints; ii++) 
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;

   } // ## end for ii-loop


 for(int ch=0;ch<NumChannels;ch++)
    {

      for (int ii = 0; ii < qtPoints; ii++)
	{ 
         SingleChannelWave = 0.0;

         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")
           {    
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

	 switch(TypePulse)
	   {

             case NRZ: 
	       
               // JZAug02: The following line was buggy.
	       // It has now been fixed
	       // if(abs(TimeRelPulseCenter) < BitLength/2.0)

	       if(TimeRelPulseCenter < BitLength/2.0 && 
                  TimeRelPulseCenter >= -BitLength/2.0)
                        Envelope = 1.0;
                     else
                        Envelope = 0.0;

                     Phase = 0.0;

                     break;

             case RZ_GAUSS:
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

	            Envelope = exp(-sq( TimeRelPulseCenter 
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    Phase = 0.0;
  
                    break;
        
	     case RZ_SECH:  
                     // ## FWHM = 2*log(1+sqrt(2)) * sigma_power  
         
                     Envelope = 1.0/cosh( TimeRelPulseCenter
                              /(FWHM_Pulse/(2*log(1+sqrt(2)) )) );

                     Phase = 0.0;

                     break;

	      case CRZ_GAUSS: 

		     Envelope   = exp( -2*log(2)*
			              sq( TimeRelPulseCenter/FWHM_Pulse ));
                   
		     // The ChirpParameter here is precisely the chirp
		     // parameter, C, in [NLO,p67].

                     Phase = -ChirpParameter*2*log(2)*
			      sq( TimeRelPulseCenter/FWHM_Pulse );
  

	             break;

	
         
              default:
		cerr << "TypePulse Error in " << endl
		     << "OptSignal::GenerateSignalStringGaussianOrSech."
		     << endl;
			 
                    exit(1);

	   } // ## end switch(TypePulse)

   
	 SingleChannelWave += Envelope*exp(jc*Phase);

    } // ## end if bit is a mark 
   } // ## end for kk-loop


      // At this point SingleChannelWave includes the contributions
      // from all the pulses at ii*DeltaTime

      // Next we incorporate the ExtinctionRatio into SingleChannelWave
      // and update  tOptSignal

  

         // ## First incorporate the ExtinctionRatio

	 // ## Incorporation of power carried by the "0"-bits using the 
	 // ## ExtinctionRatio

	 // ## See [FOCS, p. 176] for theory.

	 // ## The following equation changes the envelope for a "1" (mark)
         // ## so that
	 //
         // ## (1) It has the background power level of a "0" (space):
         // ##     OptPowerChann*LinearExtinctionRatio, and
         // ## (2) The pulse shape for a "1" is added to this background
	 // ##     power level in such as way that the peak power is still
         // ##     OptPowerChann.
	 // 
	 // ##  Note that if you want to eliminate the background level
	 // ##  you should choose LinearExtinctionRatio very small and 
	 // ## positive, i.e ExtinctionRatio in dB very large and negative
	 // ## ExtinctionRatio = -1000 dB works fine

        double AmplitudeNew = abs(SingleChannelWave);

	  if(AmplitudeNew == 0)
	    SingleChannelWave = sqrt(OptPowerChann*LinearExtinctionRatio);
	  else
	  {

           AmplitudeNew = sqrt ( sq(AmplitudeNew) * OptPowerChann
                   *(1. - LinearExtinctionRatio)
                   +OptPowerChann*LinearExtinctionRatio );

          SingleChannelWave = (AmplitudeNew/abs(SingleChannelWave))*
	                           SingleChannelWave;

	  } // ## end else


      // #### Next add the SingleChannelWave to the WDM Signal:

      // ## Get Jones vector for current channel

      // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

    //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*(ii)*DeltaTime;

	  // JZFeb03: Updated for Physics convention

      ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;


      if(ch%2==0)
	{
	  // Add overall phase shift

          ChannelPhaseOffset += 2*pi*PhaseShiftAmount/360;
	}


      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.



    //## Fixing for NRZ with nonzero raise time (Ivan Lima)
      if (RiseTime != 0 && TypePulse == NRZ) 
       { 
         sfftM_Tmp.tOptSignalX[ii] = GetInJones(ch,0)*SingleChannelWave;   
         
         if(TypeSimulation == VECTOR)
          sfftM_Tmp.tOptSignalY[ii] = GetInJones(ch,1)*SingleChannelWave;
       }	
      else 
       { //## All the other formats (or NRZ without rise time)
         sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
   	                        exp(- jc*ChannelPhaseOffset);           
       
         if(TypeSimulation == VECTOR)
           sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	                            exp(- jc*ChannelPhaseOffset);
      }
      
     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

      //## Fixing for NRZ with nonzero raise time (Ivan Lima)
     // ## Constraint of the rise time  for NRZ
     if (RiseTime != 0 && TypePulse == NRZ) 
      { 
        fft->FFT(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX); 

        if(TypeSimulation == VECTOR)
          fft->FFT(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY);
  
        optical_filter(sfftM_Tmp.tOptSignalX, sfftM_Tmp.fOptSignalX, fft,
                       3/(RiseTime*pi)/(2*log(2)) );


        if(TypeSimulation == VECTOR)
         optical_filter(sfftM_Tmp.tOptSignalY, sfftM_Tmp.fOptSignalY, fft,
                       3/(RiseTime*pi)/(2*log(2)) );

        for (int ii = 0; ii < qtPoints; ii++) {
	  //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*
	  //                      (ii)*DeltaTime;	

	  // JZFeb03: Updated for Physics convention 

           ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*
	                        (ii)*DeltaTime;	


           sfftM.tOptSignalX[ii] += sfftM_Tmp.tOptSignalX[ii]*
   	                        exp(- jc*ChannelPhaseOffset);         
                             

           if(TypeSimulation == VECTOR)
               sfftM.tOptSignalY[ii] += sfftM_Tmp.tOptSignalY[ii]*
   	                                exp(- jc*ChannelPhaseOffset);	
	}
     } // end if NRZ


   } // ## end for ch-loop


   Time2Freq();

   if (RiseTime != 0 && TypePulse == NRZ) { 
        FreeCplx(sfftM_Tmp.tOptSignalX);
        FreeCplx(sfftM_Tmp.fOptSignalX);

        if(TypeSimulation == VECTOR)
	  {
           FreeCplx(sfftM_Tmp.tOptSignalY);
           FreeCplx(sfftM_Tmp.fOptSignalY);
	  }        
     }
       
} // ## end GenerateSignalStringGaussianOrSechWithPhaseShift()

// ######################################################

void OptSignal::GenerateSignalStringRaisedCosinesWithPhaseShift(
                double PhaseShiftAmount)
{

  // Updated for Physics Convention JZFeb03

  // Written by John Zweck, Feb 20th 2001, by significantly adapting 
  // code written by Ivan Lima.

  // Fixed for finite extinction ratio by Ivan Lima (9/29/02) 

  // WARNING: THIS METHOD NEEDS TO BE REWORKED SUBSTANTIALLY
  // TO PRODUCE SMOOTH PULSES. IT IS A STOP-GAP MEASURE FOR NOW
  // But it is at least as good as GenerateSignalStringRaisedCosines

   if (! MemAllocated) 
    {
      cerr << "Error: Object of class OptSignal not initialized." 
           << endl;
      exit(1);
    }   

   
   double Envelope;
   double Phase;   // For Chirped pulses
   double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system
   cplx SingleChannelWave;

    for (int ii = 0; ii < qtPoints; ii++) 
      {
        sfftM.tOptSignalX[ii] = 0.0;
        sfftM.tOptSignalY[ii] = 0.0;
        sfftM.fOptSignalX[ii] = 0.0;
        sfftM.fOptSignalY[ii] = 0.0;

      } // ## end for ii-loop


   for(int ch=0;ch<NumChannels;ch++)
   {

   for (int ii = 0; ii < qtPoints; ii++) 
   {
      SingleChannelWave = 0;

      for(int kk=0;kk<StringLength;kk++)
	{

          double PulseCenter = (kk+0.5)*BitLength;

          //if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")
          // {    
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenter = ii*DeltaTime - PulseCenter;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenter < -TimeWindow/2)
	            TimeRelPulseCenter += TimeWindow;
             while( TimeRelPulseCenter >= TimeWindow/2)
	            TimeRelPulseCenter -= TimeWindow;

         switch(TypePulse)
	   {
	     
	      case CRZ_RCOS: //## CRZ Raised Cosine (Ruo-Mei Mu)
 
		// ## FIX: Should we take sqrt of envelope - YES!!
		// ## FIX: Add filtering for RiseTime
                // ## NOTE: Width currently determined by BitRateChann
		// ## rather than by FWHM_PULSE !!

                if(abs(TimeRelPulseCenter) < BitLength/2.0)
                   {
                     Envelope = sqrt(0.5*(1.0+cos(2*pi*BitRateChann*
                                             TimeRelPulseCenter )));
		                
                      Phase = ChirpParameter*pi*
		  	      cos(2*pi*BitRateChann*TimeRelPulseCenter);
		   }
                else
		  {
                    Envelope = 0.0;
	   	    Phase = 0.0;
		  }    

	        break;

               case CRZ_SIN_MZ: //## Sinusoidally driven Mach-Zehnder
		  //  RZ pulse format
		  // (Kath, Yang and Evangelides)
		  // "Optimal Prechirping for DM transmission of RZ pulses"
		  // A conference paper probably OFC or ECOC, ThQ4-1 / 249
		  // This pulse models the sort of RZ pulse you would get
		  // by using a sinusoidal voltage to modulate a 
		  // Mach-Zehnder interferometer.
		
                if(abs(TimeRelPulseCenter) < 1.0/(2.0*BitRateChann))
                   {
                      Envelope = sqrt(0.5*(1.0+cos(pi*
				  sin(pi*TimeRelPulseCenter*BitRateChann))));
                                             
                      Phase = ChirpParameter*pi*
		  	      cos(2*pi*TimeRelPulseCenter*BitRateChann);
                              
                   }
                else
		  {
                    Envelope = 0.0;
	   	    Phase = 0.0;
		  }    

	       break;

	   case CW:
	     break;        
 
              default:
		cerr << "OptSignal::GenerateSignalStringRaisedCosines:"
		     << endl
		     << "Invalid TypePulse"
		     << "ABORTING NOW!!"
		     << endl;

                    exit(1);

	   } // ## end switch(TypePulse)

         //## Extinction ratio fixed by Ivan Lima (9/29/02)
         if (oBitString->GetBit(ch,kk)) {    
   	    SingleChannelWave += sqrt(OptPowerChann)
	                        *Envelope*exp(jc*Phase);
	 } else { //## A space
   	    SingleChannelWave += sqrt(OptPowerChann*LinearExtinctionRatio)
	                        *Envelope*exp(jc*Phase);
	 }      
	 //SingleChannelWave += Envelope*exp(jc*Phase);

    //} // ## end if bit is a mark 
   } // ## end for kk-loop

      // At this point SingleChannelWave includes the contributions
      // from all the pulses at ii*DeltaTime

      // Next we incorporate the ExtinctionRatio into SingleChannelWave
      // and update  tOptSignal

  
      /*************
         // ## First incorporate the ExtinctionRatio

	 // ## Incorporation of power carried by the "0"-bits using the 
	 // ## ExtinctionRatio

	 // ## See [FOCS, p. 176] for theory.

	 // ## The following equation changes the envelope for a "1" (mark)
         // ## so that
	 //
         // ## (1) It has the background power level of a "0" (space):
         // ##     OptPowerChann*LinearExtinctionRatio, and
         // ## (2) The pulse shape for a "1" is added to this background
	 // ##     power level in such as way that the peak power is still
         // ##     OptPowerChann.
	 // 
	 // ##  Note that if you want to eliminate the background level
	 // ##  you should choose LinearExtinctionRatio very small and 
	 // ## positive, i.e ExtinctionRatio in dB very large and negative
	 // ## ExtinctionRatio = -1000 dB works fine

        double AmplitudeNew = abs(SingleChannelWave);

	  if(AmplitudeNew == 0)
	    SingleChannelWave = sqrt(OptPowerChann*LinearExtinctionRatio);
	  else
	  {

           AmplitudeNew = sqrt ( sq(AmplitudeNew) * OptPowerChann
                   *(1. - LinearExtinctionRatio)
                   +OptPowerChann*LinearExtinctionRatio );

          SingleChannelWave = (AmplitudeNew/abs(SingleChannelWave))*
	                           SingleChannelWave;

	  } // ## end else
      *************/

      // #### Next add the SingleChannelWave to the WDM Signal:

      // ## Get Jones vector for current channel

      // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

   //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*(ii)*DeltaTime;
  
      // JZFeb03: Updated for Physics convention
   
     ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;


      if(ch%2==0)
	{
	  // Add over phase shift

          ChannelPhaseOffset += 2*pi*PhaseShiftAmount/360;
	}


      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.

    
      if(TypePulse == CW)
	{
	  SingleChannelWave = 1.0;
	}

      sfftM.tOptSignalX[ii] += GetInJones(ch,0)*SingleChannelWave*
	                       exp(- jc*ChannelPhaseOffset); 
        
      if(TypeSimulation == VECTOR)
         sfftM.tOptSignalY[ii] += GetInJones(ch,1)*SingleChannelWave*
	                          exp(- jc*ChannelPhaseOffset);

     } // end for ii loop to incorporate Extinction Ratio and update tOptSignal

   } // ## end for ch-loop

  Time2Freq();

} // ## end GenerateSignalStringRaisedCosinesWithPhaseShift

// ######################################################


void OptSignal::InitializeFirstOrderPMDSignal(double PowerSplit, double DGD)
{
   
  // JZFeb03: Partially updated for Physics convention
  // ChannelPhase ofset is correct
  // But we need to check that the fast and slow axes are right too
  // compared to full PMD model

  cerr << "OptSignal::InitializeFirstOrderPMDSignal" << endl
       << "ABORTING: Not completely converted to Physics convention" << endl;
  exit(1);

  // See documentation in ocsOptSignal.hh
 
 double EnvelopeX, EnvelopeY;

 double PhaseX, PhaseY;   // For Chirped pulses

 double ChannelPhaseOffset; // Used to set the frequency of signal
                              // for multi-channel system

 cplx SingleChannelWaveX, SingleChannelWaveY;

 AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 0.0;

 for (int ii = 0; ii < qtPoints; ii++) 
   {
     sfftM.tOptSignalX[ii] = 0.0;
     sfftM.tOptSignalY[ii] = 0.0;
     sfftM.fOptSignalX[ii] = 0.0;
     sfftM.fOptSignalY[ii] = 0.0;

   } // ## end for ii-loop

 for(int ch=0;ch<NumChannels;ch++)
    {

      for (int ii = 0; ii < qtPoints; ii++)
	{ 
          SingleChannelWaveX = 0.0;
          SingleChannelWaveY = 0.0;

         for(int kk=0;kk<StringLength;kk++)
	 {   
	  // With this definition of PulseCenter the first
	  // bit slot goes from time = 0 to time = BitLength
	  // and the first pulse is centered at time = 0.5*BitLength.

          double PulseCenter = (kk+0.5)*BitLength;

          if (oBitString->GetBit(ch,kk)) //## Bit is a Mark ("1")
           {    
             // ## Attention: the Width FWHM is of the Power.
             // ## The terms below are of the field (not of the Power).
             // ## FWHM does not apply to NRZ, but the RiseTime
         
          
             double TimeRelPulseCenterX = ii*DeltaTime - PulseCenter - DGD/2.0;
             double TimeRelPulseCenterY = ii*DeltaTime - PulseCenter + DGD/2.0;

             // To make all pulses exact translates of a single pulse
	     // we need to wrap TimeRelPulseCenter to be in range
	     // [-TimeWindow/2,TimeWindow/2)

             while( TimeRelPulseCenterX < -TimeWindow/2)
	            TimeRelPulseCenterX += TimeWindow;
             while( TimeRelPulseCenterX >= TimeWindow/2)
	            TimeRelPulseCenterX -= TimeWindow;

             while( TimeRelPulseCenterY < -TimeWindow/2)
	            TimeRelPulseCenterY += TimeWindow;
             while( TimeRelPulseCenterY >= TimeWindow/2)
	            TimeRelPulseCenterY -= TimeWindow;


	 switch(TypePulse)
	   {

	       case NRZ: 
	      
	       if(TimeRelPulseCenterX < BitLength/2.0 && 
                  TimeRelPulseCenterX >= -BitLength/2.0)
                        EnvelopeX = 1.0;
                     else
                        EnvelopeX = 0.0;

	       if(TimeRelPulseCenterY < BitLength/2.0 && 
                  TimeRelPulseCenterY >= -BitLength/2.0)
                        EnvelopeY = 1.0;
                     else
                        EnvelopeY = 0.0;

               PhaseX = 0.0;
	       PhaseY = 0.0;

               break;

	


             case RZ_GAUSS:
	    
                     // ## FWHM = 2*sqrt(2*log(2)) * sigma_power

	            EnvelopeX = exp(-sq( TimeRelPulseCenterX
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    EnvelopeY = exp(-sq( TimeRelPulseCenterY
                               *sqrt(2*log(2))/FWHM_Pulse ));

                    PhaseX = 0.0;
                    PhaseY = 0.0;

                    break;
        
	     case RZ_SECH:  
                     // ## FWHM = 2*log(1+sqrt(2)) * sigma_power  
          
                     EnvelopeX = 1.0/cosh( TimeRelPulseCenterX
                              /(FWHM_Pulse/(2*log(1+sqrt(2)) )) );
                   
                     EnvelopeY = 1.0/cosh( TimeRelPulseCenterY
                              /(FWHM_Pulse/(2*log(1+sqrt(2)) )) );

                     PhaseX = 0.0;
                     PhaseY = 0.0;

                     break;

	   case CRZ_GAUSS:

		     EnvelopeX   = exp( -2*log(2)*
			              sq( TimeRelPulseCenterX/FWHM_Pulse ));
 
		     EnvelopeY   = exp( -2*log(2)*
			              sq( TimeRelPulseCenterY/FWHM_Pulse ));
 
                  
		     // The ChirpParameter here is precisely the chirp
		     // parameter, C, in [NLO,p67].

                     PhaseX = -ChirpParameter*2*log(2)*
			      sq( TimeRelPulseCenterX/FWHM_Pulse );
 
                     PhaseY = -ChirpParameter*2*log(2)*
			      sq( TimeRelPulseCenterY/FWHM_Pulse );
 

	             break;

	
         
              default:
		cerr << "TypePulse Error in " << endl
		     << "OptSignal::InitializeFirstOrderPMDSignal"
		     << endl;
			 
                    exit(1);

	   } // ## end switch(TypePulse)

   
	 SingleChannelWaveX += EnvelopeX*exp(jc*PhaseX);
	 SingleChannelWaveY += EnvelopeY*exp(jc*PhaseY);

    } // ## end if bit is a mark 
   } // ## end for kk-loop


      // At this point SingleChannelWave includes the contributions
      // from all the pulses at ii*DeltaTime



      // #### Next add the SingleChannelWave to the WDM Signal:

      // ## Compute the Phase Offset for current channel

      // ## We need the 2*pi factor to convert from Hz = cycles per second to
      // ## Radians per second

   //ChannelPhaseOffset = 2*pi*(CenterFreq - GetFrequency(ch))*(ii)*DeltaTime;

	 // JZFeb03: Updated for Physics convention

     ChannelPhaseOffset = 2*pi*(GetFrequency(ch) - CenterFreq)*(ii)*DeltaTime;


      // FIX: For SCALAR simulations just initialize tOptSignalX and 
      // omit Jones vector.


      sfftM.tOptSignalX[ii] += sqrt(OptPowerChann*PowerSplit)*
                               SingleChannelWaveX*
	                       exp(- jc*ChannelPhaseOffset); 
                                     
      sfftM.tOptSignalY[ii] += sqrt(OptPowerChann*(1.0 - PowerSplit))*
                               SingleChannelWaveY*
	                       exp(- jc*ChannelPhaseOffset);


     } // end for ii loop to update tOptSignal

   } // ## end for ch-loop


   Time2Freq();

// ## Constraint of the rise time  for NRZ

   if (RiseTime != 0 && TypePulse == NRZ) 
    { 

      optical_filter(sfftM.tOptSignalX, sfftM.fOptSignalX, fft,
                     3/(RiseTime*pi)/(2*log(2)) ); 
     
      optical_filter(sfftM.tOptSignalY, sfftM.fOptSignalY, fft,
                     3/(RiseTime*pi)/(2*log(2)) );   
                  
    } // end if
    

} // ## end InitializeFirstOrderPMDSignal

// ##############################################

//!   This  method shifts the signal to the right 
//!   by the amount dTime using the fft

void OptSignal::TimeShiftSignal(double dTime)
{

   fftshift();
      
   if(TypeSimulation==SCALAR)
     {
      for (int ii=0; ii<qtPoints; ii++) 
       sfftM.fOptSignalX[ii] *= 
        exp(jc*2.*pi*(ii-qtPoints/2)*dTime/TimeWindow);
     }
   else
     {

       for (int ii=0; ii<qtPoints; ii++) 
        sfftM.fOptSignalX[ii] *= 
         exp(jc*2.*pi*(ii-qtPoints/2)*dTime/TimeWindow);

       for (int ii=0; ii<qtPoints; ii++) 
        sfftM.fOptSignalY[ii] *= 
         exp(jc*2.*pi*(ii-qtPoints/2)*dTime/TimeWindow);
          
     }

   ifftshift();
   
   Freq2Time();
   
}

// ###################################################

void OptSignal::FreqShiftSignal(double dFreq)
{

   if(TypeSimulation==SCALAR)
     {
      for (int ii=0; ii<qtPoints; ii++)
       sfftM.tOptSignalX[ii] *=
        exp(-jc*2.*pi*ii*dFreq/FreqWindow);
     }
   else
     {
      for (int ii=0; ii<qtPoints; ii++)
       sfftM.tOptSignalX[ii] *=
        exp(-jc*2.*pi*ii*dFreq/FreqWindow);

      for (int ii=0; ii<qtPoints; ii++)
       sfftM.tOptSignalY[ii] *=
        exp(-jc*2.*pi*ii*dFreq/FreqWindow);


     }

   Time2Freq();

}


// #######################################################

void OptSignal::SetBitRateChannBM(double Value) 
{
  //  By: Brian Marks 01/2004
  //  Changes the bit rate of the channel of the optical signal.
  //  WARNING!!:  Be extremely careful using this function!  It was 
  //  built for a specific application!!!

  if (Value <= 0.0) 
    {
      cout << "ERROR: OptSignal::SetBitRateChannBM(double)" << endl;
      cout << "New bit rate value is 0 or less!  Not clever!" << endl;
      cout << "ABORTING NOW." << endl;
      exit(1);
    }

  AccumulatedTotalNoiseSpectralDensityGaussianStatistics = 0.0;

  BitRateChann = Value;
  BitLength = 1.0 / BitRateChann;
  TimeWindow = StringLength * BitLength;
  DeltaTime = TimeWindow / qtPoints;
  DeltaFreq = 1.0 / TimeWindow;
  FreqWindow = DeltaFreq * qtPoints;

  ResetAccumulatedFirstOrderDispersionToZero();
  ResetAccumulatedSecondOrderDispersionToZero();

  if(UsingOptFiberLocalError)
    {
      ComputeRelevantNumModesOFLE();
    }


  FreeCplx(sfftM.tOptSignalX);
  FreeCplx(sfftM.fOptSignalX);
  FreeCplx(sfftM.tOptSignalY);
  FreeCplx(sfftM.fOptSignalY);

  cout << "0.." << flush;
  DeallocateOptFiberLocalErrorWorkSpace();
  cout << "1.." << flush;

  delete fft;
  fft = new cfftw(qtPoints, TimeWindow);
  cout << "2.." << flush;
 
  sfftM.tOptSignalX = NewCplx(qtPoints);
  sfftM.fOptSignalX = NewCplx(qtPoints);
  sfftM.tOptSignalY = NewCplx(qtPoints);
  sfftM.fOptSignalY = NewCplx(qtPoints);
  
  cout << "3.." << flush;

  AllocateOptFiberLocalErrorWorkSpace();

  cout << "4.." << flush;

  RegenerateSignalString();

  cout << "Completed!" << endl;

} // ## end SetBitRateChannBM

// ##############################################

void OptSignal::AddSignalBM(OptSignal * Signal2)
{

  // SHOULDN'T NUMCHANNELS AND FREQUENCY OF Signal2 BE PRIVATE???

  if(GetQtPoints() != Signal2->GetQtPoints() ||
     GetStringLength() != Signal2->GetStringLength() ||
     GetBitRateChann() != Signal2->GetBitRateChann() ||
     GetTypeSimulation() != Signal2->GetTypeSimulation() ||
     GetCenterFreq() != Signal2->GetCenterFreq() )
    {
      cerr << "OptSignal::AddSignalBM" << endl
	   << "Input Signals incompatable" << endl
	   << "Aborting now!!" << endl;
      exit(1);
    }

  NumChannels += Signal2->NumChannels;  // NB!!
  ChannSpacing =  min(ChannSpacing,Signal2->ChannSpacing); // NB!!
  oBitString->AddBitStringBM(*Signal2->oBitString);

  /*
  if(DebugLevel >=3)
    {
     Signal2->oBitString->WriteBitString();
     oBitString->WriteBitString();
    }
  */

  StringLength = oBitString->GetStringLength();
  
  for(int ch=NumChannels-Signal2->NumChannels;ch<NumChannels;ch++)
    {
      Frequency[ch] = Signal2->Frequency[ch-NumChannels+
					     Signal2->NumChannels];
    }

   for(int ch = NumChannels-Signal2->NumChannels; ch < NumChannels; ch++)
     {
       for(int mm = 0; mm < 4; mm++)
	 SetInStokes(ch,mm,
		     Signal2->GetInStokes(ch-NumChannels+Signal2->NumChannels,mm));
       
       for(int mm = 0; mm < 2; mm++)
	 SetInJones(ch,mm,
		    Signal2->GetInJones(ch-NumChannels+Signal2->NumChannels,mm));
       
     }


  for(int jj = 0; jj < qtPoints; jj++)  {
    sfftM.fOptSignalX[jj] += Signal2->sfftM.fOptSignalX[jj];
    sfftM.fOptSignalY[jj] += Signal2->sfftM.fOptSignalY[jj];
  }
  Freq2Time();

}


// ##############################################



void OptSignal::AddCWSignal(double PowerCWChann, double FreqCWChann)
{

  // Should change this to make the CW signal a new Channel with a 
  // true channel index and everything.

  //  cout << "Relative FreqCWChann = " << CenterFreq - FreqCWChann << endl;

  for(int jj = 0; jj < qtPoints; jj++)  {
    double Time = jj*DeltaTime;
    sfftM.tOptSignalX[jj] += sqrt(PowerCWChann)*
      exp(2.0*pi*jc*(CenterFreq - FreqCWChann)*Time);
    sfftM.tOptSignalY[jj] += sqrt(PowerCWChann)*
      exp(2.0*pi*jc*(CenterFreq - FreqCWChann)*Time);
  }
  Time2Freq();
}


// ##############################################












