// $Id: ocsRamanAmp.cc,v 1.1.1.1 2004/06/11 22:56:02 zweck Exp $

//###################################################################
//
//           Optical Communication Systems Simulator
//
//       Copyright (2001):
//       Optical Fiber Communications Laboratory (OFCL)
//       Computer Science & Electrical Engineering Department (CSEE)
//       University of Maryland Baltimore County (UMBC)
//
//###################################################################

/*

relaxation method should be tested.
Z index is all from 0(left) to Z(right) 

GA should be test for DeltaZ


*/

#include "ocsRamanAmp.hh"
// JONATHON: ocs.cc and ocs.hh are not part of the ocs library
// and so you should not include ocs.hh here. 
//#include "ocs.hh"

extern ofstream *LogFile;


// FIX SCALAR case!!
// ###############################
// to solve Raman equation, please refer to our JOSAB submission. This class contains
// Relaxation method and shooting method


// ################### Methods for the class OptFilter

// #### Constructor of the class OptFilter
// ## By: Jonathan Hu
RamanAmp::RamanAmp()
{		
   ObjectInitialized = 0;
}

// ##############Destructor of the class RamanAmp
// ## By: Jonathan Hu
RamanAmp::~RamanAmp()
{
   if (ObjectInitialized)
        ReleaseMemory();
}


// ###### Releases  the memory
// ###### By: Jonathan Hu

void RamanAmp::ReleaseMemory()
{

 
 delete [] oSignalWaveLth ;
 delete [] oSignalPower ;
 delete [] oPumpWaveLth ;
 delete [] oPumpPower ;
 delete [] oPumpTarget ;
 delete [] oPumpTargetBackup;
 
// delete [] oSignalPowerPre ;
// delete [] oPumpPowerPre ; 

// delete [] DeltaPsignal ;
// delete [] DeltaPpump ;
 
// delete [] DeltaPpumpTotal ;
// delete [] DeltaPsignalTotal ;
// delete [] oSignalTrack;
 delete [] oIntgSigPower;
// delete [] oPumpTrack;
// delete [] oPumpPower45Converted;
		
 delete [] oPumpIntgTarget;
// delete [] IntgPumPower;
 
 delete [] oRamanGain;
 delete [] oLoss;
 delete [] FiberLocation;
 
// delete [] oSignalSample;
 delete [] SigFEnergy;
 
 delete [] PreCorrectP;
 
 

 delete [] DeltaZArray;
 
 
  //  delete [] CompCoeff;

}


//  #######################################

/*
if DeltaWavelength is larger than 152e-9, the gain will be assume to 0; it
should be updated later with more accurate value
*/


// ##############################################
//#####################################################
//## This initialization is for signal class input

RamanAmp::RamanAmp(string Job, string InFileName, RanNumGen *RNG2, 
	OptSignal *oOptSignal1, int NumFibersPumped2)
{
  NoFourierSignal = 0;
 oOptSignal = oOptSignal1;
  

 fft = oOptSignal->fft;
 sfftM = oOptSignal->sfftM;
 qtPoints = fft->give_N();
 DeltaFreq = fft->delta_f;
 FreqWindow = DeltaFreq*qtPoints;
 DeltaTime = 1/FreqWindow;      
 TimeWindow = oOptSignal->GetTimeWindow();


 if(oOptSignal->GetTypeSimulation() != 1)
  {
    cerr << "Error: Raman Class only work for SCALAR_NLS now!!!" << endl 
           << " Abort!!!"<< endl;
    exit(1);
  }
  // Created by Jonathan Hu 13th Nov, 2001
  // The basic constructor for the class
 char str[40]; 
 JobRaman = Job;
 RNG = RNG2;
 NumFibersPumped = NumFibersPumped2;
 ifstream InFile(InFileName.c_str(),ios::in);    
   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName
           << " could not be opened."<< endl;
    exit(1);
    }    
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   

 TypeAlgorithm = ReadInt("TypeAlgorithm"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 TypeGeneticAlgorithms = ReadInt("TypeGeneticAlgorithms"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 TypeIntergration = ReadInt("TypeIntergration"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 TypeRamanNoise  = typeRamanNoise(ReadInt("TypeRamanNoise"," ",0,2,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

*LogFile << "TypeRamanNoise = " << WriteTypeRamanNoise() << endl; 

 
 TypeCorrect = ReadInt("TypeCorrect"," ",0,-1,
                                     LOWER_ONLY,&InFile);

 WeightCorrect = ReadDouble("WeightCorrect"," ",0,-1,
                                     LOWER_ONLY,&InFile); 
 PumpPowerFactor = ReadDouble("PumpPowerFactor"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 IterationAccurateTh = ReadDouble("IterationAccurateTh"," ",0,-1,
                                     LOWER_ONLY,MANDATORY,&InFile);
// AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
 DeltaZArray = new double[NumFibersPumped];
 AmpLength = 0;

// 	AmpLength =  ReadDouble("AmpLength"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
    DeltaZConst = ReadDouble("DeltaZConst"," ",0,-1,
                                     LOWER_ONLY,&InFile);
//	NumZSteps = int (AmpLength/DeltaZConst + 0.00001 ) + 1;    // NumZSteps should +1 the calculate two end points

 // NumSigCh = 10;
 NumPumCh = ReadInt("NumPumCh"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 NumIteration = ReadInt("NumIteration"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 NumIntItr = ReadInt("NumIntItr"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 Population = ReadInt("Population"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 SelectionFixedNum = ReadInt("SelectionFixedNum"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 CrossRate = ReadDouble("CrossRate"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 MutationRate = ReadDouble("MutationRate"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 StartNumSpace = ReadInt("StartNumSpace"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 EndNumSpace = ReadInt("EndNumSpace"," ",0,-1,
                                    LOWER_ONLY,&InFile);
									
 NumSigCh =  oOptSignal->GetQtPoints();
 NumASECh = NumSigCh;
 NumSRBCh = NumSigCh;
 NumDRBCh = NumSigCh; 
 
 *LogFile << " TypeIntergration = " << TypeIntergration << endl;

 *LogFile << " TypeCorrect = " << TypeCorrect << endl;
 *LogFile << " WeightCorrect = " << WeightCorrect << endl;
 *LogFile << " IterationAccurateTh = "  << IterationAccurateTh  << endl; 
// *LogFile << " AttenuationFiber = " << AttenuationFiber << endl;
 // *LogFile << "AmpLength = "  << AmpLength  << endl;
// *LogFile << "DeltaZ = "  << DeltaZ  << endl;				    
 *LogFile << "NumSigCh = " << NumSigCh  << endl;
 *LogFile << "NumPumCh = "  <<  NumPumCh  << endl;
 *LogFile << "NumIteration = "  << NumIteration  << endl;
 
// AttenuationFiber /= 4.343;   // convert to linear
 

/* NumZSteps right now it just a integer, it will be fit for 
  different case later*/
 

 NumCh = NumSigCh + NumPumCh; // do something here to NumCh if it is changable
 
// NumZSteps = int (TAmpLength/DeltaZ + 0.00001 ) + 1;    // NumZSteps should +1 the calculate two end points

// CompCoeff = new double[NumIteration];

// set up the PreCorrectP


// a test in paper
/*fft = oOptSignal->fft;
 sfftM = oOptSignal->sfftM; 
 qtPoints = fft->give_N();
 DeltaFreq = fft->delta_f;
 FreqWindow = DeltaFreq*qtPoints;
 DeltaTime = 1/FreqWindow;      
 TimeWindow = oOptSignal->GetTimeWindow();

 fft->fftshift(sfftM.fOptSignalX);

 for(int ii=0;ii<NumSigCh;ii++)  {
   	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() );
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] ;
      oSignalPower[ii*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[ii*NumZSteps]==0)
	  		oSignalPower[ii*NumZSteps] = oSignalPower[(ii-1)*NumZSteps];
	  *LogFile << "    oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
 }
 
 fft->ifftshift(sfftM.fOptSignalX);
*/ 
 oPumpWaveLthInput = new double [NumPumCh];
 oPumpPowerInput = new double [NumPumCh];
 
 for(int ii=0;ii<NumPumCh;ii++)
   {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLthInput[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpPower%d",ii);				    
      oPumpPowerInput[ii] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
	  *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLthInput[ii] << endl;   
      *LogFile << "    oPumpPower" << ii << " = " << 
                          oPumpPowerInput[ii] << endl;  
   }

     
   InFile.close();


 oRamanGain = new double [NumFibersPumped*300];  // number 300 should be a varible later.
// PumpAttenuation = new double[NumFibersPumped];
 FiberLocation = new int[NumFibersPumped];
 oLoss = new double [NumFibersPumped*300];
 NumFiberIndex=0;


 // input the gain coefficient
/* this is for Raman Gain input
 ifstream InFile1(InFileName1.c_str(),ios::in);    
 if (!InFile1) 
  {
    cerr << "Error: Input file "<< InFileName1 
           << " could not be opened."<< endl;
    exit(1);
    }


 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
 NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile1);
 WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile1);
  
 *LogFile << "NumGainWidth = " << NumGainWidth << endl;
 *LogFile << "WavelengthStepGain = " << WavelengthStepGain << endl;
  
 for(int ii=0;ii<NumGainWidth;ii++)
    {   
       sprintf(str,"%d",ii+1452);
//         cout << str << endl;
       oRamanGain[0*300+ii] = ReadDouble(str," ",0,-1,NO_BOUNDS,&InFile1);
 //           cout << oRamanGain[ii] << endl;
       oRamanGain[0*300+ii] = oRamanGain[0*300+ii] * 2 / 1000;    

    }
   InFile1.close();

*/
/* ifstream InFile2(InFileName2.c_str(),ios::in);    
 if (!InFile2) 
  {
    cerr << "Error: Input file "<< InFileName2 
           << " could not be opened."<< endl;
    exit(1);
    }


 for(int i=0;i<NumZSteps;i++)
    {   
       sprintf(str,"%d",i*45);
 //      oSignalSample[i] = ReadDouble(str," ",0,-1,NO_BOUNDS,&InFile2);
//    cout  << oSignalSample[i] << endl;
	}
   InFile2.close();
  */ 

   AccurateSign=0;
   ObjectInitialized = 1;
   BlowupSign=0;
   FixSignalShootingSign = 0;
   FixPumpShootingSign = 0;
   if(TypeGeneticAlgorithms != 0)
   		FixPumpShootingSign =1;  
    
	
	cout << " NumPumCh=" << NumPumCh << endl;

} // ## end RamanAmp constructor

// ##############################################

// #### Initializes the object Raman Amplifier
// ## for CW signal input, mostly used for debug
// ## By: Jonathan Hu


RamanAmp::RamanAmp(string Job, string InFileName, RanNumGen *RNG2, int NumFibersPumped2)
{
  // Created by Jonathan Hu 13th Nov, 2001
  // The basic constructor for the class

  NoFourierSignal = 1;
 char str[40]; 
 JobRaman = Job;
 RNG = RNG2;
 NumFibersPumped = NumFibersPumped2;
 InFileNameSave = InFileName;

 ifstream InFile(InFileNameSave.c_str(),ios::in);    
   
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }
    
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   

 TypeAlgorithm = ReadInt("TypeAlgorithm"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 TypeGeneticAlgorithms = ReadInt("TypeGeneticAlgorithms"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 								  
 
 TypeIntergration = ReadInt("TypeIntergration"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 TypeRamanNoise  = typeRamanNoise(ReadInt("TypeRamanNoise"," ",0,2,
                               LOWER_AND_UPPER,MANDATORY,&InFile));

 *LogFile << "TypeRamanNoise = " << WriteTypeRamanNoise() << endl; 
 TypeCorrect = ReadInt("TypeCorrect"," ",0,-1,
                                     LOWER_ONLY,&InFile);

 WeightCorrect = ReadDouble("WeightCorrect"," ",0,-1,
                                     LOWER_ONLY,&InFile); 
 IterationAccurateTh = ReadDouble("IterationAccurateTh"," ",0,-1,
                                     LOWER_ONLY,MANDATORY,&InFile); 
// AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
// AmpLength =  ReadDouble("AmpLength"," ",0,-1,
 //                                   LOWER_ONLY,&InFile);
// DeltaZ = ReadDouble("DeltaZ", " ", 0,-1,
 //                                   LOWER_ONLY,&InFile);				    
// AttenuationFiber = ReadDouble("AttenuationFiber"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
// AmpLength =  ReadDouble("AmpLength"," ",0,-1,
  //                                  LOWER_ONLY,&InFile);
// DeltaZ = ReadDouble("DeltaZ", " ", 0,-1,
//                                    LOWER_ONLY,&InFile);				    
 DeltaZArray= new double[NumFibersPumped];
 AmpLength = 0;
 // 	AmpLength =  ReadDouble("AmpLength"," ",0,-1,
//                                    LOWER_ONLY,&InFile);
    DeltaZConst = ReadDouble("DeltaZConst"," ",0,-1,
                                     LOWER_ONLY,&InFile);
//	NumZSteps = int (AmpLength/DeltaZConst + 0.00001 ) + 1;    // NumZSteps should +1 the calculate two end points
  
 
 NumSigCh =  ReadInt("NumSigCh"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 NumASECh = NumSigCh;
 NumSRBCh = NumSigCh; 
 NumDRBCh = NumSigCh;// NumSigCh = 10;
 NumPumCh = ReadInt("NumPumCh"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 NumIteration = ReadInt("NumIteration"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 NumIntItr = ReadInt("NumIntItr"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 Population = ReadInt("Population"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 SelectionFixedNum = ReadInt("SelectionFixedNum"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 CrossRate = ReadDouble("CrossRate"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 MutationRate = ReadDouble("MutationRate"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 StartNumSpace = ReadInt("StartNumSpace"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 EndNumSpace = ReadInt("EndNumSpace"," ",0,-1,
                                    LOWER_ONLY,&InFile);

								
 *LogFile << " TypeIntergration = " << TypeIntergration << endl;
 *LogFile << " TypeCorrect = " << TypeCorrect << endl;
 *LogFile << " WeightCorrect = " << WeightCorrect << endl;
 *LogFile << " IterationAccurateTh = "  << IterationAccurateTh  << endl;
// *LogFile << " AttenuationFiber = " << AttenuationFiber << endl;
 *LogFile << "AmpLength = "  << AmpLength  << endl;
// *LogFile << "DeltaZ = "  << DeltaZ  << endl;				    
 *LogFile << "NumSigCh = " << NumSigCh  << endl;
 *LogFile << "NumPumCh = "  <<  NumPumCh  << endl;
 *LogFile << "hereNumIteration = "  << NumIteration  << endl;
 
// AttenuationFiber /= 4.343;   // convert to linear
 

/* NumZSteps right now it just a integer, it will be fit for 
  different case later*/
 NumCh = NumSigCh + NumPumCh;
 
// NumZSteps = int (TAmpLength/DeltaZ + 0.00001 ) + 1;    // NumZSteps should +1 the calculate two end points
// *LogFile << "NumZSteps = "  << NumZSteps  << endl;
// cout << "NumZSteps   = " << NumZSteps << endl;




// set up the PreCorrectP


// a test in paper

/*
 if(NumSigCh>8){
   for(int ii=0;ii<NumSigCh;ii++)   {
      if(NumSigCh==100){
	  		oSignalWaveLth[ii] = (1525 +  double(ii) * 80 / NumSigCh ) * 1e-9;
	  }else{
	  		oSignalWaveLth[ii] = 1550e-9 + 0.8e-9 / 2 - 0.8e-9 * NumSigCh / 2
										+ ii * 0.8e-9;
	  }
      oSignalPower[ii*NumZSteps] = 0.50118e-3;
//      *LogFile << "oSignalWaveLth" << ii << " = " << 
//                          oSignalWaveLth[ii] << endl;
     
//      *LogFile << "oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps] << endl;
   }
 }
 else{
   for(int ii=0;ii<NumSigCh;ii++)  {
   
      sprintf(str,"SignalWaveLth%d",ii);
      oSignalWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "hereoSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] << endl;
      sprintf(str,"SignalPower%d",ii);
      oSignalPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
   }
 }
 */
/* oPumpWaveLthInput = new double [NumPumCh];
 oPumpPowerInput = new double [NumPumCh];
 
 for(int ii=0;ii<NumPumCh;ii++)
   {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpPower%d",ii);				    
      oPumpPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
     *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLth[ii] << endl;   
      *LogFile << "oPumpPower" << ii << " = " << 
                          oPumpPower[ii*NumZSteps] << endl;  
   }     
   InFile.close();
*/
 oPumpWaveLthInput = new double [NumPumCh];
 oPumpPowerInput = new double [NumPumCh];
 
 for(int ii=0;ii<NumPumCh;ii++)
   {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLthInput[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpPower%d",ii);				    
      oPumpPowerInput[ii] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
	  *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLthInput[ii] << endl;   
      *LogFile << "    oPumpPower" << ii << " = " << 
                          oPumpPowerInput[ii] << endl;  
   }

   InFile.close();

 oRamanGain = new double[NumFibersPumped*300];  // number 300 should be a varible later.
// PumpAttenuation = new double[NumFibersPumped];
 FiberLocation = new int[NumFibersPumped];
 oLoss = new double[NumFibersPumped*300];
 NumFiberIndex = 0;
 // input the gain coefficient



/*
 ifstream InFile1(InFileName1.c_str(),ios::in);    
 if (!InFile1) 
  {
    cerr << "Error: Input file "<< InFileName1 
           << " could not be opened."<< endl;
    exit(1);
    }

 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
 NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile1);
 WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile1);
  
 *LogFile << "NumGainWidth = " << NumGainWidth << endl;
 *LogFile << "WavelengthStepGain = " << WavelengthStepGain << endl;

 oRamanGain = new double[300]; 
 for(int ii=0;ii<NumGainWidth;ii++)
    {   
       sprintf(str,"%d",ii+1452);
//         cout << str << endl;
       oRamanGain[ii] = ReadDouble(str," ",0,-1,NO_BOUNDS,&InFile1);
 //           cout << oRamanGain[ii] << endl;
       oRamanGain[ii] = oRamanGain[ii] * 2 / 1000;    

    }
 InFile1.close();
*/
/* ifstream InFile2(InFileName2.c_str(),ios::in);    
 if (!InFile2) 
  {
    cerr << "Error: Input file "<< InFileName2 
           << " could not be opened."<< endl;
    exit(1);
    }


 for(int i=0;i<NumZSteps;i++)
    {   
       sprintf(str,"%d",i*45);
 //      oSignalSample[i] = ReadDouble(str," ",0,-1,NO_BOUNDS,&InFile2);
//    cout  << oSignalSample[i] << endl;
	}
   InFile2.close();
*/
//   fft = oOptSignal->fft;
//   sfftM = oOptSignal->sfftM;


    AccurateSign = 0;
    ObjectInitialized = 1;
    BlowupSign=0;
    FixSignalShootingSign = 0;
	FixPumpShootingSign = 0;   
    if(TypeGeneticAlgorithms != 0)
   		FixPumpShootingSign =1;  
  
   
     

} // ## end RamanAmp constructor

// ##############################
// ## By: Jonathan Hu

void RamanAmp::RamanIteration()
{
 //   cout << "test fiber loss=" << GetAttenuationFiber(1555E-9)*4.343;
	if(NoFourierSignal == 1){
		PrepareSignalNonFourier();
	}else {
		if(NumSigCh > 10)
			PrepareSignalStart();
		else
			PrepareSignalCW();
 	}
	
	switch(TypeAlgorithm)
     {  // should do lot of correction to fit Relaxation method for Gary's loop
       case 1:   // Relaxation Method

	   //	 InitRamanIteration();
	  	 cout << "Running Raman Relaxation Method" << endl;
	     RelaxRamanIteration();
		 	  
		 break;
     
       case 2:    // Shooting Method
		 
		 switch(TypeGeneticAlgorithms)
		 	{
			case 0: //No GA
    		   	 cout << "Running Raman Shooting Method" << endl;

		 		if(NumPumCh == 1 && oPumpPower[0] < 1e-9)
				{	for(int ii = 0; ii< NumZSteps; ii++ )
					 	oPumpPower[ii]	= 0;
				}else {
					
				}
				
				ShootingRamanIteration();
				break;
				
			case 1: //Victor's GA
				 DeltaZConst = DeltaZArray[0];
				ShootingGAVictor();
			//	ShootingGeneticAlgorithmsVictor();
				break;
				
			case 2: //Equal space GA
				 //  < (1525e-9 - 1450e-9) * 2 / (NumPumCh-1)
				 DeltaZConst = DeltaZArray[0];
				ShootingGA();
			//	ShootingGeneticAlgorithms(1.39286e-08);				
				
				break;
			
			case 3: // Equal Space Vary Wc
				 DeltaZConst = DeltaZArray[0];
				ShootingGAVaryCW();
				break;
		    
			case 4:
				ShootingIntegrationExample(1);
				break;
				
			default:
         		cerr << "With TypeGeneticAlgorithms must be 0, 1 or 2" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
		 	}
			
		 break;
       	  
       default:
         cerr << "With type Algorithm must be 1 or 2" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
 
 cout << "Last" << endl;
  if(NoFourierSignal == 0){
 	CallLastRound();
	}
// for (int ii=0;ii<NumSigCh;ii++){
// cout << oSignalPower[ii*NumZSteps]  << "->" << 10*log(oSignalPower[ii*NumZSteps+NumZSteps-1]*1e3)/log(10) << "   ";
// }
 

/*
 cout  << "NumSigCh=" << NumSigCh << endl
  		<< "NumPumCh=" << NumPumCh << endl
	    << "wavelength=" << oSignalWaveLth[0] << " gain=" << GetRamanGain(oSignalWaveLth[0],0) << endl
		<< "Input0 = " << oSignalPower[0] << " = " << 10 * log(oSignalPower[0]*1000)/ log(10) << "dbm" << endl
  
  //  	 <<	"Middle7 = " << oSignalPower[FiberLocation[0]+1] << "=dbm" << 10 * log(oSignalPower[FiberLocation[0]+1]*1000)/ log(10) << endl
  // 		<< "Middle11 = " << oSignalPower[FiberLocation[1]+1] << "=dbm" << 10 * log(oSignalPower[FiberLocation[1]+1]*1000)/ log(10) << endl
   		
		
		<< "Output = " << oSignalPower[NumZSteps-1] << "=dBm" << 10 * log(oSignalPower[NumZSteps-1]*1000)/ log(10) << endl
  		<< " Gain Coeff=" << GetRamanGain(oSignalWaveLth[0],0) 
		<< "  Loss=" <<  GetAttenuationFiber(oSignalWaveLth[0],0) 
		<< " Amplength=" << AmpLength  << endl
		<< " Gain or Loss =" << 10 * log(oSignalPower[NumZSteps-1] /  
				(oSignalPower[0]  * exp(-1 * GetAttenuationFiber(oSignalWaveLth[0],1) * AmpLength) ) )/ log(10) << endl;
*/
} // end RamanIteration
// ########################################
// used for debug
// ## By: Jonathan Hu
void RamanAmp::PrepareSignalCW()
{
/*  NumSigCh = 1;
  oSignalWaveLth[0] = 1549e-9; 
  oSignalPower[0] = -7.0;
  oSignalPower[0] = exp(oSignalPower[0] / 10 * log(10)) * 1e-3;
*/

NumSigCh=1;
NumCh = NumSigCh + NumPumCh;

oSignalWaveLth[0] = 1.55252e-06;    //oSignalPower[0*NumZSteps] = 0.000341088;
/*oSignalWaveLth[1] = 1.55092e-06;    oSignalPower[1*NumZSteps] = 0.000341085;
oSignalWaveLth[2] = 1.54932e-06;    oSignalPower[2*NumZSteps] = 0.000341084;
oSignalWaveLth[3] = 1.54772e-06;    oSignalPower[3*NumZSteps] = 0.000341085;
oSignalWaveLth[4] = 1.54612e-06;    oSignalPower[4*NumZSteps] = 0.000341088;
*/

/* for(int ii=0;ii<NumPumCh;ii++)
   {
      oPumpWaveLth[ii] = oPumpWaveLthInput[ii];
      oPumpPower[ii*NumZSteps] = 0.08317 *	oPumpPowerInput[ii];
         // two spliters each at -3.6db
	  oPumpTarget[ii] = oPumpPower[ii*NumZSteps];
	  oPumpTargetBackup[ii] = oPumpPower[ii*NumZSteps];
   }
 */
}  // end of PrepareSignal



// ########################################
// to accelerate caculation, I first calculate the Raman gain for total power of each channel 
// instead of each Fourier component, after solveing the pump evolution, I run one last iteration
// to find out gain for all Fourier component.
// // ## By: Jonathan Hu
void RamanAmp::PrepareSignalStart()
{
 double MaxPower=0;
 int CountNumSigCh=0;	
	


 fft->fftshift(sfftM.fOptSignalX);
 
 
 for(int ii=0;ii<qtPoints;ii++)  {
    if( sq(abs(sfftM.fOptSignalX[ii])) > MaxPower){
		MaxPower = sq(abs(sfftM.fOptSignalX[ii]));
	}
 }
/*
 CountNumSigCh = 0; 
 
 for(int ii=0;ii<qtPoints;ii++)  {
      
	if(  sq(abs(sfftM.fOptSignalX[ii])) > MaxPower * 9e-3){
	  
	  
   	  oSignalWaveLth[CountNumSigCh] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-qtPoints/2)*oOptSignal->GetDeltaFreq() );
      *LogFile << "oSignalWaveLth" << CountNumSigCh << " = " << 
                          oSignalWaveLth[CountNumSigCh] ;
      oSignalPower[CountNumSigCh*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[CountNumSigCh*NumZSteps] < 1e-100)
			oSignalPower[CountNumSigCh*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
	  *LogFile << "    oSignalPower" << CountNumSigCh << " = " << 
                          oSignalPower[CountNumSigCh*NumZSteps]  << endl;
						  
	  CountNumSigCh ++;
	}
 }
*/

 int NFour;
 CountNumSigCh = oOptSignal->GetNumChannels();
 if(CountNumSigCh == 2 * int( CountNumSigCh/2)){
 	cerr << "With NeumChannels should be an odd number" << endl 
    	<< "Aborting now! " << endl;
    exit(1);
 }	
 NFour = int(oOptSignal->GetChannSpacing() / oOptSignal->GetDeltaFreq() / 2);
// cout << " NFour = " << NFour << endl;
 for(int ii=0;ii<CountNumSigCh;ii++)  {
      
	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-(CountNumSigCh - 1)/2)*oOptSignal->GetChannSpacing() );
//      *LogFile << "oSignalWaveLth" << ii << " = " << oSignalWaveLth[ii] ;
//	  cout << "oSignalWaveLth" << ii << " = " << oSignalWaveLth[ii] ;

      oSignalPower[ii*NumZSteps] = 0;
	  for(int jj=-NFour;jj<NFour-1;jj++){
	  	oSignalPower[ii*NumZSteps]+=sq(abs(sfftM.fOptSignalX[qtPoints/2+(ii-(CountNumSigCh-1)/2)*2*NFour+jj]));
//	  	cout << "   " << qtPoints/2+(ii-(CountNumSigCh-1)/2)*2*NFour+jj;
	  }
//      cout << "  mid = " << qtPoints/2+(ii-(CountNumSigCh-1)/2)*2*NFour+0 ;
      if(oSignalPower[ii*NumZSteps] < 1e-100)
			oSignalPower[ii*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
//	  *LogFile << "    oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps]  << endl;
//	  cout << "    oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps]  << endl;
//	  exit(1);
 }



 fft->ifftshift(sfftM.fOptSignalX);

 for(int ii=0;ii<NumPumCh;ii++)
   {
      oPumpWaveLth[ii] = oPumpWaveLthInput[ii];
      oPumpTarget[ii] = 0.08317 *	oPumpPowerInput[ii]; 
	  // two 3.6dB spliters
	   *LogFile << "    oPumpWaveLth" << ii << " = " << oPumpWaveLthInput[ii]; 
	   *LogFile << "    oPumpPower" << ii << " = " << oPumpPowerInput[ii];
	  oPumpTargetBackup[ii] = oPumpTarget[ii];
   }
   
 NumSigCh = CountNumSigCh;  
 NumCh = NumSigCh + NumPumCh;
 cout << "NumSig run = " << CountNumSigCh << endl;

 
}  // end of PrepareSignal

// #################################################
// prepare signal for CW signal input, used for debug
// ## By: Jonathan Hu
void RamanAmp::PrepareSignalNonFourier()
{
 char str[40]; 
 ifstream InFile(InFileNameSave.c_str(),ios::in);    
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileNameSave 
           << " could not be opened."<< endl;
    exit(1);
    }
    
   cout << NumSigCh << endl;

 if(NumSigCh>8){
   for(int ii=0;ii<NumSigCh;ii++)   {
      if(NumSigCh==100){
	  		oSignalWaveLth[ii] = (1525 +  double(ii) * 80 / NumSigCh ) * 1e-9;
	  }else{
	  		oSignalWaveLth[ii] = 1550e-9 + 0.8e-9 / 2 - 0.8e-9 * NumSigCh / 2
										+ ii * 0.8e-9;
	  }
      oSignalPower[ii*NumZSteps] = 0.50118e-3;
      *LogFile << "oSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] << endl;
     
//      *LogFile << "oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps] << endl;
   }
 }
 else{
   for(int ii=0;ii<NumSigCh;ii++)  {
   
      sprintf(str,"SignalWaveLth%d",ii);
      oSignalWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "hereoSignalWaveLth" << ii << " = " << 
                          oSignalWaveLth[ii] << endl;
      sprintf(str,"SignalPower%d",ii);
      oSignalPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
      *LogFile << "oSignalPower" << ii << " = " << 
                          oSignalPower[ii*NumZSteps] << endl;
   }
 } 
 
/* for(int ii=0;ii<NumPumCh;ii++)
   {
      sprintf(str,"PumpWaveLth%d",ii);
      oPumpWaveLth[ii] = ReadDouble(str, " ",0,-1,
                                    LOWER_ONLY,&InFile);
 
      sprintf(str,"PumpPower%d",ii);				    
      oPumpPower[ii*NumZSteps] = ReadDouble(str," ",0,-1,
                                    LOWER_ONLY,&InFile);
     *LogFile << "oPumpWaveLth" << ii << " = " <<
                             oPumpWaveLth[ii] << endl;   
      *LogFile << "oPumpPower" << ii << " = " << 
                          oPumpPower[ii*NumZSteps] << endl;  
	  oPumpTarget[ii] = oPumpPower[ii*NumZSteps];
	  oPumpTargetBackup[ii] = oPumpPower[ii*NumZSteps];

   }  
*/ InFile.close();

 for(int ii=0;ii<NumPumCh;ii++)
   {
      oPumpWaveLth[ii] = oPumpWaveLthInput[ii];
      oPumpTarget[ii] = 	oPumpPowerInput[ii]; 
	  // two 3.6dB spliters
	   *LogFile << "    oPumpWaveLth" << ii << " = " << oPumpWaveLthInput[ii]; 
	   *LogFile << "    oPumpPower" << ii << " = " << oPumpPowerInput[ii];
	  oPumpTargetBackup[ii] = oPumpTarget[ii];
   }
   
 NumCh = NumSigCh + NumPumCh;




 } // Prepare signal for Non Fourier 
// ######################################
// Recover all Fourier component for last iteration
// ## By: Jonathan Hu

void RamanAmp::PrepareSignalLastRound()
{

 fft->fftshift(sfftM.fOptSignalX);
 
 NumSigCh = qtPoints;
 
 for(int ii=0;ii<qtPoints;ii++)  {
   	  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
	  						+(ii-qtPoints/2)*oOptSignal->GetDeltaFreq() );
//      *LogFile << "oSignalWaveLth" << ii << " = " << 
//                          oSignalWaveLth[ii] ;
      oSignalPower[ii*NumZSteps] = sq(abs(sfftM.fOptSignalX[ii]));
      if(oSignalPower[ii*NumZSteps] < 1e-100)
			oSignalPower[ii*NumZSteps] = 1e-60;                   //oSignalPower[(ii-1)*NumZSteps];
//	  *LogFile << "    oSignalPower" << ii << " = " << 
//                          oSignalPower[ii*NumZSteps] << endl;
 }
 
 fft->ifftshift(sfftM.fOptSignalX);

 if(TypeRamanNoise != RAMAN_NOISE_OFF)
 {

  for(int ii=0;ii<NumASECh;ii++)
    {
      oASEWaveLth[ii] = oSignalWaveLth[ii];  
	  oSRBWaveLth[ii] = oSignalWaveLth[ii];
	  oDRBWaveLth[ii] = oSignalWaveLth[ii];

 	  oASEPower[ii * NumZSteps ]=0;
	    
 	  oSRBPower[ii * NumZSteps + NumZSteps -1]=0;
	  oDRBPower[ii * NumZSteps ]=0;
//	  cout << "oNoiseWaveLth[ii]=" << oNoiseWaveLth[ii]  << endl;
	}

 }



} // end of PrepareSignal

// ##########################################
// input gain data, 
// ## By: Jonathan Hu
void RamanAmp::SetGainProfile(double FiberLength, string InFileName)
{
 char str[40];
 ifstream InFile(InFileName.c_str(),ios::in);    
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
 NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile);
 WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile);
  cout << "if you want to use NumGainWidth make sure it is debuged" << endl;
 *LogFile << "NumGainWidth = " << NumGainWidth << endl;
 *LogFile << "WavelengthStepGain = " << WavelengthStepGain << endl;

 for(int ii=0;ii<NumGainWidth;ii++)
    {   
       sprintf(str,"%d",ii+1452);
//         cout << str << endl;
       oRamanGain[NumFiberIndex*300+ii] = ReadDouble(str," ",0,-1,NO_BOUNDS,&InFile);
 //           cout << oRamanGain[ii] << endl;
//       oRamanGain[NumFiberIndex*300+ii] = oRamanGain[NumFiberIndex*300+ii]   ;    
//		if(NumFiberIndex==0)	{oRamanGain[NumFiberIndex*300+ii] *= 4 ;}
//		else{oRamanGain[NumFiberIndex*300+ii] /= 4;}
//       oRamanGain[NumFiberIndex*300+ii] *=  4.7856 / 1000;   

	//	cout <<	" before=" << oRamanGain[NumFiberIndex*300+ii] ;
        oRamanGain[NumFiberIndex*300+ii] =  oRamanGain[NumFiberIndex*300+ii] * 2 / 1000 ;   
//		cout <<	"   after=" << oRamanGain[NumFiberIndex*300+ii] << endl;
       // FOR DEBUG		if(NumFiberIndex==0)	{oRamanGain[NumFiberIndex*300+ii] *= 4 ;}
       // FOR DEBUG		else{oRamanGain[NumFiberIndex*300+ii] /= 4;}
//		cout << oRamanGain[NumFiberIndex*300+ii] << "  ";

    }
 
	//cout << endl;
 InFile.close();
 cout << "run to here 3.90" << endl;
 int NumZ=0;
 do{
 	NumZ++;
 }while( FiberLength/NumZ > DeltaZConst);
 cout << "run to here 3.91" << endl;
 DeltaZArray[NumFiberIndex] = FiberLength/NumZ;
   cout << "run to here 3.92" << endl;
 if(NumFiberIndex==0){
 	FiberLocation[NumFiberIndex] = NumZ-1;
 }else{
 	FiberLocation[NumFiberIndex] = FiberLocation[NumFiberIndex-1] + NumZ;
 }

 cout 	<< "FiberLength = " << FiberLength 
 		<< "   FiberLocation[" << NumFiberIndex << "] = " <<  FiberLocation[NumFiberIndex] 
 		<< "   DeltaSave[" << NumFiberIndex << "] = " <<  FiberLength/NumZ << endl ;

 NumFiberIndex++;
 

 AmpLength += FiberLength;

 
} // ## end SetGainProfile

// #########################################################
// since we may have different fiber in one Amplifier span. so the DeltaZ steps in each 
// fiber are different
// ## By: Jonathan Hu
double RamanAmp::DeltaZ(int NumSteps)
{  int FIndex = -1;
   if(NumSteps > NumZSteps || NumSteps < 0)
   {
       cerr << "Error: Step is not in the correct range"<< endl
	   		<< "Aborting" << endl;
    exit(1);
   }
   if(NumSteps == NumZSteps || NumSteps == 0)
   {//	cout << "it return 0 for NumSteps = " << NumSteps << endl;
   		
   		return(0); 
   }else{
   		NumSteps--;
   		do{
   			FIndex++;
   		}while(NumSteps > FiberLocation[FIndex]);
   
   		return(DeltaZArray[FIndex]);
	}

} // end of DeltaZ


// #############################################################
// since we have different fiber in amp span. When we initialize the construction, we don't know how 
// long is the fiber, so we have to allocate Memory according to different fiber length in one amplifier span
// after fiber parameters are given
// ## By: Jonathan Hu
void RamanAmp::AllocateMemory()
{
 if(NumFiberIndex != NumFibersPumped)
 {
 	cerr << " Input Num of Fiber is not consistent " << endl
		<< " NumFiberIndex=" << NumFiberIndex << " NumFibersPumped=" << NumFibersPumped << endl
		<< " aborting....." << endl;
	exit(1);
 }

 
 NumZSteps = FiberLocation[NumFibersPumped-1] + 2 ;
 cout << "   FiberLocation[" << NumFibersPumped-1 << "] = " <<  FiberLocation[NumFibersPumped-1] << endl;
 *LogFile << "NumZSteps = "  << NumZSteps  << endl;
 cout << "NumZSteps   = " << NumZSteps << endl;
 cout << "NumPumCh = " << NumPumCh << endl;
 oSignalWaveLth = new double[NumSigCh];
 oSignalPower = new double[NumSigCh*(NumZSteps+1)];
 oASEWaveLth = new double[NumASECh]; 
 oSRBWaveLth = new double[NumSRBCh]; 
 oDRBWaveLth = new double[NumDRBCh]; 
 oASEPower = new double[NumASECh*(NumZSteps+1)];
 oSRBPower = new double[NumSRBCh*(NumZSteps+1)];
 oDRBPower = new double[NumDRBCh*(NumZSteps+1)];
 oPumpWaveLth = new double[NumPumCh];
 oPumpTarget = new double[NumPumCh];
 oPumpTargetBackup= new double[NumPumCh];
 oPumpIntgTarget= new double[NumPumCh];
  
 oPumpPower = new double[NumPumCh*(NumZSteps+1)];
 
// oSignalTrack = new double[NumZSteps*(NumIteration+1)];
 oIntgSigPower = new double[NumSigCh];
// oPumpTrack = new double[NumZSteps*NumPumCh];
// oPumpPower45Converted = new double[NumPumCh*(NumZSteps+1)];
 SigFEnergy = new double[NumZSteps];

// oSignalPowerPre = new double[NumSigCh*(NumZSteps+1)];
// oPumpPowerPre = new double[NumPumCh*(NumZSteps+1)]; 

// DeltaPsignal = new double[NumSigCh];
// DeltaPpump = new double[NumPumCh];
 
// DeltaPpumpTotal = new double[NumIteration];
// DeltaPsignalTotal = new double[NumIteration];
 
 PreCorrectP = new double[NumPumCh];
 
// oSignalSample = new double[NumZSteps];
 
 for(int ii=0;ii<NumPumCh;ii++)
 {
    PreCorrectP[ii] = 100;
 }

} // end of AllocateMemory




//  #####################################################
// read in Victor's gain data format
// ## By: Jonathan Hu
void RamanAmp::ReadPlainFileGainLoss(double FiberLength, string InFileName)
{
 char str[40];
  string InFileNameGain,InFileNameLoss;
  InFileNameGain = InFileName + "Gain.in";
 ifstream InFileGain(InFileNameGain.c_str(),ios::in);   
 double Keff=2; 
 if (!InFileGain) 
  {
    cerr << "Error: Input file "<< InFileNameGain 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();

 *LogFile << InFileNameGain << endl << endl;   
 
// NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile);
// WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile);


// InFile->clear(); // Resets ifstream state to good
// InFile->seekg(0); 
  FreqStartGain = 0;
  FreqStepGain = 0.2e12;
  NumGainWidth=206;

 for(int ii=0;ii<NumGainWidth;ii++)
    {   

       InFileGain >> oRamanGain[NumFiberIndex*300+ii]; 
  //	   cout << oRamanGain[NumFiberIndex*300+ii] << "  ";
	   
	   oRamanGain[NumFiberIndex*300+ii] = oRamanGain[NumFiberIndex*300+ii] / (10*80*Keff) ;
//	   cout << oRamanGain[NumFiberIndex*300+ii] << "  " ;
	   // 80 is effitive Area, should rewrite the code later for this issue.
 //      cout << oRamanGain[NumFiberIndex*300+ii] << endl;
    }

	//cout << endl;
 InFileGain.close();

 InFileNameLoss = InFileName + "Loss.in";
 ifstream InFileLoss(InFileNameLoss.c_str(),ios::in);   

 if (!InFileLoss) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

//  PumpWavelengthForGain ;
  WaveLengthStartLoss = 1.4e-6;
  WavelengthStepLoss = 10e-9;
  NumLossWidth = 34;
  
 LogFileSeparator();
 *LogFile << InFileNameLoss << endl << endl;   
 
// NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile);
// WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile);


// InFile->clear(); // Resets ifstream state to good
// InFile->seekg(0); 

 for(int ii=0;ii<NumLossWidth;ii++)
    {   

       InFileLoss >> oLoss[0*300+ii]; 
  //	   cout << oRamanGain[NumFiberIndex*300+ii] << "  ";
	   
	   oLoss[0*300+ii] = oLoss[0*300+ii] * 1e-3/ 4.343 ;
//	   *LogFile << 1370e-9+ii*10e-9 << "   "  << oLoss[0*300+ii] * 4.343 * 1e+3 << endl;
	   // 80 is effitive Area, should rewrite the code later for this issue.
 //      cout << oRamanGain[NumFiberIndex*300+ii] << endl;
    }

	//cout << endl;
 InFileLoss.close();

/*
 for(int ii=0;ii<NumGainWidth;ii++)
    {   
		*LogFile << ii*0.2;
 //		*LogFile << Frequency2Wavelength(Wavelength2Frequency(1450e-9)-ii*0.2*1e12)-1450e-9;
		*LogFile << "   " << oRamanGain[NumFiberIndex*300+ii] * (10*80*Keff) << endl; 
  

    } 

//cout << Wavelength2Frequency(1450e-9) << "   " << Wavelength2Frequency(1550e-9);
//cout << endl << Wavelength2Frequency(1450e-9) - Wavelength2Frequency(1550e-9);

exit(1);
*/


 int NumZ=0;
 do{
 	NumZ++;
 }while( FiberLength/NumZ > DeltaZConst);

 DeltaZArray[NumFiberIndex] = FiberLength/NumZ;
 if(NumFiberIndex==0){
 	FiberLocation[NumFiberIndex] = NumZ-1;
 }else{
 	FiberLocation[NumFiberIndex] = FiberLocation[NumFiberIndex-1] + NumZ;
 }


 cout 	<< "FiberLength = " << FiberLength 
 		<< "   FiberLocation[" << NumFiberIndex << "] = " <<  FiberLocation[NumFiberIndex] 
 		<< "   DeltaSave[" << NumFiberIndex << "] = " <<  FiberLength/NumZ << endl ;
 *LogFile 	<< "FiberLength = " << FiberLength 
 		<< "   FiberLocation[" << NumFiberIndex << "] = " <<  FiberLocation[NumFiberIndex] 
 		<< "   DeltaSave[" << NumFiberIndex << "] = " <<  FiberLength/NumZ << endl ;
 NumFiberIndex++;
 AmpLength += FiberLength;
 if(NumFiberIndex == NumFibersPumped)
 {
 	cout << "AmpLength = " << AmpLength << endl;
 }
   ReadPlainFileGainLossFlag = 1;
 ReadRamanGainLossWavelengthGridFlag = 0; 
} // ## end ReadPlainFileGain

//  #####################################################
//  #####################################################
// read in Gaston's experimental data format
// ## By: Jonathan Hu

void RamanAmp::ReadRamanGainLossWavelengthGrid(double FiberLength, string InFileName)
{
 char str[40];
 double temp;
 ifstream InFile(InFileName.c_str(),ios::in);   
 double Keff=2; 
 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
// NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile);
// WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile);
  InFile >>  PumpWavelengthForGain;
  InFile >>  WaveLengthStartGain;
  InFile >>  WavelengthStepGain;
  InFile >>  NumGainWidth;
//  cout <<    PumpWavelengthForGain <<  "  " <<  WaveLengthStartGain << "  " 
//  		<< WavelengthStepGain << "  " <<  NumGainWidth << endl;
 if (PumpWavelengthForGain != 1455) 
  {
    cerr << "Error: Input file :: "<< InFileName 
           << " is not in the right format."<< endl;
    exit(1);
    }
 for(int ii=0;ii<NumGainWidth;ii++)
    {   
		InFile >> temp;
	   if(temp < 1400 || temp > 1700){
	   		cerr << "Error: in wavelength input."<< endl;
    		exit(1);
    	}
	
       InFile >> oRamanGain[NumFiberIndex*300+ii]; 
	 //  oRamanGain[NumFiberIndex*300+ii] = oRamanGain[NumFiberIndex*300+ii]  ;
//	   oRamanGain[NumFiberIndex*300+ii] /= SpliterLossFactor;
//	   oRamanGain[NumFiberIndex*300+ii] *= 1.00;


//		not useful
//	   Since I use 3W Pump and -16dbm signal to assume no pump depletion to get Raman
//		gain coeff. but in the real experment. the pump depletion is there. so I do some correction
//		to reproduce the experment.	
	   
	   
	   *LogFile << oRamanGain[NumFiberIndex*300+ii] << "  " ;
	   InFile >> oLoss[NumFiberIndex*300+ii];
	   *LogFile << oLoss[NumFiberIndex*300+ii] << "  " << endl ;
	   oLoss[NumFiberIndex*300+ii] /= 4343;
//	   cout << oRamanGain[NumFiberIndex*300+ii] << endl;
    }
 WaveLengthStartLoss=WaveLengthStartGain;
 WavelengthStepLoss=WavelengthStepGain;	
 NumLossWidth = NumGainWidth;
// PumpAttenuation[NumFiberIndex] = ReadDouble("PumpAttenuation"," ",0,-1,
 PumpAttenuation = ReadDouble("PumpAttenuation"," ",0,-1,
                                    LOWER_ONLY,&InFile);
 *LogFile << " PumpAttenuation=" << PumpAttenuation << endl;
// PumpAttenuation[NumFiberIndex] /= 4343;
 PumpAttenuation/= 4343;
 
 RayleighBackscatterCoeff = ReadDouble("RayleighBackscatterCoeff"," ",0,-1,
                                    LOWER_ONLY,MANDATORY,&InFile);
 *LogFile << "RayleighBackscatterCoeff =" <<  RayleighBackscatterCoeff<< endl; 
 
 InFile.close();

 int NumZ=0;
 do{
 	NumZ++;
 }while( FiberLength/NumZ > DeltaZConst);

 DeltaZArray[NumFiberIndex] = FiberLength/NumZ;
 if(NumFiberIndex==0){
 	FiberLocation[NumFiberIndex] = NumZ-1;
 }else{
 	FiberLocation[NumFiberIndex] = FiberLocation[NumFiberIndex-1] + NumZ;
 }


// cout 	<< "FiberLength = " << FiberLength 
// 		<< "   FiberLocation[" << NumFiberIndex << "] = " <<  FiberLocation[NumFiberIndex] 
// 		<< "   DeltaSave[" << NumFiberIndex << "] = " <<  FiberLength/NumZ << endl ;
 *LogFile 	<< "FiberLength = " << FiberLength 
 		<< "   FiberLocation[" << NumFiberIndex << "] = " <<  FiberLocation[NumFiberIndex] << endl;
 *LogFile << "   DeltaSave[" << NumFiberIndex << "] = " <<  FiberLength/NumZ << endl ;
  NumFiberIndex++;
  AmpLength += FiberLength;
 if(NumFiberIndex == NumFibersPumped)
 {
 	cout << "AmpLength = " << AmpLength << endl;
 } 
 ReadPlainFileGainLossFlag = 0;
 ReadRamanGainLossWavelengthGridFlag = 1;

} // ## end ReadRamanGainWavelengthGrid

//  #####################################################





// ################################
// set all parameters 
/*
void RamanAmp::InitRamanIteration()
{

  for(int i=0;i<NumIteration;i++){
     CompCoeff = 0;
  }
  
  oSignalTrack[0] = oSignalPower[0];	 
  
  for(int i=1;i<NumZSteps;i++)
     {
     for(int NSig=0;NSig<NumSigCh;NSig++){   
      oSignalPower[NSig*NumZSteps+i] = oSignalPower[NSig*NumZSteps+i-1] *
			exp(-GetAttenuationFiber(oSignalWaveLth[NSig],i)* DeltaZ(i) );
	  oSignalPowerPre[NSig*NumZSteps+i] = oSignalPower[NSig*NumZSteps+i];
     }
	 oSignalTrack[i] = oSignalPower[i];	 
     for(int NPum=0;NPum<NumPumCh;NPum++){
	  	oPumpPower[NPum*NumZSteps+i]= oPumpPower[NPum*NumZSteps+i-1] *
			exp(-GetAttenuationFiber(oPumpWaveLth[NPum],NumZSteps-i)*DeltaZ(NumZSteps-i));
		oPumpPowerPre[NPum*NumZSteps+i] = oPumpPower[NPum*NumZSteps+i];
     }  
  }
	  			for(int i=1;i<NumZSteps;i++){
       				SignalOnlyPropagation(i);
    			}
} // end InitRamanIteration


*/


// #######################################
//############# start for Shooting method  #####################
// using shooting algorithm to solve Raman Amp, please refer JOSAB submission
// one solves all of the differential equations simultaneously in the forward 
// direction using an estimate of the pump output powers. Thus, the ODEs are 
// automatically satisfied to the order of accuracy of the numerical integration 
// scheme, and the iteration is carried out until the boundary conditions or 
// some other shooting target is satisfied. No backward integration or 
// decoupling of the system of ODEs is necessary. However, one must iterate the 
// solution until the boundary conditions or some other target conditions 
// that specify the solution uniquely are satisfied.
// ## By: Jonathan Hu

void RamanAmp::ShootingRamanIteration()
{

 double *Shot = new double[NumPumCh*NumIteration];  
 double *Target = new double[NumPumCh*NumIteration]; 
 double *oPumpGuessPre = new double[NumPumCh];
 double *rPumpPower = new double[NumPumCh];
 double *ErrorPum = new double[NumZSteps];
 double *oPumpShooting = new double[NumPumCh];
 double *oPumpTrack = new double[NumZSteps*NumPumCh]; 
 

// double Shot[NumPumCh*NumIteration],Target[NumPumCh*NumIteration];  
// double oPumpGuessPre[NumPumCh];
 char   C_String[40];
 double Energy;
 double rms;
// double rPumpPower[NumPumCh];
 int    NumItr;
 double Alpha,AlphaPre,DeltaAlpha;
// double ErrorPum[NumZSteps];
// double oPumpShooting[NumPumCh];

/* 

string FileName;
 sprintf(C_String,"ShootingPumpNew.dat");
 FileName = JobRaman + C_String; 
 ofstream OutFile(FileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << FileName  << "Could not be created." << endl;
   exit(1);
 }

for(int ii=100;ii<300;ii++) {
    oPumpPower[0] = (double) ii / 10000;
	
    for(int i=1;i<NumZSteps;i++){
          ForwardShooting(i);
    }
 OutFile << oPumpPower[0] <<  "   " << oPumpPower[NumZSteps - 1] << endl;
}
 
 OutFile.close();     
  switch(TypePerturbation){
     	case 0:     // no perturbation algorithm
			NumPumCh = NumPumCh;
			NumCh = NumSigCh + NumPumCh;
			Alpha=1;
  			for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps] = oPumpTargetBackup[NPum];
			} 
	
			for(int i=1;i<NumZSteps;i++){
       			PrePumpShooting(i);
    		}
    		for(int NPum=0;NPum<NumPumCh;NPum++) 	{			  
    			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps+NumZSteps-1] ;
//	    		cout << oPumpPower[NPum*NumZSteps] << "  ";
			}
		break;
		
		case 1:   // Inc Pum Inc Pow
			NumPumCh=2;
		 	NumCh = NumSigCh + NumPumCh;
		  	for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps] = PerturbationRatio * oPumpTargetBackup[NPum];
				} 
		  	for(int i=1;i<NumZSteps;i++){
       			PrePumpShooting(i);
    			}
  			for(int NPum=0;NPum<NumPumCh;NPum++) 	{			  
    			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps+NumZSteps-1] ;
//			    cout << oPumpPower[NPum*NumZSteps] << "  ";
			}
			NumPumCh=7;
		    Alpha = 1;			

			oPumpPower[0 * NumZSteps] = 0.00157699;
			oPumpPower[1 * NumZSteps] = 0.00186702 ;
			oPumpPower[2 * NumZSteps] = 0.00212021;
			oPumpPower[3 * NumZSteps] = 0.00262511;
			oPumpPower[4 * NumZSteps] = 0.00369277;
			oPumpPower[5 * NumZSteps] = 0.00564321;
			oPumpPower[6 * NumZSteps] = 0.0166584;
			//			 0.00157699  0.00186702  0.00212021  0.00262511  0.00369277  0.00564321  0.0166584
			//0.00157698  0.00186701  0.0021202  0.00262511  0.00369277  0.00564319  0.0166583
			//            this data is get from the result 
		break;
	   	
		case 2:   // Inc Pow
//		    cout << "run to here2";


  			for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps] = Alpha * oPumpTargetBackup[NPum];
			}    
			for(int i=1;i<NumZSteps;i++){
       			SignalOnlyPropagation(i);
    		}
			for(int NSig=0;NSig<NumSigCh;NSig++){
				for(int ii=1;ii<NumZSteps;ii++){
//						oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
//							*((oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii]-1)
			//					*ii/(NumZSteps-1)+1);
		//	
	//			oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
							*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
								*ii/(NumZSteps-1));
	
				}
			}
		  	for(int i=1;i<NumZSteps;i++){
       			BackwardStep(i);
    		}					
			


	//		exit(1);
//			for(int i=0;i<NumSigCh;i++){
  //     			cout << oSignalPower[i*NumZSteps+NumZSteps-1] << "  " ;
    //		}
		//	exit(1);
			 
			cout << "Starting oPumpPower" << endl;
  			for(int NPum=0;NPum<NumPumCh;NPum++) 	{			  
    			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps+NumZSteps-1] ;
	//		    cout << oPumpPower[NPum*NumZSteps] << "  ";
	//			cout << "run to here1";
			}
  			char  C_String[40];			
		  	sprintf(C_String,"SignalTrack.dat");
			string FileNametemp = JobRaman + C_String;
			WriteSignalTrack(FileNametemp);			
				
		break;
	  
       	default:
         		cerr << "With PerturbationType must between 0 and 2" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
		break;
      } 
 */ 	
 			AlphaPre=0;
			DeltaAlpha=1;
			Alpha=AlphaPre+DeltaAlpha;


  	for(int i=0;i<NumZSteps;i++){	 
 	 	for(int NPum=0;NPum<NumPumCh;NPum++){	
			oPumpTrack[NPum*NumZSteps+i] = 1;
		}
	}

	do{

/*	switch(TypePerturbation)
		{
     	case 0: 
			cout << "run to here0";   
			oPumpTarget[NumPumCh-1] = Alpha * oPumpTargetBackup[NumPumCh-1];

		case 1: //need to be corrected
			cout << "run to here1";   
			oPumpTarget[NumPumCh-1] = Alpha * oPumpTargetBackup[NumPumCh-1];	  
			if(Alpha <= PerturbationRatio){ 
			 	oPumpPower[(NumPumCh-1)*NumZSteps] = oPumpTarget[NumPumCh-1];
				for(int i=1;i<NumZSteps;i++){
					PrePumpShootingLastOne(i);
				}
				oPumpPower[(NumPumCh-1)*NumZSteps] = oPumpPower[(NumPumCh-1)*NumZSteps+NumZSteps-1] ;
    		}else{
				
				oPumpPower[(NumPumCh-1)*NumZSteps] = oPumpPower[(NumPumCh-1)*NumZSteps] * ( Alpha / (Alpha - PerturbationRatio));
			}
			break;
	   	
		case 2: 
*/
//			cout << "pump target";  
		  for(int NPum=0;NPum<NumPumCh;NPum++) {
	  			oPumpTarget[NPum] = Alpha * oPumpTargetBackup[NPum];
		  }
//	  		if(Alpha>PerturbationRatio){
//	  			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps] * ( Alpha / (Alpha - PerturbationRatio));
//	  		}
	  		if(AlphaPre==0){
			 	for(int NPum=0;NPum<NumPumCh;NPum++) {
		    		oPumpPower[NPum*NumZSteps+NumZSteps-1] = Alpha * oPumpTargetBackup[NPum];
				} 
	  			for(int i=1;i<NumZSteps;i++){
       				SignalOnlyPropagation(i);
    			}
				for(int NSig=0;NSig<NumSigCh;NSig++){
					for(int ii=1;ii<NumZSteps;ii++){
		//					oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
			//				*((oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii]-1)
			//					*ii/(NumZSteps-1)+1);
			
			// this is for fir inital guess
						if(oSignalPower[NSig*NumZSteps]==0)
						{	oSignalPower[NSig*NumZSteps+ii]=0;
						}else{
							oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
											*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
											*ii/(NumZSteps-1));
						}
	    
//				if(NSig==(NumSigCh/2))
	//	           cout << oSignalPower[NSig*NumZSteps+ii] << "  ";
					}
		
				}

		  		for(int i=NumZSteps-2;i>=0;i--){
       				BackwardStep(i);
					if(BlowupSign==1)	{
						i=NumZSteps;
					}
    			}  
				
/*			 	for(int NPum=0;NPum<NumPumCh;NPum++){	
					cout << "pump # " << NPum << endl;
				  	for(int i=0;i<NumZSteps;i++){	 
						cout << oPumpPower[NPum*NumZSteps+i] << "  with stepZ = " << DeltaZ(i) << endl;
					}				}
*/	  		}else{
				for(int NPum=0;NPum<NumPumCh;NPum++) {
					oPumpPower[NPum*NumZSteps] = oPumpGuessPre[NPum] * ( ( Alpha / AlphaPre - 1 )
							*0.6  + 1 );
				}

	  		}
	  	  
			
			
/*		break;
	  
      	default:
         		cerr << "With PerturbationType must between 0 and 2" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      }
*/
//	  cout << " start "; 
//	  for(int NPum=0;NPum<NumPumCh;NPum++){
//	    cout <<  oPumpPower[NPum*NumZSteps] << "    " ;		
//	  }
//	  cout << endl;

//	  AccurateSign =0;
		
// for(int i=0;i<NumZSteps;i++){
//  	SigFEnergy[i] = 0;
// }
// for(int NSig=0;NSig<NumSigCh;NSig++){
//   	SigFEnergy[0] += oSignalPower[NSig*NumZSteps];
// }
	
  for(int i=1;i<NumZSteps;i++){
    if(BlowupSign==0){
	  	ForwardShooting(i);
	}
  }
  AccurateSign=0;


  NumItr = 0;
  do{

   NumItr = NumItr +1;

//   *LogFile << NumItr << "  ";
//	cout << NumItr << "  ";
	

	if(BlowupSign==0){
 	if(NumItr>1){
		if(TypeCorrect==1) {	
      		for(int NPum=0;NPum<NumPumCh;NPum++){
       			Shot[NumItr*NumPumCh + NPum] = oPumpPower[NPum*NumZSteps];
       			Target[NumItr*NumPumCh + NPum] = oPumpPower[NPum*NumZSteps + NumZSteps - 1];

       			if (NumItr < 1) {
          			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps] - 
	            			( oPumpPower[NPum*NumZSteps + NumZSteps - 1] - oPumpTarget[NPum] ); 
       			} 
       			else{
          			oPumpPower[NPum*NumZSteps] = Shot[NumItr*NumPumCh + NPum] + 
		      			( Shot[(NumItr-1)*NumPumCh + NPum] - Shot[NumItr*NumPumCh + NPum]) 
		      			*(oPumpTarget[NPum] - Target[NumItr*NumPumCh + NPum]) 
			  			/ ( Target[ ( NumItr - 1 ) *NumPumCh + NPum] - Target[NumItr*NumPumCh + NPum] )/ NumItr ;
       			}
	   			if (oPumpPower[NPum*NumZSteps] < 0){
						oPumpPower[NPum*NumZSteps] = 1e-5;
		    	}	   
      		}
		}
		else{
			for(int NPum=0;NPum<NumPumCh;NPum++) {
				oPumpShooting[NPum] = oPumpPower[NPum*NumZSteps + NumZSteps - 1] ;	
  			}

			ParallelCorrection(oPumpShooting, oPumpTarget);
//			ParallelCorrection();
//			cout << " pump after correction" << endl;
//			for(int NPum=0;NPum<NumPumCh;NPum++) {
//				cout << oPumpPower[NPum*NumZSteps ] << "  ";	
// 			}
//			cout << endl;			
		}
	} 
	for(int i=1;i<NumZSteps;i++){
        if(BlowupSign==0){
		  	ForwardShooting(i);
		}

    }

	rms =0;
//	cout << "pump track " << endl;
 	for(int NPum=0;NPum<NumPumCh;NPum++){	
//		cout << "pump track # " << NPum << endl;
	  	for(int i=0;i<NumZSteps;i++){	 
	 		rms += sq( ( oPumpPower[NPum*NumZSteps+i] - oPumpTrack[NPum*NumZSteps+i] ) 
					/ oPumpPower[NPum*NumZSteps+i] );
			oPumpTrack[NPum*NumZSteps+i] = oPumpPower[NPum*NumZSteps+i] ;
//			cout << oPumpTrack[NPum*NumZSteps+i] << "  " ;
		}
	}
// cout << endl;


	rms = sqrt( rms / (NumPumCh*NumZSteps) );

	cout << "NumItr = " << NumItr << "   rms=" << rms << endl;
//	cout << rms << "  " << endl;

	if (abs(rms) <IterationAccurateTh){
//		cout <<  "rms="  << rms;
		AccurateSign = 1;	
		}
//	if(NumItr>15)
//		AccurateSign=1;	
	if (abs(rms)>1&&NumItr>1) {
		BlowupSign=1;	
		}
	 
	 }
	} while(AccurateSign==0&&NumItr<500&&BlowupSign==0);
 	
	cout << " BlowupSign="  << BlowupSign ;
	BlowupSign=0;

    if(AccurateSign==1){
		cout << " success with Alpha= " << Alpha  <<  "  NumItr = " << NumItr << endl;
		*LogFile << " success with Alpha= " << Alpha << "  with ShootingItrNum=" << NumItr << endl;
	 	for(int NPum=0;NPum<NumPumCh;NPum++){
		    oPumpGuessPre[NPum]=oPumpPower[NPum*NumZSteps];
		}		
	
		DeltaAlpha *= 1.2599;
		AlphaPre=Alpha;
		Alpha = DeltaAlpha + AlphaPre;
/*		if(AlphaPre>0.8){
			IterationAccurateTh=1e-3;
		}

		if (AlphaPre<0.8){
			IterationAccurateTh=1e-3;
		}	
		if (AlphaPre<0.5){
			IterationAccurateTh=1e-3;
		}	
*/		if(Alpha>=1){
			Alpha=1;
	//		IterationAccurateTh=1e-3;
		}				
	}else{
		cout << "failed with Alpha=" << Alpha << endl;
		*LogFile << "failed with Alpha=" << Alpha <<"  with ShootingItrNum=" << NumItr << endl;
		DeltaAlpha/=2;
		Alpha = DeltaAlpha + AlphaPre;
		if(Alpha>1){
			Alpha=1;
		}
	}


	}while(AlphaPre<1);
  //			AlphaPre=0;
//			DeltaAlpha=1;
//			Alpha=AlphaPre+DeltaAlpha;
	*LogFile << "Shooting Iteration Result Pump" << endl;
	LogFile->precision(15);
 	for(int NPum=0;NPum<NumPumCh;NPum++){	
 		*LogFile << oPumpWaveLth[NPum] << ":   " << oPumpPower[NPum*NumZSteps] << " --> "  << 
				oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl; 
	}
	*LogFile << "Shooting Iteration Result Signal" << endl;
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
 		*LogFile << oSignalWaveLth[NSig] 
			//	<< ":   " << oSignalPower[NSig*NumZSteps] << " --> "  << 
				<< "  " << oSignalPower[NSig*NumZSteps+NumZSteps-1] << endl; 
	}
	LogFile->precision(6);
 delete [] Shot ;  
 delete [] Target ; 
 delete [] oPumpGuessPre ;
 delete [] rPumpPower ;
 delete [] ErrorPum ;
 delete [] oPumpShooting ;
 delete [] oPumpTrack ;

}   // end of ShootingRamanIteration


// ##########################
// last run, including Noise for every Fourier component
// ## By: Jonathan Hu
void RamanAmp::CallLastRound()
{

  cout << "CallLastRound..........."  << endl;
  PrepareSignalLastRound();
  for(int i=1;i<NumZSteps;i++){
  	ForwardLastRound(i);
  } 
  for(int i=NumZSteps - 2; i>= 0;i--){
  	SingalRayleighBScattteringBackward(i);
  }
  for(int i=1;i<NumZSteps;i++){
  	DoubleRayleighBScattteringForward(i);
  }  	
    *LogFile << "compare ASE and DRB" << endl;
  for(int i=0;i<NumASECh;i++){
  	*LogFile << oASEWaveLth[i] << "  "  << oASEPower[i*NumZSteps + NumZSteps - 1] 
			<< "   " << oDRBPower[i*NumZSteps + NumZSteps - 1] << endl;
  }
  /**LogFile << "compare ASE and DRB in Z" << endl;
  double FLength=0;
  for(int i=1;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 *LogFile <<  FLength / 1000 << "   " <<  oASEPower[NumASECh/2*NumZSteps + i] 
			<< "   "  << oSRBPower[NumASECh/2*NumZSteps + i]  
			<< "  " << oDRBPower[NumASECh/2*NumZSteps + i] << endl; ;
  } 
  *LogFile << " end of compare" << endl;
*/

} // end of CallLastRound

// ###################################
// Signal RayleighBScattering Backward
// refer to JOSAB submission equation (1.c)
//// ## By: Jonathan Hu
void RamanAmp::SingalRayleighBScattteringBackward(int StepI)
{
double C1,C2,SRBStepGain;

  if(!(TypeRamanNoise == RAMAN_NOISE_OFF)){
	for(int NSRB=0;NSRB<NumSRBCh;NSRB++){ 
 	   SRBStepGain = 0;
	   C2 = 0;
    	for(int P=0;P<NumPumCh;P++){
       		SRBStepGain +=
       			GetRamanGain(oSRBWaveLth[NSRB],oPumpWaveLth[P],StepI+1)
				*	(oPumpPower[P*NumZSteps+StepI] + oPumpPower[P*NumZSteps+StepI + 1]) / 2;
	
		}
		
	C1 = SRBStepGain - GetAttenuationFiber(oSRBWaveLth[NSRB],StepI + 1);
		C2 =  RayleighBackscatterCoeff * (oSignalPower[NSRB*NumZSteps+StepI] + oSignalPower[NSRB*NumZSteps+StepI + 1]) / 2;
		oSRBPower[NSRB*NumZSteps+StepI] = oSRBPower[NSRB*NumZSteps+StepI + 1] * exp(C1 * DeltaZ(StepI+1)) 
				+ exp(C1 * DeltaZ(StepI+1)) * (  exp(C1 * DeltaZ(StepI+1)) * C2 * DeltaZ(StepI+1));

  	}
  }

} // end of SingalRayleighBScattteringBackward

// ###################################
// DoubleRayleighBScattteringForward
// refer to JOSAB submission equation (1.d)
// ## By: Jonathan Hu
void RamanAmp::DoubleRayleighBScattteringForward(int StepI)
{
	double C1,C2,DRBStepGain;
  if(!(TypeRamanNoise == RAMAN_NOISE_OFF)){	
	for(int NDRB=0;NDRB<NumDRBCh;NDRB++){ 
 	   DRBStepGain = 0;
	   C2 = 0;
    	for(int P=0;P<NumPumCh;P++){
       		DRBStepGain +=
       			GetRamanGain(oDRBWaveLth[NDRB],oPumpWaveLth[P],StepI)
				*(oPumpPower[P*NumZSteps+StepI] + oPumpPower[P*NumZSteps+StepI-1]) / 2;

		
		}
		
	C1 = DRBStepGain - GetAttenuationFiber(oDRBWaveLth[NDRB],StepI);
		C2 = RayleighBackscatterCoeff * (oSRBPower[NDRB*NumZSteps+StepI] + oSRBPower[NDRB*NumZSteps+StepI - 1]) / 2; ;
		oDRBPower[NDRB*NumZSteps+StepI] = oDRBPower[NDRB*NumZSteps+StepI - 1] * exp(C1 * DeltaZ(StepI)) 
				+ exp(C1 * DeltaZ(StepI)) * (  exp(C1 * DeltaZ(StepI)) * C2 * DeltaZ(StepI));

  	}
  }


} // end of DoubleRayleighBScattteringBackward

// ########################################
// add Raman Spontaneous Noise
// refer to JOSAB submission equation (1.b)
// ## By: Jonathan Hu
void RamanAmp::AddNoise()
{

if ( TypeRamanNoise != RAMAN_NOISE_OFF ) { 
   fft->ifftshift(sfftM.fOptSignalX);

   cplx NormalizedNoiseMode = 0;
   double NoiseAmplitudeFactor;
//   for ( int jj = 0; jj < qtPoints; jj++) {
//   		*LogFile << jj << "   " << oASEPower[jj*NumZSteps + NumZSteps - 1] << endl;
//   }
//   exit(1);
//   cout << " convet = " << sfftM.fOptSignalX[qtPoints/2] << " | | =" << sq(abs(sfftM.fOptSignalX[qtPoints/2]));
   for ( int jj = 0; jj < qtPoints; jj++) {

      //DeltaFreq*FreqFactorFFT(jj,qtPoints);        
 
		
         if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
            sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  

         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
 					sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
         	}else if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
//					NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
//					NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
//					sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
//							+ oASEPower[jj*NumZSteps + NumZSteps - 1])
//							 / oSignalPower[jj*NumZSteps]);
			}		 
	 
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
	// 	 
		 if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
         	sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
                    sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         	}else if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
	//				sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
	//						+ oDRBPower[jj*NumZSteps + NumZSteps - 1])
	//						 / oSignalPower[jj*NumZSteps]);			
			
			}		 
//	 			cout << "  SDR=" << NoiseAmplitudeFactor << endl; exit(1);
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
		 if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
					sfftM.fOptSignalX[jj] *= sqrt((sq(abs(sfftM.fOptSignalX[jj]))
							+ oDRBPower[jj*NumZSteps + NumZSteps - 1] + oASEPower[jj*NumZSteps + NumZSteps - 1])
							/ sq(abs(sfftM.fOptSignalX[jj])));			
			
		 }	
                                          
      }
  
//  cout << " conveted = " << sfftM.fOptSignalX[qtPoints/2] << " | | = " <<  sq(abs(sfftM.fOptSignalX[qtPoints/2]))  << endl;
  
  fft->ifftshift(sfftM.fOptSignalX);

  fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX);
  }


} // end of AddNoise()

// #################
// keep track the noise
// this is mainly get the noise spetrum desity use for semi analatic receiver model
// ## By: Jonathan Hu
void RamanAmp::PropagateNoise(OptSignal *Noise)
{

if ( TypeRamanNoise != RAMAN_NOISE_OFF ) { 

   fft->ifftshift(Noise->sfftM.fOptSignalX);
   for ( int jj = 0; jj < qtPoints; jj++) {
         Noise->sfftM.fOptSignalX[jj] = abs(Noise->sfftM.fOptSignalX[jj] * sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1]
		 								/ oSignalPower[jj*NumZSteps]));  
   }

   cplx NormalizedNoiseMode = 0;
   double NoiseAmplitudeFactor;
   for ( int jj = 0; jj < qtPoints; jj++) {
	
         if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
            Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  

         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
 					Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
         	}else if(TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
//					NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
//					NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
//					sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
//							+ oASEPower[jj*NumZSteps + NumZSteps - 1])
//							 / oSignalPower[jj*NumZSteps]);
			}		 
	 
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
	// 	 
		 if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
         	Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
                    Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         	}else if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
	//				sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
	//						+ oDRBPower[jj*NumZSteps + NumZSteps - 1])
	//						 / oSignalPower[jj*NumZSteps]);			
			
			}		 
//	 			cout << "  SDR=" << NoiseAmplitudeFactor << endl; exit(1);
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
		 if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
  					*LogFile << oASEWaveLth[jj] << "  "  << Noise->sfftM.fOptSignalX[jj]
							<< sq(abs(Noise->sfftM.fOptSignalX[jj]))
							<< "   " << oASEPower[jj*NumZSteps + NumZSteps - 1] ;
  
					Noise->sfftM.fOptSignalX[jj] = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj]))
							+ oDRBPower[jj*NumZSteps + NumZSteps - 1] 
							+ oASEPower[jj*NumZSteps + NumZSteps - 1]
							);
						
					*LogFile << "   " << Noise->sfftM.fOptSignalX[jj] 
							<< sq(abs(Noise->sfftM.fOptSignalX[jj])) << endl;
					
					//Noise->sfftM.fOptSignalX[jj] = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj]))
					//	+ oDRBPower[jj*NumZSteps + NumZSteps - 1] 
					//	+ oASEPower[jj*NumZSteps + NumZSteps - 1]
					//	);
					//Noise->sfftM.fOptSignalX[jj] = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
		 }	
                                          
      }
  
//  cout << " conveted = " << sfftM.fOptSignalX[qtPoints/2] << " | | = " <<  sq(abs(sfftM.fOptSignalX[qtPoints/2]))  << endl;

  fft->ifftshift(Noise->sfftM.fOptSignalX);
  fft->IFFT(Noise->sfftM.fOptSignalX,Noise->sfftM.tOptSignalX);
  }
} // end of PropagateNoise()

// #################
// calculate the Noise and signal
// ## By: Jonathan Hu
void RamanAmp::PropagateSignalandNoise(OptSignal *NFSignal, OptSignal *Noise)
{

if ( TypeRamanNoise != RAMAN_NOISE_OFF ) { 

   fft->ifftshift(NFSignal->sfftM.fOptSignalX);
   fft->ifftshift(Noise->sfftM.fOptSignalX);
   for ( int jj = 0; jj < qtPoints; jj++) {
         NFSignal->sfftM.fOptSignalX[jj] *= 
		 	sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1]/ oSignalPower[jj*NumZSteps ]); 
// 		 cout << NFSignal->sfftM.fOptSignalX[jj] << "   " ;
//       	 cout << sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1] / oSignalPower[jj*NumZSteps ]) << "   " ;
         Noise->sfftM.fOptSignalX[jj] = abs(Noise->sfftM.fOptSignalX[jj] * sqrt( oSignalPower[jj*NumZSteps + NumZSteps - 1]
		 								/ oSignalPower[jj*NumZSteps]));  
   }

   cplx NormalizedNoiseMode = 0;
   double NoiseAmplitudeFactor;
   for ( int jj = 0; jj < qtPoints; jj++) {
	
         if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
            Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  

         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
 					Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
         	}else if(TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
//					NoiseAmplitudeFactor = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
//					NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
//					sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
//							+ oASEPower[jj*NumZSteps + NumZSteps - 1])
//							 / oSignalPower[jj*NumZSteps]);
			}		 
	 
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;  
		 
	// 	 
		 if (TypeRamanNoise == RAMAN_NOISE_ON_RANDOM_PHASE ) {
      		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
	        NormalizedNoiseMode = exp(jc*2*pi*RNG->GetRanNum());
         	Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         }else if (TypeRamanNoise == RAMAN_NOISE_ON_GAUSSIAN_WHITE ) {
	        		NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
            		NormalizedNoiseMode = ( RNG->GetGaussianDeviate()
	   		      		+jc*RNG->GetGaussianDeviate() )/sqrt(2.);
                    Noise->sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
         	}else if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
	//				sfftM.fOptSignalX[jj] *= sqrt((oSignalPower[jj*NumZSteps + NumZSteps - 1] 
	//						+ oDRBPower[jj*NumZSteps + NumZSteps - 1])
	//						 / oSignalPower[jj*NumZSteps]);			
			
			}		 
//	 			cout << "  SDR=" << NoiseAmplitudeFactor << endl; exit(1);
//         sfftM.fOptSignalX[jj] += NoiseAmplitudeFactor * NormalizedNoiseMode;
		 if(	TypeRamanNoise == RAMAN_NOISE_ON_NO_RNG ) {	
					// NoiseAmplitudeFactor = sqrt(oDRBPower[jj*NumZSteps + NumZSteps - 1]);
					// NormalizedNoiseMode = sfftM.fOptSignalX[jj] / abs(sfftM.fOptSignalX[jj]);
  					*LogFile << oASEWaveLth[jj] << "  "  << Noise->sfftM.fOptSignalX[jj]
							<< sq(abs(Noise->sfftM.fOptSignalX[jj]))
							<< "   " << oASEPower[jj*NumZSteps + NumZSteps - 1] ;
  
					Noise->sfftM.fOptSignalX[jj] = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj]))
							+ oDRBPower[jj*NumZSteps + NumZSteps - 1] 
							+ oASEPower[jj*NumZSteps + NumZSteps - 1]
							);
						
					*LogFile << "   " << Noise->sfftM.fOptSignalX[jj] 
							<< sq(abs(Noise->sfftM.fOptSignalX[jj])) << endl;
					
					//Noise->sfftM.fOptSignalX[jj] = sqrt(sq(abs(Noise->sfftM.fOptSignalX[jj]))
					//	+ oDRBPower[jj*NumZSteps + NumZSteps - 1] 
					//	+ oASEPower[jj*NumZSteps + NumZSteps - 1]
					//	);
					//Noise->sfftM.fOptSignalX[jj] = sqrt(oASEPower[jj*NumZSteps + NumZSteps - 1]);
		 }	
                                          
      }
  
//  cout << " conveted = " << sfftM.fOptSignalX[qtPoints/2] << " | | = " <<  sq(abs(sfftM.fOptSignalX[qtPoints/2]))  << endl;

  fft->ifftshift(NFSignal->sfftM.fOptSignalX);
  fft->IFFT(NFSignal->sfftM.fOptSignalX,NFSignal->sfftM.tOptSignalX);
  fft->ifftshift(Noise->sfftM.fOptSignalX);
  fft->IFFT(Noise->sfftM.fOptSignalX,Noise->sfftM.tOptSignalX);
  }
} // end of PropagateSignalandNoise()

// ###############################################
// forward last Round including Noise
// ## By: Jonathan Hu
void RamanAmp::ForwardLastRound(int StepI)
{    

/*if (StepI==1)   
 for(int ii=0;ii<NumASECh;ii++)
    {
		      oASEWaveLth[ii] = oSignalWaveLth[ii];  
	  cout << "oNoiseWaveLth[ii]=" << oASEWaveLth[ii]  << "  " << oASEPower[ii*NumZSteps] << endl;
	}
*/


  double *PreSigPower = new double[NumSigCh];

//  double PreSigPower[NumSigCh]; 
  double SignalStepGain, ASEStepGain;
  double monitor;
  double C1,C2;
  
  for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
    SignalStepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       SignalStepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepI)*oPumpPower[P*NumZSteps+StepI-1];
     
//       cout << "Gain = " << GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepI) 
//          << "  oPumpPower["  << P*NumZSteps+NumZSteps-StepI-1 << "] = " << oPumpPower[P*NumZSteps+NumZSteps-StepI-1]
//	  << endl;
//      cout << "SignalStepGainfromP = " << SignalStepGain << endl;
  }

/*  for(int S=0;S<NumSigCh;S++){
    SignalStepGain +=
    GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepI)*oSignalPower[S*NumZSteps+StepI-1]; 
//       cout << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepI) << "SignalStepGainfromS = " << SignalStepGain << endl;
    }
*/
    
    oSignalPower[NSig*NumZSteps+StepI] = oSignalPower[NSig*NumZSteps+StepI-1] * 
        exp((SignalStepGain-GetAttenuationFiber(oSignalWaveLth[NSig],StepI))*DeltaZ(StepI));
	PreSigPower[NSig] = oSignalPower[NSig*NumZSteps+StepI];
  }
//  cout << " before calcu Noise" << endl;
  if(!(TypeRamanNoise == RAMAN_NOISE_OFF)){
	for(int NASE=0;NASE<NumASECh;NASE++){ 
 	   ASEStepGain = 0;
	   C2 = 0;
    	for(int P=0;P<NumPumCh;P++){
       		ASEStepGain +=
       			GetRamanGain(oASEWaveLth[NASE],oPumpWaveLth[P],StepI)*oPumpPower[P*NumZSteps+StepI-1];
			C2 += GetRamanGain(oASEWaveLth[NASE],oPumpWaveLth[P],StepI)*oPumpPower[P*NumZSteps+StepI-1]
					* GetTempDepFactor(oASEWaveLth[NASE],oPumpWaveLth[P]);
//		cout << " TempDept=" << GetTempDepFactor(oASEWaveLth[NASE],oPumpWaveLth[P]);
//		exit(1);
		
		}
		
	/*	for(int S=0;S<NumSigCh;S++){
    		ASEStepGain +=
    			GetRamanGain(oASEWaveLth[NASE],oSignalWaveLth[S],StepI)*oSignalPower[S*NumZSteps+StepI-1]; 
	    	C2 += GetRamanGain(oASEWaveLth[NASE],oSignalWaveLth[S],StepI)*oSignalPower[S*NumZSteps+StepI-1]
					* GetTempDepFactor(oASEWaveLth[NASE],oSignalWaveLth[S]);
		}
    */	C1 = ASEStepGain - GetAttenuationFiber(oASEWaveLth[NASE],StepI);
		C2 *=   2 * hPlanck * Wavelength2Frequency(oASEWaveLth[NASE]) * DeltaFreq;
		oASEPower[NASE*NumZSteps+StepI] = oASEPower[NASE*NumZSteps+StepI - 1] * exp(C1 * DeltaZ(StepI)) 
				+ exp(C1 * DeltaZ(StepI)) * (  exp(C1 * DeltaZ(StepI)) * C2 * DeltaZ(StepI));

  	}

  }
  
//  cout << " after first guess" << endl;

  for(int i=0;i<NumIntItr;i++){
//		InternalItrNum++;
    monitor = OptimizeForwardIntegrationLastRound(StepI,PreSigPower);
	//				rms = OptimizeForwardShootingIntegration(StepI, PreSigPower, PrePumPower);
	   	//		cout << endl << "InItr#" << InternalItrNum << " ";
		//		cout << "total rms = " << rms ;
	}   
//  cout << " rms =" << monitor ;
 if ( abs(monitor)<1 ){
	}else{		
	 cout << monitor << endl;
	 cerr << "Error: Internal Iteration does not converge5 " 
     		<< "  exit "<< endl;
					exit(1);
  }
/* if(StepI == (NumZSteps-1)){  
   for(int NSig=0;NSig<NumSigCh;NSig++){	
//  cout << "SignalSignalStepGainSig#"  << NSig << " = " << SignalStepGain << endl;
//      DeltaPsignal[NSig] += abs(oSignalPower[NSig*NumZSteps+StepI] - oSignalPowerPre[NSig*NumZSteps+StepI]) 
//	  						/ oSignalPower[NSig*NumZSteps+StepI];
	  oSignalPowerPre[NSig*NumZSteps+StepI] = oSignalPower[NSig*NumZSteps+StepI];
   }
 } 
*/ delete [] PreSigPower; 
 
/* for(int ii=0;ii<NumASECh;ii++)
    {
	  *LogFile << " Step = " << StepI << endl;
		      oASEWaveLth[ii] = oSignalWaveLth[ii];  
	  *LogFile << "oNoiseWaveLth[ii]=" << oASEWaveLth[ii]  << "  " << oASEPower[ii*NumZSteps+StepI] << endl;
	}
 */
}
// ##########################
/* it is not useful anymore
void RamanAmp::GetAccuratePump()
{ 
 double DeltaZAccurate=45;
 double g;
 int 	NumZStepsAccurate = int(AmpLength/DeltaZAccurate+1.0001); 
 double *TempSigPower = new double[NumSigCh];
 double *TempPumPower = new double[NumPumCh];
 double *oPumpPower45 = new double[NumPumCh*NumZStepsAccurate];
 double *oSignalPower45 = new double[NumSigCh*NumZStepsAccurate];

   

// double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
 double StepGain;   
 double Temp;

// double oPumpPower45[NumPumCh*NumZStepsAccurate],oSignalPower45[NumSigCh*NumZStepsAccurate];
 int TypeIntergrationAccurate=3;
 int InternalItrAccurate=5; 
// first initial guess
 double SignalStepGain, PumpStepGain;	      
 oPumpPower45[0*NumZStepsAccurate] = 0.002731818461717321;
 oPumpPower45[1*NumZStepsAccurate] = 0.002181072009184382;
 oPumpPower45[2*NumZStepsAccurate] = 0.001957659650272773;
 oPumpPower45[3*NumZStepsAccurate] = 0.002300419579806988;
 oPumpPower45[4*NumZStepsAccurate] = 0.01363470992052203;
//0.002731818461717321    0.002181072009184382    0.001957659650272773    0.002300419579806988    0.01363470992052203  


 oPumpPower45[0*NumZStepsAccurate] = 0.00273181846224206;
 oPumpPower45[1*NumZStepsAccurate] = 0.002181072009774254;
 oPumpPower45[2*NumZStepsAccurate] = 0.001957659650887994;
 oPumpPower45[3*NumZStepsAccurate] = 0.00230041958040452;
 oPumpPower45[4*NumZStepsAccurate] = 0.01363470991432171;
//using 3 and deltaz=45, 5 internal itr
//0.00273181846224206    0.002181072009774254    0.001957659650887994    0.00230041958040452    0.01363470991432171 
//0.00273181846224206    0.002181072009774254    0.001957659650887994    0.00230041958040452    0.01363470991432171  
//0.002727104990862845    0.002176784828099888    0.001953751564040632    0.002297470083167586    0.01372696995512106  
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	oSignalPower45[NSig*NumZStepsAccurate] = oSignalPower[NSig*NumZSteps];
	cout << oSignalPower45[NSig*NumZStepsAccurate] << "  " ;
 }

 for(int IntStep=1;IntStep<NumZStepsAccurate;IntStep++){

  	for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
    	SignalStepGain = 0;
    	for(int P=0;P<NumPumCh;P++){
       		SignalStepGain +=
       		GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],IntStep)*oPumpPower45[P*NumZStepsAccurate+IntStep-1];
       	}

    	for(int S=0;S<NumSigCh;S++){
      		SignalStepGain +=
      		GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],IntStep)*oSignalPower45[S*NumZStepsAccurate+IntStep-1]; 
      	}
    
    	oSignalPower45[NSig*NumZStepsAccurate+IntStep] = oSignalPower45[NSig*NumZStepsAccurate+IntStep-1] * 
        	exp((SignalStepGain-GetAttenuationFiber(oSignalWaveLth[NSig]))*DeltaZAccurate);
  	}
  
  	for(int NPum=0;NPum<NumPumCh;NPum++)
 		{			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    	PumpStepGain = 0;

    	for(int P=0;P<NumPumCh;P++){
       		PumpStepGain +=
       		GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],IntStep) * oPumpPower45[P*NumZStepsAccurate+IntStep-1];
    	}

    	for(int S=0;S<NumSigCh;S++){
       		PumpStepGain +=
       		GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],IntStep) * oSignalPower45[S*NumZStepsAccurate+IntStep-1]; 
    	}
  
    	oPumpPower45[NPum*NumZStepsAccurate+IntStep] = oPumpPower45[NPum*NumZStepsAccurate+IntStep-1] 
				* exp( (-1) * (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZAccurate);

    	if (oPumpPower45[NPum*NumZStepsAccurate+IntStep] < 0){
	    	cerr << "PumpPower can't be negative" << endl 
             << "Aborting now! " << endl;
        	exit(1);
		}
  
  	} 


// internal iteration
// cout << "run to here2";
	for (int i=0;i<InternalItrAccurate;i++){
  		for(int NSig=0;NSig<NumSigCh;NSig++){ 

      		switch(TypeIntergrationAccurate)
     		{
       			case 1:   // Left Boundary
         			TempSigPower[NSig]=  oSignalPower45[NSig*NumZStepsAccurate+IntStep-1]; 
         		break;
     
       			case 2:    // Forward estimate Trape Rule
         			TempSigPower[NSig] =  (oSignalPower45[NSig*NumZStepsAccurate+IntStep] + oSignalPower45[NSig*NumZStepsAccurate+IntStep-1])/2; 
         		break;
     
       			case 3:     //  Forward Distance Average
         			g = oSignalPower45[NSig*NumZStepsAccurate+IntStep] / oSignalPower45[NSig*NumZStepsAccurate+IntStep-1]; 
         			 
		 			TempSigPower[NSig] = oSignalPower45[NSig*NumZStepsAccurate+IntStep-1] *( g - 1 ) /  log(g) ;
         		break;
		 
       			default:
         			cerr << "With type Intergration must be 1 or ?" << endl 
             		<< "Aborting now! " << endl;
         			exit(1);
      		} //end switch type

  		}

  
  		for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     		switch(TypeIntergrationAccurate)
     		{
       		case 1:   // Left Boundary
         		TempPumPower[NPum]=  oPumpPower[NPum*NumZStepsAccurate+IntStep-1]; 
         	break;
     
       		case 2:    // estimate Forward Trape Rule
         		TempPumPower[NPum]=  (oPumpPower45[NPum*NumZStepsAccurate+IntStep] 
		 			+ oPumpPower45[NPum*NumZStepsAccurate+IntStep-1]) / 2; 
         	break;
     
       		case 3:     //  Forward Distance Average
         		g = oPumpPower45[NPum*NumZStepsAccurate+IntStep] / oPumpPower45[NPum*NumZStepsAccurate+IntStep-1]; 
		 		if (g==1)
		 		   cout << "g ==1" << endl;
         		TempPumPower[NPum] = oPumpPower45[NPum*NumZStepsAccurate+IntStep-1] *( g - 1 ) /   log(g) ;
        	break;

      		default:
         		cerr << "With type Intergration must be 1 or ?" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      		} //end switch type
  		}
  

   	
  		for(int NSig=0;NSig<NumSigCh;NSig++){
  
    		StepGain = 0;
 	  	 	for(int P=0;P<NumPumCh;P++){
       			StepGain +=
      	 		GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],IntStep) * TempPumPower[P];
    		}
    		for(int S=0;S<NumSigCh;S++){
   	    		StepGain +=
   	    		GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],IntStep) * TempSigPower[S]; 
   		 	}
    		if (oSignalPower45[NSig*NumZStepsAccurate+IntStep] < 0){
	  	 		cerr << "Signal Power can't be negative" << endl 
       			<< " Aborting now! " << endl;
         	exit(1);
		 }
	

		oSignalPower45[NSig*NumZStepsAccurate+IntStep] = oSignalPower45[NSig*NumZStepsAccurate+IntStep-1] * 
       		exp((StepGain-GetAttenuationFiber(oSignalWaveLth[NSig]))*DeltaZAccurate);
	
 
  		}

  		for(int NPum=0;NPum<NumPumCh;NPum++){
  
    		StepGain = 0;
   	 		for(int P=0;P<NumPumCh;P++){
     	  		StepGain +=
      	 		GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],IntStep)*TempPumPower[P];
	    	}
    		for(int S=0;S<NumSigCh;S++){
      			StepGain +=
      			GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],IntStep)*TempSigPower[S]; 
    		}
	
			oPumpPower45[NPum*NumZStepsAccurate+IntStep] = oPumpPower45[NPum*NumZStepsAccurate+IntStep-1] * 
       			exp( (-1) * (StepGain-GetAttenuationFiber(oPumpWaveLth[NPum])) * DeltaZAccurate);  

    		if (oPumpPower45[NPum*NumZStepsAccurate+IntStep] < 0){
	    		cerr << "PumpPower can't be negative" << endl 
          		   	<< " Aborting now! " << endl;
        		exit(1);
			}
  		}
 	 }
    }
//	 	cout.precision(16);
	for(int NPum=0;NPum<NumPumCh;NPum++){
		for(int ii=0;ii<NumZSteps;ii++){ 
//			cout << ii << "  " << int(DeltaZ/DeltaZAccurate + 0.0001) << "  ";	
  			oPumpPower45Converted[NPum*NumZSteps+ii] = oPumpPower45[NPum*NumZStepsAccurate+ii*int(DeltaZ/DeltaZAccurate+0.0001)];
//			cout << oPumpPower45Converted[NPum*NumZSteps+ii] << "  " ; 
		}
	//	cout << endl;
 	} 
//	for(int NPum=0;NPum<NumPumCh;NPum++){
// 		for(int ii=0;ii<NumZStepsAccurate;ii++){ 
 //			cout << oPumpPower45[NPum*NumZStepsAccurate+ii] << "  ";
 	//	}
   
	//}

	cout << "run to here2" << endl;
 delete [] TempSigPower ;
 delete [] TempPumPower ;
 delete [] oPumpPower45 ;
 delete [] oSignalPower45 ;

} // end of GetAccurate

*/
// #############################
// out of data
// ## By: Jonathan Hu
void RamanAmp::PrepareNextStep()

{
/*
  double temp=0;

  for(int ii=0;ii<NumSigCh;ii++)  {
    temp += sq(abs(sfftM.fOptSignalX[ii]))  ;
	
//	cout << oSignalPower[ii*NumZSteps] << "--->" << oSignalPower[ii*NumZSteps +NumZSteps-1] << "  " ;             
 }
 temp -= (sq(abs(sfftM.fOptSignalX[0])) + sq(abs(sfftM.fOptSignalX[NumSigCh-1]))) /2;

 cout << "fcalcus=" << temp << endl;
 
 
 temp=0;
 for(int ii=0;ii<NumSigCh;ii++)  {
    temp += sq(abs(sfftM.tOptSignalX[ii]))  ;
	
//	cout << oSignalPower[ii*NumZSteps] << "--->" << oSignalPower[ii*NumZSteps +NumZSteps-1] << "  " ;             
 }
 temp -= (sq(abs(sfftM.tOptSignalX[0])) + sq(abs(sfftM.tOptSignalX[NumSigCh-1]))) /2;
// temp *= (1 / fft->delta_f * NumSigCh);  // this is delta-time
 cout << "tcalcus=" << temp << endl;
  


 cout << "foptsignal" << endl ;
 */
 
   double Energy = 0;

   for (int ii = 0; ii < qtPoints; ii++) 
    {   
      Energy +=  sq(abs(sfftM.tOptSignalX[ii]));
    }
 

   //OutFile.precision(8);


   // ## To check Parseval's relation you can uncomment the remaining
   // ## lines:

   

  double FreqEnergy = 0;

  for (int ii = 0; ii < qtPoints; ii++) 
    {   
      FreqEnergy +=  sq(abs(sfftM.fOptSignalX[ii]));
    }

   cout << "Time domain: " << Linear2dBm(Energy*DeltaTime/TimeWindow) 
	<< " Freq domain: " << Linear2dBm(FreqEnergy) << endl << endl;
 
 
fft->ifftshift(sfftM.fOptSignalX); 
 
 for(int ii=0;ii<NumSigCh;ii++)  {
    sfftM.fOptSignalX[ii] *= sqrt(oSignalPower[ii*NumZSteps + NumZSteps - 1] / oSignalPower[ii*NumZSteps]) ;           
 }

 fft->ifftshift(sfftM.fOptSignalX);          
 fft->IFFT(sfftM.fOptSignalX,sfftM.tOptSignalX); 

// for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
//	oSignalPower[NSig*NumZSteps] = oSignalPower[NSig*NumZSteps + NumZSteps -1];
// }

} // endl of PrepareNextStep

// ###########################################
// find the correction according to algorithm, please refer JOSAB submission
// ## By: Jonathan Hu
void RamanAmp::ParallelCorrection(double * ShootingValue, double * Target)
{
double OscillateCoeff;

 double *MatrixU = new double[NumCh*NumPumCh];
 double *EpxirP = new double[NumPumCh];
 double *CorrectP = new double[NumPumCh];

//double MatrixU[NumCh*NumPumCh], EpxirP[NumPumCh], CorrectP[NumPumCh];


  for(int NPum=0;NPum<NumPumCh;NPum++) {
//	EpxirP[NPum] = (-1)* (oPumpPower[NPum*NumZSteps + NumZSteps - 1] - oPumpTarget[NPum] ) ;	
 	EpxirP[NPum] = (-1)* (ShootingValue[NPum] - Target[NPum] );
  }
//  cout << endl;
// inital MatrixU

  for(int i=0;i<NumCh;i++) {
  	for(int j=0;j<NumPumCh;j++) {
	   if (i==j){
	   	    MatrixU[i*NumPumCh+j] = 1;
	   }
	   else{
	   		MatrixU[i*NumPumCh+j] = 0;
	   }

	}
  }

//  DisplayM(MatrixU,NumPumCh,NumPumCh);  
  for(int i=1;i<NumZSteps;i++){
 
//  *LogFile << endl << "StepI = " << i << endl << " MatrixU = " << endl;

//  	*LogFile << endl;

    UEvolution(i, MatrixU);
//  DisplayM(MatrixU,NumPumCh,NumPumCh);
  }
  

  MatrixDivid(MatrixU,WeightCorrect,NumPumCh);	
  
//  *LogFile <<  endl;
//  DisplayM(MatrixU,NumPumCh,NumPumCh);  
  
  GaussianElimination(MatrixU, CorrectP, EpxirP);

  for(int NPum=0;NPum<NumPumCh;NPum++) {
//  cout << " C=" << CorrectP[NPum] << " " ;
  	oPumpPower[NPum*NumZSteps] +=  CorrectP[NPum] ;
//    cout << "  PumpPower[" << NPum << "] = " << oPumpPower[NPum*NumZSteps];
	if (oPumpPower[NPum*NumZSteps] < 0){
		oPumpPower[NPum*NumZSteps] = 1e-6;
		cout << " #" << NPum << " Ch is guess less than 0" << endl; 
	}
  }
//  OscillateCoeff = 0;
  
//  cout << "CorrectP " << "  ";
//  for(int NPum=0;NPum<NumPumCh;NPum++) {
//  	cout << CorrectP[NPum] << "  ";
	
//	OscillateCoeff +=  CorrectP[NPum] / PreCorrectP[NPum];
//  	PreCorrectP[NPum] = CorrectP[NPum];
//  }

//  OscillateCoeff = OscillateCoeff / NumPumCh;
  
//  cout <<   OscillateCoeff ; 
//  *LogFile <<   OscillateCoeff << endl;
//  if ( (OscillateCoeff<-0.9) && (WeightCorrect>0.1) ){
//     WeightCorrect -= 0.1;     // change weight
//	 cout << "Weight Change to: " << WeightCorrect <<endl;
//  }

//  double rms=0;
//  for(int NPum=0;NPum<NumPumCh;NPum++) {	
//	rms +=  sq( CorrectP[NPum] * WeightCorrect / oPumpPower[NPum*NumZSteps]);
//  }
//  rms = sqrt(rms/NumPumCh);
//  *LogFile << "rms = Jack =" << rms << endl;
//  cout << "rms = Jack =" << rms << endl;  
//  if (rms <AccurateTh){


//  	cout << "rms=" << rms << "AccurateTh=" << AccurateTh;
//  	AccurateSign = 1;	
	//exit(1);  
//  }
 delete [] MatrixU ;
 delete [] EpxirP ;
 delete [] CorrectP ;
  
}  // end of ParallelCorrection()

// #####################################################################
// find correction for distance integrated average pump  case
// ## By: Jonathan Hu
void RamanAmp::ParallelIntgCorrection(double * ShootingValue, double * Target)
{
double OscillateCoeff;

 double *MatrixU = new double[NumCh*NumPumCh];
 double *MatrixUSum = new double[NumCh*NumPumCh];
 double *EpxirP = new double[NumPumCh];
 double *CorrectP = new double[NumPumCh];

//double MatrixU[NumCh*NumPumCh], EpxirP[NumPumCh], CorrectP[NumPumCh];


  for(int NPum=0;NPum<NumPumCh;NPum++) {
//	EpxirP[NPum] = (-1)* (oPumpPower[NPum*NumZSteps + NumZSteps - 1] - oPumpTarget[NPum] ) ;	
 	EpxirP[NPum] = (-1)* (ShootingValue[NPum] - Target[NPum] );

  }

// inital MatrixU

  for(int i=0;i<NumCh;i++) {
  	for(int j=0;j<NumPumCh;j++) {
	   if (i==j){
//	   	    MatrixU[i*NumPumCh+j] = oPumpPower[i*NumZSteps];
  			MatrixU[i*NumPumCh+j] = 1;
	   }
	   else{
	   		MatrixU[i*NumPumCh+j] = 0;

	   }

	}
  }
  MatrixCopy(MatrixUSum,MatrixU,NumCh,NumPumCh);
  MatrixDivid(MatrixUSum,2,NumCh,NumPumCh);
 
  for(int i=1;i<NumZSteps;i++){
 
//  *LogFile << endl << "StepI = " << i << endl << " MatrixU = " << endl;
//  	*LogFile << endl;

    UEvolution(i, MatrixU);
	MatrixPlus(MatrixUSum,MatrixU,NumCh,NumPumCh);	

//  DisplayM(MatrixU,NumPumCh,NumPumCh);
  }


//  DisplayM(MatrixU,NumPumCh,NumPumCh);     

  MatrixDivid(MatrixU,2,NumCh,NumPumCh);
  MatrixSubtract(MatrixUSum,MatrixU,NumCh,NumPumCh);
  MatrixDivid(MatrixUSum,NumZSteps-1,NumCh,NumPumCh);
  MatrixDivid(MatrixUSum,WeightCorrect,NumPumCh);	
  
//  *LogFile <<  endl;
//  DisplayM(MatrixU,NumPumCh,NumPumCh);  
  MatrixDivid(MatrixU,WeightCorrect,NumPumCh);	
   
  GaussianElimination(MatrixU, CorrectP, EpxirP);

  for(int NPum=0;NPum<NumPumCh;NPum++) {
  cout << " C=" << CorrectP[NPum] << " " ;
  	oPumpPower[NPum*NumZSteps] +=  CorrectP[NPum]; // * oPumpPower[NPum*NumZSteps] ;

	if (oPumpPower[NPum*NumZSteps] < 0){
		oPumpPower[NPum*NumZSteps] = 1e-6;
		cout << " #" << NPum << " Ch is guess less than 0" << endl; 
	}
  }
//  OscillateCoeff = 0;
  
//  cout << "CorrectP " << "  ";
//  for(int NPum=0;NPum<NumPumCh;NPum++) {
//  	cout << CorrectP[NPum] << "  ";
	
//	OscillateCoeff +=  CorrectP[NPum] / PreCorrectP[NPum];
//  	PreCorrectP[NPum] = CorrectP[NPum];
//  }

//  OscillateCoeff = OscillateCoeff / NumPumCh;
  
//  cout <<   OscillateCoeff ; 
//  *LogFile <<   OscillateCoeff << endl;
//  if ( (OscillateCoeff<-0.9) && (WeightCorrect>0.1) ){
//     WeightCorrect -= 0.1;     // change weight
//	 cout << "Weight Change to: " << WeightCorrect <<endl;
//  }

//  double rms=0;
//  for(int NPum=0;NPum<NumPumCh;NPum++) {	
//	rms +=  sq( CorrectP[NPum] * WeightCorrect / oPumpPower[NPum*NumZSteps]);
//  }
//  rms = sqrt(rms/NumPumCh);
//  *LogFile << "rms = Jack =" << rms << endl;
//  cout << "rms = Jack =" << rms << endl;  
//  if (rms <AccurateTh){


//  	cout << "rms=" << rms << "AccurateTh=" << AccurateTh;
//  	AccurateSign = 1;	
	//exit(1);  
//  }
 delete [] MatrixU ;
 delete [] MatrixUSum;
 delete [] EpxirP ;
 delete [] CorrectP ;
  
}  // end of ParallelIntgCorrection()

// ########################################3
// find the correction for one pump each time, it is only useful for low mathmatical stablity case
// ## By: Jonathan Hu
void RamanAmp::OneByOneCorrection(double * ShootingValue, double * Target)
{
// double *MatrixU = new double[NumCh*NumPumCh];
// double *MatrixUSum = new double[NumCh*NumPumCh];
 double *EpxirP = new double[NumPumCh];
// double *CorrectP = new double[NumPumCh];
 double MaxEp=0;
 int NMaxPum;
 int Factor=2;
  for(int NPum=0;NPum<NumPumCh;NPum++) {
 	EpxirP[NPum] = (-1)* (ShootingValue[NPum] - Target[NPum] ) / Target[NPum];
	if(abs(EpxirP[NPum]) > MaxEp){
		MaxEp = abs(EpxirP[NPum]);
		NMaxPum = NPum;
	}
  }
  while((oPumpPower[NMaxPum*NumZSteps] + EpxirP[NMaxPum] * Target[NMaxPum] * WeightCorrect 
  				* exp((-1) * GetAttenuationFiber(oPumpWaveLth[NMaxPum]) * AmpLength) * Factor) < 0){
	Factor /= 4;
  }
				
  oPumpPower[NMaxPum*NumZSteps] +=  EpxirP[NMaxPum] * Target[NMaxPum] * WeightCorrect 
  				* exp((-1) * GetAttenuationFiber(oPumpWaveLth[NMaxPum]) * AmpLength) * Factor; // * oPumpPower[NPum*NumZSteps] ;
//  cout << "C[" << NMaxPum << "] = "  << EpxirP[NMaxPum] * Target[NMaxPum] * WeightCorrect 
//  				* exp((-1) * GetAttenuationFiber(oPumpWaveLth[NMaxPum]) * AmpLength) * Factor;

  if (oPumpPower[NMaxPum*NumZSteps] < 0){
	oPumpPower[NMaxPum*NumZSteps] = 1e-6;
	cout << " #" << NMaxPum << " Ch is guess less than 0" << endl; 
  }
  

// delete [] MatrixU ;
   delete [] EpxirP ;
// delete [] CorrectP ;
  
}  // end of OneByOneCorrection()


  
// ###########################
// Shooting algorithm, please refer JOSAB sumbission Eq. 2-9 
// also look at Communiction of the ACM p. 613 (1962)
// ## By: Jonathan Hu
void RamanAmp::UEvolution(int StepI, double * MatrixU)
{ 
// cout << " NumCh=" << NumCh << " NumCh*NumCh=" << NumCh*NumCh ;
// cerr << "G0";
//  int NumChsq = NumCh*NumCh;
//  double ExpW[NumCh*NumCh];

//  double MatrixW[NumCh*NumCh];

//  double MatrixW2[NumCh*NumCh];

  
//  double MatrixW3[NumCh*NumCh],MatrixW4[NumCh*NumCh];
  
  double *ExpW = new double[NumCh*NumCh];  
  double *MatrixW = new double[NumCh*NumCh];  
  double *MatrixW2 = new double[NumCh*NumCh];  
  double *MatrixW3 = new double[NumCh*NumCh];  
  double *MatrixW4 = new double[NumCh*NumCh];  
 
  for(int i=0;i<NumCh;i++) {
  	for(int j=0;j<NumCh;j++) {
	   if (i==j){
	   	    ExpW[i*NumCh+j] = 1;
	   }
	   else{
	   		ExpW[i*NumCh+j] = 0;
	   }

	}
  }
  
  GetMatrixW(StepI, MatrixW);
 

  MatrixMultiply(MatrixW,DeltaZ(StepI),NumCh);

  for(int i=0;i<NumCh;i++) {
  	for(int j=0;j<NumCh;j++) {
	   MatrixW2[i*NumCh+j] = MatrixW[i*NumCh+j];
	   MatrixW3[i*NumCh+j] = MatrixW[i*NumCh+j];
//	   MatrixW4[i*NumCh+j] = MatrixW[i*NumCh+j];
	  
	}
  } 

  // can be simply for program
  MatrixMultiply(MatrixW2,MatrixW,NumCh);

  MatrixDivid(MatrixW2,2,NumCh);

  MatrixMultiply(MatrixW3,MatrixW2,NumCh);	
  MatrixDivid(MatrixW3,6,NumCh);
// MatrixMultiply(MatrixW4,MatrixW3,NumCh);	
// MatrixDivid(MatrixW4,24,NumCh);


  MatrixPlus(ExpW,MatrixW,NumCh);
  MatrixPlus(ExpW,MatrixW2,NumCh);  
  MatrixPlus(ExpW,MatrixW3,NumCh);  
// MatrixPlus(ExpW,MatrixW4,NumCh);
   
// DisplayM(ExpW,NumCh,NumCh); 

  MatrixReducedMultiply(MatrixU,ExpW,NumCh,NumPumCh); // MatrixU is NumCh x NumPumCh, ExpW is NumCh x NumCh 

//  *LogFile << "in the function" << endl;
//  DisplayM(MatrixU,NumPumCh,NumPumCh); 	
	delete [] ExpW;
	delete [] MatrixW;
	delete [] MatrixW2;
	delete [] MatrixW3;	
	delete [] MatrixW4;
 
} // end of UEvolution

// ################################
// Shooting algorithm, please refer JOSAB sumbission and
// look at Communiction of the ACM p. 613 (1962)
// ## By: Jonathan Hu
void RamanAmp::GetMatrixW(int WStepI, double * MatrixW)
{

 double *TempSigPower = new double[NumSigCh];
 double *TempPumPower = new double[NumPumCh];
//double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
double g; 
  for(int NSig=0;NSig<NumSigCh;NSig++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+WStepI] + oSignalPower[NSig*NumZSteps+WStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+WStepI] / oSignalPower[NSig*NumZSteps+WStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+WStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
	  
       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+WStepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+WStepI] + oPumpPower[NPum*NumZSteps+WStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+WStepI] / oPumpPower[NPum*NumZSteps+WStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+WStepI-1] *( g - 1 ) /   log(g) ;
         break;

	  default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
  }
		      
  for(int NPum=0;NPum<NumPumCh;NPum++)
 	{			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    for(int P=0;P<NumPumCh;P++){
	   if(NPum==P)  {
		  MatrixW[NPum * NumCh +P] = 0;	
	      for(int PP=0;PP<NumPumCh;PP++){
		        if(PP!=NPum){
		  			MatrixW[NPum * NumCh +P] +=
					(-1) * GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[PP],WStepI) * TempPumPower[PP] ;
/*		  			if(WStepI==10){
						*LogFile << "Gain = " << GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[PP],WStepI)  
							<< "  TempPumPower[" << PP << "] = " << TempPumPower[PP] << endl
						 	<< " Gain * TempPumPower = " 
							<< (-1) * GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[PP],WStetI) * TempPumPower[PP] << endl;
					}
	*/			}
		  }
		  for(int SS=0;SS<NumSigCh;SS++){
		        MatrixW[NPum * NumCh +P] +=
       			(-1) * GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[SS],WStepI) * TempSigPower[SS];
//	  			if(WStepI==10){
//					*LogFile << (-1) * GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[SS],WStepI) * TempSigPower[SS] << endl;
//					}
		  }
		  MatrixW[NPum * NumCh +P] += GetAttenuationFiber(oPumpWaveLth[NPum],WStepI);   
	   }
	   else{
       		MatrixW[NPum * NumCh +P] =
       		(-1) * GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],WStepI) * TempPumPower[P] ;
	   }
   }
   for(int S=0;S<NumSigCh;S++){
    	   MatrixW[NPum * NumCh +NumPumCh+S] =
       		(-1) * GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],WStepI) * TempSigPower[S]; 
   }
 } 
 
 for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
    for(int P=0;P<NumPumCh;P++){
       MatrixW[(NSig+NumPumCh) * NumCh +P] =
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],WStepI)* TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
		if(NSig==S){
		  	MatrixW[(NumPumCh+NSig) * NumCh + NumPumCh + S] = 0;	
	      	for(int PP=0;PP<NumPumCh;PP++){
		        MatrixW[(NumPumCh+NSig) * NumCh + NumPumCh + S] +=
				GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[PP],WStepI) * TempPumPower[PP] ;
		  		
		  	}
		  	for(int SS=0;SS<NumSigCh;SS++){
		        if(NSig!=SS){
					MatrixW[(NumPumCh+NSig) * NumCh +NumPumCh+S] +=
       				GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[SS],WStepI) * TempSigPower[SS];
		  		}				
			}
		  	MatrixW[(NumPumCh+NSig) * NumCh + NumPumCh + S] -= GetAttenuationFiber(oSignalWaveLth[NSig],WStepI);   
	   	}
	   	else{
    		MatrixW[(NumPumCh+NSig) * NumCh + NumPumCh + S] =
    		GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],WStepI)* TempSigPower[S]; 
       	}
    }
  }
 delete [] TempSigPower ;
 delete [] TempPumPower ;  
 
} // end of GetMatrixW

// #############################
// GaussianElimination
// ## By: Jonathan Hu

void RamanAmp::GaussianElimination(double * MatrixU, double * CorrectP, double * EpxirP)
{
// in calculation, please pay attentation on the matrixU (NumCh * NumPumCh)
// but I just using GaussianElimination on left (NumPumCh* NumPumCh) Matrix
   int MaxRow;
   double MaxAbsValue;
   double Temp,m,sum;
 for(int k=0;k<NumPumCh-1;k++){
	MaxRow = k;
	MaxAbsValue = abs(MatrixU[k*NumPumCh+k]);
    for(int i=k+1;i<NumPumCh;i++){
  		if( abs(MatrixU[i*NumPumCh+k]) > MaxAbsValue) { 
	  		MaxRow = i;
			MaxAbsValue = abs(MatrixU[i*NumPumCh+k]);
		}
	}	
	if (MaxRow!=k){
		for(int i=k;i<NumPumCh;i++) {
		   Temp = MatrixU[k*NumPumCh+i];
		   MatrixU[k*NumPumCh+i] = MatrixU[MaxRow*NumPumCh+i];
		   MatrixU[MaxRow*NumPumCh+i] = Temp;
		}
		Temp = EpxirP[k];
		EpxirP[k] = EpxirP[MaxRow];
		EpxirP[MaxRow] = Temp;
	}
	if (abs(MatrixU[k*NumPumCh+k]) < 1e-10) {
         cerr << "Matrix has multiple answer" << endl 
		      << "MaxValue is " << MatrixU[k*NumPumCh+k] << " k = " << k << endl

             << "Aborting now! " << endl;
         exit(1);	   
	}
 
    // Elimination
    for(int i=k+1;i<NumPumCh;i++) {
	  m = MatrixU[i*NumPumCh+k] / MatrixU[k*NumPumCh+k];	  
	  for (int j=k+1;j<NumPumCh;j++) {
		  MatrixU[i*NumPumCh+j] -= m*MatrixU[k*NumPumCh+j];
	  }
	  EpxirP[i] -= m*EpxirP[k];
	}	
}
// Backsubstitution
CorrectP[NumPumCh-1] = EpxirP[NumPumCh-1] / MatrixU[ (NumPumCh-1) * NumPumCh + NumPumCh-1];
for(int i=NumPumCh-2;i>=0;i--) {
   sum=0;
   for(int j=i+1;j<NumPumCh;j++){
      sum += MatrixU[i*NumPumCh+j] * CorrectP[j];
   }
   CorrectP[i] = (EpxirP[i] - sum ) / MatrixU[i*NumPumCh+i];
}

}  // end of GaussianElimination

// ############################

void RamanAmp::DisplayM(double * M1 , int M , int N)
{
//	cout << endl;	

  for(int i=0;i<M;i++) {
	for(int j=0;j< N;j++) {
//		cout << M1[i*NumCh +j] << "  " ;
		*LogFile << M1[i*N +j] << "  " ;
	}
//	cout << endl;	
	*LogFile << endl;	
  }
  *LogFile << endl;

} // end of DisplayMatrix



// #################################

void RamanAmp::MatrixMultiply(double * M1, double * M2, int N)

{ 
  double *M = new double[N*N];  
//double M[N*N];
  for(int i=0;i<N;i++) {
	for(int j=0;j<N;j++) {
		M[i*N+j] =0;
		for(int k=0;k<N;k++) {
		 	M[i*N+j]  += M1[i*N+k] * M2[k*N+j];
		}
	}
  }
  for(int i=0;i<N;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] = M[i*N+j];
	}
  }
	delete [] M;
}  // end of MatrixMultiply

// ##################################

void RamanAmp::MatrixReducedMultiply(double * M1, double * M2, int N , int P)
// this is reduced Multiply :M1 is NumCh x NumPumCh, M2 is NumCh x NumCh 
// Get M1 = M2 x M1;
{   double *M = new double[N*P];  
//  double M[N*P];
  for(int i=0;i<N;i++) {
	for(int j=0;j<P;j++) {
		M[i*P+j] =0;
		for(int k=0;k<N;k++) {
		 	M[i*P+j]  += M2[i*N+k] * M1[k*P+j] ;
		}
	}
  }
  for(int i=0;i<N;i++) {
	for(int j=0;j<P;j++) {
		M1[i*P +j] = M[i*P+j];
	}
  }
	delete [] M;
}  // end of MatrixReducedMultiply

// #################################


void RamanAmp::MatrixMultiply(double * M1, double M2, int N)

{ 
  for(int i=0;i<N;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] = M1[i*N+j] * M2;
	}
  }
	
}  // end of MatrixMultiply

// #################################

void RamanAmp::MatrixDivid(double * M1, double M, int N)

{ 
  for(int i=0;i<N;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] = M1[i*N+j] / M;
	}
  }
	
}  // end of MatrixDivid	

// #####################################################

void RamanAmp::MatrixDivid(double * M1, double Num, int M, int N)

{ 
  for(int i=0;i<M;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] = M1[i*N+j] / Num;
	}
  }
	
}  // end of MatrixDivid


// #################################

void RamanAmp::MatrixPlus(double * M1, double * M2, int N)
{ 
  for(int i=0;i<N;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] += M2[i*N+j];
	}
  }
}  // end of MatrixPlus

// #################################

void RamanAmp::MatrixPlus(double * M1, double * M2, int M, int N)
{ 
  for(int i=0;i<M;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] += M2[i*N+j];
	}
  }
}  // end of MatrixPlus

// #################################

void RamanAmp::MatrixSubtract(double * M1, double * M2, int M, int N)
{ 
  for(int i=0;i<M;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] -= M2[i*N+j];
	}
  }
}  // end of MatrixSubtract

// #################################

void RamanAmp::MatrixCopy(double * M1, double * M2, int M, int N)
{ 
  for(int i=0;i<M;i++) {
	for(int j=0;j<N;j++) {
		M1[i*N +j] = M2[i*N+j];
	}
  }
}  // end of MatrixCopy

// ################################
/* 
void RamanAmp::PrePumpShooting(int PreShootingStepI)
{ double PumpStepGain;
  for(int NPum=0;NPum<NumPumCh;NPum++)
 	{			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    PumpStepGain = 0;

    for(int P=0;P<NumPumCh;P++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], NumZSteps-PreShootingStepI) * oPumpPower[P*NumZSteps+PreShootingStepI-1];
    }
  
    oPumpPower[NPum*NumZSteps+PreShootingStepI] = oPumpPower[NPum*NumZSteps+PreShootingStepI-1] 
				* exp( (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZ);
  }  
}  // end of PrePumpShooting
*/
// ################################
/*  it only used in increase number of pump
void RamanAmp::PrePumpShootingL2R(int PreShootingStepI)
{ double PumpStepGain;
  for(int NPum=0;NPum<NumPumCh;NPum++)
 	{			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    PumpStepGain = 0;

    for(int P=0;P<NumPumCh;P++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], NumZSteps-PreShootingStepI) * oPumpPower[P*NumZSteps+PreShootingStepI-1];
    }
  
    oPumpPower[NPum*NumZSteps+PreShootingStepI] = oPumpPower[NPum*NumZSteps+PreShootingStepI-1] 
				* exp( (PumpStepGain+GetAttenuationFiber(oPumpWaveLth[NPum]))*DeltaZ);
  }  
}  // end of PrePumpShootingL2R
*/

// ################################

void RamanAmp::SignalOnlyPropagation(int PreShootingStepI)
{ double SignalStepGain;
  for(int NSig=0;NSig<NumSigCh;NSig++)
 	{			
    SignalStepGain = 0;

    for(int S=0;S<NumSigCh;S++){
       SignalStepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],PreShootingStepI) * oSignalPower[S*NumZSteps+PreShootingStepI-1];
    }
  
    oSignalPower[NSig*NumZSteps+PreShootingStepI] = oSignalPower[NSig*NumZSteps+PreShootingStepI-1] 
				* exp( (SignalStepGain-GetAttenuationFiber(oSignalWaveLth[NSig],PreShootingStepI))*DeltaZ(PreShootingStepI));
//  	cout << oSignalPower[NSig*NumZSteps+PreShootingStepI] << "   ";
  }  
//  cout << endl;
 
}  // end of SignalOnlyPropagation

// ################################
/*
void RamanAmp::SignalOnlyPropagationWOA(int PreShootingStepI)
{ 
	// No Attenuation
	double SignalStepGain;
  for(int NSig=0;NSig<NumSigCh;NSig++)
 	{			
    SignalStepGain = 0;

    for(int S=0;S<NumSigCh;S++){
       SignalStepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], PreShootingStepI) * oSignalPower[S*NumZSteps+PreShootingStepI-1];
    }
  
    oSignalPower[NSig*NumZSteps+PreShootingStepI] = oSignalPower[NSig*NumZSteps+PreShootingStepI-1] 
				* exp( SignalStepGain*DeltaZ(PreShootingStepI));
  }  
}  // end of SignalOnlyPropagationWOA
*/

// ################
/*
void RamanAmp::PrePumpShootingLastOne(int StepI)
{
  double SignalStepGain, PumpStepGain;
  PumpStepGain = 0;
  for(int P=0;P<NumPumCh;P++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NumPumCh-1],oPumpWaveLth[P]) * oPumpPower[P*NumZSteps+NumZSteps-1-StepI];
  }
  for(int S=0;S<NumSigCh;S++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NumPumCh-1],oSignalWaveLth[S]) * oSignalPower[S*NumZSteps+NumZSteps-1-StepI]; 
  }
  oPumpPower[(NumPumCh-1)*NumZSteps+StepI] = oPumpPower[(NumPumCh-1)*NumZSteps+StepI-1] 
  			* exp( (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NumPumCh-1]))*DeltaZ);
  cout << " oPump= " << oPumpPower[(NumPumCh-1)*NumZSteps+StepI-1] << endl;
}  // end of PrePumpShootingLastOne
*/
// ################################
// forward shooting
// Shooting algorithm, please refer JOSAB sumbission
// ## By: Jonathan Hu

void RamanAmp::ForwardShooting(int StepI)
{ 
 double *PreSigPower = new double[NumSigCh];
 double *PrePumPower = new double[NumPumCh];
//double PreSigPower[NumSigCh], PrePumPower[NumPumCh];
  double SignalStepGain, PumpStepGain;
  double monitor;
  
// cout << "\a";
  FirstGuessForwardShooting(StepI,PreSigPower,PrePumPower);
      
  if(TypeIntergration==6){     // using Simplfy Runge-Kutta 
//       RKShootingSimple(StepI);	
  }else   if (TypeIntergration==7){   // using Runge-Kutta
//       		RKShooting(StepI);	
   		} else if (TypeIntergration==8)
		  {   // using Runge-Kutta text
//     	  RKShootingtext(StepI);	
   		  } else {
			int InternalItrNum=0;
//			cout << "StepI = " << StepI << "  " ;
   		//	do {
	//				    cout << endl <<  NumIntItr << endl;	


		    for(int i=0;i<NumIntItr;i++){

	 			InternalItrNum++;
				monitor = OptimizeForwardShootingIntegration(StepI, PreSigPower, PrePumPower);
	   	//		cout << endl << "InItr#" << InternalItrNum << " ";
		//		cout << "total rms = " << rms ;
				if(monitor>0.2){
					BlowupSign=1;
				}
			}
  
  //			if(StepI==10){
		//		cout << " monitor=" << monitor;
	//			}
  
     	//	} while( abs(rms) > 1e-3 && InternalItrNum < 25 );

//		  	Finish = clock();
//  	  		Duration = (double)(Finish - Start) / CLOCK_PER_SEC;
  
//  			cout << "Duration = " << Duration << endl;
//			exit(1);
		

//     		cout << endl;
			if ( InternalItrNum >= 25 ){
	    		cerr << "Error: Internal Iteration does not converge1 " 
           			<< "  InternalItrNum >=25 Relaxation used exit "<< endl;
    			exit(1);
				
				InternalRelaxationInForwardShooting(StepI, PreSigPower, PrePumPower);
			
				//exit(1);
	 		}
			if ( abs(monitor)<1 ){
			}
			else{
  //              FirstGuessForwardShooting(StepI,PreSigPower,PrePumPower);			
			   cout <<  "  Monitor=" << monitor << "  StepI=" << StepI << endl;
	    		cerr << "Error: Internal Iteration does not converge 2" 
           			<< " Relaxation used exit "<< endl;
				BlowupSign=1;
	
				//	exit(1);
    			//InternalRelaxationInForwardShooting(StepI, PreSigPower, PrePumPower);
				// exit(1);
	 		}
//	 		cout << endl;
  		}
 delete [] PreSigPower ;
 delete [] PrePumPower ;  
}  // end of ForwardShooting

// ##################################
// out of date, not used anymore
// ## By: Jonathan Hu

void RamanAmp::InternalRelaxationInForwardShooting(int IntStepI, double * PreSig, double * PrePum)
{
  double SignalStepGain, PumpStepGain;
  double rms, temp;
  int InternalItrNum=0;
  int i;
  
  FirstGuessForwardShooting(IntStepI, PreSig, PrePum);
  do { 
		rms=0;
		InternalItrNum++;
		i=0;
		do{
			i++;
			temp = OptimizeForwardPumpShootingIntegration(IntStepI, PreSig, PrePum);
		}while(temp>1e-3 && i<10);
		cout << "tempE = " << temp ;
		rms+=temp;
		i=0;
		do{
			i++;
			temp = OptimizeForwardSignalShootingIntegration(IntStepI, PreSig, PrePum);
		}while(temp>1e-3 && i<10);
		cout << "tempE = " << temp  ;
		rms += temp;
		cout << endl << "InItr#" << InternalItrNum << " rms = " << rms;
		
	} while(abs(rms) > 1e-8 && InternalItrNum < 45);
	if ( InternalItrNum >= 45 ){
	    		cerr << "Error: Internal Iteration does not converge3 " 
           			<< " exit "<< endl;
				
				exit(1);
	}
	if (rms<1){
	}
	else{
	   		cerr << "Error: Internal Iteration does not converge4 " 
       			<< " exit "<< endl;
			exit(1);
	}

} // end of InternalRelaxationInForwardShooting

// #############################################
// need to debug
/*

void RamanAmp::RKShootingSimple(int IntStepI)
{
 double *vSignalPower = new double[NumSigCh];
 double *vPumpPower = new double[NumPumCh];
 double *KSig = new double[NumSigCh];
 double *KPum = new double[NumPumCh];
//  double vSignalPower[NumSigCh], vPumpPower[NumPumCh] ;
//  double KSig[NumSigCh], KPum[NumPumCh];

  double SignalStepGain, PumpStepGain;
   
  for(int NSig=0;NSig<NumSigCh;NSig++){   
    vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] ;
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){
	vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] ;
  } 
   
   
  for(int i=4;i>0;i--){
   
    for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
      SignalStepGain = 0;
      for(int P=0;P<NumPumCh;P++){
         SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI)
		 		* vSignalPower[NSig] * vPumpPower[P] ;
      }

      for(int S=0;S<NumSigCh;S++){
        SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI)
				* vSignalPower[NSig] * vSignalPower[S] ; 
        }
    
      SignalStepGain -= GetAttenuationFiber(oSignalWaveLth[NSig]) * vSignalPower[NSig];
  	  KSig[NSig] = DeltaZ * SignalStepGain;
    }
  
    for(int NPum=0;NPum<NumPumCh;NPum++)
   	  {			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
      PumpStepGain = 0;

      for(int P=0;P<NumPumCh;P++){
         PumpStepGain += GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI) 
		  		* vPumpPower[NPum] * vPumpPower[P];
      }

      for(int S=0;S<NumSigCh;S++){
         PumpStepGain += GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI) 
		    	* vPumpPower[NPum] * vSignalPower[S]; 
      }
  
      PumpStepGain -= GetAttenuationFiber(oPumpWaveLth[NPum]) * vPumpPower[NPum];
	  KPum[NPum] = (-1) * PumpStepGain * DeltaZ;
    }  
	
	for(int NSig=0;NSig<NumSigCh;NSig++){   
      vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + KSig[NSig] / i;
    }

    for(int NPum=0;NPum<NumPumCh;NPum++){
  	  vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + KPum[NPum] / i;
    } 
  
//   cout << "i should be 4,3,2,1,"<< i<< endl;
  }
   	
  for(int NSig=0;NSig<NumSigCh;NSig++){   
    oSignalPower[NSig*NumZSteps+IntStepI] = vSignalPower[NSig] ;
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){
    oPumpPower[NPum*NumZSteps+IntStepI] = vPumpPower[NPum] ;
  } 

 delete [] vSignalPower ;
 delete [] vPumpPower ;
 delete [] KSig ;
 delete [] KPum ;
} // endl of RKShootingSimple


// #############################################

void RamanAmp::RKShooting(int IntStepI)
{
  double vSignalPower[NumSigCh], vPumpPower[NumPumCh] ;
  double K1Sig[NumSigCh], K2Sig[NumSigCh], K3Sig[NumSigCh], K4Sig[NumSigCh];
  double K1Pum[NumPumCh], K2Pum[NumPumCh], K3Pum[NumPumCh], K4Pum[NumPumCh];

  double SignalStepGain, PumpStepGain;
   
  for(int NSig=0;NSig<NumSigCh;NSig++){   
    vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] ;
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){
	vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] ;
  } 
     
  for(int RKStage=1;RKStage<5;RKStage++) {
  
    for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
      SignalStepGain = 0;
      for(int P=0;P<NumPumCh;P++){
         SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI)
		 		* vSignalPower[NSig] * vPumpPower[P] ;
      }

      for(int S=0;S<NumSigCh;S++){
        SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI)
				* vSignalPower[NSig] * vSignalPower[S]; 
        }
    
      SignalStepGain -= GetAttenuationFiber(oSignalWaveLth[NSig]) * vSignalPower[NSig];
      switch(RKStage){
     
	   case 1:   K1Sig[NSig] = DeltaZ * SignalStepGain;    break;
       case 2:   K2Sig[NSig] = DeltaZ * SignalStepGain;    break;
	   case 3:   K3Sig[NSig] = DeltaZ * SignalStepGain;    break;
	   case 4:   K4Sig[NSig] = DeltaZ * SignalStepGain;    break;
	  
       default:
         cerr << "With RK-Stage must between 1 and 4" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } 
  
    }
  
    for(int NPum=0;NPum<NumPumCh;NPum++)
   	  {			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
      PumpStepGain = 0;

      for(int P=0;P<NumPumCh;P++){
         PumpStepGain +=  GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI) 
		 		* vPumpPower[NPum] * vPumpPower[P];
      }

      for(int S=0;S<NumSigCh;S++){
         PumpStepGain += GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI) 
		 		* vPumpPower[NPum] * vSignalPower[S]; 
      }
  
      PumpStepGain -= GetAttenuationFiber(oPumpWaveLth[NPum]) * vPumpPower[NPum];
	  
	  switch(RKStage){
     
		   	case 1:   K1Pum[NPum] = (-1) * PumpStepGain * DeltaZ;    break;
	   		case 2:   K2Pum[NPum] = (-1) * PumpStepGain * DeltaZ;    break;
	   		case 3:   K3Pum[NPum] = (-1) * PumpStepGain * DeltaZ;    break;
	   		case 4:   K4Pum[NPum] = (-1) * PumpStepGain * DeltaZ;    break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
	  
	  
    }  
	
	for(int NSig=0;NSig<NumSigCh;NSig++){   
	  
	  	  switch(RKStage){
     
		   	case 1:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K1Sig[NSig] / 2;    break;
	   		case 2:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K2Sig[NSig] / 2;    break;
	   		case 3:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K3Sig[NSig] ;    	break;
	   		case 4:   oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] + 
	        			( K1Sig[NSig] + 2 * K2Sig[NSig] + 2 * K3Sig[NSig] + K4Sig[NSig] )/ 6;    break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
      
    }

    for(int NPum=0;NPum<NumPumCh;NPum++){
		 switch(RKStage){
     
		   	case 1:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K1Pum[NPum] / 2;    break;
	   		case 2:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K2Pum [NPum] / 2;    break;
	   		case 3:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K3Pum[NPum] ;   break;
	   		case 4:   oPumpPower[NPum*NumZSteps+IntStepI]   = oPumpPower[NPum*NumZSteps+IntStepI-1] + 
	        			( K1Pum[NPum] + 2 * K2Pum[NPum] + 2 * K3Pum[NPum] + K4Pum[NPum] )/ 6;   break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
    } 
  
  }
  
 
} // endl of RKShooting


// #############################################

void RamanAmp::RKShootingtext(int IntStepI)
{
  double vSignalPower[NumSigCh], vPumpPower[NumPumCh] ;
  double K1Sig[NumSigCh], K2Sig[NumSigCh], K3Sig[NumSigCh], K4Sig[NumSigCh];
  double K1Pum[NumPumCh], K2Pum[NumPumCh], K3Pum[NumPumCh], K4Pum[NumPumCh];

  double SignalStepGain, PumpStepGain;
   
  for(int NSig=0;NSig<NumSigCh;NSig++){   
    vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] ;
  }

  for(int NPum=0;NPum<NumPumCh;NPum++){
	vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] ;
  } 
     
  for(int RKStage=1;RKStage<5;RKStage++) {
  
    for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
      SignalStepGain = 0;
      for(int P=0;P<NumPumCh;P++){
         SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI)
		 		* vSignalPower[NSig] * vPumpPower[P] ;
      }

      for(int S=0;S<NumSigCh;S++){
        SignalStepGain += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI)
				* vSignalPower[NSig] * vSignalPower[S]; 
        }
    
      SignalStepGain -= GetAttenuationFiber(oSignalWaveLth[NSig]) * vSignalPower[NSig];
      switch(RKStage){
     
	   case 1:   K1Sig[NSig] = SignalStepGain;    break;
       case 2:   K2Sig[NSig] = SignalStepGain;    break;
	   case 3:   K3Sig[NSig] = SignalStepGain;    break;
	   case 4:   K4Sig[NSig] = SignalStepGain;    break;
	  
       default:
         cerr << "With RK-Stage must between 1 and 4" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } 
  
    }
  
    for(int NPum=0;NPum<NumPumCh;NPum++)
   	  {			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
      PumpStepGain = 0;

      for(int P=0;P<NumPumCh;P++){
         PumpStepGain +=  GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI) 
		 		* vPumpPower[NPum] * vPumpPower[P];
      }

      for(int S=0;S<NumSigCh;S++){
         PumpStepGain += GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI) 
		 		* vPumpPower[NPum] * vSignalPower[S]; 
      }
  
      PumpStepGain -= GetAttenuationFiber(oPumpWaveLth[NPum]) * vPumpPower[NPum];
	  
	  switch(RKStage){
     
		   	case 1:   K1Pum[NPum] = (-1) * PumpStepGain ;    break;
	   		case 2:   K2Pum[NPum] = (-1) * PumpStepGain ;    break;
	   		case 3:   K3Pum[NPum] = (-1) * PumpStepGain ;    break;
	   		case 4:   K4Pum[NPum] = (-1) * PumpStepGain ;    break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
	  
	  
    }  
	
	for(int NSig=0;NSig<NumSigCh;NSig++){   
	  
	  	  switch(RKStage){
     
		   	case 1:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K1Sig[NSig] * DeltaZ / 2;    break;
	   		case 2:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K2Sig[NSig] * DeltaZ / 2;    break;
	   		case 3:   vSignalPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] + K3Sig[NSig] * DeltaZ;    	break;
	   		case 4:   oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] + 
	        			( K1Sig[NSig] + 2 * K2Sig[NSig] + 2 * K3Sig[NSig] + K4Sig[NSig] )* DeltaZ / 6;    break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
      
    }

    for(int NPum=0;NPum<NumPumCh;NPum++){
		 switch(RKStage){
     
		   	case 1:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K1Pum[NPum] * DeltaZ/ 2;    break;
	   		case 2:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K2Pum [NPum] * DeltaZ/ 2;    break;
	   		case 3:   vPumpPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] + K3Pum[NPum] * DeltaZ;   break;
	   		case 4:   oPumpPower[NPum*NumZSteps+IntStepI]   = oPumpPower[NPum*NumZSteps+IntStepI-1] + 
	        			( K1Pum[NPum] + 2 * K2Pum[NPum] + 2 * K3Pum[NPum] + K4Pum[NPum] ) * DeltaZ / 6;   break;
	  
       		default:
         		cerr << "With RK-Stage must between 1 and 4" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
    } 
  
  }
  
 
} // endl of RKShootingtext


*/

// ############################
// Currectly, I only use Improved Eular method 
// so FirstGuessForwardShooting is used for guess
// ## By: Jonathan Hu

void RamanAmp::FirstGuessForwardShooting(int IntStepI, double * PreSig, double * PrePum)
{
  double SignalStepGain, PumpStepGain;
if(FixSignalShootingSign == 0){	      
  for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
    SignalStepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       SignalStepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI)*oPumpPower[P*NumZSteps+IntStepI-1];
//       cout << "Gain = " << SignalStepGain << "  test =" << GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI) << endl;
    }

    for(int S=0;S<NumSigCh;S++){
      SignalStepGain +=
      GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI)*oSignalPower[S*NumZSteps+IntStepI-1]; 
      }
    
    oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
        exp((SignalStepGain-GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI))*DeltaZ(IntStepI));
//    cout << oSignalWaveLth[NSig] << "  " << oSignalPower[NSig*NumZSteps+IntStepI] << "  " << SignalStepGain << "  " 
//			<< GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI)*4343 << endl;
 	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
  }
}
if(FixPumpShootingSign == 0){
  for(int NPum=0;NPum<NumPumCh;NPum++){			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    PumpStepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI) * oPumpPower[P*NumZSteps+IntStepI-1];
    }
    for(int S=0;S<NumSigCh;S++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI) * oSignalPower[S*NumZSteps+IntStepI-1]; 
    }  
    oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI-1] * exp( (-1) 
			* (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI))*DeltaZ(IntStepI));
  	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI];
    if (oPumpPower[NPum*NumZSteps+IntStepI] < 0){
	    cout << " StepI = "  << IntStepI << ", NPum = " << NPum << endl;
		cout << oPumpPower[NPum*NumZSteps+IntStepI] << "=" << oPumpPower[NPum*NumZSteps+IntStepI-1]
			 << "*" << exp( (-1) * (PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI))*DeltaZ(IntStepI)) << endl;  	
	    cerr << "PumpPower can't be negative" << endl 
             << "Aborting now! " << endl;
        exit(1);
		}
  }
} 

}// end of FirstGuessForwardShooting

// ##############################################
// for guess- correction as Improved Eular method
// ## By: Jonathan Hu

double RamanAmp::OptimizeForwardShootingIntegration(int IntStepI, double * PreSig, double * PrePum)
{
 double g;
 double *TempSigPower = new double[NumSigCh];
 double *TempPumPower = new double[NumPumCh];

  double StepGain;   
  double RMS;
  RMS = 0;

  for(int NSig=0;NSig<NumSigCh;NSig++){ 

      switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
	   case 4:     // using last Itr trap Rule should be corrected
//	     TempSigPower[NSig] =  (oSignalPowerPre[NSig*NumZSteps+IntStepI] + oSignalPowerPre[NSig*NumZSteps+IntStepI-1])/2;
		 		   
	     break;

       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }

  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ;
         break;

	   case 4:     // Trape Rule ?? should be corrected
	     TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
		 break;
	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
  }
  
if(FixSignalShootingSign ==0){   	
  for(int NSig=0;NSig<NumSigCh;NSig++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI) * TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI) * TempSigPower[S]; 
    }
//    cout << " oSignalPower" << NSig << "=" <<  oSignalPower[NSig*NumZSteps+IntStepI] ;
    if (oSignalPower[NSig*NumZSteps+IntStepI] < 0){
	   cerr << "Signal Power can't be negative" << endl 
       << " Aborting now! " << endl;
         exit(1);
		 }
	

	oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
       exp((StepGain-GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI))*DeltaZ(IntStepI));
	
	RMS +=  abs(oSignalPower[NSig*NumZSteps+IntStepI] - PreSig[NSig]) / oSignalPower[NSig*NumZSteps+IntStepI]; 

	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
//	cout << " RMS=" << " " << RMS ;  
  }
}
if(FixPumpShootingSign == 0){
  for(int NPum=0;NPum<NumPumCh;NPum++){
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI)*TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
      StepGain +=
      GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI)*TempSigPower[S]; 
    }
	oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI-1] * 
       exp( (-1) * (StepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI)) * DeltaZ(IntStepI));  
    if (oPumpPower[NPum*NumZSteps+IntStepI] < 0){
	    cout << IntStepI ;
		cout << oPumpPower[NPum*NumZSteps+IntStepI] << "=" << oPumpPower[NPum*NumZSteps+IntStepI-1]
			 << "*" << exp( (-1) * (StepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI)) * DeltaZ(IntStepI));  	
	    cerr << "PumpPower can't be negative" << endl 
             << " Aborting now! " << endl;
        exit(1);
		}

//	Temp = RMS;	
	RMS += abs(oPumpPower[NPum*NumZSteps+IntStepI] - PrePum[NPum]) / oPumpPower[NPum*NumZSteps+IntStepI];
	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI];
/*    if(IntStepI==10||IntStepI==9){
		Temp = RMS - Temp;
   		cout << " RMS=" << Temp ;
		//(oPumpPower[NPum*NumZSteps+IntStepI] - PrePum[NPum]) 
			//         / oPumpPower[NPum*NumZSteps+IntStepI] << endl;
  	}
*/
//	cout << " RMS=" << " " << RMS ; 
  }
}
  			
  RMS = (RMS/(NumSigCh + NumPumCh));
  delete [] TempSigPower ;
  delete [] TempPumPower ;

  return(RMS);

} // end OptimizeForwardShootingIntegration

// ############################
// out of date not used anymore
// Shooting algorithm, please refer JOSAB sumbission
// ## By: Jonathan Hu

double RamanAmp::OptimizeForwardPumpShootingIntegration(int IntStepI, double * PreSig, double * PrePum)
{
  double g;
  double *TempSigPower = new double[NumSigCh];
  double *TempPumPower = new double[NumPumCh];
  
 // double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
  double StepGain;   
  double RMS, Temp;
  RMS = 0;

  
  
  for(int NSig=0;NSig<NumSigCh;NSig++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
	   case 4:     // using last Itr trap Rule should be corrected
//	     TempSigPower[NSig] =  (oSignalPowerPre[NSig*NumZSteps+IntStepI] + oSignalPowerPre[NSig*NumZSteps+IntStepI-1])/2;
		 		   
	     break;

       case 5:     // using Improved Rule's Rule should be corrected

         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 

         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = ( 3 * TempSigPower[NSig] + 
		 		oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ) /  4;
         break;	 
		 
		 
       case 0:
          g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
          cout << " Psignal Trape-orig= " <<
	    (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2 - 
	    oSignalPower[NSig*NumZSteps+IntStepI-1] << " Ave - Orig = " << 
             oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g)  - 
	     oSignalPower[NSig*NumZSteps+IntStepI-1] << endl;
	     
          TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
	  break;
	  
       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ;
         break;

	   case 4:     // Trape Rule ?? should be corrected
	     TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
		 break;

	   
	   case 5:  // Improved rule  ?? should be corrected
	   	 TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
         g = oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] / oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = ( 3 * TempPumPower[NPum] + 
		       oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1] *( g - 1 ) /   log(g) ) / 4;
	      
	     break;

      case 0:	  
	  
	     break;
	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P], IntStepI)*TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
      StepGain +=
      GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S], IntStepI)*TempSigPower[S]; 
    }
	oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI-1] * 
       exp( (-1) * (StepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI)) * DeltaZ(IntStepI));  
	RMS += abs(oPumpPower[NPum*NumZSteps+IntStepI] - PrePum[NPum]) / oPumpPower[NPum*NumZSteps+IntStepI];
	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI];

//	cout << " RMS=" << " " << RMS ; 
  }
  RMS = (RMS/NumPumCh);
  delete [] TempSigPower ;
  delete [] TempPumPower ;
  return(RMS);

} // end OptimizeForwardPumpShootingIntegration


// #############################
// out of date, not used anymore
// ## By: Jonathan Hu

double RamanAmp::OptimizeForwardSignalShootingIntegration(int IntStepI, double * PreSig, double * PrePum)
{
  double g;

  double *TempSigPower = new double[NumSigCh];
  double *TempPumPower = new double[NumPumCh];

//  double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
  double StepGain;   
  double RMS, Temp;
  RMS = 0;

  
  
  for(int NSig=0;NSig<NumSigCh;NSig++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ;
         break;


	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
  }
  
  for(int NSig=0;NSig<NumSigCh;NSig++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P], IntStepI) * TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S], IntStepI) * TempSigPower[S]; 
    }
//    cout << " oSignalPower" << NSig << "=" <<  oSignalPower[NSig*NumZSteps+IntStepI] ;
    if (oSignalPower[NSig*NumZSteps+IntStepI] < 0){
	   cerr << "Signal Power can't be negative" << endl 
       << " Aborting now! " << endl;
         exit(1);
		 }
	

	oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
       exp((StepGain-GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI))*DeltaZ(IntStepI));
	
	RMS +=  abs(oSignalPower[NSig*NumZSteps+IntStepI] - PreSig[NSig]) / oSignalPower[NSig*NumZSteps+IntStepI]; 

	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
//	cout << " RMS=" << " " << RMS ;  
  }

  RMS = (RMS/NumSigCh);
  delete [] TempSigPower ;
  delete [] TempPumPower ;
  return(RMS);

} // end OptimizeForwardShootingIntegration



// ###################################
// RelaxRaman method integrate the signal waves forward and then integrate the pump waves
// backward to ensure that both boundary conditions are automatically satisfied. 
// This relaxation algorithm is then iterated until the solution converges
// ## By: Jonathan Hu

void RamanAmp::RelaxRamanIteration()
{   	     
 double rms;
 double *oPumpTrack = new double[NumZSteps*NumPumCh];	
 int NumItr;
 char  C_String[40];
 string FileName;

   for(int i=1;i<NumZSteps;i++){
       	SignalOnlyPropagation(i);
   }
  	for(int i=0;i<NumZSteps;i++){	 
 	 	for(int NPum=0;NPum<NumPumCh;NPum++){	
			oPumpTrack[NPum*NumZSteps+i] = 1;
		}
	}				 	
	for(int NPum=0;NPum<NumPumCh;NPum++) {
		oPumpPower[NPum*NumZSteps+NumZSteps-1] = oPumpTargetBackup[NPum];
	}   
// for(int NumItr=0;NumItr<NumIteration;NumItr++){
   
//  for(; NumPumCh<=NumPumCh;NumPumCh++){

	  AccurateSign =0;

  NumItr = 0;
  do{
 
  NumItr = NumItr +1;

   for(int i=NumZSteps-2;i>=0;i--){
    	BackwardStep(i);
   }

//   for(int i=1;i<NumZSteps;i++){
//        ForwardStep(i);
//   }

	FixPumpShootingSign = 1;
	for(int i=1;i<NumZSteps;i++){
//        if(BlowupSign==0){
		  	ForwardShooting(i);
//		}
    }	 
	FixPumpShootingSign = 0;



/*   for(int NSig=0;NSig<NumSigCh;NSig++){   
 //      cout << "Psignal#" << NumItr << "  " <<
//	   oSignalPower[NSig*NumZSteps+NumZSteps-1] << endl;
//       rmsS += DeltaPsignal[NSig];
       sprintf(C_String,"RamanItr%iSig%i.dat",NumItr,NSig);
       FileName = JobRaman + C_String;  
//       WriteSignalWaveLthPower(FileName, NSig);
       sprintf(C_String,"SignalItr%iSample%i.dat",NumItr,NSig);
       FileName = JobRaman + C_String;  
//       WriteSignalWaveLthPowerWith$(FileName, NSig);

//   *LogFile << endl;

//	   if (NSig==0) {
//	   	   		*LogFile << NumItr << "   " ;
//	   }
//	   *LogFile << oSignalPower[NSig*NumZSteps + NumZSteps -1 ] << "   " ;       
   }
 
//   *LogFile << endl;
   
   rmsS = rmsS / NumSigCh;
   rmsP=0;
   for(int NPum=0;NPum<NumPumCh;NPum++){
   
//       rmsP += DeltaPpump[NPum];
	      	      	cout << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1];
//       cout << "PPump#" << NumItr << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1];
 //      cout << "   DeltaPpump#" << NumItr << "  "  << DeltaPpump[NPum]  << endl;
       sprintf(C_String,"RamanItr%iPum%i.dat",NumItr,NPum);
       FileName = JobRaman + C_String;  
 //      WritePumpWaveLthPower(FileName, NPum);
   }
   rmsP /= NumPumCh ;
   if (rmsP < 1e-9){
	  	AccurateSign = 1;	
   }
   cout << "   Itr# " << NumItr << "  DeltaPsignalTotal = "  << rmsS;
   cout << "  DeltaPpumpTotal = "  << rmsP << endl;   
 */ 

	rms =0;
//	cout << "pump track " << endl;
 	for(int NPum=0;NPum<NumPumCh;NPum++){	
//		cout << "pump track # " << NPum << endl;
	  	for(int i=0;i<NumZSteps;i++){	 
	 		rms += sq( ( oPumpPower[NPum*NumZSteps+i] - oPumpTrack[NPum*NumZSteps+i] ) 
					/ oPumpPower[NPum*NumZSteps+i] );
			oPumpTrack[NPum*NumZSteps+i] = oPumpPower[NPum*NumZSteps+i] ;
//			cout << oPumpTrack[NPum*NumZSteps+i] << "  " ;
		}
	}
// cout << endl;


	rms = sqrt( rms / (NumPumCh*NumZSteps) );

	cout << "   rms=" << rms << endl;
//	cout << rms << "  " << endl;

	if (abs(rms) <IterationAccurateTh){
//		cout <<  "rms="  << rms;
		AccurateSign = 1;	
		}

  }while(AccurateSign==0&&NumItr<500);

  	if (NumItr>=500) {
		cout << "iteration doesn't converge, exit";
		exit(1);
	}
 cout <<  "  NumItr = " << NumItr << endl;

 *LogFile << "Raman Iteration Result Pump" << endl;
	LogFile->precision(15);
 	for(int NPum=0;NPum<NumPumCh;NPum++){	
		*LogFile << oPumpWaveLth[NPum] << ":   " << oPumpPower[NPum*NumZSteps] << " --> "  << 
				oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl; 
	}
	*LogFile << "Raman Iteration Result Signal" << endl;
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
 		*LogFile << oSignalWaveLth[NSig] 
			//	<< ":   " << oSignalPower[NSig*NumZSteps] << " --> "  << 
				<< "  " << oSignalPower[NSig*NumZSteps+NumZSteps-1] << endl; 
	}
 LogFile->precision(6);
 delete [] oPumpTrack ; 
 
}   // end of RelaxRamanIteration


// ###################################
// This is forward step
// please refer JOSAB equation (1.a)
// by Joanthan
// 
void RamanAmp::ForwardStep(int StepI)
{  
  double *PreSigPower = new double[NumSigCh];

//  double PreSigPower[NumSigCh]; 
  double SignalStepGain;
  double monitor;
  
  for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
    SignalStepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       SignalStepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepI)*oPumpPower[P*NumZSteps+StepI];
     
//       cout << "Gain = " << GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepI) 
//          << "  oPumpPower["  << P*NumZSteps+NumZSteps-StepI-1 << "] = " << oPumpPower[P*NumZSteps+NumZSteps-StepI-1]
//	  << endl;
//      cout << "SignalStepGainfromP = " << SignalStepGain << endl;
  }

  for(int S=0;S<NumSigCh;S++){
    SignalStepGain +=
    GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepI)*oSignalPower[S*NumZSteps+StepI-1]; 
//       cout << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepI) << "SignalStepGainfromS = " << SignalStepGain << endl;
    }
    
    oSignalPower[NSig*NumZSteps+StepI] = oSignalPower[NSig*NumZSteps+StepI-1] * 
        exp((SignalStepGain-GetAttenuationFiber(oSignalWaveLth[NSig],StepI))*DeltaZ(StepI));
	PreSigPower[NSig] = oSignalPower[NSig*NumZSteps+StepI];
  }

  for(int i=0;i<NumIntItr;i++){
//		InternalItrNum++;
    monitor = OptimizeForwardIntegration(StepI,PreSigPower);
	//				rms = OptimizeForwardShootingIntegration(StepI, PreSigPower, PrePumPower);
	   	//		cout << endl << "InItr#" << InternalItrNum << " ";
		//		cout << "total rms = " << rms ;
	
	}   
//  cout << " rms =" << monitor ;
 if ( abs(monitor)<1 ){
	}else{		
	 cout << monitor << endl;
	 cerr << "Error: Internal Iteration does not converge5 " 
     		<< "  exit "<< endl;
					exit(1);
  }

    
/* if(StepI == (NumZSteps-1)){  
   for(int NSig=0;NSig<NumSigCh;NSig++){	
//  cout << "SignalSignalStepGainSig#"  << NSig << " = " << SignalStepGain << endl;
//      DeltaPsignal[NSig] += abs(oSignalPower[NSig*NumZSteps+StepI] - oSignalPowerPre[NSig*NumZSteps+StepI]) 
//	  						/ oSignalPower[NSig*NumZSteps+StepI];
	  oSignalPowerPre[NSig*NumZSteps+StepI] = oSignalPower[NSig*NumZSteps+StepI];

   }
 }
*/ 
 delete [] PreSigPower; 	
}  // end of ForwardStep

// ##################################


// ##############################
// backward Step
// please refer JOSAB equation (1.a)
// Jonathan
void RamanAmp::BackwardStep(int StepI)
{ 
 double *PrePumPower = new double[NumPumCh];
// double PrePumPower[NumPumCh]; 
  double monitor;
 double PumpStepGain;

 for(int NPum=0;NPum<NumPumCh;NPum++)
 {			// for NumPumCh X (NumSigCh+NumPumCh) Matrix
    PumpStepGain = 0;

    for(int P=0;P<NumPumCh;P++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],StepI+1) * oPumpPower[P*NumZSteps+StepI+1];
//        cout << "PumpStepGaininPump = " << PumpStepGain <<  "  Gain = " 
//		 		<< GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],StepI+1) << endl;
    }

    for(int S=0;S<NumSigCh;S++){
       PumpStepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],StepI+1) * oSignalPower[S*NumZSteps+StepI]; 
//	   cout << " run to here" << endl;
//	   exit(1);
 //   cout << "getGain = " << GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],NumZSteps-StepI) << "   oSignalPower["
  //        <<  S*NumZSteps+NumZSteps-StepI-1 << "] = " << oSignalPower[S*NumZSteps+NumZSteps-StepI-1] ;
    
//    cout << " PumpStepGaininSignal  " << StepI << "  " << PumpStepGain << endl;
   }
//  cout << "PumpStepGain" << StepI << "  " << PumpStepGain << endl;
  
    oPumpPower[NPum*NumZSteps+StepI] = oPumpPower[NPum*NumZSteps+StepI+1] 
			* exp((PumpStepGain-GetAttenuationFiber(oPumpWaveLth[NPum],StepI+1))*DeltaZ(StepI));
  	PrePumPower[NPum] = oPumpPower[NPum*NumZSteps+StepI];
//	cout << PumpStepGain << "  " << GetAttenuationFiber(oPumpWaveLth[NPum],StepI+1) << "  "
//			<< DeltaZ(StepI) << "  " << oPumpPower[NPum*NumZSteps+StepI] << endl;
  }  


  for(int i=0;i<NumIntItr;i++){
//		InternalItrNum++;
    	monitor = OptimizeBackwardIntegration(StepI,PrePumPower);
	//				rms = OptimizeForwardShootingIntegration(StepI, PreSigPower, PrePumPower);
	   	//		cout << endl << "InItr#" << InternalItrNum << " ";
		//		cout << "total rms = " << rms ;
	
	}
//  cout << "rms =" << monitor ;
  if ( abs(monitor)<1 ){
	}else{		
	 cout << monitor << endl;
	 cout << "StepI=" << StepI;
	 cerr << "Error: Internal Iteration does not converge7 " 
     		<< "  exit "<< endl;
				exit(1);
				BlowupSign=1;
  }

/* if(StepI ==0 &&BlowupSign==0){ 
   for(int NPum=0;NPum<NumPumCh;NPum++){     
//    		DeltaPpump[NPum] += abs(oPumpPower[NPum*NumZSteps+StepI] - oPumpPowerPre[NPum*NumZSteps+StepI]) 
//						/ oPumpPower[NPum*NumZSteps+StepI];

//    cout << "   DeltaPpump[" << NPum << "] = " << DeltaPpump[NPum] <<
//       "   oPumpPwer["  << NPum*NumZSteps+StepI-1 << "] =" << oPumpPower[NPum*NumZSteps+StepI-1] << endl;


    		oPumpPowerPre[NPum*NumZSteps+StepI] = oPumpPower[NPum*NumZSteps+StepI];
          
	  
	}
	  
 } 
*/
 delete [] PrePumPower;

   
}  // end of BackwardStep


// #############################################
// Forward last Round (including noise)
// ## By: Jonathan Hu

double RamanAmp::OptimizeForwardIntegrationLastRound(int IntStepI, double * PreSig)
{
  double g, RMS;
  double *TempSigPower = new double[NumSigCh];
  double *TempPumPower = new double[NumPumCh];

//  double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
  double StepGain, ASEStepGain;   
  double C1,C2;    
  
  for(int NSig=0;NSig<NumSigCh;NSig++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
	   case 4:     // using last Itr trap Rule
//	     TempSigPower[NSig] =  (oSignalPowerPre[NSig*NumZSteps+IntStepI] + oSignalPowerPre[NSig*NumZSteps+IntStepI-1])/2;
		 		   
	     break;

       case 5:     // using Improved Rule's Rule

         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 

         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = ( 3 * TempSigPower[NSig] + 
		 		oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ) /  4;
         break;
		 

       case 0:
          g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
          cout << " Psignal Trape-orig= " <<
	    (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2 - 
	    oSignalPower[NSig*NumZSteps+IntStepI-1] << " Ave - Orig = " << 
             oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g)  - 
	     oSignalPower[NSig*NumZSteps+IntStepI-1] << endl;
	     
          TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
	  break;
	  
       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps + IntStepI]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ;
         break;


      case 0:	  
	  
	     break;
	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  

  
  RMS =0;
  for(int NSig=0;NSig<NumSigCh;NSig++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],IntStepI) * TempPumPower[P];
    }
 /*   for(int S=0;S<NumSigCh;S++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],IntStepI) * TempSigPower[S]; 
    }
*/
    oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
       exp((StepGain-GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI))*DeltaZ(IntStepI));  
   	RMS +=  abs(oSignalPower[NSig*NumZSteps+IntStepI] - PreSig[NSig]) / oSignalPower[NSig*NumZSteps+IntStepI]; 

	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
  }

  if(!(TypeRamanNoise == RAMAN_NOISE_OFF)){
	for(int NASE=0;NASE<NumASECh;NASE++){ 
/*    switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempASEPower[NASE]=  oASEPower[NASE * NumZSteps + StepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempASEPower[NNoi]=  (oASEPower[NNoi*NumZSteps+StepI] + oASEPower[NNoi*NumZSteps+StepI-1] ) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         cerr << "With type Intergration must not be 3 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
		// need to be recoded
       //  g = oNoisePower[NNoi*NumZSteps+StepI] / oNoisePower[NNoi*NumZSteps+StepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
       //  TempNoiPower[NNoi] = oNoisePower[NNoi*NumZSteps+StepI-1] *( g - 1 ) /   log(g) ;
         break;
  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type
    }
*/ 	   ASEStepGain = 0;
	   C2 = 0;
    	for(int P=0;P<NumPumCh;P++){
       		ASEStepGain +=
       			GetRamanGain(oASEWaveLth[NASE],oPumpWaveLth[P],IntStepI)*TempPumPower[P];
			C2 += GetRamanGain(oASEWaveLth[NASE],oPumpWaveLth[P],IntStepI)*TempPumPower[P]
					* GetTempDepFactor(oASEWaveLth[NASE],oPumpWaveLth[P]);
		} 
		
/*		for(int S=0;S<NumSigCh;S++){
    		ASEStepGain +=
    			GetRamanGain(oASEWaveLth[NASE],oSignalWaveLth[S],IntStepI)*TempSigPower[S]; 
	    	C2 += GetRamanGain(oASEWaveLth[NASE],oSignalWaveLth[S],IntStepI)*TempSigPower[S]
					* GetTempDepFactor(oASEWaveLth[NASE],oSignalWaveLth[S]);
		}
 */   	C1 = ASEStepGain - GetAttenuationFiber(oASEWaveLth[NASE],IntStepI);
		C2 *=   2 * hPlanck * Wavelength2Frequency(oASEWaveLth[NASE]) * DeltaFreq;
		oASEPower[NASE*NumZSteps+IntStepI] = oASEPower[NASE*NumZSteps + IntStepI - 1] * exp(C1 * DeltaZ(IntStepI)) 
				+ exp(C1 * DeltaZ(IntStepI)) * (  exp(C1 * DeltaZ(IntStepI)) * C2 * DeltaZ(IntStepI));
	} // ## end for NASE loop
  } // ## end for TypeRamanNoise


  RMS = (RMS / NumSigCh );
  delete [] TempSigPower ;
  delete [] TempPumPower ;

  return(RMS);


} // end OptimizeForwardIntegrationLastRound


// ####################################################
// GetTempDepFactor used in Noise calculation
// ## By: Jonathan Hu

double RamanAmp::GetTempDepFactor(double Wavelength1,double Wavelength2)
{

 double KBoltzmann = 8.617e-5 * 1.6e-19;   // eV/K
 double Temperature = 273+0;
 double Frequency1, Frequency2;
 double Nphon;

 Frequency1 = Wavelength2Frequency(Wavelength1);
 Frequency2 = Wavelength2Frequency(Wavelength2); 

 Nphon = 1/(exp(hPlanck * abs(Frequency1 - Frequency2) / KBoltzmann / Temperature) - 1);

 if (Wavelength1>Wavelength2){

     return( Nphon + 1 );
 } else {
 
     return( -1.0 * Nphon);
 }

} // end of GetTempDepFactor

// #############################################
// used for guess-correction
// ## By: Jonathan Hu

double RamanAmp::OptimizeForwardIntegration(int IntStepI, double * PreSig)
{
  double g, RMS;
  double *TempSigPower = new double[NumSigCh];
  double *TempPumPower = new double[NumPumCh];

//  double TempSigPower[NumSigCh], TempPumPower[NumPumCh];
  double StepGain;   
    
  
  for(int NSig=0;NSig<NumSigCh;NSig++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         break;
     
       case 2:    // Forward estimate Trape Rule
         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /  log(g) ;
         break;
		 
	   case 4:     // using last Itr trap Rule
//	     TempSigPower[NSig] =  (oSignalPowerPre[NSig*NumZSteps+IntStepI] + oSignalPowerPre[NSig*NumZSteps+IntStepI-1])/2;
		 		   
	     break;

       case 5:     // using Improved Rule's Rule

         TempSigPower[NSig] =  (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2; 

         g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
         		 
		 TempSigPower[NSig] = ( 3 * TempSigPower[NSig] + 
		 		oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g) ) /  4;
         break;
		 

       case 0:
          g = oSignalPower[NSig*NumZSteps+IntStepI] / oSignalPower[NSig*NumZSteps+IntStepI-1]; 
          cout << " Psignal Trape-orig= " <<
	    (oSignalPower[NSig*NumZSteps+IntStepI] + oSignalPower[NSig*NumZSteps+IntStepI-1])/2 - 
	    oSignalPower[NSig*NumZSteps+IntStepI-1] << " Ave - Orig = " << 
             oSignalPower[NSig*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g)  - 
	     oSignalPower[NSig*NumZSteps+IntStepI-1] << endl;
	     
          TempSigPower[NSig]=  oSignalPower[NSig*NumZSteps+IntStepI-1]; 
	  break;
	  
       default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  for(int NPum=0;NPum<NumPumCh;NPum++){ 
 
     switch(TypeIntergration)
     {
       case 1:   // Left Boundary
         TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]; 
         break;
     
       case 2:    // estimate Forward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
         break;
     
       case 3:     //  Forward Distance Average
         g = oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] / oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1] *( g - 1 ) /   log(g) ;
         break;

	   case 4:     // Trape Rule
	     TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
		 break;

	   
	   case 5:  // Improved rule
	   	 TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] + oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]) / 2; 
         g = oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI] / oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1]; 
		 if (g==1)
		 		   cout << "g ==1" << endl;
         TempPumPower[NPum] = ( 3 * TempPumPower[NPum] + 
		       oPumpPower[NPum*NumZSteps+NumZSteps-IntStepI-1] *( g - 1 ) /   log(g) ) / 4;
	      
	     break;

      case 0:	  
	  
	     break;
	  
      default:
         cerr << "With type Intergration must be 1 or ?" << endl 
             << "Aborting now! " << endl;
         exit(1);
      } //end switch type

  }
  
  RMS =0;
  for(int NSig=0;NSig<NumSigCh;NSig++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],IntStepI) * TempPumPower[P];
    }
    for(int S=0;S<NumSigCh;S++){
       StepGain +=
       GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],IntStepI) * TempSigPower[S]; 
    }
    oSignalPower[NSig*NumZSteps+IntStepI] = oSignalPower[NSig*NumZSteps+IntStepI-1] * 
       exp((StepGain-GetAttenuationFiber(oSignalWaveLth[NSig],IntStepI))*DeltaZ(IntStepI));  
   	RMS +=  abs(oSignalPower[NSig*NumZSteps+IntStepI] - PreSig[NSig]) / oSignalPower[NSig*NumZSteps+IntStepI]; 

	PreSig[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
  }

  RMS = (RMS / NumSigCh );
  delete [] TempSigPower ;
  delete [] TempPumPower ;

  return(RMS);


} // end OptimizeForwardIntegration


//  #################################################


// ################################################
// used for guess-correction
// ## By: Jonathan Hu

double RamanAmp::OptimizeBackwardIntegration(int IntStepI, double * PrePum)
{

  double g,RMS;
  double *TempSigPower = new double[NumSigCh];
  double *TempPumPower = new double[NumPumCh];
//  double TempPumPower[NumPumCh], TempSigPower[NumSigCh];
  double StepGain;   
  
  for(int NPum=0;NPum<NumPumCh;NPum++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
  
    switch(TypeIntergration)
    {
       case 1:
	     TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI+1];
		 break;
     
       case 2:   //  Backward Trape Rule
         TempPumPower[NPum]=  (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI+1])/2; 
  //       cout << oPumpPower[NPum*NumZSteps+IntStepI] << "  " << oPumpPower[NPum*NumZSteps+IntStepI-1];
		 break;
     
       case 3:  //        Backward Distance Average
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI+1]; 
         TempPumPower[NPum] = oPumpPower[NPum*NumZSteps+IntStepI+1] *( g - 1 ) /   log(g) ;
         break;
		 
	   
       case 0:  //  Debug
         g = oPumpPower[NPum*NumZSteps+IntStepI] / oPumpPower[NPum*NumZSteps+IntStepI-1]; 
         cout << "Pump Delta = "  <<
	   (oPumpPower[NPum*NumZSteps+IntStepI] + oPumpPower[NPum*NumZSteps+IntStepI-1])/2 -
	    oPumpPower[NPum*NumZSteps+IntStepI-1] <<
	   " Aver - orig = "  <<
	   oSignalPower[NPum*NumZSteps+IntStepI-1] *( g - 1 ) /   log(g)  - 
	    oPumpPower[NPum*NumZSteps+IntStepI-1]
	   << endl;
	 TempPumPower[NPum]=  oPumpPower[NPum*NumZSteps+IntStepI-1];
         break;
	 	 
       default:
         cerr << "With type Intergration must be 1 or ?"  
              << endl << "Aborting now! " << endl ;
         exit(1);
         break;
     
    } //end switch type

  }
  
  for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
  
    switch(TypeIntergration)
    {
       case 1:
 	     TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI];
		 break;
     
       case 2:   //  Estmate Backward Trape Rule
 		 TempSigPower[NSig] = (oSignalPower[NSig*NumZSteps+IntStepI] + 
		 			oSignalPower[NSig*NumZSteps+IntStepI+1]) / 2 ;
         break;
     
       case 3:  //        Backward Distance Average
         g = oSignalPower[NSig*NumZSteps+IntStepI+1] /
		 			oSignalPower[NSig*NumZSteps+IntStepI]; 
		 if(g==1)
		   cout << "g ==1" << endl;
         TempSigPower[NSig] = oSignalPower[NSig*NumZSteps+IntStepI] *( g - 1 ) / log(g) ;
         break;
		


 
       case 0:  //  Debug not

         break;
	 	 
       default:
         cerr << "With type Intergration must be 1 or ?"  
              << endl << "Aborting now! " << endl ;
         exit(1);
         break;
     
    } //end switch type

  } 
	RMS = 0;     
  for(int NPum=0;NPum<NumPumCh;NPum++){
  
    StepGain = 0;
    for(int P=0;P<NumPumCh;P++){
       StepGain +=
       GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],IntStepI+1)*TempPumPower[P];
//	   cout << " GetRamanGain=" << GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],NumZSteps-IntStepI);
//	   cout << "  TempPum=" << TempPumPower[P];
	 
//	cout << " StepGain=" << StepGain;
	
    }
    for(int S=0;S<NumSigCh;S++){
      StepGain +=
      GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],IntStepI+1)*TempSigPower[S]; 
//	cout << " StepGain=" << StepGain;
		//  cout << "  TempSig=" << TempSigPower[S];
    }
//	cout << " StepGain=" << StepGain;
    oPumpPower[NPum*NumZSteps+IntStepI] = oPumpPower[NPum*NumZSteps+IntStepI+1] * 
       exp((StepGain-GetAttenuationFiber(oPumpWaveLth[NPum],IntStepI+1))*DeltaZ(IntStepI+1)); 
	   
	RMS += abs(oPumpPower[NPum*NumZSteps+IntStepI] - PrePum[NPum]) / oPumpPower[NPum*NumZSteps+IntStepI];
 //   cout << "rms= " << RMS;
  	PrePum[NPum] = oPumpPower[NPum*NumZSteps+IntStepI]; 
  }
  
  RMS = (RMS/NumPumCh);
  delete [] TempSigPower ;
  delete [] TempPumPower ;

  return(RMS);

} // end OptimizeBackwardIntegration

// ###############################################################

void RamanAmp::SetAmpLength(double Value) 
{
  AmpLength=Value;
} 

// #########################################

void RamanAmp::SetPumpPower(double Value)
{  if(NumPumCh != 1){
   		 cerr << "NumPumCh = " << NumPumCh << endl;		
		 cerr << "Currently, this function works only for in the case of NumPumCh is equal to One"<< endl 
             << "Aborting now! " << endl;
         exit(1);
  } 
  if(Value < 0){
     cerr << "Pump power much large than 0"<< endl 
          << "Aborting now! " << endl;
     exit(1);
  } 
  oPumpPowerInput[0] = Value * PumpPowerFactor ;
  //* PumpPowerFactor;
}

// #############################################

void RamanAmp::SetTypeRamanNoise(typeRamanNoise TypeRamanNoise2)
{

 TypeRamanNoise = TypeRamanNoise2;

}

// ################################################
/*
double RamanAmp::GetRamanGain(double Wavelength1, double Wavelength2)
{
 double DeltaWavelength = abs(Wavelength1 - Wavelength2);
 double Interpolate;
 
 if (DeltaWavelength < 1e-12)
    return(0);
 if( DeltaWavelength > 152 * WavelengthStepGain)
    return(0);
 
 Interpolate = oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain)] + 
 	(oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain) + 1] - oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain)])
	* (DeltaWavelength/WavelengthStepGain - int(DeltaWavelength/WavelengthStepGain) ) ;

 if (Wavelength1>Wavelength2){

     return( Interpolate );
 } else {
 
     return( -Interpolate * (Wavelength1 / Wavelength2) );
 }

}  // end GetRamanGain

*/
// ################################################

double RamanAmp::GetRamanGain(double Wavelength1, double Wavelength2, int NumSteps)
{

 double Interpolate;
 double Frequency1, Frequency2;
 if(  ReadPlainFileGainLossFlag == 1){
    return( GetRamanGainFreqGrid(Wavelength1, Wavelength2, NumSteps));
 }

 
 if(ReadRamanGainLossWavelengthGridFlag == 1){ 
 	 if( NumSteps <0 || NumSteps > NumZSteps)
   	 	{ 	cerr << " NumSteps is not in range" << endl
			<< "aborting" << endl;
			exit(1);
		}

 int FIndex=-1;
 NumSteps--;
 do{
 	FIndex++;
 }while(NumSteps > FiberLocation[FIndex]);
 
// if(abs(Wavelength1 - Wavelength2)<50e-9)
// {
// 		return(GetRamanGainClose(Wavelength1, Wavelength2, FIndex));
// }

 Frequency1 = Wavelength2Frequency(Wavelength1);
 Frequency2 = Wavelength2Frequency(Wavelength2);


 double DeltaWavelength = abs(Wavelength1 - Wavelength2)- (WaveLengthStartGain - PumpWavelengthForGain * 1e-9);

  
  
 if( DeltaWavelength < 0 || DeltaWavelength > NumGainWidth * WavelengthStepGain)
    { 	// cout << " not in the range" << endl;
		return(0);
	}
 
 Interpolate = oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain)] + 
 	(oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain) + 1] 
		- oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain)])
	* ( ( DeltaWavelength - WavelengthStepGain * int(DeltaWavelength/WavelengthStepGain) ) / WavelengthStepGain ) ;
// cout << " Interpolate " <<Interpolate << endl;


 if (Wavelength1>Wavelength2){

     return( Interpolate );
 } else {
 
     return( -Interpolate * (Frequency1 / Frequency2) );
 }
}
	cout << " error no input file " << endl;
	exit(1);
}  // end GetRamanGain

// ##################################################################

double RamanAmp::GetRamanGainFreqGrid(double Wavelength1, double Wavelength2, int NumSteps)
{

 double Interpolate;
 double Frequency1, Frequency2;
 
 int FIndex=-1;
 NumSteps--;
 do{
 	FIndex++;
 }while(NumSteps > FiberLocation[FIndex]);
  
 Frequency1 = Wavelength2Frequency(Wavelength1);
 Frequency2 = Wavelength2Frequency(Wavelength2);
 
 double DeltaFrequency = abs(Frequency1 - Frequency2);
 
// cout << " Wavelength = " << Wavelength1 << "  " << Wavelength2 << endl;

 if( DeltaFrequency > NumGainWidth * FreqStepGain || DeltaFrequency ==0 )
    return(0);
 
 Interpolate = oRamanGain[FIndex*300+int(DeltaFrequency/FreqStepGain)] + 
 	(oRamanGain[FIndex*300+int(DeltaFrequency/FreqStepGain) + 1] - oRamanGain[FIndex*300+int(DeltaFrequency/FreqStepGain)])
	* ( ( DeltaFrequency - FreqStepGain * int(DeltaFrequency/FreqStepGain) ) / FreqStepGain ) ;
 
 if (Wavelength1>Wavelength2){

     return( Interpolate );
 } else {
 
     return( -Interpolate * (Frequency1 / Frequency2) );
 }

}  // end GetRamanGainFreqGrid


// #########################################

double RamanAmp::GetRamanGainClose(double Wave1, double Wave2, int Index)
{
	if(Index == 1){
 		if (Wave1>Wave2){
     		return( 0.0017 * abs(Wave1-Wave2) * 1e9 * 0.1 / 15 );
 		} else {
     		return( -0.0017 * abs(Wave1-Wave2) * 1e9 * 0.1 / 15 );
 		}
	}else{
		if (Wave1>Wave2){
     		return( 0.00045 * abs(Wave1-Wave2) * 1e9 * 0.1 / 15 );
 		} else {
     		return( -0.0045 * abs(Wave1-Wave2) * 1e9 * 0.1 / 15 );
 		}
	}




} // end of GetRamanGainClose

//#############################################

double RamanAmp::GetAttenuationFiber(double Wavelength, int NumSteps )

{ // return( 0 );

double Interpolate;
	if(Wavelength>1700E-9||Wavelength<1370E-9){
		cout << " Wavelength="  << Wavelength <<  endl;
		cout << " NumSigCh=" << NumSigCh << "  NumPumCh=" << NumPumCh << endl;
		for(int i=0;i<NumSigCh;i++){
			cout << " wavelength=" << oSignalWaveLth[i] ;
		}
		for(int i=0;i<NumPumCh;i++){
			cout << " wavelength=" << oPumpWaveLth[i] << endl;
		}		
		cerr << "Wavelength for AttenuationFiber only limited between 1370 and 1700 nm!!"  
              << endl << "Aborting now! " << endl ;
         exit(1);
	} 
 int FIndex = -1;
 NumSteps--;
 do{
 	FIndex++;
 }while(NumSteps > FiberLocation[FIndex]);
  
// Frequency1 = Wavelength2Frequency(Wavelength1);
// Frequency2 = Wavelength2Frequency(Wavelength2);

 if(ReadPlainFileGainLossFlag== 1){


  double DeltaWavelength = Wavelength - WaveLengthStartLoss;
//	if( Wavelength < 1505e-9){ 
//		return(0.5/4343);
//	}
   if( DeltaWavelength < 0 || DeltaWavelength > NumLossWidth * WavelengthStepLoss)
    { 	// cout << " not in the range" << endl;
		cout << Wavelength << "  " << WaveLengthStartLoss << "  "  << DeltaWavelength << "  " << NumLossWidth << "  " << WavelengthStepLoss;
		exit(1);
		return(0);
	}
 
 Interpolate = oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss)] + 
 	(oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss) + 1] 
		- oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss)])
	* ( ( DeltaWavelength - WavelengthStepLoss * int(DeltaWavelength/WavelengthStepLoss) ) / WavelengthStepLoss ) ;
 return( Interpolate );
 
 } 
 
 if(ReadRamanGainLossWavelengthGridFlag == 1){
 	if( abs(Wavelength - 1455e-9) < 0.000001e-9)
		return(PumpAttenuation);
 //		return(PumpAttenuation[FIndex]);
 	if( Wavelength < 1505e-9){
 		cout << " error" << endl;
		exit(1);
	}   
 double DeltaWavelength = Wavelength - WaveLengthStartLoss;
 if( DeltaWavelength < 0 || DeltaWavelength > NumLossWidth * WavelengthStepLoss)
    { 	// cout << " not in the range" << endl;
		cout << Wavelength << "  " << WaveLengthStartLoss << "  "  << DeltaWavelength << "  " << NumLossWidth << "  " << WavelengthStepLoss;
		exit(1);
		return(0);
	}
 
 Interpolate = oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss)] + 
 	(oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss) + 1] 
		- oLoss[FIndex*300+int(DeltaWavelength/WavelengthStepLoss)])
	* ( ( DeltaWavelength - WavelengthStepLoss * int(DeltaWavelength/WavelengthStepLoss) ) / WavelengthStepLoss ) ;
 return( Interpolate );
}




}  // GetAttenuationFiber

// ################################################

double RamanAmp::GetRamanGain(double Wavelength1, double Wavelength2)
{

 double Interpolate;
 double Frequency1, Frequency2;
 

 Frequency1 = Wavelength2Frequency(Wavelength1);
 Frequency2 = Wavelength2Frequency(Wavelength2);
 
 double DeltaWavelength = abs(Wavelength1 - Wavelength2)-WaveLengthStartGain;
 
 
  
 if( DeltaWavelength < 0 || DeltaWavelength > NumGainWidth * WavelengthStepGain)
    return(0);
 
 Interpolate = oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain)] + 
 	(oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain) + 1] 
		- oRamanGain[0*300+int(DeltaWavelength/WavelengthStepGain)])
	* ( ( DeltaWavelength - WavelengthStepGain * int(DeltaWavelength/WavelengthStepGain) ) / WavelengthStepGain ) ;
 
 if (Wavelength1>Wavelength2){

     return( Interpolate );
 } else {
 
     return( -Interpolate * (Frequency1 / Frequency2) );
 }

}  // end GetRamanGain
// ################################################



// ##################################################
// this is just for wavelength and for one fiber should revise later.


double RamanAmp::GetRamanGainFreqGrid(double Wavelength1, double Wavelength2)
{  

 double Interpolate;
 double FrequencyStep = 0.2e+12;
 double Frequency1, Frequency2;
 
 Frequency1 = Wavelength2Frequency(Wavelength1);
 Frequency2 = Wavelength2Frequency(Wavelength2);
 
 double DeltaFrequency = abs(Frequency1 - Frequency2);
 

 if( DeltaFrequency > 206 * FrequencyStep)
    return(0);
 
 Interpolate = oRamanGain[0*300+int(DeltaFrequency/FrequencyStep)] + 
 	(oRamanGain[0*300+int(DeltaFrequency/FrequencyStep) + 1] - oRamanGain[0*300+int(DeltaFrequency/FrequencyStep)])
	* ( ( DeltaFrequency - FrequencyStep * int(DeltaFrequency/FrequencyStep) ) / FrequencyStep ) ;
 
 if (Wavelength1>Wavelength2){

     return( Interpolate );
 } else {
 
     return( -Interpolate * (Frequency1 / Frequency2) );
 } 

}  // end GetRamanGainGrid


/*
double RamanAmp::GetRamanGain(double Wavelength1, double Wavelength2, int ZIndex)
{  cout << "run to herespecial" << endl;
cout << "testZIndex=" << ZIndex ;
cout << " WavelengthStepGain=" << WavelengthStepGain;

 double DeltaWavelength = abs(Wavelength1 - Wavelength2);
 cout << "  DeltaWavelength=" << DeltaWavelength;
 double Interpolate;
 int FIndex;
 if (DeltaWavelength < 1e-12)
    return(0);
 if( DeltaWavelength > 152 * WavelengthStepGain)
    return(0);

  FIndex=-1;
 do{	FIndex++;
 }while(ZIndex > FiberLocation[FIndex]);

 cout << "testZIndex=" << ZIndex << "FIndex=" << FIndex << endl;

 Interpolate = oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain)] + 
 	(oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain) + 1] - oRamanGain[FIndex*300+int(DeltaWavelength/WavelengthStepGain)])
	* (DeltaWavelength/WavelengthStepGain - int(DeltaWavelength/WavelengthStepGain) ) ;

 if (Wavelength1>Wavelength2){
     return( Interpolate );
 }else {
     return( -Interpolate * (Wavelength1 / Wavelength2) );
 }

}  // end GetRamanGain
 
*/
//#############################################
double RamanAmp::GetAttenuationFiber(double Wavelength)

// this is not work for DCF 
{  
return(GetAttenuationFiber( Wavelength,1));
	if(Wavelength>1700E-9||Wavelength<1370E-9){
		cout << Wavelength << endl;
		cerr << "Wavelength for AttenuationFiber only limited between 1370 and 1700 nm!!"  
              << endl << "Aborting now! " << endl ;
         exit(1);
	}

 double Interpolate;
 double WaveLtepAtt = 10e-9;
 
 Interpolate = oLoss[0*300+int((Wavelength-1370E-9)/WaveLtepAtt)] + 
 	(oLoss[0*300+int((Wavelength-1370E-9)/WaveLtepAtt) + 1] 
		- oLoss[0*300+int((Wavelength-1370E-9)/WaveLtepAtt)])
	* ( ( Wavelength-1370E-9 - WaveLtepAtt * int((Wavelength-1370E-9)/WaveLtepAtt) ) / WaveLtepAtt ) ;
 return( Interpolate );

}


// #########################################################

void RamanAmp::WriteSignalWaveLthPower(string OutFileName, int Num)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<  FLength / 1000  << "  "
                 << oSignalPower[Num * NumZSteps + i]
                 << endl;
    }
 
 
 
 OutFile.close();
}  // end  writeSignalWaveLthPower file

// #########################################################

void RamanAmp::WriteSignalTrack(string OutFileName)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<  FLength / 1000  ;
	 for(int ii=0;ii<NumSigCh;ii++){
        if (ii==10*int(ii/10))
			OutFile << "   " << oSignalPower[ii * NumZSteps + i] ;
     }                 
	 	OutFile << endl;
 }
  
 OutFile.close();
}  // end  writeSignalWaveLthPower file

// ###################################################

void RamanAmp::WriteSignalTrack(string OutFileName, int Num)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
// OutFile << "jjshift=" << Num << endl;
 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<  FLength / 1000  ;
			OutFile << "   " << oSignalPower[Num * NumZSteps + i] ;
			OutFile << endl;
     }                 
     
	 	OutFile << endl;
 
  
 OutFile.close();
}  // end  writeSignalWaveLthPower file


//   ################################################

void RamanAmp::WriteSignalWaveLthPowerWith$(string OutFileName, int Num)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
  double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<   FLength / 1000 << "   "
                 << oSignalPower[Num * NumZSteps + i]
                 << endl;
    }
 
 
 
 OutFile.close();
}  // end  writeSignalWaveLthPowerWith$ file

// ##############################################################

void RamanAmp::WriteSignalOutputPower(string OutFileName)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
 
   exit(1);
 }

 for(int NSig=0;NSig<NumSigCh;NSig++){
       OutFile <<  oSignalWaveLth[NSig]  << "   "
                 << oSignalPower[NSig * NumZSteps + NumZSteps-1] 
                 << endl;
  	//	OutFile << 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) << "  ";
		//OutFile << oSignalPower[NSig*NumZSteps+NumZSteps-1] << "  " << oSignalPower[NSig*NumZSteps] << endl;
	     

    }
 
 
 OutFile.close();
}  // end  writeSignalWaveLthPower file

// ###################################

void RamanAmp::WriteSignalNoise(OptSignal *Signal1, OptSignal *NFSignal, OptSignal *N, string OutSignalNoiseFileName)
{    
 Signal1->fft->fftshift(Signal1->sfftM.fOptSignalX);
 NFSignal->fft->fftshift(NFSignal->sfftM.fOptSignalX);
 N->fft->fftshift(N->sfftM.fOptSignalX);

 ofstream OutSignalNoiseFile(OutSignalNoiseFileName.c_str(),ios::out);
 if (!OutSignalNoiseFile)
   {
   cerr<< "Error:  "  << OutSignalNoiseFileName  << "Could not be created." << endl;
 
   exit(1);
 }

 for(int jj=0;jj<NFSignal->fft->give_N();jj++){
       OutSignalNoiseFile <<  Frequency2Wavelength( NFSignal->GetCenterFreq()
	  						+(jj- NFSignal->fft->give_N()/2)*NFSignal->GetDeltaFreq() ) << "   "
   				<< sq(abs(Signal1->sfftM.fOptSignalX[jj]))  << "   "  
   				<< sq(abs(NFSignal->sfftM.fOptSignalX[jj]))  << "   "  
				<< sq(abs(N->sfftM.fOptSignalX[jj]))  << "   "  
	            << endl;
    }

 Signal1->fft->ifftshift(Signal1->sfftM.fOptSignalX);
 NFSignal->fft->ifftshift(NFSignal->sfftM.fOptSignalX);
 N->fft->ifftshift(N->sfftM.fOptSignalX);
 OutSignalNoiseFile.close();
}

// ##############################################################

void RamanAmp::WriteSignalOutputPowerdB(string OutFileName)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
 
   exit(1);
 }

 for(int NSig=0;NSig<NumSigCh;NSig++){
       OutFile <<  oSignalWaveLth[NSig]  << "   "
//                << oSignalPower[NSig * NumZSteps + NumZSteps-1]
                 << 10*log10(oSignalPower[NSig * NumZSteps + NumZSteps-1] / oSignalPower[NSig * NumZSteps])
                 << endl;
       

    }
 
 
 OutFile.close();
}  // end  writeSignalWaveLthPower file

// ##############################################3

void RamanAmp::WritePumpWaveLthPower(string OutFileName, int Num)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(NumZSteps-i);
	 OutFile <<  FLength / 1000  << "   " 
                 << oPumpPower[Num * NumZSteps + NumZSteps
				 -1- i]
                 << endl;
    }
 
  
 OutFile.close();
}

// ##############################################

void RamanAmp::WritePumpPower(string OutFileName)

{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }

 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<  FLength / 1000  ;

// for(int i=0;i<NumZSteps;i++){
//   OutFile <<  i * DeltaZ(NumZSteps-i) ;
 
   for(int NPum=0;NPum<NumPumCh;NPum++){

           OutFile <<  "   " << oPumpPower[NPum * NumZSteps + i];
   }
   OutFile << endl;
 }
 
}
// ##########################################
 
void RamanAmp::WritePumpPowerdB(string OutFileName)

{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }

 double FLength=0;
 for(int i=0;i<NumZSteps;i++){
     FLength += DeltaZ(i);
	 OutFile <<  FLength / 1000  ;
   
     for(int NPum=0;NPum<NumPumCh;NPum++){

           OutFile <<  "   " << 10 * log( oPumpPower[NPum * NumZSteps +i] * 1000 ) / log(10);
     }
 
 
 
 
  // for(int NPum=0;NPum<NumPumCh;NPum++){

      //     OutFile <<  "   " << 10 * log( oPumpPower[NPum * NumZSteps +i] * 1000 ) / log(10) ;
   //}
//   for(int NSig=0;NSig<NumSigCh;NSig++){

     //      OutFile <<  "   " << 10 * log( oSignalPower[NSig * NumZSteps +i] * 1000 ) / log(10) ;
  // }
  
  
  
   OutFile << endl;
 }
  
 OutFile.close();
}



//  ###########################################


void RamanAmp::WriteSignalConvergence(string OutFileName)
{
 ofstream OutFile(OutFileName.c_str(),ios::out);
 if (!OutFile)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }
  for(int NumItr=0;NumItr<NumIteration;NumItr++){
   
     OutFile <<  NumItr << "   "
 //                <<    DeltaPpumpTotal[NumItr] + DeltaPsignalTotal[NumItr]
                 << endl;
    }
 
  
 OutFile.close();
}

//  ##################################

void RamanAmp::WriteAveragePowerdBm(int Numloop)
{ 

   for (int ii = 0; ii < NumZSteps; ii++) 
	   *LogFile << Numloop * AmpLength + DeltaZ(ii) * ii << "  need to be correct"
           << Linear2dBm(SigFEnergy[ii])
           << endl; 
}

// ################################
/* do not use anymore
double RamanAmp::GetRamanPowerFreq(double Z, double Freq)

// need to be debug for the B.C.
{
  double WaveLth;
  int ZInterpolate,WaveLthInterpolate;
  double ReturnPower1, ReturnPower2;
    
  WaveLth = Frequency2Wavelength(Freq);
 // cout << "WaveLth=" << WaveLth ;
  cout.precision(16);
//  if (oSignalWaveLth[0]>WaveLth){
//  	cout << " Large";
//  	if(oSignalWaveLth[0]<WaveLth){
 // 		cout << " small";
 // 	}
// }

//  if(!((oSignalWaveLth[0]>WaveLth&&WaveLth>oSignalWaveLth[NumSigCh-1])
//		||oSignalWaveLth[0]==WaveLth||WaveLth==oSignalWaveLth[NumSigCh-1])){
//  if(!(WaveLth<oSignalWaveLth[0])){
// 	 cout << "weird1";
//  	 cout << "WaveLth="  << WaveLth << "  osignal0=" << oSignalWaveLth[0];
//     cerr<< "Error:  Freq is not in the signal range." << endl;
//	 cerr << "Freq=" << Freq << "  L.B.=" << Wavelength2Frequency(oSignalWaveLth[0]) << "  R.B.=" << Wavelength2Frequency(oSignalWaveLth[NumSigCh-1]);
//   	 exit(1);
//  }
  
  if((WaveLth!=oSignalWaveLth[0])&&(WaveLth>oSignalWaveLth[0])){
  	 cout << "WaveLth="  << WaveLth << "  osignal0=" << oSignalWaveLth[0];
     cerr << "Error:  Freq is not in the signal range." << endl;
	 cerr << "Freq=" << Freq << "  L.B.=" << Wavelength2Frequency(oSignalWaveLth[0]) << "  R.B.=" << Wavelength2Frequency(oSignalWaveLth[NumSigCh-1]);
   	 exit(1);
  }
  
  if((!(WaveLth==oSignalWaveLth[NumSigCh-1]))&&(WaveLth<oSignalWaveLth[NumSigCh-1])){

  	 cout << "WaveLth="  << WaveLth << "  osignalNumSigch-1=" << oSignalWaveLth[NumSigCh-1] << endl;
     cerr << "Error:  Freq is not in the signal range." << endl;
	 cerr << "Freq=" << Freq << "  L.B.=" << Wavelength2Frequency(oSignalWaveLth[0]) << "  R.B.=" << Wavelength2Frequency(oSignalWaveLth[NumSigCh-1]);
   	 exit(1);
  }

  if(Z<0||Z>AmpLength){
     cerr<< "Error:  Z is not in the distance range." << endl;
   	 exit(1);
  }  
  
  ZInterpolate = (int)(Z/DeltaZ);
  WaveLthInterpolate = (int)(qtPoints/2 + ( Freq - oOptSignal->GetCenterFreq() ) / oOptSignal->GetDeltaFreq() );
// cout<< "Freq-***=" << ( Freq - oOptSignal->GetCenterFreq() ) / oOptSignal->GetDeltaFreq() ;
//  cout << "  qtPoints=" << qtPoints;
//  cout << " WaveLInt=" << (qtPoints/2 + ( Freq - oOptSignal->GetCenterFreq() ) / oOptSignal->GetDeltaFreq() );
//		 cout << "run to here1  ZIntp=" << ZInterpolate << "   WaveIntpe="  << WaveLthInterpolate;
  if( (( Freq - oOptSignal->GetCenterFreq() ) / oOptSignal->GetDeltaFreq()) 
					== (int)((Freq - oOptSignal->GetCenterFreq() ) / oOptSignal->GetDeltaFreq())){
		if(Z==(DeltaZ*(int)(Z/DeltaZ))){
			return(oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate]);
		}else{
			return(oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate] + 
  			(oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate+1]
			-oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate]) * (Z-DeltaZ*(int)(Z/DeltaZ))/DeltaZ );
		}
  }else{
	  	if(Z==(DeltaZ*(int)(Z/DeltaZ))){
  			ReturnPower1 = oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate]  ;
			ReturnPower2 = oSignalPower[(WaveLthInterpolate+1)*NumZSteps+ZInterpolate] ;
//		 	cout << "run to here2" << ReturnPower1;
		}else{ 
	  		ReturnPower1 = oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate] + 
  				(oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate+1]
				-oSignalPower[WaveLthInterpolate*NumZSteps+ZInterpolate]) * (Z-DeltaZ*(int)(Z/DeltaZ))/DeltaZ ;
			ReturnPower2 = oSignalPower[(WaveLthInterpolate+1)*NumZSteps+ZInterpolate] + 
  				(oSignalPower[(WaveLthInterpolate+1)*NumZSteps+ZInterpolate+1]
		 		-oSignalPower[(WaveLthInterpolate+1)*NumZSteps+ZInterpolate]) * (Z-DeltaZ*(int)(Z/DeltaZ))/DeltaZ ; 
		}
  		return(ReturnPower1 + (ReturnPower2 - ReturnPower1)	* (WaveLth-oSignalWaveLth[WaveLthInterpolate])
  		/( oSignalWaveLth[WaveLthInterpolate+1] - oSignalWaveLth[WaveLthInterpolate]));
  }

//  return(ReturnPower1 + (ReturnPower2 - ReturnPower1));
  // * (WaveLth-oSignalWaveLth[WaveLthInterpolate])
  //			/( oSignalWaveLth[WaveLthInterpolate+1] - oSignalWaveLth[WaveLthInterpolate]));
//  cout << ReturnPower1 + (ReturnPower2 - ReturnPower1);
//  return(0);
  
//  oSignalWaveLth[ii] = Frequency2Wavelength( oOptSignal->GetCenterFreq()
//		+(ii-NumSigCh/2)*oOptSignal->GetDeltaFreq() );
 
} // 
*/ 

// ########################################################


// ########################################################################

double RamanAmp::GetRamanPowerFreqIndex(double Z, int SignalFreqIndex)
{

  int ZInterpolate;
  if(SignalFreqIndex<0||SignalFreqIndex >= NumSigCh){
      cerr << "Error:  SignalFreqIndex is not in the signal range." << endl;
      cerr << " SignalFreqIndex=" << SignalFreqIndex 
	   << " NumSigCh = " << NumSigCh 
	   << "  L.B.=" 
           << Wavelength2Frequency(oSignalWaveLth[0]) 
           << "  R.B.=" << Wavelength2Frequency(oSignalWaveLth[NumSigCh-1]);
   	 exit(1);
  }
 
 
  if( Z<-1e-5 || Z > AmpLength + 1e-5){
     cerr<< "Error:  this ?Z is not in the distance range." << endl;
   	 cout << "Z = " << Z << endl << "AmpLength = " << AmpLength << endl;
	 exit(1);
  }  
  int FIndex = -1;
  double FLength=0,FLengthPre,FLengthGrid;
  
  do{
	FLengthPre = FLength;
	FIndex ++;
//	cout << "Z = "  << Z << "  FLength = " << FLength << " FIndex=" << FIndex <<endl;
	if(FIndex>=1){
		FLength += (FiberLocation[FIndex] - FiberLocation[FIndex-1]) * DeltaZArray[FIndex];
	}else{
    	FLength += (FiberLocation[FIndex] + 1) * DeltaZArray[FIndex];
	}
//	cout << "Z = "  << Z << "  FLength = " << FLength << " FIndex=" << FIndex << endl;

  }while( Z > FLength+0.000001 );
//  cout << "Z = "  << Z << "  FLength = " << FLength << " FIndex=" << FIndex << endl;
  // need to correct it.
  ZInterpolate = (int)((Z - FLengthPre) / DeltaZArray[FIndex]);
  FLengthGrid = FLengthPre + ZInterpolate * DeltaZArray[FIndex];
//    if(Z==0 && SignalFreqIndex == 10 * (int)(SignalFreqIndex/10))
 //   	cout << "FLengthPre=" << FLengthPre << " Delta=" << DeltaZArray[FIndex] << " FIndex=" 
//			<< FIndex << " ZInterpolate=" << ZInterpolate << endl ;
  if(FIndex != 0){
	  ZInterpolate += FiberLocation[FIndex-1] + 1;
  }
//  cout << " ZInterpolate=" << ZInterpolate  << "  FLengthGrid=" << FLengthGrid << endl;
  
//  if(SignalFreqIndex == 10 * (int)(SignalFreqIndex/10))
//  cout << " Z=" << Z << "  Index=" << SignalFreqIndex << "  ZInterp=" << ZInterpolate << "  return=" <<
 // 			oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate] + 
  //			(oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate+1]
//			-oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate]) * (Z-FLengthPre)/DeltaZArray[FIndex] << endl;
  
  return(oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate] + 
  			(oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate+1]
			-oSignalPower[SignalFreqIndex*NumZSteps+ZInterpolate]) * (Z-FLengthGrid)/DeltaZArray[FIndex] );
  


} 

//  ###############################################

void RamanAmp::DoubleRamanPumpPower()
{
	for(int NPum=0;NPum<NumPumCh;NPum++)
		oPumpPowerInput[NPum] *= 2;

} // end of DoubleRamanPumpPower
 
//  #####################################################

void RamanAmp::WriteEqIntgData(string InFileName)
{
 char str[40];
 double Space,PumpP;
 string name;
 ifstream InFile(InFileName.c_str(),ios::in);   

 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 
 LogFileSeparator();
 *LogFile << InFileName << endl << endl;   
 
// NumGainWidth = ReadInt("NumGainWidth"," ",0,-1,NO_BOUNDS,&InFile);
// WavelengthStepGain = ReadDouble("WavelengthStepGain"," ",0,-1,NO_BOUNDS,&InFile);


// InFile->clear(); // Resets ifstream state to good
// InFile->seekg(0); 
 for(int i=1;i<100;i++)
    {   

       InFile >> Space;
	   *LogFile << Space << "   " ; 
	   for(int j=0;j<10;j++)
	   {
	   InFile >> PumpP; //
	   } 

	   for(int j=0;j<2;j++)
	   {
 		   InFile >> name; //cout << name <<"   "; 
		   InFile >> name; //cout << name <<"   ";
	   	   InFile >> PumpP; *LogFile << PumpP <<"   ";
	   }   	   
  
	   *LogFile << endl;   	   
    }

	//cout << endl;
 InFile.close();
   
} // ## end WriteEqIntgData

//  #####################################################

void RamanAmp::WriteGainShapeData(string InFileName,string OutFileName1,string OutFileName2)
{
 char str[40];
 double Space,PumpP;
 string name; 
 double Wavelength, PumpWavelength1, PumpIntegratedPower1, PumpWavelength2, PumpIntegratedPower2;

 ifstream InFile(InFileName.c_str(),ios::in);   

 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 ofstream OutFile1(OutFileName1.c_str(),ios::out);
 if (!OutFile1)
   {
   cerr<< "Error:  "  << OutFileName1  << "Could not be created." << endl;
   exit(1);
 }
 
 ofstream OutFile2(OutFileName2.c_str(),ios::out);
 if (!OutFile2)
   {
   cerr<< "Error:  "  << OutFileName2  << "Could not be created." << endl;
   exit(1);
 }
 
 LogFileSeparator();
 *LogFile << OutFileName1 << endl << endl;   
 LogFileSeparator();
 *LogFile << OutFileName2 << endl << endl;   

 PumpWavelength1 = ReadDouble("PumpWavelength1"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 PumpIntegratedPower1 = ReadDouble("PumpIntegratedPower1"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 PumpWavelength2 = ReadDouble("PumpWavelength2"," ",0,-1,
                                     LOWER_ONLY,&InFile);
 PumpIntegratedPower2 = ReadDouble("PumpIntegratedPower2"," ",0,-1,
                                     LOWER_ONLY,&InFile); 

 for(int i=1;i<200;i++){  
		Wavelength = i * 1.5e-9 + PumpWavelength1;
		OutFile1 << Wavelength << "  only valid for one fiber  " << GetRamanGain(Wavelength,PumpWavelength1)*PumpIntegratedPower1 << endl;
		Wavelength = i * 1.5e-9 + PumpWavelength2;
		OutFile2 << Wavelength << "    " << GetRamanGain(Wavelength,PumpWavelength2)*PumpIntegratedPower2 
				<< "   " << GetRamanGain(Wavelength,PumpWavelength1)*PumpIntegratedPower1 
				+ GetRamanGain(Wavelength,PumpWavelength2)*PumpIntegratedPower2 << endl;	
	 	 
 }

	//cout << endl;
 OutFile2.close();
 OutFile1.close();
 InFile.close();
 
   
} // ## end WriteEqIntgData



//  #####################################################

void RamanAmp::WriteCentralWavelength(string InFileName,string OutFileName1)
{
 char str[40];
 double Space,PumpP;
 string name; 
 double Wavelength, PumpWavelength1, PumpIntegratedPower1, PumpWavelength2, PumpIntegratedPower2;
// double CentralWavelength;
 ifstream InFile(InFileName.c_str(),ios::in);   

 if (!InFile) 
  {
    cerr << "Error: Input file "<< InFileName 
           << " could not be opened."<< endl;
    exit(1);
    }

 ofstream OutFile1(OutFileName1.c_str(),ios::out);
 if (!OutFile1)
   {
   cerr<< "Error:  "  << OutFileName1  << "Could not be created." << endl;
   exit(1);
 }
 

 
 LogFileSeparator();
 *LogFile << OutFileName1 << endl << endl;   

 for(int Step=1;Step<100;Step++){   

       InFile >> Space;

	   for(int j=0;j<20;j++) {
	   		InFile >> PumpP; //cout << PumpP <<"   ";
	   } 
//	   cout << endl;
//	   CentralWavelength=0;
	   OutFile1 <<  Space;
	   for(int j=0;j<NumPumCh;j++){
 		   InFile >> name; InFile >> Wavelength; //cout << name <<"   "; 
		   if((Wavelength < 1370e-9)||(Wavelength>1525e-9)){
		   		cout << " error in Wavelength" << endl;
				cout << " Aborting" << endl;
				exit(1);
			}	
			
		   OutFile1 << "   " << Wavelength;
	   	   InFile >> name; InFile >> PumpP; 
		   OutFile1 << "   " << PumpP ;
		   InFile >> name; InFile >> PumpP; 
//		   CentralWavelength += Wavelength;
	   }   	   
  	   OutFile1 << endl;
//	   OutFile1 <<  Space << "   " << CentralWavelength/NumPumCh  << endl;   	   
    }

 OutFile1.close();
 InFile.close();
 
   
} // ## end WriteCentralWavelength

// ######################################

string RamanAmp::WriteTypeRamanNoise(void)
{

 switch(TypeRamanNoise)
   {

    case RAMAN_NOISE_OFF:

           return "RAMAN_NOISE_OFF";
           break;

    case RAMAN_NOISE_ON_RANDOM_PHASE:

           return "NOISE_ON_CONST_POWER_RANDOM_PHASE";
           break;

    case RAMAN_NOISE_ON_GAUSSIAN_WHITE:

           return "NOISE_ON_GAUSSIAN_WHITE";
           break;
    case RAMAN_NOISE_ON_NO_RNG:
		   return "RAMAN_NOISE_ON_NO_RNG";
		   break;
		     
 	default:
		cerr << "With TypeRamanNoise must between 0 and 3" << endl 
		<< "Aborting now! " << endl;
		exit(1);
	
	}
	

}

// ###################################

// ###################################

/*

void RamanAmp::RelaxRamanIteration()
{  
	
 double rmsP,rmsS;
 int NumItr;
 char  C_String[40];
 string FileName;
 double Alpha;
 int TypePerturbation=0;
 double PerturbationRatio=1;
 
   for(int i=1;i<NumZSteps;i++){
       	SignalOnlyPropagation(i);
   }
   switch(TypePerturbation){
     	case 0:     // no perturbation algorithm, Need to be correct
			NumPumCh = NumPumCh;
			NumCh = NumSigCh + NumPumCh;
			Alpha=1;
  			for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps] = oPumpTargetBackup[NPum];
	            cout << "  " << oPumpPower[NPum*NumZSteps];
			} 
	
			for(int i=1;i<NumZSteps;i++){
       			SignalOnlyPropagation(i);
    		}

		break;
		
		case 1:   // Inc Pum Inc Pow Need to be correct
			NumPumCh=1;
		    Alpha = PerturbationRatio;
		 	NumCh = NumSigCh + NumPumCh;
		  	for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps] = PerturbationRatio * oPumpTargetBackup[NPum];
				} 
		  	for(int i=1;i<NumZSteps;i++){
       			PrePumpShooting(i);
    			}
  			for(int NPum=0;NPum<NumPumCh;NPum++) 	{			  
    			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps+NumZSteps-1] ;
//			    cout << oPumpPower[NPum*NumZSteps] << "  ";
			}
			NumPumCh=7;
		    Alpha = 1;			

			oPumpPower[0 * NumZSteps] = 0.00157699;
			oPumpPower[1 * NumZSteps] = 0.00186702 ;
			oPumpPower[2 * NumZSteps] = 0.00212021;
			oPumpPower[3 * NumZSteps] = 0.00262511;
			oPumpPower[4 * NumZSteps] = 0.00369277;
			oPumpPower[5 * NumZSteps] = 0.00564321;
			oPumpPower[6 * NumZSteps] = 0.0166584;
			//			 0.00157699  0.00186702  0.00212021  0.00262511  0.00369277  0.00564321  0.0166584
			//0.00157698  0.00186701  0.0021202  0.00262511  0.00369277  0.00564319  0.0166583
			//            this data is get from the result 
		break;
	   	
		case 2:   // Inc Pow
//		    cout << "run to here2";
			Alpha = PerturbationRatio;
	
  			for(int NPum=0;NPum<NumPumCh;NPum++) {
		    	oPumpPower[NPum*NumZSteps+NumZSteps-1] = PerturbationRatio * oPumpTargetBackup[NPum];
			}    
		  	for(int i=1;i<NumZSteps;i++){
       			SignalOnlyPropagation(i);
    		}
			
  	//		for(int NPum=0;NPum<NumPumCh;NPum++) 	{			  
    //			oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps+NumZSteps-1] ;
	//		    cout << oPumpPower[NPum*NumZSteps] << "  ";
	//			cout << "run to here1";
	//		}
			
		break;
	  
       	default:
         		cerr << "With PerturbationType must between 0 and 2" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      } 
	  
	      cout << "NumPumCh=" << NumPumCh;
// for(int NumItr=0;NumItr<NumIteration;NumItr++){
   
//  for(; NumPumCh<=NumPumCh;NumPumCh++){
	NumCh = NumSigCh + NumPumCh;
    if(Alpha>1) 
		{ Alpha=PerturbationRatio;}
	
	for( ; Alpha <= 1; Alpha += PerturbationRatio){
	  cout << "Alpha=" << Alpha << "  " ;
//	  *LogFile <<  Alpha << "  " ;
	  
	switch(TypePerturbation)
		{
     	case 0: 
			break;

		case 1: // need to be corrected
			cout << "run to here1";   
			oPumpTarget[NumPumCh-1] = Alpha * oPumpTargetBackup[NumPumCh-1];	  
			oPumpPower[(NumPumCh-1)*NumZSteps] = oPumpTarget[NumPumCh-1];
			break;
	   	
		case 2: 
//			cout << "run to here2";  
		  for(int NPum=0;NPum<NumPumCh;NPum++) {
	  		oPumpTarget[NPum] = Alpha * oPumpTargetBackup[NPum];
	  		oPumpPower[NPum*NumZSteps+NumZSteps-1] = oPumpTarget[NPum];
	  		
	  	  }
		break;
	  
      	default:
         		cerr << "With PerturbationType must between 0 and 2" << endl 
             		<< "Aborting now! " << endl;
         		exit(1);
      }
	  AccurateSign =0;

  NumItr = 0;
  do{
  
  NumItr = NumItr +1;
   
   
   
   
//   DeltaPpumpTotal[NumItr] = 0;
//   DeltaPsignalTotal[NumItr] =0;
   
   CompCoeff = 0 ;
   for(int i=0;i<NumPumCh;i++)  {                              // clear the DeltaPpump
     DeltaPpump[i] = 0;
   }  
   for(int i=0;i<NumSigCh;i++){                               // clear the DeltaPsignal
     DeltaPsignal[i] = 0;
   }
   for(int i=NumZSteps-2;i>=0;i--){
    	BackwardStep(i);
   }
   
   for(int i=1;i<NumZSteps;i++){
        ForwardStep(i);
   }

   
//   for(int i=0;i<NumZSteps;i++){
//       oSignalTrack[(NumItr+1)*NumZSteps+i] = oSignalPower[i]; 
//   }
	rmsS=0;
   for(int NSig=0;NSig<NumSigCh;NSig++){   
 //      cout << "Psignal#" << NumItr << "  " <<
//	   oSignalPower[NSig*NumZSteps+NumZSteps-1] << endl;
 //      cout << "  DeltaPsignal#" << NumItr << "  " << DeltaPsignal[NSig] << endl;
 //      cout << NSig << endl;
       rmsS += DeltaPsignal[NSig];
       sprintf(C_String,"RamanItr%iSig%i.dat",NumItr,NSig);
       FileName = JobRaman + C_String;  
//       WriteSignalWaveLthPower(FileName, NSig);
       sprintf(C_String,"SignalItr%iSample%i.dat",NumItr,NSig);
       FileName = JobRaman + C_String;  
//       WriteSignalWaveLthPowerWith$(FileName, NSig);

	   if (NSig==0) {
	   CompCoeff = sqrt(CompCoeff / (NumZSteps - 1 ));
//	   	   		*LogFile << NumItr << "   " << CompCoeff << endl ;       
       }
   
//   *LogFile << endl;

//	   if (NSig==0) {
//	   	   		*LogFile << NumItr << "   " ;
//	   }
//	   *LogFile << oSignalPower[NSig*NumZSteps + NumZSteps -1 ] << "   " ;       
   }
   
//   *LogFile << endl;
   
   rmsS = rmsS / NumSigCh;
   rmsP=0;
   for(int NPum=0;NPum<NumPumCh;NPum++){
   
       rmsP += DeltaPpump[NPum];
	      	      	cout << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1];
//       cout << "PPump#" << NumItr << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1];
 //      cout << "   DeltaPpump#" << NumItr << "  "  << DeltaPpump[NPum]  << endl;
       sprintf(C_String,"RamanItr%iPum%i.dat",NumItr,NPum);
       FileName = JobRaman + C_String;  
 //      WritePumpWaveLthPower(FileName, NPum);
   }
   rmsP /= NumPumCh ;
   if (rmsP < 1e-9){
	  	AccurateSign = 1;	
   }
   cout << "   Itr# " << NumItr << "  DeltaPsignalTotal = "  << rmsS;
   cout << "  DeltaPpumpTotal = "  << rmsP << endl;   
 }while(AccurateSign==0);

	}
 
  
 sprintf(C_String,"Convergence.dat");
 FileName = JobRaman + C_String;  
 WriteSignalConvergence(FileName);
 cout << "numPumCh=" << NumPumCh;
 for(int NPum=0;NPum<NumPumCh;NPum++){
   	     // 	cout << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1];
       for(int i=0;i<NumZSteps;i++){
 //   		cout << "  " << oPumpPower[NPum*NumZSteps+i];
   	   }
//	   cout << endl;
 } 

 cout << endl;
 
 
}   // end of RelaxRamanIteration

*/

// ####################################################
// ############### start Genetic Algorithms  ##########
// ####################################################
// for Genetic algorithms, please refer following papers
// V. Perlin and H. Winful, ``Optimal design of flat-gain wide-band fiber Raman amplifiers,'' J. Lightwave
// Technol.,  20, 250--254, (2002).

// V. Perlin and H. Winful, ``On distributed Raman amplification for ultrabroad-band long-haul WDM system,''
// J. Lightwave Technol.,  20, 409--416, (2002).

// Z. Hu, B. S. Marks, and C. R. Menyuk, ``Design of flat-gain fiber Raman amplifiers using equally spaced
// pumps with fixed and optimized central wavelength,'' in  Proceedings of Conference on Laser and Electro Optics
// Optical Society of America, Washington, D.C., 2003, paper CThD5.


void RamanAmp::ShootingGAVaryCW()
{
 double Space ; 
 double Ripple, Wavelength;
 double MinR;
 
 char str[40];
 double *OptiPumpIntg = new double[NumPumCh];
 double *OptiPumpInput = new double[NumPumCh];
  
 string OutFileName1;
 sprintf(str,"GainRipple.dat");
 OutFileName1 = JobRaman + str;

 ofstream GROut(OutFileName1.c_str(),ios::out);
 if (!GROut){
   cerr<< "Error:  "  << OutFileName1  << "Could not be created." << endl;
   exit(1);
 }
 
 string OutFileName2;
 sprintf(str,"CentralW.dat");
 OutFileName2 = JobRaman + str;

 ofstream CWOut(OutFileName2.c_str(),ios::out);
 if (!CWOut){
   cerr<< "Error:  "  << OutFileName2  << "Could not be created." << endl;
   exit(1);
 }
 for(int NumSpace = StartNumSpace; NumSpace <= EndNumSpace; NumSpace++){
    Space = ( (double)NumSpace / 100 ) * (1525e-9 - 1450e-9) * 2 / (NumPumCh-1);
	cout << endl << "NumSpace=" << NumSpace << "  i=0" << endl;

//	Space = 32e-9;
	ShootingGeneticAlgorithmsVaryCW(Space);
	MinR = GlobleRipple;

	for(int NPum=0;NPum<NumPumCh;NPum++){
		OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
		OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
	}
	*LogFile << Space << "    " << GlobleRipple;
  
	for(int i=1;i<20;i++){
		cout << endl << endl << "NumSpace=" << NumSpace << "  i=" << i << endl;
		ShootingGeneticAlgorithmsVaryCW(Space);
		*LogFile << "   " << GlobleRipple;
		if(GlobleRipple < MinR) {
			MinR = GlobleRipple;
			for(int NPum=0;NPum<NumPumCh;NPum++){
				OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
				OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
			}
		}
	}
	*LogFile << endl;
	GROut << Space << "    " << MinR << endl;
	Wavelength = 0;
	
 	for(int NPum=0;NPum<NumPumCh;NPum++){
	 	Wavelength += oPumpWaveLth[NPum];
 		*LogFile <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
		cout <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
 	}
	CWOut << Space << "   "  << Wavelength/NumPumCh << endl;
 }

 CWOut.close();
 GROut.close();
 delete [] OptiPumpIntg;
 delete [] OptiPumpInput;

}  // end ShootingGAVaryCentralWavelength


void RamanAmp::ShootingGAVictor()
{
 double Space ; 
 double Ripple;
 double MinR;
 
 char str[40];
 double *OptiPumpIntg = new double[NumPumCh];
 double *OptiPumpInput = new double[NumPumCh];
  
 string OutFileName;
 sprintf(str,"RamanOut.dat");
 OutFileName = JobRaman + str;

 ofstream RamanOut(OutFileName.c_str(),ios::out);
 if (!RamanOut)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }

	ShootingGeneticAlgorithmsVictor();
	MinR = GlobleRipple;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
		OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
	}	
	
	*LogFile << Space << "    " << GlobleRipple << endl;
	RamanOut << Space << "    ";
	for(int i=1;i<10;i++){
		cout << endl << endl << " loop i=" << i << endl;
		ShootingGeneticAlgorithmsVictor();
		*LogFile << "   " << GlobleRipple;
		if(GlobleRipple < MinR) {
			MinR = GlobleRipple;
			for(int NPum=0;NPum<NumPumCh;NPum++){
				OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
				OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
			}
		}
	}
	*LogFile << endl;
	RamanOut << MinR << endl;
	cout << "result=" << MinR << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
 		*LogFile <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
 		cout <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
 	}

 RamanOut.close();
 delete [] OptiPumpIntg;
 delete [] OptiPumpInput;

}  //  end of ShootingGAVictor

// #####################################

void RamanAmp::ShootingGA()
{
 double Space ; 
 double Ripple;
 double MinR;
 
 char str[40];
 double *OptiPumpIntg = new double[NumPumCh];
 double *OptiPumpInput = new double[NumPumCh];
  
 string OutFileName;
 sprintf(str,"RamanOut.dat");
 OutFileName = JobRaman + str;

 ofstream RamanOut(OutFileName.c_str(),ios::out);
 if (!RamanOut)
   {
   cerr<< "Error:  "  << OutFileName  << "Could not be created." << endl;
   exit(1);
 }

 for(int NumSpace = StartNumSpace; NumSpace <= EndNumSpace; NumSpace++){
    Space = ( (double)NumSpace / 100 ) * (1525e-9 - 1450e-9) * 2 / (NumPumCh-1);
	cout << endl << "NumSpace=" << NumSpace << "  i=0" << endl;
	ShootingGeneticAlgorithms(Space);
	MinR = GlobleRipple;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
		OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
	}
	*LogFile << Space << "    " << GlobleRipple;
	RamanOut << Space << "    ";
	for(int i=1;i<10;i++){
		cout << endl << endl << "NumSpace=" << NumSpace << "  i=" << i << endl;
		ShootingGeneticAlgorithms(Space);
		*LogFile << "   " << GlobleRipple;
		if(GlobleRipple < MinR) {
			MinR = GlobleRipple;
			for(int NPum=0;NPum<NumPumCh;NPum++){
				OptiPumpIntg[NPum] = oPumpIntgTarget[NPum];
				OptiPumpInput[NPum] =  oPumpPower[NPum*NumZSteps+NumZSteps-1];
			}
		}
	}
	*LogFile << endl;
	RamanOut << MinR << endl;
 	for(int NPum=0;NPum<NumPumCh;NPum++){
 		*LogFile <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
 		cout <<  "Wavelength=  " << oPumpWaveLth[NPum] << "  PumpIntg=  " << OptiPumpIntg[NPum] 
				<< "  PumpInputpower=  " << OptiPumpInput[NPum] << endl;
 	}

 }
 RamanOut.close();
 delete [] OptiPumpIntg;
 delete [] OptiPumpInput;

} // end ShootingGA

// #######################################

void RamanAmp::ShootingGeneticAlgorithms(double VectorEqSpace)
{
 double MaxG,MinG,NetRipple,PreNetRipple;
 int NumExtItr;
 double *VectorPu = new double[Population*NumPumCh];  
 
 // double VectorPu[Population*NumPumCh*2];
	
 for(int i=1;i<NumZSteps;i++){
 	SignalOnlyPropagation(i);
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	for(int ii=1;ii<NumZSteps;ii++){
		oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
				*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
				*ii/(NumZSteps-1));
	}
 }
//  for(int i=1;i<NumZSteps;i++){
// 	cout << oSignalPower[1*NumZSteps+i] << "  ";
// }


// Population generation
 for(int i=0;i<Population;i++){
	for(int j=0;j<NumPumCh;j++){
		VectorPu[ i*NumPumCh + j ] = 30000*RNG->GetRanNum()+10;
	}

//	VectorEqSpace[i] = RNG->GetRanNum() * (1525e-9 - 1450e-9) * 2 / (NumPumCh-1) ;
	UpdateIndividualWavelength(VectorEqSpace);
 }	

 NumExtItr=0;
 NetRipple = 100;

// do{

 do{
    
 	NumExtItr++;
    GeneticAlgorithms(VectorPu);
	
/*	oPumpWaveLth[0]=1.405e-06;  
  	oPumpIntgTarget[0]=5658.4;
	oPumpWaveLth[1]=1.435e-06;   
	oPumpIntgTarget[1]=442.738;
	oPumpWaveLth[2]=1.465e-06; 
	oPumpIntgTarget[2]=3275.39;
	oPumpWaveLth[3]=1.495e-06;  
	oPumpIntgTarget[3]=177.809;
*/
/*	
	oPumpWaveLth[0]=1.41996e-06;  
  	oPumpIntgTarget[0]=2155.07  ;
	oPumpWaveLth[1]=1.4344e-06  ;   
	oPumpIntgTarget[1]=1106.33  ;
	oPumpWaveLth[2]=1.4532e-06   ; 
	oPumpIntgTarget[2]=1574.72  ;
	oPumpWaveLth[3]=1.4989e-06   ;  
	oPumpIntgTarget[3]=3268.85  ;
*/
// should work on this as one 1-D Vector
  // 
/*	if(NumExtItr<=1){
 		for(int NPum=0;NPum<NumPumCh;NPum++) {
			oPumpPower[NPum*NumZSteps] = (oPumpIntgTarget[NPum] * GetAttenuationFiber(oPumpWaveLth[NPum])) 
			/ ( exp( GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength ) -1 );
			cout << endl <<  "Wavelength=" << oPumpWaveLth[NPum] << "  InitPower=" << oPumpPower[NPum*NumZSteps] 
				<< "   IntgralTarget=" 	<< oPumpIntgTarget[NPum] ;
			cout << "  GetAttenuation=" << GetAttenuationFiber(oPumpWaveLth[NPum]) ;
 		}
	}	
*/
//	exit(1); 
	ShootingIntegration(NumExtItr);
//exit(1);
//	CalculateFrel();

	PreNetRipple = NetRipple;
	MaxG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);
	MinG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);
	for(int NSig=1;NSig<NumSigCh;NSig++){
//		 cout << 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) << "  ";
//		cout << oSignalPower[NSig*NumZSteps+NumZSteps-1] << "  " << oSignalPower[NSig*NumZSteps] << endl;
		
		if(10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) > MaxG){
			MaxG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
		}else{
			if((10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps])) < MinG){
				MinG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
			}
		}
	}	
	
	if(abs(MaxG)>abs(MinG)){
		NetRipple = abs(MaxG);
	}else{
		NetRipple = abs(MinG);
	}	
	cout << "#ExtItr=" << NumExtItr << "  NetRipple=" << NetRipple  << endl;	
// 	*LogFile << "NetRipple=" << NetRipple << " PreNetRipple=" << PreNetRipple << "  MaxG = " << MaxG << "  MinG = " << MinG ;	

 }while((abs(NetRipple-PreNetRipple)/NetRipple) > 0.01);
	 
	 GlobleRipple = NetRipple;
	 cout << " GlobleRipple=" << GlobleRipple << endl;

	   
 delete [] VectorPu;

}  // end ShootingGeneticAlgorithms

// #######################################

void RamanAmp::ShootingGeneticAlgorithmsVaryCW(double VectorEqSpace)
{
 double MaxG,MinG,NetRipple,PreNetRipple;
 int NumExtItr;
 double *VectorPu = new double[Population*NumPumCh*2];  
 double *CentralWavelength = new double[Population];
// double *PumpSpace = new double[Population]; 
// double VectorPu[Population*NumPumCh*2];
	
 for(int i=1;i<NumZSteps;i++){
 	SignalOnlyPropagation(i);
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	for(int ii=1;ii<NumZSteps;ii++){
		oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
				*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
				*ii/(NumZSteps-1));
	}
 }
//  for(int i=1;i<NumZSteps;i++){
// 	cout << oSignalPower[1*NumZSteps+i] << "  ";
// }
// Population generation
 for(int Pu=0;Pu<Population;Pu++){
	for(int j=0;j<NumPumCh;j++){
		VectorPu[ 2*Pu*NumPumCh + 2*j + 1 ] = 30000*RNG->GetRanNum()+10;
	}
	CentralWavelength[Pu] = 1375e-9 + VectorEqSpace * (NumPumCh-1) / 2
							+ (150e-9 - VectorEqSpace * (NumPumCh-1)) * RNG->GetRanNum() ;
	UpdateIndividualWavelength(Pu, VectorPu, VectorEqSpace, CentralWavelength[Pu]);
 }	

 NumExtItr=0;
 NetRipple = 100;	

 do{

 	NumExtItr++;
    GeneticAlgorithmsVaryCW(VectorPu,VectorEqSpace,CentralWavelength);


	if(NumExtItr<=1){
 		for(int NPum=0;NPum<NumPumCh;NPum++) {
			oPumpPower[NPum*NumZSteps] = (oPumpIntgTarget[NPum] * GetAttenuationFiber(oPumpWaveLth[NPum])) 
					/ ( exp( GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength ) -1 );
//			cout << endl <<  "Wavelength=" << oPumpWaveLth[NPum] << "  InitPower=" << oPumpPower[NPum*NumZSteps] 
//				<< "   IntgralTarget=" 	<< oPumpIntgTarget[NPum] ;
//			cout << "  GetAttenuation=" << GetAttenuationFiber(oPumpWaveLth[NPum]) ;
 		}
	}	

	ShootingIntegration(NumExtItr);


	PreNetRipple = NetRipple;
	MaxG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);
	MinG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);

	for(int NSig=1;NSig<NumSigCh;NSig++){
//		 cout << 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) << "  ";
//		cout << oSignalPower[NSig*NumZSteps+NumZSteps-1] << "  " << oSignalPower[NSig*NumZSteps] << endl;
		
		if(10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) > MaxG){
			MaxG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
		}else{
			if((10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps])) < MinG){
				MinG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
			}
		}
	}	

	if(abs(MaxG)>abs(MinG)){
		NetRipple = abs(MaxG); 
	}else{
		NetRipple = abs(MinG);
	}	
	cout << "NetRipple=" << NetRipple << " PreNetRipple=" << PreNetRipple << 
			"  MaxG = " << MaxG << "  MinG = " <<  MinG  << endl;	
// 	*LogFile << "NetRipple=" << NetRipple << " PreNetRipple=" << PreNetRipple << 
//			"  MaxG = " << MaxG << "  MinG = " << MinG << endl;	

 }while((abs(NetRipple-PreNetRipple)/NetRipple) > 0.01);
 
 GlobleRipple = NetRipple;
 cout << " GlobleRipple=" << GlobleRipple << endl;

 delete [] VectorPu;
 delete [] CentralWavelength;

}  // end ShootingGeneticAlgorithmsVaryCW

// #######################################

void RamanAmp::ShootingGeneticAlgorithmsVictor()
{
 double MaxG,MinG,NetRipple,PreNetRipple;
 int NumExtItr;
 double *VectorPu = new double[Population*NumPumCh*2];  
// double VectorPu[Population*NumPumCh*2];



	
 for(int i=1;i<NumZSteps;i++){
 	SignalOnlyPropagation(i);
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	for(int ii=1;ii<NumZSteps;ii++){
		oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
				*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
				*ii/(NumZSteps-1));
	}
 }
//  for(int i=1;i<NumZSteps;i++){
// 	cout << oSignalPower[1*NumZSteps+i] << "  ";
// }


// Population generation
 for(int i=0;i<Population;i++){
	for(int j=0;j<NumPumCh;j++){
		VectorPu[ 2*i*NumPumCh + 2*j ] = (120*RNG->GetRanNum()+1400)*1E-9;
		VectorPu[ 2*i*NumPumCh + 2*j + 1 ] = 30000*RNG->GetRanNum()+10;
	}

 }	

 NumExtItr=0;
 NetRipple = 100;	

// do{

 do{

 	NumExtItr++;
    GeneticAlgorithmsVictor(VectorPu);


	if(NumExtItr<=1){
 		for(int NPum=0;NPum<NumPumCh;NPum++) {
			oPumpPower[NPum*NumZSteps] = (oPumpIntgTarget[NPum] * GetAttenuationFiber(oPumpWaveLth[NPum])) 
			/ ( exp( GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength ) -1 );
			cout << endl <<  "Wavelength=" << oPumpWaveLth[NPum] << "  InitPower=" << oPumpPower[NPum*NumZSteps] 
				<< "   IntgralTarget=" 	<< oPumpIntgTarget[NPum] ;
			cout << "  GetAttenuation=" << GetAttenuationFiber(oPumpWaveLth[NPum]) ;
 		}
	}	

	ShootingIntegration(NumExtItr);
//	CalculateFrel();


	PreNetRipple = NetRipple;
	MaxG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);
	MinG = 10*log10(oSignalPower[NumZSteps-1] / oSignalPower[0]);
	for(int NSig=1;NSig<NumSigCh;NSig++){
//		 cout << 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) << "  ";
//		cout << oSignalPower[NSig*NumZSteps+NumZSteps-1] << "  " << oSignalPower[NSig*NumZSteps] << endl;
		
		if(10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]) > MaxG){
			MaxG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
		}else{
			if((10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps])) < MinG){
				MinG = 10*log10(oSignalPower[NSig*NumZSteps+NumZSteps-1] / oSignalPower[NSig*NumZSteps]);
			}
		}
	}	
	
	if(abs(MaxG)>abs(MinG)){
		NetRipple = abs(MaxG); 
	}else{
		NetRipple = abs(MinG);
	}	
	cout << "NetRipple=" << NetRipple << " PreNetRipple=" << PreNetRipple << "  MaxG = " << MaxG << "  MinG = " << MinG ;	
 	*LogFile << "NetRipple=" << NetRipple << " PreNetRipple=" << PreNetRipple << "  MaxG = " << MaxG << "  MinG = " << MinG ;	

 }while((abs(NetRipple-PreNetRipple)/NetRipple) > 0.01);
 
  GlobleRipple = NetRipple;
 delete [] VectorPu;

}  // end ShootingGeneticAlgorithmsVictor

// ###########################################

void RamanAmp::UpdateIndividualWavelength(int Num, double * VectorUpdate, double SpaceUpdate, double CW)
{
// cout << "  Space= " << VectorSpaceUpdate[Num] << "  " ;
	
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	VectorUpdate[Num * 2 * NumPumCh + 2 * NPum] = CW + SpaceUpdate / 2 - SpaceUpdate * NumPumCh / 2
										+ NPum * SpaceUpdate;
 	if(VectorUpdate[Num * 2 * NumPumCh + 2 * NPum] > 1700e-9)
	{	cout << " Num=" << Num << " SpaceUpdate=" <<  SpaceUpdate << " CW=" << CW;
		cout << "  exit";
		exit(1);		
	}
 } 
 
}    // end of UpdateIndividualWavelength

// #######################################

void RamanAmp::UpdateIndividualWavelength(double VectorSpaceUpdate)
{
// cout << "  Space= " << VectorSpaceUpdate << "  " ;
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	oPumpWaveLth[NPum] = 1450e-9 + VectorSpaceUpdate / 2 - VectorSpaceUpdate * NumPumCh / 2
										+ NPum * VectorSpaceUpdate;
	cout << oPumpWaveLth[NPum] << "  " ;
 }
// cout << endl;
		

}    // end of UpdateIndividualWavelength
// ###############################

void RamanAmp::ShootingIntegrationExample(int NExtItr)
{
 double *oPumpGuessPre = new double[NumPumCh];
// double oPumpGuessPre[NumPumCh];
 double rms;
// double rPumpPower[NumPumCh];
 int    NumItr,NumItrInternal;
 double Alpha,AlphaPre,DeltaAlpha;
 double PumpPowerAdjust;
 double *IntgPumPower = new double[NumPumCh];  
 double *oPumpIntgTargetAlpha = new double[NumPumCh];  
// double IntgPumPower[NumPumCh],oPumpIntgTargetAlpha[NumPumCh];
 double g;
 double *ShootingPower = new double[NumPumCh];  
 double *TargetPower = new double[NumPumCh];  
 double CenterPumpWavelength = 1450e-9;
 double TotalPumpPower;
 double *oPumpTrack = new double[NumZSteps*NumPumCh];
// double ShootingPower[NumPumCh],TargetPower[NumPumCh]; 
// double ErrorPum[NumZSteps];
 for(int i=0;i<NumZSteps;i++){
  	for(int NPum=0;NPum<NumPumCh;NPum++){	
		oPumpTrack[NPum*NumZSteps+i] = 1;
	}
 }
 char C_String[40];
 string FileName;
 sprintf(C_String,"ShootingExampleInput.dat");
 FileName = JobRaman + C_String; 
 ofstream OutFile1(FileName.c_str(),ios::out);
 if (!OutFile1)
   {
   cerr<< "Error:  "  << FileName  << "Could not be created." << endl;
   exit(1);
 }
 sprintf(C_String,"ShootingExampleInt.dat");
 FileName = JobRaman + C_String; 
 ofstream OutFile2(FileName.c_str(),ios::out);
 if (!OutFile2)
   {
   cerr<< "Error:  "  << FileName  << "Could not be created." << endl;
   exit(1);
 }
 
 double SpaceExample = ( (double)40 / 100 ) * (1525e-9 - 1450e-9) * 2 / (8-1) ;

 UpdateIndividualWavelength(SpaceExample);
 if(NumPumCh==8){
 	for(int NPum=0;NPum<NumPumCh;NPum++) {
		oPumpWaveLth[NPum] += 10e-9;
 	}
 }
 for(int i=1;i<NumZSteps;i++){
 	SignalOnlyPropagation(i);
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	for(int ii=1;ii<NumZSteps;ii++){
		oSignalPower[NSig*NumZSteps+ii] = oSignalPower[NSig*NumZSteps+ii]
				*exp(log(oSignalPower[NSig*NumZSteps]/oSignalPower[NSig*NumZSteps+ii])
				*ii/(NumZSteps-1));
//		cout << oSignalPower[NSig*NumZSteps+ii] << "  " ;
	} //	cout << endl;
 }
 AlphaPre = 0;
 DeltaAlpha = 1;
 Alpha = AlphaPre + DeltaAlpha;

 for(int NPum=0;NPum<NumPumCh;NPum++) {
	oPumpIntgTarget[NPum] = 1000;
 }
// NumSigCh = 0;
// NumCh = NumSigCh + NumPumCh;
 NumItr = 0;
 do{
 //			cout << "run to here2";  
		  for(int NPum=0;NPum<NumPumCh;NPum++) {
	  		oPumpIntgTargetAlpha[NPum] = Alpha * oPumpIntgTarget[NPum];
		  }

		  
	  		if(AlphaPre==0){
			 	if(!(NExtItr>1&&Alpha==1)){		// use previous data
					for(int NPum=0;NPum<NumPumCh;NPum++) {
						oPumpIntgTargetAlpha[NPum] = Alpha * oPumpIntgTarget[NPum];
						oPumpPower[NPum*NumZSteps] = (oPumpIntgTargetAlpha[NPum] * GetAttenuationFiber(oPumpWaveLth[NPum])) 
							/ ( exp( GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength ) -1 ) ;
					//	GetPumpInitGuessFromIntg(oPumpIntgTargetAlpha);	
								
						
//						oPumpPower[NPum*NumZSteps] = oPumpIntgTargetAlpha[NPum] / AmpLength;
						cout << endl <<  " Wavelength=" << oPumpWaveLth[NPum] << " Att = " << GetAttenuationFiber(oPumpWaveLth[NPum])
							<< "  InitPower=" << oPumpPower[NPum*NumZSteps] 
							<< " IntgralTarget=" << oPumpIntgTargetAlpha[NPum] << endl;
//						//*LogFile << "   Intgral=" << oPumpIntgTarget[NPum] << endl;
 					}
				}
	  		}else{
				TotalPumpPower = 0			;
				for(int NPum=0;NPum<NumPumCh;NPum++){
					
					oPumpPower[NPum*NumZSteps] = oPumpGuessPre[NPum] * ( ( Alpha / AlphaPre - 1 )*0.8  + 1 );
					TotalPumpPower += oPumpGuessPre[NPum];
				}
				
			/*	for(int NPum=0;NPum<NumPumCh;NPum++) {
					PumpPowerAdjust = 0;
					for(int P=0;P<NumPumCh;P++) {
						PumpPowerAdjust += oPumpIntgTargetAlpha[NPum] * ( oPumpWaveLth[NPum] - oPumpWaveLth[P] );
					}
					cout << " PumpPowerAdjust=" << PumpPowerAdjust  << " Alpha/AlphaPre=" << Alpha / AlphaPre << " TotalP=" << TotalPumpPower 
								<< " Opumpratio=" 
								<< oPumpPower[NPum*NumZSteps] / TotalPumpPower;
				//	cout << " fir2=" << 	PumpPowerAdjust * (Alpha / AlphaPre) *	oPumpPower[NPum*NumZSteps] / TotalPumpPower;	
					cout << "  adjust[" << NPum << "] = "  ;
					cout <<  0.1 * PumpPowerAdjust * (Alpha / AlphaPre) * (oPumpPower[NPum*NumZSteps] / TotalPumpPower) ;
					oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps] + 0.1 * PumpPowerAdjust * (Alpha / AlphaPre) 
							* (oPumpPower[NPum*NumZSteps] / TotalPumpPower) ;

					cout << endl;
				}
				*/ 
//				exit(1);
		
			}


 AccurateSign=0;
 NumItrInternal = 0;
 do{
    NumItr = NumItr +1;
	NumItrInternal++ ;
//	cout << endl << NumItr << "  " << endl;
//	cout << "BlowupSign=" << BlowupSign << endl;
//	FixSignalShootingSign = 1;
	FixSignalShootingSign = 0;
	FixPumpShootingSign = 0;
	for(int i=1;i<NumZSteps;i++){
        if(BlowupSign==0){
		  	ForwardShooting(i);
		}
    }	
/*	cout  << " Signal and Pump after shooting" << endl;
	for(int NSig=0;NSig<NumSigCh;NSig++){
 		for(int ii=0;ii<NumZSteps;ii++){
			cout << oSignalPower[NSig*NumZSteps+ii] << "  " ;
		} 
		cout << endl;
 	}
	for(int NPum=0;NPum<NumPumCh;NPum++){
 		for(int ii=0;ii<NumZSteps;ii++){
			cout << oPumpPower[NPum*NumZSteps+ii] << "  " ;
		} 
		cout << endl;
 	}	
*/
//	FixSignalShootingSign = 0;
	if(BlowupSign==0){
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] = 0;	
  		}
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			switch(TypeIntergration)
     			{
       				case 1:   // Left Boundary
         				//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         			break;
     
       				case 2:    // Forward estimate Trape Rule
						for(int i=1;i<NumZSteps;i++){
						//	cout << oPumpIntgTarget[0];
							IntgPumPower[NPum] +=  (oPumpPower[NPum*NumZSteps+i] + oPumpPower[NPum*NumZSteps+i-1])/2; 
         					
						}
					break;
     
       				case 3:     //  Forward Distance Average
						for(int i=1;i<NumZSteps;i++){
    		   	  			g = oPumpPower[NPum*NumZSteps+i] / oPumpPower[NPum*NumZSteps+i-1]; 
   	  	 					IntgPumPower[NPum] += oPumpPower[NPum*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   			}

					break;
		 
	    	   	 	default:
        	 			cerr << "With type Intergration must be 1 or ?" << endl 
        	     		<< "Aborting now! " << endl;
         				exit(1);
      			} //end switch type
			
		}
		
	    OutFile1 << NumItr-1 ;
	    OutFile2 << NumItr-1 ;
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] *= DeltaZConst ;	
  			cout << "Wavelength=" << oPumpWaveLth[NPum] << "  Shoot power=" << oPumpPower[NPum*NumZSteps] 
				<< "  Power input=" << oPumpPower[NPum*NumZSteps+NumZSteps-1] 
				<<	" Intg after shooting=" <<	IntgPumPower[NPum] / AmpLength << " -> "<< oPumpIntgTargetAlpha[NPum] / AmpLength <<  endl;
			OutFile1 << "  " << oPumpPower[NPum*NumZSteps+NumZSteps-1]   ;
			OutFile2 << "  " << IntgPumPower[NPum] / AmpLength;
		}
		OutFile1 <<  endl;
		OutFile2 <<  endl;

		//		for(int NPum=0;NPum<NumPumCh;NPum++) {
//			for(int i=0;i<NumZSteps;i++){
//				cout << oPumpPower[NPum*NumZSteps+i] << "  ";
							
//			}	
	//		cout << endl;		
//		}

		for(int NPum=0;NPum<NumPumCh;NPum++) {
			ShootingPower[NPum] = IntgPumPower[NPum] / AmpLength;
			TargetPower[NPum] = oPumpIntgTargetAlpha[NPum] / AmpLength ;	
//  			cout << "Shootng Power=" << ShootingPower[NPum] << " TargetPower[" << NPum << "]=" << TargetPower[NPum] << endl;  
		}

		ParallelIntgCorrection(ShootingPower, TargetPower);
//		OneByOneCorrection(ShootingPower, TargetPower);
	//	exit(1);


	rms =0;
  	for(int i=0;i<NumZSteps;i++){	 
 	 	for(int NPum=0;NPum<NumPumCh;NPum++){	
	 		rms += sq( ( oPumpPower[NPum*NumZSteps+i] - oPumpTrack[NPum*NumZSteps+i] ) 
					/ oPumpPower[NPum*NumZSteps+i] );
			oPumpTrack[NPum*NumZSteps+i] = oPumpPower[NPum*NumZSteps+i] ;
		}
	}


	rms = sqrt( rms / (NumPumCh*NumZSteps) );
	cout << NumItr << " rms=" << rms << endl;
//	cout << rms << "  ";
	if(Alpha < 1 ){
		if (abs(rms) <IterationAccurateTh*80){
//			cout <<  "rms="  << rms;
			AccurateSign = 1;	
		}
	}else{
		if (abs(rms) <IterationAccurateTh){
//			cout <<  "rms="  << rms;
			AccurateSign = 1;	
		}		
	}
	if(rms==0){
		AccurateSign=0;
		BlowupSign=1;
	}
//	if(NumItr>15)
//		AccurateSign=1;	
	if (abs(rms)>100&&NumItrInternal>1) {
		BlowupSign=1;	
		}
	 }
//	 exit(1);
	} while(AccurateSign==0&&NumItr<5000&&BlowupSign==0);
 	
	


	BlowupSign=0;

    if(AccurateSign==1){
		cout << " success with Alpha= " << Alpha << "  with ShootingItrNum=" << NumItr <<endl;
	//	*LogFile << " success with Alpha= " << Alpha << "  with ShootingItrNum=" << NumItr << endl;
	 	for(int NPum=0;NPum<NumPumCh;NPum++){
		    oPumpGuessPre[NPum]=oPumpPower[NPum*NumZSteps];
		}		
	
		DeltaAlpha *= 1.2599;
		AlphaPre = Alpha;
		Alpha = DeltaAlpha + AlphaPre;
		if(AlphaPre>0.8){
	//		AccurateTh=1e-3;
		}

		if (AlphaPre<0.8){
	//		AccurateTh=1e-3;
		}	
		if (AlphaPre<0.5){
	//		AccurateTh=1e-3;
		}	
		if(Alpha >= 1){
			Alpha=1;
	//		AccurateTh=1e-3;
		}				
	}else{
		cout << "failed with Alpha=" << Alpha << "  with ShootingItrNum=" << NumItr <<endl;
//		*LogFile << "failed with Alpha=" << Alpha <<"  with ShootingItrNum=" << NumItr << endl;
		DeltaAlpha/=2;
		Alpha = DeltaAlpha + AlphaPre;
		if(Alpha>1){
			Alpha=1;
		}
	}


 }while(AlphaPre<1);



 FixPumpShootingSign = 1;
 for(int i=1;i<NumZSteps;i++){
    ForwardShooting(i);
 }	 
 FixPumpShootingSign = 0;


 for(int NSig=0;NSig<NumSigCh;NSig++) {
 			cout << "Wavelength=" << oSignalWaveLth[NSig] << "  Input power=" << oSignalPower[NSig*NumZSteps] 
				<< "  Power Output=" << oSignalPower[NSig*NumZSteps+NumZSteps-1]  <<  endl;
		}

 delete [] oPumpGuessPre;  
 delete [] IntgPumPower;  
 delete [] oPumpIntgTargetAlpha;  
 delete [] ShootingPower;  
 delete [] TargetPower;  
 delete [] oPumpTrack;
} // end ShootingIntegrationExample

// ###############################

void RamanAmp::ShootingIntegration(int NExtItr)
{
 double *oPumpGuessPre = new double[NumPumCh];
// double oPumpGuessPre[NumPumCh];
 double rms;
// double rPumpPower[NumPumCh];
 int    NumItr;
 double Alpha,AlphaPre,DeltaAlpha;
 double PumpPowerAdjust;
 double *IntgPumPower = new double[NumPumCh];  
 double *oPumpIntgTargetAlpha = new double[NumPumCh];  
// double IntgPumPower[NumPumCh],oPumpIntgTargetAlpha[NumPumCh];
 double g;
 double *ShootingPower = new double[NumPumCh];  
 double *TargetPower = new double[NumPumCh];  
 double CenterPumpWavelength = 1450e-9;
 double TotalPumpPower;
 double *oPumpTrack = new double[NumZSteps*NumPumCh];
// double ShootingPower[NumPumCh],TargetPower[NumPumCh]; 
// double ErrorPum[NumZSteps];
 for(int i=0;i<NumZSteps;i++){
  	for(int NPum=0;NPum<NumPumCh;NPum++){	
		oPumpTrack[NPum*NumZSteps+i] = 1;
	}
 }

 AlphaPre = 0;
 DeltaAlpha = 1;
 Alpha = AlphaPre + DeltaAlpha;


// NumSigCh = 0;
// NumCh = NumSigCh + NumPumCh;

 do{
 //			cout << "run to here2";  
		  for(int NPum=0;NPum<NumPumCh;NPum++) {
	  		oPumpIntgTargetAlpha[NPum] = Alpha * oPumpIntgTarget[NPum];
		  }

		  
	  		if(AlphaPre==0){
			 	if(!(NExtItr>1&&Alpha==1)){		// use previous data
					for(int NPum=0;NPum<NumPumCh;NPum++) {
						oPumpIntgTargetAlpha[NPum] = Alpha * oPumpIntgTarget[NPum];
						oPumpPower[NPum*NumZSteps] = (oPumpIntgTargetAlpha[NPum] * GetAttenuationFiber(oPumpWaveLth[NPum])) 
							/ ( exp( GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength ) -1 ) ;
					//	GetPumpInitGuessFromIntg(oPumpIntgTargetAlpha);	
								
						
//						oPumpPower[NPum*NumZSteps] = oPumpIntgTargetAlpha[NPum] / AmpLength;
//						cout << endl <<  " Wavelength=" << oPumpWaveLth[NPum] << "  InitPower=" << oPumpPower[NPum*NumZSteps] 
//							<< " IntgralTarget=" << oPumpIntgTargetAlpha[NPum] ;
//						//*LogFile << "   Intgral=" << oPumpIntgTarget[NPum] << endl;
 					}
				}
	  		}else{
				TotalPumpPower = 0			;
				for(int NPum=0;NPum<NumPumCh;NPum++){
					
					oPumpPower[NPum*NumZSteps] = oPumpGuessPre[NPum] * ( ( Alpha / AlphaPre - 1 )  + 1 );
					TotalPumpPower += oPumpGuessPre[NPum];
				}
				
			/*	for(int NPum=0;NPum<NumPumCh;NPum++) {
					PumpPowerAdjust = 0;
					for(int P=0;P<NumPumCh;P++) {
						PumpPowerAdjust += oPumpIntgTargetAlpha[NPum] * ( oPumpWaveLth[NPum] - oPumpWaveLth[P] );
					}
					cout << " PumpPowerAdjust=" << PumpPowerAdjust  << " Alpha/AlphaPre=" << Alpha / AlphaPre << " TotalP=" << TotalPumpPower 
								<< " Opumpratio=" 
								<< oPumpPower[NPum*NumZSteps] / TotalPumpPower;
				//	cout << " fir2=" << 	PumpPowerAdjust * (Alpha / AlphaPre) *	oPumpPower[NPum*NumZSteps] / TotalPumpPower;	
					cout << "  adjust[" << NPum << "] = "  ;
					cout <<  0.1 * PumpPowerAdjust * (Alpha / AlphaPre) * (oPumpPower[NPum*NumZSteps] / TotalPumpPower) ;
					oPumpPower[NPum*NumZSteps] = oPumpPower[NPum*NumZSteps] + 0.1 * PumpPowerAdjust * (Alpha / AlphaPre) 
							* (oPumpPower[NPum*NumZSteps] / TotalPumpPower) ;

					cout << endl;
				}
				*/ 
//				exit(1);
		
			}


 AccurateSign=0;
 NumItr = 0;
 do{
    NumItr = NumItr +1;

//	cout << endl << NumItr << "  " << endl;
//	cout << "BlowupSign=" << BlowupSign << endl;
	FixSignalShootingSign = 1;
	for(int i=1;i<NumZSteps;i++){
        if(BlowupSign==0){
		  	ForwardShooting(i);
		}
    }	 
	FixSignalShootingSign = 0;
	if(BlowupSign==0){
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] = 0;	
  		}
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			switch(TypeIntergration)
     			{
       				case 1:   // Left Boundary
         				//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         			break;
     
       				case 2:    // Forward estimate Trape Rule
						for(int i=1;i<NumZSteps;i++){
						//	cout << oPumpIntgTarget[0];
							IntgPumPower[NPum] +=  (oPumpPower[NPum*NumZSteps+i] + oPumpPower[NPum*NumZSteps+i-1])/2; 
         					
						}
					break;
     
       				case 3:     //  Forward Distance Average
						for(int i=1;i<NumZSteps;i++){
    		   	  			g = oPumpPower[NPum*NumZSteps+i] / oPumpPower[NPum*NumZSteps+i-1]; 
   	  	 					IntgPumPower[NPum] += oPumpPower[NPum*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   			}

					break;
		 
	    	   	 	default:
        	 			cerr << "With type Intergration must be 1 or ?" << endl 
        	     		<< "Aborting now! " << endl;
         				exit(1);
      			} //end switch type
			
		}
		
	
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] *= DeltaZConst ;	
//  			cout << "Wavelength=" << oPumpWaveLth[NPum] << "  Shoot power=" << oPumpPower[NPum*NumZSteps] 
//				<< " Intg after shooting=" <<	IntgPumPower[NPum] << " -> "
//				<< oPumpIntgTargetAlpha[NPum] << "  Power=" << oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl;
		}
//		for(int NPum=0;NPum<NumPumCh;NPum++) {
//			for(int i=0;i<NumZSteps;i++){
//				cout << oPumpPower[NPum*NumZSteps+i] << "  ";
							
//			}	
	//		cout << endl;		
//		}

		for(int NPum=0;NPum<NumPumCh;NPum++) {
			ShootingPower[NPum] = IntgPumPower[NPum] / AmpLength;
			TargetPower[NPum] = oPumpIntgTargetAlpha[NPum] / AmpLength ;	
//  			cout << "Shootng Power=" << ShootingPower[NPum] << " TargetPower[" << NPum << "]=" << TargetPower[NPum] << endl;  
		}

//		ParallelIntgCorrection(ShootingPower, TargetPower);
		OneByOneCorrection(ShootingPower, TargetPower);
	//	exit(1);


	rms =0;
  	for(int i=0;i<NumZSteps;i++){	 
 	 	for(int NPum=0;NPum<NumPumCh;NPum++){	
	 		rms += sq( ( oPumpPower[NPum*NumZSteps+i] - oPumpTrack[NPum*NumZSteps+i] ) 
					/ oPumpPower[NPum*NumZSteps+i] );
			oPumpTrack[NPum*NumZSteps+i] = oPumpPower[NPum*NumZSteps+i] ;
		}
	}


	rms = sqrt( rms / (NumPumCh*NumZSteps) );
//	cout << endl << "rms=" << rms ;
//	cout << rms << "  ";
	if (abs(rms) <IterationAccurateTh){
//		cout <<  "rms="  << rms;
		AccurateSign = 1;	
		}
	if(rms==0){
		AccurateSign=0;
		BlowupSign=1;
	}
//	if(NumItr>15)
//		AccurateSign=1;	
	if (abs(rms)>100&&NumItr>1) {
		BlowupSign=1;	
		}
	 }
	} while(AccurateSign==0&&NumItr<5000&&BlowupSign==0);
 	
	


	BlowupSign=0;

    if(AccurateSign==1){
		cout << " success with Alpha= " << Alpha << "  with ShootingItrNum=" << NumItr <<endl;
	//	*LogFile << " success with Alpha= " << Alpha << "  with ShootingItrNum=" << NumItr << endl;
	 	for(int NPum=0;NPum<NumPumCh;NPum++){
		    oPumpGuessPre[NPum]=oPumpPower[NPum*NumZSteps];
		}		
	
		DeltaAlpha *= 1.2599;
		AlphaPre = Alpha;
		Alpha = DeltaAlpha + AlphaPre;
		if(AlphaPre>0.8){
	//		AccurateTh=1e-3;
		}

		if (AlphaPre<0.8){
	//		AccurateTh=1e-3;
		}	
		if (AlphaPre<0.5){
	//		AccurateTh=1e-3;
		}	
		if(Alpha >= 1){
			Alpha=1;
	//		AccurateTh=1e-3;
		}				
	}else{
		cout << "failed with Alpha=" << Alpha << "  with ShootingItrNum=" << NumItr <<endl;
//		*LogFile << "failed with Alpha=" << Alpha <<"  with ShootingItrNum=" << NumItr << endl;
		DeltaAlpha/=2;
		Alpha = DeltaAlpha + AlphaPre;
		if(Alpha>1){
			Alpha=1;
		}
	}


 }while(AlphaPre<1);

 FixPumpShootingSign = 1;
 for(int i=1;i<NumZSteps;i++){
    ForwardShooting(i);
 }	 
 FixPumpShootingSign = 0;

 delete [] oPumpGuessPre;  
 delete [] IntgPumPower;  
 delete [] oPumpIntgTargetAlpha;  
 delete [] ShootingPower;  
 delete [] TargetPower;  
 delete [] oPumpTrack;
} // end ShootingIntegration

// #######################################

void RamanAmp::GetPumpInitGuessFromIntg(double * TargetAlpha)
{
 double *PumpGain = new double[NumPumCh];  


 for(int NPum=0;NPum<NumPumCh;NPum++){ 
	PumpGain[NPum] = 0;
 } 
 for(int NPum=0;NPum<NumPumCh;NPum++){ 
 	for(int P=0;P<NumPumCh;P++){
   		PumpGain[NPum] += GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P])*TargetAlpha[P];
//		cout << PumpGain[NPum] << "  ";
	}
 	for(int S=0;S<NumSigCh;S++){
 		PumpGain[NPum] += GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S])*oIntgSigPower[S];
	}
	PumpGain[NPum] =  1 / exp(PumpGain[NPum] - GetAttenuationFiber(oPumpWaveLth[NPum]) * AmpLength);
// 	cout << PumpGain[NPum] << "  ";

	oPumpPower[NPum*NumZSteps] = ( TargetAlpha[NPum] / AmpLength ) * log(PumpGain[NPum])
 			/ ( PumpGain[NPum] -1 ) ;
 //	cout << " PumpInit[" << NPum << "]" << oPumpPower[NPum*NumZSteps] << endl;
 }
// exit(1);
 delete [] PumpGain;
 
} // end GetPumpInitGuessFromIntg
// ########################################

void RamanAmp::GeneticAlgorithmsVaryCW(double * Vector, double PumpSpace,double * CW)
{
// double Vector[Population*NumPumCh*2];
 double *IntgSigPower = new double[NumSigCh];  
// double IntgSigPower[NumSigCh];
 double *Fitness = new double[Population];  
 double *OffsetMaxMin = new double [Population];
 double *GainLK = new double[NumSigCh];  
 double *GainGK = new double[NumSigCh];  
 int *RndPerm = new int[Population];  
 int *MutationSign = new int[Population*NumPumCh*2];  
 int *RndPermPlace = new int[2*NumPumCh];  

// double Fitness[Population],GainLK[NumSigCh],GainGK[NumSigCh];
 double Max;
 int MaxNum;
 double ExTemp, MaxAveFitness=100, AveFitness;
 int PuR,PuL;
// int RndPerm[Population],
 int RndTemp;
// int MutationSign[Population*NumPumCh*2];
 int NumMutation=0;
// int PlaceMutation;
 double g,AveFitnessTh;
 int StopGA, NumItrGA;
 int PlaceCross, NumCross;
 int IntTemp;
// int RndPermPlace[2*NumPumCh];
 double MaxGgkGlk,MinGgkGlk; 
 double MaxIj; 	 
	int ChangeSign;
 int RestSelectionFixedNum;
 double Fit;
 int SameFitnessCount=0;
 	double V1,V2,V3,V4;

// Population generation
// for(int i=0;i<Population;i++){
//	for(int j=0;j<NumPumCh;j++){
//		Vector[ 2*i*NumPumCh + 2*j ] = (100*RNG->GetRanNum()+1410)*1E-9;
//		Vector[ 2*i*NumPumCh + 2*j + 1 ] = 30000*RNG->GetRanNum()+100;
//	}
// }		
//	 
//fitness calculation
// cout << "\a";
// cout << "AveFitness=?" ;
// cin >> AveFitnessTh ;
 AveFitnessTh = 10; 
// AveFitnessTh = 0.001;
 NumItrGA=0;
 StopGA=0;
 RestSelectionFixedNum = Population - SelectionFixedNum/100;

 for(int S=0;S<NumSigCh;S++){
	IntgSigPower[S]=0;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	switch(TypeIntergration){
       			case 1:   // Left Boundary
         			//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         		break;
     
       			case 2:    // Forward estimate Trape Rule
					for(int i=1;i<NumZSteps;i++){
						IntgSigPower[NSig] +=  (oSignalPower[NSig*NumZSteps+i] + oSignalPower[NSig*NumZSteps+i - 1])/2; 
         			}
				break;
     
       			case 3:     //  Forward Distance Average
					for(int i=1;i<NumZSteps;i++){
    		     		g = oSignalPower[NSig*NumZSteps+i] / oSignalPower[NSig*NumZSteps+i-1]; 
	       	  	 		IntgSigPower[NSig] += oSignalPower[NSig*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   		}
				break;
		 
	    	    default:
        	 		cerr << "With type Intergration must be 1 or ?" << endl 
        	     	<< "Aborting now! " << endl;
         			exit(1);
    } //end switch type
	
 }
// 	  cout << "run to here4" << endl;
 for(int NSig=0;NSig<NumSigCh;NSig++) {
	IntgSigPower[NSig] *= DeltaZConst ;	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 }
  for(int NSig=0;NSig<NumSigCh;NSig++) {
	oIntgSigPower[NSig] = IntgSigPower[NSig];	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 } 
 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
	GainLK[NSig] = 0;
 } 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
 	for(int S=0;S<NumSigCh;S++){
//		cout << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S]) << " * " << IntgSigPower[S]
//				<< " = " << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S]<< endl;
   		GainLK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S];
//		cout << "  GainLK=" << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S] << endl;
	}
	GainLK[NSig] =  GainLK[NSig] - GetAttenuationFiber(oSignalWaveLth[NSig]) * AmpLength;

//	cout << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
//	*LogFile << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
 }
// cout << endl;
// *LogFile << endl;
//		Vector( 2*i*NumPumCh + 2*j ) = (100*RNG->GetRanNum()+1410)*1E-9;
	//	Vector( 2*i*NumPumCh + 2*j + 1 ) = (490*RNG->GetRanNum()+10)*1E-3;

/*	for(int i=0;i<Population;i++){
		for(int NPum=0;NPum<NumPumCh;NPum++){						
							if(Vector[ 2*(Population-i)*NumPumCh + 2*NPum ] >  2000e-9){
							cout << " problem in 1" << endl;
				//			cout << " Vector=" << Vector[ 2*(Population-1)*NumPumCh + 2*7 ] << "  ";
				//			cout << " MaxNum=" << MaxNum << "  Pu=" << Pu << endl;
				//			cout << " V1=" << V1 << "  V2=" << V2 << endl;
				//			cout << " V3=" << V3 << "  V4=" << V4 << endl;							
							exit(1);
							}
							
		}
	}
*/
do{
 NumItrGA++;

// ####################
// fitness calculatation
/* for(int Pu=0;Pu<Population;Pu++){
 	Fitness[Pu]=0;
 }
 for(int Pu=0;Pu<Population;Pu++){
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
		Fitness[Pu] += sq( exp(GainLK[NSig]) * exp(GainGK[NSig] ) - 1);
	}
  	Fitness[Pu] = 1/sqrt(Fitness[Pu]);
//	cout <<"  Fitness=" << Fitness[Pu] << endl;
 }

  for(int Pu=0; Pu<Population;Pu++){

		MaxIj=0;
		for(int NPum=0;NPum<NumPumCh;NPum++){
			if(Vector[ 2*Pu*NumPumCh + 2*NPum + 1] > MaxIj)
				MaxIj =	Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
		}
		for(int NPum=0;NPum<NumPumCh;NPum++){
			Vector[ 2*Pu*NumPumCh + 2*NPum + 1] /= MaxIj;
		}
  } 
   */
 
// (3) fitness defination

  for(int Pu=0;Pu<Population;Pu++){
	for(int S=0;S<NumSigCh;S++){	
 		GainGK[S] =0;
 	}
 	for(int NSig=0;NSig<NumSigCh;NSig++){	

//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << " Gain =" << GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ]) ;
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				//	MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				    MaxGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
				//	MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
					MinGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
				}
			}
		}
	}
//    cout << MaxGgkGlk << "  " << MinGgkGlk << "  ";

	Fitness[Pu] = MaxGgkGlk - MinGgkGlk; 
//	cout << "  fitness= " << Fitness[Pu] ;
	OffsetMaxMin[Pu] = (MaxGgkGlk + MinGgkGlk) / 2 ;
	Fitness[Pu] /= OffsetMaxMin[Pu];
//	cout << "offset[" << Pu << "]=" << OffsetMaxMin[Pu] << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] /= OffsetMaxMin[Pu];
//		cout << "  "  << Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
	}	

//  	Fitness[Pu] = exp(-1*(MaxGgkGlk - MinGgkGlk));
	
		//	cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << endl;

 }
 
/*	for(int i=0;i<Population;i++){
		for(int NPum=0;NPum<NumPumCh;NPum++){						
							if(Vector[ 2*(Population-i)*NumPumCh + 2*NPum ] >  2000e-9){
							cout << " problem in 2" << endl;
				//			cout << " Vector=" << Vector[ 2*(Population-1)*NumPumCh + 2*7 ] << "  ";
				//			cout << " MaxNum=" << MaxNum << "  Pu=" << Pu << endl;
				//			cout << " V1=" << V1 << "  V2=" << V2 << endl;
				//			cout << " V3=" << V3 << "  V4=" << V4 << endl;							
							exit(1);
							}
							
		}
	}
*/ 
// cout << "run to here5" << endl;
//selection
 for(int Pu=Population-1;Pu>=1;Pu--){
 	Max=Fitness[0];
 	MaxNum = 0;
 	for(int SubPu=1;SubPu<=Pu;SubPu++){
//		cout << Max << " <>? " << Fitness[SubPu] << "  " ;
 		if(Max>Fitness[SubPu]){
			Max = Fitness[SubPu];
			MaxNum = SubPu;
//			cout << "  MaxNum= " << MaxNum << endl;
		}
 	}
//	cout << "run to here5.2" << endl;
 	//exchange
	
	

	for(int j=0;j<NumPumCh;j++){
//		cout << "maxnum=" << MaxNum << endl;
		ExTemp = Vector[ 2*MaxNum*NumPumCh + 2*j ];
		Vector[ 2*MaxNum*NumPumCh + 2*j ] = Vector[ 2*Pu*NumPumCh + 2*j ];
		Vector[ 2*Pu*NumPumCh + 2*j ] = ExTemp;
		ExTemp = Vector[ 2*MaxNum*NumPumCh + 2*j + 1];
		Vector[ 2*MaxNum*NumPumCh + 2*j + 1] = Vector[ 2*Pu*NumPumCh + 2*j + 1];
		Vector[ 2*Pu*NumPumCh + 2*j + 1 ] = ExTemp;
//		cout << "run to here5.3" << endl; 
 	}


//	cout << "run to here5.4" << endl; 
	ExTemp = Fitness[MaxNum];
	Fitness[MaxNum] = Fitness[Pu];
	Fitness[Pu] = ExTemp;
	ExTemp = OffsetMaxMin[MaxNum];
	OffsetMaxMin[MaxNum] = OffsetMaxMin[Pu];
	OffsetMaxMin[Pu] = ExTemp;	
		
 }
// cout << "run to here5.5" << endl; 
 
// if(AveFitness > 10 * Fitness[Population-1]){ MutationRate = 0.01;}
// 	else{ MutationRate = 0.05;}
 AveFitness=0;
 for(int i=0;i<Population;i++){
  	AveFitness += Fitness[i];

//    cout << Fitness[i] << "  ";
 }
 
// cin >> AveFitnessTh ; 

 AveFitness /= (double)Population;
// if(MaxAveFitness>AveFitness){
 	MaxAveFitness = AveFitness;
//	cout << "#" << NumItrGA <<  "  Best Fitness = " << Fitness[Population-1] 
//		<< "  MutationRate = " << MutationRate << " SameFitCount=" << SameFitnessCount << endl;
//		cout << " AveFitnessTh=" << AveFitnessTh << "  Fitness[Popul-1]=" << Fitness[Population-1] << endl;
//}
/* PuR=Population-1;
 PuL=0;
// cout << "run to here5.6" << endl;
 do{
 	for(int i=0;i<(int)((Fitness[PuR]/AveFitness)-1);i++){	
//	cout << "NumChild[" << PuR << "]=" << int(Fitness[PuR]/AveFitness) << endl;
//	cout << "run to here 5.7" << endl;
 		for(int j=0;j<NumPumCh;j++){
 			Vector[ 2*PuL*NumPumCh + 2*j ] = Vector[ 2*PuR*NumPumCh + 2*j ];
			Vector[ 2*PuL*NumPumCh + 2*j + 1] = Vector[ 2*PuR*NumPumCh + 2*j + 1];
		}
		PuL++;
//		cout << "PuL=" << PuL  << "  PuR=" << PuR << "  i=" << i << endl;
//		cout << "run to here 5.75" << endl;		
		if(PuL>=PuR){i=(int)(Fitness[PuR]/AveFitness);}
	}
	PuR--;
//	cout << "run to here 5.8" << endl;
 }while(PuL<PuR&&(int(Fitness[PuR]/AveFitness)-1>0));
*/



/* for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	cout <<"  Fitness[" << Pu << "]=" << Fitness[Pu]  << " Offset=" << OffsetMaxMin[Pu]  
			<< "  /=" << Fitness[Pu]/OffsetMaxMin[Pu]; // << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] = Vector[ 2*Pu*NumPumCh + 2*NPum + 1] / OffsetMaxMin[Pu];
		cout << "  "  << Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
	}
	cout << endl;
  }	

 for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	for(int S=0;S<NumSigCh;S++){	
 		GainGK[S] =0;
 	}
 	for(int NSig=0;NSig<NumSigCh;NSig++){	

//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += (oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
					MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				}
			}
		}
	}
	Fitness[Pu] = MaxGgkGlk - MinGgkGlk; 
	OffsetMaxMin[Pu] = (MaxGgkGlk + MinGgkGlk) / 2 ;
//  	Fitness[Pu] = exp(-1*(MaxGgkGlk - MinGgkGlk));
//	Fitness[Pu] = 1/(MaxGgkGlk - MinGgkGlk);
		//	cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << endl;
	cout <<"  Fitness[" << Pu << "]=" << Fitness[Pu] << endl;
 }	
	exit(1);
*/
	
	if(Fitness[Population-1] > AveFitnessTh - 1e-6){
		SameFitnessCount++;
//		AveFitnessTh = Fitness[Population-1];
	}else{
		SameFitnessCount=0;
		AveFitnessTh = Fitness[Population-1];
	}

// if(Fitness[Population-1]< AveFitnessTh && NumItrGA>=2){
  if(SameFitnessCount==30){
//	cout << "StopGA= " << StopGA;
//	cout << AveFitnessTh << " = " << AveFitnessTh << endl;
//	cout << AveFitness << " > " << AveFitnessTh << " NUmItrGA=" << NumItrGA << endl;
 	StopGA=1;
 }else{ 	
	
 for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	
	for(int NPum=0;NPum<NumPumCh;NPum++){
		ChangeSign = 0;
 	 	for(int P=0;P<(NumPumCh-NPum)-1;P++){
	 		if(Vector[ 2*Pu*NumPumCh + 2*P ] > Vector[ 2*Pu*NumPumCh + 2*(P+1) ]){
			 	ChangeSign = 1;
				ExTemp= Vector[ 2*Pu*NumPumCh + 2*P ];
				Vector[ 2*Pu*NumPumCh + 2*P ] = Vector[ 2*Pu*NumPumCh + 2*(P+1) ];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) ] = ExTemp;
				ExTemp = Vector[ 2*Pu*NumPumCh + 2*P + 1];
				Vector[ 2*Pu*NumPumCh + 2*P + 1] = Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1] = ExTemp;
			}
	 	}
		if(ChangeSign == 0)	{NPum = NumPumCh;}
   	}
 }
 for(int ii=0;ii<Population/SelectionFixedNum-1;ii++){
 	for(int i=0;i<SelectionFixedNum;i++){
		for(int j=0;j<NumPumCh;j++){
			Vector[ 2*(i+ii*SelectionFixedNum)*NumPumCh + 2*j ] = Vector[ 2*(i+Population-SelectionFixedNum)*NumPumCh + 2*j ];
			Vector[ 2*(i+ii*SelectionFixedNum)*NumPumCh + 2*j + 1] = Vector[ 2*(i+Population-SelectionFixedNum)*NumPumCh + 2*j + 1];
		}
	}
 }

//cout << "  NumItrGA=" << NumItrGA ;
//Random Permutation
// cout << "run to here6" << endl;

 	 for (int Pu=0;Pu<RestSelectionFixedNum ;Pu++){
 		RndPerm[Pu]=Pu;
	 }
	 for(int Pu=0;Pu<RestSelectionFixedNum;Pu++){
		RndTemp = (int)((RestSelectionFixedNum-Pu)*RNG->GetRanNum());
		if(RndTemp==(RestSelectionFixedNum-Pu)){RndTemp=RestSelectionFixedNum-Pu-1;}
 	    IntTemp = RndPerm[RndTemp];
		for(int i=RndTemp;i<RestSelectionFixedNum-Pu-1;i++){
			RndPerm[i] = RndPerm[i + 1];
		}
 		RndPerm[RestSelectionFixedNum - Pu - 1] = IntTemp;
	 }	
//crossover
//	cout << "Vector[Pu]before crossover" << endl;
// 	for(int i=0;i<Population;i++){
//		for(int j=0;j<NumPumCh;j++){
	//	   cout << "i,j" <<i << "  " << j;
		//	cout << Vector[ 2*i*NumPumCh + 2*j ] << "  " << Vector[ 2*i*NumPumCh + 2*j + 1 ] << "  ";
//		}
// 	}
//	cout << endl;
//	cout << "run to here6.1" << endl;


	 for(int Pu=0;Pu<RestSelectionFixedNum;Pu += 2){
			NumCross = (int)((NumPumCh-1)*RNG->GetRanNum()+1);

	 	 	for (int i=0;i<NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}
	 		for (int i=0;i<NumPumCh;i++){
				RndTemp = (int)((NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (NumPumCh-i)){RndTemp=NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
				for(int j=RndTemp;j< NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[NumPumCh - i - 1] = IntTemp;
	 		}	
			for(int NC=0;NC<NumCross;NC++){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + 2 * RndPermPlace[NC] + 1 ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + 2 * RndPermPlace[NC] +1 ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + 2 *
						RndPermPlace[NC] +1];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + 2* RndPermPlace[NC] +1] = ExTemp;
			}
			if(RNG->GetRanNum() > CrossRate){
				ExTemp = CW[RndPerm[Pu]];
				CW[RndPerm[Pu]] = CW[RndPerm[Pu+1]];
				CW[RndPerm[Pu+1]] = ExTemp;	
				
				UpdateIndividualWavelength(RndPerm[Pu], Vector, PumpSpace, CW[RndPerm[Pu]]);				
				UpdateIndividualWavelength(RndPerm[Pu+1], Vector, PumpSpace, CW[RndPerm[Pu+1]]);							
			}
 	}
 
	/*
	 for(int Pu=0;Pu<Population;Pu += 2){

			if(RNG->GetRanNum() > CrossRate){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + Place ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ] = ExTemp;
			}
	   }
 	}

*/	
/*	
	for(int Pu=0;Pu<RestSelectionFixedNum;Pu += 2){
//		if(RNG->GetRanNum()>CrossRate){
//			cout << "do crossover this time" << endl;
//			PlaceCross = (int)(NumPumCh*RNG->GetRanNum());
			NumCross = (int)(NumPumCh*RNG->GetRanNum());
//			NumCross = (int)(NumPumCh * CrossRate);
	//		if(NumCross==2*NumPumCh){ NumCross = 2*NumPumCh - 1; }
	
	 	 	for (int i=0;i<NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}

	 		for (int i=0;i<NumPumCh;i++){
		
				RndTemp = (int)((NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (NumPumCh-i)){RndTemp=NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
	//			if(IntTemp<0||IntTemp>2 * NumPumCh){
	//				cout << endl << RndTemp;
	//				exit(1);
	//			}
				for(int j=RndTemp;j<  NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[NumPumCh - i - 1] = IntTemp;
				
	 		}
	
			for(int NC=0;NC<NumCross;NC++){
		
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + 2*RndPermPlace[NC] + 1];
				Vector[ 2*RndPerm[Pu]*NumPumCh + 2*RndPermPlace[NC] + 1] = Vector[ 2*RndPerm[Pu+1]*NumPumCh +	2*RndPermPlace[NC] + 1];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + 2*RndPermPlace[NC] + 1 ] = ExTemp;
		
			//	PlaceCross++;
			//	if(PlaceCross>=(NumPumCh)){
			//		PlaceCross -= NumPumCh;
			//	}
		 
			}
//		}
 	}
	
*/

// mutation
//	cout << "run to here7" << endl;
		for(int Pu=0; Pu < RestSelectionFixedNum ; Pu++){		 
			for(int NPum=0;NPum<NumPumCh;NPum++){
				if(RNG->GetRanNum() < MutationRate){
					if(AveFitness>0.8){Vector[Pu*2*NumPumCh+2*NPum+1] = 30000*RNG->GetRanNum()+10;}
					else{if(RNG->GetRanNum()>0.95){ 
								Vector[Pu*2*NumPumCh+2*NPum+1] = 30000*RNG->GetRanNum()+10;
						 }else{
								Vector[Pu*2*NumPumCh+2*NPum+1] *= 1+0.1*(2*RNG->GetRanNum()-1);
								if(Vector[Pu*2*NumPumCh+2*NPum+1] < 0.001) {Vector[Pu*2*NumPumCh+2*NPum+1] = 0.001;}
								if(Vector[Pu*2*NumPumCh+2*NPum+1] > 300000) {Vector[Pu*2*NumPumCh+2*NPum+1] = 300000;}
						}
					}
				}
			}
			if(RNG->GetRanNum() < MutationRate){
				if(AveFitness>0.8){
//					PumpSpace[Pu] = RNG->GetRanNum()*150e-9/(NumPumCh-1);
					CW[Pu] = 1375e-9 + PumpSpace * (NumPumCh-1) / 2
								+ (150e-9 - PumpSpace * (NumPumCh-1)) * RNG->GetRanNum() ;
				}else{if(RNG->GetRanNum()>0.95){ 
//						PumpSpace[Pu] = RNG->GetRanNum()*150e-9/(NumPumCh-1);
						CW[Pu] = 1375e-9 + PumpSpace * (NumPumCh-1) / 2
								+ (150e-9 - PumpSpace * (NumPumCh-1)) * RNG->GetRanNum() ;
						}else{
	//					PumpSpace[Pu] *= 1+0.1*(2*RNG->GetRanNum()-1);
								CW[Pu] *= 1+0.1*(2*RNG->GetRanNum()-1);
	//							if(PumpSpace[Pu] > 145e-9) {PumpSpace[Pu] = 145e-9;}
								if(CW[Pu] + PumpSpace / 2 - PumpSpace * NumPumCh / 2 < 1375e-9){
									CW[Pu] = 1375e-9 + PumpSpace * (NumPumCh-1) / 2;	
								}
								if(CW[Pu] - PumpSpace / 2 + PumpSpace * NumPumCh / 2 > 1525e-9){
									CW[Pu] = 1525e-9 - PumpSpace * (NumPumCh-1) / 2;	
								}							
						}
				}
				UpdateIndividualWavelength(Pu, Vector, PumpSpace, CW[Pu]);			
			}
		}
		
		if(Vector[ 2*(Population-1)*NumPumCh + 2*7 ] >  2000e-9){
			cout << "error";
		} 
		
   }
 }while(StopGA==0);
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	 oPumpWaveLth[NPum] = Vector[ 2*(Population-1)*NumPumCh + 2*NPum ];
	 oPumpIntgTarget[NPum] = Vector[ 2*(Population-1)*NumPumCh + 2*NPum + 1] ;
//	 		OffsetMaxMin[Population-1];
 }
/* for(int NPum=0;NPum<NumPumCh;NPum++){
 	cout << " Wavelength=" << Vector[ 2*(Population-1)*NumPumCh + 2*NPum ] 
	 		<< "  " << Vector[ 2*(Population-2)*NumPumCh + 2*NPum ] << endl;
 }
 for(int i=1;i<50;i++){
 	cout << "  " << Vector[ 2*(Population-i)*NumPumCh + 2*7 ]; 
 }
*/
// for(int NPum=0;NPum<NumPumCh;NPum++){
// 	 cout << " Wavelengthnext=" << Vector[ 2*(Population-2)*NumPumCh + 2*NPum ] << endl;
// }  
 
//	cout << "run to here8" << endl;
/* for(int NPum=0;NPum<NumPumCh;NPum++){
 	 for(int P=0;P<(NumPumCh-NPum)-1;P++){
	 	if(oPumpWaveLth[P]>oPumpWaveLth[P+1]){
			ExTemp= oPumpWaveLth[P];
			oPumpWaveLth[P] = oPumpWaveLth[P+1];
			oPumpWaveLth[P+1] = ExTemp;
			ExTemp = oPumpIntgTarget[P];
			oPumpIntgTarget[P] = oPumpIntgTarget[P+1];
			oPumpIntgTarget[P+1] = ExTemp;
		}
	} 
 }

 for(int NPum=0;NPum<NumPumCh;NPum++){
 	cout <<  " PumpWavelength=" << oPumpWaveLth[NPum] << " Intg=" << oPumpIntgTarget[NPum] << endl;
 }
  */
// cout << endl;
 double FitApprove=0;
 double GainRipple=0;
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P])
				*  oPumpIntgTarget[P];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
		FitApprove  += sq( exp(GainLK[NSig]) * exp(GainGK[NSig] ) - 1);
//		cout << 10*log10( exp(GainLK[NSig]) * exp(GainGK[NSig] )) << "  ";
	//	*LogFile << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
//	cout <<  "Gain["<< NSig << "]=" 
    if (abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])))>GainRipple)
		GainRipple = abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])));
		
//	*LogFile << oSignalWaveLth[NSig] << "   " 
//		<< 10*log10(exp(GainLK[NSig])*exp(GainGK[NSig] )) << "   " 
//		<< (-1) * GainGK[NSig] / GainLK[NSig] 
//		<< endl;

	}
//	*LogFile << endl << "  OffsetMaxMin[Population-1] = " << OffsetMaxMin[Population-1] << endl;
//	*LogFile << " Gain Ripple = " << GainRipple << endl;
	
//	cout << endl << "FitApprove=" << FitApprove ;

  	FitApprove = 1/sqrt(FitApprove); 	
//	cout << endl << "FitApprove=" << FitApprove << "  Fitness[]=" << Fitness[Population-1] ;
	
//	Fitness[Pu] = 1/sqrt(Fitness[Pu]);
//	cout <<"  Fitness=" << Fitness[Pu] << endl;
 	cout << "  BestFitness=" << Fitness[Population-1] << "  GainRipple=" << GainRipple << 
			"  G.A.Itr#=" << NumItrGA << endl;
// cout << endl;
 for(int i=0;i<Population;i++){
  	AveFitness += Fitness[i];
//    cout <<  Fitness[i] << "  ";
 }
// *LogFile << "AveFitness =" << AveFitness/Population << "   Highest Fitness=" << Fitness[Population-1] << endl;
// cout << endl;
// CalculateFrel();

 delete [] IntgSigPower ;  
 delete [] Fitness ;  
 delete [] OffsetMaxMin;
 delete [] GainLK ;  
 delete [] GainGK ;  
 delete [] RndPerm ;  
 delete [] MutationSign ;  
 delete [] RndPermPlace ;  
// exit(1);

} // end GeneticAlgorithmsVaryCW

// ########################################

void RamanAmp::GeneticAlgorithmsVictor(double * Vector)
{
// double Vector[Population*NumPumCh*2];
 double *IntgSigPower = new double[NumSigCh];  
// double IntgSigPower[NumSigCh];
 double *Fitness = new double[Population];  
 double *OffsetMaxMin = new double [Population];
 double *GainLK = new double[NumSigCh];  
 double *GainGK = new double[NumSigCh];  
 int *RndPerm = new int[Population];  
 int *MutationSign = new int[Population*NumPumCh*2];  
 int *RndPermPlace = new int[2*NumPumCh];  

// double Fitness[Population],GainLK[NumSigCh],GainGK[NumSigCh];
 double Max;
 int MaxNum;
 double ExTemp, MaxAveFitness=100, AveFitness;
 int PuR,PuL;
// int RndPerm[Population],
 int RndTemp;
// int MutationSign[Population*NumPumCh*2];
 int NumMutation=0;
 int PlaceMutation;
 double g,AveFitnessTh;
 int StopGA, NumItrGA;
 int PlaceCross, NumCross;
 int IntTemp;
// int RndPermPlace[2*NumPumCh];
 double MaxGgkGlk,MinGgkGlk; 
 double MaxIj; 	 
	int ChangeSign;
 int RestSelectionFixedNum;
 double Fit;
 int SameFitnessCount=0;

// Population generation
// for(int i=0;i<Population;i++){
//	for(int j=0;j<NumPumCh;j++){
//		Vector[ 2*i*NumPumCh + 2*j ] = (100*RNG->GetRanNum()+1410)*1E-9;
//		Vector[ 2*i*NumPumCh + 2*j + 1 ] = 30000*RNG->GetRanNum()+100;
//	}
// }		
//	 
//fitness calculation
// cout << "\a";
// cout << "AveFitness=?" ;
// cin >> AveFitnessTh ;
 AveFitnessTh = 10; 
// AveFitnessTh = 0.001;
 NumItrGA=0;
 StopGA=0;
 RestSelectionFixedNum = Population - SelectionFixedNum/100;

 for(int S=0;S<NumSigCh;S++){
	IntgSigPower[S]=0;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	switch(TypeIntergration){
       			case 1:   // Left Boundary
         			//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         		break;
     
       			case 2:    // Forward estimate Trape Rule
					for(int i=1;i<NumZSteps;i++){
						IntgSigPower[NSig] +=  (oSignalPower[NSig*NumZSteps+i] + oSignalPower[NSig*NumZSteps+i - 1])/2; 
         			}
				break;
     
       			case 3:     //  Forward Distance Average
					for(int i=1;i<NumZSteps;i++){
    		     		g = oSignalPower[NSig*NumZSteps+i] / oSignalPower[NSig*NumZSteps+i-1]; 
	       	  	 		IntgSigPower[NSig] += oSignalPower[NSig*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   		}
				break;
		 
	    	    default:
        	 		cerr << "With type Intergration must be 1 or ?" << endl 
        	     	<< "Aborting now! " << endl;
         			exit(1);
    } //end switch type
	
 }
// 	  cout << "run to here4" << endl;
 for(int NSig=0;NSig<NumSigCh;NSig++) {
	IntgSigPower[NSig] *= DeltaZConst ;	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 }
  for(int NSig=0;NSig<NumSigCh;NSig++) {
	oIntgSigPower[NSig] = IntgSigPower[NSig];	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 } 
 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
	GainLK[NSig] = 0;
 } 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
 	for(int S=0;S<NumSigCh;S++){
//		cout << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S]) << " * " << IntgSigPower[S]
//				<< " = " << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S]<< endl;
   		GainLK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S];
//		cout << "  GainLK=" << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S] << endl;
	}
	GainLK[NSig] =  GainLK[NSig] - GetAttenuationFiber(oSignalWaveLth[NSig]) * AmpLength;

//	cout << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
//	*LogFile << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
 }
// cout << endl;
// *LogFile << endl;
//		Vector( 2*i*NumPumCh + 2*j ) = (100*RNG->GetRanNum()+1410)*1E-9;
	//	Vector( 2*i*NumPumCh + 2*j + 1 ) = (490*RNG->GetRanNum()+10)*1E-3;

do{
 NumItrGA++;

// ####################
// fitness calculatation
/* for(int Pu=0;Pu<Population;Pu++){
 	Fitness[Pu]=0;
 }
 for(int Pu=0;Pu<Population;Pu++){
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
		Fitness[Pu] += sq( exp(GainLK[NSig]) * exp(GainGK[NSig] ) - 1);
	}
  	Fitness[Pu] = 1/sqrt(Fitness[Pu]);
//	cout <<"  Fitness=" << Fitness[Pu] << endl;
 }

  for(int Pu=0; Pu<Population;Pu++){

		MaxIj=0;
		for(int NPum=0;NPum<NumPumCh;NPum++){
			if(Vector[ 2*Pu*NumPumCh + 2*NPum + 1] > MaxIj)
				MaxIj =	Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
		}
		for(int NPum=0;NPum<NumPumCh;NPum++){
			Vector[ 2*Pu*NumPumCh + 2*NPum + 1] /= MaxIj;
		}
  } 
   */
 
// (3) fitness defination

  for(int Pu=0;Pu<Population;Pu++){
	for(int S=0;S<NumSigCh;S++){	
 		GainGK[S] =0;
 	}
 	for(int NSig=0;NSig<NumSigCh;NSig++){	

//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << " Gain =" << GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ]) ;
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				//	MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				    MaxGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
				//	MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
					MinGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
				}
			}
		}
	}
//    cout << MaxGgkGlk << "  " << MinGgkGlk << "  ";
//	if(MaxGgkGlk>1e9) {cout
	Fitness[Pu] = MaxGgkGlk - MinGgkGlk; 
//	cout << "  fitness= " << Fitness[Pu] ;
	OffsetMaxMin[Pu] = (MaxGgkGlk + MinGgkGlk) / 2 ;
	Fitness[Pu] /= OffsetMaxMin[Pu];
//	cout << "offset[" << Pu << "]=" << OffsetMaxMin[Pu] << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] /= OffsetMaxMin[Pu];
//		cout << "  "  << Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
	}	

//  	Fitness[Pu] = exp(-1*(MaxGgkGlk - MinGgkGlk));
	
		//	cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << endl;

 }
 
// cout << "run to here5" << endl;
//selection
 for(int Pu=Population-1;Pu>=1;Pu--){
 	Max=Fitness[0];
 	MaxNum = 0;
 	for(int SubPu=1;SubPu<=Pu;SubPu++){
//		cout << Max << " <>? " << Fitness[SubPu] << "  " ;
 		if(Max>Fitness[SubPu]){
			Max = Fitness[SubPu];
			MaxNum = SubPu;
//			cout << "  MaxNum= " << MaxNum << endl;
		}
 	}
//	cout << "run to here5.2" << endl;
 	//exchange
	for(int j=0;j<NumPumCh;j++){
//		cout << "maxnum=" << MaxNum << endl;
		ExTemp = Vector[ 2*MaxNum*NumPumCh + 2*j ];
		Vector[ 2*MaxNum*NumPumCh + 2*j ] = Vector[ 2*Pu*NumPumCh + 2*j ];
		Vector[ 2*Pu*NumPumCh + 2*j ] = ExTemp;
		ExTemp = Vector[ 2*MaxNum*NumPumCh + 2*j + 1];
		Vector[ 2*MaxNum*NumPumCh + 2*j + 1] = Vector[ 2*Pu*NumPumCh + 2*j + 1];
		Vector[ 2*Pu*NumPumCh + 2*j + 1 ] = ExTemp;
//		cout << "run to here5.3" << endl; 
 	}
//	cout << "run to here5.4" << endl; 
	ExTemp = Fitness[MaxNum];
	Fitness[MaxNum] = Fitness[Pu];
	Fitness[Pu] = ExTemp;
	ExTemp = OffsetMaxMin[MaxNum];
	OffsetMaxMin[MaxNum] = OffsetMaxMin[Pu];
	OffsetMaxMin[Pu] = ExTemp;	
		
 }
// cout << "run to here5.5" << endl; 
 
// if(AveFitness > 10 * Fitness[Population-1]){ MutationRate = 0.01;}
// 	else{ MutationRate = 0.05;}
 AveFitness=0;
 for(int i=0;i<Population;i++){
  	AveFitness += Fitness[i];

//    cout << Fitness[i] << "  ";
 }
 
// cin >> AveFitnessTh ; 

 AveFitness /= (double)Population;
// if(MaxAveFitness>AveFitness){
 	MaxAveFitness = AveFitness;
	cout << "#" << NumItrGA <<  "  AveFitness=" << AveFitness << "  Best Fitness = " << Fitness[Population-1] 
		<< "  MutationRate = " << MutationRate << " SameFitnessCount=" << SameFitnessCount << endl;
//		cout << " AveFitnessTh=" << AveFitnessTh << "  Fitness[Popul-1]=" << Fitness[Population-1] << endl;
 //}
/* PuR=Population-1;
 PuL=0;
// cout << "run to here5.6" << endl;
 do{
 	for(int i=0;i<(int)((Fitness[PuR]/AveFitness)-1);i++){	
//	cout << "NumChild[" << PuR << "]=" << int(Fitness[PuR]/AveFitness) << endl;
//	cout << "run to here 5.7" << endl;
 		for(int j=0;j<NumPumCh;j++){
 			Vector[ 2*PuL*NumPumCh + 2*j ] = Vector[ 2*PuR*NumPumCh + 2*j ];
			Vector[ 2*PuL*NumPumCh + 2*j + 1] = Vector[ 2*PuR*NumPumCh + 2*j + 1];
		}
		PuL++;
//		cout << "PuL=" << PuL  << "  PuR=" << PuR << "  i=" << i << endl;
//		cout << "run to here 5.75" << endl;		
		if(PuL>=PuR){i=(int)(Fitness[PuR]/AveFitness);}
	}
	PuR--;
//	cout << "run to here 5.8" << endl;
 }while(PuL<PuR&&(int(Fitness[PuR]/AveFitness)-1>0));
*/



/* for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	cout <<"  Fitness[" << Pu << "]=" << Fitness[Pu]  << " Offset=" << OffsetMaxMin[Pu]  
			<< "  /=" << Fitness[Pu]/OffsetMaxMin[Pu]; // << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] = Vector[ 2*Pu*NumPumCh + 2*NPum + 1] / OffsetMaxMin[Pu];
		cout << "  "  << Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
	}
	cout << endl;
  }	

 for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	for(int S=0;S<NumSigCh;S++){	
 		GainGK[S] =0;
 	}
 	for(int NSig=0;NSig<NumSigCh;NSig++){	

//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],Vector[ 2*Pu*NumPumCh + 2*P ])
				* Vector[ 2*Pu*NumPumCh + 2*P + 1 ];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
					MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				}
			}
		}
	}
	Fitness[Pu] = MaxGgkGlk - MinGgkGlk; 
	OffsetMaxMin[Pu] = (MaxGgkGlk + MinGgkGlk) / 2 ;
//  	Fitness[Pu] = exp(-1*(MaxGgkGlk - MinGgkGlk));
//	Fitness[Pu] = 1/(MaxGgkGlk - MinGgkGlk);
		//	cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << endl;
	cout <<"  Fitness[" << Pu << "]=" << Fitness[Pu] << endl;
 }	
	exit(1);
*/
	
	if(Fitness[Population-1] > AveFitnessTh - 1e-6){
		SameFitnessCount++;
//		AveFitnessTh = Fitness[Population-1];
	}else{
		SameFitnessCount=0;
		AveFitnessTh = Fitness[Population-1];
	}

// if(Fitness[Population-1]< AveFitnessTh && NumItrGA>=2){
  if(SameFitnessCount==5){
//	cout << "StopGA= " << StopGA;
	cout << AveFitnessTh << " = " << AveFitnessTh << endl;
//	cout << AveFitness << " > " << AveFitnessTh << " NUmItrGA=" << NumItrGA << endl;
 	StopGA=1;
 }else{ 	
	
 for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	
//	for(int NPum=0;NPum<NumPumCh;NPum++){
//		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] = Vector[ 2*Pu*NumPumCh + 2*NPum + 1] / OffsetMaxMin[Pu];
//	}
//	MaxIj=0;
//	for(int NPum=0;NPum<NumPumCh;NPum++){
//		if(Vector[ 2*Pu*NumPumCh + 2*NPum + 1] > MaxIj)
//			MaxIj =	Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
//	}
//	for(int NPum=0;NPum<NumPumCh;NPum++){
//		Vector[ 2*Pu*NumPumCh + 2*NPum + 1] /= MaxIj;
//	}
//	OffsetMaxMin[Pu] /= MaxIj/100;

	for(int NPum=0;NPum<NumPumCh;NPum++){
		ChangeSign = 0;
 	 	for(int P=0;P<(NumPumCh-NPum)-1;P++){
	 		if(Vector[ 2*Pu*NumPumCh + 2*P ] > Vector[ 2*Pu*NumPumCh + 2*(P+1) ]){
			 	ChangeSign = 1;
				ExTemp= Vector[ 2*Pu*NumPumCh + 2*P ];
				Vector[ 2*Pu*NumPumCh + 2*P ] = Vector[ 2*Pu*NumPumCh + 2*(P+1) ];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) ] = ExTemp;
				ExTemp = Vector[ 2*Pu*NumPumCh + 2*P + 1];
				Vector[ 2*Pu*NumPumCh + 2*P + 1] = Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1] = ExTemp;
			}
	 	}
		if(ChangeSign == 0)	{NPum = NumPumCh;}
   	}
	
 }

 for(int ii=0;ii<Population/SelectionFixedNum-1;ii++){
 	for(int i=0;i<SelectionFixedNum;i++){
		for(int j=0;j<NumPumCh;j++){
			Vector[ 2*(i+ii*SelectionFixedNum)*NumPumCh + 2*j ] = Vector[ 2*(i+Population-SelectionFixedNum)*NumPumCh + 2*j ];
			Vector[ 2*(i+ii*SelectionFixedNum)*NumPumCh + 2*j + 1] = Vector[ 2*(i+Population-SelectionFixedNum)*NumPumCh + 2*j + 1];
		}
	}
 }

//cout << "  NumItrGA=" << NumItrGA ;

 //Random Permutation
// cout << "run to here6" << endl;

 	 for (int Pu=0;Pu<RestSelectionFixedNum ;Pu++){
 		RndPerm[Pu]=Pu;
	 }
	 for(int Pu=0;Pu<RestSelectionFixedNum;Pu++){
		RndTemp = (int)((RestSelectionFixedNum-Pu)*RNG->GetRanNum());
		if(RndTemp==(RestSelectionFixedNum-Pu)){RndTemp=RestSelectionFixedNum-Pu-1;}
 	    IntTemp = RndPerm[RndTemp];
		for(int i=RndTemp;i<RestSelectionFixedNum-Pu-1;i++){
			RndPerm[i] = RndPerm[i + 1];
		}
 		RndPerm[RestSelectionFixedNum - Pu - 1] = IntTemp;
//		cout << IntTemp <<  "  " ;
	 }	
//crossover
//	cout << "Vector[Pu]before crossover" << endl;
// 	for(int i=0;i<Population;i++){
//		for(int j=0;j<NumPumCh;j++){
	//	   cout << "i,j" <<i << "  " << j;
		//	cout << Vector[ 2*i*NumPumCh + 2*j ] << "  " << Vector[ 2*i*NumPumCh + 2*j + 1 ] << "  ";
//		}
// 	}
//	cout << endl;
//	cout << "run to here6.1" << endl;


	 for(int Pu=0;Pu<RestSelectionFixedNum;Pu += 2){
//		if(RNG->GetRanNum()>CrossRate){
//			cout << "do crossover this time" << endl;
//			PlaceCross = (int)(NumPumCh*RNG->GetRanNum());
			NumCross = (int)((NumPumCh-1)*RNG->GetRanNum()+1);
//			NumCross = (int)(2 * NumPumCh * CrossRate);
	//		if(NumCross==2*NumPumCh){ NumCross = 2*NumPumCh - 1; }
	
	 	 	for (int i=0;i<2*NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}
	 		for (int i=0;i<2*NumPumCh;i++){
				RndTemp = (int)((2 * NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (2*NumPumCh-i)){RndTemp=2*NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
				for(int j=RndTemp;j< 2 * NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[2 * NumPumCh - i - 1] = IntTemp;
	 		}	
			for(int NC=0;NC<NumCross;NC++){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
			}

 	}
	
	
	
	
	/*
	 for(int Pu=0;Pu<Population;Pu += 2){
		for(int Place=0;Place<2*NumPumCh;Place++){
			if(RNG->GetRanNum() > CrossRate){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + Place ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ] = ExTemp;
			}
	   }
 	}

*/	
/*	
	for(int Pu=0;Pu<RestSelectionFixedNum;Pu += 2){
//		if(RNG->GetRanNum()>CrossRate){
//			cout << "do crossover this time" << endl;
//			PlaceCross = (int)(NumPumCh*RNG->GetRanNum());
			NumCross = (int)(NumPumCh*RNG->GetRanNum());
//			NumCross = (int)(NumPumCh * CrossRate);
	//		if(NumCross==2*NumPumCh){ NumCross = 2*NumPumCh - 1; }
	
	 	 	for (int i=0;i<NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}

	 		for (int i=0;i<NumPumCh;i++){
		
				RndTemp = (int)((NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (NumPumCh-i)){RndTemp=NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
	//			if(IntTemp<0||IntTemp>2 * NumPumCh){
	//				cout << endl << RndTemp;
	//				exit(1);
	//			}
				for(int j=RndTemp;j<  NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[NumPumCh - i - 1] = IntTemp;
				
	 		}
	
			for(int NC=0;NC<NumCross;NC++){
		
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + 2*RndPermPlace[NC] + 1];
				Vector[ 2*RndPerm[Pu]*NumPumCh + 2*RndPermPlace[NC] + 1] = Vector[ 2*RndPerm[Pu+1]*NumPumCh +	2*RndPermPlace[NC] + 1];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + 2*RndPermPlace[NC] + 1 ] = ExTemp;
		
			//	PlaceCross++;
			//	if(PlaceCross>=(NumPumCh)){
			//		PlaceCross -= NumPumCh;
			//	}
		 
			}
//		}
 	}
	
*/
// mutation
//	cout << "run to here7" << endl;
	NumMutation=0;
 	for(int S=0;S< RestSelectionFixedNum*NumPumCh*2;S++){
		MutationSign[S]=0;
 	}

 	do{
 		do{
			PlaceMutation=(int)(RestSelectionFixedNum*NumPumCh*2*RNG->GetRanNum());
			if(PlaceMutation==RestSelectionFixedNum*NumPumCh*2){
				PlaceMutation -= 1;
			}
		}while(MutationSign[PlaceMutation]==1);
		MutationSign[PlaceMutation]=1;
		if((PlaceMutation == 2*(int)(PlaceMutation/2))&&(Vector[PlaceMutation]<3E-6)){
			if(AveFitness>0.8){Vector[PlaceMutation] = (120*RNG->GetRanNum()+1400)*1E-9;}
				else{if(RNG->GetRanNum()>0.95) 
							Vector[PlaceMutation] = (120*RNG->GetRanNum()+1400)*1E-9;
					 else
							Vector[PlaceMutation] *= 1+0.1*(2*RNG->GetRanNum()-1);
						if(Vector[PlaceMutation] < 1400E-9) {Vector[PlaceMutation] = 1400E-9;}
						if(Vector[PlaceMutation] > 1520E-9) {Vector[PlaceMutation] = 1520E-9;}
					}
		}else{
//			if((PlaceMutation != 2*(int)(PlaceMutation/2))&&(Vector[PlaceMutation]>0.01)){
			if(PlaceMutation != 2*(int)(PlaceMutation/2)){
				if(AveFitness>0.8){Vector[PlaceMutation] = 30000*RNG->GetRanNum()+10;}
					else{if(RNG->GetRanNum()>0.95)
							Vector[PlaceMutation] = 30000*RNG->GetRanNum()+10;
						 else
							Vector[PlaceMutation] *= 1+0.1*(2*RNG->GetRanNum()-1);
						 if(Vector[PlaceMutation] <= 0.01) {Vector[PlaceMutation] = 0.01;}
						 if(Vector[PlaceMutation] > 300000) {Vector[PlaceMutation] = 300000;}
						}
			}else{
				cout << "wavelength=" << PlaceMutation << "   vector=" << Vector[PlaceMutation] << endl;
       			cerr << "Vector value not propriate" << endl 
           			<< "Aborting now! " << endl;
       			exit(1);		
			}
		}
		NumMutation++;
 	}while(NumMutation<RestSelectionFixedNum*NumPumCh*2*MutationRate);
   }
 }while(StopGA==0);
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	 oPumpWaveLth[NPum] = Vector[ 2*(Population-1)*NumPumCh + 2*NPum ];

	 oPumpIntgTarget[NPum] = Vector[ 2*(Population-1)*NumPumCh + 2*NPum + 1] ;
//	 		OffsetMaxMin[Population-1];

 }
//	cout << "run to here8" << endl;
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	 for(int P=0;P<(NumPumCh-NPum)-1;P++){
	 	if(oPumpWaveLth[P]>oPumpWaveLth[P+1]){
			ExTemp= oPumpWaveLth[P];
			oPumpWaveLth[P] = oPumpWaveLth[P+1];
			oPumpWaveLth[P+1] = ExTemp;
			ExTemp = oPumpIntgTarget[P];
			oPumpIntgTarget[P] = oPumpIntgTarget[P+1];
			oPumpIntgTarget[P+1] = ExTemp;
		}
	} 
 }
 for(int NPum=0;NPum<NumPumCh;NPum++){
// 	cout <<  oPumpWaveLth[NPum] << "  " << oPumpIntgTarget[NPum] << endl;
 }
 cout << endl;
 double FitApprove=0;
 double GainRipple=0;
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P])
				*  oPumpIntgTarget[P];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
		FitApprove  += sq( exp(GainLK[NSig]) * exp(GainGK[NSig] ) - 1);
//		cout << 10*log10( exp(GainLK[NSig]) * exp(GainGK[NSig] )) << "  ";
	//	*LogFile << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
//	cout <<  "Gain["<< NSig << "]=" 
    if (abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])))>GainRipple)
		GainRipple = abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])));
		
//	*LogFile << oSignalWaveLth[NSig] << "   " 
//		<< 10*log10(exp(GainLK[NSig])*exp(GainGK[NSig] )) << "   " 
//		<< (-1) * GainGK[NSig] / GainLK[NSig] 
//		<< endl;

	}
//	*LogFile << endl << "  OffsetMaxMin[Population-1] = " << OffsetMaxMin[Population-1] << endl;
	*LogFile << " Gain Ripple = " << GainRipple << endl;
	
//	cout << endl << "FitApprove=" << FitApprove ;

  	FitApprove = 1/sqrt(FitApprove); 	
//	cout << endl << "FitApprove=" << FitApprove << "  Fitness[]=" << Fitness[Population-1] ;
	
//	Fitness[Pu] = 1/sqrt(Fitness[Pu]);
//	cout <<"  Fitness=" << Fitness[Pu] << endl;
 
 cout << endl;
 for(int i=0;i<Population;i++){
  	AveFitness += Fitness[i];
//    cout <<  Fitness[i] << "  ";
 }
 *LogFile << "AveFitness =" << AveFitness/Population << "   Highest Fitness=" << Fitness[Population-1] << endl;
 cout << endl;
// CalculateFrel();

 delete [] IntgSigPower ;  
 delete [] Fitness ;  
 delete [] OffsetMaxMin;
 delete [] GainLK ;  
 delete [] GainGK ;  
 delete [] RndPerm ;  
 delete [] MutationSign ;  
 delete [] RndPermPlace ;  
// exit(1);

} // end GeneticAlgorithmsVictor


void RamanAmp::GeneticAlgorithms(double * Vector)
{
// double Vector[Population*NumPumCh*2];
 double *IntgSigPower = new double[NumSigCh];  
// double IntgSigPower[NumSigCh];
 double *Fitness = new double[Population];  
 double *OffsetMaxMin = new double [Population];
 double *GainLK = new double[NumSigCh];  
 double *GainGK = new double[NumSigCh];  
 int *RndPerm = new int[Population];  
 int *RndPermPlace = new int[2*NumPumCh];  
 int ChangeSign=0;  

 double Max;
 int MaxNum;
 double ExTemp, MaxAveFitness=100, AveFitness;
 int PuR,PuL;
 int RndTemp;
 int PlaceMutation;
 double g,AveFitnessTh;
 int StopGA, NumItrGA;
 int PlaceCross, NumCross;
 int IntTemp;
// int RndPermPlace[2*NumPumCh];
 double MaxGgkGlk,MinGgkGlk; 
 double MaxIj; 	 
 int RestSelectionFixedNum;
 double Fit;
 int SameFitnessCount=0;


 AveFitnessTh = 10; 
 NumItrGA=0;
 StopGA=0;
 RestSelectionFixedNum = Population - SelectionFixedNum/100;

 for(int S=0;S<NumSigCh;S++){
	IntgSigPower[S]=0;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	switch(TypeIntergration){
       			case 1:   // Left Boundary
         			//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         		break;
     
       			case 2:    // Forward estimate Trape Rule
					for(int i=1;i<NumZSteps;i++){
						IntgSigPower[NSig] +=  (oSignalPower[NSig*NumZSteps+i] + oSignalPower[NSig*NumZSteps+i - 1])/2; 
         			}
				break;
     
       			case 3:     //  Forward Distance Average
					for(int i=1;i<NumZSteps;i++){
    		     		g = oSignalPower[NSig*NumZSteps+i] / oSignalPower[NSig*NumZSteps+i-1]; 
	       	  	 		IntgSigPower[NSig] += oSignalPower[NSig*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   		}
				break;
		 
	    	    default:
        	 		cerr << "With type Intergration must be 1 or ?" << endl 
        	     	<< "Aborting now! " << endl;
         			exit(1);
    } //end switch type
	
 }
// 	  cout << "run to here4" << endl;
 for(int NSig=0;NSig<NumSigCh;NSig++) {
	IntgSigPower[NSig] *= DeltaZConst ;	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
	GainLK[NSig] = 0;
 } 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
 	for(int S=0;S<NumSigCh;S++){
   		GainLK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S];
	}
	GainLK[NSig] =  GainLK[NSig] - GetAttenuationFiber(oSignalWaveLth[NSig]) * AmpLength;

//	cout << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
//	*LogFile << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
 }
// *LogFile << endl;

do{
 NumItrGA++;

// ####################
// fitness calculatation

 
// (3) fitness defination

  for(int Pu=0;Pu<Population;Pu++){
	for(int S=0;S<NumSigCh;S++){	
 		GainGK[S] =0;
 	}
 	for(int NSig=0;NSig<NumSigCh;NSig++){	

//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P])
				* Vector[ Pu*NumPumCh + P ];
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				//	MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				    MaxGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
				//	MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
					MinGgkGlk = -1*GainGK[NSig]/GainLK[NSig];
				}
			}
		}
	}
//    cout << MaxGgkGlk << "  " << MinGgkGlk << "  ";
//	if(MaxGgkGlk>1e9) {cout
	Fitness[Pu] = MaxGgkGlk - MinGgkGlk; 
//	cout << "  fitness= " << Fitness[Pu] << endl;
	OffsetMaxMin[Pu] = (MaxGgkGlk + MinGgkGlk) / 2 ;
	Fitness[Pu] /= OffsetMaxMin[Pu];
//	cout << "offset[" << Pu << "]=" << OffsetMaxMin[Pu] << endl;
	for(int NPum=0;NPum<NumPumCh;NPum++){
		Vector[ Pu*NumPumCh + NPum ] /= OffsetMaxMin[Pu];
//		cout << "  "  << Vector[ 2*Pu*NumPumCh + 2*NPum + 1];
	}	

//  	Fitness[Pu] = exp(-1*(MaxGgkGlk - MinGgkGlk));
	
		//	cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << endl;

 }
 
// cout << "run to here5" << endl;
//selection
 for(int Pu=Population-1;Pu>=1;Pu--){
 	Max=Fitness[0];
 	MaxNum = 0;
 	for(int SubPu=1;SubPu<=Pu;SubPu++){
//		cout << Max << " <>? " << Fitness[SubPu] << "  " ;
 		if(Max>Fitness[SubPu]){
			Max = Fitness[SubPu];
			MaxNum = SubPu;
//			cout << "  MaxNum= " << MaxNum << endl;
		}
 	}
//	cout << "run to here5.2" << endl;
 	//exchange
	for(int j=0;j<NumPumCh;j++){
//		cout << "maxnum=" << MaxNum << endl;
//		ExTemp = Vector[ 2*MaxNum*NumPumCh + 2*j ];
//		Vector[ 2*MaxNum*NumPumCh + 2*j ] = Vector[ 2*Pu*NumPumCh + 2*j ];
//		Vector[ 2*Pu*NumPumCh + 2*j ] = ExTemp;
		ExTemp = Vector[ MaxNum*NumPumCh + j ];
		Vector[ MaxNum*NumPumCh + j ] = Vector[ Pu*NumPumCh + j];
		Vector[ Pu*NumPumCh + j ] = ExTemp;
//		cout << "run to here5.3" << endl; 
 	}
//	cout << "run to here5.4" << endl; 
	ExTemp = Fitness[MaxNum];
	Fitness[MaxNum] = Fitness[Pu];
	Fitness[Pu] = ExTemp;
	ExTemp = OffsetMaxMin[MaxNum];
	OffsetMaxMin[MaxNum] = OffsetMaxMin[Pu];
	OffsetMaxMin[Pu] = ExTemp;	
		
 }
// cout << "run to here5.5" << endl; 
 
// if(AveFitness > 10 * Fitness[Population-1]){ MutationRate = 0.01;}
// 	else{ MutationRate = 0.05;}
 AveFitness=0;
 for(int i=0;i<Population;i++){
  	AveFitness += Fitness[i];

//    cout << Fitness[i] << "  ";
 }
 
// cin >> AveFitnessTh ; 

 AveFitness /= (double)Population;
// if(MaxAveFitness>AveFitness){
 	MaxAveFitness = AveFitness;
cout << "#" << NumItrGA <<  "  AveFitness=" << AveFitness << "  Best Fitness = " << Fitness[Population-1] 
		<< "  MutationRate = " << MutationRate << " SameFitnessCount=" << SameFitnessCount << endl;
		cout << " AveFitnessTh=" << AveFitnessTh << "  Fitness[Popul-1]=" << Fitness[Population-1] << endl;
 //}



	if(Fitness[Population-1] > AveFitnessTh - 1e-7){
		SameFitnessCount++;
	}else{
		SameFitnessCount=0;
		AveFitnessTh = Fitness[Population-1];
	}

// if(Fitness[Population-1]< AveFitnessTh && NumItrGA>=2){
  if(SameFitnessCount==20){
//	cout << AveFitnessTh << " = " << AveFitnessTh << endl;
 	StopGA=1;
 }else{ 	
	
/* for(int Pu=Population-1; Pu>Population-SelectionFixedNum-1;Pu--){
	for(int NPum=0;NPum<NumPumCh;NPum++){
		ChangeSign = 0;
 	 	for(int P=0;P<(NumPumCh-NPum)-1;P++){
	 		if(Vector[ 2*Pu*NumPumCh + 2*P ] > Vector[ 2*Pu*NumPumCh + 2*(P+1) ]){
			 	ChangeSign = 1;
				ExTemp= Vector[ 2*Pu*NumPumCh + 2*P ];
				Vector[ 2*Pu*NumPumCh + 2*P ] = Vector[ 2*Pu*NumPumCh + 2*(P+1) ];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) ] = ExTemp;
				ExTemp = Vector[ 2*Pu*NumPumCh + 2*P + 1];
				Vector[ 2*Pu*NumPumCh + 2*P + 1] = Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1];
				Vector[ 2*Pu*NumPumCh + 2*(P+1) + 1] = ExTemp;
			}
	 	}
		if(ChangeSign == 0)	{NPum = NumPumCh;}
   	}
 }
*/
 for(int ii=0;ii<Population/SelectionFixedNum-1;ii++){
 	for(int i=0;i<SelectionFixedNum;i++){
		for(int j=0;j<NumPumCh;j++){
			Vector[ (i+ii*SelectionFixedNum)*NumPumCh + j ] = Vector[ (i+Population-SelectionFixedNum)*NumPumCh + j];
		}
	}
 }

//cout << "  NumItrGA=" << NumItrGA ;

 //Random Permutation
// cout << "run to here6" << endl;

 	 for (int Pu=0;Pu<RestSelectionFixedNum ;Pu++){
 		RndPerm[Pu]=Pu;
	 }
	 for(int Pu=0;Pu<RestSelectionFixedNum;Pu++){
		RndTemp = (int)((RestSelectionFixedNum-Pu)*RNG->GetRanNum());
		if(RndTemp==(RestSelectionFixedNum-Pu)){RndTemp=RestSelectionFixedNum-Pu-1;}
 	    IntTemp = RndPerm[RndTemp];
		for(int i=RndTemp;i<RestSelectionFixedNum-Pu-1;i++){
			RndPerm[i] = RndPerm[i + 1];
		}
 		RndPerm[RestSelectionFixedNum - Pu - 1] = IntTemp;
//		cout << IntTemp <<  "  " ;
	 }	
//crossover
//	cout << "Vector[Pu]before crossover" << endl;
// 	for(int i=0;i<Population;i++){
//		for(int j=0;j<NumPumCh;j++){
	//	   cout << "i,j" <<i << "  " << j;
		//	cout << Vector[ 2*i*NumPumCh + 2*j ] << "  " << Vector[ 2*i*NumPumCh + 2*j + 1 ] << "  ";
//		}
// 	}
//	cout << endl;
//	cout << "run to here6.1" << endl;

/*
	 for(int Pu=0;Pu<Population;Pu += 2){
			NumCross = (int)((NumPumCh-1)*RNG->GetRanNum()+1);
	 	 	for (int i=0;i<2*NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}
	 		for (int i=0;i<2*NumPumCh;i++){
				RndTemp = (int)((2 * NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (2*NumPumCh-i)){RndTemp=2*NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
				for(int j=RndTemp;j< 2 * NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[2 * NumPumCh - i - 1] = IntTemp;
	 		}	
			for(int NC=0;NC<NumCross;NC++){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
			}

 	}
	
	*/
	
	
	/*
	 for(int Pu=0;Pu<Population;Pu += 2){
		for(int Place=0;Place<2*NumPumCh;Place++){
			if(RNG->GetRanNum() > CrossRate){
				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu]*NumPumCh + Place ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ];
				Vector[ 2*RndPerm[Pu+1]*NumPumCh + Place ] = ExTemp;
			}
	   }
 	}

*/	

	for(int Pu=0;Pu<RestSelectionFixedNum;Pu += 2){
			NumCross = (int)(NumPumCh*RNG->GetRanNum());
	 	 	for (int i=0;i<NumPumCh;i++){
 				RndPermPlace[i]=i;
	 		}

	 		for (int i=0;i<NumPumCh;i++){
		
				RndTemp = (int)((NumPumCh-i)*RNG->GetRanNum());
				if(RndTemp == (NumPumCh-i)){RndTemp=NumPumCh-i-1;}
 	    		IntTemp = RndPermPlace[RndTemp];
	//			if(IntTemp<0||IntTemp>2 * NumPumCh){
	//				cout << endl << RndTemp;
	//				exit(1);
	//			}
				for(int j=RndTemp;j<  NumPumCh - i - 1;j++){
					RndPermPlace[j] = RndPermPlace[j + 1];
				}
 				RndPermPlace[NumPumCh - i - 1] = IntTemp;
				
	 		}
	
			for(int NC=0;NC<NumCross;NC++){
		
//				ExTemp = Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
//				Vector[ 2*RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ] = Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
//				Vector[ 2*RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
				ExTemp = Vector[ RndPerm[Pu]*NumPumCh + RndPermPlace[NC] ];
				Vector[ RndPerm[ Pu ]*NumPumCh + RndPermPlace[NC] ] = Vector[ RndPerm[Pu+1]*NumPumCh + RndPermPlace[NC] ];
				Vector[ RndPerm[ Pu + 1 ]*NumPumCh + RndPermPlace[NC] ] = ExTemp;
		
			//	PlaceCross++;
			//	if(PlaceCross>=(NumPumCh)){
			//		PlaceCross -= NumPumCh;
			//	}
		 
			}

 	}
	

// mutation

	for(int PlaceMutation =0; PlaceMutation < RestSelectionFixedNum * NumPumCh ; PlaceMutation ++){
		if(RNG->GetRanNum() < MutationRate){
			if(AveFitness>0.8){
				Vector[PlaceMutation] = 30000*RNG->GetRanNum()+10;}
			else{
				if(RNG->GetRanNum()>0.95){
					Vector[PlaceMutation] = 30000*RNG->GetRanNum()+10;
				}else{
					Vector[PlaceMutation] *= 1+0.1*(2*RNG->GetRanNum()-1);
				}
				if(Vector[PlaceMutation] < 1e-6) {Vector[PlaceMutation] = 1e-6;}
				if(Vector[PlaceMutation] > 300000) {Vector[PlaceMutation] = 300000;}

			}
		}
	}
/*	for(int PlaceMutation = 0; PlaceMutation < RestSelectionFixedNum ; PlaceMutation++){
		if(RNG->GetRanNum() < MutationRate){
			VectorSpace[PlaceMutation] *= 1+0.1*(2*RNG->GetRanNum()-1);
			if(VectorSpace[PlaceMutation] > (1525e-9 - 1450e-9) * 2 / (NumPumCh-1) ){
				VectorSpace[PlaceMutation] = (1525e-9 - 1450e-9) * 2 / (NumPumCh-1);
			}
			UpdateIndividualWavelength(PlaceMutation, Vector, VectorSpace);	
		}
	}
*/	
  }

  
 }while(StopGA==0);
 for(int NPum=0;NPum<NumPumCh;NPum++){
 	// oPumpWaveLth[NPum] = Vector[ 2*(Population-1)*NumPumCh + 2*NPum ];

	oPumpIntgTarget[NPum] = Vector[ (Population-1)*NumPumCh + NPum ] ;
//	 		OffsetMaxMin[Population-1];
	

 }
//	cout << "run to here8" << endl;

 
// for(int NPum=0;NPum<NumPumCh;NPum++){
// 	cout <<  "Wavelength=" << oPumpWaveLth[NPum] << "  PumpIntg=" << oPumpIntgTarget[NPum] << endl;
// }
// cout << endl;
 double FitApprove=0;
 double GainRipple=0;
 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P])
				*  oPumpIntgTarget[P];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
		FitApprove  += sq( exp(GainLK[NSig]) * exp(GainGK[NSig] ) - 1);
//		cout << 10*log10( exp(GainLK[NSig]) * exp(GainGK[NSig] )) << "  ";
	//	*LogFile << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
//	cout <<  "Gain["<< NSig << "]=" 
    if (abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])))>GainRipple)
		GainRipple = abs(10*log10(exp(GainLK[NSig])*exp(GainGK[NSig])));
		
//	*LogFile << oSignalWaveLth[NSig] << "   " 
//		<< 10*log10(exp(GainLK[NSig])*exp(GainGK[NSig] )) << "   " 
//		<< (-1) * GainGK[NSig] / GainLK[NSig] 
//		<< endl;

	}
//	*LogFile << endl << "  OffsetMaxMin[Population-1] = " << OffsetMaxMin[Population-1] << endl;
//	*LogFile << " GainRipple= " << GainRipple << endl;
//	GlobleRipple = GainRipple;
	cout << "  BestFitness=" << Fitness[Population-1] << "  GainRipple=" << GainRipple << 
			"  G.A.Itr#=" << NumItrGA << endl;
//	cout << endl << "FitApprove=" << FitApprove ;

//  	FitApprove = 1/sqrt(FitApprove); 	
//	cout << endl << "FitApprove=" << FitApprove << "  Fitness[]=" << Fitness[Population-1] ;
	
//	Fitness[Pu] = 1/sqrt(Fitness[Pu]);
//	cout <<"  Fitness=" << Fitness[Pu] << endl;
 
// cout << endl;
// for(int i=0;i<Population;i++){
 // 	AveFitness += Fitness[i];
//    cout <<  Fitness[i] << "  ";
// }
// *LogFile << "AveFitness =" << AveFitness/Population << "   Highest Fitness=" << Fitness[Population-1] << endl;




 delete [] IntgSigPower ;  
 delete [] Fitness ;  
 delete [] OffsetMaxMin;
 delete [] GainLK ;  
 delete [] GainGK ;  
 delete [] RndPerm ;  
 delete [] RndPermPlace ;  
 
} // end GeneticAlgorithms

// ####################################

void RamanAmp::CalculateFrel()
{
 double *GainLK = new double[NumSigCh];  
 double *GainGK = new double[NumSigCh]; 
 double *IntgSigPower = new double[NumSigCh];
 double *IntgPumPower = new double[NumPumCh];
 //double GainGK[NumSigCh],GainLK[NumSigCh];
 //double IntgSigPower[NumSigCh],IntgPumPower[NumPumCh];
 double CalFitness, MaxGgkGlk, MinGgkGlk;
 double g;

 for(int S=0;S<NumSigCh;S++){
	IntgSigPower[S]=0;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){
 	switch(TypeIntergration){
       			case 1:   // Left Boundary
         			//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         		break;
     
       			case 2:    // Forward estimate Trape Rule
					for(int i=1;i<NumZSteps;i++){
						IntgSigPower[NSig] +=  (oSignalPower[NSig*NumZSteps+i] + oSignalPower[NSig*NumZSteps+i - 1])/2; 
         			}
				break;
     
       			case 3:     //  Forward Distance Average
					for(int i=1;i<NumZSteps;i++){
    		     		g = oSignalPower[NSig*NumZSteps+i] / oSignalPower[NSig*NumZSteps+i-1]; 
	       	  	 		IntgSigPower[NSig] += oSignalPower[NSig*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   		}
				break;
		 
	    	    default:
        	 		cerr << "With type Intergration must be 1 or ?" << endl 
        	     	<< "Aborting now! " << endl;
         			exit(1);
    } //end switch type
	
 }
// 	  cout << "run to here4" << endl;
 for(int NSig=0;NSig<NumSigCh;NSig++) {
	IntgSigPower[NSig] *= DeltaZConst ;	
//	cout << "  Intg=" << IntgSigPower[NSig] << endl;
 }
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
	GainLK[NSig] = 0;
 } 
 for(int NSig=0;NSig<NumSigCh;NSig++){ 
 	for(int S=0;S<NumSigCh;S++){
//		cout << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S]) << " * " << IntgSigPower[S]
//				<< " = " << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S]<< endl;
   		GainLK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S];
//		cout << "  GainLK=" << GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S])*IntgSigPower[S] << endl;
	}
	GainLK[NSig] =  GainLK[NSig] - GetAttenuationFiber(oSignalWaveLth[NSig]) * AmpLength;

//	cout << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
//	*LogFile << "  GainLK["<< NSig << "]=" << GainLK[NSig] << "  " ;
 }



		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] = 0;	
  		}
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			switch(TypeIntergration)
     			{
       				case 1:   // Left Boundary
         				//TempSigPower[S]=  oSignalPower[NSig*NumZSteps+WStepI-1]; 
         			break;
     
       				case 2:    // Forward estimate Trape Rule
						for(int i=1;i<NumZSteps;i++){
							IntgPumPower[NPum] +=  (oPumpPower[NPum*NumZSteps+i] + oPumpPower[NPum*NumZSteps+i-1])/2; 
         				}
					break;
     
       				case 3:     //  Forward Distance Average
						for(int i=1;i<NumZSteps;i++){
    		   	  			g = oPumpPower[NPum*NumZSteps+i] / oPumpPower[NPum*NumZSteps+i-1]; 
   	  	 					IntgPumPower[NPum] += oPumpPower[NPum*NumZSteps+i-1] *( g - 1 ) /  log(g) ;
      		   			}
					break;
		 
	    	   	 	default:
        	 			cerr << "With type Intergration must be 1 or ?" << endl 
        	     		<< "Aborting now! " << endl;
         				exit(1);
      			} //end switch type
			
		}
		
		for(int NPum=0;NPum<NumPumCh;NPum++) {
			IntgPumPower[NPum] *= DeltaZConst ;	
 // 			cout << "Intg after shooting" << IntgPumPower[NPum] << "  ->  "<< oPumpIntgTargetAlpha[NPum] 
//				<< "  Power" << oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl;
		}




 	for(int NSig=0;NSig<NumSigCh;NSig++){	
		for(int S=0;S<NumSigCh;S++){	
 			GainGK[S] =0;
 		}
//cout << "run to here5.1" << endl;
		for(int P=0;P<NumPumCh;P++){
			GainGK[NSig] += GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P])*IntgPumPower[P];
//			cout << "  GainGK["<< NSig << "]=" << GainGK[NSig] ;
		}
//		cout << "run to here5.3" << endl;
		if(NSig==0){
			MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
//			cout << " MaxGgkGlk=" << MaxGgkGlk  << "  MinGgkGlk=" << MinGgkGlk << "  GLK=" << GainLK[NSig] << endl;	
	//		exit(1);
		}else{
			if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) > MaxGgkGlk){
				MaxGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
			}else{
				if(log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig])) < MinGgkGlk){
					MinGgkGlk = log10(exp(GainGK[NSig]))/log10(1/exp(GainLK[NSig]));
				}
			}
		}
	}
	CalFitness = MaxGgkGlk - MinGgkGlk;
	cout << "CalFitness= " << CalFitness << endl;
    *LogFile << "CalFitness= " << CalFitness << endl;

 delete [] GainLK ;  
 delete [] GainGK ; 
 delete [] IntgSigPower ;
 delete [] IntgPumPower ;


}
// end of CalculateFrel

 

// ####################################################
// ############### end of Genetic Algorithms  ##########################
// ####################################################

// #################################################
// Transient code, it's still not working yet
// if you want to use Transient code, please talk to me
// by Jonathan Hu
void RamanAmp::RamanTransient()
{   
  oPumpWaveLth[0] = 1455e-9;
  oPumpPower[0] = 0; //300e-3;
  oSignalWaveLth[0] = 1555e-9;
  oSignalPower[0] = 0;
NumTSteps = 500000; 
 DeltaT = 1e-8;  
 double *oSignalTemp = new double[NumSigCh*NumZSteps];
 double *oPumpTemp = new double[NumPumCh*NumZSteps];
 double *dPsdt = new double[NumSigCh*NumZSteps];
 double *dPpdt = new double[NumPumCh*NumZSteps];
 double *SignalInput =  new double[NumSigCh*NumTSteps];
 double *SI =  new double[NumSigCh*NumTSteps];
 double Distance=0, Time = 0; 
 double temp;
 char   C_String[40];
 int RoundNum=15000; 
 sprintf(C_String,"RamanTransientOut.dat");
 string RamanTransientOutName = JobRaman + C_String;
 ofstream RamanTransientOut(RamanTransientOutName.c_str(),ios::out);
 if (!RamanTransientOut)
   {
   cerr<< "Error:  "  << RamanTransientOutName  << "Could not be created." << endl;
   exit(1);
 } 

 sprintf(C_String,"RamanTransientIn.dat");
 string RamanTransientInName = JobRaman + C_String;
 ofstream RamanTransientIn(RamanTransientInName.c_str(),ios::out);
 if (!RamanTransientIn)
   {
   cerr<< "Error:  "  << RamanTransientInName  << "Could not be created." << endl;
   exit(1);
 } 
 
    for(int StepT =0;StepT < NumTSteps;StepT ++){
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			if(StepT>30000){
				SI[NSig*NumTSteps + StepT] = 0.002;
				SignalInput[NSig*NumTSteps + StepT] = 0.002;
			}else{
				SI[NSig*NumTSteps + StepT] = 0.0;
				SignalInput[NSig*NumTSteps + StepT] = 0.002;
			}
		}
  	}

/*    for(int StepT =0;StepT < NumTSteps;StepT ++){
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			if(StepT>16000 && StepT < NumTSteps-16000){
				for(int ii=0;ii<RoundNum;ii++){
					SignalInput[NSig*NumTSteps + StepT] += SI[NSig*NumTSteps + StepT + ii] *(RoundNum-ii)
						+SI[NSig*NumTSteps + StepT - ii] *(RoundNum-ii);
				}
				SignalInput[NSig*NumTSteps + StepT] /= RoundNum*(RoundNum+1);
//					cout <<  StepT << "  " << SI[NSig*NumTSteps+StepT] << "   " 
//					<< SignalInput[NSig*NumTSteps+StepT] << endl ; 
			}else{
				SignalInput[NSig*NumTSteps + StepT] = SI[NSig*NumTSteps + StepT];					
//				cout <<  StepT << "   " << SI[NSig*NumTSteps+StepT] << "   " 
//					<< SignalInput[NSig*NumTSteps+StepT] << endl ; 
			}
		}
		
  	}
*/	
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
		for(int StepT = 0 ; StepT < NumTSteps; StepT+=1000){
//			RamanTransientIn << StepT*DeltaT << "  " << SI[NSig*NumTSteps+StepT] << "   " 
//					<< SignalInput[NSig*NumTSteps+StepT] << endl ; //oSignalPower[NSig*NumZSteps+StepZ] << "  " ;	
		}
	}
//	RamanTransientIn << endl;
 	
 	for(int NPum=0;NPum<NumPumCh;NPum++) {
   		oPumpPower[NPum*NumZSteps+NumZSteps-1] = oPumpTargetBackup[NPum];
		oPumpPower[NPum*NumZSteps+NumZSteps-1] = 0.0;
		cout << "oPumpPower = " << oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl;
	}    
	for(int NSig=0;NSig<NumSigCh;NSig++){
		for(int ii=0;ii<NumZSteps;ii++){
			oSignalPower[NSig*NumZSteps+ii]=0.0;
		}		
	}
	
  	for(int i=0;i<NumZSteps-1;i++){
		oPumpPower[0*NumZSteps+i] = 0;	
	} 	
	oSignalPower[0*NumZSteps]=0.002;
 	for(int i=1;i<NumZSteps;i++){
		SignalOnlyPropagation(i);
		if(BlowupSign==1)	{
			i=NumZSteps;
		} 
	}
	  
/*	for(int i=NumZSteps-2;i>=0;i--){
		BackwardStep(i);
		if(BlowupSign==1)	{
			i=NumZSteps;
		} 
	}  
*/
//		exit(1);
	*LogFile << " Start Transient data" << endl;
	for(int StepT=0;StepT<NumTSteps;StepT++){		
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			oSignalPower[NSig*NumZSteps] = SignalInput[NSig*NumTSteps + StepT]; 
		}
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				// if(StepZ!=NumZSteps-1)
					oSignalTemp[NSig*NumZSteps+StepZ] = oSignalPower[NSig*NumZSteps+StepZ];	
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				// if(StepZ!=0)
					oPumpTemp[NPum*NumZSteps+StepZ] = oPumpPower[NPum*NumZSteps+StepZ];
//				cout << "  PumpPower " << oPumpPower[NPum*NumZSteps+StepZ] << endl;
			}
		}

		if(StepT == 2*(int)(StepT/2)){
			Mac24FB(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		}else{
			Mac24FB(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		}	
		cout << endl;
		cin >> temp;
		Time += DeltaT;
		if(StepT == 10 * (int)(StepT/10))		{	

//			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
//				for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++)
//					RamanTransientOut << oSignalPower[NSig*NumZSteps+StepZ] << "  " ;	
//			}
//			RamanTransientOut << endl;

			RamanTransientOut << Time << "    " << oSignalPower[0*NumZSteps+NumZSteps-1] << "  " 
				<< oPumpPower[0*NumZSteps]<< endl;
		}
		Distance = 0;
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
 			Distance += DeltaZ(StepZ);
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				cout << Distance << "  " << oSignalPower[0*NumZSteps+StepZ] << "   " << 
					oPumpPower[0*NumZSteps+StepZ]  <<  endl;
			}
 		} 
	}
	
delete [] oSignalTemp ;
delete [] oPumpTemp ;	
delete [] dPsdt ;	
delete [] dPpdt ;	
delete [] SignalInput;
delete [] SI;
RamanTransientOut.close();	
RamanTransientIn.close();	

}

void RamanAmp::Mac24BF(double * oSignalTemp,double * oPumpTemp,double * dPsdt,double * dPpdt)
{
		CalculateBackwarddsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		
		for(int StepZ = 0 ; StepZ < NumZSteps-1; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				oSignalTemp[NSig*NumZSteps+StepZ] += dPsdt[NSig*NumZSteps+StepZ] * DeltaT;	
//				oSignalTemp[NSig*NumZSteps+StepZ] = (oSignalTemp[NSig*NumZSteps+StepZ]
//						+oSignalPower[NSig*NumZSteps+StepZ])/2;
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				oPumpTemp[NPum*NumZSteps+StepZ] += dPpdt[NPum*NumZSteps+StepZ] * DeltaT;
//				oPumpTemp[NPum*NumZSteps+StepZ] = (oPumpTemp[NPum*NumZSteps+StepZ]
//						+oPumpPower[NPum*NumZSteps+StepZ])/2;
			}
		}
//		*LogFile << "here 2" << endl;
		CalculateForwarddsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				if(StepZ!=0)
					oSignalPower[NSig*NumZSteps+StepZ] = ( oSignalTemp[NSig*NumZSteps+StepZ] +
						oSignalPower[NSig*NumZSteps+StepZ] 	+ dPsdt[NSig*NumZSteps+StepZ] * DeltaT )/2;	
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				if(StepZ!=NumZSteps-1){
					oPumpPower[NPum*NumZSteps+StepZ] = ( oPumpTemp[NPum*NumZSteps+StepZ] +
						oPumpPower[NPum*NumZSteps+StepZ] + dPpdt[NPum*NumZSteps+StepZ] * DeltaT ) /2;
//					cout << "  PumpPoweraftercal " << StepZ << "  " << oPumpPower[NPum*NumZSteps+StepZ] << endl;
					}
			}
		} 

// 	exit(1);		

}

void RamanAmp::Mac24FB(double * oSignalTemp,double * oPumpTemp,double * dPsdt,double * dPpdt)
{


/*		Distance = 0;
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
 			Distance += DeltaZ(StepZ);
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				cout << Distance << "  " << oSignalPower[0*NumZSteps+StepZ] << "   " << 
					oPumpPower[0*NumZSteps+StepZ]  <<  endl;
			}
 		} 
*/			
/*		Distance = 0;
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
 			Distance += DeltaZ(StepZ);
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				cout << Distance << " temp " << oSignalTemp[0*NumZSteps+StepZ] << "   " << 
					oPumpTemp[0*NumZSteps+StepZ]  <<  endl;
			}
 		}			
*/
//		*LogFile << NumPumCh << "  " << oPumpWaveLth[0] << "  " << oPumpPower[0] << endl;
//		*LogFile << "here 1" << endl;
		CalculateForwarddsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		
		for(int StepZ = 0 ; StepZ < NumZSteps-1; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				oSignalTemp[NSig*NumZSteps+StepZ] += dPsdt[NSig*NumZSteps+StepZ] * DeltaT;	
//				oSignalTemp[NSig*NumZSteps+StepZ] = (oSignalTemp[NSig*NumZSteps+StepZ]
//						+oSignalPower[NSig*NumZSteps+StepZ])/2;
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				oPumpTemp[NPum*NumZSteps+StepZ] += dPpdt[NPum*NumZSteps+StepZ] * DeltaT;
//				oPumpTemp[NPum*NumZSteps+StepZ] = (oPumpTemp[NPum*NumZSteps+StepZ]
//						+oPumpPower[NPum*NumZSteps+StepZ])/2;
			}
		}
//		*LogFile << "here 2" << endl;
		CalculateBackwarddsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				if(StepZ!=0)
					oSignalPower[NSig*NumZSteps+StepZ] = ( oSignalTemp[NSig*NumZSteps+StepZ] +
						oSignalPower[NSig*NumZSteps+StepZ] 	+ dPsdt[NSig*NumZSteps+StepZ] * DeltaT )/2;	
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				if(StepZ!=NumZSteps-1){
					oPumpPower[NPum*NumZSteps+StepZ] = ( oPumpTemp[NPum*NumZSteps+StepZ] +
						oPumpPower[NPum*NumZSteps+StepZ] + dPpdt[NPum*NumZSteps+StepZ] * DeltaT ) /2;
//					cout << "  PumpPoweraftercal " << StepZ << "  " << oPumpPower[NPum*NumZSteps+StepZ] << endl;
					}
			}
		} 
//		cout << endl;
//		cin >> temp;


}

//  ##########################
void RamanAmp::CalculateForwarddsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt)
{ 

 double *oSignaldudz = new double[NumSigCh*NumZSteps];
 double *oPumpdudz = new double[NumPumCh*NumZSteps];
 double GroupVelocity = LightSpeed/1.5;

 double temp;
		// calculate dudz
 for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
// 	cout << StepZ << "  PumpTemp = " << oPumTemp[0*NumZSteps+StepZ] << endl;
 
 }

/*  for(int StepZ = 1 ; StepZ < NumZSteps-1; StepZ++){
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
		oSignaldudz[NSig*NumZSteps+StepZ]	=	
				(oSignalPower[NSig*NumZSteps+StepZ]-oSignalPower[NSig*NumZSteps+StepZ-1])/DeltaZ(StepZ);
//		cout << " dudzSignal " << StepZ << "  " << oSignaldudz[NSig*NumZSteps+StepZ] << endl;
	}			
	for(int NPum = 0 ; NPum < NumPumCh; NPum++){
			oPumpdudz[NPum*NumZSteps+StepZ]	=	
				(oPumpPower[NPum*NumZSteps+StepZ+1]-oPumpPower[NPum*NumZSteps+StepZ])/DeltaZ(StepZ+1);
//		cout << " dudzPump " << StepZ << "  " << oPumpdudz[NPum*NumZSteps+StepZ] << endl;
	} 
  }
  for(int NSig = 0 ; NSig < NumSigCh; NSig++){
	oSignaldudz[NSig*NumZSteps+NumZSteps-1]	= (oSignalPower[NSig*NumZSteps+NumZSteps-1]
		-oSignalPower[NSig*NumZSteps+NumZSteps-2])/DeltaZ(NumZSteps-1);	
		
	cout << " ############dudzSignal " << NumZSteps-1 << "  " << oSignaldudz[NSig*NumZSteps+NumZSteps-1] << endl;
  
  }		
  for(int NPum = 0 ; NPum < NumPumCh; NPum++){
	oPumpdudz[NPum*NumZSteps] = (oPumpPower[NPum*NumZSteps+1]
		-oPumpPower[NPum*NumZSteps])/DeltaZ(1);
//		cout << " dudzPump " << 0 << "  " << oPumpdudz[NPum*NumZSteps] << endl;
  }
*/
  for(int NPum = 0 ; NPum < NumPumCh; NPum++){
	oPumpdudz[NPum*NumZSteps] = (-3.0 * oPumTemp[NPum*NumZSteps] + 4 * oPumTemp[NPum*NumZSteps+1]
		-oPumTemp[NPum*NumZSteps+2])/DeltaZ(1)/2;
	oPumpdudz[NPum*NumZSteps + 1] =  (oPumTemp[NPum*NumZSteps+2]- oPumTemp[NPum*NumZSteps+1])/DeltaZ(1);
	//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
//				<< "  "  << DeltaZ(1) << endl;
	oPumpdudz[NPum*NumZSteps + NumZSteps - 2] = (oPumTemp[NPum*NumZSteps+ NumZSteps - 1]
		-  oPumTemp[NPum*NumZSteps+ NumZSteps - 2])/DeltaZ(1);//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
  	oPumpdudz[NPum*NumZSteps + NumZSteps -1] =0;
  }
   
  
  for(int StepZ = 2 ; StepZ < NumZSteps-2; StepZ++){
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
//			oSignaldudz[NSig*NumZSteps+StepZ] = (8 * (oSigTemp[NSig*NumZSteps+StepZ+1]-oSigTemp[NSig*NumZSteps+StepZ-1])
//				-(oSigTemp[NSig*NumZSteps+StepZ+2]-oSigTemp[NSig*NumZSteps+StepZ-2]))/DeltaZ(StepZ)/12;
			oSignaldudz[NSig*NumZSteps+StepZ] = (-1. * oSigTemp[NSig*NumZSteps+StepZ+2] 
				+ 8 * oSigTemp[NSig*NumZSteps+StepZ+1]-7.*oSigTemp[NSig*NumZSteps+StepZ])/DeltaZ(StepZ)/6;
		
//		cout << " dudzSignal " << StepZ << "  " << oSignaldudz[NSig*NumZSteps+StepZ] << endl;
	}			
	for(int NPum = 0 ; NPum < NumPumCh; NPum++){
//			oPumpdudz[NPum*NumZSteps+StepZ]	=	
//				(8 * (oPumTemp[NPum*NumZSteps+StepZ+1]-oPumTemp[NPum*NumZSteps+StepZ-1])
//				- (oPumTemp[NPum*NumZSteps+StepZ+2]-oPumTemp[NPum*NumZSteps+StepZ-2]))/DeltaZ(StepZ)/12;
		oPumpdudz[NPum*NumZSteps+StepZ] = (-1. * oPumTemp[NPum*NumZSteps+StepZ+2] 
				+ 8 * oPumTemp[NPum*NumZSteps+StepZ+1]-7.*oPumTemp[NPum*NumZSteps+StepZ])/DeltaZ(StepZ)/6;
//		cout << " dudzPump " << StepZ << "  " << oPumTemp[NPum*NumZSteps+StepZ] << "  " << DeltaZ(StepZ) << endl;
	
	} 
  }
  for(int NSig = 0 ; NSig < NumSigCh; NSig++){
	oSignaldudz[NSig*NumZSteps]=0;
	oSignaldudz[NSig*NumZSteps+1] = (oSigTemp[NSig*NumZSteps+2]-oSigTemp[NSig*NumZSteps+1])/DeltaZ(1);		
		
//		cout << " dudzSignal " << NumZSteps-1 << "  " << oSignaldudz[NSig*NumZSteps+NumZSteps-1] << endl;
	oSignaldudz[NSig*NumZSteps+NumZSteps-2]	= (oSigTemp[NSig*NumZSteps+NumZSteps-1]
		-oSigTemp[NSig*NumZSteps+NumZSteps-2])/DeltaZ(NumZSteps-2);	;		
	oSignaldudz[NSig*NumZSteps+NumZSteps-1]	= -1 * (-3.*oSigTemp[NSig*NumZSteps+NumZSteps-1]
		+4.*oSigTemp[NSig*NumZSteps+NumZSteps-2]-oSigTemp[NSig*NumZSteps+NumZSteps-3])/DeltaZ(NumZSteps-1)/2;
  }		

 
//cin >> temp;
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
      
		cout << " dudz " << StepZ << "  " << oPumpdudz[0*NumZSteps+StepZ] 
				<< "  " << oSignaldudz[0*NumZSteps+StepZ] << "  " << "SignalTemp " << oSigTemp[0*NumZSteps+StepZ] <<  endl;
  }

// calculate dpdt
  double StepGain;
	 
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
	for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
  		if(StepZ!=0){
			StepGain = 0;
  			for(int P=0;P<NumPumCh;P++){
     			StepGain +=
     			GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepZ)*oPumpPower[P*NumZSteps+StepZ];
  			}
	  		for(int S=0;S<NumSigCh;S++){
    			StepGain +=
    			GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepZ)*oSignalPower[S*NumZSteps+StepZ-1]; 
  			}  		
			dPsigdt[NSig*NumZSteps+StepZ] =  (StepGain * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- oSignaldudz[NSig*NumZSteps+StepZ])* GroupVelocity;
			cout << " dPsdtSignal " << StepZ << "  " << StepGain * oSigTemp[NSig*NumZSteps+StepZ-1]
				<< "   " << -1 * GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1]
				<< "   " << -1 * oSignaldudz[NSig*NumZSteps+StepZ] << "  "  << dPsigdt[NSig*NumZSteps+StepZ]/GroupVelocity;		
		}else{
			dPsigdt[NSig*NumZSteps+StepZ] = 0;			
//			cout << " dPsdtSignal " << StepZ << "                " ;		

		} 
 	}
	for(int NPum=0;NPum<NumPumCh;NPum++){
		if(StepZ!=NumZSteps-1){
			StepGain = 0;
			for(int P=0;P<NumPumCh;P++){
       			StepGain +=
       				GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],StepZ+1) * oPumpPower[P*NumZSteps+StepZ+1];
   			}
    		for(int S=0;S<NumSigCh;S++){
       			StepGain +=
       			GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],StepZ+1) * oSignalPower[S*NumZSteps+StepZ]; 
  			}
  			dPpumdt[NPum*NumZSteps+StepZ] = (StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
				- GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
				+ oPumpdudz[NPum*NumZSteps+StepZ]) * GroupVelocity;
			cout << "  dPpdtPump" << StepZ << "  " << StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
				<< "   " <<  -1 * GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
				<< "   " <<   oPumpdudz[NPum*NumZSteps+StepZ]<< "   " << dPpumdt[NPum*NumZSteps+StepZ]/GroupVelocity<< endl;
    	}else{
			dPpumdt[NPum*NumZSteps+StepZ] = 0;			
//			cout << "  dPpdtPump" << StepZ << "  							" << endl;
   
		}
	}
  }  
 cin >> temp;

delete [] oSignaldudz ;
delete [] oPumpdudz ;
		
}

void RamanAmp::CalculateBackwarddsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt)
{ 

 double *oSignaldudz = new double[NumSigCh*NumZSteps];
 double *oPumpdudz = new double[NumPumCh*NumZSteps];
 double GroupVelocity = LightSpeed/1.5;

 double temp;
		// calculate dudz
 for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
// 	cout << StepZ << "  PumpTemp = " << oPumTemp[0*NumZSteps+StepZ] << endl;
 
 }

/*  for(int StepZ = 1 ; StepZ < NumZSteps-1; StepZ++){
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
		oSignaldudz[NSig*NumZSteps+StepZ]	=	
				(oSignalPower[NSig*NumZSteps+StepZ]-oSignalPower[NSig*NumZSteps+StepZ-1])/DeltaZ(StepZ);
//		cout << " dudzSignal " << StepZ << "  " << oSignaldudz[NSig*NumZSteps+StepZ] << endl;
	}			
	for(int NPum = 0 ; NPum < NumPumCh; NPum++){
			oPumpdudz[NPum*NumZSteps+StepZ]	=	
				(oPumpPower[NPum*NumZSteps+StepZ+1]-oPumpPower[NPum*NumZSteps+StepZ])/DeltaZ(StepZ+1);
//		cout << " dudzPump " << StepZ << "  " << oPumpdudz[NPum*NumZSteps+StepZ] << endl;
	} 
  }
  for(int NSig = 0 ; NSig < NumSigCh; NSig++){
	oSignaldudz[NSig*NumZSteps+NumZSteps-1]	= (oSignalPower[NSig*NumZSteps+NumZSteps-1]
		-oSignalPower[NSig*NumZSteps+NumZSteps-2])/DeltaZ(NumZSteps-1);	
		
	cout << " ############dudzSignal " << NumZSteps-1 << "  " << oSignaldudz[NSig*NumZSteps+NumZSteps-1] << endl;
  
  }		
  for(int NPum = 0 ; NPum < NumPumCh; NPum++){
	oPumpdudz[NPum*NumZSteps] = (oPumpPower[NPum*NumZSteps+1]
		-oPumpPower[NPum*NumZSteps])/DeltaZ(1);
//		cout << " dudzPump " << 0 << "  " << oPumpdudz[NPum*NumZSteps] << endl;
  }
*/
  for(int NPum = 0 ; NPum < NumPumCh; NPum++){
	oPumpdudz[NPum*NumZSteps] = (-3.0 * oPumTemp[NPum*NumZSteps] + 4 * oPumTemp[NPum*NumZSteps+1]
		-oPumTemp[NPum*NumZSteps+2])/DeltaZ(1)/2;
	oPumpdudz[NPum*NumZSteps + 1] = (oPumTemp[NPum*NumZSteps+1]- oPumTemp[NPum*NumZSteps])/DeltaZ(1);
	//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
//				<< "  "  << DeltaZ(1) << endl;
	oPumpdudz[NPum*NumZSteps + NumZSteps - 2] =( oPumTemp[NPum*NumZSteps+ NumZSteps - 2]
		- oPumTemp[NPum*NumZSteps+ NumZSteps - 3])/DeltaZ(1);//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
  	oPumpdudz[NPum*NumZSteps + NumZSteps -1] =0;
  }
   
  
  for(int StepZ = 2 ; StepZ < NumZSteps-2; StepZ++){
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			oSignaldudz[NSig*NumZSteps+StepZ] = (7* (oSigTemp[NSig*NumZSteps+StepZ]-8.*oSigTemp[NSig*NumZSteps+StepZ-1])
				+oSigTemp[NSig*NumZSteps+StepZ-2])/DeltaZ(StepZ)/6;
/*			oSignaldudz[NSig*NumZSteps+StepZ]	=	
				((oSigTemp[NSig*NumZSteps+StepZ+1]-oSigTemp[NSig*NumZSteps+StepZ])/DeltaZ(StepZ+1)
				+(oSigTemp[NSig*NumZSteps+StepZ]-oSigTemp[NSig*NumZSteps+StepZ-1])/DeltaZ(StepZ))/2;
		
*/		
//		cout << " dudzSignal " << StepZ << "  " << oSignaldudz[NSig*NumZSteps+StepZ] << endl;
	}			
	for(int NPum = 0 ; NPum < NumPumCh; NPum++){
			oPumpdudz[NPum*NumZSteps+StepZ]	=	(7 * (oPumTemp[NPum*NumZSteps+StepZ]-8.*oPumTemp[NPum*NumZSteps+StepZ-1])
				+ oPumTemp[NPum*NumZSteps+StepZ-2])/DeltaZ(StepZ)/6;

/*			oPumpdudz[NPum*NumZSteps+StepZ]	=	
				((oPumTemp[NPum*NumZSteps+StepZ+1]-oPumTemp[NPum*NumZSteps+StepZ])/DeltaZ(StepZ+1)
				+(oPumTemp[NPum*NumZSteps+StepZ]-oPumTemp[NPum*NumZSteps+StepZ-1])/DeltaZ(StepZ))/2;
*/
//		cout << " dudzPump " << StepZ << "  " << oPumTemp[NPum*NumZSteps+StepZ] << "  " << DeltaZ(StepZ) << endl;
	
	} 
  }
  for(int NSig = 0 ; NSig < NumSigCh; NSig++){
	oSignaldudz[NSig*NumZSteps]=0;
	oSignaldudz[NSig*NumZSteps+1] = (oSigTemp[NSig*NumZSteps+1]-oSigTemp[NSig*NumZSteps])/DeltaZ(1);		
	oSignaldudz[NSig*NumZSteps+NumZSteps-1]	= -1 * (-3.*oSigTemp[NSig*NumZSteps+NumZSteps-1]
		+4.*oSigTemp[NSig*NumZSteps+NumZSteps-2]-oSigTemp[NSig*NumZSteps+NumZSteps-3])/DeltaZ(NumZSteps-1)/2;		
//		cout << " dudzSignal " << NumZSteps-1 << "  " << oSignaldudz[NSig*NumZSteps+NumZSteps-1] << endl;
	oSignaldudz[NSig*NumZSteps+NumZSteps-2]	= (oSigTemp[NSig*NumZSteps+NumZSteps-2]
		-oSigTemp[NSig*NumZSteps+NumZSteps-3])/DeltaZ(NumZSteps-2);	;		

  }		

 
//cin >> temp;
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
//		cout << " dudzPump " << StepZ << "  " << oPumpdudz[0*NumZSteps+StepZ] 
//				<< "  " << oSignaldudz[0*NumZSteps+StepZ] << endl;
  }

// calculate dpdt
  double StepGain;
	 
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
	for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
  		if(StepZ!=0){
			StepGain = 0;
  			for(int P=0;P<NumPumCh;P++){
     			StepGain +=
     			GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepZ)*oPumpPower[P*NumZSteps+StepZ];
  			}
	  		for(int S=0;S<NumSigCh;S++){
    			StepGain +=
    			GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepZ)*oSignalPower[S*NumZSteps+StepZ-1]; 
  			}  		
			dPsigdt[NSig*NumZSteps+StepZ] =  (StepGain * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- oSignaldudz[NSig*NumZSteps+StepZ])* GroupVelocity;
			cout << " dPsdtSignal " << StepZ << "  " << StepGain * oSigTemp[NSig*NumZSteps+StepZ-1]
				<< "   " << -1 * GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1]
				<< "   " << -1 * oSignaldudz[NSig*NumZSteps+StepZ] << "  "  << dPsigdt[NSig*NumZSteps+StepZ]/GroupVelocity ;		
		}else{
			dPsigdt[NSig*NumZSteps+StepZ] = 0;			
//			cout << " dPsdtSignal " << StepZ << "                " ;		

		} 
 	}
	for(int NPum=0;NPum<NumPumCh;NPum++){
		if(StepZ!=NumZSteps-1){
			StepGain = 0;
			for(int P=0;P<NumPumCh;P++){
       			StepGain +=
       				GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],StepZ+1) * oPumpPower[P*NumZSteps+StepZ+1];
   			}
    		for(int S=0;S<NumSigCh;S++){
       			StepGain +=
       			GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],StepZ+1) * oSignalPower[S*NumZSteps+StepZ]; 
  			}
  			dPpumdt[NPum*NumZSteps+StepZ] = (StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
				- GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
				+ oPumpdudz[NPum*NumZSteps+StepZ]) * GroupVelocity;
			cout << "  dPpdtPump" << StepZ << "  " << StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
				<< "   " <<  -1 * GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
				<< "   " <<   oPumpdudz[NPum*NumZSteps+StepZ]<< "   " << dPpumdt[NPum*NumZSteps+StepZ]/GroupVelocity<< endl;
    	}else{
			dPpumdt[NPum*NumZSteps+StepZ] = 0;			
//			cout << "  dPpdtPump" << StepZ << "  							" << endl;
   
		}
	}
  }  
// cin >> temp;

delete [] oSignaldudz ;
delete [] oPumpdudz ;
		
}


/*

void RamanAmp::RamanTransient()
{   
  oPumpWaveLth[0] = 1455e-9;
  oPumpPower[0] = 0; //300e-3;
  oSignalWaveLth[0] = 1555e-9;
  oSignalPower[0] = 0;
 int NumTSteps = 500000; 
 double DeltaT = 1e-9;  
 double *oSignalTemp = new double[NumSigCh*NumZSteps];
 double *oPumpTemp = new double[NumPumCh*NumZSteps];
 double *dPsdt = new double[NumSigCh*NumZSteps];
 double *dPpdt = new double[NumPumCh*NumZSteps];
 double *SignalInput =  new double[NumSigCh*NumTSteps];
 double *SI =  new double[NumSigCh*NumTSteps];
 double Distance=0, Time = 0; 
 double temp;
 char   C_String[40];
 int RoundNum=15000; 
 sprintf(C_String,"RamanTransientOut.dat");
 string RamanTransientOutName = JobRaman + C_String;
 ofstream RamanTransientOut(RamanTransientOutName.c_str(),ios::out);
 if (!RamanTransientOut)
   {
   cerr<< "Error:  "  << RamanTransientOutName  << "Could not be created." << endl;
   exit(1);
 } 

 sprintf(C_String,"RamanTransientIn.dat");
 string RamanTransientInName = JobRaman + C_String;
 ofstream RamanTransientIn(RamanTransientInName.c_str(),ios::out);
 if (!RamanTransientIn)
   {
   cerr<< "Error:  "  << RamanTransientInName  << "Could not be created." << endl;
   exit(1);
 } 
 
    for(int StepT =0;StepT < NumTSteps;StepT ++){
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			if(StepT>30000){
				SI[NSig*NumTSteps + StepT] = 0.002;
				SignalInput[NSig*NumTSteps + StepT] = 0;
			}else{
				SI[NSig*NumTSteps + StepT] = 0.0;
				SignalInput[NSig*NumTSteps + StepT] = 0;
			}
		}
  	}

    for(int StepT =0;StepT < NumTSteps;StepT ++){
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			if(StepT>16000 && StepT < NumTSteps-16000){
				for(int ii=0;ii<RoundNum;ii++){
					SignalInput[NSig*NumTSteps + StepT] += SI[NSig*NumTSteps + StepT + ii] *(RoundNum-ii)
						+SI[NSig*NumTSteps + StepT - ii] *(RoundNum-ii);
				}
				SignalInput[NSig*NumTSteps + StepT] /= RoundNum*(RoundNum+1);
//					cout <<  StepT << "  " << SI[NSig*NumTSteps+StepT] << "   " 
//					<< SignalInput[NSig*NumTSteps+StepT] << endl ; 
			}else{
				SignalInput[NSig*NumTSteps + StepT] = SI[NSig*NumTSteps + StepT];					
//				cout <<  StepT << "   " << SI[NSig*NumTSteps+StepT] << "   " 
//					<< SignalInput[NSig*NumTSteps+StepT] << endl ; 
			}
		}
		
  	}	
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
		for(int StepT = 0 ; StepT < NumTSteps; StepT+=1000)
			RamanTransientIn << StepT*DeltaT << "  " << SI[NSig*NumTSteps+StepT] << "   " 
					<< SignalInput[NSig*NumTSteps+StepT] << endl ; //oSignalPower[NSig*NumZSteps+StepZ] << "  " ;	
		}
	RamanTransientIn << endl;
 	
 	for(int NPum=0;NPum<NumPumCh;NPum++) {
   		oPumpPower[NPum*NumZSteps+NumZSteps-1] = oPumpTargetBackup[NPum];
		oPumpPower[NPum*NumZSteps+NumZSteps-1] = 0.3;
		cout << "oPumpPower = " << oPumpPower[NPum*NumZSteps+NumZSteps-1] << endl;
	}    
	for(int NSig=0;NSig<NumSigCh;NSig++){
		for(int ii=0;ii<NumZSteps;ii++){
			oSignalPower[NSig*NumZSteps+ii]=0;
		}		
	}
	
	  
	for(int i=NumZSteps-2;i>=0;i--){
		BackwardStep(i);
		if(BlowupSign==1)	{
			i=NumZSteps;
		} 
	}  

//		exit(1);
	*LogFile << " Start Transient data" << endl;
	for(int StepT=0;StepT<NumTSteps;StepT++){		
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			oSignalPower[NSig*NumZSteps] = SignalInput[NSig*NumTSteps + StepT]; 
		}
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				// if(StepZ!=NumZSteps-1)
					oSignalTemp[NSig*NumZSteps+StepZ] = oSignalPower[NSig*NumZSteps+StepZ];	
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				// if(StepZ!=0)
					oPumpTemp[NPum*NumZSteps+StepZ] = oPumpPower[NPum*NumZSteps+StepZ];
//				cout << "  PumpPower " << oPumpPower[NPum*NumZSteps+StepZ] << endl;
			}
		}			

//		*LogFile << NumPumCh << "  " << oPumpWaveLth[0] << "  " << oPumpPower[0] << endl;
//		*LogFile << "here 1" << endl;
		Calculatedsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		
		for(int StepZ = 0 ; StepZ < NumZSteps-1; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				oSignalTemp[NSig*NumZSteps+StepZ] += dPsdt[NSig*NumZSteps+StepZ] * DeltaT;	
//				oSignalTemp[NSig*NumZSteps+StepZ] = (oSignalTemp[NSig*NumZSteps+StepZ]
//						+oSignalPower[NSig*NumZSteps+StepZ])/2;
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				oPumpTemp[NPum*NumZSteps+StepZ] += dPpdt[NPum*NumZSteps+StepZ] * DeltaT;
//				oPumpTemp[NPum*NumZSteps+StepZ] = (oPumpTemp[NPum*NumZSteps+StepZ]
//						+oPumpPower[NPum*NumZSteps+StepZ])/2;
			}
		}
//		*LogFile << "here 2" << endl;
		Calculatedsdt(oSignalTemp,oPumpTemp,dPsdt,dPpdt);
		for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
				if(StepZ!=0)
					oSignalPower[NSig*NumZSteps+StepZ] = ( oSignalTemp[NSig*NumZSteps+StepZ] +
						oSignalPower[NSig*NumZSteps+StepZ] 	+ dPsdt[NSig*NumZSteps+StepZ] * DeltaT )/2;	
			}				
			for(int NPum = 0 ; NPum < NumPumCh; NPum++){
				if(StepZ!=NumZSteps-1){
					oPumpPower[NPum*NumZSteps+StepZ] = ( oPumpTemp[NPum*NumZSteps+StepZ] +
						oPumpPower[NPum*NumZSteps+StepZ] + dPpdt[NPum*NumZSteps+StepZ] * DeltaT ) /2;
//					cout << "  PumpPoweraftercal " << StepZ << "  " << oPumpPower[NPum*NumZSteps+StepZ] << endl;
					}
			}
		} 
//		cout << endl;
//		cin >> temp;
		Time += DeltaT;
		if(StepT == 1000 * (int)(StepT/1000))		{	

//			for(int NSig = 0 ; NSig < NumSigCh; NSig++){
//				for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++)
//					RamanTransientOut << oSignalPower[NSig*NumZSteps+StepZ] << "  " ;	
//			}
//			RamanTransientOut << endl;

			RamanTransientOut << Time << "    " << oSignalPower[0*NumZSteps+NumZSteps-1] << "  " 
				<< oPumpPower[0*NumZSteps]<< endl;
		}
		Distance = 0;
	for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
 		Distance += DeltaZ(StepZ);
		for(int NSig = 0 ; NSig < NumSigCh; NSig++){
//			cout << Distance << "  " << oSignalPower[0*NumZSteps+StepZ] << "   " << 
//				oPumpPower[0*NumZSteps+StepZ]  <<  endl;
		}
 	} 
// 	exit(1);		

	}
	
delete [] oSignalTemp ;
delete [] oPumpTemp ;	
delete [] dPsdt ;	
delete [] dPpdt ;	
delete [] SignalInput;
delete [] SI;
RamanTransientOut.close();	
RamanTransientIn.close();	

}



//  ##########################
void RamanAmp::Calculatedsdt(double * oSigTemp,double * oPumTemp,double * dPsigdt,double * dPpumdt)
{ 

 double *oSignaldudz = new double[NumSigCh*NumZSteps];
 double *oPumpdudz = new double[NumPumCh*NumZSteps];
 double GroupVelocity = LightSpeed/1.5;

 double temp;
		// calculate dudz
 for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
// 	cout << StepZ << "  PumpTemp = " << oPumTemp[0*NumZSteps+StepZ] << endl;
 
 }


  for(int NPum = 0 ; NPum < NumPumCh; NPum++){
	oPumpdudz[NPum*NumZSteps] = (-3.0 * oPumTemp[NPum*NumZSteps] + 4 * oPumTemp[NPum*NumZSteps+1]
		-oPumTemp[NPum*NumZSteps+2])/DeltaZ(1)/2;
	oPumpdudz[NPum*NumZSteps + 1] = (-2. * oPumTemp[NPum*NumZSteps]- 3.* oPumTemp[NPum*NumZSteps+1]
		+6*oPumTemp[NPum*NumZSteps+2]-oPumTemp[NPum*NumZSteps+3])/DeltaZ(1)/6;//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
//				<< "  "  << DeltaZ(1) << endl;
	oPumpdudz[NPum*NumZSteps + NumZSteps - 2] = -1*(-2. * oPumTemp[NPum*NumZSteps+ NumZSteps - 1]
		- 3.* oPumTemp[NPum*NumZSteps+ NumZSteps - 2]
		+6*oPumTemp[NPum*NumZSteps+ NumZSteps - 3]-oPumTemp[NPum*NumZSteps+ NumZSteps - 4])/DeltaZ(1)/6;//		cout << " dudzPump " << 0 <<  "  " << oPumTemp[NPum*NumZSteps] << "  "  << oPumpdudz[NPum*NumZSteps] 
  	oPumpdudz[NPum*NumZSteps + NumZSteps -1] =0;
  }
   
  
  for(int StepZ = 2 ; StepZ < NumZSteps-2; StepZ++){
	for(int NSig = 0 ; NSig < NumSigCh; NSig++){
			oSignaldudz[NSig*NumZSteps+StepZ] = (8 * (oSigTemp[NSig*NumZSteps+StepZ+1]-oSigTemp[NSig*NumZSteps+StepZ-1])
				-(oSigTemp[NSig*NumZSteps+StepZ+2]-oSigTemp[NSig*NumZSteps+StepZ-2]))/DeltaZ(StepZ)/12;
//			oSignaldudz[NSig*NumZSteps+StepZ]	=	
//				((oSigTemp[NSig*NumZSteps+StepZ+1]-oSigTemp[NSig*NumZSteps+StepZ])/DeltaZ(StepZ+1)
//				+(oSigTemp[NSig*NumZSteps+StepZ]-oSigTemp[NSig*NumZSteps+StepZ-1])/DeltaZ(StepZ))/2;
//		
//	
//		cout << " dudzSignal " << StepZ << "  " << oSignaldudz[NSig*NumZSteps+StepZ] << endl;
	}			
	for(int NPum = 0 ; NPum < NumPumCh; NPum++){
			oPumpdudz[NPum*NumZSteps+StepZ]	=	
				(8 * (oPumTemp[NPum*NumZSteps+StepZ+1]-oPumTemp[NPum*NumZSteps+StepZ-1])
				- (oPumTemp[NPum*NumZSteps+StepZ+2]-oPumTemp[NPum*NumZSteps+StepZ-2]))/DeltaZ(StepZ)/12;

//			oPumpdudz[NPum*NumZSteps+StepZ]	=	
//				((oPumTemp[NPum*NumZSteps+StepZ+1]-oPumTemp[NPum*NumZSteps+StepZ])/DeltaZ(StepZ+1)
//				+(oPumTemp[NPum*NumZSteps+StepZ]-oPumTemp[NPum*NumZSteps+StepZ-1])/DeltaZ(StepZ))/2;

//		cout << " dudzPump " << StepZ << "  " << oPumTemp[NPum*NumZSteps+StepZ] << "  " << DeltaZ(StepZ) << endl;
	
	} 
  }
  for(int NSig = 0 ; NSig < NumSigCh; NSig++){
	oSignaldudz[NSig*NumZSteps]=0;
	oSignaldudz[NSig*NumZSteps+1] = (-2.0*oSigTemp[NSig*NumZSteps]-3.0*oSigTemp[NSig*NumZSteps+1]
		+6.*oSigTemp[NSig*NumZSteps+2]-oSigTemp[NSig*NumZSteps+3])/DeltaZ(1)/6;		
	oSignaldudz[NSig*NumZSteps+NumZSteps-1]	= -1 * (-3.*oSigTemp[NSig*NumZSteps+NumZSteps-1]
		+4.*oSigTemp[NSig*NumZSteps+NumZSteps-2]-oSigTemp[NSig*NumZSteps+NumZSteps-3])/DeltaZ(NumZSteps-1)/2;		
//		cout << " dudzSignal " << NumZSteps-1 << "  " << oSignaldudz[NSig*NumZSteps+NumZSteps-1] << endl;
	oSignaldudz[NSig*NumZSteps+NumZSteps-2]	= -1.*(-2.*oSigTemp[NSig*NumZSteps+NumZSteps-1]
		-3.*oSigTemp[NSig*NumZSteps+NumZSteps-2]
		+6.*oSigTemp[NSig*NumZSteps+NumZSteps-3]-oSigTemp[NSig*NumZSteps+NumZSteps-4])/DeltaZ(NumZSteps-2)/6;	;		

  }		

 
//cin >> temp;
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
//		cout << " dudzPump " << StepZ << "  " << oPumpdudz[0*NumZSteps+StepZ] 
//				<< "  " << oSignaldudz[0*NumZSteps+StepZ] << endl;
  }

// calculate dpdt
  double StepGain;
	 
  for(int StepZ = 0 ; StepZ < NumZSteps; StepZ++){
	for(int NSig=0;NSig<NumSigCh;NSig++){			// for NumSigCh X (NumSigCh+NumPumCh) Matrix
  		if(StepZ!=0){
			StepGain = 0;
  			for(int P=0;P<NumPumCh;P++){
     			StepGain +=
     			GetRamanGain(oSignalWaveLth[NSig],oPumpWaveLth[P],StepZ)*oPumpPower[P*NumZSteps+StepZ];
  			}
	  		for(int S=0;S<NumSigCh;S++){
    			StepGain +=
    			GetRamanGain(oSignalWaveLth[NSig],oSignalWaveLth[S],StepZ)*oSignalPower[S*NumZSteps+StepZ-1]; 
  			}  		
			dPsigdt[NSig*NumZSteps+StepZ] =  (StepGain * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1] 
				- oSignaldudz[NSig*NumZSteps+StepZ])* GroupVelocity;
//			cout << " dPsdtSignal " << StepZ << "  " << StepGain * oSigTemp[NSig*NumZSteps+StepZ-1]
//				<< "   " << -1 * GetAttenuationFiber(oSignalWaveLth[NSig],StepZ) * oSigTemp[NSig*NumZSteps+StepZ-1]
//				<< "   " << -1 * oSignaldudz[NSig*NumZSteps+StepZ] << "  "  << dPsigdt[NSig*NumZSteps+StepZ]/GroupVelocity ;		
		}else{
			dPsigdt[NSig*NumZSteps+StepZ] = 0;			
//			cout << " dPsdtSignal " << StepZ << "                " ;		

		} 
 	}
	for(int NPum=0;NPum<NumPumCh;NPum++){
		if(StepZ!=NumZSteps-1){
			StepGain = 0;
			for(int P=0;P<NumPumCh;P++){
       			StepGain +=
       				GetRamanGain(oPumpWaveLth[NPum],oPumpWaveLth[P],StepZ+1) * oPumpPower[P*NumZSteps+StepZ+1];
   			}
    		for(int S=0;S<NumSigCh;S++){
       			StepGain +=
       			GetRamanGain(oPumpWaveLth[NPum],oSignalWaveLth[S],StepZ+1) * oSignalPower[S*NumZSteps+StepZ]; 
  			}
  			dPpumdt[NPum*NumZSteps+StepZ] = (StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
				- GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
				+ oPumpdudz[NPum*NumZSteps+StepZ]) * GroupVelocity;
//			cout << "  dPpdtPump" << StepZ << "  " << StepGain * oPumTemp[NPum*NumZSteps+StepZ+1]
//				<< "   " <<  -1 * GetAttenuationFiber(oPumpWaveLth[NPum],StepZ+1) * oPumTemp[NPum*NumZSteps+StepZ+1]
//				<< "   " <<   oPumpdudz[NPum*NumZSteps+StepZ]<< "   " << dPpumdt[NPum*NumZSteps+StepZ]/GroupVelocity<< endl;
    	}else{
			dPpumdt[NPum*NumZSteps+StepZ] = 0;			
//			cout << "  dPpdtPump" << StepZ << "  							" << endl;
   
		}
	}
  }  
// cin >> temp;

delete [] oSignaldudz ;
delete [] oPumpdudz ;
		
}
*/
